(function() {
/*!
 * @overview  Ember - JavaScript Application Framework
 * @copyright Copyright 2011-2018 Tilde Inc. and contributors
 *            Portions Copyright 2006-2011 Strobe Inc.
 *            Portions Copyright 2008-2011 Apple Inc. All rights reserved.
 * @license   Licensed under MIT license
 *            See https://raw.github.com/emberjs/ember.js/master/LICENSE
 * @version   3.6.0-bugfix-ie-dual-build+e2d98aef
 */

/*globals process */
var enifed, requireModule, Ember;

// Used in @ember/-internals/environment/lib/global.js
mainContext = this; // eslint-disable-line no-undef

(function() {
  function missingModule(name, referrerName) {
    if (referrerName) {
      throw new Error('Could not find module ' + name + ' required by: ' + referrerName);
    } else {
      throw new Error('Could not find module ' + name);
    }
  }

  function internalRequire(_name, referrerName) {
    var name = _name;
    var mod = registry[name];

    if (!mod) {
      name = name + '/index';
      mod = registry[name];
    }

    var exports = seen[name];

    if (exports !== undefined) {
      return exports;
    }

    exports = seen[name] = {};

    if (!mod) {
      missingModule(_name, referrerName);
    }

    var deps = mod.deps;
    var callback = mod.callback;
    var reified = new Array(deps.length);

    for (var i = 0; i < deps.length; i++) {
      if (deps[i] === 'exports') {
        reified[i] = exports;
      } else if (deps[i] === 'require') {
        reified[i] = requireModule;
      } else {
        reified[i] = internalRequire(deps[i], name);
      }
    }

    callback.apply(this, reified);

    return exports;
  }

  var isNode =
    typeof window === 'undefined' &&
    typeof process !== 'undefined' &&
    {}.toString.call(process) === '[object process]';

  if (!isNode) {
    Ember = this.Ember = this.Ember || {};
  }

  if (typeof Ember === 'undefined') {
    Ember = {};
  }

  if (typeof Ember.__loader === 'undefined') {
    var registry = {};
    var seen = {};

    enifed = function(name, deps, callback) {
      var value = {};

      if (!callback) {
        value.deps = [];
        value.callback = deps;
      } else {
        value.deps = deps;
        value.callback = callback;
      }

      registry[name] = value;
    };

    requireModule = function(name) {
      return internalRequire(name, null);
    };

    // setup `require` module
    requireModule['default'] = requireModule;

    requireModule.has = function registryHas(moduleName) {
      return !!registry[moduleName] || !!registry[moduleName + '/index'];
    };

    requireModule._eak_seen = registry;

    Ember.__loader = {
      define: enifed,
      require: requireModule,
      registry: registry,
    };
  } else {
    enifed = Ember.__loader.define;
    requireModule = Ember.__loader.require;
  }
})();

enifed('@ember/-internals/container/tests/container_test', ['@ember/-internals/owner', '@ember/polyfills', '@ember/canary-features', '@ember/-internals/container', 'internal-test-helpers'], function (_owner, _polyfills, _canaryFeatures, _container, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Container', class extends _internalTestHelpers.AbstractTestCase {
    ['@test A registered factory returns the same instance each time'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      let postController = container.lookup('controller:post');

      assert.ok(postController instanceof PostController, 'The lookup is an instance of the factory');

      assert.equal(postController, container.lookup('controller:post'));
    }

    ['@test uses create time injections if factory has no extend'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let AppleController = (0, _internalTestHelpers.factory)();
      let PostController = (0, _internalTestHelpers.factory)();

      PostController.extend = undefined; // remove extend

      registry.register('controller:apple', AppleController);
      registry.register('controller:post', PostController);
      registry.injection('controller:post', 'apple', 'controller:apple');

      let postController = container.lookup('controller:post');

      assert.ok(postController.apple instanceof AppleController, 'instance receives an apple of instance AppleController');
    }

    ['@test A registered factory returns a fresh instance if singleton: false is passed as an option'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      let postController1 = container.lookup('controller:post');
      let postController2 = container.lookup('controller:post', {
        singleton: false
      });
      let postController3 = container.lookup('controller:post', {
        singleton: false
      });
      let postController4 = container.lookup('controller:post');

      assert.equal(postController1.toString(), postController4.toString(), 'Singleton factories looked up normally return the same value');
      assert.notEqual(postController1.toString(), postController2.toString(), 'Singleton factories are not equal to factories looked up with singleton: false');
      assert.notEqual(postController2.toString(), postController3.toString(), 'Two factories looked up with singleton: false are not equal');
      assert.notEqual(postController3.toString(), postController4.toString(), 'A singleton factory looked up after a factory called with singleton: false is not equal');

      assert.ok(postController1 instanceof PostController, 'All instances are instances of the registered factory');
      assert.ok(postController2 instanceof PostController, 'All instances are instances of the registered factory');
      assert.ok(postController3 instanceof PostController, 'All instances are instances of the registered factory');
      assert.ok(postController4 instanceof PostController, 'All instances are instances of the registered factory');
    }

    ["@test A factory type with a registered injection's instances receive that injection"](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let Store = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);
      registry.register('store:main', Store);

      registry.typeInjection('controller', 'store', 'store:main');

      let postController = container.lookup('controller:post');
      let store = container.lookup('store:main');

      assert.equal(postController.store, store);
    }

    ['@test An individual factory with a registered injection receives the injection'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let Store = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);
      registry.register('store:main', Store);

      registry.injection('controller:post', 'store', 'store:main');

      let postController = container.lookup('controller:post');
      let store = container.lookup('store:main');

      assert.equal(postController.store, store, 'has the correct store injected');
    }

    ['@test A factory with both type and individual injections'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let Store = (0, _internalTestHelpers.factory)();
      let Router = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);
      registry.register('store:main', Store);
      registry.register('router:main', Router);

      registry.injection('controller:post', 'store', 'store:main');
      registry.typeInjection('controller', 'router', 'router:main');

      let postController = container.lookup('controller:post');
      let store = container.lookup('store:main');
      let router = container.lookup('router:main');

      assert.equal(postController.store, store);
      assert.equal(postController.router, router);
    }

    ['@test A non-singleton instance is never cached'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostView = (0, _internalTestHelpers.factory)();

      registry.register('view:post', PostView, { singleton: false });

      let postView1 = container.lookup('view:post');
      let postView2 = container.lookup('view:post');

      assert.ok(postView1 !== postView2, 'Non-singletons are not cached');
    }

    ['@test A non-instantiated property is not instantiated'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let template = function () {};
      registry.register('template:foo', template, { instantiate: false });
      assert.equal(container.lookup('template:foo'), template);
    }

    ['@test A failed lookup returns undefined'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      assert.equal(container.lookup('doesnot:exist'), undefined);
    }

    ['@test An invalid factory throws an error'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      registry.register('controller:foo', {});

      assert.throws(() => {
        container.lookup('controller:foo');
      }, /Failed to create an instance of \'controller:foo\'/);
    }

    ['@test Injecting a failed lookup raises an error'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let fooInstance = {};
      let fooFactory = {};

      let Foo = {
        create() {
          return fooInstance;
        },
        extend() {
          return fooFactory;
        }
      };

      registry.register('model:foo', Foo);
      registry.injection('model:foo', 'store', 'store:main');

      assert.throws(() => {
        container.lookup('model:foo');
      });
    }

    ['@test Injecting a falsy value does not raise an error'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let ApplicationController = (0, _internalTestHelpers.factory)();

      registry.register('controller:application', ApplicationController);
      registry.register('user:current', null, { instantiate: false });
      registry.injection('controller:application', 'currentUser', 'user:current');

      assert.strictEqual(container.lookup('controller:application').currentUser, null);
    }

    ['@test The container returns same value each time even if the value is falsy'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      registry.register('falsy:value', null, { instantiate: false });

      assert.strictEqual(container.lookup('falsy:value'), container.lookup('falsy:value'));
    }

    ['@test Destroying the container destroys any cached singletons'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let PostView = (0, _internalTestHelpers.factory)();
      let template = function () {};

      registry.register('controller:post', PostController);
      registry.register('view:post', PostView, { singleton: false });
      registry.register('template:post', template, { instantiate: false });

      registry.injection('controller:post', 'postView', 'view:post');

      let postController = container.lookup('controller:post');
      let postView = postController.postView;

      assert.ok(postView instanceof PostView, 'The non-singleton was injected');

      container.destroy();

      assert.ok(postController.isDestroyed, 'Singletons are destroyed');
      assert.ok(!postView.isDestroyed, 'Non-singletons are not destroyed');
    }

    ['@test The container can use a registry hook to resolve factories lazily'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.resolver = {
        resolve(fullName) {
          if (fullName === 'controller:post') {
            return PostController;
          }
        }
      };

      let postController = container.lookup('controller:post');

      assert.ok(postController instanceof PostController, 'The correct factory was provided');
    }

    ['@test The container normalizes names before resolving'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.normalizeFullName = function () {
        return 'controller:post';
      };

      registry.register('controller:post', PostController);
      let postController = container.lookup('controller:normalized');

      assert.ok(postController instanceof PostController, 'Normalizes the name before resolving');
    }

    ['@test The container normalizes names when looking factory up'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.normalizeFullName = function () {
        return 'controller:post';
      };

      registry.register('controller:post', PostController);
      let fact = container.factoryFor('controller:normalized');

      let factInstance = fact.create();
      assert.ok(factInstance instanceof PostController, 'Normalizes the name');
    }

    ['@test Options can be registered that should be applied to a given factory'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostView = (0, _internalTestHelpers.factory)();

      registry.resolver = {
        resolve(fullName) {
          if (fullName === 'view:post') {
            return PostView;
          }
        }
      };

      registry.options('view:post', { instantiate: true, singleton: false });

      let postView1 = container.lookup('view:post');
      let postView2 = container.lookup('view:post');

      assert.ok(postView1 instanceof PostView, 'The correct factory was provided');
      assert.ok(postView2 instanceof PostView, 'The correct factory was provided');

      assert.ok(postView1 !== postView2, 'The two lookups are different');
    }

    ['@test Options can be registered that should be applied to all factories for a given type'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostView = (0, _internalTestHelpers.factory)();

      registry.resolver = {
        resolve(fullName) {
          if (fullName === 'view:post') {
            return PostView;
          }
        }
      };

      registry.optionsForType('view', { singleton: false });

      let postView1 = container.lookup('view:post');
      let postView2 = container.lookup('view:post');

      assert.ok(postView1 instanceof PostView, 'The correct factory was provided');
      assert.ok(postView2 instanceof PostView, 'The correct factory was provided');

      assert.ok(postView1 !== postView2, 'The two lookups are different');
    }

    ['@test An injected non-singleton instance is never cached'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostView = (0, _internalTestHelpers.factory)();
      let PostViewHelper = (0, _internalTestHelpers.factory)();

      registry.register('view:post', PostView, { singleton: false });
      registry.register('view_helper:post', PostViewHelper, {
        singleton: false
      });
      registry.injection('view:post', 'viewHelper', 'view_helper:post');

      let postView1 = container.lookup('view:post');
      let postView2 = container.lookup('view:post');

      assert.ok(postView1.viewHelper !== postView2.viewHelper, 'Injected non-singletons are not cached');
    }

    ['@test Factory resolves are cached'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let resolveWasCalled = [];
      registry.resolve = function (fullName) {
        resolveWasCalled.push(fullName);
        return PostController;
      };

      assert.deepEqual(resolveWasCalled, []);
      container.factoryFor('controller:post');
      assert.deepEqual(resolveWasCalled, ['controller:post']);

      container.factoryFor('controller:post');
      assert.deepEqual(resolveWasCalled, ['controller:post']);
    }

    ['@test factory for non extendables (MODEL) resolves are cached'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();
      let resolveWasCalled = [];
      registry.resolve = function (fullName) {
        resolveWasCalled.push(fullName);
        return PostController;
      };

      assert.deepEqual(resolveWasCalled, []);
      container.factoryFor('model:post');
      assert.deepEqual(resolveWasCalled, ['model:post']);

      container.factoryFor('model:post');
      assert.deepEqual(resolveWasCalled, ['model:post']);
    }

    ['@test factory for non extendables resolves are cached'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = {};
      let resolveWasCalled = [];

      registry.resolve = function (fullName) {
        resolveWasCalled.push(fullName);
        return PostController;
      };

      assert.deepEqual(resolveWasCalled, []);
      container.factoryFor('foo:post');
      assert.deepEqual(resolveWasCalled, ['foo:post']);

      container.factoryFor('foo:post');
      assert.deepEqual(resolveWasCalled, ['foo:post']);
    }

    [`@test A factory's lazy injections are validated when first instantiated`](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let Apple = (0, _internalTestHelpers.factory)();
      let Orange = (0, _internalTestHelpers.factory)();

      Apple.reopenClass({
        _lazyInjections() {
          return [{ specifier: 'orange:main' }, { specifier: 'banana:main' }];
        }
      });

      registry.register('apple:main', Apple);
      registry.register('orange:main', Orange);

      assert.throws(() => {
        container.lookup('apple:main');
      }, /Attempting to inject an unknown injection: 'banana:main'/);
    }

    ['@test Lazy injection validations are cached'](assert) {
      assert.expect(1);

      let registry = new _container.Registry();
      let container = registry.container();
      let Apple = (0, _internalTestHelpers.factory)();
      let Orange = (0, _internalTestHelpers.factory)();

      Apple.reopenClass({
        _lazyInjections: () => {
          assert.ok(true, 'should call lazy injection method');
          return [{ specifier: 'orange:main' }];
        }
      });

      registry.register('apple:main', Apple);
      registry.register('orange:main', Orange);

      container.lookup('apple:main');
      container.lookup('apple:main');
    }

    ['@test An object with its owner pre-set should be returned from ownerInjection'](assert) {
      let owner = {};
      let registry = new _container.Registry();
      let container = registry.container({ owner });

      let result = container.ownerInjection();

      assert.equal(result[_owner.OWNER], owner, 'owner is properly included');
    }

    ['@test lookup passes options through to expandlocallookup'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);
      registry.expandLocalLookup = (fullName, options) => {
        assert.ok(true, 'expandLocalLookup was called');
        assert.equal(fullName, 'foo:bar');
        assert.deepEqual(options, { source: 'baz:qux' });

        return 'controller:post';
      };

      let PostControllerLookupResult = container.lookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.ok(PostControllerLookupResult instanceof PostController);
    }

    ['@test #factoryFor class is registered class'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);

      let factoryManager = container.factoryFor('component:foo-bar');
      assert.deepEqual(factoryManager.class, Component, 'No double extend');
    }

    ['@test #factoryFor must supply a fullname']() {
      let registry = new _container.Registry();
      let container = registry.container();
      expectAssertion(() => {
        container.factoryFor('chad-bar');
      }, /fullName must be a proper full name/);
    }

    ['@test #factoryFor returns a factory manager'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);

      let factoryManager = container.factoryFor('component:foo-bar');
      assert.ok(factoryManager.create);
      assert.ok(factoryManager.class);
    }

    ['@test #factoryFor returns a cached factory manager for the same type'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);
      registry.register('component:baz-bar', Component);

      let factoryManager1 = container.factoryFor('component:foo-bar');
      let factoryManager2 = container.factoryFor('component:foo-bar');
      let factoryManager3 = container.factoryFor('component:baz-bar');

      assert.equal(factoryManager1, factoryManager2, 'cache hit');
      assert.notEqual(factoryManager1, factoryManager3, 'cache miss');
    }

    ['@test #factoryFor class returns the factory function'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);

      let factoryManager = container.factoryFor('component:foo-bar');
      assert.deepEqual(factoryManager.class, Component, 'No double extend');
    }

    ['@test #factoryFor instance have a common parent'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);

      let factoryManager1 = container.factoryFor('component:foo-bar');
      let factoryManager2 = container.factoryFor('component:foo-bar');
      let instance1 = factoryManager1.create({ foo: 'foo' });
      let instance2 = factoryManager2.create({ bar: 'bar' });

      assert.deepEqual(instance1.constructor, instance2.constructor);
    }

    ['@test can properly reset cache'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);

      let factory1 = container.factoryFor('component:foo-bar');
      let factory2 = container.factoryFor('component:foo-bar');

      let instance1 = container.lookup('component:foo-bar');
      let instance2 = container.lookup('component:foo-bar');

      assert.equal(instance1, instance2);
      assert.equal(factory1, factory2);

      container.reset();

      let factory3 = container.factoryFor('component:foo-bar');
      let instance3 = container.lookup('component:foo-bar');

      assert.notEqual(instance1, instance3);
      assert.notEqual(factory1, factory3);
    }

    ['@test #factoryFor created instances come with instance injections'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      let Ajax = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);
      registry.register('util:ajax', Ajax);
      registry.injection('component:foo-bar', 'ajax', 'util:ajax');

      let componentFactory = container.factoryFor('component:foo-bar');
      let component = componentFactory.create();

      assert.ok(component.ajax);
      assert.ok(component.ajax instanceof Ajax);
    }

    ['@test #factoryFor options passed to create clobber injections'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      let Component = (0, _internalTestHelpers.factory)();
      let Ajax = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);
      registry.register('util:ajax', Ajax);
      registry.injection('component:foo-bar', 'ajax', 'util:ajax');

      let componentFactory = container.factoryFor('component:foo-bar');
      let instrance = componentFactory.create({ ajax: 'fetch' });

      assert.equal(instrance.ajax, 'fetch');
    }

    ['@test #factoryFor does not add properties to the object being instantiated when _initFactory is present'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      class Component {
        static _initFactory() {}
        static create(options) {
          let instance = new this();
          (0, _polyfills.assign)(instance, options);
          return instance;
        }
      }
      registry.register('component:foo-bar', Component);

      let componentFactory = container.factoryFor('component:foo-bar');
      let instance = componentFactory.create();

      // note: _guid and isDestroyed are being set in the `factory` constructor
      // not via registry/container shenanigans
      assert.deepEqual(Object.keys(instance), []);
    }

    [`@test assert when calling lookup after destroy on a container`](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);
      let instance = container.lookup('component:foo-bar');
      assert.ok(instance, 'precond lookup successful');

      this.runTask(() => {
        container.destroy();
        container.finalizeDestroy();
      });

      expectAssertion(() => {
        container.lookup('component:foo-bar');
      });
    }

    [`@test assert when calling factoryFor after destroy on a container`](assert) {
      let registry = new _container.Registry();
      let container = registry.container();
      let Component = (0, _internalTestHelpers.factory)();
      registry.register('component:foo-bar', Component);
      let instance = container.factoryFor('component:foo-bar');
      assert.ok(instance, 'precond lookup successful');

      this.runTask(() => {
        container.destroy();
        container.finalizeDestroy();
      });

      expectAssertion(() => {
        container.factoryFor('component:foo-bar');
      });
    }

    // this is skipped until templates and the glimmer environment do not require `OWNER` to be
    // passed in as constructor args
    ['@skip #factoryFor does not add properties to the object being instantiated'](assert) {
      let registry = new _container.Registry();
      let container = registry.container();

      class Component {
        static create(options) {
          let instance = new this();
          (0, _polyfills.assign)(instance, options);
          return instance;
        }
      }
      registry.register('component:foo-bar', Component);

      let componentFactory = container.factoryFor('component:foo-bar');
      let instance = componentFactory.create();

      // note: _guid and isDestroyed are being set in the `factory` constructor
      // not via registry/container shenanigans
      assert.deepEqual(Object.keys(instance), []);
    }
  });

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    (0, _internalTestHelpers.moduleFor)('Container module unification', class extends _internalTestHelpers.AbstractTestCase {
      ['@test The container can expand and resolve a source to factoryFor'](assert) {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let lookup = 'component:my-input';
        let expectedSource = 'template:routes/application';
        let registry = new _container.Registry();
        let resolveCount = 0;
        let expandedKey = 'boom, special expanded key';
        registry.expandLocalLookup = (specifier, options) => {
          this.assert.strictEqual(specifier, lookup, 'specifier is expanded');
          this.assert.strictEqual(options.source, expectedSource, 'source is expanded');
          return expandedKey;
        };
        registry.resolve = function (fullName) {
          resolveCount++;
          if (fullName === expandedKey) {
            return PrivateComponent;
          }
        };

        let container = registry.container();

        assert.strictEqual(container.factoryFor(lookup, { source: expectedSource }).class, PrivateComponent, 'The correct factory was provided');
        assert.strictEqual(container.factoryFor(lookup, { source: expectedSource }).class, PrivateComponent, 'The correct factory was provided again');
        assert.equal(resolveCount, 1, 'resolve called only once and a cached factory was returned the second time');
      }

      ['@test The container can expand and resolve a source to lookup']() {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let lookup = 'component:my-input';
        let expectedSource = 'template:routes/application';
        let registry = new _container.Registry();
        let expandedKey = 'boom, special expanded key';
        registry.expandLocalLookup = (specifier, options) => {
          this.assert.strictEqual(specifier, lookup, 'specifier is expanded');
          this.assert.strictEqual(options.source, expectedSource, 'source is expanded');
          return expandedKey;
        };
        registry.resolve = function (fullName) {
          if (fullName === expandedKey) {
            return PrivateComponent;
          }
        };

        let container = registry.container();

        let result = container.lookup(lookup, { source: expectedSource });
        this.assert.ok(result instanceof PrivateComponent, 'The correct factory was provided');

        this.assert.ok(container.cache[expandedKey] instanceof PrivateComponent, 'The correct factory was stored in the cache with the correct key which includes the source.');
      }

      ['@test The container can expand and resolve a namespace to factoryFor'](assert) {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let lookup = 'component:my-input';
        let expectedNamespace = 'my-addon';
        let registry = new _container.Registry();
        let resolveCount = 0;
        let expandedKey = 'boom, special expanded key';
        registry.expandLocalLookup = (specifier, options) => {
          this.assert.strictEqual(specifier, lookup, 'specifier is expanded');
          this.assert.strictEqual(options.namespace, expectedNamespace, 'namespace is expanded');
          return expandedKey;
        };
        registry.resolve = function (fullName) {
          resolveCount++;
          if (fullName === expandedKey) {
            return PrivateComponent;
          }
        };

        let container = registry.container();

        assert.strictEqual(container.factoryFor(lookup, { namespace: expectedNamespace }).class, PrivateComponent, 'The correct factory was provided');
        assert.strictEqual(container.factoryFor(lookup, { namespace: expectedNamespace }).class, PrivateComponent, 'The correct factory was provided again');
        assert.equal(resolveCount, 1, 'resolve called only once and a cached factory was returned the second time');
      }

      ['@test The container can expand and resolve a namespace to lookup']() {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let lookup = 'component:my-input';
        let expectedNamespace = 'my-addon';
        let registry = new _container.Registry();
        let expandedKey = 'boom, special expanded key';
        registry.expandLocalLookup = (specifier, options) => {
          this.assert.strictEqual(specifier, lookup, 'specifier is expanded');
          this.assert.strictEqual(options.namespace, expectedNamespace, 'namespace is expanded');
          return expandedKey;
        };
        registry.resolve = function (fullName) {
          if (fullName === expandedKey) {
            return PrivateComponent;
          }
        };

        let container = registry.container();

        let result = container.lookup(lookup, { namespace: expectedNamespace });
        this.assert.ok(result instanceof PrivateComponent, 'The correct factory was provided');

        this.assert.ok(container.cache[expandedKey] instanceof PrivateComponent, 'The correct factory was stored in the cache with the correct key which includes the source.');
      }
    });
  }
});
enifed('@ember/-internals/container/tests/owner_test', ['@ember/-internals/owner', 'internal-test-helpers'], function (_owner, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Owner', class extends _internalTestHelpers.AbstractTestCase {
    ['@test An owner can be set with `setOwner` and retrieved with `getOwner`'](assert) {
      let owner = {};
      let obj = {};

      assert.strictEqual((0, _owner.getOwner)(obj), undefined, 'owner has not been set');

      (0, _owner.setOwner)(obj, owner);

      assert.strictEqual((0, _owner.getOwner)(obj), owner, 'owner has been set');

      assert.strictEqual(obj[_owner.OWNER], owner, 'owner has been set to the OWNER symbol');
    }
  });
});
enifed('@ember/-internals/container/tests/registry_test', ['@ember/-internals/container', 'internal-test-helpers', '@ember/canary-features'], function (_container, _internalTestHelpers, _canaryFeatures) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Registry', class extends _internalTestHelpers.AbstractTestCase {
    ['@test A registered factory is returned from resolve'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      let PostControllerFactory = registry.resolve('controller:post');

      assert.ok(PostControllerFactory, 'factory is returned');
      assert.ok(PostControllerFactory.create() instanceof PostController, 'The return of factory.create is an instance of PostController');
    }

    ['@test The registered factory returned from resolve is the same factory each time'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      assert.deepEqual(registry.resolve('controller:post'), registry.resolve('controller:post'), 'The return of resolve is always the same');
    }

    ['@test The registered value returned from resolve is the same value each time even if the value is falsy'](assert) {
      let registry = new _container.Registry();

      registry.register('falsy:value', null, { instantiate: false });

      assert.strictEqual(registry.resolve('falsy:value'), registry.resolve('falsy:value'), 'The return of resolve is always the same');
    }

    ['@test The value returned from resolver is the same value as the original value even if the value is falsy'](assert) {
      let resolver = {
        resolve(fullName) {
          if (fullName === 'falsy:value') {
            return null;
          }
        }
      };
      let registry = new _container.Registry({ resolver });

      assert.strictEqual(registry.resolve('falsy:value'), null);
    }

    ['@test A registered factory returns true for `has` if an item is registered'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      assert.equal(registry.has('controller:post'), true, 'The `has` method returned true for registered factories');
      assert.equal(registry.has('controller:posts'), false, 'The `has` method returned false for unregistered factories');
    }

    ['@test Throw exception when trying to inject `type:thing` on all type(s)']() {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.register('controller:post', PostController);

      expectAssertion(() => {
        registry.typeInjection('controller', 'injected', 'controller:post');
      }, /Cannot inject a 'controller:post' on other controller\(s\)\./);
    }

    ['@test The registry can take a hook to resolve factories lazily'](assert) {
      let PostController = (0, _internalTestHelpers.factory)();
      let resolver = {
        resolve(fullName) {
          if (fullName === 'controller:post') {
            return PostController;
          }
        }
      };
      let registry = new _container.Registry({ resolver });

      assert.strictEqual(registry.resolve('controller:post'), PostController, 'The correct factory was provided');
    }

    ['@test The registry respects the resolver hook for `has`'](assert) {
      let PostController = (0, _internalTestHelpers.factory)();
      let resolver = {
        resolve(fullName) {
          if (fullName === 'controller:post') {
            return PostController;
          }
        }
      };
      let registry = new _container.Registry({ resolver });

      assert.ok(registry.has('controller:post'), 'the `has` method uses the resolver hook');
    }

    ['@test The registry normalizes names when resolving'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.normalizeFullName = function () {
        return 'controller:post';
      };

      registry.register('controller:post', PostController);
      let type = registry.resolve('controller:normalized');

      assert.strictEqual(type, PostController, 'Normalizes the name when resolving');
    }

    ['@test The registry normalizes names when checking if the factory is registered'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();

      registry.normalizeFullName = function (fullName) {
        return fullName === 'controller:normalized' ? 'controller:post' : fullName;
      };

      registry.register('controller:post', PostController);
      let isPresent = registry.has('controller:normalized');

      assert.equal(isPresent, true, 'Normalizes the name when checking if the factory or instance is present');
    }

    ['@test The registry normalizes names when injecting'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();
      let user = { name: 'Stef' };

      registry.normalize = function () {
        return 'controller:post';
      };

      registry.register('controller:post', PostController);
      registry.register('user:post', user, { instantiate: false });
      registry.injection('controller:post', 'user', 'controller:normalized');

      assert.deepEqual(registry.resolve('controller:post'), user, 'Normalizes the name when injecting');
    }

    ['@test cannot register an `undefined` factory'](assert) {
      let registry = new _container.Registry();

      assert.throws(() => {
        registry.register('controller:apple', undefined);
      }, '');
    }

    ['@test can re-register a factory'](assert) {
      let registry = new _container.Registry();
      let FirstApple = (0, _internalTestHelpers.factory)('first');
      let SecondApple = (0, _internalTestHelpers.factory)('second');

      registry.register('controller:apple', FirstApple);
      registry.register('controller:apple', SecondApple);

      assert.ok(registry.resolve('controller:apple').create() instanceof SecondApple);
    }

    ['@test cannot re-register a factory if it has been resolved'](assert) {
      let registry = new _container.Registry();
      let FirstApple = (0, _internalTestHelpers.factory)('first');
      let SecondApple = (0, _internalTestHelpers.factory)('second');

      registry.register('controller:apple', FirstApple);
      assert.strictEqual(registry.resolve('controller:apple'), FirstApple);

      expectAssertion(function () {
        registry.register('controller:apple', SecondApple);
      }, /Cannot re-register: 'controller:apple', as it has already been resolved\./);

      assert.strictEqual(registry.resolve('controller:apple'), FirstApple);
    }

    ['@test registry.has should not accidentally cause injections on that factory to be run. (Mitigate merely on observing)'](assert) {
      assert.expect(1);

      let registry = new _container.Registry();
      let FirstApple = (0, _internalTestHelpers.factory)('first');
      let SecondApple = (0, _internalTestHelpers.factory)('second');

      SecondApple.extend = function () {
        assert.ok(false, 'should not extend or touch the injected model, merely to inspect existence of another');
      };

      registry.register('controller:apple', FirstApple);
      registry.register('controller:second-apple', SecondApple);
      registry.injection('controller:apple', 'badApple', 'controller:second-apple');

      assert.ok(registry.has('controller:apple'));
    }

    ['@test registry.has should not error for invalid fullNames'](assert) {
      let registry = new _container.Registry();

      assert.ok(!registry.has('foo:bar:baz'));
    }

    ['@test once resolved, always return the same result'](assert) {
      let registry = new _container.Registry();

      registry.resolver = {
        resolve() {
          return 'bar';
        }
      };

      let Bar = registry.resolve('models:bar');

      registry.resolver = {
        resolve() {
          return 'not bar';
        }
      };

      assert.equal(registry.resolve('models:bar'), Bar);
    }

    ['@test factory resolves are cached'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();
      let resolveWasCalled = [];

      registry.resolver = {
        resolve(fullName) {
          resolveWasCalled.push(fullName);
          return PostController;
        }
      };

      assert.deepEqual(resolveWasCalled, []);
      registry.resolve('controller:post');
      assert.deepEqual(resolveWasCalled, ['controller:post']);

      registry.resolve('controller:post');
      assert.deepEqual(resolveWasCalled, ['controller:post']);
    }

    ['@test factory for non extendables (MODEL) resolves are cached'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();
      let resolveWasCalled = [];

      registry.resolver = {
        resolve(fullName) {
          resolveWasCalled.push(fullName);
          return PostController;
        }
      };

      assert.deepEqual(resolveWasCalled, []);
      registry.resolve('model:post');
      assert.deepEqual(resolveWasCalled, ['model:post']);

      registry.resolve('model:post');
      assert.deepEqual(resolveWasCalled, ['model:post']);
    }

    ['@test factory for non extendables resolves are cached'](assert) {
      let registry = new _container.Registry();
      let PostController = {};
      let resolveWasCalled = [];

      registry.resolver = {
        resolve(fullName) {
          resolveWasCalled.push(fullName);
          return PostController;
        }
      };

      assert.deepEqual(resolveWasCalled, []);
      registry.resolve('foo:post');
      assert.deepEqual(resolveWasCalled, ['foo:post']);

      registry.resolve('foo:post');
      assert.deepEqual(resolveWasCalled, ['foo:post']);
    }

    ['@test registry.container creates a container'](assert) {
      let registry = new _container.Registry();
      let PostController = (0, _internalTestHelpers.factory)();
      registry.register('controller:post', PostController);

      let container = registry.container();
      let postController = container.lookup('controller:post');

      assert.ok(postController instanceof PostController, 'The lookup is an instance of the registered factory');
    }

    ['@test `describe` will be handled by the resolver, then by the fallback registry, if available'](assert) {
      let fallback = {
        describe(fullName) {
          return `${fullName}-fallback`;
        }
      };

      let resolver = {
        lookupDescription(fullName) {
          return `${fullName}-resolver`;
        }
      };

      let registry = new _container.Registry({ fallback, resolver });

      assert.equal(registry.describe('controller:post'), 'controller:post-resolver', '`describe` handled by the resolver first.');

      registry.resolver = null;

      assert.equal(registry.describe('controller:post'), 'controller:post-fallback', '`describe` handled by fallback registry next.');

      registry.fallback = null;

      assert.equal(registry.describe('controller:post'), 'controller:post', '`describe` by default returns argument.');
    }

    ['@test `normalizeFullName` will be handled by the resolver, then by the fallback registry, if available'](assert) {
      let fallback = {
        normalizeFullName(fullName) {
          return `${fullName}-fallback`;
        }
      };

      let resolver = {
        normalize(fullName) {
          return `${fullName}-resolver`;
        }
      };

      let registry = new _container.Registry({ fallback, resolver });

      assert.equal(registry.normalizeFullName('controller:post'), 'controller:post-resolver', '`normalizeFullName` handled by the resolver first.');

      registry.resolver = null;

      assert.equal(registry.normalizeFullName('controller:post'), 'controller:post-fallback', '`normalizeFullName` handled by fallback registry next.');

      registry.fallback = null;

      assert.equal(registry.normalizeFullName('controller:post'), 'controller:post', '`normalizeFullName` by default returns argument.');
    }

    ['@test `makeToString` will be handled by the resolver, then by the fallback registry, if available'](assert) {
      let fallback = {
        makeToString(fullName) {
          return `${fullName}-fallback`;
        }
      };

      let resolver = {
        makeToString(fullName) {
          return `${fullName}-resolver`;
        }
      };

      let registry = new _container.Registry({ fallback, resolver });

      assert.equal(registry.makeToString('controller:post'), 'controller:post-resolver', '`makeToString` handled by the resolver first.');

      registry.resolver = null;

      assert.equal(registry.makeToString('controller:post'), 'controller:post-fallback', '`makeToString` handled by fallback registry next.');

      registry.fallback = null;

      assert.equal(registry.makeToString('controller:post'), 'controller:post', '`makeToString` by default returns argument.');
    }

    ['@test `resolve` can be handled by a fallback registry'](assert) {
      let fallback = new _container.Registry();

      let registry = new _container.Registry({ fallback: fallback });
      let PostController = (0, _internalTestHelpers.factory)();

      fallback.register('controller:post', PostController);

      let PostControllerFactory = registry.resolve('controller:post');

      assert.ok(PostControllerFactory, 'factory is returned');
      assert.ok(PostControllerFactory.create() instanceof PostController, 'The return of factory.create is an instance of PostController');
    }

    ['@test `has` can be handled by a fallback registry'](assert) {
      let fallback = new _container.Registry();

      let registry = new _container.Registry({ fallback: fallback });
      let PostController = (0, _internalTestHelpers.factory)();

      fallback.register('controller:post', PostController);

      assert.equal(registry.has('controller:post'), true, 'Fallback registry is checked for registration');
    }

    ['@test `getInjections` includes injections from a fallback registry'](assert) {
      let fallback = new _container.Registry();
      let registry = new _container.Registry({ fallback: fallback });

      assert.strictEqual(registry.getInjections('model:user'), undefined, 'No injections in the primary registry');

      fallback.injection('model:user', 'post', 'model:post');

      assert.equal(registry.getInjections('model:user').length, 1, 'Injections from the fallback registry are merged');
    }

    ['@test `getTypeInjections` includes type injections from a fallback registry'](assert) {
      let fallback = new _container.Registry();
      let registry = new _container.Registry({ fallback: fallback });

      assert.strictEqual(registry.getTypeInjections('model'), undefined, 'No injections in the primary registry');

      fallback.injection('model', 'source', 'source:main');

      assert.equal(registry.getTypeInjections('model').length, 1, 'Injections from the fallback registry are merged');
    }

    ['@test `knownForType` contains keys for each item of a given type'](assert) {
      let registry = new _container.Registry();

      registry.register('foo:bar-baz', 'baz');
      registry.register('foo:qux-fez', 'fez');

      let found = registry.knownForType('foo');

      assert.deepEqual(found, {
        'foo:bar-baz': true,
        'foo:qux-fez': true
      });
    }

    ['@test `knownForType` includes fallback registry results'](assert) {
      let fallback = new _container.Registry();
      let registry = new _container.Registry({ fallback: fallback });

      registry.register('foo:bar-baz', 'baz');
      registry.register('foo:qux-fez', 'fez');
      fallback.register('foo:zurp-zorp', 'zorp');

      let found = registry.knownForType('foo');

      assert.deepEqual(found, {
        'foo:bar-baz': true,
        'foo:qux-fez': true,
        'foo:zurp-zorp': true
      });
    }

    ['@test `knownForType` is called on the resolver if present'](assert) {
      assert.expect(3);

      let resolver = {
        knownForType(type) {
          assert.ok(true, 'knownForType called on the resolver');
          assert.equal(type, 'foo', 'the type was passed through');

          return { 'foo:yorp': true };
        }
      };

      let registry = new _container.Registry({
        resolver
      });
      registry.register('foo:bar-baz', 'baz');

      let found = registry.knownForType('foo');

      assert.deepEqual(found, {
        'foo:yorp': true,
        'foo:bar-baz': true
      });
    }

    ['@test resolver.expandLocalLookup is not required'](assert) {
      let registry = new _container.Registry({
        resolver: {}
      });

      let result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, null);
    }

    ['@test expandLocalLookup is called on the resolver if present'](assert) {
      assert.expect(4);

      let resolver = {
        expandLocalLookup: (targetFullName, sourceFullName) => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');
          assert.equal(targetFullName, 'foo:bar', 'the targetFullName was passed through');
          assert.equal(sourceFullName, 'baz:qux', 'the sourceFullName was passed through');

          return 'foo:qux/bar';
        }
      };

      let registry = new _container.Registry({
        resolver
      });

      let result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar');
    }

    ['@test `expandLocalLookup` is handled by the resolver, then by the fallback registry, if available'](assert) {
      assert.expect(9);

      let fallbackResolver = {
        expandLocalLookup: (targetFullName, sourceFullName) => {
          assert.ok(true, 'expandLocalLookup is called on the fallback resolver');
          assert.equal(targetFullName, 'foo:bar', 'the targetFullName was passed through');
          assert.equal(sourceFullName, 'baz:qux', 'the sourceFullName was passed through');

          return 'foo:qux/bar-fallback';
        }
      };

      let resolver = {
        expandLocalLookup: (targetFullName, sourceFullName) => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');
          assert.equal(targetFullName, 'foo:bar', 'the targetFullName was passed through');
          assert.equal(sourceFullName, 'baz:qux', 'the sourceFullName was passed through');

          return 'foo:qux/bar-resolver';
        }
      };

      let fallbackRegistry = new _container.Registry({
        resolver: fallbackResolver
      });

      let registry = new _container.Registry({
        fallback: fallbackRegistry,
        resolver
      });

      let result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar-resolver', 'handled by the resolver');

      registry.resolver = null;

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar-fallback', 'handled by the fallback registry');

      registry.fallback = null;

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, null, 'null is returned by default when no resolver or fallback registry is present');
    }

    ['@test resolver.expandLocalLookup result is cached'](assert) {
      assert.expect(3);
      let result;

      let resolver = {
        expandLocalLookup: () => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');

          return 'foo:qux/bar';
        }
      };

      let registry = new _container.Registry({
        resolver
      });

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar');

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar');
    }

    ['@test resolver.expandLocalLookup cache is busted when any unregister is called'](assert) {
      assert.expect(4);
      let result;

      let resolver = {
        expandLocalLookup: () => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');

          return 'foo:qux/bar';
        }
      };

      let registry = new _container.Registry({
        resolver
      });

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar');

      registry.unregister('foo:bar');

      result = registry.expandLocalLookup('foo:bar', {
        source: 'baz:qux'
      });

      assert.equal(result, 'foo:qux/bar');
    }

    ['@test resolve calls expandLocallookup when it receives options.source'](assert) {
      assert.expect(3);

      let resolver = {
        resolve() {},
        expandLocalLookup: (targetFullName, sourceFullName) => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');
          assert.equal(targetFullName, 'foo:bar', 'the targetFullName was passed through');
          assert.equal(sourceFullName, 'baz:qux', 'the sourceFullName was passed through');

          return 'foo:qux/bar';
        }
      };

      let registry = new _container.Registry({
        resolver
      });

      registry.resolve('foo:bar', {
        source: 'baz:qux'
      });
    }

    ['@test has uses expandLocalLookup'](assert) {
      assert.expect(5);
      let resolvedFullNames = [];
      let result;

      let resolver = {
        resolve(name) {
          resolvedFullNames.push(name);
          return 'yippie!';
        },

        expandLocalLookup: targetFullName => {
          assert.ok(true, 'expandLocalLookup is called on the resolver');

          if (targetFullName === 'foo:bar') {
            return 'foo:qux/bar';
          } else {
            return null;
          }
        }
      };

      let registry = new _container.Registry({
        resolver
      });

      result = registry.has('foo:bar', {
        source: 'baz:qux'
      });

      assert.ok(result, 'found foo:bar/qux');

      result = registry.has('foo:baz', {
        source: 'baz:qux'
      });

      assert.ok(!result, 'foo:baz/qux not found');

      assert.deepEqual(['foo:qux/bar'], resolvedFullNames);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Registry privatize', class extends _internalTestHelpers.AbstractTestCase {
    ['@test valid format'](assert) {
      let privatized = (0, _container.privatize)(['secret:factory']);
      let matched = privatized.match(/^([^:]+):([^:]+)-(\d+)$/);

      assert.ok(matched, 'privatized format was recognized');
      assert.equal(matched[1], 'secret');
      assert.equal(matched[2], 'factory');
      assert.ok(/^\d+$/.test(matched[3]));
    }
  });

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    (0, _internalTestHelpers.moduleFor)('Registry module unification', class extends _internalTestHelpers.AbstractTestCase {
      ['@test The registry can pass a source to the resolver'](assert) {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let type = 'component';
        let name = 'my-input';
        let specifier = `${type}:${name}`;
        let source = 'template:routes/application';

        let resolver = new _internalTestHelpers.ModuleBasedTestResolver();
        resolver.add({ specifier, source }, PrivateComponent);
        let registry = new _container.Registry({ resolver });

        assert.strictEqual(registry.resolve(specifier), undefined, 'Not returned when specifier not scoped');
        assert.strictEqual(registry.resolve(specifier, { source }), PrivateComponent, 'The correct factory was provided');
        assert.strictEqual(registry.resolve(specifier, { source }), PrivateComponent, 'The correct factory was provided again');
      }

      ['@test The registry can pass a namespace to the resolver'](assert) {
        let PrivateComponent = (0, _internalTestHelpers.factory)();
        let type = 'component';
        let name = 'my-input';
        let specifier = `${type}:${name}`;
        let source = 'template:routes/application';
        let namespace = 'my-addon';

        let resolver = new _internalTestHelpers.ModuleBasedTestResolver();
        resolver.add({ specifier, source, namespace }, PrivateComponent);
        let registry = new _container.Registry({ resolver });

        assert.strictEqual(registry.resolve(specifier), undefined, 'Not returned when specifier not scoped');
        assert.strictEqual(registry.resolve(specifier, { source }), undefined, 'Not returned when specifier is missing namespace');
        assert.strictEqual(registry.resolve(specifier, { source, namespace }), PrivateComponent, 'The correct factory was provided');
        assert.strictEqual(registry.resolve(specifier, { source, namespace }), PrivateComponent, 'The correct factory was provided again');
      }
    });
  }
});
enifed('@ember/-internals/extension-support/tests/container_debug_adapter_test', ['internal-test-helpers', '@ember/polyfills', '@ember/runloop', '@ember/controller', '@ember/debug', '@ember/-internals/extension-support/index'], function (_internalTestHelpers, _polyfills, _runloop, _controller, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug'); // Must be required to export Ember.ContainerDebugAdapter.


  (0, _internalTestHelpers.moduleFor)('Container Debug Adapter', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      (0, _debug.setDebugFunction)('debug', () => {});
      super();
      this.adapter = this.application.__deprecatedInstance__.lookup('container-debug-adapter:main');
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        autoboot: true
      });
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
      (0, _runloop.run)(() => {
        this.adapter.destroy();
      });

      super.teardown();
    }

    ['@test default ContainerDebugAdapter cannot catalog certain entries by type'](assert) {
      assert.equal(this.adapter.canCatalogEntriesByType('model'), false, 'canCatalogEntriesByType should return false for model');
      assert.equal(this.adapter.canCatalogEntriesByType('template'), false, 'canCatalogEntriesByType should return false for template');
    }

    ['@test default ContainerDebugAdapter can catalog typical entries by type'](assert) {
      assert.equal(this.adapter.canCatalogEntriesByType('controller'), true, 'canCatalogEntriesByType should return true for controller');
      assert.equal(this.adapter.canCatalogEntriesByType('route'), true, 'canCatalogEntriesByType should return true for route');
      assert.equal(this.adapter.canCatalogEntriesByType('view'), true, 'canCatalogEntriesByType should return true for view');
    }

    ['@test default ContainerDebugAdapter catalogs controller entries'](assert) {
      this.application.PostController = _controller.default.extend();
      let controllerClasses = this.adapter.catalogEntriesByType('controller');

      assert.equal(controllerClasses.length, 1, 'found 1 class');
      assert.equal(controllerClasses[0], 'post', 'found the right class');
    }
  });
});
enifed('@ember/-internals/extension-support/tests/data_adapter_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/extension-support/lib/data_adapter', 'internal-test-helpers'], function (_runloop, _metal, _runtime, _data_adapter, _internalTestHelpers) {
  'use strict';

  let adapter;
  const Model = _runtime.Object.extend();

  const PostClass = Model.extend();

  const DataAdapter = _data_adapter.default.extend({
    detect(klass) {
      return klass !== Model && Model.detect(klass);
    },
    init() {
      this._super(...arguments);
      this.set('containerDebugAdapter', {
        canCatalogEntriesByType() {
          return true;
        },
        catalogEntriesByType() {
          return (0, _runtime.A)(['post']);
        }
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Data Adapter', class extends _internalTestHelpers.ApplicationTestCase {
    teardown() {
      super.teardown();
      adapter = undefined;
    }

    ['@test Model types added'](assert) {
      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return (0, _runtime.A)([1, 2, 3]);
        },
        columnsForType() {
          return [{ name: 'title', desc: 'Title' }];
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        let adapter = this.applicationInstance.lookup('data-adapter:main');

        function modelTypesAdded(types) {
          assert.equal(types.length, 1);
          let postType = types[0];
          assert.equal(postType.name, 'post', 'Correctly sets the name');
          assert.equal(postType.count, 3, 'Correctly sets the record count');
          assert.strictEqual(postType.object, PostClass, 'Correctly sets the object');
          assert.deepEqual(postType.columns, [{ name: 'title', desc: 'Title' }], 'Correctly sets the columns');
        }

        adapter.watchModelTypes(modelTypesAdded);
      });
    }

    ['@test getRecords gets a model name as second argument'](assert) {
      this.add('data-adapter:main', DataAdapter.extend({
        getRecords(klass, name) {
          assert.equal(name, 'post');
          return (0, _runtime.A)();
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');
        adapter.watchModelTypes(function () {});
      });
    }

    ['@test Model types added with custom container-debug-adapter'](assert) {
      let StubContainerDebugAdapter = _runtime.Object.extend({
        canCatalogEntriesByType() {
          return true;
        },
        catalogEntriesByType() {
          return (0, _runtime.A)(['post']);
        }
      });
      this.add('container-debug-adapter:main', StubContainerDebugAdapter);
      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return (0, _runtime.A)([1, 2, 3]);
        },
        columnsForType() {
          return [{ name: 'title', desc: 'Title' }];
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        let adapter = this.applicationInstance.lookup('data-adapter:main');

        function modelTypesAdded(types) {
          assert.equal(types.length, 1);
          let postType = types[0];
          assert.equal(postType.name, 'post', 'Correctly sets the name');
          assert.equal(postType.count, 3, 'Correctly sets the record count');
          assert.strictEqual(postType.object, PostClass, 'Correctly sets the object');
          assert.deepEqual(postType.columns, [{ name: 'title', desc: 'Title' }], 'Correctly sets the columns');
        }

        adapter.watchModelTypes(modelTypesAdded);
      });
    }

    ['@test Model Types Updated'](assert) {
      let records = (0, _runtime.A)([1, 2, 3]);
      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return records;
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');

        function modelTypesAdded() {
          (0, _runloop.run)(() => {
            records.pushObject(4);
          });
        }

        function modelTypesUpdated(types) {
          let postType = types[0];
          assert.equal(postType.count, 4, 'Correctly updates the count');
        }

        adapter.watchModelTypes(modelTypesAdded, modelTypesUpdated);
      });
    }

    ['@test Model Types Updated but Unchanged Do not Trigger Callbacks'](assert) {
      assert.expect(0);
      let records = (0, _runtime.A)([1, 2, 3]);
      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return records;
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');

        function modelTypesAdded() {
          (0, _runloop.run)(() => {
            records.arrayContentDidChange(0, 0, 0);
          });
        }

        function modelTypesUpdated() {
          assert.ok(false, "modelTypesUpdated should not be triggered if the array didn't change");
        }

        adapter.watchModelTypes(modelTypesAdded, modelTypesUpdated);
      });
    }

    ['@test Records Added'](assert) {
      let countAdded = 1;
      let post = PostClass.create();
      let recordList = (0, _runtime.A)([post]);

      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return recordList;
        },
        getRecordColor() {
          return 'blue';
        },
        getRecordColumnValues() {
          return { title: 'Post ' + countAdded };
        },
        getRecordKeywords() {
          return ['Post ' + countAdded];
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');

        function recordsAdded(records) {
          let record = records[0];
          assert.equal(record.color, 'blue', 'Sets the color correctly');
          assert.deepEqual(record.columnValues, { title: 'Post ' + countAdded }, 'Sets the column values correctly');
          assert.deepEqual(record.searchKeywords, ['Post ' + countAdded], 'Sets search keywords correctly');
          assert.strictEqual(record.object, post, 'Sets the object to the record instance');
        }

        adapter.watchRecords('post', recordsAdded);
        countAdded++;
        post = PostClass.create();
        recordList.pushObject(post);
      });
    }

    ['@test Observes and releases a record correctly'](assert) {
      let updatesCalled = 0;
      let post = PostClass.create({ title: 'Post' });
      let recordList = (0, _runtime.A)([post]);

      this.add('data-adapter:main', DataAdapter.extend({
        getRecords() {
          return recordList;
        },
        observeRecord(record, recordUpdated) {
          let self = this;
          function callback() {
            recordUpdated(self.wrapRecord(record));
          }
          (0, _metal.addObserver)(record, 'title', callback);
          return function () {
            (0, _metal.removeObserver)(record, 'title', callback);
          };
        },
        getRecordColumnValues(record) {
          return { title: (0, _metal.get)(record, 'title') };
        }
      }));
      this.add('model:post', PostClass);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');

        function recordsAdded() {
          (0, _metal.set)(post, 'title', 'Post Modified');
        }

        function recordsUpdated(records) {
          updatesCalled++;
          assert.equal(records[0].columnValues.title, 'Post Modified');
        }

        let release = adapter.watchRecords('post', recordsAdded, recordsUpdated);
        release();
        (0, _metal.set)(post, 'title', 'New Title');
        assert.equal(updatesCalled, 1, 'Release function removes observers');
      });
    }

    ['@test _nameToClass does not error when not found'](assert) {
      this.add('data-adapter:main', DataAdapter);

      return this.visit('/').then(() => {
        adapter = this.applicationInstance.lookup('data-adapter:main');

        let klass = adapter._nameToClass('foo');

        assert.equal(klass, undefined, 'returns undefined');
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/application/actions-test', ['@ember/controller', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/debug'], function (_controller, _testCase, _helpers, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');
  const noop = function () {};

  (0, _testCase.moduleFor)('Application test: actions', class extends _testCase.ApplicationTest {
    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super(...arguments);
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
    }

    ['@test actions in top level template application template target application controller'](assert) {
      assert.expect(1);

      this.add('controller:application', _controller.default.extend({
        actions: {
          handleIt() {
            assert.ok(true, 'controller received action properly');
          }
        }
      }));

      this.addTemplate('application', '<button id="handle-it" {{action "handleIt"}}>Click!</button>');

      return this.visit('/').then(() => {
        this.runTask(() => this.$('#handle-it').click());
      });
    }

    ['@test actions in nested outlet template target their controller'](assert) {
      assert.expect(1);

      this.add('controller:application', _controller.default.extend({
        actions: {
          handleIt() {
            assert.ok(false, 'application controller should not have received action!');
          }
        }
      }));

      this.add('controller:index', _controller.default.extend({
        actions: {
          handleIt() {
            assert.ok(true, 'controller received action properly');
          }
        }
      }));

      this.addTemplate('index', '<button id="handle-it" {{action "handleIt"}}>Click!</button>');

      return this.visit('/').then(() => {
        this.runTask(() => this.$('#handle-it').click());
      });
    }
  });

  (0, _testCase.moduleFor)('Rendering test: non-interactive actions', class extends _testCase.RenderingTest {
    getBootOptions() {
      return { isInteractive: false };
    }

    [`@test doesn't attatch actions`](assert) {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          actions: {
            fire() {
              assert.ok(false);
            }
          }
        }),
        template: `<button {{action 'fire'}}>Fire!</button>`
      });

      this.render('{{foo-bar tagName=""}}');

      this.assertHTML('<button>Fire!</button>');

      this.$('button').click();
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/application/engine-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/controller', '@ember/-internals/runtime', '@ember/-internals/glimmer', '@ember/engine', '@ember/-internals/routing', '@ember/runloop'], function (_testCase, _abstractTestCase, _helpers, _controller, _runtime, _glimmer, _engine, _routing, _runloop) {
  'use strict';

  (0, _testCase.moduleFor)('Application test: engine rendering', class extends _testCase.ApplicationTest {
    get routerOptions() {
      return {
        location: 'none',
        setupRouter() {
          this._super(...arguments);
          let getRoute = this._routerMicrolib.getRoute;
          this._enginePromises = Object.create(null);
          this._resolvedEngines = Object.create(null);

          this._routerMicrolib.getRoute = name => {
            let engineInfo = this._engineInfoByRoute[name];
            if (!engineInfo) {
              return getRoute(name);
            }

            let engineName = engineInfo.name;
            if (this._resolvedEngines[engineName]) {
              return getRoute(name);
            }

            let enginePromise = this._enginePromises[engineName];

            if (!enginePromise) {
              enginePromise = new _runtime.RSVP.Promise(resolve => {
                setTimeout(() => {
                  this._resolvedEngines[engineName] = true;

                  resolve();
                }, 1);
              });
              this._enginePromises[engineName] = enginePromise;
            }

            return enginePromise.then(() => getRoute(name));
          };
        }
      };
    }

    setupAppAndRoutableEngine(hooks = []) {
      let self = this;

      this.addTemplate('application', 'Application{{outlet}}');

      this.router.map(function () {
        this.mount('blog');
      });
      this.add('route-map:blog', function () {
        this.route('post', function () {
          this.route('comments');
          this.route('likes');
        });
        this.route('category', { path: 'category/:id' });
        this.route('author', { path: 'author/:id' });
      });
      this.add('route:application', _routing.Route.extend({
        model() {
          hooks.push('application - application');
        }
      }));

      this.add('engine:blog', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('controller:application', _controller.default.extend({
            queryParams: ['lang'],
            lang: ''
          }));
          this.register('controller:category', _controller.default.extend({
            queryParams: ['type']
          }));
          this.register('controller:authorKtrl', _controller.default.extend({
            queryParams: ['official']
          }));
          this.register('template:application', (0, _helpers.compile)('Engine{{lang}}{{outlet}}'));
          this.register('route:application', _routing.Route.extend({
            model() {
              hooks.push('engine - application');
            }
          }));
          this.register('route:author', _routing.Route.extend({
            controllerName: 'authorKtrl'
          }));

          if (self._additionalEngineRegistrations) {
            self._additionalEngineRegistrations.call(this);
          }
        }
      }));
    }

    setupAppAndRoutelessEngine(hooks) {
      this.setupRoutelessEngine(hooks);

      this.add('engine:chat-engine', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)('Engine'));
          this.register('controller:application', _controller.default.extend({
            init() {
              this._super(...arguments);
              hooks.push('engine - application');
            }
          }));
        }
      }));
    }

    setupAppAndRoutableEngineWithPartial(hooks) {
      this.addTemplate('application', 'Application{{outlet}}');

      this.router.map(function () {
        this.mount('blog');
      });
      this.add('route-map:blog', function () {});
      this.add('route:application', _routing.Route.extend({
        model() {
          hooks.push('application - application');
        }
      }));

      this.add('engine:blog', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:foo', (0, _helpers.compile)('foo partial'));
          this.register('template:application', (0, _helpers.compile)('Engine{{outlet}} {{partial "foo"}}'));
          this.register('route:application', _routing.Route.extend({
            model() {
              hooks.push('engine - application');
            }
          }));
        }
      }));
    }

    setupRoutelessEngine(hooks) {
      this.addTemplate('application', 'Application{{mount "chat-engine"}}');
      this.add('route:application', _routing.Route.extend({
        model() {
          hooks.push('application - application');
        }
      }));
    }

    setupAppAndRoutlessEngineWithPartial(hooks) {
      this.setupRoutelessEngine(hooks);

      this.add('engine:chat-engine', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:foo', (0, _helpers.compile)('foo partial'));
          this.register('template:application', (0, _helpers.compile)('Engine {{partial "foo"}}'));
          this.register('controller:application', _controller.default.extend({
            init() {
              this._super(...arguments);
              hooks.push('engine - application');
            }
          }));
        }
      }));
    }

    additionalEngineRegistrations(callback) {
      this._additionalEngineRegistrations = callback;
    }

    setupEngineWithAttrs() {
      this.addTemplate('application', 'Application{{mount "chat-engine"}}');

      this.add('engine:chat-engine', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:components/foo-bar', (0, _helpers.compile)(`{{partial "troll"}}`));
          this.register('template:troll', (0, _helpers.compile)('{{attrs.wat}}'));
          this.register('controller:application', _controller.default.extend({
            contextType: 'Engine'
          }));
          this.register('template:application', (0, _helpers.compile)('Engine {{foo-bar wat=contextType}}'));
        }
      }));
    }

    stringsEndWith(str, suffix) {
      return str.indexOf(suffix, str.length - suffix.length) !== -1;
    }

    ['@test attrs in an engine']() {
      this.setupEngineWithAttrs([]);

      return this.visit('/').then(() => {
        this.assertText('ApplicationEngine Engine');
      });
    }

    ['@test sharing a template between engine and application has separate refinements']() {
      this.assert.expect(1);

      let sharedTemplate = (0, _helpers.compile)(_abstractTestCase.strip`
      <h1>{{contextType}}</h1>
      {{ambiguous-curlies}}

      {{outlet}}
    `);

      this.add('template:application', sharedTemplate);
      this.add('controller:application', _controller.default.extend({
        contextType: 'Application',
        'ambiguous-curlies': 'Controller Data!'
      }));

      this.router.map(function () {
        this.mount('blog');
      });
      this.add('route-map:blog', function () {});

      this.add('engine:blog', _engine.default.extend({
        init() {
          this._super(...arguments);

          this.register('controller:application', _controller.default.extend({
            contextType: 'Engine'
          }));
          this.register('template:application', sharedTemplate);
          this.register('template:components/ambiguous-curlies', (0, _helpers.compile)(_abstractTestCase.strip`
        <p>Component!</p>
      `));
        }
      }));

      return this.visit('/blog').then(() => {
        this.assertText('ApplicationController Data!EngineComponent!');
      });
    }

    ['@test sharing a layout between engine and application has separate refinements']() {
      this.assert.expect(1);

      let sharedLayout = (0, _helpers.compile)(_abstractTestCase.strip`
      {{ambiguous-curlies}}
    `);

      let sharedComponent = _glimmer.Component.extend({
        layout: sharedLayout
      });

      this.addTemplate('application', _abstractTestCase.strip`
      <h1>Application</h1>
      {{my-component ambiguous-curlies="Local Data!"}}
      {{outlet}}
    `);

      this.add('component:my-component', sharedComponent);

      this.router.map(function () {
        this.mount('blog');
      });
      this.add('route-map:blog', function () {});

      this.add('engine:blog', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)(_abstractTestCase.strip`
          <h1>Engine</h1>
          {{my-component}}
          {{outlet}}
        `));
          this.register('component:my-component', sharedComponent);
          this.register('template:components/ambiguous-curlies', (0, _helpers.compile)(_abstractTestCase.strip`
          <p>Component!</p>
        `));
        }
      }));

      return this.visit('/blog').then(() => {
        this.assertText('ApplicationLocal Data!EngineComponent!');
      });
    }

    ['@test visit() with `shouldRender: true` returns a promise that resolves when application and engine templates have rendered'](assert) {
      assert.expect(2);

      let hooks = [];

      this.setupAppAndRoutableEngine(hooks);

      return this.visit('/blog', { shouldRender: true }).then(() => {
        this.assertText('ApplicationEngine');

        this.assert.deepEqual(hooks, ['application - application', 'engine - application'], 'the expected model hooks were fired');
      });
    }

    ['@test visit() with `shouldRender: false` returns a promise that resolves without rendering'](assert) {
      assert.expect(2);

      let hooks = [];

      this.setupAppAndRoutableEngine(hooks);

      return this.visit('/blog', { shouldRender: false }).then(() => {
        assert.strictEqual(document.getElementById('qunit-fixture').children.length, 0, `there are no elements in the qunit-fixture element`);

        this.assert.deepEqual(hooks, ['application - application', 'engine - application'], 'the expected model hooks were fired');
      });
    }

    ['@test visit() with `shouldRender: true` returns a promise that resolves when application and routeless engine templates have rendered'](assert) {
      assert.expect(2);

      let hooks = [];

      this.setupAppAndRoutelessEngine(hooks);

      return this.visit('/', { shouldRender: true }).then(() => {
        this.assertText('ApplicationEngine');

        this.assert.deepEqual(hooks, ['application - application', 'engine - application'], 'the expected hooks were fired');
      });
    }

    ['@test visit() with partials in routable engine'](assert) {
      assert.expect(2);

      let hooks = [];

      this.setupAppAndRoutableEngineWithPartial(hooks);

      return this.visit('/blog', { shouldRender: true }).then(() => {
        this.assertText('ApplicationEngine foo partial');

        this.assert.deepEqual(hooks, ['application - application', 'engine - application'], 'the expected hooks were fired');
      });
    }

    ['@test visit() with partials in non-routable engine'](assert) {
      assert.expect(2);

      let hooks = [];

      this.setupAppAndRoutlessEngineWithPartial(hooks);

      return this.visit('/', { shouldRender: true }).then(() => {
        this.assertText('ApplicationEngine foo partial');

        this.assert.deepEqual(hooks, ['application - application', 'engine - application'], 'the expected hooks were fired');
      });
    }

    ['@test deactivate should be called on Engine Routes before destruction'](assert) {
      assert.expect(3);

      this.setupAppAndRoutableEngine();

      this.add('engine:blog', _engine.default.extend({
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)('Engine{{outlet}}'));
          this.register('route:application', _routing.Route.extend({
            deactivate() {
              assert.notOk(this.isDestroyed, 'Route is not destroyed');
              assert.notOk(this.isDestroying, 'Route is not being destroyed');
            }
          }));
        }
      }));

      return this.visit('/blog').then(() => {
        this.assertText('ApplicationEngine');
      });
    }

    ['@test engine should lookup and use correct controller']() {
      this.setupAppAndRoutableEngine();

      return this.visit('/blog?lang=English').then(() => {
        this.assertText('ApplicationEngineEnglish');
      });
    }

    ['@test error substate route works for the application route of an Engine'](assert) {
      assert.expect(2);

      let errorEntered = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:application_error', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(errorEntered.resolve);
          }
        }));
        this.register('template:application_error', (0, _helpers.compile)('Error! {{model.message}}'));
        this.register('route:post', _routing.Route.extend({
          model() {
            return _runtime.RSVP.reject(new Error('Oh, noes!'));
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        return this.transitionTo('blog.post');
      }).then(() => {
        return errorEntered.promise;
      }).then(() => {
        this.assertText('ApplicationError! Oh, noes!');
      });
    }

    ['@test error route works for the application route of an Engine'](assert) {
      assert.expect(2);

      let errorEntered = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:error', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(errorEntered.resolve);
          }
        }));
        this.register('template:error', (0, _helpers.compile)('Error! {{model.message}}'));
        this.register('route:post', _routing.Route.extend({
          model() {
            return _runtime.RSVP.reject(new Error('Oh, noes!'));
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        return this.transitionTo('blog.post');
      }).then(() => {
        return errorEntered.promise;
      }).then(() => {
        this.assertText('ApplicationEngineError! Oh, noes!');
      });
    }

    ['@test error substate route works for a child route of an Engine'](assert) {
      assert.expect(2);

      let errorEntered = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:post_error', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(errorEntered.resolve);
          }
        }));
        this.register('template:post_error', (0, _helpers.compile)('Error! {{model.message}}'));
        this.register('route:post', _routing.Route.extend({
          model() {
            return _runtime.RSVP.reject(new Error('Oh, noes!'));
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        return this.transitionTo('blog.post');
      }).then(() => {
        return errorEntered.promise;
      }).then(() => {
        this.assertText('ApplicationEngineError! Oh, noes!');
      });
    }

    ['@test error route works for a child route of an Engine'](assert) {
      assert.expect(2);

      let errorEntered = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:post.error', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(errorEntered.resolve);
          }
        }));
        this.register('template:post.error', (0, _helpers.compile)('Error! {{model.message}}'));
        this.register('route:post.comments', _routing.Route.extend({
          model() {
            return _runtime.RSVP.reject(new Error('Oh, noes!'));
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        return this.transitionTo('blog.post.comments');
      }).then(() => {
        return errorEntered.promise;
      }).then(() => {
        this.assertText('ApplicationEngineError! Oh, noes!');
      });
    }

    ['@test loading substate route works for the application route of an Engine'](assert) {
      assert.expect(3);
      let done = assert.async();

      let loadingEntered = _runtime.RSVP.defer();
      let resolveLoading = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:application_loading', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(loadingEntered.resolve);
          }
        }));
        this.register('template:application_loading', (0, _helpers.compile)('Loading'));
        this.register('template:post', (0, _helpers.compile)('Post'));
        this.register('route:post', _routing.Route.extend({
          model() {
            return resolveLoading.promise;
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        let transition = this.transitionTo('blog.post');

        loadingEntered.promise.then(() => {
          this.assertText('ApplicationLoading');
          resolveLoading.resolve();

          return this.runTaskNext().then(() => {
            this.assertText('ApplicationEnginePost');
            done();
          });
        });

        return transition;
      });
    }

    ['@test loading route works for the application route of an Engine'](assert) {
      assert.expect(3);
      let done = assert.async();

      let loadingEntered = _runtime.RSVP.defer();
      let resolveLoading = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('route:loading', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(loadingEntered.resolve);
          }
        }));
        this.register('template:loading', (0, _helpers.compile)('Loading'));
        this.register('template:post', (0, _helpers.compile)('Post'));
        this.register('route:post', _routing.Route.extend({
          model() {
            return resolveLoading.promise;
          }
        }));
      });

      return this.visit('/').then(() => {
        this.assertText('Application');
        let transition = this.transitionTo('blog.post');

        loadingEntered.promise.then(() => {
          this.assertText('ApplicationEngineLoading');
          resolveLoading.resolve();

          return this.runTaskNext().then(() => {
            this.assertText('ApplicationEnginePost');
            done();
          });
        });

        return transition;
      });
    }

    ['@test loading substate route works for a child route of an Engine'](assert) {
      assert.expect(3);

      let resolveLoading;

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('template:post', (0, _helpers.compile)('{{outlet}}'));
        this.register('template:post.comments', (0, _helpers.compile)('Comments'));
        this.register('template:post.likes_loading', (0, _helpers.compile)('Loading'));
        this.register('template:post.likes', (0, _helpers.compile)('Likes'));
        this.register('route:post.likes', _routing.Route.extend({
          model() {
            return new _runtime.RSVP.Promise(resolve => {
              resolveLoading = resolve;
            });
          }
        }));
      });

      return this.visit('/blog/post/comments').then(() => {
        this.assertText('ApplicationEngineComments');
        let transition = this.transitionTo('blog.post.likes');

        this.runTaskNext().then(() => {
          this.assertText('ApplicationEngineLoading');
          resolveLoading();
        });

        return transition.then(() => this.runTaskNext()).then(() => this.assertText('ApplicationEngineLikes'));
      });
    }

    ['@test loading route works for a child route of an Engine'](assert) {
      assert.expect(3);
      let done = assert.async();

      let loadingEntered = _runtime.RSVP.defer();
      let resolveLoading = _runtime.RSVP.defer();

      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('template:post', (0, _helpers.compile)('{{outlet}}'));
        this.register('template:post.comments', (0, _helpers.compile)('Comments'));
        this.register('route:post.loading', _routing.Route.extend({
          activate() {
            (0, _runloop.next)(loadingEntered.resolve);
          }
        }));
        this.register('template:post.loading', (0, _helpers.compile)('Loading'));
        this.register('template:post.likes', (0, _helpers.compile)('Likes'));
        this.register('route:post.likes', _routing.Route.extend({
          model() {
            return resolveLoading.promise;
          }
        }));
      });

      return this.visit('/blog/post/comments').then(() => {
        this.assertText('ApplicationEngineComments');
        let transition = this.transitionTo('blog.post.likes');

        loadingEntered.promise.then(() => {
          this.assertText('ApplicationEngineLoading');
          resolveLoading.resolve();

          return this.runTaskNext().then(() => {
            this.assertText('ApplicationEngineLikes');
            done();
          });
        });

        return transition;
      });
    }

    ["@test query params don't have stickiness by default between model"](assert) {
      assert.expect(1);
      let tmpl = '{{#link-to "blog.category" 1337}}Category 1337{{/link-to}}';
      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('template:category', (0, _helpers.compile)(tmpl));
      });

      return this.visit('/blog/category/1?type=news').then(() => {
        let suffix = '/blog/category/1337';
        let href = this.element.querySelector('a').href;

        // check if link ends with the suffix
        assert.ok(this.stringsEndWith(href, suffix));
      });
    }

    ['@test query params in customized controllerName have stickiness by default between model'](assert) {
      assert.expect(2);
      let tmpl = '{{#link-to "blog.author" 1337 class="author-1337"}}Author 1337{{/link-to}}{{#link-to "blog.author" 1 class="author-1"}}Author 1{{/link-to}}';
      this.setupAppAndRoutableEngine();
      this.additionalEngineRegistrations(function () {
        this.register('template:author', (0, _helpers.compile)(tmpl));
      });

      return this.visit('/blog/author/1?official=true').then(() => {
        let suffix1 = '/blog/author/1?official=true';
        let href1 = this.element.querySelector('.author-1').href;
        let suffix1337 = '/blog/author/1337';
        let href1337 = this.element.querySelector('.author-1337').href;

        // check if link ends with the suffix
        assert.ok(this.stringsEndWith(href1, suffix1), `${href1} ends with ${suffix1}`);
        assert.ok(this.stringsEndWith(href1337, suffix1337), `${href1337} ends with ${suffix1337}`);
      });
    }

    ['@test visit() routable engine which errors on init'](assert) {
      assert.expect(1);

      let hooks = [];

      this.additionalEngineRegistrations(function () {
        this.register('route:application', _routing.Route.extend({
          init() {
            throw new Error('Whoops! Something went wrong...');
          }
        }));
      });

      this.setupAppAndRoutableEngine(hooks);

      return this.visit('/', { shouldRender: true }).then(() => {
        return this.visit('/blog');
      }).catch(error => {
        assert.equal(error.message, 'Whoops! Something went wrong...');
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/application/hot-reload-test', ['@ember/-internals/environment', '@ember/service', '@ember/-internals/glimmer', '@glimmer/util', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_environment, _service, _glimmer, _util, _testCase, _abstractTestCase) {
  'use strict';

  // This simuates what the template hot-reloading would do in development mode
  // to avoid regressions

  (0, _testCase.moduleFor)('Appliation test: template hot reloading', class extends _testCase.ApplicationTest {
    constructor() {
      super(...arguments);
      this._APPLICATION_TEMPLATE_WRAPPER = _environment.ENV._APPLICATION_TEMPLATE_WRAPPER;
      this._TEMPLATE_ONLY_GLIMMER_COMPONENTS = _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS;

      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = true;

      let didCreateReloader = reloader => {
        this.reloader = reloader;
      };

      this.add('service:reloader', _service.default.extend({
        init() {
          this._super(...arguments);
          this.revisions = {};
          this.callbacks = [];

          didCreateReloader(this);
        },

        onReload(callback) {
          this.callbacks.push(callback);
        },

        revisionFor(name) {
          return this.revisions[name];
        },

        invalidate(name) {
          let revision = this.revisions[name];

          if (revision === undefined) {
            revision = 0;
          }

          this.revisions[name] = ++revision;

          this.callbacks.forEach(callback => callback());
        }
      }));

      this.add('helper:hot-reload', _glimmer.Helper.extend({
        reloader: (0, _service.inject)(),

        init() {
          this._super(...arguments);
          this.reloader.onReload(() => this.recompute());
        },

        compute([name]) {
          let revision = this.reloader.revisionFor(name);

          if (revision === undefined) {
            return name;
          } else {
            return `${name}--hot-reload-${revision}`;
          }
        }
      }));
    }

    teardown() {
      super.teardown();
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = this._APPLICATION_TEMPLATE_WRAPPER;
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = this._TEMPLATE_ONLY_GLIMMER_COMPONENTS;
    }

    hotReload(name, template) {
      let reloader = (0, _util.expect)(this.reloader);
      let revision = (reloader.revisionFor(name) || 0) + 1;
      let ComponentClass = this.applicationInstance.resolveRegistration(`component:${name}`) || null;

      this.addComponent(`${name}--hot-reload-${revision}`, {
        ComponentClass,
        template
      });

      reloader.invalidate(name);
    }

    ['@test hot reloading template-only components']() {
      this.addTemplate('application', _abstractTestCase.strip`
          [{{component (hot-reload "x-foo") name="first"}}]
          [{{component (hot-reload "x-foo") name="second"}}]
          [{{component (hot-reload "x-bar")}}]
        `);

      this.addComponent('x-foo', {
        ComponentClass: null,
        template: 'x-foo: {{@name}}'
      });

      this.addComponent('x-bar', {
        ComponentClass: null,
        template: 'x-bar'
      });

      return this.visit('/').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
          [x-foo: first]
          [x-foo: second]
          [x-bar]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', '<h1>{{@name}}</h1>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<h1>first</h1>]
          [<h1>second</h1>]
          [x-bar]
        `);

        this.runTask(() => {
          this.hotReload('x-bar', '<h2>wow</h2>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<h1>first</h1>]
          [<h1>second</h1>]
          [<h2>wow</h2>]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', '<strong>x-foo</strong> <em>{{@name}}</em>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<strong>x-foo</strong> <em>first</em>]
          [<strong>x-foo</strong> <em>second</em>]
          [<h2>wow</h2>]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', 'x-foo: {{@name}}');
          this.hotReload('x-bar', 'x-bar');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [x-foo: first]
          [x-foo: second]
          [x-bar]
        `);
      });
    }

    ['@test hot reloading class-based components']() {
      this.addTemplate('application', _abstractTestCase.strip`
          [{{component (hot-reload "x-foo") name="first"}}]
          [{{component (hot-reload "x-foo") name="second"}}]
          [{{component (hot-reload "x-bar")}}]
        `);

      let id = 0;

      this.addComponent('x-foo', {
        ComponentClass: _glimmer.Component.extend({
          tagName: '',
          init() {
            this._super(...arguments);
            this.set('id', id++);
          }
        }),
        template: 'x-foo: {{@name}} ({{this.id}})'
      });

      this.addComponent('x-bar', {
        ComponentClass: _glimmer.Component.extend({
          tagName: '',
          init() {
            this._super(...arguments);
            this.set('id', id++);
          }
        }),
        template: 'x-bar ({{this.id}})'
      });

      return this.visit('/').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
          [x-foo: first (0)]
          [x-foo: second (1)]
          [x-bar (2)]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', '<h1>{{@name}} ({{this.id}})</h1>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<h1>first (3)</h1>]
          [<h1>second (4)</h1>]
          [x-bar (2)]
        `);

        this.runTask(() => {
          this.hotReload('x-bar', '<h2>wow ({{this.id}})</h2>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<h1>first (3)</h1>]
          [<h1>second (4)</h1>]
          [<h2>wow (5)</h2>]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', '<strong>x-foo</strong> <em>{{@name}} ({{this.id}})</em>');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [<strong>x-foo</strong> <em>first (6)</em>]
          [<strong>x-foo</strong> <em>second (7)</em>]
          [<h2>wow (5)</h2>]
        `);

        this.runTask(() => {
          this.hotReload('x-foo', 'x-foo: {{@name}} ({{this.id}})');
          this.hotReload('x-bar', 'x-bar ({{this.id}})');
        });

        this.assertInnerHTML(_abstractTestCase.strip`
          [x-foo: first (8)]
          [x-foo: second (9)]
          [x-bar (10)]
        `);
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/application/rendering-test', ['@ember/-internals/environment', '@ember/controller', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/routing', '@ember/-internals/glimmer'], function (_environment, _controller, _testCase, _abstractTestCase, _routing, _glimmer) {
  'use strict';

  (0, _testCase.moduleFor)('Application test: rendering', class extends _testCase.ApplicationTest {
    constructor() {
      super(...arguments);
      this._APPLICATION_TEMPLATE_WRAPPER = _environment.ENV._APPLICATION_TEMPLATE_WRAPPER;
    }

    teardown() {
      super.teardown();
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = this._APPLICATION_TEMPLATE_WRAPPER;
    }

    ['@test it can render the application template with a wrapper']() {
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = true;

      this.addTemplate('application', 'Hello world!');

      return this.visit('/').then(() => {
        this.assertComponentElement(this.element, { content: 'Hello world!' });
      });
    }

    ['@test it can render the application template without a wrapper']() {
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;

      this.addTemplate('application', 'Hello world!');

      return this.visit('/').then(() => {
        this.assertInnerHTML('Hello world!');
      });
    }

    ['@test it can access the model provided by the route']() {
      this.add('route:application', _routing.Route.extend({
        model() {
          return ['red', 'yellow', 'blue'];
        }
      }));

      this.addTemplate('application', _abstractTestCase.strip`
      <ul>
        {{#each model as |item|}}
          <li>{{item}}</li>
        {{/each}}
      </ul>
    `);

      return this.visit('/').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
        <ul>
          <li>red</li>
          <li>yellow</li>
          <li>blue</li>
        </ul>
      `);
      });
    }

    ['@test it can render a nested route']() {
      this.router.map(function () {
        this.route('lists', function () {
          this.route('colors', function () {
            this.route('favorite');
          });
        });
      });

      // The "favorite" route will inherit the model
      this.add('route:lists.colors', _routing.Route.extend({
        model() {
          return ['red', 'yellow', 'blue'];
        }
      }));

      this.addTemplate('lists.colors.favorite', _abstractTestCase.strip`
      <ul>
        {{#each model as |item|}}
          <li>{{item}}</li>
        {{/each}}
      </ul>
    `);

      return this.visit('/lists/colors/favorite').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
        <ul>
          <li>red</li>
          <li>yellow</li>
          <li>blue</li>
        </ul>
      `);
      });
    }

    ['@test it can render into named outlets']() {
      this.router.map(function () {
        this.route('colors');
      });

      this.addTemplate('application', _abstractTestCase.strip`
      <nav>{{outlet "nav"}}</nav>
      <main>{{outlet}}</main>
    `);

      this.addTemplate('nav', _abstractTestCase.strip`
      <a href="https://emberjs.com/">Ember</a>
    `);

      this.add('route:application', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('nav', {
            into: 'application',
            outlet: 'nav'
          });
        }
      }));

      this.add('route:colors', _routing.Route.extend({
        model() {
          return ['red', 'yellow', 'blue'];
        }
      }));

      this.addTemplate('colors', _abstractTestCase.strip`
      <ul>
        {{#each model as |item|}}
          <li>{{item}}</li>
        {{/each}}
      </ul>
    `);

      return this.visit('/colors').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
        <nav>
          <a href="https://emberjs.com/">Ember</a>
        </nav>
        <main>
          <ul>
            <li>red</li>
            <li>yellow</li>
            <li>blue</li>
          </ul>
        </main>
      `);
      });
    }

    ['@test it can render into named outlets']() {
      this.router.map(function () {
        this.route('colors');
      });

      this.addTemplate('application', _abstractTestCase.strip`
      <nav>{{outlet "nav"}}</nav>
      <main>{{outlet}}</main>
    `);

      this.addTemplate('nav', _abstractTestCase.strip`
      <a href="https://emberjs.com/">Ember</a>
    `);

      this.add('route:application', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('nav', {
            into: 'application',
            outlet: 'nav'
          });
        }
      }));

      this.add('route:colors', _routing.Route.extend({
        model() {
          return ['red', 'yellow', 'blue'];
        }
      }));

      this.addTemplate('colors', _abstractTestCase.strip`
      <ul>
        {{#each model as |item|}}
          <li>{{item}}</li>
        {{/each}}
      </ul>
    `);

      return this.visit('/colors').then(() => {
        this.assertInnerHTML(_abstractTestCase.strip`
        <nav>
          <a href="https://emberjs.com/">Ember</a>
        </nav>
        <main>
          <ul>
            <li>red</li>
            <li>yellow</li>
            <li>blue</li>
          </ul>
        </main>
      `);
      });
    }

    ['@test it should update the outlets when switching between routes']() {
      this.router.map(function () {
        this.route('a');
        this.route('b', function () {
          this.route('c');
          this.route('d');
        });
      });

      this.addTemplate('a', 'A{{outlet}}');
      this.addTemplate('b', 'B{{outlet}}');
      this.addTemplate('b.c', 'C');
      this.addTemplate('b.d', 'D');

      return this.visit('/b/c').then(() => {
        // this.assertComponentElement(this.firstChild, { content: 'BC' });
        this.assertText('BC');
        return this.visit('/a');
      }).then(() => {
        // this.assertComponentElement(this.firstChild, { content: 'A' });
        this.assertText('A');
        return this.visit('/b/d');
      }).then(() => {
        this.assertText('BD');
        // this.assertComponentElement(this.firstChild, { content: 'BD' });
      });
    }

    ['@test it should produce a stable DOM when the model changes']() {
      this.router.map(function () {
        this.route('color', { path: '/colors/:color' });
      });

      this.add('route:color', _routing.Route.extend({
        model(params) {
          return params.color;
        }
      }));

      this.addTemplate('color', 'color: {{model}}');

      return this.visit('/colors/red').then(() => {
        this.assertInnerHTML('color: red');
        this.takeSnapshot();
        return this.visit('/colors/green');
      }).then(() => {
        this.assertInnerHTML('color: green');
        this.assertInvariants();
      });
    }

    ['@test it should have the right controller in scope for the route template']() {
      this.router.map(function () {
        this.route('a');
        this.route('b');
      });

      this.add('controller:a', _controller.default.extend({
        value: 'a'
      }));

      this.add('controller:b', _controller.default.extend({
        value: 'b'
      }));

      this.addTemplate('a', '{{value}}');
      this.addTemplate('b', '{{value}}');

      return this.visit('/a').then(() => {
        this.assertText('a');
        return this.visit('/b');
      }).then(() => this.assertText('b'));
    }

    ['@test it should update correctly when the controller changes']() {
      this.router.map(function () {
        this.route('color', { path: '/colors/:color' });
      });

      this.add('route:color', _routing.Route.extend({
        model(params) {
          return { color: params.color };
        },

        renderTemplate(controller, model) {
          this.render({ controller: model.color, model });
        }
      }));

      this.add('controller:red', _controller.default.extend({
        color: 'red'
      }));

      this.add('controller:green', _controller.default.extend({
        color: 'green'
      }));

      this.addTemplate('color', 'model color: {{model.color}}, controller color: {{color}}');

      return this.visit('/colors/red').then(() => {
        this.assertInnerHTML('model color: red, controller color: red');
        return this.visit('/colors/green');
      }).then(() => {
        this.assertInnerHTML('model color: green, controller color: green');
      });
    }

    ['@test it should produce a stable DOM when two routes render the same template']() {
      this.router.map(function () {
        this.route('a');
        this.route('b');
      });

      this.add('route:a', _routing.Route.extend({
        model() {
          return 'A';
        },

        renderTemplate(controller, model) {
          this.render('common', { controller: 'common', model });
        }
      }));

      this.add('route:b', _routing.Route.extend({
        model() {
          return 'B';
        },

        renderTemplate(controller, model) {
          this.render('common', { controller: 'common', model });
        }
      }));

      this.add('controller:common', _controller.default.extend({
        prefix: 'common'
      }));

      this.addTemplate('common', '{{prefix}} {{model}}');

      return this.visit('/a').then(() => {
        this.assertInnerHTML('common A');
        this.takeSnapshot();
        return this.visit('/b');
      }).then(() => {
        this.assertInnerHTML('common B');
        this.assertInvariants();
      });
    }

    // Regression test, glimmer child outlets tried to assume the first element.
    // but the if put-args clobbered the args used by did-create-element.
    // I wish there was a way to assert that the OutletComponentManager did not
    // receive a didCreateElement.
    ['@test a child outlet is always a fragment']() {
      this.addTemplate('application', '{{outlet}}');
      this.addTemplate('index', '{{#if true}}1{{/if}}<div>2</div>');
      return this.visit('/').then(() => {
        this.assertInnerHTML('1<div>2</div>');
      });
    }

    ['@test it allows a transition during route activate']() {
      this.router.map(function () {
        this.route('a');
      });

      this.add('route:index', _routing.Route.extend({
        activate() {
          this.transitionTo('a');
        }
      }));

      this.addTemplate('a', 'Hello from A!');

      return this.visit('/').then(() => {
        this.assertInnerHTML('Hello from A!');
      });
    }

    ['@test it emits a useful backtracking re-render assertion message']() {
      this.router.map(function () {
        this.route('routeWithError');
      });

      this.add('route:routeWithError', _routing.Route.extend({
        model() {
          return { name: 'Alex' };
        }
      }));

      this.addTemplate('routeWithError', 'Hi {{model.name}} {{x-foo person=model}}');

      this.addComponent('x-foo', {
        ComponentClass: _glimmer.Component.extend({
          init() {
            this._super(...arguments);
            this.set('person.name', 'Ben');
          }
        }),
        template: 'Hi {{person.name}} from component'
      });

      let expectedBacktrackingMessage = /modified "model\.name" twice on \[object Object\] in a single render\. It was rendered in "template:my-app\/templates\/routeWithError.hbs" and modified in "component:x-foo"/;

      return this.visit('/').then(() => {
        expectAssertion(() => {
          this.visit('/routeWithError');
        }, expectedBacktrackingMessage);
      });
    }

    ['@test route templates with {{{undefined}}} [GH#14924] [GH#16172]']() {
      this.router.map(function () {
        this.route('first');
        this.route('second');
      });

      this.addTemplate('first', 'first');
      this.addTemplate('second', '{{{undefined}}}second');

      return this.visit('/first').then(() => {
        this.assertText('first');
        return this.visit('/second');
      }).then(() => {
        this.assertText('second');
        return this.visit('/first');
      }).then(() => {
        this.assertText('first');
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/angle-bracket-invocation-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-helpers'], function (_testCase, _metal, _helpers, _abstractTestCase, _testHelpers) {
  'use strict';

  if (true /* EMBER_GLIMMER_ANGLE_BRACKET_INVOCATION */) {
      (0, _testCase.moduleFor)('AngleBracket Invocation', class extends _testCase.RenderingTest {
        '@test it can resolve <XBlah /> to x-blah'() {
          this.registerComponent('x-blah', { template: 'hello' });

          this.render('<XBlah />');

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, { content: 'hello' });
        }

        '@test it can resolve <X-Blah /> to x-blah'() {
          this.registerComponent('x-blah', { template: 'hello' });

          this.render('<X-Blah />');

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, { content: 'hello' });
        }

        '@test it can render a basic template only component'() {
          this.registerComponent('foo-bar', { template: 'hello' });

          this.render('<FooBar />');

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, { content: 'hello' });
        }

        '@test it can render a basic component with template and javascript'() {
          this.registerComponent('foo-bar', {
            template: 'FIZZ BAR {{local}}',
            ComponentClass: _helpers.Component.extend({ local: 'hey' })
          });

          this.render('<FooBar />');

          this.assertComponentElement(this.firstChild, { content: 'FIZZ BAR hey' });
        }

        '@test it can render a single word component name'() {
          this.registerComponent('foo', { template: 'hello' });

          this.render('<Foo />');

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, { content: 'hello' });
        }

        '@test it can not render a component name without initial capital letter'(assert) {
          this.registerComponent('div', {
            ComponentClass: _helpers.Component.extend({
              init() {
                assert.ok(false, 'should not have created component');
              }
            })
          });

          this.render('<div></div>');

          this.assertElement(this.firstChild, { tagName: 'div', content: '' });
        }

        '@test it can have a custom id and it is not bound'() {
          this.registerComponent('foo-bar', { template: '{{id}} {{elementId}}' });

          this.render('<FooBar @id={{customId}} />', {
            customId: 'bizz'
          });

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { id: 'bizz' },
            content: 'bizz bizz'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { id: 'bizz' },
            content: 'bizz bizz'
          });

          this.runTask(() => (0, _metal.set)(this.context, 'customId', 'bar'));

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { id: 'bizz' },
            content: 'bar bizz'
          });

          this.runTask(() => (0, _metal.set)(this.context, 'customId', 'bizz'));

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { id: 'bizz' },
            content: 'bizz bizz'
          });
        }

        '@test it can have a custom tagName'() {
          let FooBarComponent = _helpers.Component.extend({
            tagName: 'foo-bar'
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: 'hello'
          });

          this.render('<FooBar></FooBar>');

          this.assertComponentElement(this.firstChild, {
            tagName: 'foo-bar',
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            tagName: 'foo-bar',
            content: 'hello'
          });
        }

        '@test it can have a custom tagName from the invocation'() {
          this.registerComponent('foo-bar', { template: 'hello' });

          this.render('<FooBar @tagName="foo-bar" />');

          this.assertComponentElement(this.firstChild, {
            tagName: 'foo-bar',
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            tagName: 'foo-bar',
            content: 'hello'
          });
        }

        '@test it can have custom classNames'() {
          let FooBarComponent = _helpers.Component.extend({
            classNames: ['foo', 'bar']
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: 'hello'
          });

          this.render('<FooBar />');

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar') },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar') },
            content: 'hello'
          });
        }

        '@test class property on components can be dynamic'() {
          this.registerComponent('foo-bar', { template: 'hello' });

          this.render('<FooBar @class={{if fooBar "foo-bar"}} />', {
            fooBar: true
          });

          this.assertComponentElement(this.firstChild, {
            content: 'hello',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            content: 'hello',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
          });

          this.runTask(() => (0, _metal.set)(this.context, 'fooBar', false));

          this.assertComponentElement(this.firstChild, {
            content: 'hello',
            attrs: { class: (0, _testHelpers.classes)('ember-view') }
          });

          this.runTask(() => (0, _metal.set)(this.context, 'fooBar', true));

          this.assertComponentElement(this.firstChild, {
            content: 'hello',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
          });
        }

        '@test it can set custom classNames from the invocation'() {
          let FooBarComponent = _helpers.Component.extend({
            classNames: ['foo']
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: 'hello'
          });

          this.render(_abstractTestCase.strip`
          <FooBar @class="bar baz" />
          <FooBar @classNames="bar baz" />
          <FooBar />
        `);

          this.assertComponentElement(this.nthChild(0), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
            content: 'hello'
          });
          this.assertComponentElement(this.nthChild(1), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
            content: 'hello'
          });
          this.assertComponentElement(this.nthChild(2), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.nthChild(0), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
            content: 'hello'
          });
          this.assertComponentElement(this.nthChild(1), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
            content: 'hello'
          });
          this.assertComponentElement(this.nthChild(2), {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
            content: 'hello'
          });
        }

        '@test it has an element'() {
          let instance;

          let FooBarComponent = _helpers.Component.extend({
            init() {
              this._super();
              instance = this;
            }
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: 'hello'
          });

          this.render('<FooBar></FooBar>');

          let element1 = instance.element;

          this.assertComponentElement(element1, { content: 'hello' });

          this.runTask(() => this.rerender());

          let element2 = instance.element;

          this.assertComponentElement(element2, { content: 'hello' });

          this.assertSameNode(element2, element1);
        }

        '@test it has the right parentView and childViews'(assert) {
          let fooBarInstance, fooBarBazInstance;

          let FooBarComponent = _helpers.Component.extend({
            init() {
              this._super();
              fooBarInstance = this;
            }
          });

          let FooBarBazComponent = _helpers.Component.extend({
            init() {
              this._super();
              fooBarBazInstance = this;
            }
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: 'foo-bar {{foo-bar-baz}}'
          });
          this.registerComponent('foo-bar-baz', {
            ComponentClass: FooBarBazComponent,
            template: 'foo-bar-baz'
          });

          this.render('<FooBar />');
          this.assertText('foo-bar foo-bar-baz');

          assert.equal(fooBarInstance.parentView, this.component);
          assert.equal(fooBarBazInstance.parentView, fooBarInstance);

          assert.deepEqual(this.component.childViews, [fooBarInstance]);
          assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);

          this.runTask(() => this.rerender());
          this.assertText('foo-bar foo-bar-baz');

          assert.equal(fooBarInstance.parentView, this.component);
          assert.equal(fooBarBazInstance.parentView, fooBarInstance);

          assert.deepEqual(this.component.childViews, [fooBarInstance]);
          assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);
        }

        '@test it renders passed named arguments'() {
          this.registerComponent('foo-bar', {
            template: '{{@foo}}'
          });

          this.render('<FooBar @foo={{model.bar}} />', {
            model: {
              bar: 'Hola'
            }
          });

          this.assertText('Hola');

          this.runTask(() => this.rerender());

          this.assertText('Hola');

          this.runTask(() => this.context.set('model.bar', 'Hello'));

          this.assertText('Hello');

          this.runTask(() => this.context.set('model', { bar: 'Hola' }));

          this.assertText('Hola');
        }

        '@test it reflects named arguments as properties'() {
          this.registerComponent('foo-bar', {
            template: '{{foo}}'
          });

          this.render('<FooBar @foo={{model.bar}} />', {
            model: {
              bar: 'Hola'
            }
          });

          this.assertText('Hola');

          this.runTask(() => this.rerender());

          this.assertText('Hola');

          this.runTask(() => this.context.set('model.bar', 'Hello'));

          this.assertText('Hello');

          this.runTask(() => this.context.set('model', { bar: 'Hola' }));

          this.assertText('Hola');
        }

        '@test it can render a basic component with a block'() {
          this.registerComponent('foo-bar', {
            template: '{{yield}} - In component'
          });

          this.render('<FooBar>hello</FooBar>');

          this.assertComponentElement(this.firstChild, {
            content: 'hello - In component'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            content: 'hello - In component'
          });
        }

        '@test it can yield internal and external properties positionally'() {
          let instance;

          let FooBarComponent = _helpers.Component.extend({
            init() {
              this._super(...arguments);
              instance = this;
            },
            greeting: 'hello'
          });

          this.registerComponent('foo-bar', {
            ComponentClass: FooBarComponent,
            template: '{{yield greeting greetee.firstName}}'
          });

          this.render('<FooBar @greetee={{person}} as |greeting name|>{{name}} {{person.lastName}}, {{greeting}}</FooBar>', {
            person: {
              firstName: 'Joel',
              lastName: 'Kang'
            }
          });

          this.assertComponentElement(this.firstChild, {
            content: 'Joel Kang, hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            content: 'Joel Kang, hello'
          });

          this.runTask(() => (0, _metal.set)(this.context, 'person', {
            firstName: 'Dora',
            lastName: 'the Explorer'
          }));

          this.assertComponentElement(this.firstChild, {
            content: 'Dora the Explorer, hello'
          });

          this.runTask(() => (0, _metal.set)(instance, 'greeting', 'hola'));

          this.assertComponentElement(this.firstChild, {
            content: 'Dora the Explorer, hola'
          });

          this.runTask(() => {
            (0, _metal.set)(instance, 'greeting', 'hello');
            (0, _metal.set)(this.context, 'person', {
              firstName: 'Joel',
              lastName: 'Kang'
            });
          });

          this.assertComponentElement(this.firstChild, {
            content: 'Joel Kang, hello'
          });
        }

        '@test positional parameters are not allowed'() {
          this.registerComponent('sample-component', {
            ComponentClass: _helpers.Component.extend().reopenClass({
              positionalParams: ['first', 'second']
            }),
            template: '{{first}}{{second}}'
          });

          // this is somewhat silly as the browser "corrects" for these as
          // attribute names, but regardless the thing we care about here is that
          // they are **not** used as positional params
          this.render('<SampleComponent one two />');

          this.assertText('');
        }

        '@test can invoke curried components with capitalized block param names'() {
          this.registerComponent('foo-bar', { template: 'hello' });

          this.render(_abstractTestCase.strip`
          {{#with (component 'foo-bar') as |Other|}}
            <Other />
          {{/with}}
        `);

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, { content: 'hello' });

          this.assertStableRerender();
        }

        '@test can invoke curried components with named args'() {
          this.registerComponent('foo-bar', { template: 'hello' });
          this.registerComponent('test-harness', { template: '<@foo />' });
          this.render(_abstractTestCase.strip`{{test-harness foo=(component 'foo-bar')}}`);

          this.assertComponentElement(this.firstChild.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild.firstChild, { content: 'hello' });

          this.assertStableRerender();
        }

        '@test can invoke curried components with a path'() {
          this.registerComponent('foo-bar', { template: 'hello' });
          this.registerComponent('test-harness', { template: '<this.foo />' });
          this.render(_abstractTestCase.strip`{{test-harness foo=(component 'foo-bar')}}`);

          this.assertComponentElement(this.firstChild.firstChild, { content: 'hello' });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild.firstChild, { content: 'hello' });

          this.assertStableRerender();
        }

        '@test can not invoke curried components with an implicit `this` path'(assert) {
          assert.expect(0);
          this.registerComponent('foo-bar', {
            template: 'hello',
            ComponentClass: _helpers.Component.extend({
              init() {
                this._super(...arguments);
                assert.ok(false, 'should not have instantiated');
              }
            })
          });
          this.registerComponent('test-harness', {
            template: '<foo.bar />'
          });
          this.render(_abstractTestCase.strip`{{test-harness foo=(hash bar=(component 'foo-bar'))}}`);
        }

        '@test has-block'() {
          this.registerComponent('check-block', {
            template: _abstractTestCase.strip`
            {{#if (has-block)}}
              Yes
            {{else}}
              No
            {{/if}}`
          });

          this.render(_abstractTestCase.strip`
          <CheckBlock />
          <CheckBlock></CheckBlock>`);

          this.assertComponentElement(this.firstChild, { content: 'No' });
          this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

          this.assertStableRerender();
        }

        '@test includes invocation specified attributes in root element ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend(),
            template: 'hello'
          });

          this.render('<FooBar data-foo={{foo}} data-bar={{bar}} />', { foo: 'foo', bar: 'bar' });

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'FOO');
            (0, _metal.set)(this.context, 'bar', undefined);
          });

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'FOO' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'foo');
            (0, _metal.set)(this.context, 'bar', 'bar');
          });

          this.assertComponentElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
        }

        '@test attributes without values passed at invocation are included in `...attributes` ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div ...attributes>hello</div>'
          });

          this.render('<FooBar data-bar />');

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-bar': '' },
            content: 'hello'
          });

          this.assertStableRerender();
        }

        '@test attributes without values at definition are included in `...attributes` ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div data-bar ...attributes>hello</div>'
          });

          this.render('<FooBar />');

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-bar': '' },
            content: 'hello'
          });

          this.assertStableRerender();
        }

        '@test includes invocation specified attributes in `...attributes` slot in tagless component ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div ...attributes>hello</div>'
          });

          this.render('<FooBar data-foo={{foo}} data-bar={{bar}} />', { foo: 'foo', bar: 'bar' });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'FOO');
            (0, _metal.set)(this.context, 'bar', undefined);
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'FOO' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'foo');
            (0, _metal.set)(this.context, 'bar', 'bar');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
        }

        '@test merges attributes with `...attributes` in tagless component ("splattributes")'() {
          let instance;
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({
              tagName: '',
              init() {
                instance = this;
                this._super(...arguments);
                this.localProp = 'qux';
              }
            }),
            template: '<div data-derp={{localProp}} ...attributes>hello</div>'
          });

          this.render('<FooBar data-foo={{foo}} data-bar={{bar}} />', { foo: 'foo', bar: 'bar' });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-derp': 'qux', 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-derp': 'qux', 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'FOO');
            (0, _metal.set)(this.context, 'bar', undefined);
            (0, _metal.set)(instance, 'localProp', 'QUZ');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-derp': 'QUZ', 'data-foo': 'FOO' },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'foo');
            (0, _metal.set)(this.context, 'bar', 'bar');
            (0, _metal.set)(instance, 'localProp', 'qux');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-derp': 'qux', 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
        }

        '@test merges class attribute with `...attributes` in tagless component ("splattributes")'() {
          let instance;
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({
              tagName: '',
              init() {
                instance = this;
                this._super(...arguments);
                this.localProp = 'qux';
              }
            }),
            template: '<div class={{localProp}} ...attributes>hello</div>'
          });

          this.render('<FooBar class={{bar}} />', { bar: 'bar' });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('qux bar') },
            content: 'hello'
          });

          this.runTask(() => this.rerender());

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('qux bar') },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'bar', undefined);
            (0, _metal.set)(instance, 'localProp', 'QUZ');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('QUZ') },
            content: 'hello'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'bar', 'bar');
            (0, _metal.set)(instance, 'localProp', 'qux');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('qux bar') },
            content: 'hello'
          });
        }

        '@test merges class attribute with `...attributes` in yielded contextual component ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '{{yield (hash baz=(component "foo-bar/baz"))}}'
          });
          this.registerComponent('foo-bar/baz', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div class="default-class" ...attributes>hello</div>'
          });

          this.render('<FooBar as |fb|><fb.baz class="custom-class" title="foo"></fb.baz></FooBar>');

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { class: (0, _testHelpers.classes)('default-class custom-class'), title: 'foo' },
            content: 'hello'
          });
        }

        '@test the attributes passed on invocation trump over the default ones on elements with `...attributes` in yielded contextual component ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '{{yield (hash baz=(component "foo-bar/baz"))}}'
          });
          this.registerComponent('foo-bar/baz', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div title="bar" ...attributes>hello</div>'
          });

          this.render('<FooBar as |fb|><fb.baz title="foo"></fb.baz></FooBar>');

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { title: 'foo' },
            content: 'hello'
          });
        }

        '@test can include `...attributes` in multiple elements in tagless component ("splattributes")'() {
          this.registerComponent('foo-bar', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: '<div ...attributes>hello</div><p ...attributes>world</p>'
          });

          this.render('<FooBar data-foo={{foo}} data-bar={{bar}} />', { foo: 'foo', bar: 'bar' });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
          this.assertElement(this.nthChild(1), {
            tagName: 'p',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'world'
          });

          this.runTask(() => this.rerender());

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
          this.assertElement(this.nthChild(1), {
            tagName: 'p',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'world'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'FOO');
            (0, _metal.set)(this.context, 'bar', undefined);
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'FOO' },
            content: 'hello'
          });
          this.assertElement(this.nthChild(1), {
            tagName: 'p',
            attrs: { 'data-foo': 'FOO' },
            content: 'world'
          });

          this.runTask(() => {
            (0, _metal.set)(this.context, 'foo', 'foo');
            (0, _metal.set)(this.context, 'bar', 'bar');
          });

          this.assertElement(this.firstChild, {
            tagName: 'div',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'hello'
          });
          this.assertElement(this.nthChild(1), {
            tagName: 'p',
            attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
            content: 'world'
          });
        }
      });
    }
});
enifed('@ember/-internals/glimmer/tests/integration/components/append-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_metal, _testCase, _helpers, _abstractTestCase) {
  'use strict';

  class AbstractAppendTest extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);

      this.components = [];
      this.ids = [];
    }

    teardown() {
      this.component = null;

      this.components.forEach(component => {
        this.runTask(() => component.destroy());
      });

      this.ids.forEach(id => {
        let $element = document.getElementById(id);
        if ($element) {
          $element.parentNode.removeChild($element);
        }
        // this.assert.strictEqual($element.length, 0, `Should not leak element: #${id}`);
      });

      super.teardown();
    }

    /* abstract append(component): Element; */

    didAppend(component) {
      this.components.push(component);
      this.ids.push(component.elementId);
    }

    ['@test lifecycle hooks during component append'](assert) {
      let hooks = [];

      let oldRegisterComponent = this.registerComponent;
      let componentsByName = {};

      // TODO: refactor/combine with other life-cycle tests
      this.registerComponent = function (name, _options) {
        function pushHook(hookName) {
          hooks.push([name, hookName]);
        }

        let options = {
          ComponentClass: _options.ComponentClass.extend({
            init() {
              this._super(...arguments);
              if (name in componentsByName) {
                throw new TypeError('Component named: ` ' + name + ' ` already registered');
              }
              componentsByName[name] = this;
              pushHook('init');
              this.on('init', () => pushHook('on(init)'));
            },

            didReceiveAttrs() {
              pushHook('didReceiveAttrs');
            },

            willInsertElement() {
              pushHook('willInsertElement');
            },

            willRender() {
              pushHook('willRender');
            },

            didInsertElement() {
              pushHook('didInsertElement');
            },

            didRender() {
              pushHook('didRender');
            },

            didUpdateAttrs() {
              pushHook('didUpdateAttrs');
            },

            willUpdate() {
              pushHook('willUpdate');
            },

            didUpdate() {
              pushHook('didUpdate');
            },

            willDestroyElement() {
              pushHook('willDestroyElement');
            },

            willClearRender() {
              pushHook('willClearRender');
            },

            didDestroyElement() {
              pushHook('didDestroyElement');
            },

            willDestroy() {
              pushHook('willDestroy');
              this._super(...arguments);
            }
          }),
          template: _options.template
        };

        oldRegisterComponent.call(this, name, options);
      };

      this.registerComponent('x-parent', {
        ComponentClass: _helpers.Component.extend({
          layoutName: 'components/x-parent'
        }),

        template: '[parent: {{foo}}]{{#x-child bar=foo}}[yielded: {{foo}}]{{/x-child}}'
      });

      this.registerComponent('x-child', {
        ComponentClass: _helpers.Component.extend({
          tagName: ''
        }),

        template: '[child: {{bar}}]{{yield}}'
      });

      let XParent;

      XParent = this.owner.factoryFor('component:x-parent');

      this.component = XParent.create({ foo: 'zomg' });

      assert.deepEqual(hooks, [['x-parent', 'init'], ['x-parent', 'on(init)']], 'creation of x-parent');

      hooks.length = 0;

      this.element = this.append(this.component);

      assert.deepEqual(hooks, [['x-parent', 'willInsertElement'], ['x-child', 'init'], ['x-child', 'on(init)'], ['x-child', 'didReceiveAttrs'], ['x-child', 'willRender'], ['x-child', 'willInsertElement'], ['x-child', 'didInsertElement'], ['x-child', 'didRender'], ['x-parent', 'didInsertElement'], ['x-parent', 'didRender']], 'appending of x-parent');

      hooks.length = 0;

      this.runTask(() => componentsByName['x-parent'].rerender());

      assert.deepEqual(hooks, [['x-parent', 'willUpdate'], ['x-parent', 'willRender'], ['x-parent', 'didUpdate'], ['x-parent', 'didRender']], 'rerender x-parent');

      hooks.length = 0;

      this.runTask(() => componentsByName['x-child'].rerender());

      assert.deepEqual(hooks, [['x-parent', 'willUpdate'], ['x-parent', 'willRender'], ['x-child', 'willUpdate'], ['x-child', 'willRender'], ['x-child', 'didUpdate'], ['x-child', 'didRender'], ['x-parent', 'didUpdate'], ['x-parent', 'didRender']], 'rerender x-child');

      hooks.length = 0;

      this.runTask(() => (0, _metal.set)(this.component, 'foo', 'wow'));

      assert.deepEqual(hooks, [['x-parent', 'willUpdate'], ['x-parent', 'willRender'], ['x-child', 'didUpdateAttrs'], ['x-child', 'didReceiveAttrs'], ['x-child', 'willUpdate'], ['x-child', 'willRender'], ['x-child', 'didUpdate'], ['x-child', 'didRender'], ['x-parent', 'didUpdate'], ['x-parent', 'didRender']], 'set foo = wow');

      hooks.length = 0;

      this.runTask(() => (0, _metal.set)(this.component, 'foo', 'zomg'));

      assert.deepEqual(hooks, [['x-parent', 'willUpdate'], ['x-parent', 'willRender'], ['x-child', 'didUpdateAttrs'], ['x-child', 'didReceiveAttrs'], ['x-child', 'willUpdate'], ['x-child', 'willRender'], ['x-child', 'didUpdate'], ['x-child', 'didRender'], ['x-parent', 'didUpdate'], ['x-parent', 'didRender']], 'set foo = zomg');

      hooks.length = 0;

      this.runTask(() => this.component.destroy());

      assert.deepEqual(hooks, [['x-parent', 'willDestroyElement'], ['x-parent', 'willClearRender'], ['x-child', 'willDestroyElement'], ['x-child', 'willClearRender'], ['x-child', 'didDestroyElement'], ['x-parent', 'didDestroyElement'], ['x-parent', 'willDestroy'], ['x-child', 'willDestroy']], 'destroy');
    }

    ['@test appending, updating and destroying a single component'](assert) {
      let willDestroyCalled = 0;

      this.registerComponent('x-parent', {
        ComponentClass: _helpers.Component.extend({
          layoutName: 'components/x-parent',
          willDestroyElement() {
            willDestroyCalled++;
          }
        }),

        template: '[parent: {{foo}}]{{#x-child bar=foo}}[yielded: {{foo}}]{{/x-child}}'
      });

      this.registerComponent('x-child', {
        ComponentClass: _helpers.Component.extend({
          tagName: ''
        }),

        template: '[child: {{bar}}]{{yield}}'
      });

      let XParent;

      XParent = this.owner.factoryFor('component:x-parent');

      this.component = XParent.create({ foo: 'zomg' });

      assert.ok(!this.component.element, 'precond - should not have an element');

      this.element = this.append(this.component);

      let componentElement = this.component.element;

      this.assertComponentElement(componentElement, {
        content: '[parent: zomg][child: zomg][yielded: zomg]'
      });

      assert.equal(componentElement.parentElement, this.element, 'It should be attached to the target');

      this.runTask(() => this.rerender());

      this.assertComponentElement(componentElement, {
        content: '[parent: zomg][child: zomg][yielded: zomg]'
      });

      assert.equal(componentElement.parentElement, this.element, 'It should be attached to the target');

      this.runTask(() => (0, _metal.set)(this.component, 'foo', 'wow'));

      this.assertComponentElement(componentElement, {
        content: '[parent: wow][child: wow][yielded: wow]'
      });

      assert.equal(componentElement.parentElement, this.element, 'It should be attached to the target');

      this.runTask(() => (0, _metal.set)(this.component, 'foo', 'zomg'));

      this.assertComponentElement(componentElement, {
        content: '[parent: zomg][child: zomg][yielded: zomg]'
      });

      assert.equal(componentElement.parentElement, this.element, 'It should be attached to the target');

      this.runTask(() => this.component.destroy());

      assert.ok(!this.component.element, 'It should not have an element');
      assert.ok(!componentElement.parentElement, 'The component element should be detached');

      this.assert.equal(willDestroyCalled, 1);
    }

    ['@test releasing a root component after it has been destroy'](assert) {
      let renderer = this.owner.lookup('renderer:-dom');

      this.registerComponent('x-component', {
        ComponentClass: _helpers.Component.extend()
      });

      this.component = this.owner.factoryFor('component:x-component').create();
      this.append(this.component);

      assert.equal(renderer._roots.length, 1, 'added a root component');

      this.runTask(() => this.component.destroy());

      assert.equal(renderer._roots.length, 0, 'released the root component');
    }

    ['@test appending, updating and destroying multiple components'](assert) {
      let willDestroyCalled = 0;

      this.registerComponent('x-first', {
        ComponentClass: _helpers.Component.extend({
          layoutName: 'components/x-first',

          willDestroyElement() {
            willDestroyCalled++;
          }
        }),

        template: 'x-first {{foo}}!'
      });

      this.registerComponent('x-second', {
        ComponentClass: _helpers.Component.extend({
          layoutName: 'components/x-second',

          willDestroyElement() {
            willDestroyCalled++;
          }
        }),

        template: 'x-second {{bar}}!'
      });

      let First, Second;

      First = this.owner.factoryFor('component:x-first');
      Second = this.owner.factoryFor('component:x-second');

      let first = First.create({ foo: 'foo' });
      let second = Second.create({ bar: 'bar' });

      this.assert.ok(!first.element, 'precond - should not have an element');
      this.assert.ok(!second.element, 'precond - should not have an element');

      let wrapper1, wrapper2;

      this.runTask(() => wrapper1 = this.append(first));
      this.runTask(() => wrapper2 = this.append(second));

      let componentElement1 = first.element;
      let componentElement2 = second.element;

      this.assertComponentElement(componentElement1, { content: 'x-first foo!' });
      this.assertComponentElement(componentElement2, {
        content: 'x-second bar!'
      });

      assert.equal(componentElement1.parentElement, wrapper1, 'The first component should be attached to the target');
      assert.equal(componentElement2.parentElement, wrapper2, 'The second component should be attached to the target');

      this.runTask(() => (0, _metal.set)(first, 'foo', 'FOO'));

      this.assertComponentElement(componentElement1, { content: 'x-first FOO!' });
      this.assertComponentElement(componentElement2, {
        content: 'x-second bar!'
      });

      assert.equal(componentElement1.parentElement, wrapper1, 'The first component should be attached to the target');
      assert.equal(componentElement2.parentElement, wrapper2, 'The second component should be attached to the target');

      this.runTask(() => (0, _metal.set)(second, 'bar', 'BAR'));

      this.assertComponentElement(componentElement1, { content: 'x-first FOO!' });
      this.assertComponentElement(componentElement2, {
        content: 'x-second BAR!'
      });

      assert.equal(componentElement1.parentElement, wrapper1, 'The first component should be attached to the target');
      assert.equal(componentElement2.parentElement, wrapper2, 'The second component should be attached to the target');

      this.runTask(() => {
        (0, _metal.set)(first, 'foo', 'foo');
        (0, _metal.set)(second, 'bar', 'bar');
      });

      this.assertComponentElement(componentElement1, { content: 'x-first foo!' });
      this.assertComponentElement(componentElement2, {
        content: 'x-second bar!'
      });

      assert.equal(componentElement1.parentElement, wrapper1, 'The first component should be attached to the target');
      assert.equal(componentElement2.parentElement, wrapper2, 'The second component should be attached to the target');

      this.runTask(() => {
        first.destroy();
        second.destroy();
      });

      assert.ok(!first.element, 'The first component should not have an element');
      assert.ok(!second.element, 'The second component should not have an element');

      assert.ok(!componentElement1.parentElement, 'The first component element should be detached');
      assert.ok(!componentElement2.parentElement, 'The second component element should be detached');

      this.assert.equal(willDestroyCalled, 2);
    }

    ['@test can appendTo while rendering']() {
      let owner = this.owner;

      let append = component => {
        return this.append(component);
      };

      let element1, element2;
      this.registerComponent('first-component', {
        ComponentClass: _helpers.Component.extend({
          layout: (0, _helpers.compile)('component-one'),

          didInsertElement() {
            element1 = this.element;

            let SecondComponent = owner.factoryFor('component:second-component');

            append(SecondComponent.create());
          }
        })
      });

      this.registerComponent('second-component', {
        ComponentClass: _helpers.Component.extend({
          layout: (0, _helpers.compile)(`component-two`),

          didInsertElement() {
            element2 = this.element;
          }
        })
      });

      let FirstComponent = this.owner.factoryFor('component:first-component');

      this.runTask(() => append(FirstComponent.create()));

      this.assertComponentElement(element1, { content: 'component-one' });
      this.assertComponentElement(element2, { content: 'component-two' });
    }

    ['@test can appendTo and remove while rendering'](assert) {
      let owner = this.owner;

      let append = component => {
        return this.append(component);
      };

      let element1, element2, element3, element4, component1, component2;
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          layout: (0, _helpers.compile)('foo-bar'),

          init() {
            this._super(...arguments);
            component1 = this;
          },

          didInsertElement() {
            element1 = this.element;

            let OtherRoot = owner.factoryFor('component:other-root');

            this._instance = OtherRoot.create({
              didInsertElement() {
                element2 = this.element;
              }
            });

            append(this._instance);
          },

          willDestroy() {
            this._instance.destroy();
          }
        })
      });

      this.registerComponent('baz-qux', {
        ComponentClass: _helpers.Component.extend({
          layout: (0, _helpers.compile)('baz-qux'),

          init() {
            this._super(...arguments);
            component2 = this;
          },

          didInsertElement() {
            element3 = this.element;

            let OtherRoot = owner.factoryFor('component:other-root');

            this._instance = OtherRoot.create({
              didInsertElement() {
                element4 = this.element;
              }
            });

            append(this._instance);
          },

          willDestroy() {
            this._instance.destroy();
          }
        })
      });

      let instantiatedRoots = 0;
      let destroyedRoots = 0;
      this.registerComponent('other-root', {
        ComponentClass: _helpers.Component.extend({
          layout: (0, _helpers.compile)(`fake-thing: {{counter}}`),
          init() {
            this._super(...arguments);
            this.counter = instantiatedRoots++;
          },
          willDestroy() {
            destroyedRoots++;
            this._super(...arguments);
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#if showFooBar}}
        {{foo-bar}}
      {{else}}
        {{baz-qux}}
      {{/if}}
    `, { showFooBar: true });

      this.assertComponentElement(element1, {});
      this.assertComponentElement(element2, { content: 'fake-thing: 0' });
      assert.equal(instantiatedRoots, 1);

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'showFooBar', false));

      assert.equal(instantiatedRoots, 2);
      assert.equal(destroyedRoots, 1);

      this.assertComponentElement(element3, {});
      this.assertComponentElement(element4, { content: 'fake-thing: 1' });

      this.runTask(() => {
        component1.destroy();
        component2.destroy();
      });

      assert.equal(instantiatedRoots, 2);
      assert.equal(destroyedRoots, 2);
    }
  }

  (0, _testCase.moduleFor)('append: no arguments (attaching to document.body)', class extends AbstractAppendTest {
    append(component) {
      this.runTask(() => component.append());
      this.didAppend(component);
      return document.body;
    }
  });

  (0, _testCase.moduleFor)('appendTo: a selector', class extends AbstractAppendTest {
    append(component) {
      this.runTask(() => component.appendTo('#qunit-fixture'));
      this.didAppend(component);
      return document.getElementById('qunit-fixture');
    }

    ['@test raises an assertion when the target does not exist in the DOM'](assert) {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          layoutName: 'components/foo-bar'
        }),
        template: 'FOO BAR!'
      });

      let FooBar = this.owner.factoryFor('component:foo-bar');

      this.component = FooBar.create();

      assert.ok(!this.component.element, 'precond - should not have an element');

      this.runTask(() => {
        expectAssertion(() => {
          this.component.appendTo('#does-not-exist-in-dom');
        }, /You tried to append to \(#does-not-exist-in-dom\) but that isn't in the DOM/);
      });

      assert.ok(!this.component.element, 'component should not have an element');
    }
  });

  (0, _testCase.moduleFor)('appendTo: an element', class extends AbstractAppendTest {
    append(component) {
      let element = document.getElementById('qunit-fixture');
      this.runTask(() => component.appendTo(element));
      this.didAppend(component);
      return element;
    }
  });

  (0, _testCase.moduleFor)('appendTo: with multiple components', class extends AbstractAppendTest {
    append(component) {
      this.runTask(() => component.appendTo('#qunit-fixture'));
      this.didAppend(component);
      return document.getElementById('qunit-fixture');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/attribute-bindings-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/metal'], function (_testCase, _helpers, _abstractTestCase, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Attribute bindings integration', class extends _testCase.RenderingTest {
    ['@test it can have attribute bindings']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['foo:data-foo', 'bar:data-bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=foo bar=bar}}', { foo: 'foo', bar: 'bar' });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'FOO');
        (0, _metal.set)(this.context, 'bar', undefined);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'FOO' },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'foo');
        (0, _metal.set)(this.context, 'bar', 'bar');
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });
    }

    ['@test it can have attribute bindings with attrs']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['attrs.foo:data-foo', 'attrs.baz.bar:data-bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=model.foo baz=model.baz}}', {
        model: { foo: undefined, baz: { bar: 'bar' } }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { 'data-bar': 'bar' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { 'data-bar': 'bar' }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.foo', 'foo');
        (0, _metal.set)(this.context, 'model.baz.bar', undefined);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'foo' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        foo: undefined,
        baz: { bar: 'bar' }
      }));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { 'data-bar': 'bar' }
      });
    }

    ['@test it can have attribute bindings with a nested path']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['foo.bar:data-foo-bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=foo}}', { foo: { bar: 'foo-bar' } });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo-bar': 'foo-bar' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo-bar': 'foo-bar' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo.bar', 'FOO-BAR'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo-bar': 'FOO-BAR' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo.bar', undefined));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', undefined));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', { bar: 'foo-bar' }));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo-bar': 'foo-bar' },
        content: 'hello'
      });
    }

    ['@test handles non-microsyntax attributeBindings']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['type']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar type=submit}}', {
        submit: 'submit'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'submit' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'submit' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'submit', 'password'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'password' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'submit', null));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'submit', 'submit'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'submit' },
        content: 'hello'
      });
    }

    ['@test non-microsyntax attributeBindings cannot contain nested paths']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['foo.bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar foo=foo}}', { foo: { bar: 'foo-bar' } });
      }, /Illegal attributeBinding: 'foo.bar' is not a valid attribute name./);
    }

    ['@test normalizes attributeBindings for property names']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['tiTLe']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar tiTLe=name}}', {
        name: 'qux'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { title: 'qux' },
        content: 'hello'
      });

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'name', null));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'qux'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { title: 'qux' },
        content: 'hello'
      });
    }

    ['@test normalizes attributeBindings for attribute names']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['foo:data-FOO']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=foo}}', {
        foo: 'qux'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'qux' },
        content: 'hello'
      });

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'foo', null));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'qux'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': 'qux' },
        content: 'hello'
      });
    }

    ['@test  attributeBindings preserves case for mixed-case attributes']() {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'svg',
        attributeBindings: ['viewBox']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: ''
      });

      this.render('{{foo-bar viewBox=foo}}', {
        foo: '0 0 100 100'
      });

      this.assert.equal(this.firstChild.getAttribute('viewBox'), '0 0 100 100', 'viewBox attribute');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'foo', null));

      this.assert.ok(!this.firstChild.hasAttribute('viewBox'), 'viewBox attribute removed');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', '0 0 100 200'));

      this.assert.equal(this.firstChild.getAttribute('viewBox'), '0 0 100 200', 'viewBox attribute');
    }

    ['@test attributeBindings handles null/undefined']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['fizz', 'bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar fizz=fizz bar=bar}}', {
        fizz: null,
        bar: undefined
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'fizz', 'fizz');
        (0, _metal.set)(this.context, 'bar', 'bar');
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { fizz: 'fizz', bar: 'bar' },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'fizz', null);
        (0, _metal.set)(this.context, 'bar', undefined);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });
    }

    ['@test attributeBindings handles number value']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['size']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar size=size}}', {
        size: 21
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { size: '21' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { size: '21' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'size', 0));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { size: '0' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'size', 21));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { size: '21' },
        content: 'hello'
      });
    }

    ['@test handles internal and external changes']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['type'],
        type: 'password',
        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'password' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'password' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(component, 'type', 'checkbox'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'checkbox' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(component, 'type', 'password'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { type: 'password' },
        content: 'hello'
      });
    }

    ['@test can set attributeBindings on component with a different tagName']() {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'input',
        attributeBindings: ['type', 'isDisabled:disabled']
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar type=type isDisabled=disabled}}', {
        type: 'password',
        disabled: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'input',
        attrs: { type: 'password' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'input',
        attrs: { type: 'password' }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'type', 'checkbox');
        (0, _metal.set)(this.context, 'disabled', true);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'input',
        attrs: { type: 'checkbox', disabled: '' }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'type', 'password');
        (0, _metal.set)(this.context, 'disabled', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'input',
        attrs: { type: 'password' }
      });
    }

    ['@test should allow namespaced attributes in micro syntax']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['xlinkHref:xlink:href']
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar type=type xlinkHref=xlinkHref}}', {
        xlinkHref: '/foo.png'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'xlink:href': '/foo.png' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'xlink:href': '/foo.png' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'xlinkHref', '/lol.png'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'xlink:href': '/lol.png' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'xlinkHref', '/foo.png'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'xlink:href': '/foo.png' }
      });
    }

    // This comes into play when using the {{#each}} helper. If the
    // passed array item is a String, it will be converted into a
    // String object instead of a normal string.
    ['@test should allow for String objects']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['foo']
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar foo=foo}}', {
        foo: function () {
          return this;
        }.call('bar')
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { foo: 'bar' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { foo: 'bar' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', function () {
        return this;
      }.call('baz')));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { foo: 'baz' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', function () {
        return this;
      }.call('bar')));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { foo: 'bar' }
      });
    }

    ['@test can set id initially via attributeBindings ']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['specialSauce:id']
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar specialSauce=sauce}}', {
        sauce: 'special-sauce'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'special-sauce' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'special-sauce' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'sauce', 'foo'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'special-sauce' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'sauce', 'special-sauce'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'special-sauce' }
      });
    }

    ['@test attributeBindings are overwritten']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['href'],
        href: 'a href'
      });

      let FizzBarComponent = FooBarComponent.extend({
        attributeBindings: ['newHref:href']
      });

      this.registerComponent('fizz-bar', { ComponentClass: FizzBarComponent });

      this.render('{{fizz-bar newHref=href}}', {
        href: 'dog.html'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { href: 'dog.html' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { href: 'dog.html' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'href', 'cat.html'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { href: 'cat.html' }
      });
    }

    ['@test it can set attribute bindings in the constructor']() {
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();

          let bindings = [];

          if (this.get('hasFoo')) {
            bindings.push('foo:data-foo');
          }

          if (this.get('hasBar')) {
            bindings.push('bar:data-bar');
          }

          this.attributeBindings = bindings;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render(_abstractTestCase.strip`
      {{foo-bar hasFoo=true foo=foo hasBar=false bar=bar}}
      {{foo-bar hasFoo=false foo=foo hasBar=true bar=bar}}
      {{foo-bar hasFoo=true foo=foo hasBar=true bar=bar}}
      {{foo-bar hasFoo=false foo=foo hasBar=false bar=bar}}
    `, { foo: 'foo', bar: 'bar' });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'FOO');
        (0, _metal.set)(this.context, 'bar', undefined);
      });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { 'data-foo': 'FOO' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { 'data-foo': 'FOO' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'bar', 'BAR'));

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { 'data-foo': 'FOO' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { 'data-bar': 'BAR' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { 'data-foo': 'FOO', 'data-bar': 'BAR' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'foo');
        (0, _metal.set)(this.context, 'bar', 'bar');
      });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { 'data-foo': 'foo', 'data-bar': 'bar' },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: {},
        content: 'hello'
      });
    }

    ['@test asserts if an attributeBinding is setup on class']() {
      let FooBarComponent = _helpers.Component.extend({
        attributeBindings: ['class']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /You cannot use class as an attributeBinding, use classNameBindings instead./i);
    }

    ['@test blacklists href bindings based on protocol']() {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'a',
        attributeBindings: ['href']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar href=xss}}', {
        xss: "javascript:alert('foo')"
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'a',
        attrs: { href: "unsafe:javascript:alert('foo')" }
      });
    }

    ['@test it can bind the role attribute (issue #14007)']() {
      let FooBarComponent = _helpers.Component.extend({ attributeBindings: ['role'] });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar role=role}}', { role: 'button' });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { role: 'button' }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { role: 'button' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'role', 'combobox'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { role: 'combobox' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'role', null));

      this.assertComponentElement(this.firstChild, { tagName: 'div' });
    }

    ['@test component with an `id` attribute binding of undefined']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['id'],

          id: undefined
        })
      });

      this.registerComponent('baz-qux', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['somethingUndefined:id'],

          somethingUndefined: undefined
        })
      });
      this.render(`{{foo-bar}}{{baz-qux}}`);

      this.assertComponentElement(this.nthChild(0), { content: '' });
      this.assertComponentElement(this.nthChild(1), { content: '' });

      this.assert.ok(this.nthChild(0).id.match(/ember\d+/), 'a valid `id` was used');
      this.assert.ok(this.nthChild(1).id.match(/ember\d+/), 'a valid `id` was used');
    }

    ['@test component with an `id` attribute binding of null']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['id'],

          id: null
        })
      });

      this.registerComponent('baz-qux', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['somethingNull:id'],

          somethingNull: null
        })
      });
      this.render(`{{foo-bar}}{{baz-qux}}`);

      this.assertComponentElement(this.nthChild(0), { content: '' });
      this.assertComponentElement(this.nthChild(1), { content: '' });

      this.assert.ok(this.nthChild(0).id.match(/ember\d+/), 'a valid `id` was used');
      this.assert.ok(this.nthChild(1).id.match(/ember\d+/), 'a valid `id` was used');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/attrs-lookup-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/test-helpers'], function (_testCase, _helpers, _metal, _testHelpers) {
  'use strict';

  (0, _testCase.moduleFor)('Components test: attrs lookup', class extends _testCase.RenderingTest {
    ['@test it should be able to lookup attrs without `attrs.` - template access']() {
      this.registerComponent('foo-bar', { template: '{{first}}' });

      this.render(`{{foo-bar first=firstAttr}}`, {
        firstAttr: 'first attr'
      });

      this.assertText('first attr');

      this.runTask(() => this.rerender());

      this.assertText('first attr');

      this.runTask(() => (0, _metal.set)(this.context, 'firstAttr', 'second attr'));

      this.assertText('second attr');

      this.runTask(() => (0, _metal.set)(this.context, 'firstAttr', 'first attr'));

      this.assertText('first attr');
    }

    ['@test it should be able to lookup attrs without `attrs.` - component access'](assert) {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        }
      });
      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{first}}'
      });

      this.render(`{{foo-bar first=firstAttr}}`, {
        firstAttr: 'first attr'
      });

      assert.equal(instance.get('first'), 'first attr');

      this.runTask(() => this.rerender());

      assert.equal(instance.get('first'), 'first attr');

      this.runTask(() => (0, _metal.set)(this.context, 'firstAttr', 'second attr'));

      assert.equal(instance.get('first'), 'second attr');

      this.runTask(() => (0, _metal.set)(this.context, 'firstAttr', 'first attr'));

      this.assertText('first attr');
    }

    ['@test should be able to modify a provided attr into local state #11571 / #11559'](assert) {
      let instance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },

        didReceiveAttrs() {
          this.set('first', this.get('first').toUpperCase());
        }
      });
      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{first}}'
      });

      this.render(`{{foo-bar first="first attr"}}`);

      assert.equal(instance.get('first'), 'FIRST ATTR', 'component lookup uses local state');
      this.assertText('FIRST ATTR');

      this.runTask(() => this.rerender());

      assert.equal(instance.get('first'), 'FIRST ATTR', 'component lookup uses local state during rerender');
      this.assertText('FIRST ATTR');

      // This is testing that passing string literals for use as initial values,
      // so there is no update step
    }

    ['@test should be able to access unspecified attr #12035'](assert) {
      let instance;
      let wootVal = 'yes';

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },

        didReceiveAttrs() {
          assert.equal(this.get('woot'), wootVal, 'found attr in didReceiveAttrs');
        }
      });
      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render(`{{foo-bar woot=woot}}`, {
        woot: wootVal
      });

      assert.equal(instance.get('woot'), 'yes', 'component found attr');

      this.runTask(() => this.rerender());

      assert.equal(instance.get('woot'), 'yes', 'component found attr after rerender');

      this.runTask(() => {
        wootVal = 'nope';
        (0, _metal.set)(this.context, 'woot', wootVal);
      });

      assert.equal(instance.get('woot'), 'nope', 'component found attr after attr change');

      this.runTask(() => {
        wootVal = 'yes';
        (0, _metal.set)(this.context, 'woot', wootVal);
      });

      assert.equal(instance.get('woot'), 'yes', 'component found attr after reset');
    }

    ['@test getAttr() should return the same value as get()'](assert) {
      assert.expect(33);

      let instance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },

        didReceiveAttrs() {
          let rootFirstPositional = this.get('firstPositional');
          let rootFirst = this.get('first');
          let rootSecond = this.get('second');
          let attrFirstPositional = this.getAttr('firstPositional');
          let attrFirst = this.getAttr('first');
          let attrSecond = this.getAttr('second');

          assert.equal(rootFirstPositional, attrFirstPositional, 'root property matches attrs value');
          assert.equal(rootFirst, attrFirst, 'root property matches attrs value');
          assert.equal(rootSecond, attrSecond, 'root property matches attrs value');
        }
      });

      FooBarComponent.reopenClass({
        positionalParams: ['firstPositional']
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render(`{{foo-bar firstPositional first=first second=second}}`, {
        firstPositional: 'firstPositional',
        first: 'first',
        second: 'second'
      });

      assert.equal(instance.get('firstPositional'), 'firstPositional', 'matches known value');
      assert.equal(instance.get('first'), 'first', 'matches known value');
      assert.equal(instance.get('second'), 'second', 'matches known value');

      this.runTask(() => this.rerender());

      assert.equal(instance.get('firstPositional'), 'firstPositional', 'matches known value');
      assert.equal(instance.get('first'), 'first', 'matches known value');
      assert.equal(instance.get('second'), 'second', 'matches known value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'first', 'third');
      });

      assert.equal(instance.get('firstPositional'), 'firstPositional', 'matches known value');
      assert.equal(instance.get('first'), 'third', 'matches known value');
      assert.equal(instance.get('second'), 'second', 'matches known value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'second', 'fourth');
      });

      assert.equal(instance.get('firstPositional'), 'firstPositional', 'matches known value');
      assert.equal(instance.get('first'), 'third', 'matches known value');
      assert.equal(instance.get('second'), 'fourth', 'matches known value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'firstPositional', 'fifth');
      });

      assert.equal(instance.get('firstPositional'), 'fifth', 'matches known value');
      assert.equal(instance.get('first'), 'third', 'matches known value');
      assert.equal(instance.get('second'), 'fourth', 'matches known value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'firstPositional', 'firstPositional');
        (0, _metal.set)(this.context, 'first', 'first');
        (0, _metal.set)(this.context, 'second', 'second');
      });

      assert.equal(instance.get('firstPositional'), 'firstPositional', 'matches known value');
      assert.equal(instance.get('first'), 'first', 'matches known value');
      assert.equal(instance.get('second'), 'second', 'matches known value');
    }

    ['@test bound computed properties can be overridden in extensions, set during init, and passed in as attrs']() {
      let FooClass = _helpers.Component.extend({
        attributeBindings: ['style'],
        style: (0, _metal.computed)('height', 'color', function () {
          let height = this.get('height');
          let color = this.get('color');
          return (0, _helpers.htmlSafe)(`height: ${height}px; background-color: ${color};`);
        }),
        color: 'red',
        height: 20
      });

      let BarClass = FooClass.extend({
        init() {
          this._super(...arguments);
          this.height = 150;
        },
        color: 'yellow'
      });

      this.registerComponent('x-foo', { ComponentClass: FooClass });
      this.registerComponent('x-bar', { ComponentClass: BarClass });

      this.render('{{x-foo}}{{x-bar}}{{x-bar color="green"}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 20px; background-color: red;') }
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 150px; background-color: yellow;') }
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 150px; background-color: green;') }
      });

      this.assertStableRerender();

      // No U-R
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/class-bindings-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_testCase, _helpers, _testHelpers, _metal, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('ClassNameBindings integration', class extends _testCase.RenderingTest {
    ['@test it can have class name bindings on the class definition']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['foo', 'isEnabled:enabled', 'isHappy:happy:sad']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=foo isEnabled=isEnabled isHappy=isHappy}}', {
        foo: 'foo',
        isEnabled: true,
        isHappy: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'FOO');
        (0, _metal.set)(this.context, 'isEnabled', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO sad') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', undefined);
        (0, _metal.set)(this.context, 'isHappy', true);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view happy') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'foo');
        (0, _metal.set)(this.context, 'isEnabled', true);
        (0, _metal.set)(this.context, 'isHappy', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });
    }

    ['@test attrs in classNameBindings']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['attrs.joker:purple:green', 'attrs.batman.robin:black:red']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar joker=model.wat batman=model.super}}', {
        model: { wat: false, super: { robin: true } }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view green black') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view green black') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.wat', true);
        (0, _metal.set)(this.context, 'model.super.robin', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view purple red') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        wat: false,
        super: { robin: true }
      }));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view green black') },
        content: 'hello'
      });
    }

    ['@test it can have class name bindings in the template']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classNameBindings="model.someInitiallyTrueProperty model.someInitiallyFalseProperty model.someInitiallyUndefinedProperty :static model.isBig:big model.isOpen:open:closed model.isUp::down model.bar:isTruthy:isFalsy"}}', {
        model: {
          someInitiallyTrueProperty: true,
          someInitiallyFalseProperty: false,
          isBig: true,
          isOpen: false,
          isUp: true,
          bar: true
        }
      });

      this.assertComponentElement(this.firstChild, {
        attrs: {
          class: (0, _testHelpers.classes)('ember-view some-initially-true-property static big closed isTruthy')
        },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        attrs: {
          class: (0, _testHelpers.classes)('ember-view some-initially-true-property static big closed isTruthy')
        },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.someInitiallyTrueProperty', false);
        (0, _metal.set)(this.context, 'model.someInitiallyFalseProperty', true);
        (0, _metal.set)(this.context, 'model.someInitiallyUndefinedProperty', true);
        (0, _metal.set)(this.context, 'model.isBig', false);
        (0, _metal.set)(this.context, 'model.isOpen', true);
        (0, _metal.set)(this.context, 'model.isUp', false);
        (0, _metal.set)(this.context, 'model.bar', false);
      });

      this.assertComponentElement(this.firstChild, {
        attrs: {
          class: (0, _testHelpers.classes)('ember-view some-initially-false-property some-initially-undefined-property static open down isFalsy')
        },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model', {
          someInitiallyTrueProperty: true,
          someInitiallyFalseProperty: false,
          someInitiallyUndefinedProperty: undefined,
          isBig: true,
          isOpen: false,
          isUp: true,
          bar: true
        });
      });

      this.assertComponentElement(this.firstChild, {
        attrs: {
          class: (0, _testHelpers.classes)('ember-view some-initially-true-property static big closed isTruthy')
        },
        content: 'hello'
      });
    }

    ['@test it can have class name bindings with nested paths']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['foo.bar', 'is.enabled:enabled', 'is.happy:happy:sad']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar foo=foo is=is}}', {
        foo: { bar: 'foo-bar' },
        is: { enabled: true, happy: false }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar enabled sad') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar enabled sad') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo.bar', 'FOO-BAR');
        (0, _metal.set)(this.context, 'is.enabled', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO-BAR sad') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo.bar', null);
        (0, _metal.set)(this.context, 'is.happy', true);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view happy') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', null);
        (0, _metal.set)(this.context, 'is', null);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view sad') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', { bar: 'foo-bar' });
        (0, _metal.set)(this.context, 'is', { enabled: true, happy: false });
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar enabled sad') },
        content: 'hello'
      });
    }

    ['@test it should dasherize the path when the it resolves to true']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['fooBar', 'nested.fooBarBaz']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar fooBar=fooBar nested=nested}}', {
        fooBar: true,
        nested: { fooBarBaz: false }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'fooBar', false);
        (0, _metal.set)(this.context, 'nested.fooBarBaz', true);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar-baz') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'fooBar', 'FOO-BAR');
        (0, _metal.set)(this.context, 'nested.fooBarBaz', null);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO-BAR') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'nested', null));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO-BAR') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'fooBar', true);
        (0, _metal.set)(this.context, 'nested', { fooBarBaz: false });
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') },
        content: 'hello'
      });
    }

    ['@test const bindings can be set as attrs']() {
      this.registerComponent('foo-bar', { template: 'hello' });
      this.render('{{foo-bar classNameBindings="foo:enabled:disabled"}}', {
        foo: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view enabled') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view enabled') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', false));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view disabled') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', true));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view enabled') },
        content: 'hello'
      });
    }

    ['@test :: class name syntax works with an empty true class']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['isEnabled::not-enabled']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar isEnabled=enabled}}', {
        enabled: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view not-enabled') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'enabled', true));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'enabled', false));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view not-enabled') },
        content: 'hello'
      });
    }

    ['@test uses all provided static class names (issue #11193)']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: [':class-one', ':class-two']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}', {
        enabled: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view class-one class-two') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'enabled', true));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view class-one class-two') },
        content: 'hello'
      });
    }

    ['@test Providing a binding with a space in it asserts']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: 'i:think:i am:so:clever'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /classNameBindings must not have spaces in them/i);
    }

    ['@test it asserts that items must be strings']() {
      let FooBarComponent = _helpers.Component.extend({
        foo: 'foo',
        bar: 'bar',
        classNameBindings: ['foo',, 'bar'] // eslint-disable-line no-sparse-arrays
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /classNameBindings must be non-empty strings/);
    }

    ['@test it asserts that items must be non-empty strings']() {
      let FooBarComponent = _helpers.Component.extend({
        foo: 'foo',
        bar: 'bar',
        classNameBindings: ['foo', '', 'bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /classNameBindings must be non-empty strings/);
    }

    ['@test it can set class name bindings in the constructor']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: ['foo'],

        init() {
          this._super();

          let bindings = this.classNameBindings = this.classNameBindings.slice();

          if (this.get('bindIsEnabled')) {
            bindings.push('isEnabled:enabled');
          }

          if (this.get('bindIsHappy')) {
            bindings.push('isHappy:happy:sad');
          }
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render(_abstractTestCase.strip`
      {{foo-bar foo=foo bindIsEnabled=true isEnabled=isEnabled bindIsHappy=false isHappy=isHappy}}
      {{foo-bar foo=foo bindIsEnabled=false isEnabled=isEnabled bindIsHappy=true isHappy=isHappy}}
      {{foo-bar foo=foo bindIsEnabled=true isEnabled=isEnabled bindIsHappy=true isHappy=isHappy}}
      {{foo-bar foo=foo bindIsEnabled=false isEnabled=isEnabled bindIsHappy=false isHappy=isHappy}}
    `, { foo: 'foo', isEnabled: true, isHappy: false });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'FOO');
        (0, _metal.set)(this.context, 'isEnabled', false);
      });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view FOO') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', undefined);
        (0, _metal.set)(this.context, 'isHappy', true);
      });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view happy') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view happy') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') },
        content: 'hello'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'foo');
        (0, _metal.set)(this.context, 'isEnabled', true);
        (0, _metal.set)(this.context, 'isHappy', false);
      });

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo enabled sad') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(3), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
        content: 'hello'
      });
    }

    ['@test using a computed property for classNameBindings triggers an assertion']() {
      let FooBarComponent = _helpers.Component.extend({
        classNameBindings: (0, _metal.computed)(function () {
          return ['isHappy:happy:sad'];
        })
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /Only arrays are allowed/);
    }
  });

  (0, _testCase.moduleFor)('ClassBinding integration', class extends _testCase.RenderingTest {
    ['@test it should apply classBinding without condition always']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding=":foo"}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo  ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo  ember-view') }
      });
    }

    ['@test it should merge classBinding with class']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="birdman:respeck" class="myName"}}', {
        birdman: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('respeck myName ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('respeck myName ember-view') }
      });
    }

    ['@test it should apply classBinding with only truthy condition']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="myName:respeck"}}', {
        myName: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('respeck  ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('respeck  ember-view') }
      });
    }

    ['@test it should apply classBinding with only falsy condition']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="myName::shade"}}', {
        myName: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('shade  ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('shade  ember-view') }
      });
    }

    ['@test it should apply nothing when classBinding is falsy but only supplies truthy class']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="myName:respeck"}}', {
        myName: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });
    }

    ['@test it should apply nothing when classBinding is truthy but only supplies falsy class']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="myName::shade"}}', { myName: true });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });
    }

    ['@test it should apply classBinding with falsy condition']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="swag:fresh:scrub"}}', {
        swag: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('scrub  ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('scrub  ember-view') }
      });
    }

    ['@test it should apply classBinding with truthy condition']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar classBinding="swag:fresh:scrub"}}', {
        swag: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fresh  ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fresh  ember-view') }
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/contextual-components-test', ['@ember/polyfills', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/runtime'], function (_polyfills, _helpers, _abstractTestCase, _testCase, _metal, _runtime) {
  'use strict';

  (0, _testCase.moduleFor)('Components test: contextual components', class extends _testCase.RenderingTest {
    ['@test renders with component helper']() {
      let expectedText = 'Hodi';

      this.registerComponent('-looked-up', {
        template: expectedText
      });

      this.render('{{component (component "-looked-up")}}');

      this.assertText(expectedText);

      this.runTask(() => this.rerender());

      this.assertText(expectedText);
    }

    ['@test renders with component helper with invocation params, hash']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{greeting}} {{name}}'
      });

      this.render(_abstractTestCase.strip`
      {{component (component "-looked-up") "Hodari" greeting="Hodi"}}`);

      this.assertText('Hodi Hodari');

      this.runTask(() => this.rerender());

      this.assertText('Hodi Hodari');
    }

    ['@test GH#13742 keeps nested rest positional parameters if rendered with no positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{component (component "-looked-up" model.greeting model.name)}}', {
        model: {
          greeting: 'Gabon ',
          name: 'Zack'
        }
      });

      this.assertText('Gabon Zack');

      this.runTask(() => this.rerender());

      this.assertText('Gabon Zack');

      this.runTask(() => this.context.set('model.greeting', 'Good morning '));

      this.assertText('Good morning Zack');

      this.runTask(() => this.context.set('model.name', 'Matthew'));

      this.assertText('Good morning Matthew');

      this.runTask(() => this.context.set('model', { greeting: 'Gabon ', name: 'Zack' }));

      this.assertText('Gabon Zack');
    }

    // Take a look at this one. Seems to pass even when currying isn't implemented.
    ['@test overwrites nested rest positional parameters if rendered with positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{component (component "-looked-up" model.greeting model.name) model.name model.greeting}}', {
        model: {
          greeting: 'Gabon ',
          name: 'Zack '
        }
      });

      this.assertText('Gabon Zack Zack Gabon ');

      this.runTask(() => this.rerender());

      this.assertText('Gabon Zack Zack Gabon ');

      this.runTask(() => this.context.set('model.greeting', 'Good morning '));

      this.assertText('Good morning Zack Zack Good morning ');

      this.runTask(() => this.context.set('model.name', 'Matthew '));

      this.assertText('Good morning Matthew Matthew Good morning ');

      this.runTask(() => this.context.set('model', { greeting: 'Gabon ', name: 'Zack ' }));

      this.assertText('Gabon Zack Zack Gabon ');
    }

    ['@test GH#13742  keeps nested rest positional parameters if nested and rendered with no positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{component (component (component "-looked-up" model.greeting model.name))}}', {
        model: {
          greeting: 'Gabon ',
          name: 'Zack'
        }
      });

      this.assertText('Gabon Zack');

      this.runTask(() => this.rerender());

      this.assertText('Gabon Zack');

      this.runTask(() => this.context.set('model.greeting', 'Good morning '));

      this.assertText('Good morning Zack');

      this.runTask(() => this.context.set('model.name', 'Matthew'));

      this.assertText('Good morning Matthew');

      this.runTask(() => this.context.set('model', { greeting: 'Gabon ', name: 'Zack' }));

      this.assertText('Gabon Zack');
    }

    ['@test overwrites nested rest positional parameters if nested with new pos params and rendered with no positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{component (component (component "-looked-up" model.greeting model.name) model.name model.greeting)}}', {
        model: {
          greeting: 'Gabon ',
          name: 'Zack '
        }
      });

      this.assertText('Gabon Zack Zack Gabon ');

      this.runTask(() => this.rerender());

      this.assertText('Gabon Zack Zack Gabon ');

      this.runTask(() => this.context.set('model.greeting', 'Good morning '));

      this.assertText('Good morning Zack Zack Good morning ');

      this.runTask(() => this.context.set('model.name', 'Matthew '));

      this.assertText('Good morning Matthew Matthew Good morning ');

      this.runTask(() => this.context.set('model', { greeting: 'Gabon ', name: 'Zack ' }));

      this.assertText('Gabon Zack Zack Gabon ');
    }

    ['@test renders with component helper with curried params, hash']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{greeting}} {{name}}'
      });

      this.render(_abstractTestCase.strip`
      {{component (component "-looked-up" "Hodari" greeting="Hodi")
                  greeting="Hola"}}`);

      this.assertText('Hola Hodari');

      this.runTask(() => this.rerender());

      this.assertText('Hola Hodari');
    }

    ['@test updates when component path is bound']() {
      this.registerComponent('-mandarin', {
        template: 'ni hao'
      });

      this.registerComponent('-hindi', {
        template: 'Namaste'
      });

      this.render('{{component (component model.lookupComponent)}}', {
        model: {
          lookupComponent: '-mandarin'
        }
      });

      this.assertText('ni hao');

      this.runTask(() => this.rerender());

      this.assertText('ni hao');

      this.runTask(() => this.context.set('model.lookupComponent', '-hindi'));

      this.assertText('Namaste');

      this.runTask(() => this.context.set('model', { lookupComponent: '-mandarin' }));

      this.assertText('ni hao');
    }

    ['@test updates when curried hash argument is bound']() {
      this.registerComponent('-looked-up', {
        template: '{{greeting}}'
      });

      this.render(`{{component (component "-looked-up" greeting=model.greeting)}}`, {
        model: {
          greeting: 'Hodi'
        }
      });

      this.assertText('Hodi');

      this.runTask(() => this.rerender());

      this.assertText('Hodi');

      this.runTask(() => this.context.set('model.greeting', 'Hola'));

      this.assertText('Hola');

      this.runTask(() => this.context.set('model', { greeting: 'Hodi' }));

      this.assertText('Hodi');
    }

    ['@test updates when curried hash arguments is bound in block form']() {
      this.registerComponent('-looked-up', {
        template: '{{greeting}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash comp=(component "-looked-up" greeting=model.greeting)) as |my|}}
        {{#my.comp}}{{/my.comp}}
      {{/with}}`, {
        model: {
          greeting: 'Hodi'
        }
      });

      this.assertText('Hodi');

      this.runTask(() => this.rerender());

      this.assertText('Hodi');

      this.runTask(() => this.context.set('model.greeting', 'Hola'));

      this.assertText('Hola');

      this.runTask(() => this.context.set('model', { greeting: 'Hodi' }));

      this.assertText('Hodi');
    }

    ['@test nested components do not overwrite positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        }),
        template: '{{name}} {{age}}'
      });

      this.render('{{component (component (component "-looked-up" "Sergio" 29) "Marvin" 21) "Hodari"}}');

      this.assertText('Sergio 29');

      this.runTask(() => this.rerender());

      this.assertText('Sergio 29');
    }

    ['@test positional parameters are combined not clobbered']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['greeting', 'name', 'age']
        }),
        template: '{{greeting}} {{name}} {{age}}'
      });

      this.render('{{component (component (component "-looked-up" "Hi") "Max") 9}}');

      this.assertText('Hi Max 9');

      this.runTask(() => this.rerender());

      this.assertText('Hi Max 9');
    }

    ['@test nested components positional parameters override named parameters [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        }),
        template: '{{name}} {{age}}'
      });

      expectDeprecation(() => {
        this.render('{{component (component (component "-looked-up" "Sergio" 29) name="Marvin" age=21)}}');
      }, 'You cannot specify both a positional param (at position 1) and the hash argument `age`.');

      this.assertText('Sergio 29');

      this.runTask(() => this.rerender());

      this.assertText('Sergio 29');
    }

    ['@test nested components with positional params at outer layer are override hash parameters [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['greeting', 'name', 'age']
        }),
        template: '{{greeting}} {{name}} {{age}}'
      });

      expectDeprecation(() => {
        this.render(_abstractTestCase.strip`
        {{#with (component "-looked-up" "Hola" "Dolores" 33) as |first|}}
          {{#with (component first greeting="Hej" name="Sigmundur") as |second|}}
            {{component second greeting=model.greeting}}
          {{/with}}
        {{/with}}`, {
          model: {
            greeting: 'Hodi'
          }
        });
      }, 'You cannot specify both a positional param (at position 1) and the hash argument `name`.');

      this.assertText('Hola Dolores 33');

      this.runTask(() => this.rerender());

      this.assertText('Hola Dolores 33');
    }

    ['@test nested components with positional params at middle layer partially override hash parameters [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['greeting', 'name', 'age']
        }),

        template: '{{greeting}} {{name}} {{age}}'
      });

      expectDeprecation(() => {
        this.render(_abstractTestCase.strip`
          {{#with (component "-looked-up" greeting="Hola" name="Dolores" age=33) as |first|}}
            {{#with (component first "Hej" "Sigmundur") as |second|}}
              {{component second greeting=model.greeting}}
            {{/with}}
          {{/with}}`, {
          model: {
            greeting: 'Hodi'
          }
        });
      }, 'You cannot specify both a positional param (at position 0) and the hash argument `greeting`.');

      this.assertText('Hej Sigmundur 33');

      this.runTask(() => this.rerender());

      this.assertText('Hej Sigmundur 33');
    }

    ['@test nested components with positional params at invocation override earlier hash parameters [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['greeting', 'name', 'age']
        }),

        template: '{{greeting}} {{name}} {{age}}'
      });

      expectDeprecation(() => {
        this.render(_abstractTestCase.strip`
          {{#with (component "-looked-up" greeting="Hola" name="Dolores" age=33) as |first|}}
            {{#with (component first greeting="Hej" name="Sigmundur") as |second|}}
              {{component second model.greeting}}
            {{/with}}
          {{/with}}`, {
          model: {
            greeting: 'Hodi'
          }
        });
      }, 'You cannot specify both a positional param (at position 0) and the hash argument `greeting`.');

      this.assertText('Hodi Sigmundur 33');

      this.runTask(() => this.rerender());

      this.assertText('Hodi Sigmundur 33');

      this.runTask(() => this.context.set('model.greeting', 'Kaixo'));

      this.assertText('Kaixo Sigmundur 33');

      this.runTask(() => this.context.set('model', { greeting: 'Hodi' }));

      this.assertText('Hodi Sigmundur 33');
    }

    ['@test nested components overwrite hash parameters']() {
      this.registerComponent('-looked-up', {
        template: '{{greeting}} {{name}} {{age}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (component "-looked-up" greeting="Hola" name="Dolores" age=33) as |first|}}
        {{#with (component first greeting="Hej" name="Sigmundur") as |second|}}
          {{component second greeting=model.greeting}}
        {{/with}}
      {{/with}}`, {
        model: {
          greeting: 'Hodi'
        }
      });

      this.assertText('Hodi Sigmundur 33');

      this.runTask(() => this.rerender());

      this.assertText('Hodi Sigmundur 33');

      this.runTask(() => this.context.set('model.greeting', 'Kaixo'));

      this.assertText('Kaixo Sigmundur 33');

      this.runTask(() => this.context.set('model', { greeting: 'Hodi' }));

      this.assertText('Hodi Sigmundur 33');
    }

    ['@test bound outer named parameters get updated in the right scope']() {
      this.registerComponent('-inner-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['comp']
        }),
        template: '{{component comp "Inner"}}'
      });

      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        }),
        template: '{{name}} {{age}}'
      });

      this.render('{{component "-inner-component" (component "-looked-up" model.outerName model.outerAge)}}', {
        model: {
          outerName: 'Outer',
          outerAge: 28
        }
      });

      this.assertText('Outer 28');

      this.runTask(() => this.rerender());

      this.assertText('Outer 28');

      this.runTask(() => this.context.set('model.outerAge', 29));

      this.assertText('Outer 29');

      this.runTask(() => this.context.set('model.outerName', 'Not outer'));

      this.assertText('Not outer 29');

      this.runTask(() => {
        this.context.set('model', {
          outerName: 'Outer',
          outerAge: 28
        });
      });

      this.assertText('Outer 28');
    }

    ['@test bound outer hash parameters get updated in the right scope']() {
      this.registerComponent('-inner-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['comp']
        }),
        template: '{{component comp name="Inner"}}'
      });

      this.registerComponent('-looked-up', {
        template: '{{name}} {{age}}'
      });

      this.render('{{component "-inner-component" (component "-looked-up" name=model.outerName age=model.outerAge)}}', {
        model: {
          outerName: 'Outer',
          outerAge: 28
        }
      });

      this.assertText('Inner 28');

      this.runTask(() => this.rerender());

      this.assertText('Inner 28');

      this.runTask(() => this.context.set('model.outerAge', 29));

      this.assertText('Inner 29');

      this.runTask(() => this.context.set('model.outerName', 'Not outer'));

      this.assertText('Inner 29');

      this.runTask(() => {
        this.context.set('model', {
          outerName: 'Outer',
          outerAge: 28
        });
      });

      this.assertText('Inner 28');
    }

    ['@test conflicting positional and hash parameters trigger a deprecation if in the same component context [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{greeting}} {{name}}'
      });

      expectDeprecation(() => {
        this.render('{{component (component "-looked-up" "Hodari" name="Sergio") "Hodari" greeting="Hodi"}}');
      }, 'You cannot specify both a positional param (at position 0) and the hash argument `name`.');
    }

    ['@test conflicting positional and hash parameters does not raise an assertion if rerendered']() {
      // In some cases, rerendering with a positional param used to cause an
      // assertion. This test checks it does not.
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{greeting}} {{name}}'
      });

      this.render('{{component (component "-looked-up" model.name greeting="Hodi")}}', {
        model: {
          name: 'Hodari'
        }
      });

      this.assertText('Hodi Hodari');

      this.runTask(() => this.rerender());

      this.assertText('Hodi Hodari');

      this.runTask(() => this.context.set('model.name', 'Sergio'));

      this.assertText('Hodi Sergio');

      this.runTask(() => this.context.set('model', { name: 'Hodari' }));

      this.assertText('Hodi Hodari');
    }

    ['@test conflicting positional and hash parameters trigger a deprecation [DEPRECATED]']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{greeting}} {{name}}'
      });

      expectDeprecation(() => {
        this.render('{{component (component "-looked-up" "Hodari") name="Sergio" greeting="Hodi"}}');
      }, 'You cannot specify both a positional param (at position 0) and the hash argument `name`.');

      this.assertText('Hodi Hodari');

      this.runTask(() => this.rerender());

      this.assertText('Hodi Hodari');
    }

    ['@test component with dynamic component name resolving to undefined, then an existing component']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component (component componentName name=name)}}', {
        componentName: undefined,
        name: 'Alex'
      });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => this.context.set('componentName', 'foo-bar'));

      this.assertText('hello Alex');

      this.runTask(() => this.context.set('componentName', undefined));

      this.assertText('');
    }

    ['@test component with dynamic component name resolving to a component, then undefined']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component (component componentName name=name)}}', {
        componentName: 'foo-bar',
        name: 'Alex'
      });

      this.assertText('hello Alex');

      this.runTask(() => this.rerender());

      this.assertText('hello Alex');

      this.runTask(() => this.context.set('componentName', undefined));

      this.assertText('');

      this.runTask(() => this.context.set('componentName', 'foo-bar'));

      this.assertText('hello Alex');
    }

    ['@test component with dynamic component name resolving to null, then an existing component']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component (component componentName name=name)}}', {
        componentName: null,
        name: 'Alex'
      });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => this.context.set('componentName', 'foo-bar'));

      this.assertText('hello Alex');

      this.runTask(() => this.context.set('componentName', null));

      this.assertText('');
    }

    ['@test component with dynamic component name resolving to a component, then null']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component (component componentName name=name)}}', {
        componentName: 'foo-bar',
        name: 'Alex'
      });

      this.assertText('hello Alex');

      this.runTask(() => this.rerender());

      this.assertText('hello Alex');

      this.runTask(() => this.context.set('componentName', null));

      this.assertText('');

      this.runTask(() => this.context.set('componentName', 'foo-bar'));

      this.assertText('hello Alex');
    }

    ['@test raises an assertion when component path is not a component name (static)']() {
      expectAssertion(() => {
        this.render('{{component (component "not-a-component")}}');
      }, 'Could not find component named "not-a-component" (no component or template with that name was found)');
    }

    ['@test raises an assertion when component path is not a component name (dynamic)']() {
      expectAssertion(() => {
        this.render('{{component (component compName)}}', {
          compName: 'not-a-component'
        });
      }, 'Could not find component named "not-a-component" (no component or template with that name was found)');
    }

    ['@test renders with dot path']() {
      let expectedText = 'Hodi';
      this.registerComponent('-looked-up', {
        template: expectedText
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash lookedup=(component "-looked-up")) as |object|}}
        {{object.lookedup}}
      {{/with}}`);

      this.assertText(expectedText);

      this.runTask(() => this.rerender());

      this.assertText(expectedText);
    }

    ['@test renders with dot path and attr']() {
      let expectedText = 'Hodi';
      this.registerComponent('-looked-up', {
        template: '{{expectedText}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash lookedup=(component "-looked-up")) as |object|}}
        {{object.lookedup expectedText=model.expectedText}}
      {{/with}}`, {
        model: {
          expectedText
        }
      });

      this.assertText(expectedText);

      this.runTask(() => this.rerender());

      this.assertText(expectedText);

      this.runTask(() => this.context.set('model.expectedText', 'Hola'));

      this.assertText('Hola');

      this.runTask(() => this.context.set('model', { expectedText }));

      this.assertText(expectedText);
    }

    ['@test renders with dot path and curried over attr']() {
      let expectedText = 'Hodi';
      this.registerComponent('-looked-up', {
        template: '{{expectedText}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash lookedup=(component "-looked-up" expectedText=model.expectedText)) as |object|}}
        {{object.lookedup}}
      {{/with}}`, {
        model: {
          expectedText
        }
      });

      this.assertText(expectedText);

      this.runTask(() => this.rerender());

      this.assertText(expectedText);

      this.runTask(() => this.context.set('model.expectedText', 'Hola'));

      this.assertText('Hola');

      this.runTask(() => this.context.set('model', { expectedText }));

      this.assertText(expectedText);
    }

    ['@test renders with dot path and with rest positional parameters']() {
      this.registerComponent('-looked-up', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{params}}'
      });

      let expectedText = 'Hodi';

      this.render(_abstractTestCase.strip`
      {{#with (hash lookedup=(component "-looked-up")) as |object|}}
        {{object.lookedup model.expectedText "Hola"}}
      {{/with}}`, {
        model: {
          expectedText
        }
      });

      this.assertText(`${expectedText},Hola`);

      this.runTask(() => this.rerender());

      this.assertText(`${expectedText},Hola`);

      this.runTask(() => this.context.set('model.expectedText', 'Kaixo'));

      this.assertText('Kaixo,Hola');

      this.runTask(() => this.context.set('model', { expectedText }));

      this.assertText(`${expectedText},Hola`);
    }

    ['@test renders with dot path and rest parameter does not leak'](assert) {
      // In the original implementation, positional parameters were not handled
      // correctly causing the first positional parameter to be the contextual
      // component itself.
      let value = false;

      this.registerComponent('my-component', {
        ComponentClass: _helpers.Component.extend({
          didReceiveAttrs() {
            value = this.getAttr('value');
          }
        }).reopenClass({
          positionalParams: ['value']
        })
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash my-component=(component 'my-component' first)) as |c|}}
        {{c.my-component}}
      {{/with}}`, { first: 'first' });

      assert.equal(value, 'first', 'value is the expected parameter');
    }

    ['@test renders with dot path and updates attributes'](assert) {
      this.registerComponent('my-nested-component', {
        ComponentClass: _helpers.Component.extend({
          didReceiveAttrs() {
            this.set('myProp', this.getAttr('my-parent-attr'));
          }
        }),
        template: '<span id="nested-prop">{{myProp}}</span>'
      });

      this.registerComponent('my-component', {
        template: '{{yield (hash my-nested-component=(component "my-nested-component" my-parent-attr=my-attr))}}'
      });

      this.registerComponent('my-action-component', {
        ComponentClass: _helpers.Component.extend({
          actions: {
            changeValue() {
              this.incrementProperty('myProp');
            }
          }
        }),
        template: _abstractTestCase.strip`
        {{#my-component my-attr=myProp as |api|}}
          {{api.my-nested-component}}
        {{/my-component}}
        <br>
        <button onclick={{action 'changeValue'}}>Change value</button>`
      });

      this.render('{{my-action-component myProp=model.myProp}}', {
        model: {
          myProp: 1
        }
      });

      assert.equal(this.$('#nested-prop').text(), '1');

      this.runTask(() => this.rerender());

      assert.equal(this.$('#nested-prop').text(), '1');

      this.runTask(() => this.$('button').click());

      assert.equal(this.$('#nested-prop').text(), '2');

      this.runTask(() => this.$('button').click());

      assert.equal(this.$('#nested-prop').text(), '3');

      this.runTask(() => this.context.set('model', { myProp: 1 }));

      assert.equal(this.$('#nested-prop').text(), '1');
    }

    ["@test adding parameters to a contextual component's instance does not add it to other instances"]() {
      // If parameters and attributes are not handled correctly, setting a value
      // in an invokation can leak to others invocation.
      this.registerComponent('select-box', {
        template: '{{yield (hash option=(component "select-box-option"))}}'
      });

      this.registerComponent('select-box-option', {
        template: '{{label}}'
      });

      this.render(_abstractTestCase.strip`
      {{#select-box as |sb|}}
        {{sb.option label="Foo"}}
        {{sb.option}}
      {{/select-box}}`);

      this.assertText('Foo');

      this.runTask(() => this.rerender());

      this.assertText('Foo');
    }

    ['@test parameters in a contextual component are mutable when value is a param'](assert) {
      // This checks that a `(mut)` is added to parameters and attributes to
      // contextual components when it is a param.

      this.registerComponent('change-button', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['val']
        }),
        template: _abstractTestCase.strip`
        <button {{action (action (mut val) 10)}} class="my-button">
          Change to 10
        </button>`
      });

      this.render(_abstractTestCase.strip`
      {{component (component "change-button" model.val2)}}
      <span class="value">{{model.val2}}</span>`, {
        model: {
          val2: 8
        }
      });

      assert.equal(this.$('.value').text(), '8');

      this.runTask(() => this.rerender());

      assert.equal(this.$('.value').text(), '8');

      this.runTask(() => this.$('.my-button').click());

      assert.equal(this.$('.value').text(), '10');

      this.runTask(() => this.context.set('model', { val2: 8 }));

      assert.equal(this.$('.value').text(), '8');
    }

    ['@test tagless blockless components render'](assert) {
      this.registerComponent('my-comp', {
        ComponentClass: _helpers.Component.extend({ tagName: '' })
      });

      this.render(`{{my-comp}}`);

      this.runTask(() => this.rerender());

      assert.equal(this.$().text(), '');
    }

    ['@test GH#13494 tagless blockless component with property binding'](assert) {
      this.registerComponent('outer-component', {
        ComponentClass: _helpers.Component.extend({
          message: 'hello',
          actions: {
            change() {
              this.set('message', 'goodbye');
            }
          }
        }),
        template: _abstractTestCase.strip`
        message: {{message}}{{inner-component message=message}}
        <button onclick={{action "change"}} />`
      });

      this.registerComponent('inner-component', {
        ComponentClass: _helpers.Component.extend({
          tagName: ''
        })
      });

      this.render(`{{outer-component}}`);

      assert.equal(this.$().text(), 'message: hello');

      this.runTask(() => this.rerender());

      assert.equal(this.$().text(), 'message: hello');

      this.runTask(() => this.$('button').click());

      assert.equal(this.$().text(), 'message: goodbye');

      this.runTask(() => this.rerender());

      assert.equal(this.$().text(), 'message: goodbye');
    }

    ['@test GH#13982 contextual component ref is stable even when bound params change'](assert) {
      let instance, previousInstance;
      let initCount = 0;

      this.registerComponent('my-comp', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            previousInstance = instance;
            instance = this;
            initCount++;
          },
          isOpen: undefined
        }),
        template: '{{if isOpen "open" "closed"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash ctxCmp=(component "my-comp" isOpen=isOpen)) as |thing|}}
        {{#thing.ctxCmp}}This is a contextual component{{/thing.ctxCmp}}
      {{/with}}
    `, {
        isOpen: true
      });

      assert.ok(!(0, _metal.isEmpty)(instance), 'a instance was created');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');

      this.runTask(() => this.context.set('isOpen', false));

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'closed', 'the component text is "closed"');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'closed', 'the component text is "closed"');

      this.runTask(() => this.context.set('isOpen', true));

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');
    }

    ['@test GH#13982 contextual component ref is stable even when bound params change (bound name param)'](assert) {
      let instance, previousInstance;
      let initCount = 0;

      this.registerComponent('my-comp', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            previousInstance = instance;
            instance = this;
            initCount++;
          },
          isOpen: undefined
        }),
        template: '{{if isOpen "open" "closed"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash ctxCmp=(component compName isOpen=isOpen)) as |thing|}}
        {{#thing.ctxCmp}}This is a contextual component{{/thing.ctxCmp}}
      {{/with}}
    `, {
        compName: 'my-comp',
        isOpen: true
      });

      assert.ok(!(0, _metal.isEmpty)(instance), 'a instance was created');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');

      this.runTask(() => this.context.set('isOpen', false));

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'closed', 'the component text is "closed"');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'closed', 'the component text is "closed"');

      this.runTask(() => this.context.set('isOpen', true));

      assert.ok(!(0, _metal.isEmpty)(instance), 'the component instance exists');
      assert.equal(previousInstance, undefined, 'no previous component exists');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'open', 'the components text is "open"');
    }

    ['@test GH#13982 contextual component ref is recomputed when component name param changes'](assert) {
      let instance, previousInstance;
      let initCount = 0;

      this.registerComponent('my-comp', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            previousInstance = instance;
            instance = this;
            initCount++;
          },
          isOpen: undefined
        }),
        template: 'my-comp: {{if isOpen "open" "closed"}}'
      });

      this.registerComponent('your-comp', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            previousInstance = instance;
            instance = this;
            initCount++;
          },
          isOpen: undefined
        }),
        template: 'your-comp: {{if isOpen "open" "closed"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with (hash ctxCmp=(component compName isOpen=isOpen)) as |thing|}}
        {{#thing.ctxCmp}}This is a contextual component{{/thing.ctxCmp}}
      {{/with}}
    `, {
        compName: 'my-comp',
        isOpen: true
      });

      assert.ok(!(0, _metal.isEmpty)(instance), 'a instance was created');
      assert.equal(previousInstance, undefined, 'there is no previous instance');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'my-comp: open');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'a instance exists after rerender');
      assert.equal(previousInstance, undefined, 'there is no previous instance after rerender');
      assert.equal(initCount, 1, 'the component was constructed exactly 1 time');
      assert.equal(this.$().text(), 'my-comp: open');

      this.runTask(() => this.context.set('compName', 'your-comp'));

      assert.ok(!(0, _metal.isEmpty)(instance), 'an instance was created after component name changed');
      assert.ok(!(0, _metal.isEmpty)(previousInstance), 'a previous instance now exists');
      assert.notEqual(instance, previousInstance, 'the instance and previous instance are not the same object');
      assert.equal(initCount, 2, 'the component was constructed exactly 2 times');
      assert.equal(this.$().text(), 'your-comp: open');

      this.runTask(() => this.rerender());

      assert.ok(!(0, _metal.isEmpty)(instance), 'an instance was created after component name changed (rerender)');
      assert.ok(!(0, _metal.isEmpty)(previousInstance), 'a previous instance now exists (rerender)');
      assert.notEqual(instance, previousInstance, 'the instance and previous instance are not the same object (rerender)');
      assert.equal(initCount, 2, 'the component was constructed exactly 2 times (rerender)');
      assert.equal(this.$().text(), 'your-comp: open');

      this.runTask(() => this.context.set('compName', 'my-comp'));

      assert.ok(!(0, _metal.isEmpty)(instance), 'an instance was created after component name changed');
      assert.ok(!(0, _metal.isEmpty)(previousInstance), 'a previous instance still exists');
      assert.notEqual(instance, previousInstance, 'the instance and previous instance are not the same object');
      assert.equal(initCount, 3, 'the component was constructed exactly 3 times (rerender)');
      assert.equal(this.$().text(), 'my-comp: open');
    }

    ['@test GH#14508 rest positional params are received when passed as named parameter']() {
      this.registerComponent('my-link', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{component (component "my-link") params=allParams}}', {
        allParams: (0, _runtime.A)(['a', 'b'])
      });

      this.assertText('ab');

      this.runTask(() => this.rerender());

      this.assertText('ab');

      this.runTask(() => this.context.get('allParams').pushObject('c'));

      this.assertText('abc');

      this.runTask(() => this.context.get('allParams').popObject());

      this.assertText('ab');

      this.runTask(() => this.context.get('allParams').clear());

      this.assertText('');

      this.runTask(() => this.context.set('allParams', (0, _runtime.A)(['1', '2'])));

      this.assertText('12');

      this.runTask(() => this.context.set('allParams', (0, _runtime.A)(['a', 'b'])));

      this.assertText('ab');
    }

    ['@test GH#14508 rest positional params are received when passed as named parameter with dot notation']() {
      this.registerComponent('my-link', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: '{{#each params as |p|}}{{p}}{{/each}}'
      });

      this.render('{{#with (hash link=(component "my-link")) as |c|}}{{c.link params=allParams}}{{/with}}', {
        allParams: (0, _runtime.A)(['a', 'b'])
      });

      this.assertText('ab');

      this.runTask(() => this.rerender());

      this.assertText('ab');

      this.runTask(() => this.context.get('allParams').pushObject('c'));

      this.assertText('abc');

      this.runTask(() => this.context.get('allParams').popObject());

      this.assertText('ab');

      this.runTask(() => this.context.get('allParams').clear());

      this.assertText('');

      this.runTask(() => this.context.set('allParams', (0, _runtime.A)(['1', '2'])));

      this.assertText('12');

      this.runTask(() => this.context.set('allParams', (0, _runtime.A)(['a', 'b'])));

      this.assertText('ab');
    }

    ['@test GH#14632 give useful warning when calling contextual components with input as a name']() {
      expectAssertion(() => {
        this.render('{{component (component "input" type="text")}}');
      }, 'You cannot use `input` as a component name.');
    }

    ['@test GH#14632 give useful warning when calling contextual components with textarea as a name']() {
      expectAssertion(() => {
        this.render('{{component (component "textarea" type="text")}}');
      }, 'You cannot use `textarea` as a component name.');
    }

    ['@test GH#17121 local variable should win over helper (without arguments)']() {
      this.registerHelper('foo', () => 'foo helper');

      this.registerComponent('foo-bar', { template: 'foo-bar component' });

      this.render(_abstractTestCase.strip`
        {{#let (component 'foo-bar') as |foo|}}
          {{foo}}
        {{/let}}
      `);

      this.assertText('foo-bar component');

      this.assertStableRerender();
    }

    ['@test GH#17121 local variable should win over helper (with arguments)']() {
      this.registerHelper('foo', params => `foo helper: ${params.join(' ')}`);

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: 'foo-bar component:{{#each params as |param|}} {{param}}{{/each}}'
      });

      this.render(_abstractTestCase.strip`
        {{#let (component 'foo-bar') as |foo|}}
          {{foo 1 2 3}}
        {{/let}}
      `);

      this.assertText('foo-bar component: 1 2 3');

      this.assertStableRerender();
    }

    ['@test GH#17121 implicit component invocations should not perform string lookup'](assert) {
      this.registerComponent('foo-bar', { template: 'foo-bar component' });

      assert.throws(() => this.render(_abstractTestCase.strip`
          {{#let 'foo-bar' as |foo|}}
            {{foo 1 2 3}}
          {{/let}}
        `), new TypeError("expected `foo` to be a contextual component but found a string. Did you mean `(component foo)`? ('-top-level' @ L1:C29) "));
    }

    ['@test RFC#311 invoking named args (without arguments)']() {
      this.registerComponent('x-outer', { template: '{{@inner}}' });
      this.registerComponent('x-inner', { template: 'inner' });

      this.render('{{x-outer inner=(component "x-inner")}}');

      this.assertText('inner');

      this.assertStableRerender();
    }

    ['@test RFC#311 invoking named args (with arguments)']() {
      this.registerComponent('x-outer', { template: '{{@inner 1 2 3}}' });

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'params'
        }),
        template: 'inner:{{#each params as |param|}} {{param}}{{/each}}'
      });

      this.render('{{x-outer inner=(component "x-inner")}}');

      this.assertText('inner: 1 2 3');

      this.assertStableRerender();
    }

    ['@test RFC#311 invoking named args (with a block)']() {
      this.registerComponent('x-outer', { template: '{{#@inner}}outer{{/@inner}}' });
      this.registerComponent('x-inner', { template: 'inner {{yield}}' });

      this.render('{{x-outer inner=(component "x-inner")}}');

      this.assertText('inner outer');

      this.assertStableRerender();
    }
  });

  class ContextualComponentMutableParamsTest extends _testCase.RenderingTest {
    render(templateStr, context = {}) {
      super.render(`${templateStr}<span class="value">{{model.val2}}</span>`, (0, _polyfills.assign)(context, { model: { val2: 8 } }));
    }
  }

  class MutableParamTestGenerator {
    constructor(cases) {
      this.cases = cases;
    }

    generate({ title, setup }) {
      return {
        [`@test parameters in a contextual component are mutable when value is a ${title}`](assert) {
          this.registerComponent('change-button', {
            ComponentClass: _helpers.Component.extend().reopenClass({
              positionalParams: ['val']
            }),
            template: _abstractTestCase.strip`
          <button {{action (action (mut val) 10)}} class="my-button">
            Change to 10
          </button>`
          });

          setup.call(this, assert);

          assert.equal(this.$('.value').text(), '8');

          this.runTask(() => this.rerender());

          assert.equal(this.$('.value').text(), '8');

          this.runTask(() => this.$('.my-button').click());

          assert.equal(this.$('.value').text(), '10');

          this.runTask(() => this.context.set('model', { val2: 8 }));

          assert.equal(this.$('.value').text(), '8');
        }
      };
    }
  }

  (0, _abstractTestCase.applyMixins)(ContextualComponentMutableParamsTest, new MutableParamTestGenerator([{
    title: 'param',
    setup() {
      this.render('{{component (component "change-button" model.val2)}}');
    }
  }, {
    title: 'nested param',
    setup() {
      this.registerComponent('my-comp', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['components']
        }),
        template: '{{component components.comp}}'
      });

      this.render('{{my-comp (hash comp=(component "change-button" model.val2))}}');
    }
  }, {
    title: 'hash value',
    setup() {
      this.registerComponent('my-comp', {
        template: '{{component component}}'
      });

      this.render('{{my-comp component=(component "change-button" val=model.val2)}}');
    }
  }, {
    title: 'nested hash value',
    setup() {
      this.registerComponent('my-comp', {
        template: '{{component components.button}}'
      });

      this.render('{{my-comp components=(hash button=(component "change-button" val=model.val2))}}');
    }
  }]));

  (0, _testCase.moduleFor)('Components test: contextual components -- mutable params', ContextualComponentMutableParamsTest);
});
enifed('@ember/-internals/glimmer/tests/integration/components/curly-components-test', ['@ember/runloop', '@ember/-internals/metal', '@ember/service', '@ember/-internals/runtime', '@ember/-internals/views', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/test-helpers'], function (_runloop, _metal, _service, _runtime, _views, _helpers, _abstractTestCase, _testCase, _testHelpers) {
  'use strict';

  /* globals EmberDev */
  (0, _testCase.moduleFor)('Components test: curly components', class extends _testCase.RenderingTest {
    ['@test it can render a basic component']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test it can have a custom id and it is not bound']() {
      this.registerComponent('foo-bar', { template: '{{id}} {{elementId}}' });

      this.render('{{foo-bar id=customId}}', {
        customId: 'bizz'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'bizz' },
        content: 'bizz bizz'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'bizz' },
        content: 'bizz bizz'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'customId', 'bar'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'bizz' },
        content: 'bar bizz'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'customId', 'bizz'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'bizz' },
        content: 'bizz bizz'
      });
    }

    ['@test elementId cannot change'](assert) {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        elementId: 'blahzorz',
        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{elementId}}'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'blahzorz' },
        content: 'blahzorz'
      });

      if (EmberDev && !EmberDev.runningProdBuild) {
        let willThrow = () => (0, _runloop.run)(null, _metal.set, component, 'elementId', 'herpyderpy');

        assert.throws(willThrow, /Changing a view's elementId after creation is not allowed/);

        this.assertComponentElement(this.firstChild, {
          tagName: 'div',
          attrs: { id: 'blahzorz' },
          content: 'blahzorz'
        });
      }
    }

    ['@test can specify template with `layoutName` property']() {
      let FooBarComponent = _helpers.Component.extend({
        elementId: 'blahzorz',
        layoutName: 'fizz-bar',
        init() {
          this._super(...arguments);
          this.local = 'hey';
        }
      });

      this.registerTemplate('fizz-bar', `FIZZ BAR {{local}}`);

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar}}');

      this.assertText('FIZZ BAR hey');
    }

    ['@test layout supports computed property']() {
      let FooBarComponent = _helpers.Component.extend({
        elementId: 'blahzorz',
        layout: (0, _metal.computed)(function () {
          return (0, _helpers.compile)('so much layout wat {{lulz}}');
        }),
        init() {
          this._super(...arguments);
          this.lulz = 'heyo';
        }
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar}}');

      this.assertText('so much layout wat heyo');
    }

    ['@test passing undefined elementId results in a default elementId'](assert) {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'h1'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'something'
      });

      this.render('{{foo-bar id=somethingUndefined}}');

      let foundId = this.$('h1').attr('id');
      assert.ok(/^ember/.test(foundId), 'Has a reasonable id attribute (found id=' + foundId + ').');

      this.runTask(() => this.rerender());

      let newFoundId = this.$('h1').attr('id');
      assert.ok(/^ember/.test(newFoundId), 'Has a reasonable id attribute (found id=' + newFoundId + ').');

      assert.equal(foundId, newFoundId);
    }

    ['@test id is an alias for elementId'](assert) {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'h1'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'something'
      });

      this.render('{{foo-bar id="custom-id"}}');

      let foundId = this.$('h1').attr('id');
      assert.equal(foundId, 'custom-id');

      this.runTask(() => this.rerender());

      let newFoundId = this.$('h1').attr('id');
      assert.equal(newFoundId, 'custom-id');

      assert.equal(foundId, newFoundId);
    }

    ['@test cannot pass both id and elementId at the same time']() {
      this.registerComponent('foo-bar', { template: '' });

      expectAssertion(() => {
        this.render('{{foo-bar id="zomg" elementId="lol"}}');
      }, /You cannot invoke a component with both 'id' and 'elementId' at the same time./);
    }

    ['@test it can have a custom tagName']() {
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'foo-bar'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });
    }

    ['@test it can have a custom tagName set in the constructor']() {
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          this.tagName = 'foo-bar';
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });
    }

    ['@test it can have a custom tagName from the invocation']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar tagName="foo-bar"}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'foo-bar',
        content: 'hello'
      });
    }

    ['@test tagName can not be a computed property']() {
      let FooBarComponent = _helpers.Component.extend({
        tagName: (0, _metal.computed)(function () {
          return 'foo-bar';
        })
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /You cannot use a computed property for the component's `tagName` \(<.+?>\)\./);
    }

    ['@test class is applied before didInsertElement'](assert) {
      let componentClass;
      let FooBarComponent = _helpers.Component.extend({
        didInsertElement() {
          componentClass = this.element.className;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar class="foo-bar"}}');

      assert.equal(componentClass, 'foo-bar ember-view');
    }

    ['@test it can have custom classNames']() {
      let FooBarComponent = _helpers.Component.extend({
        classNames: ['foo', 'bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar') },
        content: 'hello'
      });
    }

    ['@test should not apply falsy class name']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar class=somethingFalsy}}', {
        somethingFalsy: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'ember-view' },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'ember-view' },
        content: 'hello'
      });
    }

    ['@test should update class using inline if, initially false, no alternate']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar class=(if predicate "thing") }}', {
        predicate: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'ember-view' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'predicate', true));
      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view thing') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'predicate', false));
      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'ember-view' },
        content: 'hello'
      });
    }

    ['@test should update class using inline if, initially true, no alternate']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar class=(if predicate "thing") }}', {
        predicate: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view thing') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'predicate', false));
      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'ember-view' },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'predicate', true));
      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view thing') },
        content: 'hello'
      });
    }

    ['@test should apply classes of the dasherized property name when bound property specified is true']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar class=model.someTruth}}', {
        model: { someTruth: true }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view some-truth') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view some-truth') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.someTruth', false));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') },
        content: 'hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { someTruth: true }));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view some-truth') },
        content: 'hello'
      });
    }

    ['@test class property on components can be dynamic']() {
      this.registerComponent('foo-bar', { template: 'hello' });

      this.render('{{foo-bar class=(if fooBar "foo-bar")}}', {
        fooBar: true
      });

      this.assertComponentElement(this.firstChild, {
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', false));

      this.assertComponentElement(this.firstChild, {
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', true));

      this.assertComponentElement(this.firstChild, {
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo-bar') }
      });
    }

    ['@test it can have custom classNames from constructor']() {
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          this.classNames = this.classNames.slice();
          this.classNames.push('foo', 'bar', `outside-${this.get('extraClass')}`);
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar extraClass="baz"}}');

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar outside-baz') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar outside-baz') },
        content: 'hello'
      });
    }

    ['@test it can set custom classNames from the invocation']() {
      let FooBarComponent = _helpers.Component.extend({
        classNames: ['foo']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render(_abstractTestCase.strip`
      {{foo-bar class="bar baz"}}
      {{foo-bar classNames="bar baz"}}
      {{foo-bar}}
    `);

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
        content: 'hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.nthChild(0), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(1), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo bar baz') },
        content: 'hello'
      });
      this.assertComponentElement(this.nthChild(2), {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view foo') },
        content: 'hello'
      });
    }

    ['@test it has an element']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          instance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{foo-bar}}');

      let element1 = instance.element;

      this.assertComponentElement(element1, { content: 'hello' });

      this.runTask(() => this.rerender());

      let element2 = instance.element;

      this.assertComponentElement(element2, { content: 'hello' });

      this.assertSameNode(element2, element1);
    }

    ['@test an empty component does not have childNodes'](assert) {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        tagName: 'input',
        init() {
          this._super();
          fooBarInstance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: ''
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, { tagName: 'input' });

      assert.strictEqual(fooBarInstance.element.childNodes.length, 0);

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { tagName: 'input' });

      assert.strictEqual(fooBarInstance.element.childNodes.length, 0);
    }

    ['@test it has the right parentView and childViews'](assert) {
      let fooBarInstance, fooBarBazInstance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
        }
      });

      let FooBarBazComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarBazInstance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'foo-bar {{foo-bar-baz}}'
      });
      this.registerComponent('foo-bar-baz', {
        ComponentClass: FooBarBazComponent,
        template: 'foo-bar-baz'
      });

      this.render('{{foo-bar}}');
      this.assertText('foo-bar foo-bar-baz');

      assert.equal(fooBarInstance.parentView, this.component);
      assert.equal(fooBarBazInstance.parentView, fooBarInstance);

      assert.deepEqual(this.component.childViews, [fooBarInstance]);
      assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);

      this.runTask(() => this.rerender());
      this.assertText('foo-bar foo-bar-baz');

      assert.equal(fooBarInstance.parentView, this.component);
      assert.equal(fooBarBazInstance.parentView, fooBarInstance);

      assert.deepEqual(this.component.childViews, [fooBarInstance]);
      assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);
    }

    ['@feature(ember-glimmer-named-arguments) it renders passed named arguments']() {
      this.registerComponent('foo-bar', {
        template: '{{@foo}}'
      });

      this.render('{{foo-bar foo=model.bar}}', {
        model: {
          bar: 'Hola'
        }
      });

      this.assertText('Hola');

      this.runTask(() => this.rerender());

      this.assertText('Hola');

      this.runTask(() => this.context.set('model.bar', 'Hello'));

      this.assertText('Hello');

      this.runTask(() => this.context.set('model', { bar: 'Hola' }));

      this.assertText('Hola');
    }

    ['@test it reflects named arguments as properties']() {
      this.registerComponent('foo-bar', {
        template: '{{foo}}'
      });

      this.render('{{foo-bar foo=model.bar}}', {
        model: {
          bar: 'Hola'
        }
      });

      this.assertText('Hola');

      this.runTask(() => this.rerender());

      this.assertText('Hola');

      this.runTask(() => this.context.set('model.bar', 'Hello'));

      this.assertText('Hello');

      this.runTask(() => this.context.set('model', { bar: 'Hola' }));

      this.assertText('Hola');
    }

    ['@test it can render a basic component with a block']() {
      this.registerComponent('foo-bar', {
        template: '{{yield}} - In component'
      });

      this.render('{{#foo-bar}}hello{{/foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        content: 'hello - In component'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'hello - In component'
      });
    }

    ['@test it can render a basic component with a block when the yield is in a partial']() {
      this.registerPartial('_partialWithYield', 'yielded: [{{yield}}]');

      this.registerComponent('foo-bar', {
        template: '{{partial "partialWithYield"}} - In component'
      });

      this.render('{{#foo-bar}}hello{{/foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        content: 'yielded: [hello] - In component'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'yielded: [hello] - In component'
      });
    }

    ['@test it can render a basic component with a block param when the yield is in a partial']() {
      this.registerPartial('_partialWithYield', 'yielded: [{{yield "hello"}}]');

      this.registerComponent('foo-bar', {
        template: '{{partial "partialWithYield"}} - In component'
      });

      this.render('{{#foo-bar as |value|}}{{value}}{{/foo-bar}}');

      this.assertComponentElement(this.firstChild, {
        content: 'yielded: [hello] - In component'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'yielded: [hello] - In component'
      });
    }

    ['@test it renders the layout with the component instance as the context']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          instance = this;
          this.set('message', 'hello');
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{message}}'
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(instance, 'message', 'goodbye'));

      this.assertComponentElement(this.firstChild, { content: 'goodbye' });

      this.runTask(() => (0, _metal.set)(instance, 'message', 'hello'));

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test it preserves the outer context when yielding']() {
      this.registerComponent('foo-bar', { template: '{{yield}}' });

      this.render('{{#foo-bar}}{{message}}{{/foo-bar}}', { message: 'hello' });

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'goodbye'));

      this.assertComponentElement(this.firstChild, { content: 'goodbye' });

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'hello'));

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test it can yield a block param named for reserved words [GH#14096]']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },

        name: 'foo-bar'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{yield this}}'
      });

      this.render('{{#foo-bar as |component|}}{{component.name}}{{/foo-bar}}');

      this.assertComponentElement(this.firstChild, { content: 'foo-bar' });

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(instance, 'name', 'derp-qux'));

      this.assertComponentElement(this.firstChild, { content: 'derp-qux' });

      this.runTask(() => (0, _metal.set)(instance, 'name', 'foo-bar'));

      this.assertComponentElement(this.firstChild, { content: 'foo-bar' });
    }

    ['@test it can yield internal and external properties positionally']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },
        greeting: 'hello'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{yield greeting greetee.firstName}}'
      });

      this.render('{{#foo-bar greetee=person as |greeting name|}}{{name}} {{person.lastName}}, {{greeting}}{{/foo-bar}}', {
        person: {
          firstName: 'Joel',
          lastName: 'Kang'
        }
      });

      this.assertComponentElement(this.firstChild, {
        content: 'Joel Kang, hello'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'Joel Kang, hello'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'person', {
        firstName: 'Dora',
        lastName: 'the Explorer'
      }));

      this.assertComponentElement(this.firstChild, {
        content: 'Dora the Explorer, hello'
      });

      this.runTask(() => (0, _metal.set)(instance, 'greeting', 'hola'));

      this.assertComponentElement(this.firstChild, {
        content: 'Dora the Explorer, hola'
      });

      this.runTask(() => {
        (0, _metal.set)(instance, 'greeting', 'hello');
        (0, _metal.set)(this.context, 'person', {
          firstName: 'Joel',
          lastName: 'Kang'
        });
      });

      this.assertComponentElement(this.firstChild, {
        content: 'Joel Kang, hello'
      });
    }

    ['@test #11519 - block param infinite loop']() {
      let instance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          instance = this;
        },
        danger: 0
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{danger}}{{yield danger}}'
      });

      // On initial render, create streams. The bug will not have manifested yet, but at this point
      // we have created streams that create a circular invalidation.
      this.render(`{{#foo-bar as |dangerBlockParam|}}{{/foo-bar}}`);

      this.assertText('0');

      // Trigger a non-revalidating re-render. The yielded block will not be dirtied
      // nor will block param streams, and thus no infinite loop will occur.
      this.runTask(() => this.rerender());

      this.assertText('0');

      // Trigger a revalidation, which will cause an infinite loop without the fix
      // in place.  Note that we do not see the infinite loop is in testing mode,
      // because a deprecation warning about re-renders is issued, which Ember
      // treats as an exception.
      this.runTask(() => (0, _metal.set)(instance, 'danger', 1));

      this.assertText('1');

      this.runTask(() => (0, _metal.set)(instance, 'danger', 0));

      this.assertText('0');
    }

    ['@test the component and its child components are destroyed'](assert) {
      let destroyed = { 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0 };

      this.registerComponent('foo-bar', {
        template: '{{id}} {{yield}}',
        ComponentClass: _helpers.Component.extend({
          willDestroy() {
            this._super();
            destroyed[this.get('id')]++;
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#if cond1}}
        {{#foo-bar id=1}}
          {{#if cond2}}
            {{#foo-bar id=2}}{{/foo-bar}}
            {{#if cond3}}
              {{#foo-bar id=3}}
                {{#if cond4}}
                  {{#foo-bar id=4}}
                    {{#if cond5}}
                      {{#foo-bar id=5}}{{/foo-bar}}
                      {{#foo-bar id=6}}{{/foo-bar}}
                      {{#foo-bar id=7}}{{/foo-bar}}
                    {{/if}}
                    {{#foo-bar id=8}}{{/foo-bar}}
                  {{/foo-bar}}
                {{/if}}
              {{/foo-bar}}
            {{/if}}
          {{/if}}
        {{/foo-bar}}
      {{/if}}`, {
        cond1: true,
        cond2: true,
        cond3: true,
        cond4: true,
        cond5: true
      });

      this.assertText('1 2 3 4 5 6 7 8 ');

      this.runTask(() => this.rerender());

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 0,
        6: 0,
        7: 0,
        8: 0
      });

      this.runTask(() => (0, _metal.set)(this.context, 'cond5', false));

      this.assertText('1 2 3 4 8 ');

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 1,
        6: 1,
        7: 1,
        8: 0
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond3', false);
        (0, _metal.set)(this.context, 'cond5', true);
        (0, _metal.set)(this.context, 'cond4', false);
      });

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 1,
        4: 1,
        5: 1,
        6: 1,
        7: 1,
        8: 1
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond2', false);
        (0, _metal.set)(this.context, 'cond1', false);
      });

      assert.deepEqual(destroyed, {
        1: 1,
        2: 1,
        3: 1,
        4: 1,
        5: 1,
        6: 1,
        7: 1,
        8: 1
      });
    }

    ['@test should escape HTML in normal mustaches']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        output: 'you need to be more <b>bold</b>'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{output}}'
      });

      this.render('{{foo-bar}}');

      this.assertText('you need to be more <b>bold</b>');

      this.runTask(() => this.rerender());

      this.assertText('you need to be more <b>bold</b>');

      this.runTask(() => (0, _metal.set)(component, 'output', 'you are so <i>super</i>'));

      this.assertText('you are so <i>super</i>');

      this.runTask(() => (0, _metal.set)(component, 'output', 'you need to be more <b>bold</b>'));
    }

    ['@test should not escape HTML in triple mustaches']() {
      let expectedHtmlBold = 'you need to be more <b>bold</b>';
      let expectedHtmlItalic = 'you are so <i>super</i>';
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        output: expectedHtmlBold
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{{output}}}'
      });

      this.render('{{foo-bar}}');

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);

      this.runTask(() => this.rerender());

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);

      this.runTask(() => (0, _metal.set)(component, 'output', expectedHtmlItalic));

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlItalic);

      this.runTask(() => (0, _metal.set)(component, 'output', expectedHtmlBold));

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);
    }

    ['@test should not escape HTML if string is a htmlSafe']() {
      let expectedHtmlBold = 'you need to be more <b>bold</b>';
      let expectedHtmlItalic = 'you are so <i>super</i>';
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        output: (0, _helpers.htmlSafe)(expectedHtmlBold)
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{output}}'
      });

      this.render('{{foo-bar}}');

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);

      this.runTask(() => this.rerender());

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);

      this.runTask(() => (0, _metal.set)(component, 'output', (0, _helpers.htmlSafe)(expectedHtmlItalic)));

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlItalic);

      this.runTask(() => (0, _metal.set)(component, 'output', (0, _helpers.htmlSafe)(expectedHtmlBold)));

      (0, _testHelpers.equalTokens)(this.firstChild, expectedHtmlBold);
    }

    ['@test late bound layouts return the same definition'](assert) {
      let templateIds = [];

      // This is testing the scenario where you import a template and
      // set it to the layout property:
      //
      // import Component from '@ember/component';
      // import layout from './template';
      //
      // export default Component.extend({
      //   layout
      // });
      let hello = (0, _helpers.compile)('Hello');
      let bye = (0, _helpers.compile)('Bye');

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          this.layout = this.cond ? hello : bye;
          templateIds.push(this.layout.id);
        }
      });

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar cond=true}}{{foo-bar cond=false}}{{foo-bar cond=true}}{{foo-bar cond=false}}');

      let [t1, t2, t3, t4] = templateIds;
      assert.equal(t1, t3);
      assert.equal(t2, t4);
    }

    ['@test can use isStream property without conflict (#13271)']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        isStream: true,

        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,

        template: _abstractTestCase.strip`
        {{#if isStream}}
          true
        {{else}}
          false
        {{/if}}
      `
      });

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, { content: 'true' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'true' });

      this.runTask(() => (0, _metal.set)(component, 'isStream', false));

      this.assertComponentElement(this.firstChild, { content: 'false' });

      this.runTask(() => (0, _metal.set)(component, 'isStream', true));

      this.assertComponentElement(this.firstChild, { content: 'true' });
    }
    ['@test lookup of component takes priority over property']() {
      this.registerComponent('some-component', {
        template: 'some-component'
      });

      this.render('{{some-prop}} {{some-component}}', {
        'some-component': 'not-some-component',
        'some-prop': 'some-prop'
      });

      this.assertText('some-prop some-component');

      this.runTask(() => this.rerender());

      this.assertText('some-prop some-component');
    }

    ['@test component without dash is not looked up']() {
      this.registerComponent('somecomponent', {
        template: 'somecomponent'
      });

      this.render('{{somecomponent}}', {
        somecomponent: 'notsomecomponent'
      });

      this.assertText('notsomecomponent');

      this.runTask(() => this.rerender());

      this.assertText('notsomecomponent');

      this.runTask(() => this.context.set('somecomponent', 'not not notsomecomponent'));

      this.assertText('not not notsomecomponent');

      this.runTask(() => this.context.set('somecomponent', 'notsomecomponent'));

      this.assertText('notsomecomponent');
    }

    ['@test non-block with properties on attrs']() {
      this.registerComponent('non-block', {
        template: 'In layout - someProp: {{attrs.someProp}}'
      });

      this.render('{{non-block someProp=prop}}', {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.context.set('prop', 'other thing there'));

      this.assertText('In layout - someProp: other thing there');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here');
    }

    ['@feature(ember-glimmer-named-arguments) non-block with named argument']() {
      this.registerComponent('non-block', {
        template: 'In layout - someProp: {{@someProp}}'
      });

      this.render('{{non-block someProp=prop}}', {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.context.set('prop', 'other thing there'));

      this.assertText('In layout - someProp: other thing there');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here');
    }

    ['@test non-block with properties overridden in init']() {
      let instance;
      this.registerComponent('non-block', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            instance = this;
            this.someProp = 'value set in instance';
          }
        }),
        template: 'In layout - someProp: {{someProp}}'
      });

      this.render('{{non-block someProp=prop}}', {
        prop: 'something passed when invoked'
      });

      this.assertText('In layout - someProp: value set in instance');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: value set in instance');

      this.runTask(() => this.context.set('prop', 'updated something passed when invoked'));

      this.assertText('In layout - someProp: updated something passed when invoked');

      this.runTask(() => instance.set('someProp', 'update value set in instance'));

      this.assertText('In layout - someProp: update value set in instance');

      this.runTask(() => this.context.set('prop', 'something passed when invoked'));
      this.runTask(() => instance.set('someProp', 'value set in instance'));

      this.assertText('In layout - someProp: value set in instance');
    }

    ['@test rerendering component with attrs from parent'](assert) {
      let willUpdateCount = 0;
      let didReceiveAttrsCount = 0;

      function expectHooks({ willUpdate, didReceiveAttrs }, callback) {
        willUpdateCount = 0;
        didReceiveAttrsCount = 0;

        callback();

        if (willUpdate) {
          assert.strictEqual(willUpdateCount, 1, 'The willUpdate hook was fired');
        } else {
          assert.strictEqual(willUpdateCount, 0, 'The willUpdate hook was not fired');
        }

        if (didReceiveAttrs) {
          assert.strictEqual(didReceiveAttrsCount, 1, 'The didReceiveAttrs hook was fired');
        } else {
          assert.strictEqual(didReceiveAttrsCount, 0, 'The didReceiveAttrs hook was not fired');
        }
      }

      this.registerComponent('non-block', {
        ComponentClass: _helpers.Component.extend({
          didReceiveAttrs() {
            didReceiveAttrsCount++;
          },

          willUpdate() {
            willUpdateCount++;
          }
        }),
        template: 'In layout - someProp: {{someProp}}'
      });

      expectHooks({ willUpdate: false, didReceiveAttrs: true }, () => {
        this.render('{{non-block someProp=someProp}}', {
          someProp: 'wycats'
        });
      });

      this.assertText('In layout - someProp: wycats');

      // Note: Hooks are not fired in Glimmer for idempotent re-renders
      expectHooks({ willUpdate: false, didReceiveAttrs: false }, () => {
        this.runTask(() => this.rerender());
      });

      this.assertText('In layout - someProp: wycats');

      expectHooks({ willUpdate: true, didReceiveAttrs: true }, () => {
        this.runTask(() => this.context.set('someProp', 'tomdale'));
      });

      this.assertText('In layout - someProp: tomdale');

      // Note: Hooks are not fired in Glimmer for idempotent re-renders
      expectHooks({ willUpdate: false, didReceiveAttrs: false }, () => {
        this.runTask(() => this.rerender());
      });

      this.assertText('In layout - someProp: tomdale');

      expectHooks({ willUpdate: true, didReceiveAttrs: true }, () => {
        this.runTask(() => this.context.set('someProp', 'wycats'));
      });

      this.assertText('In layout - someProp: wycats');
    }

    ['@feature(!ember-glimmer-named-arguments) this.attrs.foo === attrs.foo === foo']() {
      this.registerComponent('foo-bar', {
        template: _abstractTestCase.strip`
        Args: {{this.attrs.value}} | {{attrs.value}} | {{value}}
        {{#each this.attrs.items as |item|}}
          {{item}}
        {{/each}}
        {{#each attrs.items as |item|}}
          {{item}}
        {{/each}}
        {{#each items as |item|}}
          {{item}}
        {{/each}}
      `
      });

      this.render('{{foo-bar value=model.value items=model.items}}', {
        model: {
          value: 'wat',
          items: [1, 2, 3]
        }
      });

      this.assertStableRerender();

      this.runTask(() => {
        this.context.set('model.value', 'lul');
        this.context.set('model.items', [1]);
      });

      this.assertText(_abstractTestCase.strip`Args: lul | lul | lul111`);

      this.runTask(() => this.context.set('model', { value: 'wat', items: [1, 2, 3] }));

      this.assertText('Args: wat | wat | wat123123123');
    }

    ['@feature(ember-glimmer-named-arguments) this.attrs.foo === attrs.foo === @foo === foo']() {
      this.registerComponent('foo-bar', {
        template: _abstractTestCase.strip`
        Args: {{this.attrs.value}} | {{attrs.value}} | {{@value}} | {{value}}
        {{#each this.attrs.items as |item|}}
          {{item}}
        {{/each}}
        {{#each attrs.items as |item|}}
          {{item}}
        {{/each}}
        {{#each @items as |item|}}
          {{item}}
        {{/each}}
        {{#each items as |item|}}
          {{item}}
        {{/each}}
      `
      });

      this.render('{{foo-bar value=model.value items=model.items}}', {
        model: {
          value: 'wat',
          items: [1, 2, 3]
        }
      });

      this.assertStableRerender();

      this.runTask(() => {
        this.context.set('model.value', 'lul');
        this.context.set('model.items', [1]);
      });

      this.assertText(_abstractTestCase.strip`Args: lul | lul | lul | lul1111`);

      this.runTask(() => this.context.set('model', { value: 'wat', items: [1, 2, 3] }));

      this.assertText('Args: wat | wat | wat | wat123123123123');
    }

    ['@test non-block with properties on self']() {
      this.registerComponent('non-block', {
        template: 'In layout - someProp: {{someProp}}'
      });

      this.render('{{non-block someProp=prop}}', {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here');

      this.runTask(() => this.context.set('prop', 'something else'));

      this.assertText('In layout - someProp: something else');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here');
    }

    ['@test block with properties on self']() {
      this.registerComponent('with-block', {
        template: 'In layout - someProp: {{someProp}} - {{yield}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with-block someProp=prop}}
        In template
      {{/with-block}}`, {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.context.set('prop', 'something else'));

      this.assertText('In layout - someProp: something else - In template');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here - In template');
    }

    ['@test block with properties on attrs']() {
      this.registerComponent('with-block', {
        template: 'In layout - someProp: {{attrs.someProp}} - {{yield}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with-block someProp=prop}}
        In template
      {{/with-block}}`, {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.context.set('prop', 'something else'));

      this.assertText('In layout - someProp: something else - In template');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here - In template');
    }

    ['@feature(ember-glimmer-named-arguments) block with named argument']() {
      this.registerComponent('with-block', {
        template: 'In layout - someProp: {{@someProp}} - {{yield}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with-block someProp=prop}}
        In template
      {{/with-block}}`, {
        prop: 'something here'
      });

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.rerender());

      this.assertText('In layout - someProp: something here - In template');

      this.runTask(() => this.context.set('prop', 'something else'));

      this.assertText('In layout - someProp: something else - In template');

      this.runTask(() => this.context.set('prop', 'something here'));

      this.assertText('In layout - someProp: something here - In template');
    }

    ['@test static arbitrary number of positional parameters'](assert) {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'names'
        }),
        template: _abstractTestCase.strip`
        {{#each names as |name|}}
          {{name}}
        {{/each}}`
      });

      this.render(_abstractTestCase.strip`
      {{sample-component "Foo" 4 "Bar" elementId="args-3"}}
      {{sample-component "Foo" 4 "Bar" 5 "Baz" elementId="args-5"}}`);

      assert.equal(this.$('#args-3').text(), 'Foo4Bar');
      assert.equal(this.$('#args-5').text(), 'Foo4Bar5Baz');

      this.runTask(() => this.rerender());

      assert.equal(this.$('#args-3').text(), 'Foo4Bar');
      assert.equal(this.$('#args-5').text(), 'Foo4Bar5Baz');
    }

    ['@test arbitrary positional parameter conflict with hash parameter is reported']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'names'
        }),
        template: _abstractTestCase.strip`
        {{#each names as |name|}}
          {{name}}
        {{/each}}`
      });

      expectAssertion(() => {
        this.render(`{{sample-component "Foo" 4 "Bar" names=numbers id="args-3"}}`, {
          numbers: [1, 2, 3]
        });
      }, 'You cannot specify positional parameters and the hash argument `names`.');
    }

    ['@test can use hash parameter instead of arbitrary positional param [GH #12444]']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'names'
        }),
        template: _abstractTestCase.strip`
        {{#each names as |name|}}
          {{name}}
        {{/each}}`
      });

      this.render('{{sample-component names=things}}', {
        things: (0, _runtime.A)(['Foo', 4, 'Bar'])
      });

      this.assertText('Foo4Bar');

      this.runTask(() => this.rerender());

      this.assertText('Foo4Bar');

      this.runTask(() => this.context.get('things').pushObject(5));

      this.assertText('Foo4Bar5');

      this.runTask(() => this.context.get('things').shiftObject());

      this.assertText('4Bar5');

      this.runTask(() => this.context.get('things').clear());

      this.assertText('');

      this.runTask(() => this.context.set('things', (0, _runtime.A)(['Foo', 4, 'Bar'])));

      this.assertText('Foo4Bar');
    }

    ['@test can use hash parameter instead of positional param'](assert) {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['first', 'second']
        }),
        template: '{{first}} - {{second}}'
      });

      // TODO: Fix when id is implemented
      this.render(_abstractTestCase.strip`
      {{sample-component "one" "two" elementId="two-positional"}}
      {{sample-component "one" second="two" elementId="one-positional"}}
      {{sample-component first="one" second="two" elementId="no-positional"}}`);

      assert.equal(this.$('#two-positional').text(), 'one - two');
      assert.equal(this.$('#one-positional').text(), 'one - two');
      assert.equal(this.$('#no-positional').text(), 'one - two');

      this.runTask(() => this.rerender());

      assert.equal(this.$('#two-positional').text(), 'one - two');
      assert.equal(this.$('#one-positional').text(), 'one - two');
      assert.equal(this.$('#no-positional').text(), 'one - two');
    }

    ['@test dynamic arbitrary number of positional parameters']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'n'
        }),
        template: _abstractTestCase.strip`
        {{#each n as |name|}}
          {{name}}
        {{/each}}`
      });

      this.render(`{{sample-component user1 user2}}`, {
        user1: 'Foo',
        user2: 4
      });

      this.assertText('Foo4');

      this.runTask(() => this.rerender());

      this.assertText('Foo4');

      this.runTask(() => this.context.set('user1', 'Bar'));

      this.assertText('Bar4');

      this.runTask(() => this.context.set('user2', '5'));

      this.assertText('Bar5');

      this.runTask(() => {
        this.context.set('user1', 'Foo');
        this.context.set('user2', 4);
      });

      this.assertText('Foo4');
    }

    ['@test with ariaRole specified']() {
      this.registerComponent('aria-test', {
        template: 'Here!'
      });

      this.render('{{aria-test ariaRole=role}}', {
        role: 'main'
      });

      this.assertComponentElement(this.firstChild, { attrs: { role: 'main' } });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { attrs: { role: 'main' } });

      this.runTask(() => this.context.set('role', 'input'));

      this.assertComponentElement(this.firstChild, {
        attrs: { role: 'input' }
      });

      this.runTask(() => this.context.set('role', 'main'));

      this.assertComponentElement(this.firstChild, { attrs: { role: 'main' } });
    }

    ['@test with ariaRole defined but initially falsey GH#16379']() {
      this.registerComponent('aria-test', {
        template: 'Here!'
      });

      this.render('{{aria-test ariaRole=role}}', {
        role: undefined
      });

      this.assertComponentElement(this.firstChild, { attrs: {} });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { attrs: {} });

      this.runTask(() => this.context.set('role', 'input'));

      this.assertComponentElement(this.firstChild, {
        attrs: { role: 'input' }
      });

      this.runTask(() => this.context.set('role', undefined));

      this.assertComponentElement(this.firstChild, { attrs: {} });
    }

    ['@test without ariaRole defined initially']() {
      // we are using the ability to lazily add a role as a sign that we are
      // doing extra work
      let instance;
      this.registerComponent('aria-test', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            instance = this;
          }
        }),
        template: 'Here!'
      });

      this.render('{{aria-test}}');

      this.assertComponentElement(this.firstChild, { attrs: {} });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { attrs: {} });

      this.runTask(() => instance.set('ariaRole', 'input'));

      this.assertComponentElement(this.firstChild, { attrs: {} });
    }

    ['@test `template` specified in component is overridden by block']() {
      this.registerComponent('with-template', {
        ComponentClass: _helpers.Component.extend({
          template: (0, _helpers.compile)('Should not be used')
        }),
        template: '[In layout - {{name}}] {{yield}}'
      });

      this.render(_abstractTestCase.strip`
      {{#with-template name="with-block"}}
        [In block - {{name}}]
      {{/with-template}}
      {{with-template name="without-block"}}`, {
        name: 'Whoop, whoop!'
      });

      this.assertText('[In layout - with-block] [In block - Whoop, whoop!][In layout - without-block] ');

      this.runTask(() => this.rerender());

      this.assertText('[In layout - with-block] [In block - Whoop, whoop!][In layout - without-block] ');

      this.runTask(() => this.context.set('name', 'Ole, ole'));

      this.assertText('[In layout - with-block] [In block - Ole, ole][In layout - without-block] ');

      this.runTask(() => this.context.set('name', 'Whoop, whoop!'));

      this.assertText('[In layout - with-block] [In block - Whoop, whoop!][In layout - without-block] ');
    }

    ['@test hasBlock is true when block supplied']() {
      this.registerComponent('with-block', {
        template: _abstractTestCase.strip`
        {{#if hasBlock}}
          {{yield}}
        {{else}}
          No Block!
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#with-block}}
        In template
      {{/with-block}}`);

      this.assertText('In template');

      this.runTask(() => this.rerender());

      this.assertText('In template');
    }

    ['@test hasBlock is false when no block supplied']() {
      this.registerComponent('with-block', {
        template: _abstractTestCase.strip`
        {{#if hasBlock}}
          {{yield}}
        {{else}}
          No Block!
        {{/if}}`
      });

      this.render('{{with-block}}');

      this.assertText('No Block!');

      this.runTask(() => this.rerender());

      this.assertText('No Block!');
    }

    ['@test hasBlockParams is true when block param supplied']() {
      this.registerComponent('with-block', {
        template: _abstractTestCase.strip`
        {{#if hasBlockParams}}
          {{yield this}} - In Component
        {{else}}
          {{yield}} No Block!
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#with-block as |something|}}
        In template
      {{/with-block}}`);

      this.assertText('In template - In Component');

      this.runTask(() => this.rerender());

      this.assertText('In template - In Component');
    }

    ['@test hasBlockParams is false when no block param supplied']() {
      this.registerComponent('with-block', {
        template: _abstractTestCase.strip`
        {{#if hasBlockParams}}
          {{yield this}}
        {{else}}
          {{yield}} No Block Param!
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#with-block}}
        In block
      {{/with-block}}`);

      this.assertText('In block No Block Param!');

      this.runTask(() => this.rerender());

      this.assertText('In block No Block Param!');
    }

    ['@test static named positional parameters']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        }),
        template: '{{name}}{{age}}'
      });

      this.render('{{sample-component "Quint" 4}}');

      this.assertText('Quint4');

      this.runTask(() => this.rerender());

      this.assertText('Quint4');
    }

    ['@test dynamic named positional parameters']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        }),
        template: '{{name}}{{age}}'
      });

      this.render('{{sample-component myName myAge}}', {
        myName: 'Quint',
        myAge: 4
      });

      this.assertText('Quint4');

      this.runTask(() => this.rerender());

      this.assertText('Quint4');

      this.runTask(() => this.context.set('myName', 'Sergio'));

      this.assertText('Sergio4');

      this.runTask(() => this.context.set('myAge', 2));

      this.assertText('Sergio2');

      this.runTask(() => {
        this.context.set('myName', 'Quint');
        this.context.set('myAge', 4);
      });

      this.assertText('Quint4');
    }

    ['@test if a value is passed as a non-positional parameter, it raises an assertion']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name']
        }),
        template: '{{name}}'
      });

      expectDeprecation(() => {
        this.render('{{sample-component notMyName name=myName}}', {
          myName: 'Quint',
          notMyName: 'Sergio'
        });
      }, 'You cannot specify both a positional param (at position 0) and the hash argument `name`.');
    }

    ['@test yield to inverse']() {
      this.registerComponent('my-if', {
        template: _abstractTestCase.strip`
        {{#if predicate}}
          Yes:{{yield someValue}}
        {{else}}
          No:{{yield to="inverse"}}
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#my-if predicate=activated someValue=42 as |result|}}
        Hello{{result}}
      {{else}}
        Goodbye
      {{/my-if}}`, {
        activated: true
      });

      this.assertText('Yes:Hello42');

      this.runTask(() => this.rerender());

      this.assertText('Yes:Hello42');

      this.runTask(() => this.context.set('activated', false));

      this.assertText('No:Goodbye');

      this.runTask(() => this.context.set('activated', true));

      this.assertText('Yes:Hello42');
    }

    ['@test expression hasBlock inverse']() {
      this.registerComponent('check-inverse', {
        template: _abstractTestCase.strip`
        {{#if (hasBlock "inverse")}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#check-inverse}}{{/check-inverse}}
      {{#check-inverse}}{{else}}{{/check-inverse}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test expression hasBlock default']() {
      this.registerComponent('check-block', {
        template: _abstractTestCase.strip`
        {{#if (hasBlock)}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{check-block}}
      {{#check-block}}{{/check-block}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test expression hasBlockParams inverse']() {
      this.registerComponent('check-inverse', {
        template: _abstractTestCase.strip`
        {{#if (hasBlockParams "inverse")}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#check-inverse}}{{/check-inverse}}
      {{#check-inverse as |something|}}{{/check-inverse}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'No' });

      this.assertStableRerender();
    }

    ['@test expression hasBlockParams default']() {
      this.registerComponent('check-block', {
        template: _abstractTestCase.strip`
        {{#if (hasBlockParams)}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#check-block}}{{/check-block}}
      {{#check-block as |something|}}{{/check-block}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test non-expression hasBlock']() {
      this.registerComponent('check-block', {
        template: _abstractTestCase.strip`
        {{#if hasBlock}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{check-block}}
      {{#check-block}}{{/check-block}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test expression hasBlockParams']() {
      this.registerComponent('check-params', {
        template: _abstractTestCase.strip`
        {{#if (hasBlockParams)}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#check-params}}{{/check-params}}
      {{#check-params as |foo|}}{{/check-params}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test non-expression hasBlockParams']() {
      this.registerComponent('check-params', {
        template: _abstractTestCase.strip`
        {{#if hasBlockParams}}
          Yes
        {{else}}
          No
        {{/if}}`
      });

      this.render(_abstractTestCase.strip`
      {{#check-params}}{{/check-params}}
      {{#check-params as |foo|}}{{/check-params}}`);

      this.assertComponentElement(this.firstChild, { content: 'No' });
      this.assertComponentElement(this.nthChild(1), { content: 'Yes' });

      this.assertStableRerender();
    }

    ['@test hasBlock expression in an attribute'](assert) {
      this.registerComponent('check-attr', {
        template: '<button name={{hasBlock}}></button>'
      });

      this.render(_abstractTestCase.strip`
      {{check-attr}}
      {{#check-attr}}{{/check-attr}}`);

      (0, _testHelpers.equalsElement)(assert, this.$('button')[0], 'button', { name: 'false' }, '');
      (0, _testHelpers.equalsElement)(assert, this.$('button')[1], 'button', { name: 'true' }, '');

      this.assertStableRerender();
    }

    ['@test hasBlock inverse expression in an attribute'](assert) {
      this.registerComponent('check-attr', {
        template: '<button name={{hasBlock "inverse"}}></button>'
      }, '');

      this.render(_abstractTestCase.strip`
      {{#check-attr}}{{/check-attr}}
      {{#check-attr}}{{else}}{{/check-attr}}`);

      (0, _testHelpers.equalsElement)(assert, this.$('button')[0], 'button', { name: 'false' }, '');
      (0, _testHelpers.equalsElement)(assert, this.$('button')[1], 'button', { name: 'true' }, '');

      this.assertStableRerender();
    }

    ['@test hasBlockParams expression in an attribute'](assert) {
      this.registerComponent('check-attr', {
        template: '<button name={{hasBlockParams}}></button>'
      });

      this.render(_abstractTestCase.strip`
      {{#check-attr}}{{/check-attr}}
      {{#check-attr as |something|}}{{/check-attr}}`);

      (0, _testHelpers.equalsElement)(assert, this.$('button')[0], 'button', { name: 'false' }, '');
      (0, _testHelpers.equalsElement)(assert, this.$('button')[1], 'button', { name: 'true' }, '');

      this.assertStableRerender();
    }

    ['@test hasBlockParams inverse expression in an attribute'](assert) {
      this.registerComponent('check-attr', {
        template: '<button name={{hasBlockParams "inverse"}}></button>'
      }, '');

      this.render(_abstractTestCase.strip`
      {{#check-attr}}{{/check-attr}}
      {{#check-attr as |something|}}{{/check-attr}}`);

      (0, _testHelpers.equalsElement)(assert, this.$('button')[0], 'button', { name: 'false' }, '');
      (0, _testHelpers.equalsElement)(assert, this.$('button')[1], 'button', { name: 'false' }, '');

      this.assertStableRerender();
    }

    ['@test hasBlock as a param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if hasBlock "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{check-helper}}
      {{#check-helper}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'true' });

      this.assertStableRerender();
    }

    ['@test hasBlock as an expression param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if (hasBlock) "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{check-helper}}
      {{#check-helper}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'true' });

      this.assertStableRerender();
    }

    ['@test hasBlock inverse as a param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if (hasBlock "inverse") "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#check-helper}}{{/check-helper}}
      {{#check-helper}}{{else}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'true' });

      this.assertStableRerender();
    }

    ['@test hasBlockParams as a param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if hasBlockParams "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#check-helper}}{{/check-helper}}
      {{#check-helper as |something|}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'true' });

      this.assertStableRerender();
    }

    ['@test hasBlockParams as an expression param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if (hasBlockParams) "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#check-helper}}{{/check-helper}}
      {{#check-helper as |something|}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'true' });

      this.assertStableRerender();
    }

    ['@test hasBlockParams inverse as a param to a helper']() {
      this.registerComponent('check-helper', {
        template: '{{if (hasBlockParams "inverse") "true" "false"}}'
      });

      this.render(_abstractTestCase.strip`
      {{#check-helper}}{{/check-helper}}
      {{#check-helper as |something|}}{{/check-helper}}`);

      this.assertComponentElement(this.firstChild, { content: 'false' });
      this.assertComponentElement(this.nthChild(1), { content: 'false' });

      this.assertStableRerender();
    }

    ['@test component in template of a yielding component should have the proper parentView'](assert) {
      let outer, innerTemplate, innerLayout;

      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            outer = this;
          }
        }),
        template: '{{x-inner-in-layout}}{{yield}}'
      });

      this.registerComponent('x-inner-in-template', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            innerTemplate = this;
          }
        })
      });

      this.registerComponent('x-inner-in-layout', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            innerLayout = this;
          }
        })
      });

      this.render('{{#x-outer}}{{x-inner-in-template}}{{/x-outer}}');

      assert.equal(innerTemplate.parentView, outer, 'receives the wrapping component as its parentView in template blocks');
      assert.equal(innerLayout.parentView, outer, 'receives the wrapping component as its parentView in layout');
      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView');

      this.runTask(() => this.rerender());

      assert.equal(innerTemplate.parentView, outer, 'receives the wrapping component as its parentView in template blocks');
      assert.equal(innerLayout.parentView, outer, 'receives the wrapping component as its parentView in layout');
      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView');
    }

    ['@test newly-added sub-components get correct parentView'](assert) {
      let outer, inner;

      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            outer = this;
          }
        })
      });

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            inner = this;
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#x-outer}}
        {{#if showInner}}
          {{x-inner}}
        {{/if}}
      {{/x-outer}}`, {
        showInner: false
      });

      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView');

      this.runTask(() => this.rerender());

      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView (after rerender)');

      this.runTask(() => this.context.set('showInner', true));

      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView');
      assert.equal(inner.parentView, outer, 'receives the wrapping component as its parentView in template blocks');

      this.runTask(() => this.context.set('showInner', false));

      assert.equal(outer.parentView, this.context, 'x-outer receives the ambient scope as its parentView');
    }

    ["@test when a property is changed during children's rendering"](assert) {
      let outer, middle;

      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            outer = this;
          },
          value: 1
        }),
        template: '{{#x-middle}}{{x-inner value=value}}{{/x-middle}}'
      });

      this.registerComponent('x-middle', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            middle = this;
          },
          value: null
        }),
        template: '<div id="middle-value">{{value}}</div>{{yield}}'
      });

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          value: null,
          pushDataUp: (0, _metal.observer)('value', function () {
            middle.set('value', this.get('value'));
          })
        }),
        template: '<div id="inner-value">{{value}}</div>'
      });

      this.render('{{x-outer}}');

      assert.equal(this.$('#inner-value').text(), '1', 'initial render of inner');
      assert.equal(this.$('#middle-value').text(), '', 'initial render of middle (observers do not run during init)');

      this.runTask(() => this.rerender());

      assert.equal(this.$('#inner-value').text(), '1', 'initial render of inner');
      assert.equal(this.$('#middle-value').text(), '', 'initial render of middle (observers do not run during init)');

      let expectedBacktrackingMessage = /modified "value" twice on <.+?> in a single render\. It was rendered in "component:x-middle" and modified in "component:x-inner"/;

      expectAssertion(() => {
        this.runTask(() => outer.set('value', 2));
      }, expectedBacktrackingMessage);
    }

    ["@test when a shared dependency is changed during children's rendering"]() {
      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          value: 1,
          wrapper: _runtime.Object.create({ content: null })
        }),
        template: '<div id="outer-value">{{wrapper.content}}</div> {{x-inner value=value wrapper=wrapper}}'
      });

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          didReceiveAttrs() {
            this.get('wrapper').set('content', this.get('value'));
          },
          value: null
        }),
        template: '<div id="inner-value">{{wrapper.content}}</div>'
      });

      let expectedBacktrackingMessage = /modified "wrapper\.content" twice on <.+?> in a single render\. It was rendered in "component:x-outer" and modified in "component:x-inner"/;

      expectAssertion(() => {
        this.render('{{x-outer}}');
      }, expectedBacktrackingMessage);
    }

    ['@test non-block with each rendering child components']() {
      this.registerComponent('non-block', {
        template: _abstractTestCase.strip`
        In layout. {{#each items as |item|}}
          [{{child-non-block item=item}}]
        {{/each}}`
      });

      this.registerComponent('child-non-block', {
        template: 'Child: {{item}}.'
      });

      let items = (0, _runtime.A)(['Tom', 'Dick', 'Harry']);

      this.render('{{non-block items=items}}', { items });

      this.assertText('In layout. [Child: Tom.][Child: Dick.][Child: Harry.]');

      this.runTask(() => this.rerender());

      this.assertText('In layout. [Child: Tom.][Child: Dick.][Child: Harry.]');

      this.runTask(() => this.context.get('items').pushObject('Sergio'));

      this.assertText('In layout. [Child: Tom.][Child: Dick.][Child: Harry.][Child: Sergio.]');

      this.runTask(() => this.context.get('items').shiftObject());

      this.assertText('In layout. [Child: Dick.][Child: Harry.][Child: Sergio.]');

      this.runTask(() => this.context.set('items', (0, _runtime.A)(['Tom', 'Dick', 'Harry'])));

      this.assertText('In layout. [Child: Tom.][Child: Dick.][Child: Harry.]');
    }

    ['@test specifying classNames results in correct class'](assert) {
      this.registerComponent('some-clicky-thing', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'button',
          classNames: ['foo', 'bar']
        })
      });

      this.render(_abstractTestCase.strip`
      {{#some-clicky-thing classNames="baz"}}
        Click Me
      {{/some-clicky-thing}}`);

      // TODO: ember-view is no longer viewable in the classNames array. Bug or
      // feature?
      let expectedClassNames = ['ember-view', 'foo', 'bar', 'baz'];

      assert.ok(this.$('button').is('.foo.bar.baz.ember-view'), `the element has the correct classes: ${this.$('button').attr('class')}`);
      // `ember-view` is no longer in classNames.
      // assert.deepEqual(clickyThing.get('classNames'), expectedClassNames, 'classNames are properly combined');
      this.assertComponentElement(this.firstChild, {
        tagName: 'button',
        attrs: { class: (0, _testHelpers.classes)(expectedClassNames.join(' ')) }
      });

      this.runTask(() => this.rerender());

      assert.ok(this.$('button').is('.foo.bar.baz.ember-view'), `the element has the correct classes: ${this.$('button').attr('class')} (rerender)`);
      // `ember-view` is no longer in classNames.
      // assert.deepEqual(clickyThing.get('classNames'), expectedClassNames, 'classNames are properly combined (rerender)');
      this.assertComponentElement(this.firstChild, {
        tagName: 'button',
        attrs: { class: (0, _testHelpers.classes)(expectedClassNames.join(' ')) }
      });
    }

    ['@test specifying custom concatenatedProperties avoids clobbering']() {
      this.registerComponent('some-clicky-thing', {
        ComponentClass: _helpers.Component.extend({
          concatenatedProperties: ['blahzz'],
          blahzz: ['blark', 'pory']
        }),
        template: _abstractTestCase.strip`
        {{#each blahzz as |p|}}
          {{p}}
        {{/each}}
        - {{yield}}`
      });

      this.render(_abstractTestCase.strip`
      {{#some-clicky-thing blahzz="baz"}}
        Click Me
      {{/some-clicky-thing}}`);

      this.assertText('blarkporybaz- Click Me');

      this.runTask(() => this.rerender());

      this.assertText('blarkporybaz- Click Me');
    }

    ['@test a two way binding flows upstream when consumed in the template']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,

        template: '{{bar}}'
      });

      this.render('{{localBar}} - {{foo-bar bar=localBar}}', {
        localBar: 'initial value'
      });

      this.assertText('initial value - initial value');

      this.runTask(() => this.rerender());

      this.assertText('initial value - initial value');

      if (true /* DEBUG */) {
          expectAssertion(() => {
            component.bar = 'foo-bar';
          }, /You must use set\(\) to set the `bar` property \(of .+\) to `foo-bar`\./);

          this.assertText('initial value - initial value');
        }

      this.runTask(() => {
        component.set('bar', 'updated value');
      });

      this.assertText('updated value - updated value');

      this.runTask(() => {
        component.set('bar', undefined);
      });

      this.assertText(' - ');

      this.runTask(() => {
        this.component.set('localBar', 'initial value');
      });

      this.assertText('initial value - initial value');
    }

    ['@test a two way binding flows upstream through a CP when consumed in the template']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },

        bar: (0, _metal.computed)({
          get() {
            return this._bar;
          },

          set(key, value) {
            this._bar = value;
            return this._bar;
          }
        })
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,

        template: '{{bar}}'
      });

      this.render('{{localBar}} - {{foo-bar bar=localBar}}', {
        localBar: 'initial value'
      });

      this.assertText('initial value - initial value');

      this.runTask(() => this.rerender());

      this.assertText('initial value - initial value');

      this.runTask(() => {
        component.set('bar', 'updated value');
      });

      this.assertText('updated value - updated value');

      this.runTask(() => {
        this.component.set('localBar', 'initial value');
      });

      this.assertText('initial value - initial value');
    }

    ['@test a two way binding flows upstream through a CP without template consumption']() {
      let component;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },

        bar: (0, _metal.computed)({
          get() {
            return this._bar;
          },

          set(key, value) {
            this._bar = value;
            return this._bar;
          }
        })
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: ''
      });

      this.render('{{localBar}}{{foo-bar bar=localBar}}', {
        localBar: 'initial value'
      });

      this.assertText('initial value');

      this.runTask(() => this.rerender());

      this.assertText('initial value');

      this.runTask(() => {
        component.set('bar', 'updated value');
      });

      this.assertText('updated value');

      this.runTask(() => {
        this.component.set('localBar', 'initial value');
      });

      this.assertText('initial value');
    }

    ['@test services can be injected into components']() {
      let service;
      this.registerService('name', _service.default.extend({
        init() {
          this._super(...arguments);
          service = this;
        },
        last: 'Jackson'
      }));

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          name: (0, _service.inject)()
        }),
        template: '{{name.last}}'
      });

      this.render('{{foo-bar}}');

      this.assertText('Jackson');

      this.runTask(() => this.rerender());

      this.assertText('Jackson');

      this.runTask(() => {
        service.set('last', 'McGuffey');
      });

      this.assertText('McGuffey');

      this.runTask(() => {
        service.set('last', 'Jackson');
      });

      this.assertText('Jackson');
    }

    ['@test injecting an unknown service raises an exception']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          missingService: (0, _service.inject)()
        })
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, "Attempting to inject an unknown injection: 'service:missingService'");
    }

    ['@test throws if `this._super` is not called from `init`']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {}
        })
      });

      expectAssertion(() => {
        this.render('{{foo-bar}}');
      }, /You must call `this._super\(...arguments\);` when overriding `init` on a framework object. Please update .* to call `this._super\(...arguments\);` from `init`./);
    }

    ['@test should toggle visibility with isVisible'](assert) {
      let assertStyle = expected => {
        let matcher = (0, _testHelpers.styles)(expected);
        let actual = this.firstChild.getAttribute('style');

        assert.pushResult({
          result: matcher.match(actual),
          message: matcher.message(),
          actual,
          expected
        });
      };

      this.registerComponent('foo-bar', {
        template: `<p>foo</p>`
      });

      this.render(`{{foo-bar id="foo-bar" isVisible=visible}}`, {
        visible: false
      });

      assertStyle('display: none;');

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', true);
      });
      assertStyle('');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', false);
      });
      assertStyle('display: none;');
    }

    ['@test isVisible does not overwrite component style']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['style'],
          style: (0, _helpers.htmlSafe)('color: blue;')
        }),

        template: `<p>foo</p>`
      });

      this.render(`{{foo-bar id="foo-bar" isVisible=visible}}`, {
        visible: false
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'foo-bar', style: (0, _testHelpers.styles)('color: blue; display: none;') }
      });

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', true);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'foo-bar', style: (0, _testHelpers.styles)('color: blue;') }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', false);
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { id: 'foo-bar', style: (0, _testHelpers.styles)('color: blue; display: none;') }
      });
    }

    ['@test adds isVisible binding when style binding is missing and other bindings exist'](assert) {
      let assertStyle = expected => {
        let matcher = (0, _testHelpers.styles)(expected);
        let actual = this.firstChild.getAttribute('style');

        assert.pushResult({
          result: matcher.match(actual),
          message: matcher.message(),
          actual,
          expected
        });
      };

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['foo'],
          foo: 'bar'
        }),
        template: `<p>foo</p>`
      });

      this.render(`{{foo-bar id="foo-bar" foo=foo isVisible=visible}}`, {
        visible: false,
        foo: 'baz'
      });

      assertStyle('display: none;');

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', true);
      });

      assertStyle('');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'visible', false);
        (0, _metal.set)(this.context, 'foo', 'woo');
      });

      assertStyle('display: none;');
      assert.equal(this.firstChild.getAttribute('foo'), 'woo');
    }

    ['@test it can use readDOMAttr to read input value']() {
      let component;
      let assertElement = expectedValue => {
        // value is a property, not an attribute
        this.assertHTML(`<input class="ember-view" id="${component.elementId}">`);
        this.assert.equal(this.firstChild.value, expectedValue, 'value property is correct');
        this.assert.equal((0, _metal.get)(component, 'value'), expectedValue, 'component.get("value") is correct');
      };

      this.registerComponent('one-way-input', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'input',
          attributeBindings: ['value'],

          init() {
            this._super(...arguments);
            component = this;
          },

          change() {
            let value = this.readDOMAttr('value');
            this.set('value', value);
          }
        })
      });

      this.render('{{one-way-input value=value}}', {
        value: 'foo'
      });

      assertElement('foo');

      this.assertStableRerender();

      this.runTask(() => {
        this.firstChild.value = 'bar';
        this.$('input').trigger('change');
      });

      assertElement('bar');

      this.runTask(() => {
        this.firstChild.value = 'foo';
        this.$('input').trigger('change');
      });

      assertElement('foo');

      this.runTask(() => {
        (0, _metal.set)(component, 'value', 'bar');
      });

      assertElement('bar');

      this.runTask(() => {
        this.firstChild.value = 'foo';
        this.$('input').trigger('change');
      });

      assertElement('foo');
    }

    ['@test child triggers revalidate during parent destruction (GH#13846)']() {
      this.registerComponent('x-select', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'select',

          init() {
            this._super();
            this.options = (0, _runtime.A)([]);
            this.value = null;
          },

          updateValue() {
            var newValue = this.get('options.lastObject.value');

            this.set('value', newValue);
          },

          registerOption(option) {
            this.get('options').addObject(option);
          },

          unregisterOption(option) {
            this.get('options').removeObject(option);

            this.updateValue();
          }
        }),

        template: '{{yield this}}'
      });

      this.registerComponent('x-option', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'option',
          attributeBindings: ['selected'],

          didInsertElement() {
            this._super(...arguments);

            this.get('select').registerOption(this);
          },

          selected: (0, _metal.computed)('select.value', function () {
            return this.get('value') === this.get('select.value');
          }),

          willDestroyElement() {
            this._super(...arguments);
            this.get('select').unregisterOption(this);
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#x-select value=value as |select|}}
        {{#x-option value="1" select=select}}1{{/x-option}}
        {{#x-option value="2" select=select}}2{{/x-option}}
      {{/x-select}}
    `);

      this.teardown();

      this.assert.ok(true, 'no errors during teardown');
    }

    ['@test setting a property in willDestroyElement does not assert (GH#14273)'](assert) {
      assert.expect(2);

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            this.showFoo = true;
          },

          willDestroyElement() {
            this.set('showFoo', false);
            assert.ok(true, 'willDestroyElement was fired');
            this._super(...arguments);
          }
        }),

        template: `{{#if showFoo}}things{{/if}}`
      });

      this.render(`{{foo-bar}}`);

      this.assertText('things');
    }

    ['@test didReceiveAttrs fires after .init() but before observers become active'](assert) {
      let barCopyDidChangeCount = 0;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            this.didInit = true;
          },

          didReceiveAttrs() {
            assert.ok(this.didInit, 'expected init to have run before didReceiveAttrs');
            this.set('barCopy', this.attrs.bar.value + 1);
          },

          barCopyDidChange: (0, _metal.observer)('barCopy', () => {
            barCopyDidChangeCount++;
          })
        }),

        template: '{{bar}}-{{barCopy}}'
      });

      this.render(`{{foo-bar bar=bar}}`, { bar: 3 });

      this.assertText('3-4');

      assert.strictEqual(barCopyDidChangeCount, 1, 'expected observer firing for: barCopy');

      this.runTask(() => (0, _metal.set)(this.context, 'bar', 7));

      this.assertText('7-8');

      assert.strictEqual(barCopyDidChangeCount, 2, 'expected observer firing for: barCopy');
    }

    ['@test overriding didReceiveAttrs does not trigger deprecation'](assert) {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          didReceiveAttrs() {
            assert.equal(1, this.get('foo'), 'expected attrs to have correct value');
          }
        }),

        template: '{{foo}}-{{fooCopy}}-{{bar}}-{{barCopy}}'
      });

      this.render(`{{foo-bar foo=foo bar=bar}}`, { foo: 1, bar: 3 });
    }

    ['@test overriding didUpdateAttrs does not trigger deprecation'](assert) {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          didUpdateAttrs() {
            assert.equal(5, this.get('foo'), 'expected newAttrs to have new value');
          }
        }),

        template: '{{foo}}-{{fooCopy}}-{{bar}}-{{barCopy}}'
      });

      this.render(`{{foo-bar foo=foo bar=bar}}`, { foo: 1, bar: 3 });

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 5));
    }

    ['@test returning `true` from an action does not bubble if `target` is not specified (GH#14275)'](assert) {
      this.registerComponent('display-toggle', {
        ComponentClass: _helpers.Component.extend({
          actions: {
            show() {
              assert.ok(true, 'display-toggle show action was called');
              return true;
            }
          }
        }),

        template: `<button {{action 'show'}}>Show</button>`
      });

      this.render(`{{display-toggle}}`, {
        send() {
          assert.notOk(true, 'send should not be called when action is not "subscribed" to');
        }
      });

      this.assertText('Show');

      this.runTask(() => this.$('button').click());
    }

    ['@test returning `true` from an action bubbles to the `target` if specified'](assert) {
      assert.expect(4);

      this.registerComponent('display-toggle', {
        ComponentClass: _helpers.Component.extend({
          actions: {
            show() {
              assert.ok(true, 'display-toggle show action was called');
              return true;
            }
          }
        }),

        template: `<button {{action 'show'}}>Show</button>`
      });

      this.render(`{{display-toggle target=this}}`, {
        send(actionName) {
          assert.ok(true, 'send should be called when action is "subscribed" to');
          assert.equal(actionName, 'show');
        }
      });

      this.assertText('Show');

      this.runTask(() => this.$('button').click());
    }

    ['@test triggering an event only attempts to invoke an identically named method, if it actually is a function (GH#15228)'](assert) {
      assert.expect(3);

      let payload = ['arbitrary', 'event', 'data'];

      this.registerComponent('evented-component', {
        ComponentClass: _helpers.Component.extend({
          someTruthyProperty: true,

          init() {
            this._super(...arguments);
            this.trigger('someMethod', ...payload);
            this.trigger('someTruthyProperty', ...payload);
          },

          someMethod(...data) {
            assert.deepEqual(data, payload, 'the method `someMethod` should be called, when `someMethod` is triggered');
          },

          listenerForSomeMethod: (0, _metal.on)('someMethod', function (...data) {
            assert.deepEqual(data, payload, 'the listener `listenerForSomeMethod` should be called, when `someMethod` is triggered');
          }),

          listenerForSomeTruthyProperty: (0, _metal.on)('someTruthyProperty', function (...data) {
            assert.deepEqual(data, payload, 'the listener `listenerForSomeTruthyProperty` should be called, when `someTruthyProperty` is triggered');
          })
        })
      });

      this.render(`{{evented-component}}`);
    }

    ['@test component yielding in an {{#each}} has correct block values after rerendering (GH#14284)']() {
      this.registerComponent('list-items', {
        template: `{{#each items as |item|}}{{yield item}}{{/each}}`
      });

      this.render(_abstractTestCase.strip`
      {{#list-items items=items as |thing|}}
        |{{thing}}|

        {{#if editMode}}
          Remove {{thing}}
        {{/if}}
      {{/list-items}}
    `, {
        editMode: false,
        items: ['foo', 'bar', 'qux', 'baz']
      });

      this.assertText('|foo||bar||qux||baz|');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'editMode', true));

      this.assertText('|foo|Remove foo|bar|Remove bar|qux|Remove qux|baz|Remove baz');

      this.runTask(() => (0, _metal.set)(this.context, 'editMode', false));

      this.assertText('|foo||bar||qux||baz|');
    }

    ['@test unimplimented positionalParams do not cause an error GH#14416']() {
      this.registerComponent('foo-bar', {
        template: 'hello'
      });

      this.render('{{foo-bar wat}}');
      this.assertText('hello');
    }

    ['@test using attrs for positional params']() {
      let MyComponent = _helpers.Component.extend();

      this.registerComponent('foo-bar', {
        ComponentClass: MyComponent.reopenClass({
          positionalParams: ['myVar']
        }),
        template: 'MyVar1: {{attrs.myVar}} {{myVar}} MyVar2: {{myVar2}} {{attrs.myVar2}}'
      });

      this.render('{{foo-bar 1 myVar2=2}}');

      this.assertText('MyVar1: 1 1 MyVar2: 2 2');
    }

    ['@feature(ember-glimmer-named-arguments) using named arguments for positional params']() {
      let MyComponent = _helpers.Component.extend();

      this.registerComponent('foo-bar', {
        ComponentClass: MyComponent.reopenClass({
          positionalParams: ['myVar']
        }),
        template: 'MyVar1: {{@myVar}} {{myVar}} MyVar2: {{myVar2}} {{@myVar2}}'
      });

      this.render('{{foo-bar 1 myVar2=2}}');

      this.assertText('MyVar1: 1 1 MyVar2: 2 2');
    }

    ["@test can use `{{this}}` to emit the component's toString value [GH#14581]"]() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          toString() {
            return 'special sauce goes here!';
          }
        }),
        template: '{{this}}'
      });

      this.render('{{foo-bar}}');

      this.assertText('special sauce goes here!');
    }

    ['@test can use `{{this` to access paths on current context [GH#14581]']() {
      let instance;
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);

            instance = this;
          },

          foo: {
            bar: {
              baz: 'huzzah!'
            }
          }
        }),
        template: '{{this.foo.bar.baz}}'
      });

      this.render('{{foo-bar}}');

      this.assertText('huzzah!');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(instance, 'foo.bar.baz', 'yippie!'));

      this.assertText('yippie!');

      this.runTask(() => (0, _metal.set)(instance, 'foo.bar.baz', 'huzzah!'));

      this.assertText('huzzah!');
    }

    ['@test can use custom element in component layout']() {
      this.registerComponent('foo-bar', {
        template: '<blah-zorz>Hi!</blah-zorz>'
      });

      this.render('{{foo-bar}}');

      this.assertText('Hi!');
    }

    ['@test can use nested custom element in component layout']() {
      this.registerComponent('foo-bar', {
        template: '<blah-zorz><hows-it-going>Hi!</hows-it-going></blah-zorz>'
      });

      this.render('{{foo-bar}}');

      this.assertText('Hi!');
    }

    ['@feature(!ember-glimmer-named-arguments) can access properties off of rest style positionalParams array']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'things'
        }),
        // using `attrs` here to simulate `@things.length`
        template: `{{attrs.things.length}}`
      });

      this.render('{{foo-bar "foo" "bar" "baz"}}');

      this.assertText('3');
    }

    ['@feature(ember-glimmer-named-arguments) can access properties off of rest style positionalParams array']() {
      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'things'
        }),
        template: `{{@things.length}}`
      });

      this.render('{{foo-bar "foo" "bar" "baz"}}');

      this.assertText('3');
    }

    ['@test has attrs by didReceiveAttrs with native classes'](assert) {
      class FooBarComponent extends _helpers.Component {
        constructor(injections) {
          super(injections);
          // analagous to class field defaults
          this.foo = 'bar';
        }

        didReceiveAttrs() {
          assert.equal(this.foo, 'bar', 'received default attrs correctly');
        }
      }

      this.registerComponent('foo-bar', { ComponentClass: FooBarComponent });

      this.render('{{foo-bar}}');
    }
  });

  if (_views.jQueryDisabled) {
    (0, _testCase.moduleFor)('Components test: curly components: jQuery disabled', class extends _testCase.RenderingTest {
      ['@test jQuery proxy is not available without jQuery']() {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });

        this.render('{{foo-bar}}');

        expectAssertion(() => {
          instance.$()[0];
        }, 'You cannot access this.$() with `jQuery` disabled.');
      }
    });
  } else {
    (0, _testCase.moduleFor)('Components test: curly components: jQuery enabled', class extends _testCase.RenderingTest {
      ['@test it has a jQuery proxy to the element']() {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });

        this.render('{{foo-bar}}');

        let element1 = instance.$()[0];

        this.assertComponentElement(element1, { content: 'hello' });

        this.runTask(() => this.rerender());

        let element2 = instance.$()[0];

        this.assertComponentElement(element2, { content: 'hello' });

        this.assertSameNode(element2, element1);
      }

      ['@test it scopes the jQuery proxy to the component element'](assert) {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: '<span class="inner">inner</span>'
        });

        this.render('<span class="outer">outer</span>{{foo-bar}}');

        let $span = instance.$('span');

        assert.equal($span.length, 1);
        assert.equal($span.attr('class'), 'inner');

        this.runTask(() => this.rerender());

        $span = instance.$('span');

        assert.equal($span.length, 1);
        assert.equal($span.attr('class'), 'inner');
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/components/destroy-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-case'], function (_metal, _helpers, _testCase) {
  'use strict';

  (0, _testCase.moduleFor)('Component destroy', class extends _testCase.RenderingTest {
    ['@test it correctly releases the destroyed components'](assert) {
      let FooBarComponent = _helpers.Component.extend({});

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{#if switch}}{{#foo-bar}}{{foo-bar}}{{/foo-bar}}{{/if}}', {
        switch: true
      });

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));

      this.assertText('');

      assert.equal(this.env.destroyedComponents.length, 0, 'environment.destroyedComponents should be empty');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/dynamic-components-test', ['@ember/-internals/metal', '@ember/-internals/views', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case'], function (_metal, _views, _helpers, _abstractTestCase, _testCase) {
  'use strict';

  (0, _testCase.moduleFor)('Components test: dynamic components', class extends _testCase.RenderingTest {
    ['@test it can render a basic component with a static component name argument']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component "foo-bar" name=name}}', { name: 'Sarah' });

      this.assertComponentElement(this.firstChild, { content: 'hello Sarah' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello Sarah' });

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Gavin'));

      this.assertComponentElement(this.firstChild, { content: 'hello Gavin' });

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Sarah'));

      this.assertComponentElement(this.firstChild, { content: 'hello Sarah' });
    }

    ['@test it can render a basic component with a dynamic component name argument']() {
      this.registerComponent('foo-bar', {
        template: 'hello {{name}} from foo-bar'
      });
      this.registerComponent('foo-bar-baz', {
        template: 'hello {{name}} from foo-bar-baz'
      });

      this.render('{{component componentName name=name}}', {
        componentName: 'foo-bar',
        name: 'Alex'
      });

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex from foo-bar'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex from foo-bar'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Ben'));

      this.assertComponentElement(this.firstChild, {
        content: 'hello Ben from foo-bar'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'foo-bar-baz'));

      this.assertComponentElement(this.firstChild, {
        content: 'hello Ben from foo-bar-baz'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'componentName', 'foo-bar');
        (0, _metal.set)(this.context, 'name', 'Alex');
      });

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex from foo-bar'
      });
    }

    ['@test it has an element']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          instance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{component "foo-bar"}}');

      let element1 = instance.element;

      this.assertComponentElement(element1, { content: 'hello' });

      this.runTask(() => this.rerender());

      let element2 = instance.element;

      this.assertComponentElement(element2, { content: 'hello' });

      this.assertSameNode(element2, element1);
    }

    ['@test it has the right parentView and childViews'](assert) {
      let fooBarInstance, fooBarBazInstance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
        }
      });

      let FooBarBazComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarBazInstance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'foo-bar {{foo-bar-baz}}'
      });
      this.registerComponent('foo-bar-baz', {
        ComponentClass: FooBarBazComponent,
        template: 'foo-bar-baz'
      });

      this.render('{{component "foo-bar"}}');
      this.assertText('foo-bar foo-bar-baz');

      assert.equal(fooBarInstance.parentView, this.component);
      assert.equal(fooBarBazInstance.parentView, fooBarInstance);

      assert.deepEqual(this.component.childViews, [fooBarInstance]);
      assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);

      this.runTask(() => this.rerender());
      this.assertText('foo-bar foo-bar-baz');

      assert.equal(fooBarInstance.parentView, this.component);
      assert.equal(fooBarBazInstance.parentView, fooBarInstance);

      assert.deepEqual(this.component.childViews, [fooBarInstance]);
      assert.deepEqual(fooBarInstance.childViews, [fooBarBazInstance]);
    }

    ['@test it can render a basic component with a block']() {
      this.registerComponent('foo-bar', { template: '{{yield}}' });

      this.render('{{#component "foo-bar"}}hello{{/component}}');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test it renders the layout with the component instance as the context']() {
      let instance;

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          instance = this;
          this.set('message', 'hello');
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{message}}'
      });

      this.render('{{component "foo-bar"}}');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(instance, 'message', 'goodbye'));

      this.assertComponentElement(this.firstChild, { content: 'goodbye' });

      this.runTask(() => (0, _metal.set)(instance, 'message', 'hello'));

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test it preserves the outer context when yielding']() {
      this.registerComponent('foo-bar', { template: '{{yield}}' });

      this.render('{{#component "foo-bar"}}{{message}}{{/component}}', {
        message: 'hello'
      });

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'goodbye'));

      this.assertComponentElement(this.firstChild, { content: 'goodbye' });

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'hello'));

      this.assertComponentElement(this.firstChild, { content: 'hello' });
    }

    ['@test the component and its child components are destroyed'](assert) {
      let destroyed = { 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0 };

      this.registerComponent('foo-bar', {
        template: '{{id}} {{yield}}',
        ComponentClass: _helpers.Component.extend({
          willDestroy() {
            this._super();
            destroyed[this.get('id')]++;
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#if cond1}}
        {{#component "foo-bar" id=1}}
          {{#if cond2}}
            {{#component "foo-bar" id=2}}{{/component}}
            {{#if cond3}}
              {{#component "foo-bar" id=3}}
                {{#if cond4}}
                  {{#component "foo-bar" id=4}}
                    {{#if cond5}}
                      {{#component "foo-bar" id=5}}{{/component}}
                      {{#component "foo-bar" id=6}}{{/component}}
                      {{#component "foo-bar" id=7}}{{/component}}
                    {{/if}}
                    {{#component "foo-bar" id=8}}{{/component}}
                  {{/component}}
                {{/if}}
              {{/component}}
            {{/if}}
          {{/if}}
        {{/component}}
      {{/if}}`, {
        cond1: true,
        cond2: true,
        cond3: true,
        cond4: true,
        cond5: true
      });

      this.assertText('1 2 3 4 5 6 7 8 ');

      this.runTask(() => this.rerender());

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 0,
        6: 0,
        7: 0,
        8: 0
      });

      this.runTask(() => (0, _metal.set)(this.context, 'cond5', false));

      this.assertText('1 2 3 4 8 ');

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 1,
        6: 1,
        7: 1,
        8: 0
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond3', false);
        (0, _metal.set)(this.context, 'cond5', true);
        (0, _metal.set)(this.context, 'cond4', false);
      });

      assert.deepEqual(destroyed, {
        1: 0,
        2: 0,
        3: 1,
        4: 1,
        5: 1,
        6: 1,
        7: 1,
        8: 1
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond2', false);
        (0, _metal.set)(this.context, 'cond1', false);
      });

      assert.deepEqual(destroyed, {
        1: 1,
        2: 1,
        3: 1,
        4: 1,
        5: 1,
        6: 1,
        7: 1,
        8: 1
      });
    }

    ['@test component helper destroys underlying component when it is swapped out'](assert) {
      let destroyed = { 'foo-bar': 0, 'foo-bar-baz': 0 };
      let testContext = this;

      this.registerComponent('foo-bar', {
        template: 'hello from foo-bar',
        ComponentClass: _helpers.Component.extend({
          willDestroyElement() {
            assert.equal(testContext.$(`#${this.elementId}`).length, 1, 'element is still attached to the document');
          },

          willDestroy() {
            this._super();
            destroyed['foo-bar']++;
          }
        })
      });

      this.registerComponent('foo-bar-baz', {
        template: 'hello from foo-bar-baz',
        ComponentClass: _helpers.Component.extend({
          willDestroy() {
            this._super();
            destroyed['foo-bar-baz']++;
          }
        })
      });

      this.render('{{component componentName name=name}}', {
        componentName: 'foo-bar'
      });

      assert.deepEqual(destroyed, { 'foo-bar': 0, 'foo-bar-baz': 0 });

      this.runTask(() => this.rerender());

      assert.deepEqual(destroyed, { 'foo-bar': 0, 'foo-bar-baz': 0 });

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'foo-bar-baz'));

      assert.deepEqual(destroyed, { 'foo-bar': 1, 'foo-bar-baz': 0 });

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'foo-bar'));

      assert.deepEqual(destroyed, { 'foo-bar': 1, 'foo-bar-baz': 1 });
    }

    ['@test component helper with bound properties are updating correctly in init of component']() {
      this.registerComponent('foo-bar', {
        template: 'foo-bar {{location}} {{locationCopy}} {{yield}}',
        ComponentClass: _helpers.Component.extend({
          init: function () {
            this._super(...arguments);
            this.set('locationCopy', this.get('location'));
          }
        })
      });

      this.registerComponent('foo-bar-baz', {
        template: 'foo-bar-baz {{location}} {{locationCopy}} {{yield}}',
        ComponentClass: _helpers.Component.extend({
          init: function () {
            this._super(...arguments);
            this.set('locationCopy', this.get('location'));
          }
        })
      });

      this.registerComponent('outer-component', {
        template: '{{#component componentName location=location}}arepas!{{/component}}',
        ComponentClass: _helpers.Component.extend({
          componentName: (0, _metal.computed)('location', function () {
            if (this.get('location') === 'Caracas') {
              return 'foo-bar';
            } else {
              return 'foo-bar-baz';
            }
          })
        })
      });

      this.render('{{outer-component location=location}}', {
        location: 'Caracas'
      });

      this.assertText('foo-bar Caracas Caracas arepas!');

      this.runTask(() => this.rerender());

      this.assertText('foo-bar Caracas Caracas arepas!');

      this.runTask(() => (0, _metal.set)(this.context, 'location', 'Loisaida'));

      this.assertText('foo-bar-baz Loisaida Loisaida arepas!');

      this.runTask(() => (0, _metal.set)(this.context, 'location', 'Caracas'));

      this.assertText('foo-bar Caracas Caracas arepas!');
    }

    ['@test component helper with actions'](assert) {
      this.registerComponent('inner-component', {
        template: 'inner-component {{yield}}',
        ComponentClass: _helpers.Component.extend({
          classNames: 'inner-component',
          didInsertElement() {
            // trigger action on click in absence of app's EventDispatcher
            let sendAction = this.eventHandler = () => {
              if (this.somethingClicked) {
                this.somethingClicked();
              }
            };
            this.element.addEventListener('click', sendAction);
          },
          willDestroyElement() {
            this.element.removeEventListener('click', this.eventHandler);
          }
        })
      });

      let actionTriggered = 0;
      this.registerComponent('outer-component', {
        template: '{{#component componentName somethingClicked=(action "mappedAction")}}arepas!{{/component}}',
        ComponentClass: _helpers.Component.extend({
          classNames: 'outer-component',
          componentName: 'inner-component',
          actions: {
            mappedAction() {
              actionTriggered++;
            }
          }
        })
      });

      this.render('{{outer-component}}');

      assert.equal(actionTriggered, 0, 'action was not triggered');

      this.runTask(() => {
        this.$('.inner-component').click();
      });

      assert.equal(actionTriggered, 1, 'action was triggered');
    }

    ['@test nested component helpers']() {
      this.registerComponent('foo-bar', {
        template: 'yippie! {{attrs.location}} {{yield}}'
      });
      this.registerComponent('baz-qux', {
        template: 'yummy {{attrs.location}} {{yield}}'
      });
      this.registerComponent('corge-grault', {
        template: 'delicious {{attrs.location}} {{yield}}'
      });

      this.render('{{#component componentName1 location=location}}{{#component componentName2 location=location}}arepas!{{/component}}{{/component}}', {
        componentName1: 'foo-bar',
        componentName2: 'baz-qux',
        location: 'Caracas'
      });

      this.assertText('yippie! Caracas yummy Caracas arepas!');

      this.runTask(() => this.rerender());

      this.assertText('yippie! Caracas yummy Caracas arepas!');

      this.runTask(() => (0, _metal.set)(this.context, 'location', 'Loisaida'));

      this.assertText('yippie! Loisaida yummy Loisaida arepas!');

      this.runTask(() => (0, _metal.set)(this.context, 'componentName1', 'corge-grault'));

      this.assertText('delicious Loisaida yummy Loisaida arepas!');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'componentName1', 'foo-bar');
        (0, _metal.set)(this.context, 'location', 'Caracas');
      });

      this.assertText('yippie! Caracas yummy Caracas arepas!');
    }

    ['@test component with dynamic name argument resolving to non-existent component']() {
      expectAssertion(() => {
        this.render('{{component componentName}}', {
          componentName: 'does-not-exist'
        });
      }, /Could not find component named "does-not-exist"/);
    }

    ['@test component with static name argument for non-existent component']() {
      expectAssertion(() => {
        this.render('{{component "does-not-exist"}}');
      }, /Could not find component named "does-not-exist"/);
    }

    ['@test component with dynamic component name resolving to a component, then non-existent component']() {
      this.registerComponent('foo-bar', { template: 'hello {{name}}' });

      this.render('{{component componentName name=name}}', {
        componentName: 'foo-bar',
        name: 'Alex'
      });

      this.assertText('hello Alex');

      this.runTask(() => this.rerender());

      this.assertText('hello Alex');

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', undefined));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'foo-bar'));

      this.assertText('hello Alex');
    }

    ['@test component helper properly invalidates hash params inside an {{each}} invocation #11044']() {
      this.registerComponent('foo-bar', {
        template: '[{{internalName}} - {{name}}]',
        ComponentClass: _helpers.Component.extend({
          willRender() {
            // store internally available name to ensure that the name available in `this.attrs.name`
            // matches the template lookup name
            (0, _metal.set)(this, 'internalName', this.get('name'));
          }
        })
      });

      this.render('{{#each items as |item|}}{{component "foo-bar" name=item.name}}{{/each}}', {
        items: [{ name: 'Robert' }, { name: 'Jacquie' }]
      });

      this.assertText('[Robert - Robert][Jacquie - Jacquie]');

      this.runTask(() => this.rerender());

      this.assertText('[Robert - Robert][Jacquie - Jacquie]');

      this.runTask(() => (0, _metal.set)(this.context, 'items', [{ name: 'Max' }, { name: 'James' }]));

      this.assertText('[Max - Max][James - James]');

      this.runTask(() => (0, _metal.set)(this.context, 'items', [{ name: 'Robert' }, { name: 'Jacquie' }]));

      this.assertText('[Robert - Robert][Jacquie - Jacquie]');
    }

    ['@test dashless components should not be found'](assert) {
      if (true /* EMBER_GLIMMER_ANGLE_BRACKET_INVOCATION */) {
          assert.ok(true, 'test is not applicable');
          return;
        }

      this.registerComponent('dashless2', { template: 'Do not render me!' });

      expectAssertion(() => {
        this.render('{{component "dashless"}}');
      }, /You cannot use 'dashless' as a component name. Component names must contain a hyphen./);
    }

    ['@test positional parameters does not clash when rendering different components']() {
      this.registerComponent('foo-bar', {
        template: 'hello {{name}} ({{age}}) from foo-bar',
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        })
      });

      this.registerComponent('foo-bar-baz', {
        template: 'hello {{name}} ({{age}}) from foo-bar-baz',
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['name', 'age']
        })
      });

      this.render('{{component componentName name age}}', {
        componentName: 'foo-bar',
        name: 'Alex',
        age: 29
      });

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex (29) from foo-bar'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex (29) from foo-bar'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Ben'));

      this.assertComponentElement(this.firstChild, {
        content: 'hello Ben (29) from foo-bar'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'age', 22));

      this.assertComponentElement(this.firstChild, {
        content: 'hello Ben (22) from foo-bar'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'foo-bar-baz'));

      this.assertComponentElement(this.firstChild, {
        content: 'hello Ben (22) from foo-bar-baz'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'componentName', 'foo-bar');
        (0, _metal.set)(this.context, 'name', 'Alex');
        (0, _metal.set)(this.context, 'age', 29);
      });

      this.assertComponentElement(this.firstChild, {
        content: 'hello Alex (29) from foo-bar'
      });
    }

    ['@test positional parameters does not pollute the attributes when changing components']() {
      this.registerComponent('normal-message', {
        template: 'Normal: {{something}}!',
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: ['something']
        })
      });

      this.registerComponent('alternative-message', {
        template: 'Alternative: {{something}} {{somethingElse}}!',
        ComponentClass: _helpers.Component.extend({
          something: 'Another'
        }).reopenClass({
          positionalParams: ['somethingElse']
        })
      });

      this.render('{{component componentName message}}', {
        componentName: 'normal-message',
        message: 'Hello'
      });

      this.assertComponentElement(this.firstChild, {
        content: 'Normal: Hello!'
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        content: 'Normal: Hello!'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'alternative-message'));

      this.assertComponentElement(this.firstChild, {
        content: 'Alternative: Another Hello!'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'Hi'));

      this.assertComponentElement(this.firstChild, {
        content: 'Alternative: Another Hi!'
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'componentName', 'normal-message');
        (0, _metal.set)(this.context, 'message', 'Hello');
      });

      this.assertComponentElement(this.firstChild, {
        content: 'Normal: Hello!'
      });
    }

    ['@test static arbitrary number of positional parameters']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'names'
        }),
        template: _abstractTestCase.strip`
        {{#each names as |name|}}
          {{name}}
        {{/each}}`
      });

      this.render(`{{component "sample-component" "Foo" 4 "Bar" 5 "Baz" elementId="helper"}}`);

      this.assertText('Foo4Bar5Baz');

      this.runTask(() => this.rerender());

      this.assertText('Foo4Bar5Baz');
    }

    ['@test dynamic arbitrary number of positional parameters']() {
      this.registerComponent('sample-component', {
        ComponentClass: _helpers.Component.extend().reopenClass({
          positionalParams: 'n'
        }),
        template: _abstractTestCase.strip`
        {{#each n as |name|}}
          {{name}}
        {{/each}}`
      });

      this.render(`{{component "sample-component" user1 user2}}`, {
        user1: 'Foo',
        user2: 4
      });

      this.assertText('Foo4');

      this.runTask(() => this.rerender());

      this.assertText('Foo4');

      this.runTask(() => this.context.set('user1', 'Bar'));

      this.assertText('Bar4');

      this.runTask(() => this.context.set('user2', '5'));

      this.assertText('Bar5');

      this.runTask(() => {
        this.context.set('user1', 'Foo');
        this.context.set('user2', 4);
      });

      this.assertText('Foo4');
    }

    ['@test component helper emits useful backtracking re-render assertion message']() {
      this.registerComponent('outer-component', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            this.set('person', { name: 'Alex' });
          }
        }),
        template: `Hi {{person.name}}! {{component "error-component" person=person}}`
      });

      this.registerComponent('error-component', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            this.set('person.name', { name: 'Ben' });
          }
        }),
        template: '{{person.name}}'
      });

      let expectedBacktrackingMessage = /modified "person\.name" twice on \[object Object\] in a single render\. It was rendered in "component:outer-component" and modified in "component:error-component"/;

      expectAssertion(() => {
        this.render('{{component componentName}}', {
          componentName: 'outer-component'
        });
      }, expectedBacktrackingMessage);
    }
  });

  if (_views.jQueryDisabled) {
    (0, _testCase.moduleFor)('Components test: dynamic components: jQuery disabled', class extends _testCase.RenderingTest {
      ['@test jQuery proxy is not available without jQuery']() {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });

        this.render('{{component "foo-bar"}}');

        expectAssertion(() => {
          instance.$()[0];
        }, 'You cannot access this.$() with `jQuery` disabled.');
      }
    });
  } else {
    (0, _testCase.moduleFor)('Components test: dynamic components : jQuery enabled', class extends _testCase.RenderingTest {
      ['@test it has a jQuery proxy to the element']() {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });

        this.render('{{component "foo-bar"}}');

        let element1 = instance.$()[0];

        this.assertComponentElement(element1, { content: 'hello' });

        this.runTask(() => this.rerender());

        let element2 = instance.$()[0];

        this.assertComponentElement(element2, { content: 'hello' });

        this.assertSameNode(element2, element1);
      }

      ['@test it scopes the jQuery proxy to the component element'](assert) {
        let instance;

        let FooBarComponent = _helpers.Component.extend({
          init() {
            this._super();
            instance = this;
          }
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: '<span class="inner">inner</span>'
        });

        this.render('<span class="outer">outer</span>{{component "foo-bar"}}');

        let $span = instance.$('span');

        assert.equal($span.length, 1);
        assert.equal($span.attr('class'), 'inner');

        this.runTask(() => this.rerender());

        $span = instance.$('span');

        assert.equal($span.length, 1);
        assert.equal($span.attr('class'), 'inner');
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/components/error-handling-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-case'], function (_metal, _helpers, _testCase) {
  'use strict';

  (0, _testCase.moduleFor)('Errors thrown during render', class extends _testCase.RenderingTest {
    ['@test it can recover resets the transaction when an error is thrown during initial render'](assert) {
      let shouldThrow = true;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          if (shouldThrow) {
            throw new Error('silly mistake in init!');
          }
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      assert.throws(() => {
        this.render('{{#if switch}}{{#foo-bar}}{{foo-bar}}{{/foo-bar}}{{/if}}', { switch: true });
      }, /silly mistake in init/);

      assert.equal(this.env.inTransaction, false, 'should not be in a transaction even though an error was thrown');

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));

      shouldThrow = false;

      this.runTask(() => (0, _metal.set)(this.context, 'switch', true));

      this.assertText('hello');
    }

    ['@test it can recover resets the transaction when an error is thrown during rerender'](assert) {
      let shouldThrow = false;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          if (shouldThrow) {
            throw new Error('silly mistake in init!');
          }
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{#if switch}}{{#foo-bar}}{{foo-bar}}{{/foo-bar}}{{/if}}', {
        switch: true
      });

      this.assertText('hello');

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));

      shouldThrow = true;

      assert.throws(() => {
        this.runTask(() => (0, _metal.set)(this.context, 'switch', true));
      }, /silly mistake in init/);

      assert.equal(this.env.inTransaction, false, 'should not be in a transaction even though an error was thrown');

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));
      shouldThrow = false;

      this.runTask(() => (0, _metal.set)(this.context, 'switch', true));

      this.assertText('hello');
    }

    ['@test it can recover resets the transaction when an error is thrown during didInsertElement'](assert) {
      let shouldThrow = true;
      let FooBarComponent = _helpers.Component.extend({
        didInsertElement() {
          this._super(...arguments);
          if (shouldThrow) {
            throw new Error('silly mistake!');
          }
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      assert.throws(() => {
        this.render('{{#if switch}}{{#foo-bar}}{{foo-bar}}{{/foo-bar}}{{/if}}', { switch: true });
      }, /silly mistake/);

      assert.equal(this.env.inTransaction, false, 'should not be in a transaction even though an error was thrown');

      this.assertText('hello');

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));

      this.assertText('');
    }

    ['@test it can recover resets the transaction when an error is thrown during destroy'](assert) {
      let shouldThrow = true;
      let FooBarComponent = _helpers.Component.extend({
        destroy() {
          this._super(...arguments);
          if (shouldThrow) {
            throw new Error('silly mistake!');
          }
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{#if switch}}{{#foo-bar}}{{foo-bar}}{{/foo-bar}}{{/if}}', {
        switch: true
      });

      this.assertText('hello');

      assert.throws(() => {
        this.runTask(() => (0, _metal.set)(this.context, 'switch', false));
      }, /silly mistake/);

      this.assertText('');

      shouldThrow = false;
      this.runTask(() => (0, _metal.set)(this.context, 'switch', true));

      this.assertText('hello');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/fragment-components-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal'], function (_testCase, _abstractTestCase, _helpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Components test: fragment components', class extends _testCase.RenderingTest {
    getCustomDispatcherEvents() {
      return {
        hitDem: 'folks'
      };
    }

    ['@test fragments do not render an outer tag']() {
      let instance;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        init() {
          this._super();
          instance = this;
          this.foo = true;
          this.bar = 'bar';
        }
      });

      let template = `{{#if foo}}<div>Hey</div>{{/if}}{{yield bar}}`;

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      this.render(`{{#foo-bar as |bar|}}{{bar}}{{/foo-bar}}`);

      this.assertHTML(_abstractTestCase.strip`<div>Hey</div>bar`);

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(instance, 'foo', false));

      this.assertHTML(_abstractTestCase.strip`<!---->bar`);

      this.runTask(() => (0, _metal.set)(instance, 'bar', 'bizz'));

      this.assertHTML(_abstractTestCase.strip`<!---->bizz`);

      this.runTask(() => {
        (0, _metal.set)(instance, 'bar', 'bar');
        (0, _metal.set)(instance, 'foo', true);
      });
    }

    ['@test throws an error if an event function is defined in a tagless component']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        click() {}
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You can not define a function that handles DOM events in the .* tagless component since it doesn't have any DOM element./);
    }

    ['@test throws an error if a custom defined event function is defined in a tagless component']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        folks() {}
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You can not define a function that handles DOM events in the .* tagless component since it doesn't have any DOM element./);
    }

    ['@test throws an error if `tagName` is an empty string and `classNameBindings` are specified']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        foo: true,
        classNameBindings: ['foo:is-foo:is-bar']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You cannot use `classNameBindings` on a tag-less component/);
    }

    ['@test throws an error if `tagName` is an empty string and `attributeBindings` are specified']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        attributeBindings: ['href']
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });
      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You cannot use `attributeBindings` on a tag-less component/);
    }

    ['@test throws an error if `tagName` is an empty string and `elementId` is specified via JS']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        elementId: 'turntUp'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });
      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You cannot use `elementId` on a tag-less component/);
    }

    ['@test throws an error if `tagName` is an empty string and `elementId` is specified via template']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: ''
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });
      expectAssertion(() => {
        this.render(`{{#foo-bar elementId='turntUp'}}{{/foo-bar}}`);
      }, /You cannot use `elementId` on a tag-less component/);
    }

    ['@test does not throw an error if `tagName` is an empty string and `id` is specified via JS']() {
      let template = `{{id}}`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        id: 'baz'
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });
      this.render(`{{#foo-bar}}{{/foo-bar}}`);
      this.assertText('baz');
    }

    ['@test does not throw an error if `tagName` is an empty string and `id` is specified via template']() {
      let template = `{{id}}`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: ''
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });
      this.render(`{{#foo-bar id='baz'}}{{/foo-bar}}`);
      this.assertText('baz');
    }

    ['@test does not throw an error if `tagName` is an empty string and `id` is bound property specified via template']() {
      let template = `{{id}}`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: ''
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      this.render(`{{#foo-bar id=fooBarId}}{{/foo-bar}}`, { fooBarId: 'baz' });

      this.assertText('baz');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'fooBarId', 'qux'));

      this.assertText('qux');

      this.runTask(() => (0, _metal.set)(this.context, 'fooBarId', 'baz'));

      this.assertText('baz');
    }

    ['@test does not throw an error if `tagName` is an empty string and `id` is specified via template and passed to child component']() {
      let fooBarTemplate = `{{#baz-child id=id}}{{/baz-child}}`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: ''
      });
      let BazChildComponent = _helpers.Component.extend();
      let bazChildTemplate = `{{id}}`;

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: fooBarTemplate
      });
      this.registerComponent('baz-child', {
        ComponentClass: BazChildComponent,
        template: bazChildTemplate
      });
      this.render(`{{#foo-bar id='baz'}}{{/foo-bar}}`);
      this.assertText('baz');
    }

    ['@test throws an error if when $() is accessed on component where `tagName` is an empty string']() {
      let template = `hit dem folks`;
      let FooBarComponent = _helpers.Component.extend({
        tagName: '',
        init() {
          this._super();
          this.$();
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template
      });

      expectAssertion(() => {
        this.render(`{{#foo-bar}}{{/foo-bar}}`);
      }, /You cannot access this.\$\(\) on a component with `tagName: \'\'` specified/);
    }

    ['@test renders a contained view with omitted start tag and tagless parent view context']() {
      this.registerComponent('root-component', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'section'
        }),
        template: '{{frag-ment}}'
      });

      this.registerComponent('frag-ment', {
        ComponentClass: _helpers.Component.extend({
          tagName: ''
        }),
        template: '{{my-span}}'
      });

      this.registerComponent('my-span', {
        ComponentClass: _helpers.Component.extend({
          tagName: 'span'
        }),
        template: 'dab'
      });

      this.render(`{{root-component}}`);

      this.assertElement(this.firstChild, { tagName: 'section' });
      this.assertElement(this.firstChild.firstElementChild, {
        tagName: 'span'
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, { tagName: 'section' });
      this.assertElement(this.firstChild.firstElementChild, {
        tagName: 'span'
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/instrumentation-compile-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/instrumentation'], function (_testCase, _helpers, _instrumentation) {
  'use strict';

  (0, _testCase.moduleFor)('Components compile instrumentation', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);

      this.resetEvents();

      (0, _instrumentation.subscribe)('render.getComponentDefinition', {
        before: (name, timestamp, payload) => {
          if (payload.view !== this.component) {
            this.actual.before.push(payload);
          }
        },
        after: (name, timestamp, payload) => {
          if (payload.view !== this.component) {
            this.actual.after.push(payload);
          }
        }
      });
    }

    resetEvents() {
      this.expected = {
        before: [],
        after: []
      };

      this.actual = {
        before: [],
        after: []
      };
    }

    teardown() {
      this.assert.deepEqual(this.actual.before, [], 'No unexpected events (before)');
      this.assert.deepEqual(this.actual.after, [], 'No unexpected events (after)');
      super.teardown();
      (0, _instrumentation.reset)();
    }

    ['@test it should only receive an instrumentation event for initial render']() {
      let testCase = this;

      let BaseClass = _helpers.Component.extend({
        tagName: '',

        willRender() {
          testCase.expected.before.push(this);
          testCase.expected.after.unshift(this);
        }
      });

      this.registerComponent('x-bar', {
        template: '[x-bar: {{bar}}]',
        ComponentClass: BaseClass.extend()
      });

      this.render(`[-top-level: {{foo}}] {{x-bar bar=bar}}`, {
        foo: 'foo',
        bar: 'bar'
      });

      this.assertText('[-top-level: foo] [x-bar: bar]');

      this.assertEvents('after initial render');

      this.runTask(() => this.rerender());

      this.assertEvents('after no-op rerender');
    }

    assertEvents(label) {
      let { actual, expected } = this;
      this.assert.strictEqual(actual.before.length, actual.after.length, `${label}: before and after callbacks should be balanced`);

      this._assertEvents(`${label} (before):`, actual.before, expected.before);
      this._assertEvents(`${label} (after):`, actual.before, expected.before);

      this.resetEvents();
    }

    _assertEvents(label, actual, expected) {
      this.assert.equal(actual.length, expected.length, `${label}: expected ${expected.length} and got ${actual.length}`);

      actual.forEach((payload, i) => this.assertPayload(payload, expected[i]));
    }

    assertPayload(payload, component) {
      this.assert.equal(payload.object, component._debugContainerKey, 'payload.object');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/instrumentation-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal', '@ember/instrumentation'], function (_testCase, _helpers, _metal, _instrumentation) {
  'use strict';

  (0, _testCase.moduleFor)('Components instrumentation', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);

      this.resetEvents();

      (0, _instrumentation.subscribe)('render.component', {
        before: (name, timestamp, payload) => {
          if (payload.view !== this.component) {
            this.actual.before.push(payload);
          }
        },
        after: (name, timestamp, payload) => {
          if (payload.view !== this.component) {
            this.actual.after.push(payload);
          }
        }
      });
    }

    resetEvents() {
      this.expected = {
        before: [],
        after: []
      };

      this.actual = {
        before: [],
        after: []
      };
    }

    teardown() {
      this.assert.deepEqual(this.actual.before, [], 'No unexpected events (before)');
      this.assert.deepEqual(this.actual.after, [], 'No unexpected events (after)');
      super.teardown();
      (0, _instrumentation.reset)();
    }

    ['@test zomg'](assert) {
      assert.ok(true);
    }

    ['@test it should receive an instrumentation event for both initial render and updates']() {
      let testCase = this;

      let BaseClass = _helpers.Component.extend({
        tagName: '',

        willRender() {
          testCase.expected.before.push(this);
          testCase.expected.after.unshift(this);
        }
      });

      this.registerComponent('x-bar', {
        template: '[x-bar: {{bar}}] {{yield}}',
        ComponentClass: BaseClass.extend()
      });

      this.registerComponent('x-baz', {
        template: '[x-baz: {{baz}}]',
        ComponentClass: BaseClass.extend()
      });

      this.registerComponent('x-bat', {
        template: '[x-bat: {{bat}}]',
        ComponentClass: BaseClass.extend()
      });

      this.render(`[-top-level: {{foo}}] {{#x-bar bar=bar}}{{x-baz baz=baz}}{{/x-bar}} {{x-bat bat=bat}}`, {
        foo: 'foo',
        bar: 'bar',
        baz: 'baz',
        bat: 'bat'
      });

      this.assertText('[-top-level: foo] [x-bar: bar] [x-baz: baz] [x-bat: bat]');

      this.assertEvents('after initial render', true);

      this.runTask(() => this.rerender());

      this.assertEvents('after no-op rerender');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'FOO'));

      this.assertEvents('after updating top-level');

      this.runTask(() => (0, _metal.set)(this.context, 'baz', 'BAZ'));

      this.assertEvents('after updating inner-most');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'bar', 'BAR');
        (0, _metal.set)(this.context, 'bat', 'BAT');
      });

      this.assertEvents('after updating the rest');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'FOO');
        (0, _metal.set)(this.context, 'bar', 'BAR');
        (0, _metal.set)(this.context, 'baz', 'BAZ');
        (0, _metal.set)(this.context, 'bat', 'BAT');
      });

      this.assertEvents('after reset');
    }

    assertEvents(label, initialRender = false) {
      let { actual, expected } = this;

      this.assert.strictEqual(actual.before.length, actual.after.length, `${label}: before and after callbacks should be balanced`);

      this._assertEvents(`${label} (before):`, actual.before, expected.before, initialRender);
      this._assertEvents(`${label} (after):`, actual.before, expected.before, initialRender);

      this.resetEvents();
    }

    _assertEvents(label, actual, expected, initialRender) {
      this.assert.equal(actual.length, expected.length, `${label}: expected ${expected.length} and got ${actual.length}`);

      actual.forEach((payload, i) => this.assertPayload(payload, expected[i], initialRender));
    }

    assertPayload(payload, component, initialRender) {
      this.assert.equal(payload.object, component.toString(), 'payload.object');
      this.assert.ok(payload.containerKey, 'the container key should be present');
      this.assert.equal(payload.containerKey, component._debugContainerKey, 'payload.containerKey');
      this.assert.equal(payload.view, component, 'payload.view');
      this.assert.strictEqual(payload.initialRender, initialRender, 'payload.initialRender');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/life-cycle-test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/views', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/utils', 'internal-test-helpers'], function (_runloop, _metal, _runtime, _helpers, _abstractTestCase, _testCase, _views, _testHelpers, _utils, _internalTestHelpers) {
  'use strict';

  class LifeCycleHooksTest extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);
      this.hooks = [];
      this.components = {};
      this.componentRegistry = [];
      this.teardownAssertions = [];
      this.viewRegistry = this.owner.lookup('-view-registry:main');
    }

    afterEach() {
      super.afterEach();

      for (let i = 0; i < this.teardownAssertions.length; i++) {
        this.teardownAssertions[i]();
      }
    }

    get isInteractive() {
      return true;
    }

    getBootOptions() {
      return {
        isInteractive: this.isInteractive
      };
    }

    /* abstract */
    get ComponentClass() {
      throw new Error('Not implemented: `ComponentClass`');
    }

    /* abstract */
    invocationFor() /* name, namedArgs = {} */{
      throw new Error('Not implemented: `invocationFor`');
    }

    /* abstract */
    attrFor() /* name */{
      throw new Error('Not implemented: `attrFor`');
    }

    get boundHelpers() {
      return {
        invoke: bind(this.invocationFor, this),
        attr: bind(this.attrFor, this)
      };
    }

    assertRegisteredViews(label) {
      let viewRegistry = this.viewRegistry;
      let topLevelId = (0, _views.getViewId)(this.component);
      let actual = Object.keys(viewRegistry).sort().filter(id => id !== topLevelId);

      if (this.isInteractive) {
        let expected = this.componentRegistry.sort();

        this.assert.deepEqual(actual, expected, 'registered views - ' + label);
      } else {
        this.assert.deepEqual(actual, [], 'no views should be registered for non-interactive mode');
      }
    }

    registerComponent(name, { template = null }) {
      let pushComponent = instance => {
        this.components[name] = instance;
        this.componentRegistry.push((0, _views.getViewId)(instance));
      };

      let removeComponent = instance => {
        let index = this.componentRegistry.indexOf(instance);
        this.componentRegistry.splice(index, 1);

        delete this.components[name];
      };

      let pushHook = (hookName, args) => {
        this.hooks.push(hook(name, hookName, args));
      };

      let assertParentView = (hookName, instance) => {
        this.assert.ok(instance.parentView, `parentView should be present in ${hookName}`);

        if (hookName === 'willDestroyElement') {
          this.assert.ok(instance.parentView.childViews.indexOf(instance) !== -1, `view is still connected to parentView in ${hookName}`);
        }
      };

      let assertElement = (hookName, instance, inDOM = true) => {
        if (instance.tagName === '') {
          return;
        }

        this.assert.ok((0, _views.getViewElement)(instance), `element should be present on ${instance} during ${hookName}`);

        if (this.isInteractive) {
          this.assert.ok(instance.element, `this.element should be present on ${instance} during ${hookName}`);
          this.assert.equal(document.body.contains(instance.element), inDOM, `element for ${instance} ${inDOM ? 'should' : 'should not'} be in the DOM during ${hookName}`);
        } else {
          this.assert.throws(() => instance.element, /Accessing `this.element` is not allowed in non-interactive environments/);
        }
      };

      let assertNoElement = (hookName, instance) => {
        this.assert.strictEqual((0, _views.getViewElement)(instance), null, `element should not be present in ${hookName}`);

        if (this.isInteractive) {
          this.assert.strictEqual(instance.element, null, `this.element should not be present in ${hookName}`);
        } else {
          this.assert.throws(() => instance.element, /Accessing `this.element` is not allowed in non-interactive environments/);
        }
      };

      let assertState = (hookName, expectedState, instance) => {
        this.assert.equal(instance._state, expectedState, `within ${hookName} the expected _state is ${expectedState}`);
      };

      let { isInteractive } = this;

      let ComponentClass = this.ComponentClass.extend({
        init() {
          this._super(...arguments);

          this.isInitialRender = true;
          this.componentName = name;
          pushHook('init');
          pushComponent(this);
          assertParentView('init', this);
          assertNoElement('init', this);
          assertState('init', 'preRender', this);

          this.on('init', () => pushHook('on(init)'));

          (0, _runloop.schedule)('afterRender', () => {
            this.isInitialRender = false;
          });
        },

        didReceiveAttrs(options) {
          pushHook('didReceiveAttrs', options);
          assertParentView('didReceiveAttrs', this);

          if (this.isInitialRender) {
            assertNoElement('didReceiveAttrs', this);
            assertState('didReceiveAttrs', 'preRender', this);
          } else {
            assertElement('didReceiveAttrs', this);

            if (isInteractive) {
              assertState('didReceiveAttrs', 'inDOM', this);
            } else {
              assertState('didReceiveAttrs', 'hasElement', this);
            }
          }
        },

        willInsertElement() {
          pushHook('willInsertElement');
          assertParentView('willInsertElement', this);
          assertElement('willInsertElement', this, false);
          assertState('willInsertElement', 'hasElement', this);
        },

        willRender() {
          pushHook('willRender');
          assertParentView('willRender', this);

          if (this.isInitialRender) {
            assertNoElement('willRender', this, false);
            assertState('willRender', 'preRender', this);
          } else {
            assertElement('willRender', this);
            assertState('willRender', 'inDOM', this);
          }
        },

        didInsertElement() {
          pushHook('didInsertElement');
          assertParentView('didInsertElement', this);
          assertElement('didInsertElement', this);
          assertState('didInsertElement', 'inDOM', this);
        },

        didRender() {
          pushHook('didRender');
          assertParentView('didRender', this);
          assertElement('didRender', this);
          assertState('didRender', 'inDOM', this);
        },

        didUpdateAttrs(options) {
          pushHook('didUpdateAttrs', options);
          assertParentView('didUpdateAttrs', this);

          if (isInteractive) {
            assertState('didUpdateAttrs', 'inDOM', this);
          } else {
            assertState('didUpdateAttrs', 'hasElement', this);
          }
        },

        willUpdate(options) {
          pushHook('willUpdate', options);
          assertParentView('willUpdate', this);
          assertElement('willUpdate', this);
          assertState('willUpdate', 'inDOM', this);
        },

        didUpdate(options) {
          pushHook('didUpdate', options);
          assertParentView('didUpdate', this);
          assertElement('didUpdate', this);
          assertState('didUpdate', 'inDOM', this);
        },

        willDestroyElement() {
          pushHook('willDestroyElement');
          assertParentView('willDestroyElement', this);
          assertElement('willDestroyElement', this);
          assertState('willDestroyElement', 'inDOM', this);
        },

        willClearRender() {
          pushHook('willClearRender');
          assertParentView('willClearRender', this);
          assertElement('willClearRender', this);
          assertState('willClearRender', 'inDOM', this);
        },

        didDestroyElement() {
          pushHook('didDestroyElement');
          assertNoElement('didDestroyElement', this);
          assertState('didDestroyElement', 'destroying', this);
        },

        willDestroy() {
          pushHook('willDestroy');
          removeComponent(this);

          this._super(...arguments);
        }
      });

      super.registerComponent(name, { ComponentClass, template });
    }

    assertHooks({ label, interactive, nonInteractive }) {
      let rawHooks = this.isInteractive ? interactive : nonInteractive;
      let hooks = rawHooks.map(raw => hook(...raw));
      this.assert.deepEqual(json(this.hooks), json(hooks), label);
      this.hooks = [];
    }

    ['@test lifecycle hooks are invoked in a predictable order']() {
      let { attr, invoke } = this.boundHelpers;

      this.registerComponent('the-top', {
        template: _abstractTestCase.strip`
      <div>
        Twitter: {{${attr('twitter')}}}|
        ${invoke('the-middle', { name: string('Tom Dale') })}
      </div>`
      });

      this.registerComponent('the-middle', {
        template: _abstractTestCase.strip`
      <div>
        Name: {{${attr('name')}}}|
        ${invoke('the-bottom', { website: string('tomdale.net') })}
      </div>`
      });

      this.registerComponent('the-bottom', {
        template: _abstractTestCase.strip`
      <div>
        Website: {{${attr('website')}}}
      </div>`
      });

      this.render(invoke('the-top', { twitter: expr('twitter') }), {
        twitter: '@tomdale'
      });

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');
      this.assertRegisteredViews('intial render');

      this.assertHooks({
        label: 'after initial render',

        interactive: [
        // Sync hooks

        ['the-top', 'init'], ['the-top', 'on(init)'], ['the-top', 'didReceiveAttrs'], ['the-top', 'willRender'], ['the-top', 'willInsertElement'], ['the-middle', 'init'], ['the-middle', 'on(init)'], ['the-middle', 'didReceiveAttrs'], ['the-middle', 'willRender'], ['the-middle', 'willInsertElement'], ['the-bottom', 'init'], ['the-bottom', 'on(init)'], ['the-bottom', 'didReceiveAttrs'], ['the-bottom', 'willRender'], ['the-bottom', 'willInsertElement'],

        // Async hooks

        ['the-bottom', 'didInsertElement'], ['the-bottom', 'didRender'], ['the-middle', 'didInsertElement'], ['the-middle', 'didRender'], ['the-top', 'didInsertElement'], ['the-top', 'didRender']],

        nonInteractive: [
        // Sync hooks
        ['the-top', 'init'], ['the-top', 'on(init)'], ['the-top', 'didReceiveAttrs'], ['the-middle', 'init'], ['the-middle', 'on(init)'], ['the-middle', 'didReceiveAttrs'], ['the-bottom', 'init'], ['the-bottom', 'on(init)'], ['the-bottom', 'didReceiveAttrs']]
      });

      this.runTask(() => this.components['the-bottom'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (bottom)',

        interactive: [
        // Sync hooks
        ['the-top', 'willUpdate'], ['the-top', 'willRender'], ['the-middle', 'willUpdate'], ['the-middle', 'willRender'], ['the-bottom', 'willUpdate'], ['the-bottom', 'willRender'],

        // Async hooks

        ['the-bottom', 'didUpdate'], ['the-bottom', 'didRender'], ['the-middle', 'didUpdate'], ['the-middle', 'didRender'], ['the-top', 'didUpdate'], ['the-top', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => this.components['the-middle'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (middle)',

        interactive: [
        // Sync hooks

        ['the-top', 'willUpdate'], ['the-top', 'willRender'], ['the-middle', 'willUpdate'], ['the-middle', 'willRender'],

        // Async hooks

        ['the-middle', 'didUpdate'], ['the-middle', 'didRender'], ['the-top', 'didUpdate'], ['the-top', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => this.components['the-top'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (top)',

        interactive: [
        // Sync hooks

        ['the-top', 'willUpdate'], ['the-top', 'willRender'],

        // Async hooks

        ['the-top', 'didUpdate'], ['the-top', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => (0, _metal.set)(this.context, 'twitter', '@horsetomdale'));

      this.assertText('Twitter: @horsetomdale|Name: Tom Dale|Website: tomdale.net');

      // Because the `twitter` attr is only used by the topmost component,
      // and not passed down, we do not expect to see lifecycle hooks
      // called for child components. If the `didReceiveAttrs` hook used
      // the new attribute to rerender itself imperatively, that would result
      // in lifecycle hooks being invoked for the child.

      this.assertHooks({
        label: 'after update',

        interactive: [
        // Sync hooks

        ['the-top', 'didUpdateAttrs'], ['the-top', 'didReceiveAttrs'], ['the-top', 'willUpdate'], ['the-top', 'willRender'],

        // Async hooks

        ['the-top', 'didUpdate'], ['the-top', 'didRender']],

        nonInteractive: [
        // Sync hooks
        ['the-top', 'didUpdateAttrs'], ['the-top', 'didReceiveAttrs']]
      });

      this.teardownAssertions.push(() => {
        this.assertHooks({
          label: 'destroy',

          interactive: [['the-top', 'willDestroyElement'], ['the-top', 'willClearRender'], ['the-middle', 'willDestroyElement'], ['the-middle', 'willClearRender'], ['the-bottom', 'willDestroyElement'], ['the-bottom', 'willClearRender'], ['the-top', 'didDestroyElement'], ['the-middle', 'didDestroyElement'], ['the-bottom', 'didDestroyElement'], ['the-top', 'willDestroy'], ['the-middle', 'willDestroy'], ['the-bottom', 'willDestroy']],

          nonInteractive: [['the-top', 'willDestroy'], ['the-middle', 'willDestroy'], ['the-bottom', 'willDestroy']]
        });

        this.assertRegisteredViews('after destroy');
      });
    }

    ['@test lifecycle hooks are invoked in a correct sibling order']() {
      let { attr, invoke } = this.boundHelpers;

      this.registerComponent('the-parent', {
        template: _abstractTestCase.strip`
      <div>
        ${invoke('the-first-child', { twitter: expr(attr('twitter')) })}|
        ${invoke('the-second-child', { name: expr(attr('name')) })}|
        ${invoke('the-last-child', { website: expr(attr('website')) })}
      </div>`
      });

      this.registerComponent('the-first-child', {
        template: `Twitter: {{${attr('twitter')}}}`
      });

      this.registerComponent('the-second-child', {
        template: `Name: {{${attr('name')}}}`
      });

      this.registerComponent('the-last-child', {
        template: `Website: {{${attr('website')}}}`
      });

      this.render(invoke('the-parent', {
        twitter: expr('twitter'),
        name: expr('name'),
        website: expr('website')
      }), {
        twitter: '@tomdale',
        name: 'Tom Dale',
        website: 'tomdale.net'
      });

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');
      this.assertRegisteredViews('intial render');

      this.assertHooks({
        label: 'after initial render',

        interactive: [
        // Sync hooks

        ['the-parent', 'init'], ['the-parent', 'on(init)'], ['the-parent', 'didReceiveAttrs'], ['the-parent', 'willRender'], ['the-parent', 'willInsertElement'], ['the-first-child', 'init'], ['the-first-child', 'on(init)'], ['the-first-child', 'didReceiveAttrs'], ['the-first-child', 'willRender'], ['the-first-child', 'willInsertElement'], ['the-second-child', 'init'], ['the-second-child', 'on(init)'], ['the-second-child', 'didReceiveAttrs'], ['the-second-child', 'willRender'], ['the-second-child', 'willInsertElement'], ['the-last-child', 'init'], ['the-last-child', 'on(init)'], ['the-last-child', 'didReceiveAttrs'], ['the-last-child', 'willRender'], ['the-last-child', 'willInsertElement'],

        // Async hooks

        ['the-first-child', 'didInsertElement'], ['the-first-child', 'didRender'], ['the-second-child', 'didInsertElement'], ['the-second-child', 'didRender'], ['the-last-child', 'didInsertElement'], ['the-last-child', 'didRender'], ['the-parent', 'didInsertElement'], ['the-parent', 'didRender']],

        nonInteractive: [
        // Sync hooks

        ['the-parent', 'init'], ['the-parent', 'on(init)'], ['the-parent', 'didReceiveAttrs'], ['the-first-child', 'init'], ['the-first-child', 'on(init)'], ['the-first-child', 'didReceiveAttrs'], ['the-second-child', 'init'], ['the-second-child', 'on(init)'], ['the-second-child', 'didReceiveAttrs'], ['the-last-child', 'init'], ['the-last-child', 'on(init)'], ['the-last-child', 'didReceiveAttrs']]
      });

      this.runTask(() => this.components['the-first-child'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (first child)',

        interactive: [
        // Sync hooks

        ['the-parent', 'willUpdate'], ['the-parent', 'willRender'], ['the-first-child', 'willUpdate'], ['the-first-child', 'willRender'],

        // Async hooks

        ['the-first-child', 'didUpdate'], ['the-first-child', 'didRender'], ['the-parent', 'didUpdate'], ['the-parent', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => this.components['the-second-child'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (second child)',

        interactive: [
        // Sync hooks

        ['the-parent', 'willUpdate'], ['the-parent', 'willRender'], ['the-second-child', 'willUpdate'], ['the-second-child', 'willRender'],

        // Async hooks

        ['the-second-child', 'didUpdate'], ['the-second-child', 'didRender'], ['the-parent', 'didUpdate'], ['the-parent', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => this.components['the-last-child'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (last child)',

        interactive: [
        // Sync hooks

        ['the-parent', 'willUpdate'], ['the-parent', 'willRender'], ['the-last-child', 'willUpdate'], ['the-last-child', 'willRender'],

        // Async hooks

        ['the-last-child', 'didUpdate'], ['the-last-child', 'didRender'], ['the-parent', 'didUpdate'], ['the-parent', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => this.components['the-parent'].rerender());

      this.assertText('Twitter: @tomdale|Name: Tom Dale|Website: tomdale.net');

      this.assertHooks({
        label: 'after no-op rerender (parent)',

        interactive: [
        // Sync hooks

        ['the-parent', 'willUpdate'], ['the-parent', 'willRender'],

        // Async hooks

        ['the-parent', 'didUpdate'], ['the-parent', 'didRender']],

        nonInteractive: []
      });

      this.runTask(() => (0, _metal.setProperties)(this.context, {
        twitter: '@horsetomdale',
        name: 'Horse Tom Dale',
        website: 'horsetomdale.net'
      }));

      this.assertText('Twitter: @horsetomdale|Name: Horse Tom Dale|Website: horsetomdale.net');

      this.assertHooks({
        label: 'after update',

        interactive: [
        // Sync hooks

        ['the-parent', 'didUpdateAttrs'], ['the-parent', 'didReceiveAttrs'], ['the-parent', 'willUpdate'], ['the-parent', 'willRender'], ['the-first-child', 'didUpdateAttrs'], ['the-first-child', 'didReceiveAttrs'], ['the-first-child', 'willUpdate'], ['the-first-child', 'willRender'], ['the-second-child', 'didUpdateAttrs'], ['the-second-child', 'didReceiveAttrs'], ['the-second-child', 'willUpdate'], ['the-second-child', 'willRender'], ['the-last-child', 'didUpdateAttrs'], ['the-last-child', 'didReceiveAttrs'], ['the-last-child', 'willUpdate'], ['the-last-child', 'willRender'],

        // Async hooks

        ['the-first-child', 'didUpdate'], ['the-first-child', 'didRender'], ['the-second-child', 'didUpdate'], ['the-second-child', 'didRender'], ['the-last-child', 'didUpdate'], ['the-last-child', 'didRender'], ['the-parent', 'didUpdate'], ['the-parent', 'didRender']],

        nonInteractive: [
        // Sync hooks

        ['the-parent', 'didUpdateAttrs'], ['the-parent', 'didReceiveAttrs'], ['the-first-child', 'didUpdateAttrs'], ['the-first-child', 'didReceiveAttrs'], ['the-second-child', 'didUpdateAttrs'], ['the-second-child', 'didReceiveAttrs'], ['the-last-child', 'didUpdateAttrs'], ['the-last-child', 'didReceiveAttrs']]
      });

      this.teardownAssertions.push(() => {
        this.assertHooks({
          label: 'destroy',

          interactive: [['the-parent', 'willDestroyElement'], ['the-parent', 'willClearRender'], ['the-first-child', 'willDestroyElement'], ['the-first-child', 'willClearRender'], ['the-second-child', 'willDestroyElement'], ['the-second-child', 'willClearRender'], ['the-last-child', 'willDestroyElement'], ['the-last-child', 'willClearRender'], ['the-parent', 'didDestroyElement'], ['the-first-child', 'didDestroyElement'], ['the-second-child', 'didDestroyElement'], ['the-last-child', 'didDestroyElement'], ['the-parent', 'willDestroy'], ['the-first-child', 'willDestroy'], ['the-second-child', 'willDestroy'], ['the-last-child', 'willDestroy']],

          nonInteractive: [['the-parent', 'willDestroy'], ['the-first-child', 'willDestroy'], ['the-second-child', 'willDestroy'], ['the-last-child', 'willDestroy']]
        });

        this.assertRegisteredViews('after destroy');
      });
    }

    ['@test passing values through attrs causes lifecycle hooks to fire if the attribute values have changed']() {
      let { attr, invoke } = this.boundHelpers;

      this.registerComponent('the-top', {
        template: _abstractTestCase.strip`
      <div>
        Top: ${invoke('the-middle', { twitterTop: expr(attr('twitter')) })}
      </div>`
      });

      this.registerComponent('the-middle', {
        template: _abstractTestCase.strip`
      <div>
        Middle: ${invoke('the-bottom', {
          twitterMiddle: expr(attr('twitterTop'))
        })}
      </div>`
      });

      this.registerComponent('the-bottom', {
        template: _abstractTestCase.strip`
      <div>
        Bottom: {{${attr('twitterMiddle')}}}
      </div>`
      });

      this.render(invoke('the-top', { twitter: expr('twitter') }), {
        twitter: '@tomdale'
      });

      this.assertText('Top: Middle: Bottom: @tomdale');
      this.assertRegisteredViews('intial render');

      this.assertHooks({
        label: 'after initial render',

        interactive: [
        // Sync hooks

        ['the-top', 'init'], ['the-top', 'on(init)'], ['the-top', 'didReceiveAttrs'], ['the-top', 'willRender'], ['the-top', 'willInsertElement'], ['the-middle', 'init'], ['the-middle', 'on(init)'], ['the-middle', 'didReceiveAttrs'], ['the-middle', 'willRender'], ['the-middle', 'willInsertElement'], ['the-bottom', 'init'], ['the-bottom', 'on(init)'], ['the-bottom', 'didReceiveAttrs'], ['the-bottom', 'willRender'], ['the-bottom', 'willInsertElement'],

        // Async hooks

        ['the-bottom', 'didInsertElement'], ['the-bottom', 'didRender'], ['the-middle', 'didInsertElement'], ['the-middle', 'didRender'], ['the-top', 'didInsertElement'], ['the-top', 'didRender']],

        nonInteractive: [
        // Sync hooks

        ['the-top', 'init'], ['the-top', 'on(init)'], ['the-top', 'didReceiveAttrs'], ['the-middle', 'init'], ['the-middle', 'on(init)'], ['the-middle', 'didReceiveAttrs'], ['the-bottom', 'init'], ['the-bottom', 'on(init)'], ['the-bottom', 'didReceiveAttrs']]
      });

      this.runTask(() => (0, _metal.set)(this.context, 'twitter', '@horsetomdale'));

      this.assertText('Top: Middle: Bottom: @horsetomdale');

      // Because the `twitter` attr is used by the all of the components,
      // the lifecycle hooks are invoked for all components.

      this.assertHooks({
        label: 'after updating (root)',

        interactive: [
        // Sync hooks

        ['the-top', 'didUpdateAttrs'], ['the-top', 'didReceiveAttrs'], ['the-top', 'willUpdate'], ['the-top', 'willRender'], ['the-middle', 'didUpdateAttrs'], ['the-middle', 'didReceiveAttrs'], ['the-middle', 'willUpdate'], ['the-middle', 'willRender'], ['the-bottom', 'didUpdateAttrs'], ['the-bottom', 'didReceiveAttrs'], ['the-bottom', 'willUpdate'], ['the-bottom', 'willRender'],

        // Async hooks

        ['the-bottom', 'didUpdate'], ['the-bottom', 'didRender'], ['the-middle', 'didUpdate'], ['the-middle', 'didRender'], ['the-top', 'didUpdate'], ['the-top', 'didRender']],

        nonInteractive: [
        // Sync hooks

        ['the-top', 'didUpdateAttrs'], ['the-top', 'didReceiveAttrs'], ['the-middle', 'didUpdateAttrs'], ['the-middle', 'didReceiveAttrs'], ['the-bottom', 'didUpdateAttrs'], ['the-bottom', 'didReceiveAttrs']]
      });

      this.runTask(() => this.rerender());

      this.assertText('Top: Middle: Bottom: @horsetomdale');

      // In this case, because the attrs are passed down, all child components are invoked.

      this.assertHooks({
        label: 'after no-op rernder (root)',
        interactive: [],
        nonInteractive: []
      });

      this.teardownAssertions.push(() => {
        this.assertHooks({
          label: 'destroy',

          interactive: [['the-top', 'willDestroyElement'], ['the-top', 'willClearRender'], ['the-middle', 'willDestroyElement'], ['the-middle', 'willClearRender'], ['the-bottom', 'willDestroyElement'], ['the-bottom', 'willClearRender'], ['the-top', 'didDestroyElement'], ['the-middle', 'didDestroyElement'], ['the-bottom', 'didDestroyElement'], ['the-top', 'willDestroy'], ['the-middle', 'willDestroy'], ['the-bottom', 'willDestroy']],

          nonInteractive: [['the-top', 'willDestroy'], ['the-middle', 'willDestroy'], ['the-bottom', 'willDestroy']]
        });

        this.assertRegisteredViews('after destroy');
      });
    }

    ['@test components rendered from `{{each}}` have correct life-cycle hooks to be called']() {
      let { invoke } = this.boundHelpers;

      this.registerComponent('nested-item', { template: `{{yield}}` });

      this.registerComponent('an-item', {
        template: _abstractTestCase.strip`
      {{#nested-item}}Item: {{count}}{{/nested-item}}
    `
      });

      this.registerComponent('no-items', {
        template: _abstractTestCase.strip`
      {{#nested-item}}Nothing to see here{{/nested-item}}
    `
      });

      this.render(_abstractTestCase.strip`
      {{#each items as |item|}}
        ${invoke('an-item', { count: expr('item') })}
      {{else}}
        ${invoke('no-items')}
      {{/each}}
    `, {
        items: [1, 2, 3, 4, 5]
      });

      this.assertText('Item: 1Item: 2Item: 3Item: 4Item: 5');
      this.assertRegisteredViews('intial render');

      let initialHooks = () => {
        let ret = [['an-item', 'init'], ['an-item', 'on(init)'], ['an-item', 'didReceiveAttrs']];
        if (this.isInteractive) {
          ret.push(['an-item', 'willRender'], ['an-item', 'willInsertElement']);
        }
        ret.push(['nested-item', 'init'], ['nested-item', 'on(init)'], ['nested-item', 'didReceiveAttrs']);
        if (this.isInteractive) {
          ret.push(['nested-item', 'willRender'], ['nested-item', 'willInsertElement']);
        }
        return ret;
      };

      let initialAfterRenderHooks = () => {
        if (this.isInteractive) {
          return [['nested-item', 'didInsertElement'], ['nested-item', 'didRender'], ['an-item', 'didInsertElement'], ['an-item', 'didRender']];
        } else {
          return [];
        }
      };

      this.assertHooks({
        label: 'after initial render',

        interactive: [
        // Sync hooks
        ...initialHooks(1), ...initialHooks(2), ...initialHooks(3), ...initialHooks(4), ...initialHooks(5),

        // Async hooks
        ...initialAfterRenderHooks(5), ...initialAfterRenderHooks(4), ...initialAfterRenderHooks(3), ...initialAfterRenderHooks(2), ...initialAfterRenderHooks(1)],

        nonInteractive: [
        // Sync hooks
        ...initialHooks(1), ...initialHooks(2), ...initialHooks(3), ...initialHooks(4), ...initialHooks(5),

        // Async hooks
        ...initialAfterRenderHooks(5), ...initialAfterRenderHooks(4), ...initialAfterRenderHooks(3), ...initialAfterRenderHooks(2), ...initialAfterRenderHooks(1)]
      });

      // TODO: Is this correct? Should childViews be populated in non-interactive mode?
      if (this.isInteractive) {
        this.assert.equal(this.component.childViews.length, 5, 'childViews precond');
      }

      this.runTask(() => (0, _metal.set)(this.context, 'items', []));

      // TODO: Is this correct? Should childViews be populated in non-interactive mode?
      if (this.isInteractive) {
        this.assert.equal(this.component.childViews.length, 1, 'childViews updated');
      }

      this.assertText('Nothing to see here');

      this.assertHooks({
        label: 'reset to empty array',

        interactive: [['an-item', 'willDestroyElement'], ['an-item', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['an-item', 'willDestroyElement'], ['an-item', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['an-item', 'willDestroyElement'], ['an-item', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['an-item', 'willDestroyElement'], ['an-item', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['an-item', 'willDestroyElement'], ['an-item', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['no-items', 'init'], ['no-items', 'on(init)'], ['no-items', 'didReceiveAttrs'], ['no-items', 'willRender'], ['no-items', 'willInsertElement'], ['nested-item', 'init'], ['nested-item', 'on(init)'], ['nested-item', 'didReceiveAttrs'], ['nested-item', 'willRender'], ['nested-item', 'willInsertElement'], ['an-item', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['an-item', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['an-item', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['an-item', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['an-item', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['nested-item', 'didInsertElement'], ['nested-item', 'didRender'], ['no-items', 'didInsertElement'], ['no-items', 'didRender'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy']],

        nonInteractive: [['no-items', 'init'], ['no-items', 'on(init)'], ['no-items', 'didReceiveAttrs'], ['nested-item', 'init'], ['nested-item', 'on(init)'], ['nested-item', 'didReceiveAttrs'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy'], ['an-item', 'willDestroy'], ['nested-item', 'willDestroy']]
      });

      this.teardownAssertions.push(() => {
        this.assertHooks({
          label: 'destroy',

          interactive: [['no-items', 'willDestroyElement'], ['no-items', 'willClearRender'], ['nested-item', 'willDestroyElement'], ['nested-item', 'willClearRender'], ['no-items', 'didDestroyElement'], ['nested-item', 'didDestroyElement'], ['no-items', 'willDestroy'], ['nested-item', 'willDestroy']],

          nonInteractive: [['no-items', 'willDestroy'], ['nested-item', 'willDestroy']]
        });

        this.assertRegisteredViews('after destroy');
      });
    }
  }

  class CurlyComponentsTest extends LifeCycleHooksTest {
    get ComponentClass() {
      return _helpers.Component;
    }

    invocationFor(name, namedArgs = {}) {
      let attrs = Object.keys(namedArgs).map(k => `${k}=${this.val(namedArgs[k])}`).join(' ');
      return `{{${name} ${attrs}}}`;
    }

    attrFor(name) {
      return `${name}`;
    }

    /* private */
    val(value) {
      if (value.isString) {
        return JSON.stringify(value.value);
      } else if (value.isExpr) {
        return `(readonly ${value.value})`;
      } else {
        throw new Error(`Unknown value: ${value}`);
      }
    }
  }

  (0, _testCase.moduleFor)('Components test: interactive lifecycle hooks (curly components)', class extends CurlyComponentsTest {
    get isInteractive() {
      return true;
    }
  });

  (0, _testCase.moduleFor)('Components test: non-interactive lifecycle hooks (curly components)', class extends CurlyComponentsTest {
    get isInteractive() {
      return false;
    }
  });

  (0, _testCase.moduleFor)('Components test: interactive lifecycle hooks (tagless curly components)', class extends CurlyComponentsTest {
    get ComponentClass() {
      return _helpers.Component.extend({ tagName: '' });
    }

    get isInteractive() {
      return true;
    }
  });

  (0, _testCase.moduleFor)('Components test: non-interactive lifecycle hooks (tagless curly components)', class extends CurlyComponentsTest {
    get ComponentClass() {
      return _helpers.Component.extend({ tagName: '' });
    }

    get isInteractive() {
      return false;
    }
  });

  (0, _testCase.moduleFor)('Run loop and lifecycle hooks', class extends _testCase.RenderingTest {
    ['@test afterRender set']() {
      let ComponentClass = _helpers.Component.extend({
        width: '5',
        didInsertElement() {
          (0, _runloop.schedule)('afterRender', () => {
            this.set('width', '10');
          });
        }
      });

      let template = `{{width}}`;
      this.registerComponent('foo-bar', { ComponentClass, template });

      this.render('{{foo-bar}}');

      this.assertText('10');

      this.runTask(() => this.rerender());

      this.assertText('10');
    }

    ['@test afterRender set on parent']() {
      let ComponentClass = _helpers.Component.extend({
        didInsertElement() {
          (0, _runloop.schedule)('afterRender', () => {
            let parent = this.get('parent');
            parent.set('foo', 'wat');
          });
        }
      });

      let template = `{{foo}}`;

      this.registerComponent('foo-bar', { ComponentClass, template });

      this.render('{{foo-bar parent=this foo=foo}}');

      this.assertText('wat');

      this.runTask(() => this.rerender());

      this.assertText('wat');
    }

    ['@test `willRender` can set before render (GH#14458)']() {
      let ComponentClass = _helpers.Component.extend({
        tagName: 'a',
        customHref: 'http://google.com',
        attributeBindings: ['customHref:href'],
        willRender() {
          this.set('customHref', 'http://willRender.com');
        }
      });

      let template = `Hello World`;

      this.registerComponent('foo-bar', { ComponentClass, template });

      this.render(`{{foo-bar id="foo"}}`);

      this.assertElement(this.firstChild, {
        tagName: 'a',
        attrs: {
          id: 'foo',
          href: 'http://willRender.com',
          class: (0, _testHelpers.classes)('ember-view')
        }
      });
    }

    ['@test that thing about destroying'](assert) {
      let ParentDestroyedElements = [];
      let ChildDestroyedElements = [];

      let ParentComponent = _helpers.Component.extend({
        willDestroyElement() {
          ParentDestroyedElements.push({
            id: this.itemId,
            name: 'parent-component',
            hasParent: !!this.element.parentNode,
            nextSibling: !!this.element.nextSibling,
            previousSibling: !!this.element.previousSibling
          });
        }
      });

      let PartentTemplate = _abstractTestCase.strip`
      {{yield}}
      <ul>
        {{#nested-component nestedId=(concat itemId '-A')}}A{{/nested-component}}
        {{#nested-component nestedId=(concat itemId '-B')}}B{{/nested-component}}
      </ul>
    `;

      let NestedComponent = _helpers.Component.extend({
        willDestroyElement() {
          ChildDestroyedElements.push({
            id: this.nestedId,
            name: 'nested-component',
            hasParent: !!this.element.parentNode,
            nextSibling: !!this.element.nextSibling,
            previousSibling: !!this.element.previousSibling
          });
        }
      });

      let NestedTemplate = `{{yield}}`;

      this.registerComponent('parent-component', {
        ComponentClass: ParentComponent,
        template: PartentTemplate
      });

      this.registerComponent('nested-component', {
        ComponentClass: NestedComponent,
        template: NestedTemplate
      });

      let array = (0, _runtime.A)([{ id: 1 }, { id: 2 }, { id: 3 }, { id: 4 }, { id: 5 }]);

      this.render(_abstractTestCase.strip`
      {{#each items as |item|}}
        {{#parent-component itemId=item.id}}{{item.id}}{{/parent-component}}
      {{/each}}
      {{#if model.shouldShow}}
        {{#parent-component itemId=6}}6{{/parent-component}}
      {{/if}}
      {{#if model.shouldShow}}
        {{#parent-component itemId=7}}7{{/parent-component}}
      {{/if}}
    `, {
        items: array,
        model: { shouldShow: true }
      });

      this.assertText('1AB2AB3AB4AB5AB6AB7AB');

      this.runTask(() => {
        array.removeAt(2);
        array.removeAt(2);
        (0, _metal.set)(this.context, 'model.shouldShow', false);
      });

      this.assertText('1AB2AB5AB');

      assertDestroyHooks(assert, [...ParentDestroyedElements], [{
        id: 3,
        hasParent: true,
        nextSibling: true,
        previousSibling: true
      }, {
        id: 4,
        hasParent: true,
        nextSibling: true,
        previousSibling: true
      }, {
        id: 6,
        hasParent: true,
        nextSibling: true,
        previousSibling: true
      }, {
        id: 7,
        hasParent: true,
        nextSibling: false,
        previousSibling: true
      }]);

      assertDestroyHooks(assert, [...ChildDestroyedElements], [{
        id: '3-A',
        hasParent: true,
        nextSibling: true,
        previousSibling: false
      }, {
        id: '3-B',
        hasParent: true,
        nextSibling: false,
        previousSibling: true
      }, {
        id: '4-A',
        hasParent: true,
        nextSibling: true,
        previousSibling: false
      }, {
        id: '4-B',
        hasParent: true,
        nextSibling: false,
        previousSibling: true
      }, {
        id: '6-A',
        hasParent: true,
        nextSibling: true,
        previousSibling: false
      }, {
        id: '6-B',
        hasParent: true,
        nextSibling: false,
        previousSibling: true
      }, {
        id: '7-A',
        hasParent: true,
        nextSibling: true,
        previousSibling: false
      }, {
        id: '7-B',
        hasParent: true,
        nextSibling: false,
        previousSibling: true
      }]);
    }

    ['@test lifecycle hooks exist on the base class'](assert) {
      // Make sure we get the finalized component prototype
      let prototype = _helpers.Component.proto();

      assert.equal(typeof prototype.didDestroyElement, 'function', 'didDestroyElement exists');
      assert.equal(typeof prototype.didInsertElement, 'function', 'didInsertElement exists');
      assert.equal(typeof prototype.didReceiveAttrs, 'function', 'didReceiveAttrs exists');
      assert.equal(typeof prototype.didRender, 'function', 'didRender exists');
      assert.equal(typeof prototype.didUpdate, 'function', 'didUpdate exists');
      assert.equal(typeof prototype.didUpdateAttrs, 'function', 'didUpdateAttrs exists');
      assert.equal(typeof prototype.willClearRender, 'function', 'willClearRender exists');
      assert.equal(typeof prototype.willDestroy, 'function', 'willDestroy exists');
      assert.equal(typeof prototype.willDestroyElement, 'function', 'willDestroyElement exists');
      assert.equal(typeof prototype.willInsertElement, 'function', 'willInsertElement exists');
      assert.equal(typeof prototype.willRender, 'function', 'willRender exists');
      assert.equal(typeof prototype.willUpdate, 'function', 'willUpdate exists');
    }
  });

  if (!_views.jQueryDisabled) {
    (0, _testCase.moduleFor)('Run loop and lifecycle hooks - jQuery only', class extends _testCase.RenderingTest {
      ['@test lifecycle hooks have proper access to this.$()'](assert) {
        assert.expect(6);
        let component;
        let FooBarComponent = _helpers.Component.extend({
          tagName: 'div',
          init() {
            assert.notOk(this.$(), 'no access to element via this.$() on init() enter');
            this._super(...arguments);
            assert.notOk(this.$(), 'no access to element via this.$() after init() finished');
          },
          willInsertElement() {
            component = this;
            assert.ok(this.$(), 'willInsertElement has access to element via this.$()');
          },
          didInsertElement() {
            assert.ok(this.$(), 'didInsertElement has access to element via this.$()');
          },
          willDestroyElement() {
            assert.ok(this.$(), 'willDestroyElement has access to element via this.$()');
          },
          didDestroyElement() {
            assert.notOk(this.$(), 'didDestroyElement does not have access to element via this.$()');
          }
        });
        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });
        let { owner } = this;
        let comp = owner.lookup('component:foo-bar');
        (0, _internalTestHelpers.runAppend)(comp);
        this.runTask(() => (0, _utils.tryInvoke)(component, 'destroy'));
      }
    });
  }

  function assertDestroyHooks(assert, _actual, _expected) {
    _expected.forEach((expected, i) => {
      let name = expected.name;
      assert.equal(expected.id, _actual[i].id, `${name} id is the same`);
      assert.equal(expected.hasParent, _actual[i].hasParent, `${name} has parent node`);
      assert.equal(expected.nextSibling, _actual[i].nextSibling, `${name} has next sibling node`);
      assert.equal(expected.previousSibling, _actual[i].previousSibling, `${name} has previous sibling node`);
    });
  }

  function bind(func, thisArg) {
    return (...args) => func.apply(thisArg, args);
  }

  function string(value) {
    return { isString: true, value };
  }

  function expr(value) {
    return { isExpr: true, value };
  }

  function hook(name, hook, { attrs, oldAttrs, newAttrs } = {}) {
    return { name, hook, args: { attrs, oldAttrs, newAttrs } };
  }

  function json(serializable) {
    return JSON.parse(JSON.stringify(serializable));
  }
});
enifed('@ember/-internals/glimmer/tests/integration/components/link-to-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/controller', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-helpers'], function (_testCase, _controller, _metal, _helpers, _testHelpers) {
  'use strict';

  (0, _testCase.moduleFor)('Link-to component', class extends _testCase.ApplicationTest {
    visitWithDeprecation(path, deprecation) {
      let p;

      expectDeprecation(() => {
        p = this.visit(path);
      }, deprecation);

      return p;
    }

    ['@test should be able to be inserted in DOM when the router is not present']() {
      this.addTemplate('application', `{{#link-to 'index'}}Go to Index{{/link-to}}`);

      return this.visit('/').then(() => {
        this.assertText('Go to Index');
      });
    }

    ['@test re-renders when title changes']() {
      let controller;

      this.addTemplate('application', '{{link-to title routeName}}');
      this.add('controller:application', _controller.default.extend({
        init() {
          this._super(...arguments);
          controller = this;
        },
        title: 'foo',
        routeName: 'index'
      }));

      return this.visit('/').then(() => {
        this.assertText('foo');
        this.runTask(() => (0, _metal.set)(controller, 'title', 'bar'));
        this.assertText('bar');
      });
    }

    ['@test re-computes active class when params change'](assert) {
      let controller;

      this.addTemplate('application', '{{link-to "foo" routeName}}');

      this.add('controller:application', _controller.default.extend({
        init() {
          this._super(...arguments);
          controller = this;
        },
        routeName: 'index'
      }));

      this.router.map(function () {
        this.route('bar', { path: '/bar' });
      });

      return this.visit('/bar').then(() => {
        assert.equal(this.firstChild.classList.contains('active'), false);
        this.runTask(() => (0, _metal.set)(controller, 'routeName', 'bar'));
        assert.equal(this.firstChild.classList.contains('active'), true);
      });
    }

    ['@test escaped inline form (double curlies) escapes link title']() {
      this.addTemplate('application', `{{link-to title 'index'}}`);
      this.add('controller:application', _controller.default.extend({
        title: '<b>blah</b>'
      }));

      return this.visit('/').then(() => {
        this.assertText('<b>blah</b>');
      });
    }

    ['@test escaped inline form with (-html-safe) does not escape link title'](assert) {
      this.addTemplate('application', `{{link-to (-html-safe title) 'index'}}`);
      this.add('controller:application', _controller.default.extend({
        title: '<b>blah</b>'
      }));

      return this.visit('/').then(() => {
        this.assertText('blah');
        assert.equal(this.$('b').length, 1);
      });
    }

    ['@test unescaped inline form (triple curlies) does not escape link title'](assert) {
      this.addTemplate('application', `{{{link-to title 'index'}}}`);
      this.add('controller:application', _controller.default.extend({
        title: '<b>blah</b>'
      }));

      return this.visit('/').then(() => {
        this.assertText('blah');
        assert.equal(this.$('b').length, 1);
      });
    }

    ['@test able to safely extend the built-in component and use the normal path']() {
      this.addComponent('custom-link-to', {
        ComponentClass: _helpers.LinkComponent.extend()
      });
      this.addTemplate('application', `{{#custom-link-to 'index'}}{{title}}{{/custom-link-to}}`);
      this.add('controller:application', _controller.default.extend({
        title: 'Hello'
      }));

      return this.visit('/').then(() => {
        this.assertText('Hello');
      });
    }

    ['@test [GH#13432] able to safely extend the built-in component and invoke it inline']() {
      this.addComponent('custom-link-to', {
        ComponentClass: _helpers.LinkComponent.extend()
      });
      this.addTemplate('application', `{{custom-link-to title 'index'}}`);
      this.add('controller:application', _controller.default.extend({
        title: 'Hello'
      }));

      return this.visit('/').then(() => {
        this.assertText('Hello');
      });
    }
  });

  (0, _testCase.moduleFor)('Link-to component with query-params', class extends _testCase.ApplicationTest {
    constructor() {
      super(...arguments);

      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo'],
        foo: '123',
        bar: 'yes'
      }));
    }

    ['@test populates href with fully supplied query param values']() {
      this.addTemplate('index', `{{#link-to 'index' (query-params foo='456' bar='NAW')}}Index{{/link-to}}`);

      return this.visit('/').then(() => {
        this.assertComponentElement(this.firstChild, {
          tagName: 'a',
          attrs: { href: '/?bar=NAW&foo=456' },
          content: 'Index'
        });
      });
    }

    ['@test populates href with partially supplied query param values, but omits if value is default value']() {
      this.addTemplate('index', `{{#link-to 'index' (query-params foo='123')}}Index{{/link-to}}`);

      return this.visit('/').then(() => {
        this.assertComponentElement(this.firstChild, {
          tagName: 'a',
          attrs: { href: '/', class: (0, _testHelpers.classes)('ember-view active') },
          content: 'Index'
        });
      });
    }
  });

  (0, _testCase.moduleFor)('Link-to component', class extends _testCase.RenderingTest {
    ['@test should be able to be inserted in DOM when the router is not present - block']() {
      this.render(`{{#link-to 'index'}}Go to Index{{/link-to}}`);

      this.assertText('Go to Index');
    }

    ['@test should be able to be inserted in DOM when the router is not present - inline']() {
      this.render(`{{link-to 'Go to Index' 'index'}}`);

      this.assertText('Go to Index');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/local-lookup-test', ['@ember/-internals/glimmer/tests/utils/test-case', 'ember-template-compiler', 'internal-test-helpers', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/canary-features', '@ember/-internals/glimmer'], function (_testCase, _emberTemplateCompiler, _internalTestHelpers, _helpers, _canaryFeatures, _glimmer) {
  'use strict';

  class LocalLookupTest extends _testCase.RenderingTest {
    ['@test it can lookup a local template']() {
      this.registerComponent('x-outer/x-inner', {
        template: 'Nested template says: {{yield}}'
      });
      this.registerComponent('x-outer', {
        template: '{{#x-inner}}Hi!{{/x-inner}}'
      });

      this.render('{{x-outer}}');

      this.assertText('Nested template says: Hi!', 'Initial render works');

      this.runTask(() => this.rerender());

      this.assertText('Nested template says: Hi!', 'Re-render works');
    }

    ['@test tagless blockless component can lookup local template']() {
      this.registerComponent('x-outer/x-inner', {
        template: 'Nested template says: {{yield}}'
      });
      this.registerTemplate('components/x-outer', '{{#x-inner}}Hi!{{/x-inner}}');
      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({ tagName: '' })
      });

      this.render('{{x-outer}}');

      this.assertText('Nested template says: Hi!', 'Re-render works');

      this.runTask(() => this.rerender());

      this.assertText('Nested template says: Hi!', 'Re-render works');
    }

    ['@test it can lookup a local component template']() {
      this.registerTemplate('components/x-outer/x-inner', 'Nested template says: {{yield}}');
      this.registerTemplate('components/x-outer', '{{#x-inner}}Hi!{{/x-inner}}');

      this.render('{{x-outer}}');

      this.assertText('Nested template says: Hi!', 'Initial render works');

      this.runTask(() => this.rerender());

      this.assertText('Nested template says: Hi!', 'Re-render works');
    }

    ['@test it can local lookup a dynamic component']() {
      this.registerComponent('foo-bar', {
        template: 'yall finished {{component child}}'
      });
      this.registerComponent('foo-bar/biz-baz', { template: 'or yall done?' });

      this.render('{{foo-bar child=child}}', { child: 'biz-baz' });

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall done?');
    }

    ['@test it can local lookup a dynamic component from a dynamic component']() {
      this.registerComponent('foo-bar', {
        template: 'yall finished {{component child}}'
      });
      this.registerComponent('foo-bar/biz-baz', { template: 'or yall done?' });

      this.render('{{component componentName child=child}}', {
        componentName: 'foo-bar',
        child: 'biz-baz'
      });

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall done?');
    }

    ['@test it can local lookup a dynamic component from a passed named argument']() {
      this.registerComponent('parent-foo', {
        template: `yall finished {{global-biz baz=(component 'local-bar')}}`
      });
      this.registerComponent('global-biz', { template: 'or {{component baz}}' });
      this.registerComponent('parent-foo/local-bar', { template: 'yall done?' });

      this.render('{{parent-foo}}');

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall done?');
    }

    ['@test it can local lookup a re-wrapped dynamic component from a passed named argument']() {
      this.registerComponent('parent-foo', {
        template: `yall finished {{global-x comp=(component 'local-bar')}}`
      });
      this.registerComponent('global-x', {
        template: `or {{global-y comp=(component comp phrase='done')}}`
      });
      this.registerComponent('global-y', { template: `{{component comp}}?` });
      this.registerComponent('parent-foo/local-bar', {
        template: 'yall {{phrase}}'
      });

      this.render('{{parent-foo}}');

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall done?');
    }

    ['@test it can nest local lookups of dynamic components from a passed named argument']() {
      this.registerComponent('parent-foo', {
        template: `yall finished {{global-x comp=(component 'local-bar')}}`
      });
      this.registerComponent('global-x', {
        template: `or {{global-y comp=(component comp phrase='done')}}`
      });
      this.registerComponent('global-y', {
        template: `{{component comp}}{{component 'local-bar'}}`
      });
      this.registerComponent('parent-foo/local-bar', {
        template: 'yall {{phrase}}'
      });
      this.registerComponent('global-y/local-bar', { template: `?` });

      this.render('{{parent-foo}}');

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall done?');
    }

    ['@test it can switch from local to global lookups of dynamic components from a passed named argument']() {
      this.registerComponent('parent-foo', {
        template: `yall finished {{global-x comp=(component bar)}}`
      });
      this.registerComponent('global-x', {
        template: `or yall {{component comp}}`
      });
      this.registerComponent('parent-foo/local-bar', { template: 'done?' });
      this.registerComponent('global-bar', { template: `ready?` });

      this.render('{{parent-foo bar=bar}}', { bar: 'local-bar' });

      this.assertText('yall finished or yall done?');

      this.runTask(() => this.context.set('bar', 'global-bar'));

      this.runTask(() => this.rerender());

      this.assertText('yall finished or yall ready?');
    }

    ['@test it can lookup a local helper']() {
      this.registerHelper('x-outer/x-helper', () => {
        return 'Who dis?';
      });
      this.registerComponent('x-outer', { template: 'Who dat? {{x-helper}}' });

      this.render('{{x-outer}}');

      this.assertText('Who dat? Who dis?', 'Initial render works');

      this.runTask(() => this.rerender());

      this.assertText('Who dat? Who dis?', 'Re-render works');
    }

    ['@test it overrides global helper lookup']() {
      this.registerHelper('x-outer/x-helper', () => {
        return 'Who dis?';
      });

      this.registerHelper('x-helper', () => {
        return 'I dunno';
      });

      this.registerComponent('x-outer', { template: 'Who dat? {{x-helper}}' });

      this.render('{{x-outer}} {{x-helper}}');

      this.assertText('Who dat? Who dis? I dunno', 'Initial render works');

      this.runTask(() => this.rerender());

      this.assertText('Who dat? Who dis? I dunno', 'Re-render works');
    }

    ['@test lookup without match issues standard assertion (with local helper name)']() {
      this.registerComponent('x-outer', {
        template: '{{#x-inner}}Hi!{{/x-inner}}'
      });

      expectAssertion(() => {
        this.render('{{x-outer}}');
      }, /A component or helper named "x-inner" could not be found/);
    }

    ['@test overrides global lookup']() {
      this.registerComponent('x-outer', {
        template: '{{#x-inner}}Hi!{{/x-inner}}'
      });
      this.registerComponent('x-outer/x-inner', {
        template: 'Nested template says (from local): {{yield}}'
      });
      this.registerComponent('x-inner', {
        template: 'Nested template says (from global): {{yield}}'
      });

      this.render('{{#x-inner}}Hi!{{/x-inner}} {{x-outer}} {{#x-outer/x-inner}}Hi!{{/x-outer/x-inner}}');

      this.assertText('Nested template says (from global): Hi! Nested template says (from local): Hi! Nested template says (from local): Hi!');

      this.runTask(() => this.rerender());

      this.assertText('Nested template says (from global): Hi! Nested template says (from local): Hi! Nested template says (from local): Hi!');
    }
  }

  // first run these tests with expandLocalLookup

  function buildResolver() {
    let resolver = {
      resolve() {},
      expandLocalLookup(fullName, sourceFullName) {
        if (!sourceFullName) {
          return null;
        }

        let [sourceType, sourceName] = sourceFullName.split(':');
        let [type, name] = fullName.split(':');

        sourceName = sourceName.replace('my-app/', '');

        if (sourceType === 'template' && sourceName.slice(0, 21) === 'templates/components/') {
          sourceName = sourceName.slice(21);
        }

        name = name.replace('my-app/', '');

        if (type === 'template' && name.slice(0, 11) === 'components/') {
          name = name.slice(11);
          sourceName = `components/${sourceName}`;
        }

        sourceName = sourceName.replace('.hbs', '');

        let result = `${type}:${sourceName}/${name}`;
        return result;
      }
    };

    return resolver;
  }

  (0, _testCase.moduleFor)('Components test: local lookup with expandLocalLookup feature', class extends LocalLookupTest {
    getResolver() {
      return buildResolver();
    }
  });

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    class LocalLookupTestResolver extends _internalTestHelpers.ModuleBasedTestResolver {
      expandLocalLookup(specifier, source) {
        if (source && source.indexOf('components/') !== -1) {
          let namespace = source.split('components/')[1];
          let [type, name] = specifier.split(':');
          name = name.replace('components/', '');

          namespace = namespace.replace('.hbs', '');
          return `${type}:${type === 'template' ? 'components/' : ''}${namespace}/${name}`;
        }

        return super.expandLocalLookup(specifier, source);
      }
    }

    /*
     * This sub-classing changes `registerXXX` methods to use the resolver.
     * Required for testing the module unification-friendly `resolve` call
     * with a `referrer` argument.
     *
     * In theory all these tests can be ported to use the resolver instead of
     * the registry.
     */
    (0, _testCase.moduleFor)('Components test: local lookup with resolution referrer', class extends LocalLookupTest {
      get resolver() {
        return this.owner.__registry__.fallback.resolver;
      }

      getResolver() {
        return new LocalLookupTestResolver();
      }

      registerComponent(name, { ComponentClass = _helpers.Component, template = null }) {
        let { resolver } = this;

        if (ComponentClass) {
          resolver.add(`component:${name}`, ComponentClass);
        }

        if (typeof template === 'string') {
          resolver.add(`template:components/${name}`, this.compile(template, {
            moduleName: `my-name/templates/components/${name}.hbs`
          }));
        }
      }

      registerTemplate(name, template) {
        let { resolver } = this;
        if (typeof template === 'string') {
          resolver.add(`template:${name}`, this.compile(template, {
            moduleName: `my-name/templates/${name}.hbs`
          }));
        } else {
          throw new Error(`Registered template "${name}" must be a string`);
        }
      }

      registerHelper(name, funcOrClassBody) {
        let { resolver } = this;
        let type = typeof funcOrClassBody;

        if (type === 'function') {
          resolver.add(`helper:${name}`, (0, _glimmer.helper)(funcOrClassBody));
        } else if (type === 'object' && type !== null) {
          resolver.add(`helper:${name}`, _glimmer.Helper.extend(funcOrClassBody));
        } else {
          throw new Error(`Cannot register ${funcOrClassBody} as a helper`);
        }
      }
    });
  }

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    (0, _internalTestHelpers.moduleFor)('Components test: local lookup with resolution referrer (MU)', class extends _internalTestHelpers.ApplicationTestCase {
      ['@test Ensure that the same specifier with two sources does not share a cache key'](assert) {
        this.add({
          specifier: 'template:components/x-not-shared',
          source: 'template:my-app/templates/components/x-top.hbs'
        }, (0, _emberTemplateCompiler.compile)('child-x-not-shared'));

        this.add({
          specifier: 'template:components/x-top',
          source: 'template:my-app/templates/application.hbs'
        }, (0, _emberTemplateCompiler.compile)('top-level-x-top ({{x-not-shared}})', {
          moduleName: 'my-app/templates/components/x-top.hbs'
        }));

        this.add({
          specifier: 'template:components/x-not-shared',
          source: 'template:my-app/templates/application.hbs'
        }, (0, _emberTemplateCompiler.compile)('top-level-x-not-shared'));

        this.addTemplate('application', '{{x-not-shared}} {{x-top}} {{x-not-shared}} {{x-top}}');

        return this.visit('/').then(() => {
          assert.equal(this.element.textContent, 'top-level-x-not-shared top-level-x-top (child-x-not-shared) top-level-x-not-shared top-level-x-top (child-x-not-shared)');
        });
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/components/namespaced-lookup-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/canary-features', '@ember/-internals/glimmer'], function (_testCase, _canaryFeatures, _glimmer) {
  'use strict';

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    (0, _testCase.moduleFor)('Namespaced lookup', class extends _testCase.RenderingTest {
      ['@test it can render a namespaced component']() {
        this.addTemplate({
          specifier: 'template:components/my-component',
          namespace: 'my-addon'
        }, 'namespaced template {{myProp}}');

        this.add({
          specifier: 'component:my-component',
          namespace: 'my-addon'
        }, _glimmer.Component.extend({
          myProp: 'My property'
        }));

        this.addComponent('x-outer', {
          template: '{{my-addon::my-component}}'
        });

        this.render('{{x-outer}}');

        this.assertText('namespaced template My property');

        this.runTask(() => this.rerender());

        this.assertText('namespaced template My property');
      }

      ['@test it can render a nested namespaced component']() {
        this.addTemplate({
          specifier: 'template:components/my-component',
          namespace: 'second-addon'
        }, 'second namespaced template');

        this.addTemplate({
          specifier: 'template:components/my-component',
          namespace: 'first-addon'
        }, 'first namespaced template - {{second-addon::my-component}}');

        this.addComponent('x-outer', {
          template: '{{first-addon::my-component}}'
        });

        this.render('{{x-outer}}');

        this.assertText('first namespaced template - second namespaced template');

        this.runTask(() => this.rerender());

        this.assertText('first namespaced template - second namespaced template');
      }

      ['@test it can render a nested un-namespaced component']() {
        this.addTemplate({
          specifier: 'template:components/addon-component',
          source: 'template:first-addon/src/ui/components/my-component.hbs'
        }, 'un-namespaced addon template');

        this.addTemplate({
          specifier: 'template:components/my-component',
          moduleName: 'first-addon/src/ui/components/my-component.hbs',
          namespace: 'first-addon'
        }, '{{addon-component}}');

        this.addComponent('x-outer', {
          template: '{{first-addon::my-component}}'
        });

        this.render('{{x-outer}}');

        this.assertText('un-namespaced addon template');

        this.runTask(() => this.rerender());

        this.assertText('un-namespaced addon template');
      }

      ['@test it can render a namespaced main component']() {
        this.addTemplate({
          specifier: 'template:components/addon-component',
          soruce: 'template:first-addon/src/ui/components/main.hbs'
        }, 'Nested namespaced component');

        this.addTemplate({
          specifier: 'template:components/first-addon',
          moduleName: 'first-addon/src/ui/components/main.hbs'
        }, '{{addon-component}}');

        this.addComponent('x-outer', { template: '{{first-addon}}' });

        this.render('{{x-outer}}');

        this.assertText('Nested namespaced component');

        this.runTask(() => this.rerender());

        this.assertText('Nested namespaced component');
      }

      ['@test it does not render a main component when using a namespace']() {
        this.addTemplate({
          specifier: 'template:components/main',
          namespace: 'my-addon'
        }, 'namespaced template {{myProp}}');

        this.add({
          specifier: 'component:main',
          namespace: 'my-addon'
        }, _glimmer.Component.extend({
          myProp: 'My property'
        }));

        this.add({
          specifier: 'helper:my-addon',
          namespace: 'empty-namespace'
        }, (0, _glimmer.helper)(() => 'my helper'));

        this.render('{{empty-namespace::my-addon}}');

        this.assertText('my helper'); // component should be not found

        this.runTask(() => this.rerender());

        this.assertText('my helper');
      }

      ['@test it renders a namespaced helper']() {
        this.add({
          specifier: 'helper:my-helper',
          namespace: 'my-namespace'
        }, (0, _glimmer.helper)(() => 'my helper'));

        this.render('{{my-namespace::my-helper}}');

        this.assertText('my helper');

        this.runTask(() => this.rerender());

        this.assertText('my helper');
      }
    });
  }
});
enifed("@ember/-internals/glimmer/tests/integration/components/render-to-element-test", [], function () {
  "use strict";
});
enifed('@ember/-internals/glimmer/tests/integration/components/target-action-test', ['@ember/polyfills', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/controller', '@ember/-internals/runtime', '@ember/-internals/routing'], function (_polyfills, _testCase, _abstractTestCase, _metal, _helpers, _controller, _runtime, _routing) {
  'use strict';

  function expectSendActionDeprecation(fn) {
    expectDeprecation(fn, /You called (.*).sendAction\((.*)\) but Component#sendAction is deprecated. Please use closure actions instead./);
  }

  (0, _testCase.moduleFor)('Components test: sendAction', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);
      this.actionCounts = {};
      this.sendCount = 0;
      this.actionArguments = null;

      var self = this;

      this.registerComponent('action-delegate', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            self.delegate = this;
            this.name = 'action-delegate';
          }
        })
      });
    }

    renderDelegate(template = '{{action-delegate}}', context = {}) {
      let root = this;
      context = (0, _polyfills.assign)(context, {
        send(actionName, ...args) {
          root.sendCount++;
          root.actionCounts[actionName] = root.actionCounts[actionName] || 0;
          root.actionCounts[actionName]++;
          root.actionArguments = args;
        }
      });
      this.render(template, context);
    }

    assertSendCount(count) {
      this.assert.equal(this.sendCount, count, `Send was called ${count} time(s)`);
    }

    assertNamedSendCount(actionName, count) {
      this.assert.equal(this.actionCounts[actionName], count, `An action named '${actionName}' was sent ${count} times`);
    }

    assertSentWithArgs(expected, message = 'arguments were sent with the action') {
      this.assert.deepEqual(this.actionArguments, expected, message);
    }

    ['@test Calling sendAction on a component without an action defined does nothing']() {
      this.renderDelegate();

      expectSendActionDeprecation(() => {
        this.runTask(() => this.delegate.sendAction());
      });

      this.assertSendCount(0);
    }

    ['@test Calling sendAction on a component with an action defined calls send on the controller']() {
      this.renderDelegate();

      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(this.delegate, 'action', 'addItem');
          this.delegate.sendAction();
        });
      });

      this.assertSendCount(1);
      this.assertNamedSendCount('addItem', 1);
    }

    ['@test Calling sendAction on a component with a function calls the function']() {
      this.assert.expect(2);

      this.renderDelegate();

      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(this.delegate, 'action', () => this.assert.ok(true, 'function is called'));
          this.delegate.sendAction();
        });
      });
    }

    ['@test Calling sendAction on a component with a function calls the function with arguments']() {
      this.assert.expect(2);
      let argument = {};

      this.renderDelegate();
      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(this.delegate, 'action', actualArgument => {
            this.assert.deepEqual(argument, actualArgument, 'argument is passed');
          });
          this.delegate.sendAction('action', argument);
        });
      });
    }

    // TODO consolidate these next 2 tests
    ['@test Calling sendAction on a component with a reference attr calls the function with arguments']() {
      this.renderDelegate('{{action-delegate playing=playing}}', {
        playing: null
      });

      expectSendActionDeprecation(() => {
        this.runTask(() => this.delegate.sendAction());
      });

      this.assertSendCount(0);

      this.runTask(() => {
        (0, _metal.set)(this.context, 'playing', 'didStartPlaying');
      });
      expectSendActionDeprecation(() => {
        this.runTask(() => {
          this.delegate.sendAction('playing');
        });
      });

      this.assertSendCount(1);
      this.assertNamedSendCount('didStartPlaying', 1);
    }

    ['@test Calling sendAction on a component with a {{mut}} attr calls the function with arguments']() {
      this.renderDelegate('{{action-delegate playing=(mut playing)}}', {
        playing: null
      });

      expectSendActionDeprecation(() => {
        this.runTask(() => this.delegate.sendAction('playing'));
      });

      this.assertSendCount(0);

      this.runTask(() => this.delegate.attrs.playing.update('didStartPlaying'));
      expectSendActionDeprecation(() => {
        this.runTask(() => this.delegate.sendAction('playing'));
      });

      this.assertSendCount(1);
      this.assertNamedSendCount('didStartPlaying', 1);
    }

    ["@test Calling sendAction with a named action uses the component's property as the action name"]() {
      this.renderDelegate();

      let component = this.delegate;
      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(this.delegate, 'playing', 'didStartPlaying');
          component.sendAction('playing');
        });
      });

      this.assertSendCount(1);
      this.assertNamedSendCount('didStartPlaying', 1);

      expectSendActionDeprecation(() => {
        this.runTask(() => component.sendAction('playing'));
      });

      this.assertSendCount(2);
      this.assertNamedSendCount('didStartPlaying', 2);

      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(component, 'action', 'didDoSomeBusiness');
          component.sendAction();
        });
      });

      this.assertSendCount(3);
      this.assertNamedSendCount('didDoSomeBusiness', 1);
    }

    ['@test Calling sendAction when the action name is not a string raises an exception']() {
      this.renderDelegate();

      this.runTask(() => {
        (0, _metal.set)(this.delegate, 'action', {});
        (0, _metal.set)(this.delegate, 'playing', {});
      });
      expectSendActionDeprecation(() => {
        expectAssertion(() => this.delegate.sendAction());
      });
      expectSendActionDeprecation(() => {
        expectAssertion(() => this.delegate.sendAction('playing'));
      });
    }

    ['@test Calling sendAction on a component with contexts']() {
      this.renderDelegate();

      let testContext = { song: 'She Broke My Ember' };
      let firstContext = { song: 'She Broke My Ember' };
      let secondContext = { song: 'My Achey Breaky Ember' };

      expectSendActionDeprecation(() => {
        this.runTask(() => {
          (0, _metal.set)(this.delegate, 'playing', 'didStartPlaying');
          this.delegate.sendAction('playing', testContext);
        });
      });

      this.assertSendCount(1);
      this.assertNamedSendCount('didStartPlaying', 1);
      this.assertSentWithArgs([testContext], 'context was sent with the action');

      expectSendActionDeprecation(() => {
        this.runTask(() => {
          this.delegate.sendAction('playing', firstContext, secondContext);
        });
      });

      this.assertSendCount(2);
      this.assertNamedSendCount('didStartPlaying', 2);
      this.assertSentWithArgs([firstContext, secondContext], 'multiple contexts were sent to the action');
    }

    ['@test calling sendAction on a component within a block sends to the outer scope GH#14216'](assert) {
      let testContext = this;
      // overrides default action-delegate so actions can be added
      this.registerComponent('action-delegate', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            testContext.delegate = this;
            this.name = 'action-delegate';
          },

          actions: {
            derp(arg1) {
              assert.ok(true, 'action called on action-delgate');
              assert.equal(arg1, 'something special', 'argument passed through properly');
            }
          }
        }),

        template: _abstractTestCase.strip`
        {{#component-a}}
          {{component-b bar="derp"}}
        {{/component-a}}
      `
      });

      this.registerComponent('component-a', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            this.name = 'component-a';
          },
          actions: {
            derp() {
              assert.ok(false, 'no! bad scoping!');
            }
          }
        })
      });

      let innerChild;
      this.registerComponent('component-b', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            innerChild = this;
            this.name = 'component-b';
          }
        })
      });

      this.renderDelegate();
      expectSendActionDeprecation(() => {
        this.runTask(() => innerChild.sendAction('bar', 'something special'));
      });
    }

    ['@test asserts if called on a destroyed component']() {
      let component;

      this.registerComponent('rip-alley', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            component = this;
          },

          toString() {
            return 'component:rip-alley';
          }
        })
      });

      this.render('{{#if shouldRender}}{{rip-alley}}{{/if}}', {
        shouldRender: true
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'shouldRender', false);
      });

      expectAssertion(() => {
        component.sendAction('trigger-me-dead');
      }, "Attempted to call .sendAction() with the action 'trigger-me-dead' on the destroyed object 'component:rip-alley'.");
    }
  });

  (0, _testCase.moduleFor)('Components test: sendAction to a controller', class extends _testCase.ApplicationTest {
    ["@test sendAction should trigger an action on the parent component's controller if it exists"](assert) {
      assert.expect(20);

      let component;

      this.router.map(function () {
        this.route('a');
        this.route('b');
        this.route('c', function () {
          this.route('d');
          this.route('e');
        });
      });

      this.addComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: `{{val}}`
      });

      this.add('controller:a', _controller.default.extend({
        send(actionName, actionContext) {
          assert.equal(actionName, 'poke', 'send() method was invoked from a top level controller');
          assert.equal(actionContext, 'top', 'action arguments were passed into the top level controller');
        }
      }));
      this.addTemplate('a', '{{foo-bar val="a" poke="poke"}}');

      this.add('route:b', _routing.Route.extend({
        actions: {
          poke(actionContext) {
            assert.ok(true, 'Unhandled action sent to route');
            assert.equal(actionContext, 'top no controller');
          }
        }
      }));
      this.addTemplate('b', '{{foo-bar val="b" poke="poke"}}');

      this.add('route:c', _routing.Route.extend({
        actions: {
          poke(actionContext) {
            assert.ok(true, 'Unhandled action sent to route');
            assert.equal(actionContext, 'top with nested no controller');
          }
        }
      }));
      this.addTemplate('c', '{{foo-bar val="c" poke="poke"}}{{outlet}}');

      this.add('route:c.d', _routing.Route.extend({}));

      this.add('controller:c.d', _controller.default.extend({
        send(actionName, actionContext) {
          assert.equal(actionName, 'poke', 'send() method was invoked from a nested controller');
          assert.equal(actionContext, 'nested', 'action arguments were passed into the nested controller');
        }
      }));
      this.addTemplate('c.d', '{{foo-bar val=".d" poke="poke"}}');

      this.add('route:c.e', _routing.Route.extend({
        actions: {
          poke(actionContext) {
            assert.ok(true, 'Unhandled action sent to route');
            assert.equal(actionContext, 'nested no controller');
          }
        }
      }));
      this.addTemplate('c.e', '{{foo-bar val=".e" poke="poke"}}');

      return this.visit('/a').then(() => {
        expectSendActionDeprecation(() => component.sendAction('poke', 'top'));
      }).then(() => {
        this.assertText('a');
        return this.visit('/b');
      }).then(() => {
        expectSendActionDeprecation(() => component.sendAction('poke', 'top no controller'));
      }).then(() => {
        this.assertText('b');
        return this.visit('/c');
      }).then(() => {
        expectSendActionDeprecation(() => {
          component.sendAction('poke', 'top with nested no controller');
        });
      }).then(() => {
        this.assertText('c');
        return this.visit('/c/d');
      }).then(() => {
        expectSendActionDeprecation(() => component.sendAction('poke', 'nested'));
      }).then(() => {
        this.assertText('c.d');
        return this.visit('/c/e');
      }).then(() => {
        expectSendActionDeprecation(() => component.sendAction('poke', 'nested no controller'));
      }).then(() => this.assertText('c.e'));
    }

    ["@test sendAction should not trigger an action in an outlet's controller if a parent component handles it"](assert) {
      assert.expect(2);

      let component;

      this.addComponent('x-parent', {
        ComponentClass: _helpers.Component.extend({
          actions: {
            poke() {
              assert.ok(true, 'parent component handled the aciton');
            }
          }
        }),
        template: '{{x-child poke="poke"}}'
      });

      this.addComponent('x-child', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        })
      });

      this.addTemplate('application', '{{x-parent}}');
      this.add('controller:application', _controller.default.extend({
        send() {
          throw new Error('controller action should not be called');
        }
      }));

      return this.visit('/').then(() => {
        expectSendActionDeprecation(() => component.sendAction('poke'));
      });
    }
  });

  (0, _testCase.moduleFor)('Components test: sendAction of a closure action', class extends _testCase.RenderingTest {
    ['@test action should be called'](assert) {
      assert.expect(2);
      let component;

      this.registerComponent('inner-component', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: _helpers.Component.extend({
          outerSubmit() {
            assert.ok(true, 'outerSubmit called');
          }
        }),
        template: '{{inner-component submitAction=(action outerSubmit)}}'
      });

      this.render('{{outer-component}}');
      expectSendActionDeprecation(() => {
        this.runTask(() => component.sendAction('submitAction'));
      });
    }

    ['@test contexts passed to sendAction are appended to the bound arguments on a closure action']() {
      let first = 'mitch';
      let second = 'martin';
      let third = 'matt';
      let fourth = 'wacky wycats';

      let innerComponent;
      let actualArgs;

      this.registerComponent('inner-component', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            innerComponent = this;
          }
        }),
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: _helpers.Component.extend({
          third,
          actions: {
            outerSubmit() {
              actualArgs = [...arguments];
            }
          }
        }),
        template: `{{inner-component innerSubmit=(action (action "outerSubmit" "${first}") "${second}" third)}}`
      });

      this.render('{{outer-component}}');
      expectSendActionDeprecation(() => {
        this.runTask(() => innerComponent.sendAction('innerSubmit', fourth));
      });

      this.assert.deepEqual(actualArgs, [first, second, third, fourth], 'action has the correct args');
    }
  });

  (0, _testCase.moduleFor)('Components test: send', class extends _testCase.RenderingTest {
    ['@test sending to undefined actions triggers an error'](assert) {
      assert.expect(2);

      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            component = this;
          },
          actions: {
            foo(message) {
              assert.equal('bar', message);
            }
          }
        })
      });

      this.render('{{foo-bar}}');

      this.runTask(() => component.send('foo', 'bar'));

      expectAssertion(() => {
        return component.send('baz', 'bar');
      }, /had no action handler for: baz/);
    }

    ['@test `send` will call send from a target if it is defined']() {
      let component;
      let target = {
        send: (message, payload) => {
          this.assert.equal('foo', message);
          this.assert.equal('baz', payload);
        }
      };

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            component = this;
          },
          target
        })
      });

      this.render('{{foo-bar}}');

      this.runTask(() => component.send('foo', 'baz'));
    }

    ['@test a handled action can be bubbled to the target for continued processing']() {
      this.assert.expect(2);

      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          },
          actions: {
            poke: () => {
              this.assert.ok(true, 'component action called');
              return true;
            }
          },
          target: _controller.default.extend({
            actions: {
              poke: () => {
                this.assert.ok(true, 'action bubbled to controller');
              }
            }
          }).create()
        })
      });

      this.render('{{foo-bar poke="poke"}}');

      this.runTask(() => component.send('poke'));
    }

    ["@test action can be handled by a superclass' actions object"](assert) {
      this.assert.expect(4);

      let component;

      let SuperComponent = _helpers.Component.extend({
        actions: {
          foo() {
            assert.ok(true, 'foo');
          },
          bar(msg) {
            assert.equal(msg, 'HELLO');
          }
        }
      });

      let BarViewMixin = _metal.Mixin.create({
        actions: {
          bar(msg) {
            assert.equal(msg, 'HELLO');
            this._super(msg);
          }
        }
      });

      this.registerComponent('x-index', {
        ComponentClass: SuperComponent.extend(BarViewMixin, {
          init() {
            this._super(...arguments);
            component = this;
          },
          actions: {
            baz() {
              assert.ok(true, 'baz');
            }
          }
        })
      });

      this.render('{{x-index}}');

      this.runTask(() => {
        component.send('foo');
        component.send('bar', 'HELLO');
        component.send('baz');
      });
    }

    ['@test actions cannot be provided at create time'](assert) {
      this.registerComponent('foo-bar', _helpers.Component.extend());
      let ComponentFactory = this.owner.factoryFor('component:foo-bar');

      expectAssertion(() => {
        ComponentFactory.create({
          actions: {
            foo() {
              assert.ok(true, 'foo');
            }
          }
        });
      }, /`actions` must be provided at extend time, not at create time/);
      // but should be OK on an object that doesn't mix in Ember.ActionHandler
      _runtime.Object.create({
        actions: ['foo']
      });
    }

    ['@test asserts if called on a destroyed component']() {
      let component;

      this.registerComponent('rip-alley', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super();
            component = this;
          },

          toString() {
            return 'component:rip-alley';
          }
        })
      });

      this.render('{{#if shouldRender}}{{rip-alley}}{{/if}}', {
        shouldRender: true
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'shouldRender', false);
      });

      expectAssertion(() => {
        component.send('trigger-me-dead');
      }, "Attempted to call .send() with the action 'trigger-me-dead' on the destroyed object 'component:rip-alley'.");
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/template-only-components-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/environment'], function (_testCase, _testHelpers, _environment) {
  'use strict';

  class TemplateOnlyComponentsTest extends _testCase.RenderingTest {
    registerComponent(name, template) {
      super.registerComponent(name, { template, ComponentClass: null });
    }
  }

  (0, _testCase.moduleFor)('Components test: template-only components (glimmer components)', class extends TemplateOnlyComponentsTest {
    constructor() {
      super(...arguments);
      this._TEMPLATE_ONLY_GLIMMER_COMPONENTS = _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS;
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = true;
    }

    teardown() {
      super.teardown();
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = this._TEMPLATE_ONLY_GLIMMER_COMPONENTS;
    }

    ['@test it can render a template-only component']() {
      this.registerComponent('foo-bar', 'hello');

      this.render('{{foo-bar}}');

      this.assertInnerHTML('hello');

      this.assertStableRerender();
    }

    ['@feature(ember-glimmer-named-arguments) it can render named arguments']() {
      this.registerComponent('foo-bar', '|{{@foo}}|{{@bar}}|');

      this.render('{{foo-bar foo=foo bar=bar}}', {
        foo: 'foo',
        bar: 'bar'
      });

      this.assertInnerHTML('|foo|bar|');

      this.assertStableRerender();

      this.runTask(() => this.context.set('foo', 'FOO'));

      this.assertInnerHTML('|FOO|bar|');

      this.runTask(() => this.context.set('bar', 'BAR'));

      this.assertInnerHTML('|FOO|BAR|');

      this.runTask(() => this.context.setProperties({ foo: 'foo', bar: 'bar' }));

      this.assertInnerHTML('|foo|bar|');
    }

    ['@test it does not reflected arguments as properties']() {
      this.registerComponent('foo-bar', '|{{foo}}|{{this.bar}}|');

      this.render('{{foo-bar foo=foo bar=bar}}', {
        foo: 'foo',
        bar: 'bar'
      });

      this.assertInnerHTML('|||');

      this.assertStableRerender();

      this.runTask(() => this.context.set('foo', 'FOO'));

      this.assertInnerHTML('|||');

      this.runTask(() => this.context.set('bar', null));

      this.assertInnerHTML('|||');

      this.runTask(() => this.context.setProperties({ foo: 'foo', bar: 'bar' }));

      this.assertInnerHTML('|||');
    }

    ['@test it does not have curly component features']() {
      this.registerComponent('foo-bar', 'hello');

      this.render('{{foo-bar tagName="p" class=class}}', {
        class: 'foo bar'
      });

      this.assertInnerHTML('hello');

      this.assertStableRerender();

      this.runTask(() => this.context.set('class', 'foo'));

      this.assertInnerHTML('hello');

      this.runTask(() => this.context.set('class', null));

      this.assertInnerHTML('hello');

      this.runTask(() => this.context.set('class', 'foo bar'));

      this.assertInnerHTML('hello');
    }
  });

  (0, _testCase.moduleFor)('Components test: template-only components (curly components)', class extends TemplateOnlyComponentsTest {
    constructor() {
      super(...arguments);
      this._TEMPLATE_ONLY_GLIMMER_COMPONENTS = _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS;
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = false;
    }

    teardown() {
      super.teardown();
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = this._TEMPLATE_ONLY_GLIMMER_COMPONENTS;
    }

    ['@test it can render a template-only component']() {
      this.registerComponent('foo-bar', 'hello');

      this.render('{{foo-bar}}');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.assertStableRerender();
    }

    ['@feature(ember-glimmer-named-arguments) it can render named arguments']() {
      this.registerComponent('foo-bar', '|{{@foo}}|{{@bar}}|');

      this.render('{{foo-bar foo=foo bar=bar}}', {
        foo: 'foo',
        bar: 'bar'
      });

      this.assertComponentElement(this.firstChild, { content: '|foo|bar|' });

      this.assertStableRerender();

      this.runTask(() => this.context.set('foo', 'FOO'));

      this.assertComponentElement(this.firstChild, { content: '|FOO|bar|' });

      this.runTask(() => this.context.set('bar', 'BAR'));

      this.assertComponentElement(this.firstChild, { content: '|FOO|BAR|' });

      this.runTask(() => this.context.setProperties({ foo: 'foo', bar: 'bar' }));

      this.assertComponentElement(this.firstChild, { content: '|foo|bar|' });
    }

    ['@test it renders named arguments as reflected properties']() {
      this.registerComponent('foo-bar', '|{{foo}}|{{this.bar}}|');

      this.render('{{foo-bar foo=foo bar=bar}}', {
        foo: 'foo',
        bar: 'bar'
      });

      this.assertComponentElement(this.firstChild, { content: '|foo|bar|' });

      this.assertStableRerender();

      this.runTask(() => this.context.set('foo', 'FOO'));

      this.assertComponentElement(this.firstChild, { content: '|FOO|bar|' });

      this.runTask(() => this.context.set('bar', null));

      this.assertComponentElement(this.firstChild, { content: '|FOO||' });

      this.runTask(() => this.context.setProperties({ foo: 'foo', bar: 'bar' }));

      this.assertComponentElement(this.firstChild, { content: '|foo|bar|' });
    }

    ['@test it has curly component features']() {
      this.registerComponent('foo-bar', 'hello');

      this.render('{{foo-bar tagName="p" class=class}}', {
        class: 'foo bar'
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'p',
        attrs: { class: (0, _testHelpers.classes)('foo bar ember-view') },
        content: 'hello'
      });

      this.assertStableRerender();

      this.runTask(() => this.context.set('class', 'foo'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'p',
        attrs: { class: (0, _testHelpers.classes)('foo ember-view') },
        content: 'hello'
      });

      this.runTask(() => this.context.set('class', null));

      this.assertComponentElement(this.firstChild, {
        tagName: 'p',
        attrs: { class: (0, _testHelpers.classes)('ember-view') },
        content: 'hello'
      });

      this.runTask(() => this.context.set('class', 'foo bar'));

      this.assertComponentElement(this.firstChild, {
        tagName: 'p',
        attrs: { class: (0, _testHelpers.classes)('foo bar ember-view') },
        content: 'hello'
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/to-string-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer'], function (_testCase, _glimmer) {
  'use strict';

  (0, _testCase.moduleFor)('built-in component toString', class AbstractAppendTest extends _testCase.RenderingTest {
    '@test text-field has the correct toString value'(assert) {
      assert.strictEqual(_glimmer.TextField.toString(), '@ember/component/text-field');
    }

    '@test checkbox has the correct toString value'(assert) {
      assert.strictEqual(_glimmer.Checkbox.toString(), '@ember/component/checkbox');
    }

    '@test text-area has the correct toString value'(assert) {
      assert.strictEqual(_glimmer.TextArea.toString(), '@ember/component/text-area');
    }

    '@test component has the correct toString value'(assert) {
      assert.strictEqual(_glimmer.Component.toString(), '@ember/component');
    }

    '@test LinkTo has the correct toString value'(assert) {
      assert.strictEqual(_glimmer.LinkComponent.toString(), '@ember/routing/link-component');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/utils-test', ['@ember/controller', '@ember/-internals/views', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers'], function (_controller, _views, _testCase, _helpers) {
  'use strict';

  (0, _testCase.moduleFor)('View tree tests', class extends _testCase.ApplicationTest {
    constructor() {
      super(...arguments);

      this.addComponent('x-tagless', {
        ComponentClass: _helpers.Component.extend({
          tagName: ''
        }),

        template: '<div id="{{id}}">[{{id}}] {{#if isShowing}}{{yield}}{{/if}}</div>'
      });

      this.addComponent('x-toggle', {
        ComponentClass: _helpers.Component.extend({
          isExpanded: true,

          click() {
            this.toggleProperty('isExpanded');
            return false;
          }
        }),

        template: '[{{id}}] {{#if isExpanded}}{{yield}}{{/if}}'
      });

      let ToggleController = _controller.default.extend({
        isExpanded: true,

        actions: {
          toggle: function () {
            this.toggleProperty('isExpanded');
          }
        }
      });

      this.add('controller:application', ToggleController);

      this.addTemplate('application', `
      {{x-tagless id="root-1"}}

      {{#x-toggle id="root-2"}}
        {{x-toggle id="inner-1"}}

        {{#x-toggle id="inner-2"}}
          {{x-toggle id="inner-3"}}
        {{/x-toggle}}
      {{/x-toggle}}

      <button id="toggle-application" {{action "toggle"}}>Toggle</button>

      {{#if isExpanded}}
        {{x-toggle id="root-3"}}
      {{/if}}

      {{outlet}}
    `);

      this.add('controller:index', ToggleController.extend({
        isExpanded: false
      }));

      this.addTemplate('index', `
      {{x-tagless id="root-4"}}

      {{#x-toggle id="root-5" isExpanded=false}}
        {{x-toggle id="inner-4"}}

        {{#x-toggle id="inner-5"}}
          {{x-toggle id="inner-6"}}
        {{/x-toggle}}
      {{/x-toggle}}

      <button id="toggle-index" {{action "toggle"}}>Toggle</button>

      {{#if isExpanded}}
        {{x-toggle id="root-6"}}
      {{/if}}
    `);

      this.addTemplate('zomg', `
      {{x-tagless id="root-7"}}

      {{#x-toggle id="root-8"}}
        {{x-toggle id="inner-7"}}

        {{#x-toggle id="inner-8"}}
          {{x-toggle id="inner-9"}}
        {{/x-toggle}}
      {{/x-toggle}}

      {{#x-toggle id="root-9"}}
        {{outlet}}
      {{/x-toggle}}
    `);

      this.addTemplate('zomg.lol', `
      {{x-toggle id="inner-10"}}
    `);

      this.router.map(function () {
        this.route('zomg', function () {
          this.route('lol');
        });
      });
    }

    ['@test getRootViews']() {
      return this.visit('/').then(() => {
        this.assertRootViews(['root-1', 'root-2', 'root-3', 'root-4', 'root-5']);

        this.runTask(() => this.$('#toggle-application').click());

        this.assertRootViews(['root-1', 'root-2', 'root-4', 'root-5']);

        this.runTask(() => {
          this.$('#toggle-application').click();
          this.$('#toggle-index').click();
        });

        this.assertRootViews(['root-1', 'root-2', 'root-3', 'root-4', 'root-5', 'root-6']);

        return this.visit('/zomg/lol');
      }).then(() => {
        this.assertRootViews(['root-1', 'root-2', 'root-3', 'root-7', 'root-8', 'root-9']);

        return this.visit('/');
      }).then(() => {
        this.assertRootViews(['root-1', 'root-2', 'root-3', 'root-4', 'root-5', 'root-6']);
      });
    }

    assertRootViews(ids) {
      let owner = this.applicationInstance;

      let actual = (0, _views.getRootViews)(owner).map(view => view.id).sort();
      let expected = ids.sort();

      this.assert.deepEqual(actual, expected, 'root views');
    }

    ['@test getChildViews']() {
      return this.visit('/').then(() => {
        this.assertChildViews('root-2', ['inner-1', 'inner-2']);
        this.assertChildViews('root-5', []);
        this.assertChildViews('inner-2', ['inner-3']);

        this.runTask(() => this.$('#root-2').click());

        this.assertChildViews('root-2', []);

        this.runTask(() => this.$('#root-5').click());

        this.assertChildViews('root-5', ['inner-4', 'inner-5']);
        this.assertChildViews('inner-5', ['inner-6']);

        return this.visit('/zomg');
      }).then(() => {
        this.assertChildViews('root-2', []);
        this.assertChildViews('root-8', ['inner-7', 'inner-8']);
        this.assertChildViews('inner-8', ['inner-9']);
        this.assertChildViews('root-9', []);

        this.runTask(() => this.$('#root-8').click());

        this.assertChildViews('root-8', []);

        return this.visit('/zomg/lol');
      }).then(() => {
        this.assertChildViews('root-2', []);
        this.assertChildViews('root-8', []);
        this.assertChildViews('root-9', ['inner-10']);

        return this.visit('/');
      }).then(() => {
        this.assertChildViews('root-2', []);
        this.assertChildViews('root-5', []);

        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#inner-2').click());

        this.assertChildViews('root-2', ['inner-1', 'inner-2']);
        this.assertChildViews('inner-2', []);
      });
    }

    ['@test getChildViews does not return duplicates']() {
      return this.visit('/').then(() => {
        this.assertChildViews('root-2', ['inner-1', 'inner-2']);

        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());
        this.runTask(() => this.$('#root-2').click());

        this.assertChildViews('root-2', ['inner-1', 'inner-2']);
      });
    }

    assertChildViews(parentId, childIds) {
      let parentView = this.viewFor(parentId);
      let childViews = (0, _views.getChildViews)(parentView);

      let actual = childViews.map(view => view.id).sort();
      let expected = childIds.sort();

      this.assert.deepEqual(actual, expected, `child views for #${parentId}`);
    }

    viewFor(id) {
      let owner = this.applicationInstance;
      let registry = owner.lookup('-view-registry:main');
      return registry[id];
    }
  });

  let hasGetClientRects, hasGetBoundingClientRect;
  let ClientRectListCtor, ClientRectCtor;

  (function () {
    if (document.createRange) {
      let range = document.createRange();

      if (range.getClientRects) {
        let clientRectsList = range.getClientRects();
        hasGetClientRects = true;
        ClientRectListCtor = clientRectsList && clientRectsList.constructor;
      }

      if (range.getBoundingClientRect) {
        let clientRect = range.getBoundingClientRect();
        hasGetBoundingClientRect = true;
        ClientRectCtor = clientRect && clientRect.constructor;
      }
    }
  })();

  (0, _testCase.moduleFor)('Bounds tests', class extends _testCase.RenderingTest {
    ['@test getViewBounds on a regular component'](assert) {
      let component;
      this.registerComponent('hi-mom', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: `<p>Hi, mom!</p>`
      });

      this.render(`{{hi-mom}}`);

      let { parentElement, firstNode, lastNode } = (0, _views.getViewBounds)(component);

      assert.equal(parentElement, this.element, 'a regular component should have the right parentElement');
      assert.equal(firstNode, component.element, 'a regular component should have a single node that is its element');
      assert.equal(lastNode, component.element, 'a regular component should have a single node that is its element');
    }

    ['@test getViewBounds on a tagless component'](assert) {
      let component;
      this.registerComponent('hi-mom', {
        ComponentClass: _helpers.Component.extend({
          tagName: '',
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: `<span id="start-node">Hi,</span> <em id="before-end-node">mom</em>!`
      });

      this.render(`{{hi-mom}}`);

      let { parentElement, firstNode, lastNode } = (0, _views.getViewBounds)(component);

      assert.equal(parentElement, this.element, 'a tagless component should have the right parentElement');
      assert.equal(firstNode, this.$('#start-node')[0], 'a tagless component should have a range enclosing all of its nodes');
      assert.equal(lastNode, this.$('#before-end-node')[0].nextSibling, 'a tagless component should have a range enclosing all of its nodes');
    }

    ['@test getViewClientRects'](assert) {
      if (!hasGetClientRects || !ClientRectListCtor) {
        assert.ok(true, 'The test environment does not support the DOM API required to run this test.');
        return;
      }

      let component;
      this.registerComponent('hi-mom', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: `<p>Hi, mom!</p>`
      });

      this.render(`{{hi-mom}}`);

      assert.ok((0, _views.getViewClientRects)(component) instanceof ClientRectListCtor);
    }

    ['@test getViewBoundingClientRect'](assert) {
      if (!hasGetBoundingClientRect || !ClientRectCtor) {
        assert.ok(true, 'The test environment does not support the DOM API required to run this test.');
        return;
      }

      let component;
      this.registerComponent('hi-mom', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: `<p>Hi, mom!</p>`
      });

      this.render(`{{hi-mom}}`);

      assert.ok((0, _views.getViewBoundingClientRect)(component) instanceof ClientRectCtor);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/web-component-fallback-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal'], function (_testCase, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Components test: web component fallback', class extends _testCase.RenderingTest {
    ['@test custom elements are rendered']() {
      let template = `<foo-bar some-attr="123">hello</foo-bar>`;

      this.render(template);

      this.assertHTML(template);

      this.assertStableRerender();
    }

    ['@test custom elements can have bound attributes']() {
      let template = `<foo-bar some-attr="{{name}}">hello</foo-bar>`;

      this.render(template, { name: 'Robert' });

      this.assertHTML(`<foo-bar some-attr="Robert">hello</foo-bar>`);

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Kris'));

      this.assertHTML(`<foo-bar some-attr="Kris">hello</foo-bar>`);

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Robert'));

      this.assertHTML(`<foo-bar some-attr="Robert">hello</foo-bar>`);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/components/will-destroy-element-hook-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-case'], function (_metal, _helpers, _testCase) {
  'use strict';

  (0, _testCase.moduleFor)('Component willDestroyElement hook', class extends _testCase.RenderingTest {
    ['@test it calls willDestroyElement when removed by if'](assert) {
      let didInsertElementCount = 0;
      let willDestroyElementCount = 0;
      let FooBarComponent = _helpers.Component.extend({
        didInsertElement() {
          didInsertElementCount++;
          assert.notEqual(this.element.parentNode, null, 'precond component is in DOM');
        },
        willDestroyElement() {
          willDestroyElementCount++;
          assert.notEqual(this.element.parentNode, null, 'has not been removed from DOM yet');
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: 'hello'
      });

      this.render('{{#if switch}}{{foo-bar}}{{/if}}', { switch: true });

      assert.equal(didInsertElementCount, 1, 'didInsertElement was called once');

      this.assertComponentElement(this.firstChild, { content: 'hello' });

      this.runTask(() => (0, _metal.set)(this.context, 'switch', false));

      assert.equal(willDestroyElementCount, 1, 'willDestroyElement was called once');

      this.assertText('');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/content-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/metal', '@ember/debug', '@ember/object/computed', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/views', '@ember/-internals/glimmer/tests/utils/helpers'], function (_testCase, _abstractTestCase, _metal, _debug, _computed, _runtime, _testHelpers, _views, _helpers) {
  'use strict';

  (0, _testCase.moduleFor)('Static content tests', class extends _testCase.RenderingTest {
    ['@test it can render a static text node']() {
      this.render('hello');
      let text1 = this.assertTextNode(this.firstChild, 'hello');

      this.runTask(() => this.rerender());

      let text2 = this.assertTextNode(this.firstChild, 'hello');

      this.assertSameNode(text1, text2);
    }

    ['@test it can render a static element']() {
      this.render('<p>hello</p>');
      let p1 = this.assertElement(this.firstChild, { tagName: 'p' });
      let text1 = this.assertTextNode(this.firstChild.firstChild, 'hello');

      this.runTask(() => this.rerender());

      let p2 = this.assertElement(this.firstChild, { tagName: 'p' });
      let text2 = this.assertTextNode(this.firstChild.firstChild, 'hello');

      this.assertSameNode(p1, p2);
      this.assertSameNode(text1, text2);
    }

    ['@test it can render a static template']() {
      let template = `
      <div class="header">
        <h1>Welcome to Ember.js</h1>
      </div>
      <div class="body">
        <h2>Why you should use Ember.js?</h2>
        <ol>
          <li>It's great</li>
          <li>It's awesome</li>
          <li>It's Ember.js</li>
        </ol>
      </div>
      <div class="footer">
        Ember.js is free, open source and always will be.
      </div>
    `;

      this.render(template);
      this.assertHTML(template);

      this.runTask(() => this.rerender());

      this.assertHTML(template);
    }
  }); /* globals EmberDev */


  class DynamicContentTest extends _testCase.RenderingTest {
    /* abstract */
    renderPath() /* path, context = {} */{
      throw new Error('Not implemented: `renderValues`');
    }

    assertIsEmpty() {
      this.assert.strictEqual(this.firstChild, null);
    }

    /* abstract */
    assertContent() /* content */{
      throw new Error('Not implemented: `assertContent`');
    }

    ['@test it can render a dynamic path']() {
      this.renderPath('message', { message: 'hello' });

      this.assertContent('hello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'goodbye'));

      this.assertContent('goodbye');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'hello'));

      this.assertContent('hello');
      this.assertInvariants();
    }

    ['@test resolves the string length properly']() {
      this.render('<p>{{foo.length}}</p>', { foo: undefined });

      this.assertHTML('<p></p>');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'foo'));

      this.assertHTML('<p>3</p>');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', ''));

      this.assertHTML('<p>0</p>');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', undefined));

      this.assertHTML('<p></p>');
    }

    ['@test resolves the array length properly']() {
      this.render('<p>{{foo.length}}</p>', { foo: undefined });

      this.assertHTML('<p></p>');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'foo', [1, 2, 3]));

      this.assertHTML('<p>3</p>');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', []));

      this.assertHTML('<p>0</p>');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', undefined));

      this.assertHTML('<p></p>');
    }

    ['@test it can render a capitalized path with no deprecation']() {
      expectNoDeprecation();

      this.renderPath('CaptializedPath', { CaptializedPath: 'no deprecation' });

      this.assertContent('no deprecation');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'CaptializedPath', 'still no deprecation'));

      this.assertContent('still no deprecation');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'CaptializedPath', 'no deprecation'));

      this.assertContent('no deprecation');
      this.assertInvariants();
    }

    ['@test it can render undefined dynamic paths']() {
      this.renderPath('name', {});

      this.assertIsEmpty();

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'foo-bar'));

      this.assertContent('foo-bar');

      this.runTask(() => (0, _metal.set)(this.context, 'name', undefined));

      this.assertIsEmpty();
    }

    ['@test it can render a deeply nested dynamic path']() {
      this.renderPath('a.b.c.d.e.f', {
        a: { b: { c: { d: { e: { f: 'hello' } } } } }
      });

      this.assertContent('hello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'a.b.c.d.e.f', 'goodbye'));

      this.assertContent('goodbye');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'a.b.c.d', { e: { f: 'aloha' } }));

      this.assertContent('aloha');
      this.assertInvariants();

      this.runTask(() => {
        (0, _metal.set)(this.context, 'a', { b: { c: { d: { e: { f: 'hello' } } } } });
      });

      this.assertContent('hello');
      this.assertInvariants();
    }

    ['@test it can render a computed property']() {
      let Formatter = _runtime.Object.extend({
        formattedMessage: (0, _metal.computed)('message', function () {
          return this.get('message').toUpperCase();
        })
      });

      let m = Formatter.create({ message: 'hello' });

      this.renderPath('m.formattedMessage', { m });

      this.assertContent('HELLO');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(m, 'message', 'goodbye'));

      this.assertContent('GOODBYE');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'm', Formatter.create({ message: 'hello' })));

      this.assertContent('HELLO');
      this.assertInvariants();
    }

    ['@test it can render a computed property with nested dependency']() {
      let Formatter = _runtime.Object.extend({
        formattedMessage: (0, _metal.computed)('messenger.message', function () {
          return this.get('messenger.message').toUpperCase();
        })
      });

      let m = Formatter.create({ messenger: { message: 'hello' } });

      this.renderPath('m.formattedMessage', { m });

      this.assertContent('HELLO');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(m, 'messenger.message', 'goodbye'));

      this.assertContent('GOODBYE');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'm', Formatter.create({ messenger: { message: 'hello' } })));

      this.assertContent('HELLO');
      this.assertInvariants();
    }

    ['@test it can read from a proxy object']() {
      this.renderPath('proxy.name', {
        proxy: _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })
      });

      this.assertContent('Tom Dale');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name', 'Yehuda Katz'));

      this.assertContent('Yehuda Katz');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', { name: 'Godfrey Chan' }));

      this.assertContent('Godfrey Chan');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.name', 'Stefan Penner'));

      this.assertContent('Stefan Penner');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', null));

      this.assertIsEmpty();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy', _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })));

      this.assertContent('Tom Dale');
      this.assertInvariants();
    }

    ['@test it can read from a nested path in a proxy object']() {
      this.renderPath('proxy.name.last', {
        proxy: _runtime.ObjectProxy.create({
          content: { name: { first: 'Tom', last: 'Dale' } }
        })
      });

      this.assertContent('Dale');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name.last', 'Cruise'));

      this.assertContent('Cruise');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name.first', 'Suri'));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name', { first: 'Yehuda', last: 'Katz' }));

      this.assertContent('Katz');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', {
        name: { first: 'Godfrey', last: 'Chan' }
      }));

      this.assertContent('Chan');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.name', { first: 'Stefan', last: 'Penner' }));

      this.assertContent('Penner');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy', null));

      this.assertIsEmpty();

      this.runTask(() => (0, _metal.set)(this.context, 'proxy', _runtime.ObjectProxy.create({
        content: { name: { first: 'Tom', last: 'Dale' } }
      })));

      this.assertContent('Dale');
      this.assertInvariants();
    }

    ['@test it can read from a path flipping between a proxy and a real object']() {
      this.renderPath('proxyOrObject.name.last', {
        proxyOrObject: _runtime.ObjectProxy.create({
          content: { name: { first: 'Tom', last: 'Dale' } }
        })
      });

      this.assertContent('Dale');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject', {
        name: { first: 'Tom', last: 'Dale' }
      }));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject.name.last', 'Cruise'));

      this.assertContent('Cruise');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject.name.first', 'Suri'));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject', {
        name: { first: 'Yehuda', last: 'Katz' }
      }));

      this.assertContent('Katz');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject', _runtime.ObjectProxy.create({
        content: { name: { first: 'Godfrey', last: 'Chan' } }
      })));

      this.assertContent('Chan');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject.content.name', {
        first: 'Stefan',
        last: 'Penner'
      }));

      this.assertContent('Penner');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject', null));

      this.assertIsEmpty();

      this.runTask(() => (0, _metal.set)(this.context, 'proxyOrObject', _runtime.ObjectProxy.create({
        content: { name: { first: 'Tom', last: 'Dale' } }
      })));

      this.assertContent('Dale');
      this.assertInvariants();
    }

    ['@test it can read from a path flipping between a real object and a proxy']() {
      this.renderPath('objectOrProxy.name.last', {
        objectOrProxy: { name: { first: 'Tom', last: 'Dale' } }
      });

      this.assertContent('Dale');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy', _runtime.ObjectProxy.create({
        content: { name: { first: 'Tom', last: 'Dale' } }
      })));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy.content.name.last', 'Cruise'));

      this.assertContent('Cruise');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy.content.name.first', 'Suri'));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy.content', {
        name: { first: 'Yehuda', last: 'Katz' }
      }));

      this.assertContent('Katz');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy', {
        name: { first: 'Godfrey', last: 'Chan' }
      }));

      this.assertContent('Chan');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy.name', {
        first: 'Stefan',
        last: 'Penner'
      }));

      this.assertContent('Penner');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy', null));

      this.assertIsEmpty();

      this.runTask(() => (0, _metal.set)(this.context, 'objectOrProxy', {
        name: { first: 'Tom', last: 'Dale' }
      }));

      this.assertContent('Dale');
      this.assertInvariants();
    }

    ['@test it can read from a null object']() {
      let nullObject = Object.create(null);
      nullObject['message'] = 'hello';

      this.renderPath('nullObject.message', { nullObject });

      this.assertContent('hello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(nullObject, 'message', 'goodbye'));

      this.assertContent('goodbye');
      this.assertInvariants();

      nullObject = Object.create(null);
      nullObject['message'] = 'hello';

      this.runTask(() => (0, _metal.set)(this.context, 'nullObject', nullObject));

      this.assertContent('hello');
      this.assertInvariants();
    }

    ['@test it can render a readOnly property of a path']() {
      let Messenger = _runtime.Object.extend({
        message: (0, _computed.readOnly)('a.b.c')
      });

      let messenger = Messenger.create({
        a: {
          b: {
            c: 'hello'
          }
        }
      });

      this.renderPath('messenger.message', { messenger });

      this.assertContent('hello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(messenger, 'a.b.c', 'hi'));

      this.assertContent('hi');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'messenger.a.b', {
        c: 'goodbye'
      }));

      this.assertContent('goodbye');
      this.assertInvariants();

      this.runTask(() => (0, _metal.set)(this.context, 'messenger', {
        message: 'hello'
      }));

      this.assertContent('hello');
      this.assertInvariants();
    }

    ['@test it can render a property on a function']() {
      let func = () => {};
      func.aProp = 'this is a property on a function';

      this.renderPath('func.aProp', { func });

      this.assertContent('this is a property on a function');

      this.assertStableRerender();

      // this.runTask(() => set(func, 'aProp', 'still a property on a function'));
      // this.assertContent('still a property on a function');
      // this.assertInvariants();

      // func = () => {};
      // func.aProp = 'a prop on a new function';

      // this.runTask(() => set(this.context, 'func', func));

      // this.assertContent('a prop on a new function');
      // this.assertInvariants();
    }
  }

  const EMPTY = {};

  class ContentTestGenerator {
    constructor(cases, tag = '@test') {
      this.cases = cases;
      this.tag = tag;
    }

    generate([value, expected, label]) {
      let tag = this.tag;
      label = label || value;

      if (expected === EMPTY) {
        return {
          [`${tag} rendering ${label}`]() {
            this.renderPath('value', { value });

            this.assertIsEmpty();

            this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

            this.assertContent('hello');

            this.runTask(() => (0, _metal.set)(this.context, 'value', value));

            this.assertIsEmpty();
          }
        };
      } else {
        return {
          [`${tag} rendering ${label}`]() {
            this.renderPath('value', { value });

            this.assertContent(expected);

            this.assertStableRerender();

            this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));
            this.assertContent('hello');
            this.assertInvariants();

            this.runTask(() => (0, _metal.set)(this.context, 'value', value));

            this.assertContent(expected);
            this.assertInvariants();
          }
        };
      }
    }
  }

  const SharedContentTestCases = new ContentTestGenerator([['foo', 'foo'], [0, '0'], [-0, '0', '-0'], [1, '1'], [-1, '-1'], [0.0, '0', '0.0'], [0.5, '0.5'], [undefined, EMPTY], [null, EMPTY], [true, 'true'], [false, 'false'], [NaN, 'NaN'], [new Date(2000, 0, 1), String(new Date(2000, 0, 1)), 'a Date object'], [Infinity, 'Infinity'], [1 / -0, '-Infinity'], [{ foo: 'bar' }, '[object Object]', `{ foo: 'bar' }`], [{
    toString() {
      return 'foo';
    }
  }, 'foo', 'an object with a custom toString function'], [{
    valueOf() {
      return 1;
    }
  }, '[object Object]', 'an object with a custom valueOf function'],

  // Escaping tests
  ['<b>Max</b><b>James</b>', '<b>Max</b><b>James</b>']]);

  let GlimmerContentTestCases = new ContentTestGenerator([[Object.create(null), EMPTY, 'an object with no toString']]);

  if (typeof Symbol !== 'undefined') {
    GlimmerContentTestCases.cases.push([Symbol('debug'), 'Symbol(debug)', 'a symbol']);
  }

  (0, _abstractTestCase.applyMixins)(DynamicContentTest, SharedContentTestCases, GlimmerContentTestCases);

  (0, _testCase.moduleFor)('Dynamic content tests (content position)', class extends DynamicContentTest {
    renderPath(path, context = {}) {
      this.render(`{{${path}}}`, context);
    }

    assertContent(content) {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one text node');
      this.assertTextNode(this.firstChild, content);
      // this.takeSnapshot();
    }

    ['@test it can render empty safe strings [GH#16314]']() {
      this.render('before {{value}} after', { value: (0, _helpers.htmlSafe)('hello') });

      this.assertHTML('before hello after');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'value', (0, _helpers.htmlSafe)('')));

      this.assertHTML('before <!----> after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', (0, _helpers.htmlSafe)('hello')));

      this.assertHTML('before hello after');
    }
  });

  (0, _testCase.moduleFor)('Dynamic content tests (content concat)', class extends DynamicContentTest {
    renderPath(path, context = {}) {
      this.render(`{{concat "" ${path} ""}}`, context);
    }

    assertContent(content) {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one text node');
      this.assertTextNode(this.firstChild, content);
    }
  });

  (0, _testCase.moduleFor)('Dynamic content tests (inside an element)', class extends DynamicContentTest {
    renderPath(path, context = {}) {
      this.render(`<p>{{${path}}}</p>`, context);
    }

    assertIsEmpty() {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one <p> tag');
      this.assertElement(this.firstChild, { tagName: 'p' });
      this.assertText('');
    }

    assertContent(content) {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one <p> tag');
      this.assertElement(this.firstChild, { tagName: 'p' });
      this.assertText(content);
    }
  });

  (0, _testCase.moduleFor)('Dynamic content tests (attribute position)', class extends DynamicContentTest {
    renderPath(path, context = {}) {
      this.render(`<div data-foo="{{${path}}}"></div>`, context);
    }

    assertIsEmpty() {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one <div> tag');
      this.assertElement(this.firstChild, { tagName: 'div', content: '' });
    }

    assertContent(content) {
      this.assert.strictEqual(this.nodesCount, 1, 'It should render exactly one <div> tag');
      this.assertElement(this.firstChild, {
        tagName: 'div',
        attrs: { 'data-foo': content },
        content: ''
      });
    }
  });

  class TrustedContentTest extends DynamicContentTest {
    assertIsEmpty() {
      this.assert.strictEqual(this.firstChild, null);
    }

    assertContent(content) {
      this.assertHTML(content);
    }

    assertStableRerender() {
      this.takeSnapshot();
      this.runTask(() => this.rerender());
      super.assertInvariants();
    }

    assertInvariants() {
      // If it's not stable, we will wipe out all the content and replace them,
      // so there are no invariants
    }
  }

  (0, _testCase.moduleFor)('Dynamic content tests (trusted)', class extends TrustedContentTest {
    renderPath(path, context = {}) {
      this.render(`{{{${path}}}}`, context);
    }

    ['@test updating trusted curlies']() {
      this.render('{{{htmlContent}}}{{{nested.htmlContent}}}', {
        htmlContent: '<b>Max</b>',
        nested: { htmlContent: '<b>James</b>' }
      });

      this.assertContent('<b>Max</b><b>James</b>');

      this.runTask(() => this.rerender());

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'htmlContent', '<i>M</i><u>a</u><s>x</s>'));

      this.assertContent('<i>M</i><u>a</u><s>x</s><b>James</b>');

      this.runTask(() => (0, _metal.set)(this.context, 'nested.htmlContent', 'Jammie'));

      this.assertContent('<i>M</i><u>a</u><s>x</s>Jammie');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'htmlContent', '<b>Max</b>');
        (0, _metal.set)(this.context, 'nested', { htmlContent: '<i>James</i>' });
      });

      this.assertContent('<b>Max</b><i>James</i>');
    }

    ['@test empty content in trusted curlies [GH#14978]']() {
      this.render('before {{{value}}} after', {
        value: 'hello'
      });

      this.assertContent('before hello after');

      this.runTask(() => this.rerender());

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'value', undefined));

      this.assertContent('before <!----> after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

      this.assertContent('before hello after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', null));

      this.assertContent('before <!----> after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

      this.assertContent('before hello after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', ''));

      this.assertContent('before <!----> after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

      this.assertContent('before hello after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', (0, _helpers.htmlSafe)('')));

      this.assertContent('before <!----> after');

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

      this.assertContent('before hello after');
    }
  });

  (0, _testCase.moduleFor)('Dynamic content tests (integration)', class extends _testCase.RenderingTest {
    ['@test it can render a dynamic template']() {
      let template = `
      <div class="header">
        <h1>Welcome to {{framework}}</h1>
      </div>
      <div class="body">
        <h2>Why you should use {{framework}}?</h2>
        <ol>
          <li>It's great</li>
          <li>It's awesome</li>
          <li>It's {{framework}}</li>
        </ol>
      </div>
      <div class="footer">
        {{framework}} is free, open source and always will be.
      </div>
    `;

      let ember = `
      <div class="header">
        <h1>Welcome to Ember.js</h1>
      </div>
      <div class="body">
        <h2>Why you should use Ember.js?</h2>
        <ol>
          <li>It's great</li>
          <li>It's awesome</li>
          <li>It's Ember.js</li>
        </ol>
      </div>
      <div class="footer">
        Ember.js is free, open source and always will be.
      </div>
    `;

      let react = `
      <div class="header">
        <h1>Welcome to React</h1>
      </div>
      <div class="body">
        <h2>Why you should use React?</h2>
        <ol>
          <li>It's great</li>
          <li>It's awesome</li>
          <li>It's React</li>
        </ol>
      </div>
      <div class="footer">
        React is free, open source and always will be.
      </div>
    `;

      this.render(template, {
        framework: 'Ember.js'
      });
      this.assertHTML(ember);

      this.runTask(() => this.rerender());

      this.assertHTML(ember);

      this.runTask(() => (0, _metal.set)(this.context, 'framework', 'React'));

      this.assertHTML(react);

      this.runTask(() => (0, _metal.set)(this.context, 'framework', 'Ember.js'));

      this.assertHTML(ember);
    }

    ['@test it should evaluate to nothing if part of the path is `undefined`']() {
      this.render('{{foo.bar.baz.bizz}}', {
        foo: {}
      });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: { baz: { bizz: 'Hey!' } }
      }));

      this.assertText('Hey!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: { baz: { bizz: 'Hello!' } }
      }));

      this.assertText('Hello!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('');
    }

    ['@test it should evaluate to nothing if part of the path is a primative']() {
      this.render('{{foo.bar.baz.bizz}}', {
        foo: { bar: true }
      });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: false
      }));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: 'Haha'
      }));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: null
      }));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: undefined
      }));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: 1
      }));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: { baz: { bizz: 'Hello!' } }
      }));

      this.assertText('Hello!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {
        bar: true
      }));

      this.assertText('');
    }

    ['@test can set dynamic href']() {
      this.render('<a href={{model.url}}>Example</a>', {
        model: {
          url: 'http://example.com'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'a',
        content: 'Example',
        attrs: { href: 'http://example.com' }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'a',
        content: 'Example',
        attrs: { href: 'http://example.com' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.url', 'http://linkedin.com'));

      this.assertElement(this.firstChild, {
        tagName: 'a',
        content: 'Example',
        attrs: { href: 'http://linkedin.com' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { url: 'http://example.com' }));

      this.assertElement(this.firstChild, {
        tagName: 'a',
        content: 'Example',
        attrs: { href: 'http://example.com' }
      });
    }

    ['@test quoteless class attributes update correctly']() {
      this.render('<div class={{if fooBar "foo-bar"}}>hello</div>', {
        fooBar: true
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', false));

      this.assertElement(this.firstChild, { tagName: 'div', content: 'hello' });

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', true));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });
    }

    ['@test quoted class attributes update correctly'](assert) {
      this.render('<div class="{{if fooBar "foo-bar"}}">hello</div>', {
        fooBar: true
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', false));

      assert.equal(this.firstChild.className, '');

      this.runTask(() => (0, _metal.set)(this.context, 'fooBar', true));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo-bar') }
      });
    }

    ['@test unquoted class attribute can contain multiple classes']() {
      this.render('<div class={{model.classes}}>hello</div>', {
        model: {
          classes: 'foo bar baz'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.classes', 'fizz bizz'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz bizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { classes: 'foo bar baz' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });
    }

    ['@test unquoted class attribute']() {
      this.render('<div class={{model.foo}}>hello</div>', {
        model: {
          foo: 'foo'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.foo', 'fizz'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { foo: 'foo' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });
    }

    ['@test quoted class attribute']() {
      this.render('<div class="{{model.foo}}">hello</div>', {
        model: {
          foo: 'foo'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.foo', 'fizz'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { foo: 'foo' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo') }
      });
    }

    ['@test quoted class attribute can contain multiple classes']() {
      this.render('<div class="{{model.classes}}">hello</div>', {
        model: {
          classes: 'foo bar baz'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.classes', 'fizz bizz'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz bizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { classes: 'foo bar baz' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar baz') }
      });
    }

    ['@test class attribute concats bound values']() {
      this.render('<div class="{{model.foo}} {{model.bar}} {{model.bizz}}">hello</div>', {
        model: {
          foo: 'foo',
          bar: 'bar',
          bizz: 'bizz'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar bizz') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar bizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.foo', 'fizz'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz bar bizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.bar', null));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('fizz bizz') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        foo: 'foo',
        bar: 'bar',
        bizz: 'bizz'
      }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar bizz') }
      });
    }

    ['@test class attribute accepts nested helpers, and updates']() {
      this.render(`<div class="{{if model.hasSize model.size}} {{if model.hasShape model.shape}}">hello</div>`, {
        model: {
          size: 'large',
          hasSize: true,
          hasShape: false,
          shape: 'round'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('large') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('large') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.hasShape', true));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('large round') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.hasSize', false));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('round') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        size: 'large',
        hasSize: true,
        hasShape: false,
        shape: 'round'
      }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('large') }
      });
    }

    ['@test Multiple dynamic classes']() {
      this.render('<div class="{{model.foo}} {{model.bar}} {{model.fizz}} {{model.baz}}">hello</div>', {
        model: {
          foo: 'foo',
          bar: 'bar',
          fizz: 'fizz',
          baz: 'baz'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar fizz baz') }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar fizz baz') }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.foo', null);
        (0, _metal.set)(this.context, 'model.fizz', null);
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('bar baz') }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model', {
          foo: 'foo',
          bar: 'bar',
          fizz: 'fizz',
          baz: 'baz'
        });
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: (0, _testHelpers.classes)('foo bar fizz baz') }
      });
    }

    ['@test classes are ordered: See issue #9912']() {
      this.render('<div class="{{model.foo}}  static   {{model.bar}}">hello</div>', {
        model: {
          foo: 'foo',
          bar: 'bar'
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: 'foo  static   bar' }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: 'foo  static   bar' }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.bar', null);
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: 'foo  static   ' }
      });

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model', {
          foo: 'foo',
          bar: 'bar'
        });
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: 'hello',
        attrs: { class: 'foo  static   bar' }
      });
    }
  });

  let warnings, originalWarn;
  class StyleTest extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);
      warnings = [];
      originalWarn = (0, _debug.getDebugFunction)('warn');
      (0, _debug.setDebugFunction)('warn', function (message, test) {
        if (!test) {
          warnings.push(message);
        }
      });
    }

    teardown() {
      super.teardown(...arguments);
      (0, _debug.setDebugFunction)('warn', originalWarn);
    }

    assertStyleWarning(style) {
      this.assert.deepEqual(warnings, [(0, _views.constructStyleDeprecationMessage)(style)]);
    }

    assertNoWarning() {
      this.assert.deepEqual(warnings, []);
    }
  }

  (0, _testCase.moduleFor)('Inline style tests', class extends StyleTest {
    ['@test can set dynamic style']() {
      this.render('<div style={{model.style}}></div>', {
        model: {
          style: (0, _helpers.htmlSafe)('width: 60px;')
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.style', 'height: 60px;'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'height: 60px;' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.style', null));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: {}
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { style: 'width: 60px;' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });
    }

    ['@test can set dynamic style with -html-safe']() {
      this.render('<div style={{-html-safe model.style}}></div>', {
        model: {
          style: (0, _helpers.htmlSafe)('width: 60px;')
        }
      });

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });

      this.runTask(() => this.rerender());

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.style', 'height: 60px;'));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'height: 60px;' }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { style: 'width: 60px;' }));

      this.assertElement(this.firstChild, {
        tagName: 'div',
        content: '',
        attrs: { style: 'width: 60px;' }
      });
    }
  });

  if (!EmberDev.runningProdBuild) {
    (0, _testCase.moduleFor)('Inline style tests - warnings', class extends StyleTest {
      ['@test specifying <div style={{userValue}}></div> generates a warning']() {
        let userValue = 'width: 42px';
        this.render('<div style={{userValue}}></div>', {
          userValue
        });

        this.assertStyleWarning(userValue);
      }

      ['@test specifying `attributeBindings: ["style"]` generates a warning']() {
        let FooBarComponent = _helpers.Component.extend({
          attributeBindings: ['style']
        });

        this.registerComponent('foo-bar', {
          ComponentClass: FooBarComponent,
          template: 'hello'
        });
        let userValue = 'width: 42px';
        this.render('{{foo-bar style=userValue}}', {
          userValue
        });

        this.assertStyleWarning(userValue);
      }

      ['@test specifying `<div style={{{userValue}}}></div>` works properly without a warning']() {
        this.render('<div style={{{userValue}}}></div>', {
          userValue: 'width: 42px'
        });

        this.assertNoWarning();
      }

      ['@test specifying `<div style={{userValue}}></div>` works properly with a SafeString']() {
        this.render('<div style={{userValue}}></div>', {
          userValue: new _helpers.SafeString('width: 42px')
        });

        this.assertNoWarning();
      }

      ['@test null value do not generate htmlsafe warning']() {
        this.render('<div style={{userValue}}></div>', {
          userValue: null
        });

        this.assertNoWarning();
      }

      ['@test undefined value do not generate htmlsafe warning']() {
        this.render('<div style={{userValue}}></div>');

        this.assertNoWarning();
      }

      ['@test no warnings are triggered when using `-html-safe`']() {
        this.render('<div style={{-html-safe userValue}}></div>', {
          userValue: 'width: 42px'
        });

        this.assertNoWarning();
      }

      ['@test no warnings are triggered when a safe string is quoted']() {
        this.render('<div style="{{userValue}}"></div>', {
          userValue: new _helpers.SafeString('width: 42px')
        });

        this.assertNoWarning();
      }

      ['@test binding warning is triggered when an unsafe string is quoted']() {
        let userValue = 'width: 42px';
        this.render('<div style="{{userValue}}"></div>', {
          userValue
        });

        this.assertStyleWarning(userValue);
      }

      ['@test binding warning is triggered when a safe string for a complete property is concatenated in place']() {
        let userValue = 'width: 42px';
        this.render('<div style="color: green; {{userValue}}"></div>', {
          userValue: new _helpers.SafeString('width: 42px')
        });

        this.assertStyleWarning(`color: green; ${userValue}`);
      }

      ['@test binding warning is triggered when a safe string for a value is concatenated in place']() {
        let userValue = '42px';
        this.render('<div style="color: green; width: {{userValue}}"></div>', {
          userValue: new _helpers.SafeString(userValue)
        });

        this.assertStyleWarning(`color: green; width: ${userValue}`);
      }

      ['@test binding warning is triggered when a safe string for a property name is concatenated in place']() {
        let userValue = 'width';
        this.render('<div style="color: green; {{userProperty}}: 42px"></div>', {
          userProperty: new _helpers.SafeString(userValue)
        });

        this.assertStyleWarning(`color: green; ${userValue}: 42px`);
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/custom-component-manager-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/runtime', '@ember/-internals/metal', '@ember/-internals/glimmer'], function (_testCase, _runtime, _metal, _glimmer) {
  'use strict';

  if (true /* GLIMMER_CUSTOM_COMPONENT_MANAGER */) {
      let BasicComponentManager = _runtime.Object.extend({
        capabilities: (0, _glimmer.capabilities)('3.4'),

        createComponent(factory, args) {
          return factory.create({ args });
        },

        updateComponent(component, args) {
          (0, _metal.set)(component, 'args', args);
        },

        getContext(component) {
          return component;
        }
      });

      class ComponentManagerTest extends _testCase.RenderingTest {
        constructor(assert) {
          super(...arguments);

          this.registerComponentManager('basic', BasicComponentManager);

          this.registerComponentManager('instrumented-full', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4', {
              destructor: true,
              asyncLifecycleCallbacks: true
            }),

            createComponent(factory, args) {
              assert.step('createComponent');
              return factory.create({ args });
            },

            updateComponent(component, args) {
              assert.step('updateComponent');
              (0, _metal.set)(component, 'args', args);
            },

            destroyComponent(component) {
              assert.step('destroyComponent');
              component.destroy();
            },

            getContext(component) {
              assert.step('getContext');
              return component;
            },

            didCreateComponent(component) {
              assert.step('didCreateComponent');
              component.didRender();
            },

            didUpdateComponent(component) {
              assert.step('didUpdateComponent');
              component.didUpdate();
            }
          }));
        }
      }

      (0, _testCase.moduleFor)('Component Manager - Curly Invocation', class extends ComponentManagerTest {
        ['@test it can render a basic component with custom component manager']() {
          let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
            greeting: 'hello'
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{greeting}} world</p>`,
            ComponentClass
          });

          this.render('{{foo-bar}}');

          this.assertHTML(`<p>hello world</p>`);
        }

        ['@test it can render a basic component with custom component manager with a factory']() {
          let ComponentClass = (0, _glimmer.setComponentManager)(() => BasicComponentManager.create(), _runtime.Object.extend({
            greeting: 'hello'
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{greeting}} world</p>`,
            ComponentClass
          });

          this.render('{{foo-bar}}');

          this.assertHTML(`<p>hello world</p>`);
        }

        ['@test it can have no template context']() {
          this.registerComponentManager('pseudo-template-only', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4'),

            createComponent() {
              return null;
            },

            updateComponent() {},

            getContext() {
              return null;
            }
          }));

          let ComponentClass = (0, _glimmer.setComponentManager)('pseudo-template-only', {});

          this.registerComponent('foo-bar', {
            template: `<p>{{@greeting}} world</p>`,
            ComponentClass
          });

          this.render('{{foo-bar greeting="hello"}}');

          this.assertHTML(`<p>hello world</p>`);
        }

        ['@test it can discover component manager through inheritance - ES Classes']() {
          this.registerComponentManager('test', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4'),

            createComponent(Factory, args) {
              return new Factory(args);
            },

            updateComponent() {},

            getContext(component) {
              return component;
            }
          }));

          class Base {}
          (0, _glimmer.setComponentManager)('test', Base);
          class Child extends Base {}
          class Grandchild extends Child {
            constructor() {
              super();
              this.name = 'grandchild';
            }
          }

          this.registerComponent('foo-bar', {
            template: `{{this.name}}`,
            ComponentClass: Grandchild
          });

          this.render('{{foo-bar}}');

          this.assertHTML(`grandchild`);
        }

        ['@test it can discover component manager through inheritance - Ember Object']() {
          let Parent = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend());
          let Child = Parent.extend();
          let Grandchild = Child.extend({
            init() {
              this._super(...arguments);
              this.name = 'grandchild';
            }
          });

          this.registerComponent('foo-bar', {
            template: `{{this.name}}`,
            ComponentClass: Grandchild
          });

          this.render('{{foo-bar}}');

          this.assertHTML(`grandchild`);
        }

        ['@test it can customize the template context']() {
          let customContext = {
            greeting: 'goodbye'
          };

          this.registerComponentManager('test', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4'),

            createComponent(factory) {
              return factory.create();
            },

            getContext() {
              return customContext;
            },

            updateComponent() {}
          }));

          let ComponentClass = (0, _glimmer.setComponentManager)('test', _runtime.Object.extend({
            greeting: 'hello',
            count: 1234
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{greeting}} world {{count}}</p>`,
            ComponentClass
          });

          this.render('{{foo-bar}}');

          this.assertHTML(`<p>goodbye world </p>`);

          this.runTask(() => (0, _metal.set)(customContext, 'greeting', 'sayonara'));

          this.assertHTML(`<p>sayonara world </p>`);
        }

        ['@test it can set arguments on the component instance']() {
          let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
            salutation: (0, _metal.computed)('args.named.firstName', 'args.named.lastName', function () {
              return this.args.named.firstName + ' ' + this.args.named.lastName;
            })
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{salutation}}</p>`,
            ComponentClass
          });

          this.render('{{foo-bar firstName="Yehuda" lastName="Katz"}}');

          this.assertHTML(`<p>Yehuda Katz</p>`);
        }

        ['@test arguments are updated if they change']() {
          let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
            salutation: (0, _metal.computed)('args.named.firstName', 'args.named.lastName', function () {
              return this.args.named.firstName + ' ' + this.args.named.lastName;
            })
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{salutation}}</p>`,
            ComponentClass
          });

          this.render('{{foo-bar firstName=firstName lastName=lastName}}', {
            firstName: 'Yehuda',
            lastName: 'Katz'
          });

          this.assertHTML(`<p>Yehuda Katz</p>`);

          this.runTask(() => (0, _metal.setProperties)(this.context, {
            firstName: 'Chad',
            lastName: 'Hietala'
          }));

          this.assertHTML(`<p>Chad Hietala</p>`);
        }

        ['@test it can opt-in to running destructor'](assert) {
          this.registerComponentManager('test', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4', {
              destructor: true
            }),

            createComponent(factory) {
              assert.step('createComponent');
              return factory.create();
            },

            getContext(component) {
              return component;
            },

            updateComponent() {},

            destroyComponent(component) {
              assert.step('destroyComponent');
              component.destroy();
            }
          }));

          let ComponentClass = (0, _glimmer.setComponentManager)('test', _runtime.Object.extend({
            greeting: 'hello',
            destroy() {
              assert.step('component.destroy()');
              this._super(...arguments);
            }
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{greeting}} world</p>`,
            ComponentClass
          });

          this.render('{{#if show}}{{foo-bar}}{{/if}}', { show: true });

          this.assertHTML(`<p>hello world</p>`);

          this.runTask(() => this.context.set('show', false));

          this.assertText('');

          assert.verifySteps(['createComponent', 'destroyComponent', 'component.destroy()']);
        }

        ['@test it can opt-in to running async lifecycle hooks'](assert) {
          this.registerComponentManager('test', _runtime.Object.extend({
            capabilities: (0, _glimmer.capabilities)('3.4', {
              asyncLifecycleCallbacks: true
            }),

            createComponent(factory, args) {
              assert.step('createComponent');
              return factory.create({ args });
            },

            updateComponent(component, args) {
              assert.step('updateComponent');
              (0, _metal.set)(component, 'args', args);
            },

            destroyComponent(component) {
              assert.step('destroyComponent');
              component.destroy();
            },

            getContext(component) {
              assert.step('getContext');
              return component;
            },

            didCreateComponent() {
              assert.step('didCreateComponent');
            },

            didUpdateComponent() {
              assert.step('didUpdateComponent');
            }
          }));

          let ComponentClass = (0, _glimmer.setComponentManager)('test', _runtime.Object.extend({
            greeting: 'hello'
          }));

          this.registerComponent('foo-bar', {
            template: `<p>{{greeting}} {{@name}}</p>`,
            ComponentClass
          });

          this.render('{{foo-bar name=name}}', { name: 'world' });

          this.assertHTML(`<p>hello world</p>`);
          assert.verifySteps(['createComponent', 'getContext', 'didCreateComponent']);

          this.runTask(() => this.context.set('name', 'max'));
          this.assertHTML(`<p>hello max</p>`);
          assert.verifySteps(['updateComponent', 'didUpdateComponent']);
        }
      });

      if (true /* EMBER_GLIMMER_ANGLE_BRACKET_INVOCATION */) {
          (0, _testCase.moduleFor)('Component Manager - Angle Invocation', class extends ComponentManagerTest {
            ['@test it can render a basic component with custom component manager']() {
              let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
                greeting: 'hello'
              }));

              this.registerComponent('foo-bar', {
                template: `<p>{{greeting}} world</p>`,
                ComponentClass
              });

              this.render('<FooBar />');

              this.assertHTML(`<p>hello world</p>`);
            }

            ['@test it can set arguments on the component instance']() {
              let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
                salutation: (0, _metal.computed)('args.named.firstName', 'args.named.lastName', function () {
                  return this.args.named.firstName + ' ' + this.args.named.lastName;
                })
              }));

              this.registerComponent('foo-bar', {
                template: `<p>{{salutation}}</p>`,
                ComponentClass
              });

              this.render('<FooBar @firstName="Yehuda" @lastName="Katz" />');

              this.assertHTML(`<p>Yehuda Katz</p>`);
            }

            ['@test it can pass attributes']() {
              let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend());

              this.registerComponent('foo-bar', {
                template: `<p ...attributes>Hello world!</p>`,
                ComponentClass
              });

              this.render('<FooBar data-test="foo" />');

              this.assertHTML(`<p data-test="foo">Hello world!</p>`);
            }

            ['@test arguments are updated if they change']() {
              let ComponentClass = (0, _glimmer.setComponentManager)('basic', _runtime.Object.extend({
                salutation: (0, _metal.computed)('args.named.firstName', 'args.named.lastName', function () {
                  return this.args.named.firstName + ' ' + this.args.named.lastName;
                })
              }));

              this.registerComponent('foo-bar', {
                template: `<p>{{salutation}}</p>`,
                ComponentClass
              });

              this.render('<FooBar @firstName={{firstName}} @lastName={{lastName}} />', {
                firstName: 'Yehuda',
                lastName: 'Katz'
              });

              this.assertHTML(`<p>Yehuda Katz</p>`);

              this.runTask(() => (0, _metal.setProperties)(this.context, {
                firstName: 'Chad',
                lastName: 'Hietala'
              }));

              this.assertHTML(`<p>Chad Hietala</p>`);
            }

            ['@test updating attributes triggers didUpdateComponent'](assert) {
              let ComponentClass = (0, _glimmer.setComponentManager)('instrumented-full', _runtime.Object.extend({
                didRender() {},
                didUpdate() {}
              }));

              this.registerComponent('foo-bar', {
                template: `<p ...attributes>Hello world!</p>`,
                ComponentClass
              });

              this.render('<FooBar data-test={{value}} />', { value: 'foo' });

              this.assertHTML(`<p data-test="foo">Hello world!</p>`);
              assert.verifySteps(['createComponent', 'getContext', 'didCreateComponent']);

              this.runTask(() => this.context.set('value', 'bar'));
              assert.verifySteps(['didUpdateComponent']);
            }
          });
        }
    }
});
enifed('@ember/-internals/glimmer/tests/integration/custom-modifier-manager-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/runtime', '@ember/canary-features', '@ember/-internals/glimmer', '@ember/-internals/metal'], function (_testCase, _runtime, _canaryFeatures, _glimmer, _metal) {
  'use strict';

  if (_canaryFeatures.GLIMMER_MODIFIER_MANAGER) {
    class ModifierManagerTest extends _testCase.RenderingTest {}

    class CustomModifierManager {
      constructor(owner) {
        this.owner = owner;
      }

      createModifier(factory, args) {
        return factory.create(args);
      }

      installModifier(instance, element, args) {
        instance.element = element;
        let { positional, named } = args;
        instance.didInsertElement(positional, named);
      }

      updateModifier(instance, args) {
        let { positional, named } = args;
        instance.didUpdate(positional, named);
      }

      destroyModifier(instance) {
        instance.willDestroyElement();
      }
    }

    (0, _testCase.moduleFor)('Basic Custom Modifier Manager', class extends ModifierManagerTest {
      '@test can register a custom element modifier and render it'(assert) {
        let ModifierClass = (0, _glimmer.setModifierManager)(owner => {
          return new CustomModifierManager(owner);
        }, _runtime.Object.extend({
          didInsertElement() {},
          didUpdate() {},
          willDestroyElement() {}
        }));

        this.registerModifier('foo-bar', ModifierClass.extend({
          didInsertElement() {
            assert.ok(true, 'Called didInsertElement');
          }
        }));

        this.render('<h1 {{foo-bar}}>hello world</h1>');
        this.assertHTML(`<h1>hello world</h1>`);
      }

      '@test custom lifecycle hooks'(assert) {
        assert.expect(9);
        let ModifierClass = (0, _glimmer.setModifierManager)(owner => {
          return new CustomModifierManager(owner);
        }, _runtime.Object.extend({
          didInsertElement() {},
          didUpdate() {},
          willDestroyElement() {}
        }));

        this.registerModifier('foo-bar', ModifierClass.extend({
          didUpdate([truthy]) {
            assert.ok(true, 'Called didUpdate');
            assert.equal(truthy, 'true', 'gets updated args');
          },
          didInsertElement([truthy]) {
            assert.ok(true, 'Called didInsertElement');
            assert.equal(truthy, true, 'gets initial args');
          },
          willDestroyElement() {
            assert.ok(true, 'Called willDestroyElement');
          }
        }));

        this.render('{{#if truthy}}<h1 {{foo-bar truthy}}>hello world</h1>{{/if}}', {
          truthy: true
        });
        this.assertHTML(`<h1>hello world</h1>`);

        this.runTask(() => (0, _metal.set)(this.context, 'truthy', 'true'));

        this.runTask(() => (0, _metal.set)(this.context, 'truthy', false));

        this.runTask(() => (0, _metal.set)(this.context, 'truthy', true));
      }

      '@test associates manager even through an inheritance structure'(assert) {
        assert.expect(5);
        let ModifierClass = (0, _glimmer.setModifierManager)(owner => {
          return new CustomModifierManager(owner);
        }, _runtime.Object.extend({
          didInsertElement() {},
          didUpdate() {},
          willDestroyElement() {}
        }));

        ModifierClass = ModifierClass.extend({
          didInsertElement([truthy]) {
            this._super(...arguments);
            assert.ok(true, 'Called didInsertElement');
            assert.equal(truthy, true, 'gets initial args');
          }
        });

        this.registerModifier('foo-bar', ModifierClass.extend({
          didInsertElement([truthy]) {
            this._super(...arguments);
            assert.ok(true, 'Called didInsertElement');
            assert.equal(truthy, true, 'gets initial args');
          }
        }));

        this.render('<h1 {{foo-bar truthy}}>hello world</h1>', {
          truthy: true
        });
        this.assertHTML(`<h1>hello world</h1>`);
      }

      '@test can give consistent access to underlying DOM element'(assert) {
        assert.expect(4);
        let ModifierClass = (0, _glimmer.setModifierManager)(owner => {
          return new CustomModifierManager(owner);
        }, _runtime.Object.extend({
          didInsertElement() {},
          didUpdate() {},
          willDestroyElement() {}
        }));

        this.registerModifier('foo-bar', ModifierClass.extend({
          savedElement: undefined,
          didInsertElement() {
            assert.equal(this.element.tagName, 'H1');
            this.set('savedElement', this.element);
          },
          didUpdate() {
            assert.equal(this.element, this.savedElement);
          },
          willDestroyElement() {
            assert.equal(this.element, this.savedElement);
          }
        }));

        this.render('<h1 {{foo-bar truthy}}>hello world</h1>', {
          truthy: true
        });
        this.assertHTML(`<h1>hello world</h1>`);

        this.runTask(() => (0, _metal.set)(this.context, 'truthy', 'true'));
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/event-dispatcher-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/runloop', '@ember/instrumentation', '@ember/canary-features', '@ember/-internals/views', '@ember/-internals/utils'], function (_testCase, _helpers, _runloop, _instrumentation, _canaryFeatures, _views, _utils) {
  'use strict';

  let canDataTransfer = !!document.createEvent('HTMLEvents').dataTransfer;

  function fireNativeWithDataTransfer(node, type, dataTransfer) {
    let event = document.createEvent('HTMLEvents');
    event.initEvent(type, true, true);
    event.dataTransfer = dataTransfer;
    node.dispatchEvent(event);
  }

  (0, _testCase.moduleFor)('EventDispatcher', class extends _testCase.RenderingTest {
    ['@test events bubble view hierarchy for form elements'](assert) {
      let receivedEvent;

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          change(event) {
            receivedEvent = event;
          }
        }),
        template: `<input id="is-done" type="checkbox">`
      });

      this.render(`{{x-foo}}`);

      this.runTask(() => this.$('#is-done').trigger('change'));
      assert.ok(receivedEvent, 'change event was triggered');
      assert.strictEqual(receivedEvent.target, this.$('#is-done')[0]);
    }

    ['@test case insensitive events'](assert) {
      let receivedEvent;

      this.registerComponent('x-bar', {
        ComponentClass: _helpers.Component.extend({
          clicked(event) {
            receivedEvent = event;
          }
        }),
        template: `<button id="is-done" onclick={{action clicked}}>my button</button>`
      });

      this.render(`{{x-bar}}`);

      this.runTask(() => this.$('#is-done').trigger('click'));
      assert.ok(receivedEvent, 'change event was triggered');
      assert.strictEqual(receivedEvent.target, this.$('#is-done')[0]);
    }

    ['@test case sensitive events'](assert) {
      let receivedEvent;

      this.registerComponent('x-bar', {
        ComponentClass: _helpers.Component.extend({
          clicked(event) {
            receivedEvent = event;
          }
        }),
        template: `<button id="is-done" onClick={{action clicked}}>my button</button>`
      });

      this.render(`{{x-bar}}`);

      this.runTask(() => this.$('#is-done').trigger('click'));
      assert.ok(receivedEvent, 'change event was triggered');
      assert.strictEqual(receivedEvent.target, this.$('#is-done')[0]);
    }

    ['@test events bubble to parent view'](assert) {
      let receivedEvent;

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          change(event) {
            receivedEvent = event;
          }
        }),
        template: `{{yield}}`
      });

      this.registerComponent('x-bar', {
        ComponentClass: _helpers.Component.extend({
          change() {}
        }),
        template: `<input id="is-done" type="checkbox">`
      });

      this.render(`{{#x-foo}}{{x-bar}}{{/x-foo}}`);

      this.runTask(() => this.$('#is-done').trigger('change'));
      assert.ok(receivedEvent, 'change event was triggered');
      assert.strictEqual(receivedEvent.target, this.$('#is-done')[0]);
    }

    ['@test events bubbling up can be prevented'](assert) {
      let hasReceivedEvent;

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          change() {
            hasReceivedEvent = true;
          }
        }),
        template: `{{yield}}`
      });

      this.registerComponent('x-bar', {
        ComponentClass: _helpers.Component.extend({
          change() {
            return false;
          }
        }),
        template: `<input id="is-done" type="checkbox">`
      });

      this.render(`{{#x-foo}}{{x-bar}}{{/x-foo}}`);

      this.runTask(() => this.$('#is-done').trigger('change'));
      assert.notOk(hasReceivedEvent, 'change event has not been received');
    }

    ['@test event handlers are wrapped in a run loop'](assert) {
      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          change() {
            assert.ok((0, _runloop.getCurrentRunLoop)(), 'a run loop should have started');
          }
        }),
        template: `<input id="is-done" type="checkbox">`
      });

      this.render(`{{x-foo}}`);

      this.$('#is-done').trigger('click');
    }

    ['@test delegated event listeners work for mouseEnter/Leave'](assert) {
      let receivedEnterEvents = [];
      let receivedLeaveEvents = [];

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          mouseEnter(event) {
            receivedEnterEvents.push(event);
          },
          mouseLeave(event) {
            receivedLeaveEvents.push(event);
          }
        }),
        template: `<div id="inner"></div>`
      });

      this.render(`{{x-foo id="outer"}}`);

      let parent = this.element;
      let outer = this.$('#outer')[0];
      let inner = this.$('#inner')[0];

      // mouse moves over #outer
      this.runTask(() => {
        this.$(outer).trigger('mouseenter', { canBubble: false, relatedTarget: parent });
        this.$(outer).trigger('mouseover', { relatedTarget: parent });
        this.$(parent).trigger('mouseout', { relatedTarget: outer });
      });
      assert.equal(receivedEnterEvents.length, 1, 'mouseenter event was triggered');
      assert.strictEqual(receivedEnterEvents[0].target, outer);

      // mouse moves over #inner
      this.runTask(() => {
        this.$(inner).trigger('mouseover', { relatedTarget: outer });
        this.$(outer).trigger('mouseout', { relatedTarget: inner });
      });
      assert.equal(receivedEnterEvents.length, 1, 'mouseenter event was not triggered again');

      // mouse moves out of #inner
      this.runTask(() => {
        this.$(inner).trigger('mouseout', { relatedTarget: outer });
        this.$(outer).trigger('mouseover', { relatedTarget: inner });
      });
      assert.equal(receivedLeaveEvents.length, 0, 'mouseleave event was not triggered');

      // mouse moves out of #outer
      this.runTask(() => {
        this.$(outer).trigger('mouseleave', { canBubble: false, relatedTarget: parent });
        this.$(outer).trigger('mouseout', { relatedTarget: parent });
        this.$(parent).trigger('mouseover', { relatedTarget: outer });
      });
      assert.equal(receivedLeaveEvents.length, 1, 'mouseleave event was triggered');
      assert.strictEqual(receivedLeaveEvents[0].target, outer);
    }

    ['@test delegated event listeners work for mouseEnter/Leave with skipped events'](assert) {
      let receivedEnterEvents = [];
      let receivedLeaveEvents = [];

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          mouseEnter(event) {
            receivedEnterEvents.push(event);
          },
          mouseLeave(event) {
            receivedLeaveEvents.push(event);
          }
        }),
        template: `<div id="inner"></div>`
      });

      this.render(`{{x-foo id="outer"}}`);

      let parent = this.element;
      let outer = this.$('#outer')[0];
      let inner = this.$('#inner')[0];

      // we replicate fast mouse movement, where mouseover is fired directly in #inner, skipping #outer
      this.runTask(() => {
        this.$(outer).trigger('mouseenter', { canBubble: false, relatedTarget: parent });
        this.$(inner).trigger('mouseover', { relatedTarget: parent });
        this.$(parent).trigger('mouseout', { relatedTarget: inner });
      });
      assert.equal(receivedEnterEvents.length, 1, 'mouseenter event was triggered');
      assert.strictEqual(receivedEnterEvents[0].target, inner);

      // mouse moves out of #outer
      this.runTask(() => {
        this.$(outer).trigger('mouseleave', { canBubble: false, relatedTarget: parent });
        this.$(inner).trigger('mouseout', { relatedTarget: parent });
        this.$(parent).trigger('mouseover', { relatedTarget: inner });
      });
      assert.equal(receivedLeaveEvents.length, 1, 'mouseleave event was triggered');
      assert.strictEqual(receivedLeaveEvents[0].target, inner);
    }

    ['@test delegated event listeners work for mouseEnter/Leave with skipped events and subcomponent'](assert) {
      let receivedEnterEvents = [];
      let receivedLeaveEvents = [];

      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          mouseEnter(event) {
            receivedEnterEvents.push(event);
          },
          mouseLeave(event) {
            receivedLeaveEvents.push(event);
          }
        }),
        template: `{{yield}}`
      });

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend(),
        template: ``
      });

      this.render(`{{#x-outer id="outer"}}{{x-inner id="inner"}}{{/x-outer}}`);

      let parent = this.element;
      let outer = this.$('#outer')[0];
      let inner = this.$('#inner')[0];

      // we replicate fast mouse movement, where mouseover is fired directly in #inner, skipping #outer
      this.runTask(() => {
        this.$(outer).trigger('mouseenter', { canBubble: false, relatedTarget: parent });
        this.$(inner).trigger('mouseover', { relatedTarget: parent });
        this.$(parent).trigger('mouseout', { relatedTarget: inner });
      });
      assert.equal(receivedEnterEvents.length, 1, 'mouseenter event was triggered');
      assert.strictEqual(receivedEnterEvents[0].target, inner);

      // mouse moves out of #inner
      this.runTask(() => {
        this.$(outer).trigger('mouseleave', { canBubble: false, relatedTarget: parent });
        this.$(inner).trigger('mouseout', { relatedTarget: parent });
        this.$(parent).trigger('mouseover', { relatedTarget: inner });
      });

      assert.equal(receivedLeaveEvents.length, 1, 'mouseleave event was triggered');
      assert.strictEqual(receivedLeaveEvents[0].target, inner);
    }
  });

  (0, _testCase.moduleFor)('EventDispatcher#setup', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);

      let dispatcher = this.owner.lookup('event_dispatcher:main');
      (0, _runloop.run)(dispatcher, 'destroy');
      this.owner.__container__.reset('event_dispatcher:main');
      this.dispatcher = this.owner.lookup('event_dispatcher:main');
    }

    ['@test additional events can be specified'](assert) {
      this.dispatcher.setup({ myevent: 'myEvent' });

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          myEvent() {
            assert.ok(true, 'custom event was triggered');
          }
        }),
        template: `<p>Hello!</p>`
      });

      this.render(`{{x-foo}}`);

      this.$('div').trigger('myevent');
    }

    ['@test a rootElement can be specified'](assert) {
      this.element.innerHTML = '<div id="app"></div>';
      // this.$().append('<div id="app"></div>');
      this.dispatcher.setup({ myevent: 'myEvent' }, '#app');

      assert.ok(this.$('#app').hasClass('ember-application'), 'custom rootElement was used');
      assert.equal(this.dispatcher.rootElement, '#app', 'the dispatchers rootElement was updated');
    }

    ['@test default events can be disabled via `customEvents`'](assert) {
      this.dispatcher.setup({ click: null });

      this.registerComponent('x-foo', {
        ComponentClass: _helpers.Component.extend({
          click() {
            assert.ok(false, 'click method was called');
          },

          null() {
            assert.ok(false, 'null method was called');
          },

          doubleClick() {
            assert.ok(true, 'a non-disabled event is still handled properly');
          }
        }),

        template: `<p>Hello!</p>`
      });

      this.render(`{{x-foo}}`);

      this.$('div').trigger('click');
      this.$('div').trigger('dblclick');
    }

    ['@test throws if specified rootElement does not exist'](assert) {
      assert.throws(() => {
        this.dispatcher.setup({ myevent: 'myEvent' }, '#app');
      });
    }
  });

  if (_canaryFeatures.EMBER_IMPROVED_INSTRUMENTATION) {
    (0, _testCase.moduleFor)('EventDispatcher - Instrumentation', class extends _testCase.RenderingTest {
      teardown() {
        super.teardown();
        (0, _instrumentation.reset)();
      }

      ['@test instruments triggered events'](assert) {
        let clicked = 0;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click() {
              clicked++;
            }
          }),
          template: `<p>hello</p>`
        });

        this.render(`{{x-foo}}`);

        this.$('div').trigger('click');

        assert.equal(clicked, 1, 'precond - the click handler was invoked');

        let clickInstrumented = 0;
        (0, _instrumentation.subscribe)('interaction.click', {
          before() {
            clickInstrumented++;
            assert.equal(clicked, 1, 'invoked before event is handled');
          },
          after() {
            clickInstrumented++;
            assert.equal(clicked, 2, 'invoked after event is handled');
          }
        });

        let keypressInstrumented = 0;
        (0, _instrumentation.subscribe)('interaction.keypress', {
          before() {
            keypressInstrumented++;
          },
          after() {
            keypressInstrumented++;
          }
        });

        this.$('div').trigger('click');
        this.$('div').trigger('change');
        assert.equal(clicked, 2, 'precond - The click handler was invoked');
        assert.equal(clickInstrumented, 2, 'The click was instrumented');
        assert.strictEqual(keypressInstrumented, 0, 'The keypress was not instrumented');
      }
    });
  }

  if (canDataTransfer) {
    (0, _testCase.moduleFor)('EventDispatcher - Event Properties', class extends _testCase.RenderingTest {
      ['@test dataTransfer property is added to drop event'](assert) {
        let receivedEvent;
        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            drop(event) {
              receivedEvent = event;
            }
          })
        });

        this.render(`{{x-foo}}`);

        fireNativeWithDataTransfer(this.$('div')[0], 'drop', 'success');
        assert.equal(receivedEvent.dataTransfer, 'success');
      }
    });
  }

  if (_views.jQueryDisabled) {
    (0, _testCase.moduleFor)('EventDispatcher#native-events', class extends _testCase.RenderingTest {
      ['@test native events are passed when jQuery is not present'](assert) {
        let receivedEvent;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        assert.ok(receivedEvent, 'click event was triggered');
        assert.notOk(receivedEvent.originalEvent, 'event is not a jQuery.Event');
      }
    });
  } else {
    (0, _testCase.moduleFor)('EventDispatcher#jquery-events', class extends _testCase.RenderingTest {
      beforeEach() {
        this.jqueryIntegration = window.ENV._JQUERY_INTEGRATION;
      }

      afterEach() {
        window.ENV._JQUERY_INTEGRATION = this.jqueryIntegration;
      }

      ['@test jQuery events are passed when jQuery is present'](assert) {
        let receivedEvent;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        assert.ok(receivedEvent, 'click event was triggered');
        assert.ok(receivedEvent instanceof _views.jQuery.Event, 'event is a jQuery.Event');
      }

      [`@${_utils.HAS_NATIVE_PROXY ? 'test' : 'skip'} accessing jQuery.Event#originalEvent is deprecated`](assert) {
        let receivedEvent;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        expectDeprecation(() => {
          let { originalEvent } = receivedEvent;
          assert.ok(originalEvent, 'jQuery event has originalEvent property');
          assert.equal(originalEvent.type, 'click', 'properties of originalEvent are available');
        }, 'Accessing jQuery.Event specific properties is deprecated. Either use the ember-jquery-legacy addon to normalize events to native events, or explicitly opt into jQuery integration using @ember/optional-features.');
      }

      ['@test other jQuery.Event properties do not trigger deprecation'](assert) {
        let receivedEvent;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        expectNoDeprecation(() => {
          receivedEvent.stopPropagation();
          receivedEvent.stopImmediatePropagation();
          receivedEvent.preventDefault();
          assert.ok(receivedEvent.bubbles, 'properties of jQuery event are available');
          assert.equal(receivedEvent.type, 'click', 'properties of jQuery event are available');
        });
      }

      ['@test accessing jQuery.Event#originalEvent does not trigger deprecations when jquery integration is explicitly enabled'](assert) {
        let receivedEvent;
        window.ENV._JQUERY_INTEGRATION = true;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        expectNoDeprecation(() => {
          let { originalEvent } = receivedEvent;
          assert.ok(originalEvent, 'jQuery event has originalEvent property');
          assert.equal(originalEvent.type, 'click', 'properties of originalEvent are available');
        });
      }

      [`@${_utils.HAS_NATIVE_PROXY && true /* DEBUG */ ? 'test' : 'skip'} accessing jQuery.Event#__originalEvent does not trigger deprecations to support ember-jquery-legacy`](assert) {
        let receivedEvent;

        this.registerComponent('x-foo', {
          ComponentClass: _helpers.Component.extend({
            click(event) {
              receivedEvent = event;
            }
          }),
          template: `<button id="foo">bar</button>`
        });

        this.render(`{{x-foo}}`);

        this.runTask(() => this.$('#foo').click());
        expectNoDeprecation(() => {
          let { __originalEvent: originalEvent } = receivedEvent;
          assert.ok(originalEvent, 'jQuery event has __originalEvent property');
          assert.equal(originalEvent.type, 'click', 'properties of __originalEvent are available');
        });
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/-class-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/metal'], function (_testCase, _testHelpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{-class}}', class extends _testCase.RenderingTest {
    ['@test casts binding to dasherized class']() {
      this.registerComponent('foo-bar', { template: '' });
      this.render(`{{foo-bar class=(-class someTruth "someTruth")}}`, {
        someTruth: true
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'someTruth', false));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'someTruth', true));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });
    }

    ['@tests casts leaf path of binding to dasherized class']() {
      this.registerComponent('foo-bar', { template: '' });
      this.render(`{{foo-bar class=(-class model.someTruth "someTruth")}}`, {
        model: {
          someTruth: true
        }
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });

      this.runTask(() => this.rerender());

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model.someTruth', false));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('ember-view') }
      });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { someTruth: true }));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: (0, _testHelpers.classes)('some-truth ember-view') }
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/array-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal'], function (_testCase, _abstractTestCase, _helpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{array}}', class extends _testCase.RenderingTest {
    ['@test returns an array']() {
      this.render(_abstractTestCase.strip`
        {{#with (array "Sergio") as |people|}}
          {{#each people as |personName|}}
            {{personName}}
          {{/each}}
        {{/with}}`);

      this.assertText('Sergio');

      this.assertStableRerender();
    }

    ['@test can have more than one value']() {
      this.render(_abstractTestCase.strip`
        {{#with (array "Sergio" "Robert") as |people|}}
          {{#each people as |personName|}}
            {{personName}},
          {{/each}}
        {{/with}}`);

      this.assertText('Sergio,Robert,');

      this.assertStableRerender();
    }

    ['@test binds values when variables are used']() {
      this.render(_abstractTestCase.strip`{{#with (array personOne) as |people|}}
                {{#each people as |personName|}}
                  {{personName}}
                {{/each}}
              {{/with}}`, {
        personOne: 'Tom'
      });

      this.assertText('Tom');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'personOne', 'Yehuda'));
      this.assertText('Yehuda');

      this.runTask(() => (0, _metal.set)(this.context, 'personOne', 'Tom'));
      this.assertText('Tom');
    }

    ['@test binds multiple values when variables are used']() {
      this.render(_abstractTestCase.strip`{{#with (array personOne personTwo) as |people|}}
                {{#each people as |personName|}}
                  {{personName}},
                {{/each}}
              {{/with}}`, {
        personOne: 'Tom',
        personTwo: 'Yehuda'
      });

      this.assertText('Tom,Yehuda,');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'personOne', 'Sergio'));

      this.assertText('Sergio,Yehuda,');

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Tom'));

      this.assertText('Sergio,Tom,');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'personOne', 'Tom');
        (0, _metal.set)(this.context, 'personTwo', 'Yehuda');
      });

      this.assertText('Tom,Yehuda,');
    }

    ['@test array helpers can be nested']() {
      this.render(_abstractTestCase.strip`{{#with (array (array personOne personTwo)) as |listOfPeople|}}
                {{#each listOfPeople as |people|}}
                  List:
                  {{#each people as |personName|}}
                    {{personName}},
                  {{/each}}
                {{/each}}
              {{/with}}`, {
        personOne: 'Tom',
        personTwo: 'Yehuda'
      });

      this.assertText('List:Tom,Yehuda,');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'personOne', 'Chad'));

      this.assertText('List:Chad,Yehuda,');

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Balint'));

      this.assertText('List:Chad,Balint,');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'personOne', 'Tom');
        (0, _metal.set)(this.context, 'personTwo', 'Yehuda');
      });

      this.assertText('List:Tom,Yehuda,');
    }

    ['@test should yield hash of an array of internal properties']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
          this.model = { personOne: 'Chad' };
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (hash people=(array model.personOne))}}`
      });

      this.render(_abstractTestCase.strip`
        {{#foo-bar as |values|}}
          {{#each values.people as |personName|}}
            {{personName}}
          {{/each}}
        {{/foo-bar}}`);

      this.assertText('Chad');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model.personOne', 'Godfrey'));

      this.assertText('Godfrey');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model', { personOne: 'Chad' }));

      this.assertText('Chad');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model.personOne', 'Godfrey'));

      this.assertText('Godfrey');
    }

    ['@test should yield hash of an array of internal and external properties']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
          this.model = { personOne: 'Chad' };
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (hash people=(array model.personOne personTwo))}}`
      });

      this.render(_abstractTestCase.strip`{{#foo-bar personTwo=model.personTwo as |values|}}
                {{#each values.people as |personName|}}
                  {{personName}},
                {{/each}}
              {{/foo-bar}}`, {
        model: { personTwo: 'Tom' }
      });

      this.assertText('Chad,Tom,');

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'model.personOne', 'Godfrey');
        (0, _metal.set)(this.context, 'model.personTwo', 'Yehuda');
      });

      this.assertText('Godfrey,Yehuda,');

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'model', { personOne: 'Chad' });
        (0, _metal.set)(this.context, 'model', { personTwo: 'Tom' });
      });

      this.assertText('Chad,Tom,');
    }

    ['@test should render when passing as argument to a component invocation']() {
      let FooBarComponent = _helpers.Component.extend({});

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: _abstractTestCase.strip`
          {{#each people as |personName|}}
            {{personName}},
          {{/each}}`
      });

      this.render(_abstractTestCase.strip`{{foo-bar people=(array "Tom" personTwo)}}`, { personTwo: 'Chad' });

      this.assertText('Tom,Chad,');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Godfrey'));

      this.assertText('Tom,Godfrey,');

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Chad'));

      this.assertText('Tom,Chad,');
    }

    ['@test should return an entirely new array when any argument change']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: _abstractTestCase.strip`
          {{#each people as |personName|}}
            {{personName}},
          {{/each}}`
      });

      this.render(_abstractTestCase.strip`{{foo-bar people=(array "Tom" personTwo)}}`, { personTwo: 'Chad' });

      let firstArray = fooBarInstance.people;

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Godfrey'));

      this.assert.ok(firstArray !== fooBarInstance.people, 'should have created an entirely new array');
    }

    ['@test capture array values in JS to assert deep equal']() {
      let captured;

      this.registerHelper('capture', function ([array]) {
        captured = array;
        return 'captured';
      });

      this.render(`{{capture (array 'Tom' personTwo)}}`, { personTwo: 'Godfrey' });

      this.assert.deepEqual(captured, ['Tom', 'Godfrey']);

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Robert'));

      this.assert.deepEqual(captured, ['Tom', 'Robert']);

      this.runTask(() => (0, _metal.set)(this.context, 'personTwo', 'Godfrey'));

      this.assert.deepEqual(captured, ['Tom', 'Godfrey']);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/closure-action-test', ['@ember/instrumentation', '@ember/runloop', '@ember/-internals/metal', '@ember/canary-features', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/helpers'], function (_instrumentation, _runloop, _metal, _canaryFeatures, _testCase, _abstractTestCase, _helpers) {
  'use strict';

  if (_canaryFeatures.EMBER_IMPROVED_INSTRUMENTATION) {
    (0, _testCase.moduleFor)('Helpers test: closure {{action}} improved instrumentation', class extends _testCase.RenderingTest {
      subscribe(eventName, options) {
        this.subscriber = (0, _instrumentation.subscribe)(eventName, options);
      }

      teardown() {
        if (this.subscriber) {
          (0, _instrumentation.unsubscribe)(this.subscriber);
        }

        super.teardown();
      }

      ['@test interaction event subscriber should be passed parameters']() {
        let actionParam = 'So krispy';
        let beforeParameters = [];
        let afterParameters = [];

        let InnerComponent = _helpers.Component.extend({
          actions: {
            fireAction() {
              this.attrs.submit(actionParam);
            }
          }
        });

        let OuterComponent = _helpers.Component.extend({
          outerSubmit() {}
        });

        this.registerComponent('inner-component', {
          ComponentClass: InnerComponent,
          template: '<button id="instrument-button" {{action "fireAction"}}>What it do</button>'
        });

        this.registerComponent('outer-component', {
          ComponentClass: OuterComponent,
          template: '{{inner-component submit=(action outerSubmit)}}'
        });

        this.subscribe('interaction.ember-action', {
          before(name, timestamp, payload) {
            beforeParameters.push(payload.args);
          },
          after(name, timestamp, payload) {
            afterParameters.push(payload.args);
          }
        });

        this.render(`{{outer-component}}`);

        this.runTask(() => {
          this.$('#instrument-button').trigger('click');
        });

        this.assert.deepEqual(beforeParameters, [[], [actionParam]], 'instrumentation subscriber before function was passed closure action parameters');
        this.assert.deepEqual(afterParameters, [[actionParam], []], 'instrumentation subscriber after function was passed closure action parameters');
      }

      ['@test interaction event subscriber should be passed target']() {
        let beforeParameters = [];
        let afterParameters = [];

        let InnerComponent = _helpers.Component.extend({
          myProperty: 'inner-thing',
          actions: {
            fireAction() {
              this.attrs.submit();
            }
          }
        });

        let OuterComponent = _helpers.Component.extend({
          myProperty: 'outer-thing',
          outerSubmit() {}
        });

        this.registerComponent('inner-component', {
          ComponentClass: InnerComponent,
          template: '<button id="instrument-button" {{action "fireAction"}}>What it do</button>'
        });

        this.registerComponent('outer-component', {
          ComponentClass: OuterComponent,
          template: '{{inner-component submit=(action outerSubmit)}}'
        });

        this.subscribe('interaction.ember-action', {
          before(name, timestamp, payload) {
            beforeParameters.push(payload.target.get('myProperty'));
          },
          after(name, timestamp, payload) {
            afterParameters.push(payload.target.get('myProperty'));
          }
        });

        this.render(`{{outer-component}}`);

        this.runTask(() => {
          this.$('#instrument-button').trigger('click');
        });

        this.assert.deepEqual(beforeParameters, ['inner-thing', 'outer-thing'], 'instrumentation subscriber before function was passed target');
        this.assert.deepEqual(afterParameters, ['outer-thing', 'inner-thing'], 'instrumentation subscriber after function was passed target');
      }

      ['@test instrumented action should return value']() {
        let returnedValue = 'Chris P is so krispy';
        let actualReturnedValue;

        let InnerComponent = _helpers.Component.extend({
          actions: {
            fireAction() {
              actualReturnedValue = this.attrs.submit();
            }
          }
        });

        let OuterComponent = _helpers.Component.extend({
          outerSubmit() {
            return returnedValue;
          }
        });

        this.registerComponent('inner-component', {
          ComponentClass: InnerComponent,
          template: '<button id="instrument-button" {{action "fireAction"}}>What it do</button>'
        });

        this.registerComponent('outer-component', {
          ComponentClass: OuterComponent,
          template: '{{inner-component submit=(action outerSubmit)}}'
        });

        this.subscribe('interaction.ember-action', {
          before() {},
          after() {}
        });

        this.render(`{{outer-component}}`);

        this.runTask(() => {
          this.$('#instrument-button').trigger('click');
        });

        this.assert.equal(actualReturnedValue, returnedValue, 'action can return to caller');
      }
    });
  }

  (0, _testCase.moduleFor)('Helpers test: closure {{action}}', class extends _testCase.RenderingTest {
    ['@test action should be called']() {
      let outerActionCalled = false;
      let component;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        outerSubmit() {
          outerActionCalled = true;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });
      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: '{{inner-component submit=(action outerSubmit)}}'
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        component.fireAction();
      });

      this.assert.ok(outerActionCalled, 'the action was called');
    }

    ['@test an error is triggered when bound action function is undefined']() {
      this.registerComponent('inner-component', {
        template: 'inner'
      });
      this.registerComponent('outer-component', {
        template: '{{inner-component submit=(action somethingThatIsUndefined)}}'
      });

      expectAssertion(() => {
        this.render('{{outer-component}}');
      }, /Action passed is null or undefined in \(action[^)]*\) from .*\./);
    }

    ['@test an error is triggered when bound action being passed in is a non-function']() {
      this.registerComponent('inner-component', {
        template: 'inner'
      });
      this.registerComponent('outer-component', {
        ComponentClass: _helpers.Component.extend({
          nonFunctionThing: {}
        }),
        template: '{{inner-component submit=(action nonFunctionThing)}}'
      });

      expectAssertion(() => {
        this.render('{{outer-component}}');
      }, /An action could not be made for `.*` in .*\. Please confirm that you are using either a quoted action name \(i\.e\. `\(action '.*'\)`\) or a function available in .*\./);
    }

    ['@test [#12718] a nice error is shown when a bound action function is undefined and it is passed as attrs.foo']() {
      this.registerComponent('inner-component', {
        template: '<button id="inner-button" {{action (action attrs.external-action)}}>Click me</button>'
      });

      this.registerComponent('outer-component', {
        template: '{{inner-component}}'
      });

      expectAssertion(() => {
        this.render('{{outer-component}}');
      }, /Action passed is null or undefined in \(action[^)]*\) from .*\./);
    }

    ['@test action value is returned']() {
      let expectedValue = 'terrible tom';
      let returnedValue;
      let innerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          returnedValue = this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        outerSubmit() {
          return expectedValue;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: '{{inner-component submit=(action outerSubmit)}}'
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(returnedValue, expectedValue, 'action can return to caller');
    }

    ['@test action should be called on the correct scope']() {
      let innerComponent;
      let outerComponent;
      let actualComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          outerComponent = this;
        },
        isOuterComponent: true,
        outerSubmit() {
          actualComponent = this;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: '{{inner-component submit=(action outerSubmit)}}'
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualComponent, outerComponent, 'action has the correct context');
      this.assert.ok(actualComponent.isOuterComponent, 'action has the correct context');
    }

    ['@test arguments to action are passed, curry']() {
      let first = 'mitch';
      let second = 'martin';
      let third = 'matt';
      let fourth = 'wacky wycats';

      let innerComponent;
      let actualArgs;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit(fourth);
        }
      });

      let OuterComponent = _helpers.Component.extend({
        third,
        outerSubmit() {
          // eslint-disable-line no-unused-vars
          actualArgs = [...arguments];
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action (action outerSubmit "${first}") "${second}" third)}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.deepEqual(actualArgs, [first, second, third, fourth], 'action has the correct args');
    }

    ['@test `this` can be passed as an argument']() {
      let value = {};
      let component;
      let innerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          outerAction(incomingValue) {
            value = incomingValue;
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });
      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: '{{inner-component submit=(action "outerAction" this)}}'
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.strictEqual(value, component, 'the component is passed at `this`');
    }

    ['@test arguments to action are bound']() {
      let value = 'lazy leah';

      let innerComponent;
      let outerComponent;
      let actualArg;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          outerComponent = this;
        },
        value: '',
        outerSubmit(incomingValue) {
          actualArg = incomingValue;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action outerSubmit value)}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.strictEqual(actualArg, '', 'action has the correct first arg');

      this.runTask(() => {
        outerComponent.set('value', value);
      });

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.strictEqual(actualArg, value, 'action has the correct first arg');
    }

    ['@test array arguments are passed correctly to action']() {
      let first = 'foo';
      let second = [3, 5];
      let third = [4, 9];

      let actualFirst;
      let actualSecond;
      let actualThird;

      let innerComponent;
      let outerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit(second, third);
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          outerComponent = this;
        },
        outerSubmit(incomingFirst, incomingSecond, incomingThird) {
          actualFirst = incomingFirst;
          actualSecond = incomingSecond;
          actualThird = incomingThird;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action outerSubmit first)}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        outerComponent.set('first', first);
        outerComponent.set('second', second);
      });

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualFirst, first, 'action has the correct first arg');
      this.assert.equal(actualSecond, second, 'action has the correct second arg');
      this.assert.equal(actualThird, third, 'action has the correct third arg');
    }

    ['@test mut values can be wrapped in actions, are settable']() {
      let newValue = 'trollin trek';

      let innerComponent;
      let outerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit(newValue);
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          outerComponent = this;
        },
        outerMut: 'patient peter'
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action (mut outerMut))}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(outerComponent.get('outerMut'), newValue, 'mut value is set');
    }

    ['@test mut values can be wrapped in actions, are settable with a curry']() {
      let newValue = 'trollin trek';

      let innerComponent;
      let outerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          outerComponent = this;
        },
        outerMut: 'patient peter'
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action (mut outerMut) '${newValue}')}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(outerComponent.get('outerMut'), newValue, 'mut value is set');
    }

    ['@test action can create closures over actions']() {
      let first = 'raging robert';
      let second = 'mild machty';
      let returnValue = 'butch brian';

      let actualFirst;
      let actualSecond;
      let actualReturnedValue;

      let innerComponent;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          actualReturnedValue = this.attrs.submit(second);
        }
      });

      let OuterComponent = _helpers.Component.extend({
        actions: {
          outerAction(incomingFirst, incomingSecond) {
            actualFirst = incomingFirst;
            actualSecond = incomingSecond;
            return returnValue;
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'outerAction' '${first}')}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualReturnedValue, returnValue, 'return value is present');
      this.assert.equal(actualFirst, first, 'first argument is correct');
      this.assert.equal(actualSecond, second, 'second argument is correct');
    }

    ['@test provides a helpful error if an action is not present']() {
      let InnerComponent = _helpers.Component.extend({});

      let OuterComponent = _helpers.Component.extend({
        actions: {
          something() {
            // this is present to ensure `actions` hash is present
            // a different error is triggered if `actions` is missing
            // completely
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'doesNotExist')}}`
      });

      expectAssertion(() => {
        this.render('{{outer-component}}');
      }, /An action named 'doesNotExist' was not found in /);
    }

    ['@test provides a helpful error if actions hash is not present']() {
      let InnerComponent = _helpers.Component.extend({});

      let OuterComponent = _helpers.Component.extend({});

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'doesNotExist')}}`
      });

      expectAssertion(() => {
        this.render('{{outer-component}}');
      }, /An action named 'doesNotExist' was not found in /);
    }

    ['@test action can create closures over actions with target']() {
      let innerComponent;
      let actionCalled = false;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        otherComponent: (0, _metal.computed)(function () {
          return {
            actions: {
              outerAction() {
                actionCalled = true;
              }
            }
          };
        })
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'outerAction' target=otherComponent)}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.ok(actionCalled, 'action called on otherComponent');
    }

    ['@test value can be used with action over actions']() {
      let newValue = 'yelping yehuda';

      let innerComponent;
      let actualValue;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit({
            readProp: newValue
          });
        }
      });

      let OuterComponent = _helpers.Component.extend({
        outerContent: {
          readProp: newValue
        },
        actions: {
          outerAction(incomingValue) {
            actualValue = incomingValue;
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'outerAction' value="readProp")}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualValue, newValue, 'value is read');
    }

    ['@test action will read the value of a first property']() {
      let newValue = 'irate igor';

      let innerComponent;
      let actualValue;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit({
            readProp: newValue
          });
        }
      });

      let OuterComponent = _helpers.Component.extend({
        outerAction(incomingValue) {
          actualValue = incomingValue;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action outerAction value="readProp")}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualValue, newValue, 'property is read');
    }

    ['@test action will read the value of a curried first argument property']() {
      let newValue = 'kissing kris';

      let innerComponent;
      let actualValue;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        objectArgument: {
          readProp: newValue
        },
        outerAction(incomingValue) {
          actualValue = incomingValue;
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action outerAction objectArgument value="readProp")}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actualValue, newValue, 'property is read');
    }

    ['@test action closure does not get auto-mut wrapped'](assert) {
      let first = 'raging robert';
      let second = 'mild machty';
      let returnValue = 'butch brian';

      let innerComponent;
      let actualFirst;
      let actualSecond;
      let actualReturnedValue;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.get('submit')(second);
          this.get('attrs-submit')(second);
          let attrsSubmitReturnValue = this.attrs['attrs-submit'](second);
          let submitReturnValue = this.attrs.submit(second);

          assert.equal(attrsSubmitReturnValue, submitReturnValue, 'both attrs.foo and foo should behave the same');

          return submitReturnValue;
        }
      });

      let MiddleComponent = _helpers.Component.extend({});

      let OuterComponent = _helpers.Component.extend({
        actions: {
          outerAction(incomingFirst, incomingSecond) {
            actualFirst = incomingFirst;
            actualSecond = incomingSecond;
            return returnValue;
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('middle-component', {
        ComponentClass: MiddleComponent,
        template: `{{inner-component attrs-submit=attrs.submit submit=submit}}`
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{middle-component submit=(action 'outerAction' '${first}')}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        actualReturnedValue = innerComponent.fireAction();
      });

      this.assert.equal(actualFirst, first, 'first argument is correct');
      this.assert.equal(actualSecond, second, 'second argument is correct');
      this.assert.equal(actualReturnedValue, returnValue, 'return value is present');
    }

    ['@test action should be called within a run loop']() {
      let innerComponent;
      let capturedRunLoop;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          this.attrs.submit();
        }
      });

      let OuterComponent = _helpers.Component.extend({
        actions: {
          submit() {
            capturedRunLoop = (0, _runloop.getCurrentRunLoop)();
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action 'submit')}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.ok(capturedRunLoop, 'action is called within a run loop');
    }

    ['@test objects that define INVOKE can be casted to actions']() {
      let innerComponent;
      let actionArgs;
      let invokableArgs;

      let InnerComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          innerComponent = this;
        },
        fireAction() {
          actionArgs = this.attrs.submit(4, 5, 6);
        }
      });

      let OuterComponent = _helpers.Component.extend({
        foo: 123,
        submitTask: (0, _metal.computed)(function () {
          return {
            [_helpers.INVOKE]: (...args) => {
              invokableArgs = args;
              return this.foo;
            }
          };
        })
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: 'inner'
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: `{{inner-component submit=(action submitTask 1 2 3)}}`
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        innerComponent.fireAction();
      });

      this.assert.equal(actionArgs, 123);
      this.assert.deepEqual(invokableArgs, [1, 2, 3, 4, 5, 6]);
    }

    ['@test closure action with `(mut undefinedThing)` works properly [GH#13959]']() {
      let component;

      let ExampleComponent = _helpers.Component.extend({
        label: undefined,
        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button onclick={{action (mut label) "Clicked!"}}>{{if label label "Click me"}}</button>'
      });

      this.render('{{example-component}}');

      this.assertText('Click me');

      this.assertStableRerender();

      this.runTask(() => {
        this.$('button').click();
      });

      this.assertText('Clicked!');

      this.runTask(() => {
        component.set('label', 'Dun clicked');
      });

      this.assertText('Dun clicked');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assertText('Clicked!');

      this.runTask(() => {
        component.set('label', undefined);
      });

      this.assertText('Click me');
    }

    ['@test closure actions does not cause component hooks to fire unnecessarily [GH#14305] [GH#14654]'](assert) {
      let clicked = 0;
      let didReceiveAttrsFired = 0;

      let ClickMeComponent = _helpers.Component.extend({
        tagName: 'button',

        click() {
          this.get('onClick').call(undefined, ++clicked);
        },

        didReceiveAttrs() {
          didReceiveAttrsFired++;
        }
      });

      this.registerComponent('click-me', {
        ComponentClass: ClickMeComponent
      });

      let outer;

      let OuterComponent = _helpers.Component.extend({
        clicked: 0,

        actions: {
          'on-click': function () {
            this.incrementProperty('clicked');
          }
        },

        init() {
          this._super();
          outer = this;
          this.set('onClick', () => this.incrementProperty('clicked'));
        }
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: _abstractTestCase.strip`
        <div id="counter">clicked: {{clicked}}; foo: {{foo}}</div>

        {{click-me id="string-action" onClick=(action "on-click")}}
        {{click-me id="function-action" onClick=(action onClick)}}
        {{click-me id="mut-action" onClick=(action (mut clicked))}}
      `
      });

      this.render('{{outer-component foo=foo}}', { foo: 1 });

      this.assertText('clicked: 0; foo: 1');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => this.rerender());

      this.assertText('clicked: 0; foo: 1');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 2));

      this.assertText('clicked: 0; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => this.$('#string-action').click());

      this.assertText('clicked: 1; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => this.$('#function-action').click());

      this.assertText('clicked: 2; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => (0, _metal.set)(outer, 'onClick', function () {
        outer.incrementProperty('clicked');
      }));

      this.assertText('clicked: 2; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => this.$('#function-action').click());

      this.assertText('clicked: 3; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);

      this.runTask(() => this.$('#mut-action').click());

      this.assertText('clicked: 4; foo: 2');

      assert.equal(didReceiveAttrsFired, 3);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/concat-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal'], function (_testCase, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{concat}}', class extends _testCase.RenderingTest {
    ['@test it concats static arguments']() {
      this.render(`{{concat "foo" " " "bar" " " "baz"}}`);
      this.assertText('foo bar baz');
    }

    ['@test it updates for bound arguments']() {
      this.render(`{{concat model.first model.second}}`, {
        model: { first: 'one', second: 'two' }
      });

      this.assertText('onetwo');

      this.runTask(() => this.rerender());

      this.assertText('onetwo');

      this.runTask(() => (0, _metal.set)(this.context, 'model.first', 'three'));

      this.assertText('threetwo');

      this.runTask(() => (0, _metal.set)(this.context, 'model.second', 'four'));

      this.assertText('threefour');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { first: 'one', second: 'two' }));

      this.assertText('onetwo');
    }

    ['@test it can be used as a sub-expression']() {
      this.render(`{{concat (concat model.first model.second) (concat model.third model.fourth)}}`, {
        model: {
          first: 'one',
          second: 'two',
          third: 'three',
          fourth: 'four'
        }
      });

      this.assertText('onetwothreefour');

      this.runTask(() => this.rerender());

      this.assertText('onetwothreefour');

      this.runTask(() => (0, _metal.set)(this.context, 'model.first', 'five'));

      this.assertText('fivetwothreefour');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model.second', 'six');
        (0, _metal.set)(this.context, 'model.third', 'seven');
      });

      this.assertText('fivesixsevenfour');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model', {
          first: 'one',
          second: 'two',
          third: 'three',
          fourth: 'four'
        });
      });

      this.assertText('onetwothreefour');
    }

    ['@test it can be used as input for other helpers']() {
      this.registerHelper('x-eq', ([actual, expected]) => actual === expected);

      this.render(`{{#if (x-eq (concat model.first model.second) "onetwo")}}Truthy!{{else}}False{{/if}}`, {
        model: {
          first: 'one',
          second: 'two'
        }
      });

      this.assertText('Truthy!');

      this.runTask(() => this.rerender());

      this.assertText('Truthy!');

      this.runTask(() => (0, _metal.set)(this.context, 'model.first', 'three'));

      this.assertText('False');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { first: 'one', second: 'two' }));

      this.assertText('Truthy!');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/custom-helper-test', ['@ember/-internals/glimmer/tests/utils/test-case', 'internal-test-helpers', '@ember/-internals/metal'], function (_testCase, _internalTestHelpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: custom helpers', class extends _testCase.RenderingTest {
    ['@test it cannot override built-in syntax']() {
      this.registerHelper('if', () => 'Nope');
      expectAssertion(() => {
        this.render(`{{if foo 'LOL'}}`, { foo: true });
      }, /You attempted to overwrite the built-in helper \"if\" which is not allowed. Please rename the helper./);
    }

    ['@test it can resolve custom simple helpers with or without dashes']() {
      this.registerHelper('hello', () => 'hello');
      this.registerHelper('hello-world', () => 'hello world');

      this.render('{{hello}} | {{hello-world}}');

      this.assertText('hello | hello world');

      this.runTask(() => this.rerender());

      this.assertText('hello | hello world');
    }

    ['@test it does not resolve helpers with a `.` (period)']() {
      this.registerHelper('hello.world', () => 'hello world');

      this.render('{{hello.world}}', {
        hello: {
          world: ''
        }
      });

      this.assertText('');

      this.assertStableRerender();

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'hello', { world: 'hello world!' }));

      this.assertText('hello world!');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'hello', {
          world: ''
        });
      });

      this.assertText('');
    }

    ['@test it can resolve custom class-based helpers with or without dashes']() {
      this.registerHelper('hello', {
        compute() {
          return 'hello';
        }
      });

      this.registerHelper('hello-world', {
        compute() {
          return 'hello world';
        }
      });

      this.render('{{hello}} | {{hello-world}}');

      this.assertText('hello | hello world');

      this.runTask(() => this.rerender());

      this.assertText('hello | hello world');
    }

    ['@test throws if `this._super` is not called from `init`']() {
      this.registerHelper('hello-world', {
        init() {}
      });

      expectAssertion(() => {
        this.render('{{hello-world}}');
      }, /You must call `this._super\(...arguments\);` when overriding `init` on a framework object. Please update .* to call `this._super\(...arguments\);` from `init`./);
    }

    ['@test class-based helper can recompute a new value'](assert) {
      let destroyCount = 0;
      let computeCount = 0;
      let helper;

      this.registerHelper('hello-world', {
        init() {
          this._super(...arguments);
          helper = this;
        },
        compute() {
          return ++computeCount;
        },
        destroy() {
          destroyCount++;
          this._super();
        }
      });

      this.render('{{hello-world}}');

      this.assertText('1');

      this.runTask(() => this.rerender());

      this.assertText('1');

      this.runTask(() => helper.recompute());

      this.assertText('2');

      assert.strictEqual(destroyCount, 0, 'destroy is not called on recomputation');
    }

    ['@test class-based helper with static arguments can recompute a new value'](assert) {
      let destroyCount = 0;
      let computeCount = 0;
      let helper;

      this.registerHelper('hello-world', {
        init() {
          this._super(...arguments);
          helper = this;
        },
        compute() {
          return ++computeCount;
        },
        destroy() {
          destroyCount++;
          this._super();
        }
      });

      this.render('{{hello-world "whut"}}');

      this.assertText('1');

      this.runTask(() => this.rerender());

      this.assertText('1');

      this.runTask(() => helper.recompute());

      this.assertText('2');

      assert.strictEqual(destroyCount, 0, 'destroy is not called on recomputation');
    }

    ['@test helper params can be returned']() {
      this.registerHelper('hello-world', values => {
        return values;
      });

      this.render('{{#each (hello-world model) as |item|}}({{item}}){{/each}}', {
        model: ['bob']
      });

      this.assertText('(bob)');
    }

    ['@test helper hash can be returned']() {
      this.registerHelper('hello-world', (_, hash) => {
        return hash.model;
      });

      this.render(`{{get (hello-world model=model) 'name'}}`, {
        model: { name: 'bob' }
      });

      this.assertText('bob');
    }

    ['@test simple helper is called for param changes'](assert) {
      let computeCount = 0;

      this.registerHelper('hello-world', ([value]) => {
        computeCount++;
        return `${value}-value`;
      });

      this.render('{{hello-world model.name}}', {
        model: { name: 'bob' }
      });

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 1, 'compute is called exactly 1 time');

      this.runTask(() => this.rerender());

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 1, 'compute is called exactly 1 time');

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'sal'));

      this.assertText('sal-value');

      assert.strictEqual(computeCount, 2, 'compute is called exactly 2 times');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { name: 'bob' }));

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 3, 'compute is called exactly 3 times');
    }

    ['@test class-based helper compute is called for param changes'](assert) {
      let createCount = 0;
      let computeCount = 0;

      this.registerHelper('hello-world', {
        init() {
          this._super(...arguments);
          createCount++;
        },
        compute([value]) {
          computeCount++;
          return `${value}-value`;
        }
      });

      this.render('{{hello-world model.name}}', {
        model: { name: 'bob' }
      });

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 1, 'compute is called exactly 1 time');

      this.runTask(() => this.rerender());

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 1, 'compute is called exactly 1 time');

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'sal'));

      this.assertText('sal-value');

      assert.strictEqual(computeCount, 2, 'compute is called exactly 2 times');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { name: 'bob' }));

      this.assertText('bob-value');

      assert.strictEqual(computeCount, 3, 'compute is called exactly 3 times');
      assert.strictEqual(createCount, 1, 'helper is only created once');
    }

    ['@test simple helper receives params, hash']() {
      this.registerHelper('hello-world', (_params, _hash) => {
        return `params: ${JSON.stringify(_params)}, hash: ${JSON.stringify(_hash)}`;
      });

      this.render('{{hello-world model.name "rich" first=model.age last="sam"}}', {
        model: {
          name: 'bob',
          age: 42
        }
      });

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => this.rerender());

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'sal'));

      this.assertText('params: ["sal","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model.age', 28));

      this.assertText('params: ["sal","rich"], hash: {"first":28,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { name: 'bob', age: 42 }));

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');
    }

    ['@test class-based helper receives params, hash']() {
      this.registerHelper('hello-world', {
        compute(_params, _hash) {
          return `params: ${JSON.stringify(_params)}, hash: ${JSON.stringify(_hash)}`;
        }
      });

      this.render('{{hello-world model.name "rich" first=model.age last="sam"}}', {
        model: {
          name: 'bob',
          age: 42
        }
      });

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => this.rerender());

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'sal'));

      this.assertText('params: ["sal","rich"], hash: {"first":42,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model.age', 28));

      this.assertText('params: ["sal","rich"], hash: {"first":28,"last":"sam"}');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { name: 'bob', age: 42 }));

      this.assertText('params: ["bob","rich"], hash: {"first":42,"last":"sam"}');
    }

    ['@test class-based helper usable in subexpressions']() {
      this.registerHelper('join-words', {
        compute(params) {
          return params.join(' ');
        }
      });

      this.render(`{{join-words "Who"
                   (join-words "overcomes" "by")
                   model.reason
                   (join-words (join-words "hath overcome but" "half"))
                   (join-words "his" (join-words "foe"))}}`, { model: { reason: 'force' } });

      this.assertText('Who overcomes by force hath overcome but half his foe');

      this.runTask(() => this.rerender());

      this.assertText('Who overcomes by force hath overcome but half his foe');

      this.runTask(() => (0, _metal.set)(this.context, 'model.reason', 'Nickleback'));

      this.assertText('Who overcomes by Nickleback hath overcome but half his foe');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { reason: 'force' }));

      this.assertText('Who overcomes by force hath overcome but half his foe');
    }

    ['@test parameterless helper is usable in subexpressions']() {
      this.registerHelper('should-show', () => {
        return true;
      });

      this.render(`{{#if (should-show)}}true{{/if}}`);

      this.assertText('true');

      this.runTask(() => this.rerender());

      this.assertText('true');
    }

    ['@test parameterless helper is usable in attributes']() {
      this.registerHelper('foo-bar', () => {
        return 'baz';
      });

      this.render(`<div data-foo-bar="{{foo-bar}}"></div>`);

      this.assertHTML('<div data-foo-bar="baz"></div>');

      this.runTask(() => this.rerender());

      this.assertHTML('<div data-foo-bar="baz"></div>');
    }

    ['@test simple helper not usable with a block']() {
      this.registerHelper('some-helper', () => {});

      expectAssertion(() => {
        this.render(`{{#some-helper}}{{/some-helper}}`);
      }, /Helpers may not be used in the block form/);
    }

    ['@test class-based helper not usable with a block']() {
      this.registerHelper('some-helper', {
        compute() {}
      });

      expectAssertion(() => {
        this.render(`{{#some-helper}}{{/some-helper}}`);
      }, /Helpers may not be used in the block form/);
    }

    ['@test simple helper not usable within element']() {
      this.registerHelper('some-helper', () => {});

      this.assert.throws(() => {
        this.render(`<div {{some-helper}}></div>`);
      }, /Compile Error some-helper is not a modifier: Helpers may not be used in the element form/);
    }

    ['@test class-based helper not usable within element']() {
      this.registerHelper('some-helper', {
        compute() {}
      });

      this.assert.throws(() => {
        this.render(`<div {{some-helper}}></div>`);
      }, /Compile Error some-helper is not a modifier: Helpers may not be used in the element form/);
    }

    ['@test class-based helper is torn down'](assert) {
      let destroyCalled = 0;

      this.registerHelper('some-helper', {
        destroy() {
          destroyCalled++;
          this._super(...arguments);
        },
        compute() {
          return 'must define a compute';
        }
      });

      this.render(`{{some-helper}}`);

      (0, _internalTestHelpers.runDestroy)(this.component);

      assert.strictEqual(destroyCalled, 1, 'destroy called once');
    }

    ['@test class-based helper used in subexpression can recompute']() {
      let helper;
      let phrase = 'overcomes by';

      this.registerHelper('dynamic-segment', {
        init() {
          this._super(...arguments);
          helper = this;
        },
        compute() {
          return phrase;
        }
      });

      this.registerHelper('join-words', {
        compute(params) {
          return params.join(' ');
        }
      });

      this.render(`{{join-words "Who"
                   (dynamic-segment)
                   "force"
                   (join-words (join-words "hath overcome but" "half"))
                   (join-words "his" (join-words "foe"))}}`);

      this.assertText('Who overcomes by force hath overcome but half his foe');

      this.runTask(() => this.rerender());

      this.assertText('Who overcomes by force hath overcome but half his foe');

      phrase = 'believes his';

      this.runTask(() => helper.recompute());

      this.assertText('Who believes his force hath overcome but half his foe');

      phrase = 'overcomes by';

      this.runTask(() => helper.recompute());

      this.assertText('Who overcomes by force hath overcome but half his foe');
    }

    ['@test class-based helper used in subexpression can recompute component']() {
      let helper;
      let phrase = 'overcomes by';

      this.registerHelper('dynamic-segment', {
        init() {
          this._super(...arguments);
          helper = this;
        },
        compute() {
          return phrase;
        }
      });

      this.registerHelper('join-words', {
        compute(params) {
          return params.join(' ');
        }
      });

      this.registerComponent('some-component', {
        template: '{{first}} {{second}} {{third}} {{fourth}} {{fifth}}'
      });

      this.render(`{{some-component first="Who"
                   second=(dynamic-segment)
                   third="force"
                   fourth=(join-words (join-words "hath overcome but" "half"))
                   fifth=(join-words "his" (join-words "foe"))}}`);

      this.assertText('Who overcomes by force hath overcome but half his foe');

      this.runTask(() => this.rerender());

      this.assertText('Who overcomes by force hath overcome but half his foe');

      phrase = 'believes his';

      this.runTask(() => helper.recompute());

      this.assertText('Who believes his force hath overcome but half his foe');

      phrase = 'overcomes by';

      this.runTask(() => helper.recompute());

      this.assertText('Who overcomes by force hath overcome but half his foe');
    }

    ['@test class-based helper used in subexpression is destroyed'](assert) {
      let destroyCount = 0;

      this.registerHelper('dynamic-segment', {
        phrase: 'overcomes by',
        init() {
          this._super(...arguments);
        },
        compute() {
          return this.phrase;
        },
        destroy() {
          destroyCount++;
          this._super(...arguments);
        }
      });

      this.registerHelper('join-words', {
        compute(params) {
          return params.join(' ');
        }
      });

      this.render(`{{join-words "Who"
                   (dynamic-segment)
                   "force"
                   (join-words (join-words "hath overcome but" "half"))
                   (join-words "his" (join-words "foe"))}}`);

      (0, _internalTestHelpers.runDestroy)(this.component);

      assert.equal(destroyCount, 1, 'destroy is called after a view is destroyed');
    }

    ['@test simple helper can be invoked manually via `owner.factoryFor(...).create().compute()'](assert) {
      this.registerHelper('some-helper', () => {
        assert.ok(true, 'some-helper helper invoked');
        return 'lolol';
      });

      let instance = this.owner.factoryFor('helper:some-helper').create();

      assert.equal(typeof instance.compute, 'function', 'expected instance.compute to be present');
      assert.equal(instance.compute(), 'lolol', 'can invoke `.compute`');
    }

    ['@test class-based helper can be invoked manually via `owner.factoryFor(...).create().compute()'](assert) {
      this.registerHelper('some-helper', {
        compute() {
          assert.ok(true, 'some-helper helper invoked');
          return 'lolol';
        }
      });

      let instance = this.owner.factoryFor('helper:some-helper').create();

      assert.equal(typeof instance.compute, 'function', 'expected instance.compute to be present');
      assert.equal(instance.compute(), 'lolol', 'can invoke `.compute`');
    }
  }); /* globals EmberDev */


  if (!EmberDev.runningProdBuild) {
    class HelperMutatingArgsTests extends _testCase.RenderingTest {
      buildCompute() {
        return (params, hash) => {
          this.assert.throws(() => {
            params.push('foo');

            // cannot assert error message as it varies by platform
          });

          this.assert.throws(() => {
            hash.foo = 'bar';

            // cannot assert error message as it varies by platform
          });

          this.assert.throws(() => {
            hash.someUnusedHashProperty = 'bar';

            // cannot assert error message as it varies by platform
          });
        };
      }

      ['@test cannot mutate params - no positional specified / named specified']() {
        this.render('{{test-helper foo=bar}}', { bar: 'derp' });
      }

      ['@test cannot mutate params - positional specified / no named specified']() {
        this.render('{{test-helper bar}}', { bar: 'derp' });
      }

      ['@test cannot mutate params - positional specified / named specified']() {
        this.render('{{test-helper bar foo=qux}}', { bar: 'derp', qux: 'baz' });
      }

      ['@test cannot mutate params - no positional specified / no named specified']() {
        this.render('{{test-helper}}', { bar: 'derp', qux: 'baz' });
      }
    }

    (0, _testCase.moduleFor)('Helpers test: mutation triggers errors - class based helper', class extends HelperMutatingArgsTests {
      constructor() {
        super(...arguments);

        let compute = this.buildCompute();

        this.registerHelper('test-helper', {
          compute
        });
      }
    });

    (0, _testCase.moduleFor)('Helpers test: mutation triggers errors - simple helper', class extends HelperMutatingArgsTests {
      constructor() {
        super(...arguments);

        let compute = this.buildCompute();

        this.registerHelper('test-helper', compute);
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/element-action-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal', '@ember/instrumentation', '@ember/canary-features', '@ember/-internals/runtime', '@ember/-internals/views'], function (_testCase, _abstractTestCase, _helpers, _metal, _instrumentation, _canaryFeatures, _runtime, _views) {
  'use strict';

  function getActionAttributes(element) {
    let attributes = element.attributes;
    let actionAttrs = [];

    for (let i = 0; i < attributes.length; i++) {
      let attr = attributes.item(i);

      if (attr.name.indexOf('data-ember-action-') === 0) {
        actionAttrs.push(attr.name);
      }
    }

    return actionAttrs;
  }

  function getActionIds(element) {
    return getActionAttributes(element).map(attribute => attribute.slice('data-ember-action-'.length));
  }

  const isIE11 = !window.ActiveXObject && 'ActiveXObject' in window;

  if (_canaryFeatures.EMBER_IMPROVED_INSTRUMENTATION) {
    (0, _testCase.moduleFor)('Helpers test: element action instrumentation', class extends _testCase.RenderingTest {
      teardown() {
        super.teardown();
        (0, _instrumentation.reset)();
      }

      ['@test action should fire interaction event with proper params']() {
        let subscriberCallCount = 0;
        let subscriberPayload = null;

        let ExampleComponent = _helpers.Component.extend({
          actions: {
            foo() {}
          }
        });

        this.registerComponent('example-component', {
          ComponentClass: ExampleComponent,
          template: '<button {{action "foo" "bar"}}>Click me</button>'
        });

        (0, _instrumentation.subscribe)('interaction.ember-action', {
          before() {
            subscriberCallCount++;
          },
          after(name, time, payload) {
            subscriberPayload = payload;
          }
        });

        this.render('{{example-component}}');

        this.assert.equal(subscriberCallCount, 0, 'subscriber has not been called');

        this.runTask(() => this.rerender());

        this.assert.equal(subscriberCallCount, 0, 'subscriber has not been called');

        this.runTask(() => {
          this.$('button').click();
        });

        this.assert.equal(subscriberCallCount, 1, 'subscriber has been called 1 time');
        this.assert.equal(subscriberPayload.name, 'foo', 'subscriber called with correct name');
        this.assert.equal(subscriberPayload.args[0], 'bar', 'subscriber called with correct args');
      }
    });
  }

  (0, _testCase.moduleFor)('Helpers test: element action', class extends _testCase.RenderingTest {
    ['@test it can call an action on its enclosing component']() {
      let fooCallCount = 0;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          foo() {
            fooCallCount++;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action "foo"}}>Click me</button>'
      });

      this.render('{{example-component}}');

      this.assert.equal(fooCallCount, 0, 'foo has not been called');

      this.runTask(() => this.rerender());

      this.assert.equal(fooCallCount, 0, 'foo has not been called');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.equal(fooCallCount, 1, 'foo has been called 1 time');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.equal(fooCallCount, 2, 'foo has been called 2 times');
    }

    ['@test it can call an action with parameters']() {
      let fooArgs = [];
      let component;

      let ExampleComponent = _helpers.Component.extend({
        member: 'a',
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          foo(thing) {
            fooArgs.push(thing);
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action "foo" member}}>Click me</button>'
      });

      this.render('{{example-component}}');

      this.assert.deepEqual(fooArgs, [], 'foo has not been called');

      this.runTask(() => this.rerender());

      this.assert.deepEqual(fooArgs, [], 'foo has not been called');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.deepEqual(fooArgs, ['a'], 'foo has not been called');

      this.runTask(() => {
        component.set('member', 'b');
      });

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.deepEqual(fooArgs, ['a', 'b'], 'foo has been called with an updated value');
    }

    ['@test it should output a marker attribute with a guid']() {
      this.render('<button {{action "show"}}>me the money</button>');

      let button = this.$('button');

      let attributes = getActionAttributes(button[0]);

      this.assert.ok(button.attr('data-ember-action').match(''), 'An empty data-ember-action attribute was added');
      this.assert.ok(attributes[0].match(/data-ember-action-\d+/), 'A data-ember-action-xyz attribute with a guid was added');
    }

    ['@test it should allow alternative events to be handled']() {
      let showCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          show() {
            showCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<div id="show" {{action "show" on="mouseUp"}}></div>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#show').trigger('mouseup');
      });

      this.assert.ok(showCalled, 'show action was called on mouseUp');
    }

    ['@test inside a yield, the target points at the original target']() {
      let targetWatted = false;
      let innerWatted = false;

      let TargetComponent = _helpers.Component.extend({
        actions: {
          wat() {
            targetWatted = true;
          }
        }
      });

      let InnerComponent = _helpers.Component.extend({
        actions: {
          wat() {
            innerWatted = true;
          }
        }
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: '{{yield}}'
      });

      this.registerComponent('target-component', {
        ComponentClass: TargetComponent,
        template: _abstractTestCase.strip`
        {{#inner-component}}
          <button {{action "wat"}}>Wat me!</button>
        {{/inner-component}}
      `
      });

      this.render('{{target-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.ok(targetWatted, 'the correct target was watted');
      this.assert.notOk(innerWatted, 'the inner target was not watted');
    }

    ['@test it should allow a target to be specified']() {
      let targetWatted = false;

      let TargetComponent = _helpers.Component.extend({
        actions: {
          wat() {
            targetWatted = true;
          }
        }
      });

      let OtherComponent = _helpers.Component.extend({});

      this.registerComponent('target-component', {
        ComponentClass: TargetComponent,
        template: '{{yield this}}'
      });

      this.registerComponent('other-component', {
        ComponentClass: OtherComponent,
        template: '<a {{action "wat" target=anotherTarget}}>Wat?</a>'
      });

      this.render(_abstractTestCase.strip`
          {{#target-component as |parent|}}
            {{other-component anotherTarget=parent}}
          {{/target-component}}
        `);

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.equal(targetWatted, true, 'the specified target was watted');
    }

    ['@test it should lazily evaluate the target']() {
      let firstEdit = 0;
      let secondEdit = 0;
      let component;

      let first = {
        edit() {
          firstEdit++;
        }
      };

      let second = {
        edit() {
          secondEdit++;
        }
      };

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        theTarget: first
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a {{action "edit" target=theTarget}}>Edit</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.equal(firstEdit, 1);

      this.runTask(() => {
        (0, _metal.set)(component, 'theTarget', second);
      });

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.equal(firstEdit, 1);
      this.assert.equal(secondEdit, 1);
    }

    ['@test it should register an event handler']() {
      let editHandlerWasCalled = false;
      let shortcutHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {
            editHandlerWasCalled = true;
          },
          shortcut() {
            shortcutHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit" allowedKeys="alt"}}>click me</a> <div {{action "shortcut" allowedKeys="any"}}>click me too</div>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a[data-ember-action]').trigger('click', { altKey: true });
      });

      this.assert.equal(editHandlerWasCalled, true, 'the event handler was called');

      this.runTask(() => {
        this.$('div[data-ember-action]').trigger('click', { ctrlKey: true });
      });

      this.assert.equal(shortcutHandlerWasCalled, true, 'the "any" shortcut\'s event handler was called');
    }

    ['@test it handles whitelisted bound modifier keys']() {
      let editHandlerWasCalled = false;
      let shortcutHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        altKey: 'alt',
        anyKey: 'any',
        actions: {
          edit() {
            editHandlerWasCalled = true;
          },
          shortcut() {
            shortcutHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit" allowedKeys=altKey}}>click me</a> <div {{action "shortcut" allowedKeys=anyKey}}>click me too</div>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a[data-ember-action]').trigger('click', { altKey: true });
      });

      this.assert.equal(editHandlerWasCalled, true, 'the event handler was called');

      this.runTask(() => {
        this.$('div[data-ember-action]').trigger('click', { ctrlKey: true });
      });

      this.assert.equal(shortcutHandlerWasCalled, true, 'the "any" shortcut\'s event handler was called');
    }

    ['@test it handles whitelisted bound modifier keys with current value']() {
      let editHandlerWasCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        acceptedKeys: 'alt',
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit" allowedKeys=acceptedKeys}}>click me</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a[data-ember-action]').trigger('click', { altKey: true });
      });

      this.assert.equal(editHandlerWasCalled, true, 'the event handler was called');

      editHandlerWasCalled = false;

      this.runTask(() => {
        component.set('acceptedKeys', '');
      });

      this.runTask(() => {
        this.$('div[data-ember-action]').click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the event handler was not called');
    }

    ['@test should be able to use action more than once for the same event within a view']() {
      let editHandlerWasCalled = false;
      let deleteHandlerWasCalled = false;
      let originalHandlerWasCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          edit() {
            editHandlerWasCalled = true;
          },
          delete() {
            deleteHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="edit" href="#" {{action "edit"}}>edit</a><a id="delete" href="#" {{action "delete"}}>delete</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the edit action was called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called (due to bubbling)');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.$('#delete').click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, true, 'the delete action was called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called (due to bubbling)');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.wrap(component.element).click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called');
    }

    ['@test the event should not bubble if `bubbles=false` is passed']() {
      let editHandlerWasCalled = false;
      let deleteHandlerWasCalled = false;
      let originalHandlerWasCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          edit() {
            editHandlerWasCalled = true;
          },
          delete() {
            deleteHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="edit" href="#" {{action "edit" bubbles=false}}>edit</a><a id="delete" href="#" {{action "delete" bubbles=false}}>delete</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the edit action was called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, false, 'the click handler was not called');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.$('#delete').click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, true, 'the delete action was called');
      this.assert.equal(originalHandlerWasCalled, false, 'the click handler was not called');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.wrap(component.element).click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called');
    }

    ['@test the event should not bubble if `bubbles=false` is passed bound']() {
      let editHandlerWasCalled = false;
      let deleteHandlerWasCalled = false;
      let originalHandlerWasCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        isFalse: false,
        actions: {
          edit() {
            editHandlerWasCalled = true;
          },
          delete() {
            deleteHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="edit" href="#" {{action "edit" bubbles=isFalse}}>edit</a><a id="delete" href="#" {{action "delete" bubbles=isFalse}}>delete</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the edit action was called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, false, 'the click handler was not called');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.$('#delete').click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, true, 'the delete action was called');
      this.assert.equal(originalHandlerWasCalled, false, 'the click handler was not called');

      editHandlerWasCalled = deleteHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        this.wrap(component.element).click();
      });

      this.assert.equal(editHandlerWasCalled, false, 'the edit action was not called');
      this.assert.equal(deleteHandlerWasCalled, false, 'the delete action was not called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called');
    }

    ['@test the bubbling depends on the bound parameter']() {
      let editHandlerWasCalled = false;
      let originalHandlerWasCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        shouldBubble: false,
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="edit" href="#" {{action "edit" bubbles=shouldBubble}}>edit</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the edit action was called');
      this.assert.equal(originalHandlerWasCalled, false, 'the click handler was not called');

      editHandlerWasCalled = originalHandlerWasCalled = false;

      this.runTask(() => {
        component.set('shouldBubble', true);
      });

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the edit action was called');
      this.assert.equal(originalHandlerWasCalled, true, 'the click handler was called');
    }

    ['@test it should work properly in an #each block']() {
      let editHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        items: (0, _runtime.A)([1, 2, 3, 4]),
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '{{#each items as |item|}}<a href="#" {{action "edit"}}>click me</a>{{/each}}'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the event handler was called');
    }

    ['@test it should work properly in a {{#with foo as |bar|}} block']() {
      let editHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        something: { ohai: 'there' },
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '{{#with something as |somethingElse|}}<a href="#" {{action "edit"}}>click me</a>{{/with}}'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.equal(editHandlerWasCalled, true, 'the event handler was called');
    }

    ['@test it should unregister event handlers when an element action is removed'](assert) {
      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {}
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '{{#if isActive}}<a href="#" {{action "edit"}}>click me</a>{{/if}}'
      });

      this.render('{{example-component isActive=isActive}}', {
        isActive: true
      });

      assert.equal(this.$('a[data-ember-action]').length, 1, 'The element is rendered');

      let actionId;

      actionId = getActionIds(this.$('a[data-ember-action]')[0])[0];

      assert.ok(_views.ActionManager.registeredActions[actionId], 'An action is registered');

      this.runTask(() => this.rerender());

      assert.equal(this.$('a[data-ember-action]').length, 1, 'The element is still present');

      assert.ok(_views.ActionManager.registeredActions[actionId], 'The action is still registered');

      this.runTask(() => (0, _metal.set)(this.context, 'isActive', false));

      assert.strictEqual(this.$('a[data-ember-action]').length, 0, 'The element is removed');

      assert.ok(!_views.ActionManager.registeredActions[actionId], 'The action is unregistered');

      this.runTask(() => (0, _metal.set)(this.context, 'isActive', true));

      assert.equal(this.$('a[data-ember-action]').length, 1, 'The element is rendered');

      actionId = getActionIds(this.$('a[data-ember-action]')[0])[0];

      assert.ok(_views.ActionManager.registeredActions[actionId], 'A new action is registered');
    }

    ['@test it should capture events from child elements and allow them to trigger the action']() {
      let editHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<div {{action "edit"}}><button>click me</button></div>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.ok(editHandlerWasCalled, 'event on a child target triggered the action of its parent');
    }

    ['@test it should allow bubbling of events from action helper to original parent event']() {
      let editHandlerWasCalled = false;
      let originalHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit"}}>click me</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.ok(editHandlerWasCalled && originalHandlerWasCalled, 'both event handlers were called');
    }

    ['@test it should not bubble an event from action helper to original parent event if `bubbles=false` is passed']() {
      let editHandlerWasCalled = false;
      let originalHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        },
        click() {
          originalHandlerWasCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit" bubbles=false}}>click me</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.ok(editHandlerWasCalled, 'the child event handler was called');
      this.assert.notOk(originalHandlerWasCalled, 'the parent handler was not called');
    }

    ['@test it should allow "send" as the action name (#594)']() {
      let sendHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          send() {
            sendHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "send"}}>click me</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.ok(sendHandlerWasCalled, 'the event handler was called');
    }

    ['@test it should send the view, event, and current context to the action']() {
      let passedTarget;
      let passedContext;
      let targetThis;

      let TargetComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          targetThis = this;
        },
        actions: {
          edit(context) {
            passedTarget = this === targetThis;
            passedContext = context;
          }
        }
      });

      let aContext;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          aContext = this;
        }
      });

      this.registerComponent('target-component', {
        ComponentClass: TargetComponent,
        template: '{{yield this}}'
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: _abstractTestCase.strip`
        {{#target-component as |aTarget|}}
          <a id="edit" href="#" {{action "edit" this target=aTarget}}>click me</a>
        {{/target-component}}
        `
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('#edit').click();
      });

      this.assert.ok(passedTarget, 'the action is called with the target as this');
      this.assert.strictEqual(passedContext, aContext, 'the parameter is passed along');
    }

    ['@test it should only trigger actions for the event they were registered on']() {
      let editHandlerWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          edit() {
            editHandlerWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a href="#" {{action "edit"}}>click me</a>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.ok(editHandlerWasCalled, 'the event handler was called on click');

      editHandlerWasCalled = false;

      this.runTask(() => {
        this.$('a').trigger('mouseover');
      });

      this.assert.notOk(editHandlerWasCalled, 'the event handler was not called on mouseover');
    }

    ['@test it should allow multiple contexts to be specified']() {
      let passedContexts;
      let models = [_runtime.Object.create(), _runtime.Object.create()];

      let ExampleComponent = _helpers.Component.extend({
        modelA: models[0],
        modelB: models[1],
        actions: {
          edit(...args) {
            passedContexts = args;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action "edit" modelA modelB}}>click me</button>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.deepEqual(passedContexts, models, 'the action was called with the passed contexts');
    }

    ['@test it should allow multiple contexts to be specified mixed with string args']() {
      let passedContexts;
      let model = _runtime.Object.create();

      let ExampleComponent = _helpers.Component.extend({
        model: model,
        actions: {
          edit(...args) {
            passedContexts = args;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action "edit" "herp" model}}>click me</button>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.deepEqual(passedContexts, ['herp', model], 'the action was called with the passed contexts');
    }

    ['@test it should not trigger action with special clicks']() {
      let showCalled = false;
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          show() {
            showCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action "show" href=true}}>Howdy</button>'
      });

      this.render('{{example-component}}');

      let assert = this.assert;

      let checkClick = (prop, value, expected) => {
        showCalled = false;
        let event = this.wrap(component.element).findAll('button').trigger('click', { [prop]: value })[0];
        if (expected) {
          assert.ok(showCalled, `should call action with ${prop}:${value}`);

          // IE11 does not allow simulated events to have a valid `defaultPrevented`
          if (!isIE11) {
            assert.ok(event.defaultPrevented, 'should prevent default');
          }
        } else {
          assert.notOk(showCalled, `should not call action with ${prop}:${value}`);
          assert.notOk(event.defaultPrevented, 'should not prevent default');
        }
      };

      checkClick('ctrlKey', true, false);
      checkClick('altKey', true, false);
      checkClick('metaKey', true, false);
      checkClick('shiftKey', true, false);

      checkClick('button', 0, true);
      checkClick('button', 1, false);
      checkClick('button', 2, false);
      checkClick('button', 3, false);
      checkClick('button', 4, false);
    }

    ['@test it can trigger actions for keyboard events']() {
      let showCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          show() {
            showCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<input type="text" {{action "show" on="keyUp"}}>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('input').trigger('keyup', { char: 'a', which: 65 });
      });

      this.assert.ok(showCalled, 'the action was called with keyup');
    }

    ['@test a quoteless parameter should allow dynamic lookup of the actionName']() {
      let lastAction;
      let actionOrder = [];
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        hookMeUp: 'rock',
        actions: {
          rock() {
            lastAction = 'rock';
            actionOrder.push('rock');
          },
          paper() {
            lastAction = 'paper';
            actionOrder.push('paper');
          },
          scissors() {
            lastAction = 'scissors';
            actionOrder.push('scissors');
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="bound-param" {{action hookMeUp}}>Whistle tips go woop woooop</a>'
      });

      this.render('{{example-component}}');

      let test = this;

      let testBoundAction = propertyValue => {
        test.runTask(() => {
          component.set('hookMeUp', propertyValue);
        });

        test.runTask(() => {
          this.wrap(component.element).findAll('#bound-param').click();
        });

        test.assert.ok(lastAction, propertyValue, `lastAction set to ${propertyValue}`);
      };

      testBoundAction('rock');
      testBoundAction('paper');
      testBoundAction('scissors');

      this.assert.deepEqual(actionOrder, ['rock', 'paper', 'scissors'], 'action name was looked up properly');
    }

    ['@test a quoteless string parameter should resolve actionName, including path']() {
      let lastAction;
      let actionOrder = [];
      let component;

      let ExampleComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          component = this;
        },
        allactions: (0, _runtime.A)([{ title: 'Rock', name: 'rock' }, { title: 'Paper', name: 'paper' }, { title: 'Scissors', name: 'scissors' }]),
        actions: {
          rock() {
            lastAction = 'rock';
            actionOrder.push('rock');
          },
          paper() {
            lastAction = 'paper';
            actionOrder.push('paper');
          },
          scissors() {
            lastAction = 'scissors';
            actionOrder.push('scissors');
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '{{#each allactions as |allaction|}}<a id="{{allaction.name}}" {{action allaction.name}}>{{allaction.title}}</a>{{/each}}'
      });

      this.render('{{example-component}}');

      let test = this;

      let testBoundAction = propertyValue => {
        test.runTask(() => {
          this.wrap(component.element).findAll(`#${propertyValue}`).click();
        });

        test.assert.ok(lastAction, propertyValue, `lastAction set to ${propertyValue}`);
      };

      testBoundAction('rock');
      testBoundAction('paper');
      testBoundAction('scissors');

      this.assert.deepEqual(actionOrder, ['rock', 'paper', 'scissors'], 'action name was looked up properly');
    }

    ['@test a quoteless function parameter should be called, including arguments']() {
      let submitCalled = false;
      let incomingArg;

      let arg = 'rough ray';

      let ExampleComponent = _helpers.Component.extend({
        submit(actualArg) {
          incomingArg = actualArg;
          submitCalled = true;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: `<a {{action submit '${arg}'}}>Hi</a>`
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').click();
      });

      this.assert.ok(submitCalled, 'submit function called');
      this.assert.equal(incomingArg, arg, 'argument passed');
    }

    ['@test a quoteless parameter that does not resolve to a value asserts']() {
      let ExampleComponent = _helpers.Component.extend({
        actions: {
          ohNoeNotValid() {}
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a id="oops-bound-param" {{action ohNoeNotValid}}>Hi</a>'
      });

      expectAssertion(() => {
        this.render('{{example-component}}');
      }, 'You specified a quoteless path, `ohNoeNotValid`, to the {{action}} helper ' + 'which did not resolve to an action name (a string). ' + 'Perhaps you meant to use a quoted actionName? (e.g. {{action "ohNoeNotValid"}}).');
    }

    ['@test allows multiple actions on a single element']() {
      let clickActionWasCalled = false;
      let doubleClickActionWasCalled = false;

      let ExampleComponent = _helpers.Component.extend({
        actions: {
          clicked() {
            clickActionWasCalled = true;
          },
          doubleClicked() {
            doubleClickActionWasCalled = true;
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: _abstractTestCase.strip`
        <a href="#"
          {{action "clicked" on="click"}}
          {{action "doubleClicked" on="doubleClick"}}
        >click me</a>`
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('a').trigger('click');
      });

      this.assert.ok(clickActionWasCalled, 'the clicked action was called');

      this.runTask(() => {
        this.$('a').trigger('dblclick');
      });

      this.assert.ok(doubleClickActionWasCalled, 'the doubleClicked action was called');
    }

    ['@test it should respect preventDefault option if provided']() {
      let ExampleComponent = _helpers.Component.extend({
        actions: {
          show() {}
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a {{action "show" preventDefault=false}}>Hi</a>'
      });

      this.render('{{example-component}}');

      let event;

      this.runTask(() => {
        event = this.$('a').click()[0];
      });

      this.assert.equal(event.defaultPrevented, false, 'should not preventDefault');
    }

    ['@test it should respect preventDefault option if provided bound']() {
      let component;

      let ExampleComponent = _helpers.Component.extend({
        shouldPreventDefault: false,
        init() {
          this._super(...arguments);
          component = this;
        },
        actions: {
          show() {}
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<a {{action "show" preventDefault=shouldPreventDefault}}>Hi</a>'
      });

      this.render('{{example-component}}');

      let event;

      this.runTask(() => {
        event = this.$('a').trigger(event)[0];
      });

      this.assert.equal(event.defaultPrevented, false, 'should not preventDefault');

      this.runTask(() => {
        component.set('shouldPreventDefault', true);
        event = this.$('a').trigger('click')[0];
      });

      // IE11 does not allow simulated events to have a valid `defaultPrevented`
      if (!isIE11) {
        this.assert.equal(event.defaultPrevented, true, 'should preventDefault');
      }
    }

    ['@test it should target the proper component when `action` is in yielded block [GH #12409]']() {
      let outerActionCalled = false;
      let innerClickCalled = false;

      let OuterComponent = _helpers.Component.extend({
        actions: {
          hey() {
            outerActionCalled = true;
          }
        }
      });

      let MiddleComponent = _helpers.Component.extend({});

      let InnerComponent = _helpers.Component.extend({
        click() {
          innerClickCalled = true;
          expectDeprecation(() => {
            this.sendAction();
          }, /You called (.*).sendAction\((.*)\) but Component#sendAction is deprecated. Please use closure actions instead./);
        }
      });

      this.registerComponent('outer-component', {
        ComponentClass: OuterComponent,
        template: _abstractTestCase.strip`
        {{#middle-component}}
          {{inner-component action="hey"}}
        {{/middle-component}}
      `
      });

      this.registerComponent('middle-component', {
        ComponentClass: MiddleComponent,
        template: '{{yield}}'
      });

      this.registerComponent('inner-component', {
        ComponentClass: InnerComponent,
        template: _abstractTestCase.strip`
        <button>Click Me</button>
        {{yield}}
      `
      });

      this.render('{{outer-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.ok(outerActionCalled, 'the action fired on the proper target');
      this.assert.ok(innerClickCalled, 'the click was triggered');
    }

    ['@test element action with (mut undefinedThing) works properly']() {
      let component;

      let ExampleComponent = _helpers.Component.extend({
        label: undefined,
        init() {
          this._super(...arguments);
          component = this;
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button {{action (mut label) "Clicked!"}}>{{if label label "Click me"}}</button>'
      });

      this.render('{{example-component}}');

      this.assertText('Click me');

      this.assertStableRerender();

      this.runTask(() => {
        this.$('button').click();
      });

      this.assertText('Clicked!');

      this.runTask(() => {
        component.set('label', 'Dun clicked');
      });

      this.assertText('Dun clicked');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assertText('Clicked!');

      this.runTask(() => {
        component.set('label', undefined);
      });

      this.assertText('Click me');
    }

    ['@test it supports non-registered actions [GH#14888]']() {
      this.render(`
      {{#if show}}
        <button id='ddButton' {{action (mut show) false}}>
          Show ({{show}})
        </button>
      {{/if}}
    `, { show: true });

      this.assert.equal(this.$('button').text().trim(), 'Show (true)');
      // We need to focus in to simulate an actual click.
      this.runTask(() => {
        document.getElementById('ddButton').focus();
        document.getElementById('ddButton').click();
      });
    }

    ["@test action handler that shifts element attributes doesn't trigger multiple invocations"]() {
      let actionCount = 0;
      let ExampleComponent = _helpers.Component.extend({
        selected: false,
        actions: {
          toggleSelected() {
            actionCount++;
            this.toggleProperty('selected');
          }
        }
      });

      this.registerComponent('example-component', {
        ComponentClass: ExampleComponent,
        template: '<button class="{{if selected \'selected\'}}" {{action "toggleSelected"}}>Toggle Selected</button>'
      });

      this.render('{{example-component}}');

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.equal(actionCount, 1, 'Click action only fired once.');
      this.assert.ok(this.$('button').hasClass('selected'), "Element with action handler has properly updated it's conditional class");

      this.runTask(() => {
        this.$('button').click();
      });

      this.assert.equal(actionCount, 2, 'Second click action only fired once.');
      this.assert.ok(!this.$('button').hasClass('selected'), "Element with action handler has properly updated it's conditional class");
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/get-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal'], function (_testCase, _helpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{get}}', class extends _testCase.RenderingTest {
    ['@test should be able to get an object value with a static key']() {
      this.render(`[{{get colors 'apple'}}] [{{if true (get colors 'apple')}}]`, {
        colors: { apple: 'red' }
      });

      this.assertText('[red] [red]');

      this.runTask(() => this.rerender());

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors.apple', 'green'));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', {
        apple: 'red'
      }));

      this.assertText('[red] [red]');
    }

    ['@test should be able to get an object value with nested static key']() {
      this.render(`[{{get colors "apple.gala"}}] [{{if true (get colors "apple.gala")}}]`, {
        colors: {
          apple: {
            gala: 'red and yellow'
          }
        }
      });

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => this.rerender());

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors.apple.gala', 'yellow and red striped'));

      this.assertText('[yellow and red striped] [yellow and red striped]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', { apple: { gala: 'red and yellow' } }));

      this.assertText('[red and yellow] [red and yellow]');
    }

    ['@test should be able to get an object value with a number']() {
      this.render(`[{{get items 1}}][{{get items 2}}][{{get items 3}}]`, {
        indexes: [1, 2, 3],
        items: {
          1: 'First',
          2: 'Second',
          3: 'Third'
        }
      });

      this.assertText('[First][Second][Third]');

      this.runTask(() => this.rerender());

      this.assertText('[First][Second][Third]');

      this.runTask(() => (0, _metal.set)(this.context, 'items.1', 'Qux'));

      this.assertText('[Qux][Second][Third]');

      this.runTask(() => (0, _metal.set)(this.context, 'items', { 1: 'First', 2: 'Second', 3: 'Third' }));

      this.assertText('[First][Second][Third]');
    }

    ['@test should be able to get an array value with a number']() {
      this.render(`[{{get numbers 0}}][{{get numbers 1}}][{{get numbers 2}}]`, {
        numbers: [1, 2, 3]
      });

      this.assertText('[1][2][3]');

      this.runTask(() => this.rerender());

      this.assertText('[1][2][3]');

      this.runTask(() => (0, _metal.set)(this.context, 'numbers', [3, 2, 1]));

      this.assertText('[3][2][1]');

      this.runTask(() => (0, _metal.set)(this.context, 'numbers', [1, 2, 3]));

      this.assertText('[1][2][3]');
    }

    ['@test should be able to get an object value with a path evaluating to a number']() {
      this.render(`{{#each indexes as |index|}}[{{get items index}}]{{/each}}`, {
        indexes: [1, 2, 3],
        items: {
          1: 'First',
          2: 'Second',
          3: 'Third'
        }
      });

      this.assertText('[First][Second][Third]');

      this.runTask(() => this.rerender());

      this.assertText('[First][Second][Third]');

      this.runTask(() => (0, _metal.set)(this.context, 'items.1', 'Qux'));

      this.assertText('[Qux][Second][Third]');

      this.runTask(() => (0, _metal.set)(this.context, 'items', { 1: 'First', 2: 'Second', 3: 'Third' }));

      this.assertText('[First][Second][Third]');
    }

    ['@test should be able to get an array value with a path evaluating to a number']() {
      this.render(`{{#each numbers as |num index|}}[{{get numbers index}}]{{/each}}`, {
        numbers: [1, 2, 3]
      });

      this.assertText('[1][2][3]');

      this.runTask(() => this.rerender());

      this.assertText('[1][2][3]');

      this.runTask(() => (0, _metal.set)(this.context, 'numbers', [3, 2, 1]));

      this.assertText('[3][2][1]');
    }

    ['@test should be able to get an object value with a bound/dynamic key']() {
      this.render(`[{{get colors key}}] [{{if true (get colors key)}}]`, {
        colors: { apple: 'red', banana: 'yellow' },
        key: 'apple'
      });

      this.assertText('[red] [red]');

      this.runTask(() => this.rerender());

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'banana'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'colors.apple', 'green');
        (0, _metal.set)(this.context, 'colors.banana', 'purple');
      });

      this.assertText('[purple] [purple]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple'));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', { apple: 'red' }));

      this.assertText('[red] [red]');
    }

    ['@test should be able to get an object value with nested dynamic key']() {
      this.render(`[{{get colors key}}] [{{if true (get colors key)}}]`, {
        colors: {
          apple: {
            gala: 'red and yellow',
            mcintosh: 'red'
          },
          banana: 'yellow'
        },
        key: 'apple.gala'
      });

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => this.rerender());

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple.mcintosh'));

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'banana'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple.gala'));

      this.assertText('[red and yellow] [red and yellow]');
    }

    ['@test should be able to get an object value with subexpression returning nested key']() {
      this.render(`[{{get colors (concat 'apple' '.' 'gala')}}] [{{if true (get colors (concat 'apple' '.' 'gala'))}}]`, {
        colors: {
          apple: {
            gala: 'red and yellow',
            mcintosh: 'red'
          }
        },
        key: 'apple.gala'
      });

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => this.rerender());

      this.assertText('[red and yellow] [red and yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors.apple.gala', 'yellow and red striped'));

      this.assertText('[yellow and red striped] [yellow and red striped]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors.apple.gala', 'yellow-redish'));

      this.assertText('[yellow-redish] [yellow-redish]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', {
        apple: {
          gala: 'red and yellow',
          mcintosh: 'red'
        }
      }));

      this.assertText('[red and yellow] [red and yellow]');
    }

    ['@test should be able to get an object value with a get helper as the key']() {
      this.render(`[{{get colors (get possibleKeys key)}}] [{{if true (get colors (get possibleKeys key))}}]`, {
        colors: { apple: 'red', banana: 'yellow' },
        key: 'key1',
        possibleKeys: { key1: 'apple', key2: 'banana' }
      });

      this.assertText('[red] [red]');

      this.runTask(() => this.rerender());

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'key2'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'colors.apple', 'green');
        (0, _metal.set)(this.context, 'colors.banana', 'purple');
      });

      this.assertText('[purple] [purple]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'key1'));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', { apple: 'red', banana: 'yellow' }));

      this.assertText('[red] [red]');
    }

    ['@test should be able to get an object value with a get helper value as a bound/dynamic key']() {
      this.render(`[{{get (get possibleValues objectKey) key}}] [{{if true (get (get possibleValues objectKey) key)}}]`, {
        possibleValues: {
          colors1: { apple: 'red', banana: 'yellow' },
          colors2: { apple: 'green', banana: 'purple' }
        },
        objectKey: 'colors1',
        key: 'apple'
      });

      this.assertText('[red] [red]');

      this.runTask(() => this.rerender());

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors2'));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors1'));

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'banana'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors2'));

      this.assertText('[purple] [purple]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors1'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple'));
    }

    ['@test should be able to get an object value with a get helper as the value and a get helper as the key']() {
      this.render(`[{{get (get possibleValues objectKey) (get possibleKeys key)}}] [{{if true (get (get possibleValues objectKey) (get possibleKeys key))}}]`, {
        possibleValues: {
          colors1: { apple: 'red', banana: 'yellow' },
          colors2: { apple: 'green', banana: 'purple' }
        },
        objectKey: 'colors1',
        possibleKeys: {
          key1: 'apple',
          key2: 'banana'
        },
        key: 'key1'
      });

      this.assertText('[red] [red]');

      this.runTask(() => this.rerender());

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors2'));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors1'));

      this.assertText('[red] [red]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'key2'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'objectKey', 'colors2'));

      this.assertText('[purple] [purple]');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'objectKey', 'colors1');
        (0, _metal.set)(this.context, 'key', 'key1');
      });

      this.assertText('[red] [red]');
    }

    ['@test the result of a get helper can be yielded']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
          this.mcintosh = 'red';
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (get colors mcintosh)}}`
      });

      this.render(`{{#foo-bar colors=colors as |value|}}{{value}}{{/foo-bar}}`, {
        colors: {
          red: 'banana'
        }
      });

      this.assertText('banana');

      this.runTask(() => this.rerender());

      this.assertText('banana');

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'mcintosh', 'yellow');
        (0, _metal.set)(this.context, 'colors', { yellow: 'bus' });
      });

      this.assertText('bus');

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'mcintosh', 'red');
        (0, _metal.set)(this.context, 'colors', { red: 'banana' });
      });

      this.assertText('banana');
    }

    ['@test should handle object values as nulls']() {
      this.render(`[{{get colors 'apple'}}] [{{if true (get colors 'apple')}}]`, {
        colors: null
      });

      this.assertText('[] []');

      this.runTask(() => this.rerender());

      this.assertText('[] []');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', { apple: 'green', banana: 'purple' }));

      this.assertText('[green] [green]');

      this.runTask(() => (0, _metal.set)(this.context, 'colors', null));

      this.assertText('[] []');
    }

    ['@test should handle object keys as nulls']() {
      this.render(`[{{get colors key}}] [{{if true (get colors key)}}]`, {
        colors: {
          apple: 'red',
          banana: 'yellow'
        },
        key: null
      });

      this.assertText('[] []');

      this.runTask(() => this.rerender());

      this.assertText('[] []');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'banana'));

      this.assertText('[yellow] [yellow]');

      this.runTask(() => (0, _metal.set)(this.context, 'key', null));

      this.assertText('[] []');
    }

    ['@test should handle object values and keys as nulls']() {
      this.render(`[{{get colors 'apple'}}] [{{if true (get colors key)}}]`, {
        colors: null,
        key: null
      });

      this.assertText('[] []');
    }

    ['@test get helper value should be updatable using {{input}} and (mut) - static key'](assert) {
      this.render(`{{input type='text' value=(mut (get source 'banana')) id='get-input'}}`, {
        source: {
          banana: 'banana'
        }
      });

      assert.strictEqual(this.$('#get-input').val(), 'banana');

      this.runTask(() => this.rerender());

      assert.strictEqual(this.$('#get-input').val(), 'banana');

      this.runTask(() => (0, _metal.set)(this.context, 'source.banana', 'yellow'));

      assert.strictEqual(this.$('#get-input').val(), 'yellow');

      this.runTask(() => this.$('#get-input').val('some value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'some value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.banana'), 'some value');

      this.runTask(() => (0, _metal.set)(this.context, 'source', { banana: 'banana' }));

      assert.strictEqual(this.$('#get-input').val(), 'banana');
    }

    ['@test get helper value should be updatable using {{input}} and (mut) - dynamic key'](assert) {
      this.render(`{{input type='text' value=(mut (get source key)) id='get-input'}}`, {
        source: {
          apple: 'apple',
          banana: 'banana'
        },
        key: 'banana'
      });

      assert.strictEqual(this.$('#get-input').val(), 'banana');

      this.runTask(() => this.rerender());

      assert.strictEqual(this.$('#get-input').val(), 'banana');

      this.runTask(() => (0, _metal.set)(this.context, 'source.banana', 'yellow'));

      assert.strictEqual(this.$('#get-input').val(), 'yellow');

      this.runTask(() => this.$('#get-input').val('some value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'some value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.banana'), 'some value');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple'));

      assert.strictEqual(this.$('#get-input').val(), 'apple');

      this.runTask(() => this.$('#get-input').val('some other value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'some other value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.apple'), 'some other value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'key', 'banana');
        (0, _metal.set)(this.context, 'source', { banana: 'banana' });
      });

      assert.strictEqual(this.$('#get-input').val(), 'banana');
    }

    ['@test get helper value should be updatable using {{input}} and (mut) - dynamic nested key'](assert) {
      this.render(`{{input type='text' value=(mut (get source key)) id='get-input'}}`, {
        source: {
          apple: {
            gala: 'gala',
            mcintosh: 'mcintosh'
          },
          banana: 'banana'
        },
        key: 'apple.mcintosh'
      });

      assert.strictEqual(this.$('#get-input').val(), 'mcintosh');

      this.runTask(() => this.rerender());

      assert.strictEqual(this.$('#get-input').val(), 'mcintosh');

      this.runTask(() => (0, _metal.set)(this.context, 'source.apple.mcintosh', 'red'));

      assert.strictEqual(this.$('#get-input').val(), 'red');

      this.runTask(() => this.$('#get-input').val('some value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'some value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.apple.mcintosh'), 'some value');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'apple.gala'));

      assert.strictEqual(this.$('#get-input').val(), 'gala');

      this.runTask(() => this.$('#get-input').val('some other value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'some other value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.apple.gala'), 'some other value');

      this.runTask(() => (0, _metal.set)(this.context, 'key', 'banana'));

      assert.strictEqual(this.$('#get-input').val(), 'banana');

      this.runTask(() => this.$('#get-input').val('yet another value').trigger('change'));

      assert.strictEqual(this.$('#get-input').val(), 'yet another value');
      assert.strictEqual((0, _metal.get)(this.context, 'source.banana'), 'yet another value');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'key', 'apple.mcintosh');
        (0, _metal.set)(this.context, 'source', {
          apple: {
            gala: 'gala',
            mcintosh: 'mcintosh'
          },
          banana: 'banana'
        });
      });

      assert.strictEqual(this.$('#get-input').val(), 'mcintosh');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/hash-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal'], function (_testCase, _helpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{hash}}', class extends _testCase.RenderingTest {
    ['@test returns a hash with the right key-value']() {
      this.render(`{{#with (hash name=\"Sergio\") as |person|}}{{person.name}}{{/with}}`);

      this.assertText('Sergio');

      this.runTask(() => this.rerender());

      this.assertText('Sergio');
    }

    ['@test can have more than one key-value']() {
      this.render(`{{#with (hash name="Sergio" lastName="Arbeo") as |person|}}{{person.name}} {{person.lastName}}{{/with}}`);

      this.assertText('Sergio Arbeo');

      this.runTask(() => this.rerender());

      this.assertText('Sergio Arbeo');
    }

    ['@test binds values when variables are used']() {
      this.render(`{{#with (hash name=model.firstName lastName="Arbeo") as |person|}}{{person.name}} {{person.lastName}}{{/with}}`, {
        model: {
          firstName: 'Marisa'
        }
      });

      this.assertText('Marisa Arbeo');

      this.runTask(() => this.rerender());

      this.assertText('Marisa Arbeo');

      this.runTask(() => (0, _metal.set)(this.context, 'model.firstName', 'Sergio'));

      this.assertText('Sergio Arbeo');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { firstName: 'Marisa' }));

      this.assertText('Marisa Arbeo');
    }

    ['@test binds multiple values when variables are used']() {
      this.render(`{{#with (hash name=model.firstName lastName=model.lastName) as |person|}}{{person.name}} {{person.lastName}}{{/with}}`, {
        model: {
          firstName: 'Marisa',
          lastName: 'Arbeo'
        }
      });

      this.assertText('Marisa Arbeo');

      this.runTask(() => this.rerender());

      this.assertText('Marisa Arbeo');

      this.runTask(() => (0, _metal.set)(this.context, 'model.firstName', 'Sergio'));

      this.assertText('Sergio Arbeo');

      this.runTask(() => (0, _metal.set)(this.context, 'model.lastName', 'Smith'));

      this.assertText('Sergio Smith');

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        firstName: 'Marisa',
        lastName: 'Arbeo'
      }));

      this.assertText('Marisa Arbeo');
    }

    ['@test hash helpers can be nested']() {
      this.render(`{{#with (hash person=(hash name=model.firstName)) as |ctx|}}{{ctx.person.name}}{{/with}}`, {
        model: { firstName: 'Balint' }
      });

      this.assertText('Balint');

      this.runTask(() => this.rerender());

      this.assertText('Balint');

      this.runTask(() => (0, _metal.set)(this.context, 'model.firstName', 'Chad'));

      this.assertText('Chad');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { firstName: 'Balint' }));

      this.assertText('Balint');
    }

    ['@test should yield hash of internal properties']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
          this.model = { firstName: 'Chad' };
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (hash firstName=model.firstName)}}`
      });

      this.render(`{{#foo-bar as |values|}}{{values.firstName}}{{/foo-bar}}`);

      this.assertText('Chad');

      this.runTask(() => this.rerender());

      this.assertText('Chad');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model.firstName', 'Godfrey'));

      this.assertText('Godfrey');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model', { firstName: 'Chad' }));

      this.assertText('Chad');
    }

    ['@test should yield hash of internal and external properties']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super();
          fooBarInstance = this;
          this.model = { firstName: 'Chad' };
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (hash firstName=model.firstName lastName=lastName)}}`
      });

      this.render(`{{#foo-bar lastName=model.lastName as |values|}}{{values.firstName}} {{values.lastName}}{{/foo-bar}}`, {
        model: { lastName: 'Hietala' }
      });

      this.assertText('Chad Hietala');

      this.runTask(() => this.rerender());

      this.assertText('Chad Hietala');

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'model.firstName', 'Godfrey');
        (0, _metal.set)(this.context, 'model.lastName', 'Chan');
      });

      this.assertText('Godfrey Chan');

      this.runTask(() => {
        (0, _metal.set)(fooBarInstance, 'model', { firstName: 'Chad' });
        (0, _metal.set)(this.context, 'model', { lastName: 'Hietala' });
      });

      this.assertText('Chad Hietala');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/if-unless-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/shared-conditional-tests'], function (_testCase, _sharedConditionalTests) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: inline {{if}}', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{if ${cond} ${truthy} ${falsy}}}`;
    }

    ['@test it raises when there are more than three arguments']() {
      expectAssertion(() => {
        this.render(`{{if condition 'a' 'b' 'c'}}`, { condition: true });
      }, `The inline form of the 'if' helper expects two or three arguments. ('-top-level' @ L1:C0) `);
    }

    ['@test it raises when there are less than two arguments']() {
      expectAssertion(() => {
        this.render(`{{if condition}}`, { condition: true });
      }, `The inline form of the 'if' helper expects two or three arguments. ('-top-level' @ L1:C0) `);
    }
  });

  (0, _testCase.moduleFor)('Helpers test: nested {{if}} helpers (returning truthy values)', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{if (if ${cond} ${cond} false) ${truthy} ${falsy}}}`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: nested {{if}} helpers (returning falsy values)', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{if (if ${cond} true ${cond}) ${truthy} ${falsy}}}`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: {{if}} used with another helper', class extends _sharedConditionalTests.IfUnlessHelperTest {
    wrapperFor(templates) {
      return `{{concat ${templates.join(' ')}}}`;
    }

    templateFor({ cond, truthy, falsy }) {
      return `(if ${cond} ${truthy} ${falsy})`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: {{if}} used in attribute position', class extends _sharedConditionalTests.IfUnlessHelperTest {
    wrapperFor(templates) {
      return `<div data-foo="${templates.join('')}" />`;
    }

    templateFor({ cond, truthy, falsy }) {
      return `{{if ${cond} ${truthy} ${falsy}}}`;
    }

    textValue() {
      return this.$('div').attr('data-foo');
    }
  });

  (0, _testCase.moduleFor)('Helpers test: inline {{if}} and {{unless}} without the inverse argument', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{if ${cond} ${truthy}}}{{unless ${cond} ${falsy}}}`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: inline {{unless}}', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{unless ${cond} ${falsy} ${truthy}}}`;
    }

    ['@test it raises when there are more than three arguments']() {
      expectAssertion(() => {
        this.render(`{{unless condition 'a' 'b' 'c'}}`, { condition: true });
      }, /The inline form of the `unless` helper expects two or three arguments/);
    }

    ['@test it raises when there are less than two arguments']() {
      expectAssertion(() => {
        this.render(`{{unless condition}}`, { condition: true });
      }, /The inline form of the `unless` helper expects two or three arguments/);
    }
  });

  (0, _testCase.moduleFor)('Helpers test: nested {{unless}} helpers (returning truthy values)', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{unless (unless ${cond} false ${cond}) ${falsy} ${truthy}}}`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: nested {{unless}} helpers (returning falsy values)', class extends _sharedConditionalTests.IfUnlessHelperTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{unless (unless ${cond} ${cond} true) ${falsy} ${truthy}}}`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: {{unless}} used with another helper', class extends _sharedConditionalTests.IfUnlessHelperTest {
    wrapperFor(templates) {
      return `{{concat ${templates.join(' ')}}}`;
    }

    templateFor({ cond, truthy, falsy }) {
      return `(unless ${cond} ${falsy} ${truthy})`;
    }
  });

  (0, _testCase.moduleFor)('Helpers test: {{unless}} used in attribute position', class extends _sharedConditionalTests.IfUnlessHelperTest {
    wrapperFor(templates) {
      return `<div data-foo="${templates.join('')}" />`;
    }

    templateFor({ cond, truthy, falsy }) {
      return `{{unless ${cond} ${falsy} ${truthy}}}`;
    }

    textValue() {
      return this.$('div').attr('data-foo');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/input-test', ['@ember/polyfills', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/test-case', 'internal-test-helpers', '@ember/-internals/views'], function (_polyfills, _metal, _helpers, _testCase, _internalTestHelpers, _views) {
  'use strict';

  class InputRenderingTest extends _testCase.RenderingTest {
    $input() {
      return this.$('input');
    }

    inputID() {
      return this.$input().prop('id');
    }

    assertDisabled() {
      this.assert.ok(this.$('input').prop('disabled'), 'The input is disabled');
    }

    assertNotDisabled() {
      this.assert.ok(this.$('input').is(':not(:disabled)'), 'The input is not disabled');
    }

    assertInputId(expectedId) {
      this.assert.equal(this.inputID(), expectedId, 'the input id should be `expectedId`');
    }

    assertSingleInput() {
      this.assert.equal(this.$('input').length, 1, 'A single text field was inserted');
    }

    assertSingleCheckbox() {
      this.assert.equal(this.$('input[type=checkbox]').length, 1, 'A single checkbox is added');
    }

    assertCheckboxIsChecked() {
      this.assert.equal(this.$input().prop('checked'), true, 'the checkbox is checked');
    }

    assertCheckboxIsNotChecked() {
      this.assert.equal(this.$input().prop('checked'), false, 'the checkbox is not checked');
    }

    assertValue(expected) {
      this.assert.equal(this.$input().val(), expected, `the input value should be ${expected}`);
    }

    assertAttr(name, expected) {
      this.assert.equal(this.$input().attr(name), expected, `the input ${name} attribute has the value '${expected}'`);
    }

    assertAllAttrs(names, expected) {
      names.forEach(name => this.assertAttr(name, expected));
    }

    assertSelectionRange(start, end) {
      let input = this.$input()[0];
      this.assert.equal(input.selectionStart, start, `the cursor start position should be ${start}`);
      this.assert.equal(input.selectionEnd, end, `the cursor end position should be ${end}`);
    }

    triggerEvent(type, options) {
      let event = document.createEvent('Events');
      event.initEvent(type, true, true);
      (0, _polyfills.assign)(event, options);

      let element = this.$input()[0];
      this.runTask(() => {
        element.dispatchEvent(event);
      });
    }
  }

  (0, _testCase.moduleFor)('Helpers test: {{input}}', class extends InputRenderingTest {
    ['@test a single text field is inserted into the DOM']() {
      this.render(`{{input type="text" value=value}}`, { value: 'hello' });

      let id = this.inputID();

      this.assertValue('hello');
      this.assertSingleInput();

      this.runTask(() => this.rerender());

      this.assertValue('hello');
      this.assertSingleInput();
      this.assertInputId(id);

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'goodbye'));

      this.assertValue('goodbye');
      this.assertSingleInput();
      this.assertInputId(id);

      this.runTask(() => (0, _metal.set)(this.context, 'value', 'hello'));

      this.assertValue('hello');
      this.assertSingleInput();
      this.assertInputId(id);
    }

    ['@test default type']() {
      this.render(`{{input}}`);

      this.assertAttr('type', 'text');

      this.runTask(() => this.rerender());

      this.assertAttr('type', 'text');
    }

    ['@test dynamic attributes']() {
      this.render(`
      {{input type="text"
        disabled=disabled
        value=value
        placeholder=placeholder
        name=name
        maxlength=maxlength
        minlength=minlength
        size=size
        tabindex=tabindex
      }}`, {
        disabled: false,
        value: 'Original value',
        placeholder: 'Original placeholder',
        name: 'original-name',
        maxlength: 10,
        minlength: 5,
        size: 20,
        tabindex: 30
      });

      this.assertNotDisabled();
      this.assertValue('Original value');
      this.assertAttr('placeholder', 'Original placeholder');
      this.assertAttr('name', 'original-name');
      this.assertAttr('maxlength', '10');
      this.assertAttr('minlength', '5');
      // this.assertAttr('size', '20'); //NOTE: failing in IE  (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '30'); //NOTE: failing in IE (TEST_SUITE=sauce)

      this.runTask(() => this.rerender());

      this.assertNotDisabled();
      this.assertValue('Original value');
      this.assertAttr('placeholder', 'Original placeholder');
      this.assertAttr('name', 'original-name');
      this.assertAttr('maxlength', '10');
      this.assertAttr('minlength', '5');
      // this.assertAttr('size', '20'); //NOTE: failing in IE (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '30'); //NOTE: failing in IE (TEST_SUITE=sauce)

      this.runTask(() => {
        (0, _metal.set)(this.context, 'value', 'Updated value');
        (0, _metal.set)(this.context, 'disabled', true);
        (0, _metal.set)(this.context, 'placeholder', 'Updated placeholder');
        (0, _metal.set)(this.context, 'name', 'updated-name');
        (0, _metal.set)(this.context, 'maxlength', 11);
        (0, _metal.set)(this.context, 'minlength', 6);
        // set(this.context, 'size', 21); //NOTE: failing in IE (TEST_SUITE=sauce)
        // set(this.context, 'tabindex', 31); //NOTE: failing in IE (TEST_SUITE=sauce)
      });

      this.assertDisabled();
      this.assertValue('Updated value');
      this.assertAttr('placeholder', 'Updated placeholder');
      this.assertAttr('name', 'updated-name');
      this.assertAttr('maxlength', '11');
      this.assertAttr('minlength', '6');
      // this.assertAttr('size', '21'); //NOTE: failing in IE (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '31'); //NOTE: failing in IE (TEST_SUITE=sauce)

      this.runTask(() => {
        (0, _metal.set)(this.context, 'value', 'Original value');
        (0, _metal.set)(this.context, 'disabled', false);
        (0, _metal.set)(this.context, 'placeholder', 'Original placeholder');
        (0, _metal.set)(this.context, 'name', 'original-name');
        (0, _metal.set)(this.context, 'maxlength', 10);
        (0, _metal.set)(this.context, 'minlength', 5);
        // set(this.context, 'size', 20); //NOTE: failing in IE (TEST_SUITE=sauce)
        // set(this.context, 'tabindex', 30); //NOTE: failing in IE (TEST_SUITE=sauce)
      });

      this.assertNotDisabled();
      this.assertValue('Original value');
      this.assertAttr('placeholder', 'Original placeholder');
      this.assertAttr('name', 'original-name');
      this.assertAttr('maxlength', '10');
      this.assertAttr('minlength', '5');
      // this.assertAttr('size', '20'); //NOTE: failing in IE (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '30'); //NOTE: failing in IE (TEST_SUITE=sauce)
    }

    ['@test static attributes']() {
      this.render(`
      {{input type="text"
        disabled=true
        value="Original value"
        placeholder="Original placeholder"
        name="original-name"
        maxlength=10
        minlength=5
        size=20
        tabindex=30
      }}`);

      this.assertDisabled();
      this.assertValue('Original value');
      this.assertAttr('placeholder', 'Original placeholder');
      this.assertAttr('name', 'original-name');
      this.assertAttr('maxlength', '10');
      this.assertAttr('minlength', '5');
      // this.assertAttr('size', '20');  //NOTE: failing in IE (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '30');  //NOTE: failing in IE (TEST_SUITE=sauce)

      this.runTask(() => this.rerender());

      this.assertDisabled();
      this.assertValue('Original value');
      this.assertAttr('placeholder', 'Original placeholder');
      this.assertAttr('name', 'original-name');
      this.assertAttr('maxlength', '10');
      this.assertAttr('minlength', '5');
      // this.assertAttr('size', '20');  //NOTE: failing in IE (TEST_SUITE=sauce)
      // this.assertAttr('tabindex', '30');  //NOTE: failing in IE (TEST_SUITE=sauce)
    }

    ['@test cursor selection range']() {
      // Modifying input.selectionStart, which is utilized in the cursor tests,
      // causes an event in Safari.
      (0, _internalTestHelpers.runDestroy)(this.owner.lookup('event_dispatcher:main'));

      this.render(`{{input type="text" value=value}}`, { value: 'original' });

      let input = this.$input()[0];

      // See https://ember-twiddle.com/33e506329f8176ae874422644d4cc08c?openFiles=components.input-component.js%2Ctemplates.components.input-component.hbs
      // this.assertSelectionRange(8, 8); //NOTE: this is (0, 0) on Firefox (TEST_SUITE=sauce)

      this.runTask(() => this.rerender());

      // this.assertSelectionRange(8, 8); //NOTE: this is (0, 0) on Firefox (TEST_SUITE=sauce)

      this.runTask(() => {
        input.selectionStart = 2;
        input.selectionEnd = 4;
      });

      this.assertSelectionRange(2, 4);

      this.runTask(() => this.rerender());

      this.assertSelectionRange(2, 4);

      // this.runTask(() => set(this.context, 'value', 'updated'));
      //
      // this.assertSelectionRange(7, 7); //NOTE: this fails in IE, the range is 0 -> 0 (TEST_SUITE=sauce)
      //
      // this.runTask(() => set(this.context, 'value', 'original'));
      //
      // this.assertSelectionRange(8, 8); //NOTE: this fails in IE, the range is 0 -> 0 (TEST_SUITE=sauce)
    }

    ['@test [DEPRECATED] sends an action with `{{input enter="foo"}}` when <enter> is pressed'](assert) {
      assert.expect(4);

      expectDeprecation(() => {
        this.render(`{{input enter='foo'}}`, {
          actions: {
            foo(value, event) {
              assert.ok(true, 'action was triggered');
              assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed.');
            }
          }
        });
      }, 'Please refactor `{{input enter="foo"}}` to `{{input enter=(action "foo")}}. (\'-top-level\' @ L1:C0) ');
      expectDeprecation(() => {
        this.triggerEvent('keyup', { keyCode: 13 });
      }, 'Passing actions to components as strings (like {{input enter="foo"}}) is deprecated. Please use closure actions instead ({{input enter=(action "foo")}})');
    }

    ['@test sends an action with `{{input enter=(action "foo")}}` when <enter> is pressed'](assert) {
      assert.expect(2);
      this.render(`{{input enter=(action 'foo')}}`, {
        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });

      this.triggerEvent('keyup', {
        keyCode: 13
      });
    }

    ['@test [DEPRECATED] sends an action with `{{input key-press="foo"}}` is pressed'](assert) {
      assert.expect(4);

      expectDeprecation(() => {
        this.render(`{{input value=value key-press='foo'}}`, {
          value: 'initial',

          actions: {
            foo(value, event) {
              assert.ok(true, 'action was triggered');
              assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
            }
          }
        });
      }, 'Please refactor `{{input key-press="foo"}}` to `{{input key-press=(action "foo")}}. (\'-top-level\' @ L1:C0) ');

      expectDeprecation(() => {
        this.triggerEvent('keypress', { keyCode: 65 });
      }, 'Passing actions to components as strings (like {{input key-press="foo"}}) is deprecated. Please use closure actions instead ({{input key-press=(action "foo")}})');
    }

    ['@test sends an action with `{{input key-press=(action "foo")}}` is pressed'](assert) {
      assert.expect(2);

      this.render(`{{input value=value key-press=(action 'foo')}}`, {
        value: 'initial',

        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });

      this.triggerEvent('keypress', { keyCode: 65 });
    }

    ['@test sends an action to the parent level when `bubbles=true` is provided'](assert) {
      assert.expect(1);

      let ParentComponent = _helpers.Component.extend({
        change() {
          assert.ok(true, 'bubbled upwards');
        }
      });

      this.registerComponent('x-parent', {
        ComponentClass: ParentComponent,
        template: `{{input bubbles=true}}`
      });
      this.render(`{{x-parent}}`);

      this.triggerEvent('change');
    }

    ['@test triggers `focus-in` when focused'](assert) {
      let wasFocused = false;

      this.render(`{{input focus-in=(action 'foo')}}`, {
        actions: {
          foo() {
            wasFocused = true;
          }
        }
      });

      this.runTask(() => {
        this.$input().focus();
      });

      assert.ok(wasFocused, 'action was triggered');
    }

    ['@test sends `insert-newline` when <enter> is pressed'](assert) {
      assert.expect(2);

      this.render(`{{input insert-newline=(action 'foo')}}`, {
        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });

      this.triggerEvent('keyup', {
        keyCode: 13
      });
    }

    ['@test [DEPRECATED] sends an action with `{{input escape-press="foo"}}` when <escape> is pressed'](assert) {
      assert.expect(4);

      expectDeprecation(() => {
        this.render(`{{input escape-press='foo'}}`, {
          actions: {
            foo(value, event) {
              assert.ok(true, 'action was triggered');
              assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
            }
          }
        });
      }, 'Please refactor `{{input escape-press="foo"}}` to `{{input escape-press=(action "foo")}}. (\'-top-level\' @ L1:C0) ');

      expectDeprecation(() => {
        this.triggerEvent('keyup', { keyCode: 27 });
      }, 'Passing actions to components as strings (like {{input escape-press="foo"}}) is deprecated. Please use closure actions instead ({{input escape-press=(action "foo")}})');
    }

    ['@test sends an action with `{{input escape-press=(action "foo")}}` when <escape> is pressed'](assert) {
      assert.expect(2);

      this.render(`{{input escape-press=(action 'foo')}}`, {
        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });

      this.triggerEvent('keyup', { keyCode: 27 });
    }

    ['@test [DEPRECATED] sends an action with `{{input key-down="foo"}}` when a key is pressed'](assert) {
      assert.expect(4);

      expectDeprecation(() => {
        this.render(`{{input key-down='foo'}}`, {
          actions: {
            foo(value, event) {
              assert.ok(true, 'action was triggered');
              assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
            }
          }
        });
      }, 'Please refactor `{{input key-down="foo"}}` to `{{input key-down=(action "foo")}}. (\'-top-level\' @ L1:C0) ');

      expectDeprecation(() => {
        this.triggerEvent('keydown', { keyCode: 65 });
      }, 'Passing actions to components as strings (like {{input key-down="foo"}}) is deprecated. Please use closure actions instead ({{input key-down=(action "foo")}})');
    }

    ['@test sends an action with `{{input key-down=(action "foo")}}` when a key is pressed'](assert) {
      assert.expect(2);

      this.render(`{{input key-down=(action 'foo')}}`, {
        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });

      this.triggerEvent('keydown', { keyCode: 65 });
    }

    ['@test [DEPRECATED] sends an action with `{{input key-up="foo"}}` when a key is pressed'](assert) {
      assert.expect(4);

      expectDeprecation(() => {
        this.render(`{{input key-up='foo'}}`, {
          actions: {
            foo(value, event) {
              assert.ok(true, 'action was triggered');
              assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
            }
          }
        });
      }, 'Please refactor `{{input key-up="foo"}}` to `{{input key-up=(action "foo")}}. (\'-top-level\' @ L1:C0) ');

      expectDeprecation(() => {
        this.triggerEvent('keyup', { keyCode: 65 });
      }, 'Passing actions to components as strings (like {{input key-up="foo"}}) is deprecated. Please use closure actions instead ({{input key-up=(action "foo")}})');
    }

    ['@test [DEPRECATED] sends an action with `{{input key-up=(action "foo")}}` when a key is pressed'](assert) {
      assert.expect(2);

      this.render(`{{input key-up=(action 'foo')}}`, {
        actions: {
          foo(value, event) {
            assert.ok(true, 'action was triggered');
            assert.ok(event instanceof _views.jQuery.Event, 'jQuery event was passed');
          }
        }
      });
      this.triggerEvent('keyup', { keyCode: 65 });
    }

    ['@test GH#14727 can render a file input after having had render an input of other type']() {
      this.render(`{{input type="text"}}{{input type="file"}}`);

      this.assert.equal(this.$input()[0].type, 'text');
      this.assert.equal(this.$input()[1].type, 'file');
    }
  });

  (0, _testCase.moduleFor)('Helpers test: {{input}} with dynamic type', class extends InputRenderingTest {
    ['@test a bound property can be used to determine type']() {
      this.render(`{{input type=type}}`, { type: 'password' });

      this.assertAttr('type', 'password');

      this.runTask(() => this.rerender());

      this.assertAttr('type', 'password');

      this.runTask(() => (0, _metal.set)(this.context, 'type', 'text'));

      this.assertAttr('type', 'text');

      this.runTask(() => (0, _metal.set)(this.context, 'type', 'password'));

      this.assertAttr('type', 'password');
    }

    ['@test a subexpression can be used to determine type']() {
      this.render(`{{input type=(if isTruthy trueType falseType)}}`, {
        isTruthy: true,
        trueType: 'text',
        falseType: 'password'
      });

      this.assertAttr('type', 'text');

      this.runTask(() => this.rerender());

      this.assertAttr('type', 'text');

      this.runTask(() => (0, _metal.set)(this.context, 'isTruthy', false));

      this.assertAttr('type', 'password');

      this.runTask(() => (0, _metal.set)(this.context, 'isTruthy', true));

      this.assertAttr('type', 'text');
    }

    ['@test GH16256 input macro does not modify params in place']() {
      this.registerComponent('my-input', {
        template: `{{input type=inputType}}`
      });

      this.render(`{{my-input inputType=firstType}}{{my-input inputType=secondType}}`, {
        firstType: 'password',
        secondType: 'email'
      });

      let inputs = this.element.querySelectorAll('input');
      this.assert.equal(inputs.length, 2, 'there are two inputs');
      this.assert.equal(inputs[0].getAttribute('type'), 'password');
      this.assert.equal(inputs[1].getAttribute('type'), 'email');
    }
  });

  (0, _testCase.moduleFor)(`Helpers test: {{input type='checkbox'}}`, class extends InputRenderingTest {
    ['@test dynamic attributes']() {
      this.render(`{{input
      type='checkbox'
      disabled=disabled
      name=name
      checked=checked
      tabindex=tabindex
    }}`, {
        disabled: false,
        name: 'original-name',
        checked: false,
        tabindex: 10
      });

      this.assertSingleCheckbox();
      this.assertNotDisabled();
      this.assertAttr('name', 'original-name');
      this.assertAttr('tabindex', '10');

      this.runTask(() => this.rerender());

      this.assertSingleCheckbox();
      this.assertNotDisabled();
      this.assertAttr('name', 'original-name');
      this.assertAttr('tabindex', '10');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'disabled', true);
        (0, _metal.set)(this.context, 'name', 'updated-name');
        (0, _metal.set)(this.context, 'tabindex', 11);
      });

      this.assertSingleCheckbox();
      this.assertDisabled();
      this.assertAttr('name', 'updated-name');
      this.assertAttr('tabindex', '11');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'disabled', false);
        (0, _metal.set)(this.context, 'name', 'original-name');
        (0, _metal.set)(this.context, 'tabindex', 10);
      });

      this.assertSingleCheckbox();
      this.assertNotDisabled();
      this.assertAttr('name', 'original-name');
      this.assertAttr('tabindex', '10');
    }

    ['@test `value` property assertion']() {
      expectAssertion(() => {
        this.render(`{{input type="checkbox" value=value}}`, {
          value: 'value'
        });
      }, /you must use `checked=/);
    }

    ['@test with a bound type']() {
      this.render(`{{input type=inputType checked=isChecked}}`, {
        inputType: 'checkbox',
        isChecked: true
      });

      this.assertSingleCheckbox();
      this.assertCheckboxIsChecked();

      this.runTask(() => this.rerender());

      this.assertCheckboxIsChecked();

      this.runTask(() => (0, _metal.set)(this.context, 'isChecked', false));

      this.assertCheckboxIsNotChecked();

      this.runTask(() => (0, _metal.set)(this.context, 'isChecked', true));

      this.assertCheckboxIsChecked();
    }

    ['@test native click changes check property']() {
      this.render(`{{input type="checkbox"}}`);

      this.assertSingleCheckbox();
      this.assertCheckboxIsNotChecked();
      this.$input()[0].click();
      this.assertCheckboxIsChecked();
      this.$input()[0].click();
      this.assertCheckboxIsNotChecked();
    }

    ['@test with static values']() {
      this.render(`{{input type="checkbox" disabled=false tabindex=10 name="original-name" checked=false}}`);

      this.assertSingleCheckbox();
      this.assertCheckboxIsNotChecked();
      this.assertNotDisabled();
      this.assertAttr('tabindex', '10');
      this.assertAttr('name', 'original-name');

      this.runTask(() => this.rerender());

      this.assertSingleCheckbox();
      this.assertCheckboxIsNotChecked();
      this.assertNotDisabled();
      this.assertAttr('tabindex', '10');
      this.assertAttr('name', 'original-name');
    }
  });

  (0, _testCase.moduleFor)(`Helpers test: {{input type='text'}}`, class extends InputRenderingTest {
    ['@test null values']() {
      let attributes = ['disabled', 'placeholder', 'name', 'maxlength', 'size', 'tabindex'];

      this.render(`
      {{input type="text"
        disabled=disabled
        value=value
        placeholder=placeholder
        name=name
        maxlength=maxlength
        size=size
        tabindex=tabindex
      }}`, {
        disabled: null,
        value: null,
        placeholder: null,
        name: null,
        maxlength: null,
        size: null,
        tabindex: null
      });

      this.assertValue('');
      this.assertAllAttrs(attributes, undefined);

      this.runTask(() => this.rerender());

      this.assertValue('');
      this.assertAllAttrs(attributes, undefined);

      this.runTask(() => {
        (0, _metal.set)(this.context, 'disabled', true);
        (0, _metal.set)(this.context, 'value', 'Updated value');
        (0, _metal.set)(this.context, 'placeholder', 'Updated placeholder');
        (0, _metal.set)(this.context, 'name', 'updated-name');
        (0, _metal.set)(this.context, 'maxlength', 11);
        (0, _metal.set)(this.context, 'size', 21);
        (0, _metal.set)(this.context, 'tabindex', 31);
      });

      this.assertDisabled();
      this.assertValue('Updated value');
      this.assertAttr('placeholder', 'Updated placeholder');
      this.assertAttr('name', 'updated-name');
      this.assertAttr('maxlength', '11');
      this.assertAttr('size', '21');
      this.assertAttr('tabindex', '31');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'disabled', null);
        (0, _metal.set)(this.context, 'value', null);
        (0, _metal.set)(this.context, 'placeholder', null);
        (0, _metal.set)(this.context, 'name', null);
        (0, _metal.set)(this.context, 'maxlength', null);
        // set(this.context, 'size', null); //NOTE: this fails with `Error: Failed to set the 'size' property on 'HTMLInputElement': The value provided is 0, which is an invalid size.` (TEST_SUITE=sauce)
        (0, _metal.set)(this.context, 'tabindex', null);
      });

      this.assertAttr('disabled', undefined);
      this.assertValue('');
      // this.assertAttr('placeholder', undefined); //NOTE: this fails with a value of "null" (TEST_SUITE=sauce)
      // this.assertAttr('name', undefined); //NOTE: this fails with a value of "null" (TEST_SUITE=sauce)
      this.assertAttr('maxlength', undefined);
      // this.assertAttr('size', undefined); //NOTE: re-enable once `size` bug above has been addressed
      this.assertAttr('tabindex', undefined);
    }
  });

  // These are the permutations of the set:
  // ['type="range"', 'min="-5" max="50"', 'value="%x"']
  ['type="range" min="-5" max="50" value="%x"', 'type="range" value="%x" min="-5" max="50"', 'min="-5" max="50" type="range" value="%x"', 'min="-5" max="50" value="%x" type="range"', 'value="%x" min="-5" max="50" type="range"', 'value="%x" type="range" min="-5" max="50"'].forEach(attrs => {
    (0, _testCase.moduleFor)(`[GH#15675] Helpers test: {{input ${attrs}}}`, class extends InputRenderingTest {
      renderInput(value = 25) {
        this.render(`{{input ${attrs.replace('%x', value)}}}`);
      }

      ['@test value over default max but below set max is kept']() {
        this.renderInput('25');
        this.assertValue('25');
      }

      ['@test value below default min but above set min is kept']() {
        this.renderInput('-2');
        this.assertValue('-2');
      }

      ['@test in the valid default range is kept']() {
        this.renderInput('5');
        this.assertValue('5');
      }

      ['@test value above max is reset to max']() {
        this.renderInput('55');
        this.assertValue('50');
      }

      ['@test value below min is reset to min']() {
        this.renderInput('-10');
        this.assertValue('-5');
      }
    });
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/loc-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/string'], function (_testCase, _metal, _string) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{loc}}', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);
      (0, _string._setStrings)({
        'Hello Friend': 'Hallo Freund',
        Hello: 'Hallo, %@'
      });
    }

    teardown() {
      super.teardown();
      (0, _string._setStrings)({});
    }

    ['@test it lets the original value through by default']() {
      this.render(`{{loc "Hiya buddy!"}}`);
      this.assertText('Hiya buddy!', 'the unlocalized string is correct');
      this.runTask(() => this.rerender());
      this.assertText('Hiya buddy!', 'the unlocalized string is correct after rerender');
    }

    ['@test it localizes a simple string']() {
      this.render(`{{loc "Hello Friend"}}`);
      this.assertText('Hallo Freund', 'the localized string is correct');
      this.runTask(() => this.rerender());
      this.assertText('Hallo Freund', 'the localized string is correct after rerender');
    }

    ['@test it takes passed formats into an account']() {
      this.render(`{{loc "%@, %@" "Hello" "Mr. Pitkin"}}`);
      this.assertText('Hello, Mr. Pitkin', 'the formatted string is correct');
      this.runTask(() => this.rerender());
      this.assertText('Hello, Mr. Pitkin', 'the formatted string is correct after rerender');
    }

    ['@test it updates when bound params change']() {
      this.render(`{{loc simple}} - {{loc personal 'Mr. Pitkin'}}`, {
        simple: 'Hello Friend',
        personal: 'Hello'
      });
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct');

      this.runTask(() => this.rerender());
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct after rerender');

      this.runTask(() => (0, _metal.set)(this.context, 'simple', "G'day mate"));
      this.assertText("G'day mate - Hallo, Mr. Pitkin", 'the bound value is correct after update');

      this.runTask(() => (0, _metal.set)(this.context, 'simple', 'Hello Friend'));
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct after reset');
    }

    ['@test it updates when nested bound params change']() {
      this.render(`{{loc greetings.simple}} - {{loc greetings.personal 'Mr. Pitkin'}}`, {
        greetings: {
          simple: 'Hello Friend',
          personal: 'Hello'
        }
      });
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct');

      this.runTask(() => this.rerender());
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct after rerender');

      this.runTask(() => (0, _metal.set)(this.context, 'greetings.simple', "G'day mate"));
      this.assertText("G'day mate - Hallo, Mr. Pitkin", 'the bound value is correct after interior mutation');

      this.runTask(() => (0, _metal.set)(this.context, 'greetings', {
        simple: 'Hello Friend',
        personal: 'Hello'
      }));
      this.assertText('Hallo Freund - Hallo, Mr. Pitkin', 'the bound value is correct after replacement');
    }

    ['@test it can be overriden']() {
      this.registerHelper('loc', () => 'Yup');
      this.render(`{{loc greeting}}`, {
        greeting: 'Hello Friend'
      });
      this.assertText('Yup', 'the localized string is correct');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/log-test', ['@ember/-internals/glimmer/tests/utils/test-case'], function (_testCase) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{log}}', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);
      /* eslint-disable no-console */
      this.originalLog = console.log;
      this.logCalls = [];
      console.log = (...args) => {
        this.logCalls.push(...args);
        /* eslint-enable no-console */
      };
    }

    teardown() {
      super.teardown();
      /* eslint-disable no-console */
      console.log = this.originalLog;
      /* eslint-enable no-console */
    }

    assertLog(values) {
      this.assertText('');
      this.assert.strictEqual(this.logCalls.length, values.length);

      for (let i = 0, len = values.length; i < len; i++) {
        this.assert.strictEqual(this.logCalls[i], values[i]);
      }
    }

    ['@test correctly logs primitives']() {
      this.render(`{{log "one" 1 true}}`);

      this.assertLog(['one', 1, true]);
    }

    ['@test correctly logs a property']() {
      this.render(`{{log value}}`, {
        value: 'one'
      });

      this.assertLog(['one']);
    }

    ['@test correctly logs multiple arguments']() {
      this.render(`{{log "my variable:" value}}`, {
        value: 'one'
      });

      this.assertLog(['my variable:', 'one']);
    }

    ['@test correctly logs `this`']() {
      this.render(`{{log this}}`);

      this.assertLog([this.context]);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/mut-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/test-helpers'], function (_testCase, _helpers, _metal, _testHelpers) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{mut}}', class extends _testCase.RenderingTest {
    ['@test a simple mutable binding using `mut` propagates properly']() {
      let bottom;

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{setMe}}'
      });

      this.registerComponent('middle-mut', {
        template: '{{bottom-mut setMe=value}}'
      });

      this.render('{{middle-mut value=(mut val)}}', {
        val: 12
      });

      this.assertText('12', 'the data propagated downwards');

      this.assertStableRerender();

      this.runTask(() => bottom.attrs.setMe.update(13));

      this.assertText('13', 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 13, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 13, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 13, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(bottom, 'setMe', 14));

      this.assertText('14', 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 14, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 14, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 14, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(this.context, 'val', 12));

      this.assertText('12');
    }

    ['@test a simple mutable binding using `mut` inserts into the DOM']() {
      let bottom, middle;

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{setMe}}'
      });

      this.registerComponent('middle-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            middle = this;
          }
        }),
        template: '{{bottom-mut setMe=(mut value)}}'
      });

      this.render('{{middle-mut value=(mut val)}}', {
        val: 12
      });

      this.assertText('12', 'the data propagated downwards');

      this.assertStableRerender();

      this.runTask(() => bottom.attrs.setMe.update(13));

      this.assertText('13', 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 13, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 13, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(middle, 'value'), 13, "the set propagated to middle's prop");
      this.assert.strictEqual(middle.attrs.value.value, 13, "the set propagated to middle's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 13, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(bottom, 'setMe', 14));

      this.assertText('14', 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 14, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 14, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(middle, 'value'), 14, "the set propagated to middle's prop");
      this.assert.strictEqual(middle.attrs.value.value, 14, "the set propagated to middle's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 14, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(this.context, 'val', 12));

      this.assertText('12');
    }

    ['@test passing a literal results in a assertion']() {
      this.registerComponent('bottom-mut', { template: '{{setMe}}' });

      expectAssertion(() => {
        this.render('{{bottom-mut setMe=(mut "foo bar")}}');
      }, 'You can only pass a path to mut');
    }

    ['@test passing the result of a helper invocation results in an assertion']() {
      this.registerComponent('bottom-mut', { template: '{{setMe}}' });

      expectAssertion(() => {
        this.render('{{bottom-mut setMe=(mut (concat "foo" " " "bar"))}}');
      }, 'You can only pass a path to mut');
    }

    // See https://github.com/emberjs/ember.js/commit/807a0cd for an explanation of this test
    ['@test using a string value through middle tier does not trigger assertion (due to the auto-mut transform)']() {
      let bottom;

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{stuff}}'
      });

      this.registerComponent('middle-mut', {
        template: '{{bottom-mut stuff=value}}'
      });

      this.render('{{middle-mut value="foo"}}');

      this.assert.equal((0, _metal.get)(bottom, 'stuff'), 'foo', 'the data propagated');
      this.assertText('foo');

      this.assertStableRerender();

      // No U-R for this test
    }

    ['@test {{readonly}} of a {{mut}} is converted into an immutable binding']() {
      let middle, bottom;

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{setMe}}'
      });

      this.registerComponent('middle-mut', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            middle = this;
          }
        }),
        template: '{{bottom-mut setMe=(readonly value)}}'
      });

      this.render('{{middle-mut value=(mut val)}}', {
        val: 12
      });

      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => middle.attrs.value.update(13));

      this.assert.strictEqual((0, _metal.get)(middle, 'value'), 13, "the set took effect on middle's prop");
      this.assert.strictEqual(middle.attrs.value.value, 13, "the set took effect on middle's attr");

      this.runTask(() => (0, _metal.set)(middle, 'value', 14));

      this.assert.strictEqual((0, _metal.get)(middle, 'value'), 14, "the set took effect on middle's prop");
      this.assert.strictEqual(middle.attrs.value.value, 14, "the set took effect on middle's attr");
      this.assert.strictEqual(bottom.attrs.setMe, 14, 'the mutable binding has been converted to an immutable cell');
      this.assertText('14');
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 14, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(this.context, 'val', 12));

      this.assertText('12');
    }

    ['@test mutable bindings work inside of yielded content']() {
      this.registerComponent('bottom-mut', {
        template: '{{yield}}'
      });

      this.registerComponent('middle-mut', {
        template: '{{#bottom-mut}}{{model.name}}{{/bottom-mut}}'
      });

      this.render('{{middle-mut model=(mut model)}}', {
        model: { name: 'Matthew Beale' }
      });

      this.assertText('Matthew Beale');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'Joel Kang'));

      this.assertText('Joel Kang');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { name: 'Matthew Beale' }));

      this.assertText('Matthew Beale');
    }

    ['@test a simple mutable binding using {{mut}} is available in hooks']() {
      let bottom;
      let willRender = [];
      let didInsert = [];

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          willRender() {
            willRender.push((0, _metal.get)(this, 'setMe'));
          },
          didInsertElement() {
            didInsert.push((0, _metal.get)(this, 'setMe'));
            bottom = this;
          }
        }),
        template: '{{setMe}}'
      });

      this.registerComponent('middle-mut', {
        template: '{{bottom-mut setMe=(mut value)}}'
      });

      this.render('{{middle-mut value=(mut val)}}', {
        val: 12
      });

      this.assert.deepEqual(willRender, [12], 'willReceive is [12]');
      this.assert.deepEqual(didInsert, [12], 'didInsert is [12]');
      this.assertText('12');

      this.assertStableRerender();

      this.assert.deepEqual(willRender, [12], 'willReceive is [12]');
      this.assert.deepEqual(didInsert, [12], 'didInsert is [12]');
      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 12, 'the data propagated');

      this.runTask(() => bottom.attrs.setMe.update(13));

      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 13, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 13, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 13, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(bottom, 'setMe', 14));

      this.assert.strictEqual((0, _metal.get)(bottom, 'setMe'), 14, "the set took effect on bottom's prop");
      this.assert.strictEqual(bottom.attrs.setMe.value, 14, "the set took effect on bottom's attr");
      this.assert.strictEqual((0, _metal.get)(this.context, 'val'), 14, 'the set propagated back up');

      this.runTask(() => (0, _metal.set)(this.context, 'val', 12));

      this.assertText('12');
    }

    ['@test a mutable binding with a backing computed property and attribute present in the root of the component is updated when the upstream property invalidates #11023']() {
      let bottom, middle;

      this.registerComponent('bottom-mut', {
        ComponentClass: _helpers.Component.extend({
          thingy: null,
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{thingy}}'
      });

      this.registerComponent('middle-mut', {
        ComponentClass: _helpers.Component.extend({
          baseValue: 12,
          val: (0, _metal.computed)('baseValue', function () {
            return this.get('baseValue');
          }),
          didInsertElement() {
            middle = this;
          }
        }),
        template: '{{bottom-mut thingy=(mut val)}}'
      });

      this.render('{{middle-mut}}');

      this.assert.strictEqual((0, _metal.get)(bottom, 'thingy'), 12, 'data propagated');
      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(middle, 'baseValue', 13));

      this.assert.strictEqual((0, _metal.get)(middle, 'val'), 13, 'the set took effect');
      this.assert.strictEqual(bottom.attrs.thingy.value, 13, "the set propagated down to bottom's attrs");
      this.assert.strictEqual((0, _metal.get)(bottom, 'thingy'), 13, "the set propagated down to bottom's prop");
      this.assertText('13');

      this.runTask(() => (0, _metal.set)(middle, 'baseValue', 12));

      this.assertText('12');
    }

    ['@test automatic mutable bindings exposes a mut cell in attrs']() {
      let inner;

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            inner = this;
          }
        }),
        template: '{{foo}}'
      });

      this.registerComponent('x-outer', {
        template: '{{x-inner foo=bar}}'
      });

      this.render('{{x-outer bar=baz}}', { baz: 'foo' });

      this.assertText('foo');

      this.assertStableRerender();

      this.runTask(() => inner.attrs.foo.update('bar'));

      this.assert.equal(inner.attrs.foo.value, 'bar');
      this.assert.equal((0, _metal.get)(inner, 'foo'), 'bar');
      this.assertText('bar');

      this.runTask(() => inner.attrs.foo.update('foo'));

      this.assertText('foo');
    }

    ['@test automatic mutable bindings tolerate undefined non-stream inputs and attempts to set them']() {
      let inner;

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            inner = this;
          }
        }),
        template: '{{model}}'
      });

      this.registerComponent('x-outer', {
        template: '{{x-inner model=nonexistent}}'
      });

      this.render('{{x-outer}}');

      this.assertText('');

      this.assertStableRerender();

      this.runTask(() => inner.attrs.model.update(42));

      this.assert.equal(inner.attrs.model.value, 42);
      this.assert.equal((0, _metal.get)(inner, 'model'), 42);
      this.assertText('42');

      this.runTask(() => inner.attrs.model.update(undefined));

      this.assertText('');
    }

    ['@test automatic mutable bindings tolerate constant non-stream inputs and attempts to set them']() {
      let inner;

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            inner = this;
          }
        }),
        template: 'hello{{model}}'
      });

      this.registerComponent('x-outer', {
        template: '{{x-inner model=x}}'
      });

      this.render('{{x-outer x="foo"}}');

      this.assertText('hellofoo');

      this.assertStableRerender();

      this.runTask(() => inner.attrs.model.update(42));

      this.assert.equal(inner.attrs.model.value, 42);
      this.assert.equal((0, _metal.get)(inner, 'model'), 42);
      this.assertText('hello42');

      this.runTask(() => inner.attrs.model.update('foo'));

      this.assertText('hellofoo');
    }
  });

  (0, _testCase.moduleFor)('Mutable Bindings used in Computed Properties that are bound as attributeBindings', class extends _testCase.RenderingTest {
    ['@test an attribute binding of a computed property of a 2-way bound attr recomputes when the attr changes']() {
      let input, output;

      this.registerComponent('x-input', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            input = this;
          }
        })
      });

      this.registerComponent('x-output', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['style'],
          didInsertElement() {
            output = this;
          },
          style: (0, _metal.computed)('height', function () {
            let height = this.get('height');
            return (0, _helpers.htmlSafe)(`height: ${height}px;`);
          }),
          height: 20
        }),
        template: '{{height}}'
      });

      this.render('{{x-output height=height}}{{x-input height=(mut height)}}', {
        height: 60
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 60px;') },
        content: '60'
      });

      this.assertStableRerender();

      this.runTask(() => input.attrs.height.update(35));

      this.assert.strictEqual((0, _metal.get)(output, 'height'), 35, 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(this.context, 'height'), 35, 'the set propagated back up');
      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 35px;') },
        content: '35'
      });

      this.runTask(() => (0, _metal.set)(input, 'height', 36));

      this.assert.strictEqual((0, _metal.get)(output, 'height'), 36, 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(this.context, 'height'), 36, 'the set propagated back up');
      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 36px;') },
        content: '36'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'height', 60));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 60px;') },
        content: '60'
      });
      this.assert.strictEqual((0, _metal.get)(input, 'height'), 60);
    }

    ['@test an attribute binding of a computed property with a setter of a 2-way bound attr recomputes when the attr changes']() {
      let input, output;

      this.registerComponent('x-input', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            input = this;
          }
        })
      });

      this.registerComponent('x-output', {
        ComponentClass: _helpers.Component.extend({
          attributeBindings: ['style'],
          didInsertElement() {
            output = this;
          },
          style: (0, _metal.computed)('height', 'width', function () {
            let height = this.get('height');
            let width = this.get('width');
            return (0, _helpers.htmlSafe)(`height: ${height}px; width: ${width}px;`);
          }),
          height: 20,
          width: (0, _metal.computed)('height', {
            get() {
              return this.get('height') * 2;
            },
            set(keyName, width) {
              this.set('height', width / 2);
              return width;
            }
          })
        }),
        template: '{{width}}x{{height}}'
      });

      this.render('{{x-output width=width}}{{x-input width=(mut width)}}', {
        width: 70
      });

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 35px; width: 70px;') },
        content: '70x35'
      });

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(input, 'width', 80));

      this.assert.strictEqual((0, _metal.get)(output, 'width'), 80, 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(this.context, 'width'), 80, 'the set propagated back up');
      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 40px; width: 80px;') },
        content: '80x40'
      });

      this.runTask(() => input.attrs.width.update(90));

      this.assert.strictEqual((0, _metal.get)(output, 'width'), 90, 'the set took effect');
      this.assert.strictEqual((0, _metal.get)(this.context, 'width'), 90, 'the set propagated back up');
      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 45px; width: 90px;') },
        content: '90x45'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'width', 70));

      this.assertComponentElement(this.firstChild, {
        tagName: 'div',
        attrs: { style: (0, _testHelpers.styles)('height: 35px; width: 70px;') },
        content: '70x35'
      });
      this.assert.strictEqual((0, _metal.get)(input, 'width'), 70);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/partial-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_testCase, _metal, _runtime, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{partial}}', class extends _testCase.RenderingTest {
    ['@test should render other templates registered with the container']() {
      this.registerPartial('_subTemplateFromContainer', 'sub-template');

      this.render(`This {{partial "subTemplateFromContainer"}} is pretty great.`);

      this.assertStableRerender();

      this.assertText('This sub-template is pretty great.');
    }

    ['@test should render other slash-separated templates registered with the container']() {
      this.registerPartial('child/_subTemplateFromContainer', 'sub-template');

      this.render(`This {{partial "child/subTemplateFromContainer"}} is pretty great.`);

      this.assertStableRerender();

      this.assertText('This sub-template is pretty great.');
    }

    ['@test should use the current context']() {
      this.registerPartial('_person_name', '{{model.firstName}} {{model.lastName}}');

      this.render('Who is {{partial "person_name"}}?', {
        model: {
          firstName: 'Kris',
          lastName: 'Selden'
        }
      });

      this.assertStableRerender();

      this.assertText('Who is Kris Selden?');

      this.runTask(() => (0, _metal.set)(this.context, 'model.firstName', 'Kelly'));

      this.assertText('Who is Kelly Selden?');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { firstName: 'Kris', lastName: 'Selden' }));

      this.assertText('Who is Kris Selden?');
    }

    ['@test Quoteless parameters passed to {{partial}} perform a bound property lookup of the partial name']() {
      this.registerPartial('_subTemplate', 'sub-template');
      this.registerPartial('_otherTemplate', 'other-template');

      this.render('This {{partial templates.partialName}} is pretty {{partial nonexistent}}great.', {
        templates: { partialName: 'subTemplate' }
      });

      this.assertStableRerender();

      this.assertText('This sub-template is pretty great.');

      this.runTask(() => (0, _metal.set)(this.context, 'templates.partialName', 'otherTemplate'));

      this.assertText('This other-template is pretty great.');

      this.runTask(() => (0, _metal.set)(this.context, 'templates.partialName', null));

      this.assertText('This  is pretty great.');

      this.runTask(() => (0, _metal.set)(this.context, 'templates', { partialName: 'subTemplate' }));

      this.assertText('This sub-template is pretty great.');
    }

    ['@test partial using data from {{#each}}']() {
      this.registerPartial('show-item', '{{item}}');

      this.render(_abstractTestCase.strip`
      {{#each model.items as |item|}}
        {{item}}: {{partial 'show-item'}} |
      {{/each}}`, {
        model: {
          items: (0, _runtime.A)(['apple', 'orange', 'banana'])
        }
      });

      this.assertStableRerender();

      this.assertText('apple: apple |orange: orange |banana: banana |');

      this.runTask(() => this.context.model.items.pushObject('strawberry'));

      this.assertText('apple: apple |orange: orange |banana: banana |strawberry: strawberry |');

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        items: (0, _runtime.A)(['apple', 'orange', 'banana'])
      }));

      this.assertText('apple: apple |orange: orange |banana: banana |');
    }

    ['@test partial using `{{get` on data from {{#with}}']() {
      this.registerPartial('show-id', '{{get item "id"}}');

      this.render(_abstractTestCase.strip`
      {{#with model as |item|}}
        {{item.name}}: {{partial 'show-id'}}
      {{/with}}`, {
        model: { id: 1, name: 'foo' }
      });

      this.assertStableRerender();

      this.assertText('foo: 1');

      this.runTask(() => (0, _metal.set)(this.context, 'model.id', 2));

      this.assertText('foo: 2');

      this.runTask(() => (0, _metal.set)(this.context, 'model.name', 'bar'));

      this.assertText('bar: 2');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { id: 1, name: 'foo' }));

      this.assertText('foo: 1');
    }

    ['@test partial using `{{get` on data from {{#each}}']() {
      this.registerPartial('show-item', '{{get item "id"}}');

      this.render(_abstractTestCase.strip`
      {{#each items as |item|}}
        {{item.id}}: {{partial 'show-item'}} |
      {{/each}}`, {
        items: (0, _runtime.A)([{ id: 1 }, { id: 2 }, { id: 3 }])
      });

      this.assertStableRerender();

      this.assertText('1: 1 |2: 2 |3: 3 |');

      this.runTask(() => this.context.items.pushObject({ id: 4 }));

      this.assertText('1: 1 |2: 2 |3: 3 |4: 4 |');

      this.runTask(() => (0, _metal.set)(this.context, 'items', (0, _runtime.A)([{ id: 1 }, { id: 2 }, { id: 3 }])));

      this.assertText('1: 1 |2: 2 |3: 3 |');
    }

    ['@test partial using conditional on data from {{#each}}']() {
      this.registerPartial('show-item', '{{#if item}}{{item}}{{/if}}');

      this.render(_abstractTestCase.strip`
      {{#each items as |item|}}
        {{item}}: {{partial 'show-item'}} |
      {{/each}}`, {
        items: (0, _runtime.A)(['apple', null, 'orange', 'banana'])
      });

      this.assertStableRerender();

      this.assertText('apple: apple |:  |orange: orange |banana: banana |');

      this.runTask(() => this.context.items.pushObject('strawberry'));

      this.assertText('apple: apple |:  |orange: orange |banana: banana |strawberry: strawberry |');

      this.runTask(() => (0, _metal.set)(this.context, 'items', (0, _runtime.A)(['apple', null, 'orange', 'banana'])));

      this.assertText('apple: apple |:  |orange: orange |banana: banana |');
    }

    ['@test nested partials using data from {{#each}}']() {
      this.registerPartial('_outer-partial', _abstractTestCase.strip`
      [outer: {{name}}] {{partial 'inner-partial'}}
    `);

      this.registerPartial('inner-partial', '[inner: {{name}}]');

      this.render(_abstractTestCase.strip`
      {{#each names as |name i|}}
        {{i}}: {{partial 'outer-partial'}}
      {{/each}}`, {
        names: (0, _runtime.A)(['Alex', 'Ben'])
      });

      this.assertStableRerender();

      this.assertText('0: [outer: Alex] [inner: Alex]1: [outer: Ben] [inner: Ben]');

      this.runTask(() => this.context.names.pushObject('Sophie'));

      this.assertText('0: [outer: Alex] [inner: Alex]1: [outer: Ben] [inner: Ben]2: [outer: Sophie] [inner: Sophie]');

      this.runTask(() => (0, _metal.set)(this.context, 'names', (0, _runtime.A)(['Alex', 'Ben'])));

      this.assertText('0: [outer: Alex] [inner: Alex]1: [outer: Ben] [inner: Ben]');
    }

    ['@test nested partials within nested `{{#with}}` blocks']() {
      this.registerPartial('_person2-partial', _abstractTestCase.strip`
      {{#with 'Ben' as |person2|}}
        Hi {{person1}} (aged {{age}}) and {{person2}}. {{partial 'person3-partial'}}
      {{/with}}
    `);

      this.registerPartial('_person3-partial', _abstractTestCase.strip`
      {{#with 'Alex' as |person3|}}
        Hi {{person1}} (aged {{age}}), {{person2}} and {{person3}}. {{partial 'person4-partial'}}
      {{/with}}
    `);

      this.registerPartial('_person4-partial', _abstractTestCase.strip`
      {{#with 'Sarah' as |person4|}}
        Hi {{person1}} (aged {{age}}), {{person2}}, {{person3}} and {{person4}}.
      {{/with}}
    `);

      this.render(_abstractTestCase.strip`
      {{#with 'Sophie' as |person1|}}
        Hi {{person1}} (aged {{age}}). {{partial 'person2-partial'}}
      {{/with}}`, { age: 0 });

      this.assertStableRerender();

      this.assertText('Hi Sophie (aged 0). Hi Sophie (aged 0) and Ben. Hi Sophie (aged 0), Ben and Alex. Hi Sophie (aged 0), Ben, Alex and Sarah.');

      this.runTask(() => (0, _metal.set)(this.context, 'age', 1));

      this.assertText('Hi Sophie (aged 1). Hi Sophie (aged 1) and Ben. Hi Sophie (aged 1), Ben and Alex. Hi Sophie (aged 1), Ben, Alex and Sarah.');

      this.runTask(() => (0, _metal.set)(this.context, 'age', 0));

      this.assertText('Hi Sophie (aged 0). Hi Sophie (aged 0) and Ben. Hi Sophie (aged 0), Ben and Alex. Hi Sophie (aged 0), Ben, Alex and Sarah.');
    }

    ['@test dynamic partials in {{#each}}']() {
      this.registerPartial('_odd', 'ODD{{i}}');
      this.registerPartial('_even', 'EVEN{{i}}');

      this.render(_abstractTestCase.strip`
      {{#each model.items as |template i|}}
        {{model.type}}: {{partial template}}
      {{/each}}`, {
        model: {
          items: ['even', 'odd', 'even', 'odd'],
          type: 'number'
        }
      });

      this.assertStableRerender();

      this.assertText('number: EVEN0number: ODD1number: EVEN2number: ODD3');

      this.runTask(() => (0, _metal.set)(this.context, 'model.type', 'integer'));

      this.assertText('integer: EVEN0integer: ODD1integer: EVEN2integer: ODD3');

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        items: ['even', 'odd', 'even', 'odd'],
        type: 'number'
      }));

      this.assertText('number: EVEN0number: ODD1number: EVEN2number: ODD3');
    }

    ['@test dynamic partials in {{#with}}']() {
      this.registerPartial('_thing', '{{t}}');

      this.render(_abstractTestCase.strip`
      {{#with item.thing as |t|}}
        {{partial t}}
      {{else}}
        Nothing!
      {{/with}}`, {
        item: { thing: false }
      });

      this.assertStableRerender();

      this.assertText('Nothing!');

      this.runTask(() => (0, _metal.set)(this.context, 'item.thing', 'thing'));

      this.assertText('thing');

      this.runTask(() => (0, _metal.set)(this.context, 'item', { thing: false }));

      this.assertText('Nothing!');
    }

    ['@test partials which contain contextual components']() {
      this.registerComponent('outer-component', {
        template: '{{yield (hash inner=(component "inner-component" name=name))}}'
      });

      this.registerComponent('inner-component', {
        template: '{{yield (hash name=name)}}'
      });

      this.registerPartial('_some-partial', _abstractTestCase.strip`
      {{#outer.inner as |inner|}}
        inner.name: {{inner.name}}
      {{/outer.inner}}
    `);

      this.render(_abstractTestCase.strip`
      {{#outer-component name=name as |outer|}}
        {{partial 'some-partial'}}
      {{/outer-component}}`, { name: 'Sophie' });

      this.assertStableRerender();

      this.assertText('inner.name: Sophie');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Ben'));

      this.assertText('inner.name: Ben');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Sophie'));

      this.assertText('inner.name: Sophie');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/readonly-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal'], function (_testCase, _helpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{readonly}}', class extends _testCase.RenderingTest {
    ['@test {{readonly}} of a path should work']() {
      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            component = this;
          }
        }),
        template: '{{value}}'
      });

      this.render('{{foo-bar value=(readonly val)}}', {
        val: 12
      });

      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(component, 'value', 13));
      this.assert.notOk(component.attrs.value.update);

      this.assertText('13', 'local property is updated');
      this.assert.equal((0, _metal.get)(this.context, 'val'), 12, 'upstream attribute is not updated');

      // No U-R
    }

    '@test passing an action to {{readonly}} avoids mutable cell wrapping'(assert) {
      assert.expect(4);
      let outer, inner;

      this.registerComponent('x-inner', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            inner = this;
          }
        })
      });

      this.registerComponent('x-outer', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            outer = this;
          }
        }),
        template: '{{x-inner onClick=(readonly onClick)}}'
      });

      this.render('{{x-outer onClick=(action doIt)}}', {
        doIt() {
          assert.ok(true, 'action was called');
        }
      });

      assert.equal(typeof outer.attrs.onClick, 'function', 'function itself is present in outer component attrs');
      outer.attrs.onClick();

      assert.equal(typeof inner.attrs.onClick, 'function', 'function itself is present in inner component attrs');
      inner.attrs.onClick();
    }

    '@test updating a {{readonly}} property from above works'(assert) {
      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            component = this;
          }
        }),
        template: '{{value}}'
      });

      this.render('{{foo-bar value=(readonly thing)}}', {
        thing: 'initial'
      });

      this.assertText('initial');

      this.assertStableRerender();

      assert.strictEqual(component.attrs.value, 'initial', 'no mutable cell');
      assert.strictEqual((0, _metal.get)(component, 'value'), 'initial', 'no mutable cell');
      assert.strictEqual(this.context.thing, 'initial');

      this.runTask(() => (0, _metal.set)(this.context, 'thing', 'updated!'));

      this.assertText('updated!');
      assert.strictEqual(component.attrs.value, 'updated!', 'passed down value was set in attrs');
      assert.strictEqual((0, _metal.get)(component, 'value'), 'updated!', 'passed down value was set');

      this.runTask(() => (0, _metal.set)(this.context, 'thing', 'initial'));

      this.assertText('initial');
    }

    '@test updating a nested path of a {{readonly}}'(assert) {
      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            component = this;
          }
        }),
        template: '{{value.prop}}'
      });

      this.render('{{foo-bar value=(readonly thing)}}', {
        thing: {
          prop: 'initial'
        }
      });

      this.assertText('initial');

      this.assertStableRerender();

      assert.notOk(component.attrs.value.update, 'no update available');
      assert.deepEqual((0, _metal.get)(component, 'value'), { prop: 'initial' });
      assert.deepEqual(this.context.thing, { prop: 'initial' });

      this.runTask(() => (0, _metal.set)(component, 'value.prop', 'updated!'));

      this.assertText('updated!', 'nested path is updated');
      assert.deepEqual((0, _metal.get)(component, 'value'), { prop: 'updated!' });
      assert.deepEqual(this.context.thing, { prop: 'updated!' });

      this.runTask(() => (0, _metal.set)(component, 'value.prop', 'initial'));

      this.assertText('initial');
    }

    ['@test {{readonly}} of a string renders correctly']() {
      let component;

      this.registerComponent('foo-bar', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            component = this;
          }
        }),
        template: '{{value}}'
      });

      this.render('{{foo-bar value=(readonly "12")}}');

      this.assertText('12');

      this.assertStableRerender();

      this.assert.notOk(component.attrs.value.update);
      this.assert.strictEqual((0, _metal.get)(component, 'value'), '12');

      this.runTask(() => (0, _metal.set)(component, 'value', '13'));

      this.assertText('13', 'local property is updated');
      this.assert.strictEqual((0, _metal.get)(component, 'value'), '13');

      this.runTask(() => (0, _metal.set)(component, 'value', '12'));

      this.assertText('12');
    }

    ['@test {{mut}} of a {{readonly}} mutates only the middle and bottom tiers']() {
      let middle, bottom;

      this.registerComponent('x-bottom', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            bottom = this;
          }
        }),
        template: '{{bar}}'
      });

      this.registerComponent('x-middle', {
        ComponentClass: _helpers.Component.extend({
          didInsertElement() {
            middle = this;
          }
        }),
        template: '{{foo}} {{x-bottom bar=(mut foo)}}'
      });

      this.render('{{x-middle foo=(readonly val)}}', {
        val: 12
      });

      this.assertText('12 12');

      this.assertStableRerender();

      this.assert.equal((0, _metal.get)(bottom, 'bar'), 12, "bottom's local bar received the value");
      this.assert.equal((0, _metal.get)(middle, 'foo'), 12, "middle's local foo received the value");

      // updating the mut-cell directly
      this.runTask(() => bottom.attrs.bar.update(13));

      this.assert.equal((0, _metal.get)(bottom, 'bar'), 13, "bottom's local bar was updated after set of bottom's bar");
      this.assert.equal((0, _metal.get)(middle, 'foo'), 13, "middle's local foo was updated after set of bottom's bar");
      this.assertText('13 13');
      this.assert.equal((0, _metal.get)(this.context, 'val'), 12, 'But context val is not updated');

      this.runTask(() => (0, _metal.set)(bottom, 'bar', 14));

      this.assert.equal((0, _metal.get)(bottom, 'bar'), 14, "bottom's local bar was updated after set of bottom's bar");
      this.assert.equal((0, _metal.get)(middle, 'foo'), 14, "middle's local foo was updated after set of bottom's bar");
      this.assertText('14 14');
      this.assert.equal((0, _metal.get)(this.context, 'val'), 12, 'But context val is not updated');

      this.assert.notOk(middle.attrs.foo.update, "middle's foo attr is not a mutable cell");
      this.runTask(() => (0, _metal.set)(middle, 'foo', 15));

      this.assertText('15 15');
      this.assert.equal((0, _metal.get)(middle, 'foo'), 15, "set of middle's foo took effect");
      this.assert.equal((0, _metal.get)(bottom, 'bar'), 15, "bottom's local bar was updated after set of middle's foo");
      this.assert.equal((0, _metal.get)(this.context, 'val'), 12, 'Context val remains unchanged');

      this.runTask(() => (0, _metal.set)(this.context, 'val', 10));

      this.assertText('10 10');
      this.assert.equal((0, _metal.get)(bottom, 'bar'), 10, "bottom's local bar was updated after set of context's val");
      this.assert.equal((0, _metal.get)(middle, 'foo'), 10, "middle's local foo was updated after set of context's val");

      // setting as a normal property
      this.runTask(() => (0, _metal.set)(bottom, 'bar', undefined));

      this.assertText(' ');
      this.assert.equal((0, _metal.get)(bottom, 'bar'), undefined, "bottom's local bar was updated to a falsy value");
      this.assert.equal((0, _metal.get)(middle, 'foo'), undefined, "middle's local foo was updated to a falsy value");

      this.runTask(() => (0, _metal.set)(this.context, 'val', 12));
      this.assertText('12 12', 'bottom and middle were both reset');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/text-area-test', ['@ember/polyfills', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_polyfills, _metal, _testCase, _testHelpers, _abstractTestCase) {
  'use strict';

  class TextAreaRenderingTest extends _testCase.RenderingTest {
    assertTextArea({ attrs, value } = {}) {
      let mergedAttrs = (0, _polyfills.assign)({ class: (0, _testHelpers.classes)('ember-view ember-text-area') }, attrs);
      this.assertComponentElement(this.firstChild, {
        tagName: 'textarea',
        attrs: mergedAttrs
      });

      if (value) {
        this.assert.strictEqual(value, this.firstChild.value);
      }
    }

    triggerEvent(type, options = {}) {
      let event = document.createEvent('Events');
      event.initEvent(type, true, true);
      (0, _polyfills.assign)(event, options);

      this.firstChild.dispatchEvent(event);
    }
  }

  class BoundTextAreaAttributes {
    constructor(cases) {
      this.cases = cases;
    }

    generate({ attribute, first, second }) {
      return {
        [`@test ${attribute}`]() {
          this.render(`{{textarea ${attribute}=value}}`, {
            value: first
          });
          this.assertTextArea({ attrs: { [attribute]: first } });

          this.assertStableRerender();

          this.runTask(() => (0, _metal.set)(this.context, 'value', second));
          this.assertTextArea({ attrs: { [attribute]: second } });

          this.runTask(() => (0, _metal.set)(this.context, 'value', first));
          this.assertTextArea({ attrs: { [attribute]: first } });
        }
      };
    }
  }

  (0, _abstractTestCase.applyMixins)(TextAreaRenderingTest, new BoundTextAreaAttributes([{ attribute: 'placeholder', first: 'Stuff here', second: 'Other stuff' }, { attribute: 'name', first: 'Stuff here', second: 'Other stuff' }, { attribute: 'title', first: 'Stuff here', second: 'Other stuff' }, { attribute: 'maxlength', first: '1', second: '2' }, { attribute: 'rows', first: '1', second: '2' }, { attribute: 'cols', first: '1', second: '2' }, { attribute: 'tabindex', first: '1', second: '2' }]));

  (0, _testCase.moduleFor)('Helpers test: {{textarea}}', class extends TextAreaRenderingTest {
    ['@test Should insert a textarea'](assert) {
      this.render('{{textarea}}');

      assert.equal(this.$('textarea').length, 1);

      this.assertStableRerender();
    }

    ['@test Should respect disabled'](assert) {
      this.render('{{textarea disabled=disabled}}', {
        disabled: true
      });
      assert.ok(this.$('textarea').is(':disabled'));
    }

    ['@test Should respect disabled when false'](assert) {
      this.render('{{textarea disabled=disabled}}', {
        disabled: false
      });
      assert.ok(this.$('textarea').is(':not(:disabled)'));
    }

    ['@test Should become disabled when the context changes'](assert) {
      this.render('{{textarea disabled=disabled}}');
      assert.ok(this.$('textarea').is(':not(:disabled)'));

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'disabled', true));
      assert.ok(this.$('textarea').is(':disabled'));

      this.runTask(() => (0, _metal.set)(this.context, 'disabled', false));
      assert.ok(this.$('textarea').is(':not(:disabled)'));
    }

    ['@test Should bind its contents to the specified value']() {
      this.render('{{textarea value=model.val}}', {
        model: { val: 'A beautiful day in Seattle' }
      });
      this.assertTextArea({ value: 'A beautiful day in Seattle' });

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'model.val', 'Auckland'));
      this.assertTextArea({ value: 'Auckland' });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { val: 'A beautiful day in Seattle' }));
      this.assertTextArea({ value: 'A beautiful day in Seattle' });
    }

    ['@test GH#14001 Should correctly handle an empty string bound value']() {
      this.render('{{textarea value=message}}', { message: '' });

      this.assert.strictEqual(this.firstChild.value, '');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'message', 'hello'));

      this.assert.strictEqual(this.firstChild.value, 'hello');

      this.runTask(() => (0, _metal.set)(this.context, 'message', ''));

      this.assert.strictEqual(this.firstChild.value, '');
    }

    ['@test should update the value for `cut` / `input` / `change` events']() {
      this.render('{{textarea value=model.val}}', {
        model: { val: 'A beautiful day in Seattle' }
      });
      this.assertTextArea({ value: 'A beautiful day in Seattle' });

      this.assertStableRerender();

      this.runTask(() => {
        this.firstChild.value = 'Auckland';
        this.triggerEvent('cut');
      });
      this.assertTextArea({ value: 'Auckland' });

      this.runTask(() => {
        this.firstChild.value = 'Hope';
        this.triggerEvent('paste');
      });
      this.assertTextArea({ value: 'Hope' });

      this.runTask(() => {
        this.firstChild.value = 'Boston';
        this.triggerEvent('input');
      });
      this.assertTextArea({ value: 'Boston' });

      this.runTask(() => (0, _metal.set)(this.context, 'model', { val: 'A beautiful day in Seattle' }));
      this.assertTextArea({ value: 'A beautiful day in Seattle' });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/unbound-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/runtime'], function (_testCase, _abstractTestCase, _metal, _helpers, _runtime) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{unbound}}', class extends _testCase.RenderingTest {
    ['@test should be able to output a property without binding']() {
      this.render(`<div id="first">{{unbound content.anUnboundString}}</div>`, {
        content: {
          anUnboundString: 'No spans here, son.'
        }
      });

      this.assertText('No spans here, son.');

      this.runTask(() => this.rerender());

      this.assertText('No spans here, son.');

      this.runTask(() => (0, _metal.set)(this.context, 'content.anUnboundString', 'HEY'));

      this.assertText('No spans here, son.');

      this.runTask(() => (0, _metal.set)(this.context, 'content', {
        anUnboundString: 'No spans here, son.'
      }));

      this.assertText('No spans here, son.');
    }

    ['@test should be able to use unbound helper in #each helper']() {
      this.render(`<ul>{{#each items as |item|}}<li>{{unbound item}}</li>{{/each}}</ul>`, {
        items: (0, _runtime.A)(['a', 'b', 'c', 1, 2, 3])
      });

      this.assertText('abc123');

      this.runTask(() => this.rerender());

      this.assertText('abc123');
    }

    ['@test should be able to use unbound helper in #each helper (with objects)']() {
      this.render(`<ul>{{#each items as |item|}}<li>{{unbound item.wham}}</li>{{/each}}</ul>`, {
        items: (0, _runtime.A)([{ wham: 'bam' }, { wham: 1 }])
      });

      this.assertText('bam1');

      this.runTask(() => this.rerender());

      this.assertText('bam1');

      this.runTask(() => this.context.items.setEach('wham', 'HEY'));

      this.assertText('bam1');

      this.runTask(() => (0, _metal.set)(this.context, 'items', (0, _runtime.A)([{ wham: 'bam' }, { wham: 1 }])));

      this.assertText('bam1');
    }

    ['@test it should assert unbound cannot be called with multiple arguments']() {
      let willThrow = () => {
        this.render(`{{unbound foo bar}}`, {
          foo: 'BORK',
          bar: 'BLOOP'
        });
      };

      expectAssertion(willThrow, /unbound helper cannot be called with multiple params or hash params/);
    }

    ['@test should render on attributes']() {
      this.render(`<a href="{{unbound model.foo}}"></a>`, {
        model: { foo: 'BORK' }
      });

      this.assertHTML('<a href="BORK"></a>');

      this.runTask(() => this.rerender());

      this.assertHTML('<a href="BORK"></a>');

      this.runTask(() => (0, _metal.set)(this.context, 'model.foo', 'OOF'));

      this.assertHTML('<a href="BORK"></a>');

      this.runTask(() => (0, _metal.set)(this.context, 'model', { foo: 'BORK' }));

      this.assertHTML('<a href="BORK"></a>');
    }

    ['@test should property escape unsafe hrefs']() {
      let unsafeUrls = (0, _runtime.A)([{
        name: 'Bob',
        url: 'javascript:bob-is-cool'
      }, {
        name: 'James',
        url: 'vbscript:james-is-cool'
      }, {
        name: 'Richard',
        url: 'javascript:richard-is-cool'
      }]);

      this.render(`<ul>{{#each people as |person|}}<li><a href="{{unbound person.url}}">{{person.name}}</a></li>{{/each}}</ul>`, {
        people: unsafeUrls
      });

      let escapedHtml = _abstractTestCase.strip`
      <ul>
        <li>
          <a href="unsafe:javascript:bob-is-cool">Bob</a>
        </li>
        <li>
          <a href="unsafe:vbscript:james-is-cool">James</a>
        </li>
        <li>
          <a href="unsafe:javascript:richard-is-cool">Richard</a>
        </li>
      </ul>
    `;

      this.assertHTML(escapedHtml);

      this.runTask(() => this.rerender());

      this.assertHTML(escapedHtml);

      this.runTask(() => this.context.people.setEach('url', 'http://google.com'));

      this.assertHTML(escapedHtml);

      this.runTask(() => (0, _metal.set)(this.context, 'people', unsafeUrls));

      this.assertHTML(escapedHtml);
    }

    ['@skip helper form updates on parent re-render']() {
      this.render(`{{unbound foo}}`, {
        foo: 'BORK'
      });

      this.assertText('BORK');

      this.runTask(() => this.rerender());

      this.assertText('BORK');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'OOF'));

      this.assertText('BORK');

      this.runTask(() => this.rerender());

      this.assertText('OOF');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', ''));

      this.assertText('OOF');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'BORK'));

      this.runTask(() => this.rerender());

      this.assertText('BORK');
    }

    // semantics here is not guaranteed
    ['@test sexpr form does not update no matter what']() {
      this.registerHelper('capitalize', args => args[0].toUpperCase());

      this.render(`{{capitalize (unbound foo)}}`, {
        foo: 'bork'
      });

      this.assertText('BORK');

      this.runTask(() => this.rerender());

      this.assertText('BORK');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'oof');
        this.rerender();
      });

      this.assertText('BORK');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'blip'));

      this.assertText('BORK');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'bork');
        this.rerender();
      });

      this.assertText('BORK');
    }

    ['@test sexpr in helper form does not update on parent re-render']() {
      this.registerHelper('capitalize', params => params[0].toUpperCase());

      this.registerHelper('doublize', params => `${params[0]} ${params[0]}`);

      this.render(`{{capitalize (unbound (doublize foo))}}`, {
        foo: 'bork'
      });

      this.assertText('BORK BORK');

      this.runTask(() => this.rerender());

      this.assertText('BORK BORK');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'oof');
        this.rerender();
      });

      this.assertText('BORK BORK');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', 'blip'));

      this.assertText('BORK BORK');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'foo', 'bork');
        this.rerender();
      });

      this.assertText('BORK BORK');
    }

    ['@test should be able to render an unbound helper invocation']() {
      this.registerHelper('repeat', ([value], { count }) => {
        let a = [];
        while (a.length < count) {
          a.push(value);
        }
        return a.join('');
      });

      this.render(`{{unbound (repeat foo count=bar)}} {{repeat foo count=bar}} {{unbound (repeat foo count=2)}} {{repeat foo count=4}}`, {
        foo: 'X',
        bar: 5
      });

      this.assertText('XXXXX XXXXX XX XXXX');

      this.runTask(() => this.rerender());

      this.assertText('XXXXX XXXXX XX XXXX');

      this.runTask(() => (0, _metal.set)(this.context, 'bar', 1));

      this.assertText('XXXXX X XX XXXX');

      this.runTask(() => (0, _metal.set)(this.context, 'bar', 5));

      this.assertText('XXXXX XXXXX XX XXXX');
    }

    ['@test should be able to render an bound helper invocation mixed with static values']() {
      this.registerHelper('surround', ([prefix, value, suffix]) => `${prefix}-${value}-${suffix}`);

      this.render(_abstractTestCase.strip`
      {{unbound (surround model.prefix model.value "bar")}} {{surround model.prefix model.value "bar"}} {{unbound (surround "bar" model.value model.suffix)}} {{surround "bar" model.value model.suffix}}`, {
        model: {
          prefix: 'before',
          value: 'core',
          suffix: 'after'
        }
      });

      this.assertText('before-core-bar before-core-bar bar-core-after bar-core-after');

      this.runTask(() => this.rerender());

      this.assertText('before-core-bar before-core-bar bar-core-after bar-core-after');

      this.runTask(() => {
        (0, _metal.setProperties)(this.context.model, {
          prefix: 'beforeChanged',
          value: 'coreChanged',
          suffix: 'afterChanged'
        });
      });

      this.assertText('before-core-bar beforeChanged-coreChanged-bar bar-core-after bar-coreChanged-afterChanged');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'model', {
          prefix: 'before',
          value: 'core',
          suffix: 'after'
        });
      });

      this.assertText('before-core-bar before-core-bar bar-core-after bar-core-after');
    }

    ['@test should be able to render unbound forms of multi-arg helpers']() {
      this.registerHelper('fauxconcat', params => params.join(''));

      this.render(`{{fauxconcat model.foo model.bar model.bing}} {{unbound (fauxconcat model.foo model.bar model.bing)}}`, {
        model: {
          foo: 'a',
          bar: 'b',
          bing: 'c'
        }
      });

      this.assertText('abc abc');

      this.runTask(() => this.rerender());

      this.assertText('abc abc');

      this.runTask(() => (0, _metal.set)(this.context, 'model.bar', 'X'));

      this.assertText('aXc abc');

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        foo: 'a',
        bar: 'b',
        bing: 'c'
      }));

      this.assertText('abc abc');
    }

    ['@test should be able to render an unbound helper invocation for helpers with dependent keys']() {
      this.registerHelper('capitalizeName', {
        destroy() {
          this.removeObserver('value.firstName', this, this.recompute);
          this._super(...arguments);
        },

        compute([value]) {
          if (this.get('value')) {
            this.removeObserver('value.firstName', this, this.recompute);
          }
          this.set('value', value);
          this.addObserver('value.firstName', this, this.recompute);
          return value ? (0, _metal.get)(value, 'firstName').toUpperCase() : '';
        }
      });

      this.registerHelper('concatNames', {
        destroy() {
          this.teardown();
          this._super(...arguments);
        },
        teardown() {
          this.removeObserver('value.firstName', this, this.recompute);
          this.removeObserver('value.lastName', this, this.recompute);
        },
        compute([value]) {
          if (this.get('value')) {
            this.teardown();
          }
          this.set('value', value);
          this.addObserver('value.firstName', this, this.recompute);
          this.addObserver('value.lastName', this, this.recompute);
          return (value ? (0, _metal.get)(value, 'firstName') : '') + (value ? (0, _metal.get)(value, 'lastName') : '');
        }
      });

      this.render(`{{capitalizeName person}} {{unbound (capitalizeName person)}} {{concatNames person}} {{unbound (concatNames person)}}`, {
        person: {
          firstName: 'shooby',
          lastName: 'taylor'
        }
      });

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');

      this.runTask(() => this.rerender());

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');

      this.runTask(() => (0, _metal.set)(this.context, 'person.firstName', 'sally'));

      this.assertText('SALLY SHOOBY sallytaylor shoobytaylor');

      this.runTask(() => (0, _metal.set)(this.context, 'person', {
        firstName: 'shooby',
        lastName: 'taylor'
      }));

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');
    }

    ['@test should be able to render an unbound helper invocation in #each helper']() {
      this.registerHelper('capitalize', params => params[0].toUpperCase());

      this.render(`{{#each people as |person|}}{{capitalize person.firstName}} {{unbound (capitalize person.firstName)}}{{/each}}`, {
        people: (0, _runtime.A)([{
          firstName: 'shooby',
          lastName: 'taylor'
        }, {
          firstName: 'cindy',
          lastName: 'taylor'
        }])
      });

      this.assertText('SHOOBY SHOOBYCINDY CINDY');

      this.runTask(() => this.rerender());

      this.assertText('SHOOBY SHOOBYCINDY CINDY');

      this.runTask(() => this.context.people.setEach('firstName', 'chad'));

      this.assertText('CHAD SHOOBYCHAD CINDY');

      this.runTask(() => (0, _metal.set)(this.context, 'people', (0, _runtime.A)([{
        firstName: 'shooby',
        lastName: 'taylor'
      }, {
        firstName: 'cindy',
        lastName: 'taylor'
      }])));

      this.assertText('SHOOBY SHOOBYCINDY CINDY');
    }

    ['@test should be able to render an unbound helper invocation with bound hash options']() {
      this.registerHelper('capitalizeName', {
        destroy() {
          this.removeObserver('value.firstName', this, this.recompute);
          this._super(...arguments);
        },

        compute([value]) {
          if (this.get('value')) {
            this.removeObserver('value.firstName', this, this.recompute);
          }
          this.set('value', value);
          this.addObserver('value.firstName', this, this.recompute);
          return value ? (0, _metal.get)(value, 'firstName').toUpperCase() : '';
        }
      });

      this.registerHelper('concatNames', {
        destroy() {
          this.teardown();
          this._super(...arguments);
        },
        teardown() {
          this.removeObserver('value.firstName', this, this.recompute);
          this.removeObserver('value.lastName', this, this.recompute);
        },
        compute([value]) {
          if (this.get('value')) {
            this.teardown();
          }
          this.set('value', value);
          this.addObserver('value.firstName', this, this.recompute);
          this.addObserver('value.lastName', this, this.recompute);
          return (value ? (0, _metal.get)(value, 'firstName') : '') + (value ? (0, _metal.get)(value, 'lastName') : '');
        }
      });

      this.render(`{{capitalizeName person}} {{unbound (capitalizeName person)}} {{concatNames person}} {{unbound (concatNames person)}}`, {
        person: {
          firstName: 'shooby',
          lastName: 'taylor'
        }
      });

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');

      this.runTask(() => this.rerender());

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');

      this.runTask(() => (0, _metal.set)(this.context, 'person.firstName', 'sally'));

      this.assertText('SALLY SHOOBY sallytaylor shoobytaylor');

      this.runTask(() => (0, _metal.set)(this.context, 'person', {
        firstName: 'shooby',
        lastName: 'taylor'
      }));

      this.assertText('SHOOBY SHOOBY shoobytaylor shoobytaylor');
    }

    ['@test should be able to render bound form of a helper inside unbound form of same helper']() {
      this.render(_abstractTestCase.strip`
      {{#if (unbound model.foo)}}
        {{#if model.bar}}true{{/if}}
        {{#unless model.bar}}false{{/unless}}
      {{/if}}
      {{#unless (unbound model.notfoo)}}
        {{#if model.bar}}true{{/if}}
        {{#unless model.bar}}false{{/unless}}
      {{/unless}}`, {
        model: {
          foo: true,
          notfoo: false,
          bar: true
        }
      });

      this.assertText('truetrue');

      this.runTask(() => this.rerender());

      this.assertText('truetrue');

      this.runTask(() => (0, _metal.set)(this.context, 'model.bar', false));

      this.assertText('falsefalse');

      this.runTask(() => (0, _metal.set)(this.context, 'model', {
        foo: true,
        notfoo: false,
        bar: true
      }));

      this.assertText('truetrue');
    }

    ['@test yielding unbound does not update']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          fooBarInstance = this;
        },
        model: { foo: 'bork' }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (unbound model.foo)}}`
      });

      this.render(`{{#foo-bar as |value|}}{{value}}{{/foo-bar}}`);

      this.assertText('bork');

      this.runTask(() => this.rerender());

      this.assertText('bork');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model.foo', 'oof'));

      this.assertText('bork');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model', { foo: 'bork' }));

      this.assertText('bork');
    }

    ['@test yielding unbound hash does not update']() {
      let fooBarInstance;
      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          fooBarInstance = this;
        },
        model: { foo: 'bork' }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: `{{yield (unbound (hash foo=model.foo))}}`
      });

      this.render(`{{#foo-bar as |value|}}{{value.foo}}{{/foo-bar}}`);

      this.assertText('bork');

      this.runTask(() => this.rerender());

      this.assertText('bork');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model.foo', 'oof'));

      this.assertText('bork');

      this.runTask(() => (0, _metal.set)(fooBarInstance, 'model', { foo: 'bork' }));

      this.assertText('bork');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/helpers/yield-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/helpers'], function (_testCase, _metal, _helpers) {
  'use strict';

  (0, _testCase.moduleFor)('Helpers test: {{yield}} helper', class extends _testCase.RenderingTest {
    ['@test can yield to block']() {
      this.registerComponent('yield-comp', {
        template: '[In layout:] {{yield}}'
      });

      this.render('{{#yield-comp}}[In Block:] {{object.title}}{{/yield-comp}}', {
        object: { title: 'Seattle' }
      });
      this.assertText('[In layout:] [In Block:] Seattle');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'object.title', 'Vancouver'));
      this.assertText('[In layout:] [In Block:] Vancouver');

      this.runTask(() => (0, _metal.set)(this.context, 'object', { title: 'Seattle' }));
      this.assertText('[In layout:] [In Block:] Seattle');
    }

    ['@test templates should yield to block inside a nested component']() {
      this.registerComponent('outer-comp', {
        template: '<div>[In layout:] {{yield}}</div>'
      });
      this.registerComponent('inner-comp', {
        template: '{{#outer-comp}}[In Block:] {{object.title}}{{/outer-comp}}'
      });

      this.render('{{inner-comp object=object}}', {
        object: { title: 'Seattle' }
      });
      this.assertText('[In layout:] [In Block:] Seattle');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'object.title', 'Vancouver'));
      this.assertText('[In layout:] [In Block:] Vancouver');

      this.runTask(() => (0, _metal.set)(this.context, 'object', { title: 'Seattle' }));
      this.assertText('[In layout:] [In Block:] Seattle');
    }

    ['@test templates should yield to block, when the yield is embedded in a each helper']() {
      let list = [1, 2, 3];

      this.registerComponent('outer-comp', {
        template: '{{#each list as |item|}}{{yield}}{{/each}}'
      });

      this.render('{{#outer-comp list=list}}Hello{{/outer-comp}}', {
        list: list
      });
      this.assertText('HelloHelloHello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'list', [4, 5]));
      this.assertText('HelloHello');

      this.runTask(() => (0, _metal.set)(this.context, 'list', list));
      this.assertText('HelloHelloHello');
    }

    ['@test templates should yield to block, when the yield is embedded in a if helper']() {
      this.registerComponent('outer-comp', {
        template: '{{#if boolean}}{{yield}}{{/if}}'
      });

      this.render('{{#outer-comp boolean=boolean}}Hello{{/outer-comp}}', {
        boolean: true
      });
      this.assertText('Hello');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'boolean', false));
      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'boolean', true));
      this.assertText('Hello');
    }

    ['@test simple curlies inside of a yielded clock should work when the yield is nested inside of another view']() {
      this.registerComponent('kiwi-comp', {
        template: '{{#if falsy}}{{else}}{{yield}}{{/if}}'
      });

      this.render('{{#kiwi-comp}}{{text}}{{/kiwi-comp}}', { text: 'ohai' });
      this.assertText('ohai');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'portland'));
      this.assertText('portland');

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'ohai'));
      this.assertText('ohai');
    }

    ['@test nested simple curlies inside of a yielded block should work when the yield is nested inside of another view']() {
      this.registerComponent('parent-comp', {
        template: '{{#if falsy}}{{else}}{{yield}}{{/if}}'
      });
      this.registerComponent('child-comp', {
        template: '{{#if falsy}}{{else}}{{text}}{{/if}}'
      });

      this.render('{{#parent-comp}}{{child-comp text=text}}{{/parent-comp}}', {
        text: 'ohai'
      });
      this.assertText('ohai');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'portland'));
      this.assertText('portland');

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'ohai'));
      this.assertText('ohai');
    }

    ['@test yielding to a non-existent block is not an error']() {
      this.registerComponent('yielding-comp', { template: 'Hello:{{yield}}' });
      this.registerComponent('outer-comp', {
        template: '{{yielding-comp}} {{title}}'
      });

      this.render('{{outer-comp title=title}}', { title: 'Mr. Selden' });

      this.assertText('Hello: Mr. Selden');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'title', 'Mr. Chag'));
      this.assertText('Hello: Mr. Chag');

      this.runTask(() => (0, _metal.set)(this.context, 'title', 'Mr. Selden'));
      this.assertText('Hello: Mr. Selden');
    }

    ['@test yield uses the original context']() {
      let KiwiCompComponent = _helpers.Component.extend({ boundText: 'Inner' });

      this.registerComponent('kiwi-comp', {
        ComponentClass: KiwiCompComponent,
        template: '<p>{{boundText}}</p><p>{{yield}}</p>'
      });

      this.render('{{#kiwi-comp}}{{boundText}}{{/kiwi-comp}}', {
        boundText: 'Original'
      });
      this.assertText('InnerOriginal');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Otherworld'));
      this.assertText('InnerOtherworld');

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Original'));
      this.assertText('InnerOriginal');
    }

    ["@test outer block param doesn't mask inner component property"]() {
      let KiwiCompComponent = _helpers.Component.extend({ boundText: 'Inner' });

      this.registerComponent('kiwi-comp', {
        ComponentClass: KiwiCompComponent,
        template: '<p>{{boundText}}</p><p>{{yield}}</p>'
      });

      this.render('{{#with boundText as |item|}}{{#kiwi-comp}}{{item}}{{/kiwi-comp}}{{/with}}', {
        boundText: 'Outer'
      });
      this.assertText('InnerOuter');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Otherworld'));
      this.assertText('InnerOtherworld');

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Outer'));
      this.assertText('InnerOuter');
    }

    ["@test inner block param doesn't mask yield property"]() {
      let KiwiCompComponent = _helpers.Component.extend({ boundText: 'Inner' });

      this.registerComponent('kiwi-comp', {
        ComponentClass: KiwiCompComponent,
        template: '{{#with boundText as |item|}}<p>{{item}}</p><p>{{yield}}</p>{{/with}}'
      });

      this.render('{{#kiwi-comp}}{{item}}{{/kiwi-comp}}', { item: 'Outer' });
      this.assertText('InnerOuter');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'item', 'Otherworld'));
      this.assertText('InnerOtherworld');

      this.runTask(() => (0, _metal.set)(this.context, 'item', 'Outer'));
      this.assertText('InnerOuter');
    }

    ['@test can bind a block param to a component and use it in yield']() {
      this.registerComponent('kiwi-comp', {
        template: '<p>{{content}}</p><p>{{yield}}</p>'
      });

      this.render('{{#with boundText as |item|}}{{#kiwi-comp content=item}}{{item}}{{/kiwi-comp}}{{/with}}', { boundText: 'Outer' });
      this.assertText('OuterOuter');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Update'));
      this.assertText('UpdateUpdate');

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'Outer'));
      this.assertText('OuterOuter');
    }

    // INUR not need with no data update
    ['@test yield should not introduce a view'](assert) {
      let ParentCompComponent = _helpers.Component.extend({ isParentComponent: true });

      let ChildCompComponent = _helpers.Component.extend({
        didReceiveAttrs() {
          this._super();
          let parentView = this.get('parentView');

          assert.ok(parentView.get('isParentComponent'));
        }
      });

      this.registerComponent('parent-comp', {
        ComponentClass: ParentCompComponent,
        template: '{{yield}}'
      });
      this.registerComponent('child-comp', {
        ComponentClass: ChildCompComponent
      });

      this.render('{{#parent-comp}}{{child-comp}}{{/parent-comp}}');
    }

    ['@test yield with nested components (#3220)']() {
      this.registerComponent('inner-component', { template: '{{yield}}' });
      this.registerComponent('outer-component', {
        template: '{{#inner-component}}<span>{{yield}}</span>{{/inner-component}}'
      });

      this.render('{{#outer-component}}Hello {{boundText}}{{/outer-component}}', {
        boundText: 'world'
      });
      this.assertText('Hello world');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'update'));
      this.assertText('Hello update');

      this.runTask(() => (0, _metal.set)(this.context, 'boundText', 'world'));
      this.assertText('Hello world');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/input-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal'], function (_testCase, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('Input element tests', class extends _testCase.RenderingTest {
    runAttributeTest(attributeName, values) {
      let template = `<input ${attributeName}={{value}}>`;
      this.render(template, { value: values[0] });
      this.assertAttributeHasValue(attributeName, values[0], `${attributeName} is set on initial render`);

      this.runTask(() => this.rerender());
      this.assertAttributeHasValue(attributeName, values[0], `${attributeName} is set on noop rerender`);

      this.setComponentValue(values[1]);
      this.assertAttributeHasValue(attributeName, values[1], `${attributeName} is set on rerender`);

      this.setComponentValue(values[0]);
      this.assertAttributeHasValue(attributeName, values[0], `${attributeName} can be set back to the initial value`);
    }

    runPropertyTest(propertyName, values) {
      let attributeName = propertyName;
      let template = `<input ${attributeName}={{value}}>`;
      this.render(template, { value: values[0] });
      this.assertPropertyHasValue(propertyName, values[0], `${propertyName} is set on initial render`);

      this.runTask(() => this.rerender());
      this.assertPropertyHasValue(propertyName, values[0], `${propertyName} is set on noop rerender`);

      this.setComponentValue(values[1]);
      this.assertPropertyHasValue(propertyName, values[1], `${propertyName} is set on rerender`);

      this.setComponentValue(values[0]);
      this.assertPropertyHasValue(propertyName, values[0], `${propertyName} can be set back to the initial value`);
    }

    runFalsyValueProperty(values) {
      let value = 'value';
      let template = `<input value={{value}}>`;
      this.render(template, { value: values[0] });
      this.assertPropertyHasValue(value, '', `${value} is set on initial render`);

      this.runTask(() => this.rerender());
      this.assertPropertyHasValue(value, '', `${value} is set on noop rerender`);
      this.setComponentValue(values[1]);

      this.assertPropertyHasValue(value, values[1], `${value} is set on rerender`);

      this.setComponentValue(values[0]);
      this.assertPropertyHasValue(value, '', `${value} can be set back to the initial value`);
    }

    ['@test input disabled attribute']() {
      let model = { model: { value: false } };

      this.render(`<input disabled={{model.value}}>`, model);

      this.assert.equal(this.$inputElement().prop('disabled'), false);

      this.runTask(() => this.rerender());

      this.assert.equal(this.$inputElement().prop('disabled'), false);

      this.runTask(() => this.context.set('model.value', true));

      this.assert.equal(this.$inputElement().prop('disabled'), true);
      this.assertHTML('<input disabled="">'); // Note the DOM output is <input disabled>

      this.runTask(() => this.context.set('model.value', 'wat'));

      this.assert.equal(this.$inputElement().prop('disabled'), true);
      this.assertHTML('<input disabled="">'); // Note the DOM output is <input disabled>

      this.runTask(() => this.context.set('model', { value: false }));

      this.assert.equal(this.$inputElement().prop('disabled'), false);
      this.assertHTML('<input>');
    }

    ['@test input value attribute']() {
      this.runPropertyTest('value', ['foo', 'bar']);
    }

    ['@test input placeholder attribute']() {
      this.runAttributeTest('placeholder', ['foo', 'bar']);
    }

    ['@test input name attribute']() {
      this.runAttributeTest('name', ['nam', 'name']);
    }

    ['@test input maxlength attribute']() {
      this.runAttributeTest('maxlength', [2, 3]);
    }

    ['@test input size attribute']() {
      this.runAttributeTest('size', [2, 3]);
    }

    ['@test input tabindex attribute']() {
      this.runAttributeTest('tabindex', [2, 3]);
    }

    ['@test null input value']() {
      this.runFalsyValueProperty([null, 'hello']);
    }

    ['@test undefined input value']() {
      this.runFalsyValueProperty([undefined, 'hello']);
    }

    ['@test undefined `toString` method as input value']() {
      this.runFalsyValueProperty([Object.create(null), 'hello']);
    }

    ['@test cursor position is not lost when updating content']() {
      let template = `<input value={{value}}>`;
      this.render(template, { value: 'hola' });

      this.setDOMValue('hello');
      this.setSelectionRange(1, 3);

      this.setComponentValue('hello');

      this.assertSelectionRange(1, 3);

      // Note: We should eventually get around to testing reseting, however
      // browsers handle `selectionStart` and `selectionEnd` differently
      // when are synthetically testing movement of the cursor.
    }

    ['@test input can be updated multiple times']() {
      let template = `<input value={{value}}>`;
      this.render(template, { value: 'hola' });

      this.assertValue('hola', 'Value is initialised');

      this.setComponentValue('');
      this.assertValue('', 'Value is set in the DOM');

      this.setDOMValue('hola');
      this.setComponentValue('hola');
      this.assertValue('hola', 'Value is updated the first time');

      this.setComponentValue('');
      this.assertValue('', 'Value is updated the second time');
    }

    ['@test DOM is SSOT if value is set']() {
      let template = `<input value={{value}}>`;
      this.render(template, { value: 'hola' });

      this.assertValue('hola', 'Value is initialised');

      this.setComponentValue('hello');

      this.assertValue('hello', 'Value is initialised');

      this.setDOMValue('hola');

      this.assertValue('hola', 'DOM is used');

      this.setComponentValue('bye');

      this.assertValue('bye', 'Value is used');

      // Simulates setting the input to the same value as it already is which won't cause a rerender

      this.setDOMValue('hola');

      this.assertValue('hola', 'DOM is used');

      this.setComponentValue('hola');

      this.assertValue('hola', 'Value is used');
    }

    // private helpers and assertions
    setDOMValue(value) {
      this.inputElement().value = value;
    }

    setComponentValue(value) {
      this.runTask(() => (0, _metal.set)(this.context, 'value', value));
    }

    setSelectionRange(start, end) {
      this.inputElement().selectionStart = start;
      this.inputElement().selectionEnd = end;
    }

    inputElement() {
      return this.$inputElement()[0];
    }

    $inputElement() {
      return this.$('input');
    }

    assertValue(value, message) {
      this.assertPropertyHasValue('value', value, message);
    }

    assertAttributeHasValue(attribute, value, message) {
      this.assert.equal(this.$inputElement().attr(attribute), value, `${attribute} ${message}`);
    }

    assertPropertyHasValue(property, value, message) {
      this.assert.equal(this.$inputElement().prop(property), value, `${property} ${message}`);
    }

    assertSelectionRange(start, end) {
      this.assert.equal(this.inputElement().selectionStart, start);
      this.assert.equal(this.inputElement().selectionEnd, end);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/mount-test', ['@ember/-internals/owner', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/controller', '@ember/-internals/metal', '@ember/engine'], function (_owner, _testCase, _helpers, _controller, _metal, _engine) {
  'use strict';

  if (true /* EMBER_ENGINES_MOUNT_PARAMS */) {
      (0, _testCase.moduleFor)('{{mount}} single param assertion', class extends _testCase.RenderingTest {
        ['@test it asserts that only a single param is passed']() {
          expectAssertion(() => {
            this.render('{{mount "chat" "foo"}}');
          }, /You can only pass a single positional argument to the {{mount}} helper, e.g. {{mount "chat-engine"}}./i);
        }
      });
    } else {
    (0, _testCase.moduleFor)('{{mount}} single param assertion', class extends _testCase.RenderingTest {
      ['@test it asserts that only a single param is passed']() {
        expectAssertion(() => {
          this.render('{{mount "chat" "foo"}}');
        }, /You can only pass a single argument to the {{mount}} helper, e.g. {{mount "chat-engine"}}./i);
      }
    });
  }

  (0, _testCase.moduleFor)('{{mount}} assertions', class extends _testCase.RenderingTest {
    ['@test it asserts when an invalid engine name is provided']() {
      expectAssertion(() => {
        this.render('{{mount engineName}}', { engineName: {} });
      }, /Invalid engine name '\[object Object\]' specified, engine name must be either a string, null or undefined./i);
    }

    ['@test it asserts that the specified engine is registered']() {
      expectAssertion(() => {
        this.render('{{mount "chat"}}');
      }, /You used `{{mount 'chat'}}`, but the engine 'chat' can not be found./i);
    }
  });

  (0, _testCase.moduleFor)('{{mount}} test', class extends _testCase.ApplicationTest {
    constructor() {
      super(...arguments);

      let engineRegistrations = this.engineRegistrations = {};

      this.add('engine:chat', _engine.default.extend({
        router: null,

        init() {
          this._super(...arguments);

          Object.keys(engineRegistrations).forEach(fullName => {
            this.register(fullName, engineRegistrations[fullName]);
          });
        }
      }));

      this.addTemplate('index', '{{mount "chat"}}');
    }

    ['@test it boots an engine, instantiates its application controller, and renders its application template'](assert) {
      this.engineRegistrations['template:application'] = (0, _helpers.compile)('<h2>Chat here, {{username}}</h2>', { moduleName: 'my-app/templates/application.hbs' });

      let controller;

      this.engineRegistrations['controller:application'] = _controller.default.extend({
        username: 'dgeb',

        init() {
          this._super();
          controller = this;
        }
      });

      return this.visit('/').then(() => {
        assert.ok(controller, "engine's application controller has been instantiated");

        let engineInstance = (0, _owner.getOwner)(controller);
        assert.strictEqual((0, _engine.getEngineParent)(engineInstance), this.applicationInstance, 'engine instance has the application instance as its parent');

        this.assertInnerHTML('<h2>Chat here, dgeb</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'username', 'chancancode'));

        this.assertInnerHTML('<h2>Chat here, chancancode</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'username', 'dgeb'));

        this.assertInnerHTML('<h2>Chat here, dgeb</h2>');
      });
    }

    ['@test it emits a useful backtracking re-render assertion message']() {
      this.router.map(function () {
        this.route('route-with-mount');
      });

      this.addTemplate('index', '');
      this.addTemplate('route-with-mount', '{{mount "chat"}}');

      this.engineRegistrations['template:application'] = (0, _helpers.compile)('hi {{person.name}} [{{component-with-backtracking-set person=person}}]', { moduleName: 'my-app/templates/application.hbs' });
      this.engineRegistrations['controller:application'] = _controller.default.extend({
        person: { name: 'Alex' }
      });

      this.engineRegistrations['template:components/component-with-backtracking-set'] = (0, _helpers.compile)('[component {{person.name}}]', {
        moduleName: 'my-app/templates/components/component-with-backtracking-set.hbs'
      });
      this.engineRegistrations['component:component-with-backtracking-set'] = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          this.set('person.name', 'Ben');
        }
      });

      let expectedBacktrackingMessage = /modified "person\.name" twice on \[object Object\] in a single render\. It was rendered in "template:my-app\/templates\/route-with-mount.hbs" \(in "engine:chat"\) and modified in "component:component-with-backtracking-set" \(in "engine:chat"\)/;

      return this.visit('/').then(() => {
        expectAssertion(() => {
          this.visit('/route-with-mount');
        }, expectedBacktrackingMessage);
      });
    }

    ['@test it renders with a bound engine name']() {
      this.router.map(function () {
        this.route('bound-engine-name');
      });
      let controller;
      this.add('controller:bound-engine-name', _controller.default.extend({
        engineName: null,
        init() {
          this._super();
          controller = this;
        }
      }));
      this.addTemplate('bound-engine-name', '{{mount engineName}}');

      this.add('engine:foo', _engine.default.extend({
        router: null,
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)('<h2>Foo Engine</h2>', {
            moduleName: 'my-app/templates/application.hbs'
          }));
        }
      }));
      this.add('engine:bar', _engine.default.extend({
        router: null,
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)('<h2>Bar Engine</h2>', {
            moduleName: 'my-app/templates/application.hbs'
          }));
        }
      }));

      return this.visit('/bound-engine-name').then(() => {
        this.assertInnerHTML('<!---->');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', 'foo'));

        this.assertInnerHTML('<h2>Foo Engine</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', undefined));

        this.assertInnerHTML('<!---->');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', 'foo'));

        this.assertInnerHTML('<h2>Foo Engine</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', 'bar'));

        this.assertInnerHTML('<h2>Bar Engine</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', 'foo'));

        this.assertInnerHTML('<h2>Foo Engine</h2>');

        this.runTask(() => (0, _metal.set)(controller, 'engineName', null));

        this.assertInnerHTML('<!---->');
      });
    }

    ['@test it declares the event dispatcher as a singleton']() {
      this.router.map(function () {
        this.route('engine-event-dispatcher-singleton');
      });

      let controller;
      let component;

      this.add('controller:engine-event-dispatcher-singleton', _controller.default.extend({
        init() {
          this._super(...arguments);
          controller = this;
        }
      }));
      this.addTemplate('engine-event-dispatcher-singleton', '{{mount "foo"}}');

      this.add('engine:foo', _engine.default.extend({
        router: null,
        init() {
          this._super(...arguments);
          this.register('template:application', (0, _helpers.compile)('<h2>Foo Engine: {{tagless-component}}</h2>', {
            moduleName: 'my-app/templates/application.hbs'
          }));
          this.register('component:tagless-component', _helpers.Component.extend({
            tagName: '',
            init() {
              this._super(...arguments);
              component = this;
            }
          }));
          this.register('template:components/tagless-component', (0, _helpers.compile)('Tagless Component', {
            moduleName: 'my-app/templates/components/tagless-component.hbs'
          }));
        }
      }));

      return this.visit('/engine-event-dispatcher-singleton').then(() => {
        this.assertInnerHTML('<h2>Foo Engine: Tagless Component</h2>');

        let controllerOwnerEventDispatcher = (0, _owner.getOwner)(controller).lookup('event_dispatcher:main');
        let taglessComponentOwnerEventDispatcher = (0, _owner.getOwner)(component).lookup('event_dispatcher:main');

        this.assert.strictEqual(controllerOwnerEventDispatcher, taglessComponentOwnerEventDispatcher);
      });
    }
  });

  if (true /* EMBER_ENGINES_MOUNT_PARAMS */) {
      (0, _testCase.moduleFor)('{{mount}} params tests', class extends _testCase.ApplicationTest {
        constructor() {
          super(...arguments);

          this.add('engine:paramEngine', _engine.default.extend({
            router: null,
            init() {
              this._super(...arguments);
              this.register('template:application', (0, _helpers.compile)('<h2>Param Engine: {{model.foo}}</h2>', {
                moduleName: 'my-app/templates/application.hbs'
              }));
            }
          }));
        }

        ['@test it renders with static parameters']() {
          this.router.map(function () {
            this.route('engine-params-static');
          });
          this.addTemplate('engine-params-static', '{{mount "paramEngine" model=(hash foo="bar")}}');

          return this.visit('/engine-params-static').then(() => {
            this.assertInnerHTML('<h2>Param Engine: bar</h2>');
          });
        }

        ['@test it renders with bound parameters']() {
          this.router.map(function () {
            this.route('engine-params-bound');
          });
          let controller;
          this.add('controller:engine-params-bound', _controller.default.extend({
            boundParamValue: null,
            init() {
              this._super();
              controller = this;
            }
          }));
          this.addTemplate('engine-params-bound', '{{mount "paramEngine" model=(hash foo=boundParamValue)}}');

          return this.visit('/engine-params-bound').then(() => {
            this.assertInnerHTML('<h2>Param Engine: </h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', 'bar'));

            this.assertInnerHTML('<h2>Param Engine: bar</h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', undefined));

            this.assertInnerHTML('<h2>Param Engine: </h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', 'bar'));

            this.assertInnerHTML('<h2>Param Engine: bar</h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', 'baz'));

            this.assertInnerHTML('<h2>Param Engine: baz</h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', 'bar'));

            this.assertInnerHTML('<h2>Param Engine: bar</h2>');

            this.runTask(() => (0, _metal.set)(controller, 'boundParamValue', null));

            this.assertInnerHTML('<h2>Param Engine: </h2>');
          });
        }

        ['@test it renders contextual components passed as parameter values']() {
          this.router.map(function () {
            this.route('engine-params-contextual-component');
          });

          this.addComponent('foo-component', {
            template: `foo-component rendered! - {{app-bar-component}}`
          });
          this.addComponent('app-bar-component', {
            ComponentClass: _helpers.Component.extend({ tagName: '' }),
            template: 'rendered app-bar-component from the app'
          });
          this.add('engine:componentParamEngine', _engine.default.extend({
            router: null,
            init() {
              this._super(...arguments);
              this.register('template:application', (0, _helpers.compile)('{{model.foo}}', {
                moduleName: 'my-app/templates/application.hbs'
              }));
            }
          }));
          this.addTemplate('engine-params-contextual-component', '{{mount "componentParamEngine" model=(hash foo=(component "foo-component"))}}');

          return this.visit('/engine-params-contextual-component').then(() => {
            this.assertComponentElement(this.firstChild, {
              content: 'foo-component rendered! - rendered app-bar-component from the app'
            });
          });
        }
      });
    }
});
enifed('@ember/-internals/glimmer/tests/integration/outlet-test', ['@ember/-internals/glimmer/tests/utils/test-case', 'internal-test-helpers', '@ember/-internals/metal'], function (_testCase, _internalTestHelpers, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('outlet view', class extends _testCase.RenderingTest {
    constructor() {
      super(...arguments);

      let CoreOutlet = this.owner.factoryFor('view:-outlet');

      this.component = CoreOutlet.create();
    }

    ['@test should not error when initial rendered template is undefined']() {
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: undefined,
          template: undefined
        },

        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('');
    }

    ['@test should render the outlet when set after DOM insertion']() {
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: undefined,
          template: undefined
        },

        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('');

      this.registerTemplate('application', 'HI{{outlet}}');
      outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: {},
          template: this.owner.lookup('template:application')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HI');

      this.assertStableRerender();

      this.registerTemplate('index', '<p>BYE</p>');
      outletState.outlets.main = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'index',
          controller: {},
          template: this.owner.lookup('template:index')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HIBYE');
    }

    ['@test should render the outlet when set before DOM insertion']() {
      this.registerTemplate('application', 'HI{{outlet}}');
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: {},
          template: this.owner.lookup('template:application')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('HI');

      this.assertStableRerender();

      this.registerTemplate('index', '<p>BYE</p>');
      outletState.outlets.main = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'index',
          controller: {},
          template: this.owner.lookup('template:index')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HIBYE');
    }

    ['@test should support an optional name']() {
      this.registerTemplate('application', '<h1>HI</h1>{{outlet "special"}}');
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: {},
          template: this.owner.lookup('template:application')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('HI');

      this.assertStableRerender();

      this.registerTemplate('special', '<p>BYE</p>');
      outletState.outlets.special = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'special',
          controller: {},
          template: this.owner.lookup('template:special')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HIBYE');
    }

    ['@test does not default outlet name when positional argument is present']() {
      this.registerTemplate('application', '<h1>HI</h1>{{outlet someUndefinedThing}}');
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller: {},
          template: this.owner.lookup('template:application')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('HI');

      this.assertStableRerender();

      this.registerTemplate('special', '<p>BYE</p>');
      outletState.outlets.main = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'special',
          controller: {},
          template: this.owner.lookup('template:special')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HI');
    }

    ['@test should support bound outlet name']() {
      let controller = { outletName: 'foo' };
      this.registerTemplate('application', '<h1>HI</h1>{{outlet outletName}}');
      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'application',
          controller,
          template: this.owner.lookup('template:application')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('HI');

      this.assertStableRerender();

      this.registerTemplate('foo', '<p>FOO</p>');
      outletState.outlets.foo = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'foo',
          controller: {},
          template: this.owner.lookup('template:foo')
        },
        outlets: Object.create(null)
      };

      this.registerTemplate('bar', '<p>BAR</p>');
      outletState.outlets.bar = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'bar',
          controller: {},
          template: this.owner.lookup('template:bar')
        },
        outlets: Object.create(null)
      };

      this.runTask(() => this.component.setOutletState(outletState));

      this.assertText('HIFOO');

      this.runTask(() => (0, _metal.set)(controller, 'outletName', 'bar'));

      this.assertText('HIBAR');
    }

    ['@test outletState can pass through user code (liquid-fire initimate API) ']() {
      this.registerTemplate('outer', 'A{{#-with-dynamic-vars outletState=(identity (-get-dynamic-var "outletState"))}}B{{outlet}}D{{/-with-dynamic-vars}}E');
      this.registerTemplate('inner', 'C');

      // This looks like it doesn't do anything, but its presence
      // guarantees that the outletState gets converted from a reference
      // to a value and then back to a reference. That is what we're
      // testing here.
      this.registerHelper('identity', ([a]) => a);

      let outletState = {
        render: {
          owner: this.owner,
          into: undefined,
          outlet: 'main',
          name: 'outer',
          controller: {},
          template: this.owner.lookup('template:outer')
        },
        outlets: {
          main: {
            render: {
              owner: this.owner,
              into: undefined,
              outlet: 'main',
              name: 'inner',
              controller: {},
              template: this.owner.lookup('template:inner')
            },
            outlets: Object.create(null)
          }
        }
      };

      this.runTask(() => this.component.setOutletState(outletState));

      (0, _internalTestHelpers.runAppend)(this.component);

      this.assertText('ABCDE');

      this.assertStableRerender();
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/refinements-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/metal'], function (_testCase, _abstractTestCase, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('syntax refinements', class extends _testCase.RenderingTest {
    ['@test block params should not be refined']() {
      this.registerHelper('foo', () => 'bar helper');

      this.render(_abstractTestCase.strip`
      {{#with var as |foo|}}
        {{foo}}
      {{/with}}

      ---

      {{#with var as |outlet|}}
        {{outlet}}
      {{/with}}

      ---

      {{#with var as |mount|}}
        {{mount}}
      {{/with}}

      ---

      {{#with var as |component|}}
        {{component}}
      {{/with}}

      ---

      {{#with var as |input|}}
        {{input}}
      {{/with}}

      ---

      {{#with var as |-with-dynamic-vars|}}
        {{-with-dynamic-vars}}
      {{/with}}

      ---

      {{#with var as |-in-element|}}
        {{-in-element}}
      {{/with}}`, { var: 'var' });

      this.assertText('var---var---var---var---var---var---var');

      this.runTask(() => (0, _metal.set)(this.context, 'var', 'RARRR!!!'));

      this.assertText('RARRR!!!---RARRR!!!---RARRR!!!---RARRR!!!---RARRR!!!---RARRR!!!---RARRR!!!');

      this.runTask(() => (0, _metal.set)(this.context, 'var', 'var'));

      this.assertText('var---var---var---var---var---var---var');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/render-settled-test', ['internal-test-helpers', '@ember/-internals/glimmer', 'rsvp', '@ember/runloop'], function (_internalTestHelpers, _glimmer, _rsvp, _runloop) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('renderSettled', class extends _internalTestHelpers.RenderingTestCase {
    ['@test resolves when no rendering is happening'](assert) {
      return (0, _glimmer.renderSettled)().then(() => {
        assert.ok(true, 'resolved even without rendering');
      });
    }

    ['@test resolves renderers exist but no runloops are triggered'](assert) {
      this.render(_internalTestHelpers.strip`{{foo}}`, { foo: 'bar' });

      return (0, _glimmer.renderSettled)().then(() => {
        assert.ok(true, 'resolved even without runloops');
      });
    }

    ['@test does not create extraneous promises'](assert) {
      let first = (0, _glimmer.renderSettled)();
      let second = (0, _glimmer.renderSettled)();

      assert.strictEqual(first, second);

      return (0, _rsvp.all)([first, second]);
    }

    ['@test resolves when rendering has completed (after property update)']() {
      this.render(_internalTestHelpers.strip`{{foo}}`, { foo: 'bar' });

      this.assertText('bar');
      this.component.set('foo', 'baz');
      this.assertText('bar');

      return (0, _glimmer.renderSettled)().then(() => {
        this.assertText('baz');
      });
    }

    ['@test resolves in run loop when renderer has settled'](assert) {
      assert.expect(3);

      this.render(_internalTestHelpers.strip`{{foo}}`, { foo: 'bar' });

      this.assertText('bar');
      let promise;

      return (0, _runloop.run)(() => {
        (0, _runloop.schedule)('actions', null, () => {
          this.component.set('foo', 'set in actions');

          promise = (0, _glimmer.renderSettled)().then(() => {
            this.assertText('set in afterRender');
          });

          (0, _runloop.schedule)('afterRender', null, () => {
            this.component.set('foo', 'set in afterRender');
          });
        });

        // still not updated here
        this.assertText('bar');

        return promise;
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/svg-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_testCase, _metal, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('SVG element tests', class extends _testCase.RenderingTest {
    ['@test unquoted viewBox property is output'](assert) {
      let viewBoxString = '0 0 100 100';

      this.render('<div><svg viewBox={{model.viewBoxString}}></svg></div>', {
        model: {
          viewBoxString
        }
      });

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => this.rerender());

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model.viewBoxString', null));

      assert.equal(this.firstChild.getAttribute('svg'), null);

      this.runTask(() => (0, _metal.set)(this.context, 'model', { viewBoxString }));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);
    }

    ['@test quoted viewBox property is output'](assert) {
      let viewBoxString = '0 0 100 100';

      this.render('<div><svg viewBox="{{model.viewBoxString}}"></svg></div>', {
        model: {
          viewBoxString
        }
      });

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => this.rerender());

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model.viewBoxString', null));

      assert.equal(this.firstChild.getAttribute('svg'), null);

      this.runTask(() => (0, _metal.set)(this.context, 'model', { viewBoxString }));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="${viewBoxString}"></svg>
      </div>
    `);
    }

    ['@test quoted viewBox property is concat']() {
      let viewBoxString = '100 100';

      this.render('<div><svg viewBox="0 0 {{model.viewBoxString}}"></svg></div>', {
        model: {
          viewBoxString
        }
      });

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="0 0 ${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => this.rerender());

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="0 0 ${viewBoxString}"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model.viewBoxString', '200 200'));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="0 0 200 200"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model', { viewBoxString }));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg viewBox="0 0 ${viewBoxString}"></svg>
      </div>
    `);
    }

    ['@test class is output']() {
      this.render("<div><svg class='{{model.color}} tall'></svg></div>", {
        model: {
          color: 'blue'
        }
      });

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg class="blue tall"></svg>
      </div>
    `);

      this.runTask(() => this.rerender());

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg class="blue tall"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model.color', 'yellow'));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg class="yellow tall"></svg>
      </div>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'model', { color: 'blue' }));

      this.assertInnerHTML(_abstractTestCase.strip`
      <div>
        <svg class="blue tall"></svg>
      </div>
    `);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/each-in-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/utils', '@ember/-internals/glimmer/tests/utils/shared-conditional-tests'], function (_metal, _testCase, _runtime, _abstractTestCase, _utils, _sharedConditionalTests) {
  'use strict';

  function EmptyFunction() {}
  function NonEmptyFunction() {}
  NonEmptyFunction.foo = 'bar';
  class EmptyConstructor {}
  class NonEmptyConstructor {}
  NonEmptyConstructor.foo = 'bar';

  class TogglingEachInTest extends _sharedConditionalTests.TogglingSyntaxConditionalsTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#each-in ${cond} as |key|}}${truthy}{{else}}${falsy}{{/each-in}}`;
    }
  }
  class BasicEachInTest extends TogglingEachInTest {}

  class BasicSyntaxTest extends BasicEachInTest {
    get truthyValue() {
      return { 'Not Empty': 1 };
    }

    get falsyValue() {
      return {};
    }
  }

  class EachInProxyTest extends TogglingEachInTest {}

  (0, _abstractTestCase.applyMixins)(BasicEachInTest, new _sharedConditionalTests.TruthyGenerator([{ foo: 1 }, _runtime.Object.create({ 'Not Empty': 1 }), [1], NonEmptyFunction, NonEmptyConstructor]), new _sharedConditionalTests.FalsyGenerator([null, undefined, false, '', 0, [], EmptyFunction, EmptyConstructor, {}, Object.create(null), Object.create({}), Object.create({ 'Not Empty': 1 }), _runtime.Object.create()]));

  (0, _abstractTestCase.applyMixins)(EachInProxyTest, new _sharedConditionalTests.TruthyGenerator([_runtime.ObjectProxy.create({ content: { 'Not empty': 1 } })]), new _sharedConditionalTests.FalsyGenerator([_runtime.ObjectProxy.create(), _runtime.ObjectProxy.create({ content: null }), _runtime.ObjectProxy.create({ content: {} }), _runtime.ObjectProxy.create({ content: Object.create(null) }), _runtime.ObjectProxy.create({ content: Object.create({}) }), _runtime.ObjectProxy.create({ content: Object.create({ 'Not Empty': 1 }) }), _runtime.ObjectProxy.create({ content: _runtime.Object.create() })]));

  // Truthy/Falsy tests
  (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with `ObjectProxy`', class extends EachInProxyTest {
    get truthyValue() {
      return _runtime.ObjectProxy.create({ content: { 'Not Empty': 1 } });
    }

    get falsyValue() {
      return _runtime.ObjectProxy.create({ content: null });
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each-in}}', BasicSyntaxTest);

  // Rendering tests
  class AbstractEachInTest extends _testCase.RenderingTest {
    createHash() /* hash */{
      throw new Error('Not implemented: `createHash`');
    }

    makeHash(obj) {
      let { hash, delegate } = this.createHash(obj);

      this.hash = hash;
      this.delegate = delegate;
      return hash;
    }

    replaceHash(hash) {
      this.runTask(() => (0, _metal.set)(this.context, 'hash', this.createHash(hash).hash));
    }

    clear() {
      return this.runTask(() => (0, _metal.set)(this.context, 'hash', this.createHash({}).hash));
    }

    setProp(key, value) {
      return this.runTask(() => this.delegate.setProp(this.context, key, value));
    }

    updateNestedValue(key, innerKey, value) {
      return this.runTask(() => this.delegate.updateNestedValue(this.context, key, innerKey, value));
    }

    render(template, context = {}) {
      if (this.hash !== undefined) {
        context.hash = this.hash;
      }
      if (this.type !== undefined) {
        context.type = this.type;
      }
      context.secretKey = 'asd';

      return super.render(template, context);
    }
  }

  class EachInTest extends AbstractEachInTest {
    [`@test it repeats the given block for each item in the hash`]() {
      this.makeHash({ Smartphones: 8203, 'JavaScript Frameworks': Infinity });

      this.render(`<ul>{{#each-in hash as |category count|}}<li>{{category}}: {{count}}</li>{{else}}Empty!{{/each-in}}</ul>`);

      this.assertText('Smartphones: 8203JavaScript Frameworks: Infinity');

      this.assertStableRerender();

      if (this.allowsSetProp) {
        // Not al backing data structures allow kvo tracking. Maps and Iterables don't
        this.setProp('Tweets', 100);

        this.assertText('Smartphones: 8203JavaScript Frameworks: InfinityTweets: 100');
      }

      this.clear();

      this.assertText('Empty!');
    }

    [`@test it can render sub-paths of each item`](assert) {
      this.makeHash({
        Smartphones: { reports: { unitsSold: 8203 } },
        'JavaScript Frameworks': { reports: { unitsSold: Infinity } }
      });

      this.render(`<ul>{{#each-in hash as |category data|}}<li>{{category}}: {{data.reports.unitsSold}}</li>{{else}}Empty!{{/each-in}}</ul>`);

      this.assertText('Smartphones: 8203JavaScript Frameworks: Infinity');

      this.assertStableRerender();

      if (this.allowsSetProp) {
        this.setProp('Tweets', { reports: { unitsSold: 100 } });

        this.assertText('Smartphones: 8203JavaScript Frameworks: InfinityTweets: 100');
      }

      this.runTask(() => this.updateNestedValue('Smartphones', 'reports.unitsSold', 8204));

      assert.ok(this.textValue().indexOf('Smartphones: 8204') > -1);

      this.clear();

      this.assertText('Empty!');
    }

    [`@test it can render duplicate items`]() {
      this.makeHash({
        Smartphones: 8203,
        Tablets: 8203,
        'JavaScript Frameworks': Infinity,
        Bugs: Infinity
      });

      this.render(`<ul>{{#each-in hash key='@identity' as |category count|}}<li>{{category}}: {{count}}</li>{{/each-in}}</ul>`);

      this.assertText('Smartphones: 8203Tablets: 8203JavaScript Frameworks: InfinityBugs: Infinity');

      this.assertStableRerender();

      if (this.allowsSetProp) {
        this.setProp('Smartphones', 100);
        this.setProp('Tweets', 443115);
        this.assertText('Smartphones: 100Tablets: 8203JavaScript Frameworks: InfinityBugs: InfinityTweets: 443115');
      }

      this.clear();

      this.assertText('');
    }

    [`@test it repeats the given block when the hash is dynamic`]() {
      let { hash: categories } = this.createHash({
        Smartphones: 8203,
        'JavaScript Frameworks': Infinity
      });
      let { hash: otherCategories } = this.createHash({
        Emberinios: 533462,
        Tweets: 7323
      });
      let context = {
        hashes: {
          categories,
          otherCategories,
          type: 'categories'
        }
      };
      this.render(`<ul>{{#each-in (get hashes hashes.type) as |category count|}}<li>{{category}}: {{count}}</li>{{else}}Empty!{{/each-in}}</ul>`, context);

      this.assertText('Smartphones: 8203JavaScript Frameworks: Infinity');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(context, 'hashes.type', 'otherCategories'));

      this.assertText('Emberinios: 533462Tweets: 7323');

      this.runTask(() => (0, _metal.set)(context, 'hashes.type', 'categories'));

      this.assertText('Smartphones: 8203JavaScript Frameworks: Infinity');

      this.runTask(() => (0, _metal.set)(context, 'hashes.type', 'nonExistent'));

      this.clear();

      this.assertText('Empty!');

      this.runTask(() => (0, _metal.set)(context, 'hashes.type', 'categories'));

      this.assertText('Smartphones: 8203JavaScript Frameworks: Infinity');
    }

    ['@test keying off of `undefined` does not render']() {
      this.makeHash({});

      this.render(`{{#each-in hash as |key value|}}{{key}}: {{value.baz}}{{else}}Empty!{{/each-in}}`);

      this.assertText('Empty!');

      this.assertStableRerender();

      this.replaceHash({ bar: { baz: 'Here!' } });

      this.assertText('bar: Here!');

      this.clear();

      this.assertText('Empty!');
    }

    [`@test it can render items with a key of empty string`]() {
      this.makeHash({ '': 'empty-string', a: 'a' });

      this.render(`<ul>{{#each-in hash as |key value|}}<li>{{key}}: {{value}}</li>{{else}}Empty!{{/each-in}}</ul>`);

      this.assertText(': empty-stringa: a');

      this.assertStableRerender();

      this.clear();

      this.assertText('Empty!');
    }
  }

  (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with POJOs', class extends EachInTest {
    constructor() {
      super(...arguments);
      this.allowsSetProp = true;
    }

    createHash(pojo) {
      return {
        hash: pojo,
        delegate: {
          setProp(context, key, value) {
            (0, _metal.set)(context.hash, key, value);
          },
          updateNestedValue(context, key, innerKey, value) {
            let target = context.hash[key];
            (0, _metal.set)(target, innerKey, value);
          }
        }
      };
    }

    [`@test it only iterates over an object's own properties`]() {
      let protoCategories = {
        Smartphones: 8203,
        'JavaScript Frameworks': Infinity
      };

      let categories = Object.create(protoCategories);
      categories['Televisions'] = 183;
      categories['Alarm Clocks'] = 999;

      this.render(`<ul>{{#each-in categories as |category count|}}<li>{{category}}: {{count}}</li>{{else}}Empty!{{/each-in}}</ul>`, { categories });

      this.assertText('Televisions: 183Alarm Clocks: 999');

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(protoCategories, 'Robots', 666);
        (0, _metal.set)(categories, 'Tweets', 443115);
      });

      this.assertText('Televisions: 183Alarm Clocks: 999Tweets: 443115');

      categories = Object.create(protoCategories);
      categories['Televisions'] = 183;
      categories['Alarm Clocks'] = 999;
    }

    [`@test it does not observe direct property mutations (not going through set) on the object`]() {
      this.render(_abstractTestCase.strip`
      <ul>
        {{#each-in categories as |category count|}}
          <li>{{category}}: {{count}}</li>
        {{/each-in}}
      </ul>
    `, {
        categories: {
          Smartphones: 8203,
          'JavaScript Frameworks': Infinity
        }
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 8203</li>
        <li>JavaScript Frameworks: Infinity</li>
      </ul>
    `);

      this.assertStableRerender();

      this.runTask(() => {
        let categories = (0, _metal.get)(this.context, 'categories');
        delete categories.Smartphones;
      });

      this.assertInvariants();

      this.runTask(() => {
        let categories = (0, _metal.get)(this.context, 'categories');
        categories['Emberinios'] = 123456;
      });

      this.assertInvariants();

      this.runTask(() => {
        (0, _metal.set)(this.context, 'categories', {
          Emberinios: 123456
        });
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Emberinios: 123456</li>
      </ul>
    `);

      this.runTask(() => {
        (0, _metal.set)(this.context, 'categories', {
          Smartphones: 8203,
          'JavaScript Frameworks': Infinity
        });
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 8203</li>
        <li>JavaScript Frameworks: Infinity</li>
      </ul>
    `);
    }

    ['@test it skips holes in sparse arrays']() {
      let arr = [];
      arr[5] = 'foo';
      arr[6] = 'bar';

      this.render(_abstractTestCase.strip`
      {{#each-in arr as |key value|}}
        [{{key}}:{{value}}]
      {{/each-in}}`, { arr });

      this.assertText('[5:foo][6:bar]');

      this.assertStableRerender();
    }

    ['@test it iterate over array with `in` instead of walking over elements']() {
      let arr = [1, 2, 3];
      arr.foo = 'bar';

      this.render(_abstractTestCase.strip`
      {{#each-in arr as |key value|}}
        [{{key}}:{{value}}]
      {{/each-in}}`, { arr });

      this.assertText('[0:1][1:2][2:3][foo:bar]');

      this.runTask(() => this.rerender());

      this.assertText('[0:1][1:2][2:3][foo:bar]');

      this.runTask(() => {
        (0, _metal.set)(arr, 'zomg', 'lol');
      });

      this.assertText('[0:1][1:2][2:3][foo:bar][zomg:lol]');

      arr = [1, 2, 3];
      arr.foo = 'bar';

      this.runTask(() => (0, _metal.set)(this.context, 'arr', arr));

      this.assertText('[0:1][1:2][2:3][foo:bar]');
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with EmberObjects', class extends EachInTest {
    constructor() {
      super(...arguments);
      this.allowsSetProp = true;
    }
    createHash(pojo) {
      let hash = _runtime.Object.create(pojo);
      return {
        hash,
        delegate: {
          setProp(context, key, value) {
            (0, _metal.set)(context, `hash.${key}`, value);
          },
          updateNestedValue(context, key, innerKey, value) {
            let target = (0, _metal.get)(context.hash, key);
            (0, _metal.set)(target, innerKey, value);
          }
        }
      };
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with object proxies', class extends EachInTest {
    constructor() {
      super(...arguments);
      this.allowsSetProp = true;
    }
    createHash(pojo) {
      let hash = _runtime.ObjectProxy.create({ content: pojo });
      return {
        hash,
        delegate: {
          setProp(context, key, value) {
            (0, _metal.set)(context, `hash.${key}`, value);
          },
          updateNestedValue(context, key, innerKey, value) {
            let target = (0, _metal.get)(context.hash, key);
            (0, _metal.set)(target, innerKey, value);
          }
        }
      };
    }

    ['@test it iterates over the content, not the proxy']() {
      let content = {
        Smartphones: 8203,
        'JavaScript Frameworks': Infinity
      };

      let proxy = _runtime.ObjectProxy.create({
        content,
        foo: 'bar'
      });

      this.render(_abstractTestCase.strip`
      <ul>
        {{#each-in categories as |category count|}}
          <li>{{category}}: {{count}}</li>
        {{/each-in}}
      </ul>
    `, { categories: proxy });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 8203</li>
        <li>JavaScript Frameworks: Infinity</li>
      </ul>
    `);

      this.assertStableRerender();

      this.runTask(() => {
        (0, _metal.set)(proxy, 'content.Smartphones', 100);
        (0, _metal.set)(proxy, 'content.Tweets', 443115);
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 100</li>
        <li>JavaScript Frameworks: Infinity</li>
        <li>Tweets: 443115</li>
      </ul>
    `);

      this.runTask(() => {
        (0, _metal.set)(proxy, 'content', {
          Smartphones: 100,
          Tablets: 20
        });
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 100</li>
        <li>Tablets: 20</li>
      </ul>
    `);

      this.runTask(() => (0, _metal.set)(this.context, 'categories', _runtime.ObjectProxy.create({
        content: {
          Smartphones: 8203,
          'JavaScript Frameworks': Infinity
        }
      })));

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>Smartphones: 8203</li>
        <li>JavaScript Frameworks: Infinity</li>
      </ul>
    `);
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with ES6 Maps', class extends EachInTest {
    createHash(pojo) {
      let map = new Map();
      Object.keys(pojo).forEach(key => {
        map.set(key, pojo[key]);
      });
      return {
        hash: map,
        delegate: {
          updateNestedValue(context, key, innerKey, value) {
            let target = context.hash.get(key);
            (0, _metal.set)(target, innerKey, value);
          }
        }
      };
    }

    [`@test it supports having objects as keys on ES6 Maps`]() {
      let map = new Map();
      map.set({ name: 'one' }, 'foo');
      map.set({ name: 'two' }, 'bar');

      this.render(_abstractTestCase.strip`
      <ul>
        {{#each-in map key="@identity" as |key value|}}
          <li>{{key.name}}: {{value}}</li>
        {{/each-in}}
      </ul>`, { map });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>one: foo</li>
        <li>two: bar</li>
      </ul>
    `);

      this.assertStableRerender();

      this.runTask(() => {
        let map = new Map();
        map.set({ name: 'three' }, 'qux');
        (0, _metal.set)(this.context, 'map', map);
      });

      this.assertHTML(_abstractTestCase.strip`
      <ul>
        <li>three: qux</li>
      </ul>
    `);
    }
  });

  if (_utils.HAS_NATIVE_SYMBOL) {
    (0, _testCase.moduleFor)('Syntax test: {{#each-in}} with custom iterables', class extends EachInTest {
      createHash(pojo) {
        let ary = Object.keys(pojo).reduce((accum, key) => {
          return accum.concat([[key, pojo[key]]]);
        }, []);
        let iterable = {
          [Symbol.iterator]: () => makeIterator(ary)
        };
        return {
          hash: iterable,
          delegate: {
            updateNestedValue(context, key, innerKey, value) {
              let ary = Array.from(context.hash);
              let target = ary.find(([k]) => k === key)[1];
              (0, _metal.set)(target, innerKey, value);
            }
          }
        };
      }
    });
  }

  // Utils
  function makeIterator(ary) {
    var index = 0;

    return {
      next() {
        return index < ary.length ? { value: ary[index++], done: false } : { done: true };
      }
    };
  }
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/each-test', ['@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/utils', '@ember/-internals/glimmer/tests/utils/shared-conditional-tests'], function (_metal, _abstractTestCase, _testCase, _runtime, _helpers, _utils, _sharedConditionalTests) {
  'use strict';

  class ArrayDelegate {
    constructor(content, target) {
      this._array = content;
      this._target = target || this;
    }

    // The following methods are APIs used by the tests

    toArray() {
      return this._array.slice();
    }

    objectAt(idx) {
      return this._array[idx];
    }

    clear() {
      this._array.length = 0;
      this.arrayContentDidChange();
    }

    replace(idx, del, ins) {
      this._array.splice(idx, del, ...ins);
      this.arrayContentDidChange();
    }

    unshiftObject(obj) {
      this._array.unshift(obj);
      this.arrayContentDidChange();
    }

    unshiftObjects(arr) {
      this._array.unshift(...arr);
      this.arrayContentDidChange();
    }

    pushObject(obj) {
      this._array.push(obj);
      this.arrayContentDidChange();
    }

    pushObjects(arr) {
      this._array.push(...arr);
      this.arrayContentDidChange();
    }

    shiftObject() {
      let obj = this._array.shift();
      this.arrayContentDidChange();
      return obj;
    }

    popObject() {
      let obj = this._array.pop();
      this.arrayContentDidChange();
      return obj;
    }

    insertAt(idx, obj) {
      this._array.splice(idx, 0, obj);
      this.arrayContentDidChange();
    }

    removeAt(idx, len = 1) {
      this._array.splice(idx, len);
      this.arrayContentDidChange();
    }

    arrayContentDidChange() {
      (0, _metal.notifyPropertyChange)(this._target, '[]');
      (0, _metal.notifyPropertyChange)(this._target, 'length');
    }

    toString() {
      return `#<${this.constructor.name || 'UnknownArrayDelegate'}>`;
    }

    toJSON() {
      return this.toString();
    }
  }

  const makeSet = (() => {
    // IE11 does not support `new Set(items);`
    let set = new Set([1, 2, 3]);

    if (set.size === 3) {
      return items => new Set(items);
    } else {
      return items => {
        let s = new Set();
        items.forEach(value => s.add(value));
        return s;
      };
    }
  })();

  class SetDelegate extends ArrayDelegate {
    constructor(set) {
      let array = [];
      set.forEach(value => array.push(value));
      super(array, set);
      this._set = set;
    }

    arrayContentDidChange() {
      this._set.clear();
      this._array.forEach(value => this._set.add(value));
      super.arrayContentDidChange();
    }
  }

  class ForEachable extends ArrayDelegate {
    get length() {
      return this._array.length;
    }

    forEach(callback) {
      this._array.forEach(callback);
    }
  }

  let ArrayIterable;

  if (_utils.HAS_NATIVE_SYMBOL) {
    ArrayIterable = class extends ArrayDelegate {
      [Symbol.iterator]() {
        return this._array[Symbol.iterator]();
      }
    };
  }

  class TogglingEachTest extends _sharedConditionalTests.TogglingSyntaxConditionalsTest {
    get truthyValue() {
      return ['non-empty'];
    }
    get falsyValue() {
      return [];
    }
  }

  class BasicEachTest extends TogglingEachTest {}

  const TRUTHY_CASES = [['hello'], (0, _runtime.A)(['hello']), makeSet(['hello']), new ForEachable(['hello']), _runtime.ArrayProxy.create({ content: ['hello'] }), _runtime.ArrayProxy.create({ content: (0, _runtime.A)(['hello']) })];

  const FALSY_CASES = [null, undefined, false, '', 0, [], (0, _runtime.A)([]), makeSet([]), new ForEachable([]), _runtime.ArrayProxy.create({ content: [] }), _runtime.ArrayProxy.create({ content: (0, _runtime.A)([]) })];

  if (_utils.HAS_NATIVE_SYMBOL) {
    TRUTHY_CASES.push(new ArrayIterable(['hello']));
    FALSY_CASES.push(new ArrayIterable([]));
  }

  (0, _abstractTestCase.applyMixins)(BasicEachTest, new _sharedConditionalTests.TruthyGenerator(TRUTHY_CASES), new _sharedConditionalTests.FalsyGenerator(FALSY_CASES), _sharedConditionalTests.ArrayTestCases);

  (0, _testCase.moduleFor)('Syntax test: toggling {{#each}}', class extends BasicEachTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#each ${cond}}}${truthy}{{else}}${falsy}{{/each}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: toggling {{#each as}}', class extends BasicEachTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#each ${cond} as |test|}}${truthy}{{else}}${falsy}{{/each}}`;
    }
  });

  class EachEdgeCasesTest extends TogglingEachTest {}

  (0, _abstractTestCase.applyMixins)(EachEdgeCasesTest, new _sharedConditionalTests.FalsyGenerator([true, 'hello', 1, Object, function () {}, {}, { foo: 'bar' }, Object.create(null), Object.create({}), Object.create({ foo: 'bar' })]));

  (0, _testCase.moduleFor)('Syntax test: toggling {{#each}}', class extends EachEdgeCasesTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#each ${cond}}}${truthy}{{else}}${falsy}{{/each}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: toggling {{#each as}}', class extends EachEdgeCasesTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#each ${cond} as |test|}}${truthy}{{else}}${falsy}{{/each}}`;
    }
  });

  class AbstractEachTest extends _testCase.RenderingTest {
    /* abstract */
    createList() /* items */{
      throw new Error('Not implemented: `createList`');
    }

    makeList(items) {
      let { list, delegate } = this.createList(items);

      this.list = list;
      this.delegate = delegate;
    }

    replaceList(list) {
      this.runTask(() => (0, _metal.set)(this.context, 'list', this.createList(list).list));
    }

    forEach(callback) {
      return this.delegate.toArray().forEach(callback);
    }

    objectAt(idx) {
      return this.delegate.objectAt(idx);
    }

    clear() {
      return this.delegate.clear();
    }

    replace(idx, del, ins) {
      return this.delegate.replace(idx, del, ins);
    }

    unshiftObject(obj) {
      return this.delegate.unshiftObject(obj);
    }

    unshiftObjects(arr) {
      return this.delegate.unshiftObjects(arr);
    }

    pushObject(obj) {
      return this.delegate.pushObject(obj);
    }

    pushObjects(arr) {
      return this.delegate.pushObjects(arr);
    }

    shiftObject() {
      return this.delegate.shiftObject();
    }

    popObject() {
      return this.delegate.popObject();
    }

    insertAt(idx, obj) {
      return this.delegate.insertAt(idx, obj);
    }

    removeAt(idx, len) {
      return this.delegate.removeAt(idx, len);
    }

    render(template, context = {}) {
      if (this.list !== undefined) {
        context.list = this.list;
      }

      return super.render(template, context);
    }
  }

  class EachTest extends AbstractEachTest {
    /* single each */

    ['@test it repeats the given block for each item in the array']() {
      this.makeList([{ text: 'hello' }]);

      this.render(`{{#each list as |item|}}{{item.text}}{{else}}Empty{{/each}}`);

      this.assertText('hello');

      this.runTask(() => this.rerender());

      this.assertText('hello');

      this.runTask(() => (0, _metal.set)(this.objectAt(0), 'text', 'Hello'));

      this.assertText('Hello');

      this.runTask(() => {
        this.pushObject({ text: ' ' });
        this.pushObject({ text: 'World' });
      });

      this.assertText('Hello World');

      this.runTask(() => {
        this.pushObject({ text: 'Earth' });
        this.removeAt(1);
        this.insertAt(1, { text: 'Globe' });
      });

      this.assertText('HelloGlobeWorldEarth');

      this.runTask(() => {
        this.pushObject({ text: 'Planet' });
        this.removeAt(1);
        this.insertAt(1, { text: ' ' });
        this.pushObject({ text: ' ' });
        this.pushObject({ text: 'Earth' });
        this.removeAt(3);
      });

      this.assertText('Hello WorldPlanet Earth');

      this.runTask(() => {
        this.pushObject({ text: 'Globe' });
        this.removeAt(1);
        this.insertAt(1, { text: ' ' });
        this.pushObject({ text: ' ' });
        this.pushObject({ text: 'World' });
        this.removeAt(2);
      });

      this.assertText('Hello Planet EarthGlobe World');

      this.runTask(() => this.replace(2, 4, [{ text: 'my' }]));

      this.assertText('Hello my World');

      this.runTask(() => this.clear());

      this.assertText('Empty');

      this.replaceList([{ text: 'hello' }]);

      this.assertText('hello');
    }

    ['@test it receives the index as the second parameter']() {
      this.makeList([{ text: 'hello' }, { text: 'world' }]);

      this.render(`{{#each list as |item index|}}[{{index}}. {{item.text}}]{{/each}}`);

      this.assertText('[0. hello][1. world]');

      this.assertStableRerender();

      this.runTask(() => this.insertAt(1, { text: 'my' }));

      this.assertText('[0. hello][1. my][2. world]');

      this.replaceList([{ text: 'hello' }, { text: 'world' }]);

      this.assertText('[0. hello][1. world]');
    }

    ['@test it accepts a string key']() {
      this.makeList([{ text: 'hello' }, { text: 'world' }]);

      this.render(`{{#each list key='text' as |item|}}{{item.text}}{{/each}}`);

      this.assertText('helloworld');

      this.assertStableRerender();

      this.runTask(() => this.pushObject({ text: 'again' }));

      this.assertText('helloworldagain');

      this.replaceList([{ text: 'hello' }, { text: 'world' }]);

      this.assertText('helloworld');
    }

    ['@test it accepts a numeric key']() {
      this.makeList([{ id: 1 }, { id: 2 }]);

      this.render(`{{#each list key='id' as |item|}}{{item.id}}{{/each}}`);

      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => this.pushObject({ id: 3 }));

      this.assertText('123');

      this.replaceList([{ id: 1 }, { id: 2 }]);

      this.assertText('12');
    }

    ['@test it can specify @index as the key']() {
      this.makeList([{ id: 1 }, { id: 2 }]);

      this.render(`{{#each list key='@index' as |item|}}{{item.id}}{{/each}}`);

      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => this.pushObject({ id: 3 }));

      this.assertText('123');

      this.replaceList([{ id: 1 }, { id: 2 }]);

      this.assertText('12');
    }

    ['@test it can specify @identity as the key for arrays of primitives']() {
      this.makeList([1, 2]);

      this.render(`{{#each list key='@identity' as |item|}}{{item}}{{/each}}`);

      this.assertText('12');

      this.assertStableRerender();

      this.runTask(() => this.pushObject(3));

      this.assertText('123');

      this.replaceList([1, 2]);

      this.assertText('12');
    }

    ['@test it can specify @identity as the key for mixed arrays of objects and primitives']() {
      this.makeList([1, { id: 2 }, 3]);

      this.render(`{{#each list key='@identity' as |item|}}{{if item.id item.id item}}{{/each}}`);

      this.assertText('123');

      this.assertStableRerender();

      this.runTask(() => this.insertAt(2, { id: 4 }));

      this.assertText('1243');

      this.replaceList([1, { id: 2 }, 3]);

      this.assertText('123');
    }

    ['@test it can render duplicate primitive items']() {
      this.makeList(['a', 'a', 'a']);

      this.render(`{{#each list as |item|}}{{item}}{{/each}}`);

      this.assertText('aaa');

      this.assertStableRerender();

      this.runTask(() => this.pushObject('a'));

      this.assertText('aaaa');

      this.runTask(() => this.pushObject('a'));

      this.assertText('aaaaa');

      this.replaceList(['a', 'a', 'a']);

      this.assertText('aaa');
    }

    [`@test updating and setting within #each`]() {
      this.makeList([{ value: 1 }, { value: 2 }, { value: 3 }]);

      let FooBarComponent = _helpers.Component.extend({
        init() {
          this._super(...arguments);
          this.isEven = true;
          this.tagName = 'li';
        },

        _isEven() {
          this.set('isEven', this.get('item.value') % 2 === 0);
        },

        didUpdate() {
          this._isEven();
        }
      });

      this.registerComponent('foo-bar', {
        ComponentClass: FooBarComponent,
        template: '{{#if isEven}}{{item.value}}{{/if}}'
      });

      this.render(_abstractTestCase.strip`
      {{#each list as |item|}}
        <li>Prev</li>
        {{foo-bar item=item}}
        <li>Next</li>
      {{/each}}
    `);

      this.assertText('Prev1NextPrev2NextPrev3Next');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.objectAt(0), 'value', 3));

      this.assertText('PrevNextPrev2NextPrev3Next');

      this.replaceList([{ value: 1 }, { value: 2 }, { value: 3 }]);

      this.assertText('Prev1NextPrev2NextPrev3Next');
    }

    ['@test it can render duplicate objects']() {
      let duplicateItem = { text: 'foo' };

      this.makeList([duplicateItem, duplicateItem, { text: 'bar' }, { text: 'baz' }]);

      this.render(`{{#each list as |item|}}{{item.text}}{{/each}}`);

      this.assertText('foofoobarbaz');

      this.assertStableRerender();

      this.runTask(() => this.pushObject(duplicateItem));

      this.assertText('foofoobarbazfoo');

      this.runTask(() => this.pushObject(duplicateItem));

      this.assertText('foofoobarbazfoofoo');

      this.replaceList([duplicateItem, duplicateItem, { text: 'bar' }, { text: 'baz' }]);

      this.assertText('foofoobarbaz');
    }

    [`@test it maintains DOM stability when condition changes between objects with the same keys`]() {
      this.makeList([{ text: 'Hello' }, { text: ' ' }, { text: 'world' }]);

      this.render(`{{#each list key="text" as |item|}}{{item.text}}{{/each}}`);

      this.assertText('Hello world');

      this.takeSnapshot();

      this.runTask(() => {
        this.popObject();
        this.popObject();
        this.pushObject({ text: ' ' });
        this.pushObject({ text: 'world' });
      });

      this.assertText('Hello world');

      this.assertInvariants();

      this.replaceList([{ text: 'Hello' }, { text: ' ' }, { text: 'world' }]);

      this.assertText('Hello world');

      this.assertInvariants();
    }

    [`@test it maintains DOM stability for stable keys when list is updated`]() {
      this.makeList([{ text: 'Hello' }, { text: ' ' }, { text: 'world' }]);

      this.render(`{{#each list key="text" as |item|}}{{item.text}}{{/each}}`);

      this.assertText('Hello world');

      this.assertStableRerender();

      let oldSnapshot = this.takeSnapshot();

      this.runTask(() => {
        this.unshiftObject({ text: ', ' });
        this.unshiftObject({ text: 'Hi' });
        this.pushObject({ text: '!' });
        this.pushObject({ text: 'earth' });
      });

      this.assertText('Hi, Hello world!earth');

      this.assertPartialInvariants(2, 5);

      this.replaceList([{ text: 'Hello' }, { text: ' ' }, { text: 'world' }]);

      this.assertText('Hello world');

      this.assertInvariants(oldSnapshot, this.takeSnapshot());
    }

    [`@test it renders all items with duplicate key values`]() {
      this.makeList([{ text: 'Hello' }, { text: 'Hello' }, { text: 'Hello' }]);

      this.render(`{{#each list key="text" as |item|}}{{item.text}}{{/each}}`);

      this.assertText('HelloHelloHello');

      this.runTask(() => {
        this.forEach(hash => (0, _metal.set)(hash, 'text', 'Goodbye'));
      });

      this.assertText('GoodbyeGoodbyeGoodbye');

      this.replaceList([{ text: 'Hello' }, { text: 'Hello' }, { text: 'Hello' }]);

      this.assertText('HelloHelloHello');
    }

    ['@test context is not changed to the inner scope inside an {{#each as}} block']() {
      this.makeList([{ name: 'Chad' }, { name: 'Zack' }, { name: 'Asa' }]);

      this.render(`{{name}}-{{#each list as |person|}}{{name}}{{/each}}-{{name}}`, {
        name: 'Joel'
      });

      this.assertText('Joel-JoelJoelJoel-Joel');

      this.assertStableRerender();

      this.runTask(() => this.shiftObject());

      this.assertText('Joel-JoelJoel-Joel');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Godfrey'));

      this.assertText('Godfrey-GodfreyGodfrey-Godfrey');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Joel'));
      this.replaceList([{ name: 'Chad' }, { name: 'Zack' }, { name: 'Asa' }]);

      this.assertText('Joel-JoelJoelJoel-Joel');
    }

    ['@test can access the item and the original scope']() {
      this.makeList([{ name: 'Tom Dale' }, { name: 'Yehuda Katz' }, { name: 'Godfrey Chan' }]);

      this.render(`{{#each list key="name" as |person|}}[{{title}}: {{person.name}}]{{/each}}`, {
        title: 'Señor Engineer'
      });

      this.assertText('[Señor Engineer: Tom Dale][Señor Engineer: Yehuda Katz][Señor Engineer: Godfrey Chan]');

      this.runTask(() => this.rerender());

      this.assertText('[Señor Engineer: Tom Dale][Señor Engineer: Yehuda Katz][Señor Engineer: Godfrey Chan]');

      this.runTask(() => {
        (0, _metal.set)(this.objectAt(1), 'name', 'Stefan Penner');
        this.removeAt(0);
        this.pushObject({ name: 'Tom Dale' });
        this.insertAt(1, { name: 'Chad Hietala' });
        (0, _metal.set)(this.context, 'title', 'Principal Engineer');
      });

      this.assertText('[Principal Engineer: Stefan Penner][Principal Engineer: Chad Hietala][Principal Engineer: Godfrey Chan][Principal Engineer: Tom Dale]');

      this.runTask(() => (0, _metal.set)(this.context, 'title', 'Señor Engineer'));
      this.replaceList([{ name: 'Tom Dale' }, { name: 'Yehuda Katz' }, { name: 'Godfrey Chan' }]);

      this.assertText('[Señor Engineer: Tom Dale][Señor Engineer: Yehuda Katz][Señor Engineer: Godfrey Chan]');
    }

    ['@test the scoped variable is not available outside the {{#each}} block.']() {
      this.makeList(['Yehuda']);

      this.render(`{{name}}-{{#each list as |name|}}{{name}}{{/each}}-{{name}}`, {
        name: 'Stef'
      });

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => this.rerender());

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => this.pushObjects([' ', 'Katz']));

      this.assertText('Stef-Yehuda Katz-Stef');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Tom'));

      this.assertText('Tom-Yehuda Katz-Tom');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Stef'));
      this.replaceList(['Yehuda']);

      this.assertText('Stef-Yehuda-Stef');
    }

    ['@test inverse template is displayed with context']() {
      this.makeList([]);

      this.render(`{{#each list as |thing|}}Has Thing{{else}}No Thing {{otherThing}}{{/each}}`, {
        otherThing: 'bar'
      });

      this.assertText('No Thing bar');

      this.runTask(() => this.rerender());

      this.assertText('No Thing bar');

      this.runTask(() => (0, _metal.set)(this.context, 'otherThing', 'biz'));

      this.assertText('No Thing biz');

      this.runTask(() => this.pushObject('non-empty'));

      this.assertText('Has Thing');

      this.runTask(() => (0, _metal.set)(this.context, 'otherThing', 'baz'));

      this.assertText('Has Thing');

      this.runTask(() => (0, _metal.set)(this.context, 'otherThing', 'bar'));
      this.replaceList([]);

      this.assertText('No Thing bar');
    }

    ['@test content that are not initially present updates correctly GH#13983']() {
      // The root cause of this bug is that Glimmer did not call `didInitializeChildren`
      // on the inserted `TryOpcode`, causing that `TryOpcode` to have an uninitialized
      // tag. Currently the only way to observe this the "JUMP-IF-NOT-MODIFIED", i.e. by
      // wrapping it in an component.

      this.registerComponent('x-wrapper', { template: '{{yield}}' });

      this.makeList([]);

      this.render(`{{#x-wrapper}}{{#each list as |obj|}}[{{obj.text}}]{{/each}}{{/x-wrapper}}`);

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => this.pushObject({ text: 'foo' }));

      this.assertText('[foo]');

      this.runTask(() => (0, _metal.set)(this.objectAt(0), 'text', 'FOO'));

      this.assertText('[FOO]');

      this.runTask(() => this.pushObject({ text: 'bar' }));

      this.assertText('[FOO][bar]');

      this.runTask(() => (0, _metal.set)(this.objectAt(1), 'text', 'BAR'));

      this.assertText('[FOO][BAR]');

      this.runTask(() => (0, _metal.set)(this.objectAt(1), 'text', 'baz'));

      this.assertText('[FOO][baz]');

      this.runTask(() => this.replace(1, 1, [{ text: 'BAZ' }]));

      this.assertText('[FOO][BAZ]');

      this.replaceList([]);

      this.assertText('');
    }

    ['@test empty trusted content clears properly [GH#16314]']() {
      this.makeList(['hello']);

      this.render(`before {{#each list as |value|}}{{{value}}}{{/each}} after`);

      this.assertText('before hello after');

      this.assertStableRerender();

      this.runTask(() => this.pushObjects([null, ' world']));

      this.assertText('before hello world after');

      this.runTask(() => this.replace(1, 2, [undefined, ' world!']));

      this.assertText('before hello world! after');

      this.runTask(() => this.replace(1, 2, [(0, _helpers.htmlSafe)(''), ' world!!']));

      this.assertText('before hello world!! after');

      this.replaceList(['hello']);

      this.assertText('before hello after');
    }

    /* multi each */

    ['@test re-using the same variable with different {{#each}} blocks does not override each other']() {
      let admins = this.createList([{ name: 'Tom Dale' }]);
      let users = this.createList([{ name: 'Yehuda Katz' }]);

      this.render(`Admin: {{#each admins key="name" as |person|}}[{{person.name}}]{{/each}} User: {{#each users key="name" as |person|}}[{{person.name}}]{{/each}}`, {
        admins: admins.list,
        users: users.list
      });

      this.assertText('Admin: [Tom Dale] User: [Yehuda Katz]');

      this.runTask(() => this.rerender());

      this.assertText('Admin: [Tom Dale] User: [Yehuda Katz]');

      this.runTask(() => {
        admins.delegate.pushObject({ name: 'Godfrey Chan' });
        (0, _metal.set)(users.delegate.objectAt(0), 'name', 'Stefan Penner');
      });

      this.assertText('Admin: [Tom Dale][Godfrey Chan] User: [Stefan Penner]');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'admins', this.createList([{ name: 'Tom Dale' }]).list);
        (0, _metal.set)(this.context, 'users', this.createList([{ name: 'Yehuda Katz' }]).list);
      });

      this.assertText('Admin: [Tom Dale] User: [Yehuda Katz]');
    }

    [`@test an outer {{#each}}'s scoped variable does not clobber an inner {{#each}}'s property if they share the same name - Issue #1315`]() {
      let content = this.createList(['X', 'Y']);
      let options = this.createList([{ label: 'One', value: 1 }, { label: 'Two', value: 2 }]);

      this.render(_abstractTestCase.strip`
      {{#each content as |value|}}
        {{value}}-
        {{#each options as |option|}}
          {{option.value}}:{{option.label}}
        {{/each}}
      {{/each}}
      `, {
        content: content.list,
        options: options.list
      });

      this.assertText('X-1:One2:TwoY-1:One2:Two');

      this.assertStableRerender();

      this.runTask(() => {
        content.delegate.pushObject('Z');
        (0, _metal.set)(options.delegate.objectAt(0), 'value', 0);
      });

      this.assertText('X-0:One2:TwoY-0:One2:TwoZ-0:One2:Two');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'content', this.createList(['X', 'Y']).list);
        (0, _metal.set)(this.context, 'options', this.createList([{ label: 'One', value: 1 }, { label: 'Two', value: 2 }]).list);
      });

      this.assertText('X-1:One2:TwoY-1:One2:Two');
    }

    ['@test the scoped variable is not available outside the {{#each}} block']() {
      let first = this.createList(['Limbo']);
      let fifth = this.createList(['Wrath']);
      let ninth = this.createList(['Treachery']);

      this.render(`{{ring}}-{{#each first as |ring|}}{{ring}}-{{#each fifth as |ring|}}{{ring}}-{{#each ninth as |ring|}}{{ring}}-{{/each}}{{ring}}-{{/each}}{{ring}}-{{/each}}{{ring}}`, {
        ring: 'Greed',
        first: first.list,
        fifth: fifth.list,
        ninth: ninth.list
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => this.rerender());

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'O');
        fifth.delegate.insertAt(0, 'D');
      });

      this.assertText('O-Limbo-D-Treachery-D-Wrath-Treachery-Wrath-Limbo-O');

      this.runTask(() => {
        first.delegate.pushObject('I');
        ninth.delegate.replace(0, 1, ['K']);
      });

      this.assertText('O-Limbo-D-K-D-Wrath-K-Wrath-Limbo-I-D-K-D-Wrath-K-Wrath-I-O');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'Greed');
        (0, _metal.set)(this.context, 'first', this.createList(['Limbo']).list);
        (0, _metal.set)(this.context, 'fifth', this.createList(['Wrath']).list);
        (0, _metal.set)(this.context, 'ninth', this.createList(['Treachery']).list);
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');
    }

    ['@test it should support {{#each name as |foo|}}, then {{#each foo as |bar|}}']() {
      let inner = this.createList(['caterpillar']);
      let outer = this.createList([inner.list]);

      this.render(`{{#each name key="@index" as |foo|}}{{#each foo as |bar|}}{{bar}}{{/each}}{{/each}}`, {
        name: outer.list
      });

      this.assertText('caterpillar');

      this.runTask(() => this.rerender());

      this.assertText('caterpillar');

      this.runTask(() => {
        inner.delegate.replace(0, 1, ['lady']);
        outer.delegate.pushObject(this.createList(['bird']).list);
      });

      this.assertText('ladybird');

      this.runTask(() => (0, _metal.set)(this.context, 'name', this.createList([this.createList(['caterpillar']).list]).list));

      this.assertText('caterpillar');
    }
  }

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with native arrays', class extends EachTest {
    createList(items) {
      return { list: items, delegate: new ArrayDelegate(items, items) };
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with emberA-wrapped arrays', class extends EachTest {
    createList(items) {
      let wrapped = (0, _runtime.A)(items);
      return { list: wrapped, delegate: wrapped };
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with native Set', class extends EachTest {
    createList(items) {
      let set = makeSet(items);
      return { list: set, delegate: new SetDelegate(set) };
    }

    ['@test it can render duplicate primitive items'](assert) {
      assert.ok(true, 'not supported by Set');
    }

    ['@test it can render duplicate objects'](assert) {
      assert.ok(true, 'not supported by Set');
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with array-like objects implementing forEach', class extends EachTest {
    createList(items) {
      let forEachable = new ForEachable(items);
      return { list: forEachable, delegate: forEachable };
    }
  });

  if (_utils.HAS_NATIVE_SYMBOL) {
    (0, _testCase.moduleFor)('Syntax test: {{#each}} with array-like objects implementing Symbol.iterator', class extends EachTest {
      createList(items) {
        let iterable = new ArrayIterable(items);
        return { list: iterable, delegate: iterable };
      }
    });
  }

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with array proxies, modifying itself', class extends EachTest {
    createList(items) {
      let proxty = _runtime.ArrayProxy.create({ content: (0, _runtime.A)(items) });
      return { list: proxty, delegate: proxty };
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with array proxies, replacing its content', class extends EachTest {
    createList(items) {
      let wrapped = (0, _runtime.A)(items);
      return {
        list: wrapped,
        delegate: _runtime.ArrayProxy.create({ content: wrapped })
      };
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each as}} undefined path', class extends _testCase.RenderingTest {
    ['@test keying off of `undefined` does not render']() {
      this.render(_abstractTestCase.strip`
      {{#each foo.bar.baz as |thing|}}
        {{thing}}
      {{/each}}`, { foo: {} });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', { bar: { baz: ['Here!'] } }));

      this.assertText('Here!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('');
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#each}} with sparse arrays', class extends _testCase.RenderingTest {
    ['@test it should itterate over holes']() {
      let sparseArray = [];
      sparseArray[3] = 'foo';
      sparseArray[4] = 'bar';

      this.render(_abstractTestCase.strip`
      {{#each list as |value key|}}
        [{{key}}:{{value}}]
      {{/each}}`, { list: (0, _runtime.A)(sparseArray) });

      this.assertText('[0:][1:][2:][3:foo][4:bar]');

      this.assertStableRerender();

      this.runTask(() => {
        let list = (0, _metal.get)(this.context, 'list');
        list.pushObject('baz');
      });

      this.assertText('[0:][1:][2:][3:foo][4:bar][5:baz]');
    }
  });

  /* globals MutationObserver: false */
  if (typeof MutationObserver === 'function') {
    (0, _testCase.moduleFor)('Syntax test: {{#each as}} DOM mutation test', class extends _testCase.RenderingTest {
      constructor() {
        super(...arguments);
        this.observer = null;
      }

      observe(element) {
        let observer = this.observer = new MutationObserver(function () {});
        observer.observe(element, { childList: true, characterData: true });
      }

      teardown() {
        if (this.observer) {
          this.observer.disconnect();
        }

        super.teardown();
      }

      assertNoMutation() {
        this.assert.deepEqual(this.observer.takeRecords(), [], 'Expected no mutations');
      }

      expectMutations() {
        this.assert.ok(this.observer.takeRecords().length > 0, 'Expected some mutations');
      }

      ['@test {{#each}} should not mutate a subtree when the array has not changed [GH #14332]']() {
        let page = { title: 'Blog Posts' };

        let model = [{ title: 'Rails is omakase' }, { title: 'Ember is omakase' }];

        this.render(_abstractTestCase.strip`
        <h1>{{page.title}}</h1>

        <ul id="posts">
          {{#each model as |post|}}
            <li>{{post.title}}</li>
          {{/each}}
        </ul>
      `, { page, model });

        this.assertHTML(_abstractTestCase.strip`
        <h1>Blog Posts</h1>

        <ul id="posts">
          <li>Rails is omakase</li>
          <li>Ember is omakase</li>
        </ul>
      `);

        this.observe(this.$('#posts')[0]);

        // MutationObserver is async
        return _runtime.RSVP.Promise.resolve(() => {
          this.assertStableRerender();
        }).then(() => {
          this.assertNoMutation();

          this.runTask(() => (0, _metal.set)(this.context, 'page', { title: 'Essays' }));

          this.assertHTML(_abstractTestCase.strip`
          <h1>Essays</h1>

          <ul id="posts">
            <li>Rails is omakase</li>
            <li>Ember is omakase</li>
          </ul>
        `);
        }).then(() => {
          this.assertNoMutation();

          this.runTask(() => (0, _metal.set)(this.context.page, 'title', 'Think Pieces™'));

          this.assertHTML(_abstractTestCase.strip`
          <h1>Think Pieces™</h1>

          <ul id="posts">
            <li>Rails is omakase</li>
            <li>Ember is omakase</li>
          </ul>
        `);
        }).then(() => {
          // The last set is localized to the `page` object, so we do not expect Glimmer
          // to re-iterate the list
          this.assertNoMutation();
        });
      }
    });
  }
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/experimental-syntax-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer'], function (_testCase, _abstractTestCase, _glimmer) {
  'use strict';

  (0, _testCase.moduleFor)('registerMacros', class extends _testCase.RenderingTest {
    constructor() {
      let originalMacros = _glimmer._experimentalMacros.slice();

      (0, _glimmer._registerMacros)(blocks => {
        blocks.add('-let', (params, hash, _default, inverse, builder) => {
          builder.compileParams(params);
          builder.invokeStaticBlock(_default, params.length);
        });
      });

      super(...arguments);
      this.originalMacros = originalMacros;
    }

    teardown() {
      _glimmer._experimentalMacros.length = 0;
      this.originalMacros.forEach(macro => _glimmer._experimentalMacros.push(macro));

      super.teardown();
    }

    ['@test allows registering custom syntax via private API']() {
      this.render(_abstractTestCase.strip`
      {{#-let obj as |bar|}}
        {{bar}}
      {{/-let}}
    `, { obj: 'hello world!' });

      this.assertText('hello world!');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/if-unless-test', ['@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/runtime', '@ember/-internals/metal', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/shared-conditional-tests'], function (_helpers, _runtime, _metal, _abstractTestCase, _testCase, _sharedConditionalTests) {
  'use strict';

  (0, _testCase.moduleFor)('Syntax test: {{#if}} with inverse', class extends _sharedConditionalTests.IfUnlessWithSyntaxTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#if ${cond}}}${truthy}{{else}}${falsy}{{/if}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#unless}} with inverse', class extends _sharedConditionalTests.IfUnlessWithSyntaxTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#unless ${cond}}}${falsy}{{else}}${truthy}{{/unless}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#if}} and {{#unless}} without inverse', class extends _sharedConditionalTests.IfUnlessWithSyntaxTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#if ${cond}}}${truthy}{{/if}}{{#unless ${cond}}}${falsy}{{/unless}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#if}}', class extends _testCase.RenderingTest {
    ['@test using `if` with an `{{each}}` destroys components when transitioning to and from inverse (GH #12267)'](assert) {
      let destroyedChildrenCount = 0;

      this.registerComponent('foo-bar', {
        template: '{{number}}',
        ComponentClass: _helpers.Component.extend({
          willDestroy() {
            this._super();
            destroyedChildrenCount++;
          }
        })
      });

      this.render(_abstractTestCase.strip`
      {{#if cond}}
        {{#each numbers as |number|}}
          {{foo-bar number=number}}
        {{/each}}
      {{else}}
        Nothing Here!
      {{/if}}`, { cond: true, numbers: (0, _runtime.A)([1, 2, 3]) });

      this.assertText('123');

      this.runTask(() => this.rerender());

      this.assertText('123');

      this.runTask(() => (0, _metal.set)(this.context, 'cond', false));

      this.assertText('Nothing Here!');
      assert.equal(destroyedChildrenCount, 3, 'the children were properly destroyed');

      this.runTask(() => (0, _metal.set)(this.context, 'cond', true));

      this.assertText('123');
    }

    ['@test looking up `undefined` property defaults to false']() {
      this.render(_abstractTestCase.strip`
      {{#if foo.bar.baz}}
        Here!
      {{else}}
        Nothing Here!
      {{/if}}`, { foo: {} });

      this.assertText('Nothing Here!');

      this.runTask(() => this.rerender());

      this.assertText('Nothing Here!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', { bar: { baz: true } }));

      this.assertText('Here!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('Nothing Here!');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/in-element-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/test-helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer', '@ember/-internals/metal'], function (_testCase, _testHelpers, _abstractTestCase, _glimmer, _metal) {
  'use strict';

  (0, _testCase.moduleFor)('{{-in-element}}', class extends _testCase.RenderingTest {
    ['@test using {{#in-element whatever}} asserts']() {
      // the in-element keyword is not yet public API this test should be removed
      // once https://github.com/emberjs/rfcs/pull/287 lands and is enabled

      let el = document.createElement('div');
      expectAssertion(() => {
        this.render(_abstractTestCase.strip`{{#in-element el}}{{/in-element}}`, { el });
      }, /The {{in-element}} helper cannot be used. \('-top-level' @ L1:C0\)/);
    }

    ['@test allows rendering into an external element']() {
      let someElement = document.createElement('div');

      this.render(_abstractTestCase.strip`
      {{#-in-element someElement}}
        {{text}}
      {{/-in-element}}
    `, {
        someElement,
        text: 'Whoop!'
      });

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      (0, _testHelpers.equalTokens)(someElement, 'Whoop!');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'Huzzah!!'));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      (0, _testHelpers.equalTokens)(someElement, 'Huzzah!!');

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'Whoop!'));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      (0, _testHelpers.equalTokens)(someElement, 'Whoop!');
    }

    ['@test components are cleaned up properly'](assert) {
      let hooks = [];

      let someElement = document.createElement('div');

      this.registerComponent('modal-display', {
        ComponentClass: _glimmer.Component.extend({
          didInsertElement() {
            hooks.push('didInsertElement');
          },

          willDestroyElement() {
            hooks.push('willDestroyElement');
          }
        }),

        template: `{{text}}`
      });

      this.render(_abstractTestCase.strip`
      {{#if showModal}}
        {{#-in-element someElement}}
          {{modal-display text=text}}
        {{/-in-element}}
      {{/if}}
    `, {
        someElement,
        text: 'Whoop!',
        showModal: false
      });

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      (0, _testHelpers.equalTokens)(someElement, '');

      this.assertStableRerender();

      this.runTask(() => (0, _metal.set)(this.context, 'showModal', true));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      this.assertComponentElement(someElement.firstChild, {
        content: 'Whoop!'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'Huzzah!'));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      this.assertComponentElement(someElement.firstChild, {
        content: 'Huzzah!'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'text', 'Whoop!'));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      this.assertComponentElement(someElement.firstChild, {
        content: 'Whoop!'
      });

      this.runTask(() => (0, _metal.set)(this.context, 'showModal', false));

      (0, _testHelpers.equalTokens)(this.element, '<!---->');
      (0, _testHelpers.equalTokens)(someElement, '');

      assert.deepEqual(hooks, ['didInsertElement', 'willDestroyElement']);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/let-test', ['@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_metal, _runtime, _testCase, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('Syntax test: {{#let as}}', class extends _testCase.RenderingTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#let ${cond} as |test|}}${truthy}{{else}}${falsy}{{/let}}`;
    }

    ['@test it renders the block if `undefined` is passed as an argument']() {
      this.render(_abstractTestCase.strip`
        {{#let foo.bar.baz as |thing|}}
          value: "{{thing}}"
        {{/let}}`, { foo: {} });

      this.assertText('value: ""');

      this.runTask(() => this.rerender());

      this.assertText('value: ""');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', { bar: { baz: 'Here!' } }));

      this.assertText('value: "Here!"');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('value: ""');
    }

    ['@test it renders the block if arguments are falsey']() {
      this.render(`{{#let cond1 cond2 as |cond|}}value: "{{cond1}}"{{/let}}`, {
        cond1: false
      });

      this.assertText('value: "false"');

      this.runTask(() => this.rerender());

      this.assertText('value: "false"');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', ''));

      this.assertText('value: ""');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', 0));

      this.assertText('value: "0"');
    }

    ['@test it yields multiple arguments in order']() {
      this.render(`{{#let foo bar baz.name as |a b c|}}{{a}} {{b}} {{c}}{{/let}}`, {
        foo: 'Señor Engineer',
        bar: '',
        baz: { name: 'Dale' }
      });

      this.assertText('Señor Engineer  Dale');

      this.runTask(() => (0, _metal.set)(this.context, 'bar', 'Tom'));

      this.assertText('Señor Engineer Tom Dale');
    }

    ['@test can access alias and original scope']() {
      this.render(`{{#let person as |tom|}}{{title}}: {{tom.name}}{{/let}}`, {
        title: 'Señor Engineer',
        person: { name: 'Tom Dale' }
      });

      this.assertText('Señor Engineer: Tom Dale');

      this.runTask(() => this.rerender());

      this.assertText('Señor Engineer: Tom Dale');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'person.name', 'Yehuda Katz');
        (0, _metal.set)(this.context, 'title', 'Principal Engineer');
      });

      this.assertText('Principal Engineer: Yehuda Katz');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'person', { name: 'Tom Dale' });
        (0, _metal.set)(this.context, 'title', 'Señor Engineer');
      });

      this.assertText('Señor Engineer: Tom Dale');
    }

    ['@test the scoped variable is not available outside the {{#let}} block.']() {
      this.render(`{{name}}-{{#let other as |name|}}{{name}}{{/let}}-{{name}}`, {
        name: 'Stef',
        other: 'Yehuda'
      });

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => this.rerender());

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => (0, _metal.set)(this.context, 'other', 'Chad'));

      this.assertText('Stef-Chad-Stef');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Tom'));

      this.assertText('Tom-Chad-Tom');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'name', 'Stef');
        (0, _metal.set)(this.context, 'other', 'Yehuda');
      });

      this.assertText('Stef-Yehuda-Stef');
    }

    ['@test can access alias of a proxy']() {
      this.render(`{{#let proxy as |person|}}{{person.name}}{{/let}}`, {
        proxy: _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })
      });

      this.assertText('Tom Dale');

      this.runTask(() => this.rerender());

      this.assertText('Tom Dale');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.name', 'Yehuda Katz'));

      this.assertText('Yehuda Katz');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', { name: 'Godfrey Chan' }));

      this.assertText('Godfrey Chan');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name', 'Stefan Penner'));

      this.assertText('Stefan Penner');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', null));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy', _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })));

      this.assertText('Tom Dale');
    }

    ['@test can access alias of an array']() {
      this.render(`{{#let arrayThing as |words|}}{{#each words as |word|}}{{word}}{{/each}}{{/let}}`, {
        arrayThing: (0, _runtime.A)(['Hello', ' ', 'world'])
      });

      this.assertText('Hello world');

      this.runTask(() => this.rerender());

      this.assertText('Hello world');

      this.runTask(() => {
        let array = (0, _metal.get)(this.context, 'arrayThing');
        array.replace(0, 1, ['Goodbye']);
        (0, _runtime.removeAt)(array, 1);
        array.insertAt(1, ', ');
        array.pushObject('!');
      });

      this.assertText('Goodbye, world!');

      this.runTask(() => (0, _metal.set)(this.context, 'arrayThing', ['Hello', ' ', 'world']));

      this.assertText('Hello world');
    }

    ['@test `attrs` can be used as a block param [GH#14678]']() {
      this.render('{{#let hash as |attrs|}}[{{hash.foo}}-{{attrs.foo}}]{{/let}}', {
        hash: { foo: 'foo' }
      });

      this.assertText('[foo-foo]');

      this.runTask(() => this.rerender());

      this.assertText('[foo-foo]');

      this.runTask(() => this.context.set('hash.foo', 'FOO'));

      this.assertText('[FOO-FOO]');

      this.runTask(() => this.context.set('hash.foo', 'foo'));

      this.assertText('[foo-foo]');
    }
  });

  (0, _testCase.moduleFor)('Syntax test: Multiple {{#let as}} helpers', class extends _testCase.RenderingTest {
    ['@test re-using the same variable with different {{#let}} blocks does not override each other']() {
      this.render(`Admin: {{#let admin as |person|}}{{person.name}}{{/let}} User: {{#let user as |person|}}{{person.name}}{{/let}}`, {
        admin: { name: 'Tom Dale' },
        user: { name: 'Yehuda Katz' }
      });

      this.assertText('Admin: Tom Dale User: Yehuda Katz');

      this.runTask(() => this.rerender());

      this.assertText('Admin: Tom Dale User: Yehuda Katz');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'admin.name', 'Godfrey Chan');
        (0, _metal.set)(this.context, 'user.name', 'Stefan Penner');
      });

      this.assertText('Admin: Godfrey Chan User: Stefan Penner');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'admin', { name: 'Tom Dale' });
        (0, _metal.set)(this.context, 'user', { name: 'Yehuda Katz' });
      });

      this.assertText('Admin: Tom Dale User: Yehuda Katz');
    }

    ['@test the scoped variable is not available outside the {{#let}} block']() {
      this.render(`{{ring}}-{{#let first as |ring|}}{{ring}}-{{#let fifth as |ring|}}{{ring}}-{{#let ninth as |ring|}}{{ring}}-{{/let}}{{ring}}-{{/let}}{{ring}}-{{/let}}{{ring}}`, {
        ring: 'Greed',
        first: 'Limbo',
        fifth: 'Wrath',
        ninth: 'Treachery'
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => this.rerender());

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'O');
        (0, _metal.set)(this.context, 'fifth', 'D');
      });

      this.assertText('O-Limbo-D-Treachery-D-Limbo-O');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'first', 'I');
        (0, _metal.set)(this.context, 'ninth', 'K');
      });

      this.assertText('O-I-D-K-D-I-O');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'Greed');
        (0, _metal.set)(this.context, 'first', 'Limbo');
        (0, _metal.set)(this.context, 'fifth', 'Wrath');
        (0, _metal.set)(this.context, 'ninth', 'Treachery');
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');
    }

    ['@test it should support {{#let name as |foo|}}, then {{#let foo as |bar|}}']() {
      this.render(`{{#let name as |foo|}}{{#let foo as |bar|}}{{bar}}{{/let}}{{/let}}`, {
        name: 'caterpillar'
      });

      this.assertText('caterpillar');

      this.runTask(() => this.rerender());

      this.assertText('caterpillar');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'butterfly'));

      this.assertText('butterfly');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'caterpillar'));

      this.assertText('caterpillar');
    }

    ['@test updating the context should update the alias']() {
      this.render(`{{#let this as |person|}}{{person.name}}{{/let}}`, {
        name: 'Los Pivots'
      });

      this.assertText('Los Pivots');

      this.runTask(() => this.rerender());

      this.assertText('Los Pivots');

      this.runTask(() => (0, _metal.set)(this.context, 'name', "l'Pivots"));

      this.assertText("l'Pivots");

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Los Pivots'));

      this.assertText('Los Pivots');
    }

    ['@test nested {{#let}} blocks should have access to root context']() {
      this.render(_abstractTestCase.strip`
        {{name}}
        {{#let committer1.name as |name|}}
          [{{name}}
          {{#let committer2.name as |name|}}
            [{{name}}]
          {{/let}}
          {{name}}]
        {{/let}}
        {{name}}
        {{#let committer2.name as |name|}}
          [{{name}}
          {{#let committer1.name as |name|}}
            [{{name}}]
          {{/let}}
          {{name}}]
        {{/let}}
        {{name}}
      `, {
        name: 'ebryn',
        committer1: { name: 'trek' },
        committer2: { name: 'machty' }
      });

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');

      this.runTask(() => this.rerender());

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'chancancode'));

      this.assertText('chancancode[trek[machty]trek]chancancode[machty[trek]machty]chancancode');

      this.runTask(() => (0, _metal.set)(this.context, 'committer1', { name: 'krisselden' }));

      this.assertText('chancancode[krisselden[machty]krisselden]chancancode[machty[krisselden]machty]chancancode');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'committer1.name', 'wycats');
        (0, _metal.set)(this.context, 'committer2', { name: 'rwjblue' });
      });

      this.assertText('chancancode[wycats[rwjblue]wycats]chancancode[rwjblue[wycats]rwjblue]chancancode');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'name', 'ebryn');
        (0, _metal.set)(this.context, 'committer1', { name: 'trek' });
        (0, _metal.set)(this.context, 'committer2', { name: 'machty' });
      });

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/with-dynamic-var-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_testCase, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('{{-with-dynamic-var}}', class extends _testCase.RenderingTest {
    ['@test does not allow setting values other than outletState']() {
      expectAssertion(() => {
        this.render(_abstractTestCase.strip`
        {{#-with-dynamic-vars foo="bar"}}
          {{-get-dynamic-var 'foo'}}
        {{/-with-dynamic-vars}}
      `);
      }, /Using `-with-dynamic-scope` is only supported for `outletState` \(you used `foo`\)./);
    }

    ['@test allows setting/getting outletState']() {
      // this is simply asserting that we can write and read outletState
      // the actual value being used here is not what is used in real life
      // feel free to change the value being set and asserted as needed
      this.render(_abstractTestCase.strip`
      {{#-with-dynamic-vars outletState="bar"}}
        {{-get-dynamic-var 'outletState'}}
      {{/-with-dynamic-vars}}
    `);

      this.assertText('bar');
    }

    ['@test does not allow setting values other than outletState']() {
      expectAssertion(() => {
        this.render(`{{-get-dynamic-var 'foo'}}`);
      }, /Using `-get-dynamic-scope` is only supported for `outletState` \(you used `foo`\)./);
    }
  });
});
enifed('@ember/-internals/glimmer/tests/integration/syntax/with-test', ['@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/shared-conditional-tests', '@ember/-internals/glimmer/tests/utils/abstract-test-case'], function (_metal, _runtime, _testCase, _sharedConditionalTests, _abstractTestCase) {
  'use strict';

  (0, _testCase.moduleFor)('Syntax test: {{#with}}', class extends _sharedConditionalTests.IfUnlessWithSyntaxTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#with ${cond}}}${truthy}{{else}}${falsy}{{/with}}`;
    }
  });

  (0, _testCase.moduleFor)('Syntax test: {{#with as}}', class extends _sharedConditionalTests.IfUnlessWithSyntaxTest {
    templateFor({ cond, truthy, falsy }) {
      return `{{#with ${cond} as |test|}}${truthy}{{else}}${falsy}{{/with}}`;
    }

    ['@test keying off of `undefined` does not render']() {
      this.render(_abstractTestCase.strip`
      {{#with foo.bar.baz as |thing|}}
        {{thing}}
      {{/with}}`, { foo: {} });

      this.assertText('');

      this.runTask(() => this.rerender());

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', { bar: { baz: 'Here!' } }));

      this.assertText('Here!');

      this.runTask(() => (0, _metal.set)(this.context, 'foo', {}));

      this.assertText('');
    }

    ['@test it renders and hides the given block based on the conditional']() {
      this.render(`{{#with cond1 as |cond|}}{{cond.greeting}}{{else}}False{{/with}}`, {
        cond1: { greeting: 'Hello' }
      });

      this.assertText('Hello');

      this.runTask(() => this.rerender());

      this.assertText('Hello');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1.greeting', 'Hello world'));

      this.assertText('Hello world');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', false));

      this.assertText('False');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', { greeting: 'Hello' }));

      this.assertText('Hello');
    }

    ['@test can access alias and original scope']() {
      this.render(`{{#with person as |tom|}}{{title}}: {{tom.name}}{{/with}}`, {
        title: 'Señor Engineer',
        person: { name: 'Tom Dale' }
      });

      this.assertText('Señor Engineer: Tom Dale');

      this.runTask(() => this.rerender());

      this.assertText('Señor Engineer: Tom Dale');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'person.name', 'Yehuda Katz');
        (0, _metal.set)(this.context, 'title', 'Principal Engineer');
      });

      this.assertText('Principal Engineer: Yehuda Katz');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'person', { name: 'Tom Dale' });
        (0, _metal.set)(this.context, 'title', 'Señor Engineer');
      });

      this.assertText('Señor Engineer: Tom Dale');
    }

    ['@test the scoped variable is not available outside the {{#with}} block.']() {
      this.render(`{{name}}-{{#with other as |name|}}{{name}}{{/with}}-{{name}}`, {
        name: 'Stef',
        other: 'Yehuda'
      });

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => this.rerender());

      this.assertText('Stef-Yehuda-Stef');

      this.runTask(() => (0, _metal.set)(this.context, 'other', 'Chad'));

      this.assertText('Stef-Chad-Stef');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Tom'));

      this.assertText('Tom-Chad-Tom');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'name', 'Stef');
        (0, _metal.set)(this.context, 'other', 'Yehuda');
      });

      this.assertText('Stef-Yehuda-Stef');
    }

    ['@test inverse template is displayed with context']() {
      this.render(`{{#with falsyThing as |thing|}}Has Thing{{else}}No Thing {{otherThing}}{{/with}}`, {
        falsyThing: null,
        otherThing: 'bar'
      });

      this.assertText('No Thing bar');

      this.runTask(() => this.rerender());

      this.assertText('No Thing bar');

      this.runTask(() => (0, _metal.set)(this.context, 'otherThing', 'biz'));

      this.assertText('No Thing biz');

      this.runTask(() => (0, _metal.set)(this.context, 'falsyThing', true));

      this.assertText('Has Thing');

      this.runTask(() => (0, _metal.set)(this.context, 'otherThing', 'baz'));

      this.assertText('Has Thing');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'otherThing', 'bar');
        (0, _metal.set)(this.context, 'falsyThing', null);
      });

      this.assertText('No Thing bar');
    }

    ['@test can access alias of a proxy']() {
      this.render(`{{#with proxy as |person|}}{{person.name}}{{/with}}`, {
        proxy: _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })
      });

      this.assertText('Tom Dale');

      this.runTask(() => this.rerender());

      this.assertText('Tom Dale');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.name', 'Yehuda Katz'));

      this.assertText('Yehuda Katz');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', { name: 'Godfrey Chan' }));

      this.assertText('Godfrey Chan');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content.name', 'Stefan Penner'));

      this.assertText('Stefan Penner');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy.content', null));

      this.assertText('');

      this.runTask(() => (0, _metal.set)(this.context, 'proxy', _runtime.ObjectProxy.create({ content: { name: 'Tom Dale' } })));

      this.assertText('Tom Dale');
    }

    ['@test can access alias of an array']() {
      this.render(`{{#with arrayThing as |words|}}{{#each words as |word|}}{{word}}{{/each}}{{/with}}`, {
        arrayThing: (0, _runtime.A)(['Hello', ' ', 'world'])
      });

      this.assertText('Hello world');

      this.runTask(() => this.rerender());

      this.assertText('Hello world');

      this.runTask(() => {
        let array = (0, _metal.get)(this.context, 'arrayThing');
        array.replace(0, 1, ['Goodbye']);
        (0, _runtime.removeAt)(array, 1);
        array.insertAt(1, ', ');
        array.pushObject('!');
      });

      this.assertText('Goodbye, world!');

      this.runTask(() => (0, _metal.set)(this.context, 'arrayThing', ['Hello', ' ', 'world']));

      this.assertText('Hello world');
    }

    ['@test `attrs` can be used as a block param [GH#14678]']() {
      this.render('{{#with hash as |attrs|}}[{{hash.foo}}-{{attrs.foo}}]{{/with}}', {
        hash: { foo: 'foo' }
      });

      this.assertText('[foo-foo]');

      this.runTask(() => this.rerender());

      this.assertText('[foo-foo]');

      this.runTask(() => this.context.set('hash.foo', 'FOO'));

      this.assertText('[FOO-FOO]');

      this.runTask(() => this.context.set('hash.foo', 'foo'));

      this.assertText('[foo-foo]');
    }
  });

  (0, _testCase.moduleFor)('Syntax test: Multiple {{#with as}} helpers', class extends _testCase.RenderingTest {
    ['@test re-using the same variable with different {{#with}} blocks does not override each other']() {
      this.render(`Admin: {{#with admin as |person|}}{{person.name}}{{/with}} User: {{#with user as |person|}}{{person.name}}{{/with}}`, {
        admin: { name: 'Tom Dale' },
        user: { name: 'Yehuda Katz' }
      });

      this.assertText('Admin: Tom Dale User: Yehuda Katz');

      this.runTask(() => this.rerender());

      this.assertText('Admin: Tom Dale User: Yehuda Katz');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'admin.name', 'Godfrey Chan');
        (0, _metal.set)(this.context, 'user.name', 'Stefan Penner');
      });

      this.assertText('Admin: Godfrey Chan User: Stefan Penner');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'admin', { name: 'Tom Dale' });
        (0, _metal.set)(this.context, 'user', { name: 'Yehuda Katz' });
      });

      this.assertText('Admin: Tom Dale User: Yehuda Katz');
    }

    ['@test the scoped variable is not available outside the {{#with}} block']() {
      this.render(`{{ring}}-{{#with first as |ring|}}{{ring}}-{{#with fifth as |ring|}}{{ring}}-{{#with ninth as |ring|}}{{ring}}-{{/with}}{{ring}}-{{/with}}{{ring}}-{{/with}}{{ring}}`, {
        ring: 'Greed',
        first: 'Limbo',
        fifth: 'Wrath',
        ninth: 'Treachery'
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => this.rerender());

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'O');
        (0, _metal.set)(this.context, 'fifth', 'D');
      });

      this.assertText('O-Limbo-D-Treachery-D-Limbo-O');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'first', 'I');
        (0, _metal.set)(this.context, 'ninth', 'K');
      });

      this.assertText('O-I-D-K-D-I-O');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'ring', 'Greed');
        (0, _metal.set)(this.context, 'first', 'Limbo');
        (0, _metal.set)(this.context, 'fifth', 'Wrath');
        (0, _metal.set)(this.context, 'ninth', 'Treachery');
      });

      this.assertText('Greed-Limbo-Wrath-Treachery-Wrath-Limbo-Greed');
    }

    ['@test it should support {{#with name as |foo|}}, then {{#with foo as |bar|}}']() {
      this.render(`{{#with name as |foo|}}{{#with foo as |bar|}}{{bar}}{{/with}}{{/with}}`, {
        name: 'caterpillar'
      });

      this.assertText('caterpillar');

      this.runTask(() => this.rerender());

      this.assertText('caterpillar');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'butterfly'));

      this.assertText('butterfly');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'caterpillar'));

      this.assertText('caterpillar');
    }

    ['@test updating the context should update the alias']() {
      this.render(`{{#with this as |person|}}{{person.name}}{{/with}}`, {
        name: 'Los Pivots'
      });

      this.assertText('Los Pivots');

      this.runTask(() => this.rerender());

      this.assertText('Los Pivots');

      this.runTask(() => (0, _metal.set)(this.context, 'name', "l'Pivots"));

      this.assertText("l'Pivots");

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'Los Pivots'));

      this.assertText('Los Pivots');
    }

    ['@test nested {{#with}} blocks should have access to root context']() {
      this.render(_abstractTestCase.strip`
      {{name}}
      {{#with committer1.name as |name|}}
        [{{name}}
        {{#with committer2.name as |name|}}
          [{{name}}]
        {{/with}}
        {{name}}]
      {{/with}}
      {{name}}
      {{#with committer2.name as |name|}}
        [{{name}}
        {{#with committer1.name as |name|}}
          [{{name}}]
        {{/with}}
        {{name}}]
      {{/with}}
      {{name}}
    `, {
        name: 'ebryn',
        committer1: { name: 'trek' },
        committer2: { name: 'machty' }
      });

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');

      this.runTask(() => this.rerender());

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');

      this.runTask(() => (0, _metal.set)(this.context, 'name', 'chancancode'));

      this.assertText('chancancode[trek[machty]trek]chancancode[machty[trek]machty]chancancode');

      this.runTask(() => (0, _metal.set)(this.context, 'committer1', { name: 'krisselden' }));

      this.assertText('chancancode[krisselden[machty]krisselden]chancancode[machty[krisselden]machty]chancancode');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'committer1.name', 'wycats');
        (0, _metal.set)(this.context, 'committer2', { name: 'rwjblue' });
      });

      this.assertText('chancancode[wycats[rwjblue]wycats]chancancode[rwjblue[wycats]rwjblue]chancancode');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'name', 'ebryn');
        (0, _metal.set)(this.context, 'committer1', { name: 'trek' });
        (0, _metal.set)(this.context, 'committer2', { name: 'machty' });
      });

      this.assertText('ebryn[trek[machty]trek]ebryn[machty[trek]machty]ebryn');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/unit/outlet-test', ['@ember/-internals/glimmer', '@ember/runloop', 'internal-test-helpers'], function (_glimmer, _runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Glimmer OutletView', class extends _internalTestHelpers.AbstractTestCase {
    ['@test render in the render queue'](assert) {
      let didAppendOutletView = 0;
      let expectedOutlet = '#foo.bar';

      let renderer = {
        appendOutletView(view, target) {
          didAppendOutletView++;
          assert.equal(view, outletView);
          assert.equal(target, expectedOutlet);
        }
      };

      let outletView = new _glimmer.OutletView({}, renderer);

      (0, _runloop.run)(() => {
        assert.equal(didAppendOutletView, 0, 'appendOutletView should not yet have been called (before appendTo)');
        outletView.appendTo(expectedOutlet);
        assert.equal(didAppendOutletView, 0, 'appendOutletView should not yet have been called (sync after appendTo)');

        (0, _runloop.schedule)('actions', () => assert.equal(didAppendOutletView, 0, 'appendOutletView should not yet have been called (in actions)'));
        (0, _runloop.schedule)('render', () => assert.equal(didAppendOutletView, 1, 'appendOutletView should be invoked in render'));
      });
    }
  });
});
enifed('@ember/-internals/glimmer/tests/unit/runtime-resolver-cache-test', ['@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/metal', 'internal-test-helpers'], function (_testCase, _helpers, _metal, _internalTestHelpers) {
  'use strict';

  (0, _testCase.moduleFor)('ember-glimmer runtime resolver cache', class extends _testCase.RenderingTest {
    '@test a helper definition is only generated once'(assert) {
      this.registerHelper('foo-bar', () => 'foo-bar helper!');
      this.registerHelper('baz-qux', () => 'baz-qux helper!');

      // assert precondition
      let state = this.getCacheCounters();
      assert.deepEqual(state, {
        templateCacheHits: 0,
        templateCacheMisses: 0,
        componentDefinitionCount: 0,
        helperDefinitionCount: 0
      }, 'precondition');

      this.render(`
          {{~#if cond~}}
            {{foo-bar}}
          {{~else~}}
            {{baz-qux}}
          {{~/if}}`, {
        cond: true
      });

      this.assertText('foo-bar helper!');
      state = this.expectCacheChanges({ helperDefinitionCount: 1 }, state, 'calculate foo-bar helper only');

      // show component-two for the first time
      this.runTask(() => (0, _metal.set)(this.context, 'cond', false));

      this.assertText('baz-qux helper!');
      state = this.expectCacheChanges({
        helperDefinitionCount: 1
      }, state, 'calculate baz-qux helper, misses cache');

      // show foo-bar again
      this.runTask(() => (0, _metal.set)(this.context, 'cond', true));

      this.assertText('foo-bar helper!');
      state = this.expectCacheChanges({}, state, 'toggle back to foo-bar cache hit');

      // show baz-qux again
      this.runTask(() => (0, _metal.set)(this.context, 'cond', false));

      this.assertText('baz-qux helper!');
      state = this.expectCacheChanges({}, state, 'toggle back to baz-qux cache hit');
    }

    '@test a component definition is only generated once'(assert) {
      // static layout
      this.registerComponent('component-one', { template: 'One' });
      this.registerComponent('component-two', {
        ComponentClass: _helpers.Component.extend(),
        template: 'Two'
      });

      // assert precondition
      let state = this.getCacheCounters();
      assert.deepEqual(state, {
        templateCacheHits: 0,
        templateCacheMisses: 0,
        componentDefinitionCount: 0,
        helperDefinitionCount: 0
      }, 'precondition');

      // show component-one for the first time
      this.render(`{{component componentName}}`, {
        componentName: 'component-one'
      });

      this.assertText('One');
      state = this.expectCacheChanges({ componentDefinitionCount: 1 }, state, 'test case component and component-one no change');

      // show component-two for the first time
      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'component-two'));

      this.assertText('Two');
      state = this.expectCacheChanges({ componentDefinitionCount: 1 }, state, 'component-two first render');

      // show component-one again
      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'component-one'));

      this.assertText('One');
      state = this.expectCacheChanges({}, state, 'toggle back to component-one no change');

      // show component-two again
      this.runTask(() => (0, _metal.set)(this.context, 'componentName', 'component-two'));

      this.assertText('Two');
      state = this.expectCacheChanges({}, state, 'toggle back to component-two no change');
    }

    ['@test each template is only compiled once'](assert) {
      // static layout
      this.registerComponent('component-one', { template: 'One' });

      // test directly import template factory onto late bound layout
      let Two = _helpers.Component.extend({
        layout: this.compile('Two')
      });
      this.registerComponent('component-two', { ComponentClass: Two });

      // inject layout onto component, share layout with component-one
      this.registerComponent('root-component', { ComponentClass: _helpers.Component });
      this.owner.inject('component:root-component', 'layout', 'template:components/component-one');

      // template instance shared between to template managers
      let rootFactory = this.owner.factoryFor('component:root-component');

      // assert precondition
      let state = this.getCacheCounters();
      assert.deepEqual(state, {
        templateCacheHits: 0,
        templateCacheMisses: 0,
        componentDefinitionCount: 0,
        helperDefinitionCount: 0
      }, 'precondition');

      // show component-one for the first time
      this.render(`
    {{~#if cond~}}
      {{component-one}}
    {{~else~}}
      {{component-two}}
    {{~/if}}`, {
        cond: true
      });

      this.assertText('One');
      state = this.expectCacheChanges({ componentDefinitionCount: 1 }, state, 'test case component and component-one no change');

      // show component-two for the first time
      this.runTask(() => (0, _metal.set)(this.context, 'cond', false));

      this.assertText('Two');
      state = this.expectCacheChanges({
        templateCacheMisses: 1,
        componentDefinitionCount: 1
      }, state, 'component-two first render misses template cache');

      // show component-one again
      this.runTask(() => (0, _metal.set)(this.context, 'cond', true));

      this.assertText('One');
      state = this.expectCacheChanges({}, state, 'toggle back to component-one no change');

      // show component-two again
      this.runTask(() => (0, _metal.set)(this.context, 'cond', false));

      this.assertText('Two');
      state = this.expectCacheChanges({
        templateCacheHits: 1
      }, state, 'toggle back to component-two hits template cache');

      // render new root append
      let root = rootFactory.create();
      try {
        (0, _internalTestHelpers.runAppend)(root);
        this.assertText('TwoOne');
        // roots have different capabilities so this will hit
        state = this.expectCacheChanges({}, state, 'append root with component-one no change');

        // render new root append
        let root2 = rootFactory.create();
        try {
          (0, _internalTestHelpers.runAppend)(root2);
          this.assertText('TwoOneOne');
          state = this.expectCacheChanges({}, state, 'append another root no change');
        } finally {
          (0, _internalTestHelpers.runDestroy)(root2);
        }
      } finally {
        (0, _internalTestHelpers.runDestroy)(root);
      }
    }

    getCacheCounters() {
      let {
        templateCacheHits,
        templateCacheMisses,
        componentDefinitionCount,
        helperDefinitionCount
      } = this.runtimeResolver;

      return {
        templateCacheHits,
        templateCacheMisses,
        componentDefinitionCount,
        helperDefinitionCount
      };
    }

    expectCacheChanges(expected, lastState, message) {
      let state = this.getCacheCounters();
      let actual = diff(state, lastState);
      this.assert.deepEqual(actual, expected, message);
      return state;
    }
  });

  function diff(state, lastState) {
    let res = {};
    Object.keys(state).forEach(key => {
      let delta = state[key] - lastState[key];
      if (delta !== 0) {
        res[key] = state[key] - lastState[key];
      }
    });
    return res;
  }
});
enifed('@ember/-internals/glimmer/tests/unit/template-factory-test', ['ember-template-compiler', '@ember/-internals/glimmer', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/glimmer/tests/utils/helpers'], function (_emberTemplateCompiler, _glimmer, _testCase, _helpers) {
  'use strict';

  (0, _testCase.moduleFor)('Template factory test', class extends _testCase.RenderingTest {
    ['@test the template factory returned from precompile is the same as compile'](assert) {
      let { owner } = this;
      let { runtimeResolver } = this;

      let templateStr = 'Hello {{name}}';
      let options = { moduleName: 'my-app/templates/some-module.hbs' };

      let spec = (0, _emberTemplateCompiler.precompile)(templateStr, options);
      let body = `exports.default = template(${spec});`;
      let module = new Function('exports', 'template', body);
      let exports = {};
      module(exports, _glimmer.template);
      let Precompiled = exports['default'];

      let Compiled = (0, _emberTemplateCompiler.compile)(templateStr, options);

      assert.equal(typeof Precompiled.create, 'function', 'precompiled is a factory');
      assert.ok(Precompiled.id, 'precompiled has id');

      assert.equal(typeof Compiled.create, 'function', 'compiled is a factory');
      assert.ok(Compiled.id, 'compiled has id');

      assert.equal(runtimeResolver.templateCacheMisses, 0, 'misses 0');
      assert.equal(runtimeResolver.templateCacheHits, 0, 'hits 0');

      let precompiled = runtimeResolver.createTemplate(Precompiled, owner);

      assert.equal(runtimeResolver.templateCacheMisses, 1, 'misses 1');
      assert.equal(runtimeResolver.templateCacheHits, 0, 'hits 0');

      let compiled = runtimeResolver.createTemplate(Compiled, owner);

      assert.equal(runtimeResolver.templateCacheMisses, 2, 'misses 2');
      assert.equal(runtimeResolver.templateCacheHits, 0, 'hits 0');

      assert.ok(typeof precompiled.spec !== 'string', 'Spec has been parsed');
      assert.ok(typeof compiled.spec !== 'string', 'Spec has been parsed');

      this.registerComponent('x-precompiled', {
        ComponentClass: _helpers.Component.extend({
          layout: Precompiled
        })
      });

      this.registerComponent('x-compiled', {
        ComponentClass: _helpers.Component.extend({
          layout: Compiled
        })
      });

      this.render('{{x-precompiled name="precompiled"}} {{x-compiled name="compiled"}}');

      assert.equal(runtimeResolver.templateCacheMisses, 2, 'misses 2');
      assert.equal(runtimeResolver.templateCacheHits, 2, 'hits 2');

      this.assertText('Hello precompiled Hello compiled');
    }
  });
});
enifed('@ember/-internals/glimmer/tests/unit/utils/debug-stack-test', ['@ember/-internals/glimmer', 'internal-test-helpers'], function (_glimmer, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Glimmer DebugStack', class extends _internalTestHelpers.AbstractTestCase {
    ['@test pushing and popping'](assert) {
      if (true /* DEBUG */) {
          let stack = new _glimmer.DebugStack();

          assert.equal(stack.peek(), undefined);

          stack.push('template:application');

          assert.equal(stack.peek(), '"template:application"');

          stack.push('component:top-level-component');

          assert.equal(stack.peek(), '"component:top-level-component"');

          stack.pushEngine('engine:my-engine');
          stack.push('component:component-in-engine');

          assert.equal(stack.peek(), '"component:component-in-engine" (in "engine:my-engine")');

          stack.pop();
          stack.pop();
          let item = stack.pop();

          assert.equal(item, 'component:top-level-component');
          assert.equal(stack.peek(), '"template:application"');
        } else {
        assert.expect(0);
      }
    }
  });
});
enifed('@ember/-internals/glimmer/tests/utils/abstract-test-case', ['exports', 'internal-test-helpers'], function (exports, _internalTestHelpers) {
  'use strict';

  Object.defineProperty(exports, 'TestCase', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.AbstractTestCase;
    }
  });
  Object.defineProperty(exports, 'applyMixins', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.applyMixins;
    }
  });
  Object.defineProperty(exports, 'strip', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.strip;
    }
  });
});
enifed('@ember/-internals/glimmer/tests/utils/helpers', ['exports', 'ember-template-compiler', '@ember/-internals/glimmer'], function (exports, _emberTemplateCompiler, _glimmer) {
  'use strict';

  Object.defineProperty(exports, 'compile', {
    enumerable: true,
    get: function () {
      return _emberTemplateCompiler.compile;
    }
  });
  Object.defineProperty(exports, 'precompile', {
    enumerable: true,
    get: function () {
      return _emberTemplateCompiler.precompile;
    }
  });
  Object.defineProperty(exports, 'INVOKE', {
    enumerable: true,
    get: function () {
      return _glimmer.INVOKE;
    }
  });
  Object.defineProperty(exports, 'Helper', {
    enumerable: true,
    get: function () {
      return _glimmer.Helper;
    }
  });
  Object.defineProperty(exports, 'helper', {
    enumerable: true,
    get: function () {
      return _glimmer.helper;
    }
  });
  Object.defineProperty(exports, 'Component', {
    enumerable: true,
    get: function () {
      return _glimmer.Component;
    }
  });
  Object.defineProperty(exports, 'LinkComponent', {
    enumerable: true,
    get: function () {
      return _glimmer.LinkComponent;
    }
  });
  Object.defineProperty(exports, 'InteractiveRenderer', {
    enumerable: true,
    get: function () {
      return _glimmer.InteractiveRenderer;
    }
  });
  Object.defineProperty(exports, 'InertRenderer', {
    enumerable: true,
    get: function () {
      return _glimmer.InertRenderer;
    }
  });
  Object.defineProperty(exports, 'htmlSafe', {
    enumerable: true,
    get: function () {
      return _glimmer.htmlSafe;
    }
  });
  Object.defineProperty(exports, 'SafeString', {
    enumerable: true,
    get: function () {
      return _glimmer.SafeString;
    }
  });
  Object.defineProperty(exports, 'DOMChanges', {
    enumerable: true,
    get: function () {
      return _glimmer.DOMChanges;
    }
  });
  Object.defineProperty(exports, 'isHTMLSafe', {
    enumerable: true,
    get: function () {
      return _glimmer.isHTMLSafe;
    }
  });
});
enifed('@ember/-internals/glimmer/tests/utils/shared-conditional-tests', ['exports', '@ember/polyfills', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case', '@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/glimmer/tests/utils/helpers'], function (exports, _polyfills, _abstractTestCase, _testCase, _metal, _runtime, _helpers) {
  'use strict';

  exports.IfUnlessWithSyntaxTest = exports.TogglingSyntaxConditionalsTest = exports.IfUnlessHelperTest = exports.TogglingHelperConditionalsTest = exports.TogglingConditionalsTest = exports.ArrayTestCases = exports.ObjectTestCases = exports.BasicConditionalsTest = exports.StableFalsyGenerator = exports.StableTruthyGenerator = exports.FalsyGenerator = exports.TruthyGenerator = undefined;


  class AbstractConditionalsTest extends _testCase.RenderingTest {
    get truthyValue() {
      return true;
    }

    get falsyValue() {
      return false;
    }

    wrapperFor(templates) {
      return templates.join('');
    }

    wrappedTemplateFor(options) {
      return this.wrapperFor([this.templateFor(options)]);
    }

    /* abstract */
    templateFor() /* { cond, truthy, falsy } */{
      // e.g. `{{#if ${cond}}}${truthy}{{else}}${falsy}{{/if}}`
      throw new Error('Not implemented: `templateFor`');
    }

    /* abstract */
    renderValues() /* ...values */{
      throw new Error('Not implemented: `renderValues`');
    }
  }

  class AbstractGenerator {
    constructor(cases) {
      this.cases = cases;
    }

    /* abstract */
    generate() /* value, idx */{
      throw new Error('Not implemented: `generate`');
    }
  }

  /*
    The test cases in this file generally follow the following pattern:
  
    1. Render with [ truthy, ...(other truthy variations), falsy, ...(other falsy variations) ]
    2. No-op rerender
    3. Make all of them falsy (through interior mutation)
    4. Make all of them truthy (through interior mutation, sometimes with some slight variations)
    5. Reset them to their original values (through replacement)
  */

  class TruthyGenerator extends AbstractGenerator {
    generate(value, idx) {
      return {
        [`@test it should consider ${JSON.stringify(value)} truthy [${idx}]`]() {
          this.renderValues(value);

          this.assertText('T1');

          this.runTask(() => this.rerender());

          this.assertText('T1');

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.falsyValue));

          this.assertText('F1');

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', value));

          this.assertText('T1');
        }
      };
    }
  }

  exports.TruthyGenerator = TruthyGenerator;
  class FalsyGenerator extends AbstractGenerator {
    generate(value, idx) {
      return {
        [`@test it should consider ${JSON.stringify(value)} falsy [${idx}]`]() {
          this.renderValues(value);

          this.assertText('F1');

          this.runTask(() => this.rerender());

          this.assertText('F1');

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.truthyValue));

          this.assertText('T1');

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', value));

          this.assertText('F1');
        }
      };
    }
  }

  exports.FalsyGenerator = FalsyGenerator;
  class StableTruthyGenerator extends TruthyGenerator {
    generate(value, idx) {
      return (0, _polyfills.assign)(super.generate(value, idx), {
        [`@test it maintains DOM stability when condition changes from ${value} to another truthy value and back [${idx}]`]() {
          this.renderValues(value);

          this.assertText('T1');

          this.takeSnapshot();

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.truthyValue));

          this.assertText('T1');

          this.assertInvariants();

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', value));

          this.assertText('T1');

          this.assertInvariants();
        }
      });
    }
  }

  exports.StableTruthyGenerator = StableTruthyGenerator;
  class StableFalsyGenerator extends FalsyGenerator {
    generate(value, idx) {
      return (0, _polyfills.assign)(super.generate(value), {
        [`@test it maintains DOM stability when condition changes from ${value} to another falsy value and back [${idx}]`]() {
          this.renderValues(value);

          this.assertText('F1');

          this.takeSnapshot();

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.falsyValue));

          this.assertText('F1');

          this.assertInvariants();

          this.runTask(() => (0, _metal.set)(this.context, 'cond1', value));

          this.assertText('F1');

          this.assertInvariants();
        }
      });
    }
  }

  exports.StableFalsyGenerator = StableFalsyGenerator;
  class ObjectProxyGenerator extends AbstractGenerator {
    generate(value, idx) {
      // This is inconsistent with our usual to-bool policy, but the current proxy implementation
      // simply uses !!content to determine truthiness
      if (value) {
        return {
          [`@test it should consider an object proxy with \`${JSON.stringify(value)}\` truthy [${idx}]`]() {
            this.renderValues(_runtime.ObjectProxy.create({ content: value }));

            this.assertText('T1');

            this.runTask(() => this.rerender());

            this.assertText('T1');

            this.runTask(() => (0, _metal.set)(this.context, 'cond1.content', this.falsyValue));

            this.assertText('F1');

            this.runTask(() => (0, _metal.set)(this.context, 'cond1', _runtime.ObjectProxy.create({ content: value })));

            this.assertText('T1');
          }
        };
      } else {
        return {
          [`@test it should consider an object proxy with \`${JSON.stringify(value)}\` falsy [${idx}]`]() {
            this.renderValues(_runtime.ObjectProxy.create({ content: value }));

            this.assertText('F1');

            this.runTask(() => this.rerender());

            this.assertText('F1');

            this.runTask(() => (0, _metal.set)(this.context, 'cond1.content', this.truthyValue));

            this.assertText('T1');

            this.runTask(() => (0, _metal.set)(this.context, 'cond1', _runtime.ObjectProxy.create({ content: value })));

            this.assertText('F1');
          }
        };
      }
    }
  }

  // Testing behaviors shared across all conditionals, i.e. {{#if}}, {{#unless}},
  // {{#with}}, {{#each}}, {{#each-in}}, (if) and (unless)
  class BasicConditionalsTest extends AbstractConditionalsTest {
    ['@test it renders the corresponding block based on the conditional']() {
      this.renderValues(this.truthyValue, this.falsyValue);

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.falsyValue));

      this.assertText('F1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
      });

      this.assertText('T1T2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
      });

      this.assertText('T1F2');
    }
  }

  exports.BasicConditionalsTest = BasicConditionalsTest;
  // Testing behaviors related to ember objects, object proxies, etc
  const ObjectTestCases = exports.ObjectTestCases = {
    ['@test it considers object proxies without content falsy']() {
      this.renderValues(_runtime.ObjectProxy.create({ content: {} }), _runtime.ObjectProxy.create({ content: _runtime.Object.create() }), _runtime.ObjectProxy.create({ content: null }));

      this.assertText('T1T2F3');

      this.runTask(() => this.rerender());

      this.assertText('T1T2F3');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1.content', null);
        (0, _metal.set)(this.context, 'cond2.content', null);
      });

      this.assertText('F1F2F3');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1.content', _runtime.Object.create());
        (0, _metal.set)(this.context, 'cond2.content', {});
        (0, _metal.set)(this.context, 'cond3.content', { foo: 'bar' });
      });

      this.assertText('T1T2T3');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', _runtime.ObjectProxy.create({ content: {} }));
        (0, _metal.set)(this.context, 'cond2', _runtime.ObjectProxy.create({ content: _runtime.Object.create() }));
        (0, _metal.set)(this.context, 'cond3', _runtime.ObjectProxy.create({ content: null }));
      });

      this.assertText('T1T2F3');
    }
  };

  // Testing behaviors related to arrays and array proxies
  const ArrayTestCases = exports.ArrayTestCases = {
    ['@test it considers empty arrays falsy']() {
      this.renderValues((0, _runtime.A)(['hello']), (0, _runtime.A)());

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => (0, _runtime.removeAt)((0, _metal.get)(this.context, 'cond1'), 0));

      this.assertText('F1F2');

      this.runTask(() => {
        (0, _metal.get)(this.context, 'cond1').pushObject('hello');
        (0, _metal.get)(this.context, 'cond2').pushObjects([1]);
      });

      this.assertText('T1T2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', (0, _runtime.A)(['hello']));
        (0, _metal.set)(this.context, 'cond2', (0, _runtime.A)());
      });

      this.assertText('T1F2');
    },

    ['@test it considers array proxies without content falsy']() {
      this.renderValues(_runtime.ArrayProxy.create({ content: (0, _runtime.A)(['hello']) }), _runtime.ArrayProxy.create({ content: null }));

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1.content', null);
        (0, _metal.set)(this.context, 'cond2.content', null);
      });

      this.assertText('F1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1.content', (0, _runtime.A)(['hello']));
        (0, _metal.set)(this.context, 'cond2.content', (0, _runtime.A)([1]));
      });

      this.assertText('T1T2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', _runtime.ArrayProxy.create({ content: (0, _runtime.A)(['hello']) }));
        (0, _metal.set)(this.context, 'cond2', _runtime.ArrayProxy.create({ content: null }));
      });

      this.assertText('T1F2');
    },

    ['@test it considers array proxies with empty arrays falsy']() {
      this.renderValues(_runtime.ArrayProxy.create({ content: (0, _runtime.A)(['hello']) }), _runtime.ArrayProxy.create({ content: (0, _runtime.A)() }));

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => (0, _runtime.removeAt)((0, _metal.get)(this.context, 'cond1.content'), 0));

      this.assertText('F1F2');

      this.runTask(() => {
        (0, _metal.get)(this.context, 'cond1.content').pushObject('hello');
        (0, _metal.get)(this.context, 'cond2.content').pushObjects([1]);
      });

      this.assertText('T1T2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', _runtime.ArrayProxy.create({ content: (0, _runtime.A)(['hello']) }));
        (0, _metal.set)(this.context, 'cond2', _runtime.ArrayProxy.create({ content: (0, _runtime.A)() }));
      });

      this.assertText('T1F2');
    }
  };

  const IfUnlessWithTestCases = [new StableTruthyGenerator([true, ' ', 'hello', 'false', 'null', 'undefined', 1, ['hello'], (0, _runtime.A)(['hello']), {}, { foo: 'bar' }, _runtime.Object.create(), _runtime.Object.create({ foo: 'bar' }), _runtime.ObjectProxy.create({ content: true }), Object, function () {}, new String('hello'), new String(''), new Boolean(true), new Boolean(false), new Date()]), new StableFalsyGenerator([false, null, undefined, '', 0, [], (0, _runtime.A)(), _runtime.ObjectProxy.create({ content: undefined })]), new ObjectProxyGenerator([true, ' ', 'hello', 'false', 'null', 'undefined', 1, ['hello'], (0, _runtime.A)(['hello']), _runtime.ArrayProxy.create({ content: ['hello'] }), _runtime.ArrayProxy.create({ content: [] }), {}, { foo: 'bar' }, _runtime.Object.create(), _runtime.Object.create({ foo: 'bar' }), _runtime.ObjectProxy.create({ content: true }), _runtime.ObjectProxy.create({ content: undefined }), new String('hello'), new String(''), new Boolean(true), new Boolean(false), new Date(), false, null, undefined, '', 0, [], (0, _runtime.A)()]), ObjectTestCases, ArrayTestCases];

  // Testing behaviors shared across the "toggling" conditionals, i.e. {{#if}},
  // {{#unless}}, {{#with}}, {{#each}}, {{#each-in}}, (if) and (unless)
  class TogglingConditionalsTest extends BasicConditionalsTest {}

  exports.TogglingConditionalsTest = TogglingConditionalsTest;
  // Testing behaviors shared across the (if) and (unless) helpers
  class TogglingHelperConditionalsTest extends TogglingConditionalsTest {
    renderValues(...values) {
      let templates = [];
      let context = {};

      for (let i = 1; i <= values.length; i++) {
        templates.push(this.templateFor({ cond: `cond${i}`, truthy: `t${i}`, falsy: `f${i}` }));
        context[`t${i}`] = `T${i}`;
        context[`f${i}`] = `F${i}`;
        context[`cond${i}`] = values[i - 1];
      }

      let wrappedTemplate = this.wrapperFor(templates);
      this.render(wrappedTemplate, context);
    }

    ['@test it has access to the outer scope from both templates']() {
      let template = this.wrapperFor([this.templateFor({ cond: 'cond1', truthy: 'truthy', falsy: 'falsy' }), this.templateFor({ cond: 'cond2', truthy: 'truthy', falsy: 'falsy' })]);

      this.render(template, {
        cond1: this.truthyValue,
        cond2: this.falsyValue,
        truthy: 'YES',
        falsy: 'NO'
      });

      this.assertText('YESNO');

      this.runTask(() => this.rerender());

      this.assertText('YESNO');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'truthy', 'YASS');
        (0, _metal.set)(this.context, 'falsy', 'NOPE');
      });

      this.assertText('YASSNOPE');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.falsyValue);
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
      });

      this.assertText('NOPEYASS');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'truthy', 'YES');
        (0, _metal.set)(this.context, 'falsy', 'NO');
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
      });

      this.assertText('YESNO');
    }

    ['@test it does not update when the unbound helper is used']() {
      let template = this.wrapperFor([this.templateFor({
        cond: '(unbound cond1)',
        truthy: '"T1"',
        falsy: '"F1"'
      }), this.templateFor({
        cond: '(unbound cond2)',
        truthy: '"T2"',
        falsy: '"F2"'
      })]);

      this.render(template, { cond1: this.truthyValue, cond2: this.falsyValue });

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.falsyValue));

      this.assertText('T1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
      });

      this.assertText('T1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
      });

      this.assertText('T1F2');
    }

    ['@test evaluation should be lazy'](assert) {
      let truthyEvaluated;
      let falsyEvaluated;

      let withoutEvaluatingTruthy = callback => {
        truthyEvaluated = false;
        callback();
        assert.ok(!truthyEvaluated, 'x-truthy is not evaluated');
      };

      let withoutEvaluatingFalsy = callback => {
        falsyEvaluated = false;
        callback();
        assert.ok(!falsyEvaluated, 'x-falsy is not evaluated');
      };

      this.registerHelper('x-truthy', {
        compute() {
          truthyEvaluated = true;
          return 'T';
        }
      });

      this.registerHelper('x-falsy', {
        compute() {
          falsyEvaluated = true;
          return 'F';
        }
      });

      let template = this.wrappedTemplateFor({
        cond: 'cond',
        truthy: '(x-truthy)',
        falsy: '(x-falsy)'
      });

      withoutEvaluatingFalsy(() => this.render(template, { cond: this.truthyValue }));

      this.assertText('T');

      withoutEvaluatingFalsy(() => this.runTask(() => this.rerender()));

      this.assertText('T');

      withoutEvaluatingTruthy(() => this.runTask(() => (0, _metal.set)(this.context, 'cond', this.falsyValue)));

      this.assertText('F');

      withoutEvaluatingTruthy(() => this.runTask(() => this.rerender()));

      this.assertText('F');

      withoutEvaluatingFalsy(() => this.runTask(() => (0, _metal.set)(this.context, 'cond', this.truthyValue)));

      this.assertText('T');
    }
  }

  exports.TogglingHelperConditionalsTest = TogglingHelperConditionalsTest;
  class IfUnlessHelperTest extends TogglingHelperConditionalsTest {}

  exports.IfUnlessHelperTest = IfUnlessHelperTest;
  (0, _abstractTestCase.applyMixins)(IfUnlessHelperTest, ...IfUnlessWithTestCases);

  // Testing behaviors shared across the "toggling" syntatical constructs,
  // i.e. {{#if}}, {{#unless}}, {{#with}}, {{#each}} and {{#each-in}}
  class TogglingSyntaxConditionalsTest extends TogglingConditionalsTest {
    renderValues(...values) {
      let templates = [];
      let context = {};

      for (let i = 1; i <= values.length; i++) {
        templates.push(this.templateFor({
          cond: `cond${i}`,
          truthy: `{{t}}${i}`,
          falsy: `{{f}}${i}`
        }));
        context[`cond${i}`] = values[i - 1];
      }

      let wrappedTemplate = this.wrapperFor(templates);
      this.render(wrappedTemplate, (0, _polyfills.assign)({ t: 'T', f: 'F' }, context));
    }

    ['@test it does not update when the unbound helper is used']() {
      let template = `${this.templateFor({
        cond: '(unbound cond1)',
        truthy: 'T1',
        falsy: 'F1'
      })}${this.templateFor({
        cond: '(unbound cond2)',
        truthy: 'T2',
        falsy: 'F2'
      })}`;

      this.render(template, { cond1: this.truthyValue, cond2: this.falsyValue });

      this.assertText('T1F2');

      this.runTask(() => this.rerender());

      this.assertText('T1F2');

      this.runTask(() => (0, _metal.set)(this.context, 'cond1', this.falsyValue));

      this.assertText('T1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
      });

      this.assertText('T1F2');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
      });

      this.assertText('T1F2');
    }

    ['@test it has access to the outer scope from both templates']() {
      let template = this.wrapperFor([this.templateFor({
        cond: 'cond1',
        truthy: '{{truthy}}',
        falsy: '{{falsy}}'
      }), this.templateFor({
        cond: 'cond2',
        truthy: '{{truthy}}',
        falsy: '{{falsy}}'
      })]);

      this.render(template, {
        cond1: this.truthyValue,
        cond2: this.falsyValue,
        truthy: 'YES',
        falsy: 'NO'
      });

      this.assertText('YESNO');

      this.runTask(() => this.rerender());

      this.assertText('YESNO');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'truthy', 'YASS');
        (0, _metal.set)(this.context, 'falsy', 'NOPE');
      });

      this.assertText('YASSNOPE');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond1', this.falsyValue);
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
      });

      this.assertText('NOPEYASS');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'truthy', 'YES');
        (0, _metal.set)(this.context, 'falsy', 'NO');
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
      });

      this.assertText('YESNO');
    }

    ['@test it updates correctly when enclosing another conditional']() {
      // This tests whether the outer conditional tracks its bounds correctly as its inner bounds changes
      let inner = this.templateFor({
        cond: 'inner',
        truthy: 'T-inner',
        falsy: 'F-inner'
      });
      let template = this.wrappedTemplateFor({
        cond: 'outer',
        truthy: inner,
        falsy: 'F-outer'
      });

      this.render(template, { outer: this.truthyValue, inner: this.truthyValue });

      this.assertText('T-inner');

      this.runTask(() => this.rerender());

      this.assertText('T-inner');

      // Changes the inner bounds
      this.runTask(() => (0, _metal.set)(this.context, 'inner', this.falsyValue));

      this.assertText('F-inner');

      // Now rerender the outer conditional, which require first clearing its bounds
      this.runTask(() => (0, _metal.set)(this.context, 'outer', this.falsyValue));

      this.assertText('F-outer');
    }

    ['@test it updates correctly when enclosing #each']() {
      // This tests whether the outer conditional tracks its bounds correctly as its inner bounds changes
      let template = this.wrappedTemplateFor({
        cond: 'outer',
        truthy: '{{#each inner as |text|}}{{text}}{{/each}}',
        falsy: 'F-outer'
      });

      this.render(template, {
        outer: this.truthyValue,
        inner: ['inner', '-', 'before']
      });

      this.assertText('inner-before');

      this.runTask(() => this.rerender());

      this.assertText('inner-before');

      // Changes the inner bounds
      this.runTask(() => (0, _metal.set)(this.context, 'inner', ['inner-after']));

      this.assertText('inner-after');

      // Now rerender the outer conditional, which require first clearing its bounds
      this.runTask(() => (0, _metal.set)(this.context, 'outer', this.falsyValue));

      this.assertText('F-outer');

      // Reset
      this.runTask(() => {
        (0, _metal.set)(this.context, 'inner', ['inner-again']);
        (0, _metal.set)(this.context, 'outer', this.truthyValue);
      });

      this.assertText('inner-again');

      // Now clear the inner bounds
      this.runTask(() => (0, _metal.set)(this.context, 'inner', []));

      this.assertText('');

      // Now rerender the outer conditional, which require first clearing its bounds
      this.runTask(() => (0, _metal.set)(this.context, 'outer', this.falsyValue));

      this.assertText('F-outer');
    }

    ['@test it updates correctly when enclosing triple-curlies']() {
      // This tests whether the outer conditional tracks its bounds correctly as its inner bounds changes
      let template = this.wrappedTemplateFor({
        cond: 'outer',
        truthy: '{{{inner}}}',
        falsy: 'F-outer'
      });

      this.render(template, {
        outer: this.truthyValue,
        inner: '<b>inner</b>-<b>before</b>'
      });

      this.assertText('inner-before');

      this.runTask(() => this.rerender());

      this.assertText('inner-before');

      // Changes the inner bounds
      this.runTask(() => (0, _metal.set)(this.context, 'inner', '<p>inner-after</p>'));

      this.assertText('inner-after');

      // Now rerender the outer conditional, which require first clearing its bounds
      this.runTask(() => (0, _metal.set)(this.context, 'outer', this.falsyValue));

      this.assertText('F-outer');
    }

    ['@test child conditional should not render children if parent conditional becomes false'](assert) {
      let childCreated = false;

      this.registerComponent('foo-bar', {
        template: 'foo-bar',
        ComponentClass: _helpers.Component.extend({
          init() {
            this._super(...arguments);
            childCreated = true;
          }
        })
      });

      let innerTemplate = this.templateFor({
        cond: 'cond2',
        truthy: '{{foo-bar}}',
        falsy: ''
      });
      let wrappedTemplate = this.wrappedTemplateFor({
        cond: 'cond1',
        truthy: innerTemplate,
        falsy: ''
      });

      this.render(wrappedTemplate, {
        cond1: this.truthyValue,
        cond2: this.falsyValue
      });

      assert.ok(!childCreated);
      this.assertText('');

      this.runTask(() => this.rerender());

      assert.ok(!childCreated);
      this.assertText('');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond2', this.truthyValue);
        (0, _metal.set)(this.context, 'cond1', this.falsyValue);
      });

      assert.ok(!childCreated);
      this.assertText('');

      this.runTask(() => {
        (0, _metal.set)(this.context, 'cond2', this.falsyValue);
        (0, _metal.set)(this.context, 'cond1', this.truthyValue);
      });

      assert.ok(!childCreated);
      this.assertText('');
    }

    ['@test evaluation should be lazy'](assert) {
      let truthyEvaluated;
      let falsyEvaluated;

      let withoutEvaluatingTruthy = callback => {
        truthyEvaluated = false;
        callback();
        assert.ok(!truthyEvaluated, 'x-truthy is not evaluated');
      };

      let withoutEvaluatingFalsy = callback => {
        falsyEvaluated = false;
        callback();
        assert.ok(!falsyEvaluated, 'x-falsy is not evaluated');
      };

      this.registerHelper('x-truthy', {
        compute() {
          truthyEvaluated = true;
          return 'T';
        }
      });

      this.registerHelper('x-falsy', {
        compute() {
          falsyEvaluated = true;
          return 'F';
        }
      });

      let template = this.wrappedTemplateFor({
        cond: 'cond',
        truthy: '{{x-truthy}}',
        falsy: '{{x-falsy}}'
      });

      withoutEvaluatingFalsy(() => this.render(template, { cond: this.truthyValue }));

      this.assertText('T');

      withoutEvaluatingFalsy(() => this.runTask(() => this.rerender()));

      this.assertText('T');

      withoutEvaluatingTruthy(() => this.runTask(() => (0, _metal.set)(this.context, 'cond', this.falsyValue)));

      this.assertText('F');

      withoutEvaluatingTruthy(() => this.runTask(() => this.rerender()));

      this.assertText('F');

      withoutEvaluatingFalsy(() => this.runTask(() => (0, _metal.set)(this.context, 'cond', this.truthyValue)));

      this.assertText('T');
    }
  }

  exports.TogglingSyntaxConditionalsTest = TogglingSyntaxConditionalsTest;
  class IfUnlessWithSyntaxTest extends TogglingSyntaxConditionalsTest {}

  exports.IfUnlessWithSyntaxTest = IfUnlessWithSyntaxTest;
  (0, _abstractTestCase.applyMixins)(IfUnlessWithSyntaxTest, ...IfUnlessWithTestCases);
});
enifed('@ember/-internals/glimmer/tests/utils/string-test', ['@ember/-internals/glimmer/tests/utils/helpers', '@ember/-internals/glimmer/tests/utils/abstract-test-case', '@ember/-internals/glimmer/tests/utils/test-case'], function (_helpers, _abstractTestCase, _testCase) {
  'use strict';

  (0, _testCase.moduleFor)('SafeString', class extends _abstractTestCase.TestCase {
    ['@test htmlSafe should return an instance of SafeString']() {
      let safeString = (0, _helpers.htmlSafe)('you need to be more <b>bold</b>');

      this.assert.ok(safeString instanceof _helpers.SafeString, 'should be a SafeString');
    }

    ['@test htmlSafe should return an empty string for null']() {
      let safeString = (0, _helpers.htmlSafe)(null);

      this.assert.equal(safeString instanceof _helpers.SafeString, true, 'should be a SafeString');
      this.assert.equal(safeString.toString(), '', 'should return an empty string');
    }

    ['@test htmlSafe should return an instance of SafeString']() {
      let safeString = (0, _helpers.htmlSafe)();

      this.assert.equal(safeString instanceof _helpers.SafeString, true, 'should be a SafeString');
      this.assert.equal(safeString.toString(), '', 'should return an empty string');
    }
  });

  (0, _testCase.moduleFor)('SafeString isHTMLSafe', class extends _abstractTestCase.TestCase {
    ['@test isHTMLSafe should detect SafeString']() {
      let safeString = (0, _helpers.htmlSafe)('<em>Emphasize</em> the important things.');

      this.assert.ok((0, _helpers.isHTMLSafe)(safeString));
    }

    ['@test isHTMLSafe should not detect SafeString on primatives']() {
      this.assert.notOk((0, _helpers.isHTMLSafe)('Hello World'));
      this.assert.notOk((0, _helpers.isHTMLSafe)({}));
      this.assert.notOk((0, _helpers.isHTMLSafe)([]));
      this.assert.notOk((0, _helpers.isHTMLSafe)(10));
      this.assert.notOk((0, _helpers.isHTMLSafe)(null));
    }
  });
});
enifed('@ember/-internals/glimmer/tests/utils/test-case', ['exports', 'internal-test-helpers'], function (exports, _internalTestHelpers) {
  'use strict';

  Object.defineProperty(exports, 'TestCase', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.AbstractTestCase;
    }
  });
  Object.defineProperty(exports, 'ApplicationTest', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.ApplicationTestCase;
    }
  });
  Object.defineProperty(exports, 'RenderingTest', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.RenderingTestCase;
    }
  });
  Object.defineProperty(exports, 'moduleFor', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.moduleFor;
    }
  });
});
enifed('@ember/-internals/glimmer/tests/utils/test-helpers', ['exports', 'internal-test-helpers'], function (exports, _internalTestHelpers) {
  'use strict';

  Object.defineProperty(exports, 'styles', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.styles;
    }
  });
  Object.defineProperty(exports, 'classes', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.classes;
    }
  });
  Object.defineProperty(exports, 'equalTokens', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.equalTokens;
    }
  });
  Object.defineProperty(exports, 'equalsElement', {
    enumerable: true,
    get: function () {
      return _internalTestHelpers.equalsElement;
    }
  });
});
enifed('@ember/-internals/meta/tests/listeners_test', ['internal-test-helpers', '@ember/-internals/meta'], function (_internalTestHelpers, _meta) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.meta listeners', class extends _internalTestHelpers.AbstractTestCase {
    ['@test basics'](assert) {
      let t = {};
      let m = (0, _meta.meta)({});
      m.addToListeners('hello', t, 'm', 0);
      let matching = m.matchingListeners('hello');
      assert.equal(matching.length, 3);
      assert.equal(matching[0], t);
      m.removeFromListeners('hello', t, 'm');
      matching = m.matchingListeners('hello');
      assert.equal(matching, undefined);
    }

    ['@test inheritance'](assert) {
      let target = {};
      let parent = {};
      let parentMeta = (0, _meta.meta)(parent);
      parentMeta.addToListeners('hello', target, 'm', 0);

      let child = Object.create(parent);
      let m = (0, _meta.meta)(child);

      let matching = m.matchingListeners('hello');
      assert.equal(matching.length, 3);
      assert.equal(matching[0], target);
      assert.equal(matching[1], 'm');
      assert.equal(matching[2], 0);
      m.removeFromListeners('hello', target, 'm');
      matching = m.matchingListeners('hello');
      assert.equal(matching, undefined);
      matching = parentMeta.matchingListeners('hello');
      assert.equal(matching.length, 3);
    }

    ['@test deduplication'](assert) {
      let t = {};
      let m = (0, _meta.meta)({});
      m.addToListeners('hello', t, 'm', 0);
      m.addToListeners('hello', t, 'm', 0);
      let matching = m.matchingListeners('hello');
      assert.equal(matching.length, 3);
      assert.equal(matching[0], t);
    }

    ['@test parent caching'](assert) {
      _meta.counters.flattenedListenersCalls = 0;
      _meta.counters.parentListenersUsed = 0;

      class Class {}
      let classMeta = (0, _meta.meta)(Class.prototype);
      classMeta.addToListeners('hello', null, 'm', 0);

      let instance = new Class();
      let m = (0, _meta.meta)(instance);

      let matching = m.matchingListeners('hello');

      assert.equal(matching.length, 3);
      assert.equal(_meta.counters.flattenedListenersCalls, 2);
      assert.equal(_meta.counters.parentListenersUsed, 1);

      matching = m.matchingListeners('hello');

      assert.equal(matching.length, 3);
      assert.equal(_meta.counters.flattenedListenersCalls, 3);
      assert.equal(_meta.counters.parentListenersUsed, 1);
    }

    ['@test parent cache invalidation'](assert) {
      _meta.counters.flattenedListenersCalls = 0;
      _meta.counters.parentListenersUsed = 0;
      _meta.counters.listenersInherited = 0;

      class Class {}
      let classMeta = (0, _meta.meta)(Class.prototype);
      classMeta.addToListeners('hello', null, 'm', 0);

      let instance = new Class();
      let m = (0, _meta.meta)(instance);

      let matching = m.matchingListeners('hello');

      assert.equal(matching.length, 3);
      assert.equal(_meta.counters.flattenedListenersCalls, 2);
      assert.equal(_meta.counters.parentListenersUsed, 1);
      assert.equal(_meta.counters.listenersInherited, 0);

      m.addToListeners('hello', null, 'm2');

      matching = m.matchingListeners('hello');

      assert.equal(matching.length, 6);
      assert.equal(_meta.counters.flattenedListenersCalls, 4);
      assert.equal(_meta.counters.parentListenersUsed, 1);
      assert.equal(_meta.counters.listenersInherited, 1);
    }

    ['@test reopen after flatten'](assert) {
      // Ensure counter is zeroed
      _meta.counters.reopensAfterFlatten = 0;

      class Class1 {}
      let class1Meta = (0, _meta.meta)(Class1.prototype);
      class1Meta.addToListeners('hello', null, 'm', 0);

      let instance1 = new Class1();
      let m1 = (0, _meta.meta)(instance1);

      class Class2 {}
      let class2Meta = (0, _meta.meta)(Class2.prototype);
      class2Meta.addToListeners('hello', null, 'm', 0);

      let instance2 = new Class2();
      let m2 = (0, _meta.meta)(instance2);

      m1.matchingListeners('hello');
      m2.matchingListeners('hello');

      assert.equal(_meta.counters.reopensAfterFlatten, 0, 'no reopen calls yet');

      m1.addToListeners('world', null, 'm', 0);
      m2.addToListeners('world', null, 'm', 0);
      m1.matchingListeners('world');
      m2.matchingListeners('world');

      assert.equal(_meta.counters.reopensAfterFlatten, 1, 'reopen calls after invalidating parent cache');

      m1.addToListeners('world', null, 'm', 0);
      m2.addToListeners('world', null, 'm', 0);
      m1.matchingListeners('world');
      m2.matchingListeners('world');

      assert.equal(_meta.counters.reopensAfterFlatten, 1, 'no reopen calls after mutating leaf nodes');

      class1Meta.removeFromListeners('hello', null, 'm');
      class2Meta.removeFromListeners('hello', null, 'm');
      m1.matchingListeners('hello');
      m2.matchingListeners('hello');

      assert.equal(_meta.counters.reopensAfterFlatten, 2, 'one reopen call after mutating parents');

      class1Meta.addToListeners('hello', null, 'm', 0);
      m1.matchingListeners('hello');
      class2Meta.addToListeners('hello', null, 'm', 0);
      m2.matchingListeners('hello');

      assert.equal(_meta.counters.reopensAfterFlatten, 3, 'one reopen call after mutating parents and flattening out of order');
    }

    ['@test REMOVE_ALL does not interfere with future adds'](assert) {
      expectDeprecation(() => {
        let t = {};
        let m = (0, _meta.meta)({});

        m.addToListeners('hello', t, 'm', 0);
        let matching = m.matchingListeners('hello');

        assert.equal(matching.length, 3);
        assert.equal(matching[0], t);

        // Remove all listeners
        m.removeAllListeners('hello');

        matching = m.matchingListeners('hello');
        assert.equal(matching, undefined);

        m.addToListeners('hello', t, 'm', 0);
        matching = m.matchingListeners('hello');

        // listener was added back successfully
        assert.equal(matching.length, 3);
        assert.equal(matching[0], t);
      });
    }
  });
});
enifed('@ember/-internals/meta/tests/meta_test', ['internal-test-helpers', '@ember/-internals/meta'], function (_internalTestHelpers, _meta) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.meta', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should return the same hash for an object'](assert) {
      let obj = {};

      (0, _meta.meta)(obj).foo = 'bar';

      assert.equal((0, _meta.meta)(obj).foo, 'bar', 'returns same hash with multiple calls to Ember.meta()');
    }

    ['@test meta is not enumerable'](assert) {
      let proto, obj, props, prop;
      proto = { foo: 'bar' };
      (0, _meta.meta)(proto);
      obj = Object.create(proto);
      (0, _meta.meta)(obj);
      obj.bar = 'baz';
      props = [];
      for (prop in obj) {
        props.push(prop);
      }
      assert.deepEqual(props.sort(), ['bar', 'foo']);
      if (typeof JSON !== 'undefined' && 'stringify' in JSON) {
        try {
          JSON.stringify(obj);
        } catch (e) {
          assert.ok(false, 'meta should not fail JSON.stringify');
        }
      }
    }

    ['@test meta.writeWatching issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writeWatching('hello', 1);
      }, 'Cannot update watchers for `hello` on `<special-sauce:123>` after it has been destroyed.');
    }

    ['@test meta.writableTag issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writableTag(() => {});
      }, 'Cannot create a new tag for `<special-sauce:123>` after it has been destroyed.');
    }

    ['@test meta.writableChainWatchers issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writableChainWatchers(() => {});
      }, 'Cannot create a new chain watcher for `<special-sauce:123>` after it has been destroyed.');
    }

    ['@test meta.writableChains issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writableChains(() => {});
      }, 'Cannot create a new chains for `<special-sauce:123>` after it has been destroyed.');
    }

    ['@test meta.writeValues issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writeValues('derp', 'ohai');
      }, 'Cannot set the value of `derp` on `<special-sauce:123>` after it has been destroyed.');
    }

    ['@test meta.writeDeps issues useful error after destroy']() {
      let target = {
        toString() {
          return '<special-sauce:123>';
        }
      };
      let targetMeta = (0, _meta.meta)(target);

      targetMeta.destroy();

      expectAssertion(() => {
        targetMeta.writeDeps('derp', 'ohai', 1);
      }, 'Cannot modify dependent keys for `ohai` on `<special-sauce:123>` after it has been destroyed.');
    }
  });
});
enifed('@ember/-internals/metal/tests/accessors/get_path_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  let obj;

  (0, _internalTestHelpers.moduleFor)('get with path', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      obj = {
        foo: {
          bar: {
            baz: { biff: 'BIFF' }
          }
        },
        foothis: {
          bar: {
            baz: { biff: 'BIFF' }
          }
        },
        falseValue: false,
        emptyString: '',
        Wuz: {
          nar: 'foo'
        },
        nullValue: null
      };
    }

    teardown() {
      obj = undefined;
    }

    // ..........................................................
    // LOCAL PATHS
    //
    ['@test [obj, foo] -> obj.foo'](assert) {
      assert.deepEqual((0, _metal.get)(obj, 'foo'), obj.foo);
    }

    ['@test [obj, foo.bar] -> obj.foo.bar'](assert) {
      assert.deepEqual((0, _metal.get)(obj, 'foo.bar'), obj.foo.bar);
    }

    ['@test [obj, foothis.bar] -> obj.foothis.bar'](assert) {
      assert.deepEqual((0, _metal.get)(obj, 'foothis.bar'), obj.foothis.bar);
    }

    ['@test [obj, falseValue.notDefined] -> (undefined)'](assert) {
      assert.strictEqual((0, _metal.get)(obj, 'falseValue.notDefined'), undefined);
    }

    ['@test [obj, emptyString.length] -> 0'](assert) {
      assert.strictEqual((0, _metal.get)(obj, 'emptyString.length'), 0);
    }

    ['@test [obj, nullValue.notDefined] -> (undefined)'](assert) {
      assert.strictEqual((0, _metal.get)(obj, 'nullValue.notDefined'), undefined);
    }

    // ..........................................................
    // GLOBAL PATHS TREATED LOCAL WITH GET
    //

    ['@test [obj, Wuz] -> obj.Wuz'](assert) {
      assert.deepEqual((0, _metal.get)(obj, 'Wuz'), obj.Wuz);
    }

    ['@test [obj, Wuz.nar] -> obj.Wuz.nar'](assert) {
      assert.deepEqual((0, _metal.get)(obj, 'Wuz.nar'), obj.Wuz.nar);
    }

    ['@test [obj, Foo] -> (undefined)'](assert) {
      assert.strictEqual((0, _metal.get)(obj, 'Foo'), undefined);
    }

    ['@test [obj, Foo.bar] -> (undefined)'](assert) {
      assert.strictEqual((0, _metal.get)(obj, 'Foo.bar'), undefined);
    }
  });
});
enifed('@ember/-internals/metal/tests/accessors/get_properties_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('getProperties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test can retrieve a hash of properties from an object via an argument list or array of property names'](assert) {
      let obj = {
        firstName: 'Steve',
        lastName: 'Jobs',
        companyName: 'Apple, Inc.'
      };

      assert.deepEqual((0, _metal.getProperties)(obj, 'firstName', 'lastName'), {
        firstName: 'Steve',
        lastName: 'Jobs'
      });
      assert.deepEqual((0, _metal.getProperties)(obj, 'firstName', 'lastName'), {
        firstName: 'Steve',
        lastName: 'Jobs'
      });
      assert.deepEqual((0, _metal.getProperties)(obj, 'lastName'), { lastName: 'Jobs' });
      assert.deepEqual((0, _metal.getProperties)(obj), {});
      assert.deepEqual((0, _metal.getProperties)(obj, ['firstName', 'lastName']), {
        firstName: 'Steve',
        lastName: 'Jobs'
      });
      assert.deepEqual((0, _metal.getProperties)(obj, ['firstName']), {
        firstName: 'Steve'
      });
      assert.deepEqual((0, _metal.getProperties)(obj, []), {});
    }
  });
});
enifed('@ember/-internals/metal/tests/accessors/get_test', ['@ember/-internals/environment', '@ember/-internals/runtime', '@ember/-internals/metal', 'internal-test-helpers', '@ember/runloop'], function (_environment, _runtime, _metal, _internalTestHelpers, _runloop) {
  'use strict';

  function aget(x, y) {
    return x[y];
  }

  (0, _internalTestHelpers.moduleFor)('get', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should get arbitrary properties on an object'](assert) {
      let obj = {
        string: 'string',
        number: 23,
        boolTrue: true,
        boolFalse: false,
        nullValue: null
      };

      for (let key in obj) {
        if (!obj.hasOwnProperty(key)) {
          continue;
        }
        assert.equal((0, _metal.get)(obj, key), obj[key], key);
      }
    }

    ['@test should retrieve a number key on an object'](assert) {
      let obj = { 1: 'first' };

      assert.equal((0, _metal.get)(obj, 1), 'first');
    }

    ['@test should retrieve an array index'](assert) {
      let arr = ['first', 'second'];

      assert.equal((0, _metal.get)(arr, 0), 'first');
      assert.equal((0, _metal.get)(arr, 1), 'second');
    }

    ['@test should retrieve an empty string key on an object'](assert) {
      let obj = { '': 'empty-string' };

      assert.equal((0, _metal.get)(obj, ''), 'empty-string');
    }

    ['@test should return undefined when passed an empty string if that key does not exist on an object'](assert) {
      let obj = { tomster: true };

      assert.equal((0, _metal.get)(obj, ''), undefined);
    }

    ['@test should not access a property more than once'](assert) {
      let count = 0;
      let obj = {
        get id() {
          return ++count;
        }
      };

      (0, _metal.get)(obj, 'id');

      assert.equal(count, 1);
    }

    ['@test should call unknownProperty on watched values if the value is undefined using getFromEmberMetal()/set()'](assert) {
      let obj = {
        unknownProperty(key) {
          assert.equal(key, 'foo', 'should pass key');
          return 'FOO';
        }
      };
      assert.equal((0, _metal.get)(obj, 'foo'), 'FOO', 'should return value from unknown');
    }

    ['@test should call unknownProperty on watched values if the value is undefined using accessors'](assert) {
      if (_environment.ENV.USES_ACCESSORS) {
        let obj = {
          unknownProperty(key) {
            assert.equal(key, 'foo', 'should pass key');
            return 'FOO';
          }
        };
        assert.equal(aget(obj, 'foo'), 'FOO', 'should return value from unknown');
      } else {
        assert.ok('SKIPPING ACCESSORS');
      }
    }

    ['@test get works with paths correctly'](assert) {
      let func = function () {};
      func.bar = 'awesome';

      let destroyedObj = _runtime.Object.create({ bar: 'great' });
      (0, _runloop.run)(() => destroyedObj.destroy());

      assert.equal((0, _metal.get)({ foo: null }, 'foo.bar'), undefined);
      assert.equal((0, _metal.get)({ foo: { bar: 'hello' } }, 'foo.bar.length'), 5);
      assert.equal((0, _metal.get)({ foo: func }, 'foo.bar'), 'awesome');
      assert.equal((0, _metal.get)({ foo: func }, 'foo.bar.length'), 7);
      assert.equal((0, _metal.get)({}, 'foo.bar.length'), undefined);
      assert.equal((0, _metal.get)(function () {}, 'foo.bar.length'), undefined);
      assert.equal((0, _metal.get)('', 'foo.bar.length'), undefined);
      assert.equal((0, _metal.get)({ foo: destroyedObj }, 'foo.bar'), undefined);
    }

    ['@test warn on attempts to call get with no arguments']() {
      expectAssertion(function () {
        (0, _metal.get)('aProperty');
      }, /Get must be called with two arguments;/i);
    }

    ['@test warn on attempts to call get with only one argument']() {
      expectAssertion(function () {
        (0, _metal.get)('aProperty');
      }, /Get must be called with two arguments;/i);
    }

    ['@test warn on attempts to call get with more then two arguments']() {
      expectAssertion(function () {
        (0, _metal.get)({}, 'aProperty', true);
      }, /Get must be called with two arguments;/i);
    }

    ['@test warn on attempts to get a property of undefined']() {
      expectAssertion(function () {
        (0, _metal.get)(undefined, 'aProperty');
      }, /Cannot call get with 'aProperty' on an undefined object/i);
    }

    ['@test warn on attempts to get a property path of undefined']() {
      expectAssertion(function () {
        (0, _metal.get)(undefined, 'aProperty.on.aPath');
      }, /Cannot call get with 'aProperty.on.aPath' on an undefined object/);
    }

    ['@test warn on attempts to get a property of null']() {
      expectAssertion(function () {
        (0, _metal.get)(null, 'aProperty');
      }, /Cannot call get with 'aProperty' on an undefined object/);
    }

    ['@test warn on attempts to get a property path of null']() {
      expectAssertion(function () {
        (0, _metal.get)(null, 'aProperty.on.aPath');
      }, /Cannot call get with 'aProperty.on.aPath' on an undefined object/);
    }

    ['@test warn on attempts to use get with an unsupported property path']() {
      let obj = {};
      expectAssertion(() => (0, _metal.get)(obj, null), /The key provided to get must be a string or number, you passed null/);
      expectAssertion(() => (0, _metal.get)(obj, NaN), /The key provided to get must be a string or number, you passed NaN/);
      expectAssertion(() => (0, _metal.get)(obj, undefined), /The key provided to get must be a string or number, you passed undefined/);
      expectAssertion(() => (0, _metal.get)(obj, false), /The key provided to get must be a string or number, you passed false/);
    }

    // ..........................................................
    // BUGS
    //

    ['@test (regression) watched properties on unmodified inherited objects should still return their original value'](assert) {
      let MyMixin = _metal.Mixin.create({
        someProperty: 'foo',
        propertyDidChange: (0, _metal.observer)('someProperty', () => {})
      });

      let baseObject = MyMixin.apply({});
      let theRealObject = Object.create(baseObject);

      assert.equal((0, _metal.get)(theRealObject, 'someProperty'), 'foo', 'should return the set value, not false');
    }
  });

  (0, _internalTestHelpers.moduleFor)('getWithDefault', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should get arbitrary properties on an object'](assert) {
      let obj = {
        string: 'string',
        number: 23,
        boolTrue: true,
        boolFalse: false,
        nullValue: null
      };

      for (let key in obj) {
        if (!obj.hasOwnProperty(key)) {
          continue;
        }
        assert.equal((0, _metal.getWithDefault)(obj, key, 'fail'), obj[key], key);
      }

      obj = {
        undef: undefined
      };

      assert.equal((0, _metal.getWithDefault)(obj, 'undef', 'default'), 'default', 'explicit undefined retrieves the default');
      assert.equal((0, _metal.getWithDefault)(obj, 'not-present', 'default'), 'default', 'non-present key retrieves the default');
    }

    ['@test should call unknownProperty if defined and value is undefined'](assert) {
      let obj = {
        count: 0,
        unknownProperty(key) {
          assert.equal(key, 'foo', 'should pass key');
          this.count++;
          return 'FOO';
        }
      };

      assert.equal((0, _metal.get)(obj, 'foo'), 'FOO', 'should return value from unknown');
      assert.equal(obj.count, 1, 'should have invoked');
    }

    ['@test if unknownProperty is present, it is called using getFromEmberMetal()/set()'](assert) {
      let obj = {
        unknownProperty(key) {
          if (key === 'foo') {
            assert.equal(key, 'foo', 'should pass key');
            return 'FOO';
          }
        }
      };
      assert.equal((0, _metal.getWithDefault)(obj, 'foo', 'fail'), 'FOO', 'should return value from unknownProperty');
      assert.equal((0, _metal.getWithDefault)(obj, 'bar', 'default'), 'default', 'should convert undefined from unknownProperty into default');
    }

    ['@test if unknownProperty is present, it is called using accessors'](assert) {
      if (_environment.ENV.USES_ACCESSORS) {
        let obj = {
          unknownProperty(key) {
            if (key === 'foo') {
              assert.equal(key, 'foo', 'should pass key');
              return 'FOO';
            }
          }
        };
        assert.equal(aget(obj, 'foo', 'fail'), 'FOO', 'should return value from unknownProperty');
        assert.equal(aget(obj, 'bar', 'default'), 'default', 'should convert undefined from unknownProperty into default');
      } else {
        assert.ok('SKIPPING ACCESSORS');
      }
    }

    // ..........................................................
    // BUGS
    //

    ['@test (regression) watched properties on unmodified inherited objects should still return their original value'](assert) {
      let MyMixin = _metal.Mixin.create({
        someProperty: 'foo',
        propertyDidChange: (0, _metal.observer)('someProperty', () => {
          /* nothing to do */
        })
      });

      let baseObject = MyMixin.apply({});
      let theRealObject = Object.create(baseObject);

      assert.equal((0, _metal.getWithDefault)(theRealObject, 'someProperty', 'fail'), 'foo', 'should return the set value, not false');
    }
  });
});
enifed('@ember/-internals/metal/tests/accessors/mandatory_setters_test', ['@ember/-internals/metal', '@ember/-internals/meta', 'internal-test-helpers'], function (_metal, _meta, _internalTestHelpers) {
  'use strict';

  function hasMandatorySetter(object, property) {
    try {
      return Object.getOwnPropertyDescriptor(object, property).set.isMandatorySetter === true;
    } catch (e) {
      return false;
    }
  }

  function hasMetaValue(object, property) {
    return (0, _meta.meta)(object).peekValues(property) !== undefined;
  }

  if (true /* DEBUG */) {
      (0, _internalTestHelpers.moduleFor)('mandory-setters', class extends _internalTestHelpers.AbstractTestCase {
        ['@test does not assert if property is not being watched'](assert) {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          obj.someProp = 'blastix';
          assert.equal((0, _metal.get)(obj, 'someProp'), 'blastix');
        }

        ['@test should not setup mandatory-setter if property is not writable'](assert) {
          assert.expect(6);

          let obj = {};

          Object.defineProperty(obj, 'a', { value: true });
          Object.defineProperty(obj, 'b', { value: false });
          Object.defineProperty(obj, 'c', { value: undefined });
          Object.defineProperty(obj, 'd', { value: undefined, writable: false });
          Object.defineProperty(obj, 'e', {
            value: undefined,
            configurable: false
          });
          Object.defineProperty(obj, 'f', {
            value: undefined,
            configurable: true
          });

          (0, _metal.watch)(obj, 'a');
          (0, _metal.watch)(obj, 'b');
          (0, _metal.watch)(obj, 'c');
          (0, _metal.watch)(obj, 'd');
          (0, _metal.watch)(obj, 'e');
          (0, _metal.watch)(obj, 'f');

          assert.ok(!hasMandatorySetter(obj, 'a'), 'mandatory-setter should not be installed');
          assert.ok(!hasMandatorySetter(obj, 'b'), 'mandatory-setter should not be installed');
          assert.ok(!hasMandatorySetter(obj, 'c'), 'mandatory-setter should not be installed');
          assert.ok(!hasMandatorySetter(obj, 'd'), 'mandatory-setter should not be installed');
          assert.ok(!hasMandatorySetter(obj, 'e'), 'mandatory-setter should not be installed');
          assert.ok(!hasMandatorySetter(obj, 'f'), 'mandatory-setter should not be installed');
        }

        ['@test should not teardown non mandatory-setter descriptor'](assert) {
          assert.expect(1);

          let obj = {
            get a() {
              return 'hi';
            }
          };

          (0, _metal.watch)(obj, 'a');
          (0, _metal.unwatch)(obj, 'a');

          assert.equal(obj.a, 'hi');
        }

        ['@test should not confuse non descriptor watched gets'](assert) {
          assert.expect(2);

          let obj = {
            get a() {
              return 'hi';
            }
          };

          (0, _metal.watch)(obj, 'a');
          assert.equal((0, _metal.get)(obj, 'a'), 'hi');
          assert.equal(obj.a, 'hi');
        }

        ['@test should not setup mandatory-setter if setter is already setup on property'](assert) {
          assert.expect(2);

          let obj = { someProp: null };

          Object.defineProperty(obj, 'someProp', {
            get() {
              return null;
            },

            set(value) {
              assert.equal(value, 'foo-bar', 'custom setter was called');
            }
          });

          (0, _metal.watch)(obj, 'someProp');
          assert.ok(!hasMandatorySetter(obj, 'someProp'), 'mandatory-setter should not be installed');

          obj.someProp = 'foo-bar';
        }

        ['@test watched ES5 setter should not be smashed by mandatory setter'](assert) {
          let value;
          let obj = {
            get foo() {},
            set foo(_value) {
              value = _value;
            }
          };

          (0, _metal.watch)(obj, 'foo');

          (0, _metal.set)(obj, 'foo', 2);
          assert.equal(value, 2);
        }

        ['@test should not setup mandatory-setter if setter is already setup on property in parent prototype'](assert) {
          assert.expect(2);

          function Foo() {}

          Object.defineProperty(Foo.prototype, 'someProp', {
            get() {
              return null;
            },

            set(value) {
              assert.equal(value, 'foo-bar', 'custom setter was called');
            }
          });

          let obj = new Foo();

          (0, _metal.watch)(obj, 'someProp');
          assert.ok(!hasMandatorySetter(obj, 'someProp'), 'mandatory-setter should not be installed');

          obj.someProp = 'foo-bar';
        }

        ['@test should not setup mandatory-setter if setter is already setup on property in grandparent prototype'](assert) {
          assert.expect(2);

          function Foo() {}

          Object.defineProperty(Foo.prototype, 'someProp', {
            get() {
              return null;
            },

            set(value) {
              assert.equal(value, 'foo-bar', 'custom setter was called');
            }
          });

          function Bar() {}
          Bar.prototype = Object.create(Foo.prototype);
          Bar.prototype.constructor = Bar;

          let obj = new Bar();

          (0, _metal.watch)(obj, 'someProp');
          assert.ok(!hasMandatorySetter(obj, 'someProp'), 'mandatory-setter should not be installed');

          obj.someProp = 'foo-bar';
        }

        ['@test should not setup mandatory-setter if setter is already setup on property in great grandparent prototype'](assert) {
          assert.expect(2);

          function Foo() {}

          Object.defineProperty(Foo.prototype, 'someProp', {
            get() {
              return null;
            },

            set(value) {
              assert.equal(value, 'foo-bar', 'custom setter was called');
            }
          });

          function Bar() {}
          Bar.prototype = Object.create(Foo.prototype);
          Bar.prototype.constructor = Bar;

          function Qux() {}
          Qux.prototype = Object.create(Bar.prototype);
          Qux.prototype.constructor = Qux;

          let obj = new Qux();

          (0, _metal.watch)(obj, 'someProp');
          assert.ok(!hasMandatorySetter(obj, 'someProp'), 'mandatory-setter should not be installed');

          obj.someProp = 'foo-bar';
        }

        ['@test should assert if set without set when property is being watched']() {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          (0, _metal.watch)(obj, 'someProp');

          expectAssertion(function () {
            obj.someProp = 'foo-bar';
          }, 'You must use set() to set the `someProp` property (of custom-object) to `foo-bar`.');
        }

        ['@test should not assert if set with set when property is being watched'](assert) {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          (0, _metal.watch)(obj, 'someProp');
          (0, _metal.set)(obj, 'someProp', 'foo-bar');

          assert.equal((0, _metal.get)(obj, 'someProp'), 'foo-bar');
        }

        ['@test does not setup mandatory-setter if non-configurable'](assert) {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          Object.defineProperty(obj, 'someProp', {
            configurable: false,
            enumerable: true,
            value: 'blastix'
          });

          (0, _metal.watch)(obj, 'someProp');
          assert.ok(!hasMandatorySetter(obj, 'someProp'), 'blastix');
        }

        ['@test ensure after watch the property is restored (and the value is no-longer stored in meta) [non-enumerable]'](assert) {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          Object.defineProperty(obj, 'someProp', {
            configurable: true,
            enumerable: false,
            value: 'blastix'
          });

          (0, _metal.watch)(obj, 'someProp');
          assert.equal(hasMandatorySetter(obj, 'someProp'), true, 'should have a mandatory setter');

          let descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(descriptor.enumerable, false, 'property should remain non-enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(obj.someProp, 'blastix', 'expected value to be the getter');

          assert.equal(descriptor.value, undefined, 'expected existing value to NOT remain');

          assert.ok(hasMetaValue(obj, 'someProp'), 'someProp is stored in meta.values');

          (0, _metal.unwatch)(obj, 'someProp');

          assert.ok(!hasMetaValue(obj, 'someProp'), 'someProp is no longer stored in meta.values');

          descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(hasMandatorySetter(obj, 'someProp'), false, 'should no longer have a mandatory setter');

          assert.equal(descriptor.enumerable, false, 'property should remain non-enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(obj.someProp, 'blastix', 'expected value to be the getter');
          assert.equal(descriptor.value, 'blastix', 'expected existing value to remain');

          obj.someProp = 'new value';

          // make sure the descriptor remains correct (nothing funky, like a redefined, happened in the setter);
          descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(descriptor.enumerable, false, 'property should remain non-enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(descriptor.value, 'new value', 'expected existing value to NOT remain');
          assert.equal(obj.someProp, 'new value', 'expected value to be the getter');
          assert.equal(obj.someProp, 'new value');
        }

        ['@test ensure after watch the property is restored (and the value is no-longer stored in meta) [enumerable]'](assert) {
          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          Object.defineProperty(obj, 'someProp', {
            configurable: true,
            enumerable: true,
            value: 'blastix'
          });

          (0, _metal.watch)(obj, 'someProp');
          assert.equal(hasMandatorySetter(obj, 'someProp'), true, 'should have a mandatory setter');

          let descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(descriptor.enumerable, true, 'property should remain enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(obj.someProp, 'blastix', 'expected value to be the getter');

          assert.equal(descriptor.value, undefined, 'expected existing value to NOT remain');

          assert.ok(hasMetaValue(obj, 'someProp'), 'someProp is stored in meta.values');

          (0, _metal.unwatch)(obj, 'someProp');

          assert.ok(!hasMetaValue(obj, 'someProp'), 'someProp is no longer stored in meta.values');

          descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(hasMandatorySetter(obj, 'someProp'), false, 'should no longer have a mandatory setter');

          assert.equal(descriptor.enumerable, true, 'property should remain enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(obj.someProp, 'blastix', 'expected value to be the getter');
          assert.equal(descriptor.value, 'blastix', 'expected existing value to remain');

          obj.someProp = 'new value';

          // make sure the descriptor remains correct (nothing funky, like a redefined, happened in the setter);
          descriptor = Object.getOwnPropertyDescriptor(obj, 'someProp');

          assert.equal(descriptor.enumerable, true, 'property should remain enumerable');
          assert.equal(descriptor.configurable, true, 'property should remain configurable');
          assert.equal(descriptor.value, 'new value', 'expected existing value to NOT remain');
          assert.equal(obj.someProp, 'new value');
        }

        ['@test sets up mandatory-setter if property comes from prototype'](assert) {
          assert.expect(2);

          let obj = {
            someProp: null,
            toString() {
              return 'custom-object';
            }
          };

          let obj2 = Object.create(obj);

          (0, _metal.watch)(obj2, 'someProp');

          assert.ok(hasMandatorySetter(obj2, 'someProp'), 'mandatory setter has been setup');

          expectAssertion(function () {
            obj2.someProp = 'foo-bar';
          }, 'You must use set() to set the `someProp` property (of custom-object) to `foo-bar`.');
        }

        ['@test inheritance remains live'](assert) {
          function Parent() {}
          Parent.prototype.food = 'chips';

          let child = new Parent();

          assert.equal(child.food, 'chips');

          (0, _metal.watch)(child, 'food');

          assert.equal(child.food, 'chips');

          Parent.prototype.food = 'icecreame';

          assert.equal(child.food, 'icecreame');

          (0, _metal.unwatch)(child, 'food');

          assert.equal(child.food, 'icecreame');

          Parent.prototype.food = 'chips';

          assert.equal(child.food, 'chips');
        }

        ['@test inheritance remains live and preserves this'](assert) {
          function Parent(food) {
            this._food = food;
          }

          Object.defineProperty(Parent.prototype, 'food', {
            get() {
              return this._food;
            }
          });

          let child = new Parent('chips');

          assert.equal(child.food, 'chips');

          (0, _metal.watch)(child, 'food');

          assert.equal(child.food, 'chips');

          child._food = 'icecreame';

          assert.equal(child.food, 'icecreame');

          (0, _metal.unwatch)(child, 'food');

          assert.equal(child.food, 'icecreame');

          let foodDesc = Object.getOwnPropertyDescriptor(Parent.prototype, 'food');
          assert.ok(!foodDesc.configurable, 'Parent.prototype.food desc should be non configable');
          assert.ok(!foodDesc.enumerable, 'Parent.prototype.food desc should be non enumerable');

          assert.equal(foodDesc.get.call({
            _food: 'hi'
          }), 'hi');
          assert.equal(foodDesc.set, undefined);

          assert.equal(child.food, 'icecreame');
        }
      });
    }
});
enifed('@ember/-internals/metal/tests/accessors/set_path_test', ['@ember/-internals/environment', '@ember/-internals/metal', 'internal-test-helpers'], function (_environment, _metal, _internalTestHelpers) {
  'use strict';

  let originalLookup = _environment.context.lookup;
  let lookup;

  let obj;
  function commonSetup() {
    _environment.context.lookup = lookup = {};
    obj = {
      foo: {
        bar: {
          baz: { biff: 'BIFF' }
        }
      }
    };
  }

  function commonTeardown() {
    _environment.context.lookup = originalLookup;
    obj = null;
  }

  (0, _internalTestHelpers.moduleFor)('set with path', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      commonSetup();
    }

    teardown() {
      commonTeardown();
    }

    ['@test [Foo, bar] -> Foo.bar'](assert) {
      lookup.Foo = {
        toString() {
          return 'Foo';
        }
      }; // Behave like an Ember.Namespace

      (0, _metal.set)(lookup.Foo, 'bar', 'baz');
      assert.equal((0, _metal.get)(lookup.Foo, 'bar'), 'baz');
    }

    // ..........................................................
    //
    // LOCAL PATHS

    ['@test [obj, foo] -> obj.foo'](assert) {
      (0, _metal.set)(obj, 'foo', 'BAM');
      assert.equal((0, _metal.get)(obj, 'foo'), 'BAM');
    }

    ['@test [obj, foo.bar] -> obj.foo.bar'](assert) {
      (0, _metal.set)(obj, 'foo.bar', 'BAM');
      assert.equal((0, _metal.get)(obj, 'foo.bar'), 'BAM');
    }
  });

  (0, _internalTestHelpers.moduleFor)('set with path - deprecated', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      commonSetup();
    }

    teardown() {
      commonTeardown();
    }

    // ..........................................................
    // DEPRECATED
    //
    ['@test [obj, bla.bla] gives a proper exception message'](assert) {
      let exceptionMessage = 'Property set failed: object in path "bla" could not be found.';
      try {
        (0, _metal.set)(obj, 'bla.bla', 'BAM');
      } catch (ex) {
        assert.equal(ex.message, exceptionMessage);
      }
    }

    ['@test [obj, foo.baz.bat] -> EXCEPTION'](assert) {
      assert.throws(() => (0, _metal.set)(obj, 'foo.baz.bat', 'BAM'));
    }

    ['@test [obj, foo.baz.bat] -> EXCEPTION'](assert) {
      (0, _metal.trySet)(obj, 'foo.baz.bat', 'BAM');
      assert.ok(true, 'does not raise');
    }
  });
});
enifed('@ember/-internals/metal/tests/accessors/set_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('set', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      (0, _metal.setHasViews)(() => false);
    }

    ['@test should set arbitrary properties on an object'](assert) {
      let obj = {
        string: 'string',
        number: 23,
        boolTrue: true,
        boolFalse: false,
        nullValue: null,
        undefinedValue: undefined
      };

      let newObj = {};

      for (let key in obj) {
        if (!obj.hasOwnProperty(key)) {
          continue;
        }

        assert.equal((0, _metal.set)(newObj, key, obj[key]), obj[key], 'should return value');
        assert.ok(key in newObj, 'should have key');
        assert.ok(newObj.hasOwnProperty(key), 'should have key');
        assert.equal((0, _metal.get)(newObj, key), obj[key], 'should set value');
      }
    }

    ['@test should set a number key on an object'](assert) {
      let obj = {};

      (0, _metal.set)(obj, 1, 'first');
      assert.equal(obj[1], 'first');
    }

    ['@test should set an array index'](assert) {
      let arr = ['first', 'second'];

      (0, _metal.set)(arr, 1, 'lol');
      assert.deepEqual(arr, ['first', 'lol']);
    }

    ['@test should call setUnknownProperty if defined and value is undefined'](assert) {
      let obj = {
        count: 0,

        unknownProperty() {
          assert.ok(false, 'should not invoke unknownProperty if setUnknownProperty is defined');
        },

        setUnknownProperty(key, value) {
          assert.equal(key, 'foo', 'should pass key');
          assert.equal(value, 'BAR', 'should pass key');
          this.count++;
          return 'FOO';
        }
      };

      assert.equal((0, _metal.set)(obj, 'foo', 'BAR'), 'BAR', 'should return set value');
      assert.equal(obj.count, 1, 'should have invoked');
    }

    ['@test warn on attempts to call set with undefined as object']() {
      expectAssertion(() => (0, _metal.set)(undefined, 'aProperty', 'BAM'), /Cannot call set with 'aProperty' on an undefined object./);
    }

    ['@test warn on attempts to call set with null as object']() {
      expectAssertion(() => (0, _metal.set)(null, 'aProperty', 'BAM'), /Cannot call set with 'aProperty' on an undefined object./);
    }

    ['@test warn on attempts to use set with an unsupported property path']() {
      let obj = {};
      expectAssertion(() => (0, _metal.set)(obj, null, 42), /The key provided to set must be a string or number, you passed null/);
      expectAssertion(() => (0, _metal.set)(obj, NaN, 42), /The key provided to set must be a string or number, you passed NaN/);
      expectAssertion(() => (0, _metal.set)(obj, undefined, 42), /The key provided to set must be a string or number, you passed undefined/);
      expectAssertion(() => (0, _metal.set)(obj, false, 42), /The key provided to set must be a string or number, you passed false/);
    }

    ['@test warn on attempts of calling set on a destroyed object']() {
      let obj = { isDestroyed: true };

      expectAssertion(() => (0, _metal.set)(obj, 'favoriteFood', 'hot dogs'), 'calling set on destroyed object: [object Object].favoriteFood = hot dogs');
    }

    ['@test does not trigger auto-run assertion for objects that have not been tagged'](assert) {
      (0, _metal.setHasViews)(() => true);
      let obj = {};

      (0, _metal.set)(obj, 'foo', 'bar');

      assert.equal(obj.foo, 'bar');
    }

    ['@test does not warn on attempts of calling set on a destroyed object with `trySet`'](assert) {
      let obj = { isDestroyed: true };

      (0, _metal.trySet)(obj, 'favoriteFood', 'hot dogs');
      assert.equal(obj.favoriteFood, undefined, 'does not set and does not error');
    }
  });
});
enifed('@ember/-internals/metal/tests/alias_test', ['@ember/-internals/metal', '@ember/-internals/meta', 'internal-test-helpers'], function (_metal, _meta, _internalTestHelpers) {
  'use strict';

  let obj, count;

  function incrementCount() {
    count++;
  }

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/metal/alias', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = { foo: { faz: 'FOO' } };
      count = 0;
    }

    afterEach() {
      obj = null;
    }

    ['@test should proxy get to alt key'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));
      assert.equal((0, _metal.get)(obj, 'bar'), 'FOO');
    }

    ['@test should proxy set to alt key'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));
      (0, _metal.set)(obj, 'bar', 'BAR');
      assert.equal((0, _metal.get)(obj, 'foo.faz'), 'BAR');
    }

    ['@test old dependent keys should not trigger property changes'](assert) {
      let obj1 = Object.create(null);
      (0, _metal.defineProperty)(obj1, 'foo', null, null);
      (0, _metal.defineProperty)(obj1, 'bar', (0, _metal.alias)('foo'));
      (0, _metal.defineProperty)(obj1, 'baz', (0, _metal.alias)('foo'));
      (0, _metal.defineProperty)(obj1, 'baz', (0, _metal.alias)('bar')); // redefine baz
      (0, _metal.addObserver)(obj1, 'baz', incrementCount);

      (0, _metal.set)(obj1, 'foo', 'FOO');
      assert.equal(count, 1);

      (0, _metal.removeObserver)(obj1, 'baz', incrementCount);

      (0, _metal.set)(obj1, 'foo', 'OOF');
      assert.equal(count, 1);
    }

    [`@test inheriting an observer of the alias from the prototype then
    redefining the alias on the instance to another property dependent on same key
    does not call the observer twice`](assert) {
      let obj1 = Object.create(null);
      obj1.incrementCount = incrementCount;

      (0, _meta.meta)(obj1).proto = obj1;

      (0, _metal.defineProperty)(obj1, 'foo', null, null);
      (0, _metal.defineProperty)(obj1, 'bar', (0, _metal.alias)('foo'));
      (0, _metal.defineProperty)(obj1, 'baz', (0, _metal.alias)('foo'));
      (0, _metal.addObserver)(obj1, 'baz', null, 'incrementCount');

      let obj2 = Object.create(obj1);
      (0, _metal.defineProperty)(obj2, 'baz', (0, _metal.alias)('bar')); // override baz

      (0, _metal.set)(obj2, 'foo', 'FOO');
      assert.equal(count, 1);

      (0, _metal.removeObserver)(obj2, 'baz', null, 'incrementCount');

      (0, _metal.set)(obj2, 'foo', 'OOF');
      assert.equal(count, 1);
    }

    ['@test an observer of the alias works if added after defining the alias'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));
      (0, _metal.addObserver)(obj, 'bar', incrementCount);
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));
      (0, _metal.set)(obj, 'foo.faz', 'BAR');
      assert.equal(count, 1);
    }

    ['@test an observer of the alias works if added before defining the alias'](assert) {
      (0, _metal.addObserver)(obj, 'bar', incrementCount);
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));
      (0, _metal.set)(obj, 'foo.faz', 'BAR');
      assert.equal(count, 1);
    }

    ['@test object with alias is dirtied if interior object of alias is set after consumption'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));
      (0, _metal.get)(obj, 'bar');

      let tag = (0, _metal.tagFor)(obj);
      let tagValue = tag.value();
      (0, _metal.set)(obj, 'foo.faz', 'BAR');

      assert.ok(!tag.validate(tagValue), 'setting the aliased key should dirty the object');
    }

    ['@test setting alias on self should fail assertion']() {
      expectAssertion(() => (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('bar')), "Setting alias 'bar' on self");
    }

    ['@test destroyed alias does not disturb watch count'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));

      assert.equal((0, _metal.get)(obj, 'bar'), 'FOO');
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));

      (0, _metal.defineProperty)(obj, 'bar', null);

      assert.notOk((0, _metal.isWatching)(obj, 'foo.faz'));
    }

    ['@test setting on oneWay alias does not disturb watch count'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz').oneWay());

      assert.equal((0, _metal.get)(obj, 'bar'), 'FOO');
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));

      (0, _metal.set)(obj, 'bar', null);

      assert.notOk((0, _metal.isWatching)(obj, 'foo.faz'));
    }

    ['@test redefined alias with observer does not disturb watch count'](assert) {
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz').oneWay());

      assert.equal((0, _metal.get)(obj, 'bar'), 'FOO');
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));

      (0, _metal.addObserver)(obj, 'bar', incrementCount);

      assert.equal(count, 0);

      (0, _metal.set)(obj, 'bar', null);

      assert.equal(count, 1);
      assert.notOk((0, _metal.isWatching)(obj, 'foo.faz'));

      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.alias)('foo.faz'));

      assert.equal(count, 1);
      assert.ok((0, _metal.isWatching)(obj, 'foo.faz'));

      (0, _metal.set)(obj, 'foo.faz', 'great');

      assert.equal(count, 2);
    }
  });
});
enifed('@ember/-internals/metal/tests/chains_test', ['@ember/-internals/metal', '@ember/-internals/meta', 'internal-test-helpers'], function (_metal, _meta, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Chains', class extends _internalTestHelpers.AbstractTestCase {
    ['@test finishChains should properly copy chains from prototypes to instances'](assert) {
      function didChange() {}

      let obj = {};
      (0, _metal.addObserver)(obj, 'foo.bar', null, didChange);

      let childObj = Object.create(obj);

      let parentMeta = (0, _meta.meta)(obj);
      let childMeta = (0, _meta.meta)(childObj);

      (0, _metal.finishChains)(childMeta);

      assert.ok(parentMeta.readableChains() !== childMeta.readableChains(), 'The chains object is copied');
    }

    ['@test does not observe primitive values'](assert) {
      let obj = {
        foo: { bar: 'STRING' }
      };

      (0, _metal.addObserver)(obj, 'foo.bar.baz', null, function () {});
      let meta = (0, _meta.peekMeta)(obj);
      assert.notOk(meta._object);
    }

    ['@test observer and CP chains'](assert) {
      let obj = {};

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)('qux.[]', function () {}));
      (0, _metal.defineProperty)(obj, 'qux', (0, _metal.computed)(function () {}));

      // create DK chains
      (0, _metal.get)(obj, 'foo');

      // create observer chain
      (0, _metal.addObserver)(obj, 'qux.length', function () {});

      /*
             +-----+
             | qux |   root CP
             +-----+
                ^
         +------+-----+
         |            |
      +--------+    +----+
      | length |    | [] |  chainWatchers
      +--------+    +----+
      observer       CP(foo, 'qux.[]')
      */

      // invalidate qux
      (0, _metal.notifyPropertyChange)(obj, 'qux');

      // CP chain is blown away

      /*
             +-----+
             | qux |   root CP
             +-----+
                ^
         +------+xxxxxx
         |            x
      +--------+    xxxxxx
      | length |    x [] x  chainWatchers
      +--------+    xxxxxx
      observer       CP(foo, 'qux.[]')
      */

      (0, _metal.get)(obj, 'qux'); // CP chain re-recreated
      assert.ok(true, 'no crash');
    }

    ['@test checks cache correctly'](assert) {
      let obj = {};
      let parentChainNode = new _metal.ChainNode(null, null, obj);
      let chainNode = new _metal.ChainNode(parentChainNode, 'foo');

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        return undefined;
      }));
      (0, _metal.get)(obj, 'foo');

      assert.strictEqual(chainNode.value(), undefined);
    }

    ['@test chains are watched correctly'](assert) {
      let obj = { foo: { bar: { baz: 1 } } };

      (0, _metal.watch)(obj, 'foo.bar.baz');

      assert.equal((0, _metal.watcherCount)(obj, 'foo'), 1);
      assert.equal((0, _metal.watcherCount)(obj, 'foo.bar'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'foo.bar.baz'), 1);
      assert.equal((0, _metal.watcherCount)(obj.foo, 'bar'), 1);
      assert.equal((0, _metal.watcherCount)(obj.foo, 'bar.baz'), 0);
      assert.equal((0, _metal.watcherCount)(obj.foo.bar, 'baz'), 1);

      (0, _metal.unwatch)(obj, 'foo.bar.baz');

      assert.equal((0, _metal.watcherCount)(obj, 'foo'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'foo.bar'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'foo.bar.baz'), 0);
      assert.equal((0, _metal.watcherCount)(obj.foo, 'bar'), 0);
      assert.equal((0, _metal.watcherCount)(obj.foo, 'bar.baz'), 0);
      assert.equal((0, _metal.watcherCount)(obj.foo.bar, 'baz'), 0);
    }

    ['@test chains with single character keys are watched correctly'](assert) {
      let obj = { a: { b: { c: 1 } } };

      (0, _metal.watch)(obj, 'a.b.c');

      assert.equal((0, _metal.watcherCount)(obj, 'a'), 1);
      assert.equal((0, _metal.watcherCount)(obj, 'a.b'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'a.b.c'), 1);
      assert.equal((0, _metal.watcherCount)(obj.a, 'b'), 1);
      assert.equal((0, _metal.watcherCount)(obj.a, 'b.c'), 0);
      assert.equal((0, _metal.watcherCount)(obj.a.b, 'c'), 1);

      (0, _metal.unwatch)(obj, 'a.b.c');

      assert.equal((0, _metal.watcherCount)(obj, 'a'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'a.b'), 0);
      assert.equal((0, _metal.watcherCount)(obj, 'a.b.c'), 0);
      assert.equal((0, _metal.watcherCount)(obj.a, 'b'), 0);
      assert.equal((0, _metal.watcherCount)(obj.a, 'b.c'), 0);
      assert.equal((0, _metal.watcherCount)(obj.a.b, 'c'), 0);
    }

    ['@test writable chains is not defined more than once'](assert) {
      assert.expect(0);

      class Base {
        constructor() {
          (0, _metal.finishChains)((0, _meta.meta)(this));
        }

        didChange() {}
      }

      Base.prototype.foo = {
        bar: {
          baz: {
            value: 123
          }
        }
      };

      // Define a standard computed property, which will eventually setup dependencies
      (0, _metal.defineProperty)(Base.prototype, 'bar', (0, _metal.computed)('foo.bar', {
        get() {
          return this.foo.bar;
        }
      }));

      // Define some aliases, which will proxy chains along
      (0, _metal.defineProperty)(Base.prototype, 'baz', (0, _metal.alias)('bar.baz'));
      (0, _metal.defineProperty)(Base.prototype, 'value', (0, _metal.alias)('baz.value'));

      // Define an observer, which will eagerly attempt to setup chains and watch
      // their values. This follows the aliases eagerly, and forces the first
      // computed to actually set up its values/dependencies for chains. If
      // writableChains was not already defined, this results in multiple root
      // chain nodes being defined on the same object meta.
      (0, _metal.addObserver)(Base.prototype, 'value', null, 'didChange');

      class Child extends Base {}

      let childObj = new Child();

      (0, _metal.set)(childObj, 'foo.bar', {
        baz: {
          value: 456
        }
      });
    }
  });
});
enifed('@ember/-internals/metal/tests/computed_test', ['@ember/-internals/runtime', '@ember/-internals/metal', '@ember/-internals/meta', 'internal-test-helpers'], function (_runtime, _metal, _meta, _internalTestHelpers) {
  'use strict';

  let obj, count;

  (0, _internalTestHelpers.moduleFor)('computed', class extends _internalTestHelpers.AbstractTestCase {
    ['@test computed property should be an instance of descriptor'](assert) {
      assert.ok((0, _metal.computed)(function () {}) instanceof _metal.Descriptor);
    }

    ['@test computed properties assert the presence of a getter or setter function']() {
      expectAssertion(function () {
        (0, _metal.computed)('nogetternorsetter', {});
      }, 'Computed properties must receive a getter or a setter, you passed none.');
    }

    ['@test computed properties check for the presence of a function or configuration object']() {
      expectAssertion(function () {
        (0, _metal.computed)('nolastargument');
      }, 'computed expects a function or an object as last argument.');
    }

    ['@test computed properties defined with an object only allow `get` and `set` keys']() {
      expectAssertion(function () {
        (0, _metal.computed)({
          get() {},
          set() {},
          other() {}
        });
      }, 'Config object passed to computed can only contain `get` and `set` keys.');
    }

    ['@test computed property can be accessed without `get`'](assert) {
      let obj = {};
      let count = 0;
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function (key) {
        count++;
        return 'computed ' + key;
      }));

      assert.equal(obj.foo, 'computed foo', 'should return value');
      assert.equal(count, 1, 'should have invoked computed property');
    }

    ['@test defining computed property should invoke property on get'](assert) {
      let obj = {};
      let count = 0;
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function (key) {
        count++;
        return 'computed ' + key;
      }));

      assert.equal((0, _metal.get)(obj, 'foo'), 'computed foo', 'should return value');
      assert.equal(count, 1, 'should have invoked computed property');
    }

    ['@test defining computed property should invoke property on set'](assert) {
      let obj = {};
      let count = 0;
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get(key) {
          return this['__' + key];
        },
        set(key, value) {
          count++;
          this['__' + key] = 'computed ' + value;
          return this['__' + key];
        }
      }));

      assert.equal((0, _metal.set)(obj, 'foo', 'bar'), 'bar', 'should return set value');
      assert.equal(count, 1, 'should have invoked computed property');
      assert.equal((0, _metal.get)(obj, 'foo'), 'computed bar', 'should return new value');
    }

    ['@test defining a computed property with a dependent key ending with @each is expanded to []'](assert) {
      let cp = (0, _metal.computed)('blazo.@each', function () {});

      assert.deepEqual(cp._dependentKeys, ['blazo.[]']);

      cp = (0, _metal.computed)('qux', 'zoopa.@each', function () {});

      assert.deepEqual(cp._dependentKeys, ['qux', 'zoopa.[]']);
    }

    ['@test defining a computed property with a dependent key more than one level deep beyond @each is not supported']() {
      expectNoWarning(() => {
        (0, _metal.computed)('todos', () => {});
      });

      expectNoWarning(() => {
        (0, _metal.computed)('todos.@each.owner', () => {});
      });

      expectWarning(() => {
        (0, _metal.computed)('todos.@each.owner.name', () => {});
      }, /You used the key "todos\.@each\.owner\.name" which is invalid\. /);

      expectWarning(() => {
        (0, _metal.computed)('todos.@each.owner.@each.name', () => {});
      }, /You used the key "todos\.@each\.owner\.@each\.name" which is invalid\. /);
    }
  });

  let objA, objB;
  (0, _internalTestHelpers.moduleFor)('computed should inherit through prototype', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      objA = { __foo: 'FOO' };
      (0, _metal.defineProperty)(objA, 'foo', (0, _metal.computed)({
        get(key) {
          return this['__' + key];
        },
        set(key, value) {
          this['__' + key] = 'computed ' + value;
          return this['__' + key];
        }
      }));

      objB = Object.create(objA);
      objB.__foo = 'FOO'; // make a copy;
    }

    afterEach() {
      objA = objB = null;
    }

    ['@test using get() and set()'](assert) {
      assert.equal((0, _metal.get)(objA, 'foo'), 'FOO', 'should get FOO from A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'FOO', 'should get FOO from B');

      (0, _metal.set)(objA, 'foo', 'BIFF');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BIFF', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'FOO', 'should NOT change B');

      (0, _metal.set)(objB, 'foo', 'bar');
      assert.equal((0, _metal.get)(objB, 'foo'), 'computed bar', 'should change B');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BIFF', 'should NOT change A');

      (0, _metal.set)(objA, 'foo', 'BAZ');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BAZ', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'computed bar', 'should NOT change B');
    }
  });

  (0, _internalTestHelpers.moduleFor)('redefining computed property to normal', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      objA = { __foo: 'FOO' };
      (0, _metal.defineProperty)(objA, 'foo', (0, _metal.computed)({
        get(key) {
          return this['__' + key];
        },
        set(key, value) {
          this['__' + key] = 'computed ' + value;
          return this['__' + key];
        }
      }));

      objB = Object.create(objA);
      (0, _metal.defineProperty)(objB, 'foo'); // make this just a normal property.
    }

    afterEach() {
      objA = objB = null;
    }

    ['@test using get() and set()'](assert) {
      assert.equal((0, _metal.get)(objA, 'foo'), 'FOO', 'should get FOO from A');
      assert.equal((0, _metal.get)(objB, 'foo'), undefined, 'should get undefined from B');

      (0, _metal.set)(objA, 'foo', 'BIFF');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BIFF', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), undefined, 'should NOT change B');

      (0, _metal.set)(objB, 'foo', 'bar');
      assert.equal((0, _metal.get)(objB, 'foo'), 'bar', 'should change B');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BIFF', 'should NOT change A');

      (0, _metal.set)(objA, 'foo', 'BAZ');
      assert.equal((0, _metal.get)(objA, 'foo'), 'computed BAZ', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'bar', 'should NOT change B');
    }
  });

  (0, _internalTestHelpers.moduleFor)('redefining computed property to another property', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      objA = { __foo: 'FOO' };
      (0, _metal.defineProperty)(objA, 'foo', (0, _metal.computed)({
        get(key) {
          return this['__' + key];
        },
        set(key, value) {
          this['__' + key] = 'A ' + value;
          return this['__' + key];
        }
      }));

      objB = Object.create(objA);
      objB.__foo = 'FOO';
      (0, _metal.defineProperty)(objB, 'foo', (0, _metal.computed)({
        get(key) {
          return this['__' + key];
        },
        set(key, value) {
          this['__' + key] = 'B ' + value;
          return this['__' + key];
        }
      }));
    }

    afterEach() {
      objA = objB = null;
    }

    ['@test using get() and set()'](assert) {
      assert.equal((0, _metal.get)(objA, 'foo'), 'FOO', 'should get FOO from A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'FOO', 'should get FOO from B');

      (0, _metal.set)(objA, 'foo', 'BIFF');
      assert.equal((0, _metal.get)(objA, 'foo'), 'A BIFF', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'FOO', 'should NOT change B');

      (0, _metal.set)(objB, 'foo', 'bar');
      assert.equal((0, _metal.get)(objB, 'foo'), 'B bar', 'should change B');
      assert.equal((0, _metal.get)(objA, 'foo'), 'A BIFF', 'should NOT change A');

      (0, _metal.set)(objA, 'foo', 'BAZ');
      assert.equal((0, _metal.get)(objA, 'foo'), 'A BAZ', 'should change A');
      assert.equal((0, _metal.get)(objB, 'foo'), 'B bar', 'should NOT change B');
    }
  });

  (0, _internalTestHelpers.moduleFor)('computed - metadata', class extends _internalTestHelpers.AbstractTestCase {
    ['@test can set metadata on a computed property'](assert) {
      let computedProperty = (0, _metal.computed)(function () {});
      computedProperty.meta({ key: 'keyValue' });

      assert.equal(computedProperty.meta().key, 'keyValue', 'saves passed meta hash to the _meta property');
    }

    ['@test meta should return an empty hash if no meta is set'](assert) {
      let computedProperty = (0, _metal.computed)(function () {});
      assert.deepEqual(computedProperty.meta(), {}, 'returned value is an empty hash');
    }
  });

  // ..........................................................
  // CACHEABLE
  //

  (0, _internalTestHelpers.moduleFor)('computed - cacheable', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = {};
      count = 0;
      let func = function () {
        count++;
        return 'bar ' + count;
      };
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({ get: func, set: func }));
    }

    afterEach() {
      obj = count = null;
    }
    ['@test cacheable should cache'](assert) {
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'first get');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'second get');
      assert.equal(count, 1, 'should only invoke once');
    }

    ['@test modifying a cacheable property should update cache'](assert) {
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'first get');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'second get');

      assert.equal((0, _metal.set)(obj, 'foo', 'baz'), 'baz', 'setting');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 2', 'third get');
      assert.equal(count, 2, 'should not invoke again');
    }

    ['@test inherited property should not pick up cache'](assert) {
      let objB = Object.create(obj);

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'obj first get');
      assert.equal((0, _metal.get)(objB, 'foo'), 'bar 2', 'objB first get');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'obj second get');
      assert.equal((0, _metal.get)(objB, 'foo'), 'bar 2', 'objB second get');

      (0, _metal.set)(obj, 'foo', 'baz'); // modify A
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 3', 'obj third get');
      assert.equal((0, _metal.get)(objB, 'foo'), 'bar 2', 'objB third get');
    }

    ['@test getCachedValueFor should return the cached value'](assert) {
      assert.equal((0, _metal.getCachedValueFor)(obj, 'foo'), undefined, 'should not yet be a cached value');

      (0, _metal.get)(obj, 'foo');

      assert.equal((0, _metal.getCachedValueFor)(obj, 'foo'), 'bar 1', 'should retrieve cached value');
    }

    ['@test getCachedValueFor should return falsy cached values'](assert) {
      (0, _metal.defineProperty)(obj, 'falsy', (0, _metal.computed)(function () {
        return false;
      }));

      assert.equal((0, _metal.getCachedValueFor)(obj, 'falsy'), undefined, 'should not yet be a cached value');

      (0, _metal.get)(obj, 'falsy');

      assert.equal((0, _metal.getCachedValueFor)(obj, 'falsy'), false, 'should retrieve cached value');
    }

    ['@test setting a cached computed property passes the old value as the third argument'](assert) {
      let obj = {
        foo: 0
      };

      let receivedOldValue;

      (0, _metal.defineProperty)(obj, 'plusOne', (0, _metal.computed)({
        get() {},
        set(key, value, oldValue) {
          receivedOldValue = oldValue;
          return value;
        }
      }).property('foo'));

      (0, _metal.set)(obj, 'plusOne', 1);
      assert.strictEqual(receivedOldValue, undefined, 'oldValue should be undefined');

      (0, _metal.set)(obj, 'plusOne', 2);
      assert.strictEqual(receivedOldValue, 1, 'oldValue should be 1');

      (0, _metal.set)(obj, 'plusOne', 3);
      assert.strictEqual(receivedOldValue, 2, 'oldValue should be 2');
    }
  });

  // ..........................................................
  // DEPENDENT KEYS
  //

  (0, _internalTestHelpers.moduleFor)('computed - dependentkey', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = { bar: 'baz' };
      count = 0;
      let getterAndSetter = function () {
        count++;
        (0, _metal.get)(this, 'bar');
        return 'bar ' + count;
      };
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get: getterAndSetter,
        set: getterAndSetter
      }).property('bar'));
    }

    afterEach() {
      obj = count = null;
    }

    ['@test should lazily watch dependent keys on set'](assert) {
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily watching dependent key');
    }

    ['@test should lazily watch dependent keys on get'](assert) {
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');
      (0, _metal.get)(obj, 'foo');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily watching dependent key');
    }

    ['@test local dependent key should invalidate cache'](assert) {
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'get once');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily setup watching dependent key');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'cached retrieve');

      (0, _metal.set)(obj, 'bar', 'BIFF'); // should invalidate foo

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 2', 'should recache');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 2', 'cached retrieve');
    }

    ['@test should invalidate multiple nested dependent keys'](assert) {
      let count = 0;
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.computed)(function () {
        count++;
        (0, _metal.get)(this, 'baz');
        return 'baz ' + count;
      }).property('baz'));

      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');
      assert.equal((0, _metal.isWatching)(obj, 'baz'), false, 'precond not watching dependent key');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'get once');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily setup watching dependent key');
      assert.equal((0, _metal.isWatching)(obj, 'baz'), true, 'lazily setup watching dependent key');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1', 'cached retrieve');

      (0, _metal.set)(obj, 'baz', 'BIFF'); // should invalidate bar -> foo
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'should not be watching dependent key after cache cleared');
      assert.equal((0, _metal.isWatching)(obj, 'baz'), false, 'should not be watching dependent key after cache cleared');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 2', 'should recache');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 2', 'cached retrieve');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily setup watching dependent key');
      assert.equal((0, _metal.isWatching)(obj, 'baz'), true, 'lazily setup watching dependent key');
    }

    ['@test circular keys should not blow up'](assert) {
      let func = function () {
        count++;
        return 'bar ' + count;
      };
      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.computed)({ get: func, set: func }).property('foo'));

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        count++;
        return 'foo ' + count;
      }).property('bar'));

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 1', 'get once');
      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 1', 'cached retrieve');

      (0, _metal.set)(obj, 'bar', 'BIFF'); // should invalidate bar -> foo -> bar

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 3', 'should recache');
      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 3', 'cached retrieve');
    }

    ['@test redefining a property should undo old dependent keys'](assert) {
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar 1');
      assert.equal((0, _metal.isWatching)(obj, 'bar'), true, 'lazily watching dependent key');

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        count++;
        return 'baz ' + count;
      }).property('baz'));

      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'after redefining should not be watching dependent key');

      assert.equal((0, _metal.get)(obj, 'foo'), 'baz 2');

      (0, _metal.set)(obj, 'bar', 'BIFF'); // should not kill cache
      assert.equal((0, _metal.get)(obj, 'foo'), 'baz 2');

      (0, _metal.set)(obj, 'baz', 'BOP');
      assert.equal((0, _metal.get)(obj, 'foo'), 'baz 3');
    }

    ['@test can watch multiple dependent keys specified declaratively via brace expansion'](assert) {
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        count++;
        return 'foo ' + count;
      }).property('qux.{bar,baz}'));

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 1', 'get once');
      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 1', 'cached retrieve');

      (0, _metal.set)(obj, 'qux', {});
      (0, _metal.set)(obj, 'qux.bar', 'bar'); // invalidate foo

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 2', 'foo invalidated from bar');

      (0, _metal.set)(obj, 'qux.baz', 'baz'); // invalidate foo

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 3', 'foo invalidated from baz');

      (0, _metal.set)(obj, 'qux.quux', 'quux'); // do not invalidate foo

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo 3', 'foo not invalidated by quux');
    }

    ['@test throws assertion if brace expansion notation has spaces']() {
      expectAssertion(function () {
        (0, _metal.defineProperty)(obj, 'roo', (0, _metal.computed)(function () {
          count++;
          return 'roo ' + count;
        }).property('fee.{bar, baz,bop , }'));
      }, /cannot contain spaces/);
    }

    ['@test throws an assertion if an uncached `get` is called after object is destroyed'](assert) {
      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'precond not watching dependent key');

      let meta = (0, _meta.meta)(obj);
      meta.destroy();

      obj.toString = () => '<custom-obj:here>';

      expectAssertion(() => {
        (0, _metal.get)(obj, 'foo');
      }, 'Cannot modify dependent keys for `foo` on `<custom-obj:here>` after it has been destroyed.');

      assert.equal((0, _metal.isWatching)(obj, 'bar'), false, 'deps were not updated');
    }
  });

  // ..........................................................
  // CHAINED DEPENDENT KEYS
  //

  let func;

  (0, _internalTestHelpers.moduleFor)('computed - dependentkey with chained properties', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = {
        foo: {
          bar: {
            baz: {
              biff: 'BIFF'
            }
          }
        }
      };

      count = 0;
      func = function () {
        count++;
        return (0, _metal.get)(obj, 'foo.bar.baz.biff') + ' ' + count;
      };
    }

    afterEach() {
      obj = count = func = null;
    }

    ['@test depending on simple chain'](assert) {
      // assign computed property
      (0, _metal.defineProperty)(obj, 'prop', (0, _metal.computed)(func).property('foo.bar.baz.biff'));

      assert.equal((0, _metal.get)(obj, 'prop'), 'BIFF 1');

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 2');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 2');

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar'), 'baz', { biff: 'BLOB' });
      assert.equal((0, _metal.get)(obj, 'prop'), 'BLOB 3');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BLOB 3');

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 4');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 4');

      (0, _metal.set)((0, _metal.get)(obj, 'foo'), 'bar', { baz: { biff: 'BOOM' } });
      assert.equal((0, _metal.get)(obj, 'prop'), 'BOOM 5');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BOOM 5');

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 6');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 6');

      (0, _metal.set)(obj, 'foo', { bar: { baz: { biff: 'BLARG' } } });
      assert.equal((0, _metal.get)(obj, 'prop'), 'BLARG 7');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BLARG 7');

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 8');
      assert.equal((0, _metal.get)(obj, 'prop'), 'BUZZ 8');

      (0, _metal.defineProperty)(obj, 'prop');
      (0, _metal.set)(obj, 'prop', 'NONE');
      assert.equal((0, _metal.get)(obj, 'prop'), 'NONE');

      (0, _metal.set)(obj, 'foo', { bar: { baz: { biff: 'BLARG' } } });
      assert.equal((0, _metal.get)(obj, 'prop'), 'NONE'); // should do nothing
      assert.equal(count, 8, 'should be not have invoked computed again');
    }

    ['@test chained dependent keys should evaluate computed properties lazily'](assert) {
      (0, _metal.defineProperty)(obj.foo.bar, 'b', (0, _metal.computed)(func));
      (0, _metal.defineProperty)(obj.foo, 'c', (0, _metal.computed)(function () {}).property('bar.b'));
      assert.equal(count, 0, 'b should not run');
    }
  });

  // ..........................................................
  // improved-cp-syntax
  //

  (0, _internalTestHelpers.moduleFor)('computed - improved cp syntax', class extends _internalTestHelpers.AbstractTestCase {
    ['@test setter and getters are passed using an object'](assert) {
      let testObj = _runtime.Object.extend({
        a: '1',
        b: '2',
        aInt: (0, _metal.computed)('a', {
          get(keyName) {
            assert.equal(keyName, 'aInt', 'getter receives the keyName');
            return parseInt(this.get('a'));
          },
          set(keyName, value, oldValue) {
            assert.equal(keyName, 'aInt', 'setter receives the keyName');
            assert.equal(value, 123, 'setter receives the new value');
            assert.equal(oldValue, 1, 'setter receives the old value');
            this.set('a', '' + value); // side effect
            return parseInt(this.get('a'));
          }
        })
      }).create();

      assert.ok(testObj.get('aInt') === 1, 'getter works');
      testObj.set('aInt', 123);
      assert.ok(testObj.get('a') === '123', 'setter works');
      assert.ok(testObj.get('aInt') === 123, 'cp has been updated too');
    }

    ['@test setter can be omited'](assert) {
      let testObj = _runtime.Object.extend({
        a: '1',
        b: '2',
        aInt: (0, _metal.computed)('a', {
          get(keyName) {
            assert.equal(keyName, 'aInt', 'getter receives the keyName');
            return parseInt(this.get('a'));
          }
        })
      }).create();

      assert.ok(testObj.get('aInt') === 1, 'getter works');
      assert.ok(testObj.get('a') === '1');
      testObj.set('aInt', '123');
      assert.ok(testObj.get('aInt') === '123', 'cp has been updated too');
    }

    ['@test getter can be omited'](assert) {
      let testObj = _runtime.Object.extend({
        com: (0, _metal.computed)({
          set(key, value) {
            return value;
          }
        })
      }).create();

      assert.ok(testObj.get('com') === undefined);
      testObj.set('com', '123');
      assert.ok(testObj.get('com') === '123', 'cp has been updated');
    }

    ['@test the return value of the setter gets cached'](assert) {
      let testObj = _runtime.Object.extend({
        a: '1',
        sampleCP: (0, _metal.computed)('a', {
          get() {
            assert.ok(false, 'The getter should not be invoked');
            return 'get-value';
          },
          set() {
            return 'set-value';
          }
        })
      }).create();

      testObj.set('sampleCP', 'abcd');
      assert.ok(testObj.get('sampleCP') === 'set-value', 'The return value of the CP was cached');
    }
  });

  // ..........................................................
  // BUGS
  //

  (0, _internalTestHelpers.moduleFor)('computed edge cases', class extends _internalTestHelpers.AbstractTestCase {
    ['@test adding a computed property should show up in key iteration'](assert) {
      let obj = {};
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {}));

      let found = [];
      for (let key in obj) {
        found.push(key);
      }
      assert.ok(found.indexOf('foo') >= 0, 'should find computed property in iteration found=' + found);
      assert.ok('foo' in obj, 'foo in obj should pass');
    }

    ["@test when setting a value after it had been retrieved empty don't pass function UNDEFINED as oldValue"](assert) {
      let obj = {};
      let oldValueIsNoFunction = true;

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get() {},
        set(key, value, oldValue) {
          if (typeof oldValue === 'function') {
            oldValueIsNoFunction = false;
          }
          return undefined;
        }
      }));

      (0, _metal.get)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', undefined);

      assert.ok(oldValueIsNoFunction);
    }
  });

  (0, _internalTestHelpers.moduleFor)('computed - setter', class extends _internalTestHelpers.AbstractTestCase {
    ['@test setting a watched computed property'](assert) {
      let obj = {
        firstName: 'Yehuda',
        lastName: 'Katz'
      };

      (0, _metal.defineProperty)(obj, 'fullName', (0, _metal.computed)({
        get() {
          return (0, _metal.get)(this, 'firstName') + ' ' + (0, _metal.get)(this, 'lastName');
        },
        set(key, value) {
          let values = value.split(' ');
          (0, _metal.set)(this, 'firstName', values[0]);
          (0, _metal.set)(this, 'lastName', values[1]);
          return value;
        }
      }).property('firstName', 'lastName'));

      let fullNameDidChange = 0;
      let firstNameDidChange = 0;
      let lastNameDidChange = 0;
      (0, _metal.addObserver)(obj, 'fullName', function () {
        fullNameDidChange++;
      });
      (0, _metal.addObserver)(obj, 'firstName', function () {
        firstNameDidChange++;
      });
      (0, _metal.addObserver)(obj, 'lastName', function () {
        lastNameDidChange++;
      });

      assert.equal((0, _metal.get)(obj, 'fullName'), 'Yehuda Katz');

      (0, _metal.set)(obj, 'fullName', 'Yehuda Katz');

      (0, _metal.set)(obj, 'fullName', 'Kris Selden');

      assert.equal((0, _metal.get)(obj, 'fullName'), 'Kris Selden');
      assert.equal((0, _metal.get)(obj, 'firstName'), 'Kris');
      assert.equal((0, _metal.get)(obj, 'lastName'), 'Selden');

      assert.equal(fullNameDidChange, 1);
      assert.equal(firstNameDidChange, 1);
      assert.equal(lastNameDidChange, 1);
    }

    ['@test setting a cached computed property that modifies the value you give it'](assert) {
      let obj = {
        foo: 0
      };

      (0, _metal.defineProperty)(obj, 'plusOne', (0, _metal.computed)({
        get() {
          return (0, _metal.get)(this, 'foo') + 1;
        },
        set(key, value) {
          (0, _metal.set)(this, 'foo', value);
          return value + 1;
        }
      }).property('foo'));

      let plusOneDidChange = 0;
      (0, _metal.addObserver)(obj, 'plusOne', function () {
        plusOneDidChange++;
      });

      assert.equal((0, _metal.get)(obj, 'plusOne'), 1);
      (0, _metal.set)(obj, 'plusOne', 1);
      assert.equal((0, _metal.get)(obj, 'plusOne'), 2);
      (0, _metal.set)(obj, 'plusOne', 1);
      assert.equal((0, _metal.get)(obj, 'plusOne'), 2);

      assert.equal(plusOneDidChange, 1);

      (0, _metal.set)(obj, 'foo', 5);
      assert.equal((0, _metal.get)(obj, 'plusOne'), 6);

      assert.equal(plusOneDidChange, 2);
    }
  });

  (0, _internalTestHelpers.moduleFor)('computed - default setter', class extends _internalTestHelpers.AbstractTestCase {
    ["@test when setting a value on a computed property that doesn't handle sets"](assert) {
      let obj = {};
      let observerFired = false;

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        return 'foo';
      }));

      (0, _metal.addObserver)(obj, 'foo', null, () => observerFired = true);

      (0, _metal.set)(obj, 'foo', 'bar');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'The set value is properly returned');
      assert.ok(typeof obj.foo === 'string', 'The computed property was removed');
      assert.ok(observerFired, 'The observer was still notified');
    }
  });

  (0, _internalTestHelpers.moduleFor)('computed - readOnly', class extends _internalTestHelpers.AbstractTestCase {
    ['@test is chainable'](assert) {
      let cp = (0, _metal.computed)(function () {}).readOnly();

      assert.ok(cp instanceof _metal.Descriptor);
      assert.ok(cp instanceof _metal.ComputedProperty);
    }

    ['@test throws assertion if called over a CP with a setter defined with the new syntax']() {
      expectAssertion(() => {
        (0, _metal.computed)({
          get() {},
          set() {}
        }).readOnly();
      }, /Computed properties that define a setter using the new syntax cannot be read-only/);
    }

    ['@test protects against setting'](assert) {
      let obj = {};

      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.computed)(function () {
        return 'barValue';
      }).readOnly());

      assert.equal((0, _metal.get)(obj, 'bar'), 'barValue');

      assert.throws(() => {
        (0, _metal.set)(obj, 'bar', 'newBar');
      }, /Cannot set read\-only property "bar" on object:/);

      assert.equal((0, _metal.get)(obj, 'bar'), 'barValue');
    }
  });
});
enifed('@ember/-internals/metal/tests/descriptor_test', ['@ember/-internals/runtime', '@ember/-internals/metal', 'internal-test-helpers'], function (_runtime, _metal, _internalTestHelpers) {
  'use strict';

  let classes = [class {
    static module(title) {
      return `${title}: using defineProperty on an object directly`;
    }

    constructor() {
      this.object = {};
    }

    install(key, desc, assert) {
      let { object } = this;

      (0, _metal.defineProperty)(object, key, desc);

      assert.ok(object.hasOwnProperty(key));
    }

    set(key, value) {
      this.object[key] = value;
    }

    finalize() {
      return this.object;
    }

    source() {
      return this.object;
    }
  }, class {
    static module(title) {
      return `${title}: using defineProperty on a prototype`;
    }

    constructor() {
      this.proto = {};
    }

    install(key, desc, assert) {
      let { proto } = this;

      (0, _metal.defineProperty)(proto, key, desc);

      assert.ok(proto.hasOwnProperty(key));
    }

    set(key, value) {
      this.proto[key] = value;
    }

    finalize() {
      return Object.create(this.proto);
    }

    source() {
      return this.proto;
    }
  }, class {
    static module(title) {
      return `${title}: in EmberObject.extend()`;
    }

    constructor() {
      this.klass = null;
      this.props = {};
    }

    install(key, desc) {
      this.props[key] = desc;
    }

    set(key, value) {
      this.props[key] = value;
    }

    finalize() {
      this.klass = _runtime.Object.extend(this.props);
      return this.klass.create();
    }

    source() {
      return this.klass.prototype;
    }
  }, class {
    static module(title) {
      return `${title}: in EmberObject.extend() through a mixin`;
    }

    constructor() {
      this.klass = null;
      this.props = {};
    }

    install(key, desc) {
      this.props[key] = desc;
    }

    set(key, value) {
      this.props[key] = value;
    }

    finalize() {
      this.klass = _runtime.Object.extend(_metal.Mixin.create(this.props));
      return this.klass.create();
    }

    source() {
      return this.klass.prototype;
    }
  }, class {
    static module(title) {
      return `${title}: inherited from another EmberObject super class`;
    }

    constructor() {
      this.superklass = null;
      this.props = {};
    }

    install(key, desc) {
      this.props[key] = desc;
    }

    set(key, value) {
      this.props[key] = value;
    }

    finalize() {
      this.superklass = _runtime.Object.extend(this.props);
      return this.superklass.extend().create();
    }

    source() {
      return this.superklass.prototype;
    }
  }];

  classes.forEach((TestClass, testCaseIndex) => {
    (0, _internalTestHelpers.moduleFor)(TestClass.module('@ember/-internals/metal/descriptor'), class extends _internalTestHelpers.AbstractTestCase {
      ['@test defining a configurable property'](assert) {
        let factory = new TestClass(assert);

        factory.install('foo', (0, _metal.descriptor)({ configurable: true, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        delete source.foo;

        assert.strictEqual(obj.foo, undefined);

        Object.defineProperty(source, 'foo', { configurable: true, value: 'baz' });

        assert.equal(obj.foo, 'baz');
      }

      ['@test defining a non-configurable property'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({ configurable: false, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        if (_internalTestHelpers.isEdge && testCaseIndex === 0) {
          // https://github.com/emberjs/ember.js/pull/16741#issuecomment-420963181
          assert.equal(delete source.foo, false);
        } else {
          assert.throws(() => delete source.foo, TypeError);
        }

        assert.throws(() => Object.defineProperty(source, 'foo', {
          configurable: true,
          value: 'baz'
        }), TypeError);

        assert.equal(obj.foo, 'bar');
      }

      ['@test defining an enumerable property'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({ enumerable: true, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        assert.ok(Object.keys(source).indexOf('foo') !== -1);
      }

      ['@test defining a non-enumerable property'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({ enumerable: false, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        assert.ok(Object.keys(source).indexOf('foo') === -1);
      }

      ['@test defining a writable property'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({ writable: true, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        source.foo = 'baz';

        assert.equal(obj.foo, 'baz');

        obj.foo = 'bat';

        assert.equal(obj.foo, 'bat');
      }

      ['@test defining a non-writable property'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({ writable: false, value: 'bar' }), assert);

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        let source = factory.source();

        assert.throws(() => source.foo = 'baz', TypeError);
        assert.throws(() => obj.foo = 'baz', TypeError);

        assert.equal(obj.foo, 'bar');
      }

      ['@test defining a getter'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({
          get: function () {
            return this.__foo__;
          }
        }), assert);

        factory.set('__foo__', 'bar');

        let obj = factory.finalize();

        assert.equal(obj.foo, 'bar');

        obj.__foo__ = 'baz';

        assert.equal(obj.foo, 'baz');
      }

      ['@test defining a setter'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({
          set: function (value) {
            this.__foo__ = value;
          }
        }), assert);

        factory.set('__foo__', 'bar');

        let obj = factory.finalize();

        assert.equal(obj.__foo__, 'bar');

        obj.foo = 'baz';

        assert.equal(obj.__foo__, 'baz');
      }

      ['@test combining multiple setter and getters'](assert) {
        let factory = new TestClass(assert);
        factory.install('foo', (0, _metal.descriptor)({
          get: function () {
            return this.__foo__;
          },

          set: function (value) {
            this.__foo__ = value;
          }
        }), assert);

        factory.set('__foo__', 'foo');

        factory.install('bar', (0, _metal.descriptor)({
          get: function () {
            return this.__bar__;
          },

          set: function (value) {
            this.__bar__ = value;
          }
        }), assert);

        factory.set('__bar__', 'bar');

        factory.install('fooBar', (0, _metal.descriptor)({
          get: function () {
            return this.foo + '-' + this.bar;
          }
        }), assert);

        let obj = factory.finalize();

        assert.equal(obj.fooBar, 'foo-bar');

        obj.foo = 'FOO';

        assert.equal(obj.fooBar, 'FOO-bar');

        obj.__bar__ = 'BAR';

        assert.equal(obj.fooBar, 'FOO-BAR');

        assert.throws(() => obj.fooBar = 'foobar', TypeError);

        assert.equal(obj.fooBar, 'FOO-BAR');
      }
    });
  });
});
enifed('@ember/-internals/metal/tests/events_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/props/events_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test listener should receive event - removing should remove'](assert) {
      let obj = {};
      let count = 0;

      function F() {
        count++;
      }

      (0, _metal.addListener)(obj, 'event!', F);
      assert.equal(count, 0, 'nothing yet');

      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(count, 1, 'received event');

      (0, _metal.removeListener)(obj, 'event!', F);

      count = 0;
      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(count, 0, 'received event');
    }

    ['@test listeners should be inherited'](assert) {
      let count = 0;

      let obj = {
        func() {
          count++;
        }
      };

      (0, _metal.addListener)(obj, 'event!', null, 'func');

      let obj2 = Object.create(obj);

      assert.equal(count, 0, 'nothing yet');

      (0, _metal.sendEvent)(obj2, 'event!');
      assert.equal(count, 1, 'received event');

      (0, _metal.removeListener)(obj2, 'event!', null, 'func');

      count = 0;
      (0, _metal.sendEvent)(obj2, 'event!');
      assert.equal(count, 0, 'did not receive event');

      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(count, 1, 'should still invoke on parent');
    }

    ['@test adding a listener more than once should only invoke once'](assert) {
      let count = 0;
      let obj = {
        func() {
          count++;
        }
      };

      (0, _metal.addListener)(obj, 'event!', null, 'func');
      (0, _metal.addListener)(obj, 'event!', null, 'func');

      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(count, 1, 'should only invoke once');
    }

    ['@test adding a listener with a target should invoke with target'](assert) {
      let obj = {};
      let target;

      target = {
        count: 0,
        method() {
          this.count++;
        }
      };

      (0, _metal.addListener)(obj, 'event!', target, target.method);
      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(target.count, 1, 'should invoke');
    }

    ['@test adding a listener with string method should lookup method on event delivery'](assert) {
      let obj = {};
      let target;

      target = {
        count: 0,
        method() {}
      };

      (0, _metal.addListener)(obj, 'event!', target, 'method');
      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(target.count, 0, 'should invoke but do nothing');

      target.method = function () {
        this.count++;
      };
      (0, _metal.sendEvent)(obj, 'event!');
      assert.equal(target.count, 1, 'should invoke now');
    }

    ['@test calling sendEvent with extra params should be passed to listeners'](assert) {
      let obj = {};
      let params = null;
      (0, _metal.addListener)(obj, 'event!', function () {
        params = Array.prototype.slice.call(arguments);
      });

      (0, _metal.sendEvent)(obj, 'event!', ['foo', 'bar']);
      assert.deepEqual(params, ['foo', 'bar'], 'params should be saved');
    }

    ['@test hasListeners tells you if there are listeners for a given event'](assert) {
      let obj = {};

      function F() {}
      function F2() {}

      assert.equal((0, _metal.hasListeners)(obj, 'event!'), false, 'no listeners at first');

      (0, _metal.addListener)(obj, 'event!', F);
      (0, _metal.addListener)(obj, 'event!', F2);

      assert.equal((0, _metal.hasListeners)(obj, 'event!'), true, 'has listeners');

      (0, _metal.removeListener)(obj, 'event!', F);
      assert.equal((0, _metal.hasListeners)(obj, 'event!'), true, 'has listeners');

      (0, _metal.removeListener)(obj, 'event!', F2);
      assert.equal((0, _metal.hasListeners)(obj, 'event!'), false, 'has no more listeners');

      (0, _metal.addListener)(obj, 'event!', F);
      assert.equal((0, _metal.hasListeners)(obj, 'event!'), true, 'has listeners');
    }

    ['@test calling removeListener without method should remove all listeners'](assert) {
      expectDeprecation(() => {
        let obj = {};
        function F() {}
        function F2() {}

        assert.equal((0, _metal.hasListeners)(obj, 'event!'), false, 'no listeners at first');

        (0, _metal.addListener)(obj, 'event!', F);
        (0, _metal.addListener)(obj, 'event!', F2);

        assert.equal((0, _metal.hasListeners)(obj, 'event!'), true, 'has listeners');
        (0, _metal.removeListener)(obj, 'event!');

        assert.equal((0, _metal.hasListeners)(obj, 'event!'), false, 'has no more listeners');
      }, /The remove all functionality of removeListener and removeObserver has been deprecated/);
    }

    ['@test a listener can be added as part of a mixin'](assert) {
      let triggered = 0;
      let MyMixin = _metal.Mixin.create({
        foo1: (0, _metal.on)('bar', function () {
          triggered++;
        }),

        foo2: (0, _metal.on)('bar', function () {
          triggered++;
        })
      });

      let obj = {};
      MyMixin.apply(obj);

      (0, _metal.sendEvent)(obj, 'bar');
      assert.equal(triggered, 2, 'should invoke listeners');
    }

    [`@test 'on' asserts for invalid arguments`]() {
      expectAssertion(() => {
        _metal.Mixin.create({
          foo1: (0, _metal.on)('bar')
        });
      }, 'on expects function as last argument');

      expectAssertion(() => {
        _metal.Mixin.create({
          foo1: (0, _metal.on)(function () {})
        });
      }, 'on called without valid event names');
    }

    ['@test a listener added as part of a mixin may be overridden'](assert) {
      let triggered = 0;
      let FirstMixin = _metal.Mixin.create({
        foo: (0, _metal.on)('bar', function () {
          triggered++;
        })
      });
      let SecondMixin = _metal.Mixin.create({
        foo: (0, _metal.on)('baz', function () {
          triggered++;
        })
      });

      let obj = {};
      FirstMixin.apply(obj);
      SecondMixin.apply(obj);

      (0, _metal.sendEvent)(obj, 'bar');
      assert.equal(triggered, 0, 'should not invoke from overridden property');

      (0, _metal.sendEvent)(obj, 'baz');
      assert.equal(triggered, 1, 'should invoke from subclass property');
    }
  });
});
enifed('@ember/-internals/metal/tests/expand_properties_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  let foundProperties = [];

  function addProperty(property) {
    foundProperties.push(property);
  }

  (0, _internalTestHelpers.moduleFor)('Property Brace Expansion Test', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      foundProperties = [];
    }

    ['@test Properties without expansions are unaffected'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('a', addProperty);
      (0, _metal.expandProperties)('a.b', addProperty);
      (0, _metal.expandProperties)('a.b.[]', addProperty);
      (0, _metal.expandProperties)('a.b.@each.c', addProperty);

      assert.deepEqual(['a', 'a.b', 'a.b.[]', 'a.b.@each.c'].sort(), foundProperties.sort());
    }

    ['@test A single expansion at the end expands properly'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('a.b.{c,d}', addProperty);

      assert.deepEqual(['a.b.c', 'a.b.d'].sort(), foundProperties.sort());
    }

    ['@test A property with only a brace expansion expands correctly'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('{a,b,c}', addProperty);

      let expected = ['a', 'b', 'c'];
      assert.deepEqual(expected.sort(), foundProperties.sort());
    }

    ['@test Expansions with single properties only expand once'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('a.b.{c}.d.{e}', addProperty);

      assert.deepEqual(['a.b.c.d.e'], foundProperties);
    }

    ['@test A single brace expansion expands correctly'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('a.{b,c,d}.e', addProperty);

      let expected = ['a.b.e', 'a.c.e', 'a.d.e'];
      assert.deepEqual(expected.sort(), foundProperties.sort());
    }

    ['@test Multiple brace expansions work correctly'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('{a,b,c}.d.{e,f}.g', addProperty);

      let expected = ['a.d.e.g', 'a.d.f.g', 'b.d.e.g', 'b.d.f.g', 'c.d.e.g', 'c.d.f.g'];
      assert.deepEqual(expected.sort(), foundProperties.sort());
    }

    ['@test A property with only brace expansions expands correctly'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('{a,b,c}.{d}.{e,f}', addProperty);

      let expected = ['a.d.e', 'a.d.f', 'b.d.e', 'b.d.f', 'c.d.e', 'c.d.f'];
      assert.deepEqual(expected.sort(), foundProperties.sort());
    }

    ['@test Nested brace expansions are not allowed']() {
      let nestedBraceProperties = ['a.{b.{c,d}}', 'a.{{b}.c}', 'a.{b,c}.{d.{e,f}.g', 'a.{b.{c}', 'a.{b,c}}', 'model.{bar,baz'];

      nestedBraceProperties.forEach(invalidProperties => {
        expectAssertion(() => (0, _metal.expandProperties)(invalidProperties, addProperty));
      }, /Brace expanded properties have to be balanced and cannot be nested/);
    }

    ['@test A property with no braces does not expand'](assert) {
      assert.expect(1);

      (0, _metal.expandProperties)('a,b,c.d.e,f', addProperty);

      assert.deepEqual(foundProperties, ['a,b,c.d.e,f']);
    }

    ['@test A pattern must be a string'](assert) {
      assert.expect(1);

      expectAssertion(() => {
        (0, _metal.expandProperties)([1, 2], addProperty);
      }, /A computed property key must be a string/);
    }

    ['@test A pattern must not contain a space'](assert) {
      assert.expect(1);

      expectAssertion(function () {
        (0, _metal.expandProperties)('{a, b}', addProperty);
      }, /Brace expanded properties cannot contain spaces, e.g. "user.{firstName, lastName}" should be "user.{firstName,lastName}"/);
    }
  });
});
enifed('@ember/-internals/metal/tests/injected_property_test', ['@ember/-internals/owner', '@ember/-internals/metal', 'internal-test-helpers'], function (_owner, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('InjectedProperty', class extends _internalTestHelpers.AbstractTestCase {
    ['@test injected properties should be descriptors'](assert) {
      assert.ok(new _metal.InjectedProperty() instanceof _metal.Descriptor);
    }

    ['@test injected properties should be overridable'](assert) {
      let obj = {};
      (0, _metal.defineProperty)(obj, 'foo', new _metal.InjectedProperty());

      (0, _metal.set)(obj, 'foo', 'bar');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'should return the overridden value');
    }

    ['@test getting on an object without an owner or container should fail assertion']() {
      let obj = {};
      (0, _metal.defineProperty)(obj, 'foo', new _metal.InjectedProperty('type', 'name'));

      expectAssertion(function () {
        (0, _metal.get)(obj, 'foo');
      }, /Attempting to lookup an injected property on an object without a container, ensure that the object was instantiated via a container./);
    }

    ['@test getting on an object without an owner but with a container should not fail'](assert) {
      let obj = {
        container: {
          lookup(key) {
            assert.ok(true, 'should call container.lookup');
            return key;
          }
        }
      };

      (0, _metal.defineProperty)(obj, 'foo', new _metal.InjectedProperty('type', 'name'));

      assert.equal((0, _metal.get)(obj, 'foo'), 'type:name', 'should return the value of container.lookup');
    }

    ['@test getting should return a lookup on the container'](assert) {
      assert.expect(2);

      let obj = {};

      (0, _owner.setOwner)(obj, {
        lookup(key) {
          assert.ok(true, 'should call container.lookup');
          return key;
        }
      });

      (0, _metal.defineProperty)(obj, 'foo', new _metal.InjectedProperty('type', 'name'));

      assert.equal((0, _metal.get)(obj, 'foo'), 'type:name', 'should return the value of container.lookup');
    }

    ['@test omitting the lookup name should default to the property name'](assert) {
      let obj = {};

      (0, _owner.setOwner)(obj, {
        lookup(key) {
          return key;
        }
      });

      (0, _metal.defineProperty)(obj, 'foo', new _metal.InjectedProperty('type'));

      assert.equal((0, _metal.get)(obj, 'foo'), 'type:foo', 'should lookup the type using the property name');
    }
  });
});
enifed('@ember/-internals/metal/tests/is_blank_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('isBlank', class extends _internalTestHelpers.AbstractTestCase {
    ['@test isBlank'](assert) {
      let string = 'string';
      let fn = function () {};
      let object = { length: 0 };

      assert.equal(true, (0, _metal.isBlank)(null), 'for null');
      assert.equal(true, (0, _metal.isBlank)(undefined), 'for undefined');
      assert.equal(true, (0, _metal.isBlank)(''), 'for an empty String');
      assert.equal(true, (0, _metal.isBlank)('  '), 'for a whitespace String');
      assert.equal(true, (0, _metal.isBlank)('\n\t'), 'for another whitespace String');
      assert.equal(false, (0, _metal.isBlank)('\n\t Hi'), 'for a String with whitespaces');
      assert.equal(false, (0, _metal.isBlank)(true), 'for true');
      assert.equal(false, (0, _metal.isBlank)(false), 'for false');
      assert.equal(false, (0, _metal.isBlank)(string), 'for a String');
      assert.equal(false, (0, _metal.isBlank)(fn), 'for a Function');
      assert.equal(false, (0, _metal.isBlank)(0), 'for 0');
      assert.equal(true, (0, _metal.isBlank)([]), 'for an empty Array');
      assert.equal(false, (0, _metal.isBlank)({}), 'for an empty Object');
      assert.equal(true, (0, _metal.isBlank)(object), "for an Object that has zero 'length'");
      assert.equal(false, (0, _metal.isBlank)([1, 2, 3]), 'for a non-empty array');
    }
  });
});
enifed('@ember/-internals/metal/tests/is_empty_test', ['@ember/-internals/metal', '@ember/map', '@ember/map/lib/ordered-set', 'internal-test-helpers'], function (_metal, _map, _orderedSet, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('isEmpty', class extends _internalTestHelpers.AbstractTestCase {
    ['@test isEmpty'](assert) {
      let string = 'string';
      let fn = function () {};
      let object = { length: 0 };

      assert.equal(true, (0, _metal.isEmpty)(null), 'for null');
      assert.equal(true, (0, _metal.isEmpty)(undefined), 'for undefined');
      assert.equal(true, (0, _metal.isEmpty)(''), 'for an empty String');
      assert.equal(false, (0, _metal.isEmpty)('  '), 'for a whitespace String');
      assert.equal(false, (0, _metal.isEmpty)('\n\t'), 'for another whitespace String');
      assert.equal(false, (0, _metal.isEmpty)(true), 'for true');
      assert.equal(false, (0, _metal.isEmpty)(false), 'for false');
      assert.equal(false, (0, _metal.isEmpty)(string), 'for a String');
      assert.equal(false, (0, _metal.isEmpty)(fn), 'for a Function');
      assert.equal(false, (0, _metal.isEmpty)(0), 'for 0');
      assert.equal(true, (0, _metal.isEmpty)([]), 'for an empty Array');
      assert.equal(false, (0, _metal.isEmpty)({}), 'for an empty Object');
      assert.equal(true, (0, _metal.isEmpty)(object), "for an Object that has zero 'length'");
    }

    ['@test isEmpty Map'](assert) {
      expectDeprecation(() => {
        let map = new _map.default();
        assert.equal(true, (0, _metal.isEmpty)(map), 'Empty map is empty');
        map.set('foo', 'bar');
        assert.equal(false, (0, _metal.isEmpty)(map), 'Map is not empty');
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
    }

    ['@test isEmpty Ember.OrderedSet'](assert) {
      expectDeprecation(() => {
        let orderedSet = new _orderedSet.default();
        assert.equal(true, (0, _metal.isEmpty)(orderedSet), 'Empty ordered set is empty');
        orderedSet.add('foo');
        assert.equal(false, (0, _metal.isEmpty)(orderedSet), 'Ordered set is not empty');
      }, 'Use of @ember/OrderedSet is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
    }
  });
});
enifed('@ember/-internals/metal/tests/is_none_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('isNone', class extends _internalTestHelpers.AbstractTestCase {
    ['@test isNone'](assert) {
      let string = 'string';
      let fn = function () {};

      assert.equal(true, (0, _metal.isNone)(null), 'for null');
      assert.equal(true, (0, _metal.isNone)(undefined), 'for undefined');
      assert.equal(false, (0, _metal.isNone)(''), 'for an empty String');
      assert.equal(false, (0, _metal.isNone)(true), 'for true');
      assert.equal(false, (0, _metal.isNone)(false), 'for false');
      assert.equal(false, (0, _metal.isNone)(string), 'for a String');
      assert.equal(false, (0, _metal.isNone)(fn), 'for a Function');
      assert.equal(false, (0, _metal.isNone)(0), 'for 0');
      assert.equal(false, (0, _metal.isNone)([]), 'for an empty Array');
      assert.equal(false, (0, _metal.isNone)({}), 'for an empty Object');
    }
  });
});
enifed('@ember/-internals/metal/tests/is_present_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('isPresent', class extends _internalTestHelpers.AbstractTestCase {
    ['@test isPresent'](assert) {
      let string = 'string';
      let fn = function () {};
      let object = { length: 0 };

      assert.equal(false, (0, _metal.isPresent)(), 'for no params');
      assert.equal(false, (0, _metal.isPresent)(null), 'for null');
      assert.equal(false, (0, _metal.isPresent)(undefined), 'for undefined');
      assert.equal(false, (0, _metal.isPresent)(''), 'for an empty String');
      assert.equal(false, (0, _metal.isPresent)('  '), 'for a whitespace String');
      assert.equal(false, (0, _metal.isPresent)('\n\t'), 'for another whitespace String');
      assert.equal(true, (0, _metal.isPresent)('\n\t Hi'), 'for a String with whitespaces');
      assert.equal(true, (0, _metal.isPresent)(true), 'for true');
      assert.equal(true, (0, _metal.isPresent)(false), 'for false');
      assert.equal(true, (0, _metal.isPresent)(string), 'for a String');
      assert.equal(true, (0, _metal.isPresent)(fn), 'for a Function');
      assert.equal(true, (0, _metal.isPresent)(0), 'for 0');
      assert.equal(false, (0, _metal.isPresent)([]), 'for an empty Array');
      assert.equal(true, (0, _metal.isPresent)({}), 'for an empty Object');
      assert.equal(false, (0, _metal.isPresent)(object), "for an Object that has zero 'length'");
      assert.equal(true, (0, _metal.isPresent)([1, 2, 3]), 'for a non-empty array');
    }
  });
});
enifed('@ember/-internals/metal/tests/libraries_test', ['@ember/debug', '@ember/-internals/metal', '@ember/canary-features', 'internal-test-helpers'], function (_debug, _metal, _canaryFeatures, _internalTestHelpers) {
  'use strict';

  /* globals EmberDev */
  let libs, registry;
  let originalWarn = (0, _debug.getDebugFunction)('warn');
  function noop() {}

  (0, _internalTestHelpers.moduleFor)('Libraries registry', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      libs = new _metal.Libraries();
      registry = libs._registry;
    }

    afterEach() {
      libs = null;
      registry = null;

      (0, _debug.setDebugFunction)('warn', originalWarn);
    }

    ['@test core libraries come before other libraries'](assert) {
      assert.expect(2);

      libs.register('my-lib', '2.0.0a');
      libs.registerCoreLibrary('DS', '1.0.0-beta.2');

      assert.equal(registry[0].name, 'DS');
      assert.equal(registry[1].name, 'my-lib');
    }

    ['@test only the first registration of a library is stored'](assert) {
      assert.expect(3);

      // overwrite warn to supress the double registration warning (see https://github.com/emberjs/ember.js/issues/16391)
      (0, _debug.setDebugFunction)('warn', noop);
      libs.register('magic', 1.23);
      libs.register('magic', 2.23);

      assert.equal(registry[0].name, 'magic');
      assert.equal(registry[0].version, 1.23);
      assert.equal(registry.length, 1);
    }

    ['@test isRegistered returns correct value'](assert) {
      if (_canaryFeatures.EMBER_LIBRARIES_ISREGISTERED) {
        assert.expect(3);

        assert.equal(libs.isRegistered('magic'), false);

        libs.register('magic', 1.23);
        assert.equal(libs.isRegistered('magic'), true);

        libs.deRegister('magic');
        assert.equal(libs.isRegistered('magic'), false);
      } else {
        assert.expect(0);
      }
    }

    ['@test attempting to register a library that is already registered warns you'](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      assert.expect(1);

      libs.register('magic', 1.23);

      (0, _debug.setDebugFunction)('warn', function (msg, test) {
        if (!test) {
          assert.equal(msg, 'Library "magic" is already registered with Ember.');
        }
      });

      // Should warn us
      libs.register('magic', 2.23);
    }

    ['@test libraries can be de-registered'](assert) {
      assert.expect(2);

      libs.register('lib1', '1.0.0b');
      libs.register('lib2', '1.0.0b');
      libs.register('lib3', '1.0.0b');

      libs.deRegister('lib1');
      libs.deRegister('lib3');

      assert.equal(registry[0].name, 'lib2');
      assert.equal(registry.length, 1);
    }
  });
});
enifed('@ember/-internals/metal/tests/main_test', ['ember/version', 'internal-test-helpers'], function (_version, _internalTestHelpers) {
  'use strict';

  // From https://github.com/semver/semver.org/issues/59 & https://regex101.com/r/vW1jA8/6
  const SEMVER_REGEX = /^((?:0|(?:[1-9]\d*)))\.((?:0|(?:[1-9]\d*)))\.((?:0|(?:[1-9]\d*)))(?:-([0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*))?(?:\+([0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*))?$/;

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/metal/core/main', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.VERSION is in alignment with SemVer v2.0.0'](assert) {
      assert.ok(SEMVER_REGEX.test(_version.default), `Ember.VERSION (${_version.default})is valid SemVer v2.0.0`);
    }

    ['@test SEMVER_REGEX properly validates and invalidates version numbers'](assert) {
      function validateVersionString(versionString, expectedResult) {
        assert.equal(SEMVER_REGEX.test(versionString), expectedResult);
      }

      // Positive test cases
      validateVersionString('1.11.3', true);
      validateVersionString('1.0.0-beta.16.1', true);
      validateVersionString('1.12.1+canary.aba1412', true);
      validateVersionString('2.0.0-beta.1+canary.bb344775', true);
      validateVersionString('3.1.0-foobarBaz+30d70bd3', true);

      // Negative test cases
      validateVersionString('1.11.3.aba18a', false);
      validateVersionString('1.11', false);
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/alias_method_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  function validateAliasMethod(assert, obj) {
    assert.equal(obj.fooMethod(), 'FOO', 'obj.fooMethod()');
    assert.equal(obj.barMethod(), 'FOO', 'obj.barMethod should be a copy of foo');
  }

  (0, _internalTestHelpers.moduleFor)('aliasMethod', class extends _internalTestHelpers.AbstractTestCase {
    ['@test methods of another name are aliased when the mixin is applied'](assert) {
      let MyMixin = _metal.Mixin.create({
        fooMethod() {
          return 'FOO';
        },
        barMethod: (0, _metal.aliasMethod)('fooMethod')
      });

      let obj = MyMixin.apply({});
      validateAliasMethod(assert, obj);
    }

    ['@test should follow aliasMethods all the way down'](assert) {
      let MyMixin = _metal.Mixin.create({
        bar: (0, _metal.aliasMethod)('foo'), // put first to break ordered iteration
        baz() {
          return 'baz';
        },
        foo: (0, _metal.aliasMethod)('baz')
      });

      let obj = MyMixin.apply({});
      assert.equal((0, _metal.get)(obj, 'bar')(), 'baz', 'should have followed aliasMethods');
    }

    ['@test should alias methods from other dependent mixins'](assert) {
      let BaseMixin = _metal.Mixin.create({
        fooMethod() {
          return 'FOO';
        }
      });

      let MyMixin = _metal.Mixin.create(BaseMixin, {
        barMethod: (0, _metal.aliasMethod)('fooMethod')
      });

      let obj = MyMixin.apply({});
      validateAliasMethod(assert, obj);
    }

    ['@test should alias methods from other mixins applied at same time'](assert) {
      let BaseMixin = _metal.Mixin.create({
        fooMethod() {
          return 'FOO';
        }
      });

      let MyMixin = _metal.Mixin.create({
        barMethod: (0, _metal.aliasMethod)('fooMethod')
      });

      let obj = (0, _metal.mixin)({}, BaseMixin, MyMixin);
      validateAliasMethod(assert, obj);
    }

    ['@test should alias methods from mixins already applied on object'](assert) {
      let BaseMixin = _metal.Mixin.create({
        quxMethod() {
          return 'qux';
        }
      });

      let MyMixin = _metal.Mixin.create({
        bar: (0, _metal.aliasMethod)('foo'),
        barMethod: (0, _metal.aliasMethod)('fooMethod')
      });

      let obj = {
        fooMethod() {
          return 'FOO';
        }
      };

      BaseMixin.apply(obj);
      MyMixin.apply(obj);

      validateAliasMethod(assert, obj);
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/apply_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  function K() {}

  (0, _internalTestHelpers.moduleFor)('Mixin.apply', class extends _internalTestHelpers.AbstractTestCase {
    ['@test using apply() should apply properties'](assert) {
      let MixinA = _metal.Mixin.create({ foo: 'FOO', baz: K });
      let obj = {};
      (0, _metal.mixin)(obj, MixinA);

      assert.equal((0, _metal.get)(obj, 'foo'), 'FOO', 'should apply foo');
      assert.equal((0, _metal.get)(obj, 'baz'), K, 'should apply foo');
    }

    ['@test applying anonymous properties'](assert) {
      let obj = {};
      (0, _metal.mixin)(obj, {
        foo: 'FOO',
        baz: K
      });

      assert.equal((0, _metal.get)(obj, 'foo'), 'FOO', 'should apply foo');
      assert.equal((0, _metal.get)(obj, 'baz'), K, 'should apply foo');
    }

    ['@test applying null values']() {
      expectAssertion(() => (0, _metal.mixin)({}, null));
    }

    ['@test applying a property with an undefined value'](assert) {
      let obj = { tagName: '' };
      (0, _metal.mixin)(obj, { tagName: undefined });

      assert.strictEqual((0, _metal.get)(obj, 'tagName'), '');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/computed_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  function K() {
    return this;
  }

  (0, _internalTestHelpers.moduleFor)('Mixin Computed Properties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test overriding computed properties'](assert) {
      let MixinA, MixinB, MixinC, MixinD;
      let obj;

      MixinA = _metal.Mixin.create({
        aProp: (0, _metal.computed)(function () {
          return 'A';
        })
      });

      MixinB = _metal.Mixin.create(MixinA, {
        aProp: (0, _metal.computed)(function () {
          return this._super(...arguments) + 'B';
        })
      });

      MixinC = _metal.Mixin.create(MixinA, {
        aProp: (0, _metal.computed)(function () {
          return this._super(...arguments) + 'C';
        })
      });

      MixinD = _metal.Mixin.create({
        aProp: (0, _metal.computed)(function () {
          return this._super(...arguments) + 'D';
        })
      });

      obj = {};
      MixinB.apply(obj);
      assert.equal((0, _metal.get)(obj, 'aProp'), 'AB', 'should expose super for B');

      obj = {};
      MixinC.apply(obj);
      assert.equal((0, _metal.get)(obj, 'aProp'), 'AC', 'should expose super for C');

      obj = {};

      MixinA.apply(obj);
      MixinD.apply(obj);
      assert.equal((0, _metal.get)(obj, 'aProp'), 'AD', 'should define super for D');

      obj = {};
      (0, _metal.defineProperty)(obj, 'aProp', (0, _metal.computed)(function () {
        return 'obj';
      }));
      MixinD.apply(obj);
      assert.equal((0, _metal.get)(obj, 'aProp'), 'objD', 'should preserve original computed property');
    }

    ['@test calling set on overridden computed properties'](assert) {
      let SuperMixin, SubMixin;
      let obj;

      let superGetOccurred = false;
      let superSetOccurred = false;

      SuperMixin = _metal.Mixin.create({
        aProp: (0, _metal.computed)({
          get() {
            superGetOccurred = true;
          },
          set() {
            superSetOccurred = true;
          }
        })
      });

      SubMixin = _metal.Mixin.create(SuperMixin, {
        aProp: (0, _metal.computed)({
          get() {
            return this._super(...arguments);
          },
          set() {
            return this._super(...arguments);
          }
        })
      });

      obj = {};
      SubMixin.apply(obj);

      (0, _metal.set)(obj, 'aProp', 'set thyself');
      assert.ok(superSetOccurred, 'should pass set to _super');

      superSetOccurred = false; // reset the set assertion

      obj = {};
      SubMixin.apply(obj);

      (0, _metal.get)(obj, 'aProp');
      assert.ok(superGetOccurred, 'should pass get to _super');

      (0, _metal.set)(obj, 'aProp', 'set thyself');
      assert.ok(superSetOccurred, 'should pass set to _super after getting');
    }

    ['@test setter behavior works properly when overriding computed properties'](assert) {
      let obj = {};

      let MixinA = _metal.Mixin.create({
        cpWithSetter2: (0, _metal.computed)(K),
        cpWithSetter3: (0, _metal.computed)(K),
        cpWithoutSetter: (0, _metal.computed)(K)
      });

      let cpWasCalled = false;

      let MixinB = _metal.Mixin.create({
        cpWithSetter2: (0, _metal.computed)({
          get: K,
          set() {
            cpWasCalled = true;
          }
        }),

        cpWithSetter3: (0, _metal.computed)({
          get: K,
          set() {
            cpWasCalled = true;
          }
        }),

        cpWithoutSetter: (0, _metal.computed)(function () {
          cpWasCalled = true;
        })
      });

      MixinA.apply(obj);
      MixinB.apply(obj);

      (0, _metal.set)(obj, 'cpWithSetter2', 'test');
      assert.ok(cpWasCalled, 'The computed property setter was called when defined with two args');
      cpWasCalled = false;

      (0, _metal.set)(obj, 'cpWithSetter3', 'test');
      assert.ok(cpWasCalled, 'The computed property setter was called when defined with three args');
      cpWasCalled = false;

      (0, _metal.set)(obj, 'cpWithoutSetter', 'test');
      assert.equal((0, _metal.get)(obj, 'cpWithoutSetter'), 'test', 'The default setter was called, the value is correct');
      assert.ok(!cpWasCalled, 'The default setter was called, not the CP itself');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/concatenated_properties_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin concatenatedProperties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test defining concatenated properties should concat future version'](assert) {
      let MixinA = _metal.Mixin.create({
        concatenatedProperties: ['foo'],
        foo: ['a', 'b', 'c']
      });

      let MixinB = _metal.Mixin.create({
        foo: ['d', 'e', 'f']
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), ['a', 'b', 'c', 'd', 'e', 'f']);
    }

    ['@test defining concatenated properties should concat future version'](assert) {
      let MixinA = _metal.Mixin.create({
        concatenatedProperties: null
      });

      let MixinB = _metal.Mixin.create({
        concatenatedProperties: null
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);

      assert.deepEqual(obj.concatenatedProperties, []);
    }

    ['@test concatenatedProperties should be concatenated'](assert) {
      let MixinA = _metal.Mixin.create({
        concatenatedProperties: ['foo'],
        foo: ['a', 'b', 'c']
      });

      let MixinB = _metal.Mixin.create({
        concatenatedProperties: 'bar',
        foo: ['d', 'e', 'f'],
        bar: [1, 2, 3]
      });

      let MixinC = _metal.Mixin.create({
        bar: [4, 5, 6]
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB, MixinC);
      assert.deepEqual((0, _metal.get)(obj, 'concatenatedProperties'), ['foo', 'bar'], 'get concatenatedProperties');
      assert.deepEqual((0, _metal.get)(obj, 'foo'), ['a', 'b', 'c', 'd', 'e', 'f'], 'get foo');
      assert.deepEqual((0, _metal.get)(obj, 'bar'), [1, 2, 3, 4, 5, 6], 'get bar');
    }

    ['@test adding a prop that is not an array should make array'](assert) {
      let MixinA = _metal.Mixin.create({
        concatenatedProperties: ['foo'],
        foo: [1, 2, 3]
      });

      let MixinB = _metal.Mixin.create({
        foo: 4
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), [1, 2, 3, 4]);
    }

    ['@test adding a prop that is not an array should make array'](assert) {
      let MixinA = _metal.Mixin.create({
        concatenatedProperties: ['foo'],
        foo: 'bar'
      });

      let obj = (0, _metal.mixin)({}, MixinA);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), ['bar']);
    }

    ['@test adding a non-concatenable property that already has a defined value should result in an array with both values'](assert) {
      let mixinA = _metal.Mixin.create({
        foo: 1
      });

      let mixinB = _metal.Mixin.create({
        concatenatedProperties: ['foo'],
        foo: 2
      });

      let obj = (0, _metal.mixin)({}, mixinA, mixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), [1, 2]);
    }

    ['@test adding a concatenable property that already has a defined value should result in a concatenated value'](assert) {
      let mixinA = _metal.Mixin.create({
        foobar: 'foo'
      });

      let mixinB = _metal.Mixin.create({
        concatenatedProperties: ['foobar'],
        foobar: 'bar'
      });

      let obj = (0, _metal.mixin)({}, mixinA, mixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foobar'), ['foo', 'bar']);
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/detect_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin.detect', class extends _internalTestHelpers.AbstractTestCase {
    ['@test detect() finds a directly applied mixin'](assert) {
      let MixinA = _metal.Mixin.create();
      let obj = {};

      assert.equal(MixinA.detect(obj), false, 'MixinA.detect(obj) before apply()');

      MixinA.apply(obj);
      assert.equal(MixinA.detect(obj), true, 'MixinA.detect(obj) after apply()');
    }

    ['@test detect() finds nested mixins'](assert) {
      let MixinA = _metal.Mixin.create({});
      let MixinB = _metal.Mixin.create(MixinA);
      let obj = {};

      assert.equal(MixinA.detect(obj), false, 'MixinA.detect(obj) before apply()');

      MixinB.apply(obj);
      assert.equal(MixinA.detect(obj), true, 'MixinA.detect(obj) after apply()');
    }

    ['@test detect() finds mixins on other mixins'](assert) {
      let MixinA = _metal.Mixin.create({});
      let MixinB = _metal.Mixin.create(MixinA);
      assert.equal(MixinA.detect(MixinB), true, 'MixinA is part of MixinB');
      assert.equal(MixinB.detect(MixinA), false, 'MixinB is not part of MixinA');
    }

    ['@test detect handles null values'](assert) {
      let MixinA = _metal.Mixin.create();
      assert.equal(MixinA.detect(null), false);
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/introspection_test', ['@ember/-internals/utils', '@ember/-internals/metal', 'internal-test-helpers'], function (_utils, _metal, _internalTestHelpers) {
  'use strict';

  const PrivateProperty = _metal.Mixin.create({
    _foo: '_FOO'
  }); // NOTE: A previous iteration differentiated between public and private props
  // as well as methods vs props.  We are just keeping these for testing; the
  // current impl doesn't care about the differences as much...

  const PublicProperty = _metal.Mixin.create({
    foo: 'FOO'
  });
  const PrivateMethod = _metal.Mixin.create({
    _fooMethod() {}
  });
  const PublicMethod = _metal.Mixin.create({
    fooMethod() {}
  });
  const BarProperties = _metal.Mixin.create({
    _bar: '_BAR',
    bar: 'bar'
  });
  const BarMethods = _metal.Mixin.create({
    _barMethod() {},
    barMethod() {}
  });

  const Combined = _metal.Mixin.create(BarProperties, BarMethods);

  let obj;

  (0, _internalTestHelpers.moduleFor)('Basic introspection', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = {};
      (0, _metal.mixin)(obj, PrivateProperty, PublicProperty, PrivateMethod, PublicMethod, Combined);
    }

    ['@test Ember.mixins()'](assert) {
      function mapGuids(ary) {
        return ary.map(x => (0, _utils.guidFor)(x));
      }

      assert.deepEqual(mapGuids(_metal.Mixin.mixins(obj)), mapGuids([PrivateProperty, PublicProperty, PrivateMethod, PublicMethod, Combined, BarProperties, BarMethods]), 'should return included mixins');
    }

    ['@test setting a NAME_KEY on a mixin does not error'](assert) {
      assert.expect(0);

      let instance = _metal.Mixin.create();
      instance[_utils.NAME_KEY] = 'My special name!';
    }

    ['@test setting a NAME_KEY on a mixin instance does not error'](assert) {
      assert.expect(0);

      _metal.Mixin.create({ [_utils.NAME_KEY]: 'My special name' });
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/merged_properties_test', ['@ember/-internals/runtime', '@ember/-internals/metal', 'internal-test-helpers'], function (_runtime, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin mergedProperties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test defining mergedProperties should merge future version'](assert) {
      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: { a: true, b: true, c: true }
      });

      let MixinB = _metal.Mixin.create({
        foo: { d: true, e: true, f: true }
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), {
        a: true,
        b: true,
        c: true,
        d: true,
        e: true,
        f: true
      });
    }

    ['@test defining mergedProperties on future mixin should merged into past'](assert) {
      let MixinA = _metal.Mixin.create({
        foo: { a: true, b: true, c: true }
      });

      let MixinB = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: { d: true, e: true, f: true }
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), {
        a: true,
        b: true,
        c: true,
        d: true,
        e: true,
        f: true
      });
    }

    ['@test defining mergedProperties with null properties should keep properties null'](assert) {
      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: null
      });

      let MixinB = _metal.Mixin.create({
        foo: null
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.equal((0, _metal.get)(obj, 'foo'), null);
    }

    ["@test mergedProperties' properties can get overwritten"](assert) {
      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: { a: 1 }
      });

      let MixinB = _metal.Mixin.create({
        foo: { a: 2 }
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB);
      assert.deepEqual((0, _metal.get)(obj, 'foo'), { a: 2 });
    }

    ['@test mergedProperties should be concatenated'](assert) {
      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: { a: true, b: true, c: true }
      });

      let MixinB = _metal.Mixin.create({
        mergedProperties: 'bar',
        foo: { d: true, e: true, f: true },
        bar: { a: true, l: true }
      });

      let MixinC = _metal.Mixin.create({
        bar: { e: true, x: true }
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB, MixinC);
      assert.deepEqual((0, _metal.get)(obj, 'mergedProperties'), ['foo', 'bar'], 'get mergedProperties');
      assert.deepEqual((0, _metal.get)(obj, 'foo'), { a: true, b: true, c: true, d: true, e: true, f: true }, 'get foo');
      assert.deepEqual((0, _metal.get)(obj, 'bar'), { a: true, l: true, e: true, x: true }, 'get bar');
    }

    ['@test mergedProperties should exist even if not explicitly set on create'](assert) {
      let AnObj = _runtime.Object.extend({
        mergedProperties: ['options'],
        options: {
          a: 'a',
          b: {
            c: 'ccc'
          }
        }
      });

      let obj = AnObj.create({
        options: {
          a: 'A'
        }
      });

      assert.equal((0, _metal.get)(obj, 'options').a, 'A');
      assert.equal((0, _metal.get)(obj, 'options').b.c, 'ccc');
    }

    ['@test defining mergedProperties at create time should not modify the prototype'](assert) {
      let AnObj = _runtime.Object.extend({
        mergedProperties: ['options'],
        options: {
          a: 1
        }
      });

      let objA = AnObj.create({
        options: {
          a: 2
        }
      });
      let objB = AnObj.create({
        options: {
          a: 3
        }
      });

      assert.equal((0, _metal.get)(objA, 'options').a, 2);
      assert.equal((0, _metal.get)(objB, 'options').a, 3);
    }

    ["@test mergedProperties' overwriting methods can call _super"](assert) {
      assert.expect(4);

      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: {
          meth(a) {
            assert.equal(a, 'WOOT', "_super successfully called MixinA's `foo.meth` method");
            return 'WAT';
          }
        }
      });

      let MixinB = _metal.Mixin.create({
        foo: {
          meth() {
            assert.ok(true, "MixinB's `foo.meth` method called");
            return this._super(...arguments);
          }
        }
      });

      let MixinC = _metal.Mixin.create({
        foo: {
          meth(a) {
            assert.ok(true, "MixinC's `foo.meth` method called");
            return this._super(a);
          }
        }
      });

      let obj = (0, _metal.mixin)({}, MixinA, MixinB, MixinC);
      assert.equal(obj.foo.meth('WOOT'), 'WAT');
    }

    ['@test Merging an Array should raise an error'](assert) {
      assert.expect(1);

      let MixinA = _metal.Mixin.create({
        mergedProperties: ['foo'],
        foo: { a: true, b: true, c: true }
      });

      let MixinB = _metal.Mixin.create({
        foo: ['a']
      });

      expectAssertion(() => {
        (0, _metal.mixin)({}, MixinA, MixinB);
      }, 'You passed in `["a"]` as the value for `foo` but `foo` cannot be an Array');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/method_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin Methods', class extends _internalTestHelpers.AbstractTestCase {
    ['@test defining simple methods'](assert) {
      let MixinA, obj, props;

      props = {
        publicMethod() {
          return 'publicMethod';
        },
        _privateMethod() {
          return 'privateMethod';
        }
      };

      MixinA = _metal.Mixin.create(props);
      obj = {};
      MixinA.apply(obj);

      // but should be defined
      assert.equal(props.publicMethod(), 'publicMethod', 'publicMethod is func');
      assert.equal(props._privateMethod(), 'privateMethod', 'privateMethod is func');
    }

    ['@test overriding public methods'](assert) {
      let MixinA, MixinB, MixinD, MixinF, obj;

      MixinA = _metal.Mixin.create({
        publicMethod() {
          return 'A';
        }
      });

      MixinB = _metal.Mixin.create(MixinA, {
        publicMethod() {
          return this._super(...arguments) + 'B';
        }
      });

      MixinD = _metal.Mixin.create(MixinA, {
        publicMethod() {
          return this._super(...arguments) + 'D';
        }
      });

      MixinF = _metal.Mixin.create({
        publicMethod() {
          return this._super(...arguments) + 'F';
        }
      });

      obj = {};
      MixinB.apply(obj);
      assert.equal(obj.publicMethod(), 'AB', 'should define super for A and B');

      obj = {};
      MixinD.apply(obj);
      assert.equal(obj.publicMethod(), 'AD', 'should define super for A and B');

      obj = {};
      MixinA.apply(obj);
      MixinF.apply(obj);
      assert.equal(obj.publicMethod(), 'AF', 'should define super for A and F');

      obj = {
        publicMethod() {
          return 'obj';
        }
      };
      MixinF.apply(obj);
      assert.equal(obj.publicMethod(), 'objF', 'should define super for F');
    }

    ['@test overriding inherited objects'](assert) {
      let cnt = 0;
      let MixinA = _metal.Mixin.create({
        foo() {
          cnt++;
        }
      });

      let MixinB = _metal.Mixin.create({
        foo() {
          this._super(...arguments);
          cnt++;
        }
      });

      let objA = {};
      MixinA.apply(objA);

      let objB = Object.create(objA);
      MixinB.apply(objB);

      cnt = 0;
      objB.foo();
      assert.equal(cnt, 2, 'should invoke both methods');

      cnt = 0;
      objA.foo();
      assert.equal(cnt, 1, 'should not screw w/ parent obj');
    }

    ['@test Including the same mixin more than once will only run once'](assert) {
      let cnt = 0;
      let MixinA = _metal.Mixin.create({
        foo() {
          cnt++;
        }
      });

      let MixinB = _metal.Mixin.create(MixinA, {
        foo() {
          this._super(...arguments);
        }
      });

      let MixinC = _metal.Mixin.create(MixinA, {
        foo() {
          this._super(...arguments);
        }
      });

      let MixinD = _metal.Mixin.create(MixinB, MixinC, MixinA, {
        foo() {
          this._super(...arguments);
        }
      });

      let obj = {};
      MixinD.apply(obj);
      MixinA.apply(obj); // try to apply again..

      cnt = 0;
      obj.foo();

      assert.equal(cnt, 1, 'should invoke MixinA.foo one time');
    }

    ['@test _super from a single mixin with no superclass does not error'](assert) {
      let MixinA = _metal.Mixin.create({
        foo() {
          this._super(...arguments);
        }
      });

      let obj = {};
      MixinA.apply(obj);

      obj.foo();
      assert.ok(true);
    }

    ['@test _super from a first-of-two mixins with no superclass function does not error'](assert) {
      // _super was previously calling itself in the second assertion.
      // Use remaining count of calls to ensure it doesn't loop indefinitely.
      let remaining = 3;
      let MixinA = _metal.Mixin.create({
        foo() {
          if (remaining-- > 0) {
            this._super(...arguments);
          }
        }
      });

      let MixinB = _metal.Mixin.create({
        foo() {
          this._super(...arguments);
        }
      });

      let obj = {};
      MixinA.apply(obj);
      MixinB.apply(obj);

      obj.foo();
      assert.ok(true);
    }
  });

  // ..........................................................
  // CONFLICTS
  //
  (0, _internalTestHelpers.moduleFor)('Method Conflicts', class extends _internalTestHelpers.AbstractTestCase {
    ['@test overriding toString'](assert) {
      let MixinA = _metal.Mixin.create({
        toString() {
          return 'FOO';
        }
      });

      let obj = {};
      MixinA.apply(obj);
      assert.equal(obj.toString(), 'FOO', 'should override toString w/o error');

      obj = {};
      (0, _metal.mixin)(obj, {
        toString() {
          return 'FOO';
        }
      });
      assert.equal(obj.toString(), 'FOO', 'should override toString w/o error');
    }
  });

  // ..........................................................
  // BUGS
  //
  (0, _internalTestHelpers.moduleFor)('system/mixin/method_test BUGS', class extends _internalTestHelpers.AbstractTestCase {
    ['@test applying several mixins at once with sup already defined causes infinite loop'](assert) {
      let cnt = 0;
      let MixinA = _metal.Mixin.create({
        foo() {
          cnt++;
        }
      });

      let MixinB = _metal.Mixin.create({
        foo() {
          this._super(...arguments);
          cnt++;
        }
      });

      let MixinC = _metal.Mixin.create({
        foo() {
          this._super(...arguments);
          cnt++;
        }
      });

      let obj = {};
      (0, _metal.mixin)(obj, MixinA); // sup already exists
      (0, _metal.mixin)(obj, MixinB, MixinC); // must be more than one mixin

      cnt = 0;
      obj.foo();
      assert.equal(cnt, 3, 'should invoke all 3 methods');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/observer_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin observer', class extends _internalTestHelpers.AbstractTestCase {
    ['@test global observer helper'](assert) {
      let MyMixin = _metal.Mixin.create({
        count: 0,

        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test global observer helper takes multiple params'](assert) {
      let MyMixin = _metal.Mixin.create({
        count: 0,

        foo: (0, _metal.observer)('bar', 'baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      (0, _metal.set)(obj, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 2, 'should invoke observer after change');
    }

    ['@test replacing observer should remove old observer'](assert) {
      let MyMixin = _metal.Mixin.create({
        count: 0,

        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let Mixin2 = _metal.Mixin.create({
        foo: (0, _metal.observer)('baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 10);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin, Mixin2);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer after change');

      (0, _metal.set)(obj, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 10, 'should invoke observer after change');
    }

    ['@test observing chain with property before'](assert) {
      let obj2 = { baz: 'baz' };

      let MyMixin = _metal.Mixin.create({
        count: 0,
        bar: obj2,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with property after'](assert) {
      let obj2 = { baz: 'baz' };

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        }),
        bar: obj2
      });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with property in mixin applied later'](assert) {
      let obj2 = { baz: 'baz' };

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let MyMixin2 = _metal.Mixin.create({ bar: obj2 });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      MyMixin2.apply(obj);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with existing property'](assert) {
      let obj2 = { baz: 'baz' };

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({ bar: obj2 }, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with property in mixin before'](assert) {
      let obj2 = { baz: 'baz' };
      let MyMixin2 = _metal.Mixin.create({ bar: obj2 });

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin2, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with property in mixin after'](assert) {
      let obj2 = { baz: 'baz' };
      let MyMixin2 = _metal.Mixin.create({ bar: obj2 });

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({}, MyMixin, MyMixin2);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observing chain with overridden property'](assert) {
      let obj2 = { baz: 'baz' };
      let obj3 = { baz: 'foo' };

      let MyMixin2 = _metal.Mixin.create({ bar: obj3 });

      let MyMixin = _metal.Mixin.create({
        count: 0,
        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = (0, _metal.mixin)({ bar: obj2 }, MyMixin, MyMixin2);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      assert.equal((0, _metal.isWatching)(obj2, 'baz'), false, 'should not be watching baz');
      assert.equal((0, _metal.isWatching)(obj3, 'baz'), true, 'should be watching baz');

      (0, _metal.set)(obj2, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer after change');

      (0, _metal.set)(obj3, 'baz', 'BEAR');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/reopen_test', ['@ember/-internals/runtime', '@ember/-internals/metal', '@ember/runloop', 'internal-test-helpers'], function (_runtime, _metal, _runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Mixin#reopen', class extends _internalTestHelpers.AbstractTestCase {
    ['@test using reopen() to add more properties to a simple'](assert) {
      let MixinA = _metal.Mixin.create({ foo: 'FOO', baz: 'BAZ' });
      MixinA.reopen({ bar: 'BAR', foo: 'FOO2' });
      let obj = {};
      MixinA.apply(obj);

      assert.equal((0, _metal.get)(obj, 'foo'), 'FOO2', 'mixin() should override');
      assert.equal((0, _metal.get)(obj, 'baz'), 'BAZ', 'preserve MixinA props');
      assert.equal((0, _metal.get)(obj, 'bar'), 'BAR', 'include MixinB props');
    }

    ['@test using reopen() and calling _super where there is not a super function does not cause infinite recursion'](assert) {
      let Taco = _runtime.Object.extend({
        createBreakfast() {
          // There is no original createBreakfast function.
          // Calling the wrapped _super function here
          // used to end in an infinite call loop
          this._super(...arguments);
          return 'Breakfast!';
        }
      });

      Taco.reopen({
        createBreakfast() {
          return this._super(...arguments);
        }
      });

      let taco = Taco.create();

      let result;
      (0, _runloop.run)(() => {
        try {
          result = taco.createBreakfast();
        } catch (e) {
          result = 'Your breakfast was interrupted by an infinite stack error.';
        }
      });

      assert.equal(result, 'Breakfast!');
    }
  });
});
enifed('@ember/-internals/metal/tests/mixin/without_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('without', class extends _internalTestHelpers.AbstractTestCase {
    ['@test without should create a new mixin excluding named properties'](assert) {
      let MixinA = _metal.Mixin.create({
        foo: 'FOO',
        bar: 'BAR'
      });

      let MixinB = MixinA.without('bar');

      let obj = {};
      MixinB.apply(obj);

      assert.equal(obj.foo, 'FOO', 'should defined foo');
      assert.equal(obj.bar, undefined, 'should not define bar');
    }
  });
});
enifed('@ember/-internals/metal/tests/namespace_search_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('NamespaceSearch', class extends _internalTestHelpers.AbstractTestCase {
    ['@test classToString: null as this inside class must not throw error'](assert) {
      let mixin = _metal.Mixin.create();
      assert.equal(mixin.toString(), '(unknown)', 'this = null should be handled on Mixin.toString() call');
    }
  });
});
enifed('@ember/-internals/metal/tests/observer_test', ['@ember/-internals/environment', '@ember/-internals/metal', 'internal-test-helpers'], function (_environment, _metal, _internalTestHelpers) {
  'use strict';

  function K() {}

  // ..........................................................
  // ADD OBSERVER
  //

  (0, _internalTestHelpers.moduleFor)('addObserver', class extends _internalTestHelpers.AbstractTestCase {
    ['@test observer should assert to invalid input']() {
      expectAssertion(() => {
        (0, _metal.observer)(() => {});
      }, 'observer called without valid path');

      expectAssertion(() => {
        (0, _metal.observer)(null);
      }, 'observer called without a function');
    }

    ['@test observer should fire when property is modified'](assert) {
      let obj = {};
      let count = 0;

      (0, _metal.addObserver)(obj, 'foo', function () {
        assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'should invoke AFTER value changed');
        count++;
      });

      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(count, 1, 'should have invoked observer');
    }

    ['@test observer should fire when dependent property is modified'](assert) {
      let obj = { bar: 'bar' };
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        return (0, _metal.get)(this, 'bar').toUpperCase();
      }).property('bar'));

      (0, _metal.get)(obj, 'foo');

      let count = 0;
      (0, _metal.addObserver)(obj, 'foo', function () {
        assert.equal((0, _metal.get)(obj, 'foo'), 'BAZ', 'should have invoked after prop change');
        count++;
      });

      (0, _metal.set)(obj, 'bar', 'baz');
      assert.equal(count, 1, 'should have invoked observer');
    }

    ['@test observer should continue to fire after dependent properties are accessed'](assert) {
      let observerCount = 0;
      let obj = {};

      (0, _metal.defineProperty)(obj, 'prop', (0, _metal.computed)(function () {
        return Math.random();
      }));
      (0, _metal.defineProperty)(obj, 'anotherProp', (0, _metal.computed)('prop', function () {
        return (0, _metal.get)(this, 'prop') + Math.random();
      }));

      (0, _metal.addObserver)(obj, 'prop', function () {
        observerCount++;
      });

      (0, _metal.get)(obj, 'anotherProp');

      for (let i = 0; i < 10; i++) {
        (0, _metal.notifyPropertyChange)(obj, 'prop');
      }

      assert.equal(observerCount, 10, 'should continue to fire indefinitely');
    }

    ['@test observer added declaratively via brace expansion should fire when property changes'](assert) {
      if (_environment.ENV.EXTEND_PROTOTYPES.Function) {
        let obj = {};
        let count = 0;

        (0, _metal.mixin)(obj, {
          observeFooAndBar: function () {
            count++;
          }.observes('{foo,bar}')
        });

        (0, _metal.set)(obj, 'foo', 'foo');
        assert.equal(count, 1, 'observer specified via brace expansion invoked on property change');

        (0, _metal.set)(obj, 'bar', 'bar');
        assert.equal(count, 2, 'observer specified via brace expansion invoked on property change');

        (0, _metal.set)(obj, 'baz', 'baz');
        assert.equal(count, 2, 'observer not invoked on unspecified property');
      } else {
        assert.expect(0);
      }
    }

    ['@test observer specified declaratively via brace expansion should fire when dependent property changes'](assert) {
      if (_environment.ENV.EXTEND_PROTOTYPES.Function) {
        let obj = { baz: 'Initial' };
        let count = 0;

        (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
          return (0, _metal.get)(this, 'bar').toLowerCase();
        }).property('bar'));

        (0, _metal.defineProperty)(obj, 'bar', (0, _metal.computed)(function () {
          return (0, _metal.get)(this, 'baz').toUpperCase();
        }).property('baz'));

        (0, _metal.mixin)(obj, {
          fooAndBarWatcher: function () {
            count++;
          }.observes('{foo,bar}')
        });

        (0, _metal.get)(obj, 'foo');
        (0, _metal.set)(obj, 'baz', 'Baz');
        // fire once for foo, once for bar
        assert.equal(count, 2, 'observer specified via brace expansion invoked on dependent property change');

        (0, _metal.set)(obj, 'quux', 'Quux');
        assert.equal(count, 2, 'observer not fired on unspecified property');
      } else {
        assert.expect(0);
      }
    }

    ['@test observers watching multiple properties via brace expansion should fire when the properties change'](assert) {
      let obj = {};
      let count = 0;

      (0, _metal.mixin)(obj, {
        observeFooAndBar: (0, _metal.observer)('{foo,bar}', function () {
          count++;
        })
      });

      (0, _metal.set)(obj, 'foo', 'foo');
      assert.equal(count, 1, 'observer specified via brace expansion invoked on property change');

      (0, _metal.set)(obj, 'bar', 'bar');
      assert.equal(count, 2, 'observer specified via brace expansion invoked on property change');

      (0, _metal.set)(obj, 'baz', 'baz');
      assert.equal(count, 2, 'observer not invoked on unspecified property');
    }

    ['@test observers watching multiple properties via brace expansion should fire when dependent properties change'](assert) {
      let obj = { baz: 'Initial' };
      let count = 0;

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)(function () {
        return (0, _metal.get)(this, 'bar').toLowerCase();
      }).property('bar'));

      (0, _metal.defineProperty)(obj, 'bar', (0, _metal.computed)(function () {
        return (0, _metal.get)(this, 'baz').toUpperCase();
      }).property('baz'));

      (0, _metal.mixin)(obj, {
        fooAndBarWatcher: (0, _metal.observer)('{foo,bar}', function () {
          count++;
        })
      });

      (0, _metal.get)(obj, 'foo');
      (0, _metal.set)(obj, 'baz', 'Baz');
      // fire once for foo, once for bar
      assert.equal(count, 2, 'observer specified via brace expansion invoked on dependent property change');

      (0, _metal.set)(obj, 'quux', 'Quux');
      assert.equal(count, 2, 'observer not fired on unspecified property');
    }

    ['@test nested observers should fire in order'](assert) {
      let obj = { foo: 'foo', bar: 'bar' };
      let fooCount = 0;
      let barCount = 0;

      (0, _metal.addObserver)(obj, 'foo', function () {
        fooCount++;
      });
      (0, _metal.addObserver)(obj, 'bar', function () {
        (0, _metal.set)(obj, 'foo', 'BAZ');
        assert.equal(fooCount, 1, 'fooCount should have fired already');
        barCount++;
      });

      (0, _metal.set)(obj, 'bar', 'BIFF');
      assert.equal(barCount, 1, 'barCount should have fired');
      assert.equal(fooCount, 1, 'foo should have fired');
    }

    ['@test removing an chain observer on change should not fail'](assert) {
      let foo = { bar: 'bar' };
      let obj1 = { foo: foo };
      let obj2 = { foo: foo };
      let obj3 = { foo: foo };
      let obj4 = { foo: foo };
      let count1 = 0;
      let count2 = 0;
      let count3 = 0;
      let count4 = 0;

      function observer1() {
        count1++;
      }
      function observer2() {
        count2++;
      }
      function observer3() {
        count3++;
        (0, _metal.removeObserver)(obj1, 'foo.bar', observer1);
        (0, _metal.removeObserver)(obj2, 'foo.bar', observer2);
        (0, _metal.removeObserver)(obj4, 'foo.bar', observer4);
      }
      function observer4() {
        count4++;
      }

      (0, _metal.addObserver)(obj1, 'foo.bar', observer1);
      (0, _metal.addObserver)(obj2, 'foo.bar', observer2);
      (0, _metal.addObserver)(obj3, 'foo.bar', observer3);
      (0, _metal.addObserver)(obj4, 'foo.bar', observer4);

      (0, _metal.set)(foo, 'bar', 'baz');

      assert.equal(count1, 1, 'observer1 fired');
      assert.equal(count2, 1, 'observer2 fired');
      assert.equal(count3, 1, 'observer3 fired');
      assert.equal(count4, 0, 'observer4 did not fire');
    }

    ['@test deferring property change notifications'](assert) {
      let obj = { foo: 'foo' };
      let fooCount = 0;

      (0, _metal.addObserver)(obj, 'foo', function () {
        fooCount++;
      });

      (0, _metal.beginPropertyChanges)();
      (0, _metal.set)(obj, 'foo', 'BIFF');
      (0, _metal.set)(obj, 'foo', 'BAZ');
      (0, _metal.endPropertyChanges)();

      assert.equal(fooCount, 1, 'foo should have fired once');
    }

    ['@test deferring property change notifications safely despite exceptions'](assert) {
      let obj = { foo: 'foo' };
      let fooCount = 0;
      let exc = new Error('Something unexpected happened!');

      assert.expect(2);
      (0, _metal.addObserver)(obj, 'foo', function () {
        fooCount++;
      });

      try {
        (0, _metal.changeProperties)(function () {
          (0, _metal.set)(obj, 'foo', 'BIFF');
          (0, _metal.set)(obj, 'foo', 'BAZ');
          throw exc;
        });
      } catch (err) {
        if (err !== exc) {
          throw err;
        }
      }

      assert.equal(fooCount, 1, 'foo should have fired once');

      (0, _metal.changeProperties)(function () {
        (0, _metal.set)(obj, 'foo', 'BIFF2');
        (0, _metal.set)(obj, 'foo', 'BAZ2');
      });

      assert.equal(fooCount, 2, 'foo should have fired again once');
    }

    ['@test addObserver should propagate through prototype'](assert) {
      let obj = { foo: 'foo', count: 0 };
      let obj2;

      (0, _metal.addObserver)(obj, 'foo', function () {
        this.count++;
      });
      obj2 = Object.create(obj);

      (0, _metal.set)(obj2, 'foo', 'bar');

      assert.equal(obj2.count, 1, 'should have invoked observer on inherited');
      assert.equal(obj.count, 0, 'should not have invoked observer on parent');

      obj2.count = 0;
      (0, _metal.set)(obj, 'foo', 'baz');
      assert.equal(obj.count, 0, 'should not have invoked observer on parent');
      assert.equal(obj2.count, 0, 'should not have invoked observer on inherited');
    }

    ['@test addObserver should respect targets with methods'](assert) {
      let observed = { foo: 'foo' };

      let target1 = {
        count: 0,

        didChange(obj, keyName) {
          let value = (0, _metal.get)(obj, keyName);
          assert.equal(this, target1, 'should invoke with this');
          assert.equal(obj, observed, 'param1 should be observed object');
          assert.equal(keyName, 'foo', 'param2 should be keyName');
          assert.equal(value, 'BAZ', 'param3 should new value');
          this.count++;
        }
      };

      let target2 = {
        count: 0,

        didChange(obj, keyName) {
          let value = (0, _metal.get)(obj, keyName);
          assert.equal(this, target2, 'should invoke with this');
          assert.equal(obj, observed, 'param1 should be observed object');
          assert.equal(keyName, 'foo', 'param2 should be keyName');
          assert.equal(value, 'BAZ', 'param3 should new value');
          this.count++;
        }
      };

      (0, _metal.addObserver)(observed, 'foo', target1, 'didChange');
      (0, _metal.addObserver)(observed, 'foo', target2, target2.didChange);

      (0, _metal.set)(observed, 'foo', 'BAZ');
      assert.equal(target1.count, 1, 'target1 observer should have fired');
      assert.equal(target2.count, 1, 'target2 observer should have fired');
    }

    ['@test addObserver should allow multiple objects to observe a property'](assert) {
      let observed = { foo: 'foo' };

      let target1 = {
        count: 0,

        didChange() {
          this.count++;
        }
      };

      let target2 = {
        count: 0,

        didChange() {
          this.count++;
        }
      };

      (0, _metal.addObserver)(observed, 'foo', target1, 'didChange');
      (0, _metal.addObserver)(observed, 'foo', target2, 'didChange');

      (0, _metal.set)(observed, 'foo', 'BAZ');
      assert.equal(target1.count, 1, 'target1 observer should have fired');
      assert.equal(target2.count, 1, 'target2 observer should have fired');
    }
  });

  // ..........................................................
  // REMOVE OBSERVER
  //

  (0, _internalTestHelpers.moduleFor)('removeObserver', class extends _internalTestHelpers.AbstractTestCase {
    ['@test removing observer should stop firing'](assert) {
      let obj = {};
      let count = 0;
      function F() {
        count++;
      }
      (0, _metal.addObserver)(obj, 'foo', F);

      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(count, 1, 'should have invoked observer');

      (0, _metal.removeObserver)(obj, 'foo', F);

      (0, _metal.set)(obj, 'foo', 'baz');
      assert.equal(count, 1, "removed observer shouldn't fire");
    }

    ['@test local observers can be removed'](assert) {
      let barObserved = 0;

      let MyMixin = _metal.Mixin.create({
        foo1: (0, _metal.observer)('bar', function () {
          barObserved++;
        }),

        foo2: (0, _metal.observer)('bar', function () {
          barObserved++;
        })
      });

      let obj = {};
      MyMixin.apply(obj);

      (0, _metal.set)(obj, 'bar', 'HI!');
      assert.equal(barObserved, 2, 'precond - observers should be fired');

      (0, _metal.removeObserver)(obj, 'bar', null, 'foo1');

      barObserved = 0;
      (0, _metal.set)(obj, 'bar', 'HI AGAIN!');

      assert.equal(barObserved, 1, 'removed observers should not be called');
    }

    ['@test removeObserver should respect targets with methods'](assert) {
      let observed = { foo: 'foo' };

      let target1 = {
        count: 0,

        didChange() {
          this.count++;
        }
      };

      let target2 = {
        count: 0,

        didChange() {
          this.count++;
        }
      };

      (0, _metal.addObserver)(observed, 'foo', target1, 'didChange');
      (0, _metal.addObserver)(observed, 'foo', target2, target2.didChange);

      (0, _metal.set)(observed, 'foo', 'BAZ');
      assert.equal(target1.count, 1, 'target1 observer should have fired');
      assert.equal(target2.count, 1, 'target2 observer should have fired');

      (0, _metal.removeObserver)(observed, 'foo', target1, 'didChange');
      (0, _metal.removeObserver)(observed, 'foo', target2, target2.didChange);

      target1.count = target2.count = 0;
      (0, _metal.set)(observed, 'foo', 'BAZ');
      assert.equal(target1.count, 0, 'target1 observer should not fire again');
      assert.equal(target2.count, 0, 'target2 observer should not fire again');
    }
  });

  // ..........................................................
  // CHAINED OBSERVERS
  //

  let obj, count;

  (0, _internalTestHelpers.moduleFor)('addObserver - dependentkey with chained properties', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = {
        foo: {
          bar: {
            baz: {
              biff: 'BIFF'
            }
          }
        },
        Capital: {
          foo: {
            bar: {
              baz: {
                biff: 'BIFF'
              }
            }
          }
        }
      };

      count = 0;
    }

    afterEach() {
      obj = count = null;
    }

    ['@test depending on a chain with a computed property'](assert) {
      (0, _metal.defineProperty)(obj, 'computed', (0, _metal.computed)(function () {
        return { foo: 'bar' };
      }));

      let changed = 0;
      (0, _metal.addObserver)(obj, 'computed.foo', function () {
        changed++;
      });

      assert.equal((0, _metal.getCachedValueFor)(obj, 'computed'), undefined, 'addObserver should not compute CP');

      (0, _metal.set)(obj, 'computed.foo', 'baz');

      assert.equal(changed, 1, 'should fire observer');
    }

    ['@test depending on a simple chain'](assert) {
      let val;
      (0, _metal.addObserver)(obj, 'foo.bar.baz.biff', function (target, key) {
        val = (0, _metal.get)(target, key);
        count++;
      });

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal(val, 'BUZZ');
      assert.equal(count, 1);

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar'), 'baz', { biff: 'BLARG' });
      assert.equal(val, 'BLARG');
      assert.equal(count, 2);

      (0, _metal.set)((0, _metal.get)(obj, 'foo'), 'bar', { baz: { biff: 'BOOM' } });
      assert.equal(val, 'BOOM');
      assert.equal(count, 3);

      (0, _metal.set)(obj, 'foo', { bar: { baz: { biff: 'BLARG' } } });
      assert.equal(val, 'BLARG');
      assert.equal(count, 4);

      (0, _metal.set)((0, _metal.get)(obj, 'foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal(val, 'BUZZ');
      assert.equal(count, 5);

      let foo = (0, _metal.get)(obj, 'foo');

      (0, _metal.set)(obj, 'foo', 'BOO');
      assert.equal(val, undefined);
      assert.equal(count, 6);

      (0, _metal.set)(foo.bar.baz, 'biff', 'BOOM');
      assert.equal(count, 6, 'should be not have invoked observer');
    }

    ['@test depending on a chain with a capitalized first key'](assert) {
      let val;

      (0, _metal.addObserver)(obj, 'Capital.foo.bar.baz.biff', function (target, key) {
        val = (0, _metal.get)(obj, key);
        count++;
      });

      (0, _metal.set)((0, _metal.get)(obj, 'Capital.foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal(val, 'BUZZ');
      assert.equal(count, 1);

      (0, _metal.set)((0, _metal.get)(obj, 'Capital.foo.bar'), 'baz', { biff: 'BLARG' });
      assert.equal(val, 'BLARG');
      assert.equal(count, 2);

      (0, _metal.set)((0, _metal.get)(obj, 'Capital.foo'), 'bar', { baz: { biff: 'BOOM' } });
      assert.equal(val, 'BOOM');
      assert.equal(count, 3);

      (0, _metal.set)(obj, 'Capital.foo', { bar: { baz: { biff: 'BLARG' } } });
      assert.equal(val, 'BLARG');
      assert.equal(count, 4);

      (0, _metal.set)((0, _metal.get)(obj, 'Capital.foo.bar.baz'), 'biff', 'BUZZ');
      assert.equal(val, 'BUZZ');
      assert.equal(count, 5);

      let foo = (0, _metal.get)(obj, 'foo');

      (0, _metal.set)(obj, 'Capital.foo', 'BOO');
      assert.equal(val, undefined);
      assert.equal(count, 6);

      (0, _metal.set)(foo.bar.baz, 'biff', 'BOOM');
      assert.equal(count, 6, 'should be not have invoked observer');
    }
  });

  // ..........................................................
  // SETTING IDENTICAL VALUES
  //

  (0, _internalTestHelpers.moduleFor)('props/observer_test - setting identical values', class extends _internalTestHelpers.AbstractTestCase {
    ['@test setting simple prop should not trigger'](assert) {
      let obj = { foo: 'bar' };
      let count = 0;

      (0, _metal.addObserver)(obj, 'foo', function () {
        count++;
      });

      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(count, 0, 'should not trigger observer');

      (0, _metal.set)(obj, 'foo', 'baz');
      assert.equal(count, 1, 'should trigger observer');

      (0, _metal.set)(obj, 'foo', 'baz');
      assert.equal(count, 1, 'should not trigger observer again');
    }

    // The issue here is when a computed property is directly set with a value, then has a
    // dependent key change (which triggers a cache expiration and recomputation), observers will
    // not be fired if the CP setter is called with the last set value.
    ['@test setting a cached computed property whose value has changed should trigger'](assert) {
      let obj = {};

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get: function () {
          return (0, _metal.get)(this, 'baz');
        },
        set: function (key, value) {
          return value;
        }
      }).property('baz'));

      let count = 0;

      (0, _metal.addObserver)(obj, 'foo', function () {
        count++;
      });

      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(count, 1);
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar');

      (0, _metal.set)(obj, 'baz', 'qux');
      assert.equal(count, 2);
      assert.equal((0, _metal.get)(obj, 'foo'), 'qux');

      (0, _metal.get)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(count, 3);
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar');
    }
  });

  (0, _internalTestHelpers.moduleFor)('changeProperties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test observers added/removed during changeProperties should do the right thing.'](assert) {
      let obj = {
        foo: 0
      };
      function Observer() {
        this.didChangeCount = 0;
      }
      Observer.prototype = {
        add() {
          (0, _metal.addObserver)(obj, 'foo', this, 'didChange');
        },
        remove() {
          (0, _metal.removeObserver)(obj, 'foo', this, 'didChange');
        },
        didChange() {
          this.didChangeCount++;
        }
      };
      let addedBeforeFirstChangeObserver = new Observer();
      let addedAfterFirstChangeObserver = new Observer();
      let addedAfterLastChangeObserver = new Observer();
      let removedBeforeFirstChangeObserver = new Observer();
      let removedBeforeLastChangeObserver = new Observer();
      let removedAfterLastChangeObserver = new Observer();
      removedBeforeFirstChangeObserver.add();
      removedBeforeLastChangeObserver.add();
      removedAfterLastChangeObserver.add();
      (0, _metal.changeProperties)(function () {
        removedBeforeFirstChangeObserver.remove();
        addedBeforeFirstChangeObserver.add();

        (0, _metal.set)(obj, 'foo', 1);

        assert.equal(addedBeforeFirstChangeObserver.didChangeCount, 0, 'addObserver called before the first change is deferred');

        addedAfterFirstChangeObserver.add();
        removedBeforeLastChangeObserver.remove();

        (0, _metal.set)(obj, 'foo', 2);

        assert.equal(addedAfterFirstChangeObserver.didChangeCount, 0, 'addObserver called after the first change is deferred');

        addedAfterLastChangeObserver.add();
        removedAfterLastChangeObserver.remove();
      });

      assert.equal(removedBeforeFirstChangeObserver.didChangeCount, 0, 'removeObserver called before the first change sees none');
      assert.equal(addedBeforeFirstChangeObserver.didChangeCount, 1, 'addObserver called before the first change sees only 1');
      assert.equal(addedAfterFirstChangeObserver.didChangeCount, 1, 'addObserver called after the first change sees 1');
      assert.equal(addedAfterLastChangeObserver.didChangeCount, 1, 'addObserver called after the last change sees 1');
      assert.equal(removedBeforeLastChangeObserver.didChangeCount, 0, 'removeObserver called before the last change sees none');
      assert.equal(removedAfterLastChangeObserver.didChangeCount, 0, 'removeObserver called after the last change sees none');
    }

    ['@test calling changeProperties while executing deferred observers works correctly'](assert) {
      let obj = { foo: 0 };
      let fooDidChange = 0;

      (0, _metal.addObserver)(obj, 'foo', () => {
        fooDidChange++;
        (0, _metal.changeProperties)(() => {});
      });

      (0, _metal.changeProperties)(() => {
        (0, _metal.set)(obj, 'foo', 1);
      });

      assert.equal(fooDidChange, 1);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Keys behavior with observers', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should not leak properties on the prototype'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();

      (0, _metal.addObserver)(beer, 'type', K);
      assert.deepEqual(Object.keys(beer), []);
      (0, _metal.removeObserver)(beer, 'type', K);
    }

    ['@test observing a non existent property'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();

      (0, _metal.addObserver)(beer, 'brand', K);

      assert.deepEqual(Object.keys(beer), []);

      (0, _metal.set)(beer, 'brand', 'Corona');
      assert.deepEqual(Object.keys(beer), ['brand']);

      (0, _metal.removeObserver)(beer, 'brand', K);
    }

    ['@test with observers switched on and off'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();

      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);

      assert.deepEqual(Object.keys(beer), []);
    }

    ['@test observers switched on and off with setter in between'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();

      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.set)(beer, 'type', 'ale');
      (0, _metal.removeObserver)(beer, 'type', K);

      assert.deepEqual(Object.keys(beer), ['type']);
    }

    ['@test observer switched on and off and then setter'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();

      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);
      (0, _metal.set)(beer, 'type', 'ale');

      assert.deepEqual(Object.keys(beer), ['type']);
    }

    ['@test observers switched on and off with setter in between (observed property is not shadowing)'](assert) {
      function Beer() {}

      let beer = new Beer();
      (0, _metal.set)(beer, 'type', 'ale');
      assert.deepEqual(Object.keys(beer), ['type'], 'only set');

      let otherBeer = new Beer();
      (0, _metal.addObserver)(otherBeer, 'type', K);
      (0, _metal.set)(otherBeer, 'type', 'ale');
      assert.deepEqual(Object.keys(otherBeer), ['type'], 'addObserver -> set');

      let yetAnotherBeer = new Beer();
      (0, _metal.addObserver)(yetAnotherBeer, 'type', K);
      (0, _metal.set)(yetAnotherBeer, 'type', 'ale');
      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);
      assert.deepEqual(Object.keys(yetAnotherBeer), ['type'], 'addObserver -> set -> removeObserver');

      let itsMyLastBeer = new Beer();
      (0, _metal.set)(itsMyLastBeer, 'type', 'ale');
      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);
      assert.deepEqual(Object.keys(itsMyLastBeer), ['type'], 'set -> removeObserver');
    }

    ['@test observers switched on and off with setter in between (observed property is shadowing one on the prototype)'](assert) {
      function Beer() {}
      Beer.prototype.type = 'ipa';

      let beer = new Beer();
      (0, _metal.set)(beer, 'type', 'ale');
      assert.deepEqual(Object.keys(beer), ['type'], 'after set');

      let otherBeer = new Beer();
      (0, _metal.addObserver)(otherBeer, 'type', K);
      (0, _metal.set)(otherBeer, 'type', 'ale');
      assert.deepEqual(Object.keys(otherBeer), ['type'], 'addObserver -> set');

      let yetAnotherBeer = new Beer();
      (0, _metal.addObserver)(yetAnotherBeer, 'type', K);
      (0, _metal.set)(yetAnotherBeer, 'type', 'ale');
      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);
      assert.deepEqual(Object.keys(yetAnotherBeer), ['type'], 'addObserver -> set -> removeObserver');

      let itsMyLastBeer = new Beer();
      (0, _metal.set)(itsMyLastBeer, 'type', 'ale');
      (0, _metal.addObserver)(beer, 'type', K);
      (0, _metal.removeObserver)(beer, 'type', K);
      assert.deepEqual(Object.keys(itsMyLastBeer), ['type'], 'set -> removeObserver');
    }
  });
});
enifed('@ember/-internals/metal/tests/performance_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  /*
    This test file is designed to capture performance regressions related to
    deferred computation. Things like run loops, computed properties, and bindings
    should run the minimum amount of times to achieve best performance, so any
    bugs that cause them to get evaluated more than necessary should be put here.
  */

  (0, _internalTestHelpers.moduleFor)('Computed Properties - Number of times evaluated', class extends _internalTestHelpers.AbstractTestCase {
    ['@test computed properties that depend on multiple properties should run only once per run loop'](assert) {
      let obj = { a: 'a', b: 'b', c: 'c' };
      let cpCount = 0;
      let obsCount = 0;

      (0, _metal.defineProperty)(obj, 'abc', (0, _metal.computed)(function (key) {
        cpCount++;
        return 'computed ' + key;
      }).property('a', 'b', 'c'));

      (0, _metal.get)(obj, 'abc');

      cpCount = 0;

      (0, _metal.addObserver)(obj, 'abc', function () {
        obsCount++;
      });

      (0, _metal.beginPropertyChanges)();
      (0, _metal.set)(obj, 'a', 'aa');
      (0, _metal.set)(obj, 'b', 'bb');
      (0, _metal.set)(obj, 'c', 'cc');
      (0, _metal.endPropertyChanges)();

      (0, _metal.get)(obj, 'abc');

      assert.equal(cpCount, 1, 'The computed property is only invoked once');
      assert.equal(obsCount, 1, 'The observer is only invoked once');
    }

    ['@test computed properties are not executed if they are the last segment of an observer chain pain'](assert) {
      let foo = { bar: { baz: {} } };

      let count = 0;

      (0, _metal.defineProperty)(foo.bar.baz, 'bam', (0, _metal.computed)(function () {
        count++;
      }));

      (0, _metal.addObserver)(foo, 'bar.baz.bam', function () {});

      (0, _metal.notifyPropertyChange)((0, _metal.get)(foo, 'bar.baz'), 'bam');

      assert.equal(count, 0, 'should not have recomputed property');
    }
  });
});
enifed('@ember/-internals/metal/tests/properties_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('defineProperty', class extends _internalTestHelpers.AbstractTestCase {
    ['@test toString'](assert) {
      let obj = {};
      (0, _metal.defineProperty)(obj, 'toString', undefined, function () {
        return 'FOO';
      });
      assert.equal(obj.toString(), 'FOO', 'should replace toString');
    }

    ['@test for data properties, didDefineProperty hook should be called if implemented'](assert) {
      assert.expect(2);

      let obj = {
        didDefineProperty(obj, keyName, value) {
          assert.equal(keyName, 'foo', 'key name should be foo');
          assert.equal(value, 'bar', 'value should be bar');
        }
      };

      (0, _metal.defineProperty)(obj, 'foo', undefined, 'bar');
    }

    ['@test for computed properties, didDefineProperty hook should be called if implemented'](assert) {
      assert.expect(2);

      let computedProperty = (0, _metal.computed)(function () {
        return this;
      });

      let obj = {
        didDefineProperty(obj, keyName, value) {
          assert.equal(keyName, 'foo', 'key name should be foo');
          assert.strictEqual(value, computedProperty, 'value should be passed as computed property');
        }
      };

      (0, _metal.defineProperty)(obj, 'foo', computedProperty);
    }

    ['@test for descriptor properties, didDefineProperty hook should be called if implemented'](assert) {
      assert.expect(2);

      let descriptor = {
        writable: true,
        configurable: false,
        enumerable: true,
        value: 42
      };

      let obj = {
        didDefineProperty(obj, keyName, value) {
          assert.equal(keyName, 'answer', 'key name should be answer');
          assert.strictEqual(value, descriptor, 'value should be passed as descriptor');
        }
      };

      (0, _metal.defineProperty)(obj, 'answer', descriptor);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember.deprecateProperty', class extends _internalTestHelpers.AbstractTestCase {
    ['@test enables access to deprecated property and returns the value of the new property'](assert) {
      assert.expect(3);
      let obj = { foo: 'bar' };

      (0, _metal.deprecateProperty)(obj, 'baz', 'foo', { id: 'baz-deprecation', until: 'some.version' });

      expectDeprecation();
      assert.equal(obj.baz, obj.foo, 'baz and foo are equal');

      obj.foo = 'blammo';
      assert.equal(obj.baz, obj.foo, 'baz and foo are equal');
    }

    ['@test deprecatedKey is not enumerable'](assert) {
      assert.expect(2);
      let obj = { foo: 'bar', blammo: 'whammy' };

      (0, _metal.deprecateProperty)(obj, 'baz', 'foo', { id: 'baz-deprecation', until: 'some.version' });

      for (let prop in obj) {
        if (obj.hasOwnProperty(prop)) {
          assert.notEqual(prop, 'baz');
        }
      }
    }

    ['@test enables setter to deprecated property and updates the value of the new property'](assert) {
      assert.expect(3);
      let obj = { foo: 'bar' };

      (0, _metal.deprecateProperty)(obj, 'baz', 'foo', { id: 'baz-deprecation', until: 'some.version' });

      expectDeprecation();
      obj.baz = 'bloop';
      assert.equal(obj.foo, 'bloop', 'updating baz updates foo');
      assert.equal(obj.baz, obj.foo, 'baz and foo are equal');
    }
  });
});
enifed('@ember/-internals/metal/tests/property_did_change_hook', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('PROPERTY_DID_CHANGE', class extends _internalTestHelpers.AbstractTestCase {
    ['@test alias and cp'](assert) {
      let counts = {};
      let obj = {
        child: {},
        [_metal.PROPERTY_DID_CHANGE](keyName) {
          counts[keyName] = (counts[keyName] || 0) + 1;
        }
      };

      (0, _metal.defineProperty)(obj, 'cost', (0, _metal.alias)('child.cost'));
      (0, _metal.defineProperty)(obj, 'tax', (0, _metal.alias)('child.tax'));

      (0, _metal.defineProperty)(obj, 'total', (0, _metal.computed)('cost', 'tax', {
        get() {
          return (0, _metal.get)(this, 'cost') + (0, _metal.get)(this, 'tax');
        }
      }));

      assert.ok(!(0, _metal.isWatching)(obj, 'child.cost'), 'precond alias target `child.cost` is not watched');
      assert.equal((0, _metal.get)(obj, 'cost'), undefined);
      // this is how PROPERTY_DID_CHANGE will get notified
      assert.ok((0, _metal.isWatching)(obj, 'child.cost'), 'alias target `child.cost` is watched after consumption');

      assert.ok(!(0, _metal.isWatching)(obj, 'child.tax'), 'precond alias target `child.tax` is not watched');
      assert.equal((0, _metal.get)(obj, 'tax'), undefined);
      // this is how PROPERTY_DID_CHANGE will get notified
      assert.ok((0, _metal.isWatching)(obj, 'child.tax'), 'alias target `child.cost` is watched after consumption');

      // increments the watching count on the alias itself to 1
      assert.ok(isNaN((0, _metal.get)(obj, 'total')), 'total is initialized');

      // decrements the watching count on the alias itself to 0
      (0, _metal.set)(obj, 'child', {
        cost: 399.0,
        tax: 32.93
      });

      // this should have called PROPERTY_DID_CHANGE for all of them
      assert.equal(counts['cost'], 1, 'PROPERTY_DID_CHANGE called with cost');
      assert.equal(counts['tax'], 1, 'PROPERTY_DID_CHANGE called with tax');
      assert.equal(counts['total'], 1, 'PROPERTY_DID_CHANGE called with total');

      // we should still have a dependency installed
      assert.ok((0, _metal.isWatching)(obj, 'child.cost'), 'watching child.cost');
      assert.ok((0, _metal.isWatching)(obj, 'child.tax'), 'watching child.tax');

      (0, _metal.set)(obj, 'child', {
        cost: 100.0,
        tax: 10.0
      });

      assert.equal(counts['cost'], 2, 'PROPERTY_DID_CHANGE called with cost');
      assert.equal(counts['tax'], 2, 'PROPERTY_DID_CHANGE called with tax');
      assert.equal(counts['total'], 1, 'PROPERTY_DID_CHANGE called with total');
    }
  });
});
enifed('@ember/-internals/metal/tests/property_events_test', ['@ember/-internals/meta', '@ember/-internals/metal', 'internal-test-helpers'], function (_meta, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('notifyPropertyChange', class extends _internalTestHelpers.AbstractTestCase {
    ['@test notifies property changes on instances'](assert) {
      class Foo {
        [_metal.PROPERTY_DID_CHANGE](prop) {
          assert.equal(prop, 'bar', 'property change notified');
        }
      }

      let foo = new Foo();

      (0, _metal.notifyPropertyChange)(foo, 'bar');
    }

    ['@test notifies property changes on instances with meta'](assert) {
      class Foo {
        [_metal.PROPERTY_DID_CHANGE](prop) {
          assert.equal(prop, 'bar', 'property change notified');
        }
      }

      let foo = new Foo();

      (0, _meta.meta)(foo); // setup meta

      (0, _metal.notifyPropertyChange)(foo, 'bar');
    }

    ['@test does not notify on class prototypes with meta'](assert) {
      assert.expect(0);

      class Foo {
        [_metal.PROPERTY_DID_CHANGE](prop) {
          assert.equal(prop, 'bar', 'property change notified');
        }
      }

      let foo = new Foo();

      (0, _meta.meta)(foo.constructor.prototype); // setup meta for prototype

      (0, _metal.notifyPropertyChange)(foo.constructor.prototype, 'bar');
    }

    ['@test does not notify on non-class prototypes with meta'](assert) {
      assert.expect(0);

      let foo = {
        [_metal.PROPERTY_DID_CHANGE](prop) {
          assert.equal(prop, 'baz', 'property change notified');
        }
      };

      let bar = Object.create(foo);

      (0, _meta.meta)(foo); // setup meta for prototype
      (0, _meta.meta)(bar); // setup meta for instance, switch prototype

      (0, _metal.notifyPropertyChange)(foo, 'baz');
    }
  });
});
enifed('@ember/-internals/metal/tests/set_properties_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('setProperties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test supports setting multiple attributes at once'](assert) {
      assert.deepEqual((0, _metal.setProperties)(null, null), null, 'noop for null properties and null object');
      assert.deepEqual((0, _metal.setProperties)(undefined, undefined), undefined, 'noop for undefined properties and undefined object');

      assert.deepEqual((0, _metal.setProperties)({}), undefined, 'noop for no properties');
      assert.deepEqual((0, _metal.setProperties)({}, undefined), undefined, 'noop for undefined');
      assert.deepEqual((0, _metal.setProperties)({}, null), null, 'noop for null');
      assert.deepEqual((0, _metal.setProperties)({}, NaN), NaN, 'noop for NaN');
      assert.deepEqual((0, _metal.setProperties)({}, {}), {}, 'meh');

      let props = (0, _metal.setProperties)({}, { foo: undefined });
      assert.deepEqual(props, { foo: undefined }, 'Setting undefined value');
      assert.ok('foo' in props, 'Setting undefined value');
      assert.deepEqual(Object.keys(props), ['foo'], 'Setting undefined value');

      assert.deepEqual((0, _metal.setProperties)({}, { foo: 1 }), { foo: 1 }, 'Set a single property');

      assert.deepEqual((0, _metal.setProperties)({}, { foo: 1, bar: 1 }), { foo: 1, bar: 1 }, 'Set multiple properties');

      assert.deepEqual((0, _metal.setProperties)({ foo: 2, baz: 2 }, { foo: 1 }), { foo: 1 }, 'Set one of multiple properties');

      assert.deepEqual((0, _metal.setProperties)({ foo: 2, baz: 2 }, { bar: 2 }), {
        bar: 2
      }, 'Set an additional, previously unset property');
    }
  });
});
enifed("@ember/-internals/metal/tests/tracked/computed_test", ["@ember/-internals/metal", "@ember/canary-features", "internal-test-helpers"], function (_metal, _canaryFeatures, _internalTestHelpers) {
    "use strict";

    var __decorate = undefined && undefined.__decorate || function (decorators, target, key, desc) {
        var c = arguments.length,
            r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc,
            d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    };

    if (_canaryFeatures.EMBER_METAL_TRACKED_PROPERTIES) {
        (0, _internalTestHelpers.moduleFor)('@tracked getters', class extends _internalTestHelpers.AbstractTestCase {
            ['@test works without get'](assert) {
                let count = 0;
                class Count {
                    get foo() {
                        count++;
                        return `computed foo`;
                    }
                }
                __decorate([_metal.tracked], Count.prototype, "foo", null);
                let obj = new Count();
                assert.equal(obj.foo, 'computed foo', 'should return value');
                assert.equal(count, 1, 'should have invoked computed property');
            }
            ['@test defining computed property should invoke property on get'](assert) {
                let count = 0;
                class Count {
                    get foo() {
                        count++;
                        return `computed foo`;
                    }
                }
                __decorate([_metal.tracked], Count.prototype, "foo", null);
                let obj = new Count();
                assert.equal((0, _metal.get)(obj, 'foo'), 'computed foo', 'should return value');
                assert.equal(count, 1, 'should have invoked computed property');
            }
            ['@test defining computed property should invoke property on set'](assert) {
                let count = 0;
                class Foo {
                    constructor() {
                        this.__foo = '';
                    }
                    get foo() {
                        return this.__foo;
                    }
                    set foo(value) {
                        count++;
                        this.__foo = `computed ${value}`;
                    }
                }
                __decorate([_metal.tracked], Foo.prototype, "foo", null);
                let obj = new Foo();
                assert.equal((0, _metal.set)(obj, 'foo', 'bar'), 'bar', 'should return set value');
                assert.equal(count, 1, 'should have invoked computed property');
                assert.equal((0, _metal.get)(obj, 'foo'), 'computed bar', 'should return new value');
            }
        });
    }
});
enifed("@ember/-internals/metal/tests/tracked/get_test", ["@ember/canary-features", "internal-test-helpers", "@ember/-internals/metal"], function (_canaryFeatures, _internalTestHelpers, _metal) {
    "use strict";

    var __decorate = undefined && undefined.__decorate || function (decorators, target, key, desc) {
        var c = arguments.length,
            r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc,
            d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    };

    if (_canaryFeatures.EMBER_METAL_TRACKED_PROPERTIES) {
        function createObj() {
            class Obj {
                constructor() {
                    this.string = 'string';
                    this.number = 23;
                    this.boolTrue = true;
                    this.boolFalse = false;
                    this.nullValue = null;
                }
            }
            __decorate([_metal.tracked], Obj.prototype, "string", void 0);
            __decorate([_metal.tracked], Obj.prototype, "number", void 0);
            __decorate([_metal.tracked], Obj.prototype, "boolTrue", void 0);
            __decorate([_metal.tracked], Obj.prototype, "boolFalse", void 0);
            __decorate([_metal.tracked], Obj.prototype, "nullValue", void 0);
            return new Obj();
        }
        (0, _internalTestHelpers.moduleFor)('@tracked decorator: get', class extends _internalTestHelpers.AbstractTestCase {
            '@test should get arbitrary properties on an object'() {
                let obj = createObj();
                for (let key in obj) {
                    this.assert.equal((0, _metal.get)(obj, key), obj[key], key);
                }
            }
            '@test should retrieve a number key on an object'() {
                class Obj {
                    constructor() {
                        this[1] = 'first';
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, 1, void 0);
                let obj = new Obj();
                this.assert.equal((0, _metal.get)(obj, '1'), 'first');
            }
            '@test should retrieve an empty key on an object'() {
                class Obj {
                    constructor() {
                        this[''] = 'empty';
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, "", void 0);
                let obj = new Obj();
                this.assert.equal((0, _metal.get)(obj, ''), 'empty');
            }
            '@test should get a @tracked path'() {
                class Key {
                    constructor() {
                        this.value = 'value';
                    }
                }
                __decorate([_metal.tracked], Key.prototype, "value", void 0);
                class Path {
                    constructor() {
                        this.key = new Key();
                    }
                }
                __decorate([_metal.tracked], Path.prototype, "key", void 0);
                class Obj {
                    constructor() {
                        this.path = new Path();
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, "path", void 0);
                let obj = new Obj();
                this.assert.equal((0, _metal.get)(obj, 'path.key.value'), 'value');
            }
            '@test should not access a property more than once'() {
                let count = 20;
                class Count {
                    get id() {
                        return ++count;
                    }
                }
                __decorate([_metal.tracked], Count.prototype, "id", null);
                let obj = new Count();
                (0, _metal.get)(obj, 'id');
                this.assert.equal(count, 21);
            }
        });
        (0, _internalTestHelpers.moduleFor)('@tracked decorator: getWithDefault', class extends _internalTestHelpers.AbstractTestCase {
            ['@test should get arbitrary properties on an object']() {
                let obj = createObj();
                for (let key in obj) {
                    this.assert.equal((0, _metal.getWithDefault)(obj, key, 'fail'), obj[key], key);
                }
                class Obj {
                    constructor() {
                        this.undef = undefined;
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, "undef", void 0);
                let obj2 = new Obj();
                this.assert.equal((0, _metal.getWithDefault)(obj2, 'undef', 'default'), 'default', 'explicit undefined retrieves the default');
                this.assert.equal((0, _metal.getWithDefault)(obj2, 'not-present', 'default'), 'default', 'non-present key retrieves the default');
            }
        });
    }
});
enifed("@ember/-internals/metal/tests/tracked/set_test", ["internal-test-helpers", "@ember/-internals/metal", "@ember/canary-features"], function (_internalTestHelpers, _metal, _canaryFeatures) {
    "use strict";

    var __decorate = undefined && undefined.__decorate || function (decorators, target, key, desc) {
        var c = arguments.length,
            r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc,
            d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    };

    if (_canaryFeatures.EMBER_METAL_TRACKED_PROPERTIES) {
        const createObj = () => {
            class Obj {
                constructor() {
                    this.string = 'string';
                    this.number = 23;
                    this.boolTrue = true;
                    this.boolFalse = false;
                    this.nullValue = null;
                    this.undefinedValue = undefined;
                }
            }
            __decorate([_metal.tracked], Obj.prototype, "string", void 0);
            __decorate([_metal.tracked], Obj.prototype, "number", void 0);
            __decorate([_metal.tracked], Obj.prototype, "boolTrue", void 0);
            __decorate([_metal.tracked], Obj.prototype, "boolFalse", void 0);
            __decorate([_metal.tracked], Obj.prototype, "nullValue", void 0);
            __decorate([_metal.tracked], Obj.prototype, "undefinedValue", void 0);
            return new Obj();
        };
        (0, _internalTestHelpers.moduleFor)('@tracked set', class extends _internalTestHelpers.AbstractTestCase {
            ['@test should set arbitrary properties on an object'](assert) {
                let obj = createObj();
                class Obj {
                    constructor() {
                        this.undefinedValue = 'emberjs';
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, "undefinedValue", void 0);
                let newObj = new Obj();
                for (let key in obj) {
                    assert.equal((0, _metal.set)(newObj, key, obj[key]), obj[key], 'should return value');
                    assert.equal((0, _metal.get)(newObj, key), obj[key], 'should set value');
                }
            }
            ['@test should set a number key on an object'](assert) {
                class Obj {
                    constructor() {
                        this[1] = 'original';
                    }
                }
                __decorate([_metal.tracked], Obj.prototype, 1, void 0);
                let obj = new Obj();
                (0, _metal.set)(obj, '1', 'first');
                assert.equal(obj[1], 'first');
            }
        });
    }
});
enifed('@ember/-internals/metal/tests/tracked/support', ['exports', '@ember/-internals/metal'], function (exports, _metal) {
  'use strict';

  exports.createTracked = createTracked;
  exports.createWithDescriptors = createWithDescriptors;
  function createTracked(values, proto = {}) {
    function Class() {
      for (let prop in values) {
        this[prop] = values[prop];
      }
    }

    for (let prop in values) {
      Object.defineProperty(proto, prop, (0, _metal.tracked)(proto, prop, {
        enumerable: true,
        configurable: true,
        writable: true,
        value: values[prop]
      }));
    }

    Class.prototype = proto;

    return new Class();
  }

  function createWithDescriptors(values) {
    function Class() {}

    for (let prop in values) {
      let descriptor = Object.getOwnPropertyDescriptor(values, prop);
      Object.defineProperty(Class.prototype, prop, (0, _metal.tracked)(Class.prototype, prop, descriptor));
    }

    return new Class();
  }
});
enifed("@ember/-internals/metal/tests/tracked/validation_test", ["@ember/-internals/metal", "@ember/canary-features", "internal-test-helpers"], function (_metal, _canaryFeatures, _internalTestHelpers) {
    "use strict";

    var __decorate = undefined && undefined.__decorate || function (decorators, target, key, desc) {
        var c = arguments.length,
            r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc,
            d;
        if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
        return c > 3 && r && Object.defineProperty(target, key, r), r;
    };

    if (_canaryFeatures.EMBER_METAL_TRACKED_PROPERTIES) {
        (0, _internalTestHelpers.moduleFor)('@tracked get validation', class extends _internalTestHelpers.AbstractTestCase {
            [`@test validators for tracked getters with dependencies should invalidate when the dependencies invalidate`](assert) {
                class Tracked {
                    constructor(first, last) {
                        this.first = undefined;
                        this.last = undefined;
                        this.first = first;
                        this.last = last;
                    }
                    get full() {
                        return `${this.first} ${this.last}`;
                    }
                }
                __decorate([_metal.tracked], Tracked.prototype, "first", void 0);
                __decorate([_metal.tracked], Tracked.prototype, "last", void 0);
                __decorate([_metal.tracked], Tracked.prototype, "full", null);
                let obj = new Tracked('Tom', 'Dale');
                let tag = (0, _metal.tagForProperty)(obj, 'full');
                let snapshot = tag.value();
                let full = obj.full;
                assert.equal(full, 'Tom Dale', 'The full name starts correct');
                assert.equal(tag.validate(snapshot), true);
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
                obj.first = 'Thomas';
                assert.equal(tag.validate(snapshot), false);
                assert.equal(obj.full, 'Thomas Dale');
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
            }
            [`@test interaction with Ember object model (tracked property depending on Ember property)`](assert) {
                class Tracked {
                    constructor(name) {
                        this.name = name;
                    }
                    get full() {
                        return `${(0, _metal.get)(this.name, 'first')} ${(0, _metal.get)(this.name, 'last')}`;
                    }
                }
                __decorate([_metal.tracked], Tracked.prototype, "name", void 0);
                __decorate([_metal.tracked], Tracked.prototype, "full", null);
                let tom = { first: 'Tom', last: 'Dale' };
                let obj = new Tracked(tom);
                let tag = (0, _metal.tagForProperty)(obj, 'full');
                let snapshot = tag.value();
                let full = obj.full;
                assert.equal(full, 'Tom Dale');
                assert.equal(tag.validate(snapshot), true);
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
                (0, _metal.set)(tom, 'first', 'Thomas');
                assert.equal(tag.validate(snapshot), false, 'invalid after setting with Ember set');
                assert.equal(obj.full, 'Thomas Dale');
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
            }
            [`@test interaction with Ember object model (Ember computed property depending on tracked property)`](assert) {
                class EmberObject {
                    constructor(name) {
                        this.name = name;
                    }
                }
                (0, _metal.defineProperty)(EmberObject.prototype, 'full', (0, _metal.computed)('name', function () {
                    let name = (0, _metal.get)(this, 'name');
                    return `${name.first} ${name.last}`;
                }));
                class Name {
                    constructor(first, last) {
                        this.first = first;
                        this.last = last;
                    }
                }
                __decorate([_metal.tracked], Name.prototype, "first", void 0);
                __decorate([_metal.tracked], Name.prototype, "last", void 0);
                let tom = new Name('Tom', 'Dale');
                let obj = new EmberObject(tom);
                let tag = (0, _metal.tagForProperty)(obj, 'full');
                let snapshot = tag.value();
                let full = (0, _metal.get)(obj, 'full');
                assert.equal(full, 'Tom Dale');
                assert.equal(tag.validate(snapshot), true);
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
                tom.first = 'Thomas';
                assert.equal(tag.validate(snapshot), false, 'invalid after setting with tracked properties');
                assert.equal((0, _metal.get)(obj, 'full'), 'Thomas Dale');
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
            }
            ['@test interaction with the Ember object model (paths going through tracked properties)'](assert) {
                let self;
                class EmberObject {
                    constructor(contact) {
                        this.contact = contact;
                        self = this;
                    }
                }
                (0, _metal.defineProperty)(EmberObject.prototype, 'full', (0, _metal.computed)('contact.name.first', 'contact.name.last', function () {
                    let contact = (0, _metal.get)(self, 'contact');
                    return `${(0, _metal.get)(contact.name, 'first')} ${(0, _metal.get)(contact.name, 'last')}`;
                }));
                class Contact {
                    constructor(name) {
                        this.name = undefined;
                        this.name = name;
                    }
                }
                __decorate([_metal.tracked], Contact.prototype, "name", void 0);
                class EmberName {
                    constructor(first, last) {
                        this.first = first;
                        this.last = last;
                    }
                }
                let tom = new EmberName('Tom', 'Dale');
                let contact = new Contact(tom);
                let obj = new EmberObject(contact);
                let tag = (0, _metal.tagForProperty)(obj, 'full');
                let snapshot = tag.value();
                let full = (0, _metal.get)(obj, 'full');
                assert.equal(full, 'Tom Dale');
                assert.equal(tag.validate(snapshot), true);
                snapshot = tag.value();
                assert.equal(tag.validate(snapshot), true);
                (0, _metal.set)(tom, 'first', 'Thomas');
                assert.equal(tag.validate(snapshot), false, 'invalid after setting with Ember.set');
                assert.equal((0, _metal.get)(obj, 'full'), 'Thomas Dale');
                snapshot = tag.value();
                tom = contact.name = new EmberName('T', 'Dale');
                assert.equal(tag.validate(snapshot), false, 'invalid after setting with Ember.set');
                assert.equal((0, _metal.get)(obj, 'full'), 'T Dale');
                snapshot = tag.value();
                (0, _metal.set)(tom, 'first', 'Tizzle');
                assert.equal(tag.validate(snapshot), false, 'invalid after setting with Ember.set');
                assert.equal((0, _metal.get)(obj, 'full'), 'Tizzle Dale');
            }
        });
    }
});
enifed('@ember/-internals/metal/tests/watching/is_watching_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  function testObserver(assert, setup, teardown, key = 'key') {
    let obj = {};

    assert.equal((0, _metal.isWatching)(obj, key), false, 'precond - isWatching is false by default');
    setup(obj, key, 'fn');
    assert.equal((0, _metal.isWatching)(obj, key), true, 'isWatching is true when observers are added');
    teardown(obj, key, 'fn');
    assert.equal((0, _metal.isWatching)(obj, key), false, 'isWatching is false after observers are removed');
  }

  (0, _internalTestHelpers.moduleFor)('isWatching', class extends _internalTestHelpers.AbstractTestCase {
    ['@test isWatching is true for regular local observers'](assert) {
      testObserver(assert, (obj, key, fn) => {
        _metal.Mixin.create({
          [fn]: (0, _metal.observer)(key, function () {})
        }).apply(obj);
      }, (obj, key, fn) => (0, _metal.removeObserver)(obj, key, obj, fn));
    }

    ['@test isWatching is true for nonlocal observers'](assert) {
      testObserver(assert, (obj, key, fn) => {
        (0, _metal.addObserver)(obj, key, obj, fn);
      }, (obj, key, fn) => (0, _metal.removeObserver)(obj, key, obj, fn));
    }

    ['@test isWatching is true for chained observers'](assert) {
      testObserver(assert, function (obj, key, fn) {
        (0, _metal.addObserver)(obj, key + '.bar', obj, fn);
      }, function (obj, key, fn) {
        (0, _metal.removeObserver)(obj, key + '.bar', obj, fn);
      });
    }

    ['@test isWatching is true for computed properties'](assert) {
      testObserver(assert, (obj, key, fn) => {
        (0, _metal.defineProperty)(obj, fn, (0, _metal.computed)(function () {}).property(key));
        (0, _metal.get)(obj, fn);
      }, (obj, key, fn) => (0, _metal.defineProperty)(obj, fn, null));
    }

    ['@test isWatching is true for chained computed properties'](assert) {
      testObserver(assert, (obj, key, fn) => {
        (0, _metal.defineProperty)(obj, fn, (0, _metal.computed)(function () {}).property(key + '.bar'));
        (0, _metal.get)(obj, fn);
      }, (obj, key, fn) => (0, _metal.defineProperty)(obj, fn, null));
    }

    // can't watch length on Array - it is special...
    // But you should be able to watch a length property of an object
    ["@test isWatching is true for 'length' property on object"](assert) {
      testObserver(assert, (obj, key, fn) => {
        (0, _metal.defineProperty)(obj, 'length', null, '26.2 miles');
        (0, _metal.addObserver)(obj, 'length', obj, fn);
      }, (obj, key, fn) => (0, _metal.removeObserver)(obj, 'length', obj, fn), 'length');
    }
  });
});
enifed('@ember/-internals/metal/tests/watching/unwatch_test', ['@ember/-internals/metal', 'internal-test-helpers'], function (_metal, _internalTestHelpers) {
  'use strict';

  let didCount;

  function addListeners(obj, keyPath) {
    (0, _metal.addListener)(obj, keyPath + ':change', () => didCount++);
  }

  (0, _internalTestHelpers.moduleFor)('unwatch', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      didCount = 0;
    }

    ['@test unwatching a computed property - regular get/set'](assert) {
      let obj = {};

      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get() {
          return this.__foo;
        },
        set(keyName, value) {
          this.__foo = value;
          return this.__foo;
        }
      }));
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(didCount, 1, 'should have invoked didCount');

      (0, _metal.unwatch)(obj, 'foo');
      didCount = 0;
      (0, _metal.set)(obj, 'foo', 'BAZ');
      assert.equal(didCount, 0, 'should NOT have invoked didCount');
    }

    ['@test unwatching a regular property - regular get/set'](assert) {
      let obj = { foo: 'BIFF' };
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(didCount, 1, 'should have invoked didCount');

      (0, _metal.unwatch)(obj, 'foo');
      didCount = 0;
      (0, _metal.set)(obj, 'foo', 'BAZ');
      assert.equal(didCount, 0, 'should NOT have invoked didCount');
    }

    ['@test unwatching should be nested'](assert) {
      let obj = { foo: 'BIFF' };
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');
      (0, _metal.watch)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(didCount, 1, 'should have invoked didCount');

      (0, _metal.unwatch)(obj, 'foo');
      didCount = 0;
      (0, _metal.set)(obj, 'foo', 'BAZ');
      assert.equal(didCount, 1, 'should NOT have invoked didCount');

      (0, _metal.unwatch)(obj, 'foo');
      didCount = 0;
      (0, _metal.set)(obj, 'foo', 'BAZ');
      assert.equal(didCount, 0, 'should NOT have invoked didCount');
    }

    ['@test unwatching "length" property on an object'](assert) {
      let obj = { foo: 'RUN' };
      addListeners(obj, 'length');

      // Can watch length when it is undefined
      (0, _metal.watch)(obj, 'length');
      (0, _metal.set)(obj, 'length', '10k');
      assert.equal(didCount, 1, 'should have invoked didCount');

      // Should stop watching despite length now being defined (making object 'array-like')
      (0, _metal.unwatch)(obj, 'length');
      didCount = 0;
      (0, _metal.set)(obj, 'length', '5k');
      assert.equal(didCount, 0, 'should NOT have invoked didCount');
    }

    ['@test unwatching should not destroy non MANDATORY_SETTER descriptor'](assert) {
      let obj = {
        get foo() {
          return 'RUN';
        }
      };

      assert.equal(obj.foo, 'RUN', 'obj.foo');
      (0, _metal.watch)(obj, 'foo');
      assert.equal(obj.foo, 'RUN', 'obj.foo after watch');
      (0, _metal.unwatch)(obj, 'foo');
      assert.equal(obj.foo, 'RUN', 'obj.foo after unwatch');
    }
  });
});
enifed('@ember/-internals/metal/tests/watching/watch_test', ['@ember/-internals/environment', '@ember/-internals/metal', '@ember/-internals/meta', 'internal-test-helpers'], function (_environment, _metal, _meta, _internalTestHelpers) {
  'use strict';

  let didCount, didKeys, originalLookup;

  function addListeners(obj, keyPath) {
    (0, _metal.addListener)(obj, keyPath + ':change', function () {
      didCount++;
      didKeys.push(keyPath);
    });
  }

  (0, _internalTestHelpers.moduleFor)('watch', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      didCount = 0;
      didKeys = [];

      originalLookup = _environment.context.lookup;
      _environment.context.lookup = {};
    }

    afterEach() {
      _environment.context.lookup = originalLookup;
    }

    ['@test watching a computed property'](assert) {
      let obj = {};
      (0, _metal.defineProperty)(obj, 'foo', (0, _metal.computed)({
        get() {
          return this.__foo;
        },
        set(keyName, value) {
          if (value !== undefined) {
            this.__foo = value;
          }
          return this.__foo;
        }
      }));
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(didCount, 1, 'should have invoked didCount');
    }

    ['@test watching a regular defined property'](assert) {
      let obj = { foo: 'baz' };
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');
      assert.equal((0, _metal.get)(obj, 'foo'), 'baz', 'should have original prop');

      (0, _metal.set)(obj, 'foo', 'bar');
      assert.equal(didCount, 1, 'should have invoked didCount');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'should get new value');
      assert.equal(obj.foo, 'bar', 'property should be accessible on obj');
    }

    ['@test watching a regular undefined property'](assert) {
      let obj = {};
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');

      assert.equal('foo' in obj, false, 'precond undefined');

      (0, _metal.set)(obj, 'foo', 'bar');

      assert.equal(didCount, 1, 'should have invoked didCount');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'should get new value');
      assert.equal(obj.foo, 'bar', 'property should be accessible on obj');
    }

    ['@test watches should inherit'](assert) {
      let obj = { foo: 'baz' };
      let objB = Object.create(obj);

      addListeners(obj, 'foo');
      (0, _metal.watch)(obj, 'foo');
      assert.equal((0, _metal.get)(obj, 'foo'), 'baz', 'should have original prop');

      (0, _metal.set)(objB, 'foo', 'bar');
      (0, _metal.set)(obj, 'foo', 'baz');
      assert.equal(didCount, 1, 'should have invoked didCount once only');
    }

    ['@test watching an object THEN defining it should work also'](assert) {
      let obj = {};
      addListeners(obj, 'foo');

      (0, _metal.watch)(obj, 'foo');

      (0, _metal.defineProperty)(obj, 'foo');
      (0, _metal.set)(obj, 'foo', 'bar');

      assert.equal((0, _metal.get)(obj, 'foo'), 'bar', 'should have set');
      assert.equal(didCount, 1, 'should have invoked didChange once');
    }

    ['@test watching a chain then defining the property'](assert) {
      let obj = {};
      let foo = { bar: 'bar' };
      addListeners(obj, 'foo.bar');
      addListeners(foo, 'bar');

      (0, _metal.watch)(obj, 'foo.bar');

      (0, _metal.defineProperty)(obj, 'foo', undefined, foo);
      (0, _metal.set)(foo, 'bar', 'baz');

      assert.deepEqual(didKeys, ['foo.bar', 'bar'], 'should have invoked didChange with bar, foo.bar');
      assert.equal(didCount, 2, 'should have invoked didChange twice');
    }

    ['@test watching a chain then defining the nested property'](assert) {
      let bar = {};
      let obj = { foo: bar };
      let baz = { baz: 'baz' };
      addListeners(obj, 'foo.bar.baz');
      addListeners(baz, 'baz');

      (0, _metal.watch)(obj, 'foo.bar.baz');

      (0, _metal.defineProperty)(bar, 'bar', undefined, baz);
      (0, _metal.set)(baz, 'baz', 'BOO');

      assert.deepEqual(didKeys, ['foo.bar.baz', 'baz'], 'should have invoked didChange with bar, foo.bar');
      assert.equal(didCount, 2, 'should have invoked didChange twice');
    }

    ['@test watching an object value then unwatching should restore old value'](assert) {
      let obj = { foo: { bar: { baz: { biff: 'BIFF' } } } };
      addListeners(obj, 'foo.bar.baz.biff');

      (0, _metal.watch)(obj, 'foo.bar.baz.biff');

      let foo = (0, _metal.get)(obj, 'foo');
      assert.equal((0, _metal.get)((0, _metal.get)((0, _metal.get)(foo, 'bar'), 'baz'), 'biff'), 'BIFF', 'biff should exist');

      (0, _metal.unwatch)(obj, 'foo.bar.baz.biff');
      assert.equal((0, _metal.get)((0, _metal.get)((0, _metal.get)(foo, 'bar'), 'baz'), 'biff'), 'BIFF', 'biff should exist');
    }

    ['@test when watching another object, destroy should remove chain watchers from the other object'](assert) {
      let objA = {};
      let objB = { foo: 'bar' };
      objA.b = objB;
      addListeners(objA, 'b.foo');

      (0, _metal.watch)(objA, 'b.foo');

      let meta_objB = (0, _meta.meta)(objB);
      let chainNode = (0, _meta.meta)(objA).readableChains().chains.b.chains.foo;

      assert.equal(meta_objB.peekWatching('foo'), 1, 'should be watching foo');
      assert.equal(meta_objB.readableChainWatchers().has('foo', chainNode), true, 'should have chain watcher');

      (0, _meta.deleteMeta)(objA);

      assert.equal(meta_objB.peekWatching('foo'), 0, 'should not be watching foo');
      assert.equal(meta_objB.readableChainWatchers().has('foo', chainNode), false, 'should not have chain watcher');
    }

    // TESTS for length property

    ['@test watching "length" property on an object'](assert) {
      let obj = { length: '26.2 miles' };
      addListeners(obj, 'length');

      (0, _metal.watch)(obj, 'length');
      assert.equal((0, _metal.get)(obj, 'length'), '26.2 miles', 'should have original prop');

      (0, _metal.set)(obj, 'length', '10k');
      assert.equal(didCount, 1, 'should have invoked didCount');

      assert.equal((0, _metal.get)(obj, 'length'), '10k', 'should get new value');
      assert.equal(obj.length, '10k', 'property should be accessible on obj');
    }

    ['@test watching "length" property on an array'](assert) {
      let arr = [];
      addListeners(arr, 'length');

      (0, _metal.watch)(arr, 'length');
      assert.equal((0, _metal.get)(arr, 'length'), 0, 'should have original prop');

      (0, _metal.set)(arr, 'length', '10');
      assert.equal(didCount, 1, 'should NOT have invoked didCount');

      assert.equal((0, _metal.get)(arr, 'length'), 10, 'should get new value');
      assert.equal(arr.length, 10, 'property should be accessible on arr');
    }

    ['@test watch + ES5 getter'](assert) {
      let parent = { b: 1 };
      let child = {
        get b() {
          return parent.b;
        }
      };

      assert.equal(parent.b, 1, 'parent.b should be 1');
      assert.equal(child.b, 1, 'child.b should be 1');
      assert.equal((0, _metal.get)(child, 'b'), 1, 'get(child, "b") should be 1');

      (0, _metal.watch)(child, 'b');

      assert.equal(parent.b, 1, 'parent.b should be 1 (after watch)');
      assert.equal(child.b, 1, 'child.b should be 1  (after watch)');

      assert.equal((0, _metal.get)(child, 'b'), 1, 'get(child, "b") should be 1 (after watch)');
    }

    ['@test watch + set + no-descriptor'](assert) {
      let child = {};

      assert.equal(child.b, undefined, 'child.b ');
      assert.equal((0, _metal.get)(child, 'b'), undefined, 'get(child, "b")');

      (0, _metal.watch)(child, 'b');
      (0, _metal.set)(child, 'b', 1);

      assert.equal(child.b, 1, 'child.b (after watch)');
      assert.equal((0, _metal.get)(child, 'b'), 1, 'get(child, "b") (after watch)');
    }
  });
});
enifed('@ember/-internals/routing/tests/ext/controller_test', ['@ember/-internals/owner', '@ember/controller', 'internal-test-helpers'], function (_owner, _controller, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/routing/ext/controller', class extends _internalTestHelpers.AbstractTestCase {
    ["@test transitionToRoute considers an engine's mountPoint"](assert) {
      let router = {
        transitionTo(route) {
          return route;
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,
          mountPoint: 'foo.bar'
        }
      });

      let controller = _controller.default.create({ target: router });
      (0, _owner.setOwner)(controller, engineInstance);

      assert.strictEqual(controller.transitionToRoute('application'), 'foo.bar.application', 'properly prefixes application route');
      assert.strictEqual(controller.transitionToRoute('posts'), 'foo.bar.posts', 'properly prefixes child routes');
      assert.throws(() => controller.transitionToRoute('/posts'), 'throws when trying to use a url');

      let queryParams = {};
      assert.strictEqual(controller.transitionToRoute(queryParams), queryParams, 'passes query param only transitions through');
    }

    ["@test replaceRoute considers an engine's mountPoint"](assert) {
      let router = {
        replaceWith(route) {
          return route;
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,
          mountPoint: 'foo.bar'
        }
      });

      let controller = _controller.default.create({ target: router });
      (0, _owner.setOwner)(controller, engineInstance);

      assert.strictEqual(controller.replaceRoute('application'), 'foo.bar.application', 'properly prefixes application route');
      assert.strictEqual(controller.replaceRoute('posts'), 'foo.bar.posts', 'properly prefixes child routes');
      assert.throws(() => controller.replaceRoute('/posts'), 'throws when trying to use a url');

      let queryParams = {};
      assert.strictEqual(controller.replaceRoute(queryParams), queryParams, 'passes query param only transitions through');
    }
  });
});
enifed('@ember/-internals/routing/tests/location/auto_location_test', ['@ember/-internals/owner', '@ember/polyfills', '@ember/-internals/browser-environment', '@ember/runloop', '@ember/-internals/metal', '@ember/-internals/routing/lib/location/auto_location', '@ember/-internals/routing/lib/location/history_location', '@ember/-internals/routing/lib/location/hash_location', '@ember/-internals/routing/lib/location/none_location', 'internal-test-helpers'], function (_owner, _polyfills, _browserEnvironment, _runloop, _metal, _auto_location, _history_location, _hash_location, _none_location, _internalTestHelpers) {
  'use strict';

  function mockBrowserLocation(overrides, assert) {
    return (0, _polyfills.assign)({
      href: 'http://test.com/',
      pathname: '/',
      hash: '',
      search: '',
      replace() {
        assert.ok(false, 'location.replace should not be called during testing');
      }
    }, overrides);
  }

  function mockBrowserHistory(overrides, assert) {
    return (0, _polyfills.assign)({
      pushState() {
        assert.ok(false, 'history.pushState should not be called during testing');
      },
      replaceState() {
        assert.ok(false, 'history.replaceState should not be called during testing');
      }
    }, overrides);
  }

  function createLocation(location, history) {
    owner = (0, _internalTestHelpers.buildOwner)();

    owner.register('location:history', _history_location.default);
    owner.register('location:hash', _hash_location.default);
    owner.register('location:none', _none_location.default);

    let autolocation = _auto_location.default.create({
      [_owner.OWNER]: owner,
      location: location,
      history: history,
      global: {}
    });

    return autolocation;
  }

  let location, owner;

  (0, _internalTestHelpers.moduleFor)('AutoLocation', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      if (owner) {
        (0, _runloop.run)(owner, 'destroy');
        owner = location = undefined;
      }
    }

    ['@test AutoLocation should have the `global`'](assert) {
      let location = _auto_location.default.create();

      assert.ok(location.global, 'has a global defined');
      assert.strictEqual(location.global, _browserEnvironment.window, 'has the environments window global');
    }

    ["@test AutoLocation should return concrete implementation's value for `getURL`"](assert) {
      let browserLocation = mockBrowserLocation({}, assert);
      let browserHistory = mockBrowserHistory({}, assert);

      location = createLocation(browserLocation, browserHistory);
      location.detect();

      let concreteImplementation = (0, _metal.get)(location, 'concreteImplementation');

      concreteImplementation.getURL = function () {
        return '/lincoln/park';
      };

      assert.equal(location.getURL(), '/lincoln/park');
    }

    ['@test AutoLocation should use a HistoryLocation instance when pushStates is supported'](assert) {
      let browserLocation = mockBrowserLocation({}, assert);
      let browserHistory = mockBrowserHistory({}, assert);

      location = createLocation(browserLocation, browserHistory);
      location.detect();

      assert.ok((0, _metal.get)(location, 'concreteImplementation') instanceof _history_location.default);
    }

    ['@test AutoLocation should use a HashLocation instance when pushStates are not supported, but hashchange events are and the URL is already in the HashLocation format'](assert) {
      let browserLocation = mockBrowserLocation({
        hash: '#/testd'
      }, assert);

      location = createLocation(browserLocation);
      location.global = {
        onhashchange() {}
      };

      location.detect();
      assert.ok((0, _metal.get)(location, 'concreteImplementation') instanceof _hash_location.default);
    }

    ['@test AutoLocation should use a NoneLocation instance when neither history nor hashchange are supported.'](assert) {
      location = createLocation(mockBrowserLocation({}, assert));
      location.detect();

      assert.ok((0, _metal.get)(location, 'concreteImplementation') instanceof _none_location.default);
    }

    ["@test AutoLocation should use an index path (i.e. '/') without any location.hash as OK for HashLocation"](assert) {
      let browserLocation = mockBrowserLocation({
        href: 'http://test.com/',
        pathname: '/',
        hash: '',
        search: '',
        replace() {
          assert.ok(false, 'location.replace should not be called');
        }
      }, assert);

      location = createLocation(browserLocation);
      location.global = {
        onhashchange() {}
      };

      location.detect();

      assert.ok((0, _metal.get)(location, 'concreteImplementation') instanceof _hash_location.default, 'uses a HashLocation');
    }

    ['@test AutoLocation should transform the URL for hashchange-only browsers viewing a HistoryLocation-formatted path'](assert) {
      assert.expect(3);

      let browserLocation = mockBrowserLocation({
        hash: '',
        hostname: 'test.com',
        href: 'http://test.com/test',
        pathname: '/test',
        protocol: 'http:',
        port: '',
        search: '',
        replace(path) {
          assert.equal(path, 'http://test.com/#/test', 'location.replace should be called with normalized HashLocation path');
        }
      }, assert);

      let location = createLocation(browserLocation);
      location.global = {
        onhashchange() {}
      };

      location.detect();

      assert.ok((0, _metal.get)(location, 'concreteImplementation') instanceof _none_location.default, 'NoneLocation should be used while we attempt to location.replace()');
      assert.equal((0, _metal.get)(location, 'cancelRouterSetup'), true, 'cancelRouterSetup should be set so the router knows.');
    }

    ['@test AutoLocation should replace the URL for pushState-supported browsers viewing a HashLocation-formatted url'](assert) {
      assert.expect(2);
      let browserLocation = mockBrowserLocation({
        hash: '#/test',
        hostname: 'test.com',
        href: 'http://test.com/#/test',
        pathname: '/',
        protocol: 'http:',
        port: '',
        search: ''
      }, assert);

      let browserHistory = mockBrowserHistory({
        replaceState(state, title, path) {
          assert.equal(path, '/test', 'history.replaceState should be called with normalized HistoryLocation url');
        }
      }, assert);

      let location = createLocation(browserLocation, browserHistory);
      location.detect();

      assert.ok((0, _metal.get)(location, 'concreteImplementation'), _history_location.default);
    }

    ['@test AutoLocation requires any rootURL given to end in a trailing forward slash'](assert) {
      let browserLocation = mockBrowserLocation({}, assert);
      let expectedMsg = /rootURL must end with a trailing forward slash e.g. "\/app\/"/;

      location = createLocation(browserLocation);
      location.rootURL = 'app';

      expectAssertion(function () {
        location.detect();
      }, expectedMsg);

      location.rootURL = '/app';
      expectAssertion(function () {
        location.detect();
      }, expectedMsg);

      // Note the trailing whitespace
      location.rootURL = '/app/ ';
      expectAssertion(function () {
        location.detect();
      }, expectedMsg);
    }

    ['@test AutoLocation provides its rootURL to the concreteImplementation'](assert) {
      let browserLocation = mockBrowserLocation({
        pathname: '/some/subdir/derp'
      }, assert);
      let browserHistory = mockBrowserHistory({}, assert);

      location = createLocation(browserLocation, browserHistory);
      location.rootURL = '/some/subdir/';

      location.detect();

      let concreteLocation = (0, _metal.get)(location, 'concreteImplementation');
      assert.equal(location.rootURL, concreteLocation.rootURL);
    }

    ['@test getHistoryPath() should return a normalized, HistoryLocation-supported path'](assert) {
      let browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/about?foo=bar#foo',
        pathname: '/app/about',
        search: '?foo=bar',
        hash: '#foo'
      }, assert);

      assert.equal((0, _auto_location.getHistoryPath)('/app/', browserLocation), '/app/about?foo=bar#foo', 'URLs already in HistoryLocation form should come out the same');

      browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/#/about?foo=bar#foo',
        pathname: '/app/',
        search: '',
        hash: '#/about?foo=bar#foo'
      }, assert);
      assert.equal((0, _auto_location.getHistoryPath)('/app/', browserLocation), '/app/about?foo=bar#foo', 'HashLocation formed URLs should be normalized');

      browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/#about?foo=bar#foo',
        pathname: '/app/',
        search: '',
        hash: '#about?foo=bar#foo'
      }, assert);
      assert.equal((0, _auto_location.getHistoryPath)('/app', browserLocation), '/app/#about?foo=bar#foo', "URLs with a hash not following #/ convention shouldn't be normalized as a route");
    }

    ['@test getHashPath() should return a normalized, HashLocation-supported path'](assert) {
      let browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/#/about?foo=bar#foo',
        pathname: '/app/',
        search: '',
        hash: '#/about?foo=bar#foo'
      }, assert);
      assert.equal((0, _auto_location.getHashPath)('/app/', browserLocation), '/app/#/about?foo=bar#foo', 'URLs already in HistoryLocation form should come out the same');

      browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/about?foo=bar#foo',
        pathname: '/app/about',
        search: '?foo=bar',
        hash: '#foo'
      }, assert);
      assert.equal((0, _auto_location.getHashPath)('/app/', browserLocation), '/app/#/about?foo=bar#foo', 'HistoryLocation formed URLs should be normalized');

      browserLocation = mockBrowserLocation({
        href: 'http://test.com/app/#about?foo=bar#foo',
        pathname: '/app/',
        search: '',
        hash: '#about?foo=bar#foo'
      }, assert);

      assert.equal((0, _auto_location.getHashPath)('/app/', browserLocation), '/app/#/#about?foo=bar#foo', "URLs with a hash not following #/ convention shouldn't be normalized as a route");
    }
  });
});
enifed('@ember/-internals/routing/tests/location/hash_location_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/routing/lib/location/hash_location', 'internal-test-helpers'], function (_runloop, _metal, _hash_location, _internalTestHelpers) {
  'use strict';

  let location;

  function createLocation(options, assert) {
    let HashTestLocation = _hash_location.default.extend({
      _location: {
        href: 'http://test.com/',
        pathname: '/',
        hash: '',
        search: '',
        replace() {
          assert.ok(false, 'location.replace should not be called during testing');
        }
      }
    });

    if (!options) {
      options = {};
    }
    location = HashTestLocation.create(options);
  }

  function mockBrowserLocation(path) {
    // This is a neat trick to auto-magically extract the hostname from any
    // url by letting the browser do the work ;)
    let tmp = document.createElement('a');
    tmp.href = path;

    let protocol = !tmp.protocol || tmp.protocol === ':' ? 'http' : tmp.protocol;
    let pathname = tmp.pathname.match(/^\//) ? tmp.pathname : '/' + tmp.pathname;

    return {
      hash: tmp.hash,
      host: tmp.host || 'localhost',
      hostname: tmp.hostname || 'localhost',
      href: tmp.href,
      pathname: pathname,
      port: tmp.port || '',
      protocol: protocol,
      search: tmp.search
    };
  }

  function triggerHashchange() {
    var event = document.createEvent('HTMLEvents');
    event.initEvent('hashchange', true, false);
    window.dispatchEvent(event);
  }

  (0, _internalTestHelpers.moduleFor)('HashLocation', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      (0, _runloop.run)(function () {
        if (location) {
          location.destroy();
        }
      });
    }

    ['@test HashLocation.getURL() returns the current url'](assert) {
      createLocation({
        _location: mockBrowserLocation('/#/foo/bar')
      }, assert);

      assert.equal(location.getURL(), '/foo/bar');
    }

    ['@test HashLocation.getURL() includes extra hashes'](assert) {
      createLocation({
        _location: mockBrowserLocation('/#/foo#bar#car')
      }, assert);

      assert.equal(location.getURL(), '/foo#bar#car');
    }

    ['@test HashLocation.getURL() assumes location.hash without #/ prefix is not a route path'](assert) {
      createLocation({
        _location: mockBrowserLocation('/#foo#bar')
      }, assert);

      assert.equal(location.getURL(), '/#foo#bar');
    }

    ['@test HashLocation.getURL() returns a normal forward slash when there is no location.hash'](assert) {
      createLocation({
        _location: mockBrowserLocation('/')
      }, assert);

      assert.equal(location.getURL(), '/');
    }

    ['@test HashLocation.setURL() correctly sets the url'](assert) {
      createLocation({}, assert);

      location.setURL('/bar');

      assert.equal((0, _metal.get)(location, 'location.hash'), '/bar');
      assert.equal((0, _metal.get)(location, 'lastSetURL'), '/bar');
    }

    ['@test HashLocation.replaceURL() correctly replaces to the path with a page reload'](assert) {
      assert.expect(2);

      createLocation({
        _location: {
          replace(path) {
            assert.equal(path, '#/foo');
          }
        }
      }, assert);

      location.replaceURL('/foo');

      assert.equal((0, _metal.get)(location, 'lastSetURL'), '/foo');
    }

    ['@test HashLocation.onUpdateURL callback executes as expected'](assert) {
      assert.expect(1);

      createLocation({
        _location: mockBrowserLocation('/#/foo/bar')
      }, assert);

      let callback = function (param) {
        assert.equal(param, '/foo/bar', 'path is passed as param');
      };

      location.onUpdateURL(callback);

      triggerHashchange();
    }

    ["@test HashLocation.onUpdateURL doesn't execute callback if lastSetURL === path"](assert) {
      assert.expect(0);

      createLocation({
        _location: {
          hash: '#/foo/bar'
        },
        lastSetURL: '/foo/bar'
      }, assert);

      let callback = function () {
        assert.ok(false, 'callback should not be called');
      };

      location.onUpdateURL(callback);

      triggerHashchange();
    }

    ['@test HashLocation.formatURL() prepends a # to the provided string'](assert) {
      createLocation({}, assert);

      assert.equal(location.formatURL('/foo#bar'), '#/foo#bar');
    }

    ['@test HashLocation.willDestroy() cleans up hashchange event listener'](assert) {
      assert.expect(1);

      createLocation({}, assert);

      let callback = function () {
        assert.ok(true, 'should invoke callback once');
      };

      location.onUpdateURL(callback);

      triggerHashchange();

      (0, _runloop.run)(location, 'destroy');
      location = null;

      triggerHashchange();
    }
  });
});
enifed('@ember/-internals/routing/tests/location/history_location_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/routing/lib/location/history_location', 'internal-test-helpers'], function (_runloop, _metal, _history_location, _internalTestHelpers) {
  'use strict';

  let FakeHistory, HistoryTestLocation, location;

  function createLocation(options) {
    if (!options) {
      options = {};
    }
    location = HistoryTestLocation.create(options);
  }

  function mockBrowserLocation(path) {
    // This is a neat trick to auto-magically extract the hostname from any
    // url by letting the browser do the work ;)
    let tmp = document.createElement('a');
    tmp.href = path;

    let protocol = !tmp.protocol || tmp.protocol === ':' ? 'http' : tmp.protocol;
    let pathname = tmp.pathname.match(/^\//) ? tmp.pathname : '/' + tmp.pathname;

    return {
      hash: tmp.hash,
      host: tmp.host || 'localhost',
      hostname: tmp.hostname || 'localhost',
      href: tmp.href,
      pathname: pathname,
      port: tmp.port || '',
      protocol: protocol,
      search: tmp.search
    };
  }

  (0, _internalTestHelpers.moduleFor)('HistoryLocation', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      FakeHistory = {
        state: null,
        _states: [],
        replaceState(state) {
          this.state = state;
          this._states[0] = state;
        },
        pushState(state) {
          this.state = state;
          this._states.unshift(state);
        }
      };

      HistoryTestLocation = _history_location.default.extend({
        history: FakeHistory
      });
    }

    teardown() {
      (0, _runloop.run)(() => {
        if (location) {
          location.destroy();
        }
      });
    }

    ['@test HistoryLocation initState does not get fired on init'](assert) {
      assert.expect(1);

      HistoryTestLocation.reopen({
        init() {
          assert.ok(true, 'init was called');
          this._super(...arguments);
        },
        initState() {
          assert.ok(false, 'initState() should not be called automatically');
        }
      });

      createLocation();
    }

    ["@test webkit doesn't fire popstate on page load"](assert) {
      assert.expect(1);

      HistoryTestLocation.reopen({
        initState() {
          this._super(...arguments);
          // these two should be equal to be able
          // to successfully detect webkit initial popstate
          assert.equal(this._previousURL, this.getURL());
        }
      });

      createLocation();
      location.initState();
    }

    ['@test base URL is removed when retrieving the current pathname'](assert) {
      assert.expect(1);

      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);

          (0, _metal.set)(this, 'location', mockBrowserLocation('/base/foo/bar'));
          (0, _metal.set)(this, 'baseURL', '/base/');
        },

        initState() {
          this._super(...arguments);

          assert.equal(this.getURL(), '/foo/bar');
        }
      });

      createLocation();
      location.initState();
    }

    ['@test base URL is preserved when moving around'](assert) {
      assert.expect(2);

      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);

          (0, _metal.set)(this, 'location', mockBrowserLocation('/base/foo/bar'));
          (0, _metal.set)(this, 'baseURL', '/base/');
        }
      });

      createLocation();
      location.initState();
      location.setURL('/one/two');

      assert.equal(location._historyState.path, '/base/one/two');
      assert.ok(location._historyState.uuid);
    }

    ['@test setURL continues to set even with a null state (iframes may set this)'](assert) {
      createLocation();
      location.initState();

      FakeHistory.pushState(null);
      location.setURL('/three/four');

      assert.equal(location._historyState.path, '/three/four');
      assert.ok(location._historyState.uuid);
    }

    ['@test replaceURL continues to set even with a null state (iframes may set this)'](assert) {
      createLocation();
      location.initState();

      FakeHistory.pushState(null);
      location.replaceURL('/three/four');

      assert.equal(location._historyState.path, '/three/four');
      assert.ok(location._historyState.uuid);
    }

    ['@test HistoryLocation.getURL() returns the current url, excluding both rootURL and baseURL'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);

          (0, _metal.set)(this, 'location', mockBrowserLocation('/base/foo/bar'));
          (0, _metal.set)(this, 'rootURL', '/app/');
          (0, _metal.set)(this, 'baseURL', '/base/');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar');
    }

    ['@test HistoryLocation.getURL() returns the current url, does not remove rootURL if its not at start of url'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);

          (0, _metal.set)(this, 'location', mockBrowserLocation('/foo/bar/baz'));
          (0, _metal.set)(this, 'rootURL', '/bar/');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar/baz');
    }

    ['@test HistoryLocation.getURL() will not remove the rootURL when only a partial match'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/bars/baz'));
          (0, _metal.set)(this, 'rootURL', '/bar/');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/bars/baz');
    }

    ['@test HistoryLocation.getURL() returns the current url, does not remove baseURL if its not at start of url'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);

          (0, _metal.set)(this, 'location', mockBrowserLocation('/foo/bar/baz'));
          (0, _metal.set)(this, 'baseURL', '/bar/');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar/baz');
    }

    ['@test HistoryLocation.getURL() will not remove the baseURL when only a partial match'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/bars/baz'));
          (0, _metal.set)(this, 'baseURL', '/bar/');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/bars/baz');
    }

    ['@test HistoryLocation.getURL() includes location.search'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/foo/bar?time=morphin'));
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar?time=morphin');
    }

    ['@test HistoryLocation.getURL() includes location.hash'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/foo/bar#pink-power-ranger'));
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar#pink-power-ranger');
    }

    ['@test HistoryLocation.getURL() includes location.hash and location.search'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/foo/bar?time=morphin#pink-power-ranger'));
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar?time=morphin#pink-power-ranger');
    }

    ['@test HistoryLocation.getURL() drops duplicate slashes'](assert) {
      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          let location = mockBrowserLocation('//');
          location.pathname = '//'; // mockBrowserLocation does not allow for `//`, so force it
          (0, _metal.set)(this, 'location', location);
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/');
    }

    ['@test Existing state is preserved on init'](assert) {
      let existingState = {
        path: '/route/path',
        uuid: 'abcd'
      };

      FakeHistory.state = existingState;

      HistoryTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'location', mockBrowserLocation('/route/path'));
        }
      });

      createLocation();
      location.initState();
      assert.deepEqual(location.getState(), existingState);
    }
  });
});
enifed('@ember/-internals/routing/tests/location/none_location_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/routing/lib/location/none_location', 'internal-test-helpers'], function (_runloop, _metal, _none_location, _internalTestHelpers) {
  'use strict';

  let NoneTestLocation, location;

  function createLocation(options) {
    if (!options) {
      options = {};
    }
    location = NoneTestLocation.create(options);
  }

  (0, _internalTestHelpers.moduleFor)('NoneLocation', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      NoneTestLocation = _none_location.default.extend({});
    }

    teardown() {
      (0, _runloop.run)(() => {
        if (location) {
          location.destroy();
        }
      });
    }

    ['@test NoneLocation.formatURL() returns the current url always appending rootURL'](assert) {
      NoneTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'rootURL', '/en/');
        }
      });

      createLocation();

      assert.equal(location.formatURL('/foo/bar'), '/en/foo/bar');
    }

    ['@test NoneLocation.getURL() returns the current path minus rootURL'](assert) {
      NoneTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'rootURL', '/foo/');
          (0, _metal.set)(this, 'path', '/foo/bar');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/bar');
    }

    ['@test NoneLocation.getURL() will remove the rootURL only from the beginning of a url'](assert) {
      NoneTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'rootURL', '/bar/');
          (0, _metal.set)(this, 'path', '/foo/bar/baz');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/foo/bar/baz');
    }

    ['@test NoneLocation.getURL() will not remove the rootURL when only a partial match'](assert) {
      NoneTestLocation.reopen({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'rootURL', '/bar/');
          (0, _metal.set)(this, 'path', '/bars/baz');
        }
      });

      createLocation();

      assert.equal(location.getURL(), '/bars/baz');
    }
  });
});
enifed('@ember/-internals/routing/tests/location/util_test', ['@ember/polyfills', '@ember/-internals/routing/lib/location/util', 'internal-test-helpers'], function (_polyfills, _util, _internalTestHelpers) {
  'use strict';

  function mockBrowserLocation(overrides, assert) {
    return (0, _polyfills.assign)({
      href: 'http://test.com/',
      pathname: '/',
      hash: '',
      search: '',
      replace() {
        assert.ok(false, 'location.replace should not be called during testing');
      }
    }, overrides);
  }

  (0, _internalTestHelpers.moduleFor)('Location Utilities', class extends _internalTestHelpers.AbstractTestCase {
    ['@test replacePath cannot be used to redirect to a different origin'](assert) {
      assert.expect(1);

      let expectedURL;

      let location = {
        protocol: 'http:',
        hostname: 'emberjs.com',
        port: '1337',

        replace(url) {
          assert.equal(url, expectedURL);
        }
      };

      expectedURL = 'http://emberjs.com:1337//google.com';
      (0, _util.replacePath)(location, '//google.com');
    }

    ['@test getPath() should normalize location.pathname, making sure it always returns a leading slash'](assert) {
      let location = mockBrowserLocation({ pathname: 'test' }, assert);
      assert.equal((0, _util.getPath)(location), '/test', 'When there is no leading slash, one is added.');

      location = mockBrowserLocation({ pathname: '/test' }, assert);
      assert.equal((0, _util.getPath)(location), '/test', "When a leading slash is already there, it isn't added again");
    }

    ['@test getQuery() should return location.search as-is'](assert) {
      let location = mockBrowserLocation({ search: '?foo=bar' }, assert);
      assert.equal((0, _util.getQuery)(location), '?foo=bar');
    }

    ['@test getFullPath() should return full pathname including query and hash'](assert) {
      let location = mockBrowserLocation({
        href: 'http://test.com/about?foo=bar#foo',
        pathname: '/about',
        search: '?foo=bar',
        hash: '#foo'
      }, assert);

      assert.equal((0, _util.getFullPath)(location), '/about?foo=bar#foo');
    }

    ['@test Feature-Detecting onhashchange'](assert) {
      assert.equal((0, _util.supportsHashChange)(undefined, { onhashchange() {} }), true, 'When not in IE, use onhashchange existence as evidence of the feature');
      assert.equal((0, _util.supportsHashChange)(undefined, {}), false, 'When not in IE, use onhashchange absence as evidence of the feature absence');
      assert.equal((0, _util.supportsHashChange)(7, { onhashchange() {} }), false, 'When in IE7 compatibility mode, never report existence of the feature');
      assert.equal((0, _util.supportsHashChange)(8, { onhashchange() {} }), true, 'When in IE8+, use onhashchange existence as evidence of the feature');
    }

    ['@test Feature-detecting the history API'](assert) {
      assert.equal((0, _util.supportsHistory)('', { pushState: true }), true, 'returns true if not Android Gingerbread and history.pushState exists');
      assert.equal((0, _util.supportsHistory)('', {}), false, "returns false if history.pushState doesn't exist");
      assert.equal((0, _util.supportsHistory)('', undefined), false, "returns false if history doesn't exist");

      assert.equal((0, _util.supportsHistory)('Mozilla/5.0 (Linux; U; Android 2.3.5; en-us; HTC Vision Build/GRI40) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1', { pushState: true }), false, 'returns false if Android 2.x stock browser (not Chrome) claiming to support pushState');

      assert.equal((0, _util.supportsHistory)('Mozilla/5.0 (Linux; U; Android 4.0.3; nl-nl; GT-N7000 Build/IML74K) AppleWebKit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30', { pushState: true }), false, 'returns false for Android 4.0.x stock browser (not Chrome) claiming to support pushState');

      assert.equal((0, _util.supportsHistory)('Mozilla/5.0 (Linux; U; Android 20.3.5; en-us; HTC Vision Build/GRI40) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1', { pushState: true }), true, 'returns true if Android version begins with 2, but is greater than 2');

      assert.equal((0, _util.supportsHistory)('Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.133 Mobile Safari/535.19', { pushState: true }), true, 'returns true for Chrome (not stock browser) on Android 4.0.x');

      // Windows Phone UA and History API: https://github.com/Modernizr/Modernizr/issues/1471
      assert.equal((0, _util.supportsHistory)('Mozilla/5.0 (Mobile; Windows Phone 8.1; Android 4.0; ARM; Trident/7.0; Touch; rv:11.0; IEMobile/11.0; Microsoft; Virtual) like iPhone OS 7_0_3 Mac OS X AppleWebKit/537 (KHTML, like Gecko) Mobile Safari/537', { pushState: true }), true, 'returns true for Windows Phone 8.1 with misleading user agent string');
    }
  });
});
enifed('@ember/-internals/routing/tests/system/cache_test', ['@ember/-internals/routing/lib/system/cache', 'internal-test-helpers'], function (_cache, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('BucketCache', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      this.cache = new _cache.default();
    }

    ['@test has - returns false when bucket is not in cache'](assert) {
      assert.strictEqual(this.cache.has('foo'), false);
      assert.strictEqual(this.cache.has('constructor'), false);
    }

    ['@test has - returns true when bucket is in cache'](assert) {
      let token = {};

      this.cache.stash('foo', 'bar', token);
      this.cache.stash('constructor', 'bar', token);

      assert.strictEqual(this.cache.has('foo'), true);
      assert.strictEqual(this.cache.has('constructor'), true);
    }

    ['@test lookup - returns stashed value if key does exist in bucket'](assert) {
      let token = {};
      let defaultValue = {};

      this.cache.stash('foo', 'bar', token);

      assert.strictEqual(this.cache.lookup('foo', 'bar', defaultValue), token);
    }

    ['@test lookup - returns default value if key does not exist in bucket'](assert) {
      let token = {};
      let defaultValue = {};

      this.cache.stash('foo', 'bar', token);

      assert.strictEqual(this.cache.lookup('foo', 'boo', defaultValue), defaultValue);
      assert.strictEqual(this.cache.lookup('foo', 'constructor', defaultValue), defaultValue);
    }

    ['@test lookup - returns default value if bucket does not exist'](assert) {
      let defaultValue = {};

      assert.strictEqual(this.cache.lookup('boo', 'bar', defaultValue), defaultValue);
      assert.strictEqual(this.cache.lookup('constructor', 'bar', defaultValue), defaultValue);
    }
  });
});
enifed('@ember/-internals/routing/tests/system/controller_for_test', ['@ember/controller', '@ember/-internals/routing/lib/system/controller_for', '@ember/-internals/routing/lib/system/generate_controller', 'internal-test-helpers', '@ember/debug'], function (_controller, _controller_for, _generate_controller, _internalTestHelpers, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');
  const noop = function () {};

  (0, _internalTestHelpers.moduleFor)('controllerFor', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super();
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
    }

    ['@test controllerFor should lookup for registered controllers'](assert) {
      this.add('controller:app', _controller.default.extend());

      return this.visit('/').then(() => {
        let appInstance = this.applicationInstance;
        let appController = appInstance.lookup('controller:app');
        let controller = (0, _controller_for.default)(appInstance, 'app');
        assert.equal(appController, controller, 'should find app controller');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('generateController', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super();
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
    }

    ['@test generateController should return Controller'](assert) {
      return this.visit('/').then(() => {
        let controller = (0, _generate_controller.default)(this.applicationInstance, 'home');
        assert.ok(controller instanceof _controller.default, 'should return controller');
      });
    }

    ['@test generateController should return controller:basic if resolved'](assert) {
      let BasicController = _controller.default.extend();
      this.add('controller:basic', BasicController);

      return this.visit('/').then(() => {
        let controller = (0, _generate_controller.default)(this.applicationInstance, 'home');
        assert.ok(controller instanceof BasicController, 'should return controller');
      });
    }

    ['@test generateController should return controller:basic if registered'](assert) {
      let BasicController = _controller.default.extend();
      this.application.register('controller:basic', BasicController);

      return this.visit('/').then(() => {
        let controller = (0, _generate_controller.default)(this.applicationInstance, 'home');

        assert.ok(controller instanceof BasicController, 'should return base class of controller');
      });
    }
  });
});
enifed('@ember/-internals/routing/tests/system/dsl_test', ['@ember/-internals/owner', '@ember/-internals/routing/lib/system/router', 'internal-test-helpers'], function (_owner, _router, _internalTestHelpers) {
  'use strict';

  let Router;

  (0, _internalTestHelpers.moduleFor)('Ember Router DSL', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      Router = _router.default.extend();
    }

    teardown() {
      Router = null;
    }

    ['@test should fail when using a reserved route name'](assert) {
      let reservedNames = ['basic', 'application'];

      assert.expect(reservedNames.length);

      reservedNames.forEach(reservedName => {
        expectAssertion(() => {
          Router = _router.default.extend();

          Router.map(function () {
            this.route(reservedName);
          });

          let router = Router.create();
          router._initRouterJs();
        }, "'" + reservedName + "' cannot be used as a route name.");
      });
    }

    ['@test [GH#16642] better error when using a colon in a route name']() {
      expectAssertion(() => {
        Router = _router.default.extend();

        Router.map(function () {
          this.route('resource/:id');
        });

        let router = Router.create();
        router._initRouterJs();
      }, "'resource/:id' is not a valid route name. It cannot contain a ':'. You may want to use the 'path' option instead.");
    }

    ['@test should retain resource namespace if nested with routes'](assert) {
      Router = Router.map(function () {
        this.route('bleep', function () {
          this.route('bloop', function () {
            this.route('blork');
          });
        });
      });

      let router = Router.create();
      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['bleep'], 'parent name was used as base of nested routes');
      assert.ok(router._routerMicrolib.recognizer.names['bleep.bloop'], 'parent name was used as base of nested routes');
      assert.ok(router._routerMicrolib.recognizer.names['bleep.bloop.blork'], 'parent name was used as base of nested routes');
    }

    ['@test should add loading and error routes if _isRouterMapResult is true'](assert) {
      Router.map(function () {
        this.route('blork');
      });

      let router = Router.create({
        _hasModuleBasedResolver() {
          return true;
        }
      });

      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['blork'], 'main route was created');
      assert.ok(router._routerMicrolib.recognizer.names['blork_loading'], 'loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['blork_error'], 'error route was added');
    }

    ['@test should not add loading and error routes if _isRouterMapResult is false'](assert) {
      Router.map(function () {
        this.route('blork');
      });

      let router = Router.create();
      router._initRouterJs(false);

      assert.ok(router._routerMicrolib.recognizer.names['blork'], 'main route was created');
      assert.ok(!router._routerMicrolib.recognizer.names['blork_loading'], 'loading route was not added');
      assert.ok(!router._routerMicrolib.recognizer.names['blork_error'], 'error route was not added');
    }

    ['@test should reset namespace of loading and error routes for routes with resetNamespace'](assert) {
      Router.map(function () {
        this.route('blork', function () {
          this.route('blorp');
          this.route('bleep', { resetNamespace: true });
        });
      });

      let router = Router.create({
        _hasModuleBasedResolver() {
          return true;
        }
      });

      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['blork.blorp'], 'nested route was created');
      assert.ok(router._routerMicrolib.recognizer.names['blork.blorp_loading'], 'nested loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['blork.blorp_error'], 'nested error route was added');

      assert.ok(router._routerMicrolib.recognizer.names['bleep'], 'reset route was created');
      assert.ok(router._routerMicrolib.recognizer.names['bleep_loading'], 'reset loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['bleep_error'], 'reset error route was added');

      assert.ok(!router._routerMicrolib.recognizer.names['blork.bleep'], 'nested reset route was not created');
      assert.ok(!router._routerMicrolib.recognizer.names['blork.bleep_loading'], 'nested reset loading route was not added');
      assert.ok(!router._routerMicrolib.recognizer.names['blork.bleep_error'], 'nested reset error route was not added');
    }

    ['@test should throw an error when defining a route serializer outside an engine'](assert) {
      Router.map(function () {
        assert.throws(() => {
          this.route('posts', { serialize: function () {} });
        }, /Defining a route serializer on route 'posts' outside an Engine is not allowed/);
      });

      Router.create()._initRouterJs();
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember Router DSL with engines', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      Router = _router.default.extend();
    }

    teardown() {
      Router = null;
    }

    ['@test should allow mounting of engines'](assert) {
      assert.expect(3);

      Router = Router.map(function () {
        this.route('bleep', function () {
          this.route('bloop', function () {
            this.mount('chat');
          });
        });
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create();
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['bleep'], 'parent name was used as base of nested routes');
      assert.ok(router._routerMicrolib.recognizer.names['bleep.bloop'], 'parent name was used as base of nested routes');
      assert.ok(router._routerMicrolib.recognizer.names['bleep.bloop.chat'], 'parent name was used as base of mounted engine');
    }

    ['@test should allow mounting of engines at a custom path'](assert) {
      assert.expect(1);

      Router = Router.map(function () {
        this.route('bleep', function () {
          this.route('bloop', function () {
            this.mount('chat', { path: 'custom-chat' });
          });
        });
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create();
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.deepEqual(router._routerMicrolib.recognizer.names['bleep.bloop.chat'].segments.slice(1, 4).map(s => s.value), ['bleep', 'bloop', 'custom-chat'], 'segments are properly associated with mounted engine');
    }

    ['@test should allow aliasing of engine names with `as`'](assert) {
      assert.expect(1);

      Router = Router.map(function () {
        this.route('bleep', function () {
          this.route('bloop', function () {
            this.mount('chat', { as: 'blork' });
          });
        });
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create();
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.deepEqual(router._routerMicrolib.recognizer.names['bleep.bloop.blork'].segments.slice(1, 4).map(s => s.value), ['bleep', 'bloop', 'blork'], 'segments are properly associated with mounted engine with aliased name');
    }

    ['@test should add loading and error routes to a mount if _isRouterMapResult is true'](assert) {
      Router.map(function () {
        this.mount('chat');
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create({
        _hasModuleBasedResolver() {
          return true;
        }
      });
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['chat'], 'main route was created');
      assert.ok(router._routerMicrolib.recognizer.names['chat_loading'], 'loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['chat_error'], 'error route was added');
    }

    ['@test should add loading and error routes to a mount alias if _isRouterMapResult is true'](assert) {
      Router.map(function () {
        this.mount('chat', { as: 'shoutbox' });
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create({
        _hasModuleBasedResolver() {
          return true;
        }
      });
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['shoutbox'], 'main route was created');
      assert.ok(router._routerMicrolib.recognizer.names['shoutbox_loading'], 'loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['shoutbox_error'], 'error route was added');
    }

    ['@test should not add loading and error routes to a mount if _isRouterMapResult is false'](assert) {
      Router.map(function () {
        this.mount('chat');
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create();
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs(false);

      assert.ok(router._routerMicrolib.recognizer.names['chat'], 'main route was created');
      assert.ok(!router._routerMicrolib.recognizer.names['chat_loading'], 'loading route was not added');
      assert.ok(!router._routerMicrolib.recognizer.names['chat_error'], 'error route was not added');
    }

    ['@test should reset namespace of loading and error routes for mounts with resetNamespace'](assert) {
      Router.map(function () {
        this.route('news', function () {
          this.mount('chat');
          this.mount('blog', { resetNamespace: true });
        });
      });

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: { routable: true }
      });

      let router = Router.create({
        _hasModuleBasedResolver() {
          return true;
        }
      });
      (0, _owner.setOwner)(router, engineInstance);
      router._initRouterJs();

      assert.ok(router._routerMicrolib.recognizer.names['news.chat'], 'nested route was created');
      assert.ok(router._routerMicrolib.recognizer.names['news.chat_loading'], 'nested loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['news.chat_error'], 'nested error route was added');

      assert.ok(router._routerMicrolib.recognizer.names['blog'], 'reset route was created');
      assert.ok(router._routerMicrolib.recognizer.names['blog_loading'], 'reset loading route was added');
      assert.ok(router._routerMicrolib.recognizer.names['blog_error'], 'reset error route was added');

      assert.ok(!router._routerMicrolib.recognizer.names['news.blog'], 'nested reset route was not created');
      assert.ok(!router._routerMicrolib.recognizer.names['news.blog_loading'], 'nested reset loading route was not added');
      assert.ok(!router._routerMicrolib.recognizer.names['news.blog_error'], 'nested reset error route was not added');
    }
  });
});
enifed('@ember/-internals/routing/tests/system/route_test', ['@ember/-internals/owner', 'internal-test-helpers', '@ember/service', '@ember/-internals/runtime', '@ember/-internals/routing/lib/system/route'], function (_owner, _internalTestHelpers, _service, _runtime, _route) {
  'use strict';

  let route, routeOne, routeTwo, lookupHash;

  (0, _internalTestHelpers.moduleFor)('Route', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      route = _route.default.create();
    }

    teardown() {
      super.teardown();
      (0, _internalTestHelpers.runDestroy)(route);
      route = routeOne = routeTwo = lookupHash = undefined;
    }

    ['@test default store utilizes the container to acquire the model factory'](assert) {
      assert.expect(4);

      let Post = _runtime.Object.extend();
      let post = {};

      Post.reopenClass({
        find() {
          return post;
        }
      });

      let ownerOptions = {
        ownerOptions: {
          hasRegistration() {
            return true;
          },
          factoryFor(fullName) {
            assert.equal(fullName, 'model:post', 'correct factory was looked up');

            return {
              class: Post,
              create() {
                return Post.create();
              }
            };
          }
        }
      };

      let owner = (0, _internalTestHelpers.buildOwner)(ownerOptions);
      (0, _owner.setOwner)(route, owner);

      route.set('_qp', null);

      assert.equal(route.model({ post_id: 1 }), post);
      assert.equal(route.findModel('post', 1), post, '#findModel returns the correct post');

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ["@test 'store' can be injected by data persistence frameworks"](assert) {
      assert.expect(8);
      (0, _internalTestHelpers.runDestroy)(route);

      let owner = (0, _internalTestHelpers.buildOwner)();

      let post = {
        id: 1
      };

      let Store = _runtime.Object.extend({
        find(type, value) {
          assert.ok(true, 'injected model was called');
          assert.equal(type, 'post', 'correct type was called');
          assert.equal(value, 1, 'correct value was called');
          return post;
        }
      });

      owner.register('route:index', _route.default);
      owner.register('store:main', Store);

      owner.inject('route', 'store', 'store:main');

      route = owner.lookup('route:index');

      assert.equal(route.model({ post_id: 1 }), post, '#model returns the correct post');
      assert.equal(route.findModel('post', 1), post, '#findModel returns the correct post');

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ["@test assert if 'store.find' method is not found"]() {
      (0, _internalTestHelpers.runDestroy)(route);

      let owner = (0, _internalTestHelpers.buildOwner)();
      let Post = _runtime.Object.extend();

      owner.register('route:index', _route.default);
      owner.register('model:post', Post);

      route = owner.lookup('route:index');

      expectAssertion(function () {
        route.findModel('post', 1);
      }, 'Post has no method `find`.');

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ['@test asserts if model class is not found']() {
      (0, _internalTestHelpers.runDestroy)(route);

      let owner = (0, _internalTestHelpers.buildOwner)();
      owner.register('route:index', _route.default);

      route = owner.lookup('route:index');

      expectAssertion(function () {
        route.model({ post_id: 1 });
      }, /You used the dynamic segment post_id in your route undefined, but <Ember.Object:ember\d+>.Post did not exist and you did not override your route\'s `model` hook./);

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ["@test 'store' does not need to be injected"](assert) {
      (0, _internalTestHelpers.runDestroy)(route);

      let owner = (0, _internalTestHelpers.buildOwner)();

      owner.register('route:index', _route.default);

      route = owner.lookup('route:index');

      ignoreAssertion(function () {
        route.model({ post_id: 1 });
      });

      assert.ok(true, 'no error was raised');

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ["@test modelFor doesn't require the router"](assert) {
      let owner = (0, _internalTestHelpers.buildOwner)();
      (0, _owner.setOwner)(route, owner);

      let foo = { name: 'foo' };

      let FooRoute = _route.default.extend({
        currentModel: foo
      });

      owner.register('route:foo', FooRoute);

      assert.strictEqual(route.modelFor('foo'), foo);

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ["@test modelFor doesn't require the routerMicrolib"](assert) {
      let route = _route.default.create({
        _router: { _routerMicrolib: null }
      });

      let owner = (0, _internalTestHelpers.buildOwner)();
      (0, _owner.setOwner)(route, owner);

      let foo = { name: 'foo' };

      let FooRoute = _route.default.extend({
        currentModel: foo
      });

      owner.register('route:foo', FooRoute);

      assert.strictEqual(route.modelFor('foo'), foo);

      (0, _internalTestHelpers.runDestroy)(owner);
    }

    ['@test .send just calls an action if the router is absent'](assert) {
      assert.expect(7);
      let route = _route.default.extend({
        actions: {
          returnsTrue(foo, bar) {
            assert.equal(foo, 1);
            assert.equal(bar, 2);
            assert.equal(this, route);
            return true;
          },

          returnsFalse() {
            assert.ok(true, 'returnsFalse was called');
            return false;
          }
        }
      }).create();

      assert.equal(route.send('returnsTrue', 1, 2), true);
      assert.equal(route.send('returnsFalse'), false);
      assert.equal(route.send('nonexistent', 1, 2, 3), undefined);

      (0, _internalTestHelpers.runDestroy)(route);
    }

    ['@test .send just calls an action if the routers internal router property is absent'](assert) {
      assert.expect(7);
      let route = _route.default.extend({
        router: {},
        actions: {
          returnsTrue(foo, bar) {
            assert.equal(foo, 1);
            assert.equal(bar, 2);
            assert.equal(this, route);
            return true;
          },

          returnsFalse() {
            assert.ok(true, 'returnsFalse was called');
            return false;
          }
        }
      }).create();

      assert.equal(true, route.send('returnsTrue', 1, 2));
      assert.equal(false, route.send('returnsFalse'));
      assert.equal(undefined, route.send('nonexistent', 1, 2, 3));

      (0, _internalTestHelpers.runDestroy)(route);
    }

    ['@test .send asserts if called on a destroyed route']() {
      route.routeName = 'rip-alley';
      (0, _internalTestHelpers.runDestroy)(route);

      expectAssertion(() => {
        route.send('trigger-me-dead');
      }, "Attempted to call .send() with the action 'trigger-me-dead' on the destroyed route 'rip-alley'.");
    }
  });

  (0, _internalTestHelpers.moduleFor)('Route serialize', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      route = _route.default.create();
    }

    teardown() {
      (0, _internalTestHelpers.runDestroy)(route);
    }

    ['@test returns the models properties if params does not include *_id'](assert) {
      let model = { id: 2, firstName: 'Ned', lastName: 'Ryerson' };

      assert.deepEqual(route.serialize(model, ['firstName', 'lastName']), { firstName: 'Ned', lastName: 'Ryerson' }, 'serialized correctly');
    }

    ['@test returns model.id if params include *_id'](assert) {
      let model = { id: 2 };

      assert.deepEqual(route.serialize(model, ['post_id']), { post_id: 2 }, 'serialized correctly');
    }

    ['@test returns checks for existence of model.post_id before trying model.id'](assert) {
      let model = { post_id: 3 };

      assert.deepEqual(route.serialize(model, ['post_id']), { post_id: 3 }, 'serialized correctly');
    }

    ['@test returns undefined if model is not set'](assert) {
      assert.equal(route.serialize(undefined, ['post_id']), undefined, 'serialized correctly');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Route interaction', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      let owner = {
        lookup(fullName) {
          return lookupHash[fullName];
        }
      };

      routeOne = _route.default.create({ routeName: 'one' });
      routeTwo = _route.default.create({ routeName: 'two' });

      (0, _owner.setOwner)(routeOne, owner);
      (0, _owner.setOwner)(routeTwo, owner);

      lookupHash = {
        'route:one': routeOne,
        'route:two': routeTwo
      };
    }

    teardown() {
      (0, _internalTestHelpers.runDestroy)(routeOne);
      (0, _internalTestHelpers.runDestroy)(routeTwo);
    }

    ['@test route._qp does not crash if the controller has no QP, or setProperties'](assert) {
      lookupHash['controller:test'] = {};

      routeOne.controllerName = 'test';
      let qp = routeOne.get('_qp');

      assert.deepEqual(qp.map, {}, 'map should be empty');
      assert.deepEqual(qp.propertyNames, [], 'property names should be empty');
      assert.deepEqual(qp.qps, [], 'qps is should be empty');
    }

    ["@test controllerFor uses route's controllerName if specified"](assert) {
      let testController = {};
      lookupHash['controller:test'] = testController;

      routeOne.controllerName = 'test';

      assert.equal(routeTwo.controllerFor('one'), testController);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Route injected properties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test services can be injected into routes'](assert) {
      let owner = (0, _internalTestHelpers.buildOwner)();

      owner.register('route:application', _route.default.extend({
        authService: (0, _service.inject)('auth')
      }));

      owner.register('service:auth', _service.default.extend());

      let appRoute = owner.lookup('route:application');
      let authService = owner.lookup('service:auth');

      assert.equal(authService, appRoute.get('authService'), 'service.auth is injected');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Route with engines', class extends _internalTestHelpers.AbstractTestCase {
    ["@test paramsFor considers an engine's mountPoint"](assert) {
      let router = {
        _deserializeQueryParams() {},
        _routerMicrolib: {
          state: {
            routeInfos: [{ name: 'posts' }],
            params: {
              'foo.bar': { a: 'b' },
              'foo.bar.posts': { c: 'd' }
            }
          }
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,

          mountPoint: 'foo.bar',

          lookup(name) {
            if (name === 'route:posts') {
              return postsRoute;
            } else if (name === 'route:application') {
              return applicationRoute;
            }
          }
        }
      });

      let applicationRoute = _route.default.create({
        _router: router,
        routeName: 'application',
        fullRouteName: 'foo.bar'
      });
      let postsRoute = _route.default.create({
        _router: router,
        routeName: 'posts',
        fullRouteName: 'foo.bar.posts'
      });
      let route = _route.default.create({ _router: router });

      (0, _owner.setOwner)(applicationRoute, engineInstance);
      (0, _owner.setOwner)(postsRoute, engineInstance);
      (0, _owner.setOwner)(route, engineInstance);

      assert.deepEqual(route.paramsFor('application'), { a: 'b' }, 'params match for root `application` route in engine');
      assert.deepEqual(route.paramsFor('posts'), { c: 'd' }, 'params match for `posts` route in engine');
    }

    ["@test modelFor considers an engine's mountPoint"](assert) {
      let applicationModel = { id: '1' };
      let postsModel = { id: '2' };

      let router = {
        _routerMicrolib: {
          activeTransition: {
            resolvedModels: {
              'foo.bar': applicationModel,
              'foo.bar.posts': postsModel
            }
          }
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,

          mountPoint: 'foo.bar',

          lookup(name) {
            if (name === 'route:posts') {
              return postsRoute;
            } else if (name === 'route:application') {
              return applicationRoute;
            }
          }
        }
      });

      let applicationRoute = _route.default.create({
        _router: router,
        routeName: 'application'
      });
      let postsRoute = _route.default.create({
        _router: router,
        routeName: 'posts'
      });
      let route = _route.default.create({ _router: router });

      (0, _owner.setOwner)(applicationRoute, engineInstance);
      (0, _owner.setOwner)(postsRoute, engineInstance);
      (0, _owner.setOwner)(route, engineInstance);

      assert.strictEqual(route.modelFor('application'), applicationModel);
      assert.strictEqual(route.modelFor('posts'), postsModel);
    }

    ["@test transitionTo considers an engine's mountPoint"](assert) {
      let router = {
        transitionTo(route) {
          return route;
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,
          mountPoint: 'foo.bar'
        }
      });

      let route = _route.default.create({ _router: router });
      (0, _owner.setOwner)(route, engineInstance);

      assert.strictEqual(route.transitionTo('application'), 'foo.bar.application', 'properly prefixes application route');
      assert.strictEqual(route.transitionTo('posts'), 'foo.bar.posts', 'properly prefixes child routes');
      assert.throws(() => route.transitionTo('/posts'), 'throws when trying to use a url');

      let queryParams = {};
      assert.strictEqual(route.transitionTo(queryParams), queryParams, 'passes query param only transitions through');
    }

    ["@test intermediateTransitionTo considers an engine's mountPoint"](assert) {
      let lastRoute;
      let router = {
        intermediateTransitionTo(route) {
          lastRoute = route;
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,
          mountPoint: 'foo.bar'
        }
      });

      let route = _route.default.create({ _router: router });
      (0, _owner.setOwner)(route, engineInstance);

      route.intermediateTransitionTo('application');
      assert.strictEqual(lastRoute, 'foo.bar.application', 'properly prefixes application route');

      route.intermediateTransitionTo('posts');
      assert.strictEqual(lastRoute, 'foo.bar.posts', 'properly prefixes child routes');

      assert.throws(() => route.intermediateTransitionTo('/posts'), 'throws when trying to use a url');

      let queryParams = {};
      route.intermediateTransitionTo(queryParams);
      assert.strictEqual(lastRoute, queryParams, 'passes query param only transitions through');
    }

    ["@test replaceWith considers an engine's mountPoint"](assert) {
      let router = {
        replaceWith(route) {
          return route;
        }
      };

      let engineInstance = (0, _internalTestHelpers.buildOwner)({
        ownerOptions: {
          routable: true,
          mountPoint: 'foo.bar'
        }
      });

      let route = _route.default.create({ _router: router });
      (0, _owner.setOwner)(route, engineInstance);

      assert.strictEqual(route.replaceWith('application'), 'foo.bar.application', 'properly prefixes application route');
      assert.strictEqual(route.replaceWith('posts'), 'foo.bar.posts', 'properly prefixes child routes');
      assert.throws(() => route.replaceWith('/posts'), 'throws when trying to use a url');

      let queryParams = {};
      assert.strictEqual(route.replaceWith(queryParams), queryParams, 'passes query param only transitions through');
    }
  });
});
enifed('@ember/-internals/routing/tests/system/router_test', ['@ember/-internals/owner', '@ember/-internals/routing/lib/location/hash_location', '@ember/-internals/routing/lib/location/history_location', '@ember/-internals/routing/lib/location/auto_location', '@ember/-internals/routing/lib/location/none_location', '@ember/-internals/routing/lib/system/router', 'internal-test-helpers'], function (_owner, _hash_location, _history_location, _auto_location, _none_location, _router, _internalTestHelpers) {
  'use strict';

  let owner;

  function createRouter(settings, options = {}) {
    let CustomRouter = _router.default.extend();
    let router = CustomRouter.create(settings);

    if (!options.skipOwner) {
      (0, _owner.setOwner)(router, owner);
    }

    if (!options.disableSetup) {
      router.setupRouter();
    }

    return router;
  }

  (0, _internalTestHelpers.moduleFor)('Ember Router', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      owner = (0, _internalTestHelpers.buildOwner)();

      //register the HashLocation (the default)
      owner.register('location:hash', _hash_location.default);
      owner.register('location:history', _history_location.default);
      owner.register('location:auto', _auto_location.default);
      owner.register('location:none', _none_location.default);
    }

    teardown() {
      (0, _internalTestHelpers.runDestroy)(owner);
      owner = null;
    }

    ['@test can create a router without an owner'](assert) {
      createRouter(undefined, { disableSetup: true, skipOwner: true });

      assert.ok(true, 'no errors were thrown when creating without a container');
    }

    ['@test [GH#15237] EmberError is imported correctly'](assert) {
      // If we get the right message it means Error is being imported correctly.
      assert.throws(function () {
        (0, _router.triggerEvent)(null, false, []);
      }, /because your app hasn't finished transitioning/);
    }

    ['@test should not create a router.js instance upon init'](assert) {
      let router = createRouter(undefined, { disableSetup: true });

      assert.ok(!router._routerMicrolib);
    }

    ['@test should not reify location until setupRouter is called'](assert) {
      let router = createRouter(undefined, { disableSetup: true });
      assert.equal(typeof router.location, 'string', 'location is specified as a string');

      router.setupRouter();

      assert.equal(typeof router.location, 'object', 'location is reified into an object');
    }

    ['@test should destroy its location upon destroying the routers owner.'](assert) {
      let router = createRouter();
      let location = router.get('location');

      (0, _internalTestHelpers.runDestroy)(owner);

      assert.ok(location.isDestroyed, 'location should be destroyed');
    }

    ['@test should instantiate its location with its `rootURL`'](assert) {
      let router = createRouter({
        rootURL: '/rootdir/'
      });
      let location = router.get('location');

      assert.equal(location.get('rootURL'), '/rootdir/');
    }

    ['@test replacePath should be called with the right path'](assert) {
      assert.expect(1);

      let location = owner.lookup('location:auto');

      let browserLocation = {
        href: 'http://test.com/rootdir/welcome',
        origin: 'http://test.com',
        pathname: '/rootdir/welcome',
        hash: '',
        search: '',
        replace(url) {
          assert.equal(url, 'http://test.com/rootdir/#/welcome');
        }
      };

      location.location = browserLocation;
      location.global = { onhashchange() {} };
      location.history = null;

      createRouter({
        location: 'auto',
        rootURL: '/rootdir/'
      });
    }

    ['@test Router._routePath should consume identical prefixes'](assert) {
      createRouter();

      function routePath() {
        let routeInfos = Array.prototype.slice.call(arguments).map(function (s) {
          return { name: s };
        });
        routeInfos.unshift({ name: 'ignored' });

        return _router.default._routePath(routeInfos);
      }

      assert.equal(routePath('foo'), 'foo');
      assert.equal(routePath('foo', 'bar', 'baz'), 'foo.bar.baz');
      assert.equal(routePath('foo', 'foo.bar'), 'foo.bar');
      assert.equal(routePath('foo', 'foo.bar', 'foo.bar.baz'), 'foo.bar.baz');
      assert.equal(routePath('foo', 'foo.bar', 'foo.bar.baz.wow'), 'foo.bar.baz.wow');
      assert.equal(routePath('foo', 'foo.bar.baz.wow'), 'foo.bar.baz.wow');
      assert.equal(routePath('foo.bar', 'bar.baz.wow'), 'foo.bar.baz.wow');

      // This makes no sense, not trying to handle it, just
      // making sure it doesn't go boom.
      assert.equal(routePath('foo.bar.baz', 'foo'), 'foo.bar.baz.foo');
    }

    ['@test Router should cancel routing setup when the Location class says so via cancelRouterSetup'](assert) {
      assert.expect(0);

      let router;
      let FakeLocation = {
        cancelRouterSetup: true,
        create() {
          return this;
        }
      };

      owner.register('location:fake', FakeLocation);

      router = createRouter({
        location: 'fake',

        _setupRouter() {
          assert.ok(false, '_setupRouter should not be called');
        }
      });

      router.startRouting();
    }

    ["@test AutoLocation should replace the url when it's not in the preferred format"](assert) {
      assert.expect(1);

      let location = owner.lookup('location:auto');

      location.location = {
        href: 'http://test.com/rootdir/welcome',
        origin: 'http://test.com',
        pathname: '/rootdir/welcome',
        hash: '',
        search: '',
        replace(url) {
          assert.equal(url, 'http://test.com/rootdir/#/welcome');
        }
      };
      location.history = null;
      location.global = {
        onhashchange() {}
      };

      createRouter({
        location: 'auto',
        rootURL: '/rootdir/'
      });
    }

    ['@test Router#handleURL should remove any #hashes before doing URL transition'](assert) {
      assert.expect(2);

      let router = createRouter({
        _doURLTransition(routerJsMethod, url) {
          assert.equal(routerJsMethod, 'handleURL');
          assert.equal(url, '/foo/bar?time=morphin');
        }
      });

      router.handleURL('/foo/bar?time=morphin#pink-power-ranger');
    }

    ['@test Router#triggerEvent allows actions to bubble when returning true'](assert) {
      assert.expect(2);

      let routeInfos = [{
        name: 'application',
        route: {
          actions: {
            loading() {
              assert.ok(false, 'loading not handled by application route');
            }
          }
        }
      }, {
        name: 'about',
        route: {
          actions: {
            loading() {
              assert.ok(true, 'loading handled by about route');
              return false;
            }
          }
        }
      }, {
        name: 'about.me',
        route: {
          actions: {
            loading() {
              assert.ok(true, 'loading handled by about.me route');
              return true;
            }
          }
        }
      }];

      (0, _router.triggerEvent)(routeInfos, false, ['loading']);
    }

    ['@test Router#triggerEvent ignores handlers that have not loaded yet'](assert) {
      assert.expect(1);

      let routeInfos = [{
        name: 'about',
        route: {
          actions: {
            loading() {
              assert.ok(true, 'loading handled by about route');
            }
          }
        }
      }, {
        name: 'about.me',
        route: undefined
      }];

      (0, _router.triggerEvent)(routeInfos, false, ['loading']);
    }

    ['@test transitionTo should throw an error when called after owner is destroyed']() {
      let router = createRouter();

      (0, _internalTestHelpers.runDestroy)(router);

      router.currentRouteName = 'route-a';

      expectAssertion(function () {
        router.transitionTo('route-b');
      }, "A transition was attempted from 'route-a' to 'route-b' but the application instance has already been destroyed.");

      expectAssertion(function () {
        router.transitionTo('./route-b/1');
      }, "A transition was attempted from 'route-a' to './route-b/1' but the application instance has already been destroyed.");
    }
  });
});
enifed('@ember/-internals/routing/tests/utils_test', ['@ember/-internals/routing/lib/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Routing query parameter utils - normalizeControllerQueryParams', class extends _internalTestHelpers.AbstractTestCase {
    ['@test converts array style into verbose object style'](assert) {
      let paramName = 'foo';
      let params = [paramName];
      let normalized = (0, _utils.normalizeControllerQueryParams)(params);

      assert.ok(normalized[paramName], 'turns the query param name into key');
      assert.equal(normalized[paramName].as, null, "includes a blank alias in 'as' key");
      assert.equal(normalized[paramName].scope, 'model', 'defaults scope to model');
    }

    ["@test converts object style [{foo: 'an_alias'}]"](assert) {
      let paramName = 'foo';
      let params = [{ foo: 'an_alias' }];
      let normalized = (0, _utils.normalizeControllerQueryParams)(params);

      assert.ok(normalized[paramName], 'retains the query param name as key');
      assert.equal(normalized[paramName].as, 'an_alias', "includes the provided alias in 'as' key");
      assert.equal(normalized[paramName].scope, 'model', 'defaults scope to model');
    }

    ["@test retains maximally verbose object style [{foo: {as: 'foo'}}]"](assert) {
      let paramName = 'foo';
      let params = [{ foo: { as: 'an_alias' } }];
      let normalized = (0, _utils.normalizeControllerQueryParams)(params);

      assert.ok(normalized[paramName], 'retains the query param name as key');
      assert.equal(normalized[paramName].as, 'an_alias', "includes the provided alias in 'as' key");
      assert.equal(normalized[paramName].scope, 'model', 'defaults scope to model');
    }
  });
});
enifed('@ember/-internals/runtime/tests/array/any-test', ['@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_array, _internalTestHelpers, _array2) {
  'use strict';

  class AnyTests extends _internalTestHelpers.AbstractTestCase {
    '@test any should should invoke callback on each item as long as you return false'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let found = [];
      let result;

      result = obj.any(function (i) {
        found.push(i);
        return false;
      });

      this.assert.equal(result, false, 'return value of obj.any');
      this.assert.deepEqual(found, ary, 'items passed during any() should match');
    }

    '@test any should stop invoking when you return true'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let cnt = ary.length - 2;
      let exp = cnt;
      let found = [];
      let result;

      result = obj.any(function (i) {
        found.push(i);
        return --cnt <= 0;
      });
      this.assert.equal(result, true, 'return value of obj.any');
      this.assert.equal(found.length, exp, 'should invoke proper number of times');
      this.assert.deepEqual(found, ary.slice(0, -2), 'items passed during any() should match');
    }

    '@test any should return true if any object matches the callback'() {
      let obj = (0, _array.A)([0, 1, 2]);
      let result;

      result = obj.any(i => !!i);
      this.assert.equal(result, true, 'return value of obj.any');
    }

    '@test any should produce correct results even if the matching element is undefined'(assert) {
      let obj = (0, _array.A)([undefined]);
      let result;

      result = obj.any(() => true);
      assert.equal(result, true, 'return value of obj.any');
    }
  }

  (0, _array2.runArrayTests)('any', AnyTests);
});
enifed('@ember/-internals/runtime/tests/array/compact-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class CompactTests extends _internalTestHelpers.AbstractTestCase {
    '@test removes null and undefined values from enumerable'() {
      let obj = this.newObject([null, 1, false, '', undefined, 0, null]);
      let ary = obj.compact();
      this.assert.deepEqual(ary, [1, false, '', 0]);
    }
  }

  (0, _array.runArrayTests)('compact', CompactTests);
});
enifed('@ember/-internals/runtime/tests/array/every-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/runtime/lib/system/object'], function (_internalTestHelpers, _array, _object) {
  'use strict';

  class EveryTest extends _internalTestHelpers.AbstractTestCase {
    '@test every should should invoke callback on each item as long as you return true'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let found = [];
      let result;

      result = obj.every(function (i) {
        found.push(i);
        return true;
      });
      this.assert.equal(result, true, 'return value of obj.every');
      this.assert.deepEqual(found, ary, 'items passed during every() should match');
    }

    '@test every should stop invoking when you return false'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let cnt = ary.length - 2;
      let exp = cnt;
      let found = [];
      let result;

      result = obj.every(function (i) {
        found.push(i);
        return --cnt > 0;
      });
      this.assert.equal(result, false, 'return value of obj.every');
      this.assert.equal(found.length, exp, 'should invoke proper number of times');
      this.assert.deepEqual(found, ary.slice(0, -2), 'items passed during every() should match');
    }
  }

  class IsEveryTest extends _internalTestHelpers.AbstractTestCase {
    '@test should return true of every property matches'() {
      let obj = this.newObject([{ foo: 'foo', bar: 'BAZ' }, _object.default.create({ foo: 'foo', bar: 'bar' })]);

      this.assert.equal(obj.isEvery('foo', 'foo'), true, 'isEvery(foo)');
      this.assert.equal(obj.isEvery('bar', 'bar'), false, 'isEvery(bar)');
    }

    '@test should return true of every property is true'() {
      let obj = this.newObject([{ foo: 'foo', bar: true }, _object.default.create({ foo: 'bar', bar: false })]);

      // different values - all eval to true
      this.assert.equal(obj.isEvery('foo'), true, 'isEvery(foo)');
      this.assert.equal(obj.isEvery('bar'), false, 'isEvery(bar)');
    }

    '@test should return true if every property matches null'() {
      let obj = this.newObject([{ foo: null, bar: 'BAZ' }, _object.default.create({ foo: null, bar: null })]);

      this.assert.equal(obj.isEvery('foo', null), true, "isEvery('foo', null)");
      this.assert.equal(obj.isEvery('bar', null), false, "isEvery('bar', null)");
    }

    '@test should return true if every property is undefined'() {
      let obj = this.newObject([{ foo: undefined, bar: 'BAZ' }, _object.default.create({ bar: undefined })]);

      this.assert.equal(obj.isEvery('foo', undefined), true, "isEvery('foo', undefined)");
      this.assert.equal(obj.isEvery('bar', undefined), false, "isEvery('bar', undefined)");
    }
  }

  (0, _array.runArrayTests)('every', EveryTest);
  (0, _array.runArrayTests)('isEvery', IsEveryTest);
});
enifed('@ember/-internals/runtime/tests/array/filter-test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_object, _internalTestHelpers, _array) {
  'use strict';

  class FilterTest extends _internalTestHelpers.AbstractTestCase {
    '@test filter should invoke on each item'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let cnt = ary.length - 2;
      let found = [];
      let result;

      // return true on all but the last two
      result = obj.filter(function (i) {
        found.push(i);
        return --cnt >= 0;
      });
      this.assert.deepEqual(found, ary, 'should have invoked on each item');
      this.assert.deepEqual(result, ary.slice(0, -2), 'filtered array should exclude items');
    }
  }

  class FilterByTest extends _internalTestHelpers.AbstractTestCase {
    '@test should filter based on object'() {
      let obj, ary;

      ary = [{ foo: 'foo', bar: 'BAZ' }, _object.default.create({ foo: 'foo', bar: 'bar' })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo', 'foo'), ary, 'filterBy(foo)');
      this.assert.deepEqual(obj.filterBy('bar', 'bar'), [ary[1]], 'filterBy(bar)');
    }

    '@test should include in result if property is true'() {
      let obj, ary;

      ary = [{ foo: 'foo', bar: true }, _object.default.create({ foo: 'bar', bar: false })];

      obj = this.newObject(ary);

      // different values - all eval to true
      this.assert.deepEqual(obj.filterBy('foo'), ary, 'filterBy(foo)');
      this.assert.deepEqual(obj.filterBy('bar'), [ary[0]], 'filterBy(bar)');
    }

    '@test should filter on second argument if provided'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 2 }), { name: 'obj3', foo: 2 }, _object.default.create({ name: 'obj4', foo: 3 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo', 3), [ary[0], ary[3]], "filterBy('foo', 3)')");
    }

    '@test should correctly filter null second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: null }), { name: 'obj3', foo: null }, _object.default.create({ name: 'obj4', foo: 3 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo', null), [ary[1], ary[2]], "filterBy('foo', 3)')");
    }

    '@test should not return all objects on undefined second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 2 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo', undefined), [], "filterBy('foo', 3)')");
    }

    '@test should correctly filter explicit undefined second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 3 }), { name: 'obj3', foo: undefined }, _object.default.create({ name: 'obj4', foo: undefined }), { name: 'obj5' }, _object.default.create({ name: 'obj6' })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo', undefined), ary.slice(2), "filterBy('foo', 3)')");
    }

    '@test should not match undefined properties without second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 3 }), { name: 'obj3', foo: undefined }, _object.default.create({ name: 'obj4', foo: undefined }), { name: 'obj5' }, _object.default.create({ name: 'obj6' })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.filterBy('foo'), ary.slice(0, 2), "filterBy('foo', 3)')");
    }
  }

  (0, _array.runArrayTests)('filter', FilterTest);
  (0, _array.runArrayTests)('filter', FilterByTest);
});
enifed('@ember/-internals/runtime/tests/array/find-test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_object, _internalTestHelpers, _array) {
  'use strict';

  class FindTests extends _internalTestHelpers.AbstractTestCase {
    '@test find should invoke callback on each item as long as you return false'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let found = [];
      let result;

      result = obj.find(function (i) {
        found.push(i);
        return false;
      });
      this.assert.equal(result, undefined, 'return value of obj.find');
      this.assert.deepEqual(found, ary, 'items passed during find() should match');
    }

    '@test every should stop invoking when you return true'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let cnt = ary.length - 2;
      let exp = cnt;
      let found = [];
      let result;

      result = obj.find(function (i) {
        found.push(i);
        return --cnt >= 0;
      });
      this.assert.equal(result, ary[exp - 1], 'return value of obj.find');
      this.assert.equal(found.length, exp, 'should invoke proper number of times');
      this.assert.deepEqual(found, ary.slice(0, -2), 'items passed during find() should match');
    }
  }

  class FindByTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return first object of property matches'() {
      let ary, obj;

      ary = [{ foo: 'foo', bar: 'BAZ' }, _object.default.create({ foo: 'foo', bar: 'bar' })];

      obj = this.newObject(ary);

      this.assert.equal(obj.findBy('foo', 'foo'), ary[0], 'findBy(foo)');
      this.assert.equal(obj.findBy('bar', 'bar'), ary[1], 'findBy(bar)');
    }

    '@test should return first object with truthy prop'() {
      let ary, obj;

      ary = [{ foo: 'foo', bar: false }, _object.default.create({ foo: 'bar', bar: true })];

      obj = this.newObject(ary);

      // different values - all eval to true
      this.assert.equal(obj.findBy('foo'), ary[0], 'findBy(foo)');
      this.assert.equal(obj.findBy('bar'), ary[1], 'findBy(bar)');
    }

    '@test should return first null property match'() {
      let ary, obj;

      ary = [{ foo: null, bar: 'BAZ' }, _object.default.create({ foo: null, bar: null })];

      obj = this.newObject(ary);

      this.assert.equal(obj.findBy('foo', null), ary[0], "findBy('foo', null)");
      this.assert.equal(obj.findBy('bar', null), ary[1], "findBy('bar', null)");
    }

    '@test should return first undefined property match'() {
      let ary, obj;

      ary = [{ foo: undefined, bar: 'BAZ' }, _object.default.create({})];

      obj = this.newObject(ary);

      this.assert.equal(obj.findBy('foo', undefined), ary[0], "findBy('foo', undefined)");
      this.assert.equal(obj.findBy('bar', undefined), ary[1], "findBy('bar', undefined)");
    }
  }

  (0, _array.runArrayTests)('find', FindTests);
  (0, _array.runArrayTests)('findBy', FindByTests);
});
enifed('@ember/-internals/runtime/tests/array/firstObject-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class FirstObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test returns first item in enumerable'() {
      let obj = this.newObject();
      this.assert.equal((0, _metal.get)(obj, 'firstObject'), this.toArray(obj)[0]);
    }

    '@test returns undefined if enumerable is empty'() {
      let obj = this.newObject([]);
      this.assert.equal((0, _metal.get)(obj, 'firstObject'), undefined);
    }

    '@test can not be set'() {
      let obj = this.newObject([]);

      this.assert.equal((0, _metal.get)(obj, 'firstObject'), this.toArray(obj)[0]);

      this.assert.throws(() => {
        (0, _metal.set)(obj, 'firstObject', 'foo!');
      }, /Cannot set read-only property "firstObject" on object/);
    }
  }

  (0, _array.runArrayTests)('firstObject', FirstObjectTests);
});
enifed('@ember/-internals/runtime/tests/array/forEach-test', ['@ember/-internals/utils', '@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_utils, _metal, _internalTestHelpers, _array) {
  'use strict';

  class ForEachTests extends _internalTestHelpers.AbstractTestCase {
    '@test forEach should iterate over list'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let found = [];

      obj.forEach(i => found.push(i));
      this.assert.deepEqual(found, ary, 'items passed during forEach should match');
    }

    '@test forEach should iterate over list after mutation'() {
      if ((0, _metal.get)(this, 'canTestMutation')) {
        this.assert.expect(0);
        return;
      }

      let obj = this.newObject();
      let ary = this.toArray(obj);
      let found = [];

      obj.forEach(i => found.push(i));
      this.assert.deepEqual(found, ary, 'items passed during forEach should match');

      this.mutate(obj);
      ary = this.toArray(obj);
      found = [];

      obj.forEach(i => found.push(i));
      this.assert.deepEqual(found, ary, 'items passed during forEach should match');
    }

    '@test 2nd target parameter'() {
      let obj = this.newObject();
      let target = this;

      obj.forEach(() => {
        // ES6TODO: When transpiled we will end up with "use strict" which disables automatically binding to the global context.
        // Therefore, the following test can never pass in strict mode unless we modify the `map` function implementation to
        // use `Ember.lookup` if target is not specified.
        //
        // equal(guidFor(this), guidFor(global), 'should pass the global object as this if no context');
      });

      obj.forEach(() => {
        this.assert.equal((0, _utils.guidFor)(this), (0, _utils.guidFor)(target), 'should pass target as this if context');
      }, target);
    }

    '@test callback params'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let loc = 0;

      obj.forEach((item, idx, enumerable) => {
        this.assert.equal(item, ary[loc], 'item param');
        this.assert.equal(idx, loc, 'idx param');
        this.assert.equal((0, _utils.guidFor)(enumerable), (0, _utils.guidFor)(obj), 'enumerable param');
        loc++;
      });
    }
  }

  (0, _array.runArrayTests)('forEach', ForEachTests);
});
enifed('@ember/-internals/runtime/tests/array/includes-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class IncludesTests extends _internalTestHelpers.AbstractTestCase {
    '@test includes returns correct value if startAt is positive'() {
      let data = (0, _array.newFixture)(3);
      let obj = this.newObject(data);

      this.assert.equal(obj.includes(data[1], 1), true, 'should return true if included');
      this.assert.equal(obj.includes(data[0], 1), false, 'should return false if not included');
    }

    '@test includes returns correct value if startAt is negative'() {
      let data = (0, _array.newFixture)(3);
      let obj = this.newObject(data);

      this.assert.equal(obj.includes(data[1], -2), true, 'should return true if included');
      this.assert.equal(obj.includes(data[0], -2), false, 'should return false if not included');
    }

    '@test includes returns true if startAt + length is still negative'() {
      let data = (0, _array.newFixture)(1);
      let obj = this.newObject(data);

      this.assert.equal(obj.includes(data[0], -2), true, 'should return true if included');
      this.assert.equal(obj.includes((0, _array.newFixture)(1), -2), false, 'should return false if not included');
    }

    '@test includes returns false if startAt out of bounds'() {
      let data = (0, _array.newFixture)(1);
      let obj = this.newObject(data);

      this.assert.equal(obj.includes(data[0], 2), false, 'should return false if startAt >= length');
      this.assert.equal(obj.includes((0, _array.newFixture)(1), 2), false, 'should return false if startAt >= length');
    }
  }

  (0, _array.runArrayTests)('includes', IncludesTests);
});
enifed('@ember/-internals/runtime/tests/array/indexOf-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class IndexOfTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return index of object'() {
      let expected = (0, _array.newFixture)(3);
      let obj = this.newObject(expected);
      let len = 3;

      for (let idx = 0; idx < len; idx++) {
        this.assert.equal(obj.indexOf(expected[idx]), idx, `obj.indexOf(${expected[idx]}) should match idx`);
      }
    }

    '@test should return -1 when requesting object not in index'() {
      let obj = this.newObject((0, _array.newFixture)(3));
      let foo = {};

      this.assert.equal(obj.indexOf(foo), -1, 'obj.indexOf(foo) should be < 0');
    }
  }

  (0, _array.runArrayTests)('indexOf', IndexOfTests);
});
enifed('@ember/-internals/runtime/tests/array/invoke-test', ['@ember/-internals/runtime/index', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_index, _internalTestHelpers, _array) {
  'use strict';

  class InvokeTests extends _internalTestHelpers.AbstractTestCase {
    '@test invoke should call on each object that implements'() {
      let cnt, ary, obj;

      function F(amt) {
        cnt += amt === undefined ? 1 : amt;
      }
      cnt = 0;
      ary = [{ foo: F }, _index.Object.create({ foo: F }),

      // NOTE: does not impl foo - invoke should just skip
      _index.Object.create({ bar: F }), { foo: F }];

      obj = this.newObject(ary);
      obj.invoke('foo');
      this.assert.equal(cnt, 3, 'should have invoked 3 times');

      cnt = 0;
      obj.invoke('foo', 2);
      this.assert.equal(cnt, 6, 'should have invoked 3 times, passing param');
    }

    '@test invoke should return an array containing the results of each invoked method'(assert) {
      let obj = this.newObject([{
        foo() {
          return 'one';
        }
      }, {}, // intentionally not including `foo` method
      {
        foo() {
          return 'two';
        }
      }]);

      let result = obj.invoke('foo');
      assert.deepEqual(result, ['one', undefined, 'two']);
    }

    '@test invoke should return an extended array (aka Ember.A)'(assert) {
      let obj = this.newObject([{ foo() {} }, { foo() {} }]);

      let result = obj.invoke('foo');

      assert.ok(_index.NativeArray.detect(result), 'NativeArray has been applied');
    }
  }

  (0, _array.runArrayTests)('invoke', InvokeTests);
});
enifed('@ember/-internals/runtime/tests/array/isAny-test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_object, _internalTestHelpers, _array) {
  'use strict';

  class IsAnyTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return true of any property matches'() {
      let obj = this.newObject([{ foo: 'foo', bar: 'BAZ' }, _object.default.create({ foo: 'foo', bar: 'bar' })]);

      this.assert.equal(obj.isAny('foo', 'foo'), true, 'isAny(foo)');
      this.assert.equal(obj.isAny('bar', 'bar'), true, 'isAny(bar)');
      this.assert.equal(obj.isAny('bar', 'BIFF'), false, 'isAny(BIFF)');
    }

    '@test should return true of any property is true'() {
      let obj = this.newObject([{ foo: 'foo', bar: true }, _object.default.create({ foo: 'bar', bar: false })]);

      // different values - all eval to true
      this.assert.equal(obj.isAny('foo'), true, 'isAny(foo)');
      this.assert.equal(obj.isAny('bar'), true, 'isAny(bar)');
      this.assert.equal(obj.isAny('BIFF'), false, 'isAny(biff)');
    }

    '@test should return true if any property matches null'() {
      let obj = this.newObject([{ foo: null, bar: 'bar' }, _object.default.create({ foo: 'foo', bar: null })]);

      this.assert.equal(obj.isAny('foo', null), true, "isAny('foo', null)");
      this.assert.equal(obj.isAny('bar', null), true, "isAny('bar', null)");
    }

    '@test should return true if any property is undefined'() {
      let obj = this.newObject([{ foo: undefined, bar: 'bar' }, _object.default.create({ foo: 'foo' })]);

      this.assert.equal(obj.isAny('foo', undefined), true, "isAny('foo', undefined)");
      this.assert.equal(obj.isAny('bar', undefined), true, "isAny('bar', undefined)");
    }

    '@test should not match undefined properties without second argument'() {
      let obj = this.newObject([{ foo: undefined }, _object.default.create({})]);

      this.assert.equal(obj.isAny('foo'), false, "isAny('foo', undefined)");
    }
  }

  (0, _array.runArrayTests)('isAny', IsAnyTests);
});
enifed('@ember/-internals/runtime/tests/array/lastIndexOf-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class LastIndexOfTests extends _internalTestHelpers.AbstractTestCase {
    "@test should return index of object's last occurrence"() {
      let expected = (0, _array.newFixture)(3);
      let obj = this.newObject(expected);
      let len = 3;

      for (let idx = 0; idx < len; idx++) {
        this.assert.equal(obj.lastIndexOf(expected[idx]), idx, `obj.lastIndexOf(${expected[idx]}) should match idx`);
      }
    }

    "@test should return index of object's last occurrence even startAt search location is equal to length"() {
      let expected = (0, _array.newFixture)(3);
      let obj = this.newObject(expected);
      let len = 3;

      for (let idx = 0; idx < len; idx++) {
        this.assert.equal(obj.lastIndexOf(expected[idx], len), idx, `obj.lastIndexOfs(${expected[idx]}) should match idx`);
      }
    }

    "@test should return index of object's last occurrence even startAt search location is greater than length"() {
      let expected = (0, _array.newFixture)(3);
      let obj = this.newObject(expected);
      let len = 3;

      for (let idx = 0; idx < len; idx++) {
        this.assert.equal(obj.lastIndexOf(expected[idx], len + 1), idx, `obj.lastIndexOf(${expected[idx]}) should match idx`);
      }
    }

    '@test should return -1 when no match is found'() {
      let obj = this.newObject((0, _array.newFixture)(3));
      let foo = {};

      this.assert.equal(obj.lastIndexOf(foo), -1, 'obj.lastIndexOf(foo) should be -1');
    }

    '@test should return -1 when no match is found even startAt search location is equal to length'() {
      let obj = this.newObject((0, _array.newFixture)(3));
      let foo = {};

      this.assert.equal(obj.lastIndexOf(foo, (0, _metal.get)(obj, 'length')), -1, 'obj.lastIndexOf(foo) should be -1');
    }

    '@test should return -1 when no match is found even startAt search location is greater than length'() {
      let obj = this.newObject((0, _array.newFixture)(3));
      let foo = {};

      this.assert.equal(obj.lastIndexOf(foo, (0, _metal.get)(obj, 'length') + 1), -1, 'obj.lastIndexOf(foo) should be -1');
    }
  }

  (0, _array.runArrayTests)('lastIndexOf', LastIndexOfTests);
});
enifed('@ember/-internals/runtime/tests/array/lastObject-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/metal'], function (_internalTestHelpers, _array, _metal) {
  'use strict';

  class LastObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test returns last item in enumerable'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);

      this.assert.equal((0, _metal.get)(obj, 'lastObject'), ary[ary.length - 1]);
    }

    '@test returns undefined if enumerable is empty'() {
      let obj = this.newObject([]);

      this.assert.equal((0, _metal.get)(obj, 'lastObject'), undefined);
    }

    '@test can not be set'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);

      this.assert.equal((0, _metal.get)(obj, 'lastObject'), ary[ary.length - 1]);

      this.assert.throws(function () {
        (0, _metal.set)(obj, 'lastObject', 'foo!');
      }, /Cannot set read-only property "lastObject" on object/);
    }
  }

  (0, _array.runArrayTests)('lastObject', LastObjectTests);
});
enifed('@ember/-internals/runtime/tests/array/map-test', ['@ember/-internals/utils', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/metal'], function (_utils, _internalTestHelpers, _array, _metal) {
  'use strict';

  const mapFunc = item => item ? item.toString() : null;

  class MapTests extends _internalTestHelpers.AbstractTestCase {
    '@test map should iterate over list'() {
      let obj = this.newObject();
      let ary = this.toArray(obj).map(mapFunc);
      let found = [];

      found = obj.map(mapFunc);
      this.assert.deepEqual(found, ary, 'mapped arrays should match');
    }

    '@test map should iterate over list after mutation'() {
      if ((0, _metal.get)(this, 'canTestMutation')) {
        this.assert.expect(0);
        return;
      }

      let obj = this.newObject();
      let ary = this.toArray(obj).map(mapFunc);
      let found;

      found = obj.map(mapFunc);
      this.assert.deepEqual(found, ary, 'items passed during forEach should match');

      this.mutate(obj);
      ary = this.toArray(obj).map(mapFunc);
      found = obj.map(mapFunc);
      this.assert.deepEqual(found, ary, 'items passed during forEach should match');
    }

    '@test 2nd target parameter'() {
      let obj = this.newObject();
      let target = this;

      obj.map(() => {
        // ES6TODO: When transpiled we will end up with "use strict" which disables automatically binding to the global context.
        // Therefore, the following test can never pass in strict mode unless we modify the `map` function implementation to
        // use `Ember.lookup` if target is not specified.
        //
        // equal(guidFor(this), guidFor(global), 'should pass the global object as this if no context');
      });

      obj.map(() => {
        this.assert.equal((0, _utils.guidFor)(this), (0, _utils.guidFor)(target), 'should pass target as this if context');
      }, target);
    }

    '@test callback params'() {
      let obj = this.newObject();
      let ary = this.toArray(obj);
      let loc = 0;

      obj.map((item, idx, enumerable) => {
        this.assert.equal(item, ary[loc], 'item param');
        this.assert.equal(idx, loc, 'idx param');
        this.assert.equal((0, _utils.guidFor)(enumerable), (0, _utils.guidFor)(obj), 'enumerable param');
        loc++;
      });
    }
  }

  (0, _array.runArrayTests)('map', MapTests);
});
enifed('@ember/-internals/runtime/tests/array/mapBy-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class MapByTests extends _internalTestHelpers.AbstractTestCase {
    '@test get value of each property'() {
      let obj = this.newObject([{ a: 1 }, { a: 2 }]);
      this.assert.equal(obj.mapBy('a').join(''), '12');
    }

    '@test should work also through getEach alias'() {
      let obj = this.newObject([{ a: 1 }, { a: 2 }]);
      this.assert.equal(obj.getEach('a').join(''), '12');
    }
  }

  (0, _array.runArrayTests)('mapBy', MapByTests);
});
enifed('@ember/-internals/runtime/tests/array/objectAt-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class ObjectAtTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return object at specified index'() {
      let expected = (0, _array.newFixture)(3);
      let obj = this.newObject(expected);
      let len = expected.length;

      for (let idx = 0; idx < len; idx++) {
        this.assert.equal(obj.objectAt(idx), expected[idx], `obj.objectAt(${idx}) should match`);
      }
    }

    '@test should return undefined when requesting objects beyond index'() {
      let obj;

      obj = this.newObject((0, _array.newFixture)(3));
      this.assert.equal(obj.objectAt(obj, 5), undefined, 'should return undefined for obj.objectAt(5) when len = 3');

      obj = this.newObject([]);
      this.assert.equal(obj.objectAt(obj, 0), undefined, 'should return undefined for obj.objectAt(0) when len = 0');
    }
  }

  (0, _array.runArrayTests)('objectAt', ObjectAtTests);
});
enifed('@ember/-internals/runtime/tests/array/reduce-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class ReduceTests extends _internalTestHelpers.AbstractTestCase {
    '@test collects a summary value from an enumeration'() {
      let obj = this.newObject([1, 2, 3]);
      let res = obj.reduce((previousValue, item) => previousValue + item, 0);
      this.assert.equal(res, 6);
    }

    '@test passes index of item to callback'() {
      let obj = this.newObject([1, 2, 3]);
      let res = obj.reduce((previousValue, item, index) => previousValue + index, 0);
      this.assert.equal(res, 3);
    }

    '@test passes enumerable object to callback'() {
      let obj = this.newObject([1, 2, 3]);
      let res = obj.reduce((previousValue, item, index, enumerable) => enumerable, 0);
      this.assert.equal(res, obj);
    }
  }

  (0, _array.runArrayTests)('reduce', ReduceTests);
});
enifed('@ember/-internals/runtime/tests/array/reject-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/runtime/lib/system/object'], function (_internalTestHelpers, _array, _object) {
  'use strict';

  class RejectTest extends _internalTestHelpers.AbstractTestCase {
    '@test should reject any item that does not meet the condition'() {
      let obj = this.newObject([1, 2, 3, 4]);
      let result;

      result = obj.reject(i => i < 3);
      this.assert.deepEqual(result, [3, 4], 'reject the correct items');
    }

    '@test should be the inverse of filter'() {
      let obj = this.newObject([1, 2, 3, 4]);
      let isEven = i => i % 2 === 0;
      let filtered, rejected;

      filtered = obj.filter(isEven);
      rejected = obj.reject(isEven);

      this.assert.deepEqual(filtered, [2, 4], 'filtered evens');
      this.assert.deepEqual(rejected, [1, 3], 'rejected evens');
    }
  }

  class RejectByTest extends _internalTestHelpers.AbstractTestCase {
    '@test should reject based on object'() {
      let obj, ary;

      ary = [{ foo: 'foo', bar: 'BAZ' }, _object.default.create({ foo: 'foo', bar: 'bar' })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo', 'foo'), [], 'rejectBy(foo)');
      this.assert.deepEqual(obj.rejectBy('bar', 'bar'), [ary[0]], 'rejectBy(bar)');
    }

    '@test should include in result if property is false'() {
      let obj, ary;

      ary = [{ foo: false, bar: true }, _object.default.create({ foo: false, bar: false })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo'), ary, 'rejectBy(foo)');
      this.assert.deepEqual(obj.rejectBy('bar'), [ary[1]], 'rejectBy(bar)');
    }

    '@test should reject on second argument if provided'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 2 }), { name: 'obj3', foo: 2 }, _object.default.create({ name: 'obj4', foo: 3 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo', 3), [ary[1], ary[2]], "rejectBy('foo', 3)')");
    }

    '@test should correctly reject null second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: null }), { name: 'obj3', foo: null }, _object.default.create({ name: 'obj4', foo: 3 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo', null), [ary[0], ary[3]], "rejectBy('foo', null)')");
    }

    '@test should correctly reject undefined second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 2 })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('bar', undefined), [], "rejectBy('bar', undefined)')");
    }

    '@test should correctly reject explicit undefined second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 3 }), { name: 'obj3', foo: undefined }, _object.default.create({ name: 'obj4', foo: undefined }), { name: 'obj5' }, _object.default.create({ name: 'obj6' })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo', undefined), ary.slice(0, 2), "rejectBy('foo', undefined)')");
    }

    '@test should match undefined, null, or false properties without second argument'() {
      let obj, ary;

      ary = [{ name: 'obj1', foo: 3 }, _object.default.create({ name: 'obj2', foo: 3 }), { name: 'obj3', foo: undefined }, _object.default.create({ name: 'obj4', foo: undefined }), { name: 'obj5' }, _object.default.create({ name: 'obj6' }), { name: 'obj7', foo: null }, _object.default.create({ name: 'obj8', foo: null }), { name: 'obj9', foo: false }, _object.default.create({ name: 'obj10', foo: false })];

      obj = this.newObject(ary);

      this.assert.deepEqual(obj.rejectBy('foo'), ary.slice(2), "rejectBy('foo')')");
    }
  }

  (0, _array.runArrayTests)('reject', RejectTest);
  (0, _array.runArrayTests)('rejectBy', RejectByTest);
});
enifed('@ember/-internals/runtime/tests/array/sortBy-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class SortByTests extends _internalTestHelpers.AbstractTestCase {
    '@test sort by value of property'() {
      let obj = this.newObject([{ a: 2 }, { a: 1 }]);
      let sorted = obj.sortBy('a');

      this.assert.equal((0, _metal.get)(sorted[0], 'a'), 1);
      this.assert.equal((0, _metal.get)(sorted[1], 'a'), 2);
    }

    '@test supports multiple propertyNames'() {
      let obj = this.newObject([{ a: 1, b: 2 }, { a: 1, b: 1 }]);
      let sorted = obj.sortBy('a', 'b');

      this.assert.equal((0, _metal.get)(sorted[0], 'b'), 1);
      this.assert.equal((0, _metal.get)(sorted[1], 'b'), 2);
    }
  }

  (0, _array.runArrayTests)('sortBy', SortByTests);
});
enifed('@ember/-internals/runtime/tests/array/toArray-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class ToArrayTests extends _internalTestHelpers.AbstractTestCase {
    '@test toArray should convert to an array'() {
      let obj = this.newObject();
      this.assert.deepEqual(obj.toArray(), this.toArray(obj));
    }
  }

  (0, _array.runArrayTests)('toArray', ToArrayTests);
});
enifed('@ember/-internals/runtime/tests/array/uniq-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class UniqTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return new instance with duplicates removed'() {
      let before, after, obj, ret;

      after = (0, _array.newFixture)(3);
      before = [after[0], after[1], after[2], after[1], after[0]];
      obj = this.newObject(before);
      before = obj.toArray(); // in case of set before will be different...

      ret = obj.uniq();
      this.assert.deepEqual(this.toArray(ret), after, 'should have removed item');
      this.assert.deepEqual(this.toArray(obj), before, 'should not have changed original');
    }

    '@test should return duplicate of same content if no duplicates found'() {
      let item, obj, ret;
      obj = this.newObject((0, _array.newFixture)(3));
      ret = obj.uniq(item);
      this.assert.ok(ret !== obj, 'should not be same object');
      this.assert.deepEqual(this.toArray(ret), this.toArray(obj), 'should be the same content');
    }
  }

  (0, _array.runArrayTests)('uniq', UniqTests);
});
enifed('@ember/-internals/runtime/tests/array/uniqBy-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class UniqByTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return new instance with duplicates removed'() {
      let numbers = this.newObject([{ id: 1, value: 'one' }, { id: 2, value: 'two' }, { id: 1, value: 'one' }]);
      this.assert.deepEqual(numbers.uniqBy('id'), [{ id: 1, value: 'one' }, { id: 2, value: 'two' }]);
    }

    '@test supports function as key'() {
      let numbers = this.newObject([{ id: 1, value: 'boom' }, { id: 2, value: 'boom' }, { id: 1, value: 'doom' }]);

      let keyFunction = val => {
        this.assert.equal(arguments.length, 1);
        return val.value;
      };

      this.assert.deepEqual(numbers.uniqBy(keyFunction), [{ id: 1, value: 'boom' }, { id: 1, value: 'doom' }]);
    }
  }

  (0, _array.runArrayTests)('uniqBy', UniqByTests);
});
enifed('@ember/-internals/runtime/tests/array/without-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class WithoutTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return new instance with item removed'() {
      let before, after, obj, ret;

      before = (0, _array.newFixture)(3);
      after = [before[0], before[2]];
      obj = this.newObject(before);

      ret = obj.without(before[1]);
      this.assert.deepEqual(this.toArray(ret), after, 'should have removed item');
      this.assert.deepEqual(this.toArray(obj), before, 'should not have changed original');
    }

    '@test should remove NaN value'() {
      let before, after, obj, ret;

      before = [...(0, _array.newFixture)(2), NaN];
      after = [before[0], before[1]];
      obj = this.newObject(before);

      ret = obj.without(NaN);
      this.assert.deepEqual(this.toArray(ret), after, 'should have removed item');
    }

    '@test should return same instance if object not found'() {
      let item, obj, ret;

      item = (0, _array.newFixture)(1)[0];
      obj = this.newObject((0, _array.newFixture)(3));

      ret = obj.without(item);
      this.assert.equal(ret, obj, 'should be same instance');
    }
  }

  (0, _array.runArrayTests)('without', WithoutTests);
});
enifed('@ember/-internals/runtime/tests/copyable-array/copy-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class CopyTest extends _internalTestHelpers.AbstractTestCase {
    '@test should return an equivalent copy'() {
      let obj = this.newObject();
      let copy = obj.copy();
      this.assert.ok(this.isEqual(obj, copy), 'old object and new object should be equivalent');
    }
  }

  (0, _array.runArrayTests)('copy', CopyTest, 'CopyableNativeArray', 'CopyableArray');
});
enifed('@ember/-internals/runtime/tests/core/compare_test', ['@ember/-internals/runtime/lib/type-of', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/compare', '@ember/-internals/runtime/lib/mixins/comparable', 'internal-test-helpers'], function (_typeOf, _object, _compare, _comparable, _internalTestHelpers) {
  'use strict';

  let data = [];
  let Comp = _object.default.extend(_comparable.default);

  Comp.reopenClass({
    compare(obj) {
      return obj.get('val');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember.compare()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      data[0] = null;
      data[1] = false;
      data[2] = true;
      data[3] = -12;
      data[4] = 3.5;
      data[5] = 'a string';
      data[6] = 'another string';
      data[7] = 'last string';
      data[8] = [1, 2];
      data[9] = [1, 2, 3];
      data[10] = [1, 3];
      data[11] = { a: 'hash' };
      data[12] = _object.default.create();
      data[13] = function (a) {
        return a;
      };
      data[14] = new Date('2012/01/01');
      data[15] = new Date('2012/06/06');
    }

    ['@test ordering should work'](assert) {
      let suspect, comparable, failureMessage, suspectIndex, comparableIndex;

      for (suspectIndex = 0; suspectIndex < data.length; suspectIndex++) {
        suspect = data[suspectIndex];

        assert.equal((0, _compare.default)(suspect, suspect), 0, suspectIndex + ' should equal itself');

        for (comparableIndex = suspectIndex + 1; comparableIndex < data.length; comparableIndex++) {
          comparable = data[comparableIndex];

          failureMessage = 'data[' + suspectIndex + '] (' + (0, _typeOf.typeOf)(suspect) + ') should be smaller than data[' + comparableIndex + '] (' + (0, _typeOf.typeOf)(comparable) + ')';

          assert.equal((0, _compare.default)(suspect, comparable), -1, failureMessage);
        }
      }
    }

    ['@test comparables should return values in the range of -1, 0, 1'](assert) {
      let negOne = Comp.create({
        val: -1
      });

      let zero = Comp.create({
        val: 0
      });

      let one = Comp.create({
        val: 1
      });

      assert.equal((0, _compare.default)(negOne, 'a'), -1, 'First item comparable - returns -1 (not negated)');
      assert.equal((0, _compare.default)(zero, 'b'), 0, 'First item comparable - returns  0 (not negated)');
      assert.equal((0, _compare.default)(one, 'c'), 1, 'First item comparable - returns  1 (not negated)');

      assert.equal((0, _compare.default)('a', negOne), 1, 'Second item comparable - returns -1 (negated)');
      assert.equal((0, _compare.default)('b', zero), 0, 'Second item comparable - returns  0 (negated)');
      assert.equal((0, _compare.default)('c', one), -1, 'Second item comparable - returns  1 (negated)');
    }
  });
});
enifed('@ember/-internals/runtime/tests/core/copy_test', ['@ember/-internals/runtime/lib/copy', 'internal-test-helpers'], function (_copy, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember Copy Method', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.copy null'](assert) {
      let obj = { field: null };
      let copied = null;
      expectDeprecation(() => {
        copied = (0, _copy.default)(obj, true);
      }, 'Use ember-copy addon instead of copy method and Copyable mixin.');
      assert.equal(copied.field, null, 'null should still be null');
    }

    ['@test Ember.copy date'](assert) {
      let date = new Date(2014, 7, 22);
      let dateCopy = null;
      expectDeprecation(() => {
        dateCopy = (0, _copy.default)(date);
      }, 'Use ember-copy addon instead of copy method and Copyable mixin.');
      assert.equal(date.getTime(), dateCopy.getTime(), 'dates should be equivalent');
    }

    ['@test Ember.copy null prototype object'](assert) {
      let obj = Object.create(null);

      obj.foo = 'bar';
      let copied = null;
      expectDeprecation(() => {
        copied = (0, _copy.default)(obj);
      }, 'Use ember-copy addon instead of copy method and Copyable mixin.');

      assert.equal(copied.foo, 'bar', 'bar should still be bar');
    }

    ['@test Ember.copy Array'](assert) {
      let array = [1, null, new Date(2015, 9, 9), 'four'];
      let arrayCopy = null;
      expectDeprecation(() => {
        arrayCopy = (0, _copy.default)(array);
      }, 'Use ember-copy addon instead of copy method and Copyable mixin.');

      assert.deepEqual(array, arrayCopy, 'array content cloned successfully in new array');
    }
  });
});
enifed('@ember/-internals/runtime/tests/core/isEqual_test', ['@ember/-internals/runtime/lib/is-equal', 'internal-test-helpers'], function (_isEqual, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('isEqual', class extends _internalTestHelpers.AbstractTestCase {
    ['@test undefined and null'](assert) {
      assert.ok((0, _isEqual.default)(undefined, undefined), 'undefined is equal to undefined');
      assert.ok(!(0, _isEqual.default)(undefined, null), 'undefined is not equal to null');
      assert.ok((0, _isEqual.default)(null, null), 'null is equal to null');
      assert.ok(!(0, _isEqual.default)(null, undefined), 'null is not equal to undefined');
    }

    ['@test strings should be equal'](assert) {
      assert.ok(!(0, _isEqual.default)('Hello', 'Hi'), 'different Strings are unequal');
      assert.ok((0, _isEqual.default)('Hello', 'Hello'), 'same Strings are equal');
    }

    ['@test numericals should be equal'](assert) {
      assert.ok((0, _isEqual.default)(24, 24), 'same numbers are equal');
      assert.ok(!(0, _isEqual.default)(24, 21), 'different numbers are inequal');
    }

    ['@test dates should be equal'](assert) {
      assert.ok((0, _isEqual.default)(new Date(1985, 7, 22), new Date(1985, 7, 22)), 'same dates are equal');
      assert.ok(!(0, _isEqual.default)(new Date(2014, 7, 22), new Date(1985, 7, 22)), 'different dates are not equal');
    }

    ['@test array should be equal'](assert) {
      // NOTE: We don't test for array contents -- that would be too expensive.
      assert.ok(!(0, _isEqual.default)([1, 2], [1, 2]), 'two array instances with the same values should not be equal');
      assert.ok(!(0, _isEqual.default)([1, 2], [1]), 'two array instances with different values should not be equal');
    }

    ['@test first object implements isEqual should use it'](assert) {
      assert.ok((0, _isEqual.default)({
        isEqual() {
          return true;
        }
      }, null), 'should return true always');

      let obj = {
        isEqual() {
          return false;
        }
      };
      assert.equal((0, _isEqual.default)(obj, obj), false, 'should return false because isEqual returns false');
    }
  });
});
enifed('@ember/-internals/runtime/tests/core/is_array_test', ['@ember/-internals/runtime/lib/mixins/array', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/browser-environment', 'internal-test-helpers'], function (_array, _array_proxy, _object, _browserEnvironment, _internalTestHelpers) {
  'use strict';

  const global = undefined;

  (0, _internalTestHelpers.moduleFor)('Ember Type Checking', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.isArray'](assert) {
      let numarray = [1, 2, 3];
      let number = 23;
      let strarray = ['Hello', 'Hi'];
      let string = 'Hello';
      let object = {};
      let length = { length: 12 };
      let strangeLength = { length: 'yes' };
      let fn = function () {};
      let arrayProxy = _array_proxy.default.create({ content: (0, _array.A)() });

      assert.equal((0, _array.isArray)(numarray), true, '[1,2,3]');
      assert.equal((0, _array.isArray)(number), false, '23');
      assert.equal((0, _array.isArray)(strarray), true, '["Hello", "Hi"]');
      assert.equal((0, _array.isArray)(string), false, '"Hello"');
      assert.equal((0, _array.isArray)(object), false, '{}');
      assert.equal((0, _array.isArray)(length), true, '{ length: 12 }');
      assert.equal((0, _array.isArray)(strangeLength), false, '{ length: "yes" }');
      assert.equal((0, _array.isArray)(global), false, 'global');
      assert.equal((0, _array.isArray)(fn), false, 'function() {}');
      assert.equal((0, _array.isArray)(arrayProxy), true, '[]');
    }

    '@test Ember.isArray does not trigger proxy assertion when probing for length GH#16495'(assert) {
      let instance = _object.default.extend({
        // intentionally returning non-null / non-undefined
        unknownProperty() {
          return false;
        }
      }).create();

      assert.equal((0, _array.isArray)(instance), false);
    }

    ['@test Ember.isArray(fileList)'](assert) {
      if (_browserEnvironment.window && typeof _browserEnvironment.window.FileList === 'function') {
        let fileListElement = document.createElement('input');
        fileListElement.type = 'file';
        let fileList = fileListElement.files;
        assert.equal((0, _array.isArray)(fileList), false, 'fileList');
      } else {
        assert.ok(true, 'FileList is not present on window');
      }
    }
  });
});
enifed('@ember/-internals/runtime/tests/core/is_empty_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_metal, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.isEmpty', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.isEmpty'](assert) {
      let arrayProxy = _array_proxy.default.create({ content: (0, _array.A)() });

      assert.equal(true, (0, _metal.isEmpty)(arrayProxy), 'for an ArrayProxy that has empty content');
    }
  });
});
enifed('@ember/-internals/runtime/tests/core/type_of_test', ['@ember/-internals/runtime/lib/type-of', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/browser-environment', 'internal-test-helpers'], function (_typeOf, _object, _browserEnvironment, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember Type Checking', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.typeOf'](assert) {
      let MockedDate = function () {};
      MockedDate.prototype = new Date();

      let mockedDate = new MockedDate();
      let date = new Date();
      let error = new Error('boum');
      let object = { a: 'b' };
      let a = null;
      let arr = [1, 2, 3];
      let obj = {};
      let instance = _object.default.create({ method() {} });

      assert.equal((0, _typeOf.typeOf)(), 'undefined', 'undefined');
      assert.equal((0, _typeOf.typeOf)(null), 'null', 'null');
      assert.equal((0, _typeOf.typeOf)('Cyril'), 'string', 'Cyril');
      assert.equal((0, _typeOf.typeOf)(101), 'number', '101');
      assert.equal((0, _typeOf.typeOf)(true), 'boolean', 'true');
      assert.equal((0, _typeOf.typeOf)([1, 2, 90]), 'array', '[1,2,90]');
      assert.equal((0, _typeOf.typeOf)(/abc/), 'regexp', '/abc/');
      assert.equal((0, _typeOf.typeOf)(date), 'date', 'new Date()');
      assert.equal((0, _typeOf.typeOf)(mockedDate), 'date', 'mocked date');
      assert.equal((0, _typeOf.typeOf)(error), 'error', 'error');
      assert.equal((0, _typeOf.typeOf)(object), 'object', 'object');
      assert.equal((0, _typeOf.typeOf)(undefined), 'undefined', 'item of type undefined');
      assert.equal((0, _typeOf.typeOf)(a), 'null', 'item of type null');
      assert.equal((0, _typeOf.typeOf)(arr), 'array', 'item of type array');
      assert.equal((0, _typeOf.typeOf)(obj), 'object', 'item of type object');
      assert.equal((0, _typeOf.typeOf)(instance), 'instance', 'item of type instance');
      assert.equal((0, _typeOf.typeOf)(instance.method), 'function', 'item of type function');
      assert.equal((0, _typeOf.typeOf)(_object.default.extend()), 'class', 'item of type class');
      assert.equal((0, _typeOf.typeOf)(new Error()), 'error', 'item of type error');
    }

    ['@test Ember.typeOf(fileList)'](assert) {
      if (_browserEnvironment.window && typeof _browserEnvironment.window.FileList === 'function') {
        let fileListElement = document.createElement('input');
        fileListElement.type = 'file';
        let fileList = fileListElement.files;
        assert.equal((0, _typeOf.typeOf)(fileList), 'filelist', 'item of type filelist');
      } else {
        assert.ok(true, 'FileList is not present on window');
      }
    }
  });
});
enifed('@ember/-internals/runtime/tests/ext/function_test', ['@ember/-internals/environment', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/evented', 'internal-test-helpers'], function (_environment, _metal, _object, _evented, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Function.prototype.observes() helper', class extends _internalTestHelpers.AbstractTestCase {
    ['@test global observer helper takes multiple params'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.Function) {
        assert.ok('undefined' === typeof Function.prototype.observes, 'Function.prototype helper disabled');
        return;
      }

      let MyMixin = _metal.Mixin.create({
        count: 0,

        foo: function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        }.observes('bar', 'baz')
      });

      let obj = (0, _metal.mixin)({}, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      (0, _metal.set)(obj, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 2, 'should invoke observer after change');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Function.prototype.on() helper', class extends _internalTestHelpers.AbstractTestCase {
    ['@test sets up an event listener, and can trigger the function on multiple events'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.Function) {
        assert.ok('undefined' === typeof Function.prototype.on, 'Function.prototype helper disabled');
        return;
      }

      let MyMixin = _metal.Mixin.create({
        count: 0,

        foo: function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        }.on('bar', 'baz')
      });

      let obj = (0, _metal.mixin)({}, _evented.default, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke listener immediately');

      obj.trigger('bar');
      obj.trigger('baz');
      assert.equal((0, _metal.get)(obj, 'count'), 2, 'should invoke listeners when events trigger');
    }

    ['@test can be chained with observes'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.Function) {
        assert.ok('Function.prototype helper disabled');
        return;
      }

      let MyMixin = _metal.Mixin.create({
        count: 0,
        bay: 'bay',
        foo: function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        }.observes('bay').on('bar')
      });

      let obj = (0, _metal.mixin)({}, _evented.default, MyMixin);
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke listener immediately');

      (0, _metal.set)(obj, 'bay', 'BAY');
      obj.trigger('bar');
      assert.equal((0, _metal.get)(obj, 'count'), 2, 'should invoke observer and listener');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Function.prototype.property() helper', class extends _internalTestHelpers.AbstractTestCase {
    ['@test sets up a ComputedProperty'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.Function) {
        assert.ok('undefined' === typeof Function.prototype.property, 'Function.prototype helper disabled');
        return;
      }

      let MyClass = _object.default.extend({
        firstName: null,
        lastName: null,
        fullName: function () {
          return (0, _metal.get)(this, 'firstName') + ' ' + (0, _metal.get)(this, 'lastName');
        }.property('firstName', 'lastName')
      });

      let obj = MyClass.create({ firstName: 'Fred', lastName: 'Flinstone' });
      assert.equal((0, _metal.get)(obj, 'fullName'), 'Fred Flinstone', 'should return the computed value');

      (0, _metal.set)(obj, 'firstName', 'Wilma');
      assert.equal((0, _metal.get)(obj, 'fullName'), 'Wilma Flinstone', 'should return the new computed value');

      (0, _metal.set)(obj, 'lastName', '');
      assert.equal((0, _metal.get)(obj, 'fullName'), 'Wilma ', 'should return the new computed value');
    }
  });
});
enifed('@ember/-internals/runtime/tests/ext/rsvp_test', ['@ember/runloop', '@ember/-internals/error-handling', '@ember/-internals/runtime/lib/ext/rsvp', '@ember/debug', 'internal-test-helpers'], function (_runloop, _errorHandling, _rsvp, _debug, _internalTestHelpers) {
  'use strict';

  const ORIGINAL_ONERROR = (0, _errorHandling.getOnerror)();

  (0, _internalTestHelpers.moduleFor)('Ember.RSVP', class extends _internalTestHelpers.AbstractTestCase {
    afterEach() {
      (0, _errorHandling.setOnerror)(ORIGINAL_ONERROR);
    }

    ['@test Ensure that errors thrown from within a promise are sent to the console'](assert) {
      let error = new Error('Error thrown in a promise for testing purposes.');

      try {
        (0, _runloop.run)(function () {
          new _rsvp.default.Promise(function () {
            throw error;
          });
        });
        assert.ok(false, 'expected assertion to be thrown');
      } catch (e) {
        assert.equal(e, error, 'error was re-thrown');
      }
    }

    ['@test TransitionAborted errors are not re-thrown'](assert) {
      assert.expect(1);
      let fakeTransitionAbort = { name: 'TransitionAborted' };

      (0, _runloop.run)(_rsvp.default, 'reject', fakeTransitionAbort);

      assert.ok(true, 'did not throw an error when dealing with TransitionAborted');
    }

    ['@test Can reject with non-Error object'](assert) {
      let wasEmberTesting = (0, _debug.isTesting)();
      (0, _debug.setTesting)(false);
      assert.expect(1);

      try {
        (0, _runloop.run)(_rsvp.default, 'reject', 'foo');
      } catch (e) {
        assert.equal(e, 'foo', 'should throw with rejection message');
      } finally {
        (0, _debug.setTesting)(wasEmberTesting);
      }
    }

    ['@test Can reject with no arguments'](assert) {
      let wasEmberTesting = (0, _debug.isTesting)();
      (0, _debug.setTesting)(false);
      assert.expect(1);

      try {
        (0, _runloop.run)(_rsvp.default, 'reject');
      } catch (e) {
        assert.ok(false, 'should not throw');
      } finally {
        (0, _debug.setTesting)(wasEmberTesting);
      }

      assert.ok(true);
    }

    ['@test rejections like jqXHR which have errorThrown property work'](assert) {
      assert.expect(2);

      let wasEmberTesting = (0, _debug.isTesting)();
      let wasOnError = (0, _errorHandling.getOnerror)();

      try {
        (0, _debug.setTesting)(false);
        (0, _errorHandling.setOnerror)(error => {
          assert.equal(error, actualError, 'expected the real error on the jqXHR');
          assert.equal(error.__reason_with_error_thrown__, jqXHR, 'also retains a helpful reference to the rejection reason');
        });

        let actualError = new Error('OMG what really happened');
        let jqXHR = {
          errorThrown: actualError
        };

        (0, _runloop.run)(_rsvp.default, 'reject', jqXHR);
      } finally {
        (0, _errorHandling.setOnerror)(wasOnError);
        (0, _debug.setTesting)(wasEmberTesting);
      }
    }

    ['@test rejections where the errorThrown is a string should wrap the sting in an error object'](assert) {
      assert.expect(2);

      let wasEmberTesting = (0, _debug.isTesting)();
      let wasOnError = (0, _errorHandling.getOnerror)();

      try {
        (0, _debug.setTesting)(false);
        (0, _errorHandling.setOnerror)(error => {
          assert.equal(error.message, actualError, 'expected the real error on the jqXHR');
          assert.equal(error.__reason_with_error_thrown__, jqXHR, 'also retains a helpful reference to the rejection reason');
        });

        let actualError = 'OMG what really happened';
        let jqXHR = {
          errorThrown: actualError
        };

        (0, _runloop.run)(_rsvp.default, 'reject', jqXHR);
      } finally {
        (0, _errorHandling.setOnerror)(wasOnError);
        (0, _debug.setTesting)(wasEmberTesting);
      }
    }

    ['@test rejections can be serialized to JSON'](assert) {
      assert.expect(2);

      let wasEmberTesting = (0, _debug.isTesting)();
      let wasOnError = (0, _errorHandling.getOnerror)();

      try {
        (0, _debug.setTesting)(false);
        (0, _errorHandling.setOnerror)(error => {
          assert.equal(error.message, 'a fail');
          assert.ok(JSON.stringify(error), 'Error can be serialized');
        });

        let jqXHR = {
          errorThrown: new Error('a fail')
        };

        (0, _runloop.run)(_rsvp.default, 'reject', jqXHR);
      } finally {
        (0, _errorHandling.setOnerror)(wasOnError);
        (0, _debug.setTesting)(wasEmberTesting);
      }
    }
  });

  const reason = 'i failed';
  function ajax() {
    return new _rsvp.default.Promise(function (resolve) {
      setTimeout(resolve, 0); // fake true / foreign async
    });
  }

  (0, _internalTestHelpers.moduleFor)('Ember.test: rejection assertions', class extends _internalTestHelpers.AbstractTestCase {
    ['@test unambigiously unhandled rejection'](assert) {
      assert.throws(function () {
        (0, _runloop.run)(function () {
          _rsvp.default.Promise.reject(reason);
        }); // something is funky, we should likely assert
      }, reason);
    }

    ['@test sync handled'](assert) {
      (0, _runloop.run)(function () {
        _rsvp.default.Promise.reject(reason).catch(function () {});
      }); // handled, we shouldn't need to assert.
      assert.ok(true, 'reached end of test');
    }

    ['@test handled within the same micro-task (via Ember.RVP.Promise)'](assert) {
      (0, _runloop.run)(function () {
        let rejection = _rsvp.default.Promise.reject(reason);
        _rsvp.default.Promise.resolve(1).then(() => rejection.catch(function () {}));
      }); // handled, we shouldn't need to assert.
      assert.ok(true, 'reached end of test');
    }

    ['@test handled within the same micro-task (via direct run-loop)'](assert) {
      (0, _runloop.run)(function () {
        let rejection = _rsvp.default.Promise.reject(reason);
        (0, _runloop.schedule)('afterRender', () => rejection.catch(function () {}));
      }); // handled, we shouldn't need to assert.
      assert.ok(true, 'reached end of test');
    }

    ['@test handled in the next microTask queue flush (next)'](assert) {
      assert.expect(2);
      let done = assert.async();

      assert.throws(function () {
        (0, _runloop.run)(function () {
          let rejection = _rsvp.default.Promise.reject(reason);

          (0, _runloop.next)(() => {
            rejection.catch(function () {});
            assert.ok(true, 'reached end of test');
            done();
          });
        });
      }, reason);

      // a promise rejection survived a full flush of the run-loop without being handled
      // this is very likely an issue.
    }

    ['@test handled in the same microTask Queue flush do to data locality'](assert) {
      // an ambiguous scenario, this may or may not assert
      // it depends on the locality of `user#1`
      let store = {
        find() {
          return _rsvp.default.Promise.resolve(1);
        }
      };
      (0, _runloop.run)(function () {
        let rejection = _rsvp.default.Promise.reject(reason);
        store.find('user', 1).then(() => rejection.catch(function () {}));
      });

      assert.ok(true, 'reached end of test');
    }

    ['@test handled in a different microTask Queue flush do to data locality'](assert) {
      let done = assert.async();
      // an ambiguous scenario, this may or may not assert
      // it depends on the locality of `user#1`
      let store = {
        find() {
          return ajax();
        }
      };
      assert.throws(function () {
        (0, _runloop.run)(function () {
          let rejection = _rsvp.default.Promise.reject(reason);
          store.find('user', 1).then(() => {
            rejection.catch(function () {});
            assert.ok(true, 'reached end of test');
            done();
          });
        });
      }, reason);
    }

    ['@test handled in the next microTask queue flush (ajax example)'](assert) {
      let done = assert.async();

      assert.throws(function () {
        (0, _runloop.run)(function () {
          let rejection = _rsvp.default.Promise.reject(reason);
          ajax().then(() => {
            rejection.catch(function () {});
            assert.ok(true, 'reached end of test');
            done();
          });
        });
      }, reason);
    }
  });
});
enifed('@ember/-internals/runtime/tests/helpers/array', ['exports', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', '@ember/-internals/utils', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/copyable', 'internal-test-helpers'], function (exports, _array_proxy, _array, _utils, _metal, _object, _copyable, _internalTestHelpers) {
  'use strict';

  exports.newFixture = newFixture;
  exports.newObjectsFixture = newObjectsFixture;
  exports.runArrayTests = runArrayTests;
  function newFixture(cnt) {
    let ret = [];
    while (--cnt >= 0) {
      ret.push((0, _utils.generateGuid)());
    }

    return ret;
  }

  function newObjectsFixture(cnt) {
    let ret = [];
    let item;
    while (--cnt >= 0) {
      item = {};
      (0, _utils.guidFor)(item);
      ret.push(item);
    }
    return ret;
  }

  const ArrayTestsObserverClass = _object.default.extend({
    init() {
      this._super(...arguments);
      this.isEnabled = true;
      this.reset();
    },

    reset() {
      this._keys = {};
      this._values = {};
      this._before = null;
      this._after = null;
      return this;
    },

    observe(obj, ...keys) {
      if (obj.addObserver) {
        let loc = keys.length;

        while (--loc >= 0) {
          obj.addObserver(keys[loc], this, 'propertyDidChange');
        }
      } else {
        this.isEnabled = false;
      }
      return this;
    },

    observeArray(obj) {
      (0, _metal.addArrayObserver)(obj, this);
      return this;
    },

    stopObserveArray(obj) {
      (0, _metal.removeArrayObserver)(obj, this);
      return this;
    },

    propertyDidChange(target, key, value) {
      if (this._keys[key] === undefined) {
        this._keys[key] = 0;
      }
      this._keys[key]++;
      this._values[key] = value;
    },

    arrayWillChange() {
      this.assert.equal(this._before, null, 'should only call once');
      this._before = Array.prototype.slice.call(arguments);
    },

    arrayDidChange() {
      this.assert.equal(this._after, null, 'should only call once');
      this._after = Array.prototype.slice.call(arguments);
    },

    validate(key, value) {
      if (!this.isEnabled) {
        return true;
      }

      if (!this._keys[key]) {
        return false;
      }

      if (arguments.length > 1) {
        return this._values[key] === value;
      } else {
        return true;
      }
    },

    timesCalled(key) {
      return this._keys[key] || 0;
    }
  });

  class AbstractArrayHelper {
    beforeEach(assert) {
      this.assert = assert;
    }

    newObject(ary) {
      return ary ? ary.slice() : newFixture(3);
    }

    toArray(obj) {
      return obj.slice();
    }

    newObserver() {
      let ret = ArrayTestsObserverClass.create({
        assert: this.assert
      });

      if (arguments.length > 0) {
        ret.observe.apply(ret, arguments);
      }

      return ret;
    }
  }

  class NativeArrayHelpers extends AbstractArrayHelper {
    newObject(ary) {
      return (0, _array.A)(super.newObject(ary));
    }

    mutate(obj) {
      obj.pushObject(obj.length + 1);
    }
  }

  class CopyableNativeArray extends AbstractArrayHelper {
    newObject() {
      return (0, _array.A)([(0, _utils.generateGuid)()]);
    }

    isEqual(a, b) {
      if (!(a instanceof Array)) {
        return false;
      }

      if (!(b instanceof Array)) {
        return false;
      }

      if (a.length !== b.length) {
        return false;
      }

      return a[0] === b[0];
    }
  }

  class CopyableArray extends AbstractArrayHelper {
    newObject() {
      return CopyableObject.create();
    }

    isEqual(a, b) {
      if (!(a instanceof CopyableObject) || !(b instanceof CopyableObject)) {
        return false;
      }

      return (0, _metal.get)(a, 'id') === (0, _metal.get)(b, 'id');
    }
  }

  class ArrayProxyHelpers extends AbstractArrayHelper {
    newObject(ary) {
      return _array_proxy.default.create({ content: (0, _array.A)(super.newObject(ary)) });
    }

    mutate(obj) {
      obj.pushObject((0, _metal.get)(obj, 'length') + 1);
    }

    toArray(obj) {
      return obj.toArray ? obj.toArray() : obj.slice();
    }
  }

  /*
    Implement a basic fake mutable array.  This validates that any non-native
    enumerable can impl this API.
  */
  const TestArray = _object.default.extend(_array.default, {
    _content: null,

    init() {
      this._content = this._content || [];
    },

    // some methods to modify the array so we can test changes.  Note that
    // arrays can be modified even if they don't implement MutableArray.  The
    // MutableArray is just a standard API for mutation but not required.
    addObject(obj) {
      let idx = this._content.length;
      (0, _metal.arrayContentWillChange)(this, idx, 0, 1);
      this._content.push(obj);
      (0, _metal.arrayContentDidChange)(this, idx, 0, 1);
    },

    removeFirst() {
      (0, _metal.arrayContentWillChange)(this, 0, 1, 0);
      this._content.shift();
      (0, _metal.arrayContentDidChange)(this, 0, 1, 0);
    },

    objectAt(idx) {
      return this._content[idx];
    },

    length: (0, _metal.computed)(function () {
      return this._content.length;
    })
  });

  /*
    Implement a basic fake mutable array.  This validates that any non-native
    enumerable can impl this API.
  */
  const TestMutableArray = _object.default.extend(_array.MutableArray, {
    _content: null,

    init(ary = []) {
      this._content = (0, _array.A)(ary);
    },

    replace(idx, amt, objects) {
      let args = objects ? objects.slice() : [];
      let removeAmt = amt;
      let addAmt = args.length;

      (0, _metal.arrayContentWillChange)(this, idx, removeAmt, addAmt);

      args.unshift(amt);
      args.unshift(idx);
      this._content.splice.apply(this._content, args);
      (0, _metal.arrayContentDidChange)(this, idx, removeAmt, addAmt);
      return this;
    },

    objectAt(idx) {
      return this._content[idx];
    },

    length: (0, _metal.computed)(function () {
      return this._content.length;
    }),

    slice() {
      return this._content.slice();
    }
  });

  const CopyableObject = _object.default.extend(_copyable.default, {
    id: null,

    init() {
      this._super(...arguments);
      (0, _metal.set)(this, 'id', (0, _utils.generateGuid)());
    },

    copy() {
      let ret = CopyableObject.create();
      (0, _metal.set)(ret, 'id', (0, _metal.get)(this, 'id'));
      return ret;
    }
  });

  class MutableArrayHelpers extends NativeArrayHelpers {
    newObject(ary) {
      return TestMutableArray.create(super.newObject(ary));
    }

    // allows for testing of the basic enumerable after an internal mutation
    mutate(obj) {
      obj.addObject(this.getFixture(1)[0]);
    }
  }

  class EmberArrayHelpers extends MutableArrayHelpers {
    newObject(ary) {
      return TestArray.create(super.newObject(ary));
    }
  }

  function runArrayTests(name, Tests, ...types) {
    if (types.length > 0) {
      types.forEach(type => {
        switch (type) {
          case 'ArrayProxy':
            (0, _internalTestHelpers.moduleFor)(`ArrayProxy: ${name}`, Tests, ArrayProxyHelpers);
            break;
          case 'EmberArray':
            (0, _internalTestHelpers.moduleFor)(`EmberArray: ${name}`, Tests, EmberArrayHelpers);
            break;
          case 'MutableArray':
            (0, _internalTestHelpers.moduleFor)(`MutableArray: ${name}`, Tests, EmberArrayHelpers);
            break;
          case 'CopyableArray':
            (0, _internalTestHelpers.moduleFor)(`CopyableArray: ${name}`, Tests, CopyableArray);
            break;
          case 'CopyableNativeArray':
            (0, _internalTestHelpers.moduleFor)(`CopyableNativeArray: ${name}`, Tests, CopyableNativeArray);
            break;
          case 'NativeArray':
            (0, _internalTestHelpers.moduleFor)(`NativeArray: ${name}`, Tests, EmberArrayHelpers);
            break;
        }
      });
    } else {
      (0, _internalTestHelpers.moduleFor)(`ArrayProxy: ${name}`, Tests, ArrayProxyHelpers);
      (0, _internalTestHelpers.moduleFor)(`EmberArray: ${name}`, Tests, EmberArrayHelpers);
      (0, _internalTestHelpers.moduleFor)(`MutableArray: ${name}`, Tests, MutableArrayHelpers);
      (0, _internalTestHelpers.moduleFor)(`NativeArray: ${name}`, Tests, NativeArrayHelpers);
    }
  }
});
enifed('@ember/-internals/runtime/tests/inject_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('inject', class extends _internalTestHelpers.AbstractTestCase {
    ['@test attempting to inject a nonexistent container key should error']() {
      let owner = (0, _internalTestHelpers.buildOwner)();
      let AnObject = _object.default.extend({
        foo: new _metal.InjectedProperty('bar', 'baz')
      });

      owner.register('foo:main', AnObject);

      expectAssertion(() => {
        owner.lookup('foo:main');
      }, /Attempting to inject an unknown injection: 'bar:baz'/);
    }

    ['@test factories should return a list of lazy injection full names'](assert) {
      if (true /* DEBUG */) {
          let AnObject = _object.default.extend({
            foo: new _metal.InjectedProperty('foo', 'bar'),
            bar: new _metal.InjectedProperty('quux')
          });

          assert.deepEqual(AnObject._lazyInjections(), {
            foo: {
              specifier: 'foo:bar',
              source: undefined,
              namespace: undefined
            },
            bar: {
              specifier: 'quux:bar',
              source: undefined,
              namespace: undefined
            }
          }, 'should return injected container keys');
        } else {
        assert.expect(0);
      }
    }
  });
});
enifed('@ember/-internals/runtime/tests/legacy_1x/mixins/observable/chained_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_runloop, _metal, _object, _array, _internalTestHelpers) {
  'use strict';

  /*
    NOTE: This test is adapted from the 1.x series of unit tests.  The tests
    are the same except for places where we intend to break the API we instead
    validate that we warn the developer appropriately.
  
    CHANGES FROM 1.6:
  
    * changed obj.set() and obj.get() to Ember.set() and Ember.get()
    * changed obj.addObserver() to addObserver()
  */

  (0, _internalTestHelpers.moduleFor)('Ember.Observable - Observing with @each', class extends _internalTestHelpers.AbstractTestCase {
    ['@test chained observers on enumerable properties are triggered when the observed property of any item changes'](assert) {
      let family = _object.default.create({ momma: null });
      let momma = _object.default.create({ children: [] });

      let child1 = _object.default.create({ name: 'Bartholomew' });
      let child2 = _object.default.create({ name: 'Agnes' });
      let child3 = _object.default.create({ name: 'Dan' });
      let child4 = _object.default.create({ name: 'Nancy' });

      (0, _metal.set)(family, 'momma', momma);
      (0, _metal.set)(momma, 'children', (0, _array.A)([child1, child2, child3]));

      let observerFiredCount = 0;
      (0, _metal.addObserver)(family, 'momma.children.@each.name', this, function () {
        observerFiredCount++;
      });

      observerFiredCount = 0;
      (0, _runloop.run)(() => (0, _metal.get)(momma, 'children').setEach('name', 'Juan'));
      assert.equal(observerFiredCount, 3, 'observer fired after changing child names');

      observerFiredCount = 0;
      (0, _runloop.run)(() => (0, _metal.get)(momma, 'children').pushObject(child4));
      assert.equal(observerFiredCount, 1, 'observer fired after adding a new item');

      observerFiredCount = 0;
      (0, _runloop.run)(() => (0, _metal.set)(child4, 'name', 'Herbert'));
      assert.equal(observerFiredCount, 1, 'observer fired after changing property on new object');

      (0, _metal.set)(momma, 'children', []);

      observerFiredCount = 0;
      (0, _runloop.run)(() => (0, _metal.set)(child1, 'name', 'Hanna'));
      assert.equal(observerFiredCount, 0, 'observer did not fire after removing changing property on a removed object');
    }
  });
});
enifed('@ember/-internals/runtime/tests/legacy_1x/mixins/observable/observable_test', ['@ember/-internals/environment', '@ember/runloop', '@ember/-internals/metal', '@ember/string', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/observable', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_environment, _runloop, _metal, _string, _object, _observable, _array, _internalTestHelpers) {
  'use strict';

  /*
    NOTE: This test is adapted from the 1.x series of unit tests.  The tests
    are the same except for places where we intend to break the API we instead
    validate that we warn the developer appropriately.
  
    CHANGES FROM 1.6:
  
    * Added ObservableObject which applies the Ember.Observable mixin.
    * Changed reference to Ember.T_FUNCTION to 'function'
    * Changed all references to sc_super to this._super(...arguments)
    * Changed Ember.objectForPropertyPath() to Ember.getPath()
    * Removed allPropertiesDidChange test - no longer supported
    * Changed test that uses 'ObjectE' as path to 'objectE' to reflect new
      rule on using capital letters for property paths.
    * Removed test passing context to addObserver.  context param is no longer
      supported.
    * removed test in observer around line 862 that expected key/value to be
      the last item in the chained path.  Should be root and chained path
  
  */

  // ========================================================================
  // Ember.Observable Tests
  // ========================================================================

  var object, ObjectC, ObjectD, objectA, objectB, lookup;

  var ObservableObject = _object.default.extend(_observable.default);
  var originalLookup = _environment.context.lookup;

  // ..........................................................
  // GET()
  //

  (0, _internalTestHelpers.moduleFor)('object.get()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      object = ObservableObject.extend(_observable.default, {
        computed: (0, _metal.computed)(function () {
          return 'value';
        }).volatile(),
        method() {
          return 'value';
        },
        unknownProperty(key) {
          this.lastUnknownProperty = key;
          return 'unknown';
        }
      }).create({
        normal: 'value',
        numberVal: 24,
        toggleVal: true,
        nullProperty: null
      });
    }

    ['@test should get normal properties'](assert) {
      assert.equal(object.get('normal'), 'value');
    }

    ['@test should call computed properties and return their result'](assert) {
      assert.equal(object.get('computed'), 'value');
    }

    ['@test should return the function for a non-computed property'](assert) {
      var value = object.get('method');
      assert.equal(typeof value, 'function');
    }

    ['@test should return null when property value is null'](assert) {
      assert.equal(object.get('nullProperty'), null);
    }

    ['@test should call unknownProperty when value is undefined'](assert) {
      assert.equal(object.get('unknown'), 'unknown');
      assert.equal(object.lastUnknownProperty, 'unknown');
    }
  });
  // ..........................................................
  // Ember.GET()
  //
  (0, _internalTestHelpers.moduleFor)('Ember.get()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      objectA = ObservableObject.extend({
        computed: (0, _metal.computed)(function () {
          return 'value';
        }).volatile(),
        method() {
          return 'value';
        },
        unknownProperty(key) {
          this.lastUnknownProperty = key;
          return 'unknown';
        }
      }).create({
        normal: 'value',
        numberVal: 24,
        toggleVal: true,
        nullProperty: null
      });

      objectB = {
        normal: 'value',
        nullProperty: null
      };
    }

    ['@test should get normal properties on Ember.Observable'](assert) {
      assert.equal((0, _metal.get)(objectA, 'normal'), 'value');
    }

    ['@test should call computed properties on Ember.Observable and return their result'](assert) {
      assert.equal((0, _metal.get)(objectA, 'computed'), 'value');
    }

    ['@test should return the function for a non-computed property on Ember.Observable'](assert) {
      var value = (0, _metal.get)(objectA, 'method');
      assert.equal(typeof value, 'function');
    }

    ['@test should return null when property value is null on Ember.Observable'](assert) {
      assert.equal((0, _metal.get)(objectA, 'nullProperty'), null);
    }

    ['@test should call unknownProperty when value is undefined on Ember.Observable'](assert) {
      assert.equal((0, _metal.get)(objectA, 'unknown'), 'unknown');
      assert.equal(objectA.lastUnknownProperty, 'unknown');
    }

    ['@test should get normal properties on standard objects'](assert) {
      assert.equal((0, _metal.get)(objectB, 'normal'), 'value');
    }

    ['@test should return null when property is null on standard objects'](assert) {
      assert.equal((0, _metal.get)(objectB, 'nullProperty'), null);
    }

    ['@test raise if the provided object is undefined']() {
      expectAssertion(function () {
        (0, _metal.get)(undefined, 'key');
      }, /Cannot call get with 'key' on an undefined object/i);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember.get() with paths', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should return a property at a given path relative to the passed object'](assert) {
      var foo = ObservableObject.create({
        bar: ObservableObject.extend({
          baz: (0, _metal.computed)(function () {
            return 'blargh';
          }).volatile()
        }).create()
      });

      assert.equal((0, _metal.get)(foo, 'bar.baz'), 'blargh');
    }

    ['@test should return a property at a given path relative to the passed object - JavaScript hash'](assert) {
      var foo = {
        bar: {
          baz: 'blargh'
        }
      };

      assert.equal((0, _metal.get)(foo, 'bar.baz'), 'blargh');
    }
  });

  // ..........................................................
  // SET()
  //

  (0, _internalTestHelpers.moduleFor)('object.set()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      object = ObservableObject.extend({
        computed: (0, _metal.computed)({
          get() {
            return this._computed;
          },
          set(key, value) {
            this._computed = value;
            return this._computed;
          }
        }).volatile(),

        method(key, value) {
          if (value !== undefined) {
            this._method = value;
          }
          return this._method;
        },

        unknownProperty() {
          return this._unknown;
        },

        setUnknownProperty(key, value) {
          this._unknown = value;
          return this._unknown;
        },

        // normal property
        normal: 'value',

        // computed property
        _computed: 'computed',
        // method, but not a property
        _method: 'method',
        // null property
        nullProperty: null,

        // unknown property
        _unknown: 'unknown'
      }).create();
    }

    ['@test should change normal properties and return the value'](assert) {
      var ret = object.set('normal', 'changed');
      assert.equal(object.get('normal'), 'changed');
      assert.equal(ret, 'changed');
    }

    ['@test should call computed properties passing value and return the value'](assert) {
      var ret = object.set('computed', 'changed');
      assert.equal(object.get('_computed'), 'changed');
      assert.equal(ret, 'changed');
    }

    ['@test should change normal properties when passing undefined'](assert) {
      var ret = object.set('normal', undefined);
      assert.equal(object.get('normal'), undefined);
      assert.equal(ret, undefined);
    }

    ['@test should replace the function for a non-computed property and return the value'](assert) {
      var ret = object.set('method', 'changed');
      assert.equal(object.get('_method'), 'method'); // make sure this was NOT run
      assert.ok(typeof object.get('method') !== 'function');
      assert.equal(ret, 'changed');
    }

    ['@test should replace prover when property value is null'](assert) {
      var ret = object.set('nullProperty', 'changed');
      assert.equal(object.get('nullProperty'), 'changed');
      assert.equal(ret, 'changed');
    }

    ['@test should call unknownProperty with value when property is undefined'](assert) {
      var ret = object.set('unknown', 'changed');
      assert.equal(object.get('_unknown'), 'changed');
      assert.equal(ret, 'changed');
    }
  });

  // ..........................................................
  // COMPUTED PROPERTIES
  //

  (0, _internalTestHelpers.moduleFor)('Computed properties', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      lookup = _environment.context.lookup = {};

      object = ObservableObject.extend({
        computed: (0, _metal.computed)({
          get() {
            this.computedCalls.push('getter-called');
            return 'computed';
          },
          set(key, value) {
            this.computedCalls.push(value);
          }
        }).volatile(),

        computedCached: (0, _metal.computed)({
          get() {
            this.computedCachedCalls.push('getter-called');
            return 'computedCached';
          },
          set: function (key, value) {
            this.computedCachedCalls.push(value);
          }
        }),

        dependent: (0, _metal.computed)({
          get() {
            this.dependentCalls.push('getter-called');
            return 'dependent';
          },
          set(key, value) {
            this.dependentCalls.push(value);
          }
        }).property('changer').volatile(),
        dependentFront: (0, _metal.computed)('changer', {
          get() {
            this.dependentFrontCalls.push('getter-called');
            return 'dependentFront';
          },
          set(key, value) {
            this.dependentFrontCalls.push(value);
          }
        }).volatile(),
        dependentCached: (0, _metal.computed)({
          get() {
            this.dependentCachedCalls.push('getter-called!');
            return 'dependentCached';
          },
          set(key, value) {
            this.dependentCachedCalls.push(value);
          }
        }).property('changer'),

        inc: (0, _metal.computed)('changer', function () {
          return this.incCallCount++;
        }),

        nestedInc: (0, _metal.computed)(function () {
          (0, _metal.get)(this, 'inc');
          return this.nestedIncCallCount++;
        }).property('inc'),

        isOn: (0, _metal.computed)({
          get() {
            return this.get('state') === 'on';
          },
          set() {
            this.set('state', 'on');
            return this.get('state') === 'on';
          }
        }).property('state').volatile(),

        isOff: (0, _metal.computed)({
          get() {
            return this.get('state') === 'off';
          },
          set() {
            this.set('state', 'off');
            return this.get('state') === 'off';
          }
        }).property('state').volatile()
      }).create({
        computedCalls: [],
        computedCachedCalls: [],
        changer: 'foo',
        dependentCalls: [],
        dependentFrontCalls: [],
        dependentCachedCalls: [],
        incCallCount: 0,
        nestedIncCallCount: 0,
        state: 'on'
      });
    }
    afterEach() {
      _environment.context.lookup = originalLookup;
    }

    ['@test getting values should call function return value'](assert) {
      // get each property twice. Verify return.
      var keys = (0, _string.w)('computed computedCached dependent dependentFront dependentCached');

      keys.forEach(function (key) {
        assert.equal(object.get(key), key, `Try #1: object.get(${key}) should run function`);
        assert.equal(object.get(key), key, `Try #2: object.get(${key}) should run function`);
      });

      // verify each call count.  cached should only be called once
      (0, _string.w)('computedCalls dependentFrontCalls dependentCalls').forEach(key => {
        assert.equal(object[key].length, 2, `non-cached property ${key} should be called 2x`);
      });

      (0, _string.w)('computedCachedCalls dependentCachedCalls').forEach(key => {
        assert.equal(object[key].length, 1, `non-cached property ${key} should be called 1x`);
      });
    }

    ['@test setting values should call function return value'](assert) {
      // get each property twice. Verify return.
      var keys = (0, _string.w)('computed dependent dependentFront computedCached dependentCached');
      var values = (0, _string.w)('value1 value2');

      keys.forEach(key => {
        assert.equal(object.set(key, values[0]), values[0], `Try #1: object.set(${key}, ${values[0]}) should run function`);

        assert.equal(object.set(key, values[1]), values[1], `Try #2: object.set(${key}, ${values[1]}) should run function`);

        assert.equal(object.set(key, values[1]), values[1], `Try #3: object.set(${key}, ${values[1]}) should not run function since it is setting same value as before`);
      });

      // verify each call count.  cached should only be called once
      keys.forEach(key => {
        var calls = object[key + 'Calls'];
        var idx, expectedLength;

        // Cached properties first check their cached value before setting the
        // property. Other properties blindly call set.
        expectedLength = 3;
        assert.equal(calls.length, expectedLength, `set(${key}) should be called the right amount of times`);
        for (idx = 0; idx < 2; idx++) {
          assert.equal(calls[idx], values[idx], `call #${idx + 1} to set(${key}) should have passed value ${values[idx]}`);
        }
      });
    }

    ['@test notify change should clear cache'](assert) {
      // call get several times to collect call count
      object.get('computedCached'); // should run func
      object.get('computedCached'); // should not run func

      object.notifyPropertyChange('computedCached');

      object.get('computedCached'); // should run again
      assert.equal(object.computedCachedCalls.length, 2, 'should have invoked method 2x');
    }

    ['@test change dependent should clear cache'](assert) {
      // call get several times to collect call count
      var ret1 = object.get('inc'); // should run func
      assert.equal(object.get('inc'), ret1, 'multiple calls should not run cached prop');

      object.set('changer', 'bar');

      assert.equal(object.get('inc'), ret1 + 1, 'should increment after dependent key changes'); // should run again
    }

    ['@test just notifying change of dependent should clear cache'](assert) {
      // call get several times to collect call count
      var ret1 = object.get('inc'); // should run func
      assert.equal(object.get('inc'), ret1, 'multiple calls should not run cached prop');

      object.notifyPropertyChange('changer');

      assert.equal(object.get('inc'), ret1 + 1, 'should increment after dependent key changes'); // should run again
    }

    ['@test changing dependent should clear nested cache'](assert) {
      // call get several times to collect call count
      var ret1 = object.get('nestedInc'); // should run func
      assert.equal(object.get('nestedInc'), ret1, 'multiple calls should not run cached prop');

      object.set('changer', 'bar');

      assert.equal(object.get('nestedInc'), ret1 + 1, 'should increment after dependent key changes'); // should run again
    }

    ['@test just notifying change of dependent should clear nested cache'](assert) {
      // call get several times to collect call count
      var ret1 = object.get('nestedInc'); // should run func
      assert.equal(object.get('nestedInc'), ret1, 'multiple calls should not run cached prop');

      object.notifyPropertyChange('changer');

      assert.equal(object.get('nestedInc'), ret1 + 1, 'should increment after dependent key changes'); // should run again
    }

    // This verifies a specific bug encountered where observers for computed
    // properties would fire before their prop caches were cleared.
    ['@test change dependent should clear cache when observers of dependent are called'](assert) {
      // call get several times to collect call count
      var ret1 = object.get('inc'); // should run func
      assert.equal(object.get('inc'), ret1, 'multiple calls should not run cached prop');

      // add observer to verify change...
      object.addObserver('inc', this, function () {
        assert.equal(object.get('inc'), ret1 + 1, 'should increment after dependent key changes'); // should run again
      });

      // now run
      object.set('changer', 'bar');
    }

    ['@test setting one of two computed properties that depend on a third property should clear the kvo cache'](assert) {
      // we have to call set twice to fill up the cache
      object.set('isOff', true);
      object.set('isOn', true);

      // setting isOff to true should clear the kvo cache
      object.set('isOff', true);
      assert.equal(object.get('isOff'), true, 'object.isOff should be true');
      assert.equal(object.get('isOn'), false, 'object.isOn should be false');
    }

    ['@test dependent keys should be able to be specified as property paths'](assert) {
      var depObj = ObservableObject.extend({
        menuPrice: (0, _metal.computed)(function () {
          return this.get('menu.price');
        }).property('menu.price')
      }).create({
        menu: ObservableObject.create({
          price: 5
        })
      });

      assert.equal(depObj.get('menuPrice'), 5, 'precond - initial value returns 5');

      depObj.set('menu.price', 6);

      assert.equal(depObj.get('menuPrice'), 6, 'cache is properly invalidated after nested property changes');
    }

    ['@test cacheable nested dependent keys should clear after their dependencies update'](assert) {
      assert.ok(true);

      var DepObj;

      (0, _runloop.run)(function () {
        lookup.DepObj = DepObj = ObservableObject.extend({
          price: (0, _metal.computed)('restaurant.menu.price', function () {
            return this.get('restaurant.menu.price');
          })
        }).create({
          restaurant: ObservableObject.create({
            menu: ObservableObject.create({
              price: 5
            })
          })
        });
      });

      assert.equal(DepObj.get('price'), 5, 'precond - computed property is correct');

      (0, _runloop.run)(function () {
        DepObj.set('restaurant.menu.price', 10);
      });
      assert.equal(DepObj.get('price'), 10, 'cacheable computed properties are invalidated even if no run loop occurred');

      (0, _runloop.run)(function () {
        DepObj.set('restaurant.menu.price', 20);
      });
      assert.equal(DepObj.get('price'), 20, 'cacheable computed properties are invalidated after a second get before a run loop');
      assert.equal(DepObj.get('price'), 20, 'precond - computed properties remain correct after a run loop');

      (0, _runloop.run)(function () {
        DepObj.set('restaurant.menu', ObservableObject.create({
          price: 15
        }));
      });

      assert.equal(DepObj.get('price'), 15, 'cacheable computed properties are invalidated after a middle property changes');

      (0, _runloop.run)(function () {
        DepObj.set('restaurant.menu', ObservableObject.create({
          price: 25
        }));
      });

      assert.equal(DepObj.get('price'), 25, 'cacheable computed properties are invalidated after a middle property changes again, before a run loop');
    }
  });

  // ..........................................................
  // OBSERVABLE OBJECTS
  //

  (0, _internalTestHelpers.moduleFor)('Observable objects & object properties ', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      object = ObservableObject.extend({
        getEach() {
          var keys = ['normal', 'abnormal'];
          var ret = [];
          for (var idx = 0; idx < keys.length; idx++) {
            ret[ret.length] = this.get(keys[idx]);
          }
          return ret;
        },

        newObserver() {
          this.abnormal = 'changedValueObserved';
        },

        testObserver: (0, _metal.observer)('normal', function () {
          this.abnormal = 'removedObserver';
        }),

        testArrayObserver: (0, _metal.observer)('normalArray.[]', function () {
          this.abnormal = 'notifiedObserver';
        })
      }).create({
        normal: 'value',
        abnormal: 'zeroValue',
        numberVal: 24,
        toggleVal: true,
        observedProperty: 'beingWatched',
        testRemove: 'observerToBeRemoved',
        normalArray: (0, _array.A)([1, 2, 3, 4, 5])
      });
    }

    ['@test incrementProperty and decrementProperty'](assert) {
      var newValue = object.incrementProperty('numberVal');

      assert.equal(25, newValue, 'numerical value incremented');
      object.numberVal = 24;
      newValue = object.decrementProperty('numberVal');
      assert.equal(23, newValue, 'numerical value decremented');
      object.numberVal = 25;
      newValue = object.incrementProperty('numberVal', 5);
      assert.equal(30, newValue, 'numerical value incremented by specified increment');
      object.numberVal = 25;
      newValue = object.incrementProperty('numberVal', -5);
      assert.equal(20, newValue, 'minus numerical value incremented by specified increment');
      object.numberVal = 25;
      newValue = object.incrementProperty('numberVal', 0);
      assert.equal(25, newValue, 'zero numerical value incremented by specified increment');

      expectAssertion(function () {
        newValue = object.incrementProperty('numberVal', 0 - void 0); // Increment by NaN
      }, /Must pass a numeric value to incrementProperty/i);

      expectAssertion(function () {
        newValue = object.incrementProperty('numberVal', 'Ember'); // Increment by non-numeric String
      }, /Must pass a numeric value to incrementProperty/i);

      expectAssertion(function () {
        newValue = object.incrementProperty('numberVal', 1 / 0); // Increment by Infinity
      }, /Must pass a numeric value to incrementProperty/i);

      assert.equal(25, newValue, 'Attempting to increment by non-numeric values should not increment value');

      object.numberVal = 25;
      newValue = object.decrementProperty('numberVal', 5);
      assert.equal(20, newValue, 'numerical value decremented by specified increment');
      object.numberVal = 25;
      newValue = object.decrementProperty('numberVal', -5);
      assert.equal(30, newValue, 'minus numerical value decremented by specified increment');
      object.numberVal = 25;
      newValue = object.decrementProperty('numberVal', 0);
      assert.equal(25, newValue, 'zero numerical value decremented by specified increment');

      expectAssertion(function () {
        newValue = object.decrementProperty('numberVal', 0 - void 0); // Decrement by NaN
      }, /Must pass a numeric value to decrementProperty/i);

      expectAssertion(function () {
        newValue = object.decrementProperty('numberVal', 'Ember'); // Decrement by non-numeric String
      }, /Must pass a numeric value to decrementProperty/i);

      expectAssertion(function () {
        newValue = object.decrementProperty('numberVal', 1 / 0); // Decrement by Infinity
      }, /Must pass a numeric value to decrementProperty/i);

      assert.equal(25, newValue, 'Attempting to decrement by non-numeric values should not decrement value');
    }

    ['@test toggle function, should be boolean'](assert) {
      assert.equal(object.toggleProperty('toggleVal', true, false), object.get('toggleVal'));
      assert.equal(object.toggleProperty('toggleVal', true, false), object.get('toggleVal'));
      assert.equal(object.toggleProperty('toggleVal', undefined, undefined), object.get('toggleVal'));
    }

    ['@test should notify array observer when array changes'](assert) {
      (0, _metal.get)(object, 'normalArray').replace(0, 0, [6]);
      assert.equal(object.abnormal, 'notifiedObserver', 'observer should be notified');
    }
  });

  (0, _internalTestHelpers.moduleFor)('object.addObserver()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      ObjectC = ObservableObject.create({
        objectE: ObservableObject.create({
          propertyVal: 'chainedProperty'
        }),

        normal: 'value',
        normal1: 'zeroValue',
        normal2: 'dependentValue',
        incrementor: 10,

        action() {
          this.normal1 = 'newZeroValue';
        },

        observeOnceAction() {
          this.incrementor = this.incrementor + 1;
        },

        chainedObserver() {
          this.normal2 = 'chainedPropertyObserved';
        }
      });
    }

    ['@test should register an observer for a property'](assert) {
      ObjectC.addObserver('normal', ObjectC, 'action');
      ObjectC.set('normal', 'newValue');
      assert.equal(ObjectC.normal1, 'newZeroValue');
    }

    ['@test should register an observer for a property - Special case of chained property'](assert) {
      ObjectC.addObserver('objectE.propertyVal', ObjectC, 'chainedObserver');
      ObjectC.objectE.set('propertyVal', 'chainedPropertyValue');
      assert.equal('chainedPropertyObserved', ObjectC.normal2);
      ObjectC.normal2 = 'dependentValue';
      ObjectC.set('objectE', '');
      assert.equal('chainedPropertyObserved', ObjectC.normal2);
    }
  });

  (0, _internalTestHelpers.moduleFor)('object.removeObserver()', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      ObjectD = ObservableObject.create({
        objectF: ObservableObject.create({
          propertyVal: 'chainedProperty'
        }),

        normal: 'value',
        normal1: 'zeroValue',
        normal2: 'dependentValue',
        ArrayKeys: ['normal', 'normal1'],

        addAction() {
          this.normal1 = 'newZeroValue';
        },
        removeAction() {
          this.normal2 = 'newDependentValue';
        },
        removeChainedObserver() {
          this.normal2 = 'chainedPropertyObserved';
        },

        observableValue: 'hello world',

        observer1() {
          // Just an observer
        },
        observer2() {
          this.removeObserver('observableValue', null, 'observer1');
          this.removeObserver('observableValue', null, 'observer2');
          this.hasObserverFor('observableValue'); // Tickle 'getMembers()'
          this.removeObserver('observableValue', null, 'observer3');
        },
        observer3() {
          // Just an observer
        }
      });
    }

    ['@test should unregister an observer for a property'](assert) {
      ObjectD.addObserver('normal', ObjectD, 'addAction');
      ObjectD.set('normal', 'newValue');
      assert.equal(ObjectD.normal1, 'newZeroValue');

      ObjectD.set('normal1', 'zeroValue');

      ObjectD.removeObserver('normal', ObjectD, 'addAction');
      ObjectD.set('normal', 'newValue');
      assert.equal(ObjectD.normal1, 'zeroValue');
    }

    ["@test should unregister an observer for a property - special case when key has a '.' in it."](assert) {
      ObjectD.addObserver('objectF.propertyVal', ObjectD, 'removeChainedObserver');
      ObjectD.objectF.set('propertyVal', 'chainedPropertyValue');
      ObjectD.removeObserver('objectF.propertyVal', ObjectD, 'removeChainedObserver');
      ObjectD.normal2 = 'dependentValue';
      ObjectD.objectF.set('propertyVal', 'removedPropertyValue');
      assert.equal('dependentValue', ObjectD.normal2);
      ObjectD.set('objectF', '');
      assert.equal('dependentValue', ObjectD.normal2);
    }

    ['@test removing an observer inside of an observer shouldn’t cause any problems'](assert) {
      // The observable system should be protected against clients removing
      // observers in the middle of observer notification.
      var encounteredError = false;
      try {
        ObjectD.addObserver('observableValue', null, 'observer1');
        ObjectD.addObserver('observableValue', null, 'observer2');
        ObjectD.addObserver('observableValue', null, 'observer3');
        (0, _runloop.run)(function () {
          ObjectD.set('observableValue', 'hi world');
        });
      } catch (e) {
        encounteredError = true;
      }
      assert.equal(encounteredError, false);
    }
  });
});
enifed('@ember/-internals/runtime/tests/legacy_1x/mixins/observable/propertyChanges_test', ['@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/observable', '@ember/-internals/metal', 'internal-test-helpers'], function (_object, _observable, _metal, _internalTestHelpers) {
  'use strict';

  /*
    NOTE: This test is adapted from the 1.x series of unit tests.  The tests
    are the same except for places where we intend to break the API we instead
    validate that we warn the developer appropriately.
  
    CHANGES FROM 1.6:
  
    * Create ObservableObject which includes Ember.Observable
    * Remove test that tests internal _kvo_changeLevel property.  This is an
      implementation detail.
    * Remove test for allPropertiesDidChange
    * Removed star observer test.  no longer supported
    * Removed property revision test.  no longer supported
  */

  // ========================================================================
  // Ember.Observable Tests
  // ========================================================================

  const ObservableObject = _object.default.extend(_observable.default);

  let ObjectA;

  (0, _internalTestHelpers.moduleFor)('object.propertyChanges', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      ObjectA = ObservableObject.extend({
        action: (0, _metal.observer)('foo', function () {
          this.set('prop', 'changedPropValue');
        }),
        notifyAction: (0, _metal.observer)('newFoo', function () {
          this.set('newProp', 'changedNewPropValue');
        }),

        notifyAllAction: (0, _metal.observer)('prop', function () {
          this.set('newFoo', 'changedNewFooValue');
        }),

        starObserver(target, key) {
          this.starProp = key;
        }
      }).create({
        starProp: null,

        foo: 'fooValue',
        prop: 'propValue',

        newFoo: 'newFooValue',
        newProp: 'newPropValue'
      });
    }

    ['@test should observe the changes within the nested begin / end property changes'](assert) {
      //start the outer nest
      ObjectA.beginPropertyChanges();

      // Inner nest
      ObjectA.beginPropertyChanges();
      ObjectA.set('foo', 'changeFooValue');

      assert.equal(ObjectA.prop, 'propValue');
      ObjectA.endPropertyChanges();

      //end inner nest
      ObjectA.set('prop', 'changePropValue');
      assert.equal(ObjectA.newFoo, 'newFooValue');

      //close the outer nest
      ObjectA.endPropertyChanges();

      assert.equal(ObjectA.prop, 'changedPropValue');
      assert.equal(ObjectA.newFoo, 'changedNewFooValue');
    }

    ['@test should observe the changes within the begin and end property changes'](assert) {
      ObjectA.beginPropertyChanges();
      ObjectA.set('foo', 'changeFooValue');

      assert.equal(ObjectA.prop, 'propValue');
      ObjectA.endPropertyChanges();

      assert.equal(ObjectA.prop, 'changedPropValue');
    }

    ['@test should indicate that the property of an object has just changed'](assert) {
      //change the value of foo.
      ObjectA.set('foo', 'changeFooValue');

      // Indicate the subscribers of foo that the value has just changed
      ObjectA.notifyPropertyChange('foo', null);

      // Values of prop has just changed
      assert.equal(ObjectA.prop, 'changedPropValue');
    }

    ['@test should notify that the property of an object has changed'](assert) {
      // Notify to its subscriber that the values of 'newFoo' will be changed. In this
      // case the observer is "newProp". Therefore this will call the notifyAction function
      // and value of "newProp" will be changed.
      ObjectA.notifyPropertyChange('newFoo', 'fooValue');

      //value of newProp changed.
      assert.equal(ObjectA.newProp, 'changedNewPropValue');
    }

    ['@test should invalidate function property cache when notifyPropertyChange is called'](assert) {
      let a = ObservableObject.extend({
        b: (0, _metal.computed)({
          get() {
            return this._b;
          },
          set(key, value) {
            this._b = value;
            return this;
          }
        }).volatile()
      }).create({
        _b: null
      });

      a.set('b', 'foo');
      assert.equal(a.get('b'), 'foo', 'should have set the correct value for property b');

      a._b = 'bar';
      a.notifyPropertyChange('b');
      a.set('b', 'foo');
      assert.equal(a.get('b'), 'foo', 'should have invalidated the cache so that the newly set value is actually set');
    }
  });
});
enifed('@ember/-internals/runtime/tests/legacy_1x/system/object/base_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  /*
    NOTE: This test is adapted from the 1.x series of unit tests.  The tests
    are the same except for places where we intend to break the API we instead
    validate that we warn the developer appropriately.
  
    CHANGES FROM 1.6:
  
    * Changed get(obj, ) and set(obj, ) to Ember.get() and Ember.set()
    * Removed obj.instanceOf() and obj.kindOf() tests.  use obj instanceof Foo
      instead
    * Removed respondsTo() and tryToPerform() tests.  Can be brought back in a
      utils package.
    * Removed destroy() test.  You can impl yourself but not built in
    * Changed Class.subclassOf() test to Class.detect()
    * Remove broken test for 'superclass' property.
    * Removed obj.didChangeFor()
  */

  // ========================================================================
  // EmberObject Base Tests
  // ========================================================================

  let obj, obj1; // global variables

  (0, _internalTestHelpers.moduleFor)('A new EmberObject instance', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _object.default.create({
        foo: 'bar',
        total: 12345,
        aMethodThatExists() {},
        aMethodThatReturnsTrue() {
          return true;
        },
        aMethodThatReturnsFoobar() {
          return 'Foobar';
        },
        aMethodThatReturnsFalse() {
          return false;
        }
      });
    }

    afterEach() {
      obj = undefined;
    }

    ['@test Should return its properties when requested using EmberObject#get'](assert) {
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar');
      assert.equal((0, _metal.get)(obj, 'total'), 12345);
    }

    ['@test Should allow changing of those properties by calling EmberObject#set'](assert) {
      assert.equal((0, _metal.get)(obj, 'foo'), 'bar');
      assert.equal((0, _metal.get)(obj, 'total'), 12345);

      (0, _metal.set)(obj, 'foo', 'Chunky Bacon');
      (0, _metal.set)(obj, 'total', 12);

      assert.equal((0, _metal.get)(obj, 'foo'), 'Chunky Bacon');
      assert.equal((0, _metal.get)(obj, 'total'), 12);
    }
  });

  (0, _internalTestHelpers.moduleFor)('EmberObject superclass and subclasses', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _object.default.extend({
        method1() {
          return 'hello';
        }
      });
      obj1 = obj.extend();
    }

    afterEach() {
      obj = undefined;
      obj1 = undefined;
    }

    ['@test Checking the detect() function on an object and its subclass'](assert) {
      assert.equal(obj.detect(obj1), true);
      assert.equal(obj1.detect(obj), false);
    }

    ['@test Checking the detectInstance() function on an object and its subclass'](assert) {
      assert.ok(_object.default.detectInstance(obj.create()));
      assert.ok(obj.detectInstance(obj.create()));
    }
  });
});
enifed('@ember/-internals/runtime/tests/legacy_1x/system/object/concatenated_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  /*
    NOTE: This test is adapted from the 1.x series of unit tests.  The tests
    are the same except for places where we intend to break the API we instead
    validate that we warn the developer appropriately.
  
    CHANGES FROM 1.6:
  
    * changed get(obj, ) and set(obj, ) to Ember.get() and Ember.set()
    * converted uses of obj.isEqual() to use deepEqual() test since isEqual is not
      always defined
  */

  function K() {
    return this;
  }

  let klass;

  (0, _internalTestHelpers.moduleFor)('EmberObject Concatenated Properties', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      klass = _object.default.extend({
        concatenatedProperties: ['values', 'functions'],
        values: ['a', 'b', 'c'],
        functions: [K]
      });
    }

    ['@test concatenates instances'](assert) {
      let obj = klass.create({
        values: ['d', 'e', 'f']
      });

      let values = (0, _metal.get)(obj, 'values');
      let expected = ['a', 'b', 'c', 'd', 'e', 'f'];

      assert.deepEqual(values, expected, `should concatenate values property (expected: ${expected}, got: ${values})`);
    }

    ['@test concatenates subclasses'](assert) {
      let subKlass = klass.extend({
        values: ['d', 'e', 'f']
      });
      let obj = subKlass.create();

      let values = (0, _metal.get)(obj, 'values');
      let expected = ['a', 'b', 'c', 'd', 'e', 'f'];

      assert.deepEqual(values, expected, `should concatenate values property (expected: ${expected}, got: ${values})`);
    }

    ['@test concatenates reopen'](assert) {
      klass.reopen({
        values: ['d', 'e', 'f']
      });
      let obj = klass.create();

      let values = (0, _metal.get)(obj, 'values');
      let expected = ['a', 'b', 'c', 'd', 'e', 'f'];

      assert.deepEqual(values, expected, `should concatenate values property (expected: ${expected}, got: ${values})`);
    }

    ['@test concatenates mixin'](assert) {
      let mixin = {
        values: ['d', 'e']
      };
      let subKlass = klass.extend(mixin, {
        values: ['f']
      });
      let obj = subKlass.create();

      let values = (0, _metal.get)(obj, 'values');
      let expected = ['a', 'b', 'c', 'd', 'e', 'f'];

      assert.deepEqual(values, expected, `should concatenate values property (expected: ${expected}, got: ${values})`);
    }

    ['@test concatenates reopen, subclass, and instance'](assert) {
      klass.reopen({ values: ['d'] });
      let subKlass = klass.extend({ values: ['e'] });
      let obj = subKlass.create({ values: ['f'] });

      let values = (0, _metal.get)(obj, 'values');
      let expected = ['a', 'b', 'c', 'd', 'e', 'f'];

      assert.deepEqual(values, expected, `should concatenate values property (expected: ${expected}, got: ${values})`);
    }

    ['@test concatenates subclasses when the values are functions'](assert) {
      let subKlass = klass.extend({
        functions: K
      });
      let obj = subKlass.create();

      let values = (0, _metal.get)(obj, 'functions');
      let expected = [K, K];

      assert.deepEqual(values, expected, `should concatenate functions property (expected: ${expected}, got: ${values})`);
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/array_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_metal, _object, _array, _internalTestHelpers) {
  'use strict';

  /*
    Implement a basic fake mutable array.  This validates that any non-native
    enumerable can impl this API.
  */
  const TestArray = _object.default.extend(_array.default, {
    _content: null,

    init() {
      this._content = this._content || [];
    },

    // some methods to modify the array so we can test changes.  Note that
    // arrays can be modified even if they don't implement MutableArray.  The
    // MutableArray is just a standard API for mutation but not required.
    addObject(obj) {
      let idx = this._content.length;
      (0, _metal.arrayContentWillChange)(this, idx, 0, 1);
      this._content.push(obj);
      (0, _metal.arrayContentDidChange)(this, idx, 0, 1);
    },

    removeFirst() {
      (0, _metal.arrayContentWillChange)(this, 0, 1, 0);
      this._content.shift();
      (0, _metal.arrayContentDidChange)(this, 0, 1, 0);
    },

    objectAt(idx) {
      return this._content[idx];
    },

    length: (0, _metal.computed)(function () {
      return this._content.length;
    })
  });

  (0, _internalTestHelpers.moduleFor)('Ember.Array', class extends _internalTestHelpers.AbstractTestCase {
    ['@test the return value of slice has Ember.Array applied'](assert) {
      let x = _object.default.extend(_array.default).create({
        length: 0
      });
      let y = x.slice(1);
      assert.equal(_array.default.detect(y), true, 'mixin should be applied');
    }

    ['@test slice supports negative index arguments'](assert) {
      let testArray = TestArray.create({ _content: [1, 2, 3, 4] });

      assert.deepEqual(testArray.slice(-2), [3, 4], 'slice(-2)');
      assert.deepEqual(testArray.slice(-2, -1), [3], 'slice(-2, -1');
      assert.deepEqual(testArray.slice(-2, -2), [], 'slice(-2, -2)');
      assert.deepEqual(testArray.slice(-1, -2), [], 'slice(-1, -2)');

      assert.deepEqual(testArray.slice(-4, 1), [1], 'slice(-4, 1)');
      assert.deepEqual(testArray.slice(-4, 5), [1, 2, 3, 4], 'slice(-4, 5)');
      assert.deepEqual(testArray.slice(-4), [1, 2, 3, 4], 'slice(-4)');

      assert.deepEqual(testArray.slice(0, -1), [1, 2, 3], 'slice(0, -1)');
      assert.deepEqual(testArray.slice(0, -4), [], 'slice(0, -4)');
      assert.deepEqual(testArray.slice(0, -3), [1], 'slice(0, -3)');
    }
  });

  // ..........................................................
  // CONTENT DID CHANGE
  //

  const DummyArray = _object.default.extend(_array.default, {
    length: 0,
    objectAt(idx) {
      return 'ITEM-' + idx;
    }
  });

  let obj, observer;

  // ..........................................................
  // NOTIFY ARRAY OBSERVERS
  //

  (0, _internalTestHelpers.moduleFor)('mixins/array/arrayContent[Will|Did]Change', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should notify observers of []'](assert) {
      obj = DummyArray.extend({
        enumerablePropertyDidChange: (0, _metal.observer)('[]', function () {
          this._count++;
        })
      }).create({
        _count: 0
      });

      assert.equal(obj._count, 0, 'should not have invoked yet');

      (0, _metal.arrayContentWillChange)(obj, 0, 1, 1);
      (0, _metal.arrayContentDidChange)(obj, 0, 1, 1);

      assert.equal(obj._count, 1, 'should have invoked');
    }
  });

  // ..........................................................
  // NOTIFY CHANGES TO LENGTH
  //

  (0, _internalTestHelpers.moduleFor)('notify observers of length', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach(assert) {
      obj = DummyArray.extend({
        lengthDidChange: (0, _metal.observer)('length', function () {
          this._after++;
        })
      }).create({
        _after: 0
      });

      assert.equal(obj._after, 0, 'should not have fired yet');
    }

    afterEach() {
      obj = null;
    }

    ['@test should notify observers when call with no params'](assert) {
      (0, _metal.arrayContentWillChange)(obj);
      assert.equal(obj._after, 0);

      (0, _metal.arrayContentDidChange)(obj);
      assert.equal(obj._after, 1);
    }

    // API variation that included items only
    ['@test should not notify when passed lengths are same'](assert) {
      (0, _metal.arrayContentWillChange)(obj, 0, 1, 1);
      assert.equal(obj._after, 0);

      (0, _metal.arrayContentDidChange)(obj, 0, 1, 1);
      assert.equal(obj._after, 0);
    }

    ['@test should notify when passed lengths are different'](assert) {
      (0, _metal.arrayContentWillChange)(obj, 0, 1, 2);
      assert.equal(obj._after, 0);

      (0, _metal.arrayContentDidChange)(obj, 0, 1, 2);
      assert.equal(obj._after, 1);
    }
  });

  // ..........................................................
  // NOTIFY ARRAY OBSERVER
  //

  (0, _internalTestHelpers.moduleFor)('notify array observers', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach(assert) {
      obj = DummyArray.create();

      observer = _object.default.extend({
        arrayWillChange() {
          assert.equal(this._before, null); // should only call once
          this._before = Array.prototype.slice.call(arguments);
        },

        arrayDidChange() {
          assert.equal(this._after, null); // should only call once
          this._after = Array.prototype.slice.call(arguments);
        }
      }).create({
        _before: null,
        _after: null
      });

      (0, _metal.addArrayObserver)(obj, observer);
    }

    afterEach() {
      obj = observer = null;
    }

    ['@test should notify array observers when called with no params'](assert) {
      (0, _metal.arrayContentWillChange)(obj);
      assert.deepEqual(observer._before, [obj, 0, -1, -1]);

      (0, _metal.arrayContentDidChange)(obj);
      assert.deepEqual(observer._after, [obj, 0, -1, -1]);
    }

    // API variation that included items only
    ['@test should notify when called with same length items'](assert) {
      (0, _metal.arrayContentWillChange)(obj, 0, 1, 1);
      assert.deepEqual(observer._before, [obj, 0, 1, 1]);

      (0, _metal.arrayContentDidChange)(obj, 0, 1, 1);
      assert.deepEqual(observer._after, [obj, 0, 1, 1]);
    }

    ['@test should notify when called with diff length items'](assert) {
      (0, _metal.arrayContentWillChange)(obj, 0, 2, 1);
      assert.deepEqual(observer._before, [obj, 0, 2, 1]);

      (0, _metal.arrayContentDidChange)(obj, 0, 2, 1);
      assert.deepEqual(observer._after, [obj, 0, 2, 1]);
    }

    ['@test removing array observer should disable'](assert) {
      (0, _metal.removeArrayObserver)(obj, observer);
      (0, _metal.arrayContentWillChange)(obj);
      assert.deepEqual(observer._before, null);

      (0, _metal.arrayContentDidChange)(obj);
      assert.deepEqual(observer._after, null);
    }
  });

  // ..........................................................
  // @each
  //

  let ary;

  (0, _internalTestHelpers.moduleFor)('EmberArray.@each support', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      ary = TestArray.create({
        _content: [{ isDone: true, desc: 'Todo 1' }, { isDone: false, desc: 'Todo 2' }, { isDone: true, desc: 'Todo 3' }, { isDone: false, desc: 'Todo 4' }]
      });
    }

    afterEach() {
      ary = null;
    }

    ['@test adding an object should notify (@each.isDone)'](assert) {
      let called = 0;

      let observerObject = _object.default.create({
        wasCalled() {
          called++;
        }
      });

      (0, _metal.addObserver)(ary, '@each.isDone', observerObject, 'wasCalled');

      ary.addObject(_object.default.create({
        desc: 'foo',
        isDone: false
      }));

      assert.equal(called, 1, 'calls observer when object is pushed');
    }

    ['@test getting @each is deprecated'](assert) {
      assert.expect(1);

      expectDeprecation(() => {
        (0, _metal.get)(ary, '@each');
      }, /Getting the '@each' property on object .* is deprecated/);
    }

    ['@test @each is readOnly'](assert) {
      assert.expect(1);

      assert.throws(function () {
        (0, _metal.set)(ary, '@each', 'foo');
      }, /Cannot set read-only property "@each"/);
    }

    ['@test using @each to observe arrays that does not return objects raise error'](assert) {
      let called = 0;

      let observerObject = _object.default.create({
        wasCalled() {
          called++;
        }
      });

      ary = TestArray.create({
        objectAt(idx) {
          return (0, _metal.get)(this._content[idx], 'desc');
        }
      });

      (0, _metal.addObserver)(ary, '@each.isDone', observerObject, 'wasCalled');

      expectAssertion(() => {
        ary.addObject(_object.default.create({
          desc: 'foo',
          isDone: false
        }));
      }, /When using @each to observe the array/);

      assert.equal(called, 0, 'not calls observer when object is pushed');
    }

    ['@test modifying the array should also indicate the isDone prop itself has changed'](assert) {
      // NOTE: we never actually get the '@each.isDone' property here.  This is
      // important because it tests the case where we don't have an isDone
      // EachArray materialized but just want to know when the property has
      // changed.
      let each;
      expectDeprecation(() => {
        each = (0, _metal.get)(ary, '@each');
      });
      let count = 0;

      (0, _metal.addObserver)(each, 'isDone', () => count++);

      count = 0;
      let item = (0, _metal.objectAt)(ary, 2);
      (0, _metal.set)(item, 'isDone', !(0, _metal.get)(item, 'isDone'));
      assert.equal(count, 1, '@each.isDone should have notified');
    }

    ['@test `objectAt` returns correct object'](assert) {
      let arr = ['first', 'second', 'third', 'fourth'];
      assert.equal((0, _metal.objectAt)(arr, 2), 'third');
      assert.equal((0, _metal.objectAt)(arr, 4), undefined);
    }

    ['@test should be clear caches for computed properties that have dependent keys on arrays that are changed after object initialization'](assert) {
      let obj = _object.default.extend({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'resources', (0, _array.A)());
        },

        common: (0, _metal.computed)('resources.@each.common', function () {
          return (0, _metal.get)((0, _metal.objectAt)((0, _metal.get)(this, 'resources'), 0), 'common');
        })
      }).create();

      (0, _metal.get)(obj, 'resources').pushObject(_object.default.create({ common: 'HI!' }));
      assert.equal('HI!', (0, _metal.get)(obj, 'common'));

      (0, _metal.set)((0, _metal.objectAt)((0, _metal.get)(obj, 'resources'), 0), 'common', 'BYE!');
      assert.equal('BYE!', (0, _metal.get)(obj, 'common'));
    }

    ['@test observers that contain @each in the path should fire only once the first time they are accessed'](assert) {
      let count = 0;

      let obj = _object.default.extend({
        init() {
          this._super(...arguments);
          // Observer does not fire on init
          (0, _metal.set)(this, 'resources', (0, _array.A)());
        },

        commonDidChange: (0, _metal.observer)('resources.@each.common', () => count++)
      }).create();

      // Observer fires second time when new object is added
      (0, _metal.get)(obj, 'resources').pushObject(_object.default.create({ common: 'HI!' }));
      // Observer fires third time when property on an object is changed
      (0, _metal.set)((0, _metal.objectAt)((0, _metal.get)(obj, 'resources'), 0), 'common', 'BYE!');

      assert.equal(count, 2, 'observers should only be called once');
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/comparable_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/compare', '@ember/-internals/runtime/lib/mixins/comparable', 'internal-test-helpers'], function (_metal, _object, _compare, _comparable, _internalTestHelpers) {
  'use strict';

  const Rectangle = _object.default.extend(_comparable.default, {
    length: 0,
    width: 0,

    area() {
      return (0, _metal.get)(this, 'length') * (0, _metal.get)(this, 'width');
    },

    compare(a, b) {
      return (0, _compare.default)(a.area(), b.area());
    }
  });

  let r1, r2;

  (0, _internalTestHelpers.moduleFor)('Comparable', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      r1 = Rectangle.create({ length: 6, width: 12 });
      r2 = Rectangle.create({ length: 6, width: 13 });
    }

    ['@test should be comparable and return the correct result'](assert) {
      assert.equal(_comparable.default.detect(r1), true);
      assert.equal((0, _compare.default)(r1, r1), 0);
      assert.equal((0, _compare.default)(r1, r2), -1);
      assert.equal((0, _compare.default)(r2, r1), 1);
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/container_proxy_test', ['@ember/-internals/owner', '@ember/-internals/container', '@ember/-internals/runtime/lib/mixins/container_proxy', '@ember/-internals/runtime/lib/system/object', '@ember/runloop', 'internal-test-helpers'], function (_owner, _container, _container_proxy, _object, _runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/runtime/mixins/container_proxy', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      this.Owner = _object.default.extend(_container_proxy.default);
      this.instance = this.Owner.create();

      this.registry = new _container.Registry();

      this.instance.__container__ = new _container.Container(this.registry, {
        owner: this.instance
      });
    }

    ['@test provides ownerInjection helper method'](assert) {
      let result = this.instance.ownerInjection();

      assert.equal(result[_owner.OWNER], this.instance, 'returns an object with the OWNER symbol');
    }

    ['@test actions queue completes before destruction'](assert) {
      assert.expect(1);

      this.registry.register('service:auth', _object.default.extend({
        willDestroy() {
          assert.ok((0, _owner.getOwner)(this).lookup('service:auth'), 'can still lookup');
        }
      }));

      let service = this.instance.lookup('service:auth');

      (0, _runloop.run)(() => {
        (0, _runloop.schedule)('actions', service, 'destroy');
        this.instance.destroy();
      });
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/enumerable_test', ['@ember/-internals/runtime/lib/mixins/enumerable', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_enumerable, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Enumerable', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should be mixed into A()'](assert) {
      assert.ok(_enumerable.default.detect((0, _array.A)()));
    }

    ['@test should be mixed into ArrayProxy'](assert) {
      assert.ok(_enumerable.default.detect(_array_proxy.default.create()));
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/evented_test', ['@ember/-internals/runtime/lib/mixins/evented', '@ember/-internals/runtime/lib/system/core_object', 'internal-test-helpers'], function (_evented, _core_object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.Evented', class extends _internalTestHelpers.AbstractTestCase {
    ['@test works properly on proxy-ish objects'](assert) {
      let eventedProxyObj = _core_object.default.extend(_evented.default, {
        unknownProperty() {
          return true;
        }
      }).create();

      let noop = function () {};

      eventedProxyObj.on('foo', noop);
      eventedProxyObj.off('foo', noop);

      assert.ok(true, 'An assertion was triggered');
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/mutable_enumerable_test', ['@ember/-internals/runtime/lib/mixins/mutable_enumerable', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_mutable_enumerable, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('MutableEnumerable', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should be mixed into A()'](assert) {
      assert.ok(_mutable_enumerable.default.detect((0, _array.A)()));
    }

    ['@test should be mixed into ArrayProxy'](assert) {
      assert.ok(_mutable_enumerable.default.detect(_array_proxy.default.create()));
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/observable_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('mixins/observable', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should be able to use getProperties to get a POJO of provided keys'](assert) {
      let obj = _object.default.create({
        firstName: 'Steve',
        lastName: 'Jobs',
        companyName: 'Apple, Inc.'
      });

      let pojo = obj.getProperties('firstName', 'lastName');
      assert.equal('Steve', pojo.firstName);
      assert.equal('Jobs', pojo.lastName);
    }

    ['@test should be able to use getProperties with array parameter to get a POJO of provided keys'](assert) {
      let obj = _object.default.create({
        firstName: 'Steve',
        lastName: 'Jobs',
        companyName: 'Apple, Inc.'
      });

      let pojo = obj.getProperties(['firstName', 'lastName']);
      assert.equal('Steve', pojo.firstName);
      assert.equal('Jobs', pojo.lastName);
    }

    ['@test should be able to use setProperties to set multiple properties at once'](assert) {
      let obj = _object.default.create({
        firstName: 'Steve',
        lastName: 'Jobs',
        companyName: 'Apple, Inc.'
      });

      obj.setProperties({ firstName: 'Tim', lastName: 'Cook' });
      assert.equal('Tim', obj.get('firstName'));
      assert.equal('Cook', obj.get('lastName'));
    }

    ['@test calling setProperties completes safely despite exceptions'](assert) {
      let exc = new Error('Something unexpected happened!');
      let obj = _object.default.extend({
        companyName: (0, _metal.computed)({
          get() {
            return 'Apple, Inc.';
          },
          set() {
            throw exc;
          }
        })
      }).create({
        firstName: 'Steve',
        lastName: 'Jobs'
      });

      let firstNameChangedCount = 0;

      (0, _metal.addObserver)(obj, 'firstName', () => firstNameChangedCount++);

      try {
        obj.setProperties({
          firstName: 'Tim',
          lastName: 'Cook',
          companyName: 'Fruit Co., Inc.'
        });
      } catch (err) {
        if (err !== exc) {
          throw err;
        }
      }

      assert.equal(firstNameChangedCount, 1, 'firstName should have fired once');
    }

    ['@test should be able to retrieve cached values of computed properties without invoking the computed property'](assert) {
      let obj = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'foo';
        })
      }).create({
        bar: 'bar'
      });

      assert.equal(obj.cacheFor('foo'), undefined, 'should return undefined if no value has been cached');
      (0, _metal.get)(obj, 'foo');

      assert.equal((0, _metal.get)(obj, 'foo'), 'foo', 'precond - should cache the value');
      assert.equal(obj.cacheFor('foo'), 'foo', 'should return the cached value after it is invoked');

      assert.equal(obj.cacheFor('bar'), undefined, 'returns undefined if the value is not a computed property');
    }

    ['@test incrementProperty should work even if value is number in string'](assert) {
      let obj = _object.default.create({
        age: '24'
      });
      obj.incrementProperty('age');
      assert.equal(25, obj.get('age'));
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/promise_proxy_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object_proxy', '@ember/-internals/runtime/lib/mixins/promise_proxy', '@ember/-internals/runtime/lib/ext/rsvp', 'rsvp', 'internal-test-helpers'], function (_runloop, _metal, _object_proxy, _promise_proxy, _rsvp, _rsvp2, _internalTestHelpers) {
  'use strict';

  let ObjectPromiseProxy;

  (0, _internalTestHelpers.moduleFor)('Ember.PromiseProxy - ObjectProxy', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      ObjectPromiseProxy = _object_proxy.default.extend(_promise_proxy.default);
    }

    afterEach() {
      _rsvp2.on('error', _rsvp.onerrorDefault);
    }

    ['@test present on ember namespace'](assert) {
      assert.ok(_promise_proxy.default, 'expected PromiseProxyMixin to exist');
    }

    ['@test no promise, invoking then should raise'](assert) {
      let proxy = ObjectPromiseProxy.create();

      assert.throws(function () {
        proxy.then(function () {
          return this;
        }, function () {
          return this;
        });
      }, new RegExp("PromiseProxy's promise must be set"));
    }

    ['@test fulfillment'](assert) {
      let value = {
        firstName: 'stef',
        lastName: 'penner'
      };

      let deferred = _rsvp2.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      let didFulfillCount = 0;
      let didRejectCount = 0;

      proxy.then(() => didFulfillCount++, () => didRejectCount++);

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy to have no reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), true, 'expects the proxy to indicate that it is loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), false, 'expects the proxy to indicate that it is not settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      assert.equal(didFulfillCount, 0, 'should not yet have been fulfilled');
      assert.equal(didRejectCount, 0, 'should not yet have been rejected');

      (0, _runloop.run)(deferred, 'resolve', value);

      assert.equal(didFulfillCount, 1, 'should have been fulfilled');
      assert.equal(didRejectCount, 0, 'should not have been rejected');

      assert.equal((0, _metal.get)(proxy, 'content'), value, 'expects the proxy to have content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy to still have no reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is no longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), true, 'expects the proxy to indicate that it is fulfilled');

      (0, _runloop.run)(deferred, 'resolve', value);

      assert.equal(didFulfillCount, 1, 'should still have been only fulfilled once');
      assert.equal(didRejectCount, 0, 'should still not have been rejected');

      (0, _runloop.run)(deferred, 'reject', value);

      assert.equal(didFulfillCount, 1, 'should still have been only fulfilled once');
      assert.equal(didRejectCount, 0, 'should still not have been rejected');

      assert.equal((0, _metal.get)(proxy, 'content'), value, 'expects the proxy to have still have same content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy still to have no reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is no longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), true, 'expects the proxy to indicate that it is fulfilled');

      // rest of the promise semantics are tested in directly in RSVP
    }

    ['@test rejection'](assert) {
      let reason = new Error('failure');
      let deferred = _rsvp2.defer();
      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      let didFulfillCount = 0;
      let didRejectCount = 0;

      proxy.then(() => didFulfillCount++, () => didRejectCount++);

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy to have no reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), true, 'expects the proxy to indicate that it is loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), false, 'expects the proxy to indicate that it is not settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      assert.equal(didFulfillCount, 0, 'should not yet have been fulfilled');
      assert.equal(didRejectCount, 0, 'should not yet have been rejected');

      (0, _runloop.run)(deferred, 'reject', reason);

      assert.equal(didFulfillCount, 0, 'should not yet have been fulfilled');
      assert.equal(didRejectCount, 1, 'should have been rejected');

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), reason, 'expects the proxy to have a reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is not longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), true, 'expects the proxy to indicate that it is  rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      (0, _runloop.run)(deferred, 'reject', reason);

      assert.equal(didFulfillCount, 0, 'should stll not yet have been fulfilled');
      assert.equal(didRejectCount, 1, 'should still remain rejected');

      (0, _runloop.run)(deferred, 'resolve', 1);

      assert.equal(didFulfillCount, 0, 'should stll not yet have been fulfilled');
      assert.equal(didRejectCount, 1, 'should still remain rejected');

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), reason, 'expects the proxy to have a reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is not longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), true, 'expects the proxy to indicate that it is  rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');
    }

    // https://github.com/emberjs/ember.js/issues/15694
    ['@test rejection without specifying reason'](assert) {
      let deferred = _rsvp2.defer();
      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      let didFulfillCount = 0;
      let didRejectCount = 0;

      proxy.then(() => didFulfillCount++, () => didRejectCount++);

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy to have no reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), true, 'expects the proxy to indicate that it is loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), false, 'expects the proxy to indicate that it is not settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      assert.equal(didFulfillCount, 0, 'should not yet have been fulfilled');
      assert.equal(didRejectCount, 0, 'should not yet have been rejected');

      (0, _runloop.run)(deferred, 'reject');

      assert.equal(didFulfillCount, 0, 'should not yet have been fulfilled');
      assert.equal(didRejectCount, 1, 'should have been rejected');

      assert.equal((0, _metal.get)(proxy, 'content'), undefined, 'expects the proxy to have no content');
      assert.equal((0, _metal.get)(proxy, 'reason'), undefined, 'expects the proxy to have a reason');
      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is not longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), true, 'expects the proxy to indicate that it is  rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');
    }

    ["@test unhandled rejects still propagate to RSVP.on('error', ...) "](assert) {
      assert.expect(1);

      _rsvp2.on('error', onerror);
      _rsvp2.off('error', _rsvp.onerrorDefault);

      let expectedReason = new Error('failure');
      let deferred = _rsvp2.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.get('promise');

      function onerror(reason) {
        assert.equal(reason, expectedReason, 'expected reason');
      }

      _rsvp2.on('error', onerror);
      _rsvp2.off('error', _rsvp.onerrorDefault);

      (0, _runloop.run)(deferred, 'reject', expectedReason);

      _rsvp2.on('error', _rsvp.onerrorDefault);
      _rsvp2.off('error', onerror);

      (0, _runloop.run)(deferred, 'reject', expectedReason);

      _rsvp2.on('error', _rsvp.onerrorDefault);
      _rsvp2.off('error', onerror);
    }

    ['@test should work with promise inheritance'](assert) {
      class PromiseSubclass extends _rsvp2.Promise {}

      let proxy = ObjectPromiseProxy.create({
        promise: new PromiseSubclass(() => {})
      });

      assert.ok(proxy.then() instanceof PromiseSubclass, 'promise proxy respected inheritance');
    }

    ['@test should reset isFulfilled and isRejected when promise is reset'](assert) {
      let deferred = _rsvp.default.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      assert.equal((0, _metal.get)(proxy, 'isPending'), true, 'expects the proxy to indicate that it is loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), false, 'expects the proxy to indicate that it is not settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      (0, _runloop.run)(deferred, 'resolve');

      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is no longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), true, 'expects the proxy to indicate that it is fulfilled');

      let anotherDeferred = _rsvp.default.defer();
      proxy.set('promise', anotherDeferred.promise);

      assert.equal((0, _metal.get)(proxy, 'isPending'), true, 'expects the proxy to indicate that it is loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), false, 'expects the proxy to indicate that it is not settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), false, 'expects the proxy to indicate that it is not rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');

      (0, _runloop.run)(anotherDeferred, 'reject');

      assert.equal((0, _metal.get)(proxy, 'isPending'), false, 'expects the proxy to indicate that it is not longer loading');
      assert.equal((0, _metal.get)(proxy, 'isSettled'), true, 'expects the proxy to indicate that it is settled');
      assert.equal((0, _metal.get)(proxy, 'isRejected'), true, 'expects the proxy to indicate that it is  rejected');
      assert.equal((0, _metal.get)(proxy, 'isFulfilled'), false, 'expects the proxy to indicate that it is not fulfilled');
    }

    ['@test should have content when isFulfilled is set'](assert) {
      let deferred = _rsvp.default.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.addObserver('isFulfilled', () => assert.equal((0, _metal.get)(proxy, 'content'), true));

      (0, _runloop.run)(deferred, 'resolve', true);
    }

    ['@test should have reason when isRejected is set'](assert) {
      let error = new Error('Y U REJECT?!?');
      let deferred = _rsvp.default.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.addObserver('isRejected', () => assert.equal((0, _metal.get)(proxy, 'reason'), error));

      try {
        (0, _runloop.run)(deferred, 'reject', error);
      } catch (e) {
        assert.equal(e, error);
      }
    }

    ['@test should not error if promise is resolved after proxy has been destroyed'](assert) {
      let deferred = _rsvp.default.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.then(() => {}, () => {});

      (0, _runloop.run)(proxy, 'destroy');

      (0, _runloop.run)(deferred, 'resolve', true);

      assert.ok(true, 'resolving the promise after the proxy has been destroyed does not raise an error');
    }

    ['@test should not error if promise is rejected after proxy has been destroyed'](assert) {
      let deferred = _rsvp.default.defer();

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.then(() => {}, () => {});

      (0, _runloop.run)(proxy, 'destroy');

      (0, _runloop.run)(deferred, 'reject', 'some reason');

      assert.ok(true, 'rejecting the promise after the proxy has been destroyed does not raise an error');
    }

    ['@test promise chain is not broken if promised is resolved after proxy has been destroyed'](assert) {
      let deferred = _rsvp.default.defer();
      let expectedValue = {};
      let receivedValue;
      let didResolveCount = 0;

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.then(value => {
        receivedValue = value;
        didResolveCount++;
      }, () => {});

      (0, _runloop.run)(proxy, 'destroy');

      (0, _runloop.run)(deferred, 'resolve', expectedValue);

      assert.equal(didResolveCount, 1, 'callback called');
      assert.equal(receivedValue, expectedValue, 'passed value is the value the promise was resolved with');
    }

    ['@test promise chain is not broken if promised is rejected after proxy has been destroyed'](assert) {
      let deferred = _rsvp.default.defer();
      let expectedReason = 'some reason';
      let receivedReason;
      let didRejectCount = 0;

      let proxy = ObjectPromiseProxy.create({
        promise: deferred.promise
      });

      proxy.then(() => {}, reason => {
        receivedReason = reason;
        didRejectCount++;
      });

      (0, _runloop.run)(proxy, 'destroy');

      (0, _runloop.run)(deferred, 'reject', expectedReason);

      assert.equal(didRejectCount, 1, 'callback called');
      assert.equal(receivedReason, expectedReason, 'passed reason is the reason the promise was rejected for');
    }
  });
});
enifed('@ember/-internals/runtime/tests/mixins/target_action_support_test', ['@ember/-internals/environment', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/target_action_support', 'internal-test-helpers'], function (_environment, _object, _target_action_support, _internalTestHelpers) {
  'use strict';

  let originalLookup = _environment.context.lookup;
  let lookup;

  (0, _internalTestHelpers.moduleFor)('TargetActionSupport', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      _environment.context.lookup = lookup = {};
    }

    afterEach() {
      _environment.context.lookup = originalLookup;
    }

    ['@test it should return false if no target or action are specified'](assert) {
      assert.expect(1);

      let obj = _object.default.extend(_target_action_support.default).create();

      assert.ok(false === obj.triggerAction(), 'no target or action was specified');
    }

    ['@test it should support actions specified as strings'](assert) {
      assert.expect(2);

      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          anEvent() {
            assert.ok(true, 'anEvent method was called');
          }
        }),

        action: 'anEvent'
      });

      assert.ok(true === obj.triggerAction(), 'a valid target and action were specified');
    }

    ['@test it should invoke the send() method on objects that implement it'](assert) {
      assert.expect(3);

      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          send(evt, context) {
            assert.equal(evt, 'anEvent', 'send() method was invoked with correct event name');
            assert.equal(context, obj, 'send() method was invoked with correct context');
          }
        }),

        action: 'anEvent'
      });

      assert.ok(true === obj.triggerAction(), 'a valid target and action were specified');
    }

    ['@test it should find targets specified using a property path'](assert) {
      assert.expect(2);

      let Test = {};
      lookup.Test = Test;

      Test.targetObj = _object.default.create({
        anEvent() {
          assert.ok(true, 'anEvent method was called on global object');
        }
      });

      let myObj = _object.default.extend(_target_action_support.default).create({
        target: 'Test.targetObj',
        action: 'anEvent'
      });

      assert.ok(true === myObj.triggerAction(), 'a valid target and action were specified');
    }

    ['@test it should use an actionContext object specified as a property on the object'](assert) {
      assert.expect(2);
      let obj = _object.default.extend(_target_action_support.default).create({
        action: 'anEvent',
        actionContext: {},
        target: _object.default.create({
          anEvent(ctx) {
            assert.ok(obj.actionContext === ctx, 'anEvent method was called with the expected context');
          }
        })
      });
      assert.ok(true === obj.triggerAction(), 'a valid target and action were specified');
    }

    ['@test it should raise a deprecation warning when targetObject is specified and used'](assert) {
      assert.expect(4);
      let obj;
      expectDeprecation(() => {
        obj = _object.default.extend(_target_action_support.default).create({
          action: 'anEvent',
          actionContext: {},
          targetObject: _object.default.create({
            anEvent(ctx) {
              assert.ok(obj.actionContext === ctx, 'anEvent method was called with the expected context');
            }
          })
        });
      }, /Usage of `targetObject` is deprecated. Please use `target` instead./);
      assert.ok(true === obj.triggerAction(), 'a valid targetObject and action were specified');
      expectDeprecation(() => obj.get('targetObject'), /Usage of `targetObject` is deprecated. Please use `target` instead./);
    }

    ['@test it should find an actionContext specified as a property path'](assert) {
      assert.expect(2);

      let Test = {};
      lookup.Test = Test;
      Test.aContext = {};

      let obj = _object.default.extend(_target_action_support.default).create({
        action: 'anEvent',
        actionContext: 'Test.aContext',
        target: _object.default.create({
          anEvent(ctx) {
            assert.ok(Test.aContext === ctx, 'anEvent method was called with the expected context');
          }
        })
      });

      assert.ok(true === obj.triggerAction(), 'a valid target and action were specified');
    }

    ['@test it should use the target specified in the argument'](assert) {
      assert.expect(2);
      let targetObj = _object.default.create({
        anEvent() {
          assert.ok(true, 'anEvent method was called');
        }
      });
      let obj = _object.default.extend(_target_action_support.default).create({
        action: 'anEvent'
      });

      assert.ok(true === obj.triggerAction({ target: targetObj }), 'a valid target and action were specified');
    }

    ['@test it should use the action specified in the argument'](assert) {
      assert.expect(2);

      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          anEvent() {
            assert.ok(true, 'anEvent method was called');
          }
        })
      });
      assert.ok(true === obj.triggerAction({ action: 'anEvent' }), 'a valid target and action were specified');
    }

    ['@test it should use the actionContext specified in the argument'](assert) {
      assert.expect(2);
      let context = {};
      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          anEvent(ctx) {
            assert.ok(context === ctx, 'anEvent method was called with the expected context');
          }
        }),
        action: 'anEvent'
      });

      assert.ok(true === obj.triggerAction({ actionContext: context }), 'a valid target and action were specified');
    }

    ['@test it should allow multiple arguments from actionContext'](assert) {
      assert.expect(3);
      let param1 = 'someParam';
      let param2 = 'someOtherParam';
      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          anEvent(first, second) {
            assert.ok(first === param1, 'anEvent method was called with the expected first argument');
            assert.ok(second === param2, 'anEvent method was called with the expected second argument');
          }
        }),
        action: 'anEvent'
      });

      assert.ok(true === obj.triggerAction({ actionContext: [param1, param2] }), 'a valid target and action were specified');
    }

    ['@test it should use a null value specified in the actionContext argument'](assert) {
      assert.expect(2);
      let obj = _object.default.extend(_target_action_support.default).create({
        target: _object.default.create({
          anEvent(ctx) {
            assert.ok(null === ctx, 'anEvent method was called with the expected context (null)');
          }
        }),
        action: 'anEvent'
      });
      assert.ok(true === obj.triggerAction({ actionContext: null }), 'a valid target and action were specified');
    }
  });
});
enifed('@ember/-internals/runtime/tests/mutable-array/addObject-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class AddObjectTest extends _internalTestHelpers.AbstractTestCase {
    '@test should return receiver'() {
      let before = (0, _array.newFixture)(3);
      let obj = this.newObject(before);
      this.assert.equal(obj.addObject(before[1]), obj, 'should return receiver');
    }

    '@test [A,B].addObject(C) => [A,B,C] + notify'() {
      let before = (0, _array.newFixture)(2);
      let item = (0, _array.newFixture)(1)[0];
      let after = [before[0], before[1], item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.addObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
        this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      }
    }

    '@test [A,B,C].addObject(A) => [A,B,C] + NO notify'() {
      let before = (0, _array.newFixture)(3);
      let after = before;
      let item = before[0];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.addObject(item); // note: item in set

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.validate('[]'), false, 'should NOT have notified []');
        this.assert.equal(observer.validate('@each'), false, 'should NOT have notified @each');
        this.assert.equal(observer.validate('length'), false, 'should NOT have notified length');
        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
      }
    }
  }

  (0, _array.runArrayTests)('addObject', AddObjectTest, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/clear-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class ClearTests extends _internalTestHelpers.AbstractTestCase {
    '@test [].clear() => [] + notify'() {
      let before = [];
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.clear(), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.validate('[]'), false, 'should NOT have notified [] once');
      this.assert.equal(observer.validate('@each'), false, 'should NOT have notified @each once');
      this.assert.equal(observer.validate('length'), false, 'should NOT have notified length once');
      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [X].clear() => [] + notify'() {
      var obj, before, after, observer;

      before = (0, _array.newFixture)(1);
      after = [];
      obj = this.newObject(before);
      observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');
      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.clear(), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }
  }

  (0, _array.runArrayTests)('clear', ClearTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/insertAt-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class InsertAtTests extends _internalTestHelpers.AbstractTestCase {
    '@test [].insertAt(0, X) => [X] + notify'() {
      let after = (0, _array.newFixture)(1);
      let obj = this.newObject([]);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.insertAt(0, after[0]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] did change once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each did change once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length did change once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject did change once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject did change once');
    }

    '@test [].insertAt(200,X) => OUT_OF_RANGE_EXCEPTION exception'() {
      let obj = this.newObject([]);
      let that = this;

      this.assert.throws(() => obj.insertAt(200, that.newFixture(1)[0]), Error);
    }

    '@test [A].insertAt(0, X) => [X,A] + notify'() {
      let item = (0, _array.newFixture)(1)[0];
      let before = (0, _array.newFixture)(1);
      let after = [item, before[0]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.insertAt(0, item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [A].insertAt(1, X) => [A,X] + notify'() {
      let item = (0, _array.newFixture)(1)[0];
      let before = (0, _array.newFixture)(1);
      let after = [before[0], item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.insertAt(1, item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
    }

    '@test [A].insertAt(200,X) => OUT_OF_RANGE exception'() {
      let obj = this.newObject((0, _array.newFixture)(1));
      let that = this;

      this.assert.throws(() => obj.insertAt(200, that.newFixture(1)[0]), Error);
    }

    '@test [A,B,C].insertAt(0,X) => [X,A,B,C] + notify'() {
      let item = (0, _array.newFixture)(1)[0];
      let before = (0, _array.newFixture)(3);
      let after = [item, before[0], before[1], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.insertAt(0, item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [A,B,C].insertAt(1,X) => [A,X,B,C] + notify'() {
      let item = (0, _array.newFixture)(1)[0];
      let before = (0, _array.newFixture)(3);
      let after = [before[0], item, before[1], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');
      let objectAtCalls = [];

      let objectAt = obj.objectAt;
      obj.objectAt = ix => {
        objectAtCalls.push(ix);
        return objectAt.call(obj, ix);
      };

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      objectAtCalls.splice(0, objectAtCalls.length);

      obj.insertAt(1, item);
      this.assert.deepEqual(objectAtCalls, [], 'objectAt is not called when only inserting items');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [A,B,C].insertAt(3,X) => [A,B,C,X] + notify'() {
      let item = (0, _array.newFixture)(1)[0];
      let before = (0, _array.newFixture)(3);
      let after = [before[0], before[1], before[2], item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.insertAt(3, item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
    }
  }

  (0, _array.runArrayTests)('instertAt', InsertAtTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/popObject-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class PopObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test [].popObject() => [] + returns undefined + NO notify'() {
      let obj = this.newObject([]);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.popObject(), undefined, 'popObject results');

      this.assert.deepEqual(this.toArray(obj), [], 'post item results');

      this.assert.equal(observer.validate('[]'), false, 'should NOT have notified []');
      this.assert.equal(observer.validate('@each'), false, 'should NOT have notified @each');
      this.assert.equal(observer.validate('length'), false, 'should NOT have notified length');
      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [X].popObject() => [] + notify'() {
      let before = (0, _array.newFixture)(1);
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      let ret = obj.popObject();

      this.assert.equal(ret, before[0], 'return object');
      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].popObject() => [A,B] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [before[0], before[1]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      let ret = obj.popObject();

      this.assert.equal(ret, before[2], 'return object');
      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
    }
  }

  (0, _array.runArrayTests)('popObject', PopObjectTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/pushObject-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class PushObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test returns pushed object'() {
      let exp = (0, _array.newFixture)(1)[0];
      let obj = this.newObject([]);

      this.assert.equal(obj.pushObject(exp), exp, 'should return pushed object');
    }

    '@test [].pushObject(X) => [X] + notify'() {
      let before = [];
      let after = (0, _array.newFixture)(1);
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.pushObject(after[0]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].pushObject(X) => [A,B,C,X] + notify'() {
      let before = (0, _array.newFixture)(3);
      let item = (0, _array.newFixture)(1)[0];
      let after = [before[0], before[1], before[2], item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.pushObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
    }

    '@test [A,B,C,C].pushObject(A) => [A,B,C,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let item = before[2]; // note same object as current tail. should end up twice
      let after = [before[0], before[1], before[2], item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.pushObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
      this.assert.equal(observer.validate('lastObject'), true, 'should have notified lastObject');
    }
  }

  (0, _array.runArrayTests)('pushObject', PushObjectTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/pushObjects-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class PushObjectsTests extends _internalTestHelpers.AbstractTestCase {
    '@test should raise exception if not Ember.Enumerable is passed to pushObjects'() {
      let obj = this.newObject([]);

      expectAssertion(() => obj.pushObjects('string'));
    }
  }

  (0, _array.runArrayTests)('pushObjects', PushObjectsTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/removeAt-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/runtime/lib/mixins/array', '@ember/-internals/metal'], function (_internalTestHelpers, _array, _array2, _metal) {
  'use strict';

  class RemoveAtTests extends _internalTestHelpers.AbstractTestCase {
    '@test removeAt([X], 0) => [] + notify'() {
      let before = (0, _array.newFixture)(1);
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal((0, _array2.removeAt)(obj, 0), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test removeAt([], 200) => OUT_OF_RANGE_EXCEPTION exception'() {
      let obj = this.newObject([]);
      this.assert.throws(() => (0, _array2.removeAt)(obj, 200), Error);
    }

    '@test removeAt([A,B], 0) => [B] + notify'() {
      let before = (0, _array.newFixture)(2);
      let after = [before[1]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal((0, _array2.removeAt)(obj, 0), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test removeAt([A,B], 1) => [A] + notify'() {
      let before = (0, _array.newFixture)(2);
      let after = [before[0]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal((0, _array2.removeAt)(obj, 1), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
    }

    '@test removeAt([A,B,C], 1) => [A,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [before[0], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal((0, _array2.removeAt)(obj, 1), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test removeAt([A,B,C,D], 1,2) => [A,D] + notify'() {
      let before = (0, _array.newFixture)(4);
      let after = [before[0], before[3]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal((0, _array2.removeAt)(obj, 1, 2), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [A,B,C,D].removeAt(1,2) => [A,D] + notify'() {
      var obj, before, after, observer;

      before = (0, _array.newFixture)(4);
      after = [before[0], before[3]];
      obj = this.newObject(before);
      observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');
      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.removeAt(1, 2), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }
  }

  (0, _array.runArrayTests)('removeAt', RemoveAtTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/removeObject-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_metal, _internalTestHelpers, _array) {
  'use strict';

  class RemoveObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return receiver'() {
      let before = (0, _array.newFixture)(3);
      let obj = this.newObject(before);

      this.assert.equal(obj.removeObject(before[1]), obj, 'should return receiver');
    }

    '@test [A,B,C].removeObject(B) => [A,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [before[0], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.removeObject(before[1]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
      }
    }

    '@test [A,B,C].removeObject(D) => [A,B,C]'() {
      let before = (0, _array.newFixture)(3);
      let after = before;
      let item = (0, _array.newFixture)(1)[0];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.removeObject(item); // note: item not in set

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.validate('[]'), false, 'should NOT have notified []');
        this.assert.equal(observer.validate('@each'), false, 'should NOT have notified @each');
        this.assert.equal(observer.validate('length'), false, 'should NOT have notified length');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
      }
    }
  }

  (0, _array.runArrayTests)('removeObject', RemoveObjectTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/removeObjects-test', ['@ember/-internals/metal', 'internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/runtime/lib/mixins/array'], function (_metal, _internalTestHelpers, _array, _array2) {
  'use strict';

  class RemoveObjectsTests extends _internalTestHelpers.AbstractTestCase {
    '@test should return receiver'() {
      let before = (0, _array2.A)((0, _array.newFixture)(3));
      let obj = before;

      this.assert.equal(obj.removeObjects(before[1]), obj, 'should return receiver');
    }

    '@test [A,B,C].removeObjects([B]) => [A,C] + notify'() {
      let before = (0, _array2.A)((0, _array.newFixture)(3));
      let after = [before[0], before[2]];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([before[1]]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
      }
    }

    '@test [{A},{B},{C}].removeObjects([{B}]) => [{A},{C}] + notify'() {
      let before = (0, _array2.A)((0, _array.newObjectsFixture)(3));
      let after = [before[0], before[2]];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([before[1]]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
      }
    }

    '@test [A,B,C].removeObjects([A,B]) => [C] + notify'() {
      let before = (0, _array2.A)((0, _array.newFixture)(3));
      let after = [before[2]];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([before[0], before[1]]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
      }
    }

    '@test [{A},{B},{C}].removeObjects([{A},{B}]) => [{C}] + notify'() {
      let before = (0, _array2.A)((0, _array.newObjectsFixture)(3));
      let after = [before[2]];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([before[0], before[1]]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
      }
    }

    '@test [A,B,C].removeObjects([A,B,C]) => [] + notify'() {
      let before = (0, _array2.A)((0, _array.newFixture)(3));
      let after = [];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([before[0], before[1], before[2]]);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject');
        this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject');
      }
    }

    '@test [{A},{B},{C}].removeObjects([{A},{B},{C}]) => [] + notify'() {
      let before = (0, _array2.A)((0, _array.newObjectsFixture)(3));
      let after = [];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects(before);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
        this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

        this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), 1, 'should have notified lastObject');
      }
    }

    '@test [A,B,C].removeObjects([D]) => [A,B,C]'() {
      let before = (0, _array2.A)((0, _array.newFixture)(3));
      let after = before;
      let item = (0, _array.newFixture)(1)[0];
      let obj = before;
      let observer = this.newObserver(obj, '[]', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); // Prime the cache

      obj.removeObjects([item]); // Note: item not in set

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      if (observer.isEnabled) {
        this.assert.equal(observer.validate('[]'), false, 'should NOT have notified []');
        this.assert.equal(observer.validate('length'), false, 'should NOT have notified length');

        this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject');
        this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
      }
    }
  }

  (0, _array.runArrayTests)('removeObjects', RemoveObjectsTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/replace-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _array) {
  'use strict';

  class ReplaceTests extends _internalTestHelpers.AbstractTestCase {
    "@test [].replace(0,0,'X') => ['X'] + notify"() {
      let exp = (0, _array.newFixture)(1);
      let obj = this.newObject([]);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(0, 0, exp);

      this.assert.deepEqual(this.toArray(obj), exp, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [].replace(0,0,"X") => ["X"] + avoid calling objectAt and notifying fistObject/lastObject when not in cache'() {
      var obj, exp, observer;
      var called = 0;
      exp = (0, _array.newFixture)(1);
      obj = this.newObject([]);
      obj.objectAt = function () {
        called++;
      };
      observer = this.newObserver(obj, 'firstObject', 'lastObject');

      obj.replace(0, 0, exp);

      this.assert.equal(called, 0, 'should NOT have called objectAt upon replace when firstObject/lastObject are not cached');
      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject since not cached');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject since not cached');
    }

    '@test [A,B,C,D].replace(1,2,X) => [A,X,D] + notify'() {
      let before = (0, _array.newFixture)(4);
      let replace = (0, _array.newFixture)(1);
      let after = [before[0], replace[0], before[3]];

      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(1, 2, replace);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [A,B,C,D].replace(1,2,[X,Y]) => [A,X,Y,D] + notify'() {
      let before = (0, _array.newFixture)(4);
      let replace = (0, _array.newFixture)(2);
      let after = [before[0], replace[0], replace[1], before[3]];

      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(1, 2, replace);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.validate('length'), false, 'should NOT have notified length');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [A,B].replace(1,0,[X,Y]) => [A,X,Y,B] + notify'() {
      let before = (0, _array.newFixture)(2);
      let replace = (0, _array.newFixture)(2);
      let after = [before[0], replace[0], replace[1], before[1]];

      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(1, 0, replace);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [A,B,C,D].replace(2,2) => [A,B] + notify'() {
      let before = (0, _array.newFixture)(4);
      let after = [before[0], before[1]];

      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(2, 2);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
    }

    '@test [A,B,C,D].replace(-1,1) => [A,B,C] + notify'() {
      let before = (0, _array.newFixture)(4);
      let after = [before[0], before[1], before[2]];

      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.replace(-1, 1);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
    }

    '@test Adding object should notify array observer'() {
      let fixtures = (0, _array.newFixture)(4);
      let obj = this.newObject(fixtures);
      let observer = this.newObserver(obj).observeArray(obj);
      let item = (0, _array.newFixture)(1)[0];

      obj.replace(2, 2, [item]);

      this.assert.deepEqual(observer._before, [obj, 2, 2, 1], 'before');
      this.assert.deepEqual(observer._after, [obj, 2, 2, 1], 'after');
    }
  }

  (0, _array.runArrayTests)('replace', ReplaceTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/reverseObjects-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/metal'], function (_internalTestHelpers, _array, _metal) {
  'use strict';

  class ReverseObjectsTests extends _internalTestHelpers.AbstractTestCase {
    '@test [A,B,C].reverseObjects() => [] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [before[2], before[1], before[0]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.reverseObjects(), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 0, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }
  }

  (0, _array.runArrayTests)('reverseObjects', ReverseObjectsTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/setObjects-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/metal'], function (_internalTestHelpers, _array, _metal) {
  'use strict';

  class SetObjectsTests extends _internalTestHelpers.AbstractTestCase {
    '@test [A,B,C].setObjects([]) = > [] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.setObjects(after), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].setObjects([D, E, F, G]) = > [D, E, F, G] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = (0, _array.newFixture)(4);
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.setObjects(after), obj, 'return self');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }
  }

  (0, _array.runArrayTests)('setObjects', SetObjectsTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/shiftObject-test', ['internal-test-helpers', '@ember/-internals/runtime/tests/helpers/array', '@ember/-internals/metal'], function (_internalTestHelpers, _array, _metal) {
  'use strict';

  class ShiftObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test [].shiftObject() => [] + returns undefined + NO notify'() {
      let before = [];
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.shiftObject(), undefined);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.validate('[]', undefined, 1), false, 'should NOT have notified [] once');
      this.assert.equal(observer.validate('@each', undefined, 1), false, 'should NOT have notified @each once');
      this.assert.equal(observer.validate('length', undefined, 1), false, 'should NOT have notified length once');

      this.assert.equal(observer.validate('firstObject'), false, 'should NOT have notified firstObject once');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }

    '@test [X].shiftObject() => [] + notify'() {
      let before = (0, _array.newFixture)(1);
      let after = [];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.shiftObject(), before[0], 'should return object');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].shiftObject() => [B,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let after = [before[1], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      this.assert.equal(obj.shiftObject(), before[0], 'should return object');

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject once');
    }
  }

  (0, _array.runArrayTests)('shiftObject', ShiftObjectTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/unshiftObject-test', ['internal-test-helpers', '@ember/-internals/metal', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _metal, _array) {
  'use strict';

  class UnshiftObjectTests extends _internalTestHelpers.AbstractTestCase {
    '@test returns unshifted object'() {
      let obj = this.newObject([]);
      let item = (0, _array.newFixture)(1)[0];

      this.assert.equal(obj.unshiftObject(item), item, 'should return unshifted object');
    }

    '@test [].unshiftObject(X) => [X] + notify'() {
      let before = [];
      let item = (0, _array.newFixture)(1)[0];
      let after = [item];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].unshiftObject(X) => [X,A,B,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let item = (0, _array.newFixture)(1)[0];
      let after = [item, before[0], before[1], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [A,B,C].unshiftObject(A) => [A,A,B,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let item = before[0]; // note same object as current head. should end up twice
      let after = [item, before[0], before[1], before[2]];
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObject(item);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), true, 'should have notified firstObject');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }
  }

  (0, _array.runArrayTests)('unshiftObject', UnshiftObjectTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/mutable-array/unshiftObjects-test', ['internal-test-helpers', '@ember/-internals/metal', '@ember/-internals/runtime/tests/helpers/array'], function (_internalTestHelpers, _metal, _array) {
  'use strict';

  class UnshiftObjectsTests extends _internalTestHelpers.AbstractTestCase {
    '@test returns receiver'() {
      let obj = this.newObject([]);
      let items = (0, _array.newFixture)(3);

      this.assert.equal(obj.unshiftObjects(items), obj, 'should return receiver');
    }

    '@test [].unshiftObjects([A,B,C]) => [A,B,C] + notify'() {
      let before = [];
      let items = (0, _array.newFixture)(3);
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObjects(items);

      this.assert.deepEqual(this.toArray(obj), items, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), items.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');
      this.assert.equal(observer.timesCalled('lastObject'), 1, 'should have notified lastObject once');
    }

    '@test [A,B,C].unshiftObjects([X,Y]) => [X,Y,A,B,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let items = (0, _array.newFixture)(2);
      let after = items.concat(before);
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObjects(items);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');
      this.assert.equal(observer.timesCalled('firstObject'), 1, 'should have notified firstObject once');

      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }

    '@test [A,B,C].unshiftObjects([A,B]) => [A,B,A,B,C] + notify'() {
      let before = (0, _array.newFixture)(3);
      let items = [before[0], before[1]]; // note same object as current head. should end up twice
      let after = items.concat(before);
      let obj = this.newObject(before);
      let observer = this.newObserver(obj, '[]', '@each', 'length', 'firstObject', 'lastObject');

      obj.getProperties('firstObject', 'lastObject'); /* Prime the cache */

      obj.unshiftObjects(items);

      this.assert.deepEqual(this.toArray(obj), after, 'post item results');
      this.assert.equal((0, _metal.get)(obj, 'length'), after.length, 'length');

      this.assert.equal(observer.timesCalled('[]'), 1, 'should have notified [] once');
      this.assert.equal(observer.timesCalled('@each'), 0, 'should not have notified @each once');
      this.assert.equal(observer.timesCalled('length'), 1, 'should have notified length once');

      this.assert.equal(observer.validate('firstObject'), true, 'should NOT have notified firstObject');
      this.assert.equal(observer.validate('lastObject'), false, 'should NOT have notified lastObject');
    }
  }

  (0, _array.runArrayTests)('unshiftObjects', UnshiftObjectsTests, 'MutableArray', 'NativeArray', 'ArrayProxy');
});
enifed('@ember/-internals/runtime/tests/system/array_proxy/arranged_content_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_runloop, _metal, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  let array;

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - arrangedContent', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      (0, _runloop.run)(() => {
        array = _array_proxy.default.extend({
          arrangedContent: (0, _metal.computed)('content.[]', function () {
            let content = this.get('content');
            return content && (0, _array.A)(content.slice().sort((a, b) => {
              if (a == null) {
                a = -1;
              }
              if (b == null) {
                b = -1;
              }
              return b - a;
            }));
          })
        }).create({
          content: (0, _array.A)([1, 2, 4, 5])
        });
      });
    }

    afterEach() {
      (0, _runloop.run)(() => array.destroy());
    }

    ['@test compact - returns arrangedContent without nulls and undefined'](assert) {
      (0, _runloop.run)(() => array.set('content', (0, _array.A)([1, 3, null, 2, undefined])));

      assert.deepEqual(array.compact(), [3, 2, 1]);
    }

    ['@test indexOf - returns index of object in arrangedContent'](assert) {
      assert.equal(array.indexOf(4), 1, 'returns arranged index');
    }

    ['@test lastIndexOf - returns last index of object in arrangedContent'](assert) {
      array.get('content').pushObject(4);
      assert.equal(array.lastIndexOf(4), 2, 'returns last arranged index');
    }

    ['@test objectAt - returns object at index in arrangedContent'](assert) {
      assert.equal((0, _metal.objectAt)(array, 1), 4, 'returns object at index');
    }

    // Not sure if we need a specific test for it, since it's internal
    ['@test objectAtContent - returns object at index in arrangedContent'](assert) {
      assert.equal(array.objectAtContent(1), 4, 'returns object at index');
    }

    ['@test objectsAt - returns objects at indices in arrangedContent'](assert) {
      assert.deepEqual(array.objectsAt([0, 2, 4]), [5, 2, undefined], 'returns objects at indices');
    }

    ['@test replace - mutating an arranged ArrayProxy is not allowed']() {
      expectAssertion(() => {
        array.replace(0, 0, [3]);
      }, /Mutating an arranged ArrayProxy is not allowed/);
    }

    ['@test replaceContent - does a standard array replace on content'](assert) {
      (0, _runloop.run)(() => array.replaceContent(1, 2, [3]));
      assert.deepEqual(array.get('content'), [1, 3, 5]);
    }

    ['@test slice - returns a slice of the arrangedContent'](assert) {
      assert.deepEqual(array.slice(1, 3), [4, 2], 'returns sliced arrangedContent');
    }

    ['@test toArray - returns copy of arrangedContent'](assert) {
      assert.deepEqual(array.toArray(), [5, 4, 2, 1]);
    }

    ['@test without - returns arrangedContent without object'](assert) {
      assert.deepEqual(array.without(2), [5, 4, 1], 'returns arranged without object');
    }

    ['@test lastObject - returns last arranged object'](assert) {
      assert.equal(array.get('lastObject'), 1, 'returns last arranged object');
    }

    ['@test firstObject - returns first arranged object'](assert) {
      assert.equal(array.get('firstObject'), 5, 'returns first arranged object');
    }
  });

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - arrangedContent matching content', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      (0, _runloop.run)(function () {
        array = _array_proxy.default.create({
          content: (0, _array.A)([1, 2, 4, 5])
        });
      });
    }

    afterEach() {
      (0, _runloop.run)(function () {
        array.destroy();
      });
    }

    ['@test insertAt - inserts object at specified index'](assert) {
      (0, _runloop.run)(function () {
        array.insertAt(2, 3);
      });
      assert.deepEqual(array.get('content'), [1, 2, 3, 4, 5]);
    }

    ['@test replace - does a standard array replace'](assert) {
      (0, _runloop.run)(function () {
        array.replace(1, 2, [3]);
      });
      assert.deepEqual(array.get('content'), [1, 3, 5]);
    }

    ['@test reverseObjects - reverses content'](assert) {
      (0, _runloop.run)(function () {
        array.reverseObjects();
      });
      assert.deepEqual(array.get('content'), [5, 4, 2, 1]);
    }
  });

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - arrangedContent with transforms', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      (0, _runloop.run)(function () {
        array = _array_proxy.default.extend({
          arrangedContent: (0, _metal.computed)(function () {
            let content = this.get('content');
            return content && (0, _array.A)(content.slice().sort(function (a, b) {
              if (a == null) {
                a = -1;
              }
              if (b == null) {
                b = -1;
              }
              return b - a;
            }));
          }).property('content.[]'),

          objectAtContent(idx) {
            let obj = (0, _metal.objectAt)(this.get('arrangedContent'), idx);
            return obj && obj.toString();
          }
        }).create({
          content: (0, _array.A)([1, 2, 4, 5])
        });
      });
    }

    afterEach() {
      (0, _runloop.run)(function () {
        array.destroy();
      });
    }

    ['@test indexOf - returns index of object in arrangedContent'](assert) {
      assert.equal(array.indexOf('4'), 1, 'returns arranged index');
    }

    ['@test lastIndexOf - returns last index of object in arrangedContent'](assert) {
      array.get('content').pushObject(4);
      assert.equal(array.lastIndexOf('4'), 2, 'returns last arranged index');
    }

    ['@test objectAt - returns object at index in arrangedContent'](assert) {
      assert.equal((0, _metal.objectAt)(array, 1), '4', 'returns object at index');
    }

    // Not sure if we need a specific test for it, since it's internal
    ['@test objectAtContent - returns object at index in arrangedContent'](assert) {
      assert.equal(array.objectAtContent(1), '4', 'returns object at index');
    }

    ['@test objectsAt - returns objects at indices in arrangedContent'](assert) {
      assert.deepEqual(array.objectsAt([0, 2, 4]), ['5', '2', undefined], 'returns objects at indices');
    }

    ['@test slice - returns a slice of the arrangedContent'](assert) {
      assert.deepEqual(array.slice(1, 3), ['4', '2'], 'returns sliced arrangedContent');
    }

    ['@test toArray - returns copy of arrangedContent'](assert) {
      assert.deepEqual(array.toArray(), ['5', '4', '2', '1']);
    }

    ['@test without - returns arrangedContent without object'](assert) {
      assert.deepEqual(array.without('2'), ['5', '4', '1'], 'returns arranged without object');
    }

    ['@test lastObject - returns last arranged object'](assert) {
      assert.equal(array.get('lastObject'), '1', 'returns last arranged object');
    }

    ['@test firstObject - returns first arranged object'](assert) {
      assert.equal(array.get('firstObject'), '5', 'returns first arranged object');
    }
  });

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - with transforms', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      (0, _runloop.run)(function () {
        array = _array_proxy.default.extend({
          objectAtContent(idx) {
            let obj = (0, _metal.objectAt)(this.get('arrangedContent'), idx);
            return obj && obj.toString();
          }
        }).create({
          content: (0, _array.A)([1, 2, 4, 5])
        });
      });
    }

    afterEach() {
      (0, _runloop.run)(function () {
        array.destroy();
      });
    }

    ['@test popObject - removes last object in arrangedContent'](assert) {
      let popped = array.popObject();
      assert.equal(popped, '5', 'returns last object');
      assert.deepEqual(array.toArray(), ['1', '2', '4'], 'removes from content');
    }

    ['@test removeObject - removes object from content'](assert) {
      array.removeObject('2');
      assert.deepEqual(array.toArray(), ['1', '4', '5']);
    }

    ['@test removeObjects - removes objects from content'](assert) {
      array.removeObjects(['2', '4', '6']);
      assert.deepEqual(array.toArray(), ['1', '5']);
    }

    ['@test shiftObject - removes from start of arrangedContent'](assert) {
      let shifted = array.shiftObject();
      assert.equal(shifted, '1', 'returns first object');
      assert.deepEqual(array.toArray(), ['2', '4', '5'], 'removes object from content');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/array_proxy/array_observer_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_metal, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - array observers', class extends _internalTestHelpers.AbstractTestCase {
    ['@test mutating content'](assert) {
      assert.expect(4);

      let content = (0, _array.A)(['x', 'y', 'z']);
      let proxy = _array_proxy.default.create({ content });

      proxy.addArrayObserver({
        arrayWillChange(proxy, startIndex, removeCount, addCount) {
          assert.deepEqual([startIndex, removeCount, addCount], [1, 1, 3]);
          assert.deepEqual(proxy.toArray(), ['x', 'y', 'z']);
        },
        arrayDidChange(proxy, startIndex, removeCount, addCount) {
          assert.deepEqual([startIndex, removeCount, addCount], [1, 1, 3]);
          assert.deepEqual(proxy.toArray(), ['x', 'a', 'b', 'c', 'z']);
        }
      });

      proxy.toArray();
      content.replace(1, 1, ['a', 'b', 'c']);
    }

    ['@test assigning content'](assert) {
      assert.expect(4);

      let content = (0, _array.A)(['x', 'y', 'z']);
      let proxy = _array_proxy.default.create({ content });

      proxy.addArrayObserver({
        arrayWillChange(proxy, startIndex, removeCount, addCount) {
          assert.deepEqual([startIndex, removeCount, addCount], [0, 3, 5]);
          assert.deepEqual(proxy.toArray(), ['x', 'y', 'z']);
        },
        arrayDidChange(proxy, startIndex, removeCount, addCount) {
          assert.deepEqual([startIndex, removeCount, addCount], [0, 3, 5]);
          assert.deepEqual(proxy.toArray(), ['a', 'b', 'c', 'd', 'e']);
        }
      });

      proxy.toArray();
      (0, _metal.set)(proxy, 'content', (0, _array.A)(['a', 'b', 'c', 'd', 'e']));
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/array_proxy/content_change_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_runloop, _metal, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - content change', class extends _internalTestHelpers.AbstractTestCase {
    ["@test The ArrayProxy doesn't explode when assigned a destroyed object"](assert) {
      let proxy1 = _array_proxy.default.create();
      let proxy2 = _array_proxy.default.create();

      (0, _runloop.run)(() => proxy1.destroy());

      (0, _metal.set)(proxy2, 'content', proxy1);

      assert.ok(true, 'No exception was raised');
    }

    ['@test should update if content changes while change events are deferred'](assert) {
      let proxy = _array_proxy.default.create();

      assert.deepEqual(proxy.toArray(), []);

      (0, _metal.changeProperties)(() => {
        proxy.set('content', (0, _array.A)([1, 2, 3]));
        assert.deepEqual(proxy.toArray(), [1, 2, 3]);
      });
    }

    ['@test objectAt recomputes the object cache correctly'](assert) {
      let indexes = [];

      let proxy = _array_proxy.default.extend({
        objectAtContent(index) {
          indexes.push(index);
          return this.content[index];
        }
      }).create({
        content: (0, _array.A)([1, 2, 3, 4, 5])
      });

      assert.deepEqual(indexes, []);
      assert.deepEqual(proxy.objectAt(0), 1);
      assert.deepEqual(indexes, [0, 1, 2, 3, 4]);

      indexes.length = 0;
      proxy.set('content', (0, _array.A)([1, 2, 3]));
      assert.deepEqual(proxy.objectAt(0), 1);
      assert.deepEqual(indexes, [0, 1, 2]);

      indexes.length = 0;
      proxy.content.replace(2, 0, [4, 5]);
      assert.deepEqual(proxy.objectAt(0), 1);
      assert.deepEqual(proxy.objectAt(1), 2);
      assert.deepEqual(indexes, []);
      assert.deepEqual(proxy.objectAt(2), 4);
      assert.deepEqual(indexes, [2, 3, 4]);
    }

    ['@test negative indexes are handled correctly'](assert) {
      let indexes = [];

      let proxy = _array_proxy.default.extend({
        objectAtContent(index) {
          indexes.push(index);
          return this.content[index];
        }
      }).create({
        content: (0, _array.A)([1, 2, 3, 4, 5])
      });

      assert.deepEqual(proxy.toArray(), [1, 2, 3, 4, 5]);

      indexes.length = 0;

      proxy.content.replace(-1, 0, [7]);
      proxy.content.replace(-2, 0, [6]);

      assert.deepEqual(proxy.toArray(), [1, 2, 3, 4, 6, 7, 5]);
      assert.deepEqual(indexes, [4, 5, 6]);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/array_proxy/length_test', ['@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/metal', '@ember/object/computed', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_array_proxy, _object, _metal, _computed, _array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.ArrayProxy - content change (length)', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should update length for null content'](assert) {
      let proxy = _array_proxy.default.create({
        content: (0, _array.A)([1, 2, 3])
      });

      assert.equal(proxy.get('length'), 3, 'precond - length is 3');

      proxy.set('content', null);

      assert.equal(proxy.get('length'), 0, 'length updates');
    }

    ['@test should update length for null content when there is a computed property watching length'](assert) {
      let proxy = _array_proxy.default.extend({
        isEmpty: (0, _computed.not)('length')
      }).create({
        content: (0, _array.A)([1, 2, 3])
      });

      assert.equal(proxy.get('length'), 3, 'precond - length is 3');

      // Consume computed property that depends on length
      proxy.get('isEmpty');

      // update content
      proxy.set('content', null);

      assert.equal(proxy.get('length'), 0, 'length updates');
    }

    ['@test getting length does not recompute the object cache'](assert) {
      let indexes = [];

      let proxy = _array_proxy.default.extend({
        objectAtContent(index) {
          indexes.push(index);
          return this.content[index];
        }
      }).create({
        content: (0, _array.A)([1, 2, 3, 4, 5])
      });

      assert.equal((0, _metal.get)(proxy, 'length'), 5);
      assert.deepEqual(indexes, []);

      indexes.length = 0;
      proxy.set('content', (0, _array.A)([6, 7, 8]));
      assert.equal((0, _metal.get)(proxy, 'length'), 3);
      assert.deepEqual(indexes, []);

      indexes.length = 0;
      proxy.content.replace(1, 0, [1, 2, 3]);
      assert.equal((0, _metal.get)(proxy, 'length'), 6);
      assert.deepEqual(indexes, []);
    }

    '@test accessing length after content set to null'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'content', null);

      assert.equal(obj.length, 0, 'length is 0 without content');
      assert.deepEqual(obj.content, null, 'content was updated');
    }

    '@test accessing length after content set to null in willDestroy'(assert) {
      let obj = _array_proxy.default.extend({
        willDestroy() {
          this.set('content', null);
          this._super(...arguments);
        }
      }).create({
        content: ['foo', 'bar']
      });

      assert.equal(obj.length, 2, 'precond');

      this.runTask(() => obj.destroy());

      assert.equal(obj.length, 0, 'length is 0 without content');
      assert.deepEqual(obj.content, null, 'content was updated');
    }

    '@test setting length to 0'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'length', 0);

      assert.equal(obj.length, 0, 'length was updated');
      assert.deepEqual(obj.content, [], 'content length was truncated');
    }

    '@test setting length to smaller value'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'length', 1);

      assert.equal(obj.length, 1, 'length was updated');
      assert.deepEqual(obj.content, ['foo'], 'content length was truncated');
    }

    '@test setting length to larger value'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'length', 3);

      assert.equal(obj.length, 3, 'length was updated');
      assert.deepEqual(obj.content, ['foo', 'bar', undefined], 'content length was updated');
    }

    '@test setting length after content set to null'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'content', null);
      assert.equal(obj.length, 0, 'length was updated');

      (0, _metal.set)(obj, 'length', 0);
      assert.equal(obj.length, 0, 'length is still updated');
    }

    '@test setting length to greater than zero'(assert) {
      let obj = _array_proxy.default.create({ content: ['foo', 'bar'] });

      assert.equal(obj.length, 2, 'precond');

      (0, _metal.set)(obj, 'length', 1);

      assert.equal(obj.length, 1, 'length was updated');
      assert.deepEqual(obj.content, ['foo'], 'content length was truncated');
    }

    ['@test array proxy + aliasedProperty complex test'](assert) {
      let aCalled, bCalled, cCalled, dCalled, eCalled;

      aCalled = bCalled = cCalled = dCalled = eCalled = 0;

      let obj = _object.default.extend({
        colors: (0, _computed.oneWay)('model'),
        length: (0, _computed.oneWay)('colors.length'),

        a: (0, _metal.observer)('length', () => aCalled++),
        b: (0, _metal.observer)('colors.length', () => bCalled++),
        c: (0, _metal.observer)('colors.content.length', () => cCalled++),
        d: (0, _metal.observer)('colors.[]', () => dCalled++),
        e: (0, _metal.observer)('colors.content.[]', () => eCalled++)
      }).create();

      obj.set('model', _array_proxy.default.create({
        content: (0, _array.A)(['red', 'yellow', 'blue'])
      }));

      assert.equal(obj.get('colors.content.length'), 3);
      assert.equal(obj.get('colors.length'), 3);
      assert.equal(obj.get('length'), 3);

      assert.equal(aCalled, 1, 'expected observer `length` to be called ONCE');
      assert.equal(bCalled, 1, 'expected observer `colors.length` to be called ONCE');
      assert.equal(cCalled, 1, 'expected observer `colors.content.length` to be called ONCE');
      assert.equal(dCalled, 1, 'expected observer `colors.[]` to be called ONCE');
      assert.equal(eCalled, 1, 'expected observer `colors.content.[]` to be called ONCE');

      obj.get('colors').pushObjects(['green', 'red']);

      assert.equal(obj.get('colors.content.length'), 5);
      assert.equal(obj.get('colors.length'), 5);
      assert.equal(obj.get('length'), 5);

      assert.equal(aCalled, 2, 'expected observer `length` to be called TWICE');
      assert.equal(bCalled, 2, 'expected observer `colors.length` to be called TWICE');
      assert.equal(cCalled, 2, 'expected observer `colors.content.length` to be called TWICE');
      assert.equal(dCalled, 2, 'expected observer `colors.[]` to be called TWICE');
      assert.equal(eCalled, 2, 'expected observer `colors.content.[]` to be called TWICE');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/array_proxy/watching_and_listening_test', ['@ember/-internals/meta', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/array_proxy', '@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_meta, _metal, _array_proxy, _array, _internalTestHelpers) {
  'use strict';

  function sortedListenersFor(obj, eventName) {
    let listeners = (0, _meta.peekMeta)(obj).matchingListeners(eventName) || [];

    let keys = [];
    for (let i = 0; i < listeners.length; i += 3) {
      keys.push(listeners[i + 1]);
    }
    return keys.sort();
  }

  (0, _internalTestHelpers.moduleFor)('ArrayProxy - watching and listening', class extends _internalTestHelpers.AbstractTestCase {
    [`@test setting 'content' adds listeners correctly`](assert) {
      let content = (0, _array.A)();
      let proxy = _array_proxy.default.create();

      assert.deepEqual(sortedListenersFor(content, '@array:before'), []);
      assert.deepEqual(sortedListenersFor(content, '@array:change'), []);

      proxy.set('content', content);

      assert.deepEqual(sortedListenersFor(content, '@array:before'), ['_arrangedContentArrayWillChange']);
      assert.deepEqual(sortedListenersFor(content, '@array:change'), ['_arrangedContentArrayDidChange']);
    }

    [`@test changing 'content' adds and removes listeners correctly`](assert) {
      let content1 = (0, _array.A)();
      let content2 = (0, _array.A)();
      let proxy = _array_proxy.default.create({ content: content1 });

      assert.deepEqual(sortedListenersFor(content1, '@array:before'), ['_arrangedContentArrayWillChange']);
      assert.deepEqual(sortedListenersFor(content1, '@array:change'), ['_arrangedContentArrayDidChange']);

      proxy.set('content', content2);

      assert.deepEqual(sortedListenersFor(content1, '@array:before'), []);
      assert.deepEqual(sortedListenersFor(content1, '@array:change'), []);
      assert.deepEqual(sortedListenersFor(content2, '@array:before'), ['_arrangedContentArrayWillChange']);
      assert.deepEqual(sortedListenersFor(content2, '@array:change'), ['_arrangedContentArrayDidChange']);
    }

    [`@test regression test for https://github.com/emberjs/ember.js/issues/12475`](assert) {
      let item1a = { id: 1 };
      let item1b = { id: 2 };
      let item1c = { id: 3 };
      let content1 = (0, _array.A)([item1a, item1b, item1c]);

      let proxy = _array_proxy.default.create({ content: content1 });
      let obj = { proxy };

      (0, _metal.defineProperty)(obj, 'ids', (0, _metal.computed)('proxy.@each.id', function () {
        return (0, _metal.get)(this, 'proxy').mapBy('id');
      }));

      // These manually added observers are to simulate the observers added by the
      // rendering process in a template like:
      //
      // {{#each items as |item|}}
      //   {{item.id}}
      // {{/each}}
      (0, _metal.addObserver)(item1a, 'id', function () {});
      (0, _metal.addObserver)(item1b, 'id', function () {});
      (0, _metal.addObserver)(item1c, 'id', function () {});

      // The EachProxy has not yet been consumed. Only the manually added
      // observers are watching.
      assert.equal((0, _metal.watcherCount)(item1a, 'id'), 1);
      assert.equal((0, _metal.watcherCount)(item1b, 'id'), 1);
      assert.equal((0, _metal.watcherCount)(item1c, 'id'), 1);

      // Consume the each proxy. This causes the EachProxy to add two observers
      // per item: one for "before" events and one for "after" events.
      assert.deepEqual((0, _metal.get)(obj, 'ids'), [1, 2, 3]);

      // For each item, the two each proxy observers and one manual added observer
      // are watching.
      assert.equal((0, _metal.watcherCount)(item1a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1c, 'id'), 2);

      // This should be a no-op because observers do not fire if the value
      // 1. is an object and 2. is the same as the old value.
      proxy.set('content', content1);

      assert.equal((0, _metal.watcherCount)(item1a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1c, 'id'), 2);

      // This is repeated to catch the regression. It should still be a no-op.
      proxy.set('content', content1);

      assert.equal((0, _metal.watcherCount)(item1a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item1c, 'id'), 2);

      // Set the content to a new array with completely different items and
      // repeat the process.
      let item2a = { id: 4 };
      let item2b = { id: 5 };
      let item2c = { id: 6 };
      let content2 = (0, _array.A)([item2a, item2b, item2c]);

      (0, _metal.addObserver)(item2a, 'id', function () {});
      (0, _metal.addObserver)(item2b, 'id', function () {});
      (0, _metal.addObserver)(item2c, 'id', function () {});

      proxy.set('content', content2);

      assert.deepEqual((0, _metal.get)(obj, 'ids'), [4, 5, 6]);

      assert.equal((0, _metal.watcherCount)(item2a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2c, 'id'), 2);

      // Ensure that the observers added by the EachProxy on all items in the
      // first content array have been torn down.
      assert.equal((0, _metal.watcherCount)(item1a, 'id'), 1);
      assert.equal((0, _metal.watcherCount)(item1b, 'id'), 1);
      assert.equal((0, _metal.watcherCount)(item1c, 'id'), 1);

      proxy.set('content', content2);

      assert.equal((0, _metal.watcherCount)(item2a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2c, 'id'), 2);

      proxy.set('content', content2);

      assert.equal((0, _metal.watcherCount)(item2a, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2b, 'id'), 2);
      assert.equal((0, _metal.watcherCount)(item2c, 'id'), 2);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/core_object_test', ['@ember/-internals/owner', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/core_object', 'internal-test-helpers'], function (_owner, _metal, _core_object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.CoreObject', class extends _internalTestHelpers.AbstractTestCase {
    ['@test throws deprecation with new (one arg)']() {
      expectDeprecation(() => {
        new _core_object.default({
          firstName: 'Stef',
          lastName: 'Penner'
        });
      }, /using `new` with EmberObject has been deprecated/);
    }

    ['@test throws deprecation with new (> 1 arg)']() {
      expectDeprecation(() => {
        new _core_object.default({
          firstName: 'Stef',
          lastName: 'Penner'
        }, {
          other: 'name'
        });
      }, /using `new` with EmberObject has been deprecated/);
    }

    ['@test toString should be not be added as a property when calling toString()'](assert) {
      let obj = _core_object.default.create({
        firstName: 'Foo',
        lastName: 'Bar'
      });

      obj.toString();

      assert.notOk(obj.hasOwnProperty('toString'), 'Calling toString() should not create a toString class property');
    }

    ['@test should not trigger proxy assertion when retrieving a proxy with (GH#16263)'](assert) {
      let someProxyishThing = _core_object.default.extend({
        unknownProperty() {
          return true;
        }
      }).create();

      let obj = _core_object.default.create({
        someProxyishThing
      });

      let proxy = (0, _metal.get)(obj, 'someProxyishThing');
      assert.equal((0, _metal.get)(proxy, 'lolol'), true, 'should be able to get data from a proxy');
    }

    ['@test should not trigger proxy assertion when retrieving a re-registered proxy (GH#16610)'](assert) {
      let owner = (0, _internalTestHelpers.buildOwner)();

      let someProxyishThing = _core_object.default.extend({
        unknownProperty() {
          return true;
        }
      }).create();

      // emulates ember-engines's process of registering services provided
      // by the host app down to the engine
      owner.register('thing:one', someProxyishThing, { instantiate: false });

      assert.equal(owner.lookup('thing:one'), someProxyishThing);
    }

    ['@test should not trigger proxy assertion when probing for a "symbol"'](assert) {
      let proxy = _core_object.default.extend({
        unknownProperty() {
          return true;
        }
      }).create();

      assert.equal((0, _metal.get)(proxy, 'lolol'), true, 'should be able to get data from a proxy');

      // should not trigger an assertion
      (0, _owner.getOwner)(proxy);
    }

    ['@test can use getOwner in a proxy init GH#16484'](assert) {
      let owner = {};
      let options = {};
      (0, _owner.setOwner)(options, owner);

      _core_object.default.extend({
        init() {
          this._super(...arguments);
          let localOwner = (0, _owner.getOwner)(this);

          assert.equal(localOwner, owner, 'should be able to `getOwner` in init');
        },
        unknownProperty() {
          return undefined;
        }
      }).create(options);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/namespace/base_test', ['@ember/-internals/environment', '@ember/runloop', '@ember/-internals/metal', '@ember/-internals/utils', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/system/namespace', 'internal-test-helpers'], function (_environment, _runloop, _metal, _utils, _object, _namespace, _internalTestHelpers) {
  'use strict';

  const originalLookup = _environment.context.lookup;
  let lookup;

  (0, _internalTestHelpers.moduleFor)('Namespace', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      (0, _metal.setNamespaceSearchDisabled)(false);

      lookup = _environment.context.lookup = {};
    }

    afterEach() {
      (0, _metal.setNamespaceSearchDisabled)(false);

      for (let prop in lookup) {
        if (lookup[prop]) {
          (0, _runloop.run)(lookup[prop], 'destroy');
        }
      }

      _environment.context.lookup = originalLookup;
    }

    ['@test Namespace should be a subclass of EmberObject'](assert) {
      assert.ok(_object.default.detect(_namespace.default));
    }

    ['@test Namespace should be duck typed'](assert) {
      let namespace = _namespace.default.create();
      try {
        assert.ok((0, _metal.get)(namespace, 'isNamespace'), 'isNamespace property is true');
      } finally {
        (0, _runloop.run)(namespace, 'destroy');
      }
    }

    ['@test Namespace is found and named'](assert) {
      let nsA = lookup.NamespaceA = _namespace.default.create();
      assert.equal(nsA.toString(), 'NamespaceA', 'namespaces should have a name if they are on lookup');

      let nsB = lookup.NamespaceB = _namespace.default.create();
      assert.equal(nsB.toString(), 'NamespaceB', 'namespaces work if created after the first namespace processing pass');
    }

    ['@test Classes under an Namespace are properly named'](assert) {
      let nsA = lookup.NamespaceA = _namespace.default.create();
      nsA.Foo = _object.default.extend();
      assert.equal(nsA.Foo.toString(), 'NamespaceA.Foo', 'Classes pick up their parent namespace');

      nsA.Bar = _object.default.extend();
      assert.equal(nsA.Bar.toString(), 'NamespaceA.Bar', 'New Classes get the naming treatment too');

      let nsB = lookup.NamespaceB = _namespace.default.create();
      nsB.Foo = _object.default.extend();
      assert.equal(nsB.Foo.toString(), 'NamespaceB.Foo', 'Classes in new namespaces get the naming treatment');
    }

    //test("Classes under Ember are properly named", function() {
    //  // ES6TODO: This test does not work reliably when running independent package build with Broccoli config.
    //  Ember.TestObject = EmberObject.extend({});
    //  equal(Ember.TestObject.toString(), "Ember.TestObject", "class under Ember is given a string representation");
    //});

    ['@test Lowercase namespaces are no longer supported'](assert) {
      let nsC = lookup.namespaceC = _namespace.default.create();
      assert.equal(nsC.toString(), (0, _utils.guidFor)(nsC));
    }

    ['@test A namespace can be assigned a custom name'](assert) {
      let nsA = _namespace.default.create({
        name: 'NamespaceA'
      });

      try {
        let nsB = lookup.NamespaceB = _namespace.default.create({
          name: 'CustomNamespaceB'
        });

        nsA.Foo = _object.default.extend();
        nsB.Foo = _object.default.extend();

        assert.equal(nsA.Foo.toString(), 'NamespaceA.Foo', "The namespace's name is used when the namespace is not in the lookup object");
        assert.equal(nsB.Foo.toString(), 'CustomNamespaceB.Foo', "The namespace's name is used when the namespace is in the lookup object");
      } finally {
        (0, _runloop.run)(nsA, 'destroy');
      }
    }

    ['@test Calling namespace.nameClasses() eagerly names all classes'](assert) {
      (0, _metal.setNamespaceSearchDisabled)(true);

      let namespace = lookup.NS = _namespace.default.create();

      namespace.ClassA = _object.default.extend();
      namespace.ClassB = _object.default.extend();

      _namespace.default.processAll();

      assert.equal(namespace.ClassA.toString(), 'NS.ClassA');
      assert.equal(namespace.ClassB.toString(), 'NS.ClassB');
    }

    ['@test A namespace can be looked up by its name'](assert) {
      let NS = lookup.NS = _namespace.default.create();
      let UI = lookup.UI = _namespace.default.create();
      let CF = lookup.CF = _namespace.default.create();

      assert.equal(_namespace.default.byName('NS'), NS);
      assert.equal(_namespace.default.byName('UI'), UI);
      assert.equal(_namespace.default.byName('CF'), CF);
    }

    ['@test A nested namespace can be looked up by its name'](assert) {
      let UI = lookup.UI = _namespace.default.create();
      UI.Nav = _namespace.default.create();

      assert.equal(_namespace.default.byName('UI.Nav'), UI.Nav);

      (0, _runloop.run)(UI.Nav, 'destroy');
    }

    ['@test Destroying a namespace before caching lookup removes it from the list of namespaces'](assert) {
      let CF = lookup.CF = _namespace.default.create();

      (0, _runloop.run)(CF, 'destroy');
      assert.equal(_namespace.default.byName('CF'), undefined, 'namespace can not be found after destroyed');
    }

    ['@test Destroying a namespace after looking up removes it from the list of namespaces'](assert) {
      let CF = lookup.CF = _namespace.default.create();

      assert.equal(_namespace.default.byName('CF'), CF, 'precondition - namespace can be looked up by name');

      (0, _runloop.run)(CF, 'destroy');
      assert.equal(_namespace.default.byName('CF'), undefined, 'namespace can not be found after destroyed');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/native_array/a_test', ['@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.A', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.A'](assert) {
      assert.deepEqual((0, _array.A)([1, 2]), [1, 2], 'array values were not be modified');
      assert.deepEqual((0, _array.A)(), [], 'returned an array with no arguments');
      assert.deepEqual((0, _array.A)(null), [], 'returned an array with a null argument');
      assert.ok(_array.default.detect((0, _array.A)()), 'returned an ember array');
      assert.ok(_array.default.detect((0, _array.A)([1, 2])), 'returned an ember array');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/native_array/copyable_suite_test', ['@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('NativeArray Copyable', class extends _internalTestHelpers.AbstractTestCase {
    ['@test deep copy is respected'](assert) {
      let array = (0, _array.A)([{ id: 1 }, { id: 2 }, { id: 3 }]);
      let copiedArray;

      expectDeprecation(() => {
        copiedArray = array.copy(true);
      }, `Using \`NativeArray#copy\` is deprecated`);

      assert.deepEqual(copiedArray, array, 'copied array is equivalent');
      assert.ok(copiedArray[0] !== array[0], 'objects inside should be unique');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/native_array/replace_test', ['@ember/-internals/runtime/lib/mixins/array', 'internal-test-helpers'], function (_array, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('NativeArray.replace', class extends _internalTestHelpers.AbstractTestCase {
    ['@test raises assertion if third argument is not an array']() {
      expectAssertion(function () {
        (0, _array.A)([1, 2, 3]).replace(1, 1, '');
      }, 'The third argument to replace needs to be an array.');
    }

    ['@test it does not raise an assertion if third parameter is not passed'](assert) {
      assert.deepEqual((0, _array.A)([1, 2, 3]).replace(1, 2), (0, _array.A)([1]), 'no assertion raised');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/computed_test', ['@ember/-internals/metal', '@ember/object/computed', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _computed, _object, _internalTestHelpers) {
  'use strict';

  function K() {
    return this;
  }

  function testWithDefault(assert, expect, x, y, z) {
    assert.equal((0, _metal.get)(x, y), expect);
    assert.equal((0, _metal.getWithDefault)(x, y, z), expect);
    assert.equal(x.getWithDefault(y, z), expect);
  }

  (0, _internalTestHelpers.moduleFor)('EmberObject computed property', class extends _internalTestHelpers.AbstractTestCase {
    ['@test computed property on instance'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      testWithDefault(assert, 'FOO', MyClass.create(), 'foo');
    }

    ['@test computed property on subclass'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      let Subclass = MyClass.extend({
        foo: (0, _metal.computed)(function () {
          return 'BAR';
        })
      });

      testWithDefault(assert, 'BAR', Subclass.create(), 'foo');
    }

    ['@test replacing computed property with regular val'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      let Subclass = MyClass.extend({
        foo: 'BAR'
      });

      testWithDefault(assert, 'BAR', Subclass.create(), 'foo');
    }

    ['@test complex depndent keys'](assert) {
      let MyClass = _object.default.extend({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'bar', { baz: 'BIFF' });
        },

        count: 0,

        foo: (0, _metal.computed)(function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
          return (0, _metal.get)((0, _metal.get)(this, 'bar'), 'baz') + ' ' + (0, _metal.get)(this, 'count');
        }).property('bar.baz')
      });

      let Subclass = MyClass.extend({
        count: 20
      });

      let obj1 = MyClass.create();
      let obj2 = Subclass.create();

      testWithDefault(assert, 'BIFF 1', obj1, 'foo');
      testWithDefault(assert, 'BIFF 21', obj2, 'foo');

      (0, _metal.set)((0, _metal.get)(obj1, 'bar'), 'baz', 'BLARG');

      testWithDefault(assert, 'BLARG 2', obj1, 'foo');
      testWithDefault(assert, 'BIFF 21', obj2, 'foo');

      (0, _metal.set)((0, _metal.get)(obj2, 'bar'), 'baz', 'BOOM');

      testWithDefault(assert, 'BLARG 2', obj1, 'foo');
      testWithDefault(assert, 'BOOM 22', obj2, 'foo');
    }

    ['@test complex dependent keys changing complex dependent keys'](assert) {
      let MyClass = _object.default.extend({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'bar', { baz: 'BIFF' });
        },

        count: 0,

        foo: (0, _metal.computed)(function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
          return (0, _metal.get)((0, _metal.get)(this, 'bar'), 'baz') + ' ' + (0, _metal.get)(this, 'count');
        }).property('bar.baz')
      });

      let Subclass = MyClass.extend({
        init() {
          this._super(...arguments);
          (0, _metal.set)(this, 'bar2', { baz: 'BIFF2' });
        },

        count: 0,

        foo: (0, _metal.computed)(function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
          return (0, _metal.get)((0, _metal.get)(this, 'bar2'), 'baz') + ' ' + (0, _metal.get)(this, 'count');
        }).property('bar2.baz')
      });

      let obj2 = Subclass.create();

      testWithDefault(assert, 'BIFF2 1', obj2, 'foo');

      (0, _metal.set)((0, _metal.get)(obj2, 'bar'), 'baz', 'BLARG');
      testWithDefault(assert, 'BIFF2 1', obj2, 'foo'); // should not invalidate property

      (0, _metal.set)((0, _metal.get)(obj2, 'bar2'), 'baz', 'BLARG');
      testWithDefault(assert, 'BLARG 2', obj2, 'foo'); // should not invalidate property
    }

    ['@test can retrieve metadata for a computed property'](assert) {
      let MyClass = _object.default.extend({
        computedProperty: (0, _metal.computed)(function () {}).meta({ key: 'keyValue' })
      });

      assert.equal((0, _metal.get)(MyClass.metaForProperty('computedProperty'), 'key'), 'keyValue', 'metadata saved on the computed property can be retrieved');

      let ClassWithNoMetadata = _object.default.extend({
        computedProperty: (0, _metal.computed)(function () {}).volatile(),

        staticProperty: 12
      });

      assert.equal(typeof ClassWithNoMetadata.metaForProperty('computedProperty'), 'object', 'returns empty hash if no metadata has been saved');

      expectAssertion(function () {
        ClassWithNoMetadata.metaForProperty('nonexistentProperty');
      }, "metaForProperty() could not find a computed property with key 'nonexistentProperty'.");

      expectAssertion(function () {
        ClassWithNoMetadata.metaForProperty('staticProperty');
      }, "metaForProperty() could not find a computed property with key 'staticProperty'.");
    }

    ['@test overriding a computed property with null removes it from eachComputedProperty iteration'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {}),

        fooDidChange: (0, _metal.observer)('foo', function () {}),

        bar: (0, _metal.computed)(function () {})
      });

      let SubClass = MyClass.extend({
        foo: null
      });

      let list = [];

      SubClass.eachComputedProperty(name => list.push(name));

      assert.deepEqual(list.sort(), ['bar'], 'overridding with null removes from eachComputedProperty listing');
    }

    ['@test can iterate over a list of computed properties for a class'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {}),

        fooDidChange: (0, _metal.observer)('foo', function () {}),

        bar: (0, _metal.computed)(function () {}),

        qux: (0, _metal.alias)('foo')
      });

      let SubClass = MyClass.extend({
        baz: (0, _metal.computed)(function () {})
      });

      SubClass.reopen({
        bat: (0, _metal.computed)(function () {}).meta({ iAmBat: true })
      });

      let list = [];

      MyClass.eachComputedProperty(function (name) {
        list.push(name);
      });

      assert.deepEqual(list.sort(), ['bar', 'foo', 'qux'], 'watched and unwatched computed properties are iterated');

      list = [];

      SubClass.eachComputedProperty(function (name, meta) {
        list.push(name);

        if (name === 'bat') {
          assert.deepEqual(meta, { iAmBat: true });
        } else {
          assert.deepEqual(meta, {});
        }
      });

      assert.deepEqual(list.sort(), ['bar', 'bat', 'baz', 'foo', 'qux'], 'all inherited properties are included');
    }

    ['@test list of properties updates when an additional property is added (such cache busting)'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(K),

        fooDidChange: (0, _metal.observer)('foo', function () {}),

        bar: (0, _metal.computed)(K)
      });

      let list = [];

      MyClass.eachComputedProperty(function (name) {
        list.push(name);
      });

      assert.deepEqual(list.sort(), ['bar', 'foo'].sort(), 'expected two computed properties');

      MyClass.reopen({
        baz: (0, _metal.computed)(K)
      });

      MyClass.create(); // force apply mixins

      list = [];

      MyClass.eachComputedProperty(function (name) {
        list.push(name);
      });

      assert.deepEqual(list.sort(), ['bar', 'foo', 'baz'].sort(), 'expected three computed properties');

      (0, _metal.defineProperty)(MyClass.prototype, 'qux', (0, _metal.computed)(K));

      list = [];

      MyClass.eachComputedProperty(function (name) {
        list.push(name);
      });

      assert.deepEqual(list.sort(), ['bar', 'foo', 'baz', 'qux'].sort(), 'expected four computed properties');
    }

    ['@test Calling _super in call outside the immediate function of a CP getter works'](assert) {
      function macro(callback) {
        return (0, _metal.computed)(function () {
          return callback.call(this);
        });
      }

      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      let SubClass = MyClass.extend({
        foo: macro(function () {
          return this._super();
        })
      });

      assert.ok((0, _metal.get)(SubClass.create(), 'foo'), 'FOO', 'super value is fetched');
    }

    ['@test Calling _super in apply outside the immediate function of a CP getter works'](assert) {
      function macro(callback) {
        return (0, _metal.computed)(function () {
          return callback.apply(this);
        });
      }

      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      let SubClass = MyClass.extend({
        foo: macro(function () {
          return this._super();
        })
      });

      assert.ok((0, _metal.get)(SubClass.create(), 'foo'), 'FOO', 'super value is fetched');
    }

    ['@test observing computed.reads prop and overriding it in create() works'](assert) {
      let Obj = _object.default.extend({
        name: (0, _computed.oneWay)('model.name'),
        nameDidChange: (0, _metal.observer)('name', function () {})
      });

      let obj1 = Obj.create({ name: '1' });
      let obj2 = Obj.create({ name: '2' });

      assert.equal(obj1.get('name'), '1');
      assert.equal(obj2.get('name'), '2');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/create_test', ['@ember/-internals/owner', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_owner, _metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('EmberObject.create', class extends _internalTestHelpers.AbstractTestCase {
    ['@test simple properties are set'](assert) {
      let o = _object.default.create({ ohai: 'there' });
      assert.equal(o.get('ohai'), 'there');
    }

    ['@test calls computed property setters'](assert) {
      let MyClass = _object.default.extend({
        foo: (0, _metal.computed)({
          get() {
            return "this is not the value you're looking for";
          },
          set(key, value) {
            return value;
          }
        })
      });

      let o = MyClass.create({ foo: 'bar' });
      assert.equal(o.get('foo'), 'bar');
    }

    ['@test sets up mandatory setters for watched simple properties'](assert) {
      if (true /* DEBUG */) {
          let MyClass = _object.default.extend({
            foo: null,
            bar: null,
            fooDidChange: (0, _metal.observer)('foo', function () {})
          });

          let o = MyClass.create({ foo: 'bar', bar: 'baz' });
          assert.equal(o.get('foo'), 'bar');

          let descriptor = Object.getOwnPropertyDescriptor(o, 'foo');
          assert.ok(descriptor.set, 'Mandatory setter was setup');

          descriptor = Object.getOwnPropertyDescriptor(o, 'bar');
          assert.ok(!descriptor.set, 'Mandatory setter was not setup');
        } else {
        assert.expect(0);
      }
    }

    ['@test calls setUnknownProperty if defined'](assert) {
      let setUnknownPropertyCalled = false;

      let MyClass = _object.default.extend({
        setUnknownProperty() /* key, value */{
          setUnknownPropertyCalled = true;
        }
      });

      MyClass.create({ foo: 'bar' });
      assert.ok(setUnknownPropertyCalled, 'setUnknownProperty was called');
    }

    ['@test throws if you try to define a computed property']() {
      expectAssertion(function () {
        _object.default.create({
          foo: (0, _metal.computed)(function () {})
        });
      }, 'EmberObject.create no longer supports defining computed properties. Define computed properties using extend() or reopen() before calling create().');
    }

    ['@test throws if you try to call _super in a method']() {
      expectAssertion(function () {
        _object.default.create({
          foo() {
            this._super(...arguments);
          }
        });
      }, 'EmberObject.create no longer supports defining methods that call _super.');
    }

    ["@test throws if you try to 'mixin' a definition"]() {
      let myMixin = _metal.Mixin.create({
        adder(arg1, arg2) {
          return arg1 + arg2;
        }
      });

      expectAssertion(function () {
        _object.default.create(myMixin);
      }, 'EmberObject.create no longer supports mixing in other definitions, use .extend & .create separately instead.');
    }

    ['@test inherits properties from passed in EmberObject'](assert) {
      let baseObj = _object.default.create({ foo: 'bar' });
      let secondaryObj = _object.default.create(baseObj);

      assert.equal(secondaryObj.foo, baseObj.foo, 'Em.O.create inherits properties from EmberObject parameter');
    }

    ['@test throws if you try to pass anything a string as a parameter']() {
      let expected = 'EmberObject.create only accepts objects.';

      expectAssertion(() => _object.default.create('some-string'), expected);
    }

    ['@test EmberObject.create can take undefined as a parameter'](assert) {
      let o = _object.default.create(undefined);
      assert.deepEqual(_object.default.create(), o);
    }

    ['@test can use getOwner in a proxy init GH#16484'](assert) {
      let owner = {};
      let options = {};
      (0, _owner.setOwner)(options, owner);

      _object.default.extend({
        init() {
          this._super(...arguments);
          let localOwner = (0, _owner.getOwner)(this);

          assert.equal(localOwner, owner, 'should be able to `getOwner` in init');
        },
        unknownProperty() {
          return undefined;
        }
      }).create(options);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/destroy_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/meta', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_runloop, _metal, _meta, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/runtime/system/object/destroy_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should schedule objects to be destroyed at the end of the run loop'](assert) {
      let obj = _object.default.create();
      let meta;

      (0, _runloop.run)(() => {
        obj.destroy();
        meta = (0, _meta.peekMeta)(obj);
        assert.ok(meta, 'meta is not destroyed immediately');
        assert.ok((0, _metal.get)(obj, 'isDestroying'), 'object is marked as destroying immediately');
        assert.ok(!(0, _metal.get)(obj, 'isDestroyed'), 'object is not destroyed immediately');
      });

      meta = (0, _meta.peekMeta)(obj);
      assert.ok((0, _metal.get)(obj, 'isDestroyed'), 'object is destroyed after run loop finishes');
    }

    // MANDATORY_SETTER moves value to meta.values
    // a destroyed object removes meta but leaves the accessor
    // that looks it up
    ['@test should raise an exception when modifying watched properties on a destroyed object'](assert) {
      if (true /* DEBUG */) {
          let obj = _object.default.extend({
            fooDidChange: (0, _metal.observer)('foo', function () {})
          }).create({
            foo: 'bar'
          });

          (0, _runloop.run)(() => obj.destroy());

          assert.throws(() => (0, _metal.set)(obj, 'foo', 'baz'), Error, 'raises an exception');
        } else {
        assert.expect(0);
      }
    }

    ['@test observers should not fire after an object has been destroyed'](assert) {
      let count = 0;
      let obj = _object.default.extend({
        fooDidChange: (0, _metal.observer)('foo', function () {
          count++;
        })
      }).create();

      obj.set('foo', 'bar');

      assert.equal(count, 1, 'observer was fired once');

      (0, _runloop.run)(() => {
        (0, _metal.beginPropertyChanges)();
        obj.set('foo', 'quux');
        obj.destroy();
        (0, _metal.endPropertyChanges)();
      });

      assert.equal(count, 1, 'observer was not called after object was destroyed');
    }

    ['@test destroyed objects should not see each others changes during teardown but a long lived object should'](assert) {
      let shouldChange = 0;
      let shouldNotChange = 0;

      let objs = {};

      let A = _object.default.extend({
        objs: objs,
        isAlive: true,
        willDestroy() {
          this.set('isAlive', false);
        },
        bDidChange: (0, _metal.observer)('objs.b.isAlive', function () {
          shouldNotChange++;
        }),
        cDidChange: (0, _metal.observer)('objs.c.isAlive', function () {
          shouldNotChange++;
        })
      });

      let B = _object.default.extend({
        objs: objs,
        isAlive: true,
        willDestroy() {
          this.set('isAlive', false);
        },
        aDidChange: (0, _metal.observer)('objs.a.isAlive', function () {
          shouldNotChange++;
        }),
        cDidChange: (0, _metal.observer)('objs.c.isAlive', function () {
          shouldNotChange++;
        })
      });

      let C = _object.default.extend({
        objs: objs,
        isAlive: true,
        willDestroy() {
          this.set('isAlive', false);
        },
        aDidChange: (0, _metal.observer)('objs.a.isAlive', function () {
          shouldNotChange++;
        }),
        bDidChange: (0, _metal.observer)('objs.b.isAlive', function () {
          shouldNotChange++;
        })
      });

      let LongLivedObject = _object.default.extend({
        objs: objs,
        isAliveDidChange: (0, _metal.observer)('objs.a.isAlive', function () {
          shouldChange++;
        })
      });

      objs.a = A.create();

      objs.b = B.create();

      objs.c = C.create();

      LongLivedObject.create();

      (0, _runloop.run)(() => {
        let keys = Object.keys(objs);
        for (let i = 0; i < keys.length; i++) {
          objs[keys[i]].destroy();
        }
      });

      assert.equal(shouldNotChange, 0, 'destroyed graph objs should not see change in willDestroy');
      assert.equal(shouldChange, 1, 'long lived should see change in willDestroy');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/detectInstance_test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/object/detectInstance', class extends _internalTestHelpers.AbstractTestCase {
    ['@test detectInstance detects instances correctly'](assert) {
      let A = _object.default.extend();
      let B = A.extend();
      let C = A.extend();

      let o = _object.default.create();
      let a = A.create();
      let b = B.create();
      let c = C.create();

      assert.ok(_object.default.detectInstance(o), 'o is an instance of EmberObject');
      assert.ok(_object.default.detectInstance(a), 'a is an instance of EmberObject');
      assert.ok(_object.default.detectInstance(b), 'b is an instance of EmberObject');
      assert.ok(_object.default.detectInstance(c), 'c is an instance of EmberObject');

      assert.ok(!A.detectInstance(o), 'o is not an instance of A');
      assert.ok(A.detectInstance(a), 'a is an instance of A');
      assert.ok(A.detectInstance(b), 'b is an instance of A');
      assert.ok(A.detectInstance(c), 'c is an instance of A');

      assert.ok(!B.detectInstance(o), 'o is not an instance of B');
      assert.ok(!B.detectInstance(a), 'a is not an instance of B');
      assert.ok(B.detectInstance(b), 'b is an instance of B');
      assert.ok(!B.detectInstance(c), 'c is not an instance of B');

      assert.ok(!C.detectInstance(o), 'o is not an instance of C');
      assert.ok(!C.detectInstance(a), 'a is not an instance of C');
      assert.ok(!C.detectInstance(b), 'b is not an instance of C');
      assert.ok(C.detectInstance(c), 'c is an instance of C');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/detect_test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/object/detect', class extends _internalTestHelpers.AbstractTestCase {
    ['@test detect detects classes correctly'](assert) {
      let A = _object.default.extend();
      let B = A.extend();
      let C = A.extend();

      assert.ok(_object.default.detect(_object.default), 'EmberObject is an EmberObject class');
      assert.ok(_object.default.detect(A), 'A is an EmberObject class');
      assert.ok(_object.default.detect(B), 'B is an EmberObject class');
      assert.ok(_object.default.detect(C), 'C is an EmberObject class');

      assert.ok(!A.detect(_object.default), 'EmberObject is not an A class');
      assert.ok(A.detect(A), 'A is an A class');
      assert.ok(A.detect(B), 'B is an A class');
      assert.ok(A.detect(C), 'C is an A class');

      assert.ok(!B.detect(_object.default), 'EmberObject is not a B class');
      assert.ok(!B.detect(A), 'A is not a B class');
      assert.ok(B.detect(B), 'B is a B class');
      assert.ok(!B.detect(C), 'C is not a B class');

      assert.ok(!C.detect(_object.default), 'EmberObject is not a C class');
      assert.ok(!C.detect(A), 'A is not a C class');
      assert.ok(!C.detect(B), 'B is not a C class');
      assert.ok(C.detect(C), 'C is a C class');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/es-compatibility-test', ['@ember/-internals/runtime/lib/system/object', '@ember/-internals/metal', 'internal-test-helpers'], function (_object, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('EmberObject ES Compatibility', class extends _internalTestHelpers.AbstractTestCase {
    ['@test extending an Ember.Object'](assert) {
      let calls = [];

      class MyObject extends _object.default {
        constructor() {
          calls.push('constructor');
          super(...arguments);
          this.postInitProperty = 'post-init-property';
        }

        init() {
          calls.push('init');
          super.init(...arguments);
          this.initProperty = 'init-property';
        }
      }

      let myObject = MyObject.create({ passedProperty: 'passed-property' });

      assert.deepEqual(calls, ['constructor', 'init'], 'constructor then init called (create)');
      assert.equal(myObject.postInitProperty, 'post-init-property', 'constructor property available on instance (create)');
      assert.equal(myObject.initProperty, 'init-property', 'init property available on instance (create)');
      assert.equal(myObject.passedProperty, 'passed-property', 'passed property available on instance (create)');
    }

    ['@test normal method super'](assert) {
      let calls = [];

      let Foo = _object.default.extend({
        method() {
          calls.push('foo');
        }
      });

      let Bar = Foo.extend({
        method() {
          this._super();
          calls.push('bar');
        }
      });

      class Baz extends Bar {
        method() {
          super.method();
          calls.push('baz');
        }
      }

      let Qux = Baz.extend({
        method() {
          this._super();
          calls.push('qux');
        }
      });

      let Quux = Qux.extend({
        method() {
          this._super();
          calls.push('quux');
        }
      });

      class Corge extends Quux {
        method() {
          super.method();
          calls.push('corge');
        }
      }

      let callValues = ['foo', 'bar', 'baz', 'qux', 'quux', 'corge'];

      [Foo, Bar, Baz, Qux, Quux, Corge].forEach((Class, index) => {
        calls = [];
        Class.create().method();

        assert.deepEqual(calls, callValues.slice(0, index + 1), 'chain of static methods called with super');
      });
    }

    ['@test static method super'](assert) {
      let calls;

      let Foo = _object.default.extend();
      Foo.reopenClass({
        method() {
          calls.push('foo');
        }
      });

      let Bar = Foo.extend();
      Bar.reopenClass({
        method() {
          this._super();
          calls.push('bar');
        }
      });

      class Baz extends Bar {
        static method() {
          super.method();
          calls.push('baz');
        }
      }

      let Qux = Baz.extend();
      Qux.reopenClass({
        method() {
          this._super();
          calls.push('qux');
        }
      });

      let Quux = Qux.extend();
      Quux.reopenClass({
        method() {
          this._super();
          calls.push('quux');
        }
      });

      class Corge extends Quux {
        static method() {
          super.method();
          calls.push('corge');
        }
      }

      let callValues = ['foo', 'bar', 'baz', 'qux', 'quux', 'corge'];

      [Foo, Bar, Baz, Qux, Quux, Corge].forEach((Class, index) => {
        calls = [];
        Class.method();

        assert.deepEqual(calls, callValues.slice(0, index + 1), 'chain of static methods called with super');
      });
    }

    ['@test using mixins'](assert) {
      let Mixin1 = _metal.Mixin.create({
        property1: 'data-1'
      });

      let Mixin2 = _metal.Mixin.create({
        property2: 'data-2'
      });

      class MyObject extends _object.default.extend(Mixin1, Mixin2) {}

      let myObject = MyObject.create();
      assert.equal(myObject.property1, 'data-1', 'includes the first mixin');
      assert.equal(myObject.property2, 'data-2', 'includes the second mixin');
    }

    ['@test using instanceof'](assert) {
      class MyObject extends _object.default {}

      let myObject = MyObject.create();

      assert.ok(myObject instanceof MyObject);
      assert.ok(myObject instanceof _object.default);
    }

    ['@test using Ember.Object#detect'](assert) {
      let Parent = _object.default.extend();
      class Child extends Parent {}
      let Grandchild = Child.extend();

      assert.ok(Parent.detect(Child), 'Parent.detect(Child)');
      assert.ok(Child.detect(Grandchild), 'Child.detect(Grandchild)');
    }

    ['@test extending an ES subclass of EmberObject'](assert) {
      let calls = [];

      class SubEmberObject extends _object.default {
        constructor() {
          calls.push('constructor');
          super(...arguments);
        }

        init() {
          calls.push('init');
          super.init(...arguments);
        }
      }

      class MyObject extends SubEmberObject {}

      MyObject.create();
      assert.deepEqual(calls, ['constructor', 'init'], 'constructor then init called (create)');
    }

    ['@test calling extend on an ES subclass of EmberObject'](assert) {
      let calls = [];

      class SubEmberObject extends _object.default {
        constructor() {
          calls.push('before constructor');
          super(...arguments);
          calls.push('after constructor');
          this.foo = 123;
        }

        init() {
          calls.push('init');
          super.init(...arguments);
        }
      }

      let MyObject = SubEmberObject.extend({});

      MyObject.create();
      assert.deepEqual(calls, ['before constructor', 'after constructor', 'init'], 'constructor then init called (create)');

      let obj = MyObject.create({
        foo: 456,
        bar: 789
      });

      assert.equal(obj.foo, 456, 'sets class fields on instance correctly');
      assert.equal(obj.bar, 789, 'sets passed in properties on instance correctly');
    }

    ['@test calling metaForProperty on a native class works'](assert) {
      assert.expect(0);

      class SubEmberObject extends _object.default {}

      (0, _metal.defineProperty)(SubEmberObject.prototype, 'foo', (0, _metal.computed)('foo', {
        get() {
          return 'bar';
        }
      }));

      // able to get meta without throwing an error
      SubEmberObject.metaForProperty('foo');
    }

    '@test observes / removeObserver on / removeListener interop'(assert) {
      let fooDidChangeBase = 0;
      let fooDidChangeA = 0;
      let fooDidChangeB = 0;
      let someEventBase = 0;
      let someEventA = 0;
      let someEventB = 0;
      class A extends _object.default.extend({
        fooDidChange: (0, _metal.observer)('foo', function () {
          fooDidChangeBase++;
        }),

        onSomeEvent: (0, _metal.on)('someEvent', function () {
          someEventBase++;
        })
      }) {
        init() {
          super.init();
          this.foo = 'bar';
        }

        fooDidChange() {
          super.fooDidChange();
          fooDidChangeA++;
        }

        onSomeEvent() {
          super.onSomeEvent();
          someEventA++;
        }
      }

      class B extends A {
        fooDidChange() {
          super.fooDidChange();
          fooDidChangeB++;
        }

        onSomeEvent() {
          super.onSomeEvent();
          someEventB++;
        }
      }

      (0, _metal.removeObserver)(B.prototype, 'foo', null, 'fooDidChange');
      (0, _metal.removeListener)(B.prototype, 'someEvent', null, 'onSomeEvent');

      assert.equal(fooDidChangeBase, 0);
      assert.equal(fooDidChangeA, 0);
      assert.equal(fooDidChangeB, 0);

      assert.equal(someEventBase, 0);
      assert.equal(someEventA, 0);
      assert.equal(someEventB, 0);

      let a = A.create();
      a.set('foo', 'something');
      assert.equal(fooDidChangeBase, 1);
      assert.equal(fooDidChangeA, 1);
      assert.equal(fooDidChangeB, 0);

      (0, _metal.sendEvent)(a, 'someEvent');
      assert.equal(someEventBase, 1);
      assert.equal(someEventA, 1);
      assert.equal(someEventB, 0);

      let b = B.create();
      b.set('foo', 'something');
      assert.equal(fooDidChangeBase, 1);
      assert.equal(fooDidChangeA, 1);
      assert.equal(fooDidChangeB, 0);

      (0, _metal.sendEvent)(b, 'someEvent');
      assert.equal(someEventBase, 1);
      assert.equal(someEventA, 1);
      assert.equal(someEventB, 0);
    }

    '@test super and _super interop between old and new methods'(assert) {
      let calls = [];
      let changes = [];
      let events = [];
      let lastProps;

      class A extends _object.default {
        init(props) {
          calls.push('A init');
          lastProps = props;
        }
      }

      let Mixin1 = _metal.Mixin.create({
        init() {
          calls.push('Mixin1 init before _super');
          this._super(...arguments);
          calls.push('Mixin1 init after _super');
        }
      });

      let Mixin2 = _metal.Mixin.create({
        init() {
          calls.push('Mixin2 init before _super');
          this._super(...arguments);
          calls.push('Mixin2 init after _super');
        }
      });

      class B extends A.extend(Mixin1, Mixin2) {
        init() {
          calls.push('B init before super.init');
          super.init(...arguments);
          calls.push('B init after super.init');
        }

        onSomeEvent(evt) {
          events.push(`B onSomeEvent ${evt}`);
        }

        fullNameDidChange() {
          changes.push('B fullNameDidChange');
        }
      }

      // // define a CP
      (0, _metal.defineProperty)(B.prototype, 'full', (0, _metal.computed)('first', 'last', {
        get() {
          return this.first + ' ' + this.last;
        }
      }));

      // Only string observers are allowed for prototypes
      (0, _metal.addObserver)(B.prototype, 'full', null, 'fullNameDidChange');

      // Only string listeners are allowed for prototypes
      (0, _metal.addListener)(B.prototype, 'someEvent', null, 'onSomeEvent');

      B.reopen({
        init() {
          calls.push('reopen init before _super');
          this._super(...arguments);
          calls.push('reopen init after _super');
        }
      });

      let C = B.extend({
        init() {
          calls.push('C init before _super');
          this._super(...arguments);
          calls.push('C init after _super');
        },

        onSomeEvent(evt) {
          calls.push('C onSomeEvent before _super');
          this._super(evt);
          calls.push('C onSomeEvent after _super');
        },

        fullNameDidChange() {
          calls.push('C fullNameDidChange before _super');
          this._super();
          calls.push('C fullNameDidChange after _super');
        }
      });

      class D extends C {
        init() {
          calls.push('D init before super.init');
          super.init(...arguments);
          calls.push('D init after super.init');
        }

        onSomeEvent(evt) {
          events.push('D onSomeEvent before super.onSomeEvent');
          super.onSomeEvent(evt);
          events.push('D onSomeEvent after super.onSomeEvent');
        }

        fullNameDidChange() {
          changes.push('D fullNameDidChange before super.fullNameDidChange');
          super.fullNameDidChange();
          changes.push('D fullNameDidChange after super.fullNameDidChange');
        }

        triggerSomeEvent(...args) {
          (0, _metal.sendEvent)(this, 'someEvent', args);
        }
      }

      assert.deepEqual(calls, [], 'nothing has been called');
      assert.deepEqual(changes, [], 'full has not changed');
      assert.deepEqual(events, [], 'onSomeEvent has not been triggered');

      let d = D.create({ first: 'Robert', last: 'Jackson' });

      assert.deepEqual(calls, ['D init before super.init', 'C init before _super', 'reopen init before _super', 'B init before super.init', 'Mixin2 init before _super', 'Mixin1 init before _super', 'A init', 'Mixin1 init after _super', 'Mixin2 init after _super', 'B init after super.init', 'reopen init after _super', 'C init after _super', 'D init after super.init']);
      assert.deepEqual(changes, [], 'full has not changed');
      assert.deepEqual(events, [], 'onSomeEvent has not been triggered');

      assert.deepEqual(lastProps, {
        first: 'Robert',
        last: 'Jackson'
      });

      assert.equal(d.full, 'Robert Jackson');

      d.setProperties({ first: 'Kris', last: 'Selden' });
      assert.deepEqual(changes, ['D fullNameDidChange before super.fullNameDidChange', 'B fullNameDidChange', 'D fullNameDidChange after super.fullNameDidChange']);

      assert.equal(d.full, 'Kris Selden');

      d.triggerSomeEvent('event arg');
      assert.deepEqual(events, ['D onSomeEvent before super.onSomeEvent', 'B onSomeEvent event arg', 'D onSomeEvent after super.onSomeEvent']);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/events_test', ['@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/mixins/evented', 'internal-test-helpers'], function (_object, _evented, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Object events', class extends _internalTestHelpers.AbstractTestCase {
    ['@test a listener can be added to an object'](assert) {
      let count = 0;
      let F = function () {
        count++;
      };

      let obj = _object.default.extend(_evented.default).create();

      obj.on('event!', F);
      obj.trigger('event!');

      assert.equal(count, 1, 'the event was triggered');

      obj.trigger('event!');

      assert.equal(count, 2, 'the event was triggered');
    }

    ['@test a listener can be added and removed automatically the first time it is triggered'](assert) {
      let count = 0;
      let F = function () {
        count++;
      };

      let obj = _object.default.extend(_evented.default).create();

      obj.one('event!', F);
      obj.trigger('event!');

      assert.equal(count, 1, 'the event was triggered');

      obj.trigger('event!');

      assert.equal(count, 1, 'the event was not triggered again');
    }

    ['@test triggering an event can have arguments'](assert) {
      let self, args;

      let obj = _object.default.extend(_evented.default).create();

      obj.on('event!', function () {
        args = [].slice.call(arguments);
        self = this;
      });

      obj.trigger('event!', 'foo', 'bar');

      assert.deepEqual(args, ['foo', 'bar']);
      assert.equal(self, obj);
    }

    ['@test a listener can be added and removed automatically and have arguments'](assert) {
      let self, args;
      let count = 0;

      let obj = _object.default.extend(_evented.default).create();

      obj.one('event!', function () {
        args = [].slice.call(arguments);
        self = this;
        count++;
      });

      obj.trigger('event!', 'foo', 'bar');

      assert.deepEqual(args, ['foo', 'bar']);
      assert.equal(self, obj);
      assert.equal(count, 1, 'the event is triggered once');

      obj.trigger('event!', 'baz', 'bat');

      assert.deepEqual(args, ['foo', 'bar']);
      assert.equal(count, 1, 'the event was not triggered again');
      assert.equal(self, obj);
    }

    ['@test binding an event can specify a different target'](assert) {
      let self, args;

      let obj = _object.default.extend(_evented.default).create();
      let target = {};

      obj.on('event!', target, function () {
        args = [].slice.call(arguments);
        self = this;
      });

      obj.trigger('event!', 'foo', 'bar');

      assert.deepEqual(args, ['foo', 'bar']);
      assert.equal(self, target);
    }

    ['@test a listener registered with one can take method as string and can be added with different target'](assert) {
      let count = 0;
      let target = {};
      target.fn = function () {
        count++;
      };

      let obj = _object.default.extend(_evented.default).create();

      obj.one('event!', target, 'fn');
      obj.trigger('event!');

      assert.equal(count, 1, 'the event was triggered');

      obj.trigger('event!');

      assert.equal(count, 1, 'the event was not triggered again');
    }

    ['@test a listener registered with one can be removed with off'](assert) {
      let obj = _object.default.extend(_evented.default, {
        F() {}
      }).create();
      let F = function () {};

      obj.one('event!', F);
      obj.one('event!', obj, 'F');

      assert.equal(obj.has('event!'), true, 'has events');

      obj.off('event!', F);
      obj.off('event!', obj, 'F');

      assert.equal(obj.has('event!'), false, 'has no more events');
    }

    ['@test adding and removing listeners should be chainable'](assert) {
      let obj = _object.default.extend(_evented.default).create();
      let F = function () {};

      let ret = obj.on('event!', F);
      assert.equal(ret, obj, '#on returns self');

      ret = obj.off('event!', F);
      assert.equal(ret, obj, '#off returns self');

      ret = obj.one('event!', F);
      assert.equal(ret, obj, '#one returns self');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/extend_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('EmberObject.extend', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Basic extend'](assert) {
      let SomeClass = _object.default.extend({ foo: 'BAR' });
      assert.ok(SomeClass.isClass, 'A class has isClass of true');
      let obj = SomeClass.create();
      assert.equal(obj.foo, 'BAR');
    }

    ['@test Sub-subclass'](assert) {
      let SomeClass = _object.default.extend({ foo: 'BAR' });
      let AnotherClass = SomeClass.extend({ bar: 'FOO' });
      let obj = AnotherClass.create();
      assert.equal(obj.foo, 'BAR');
      assert.equal(obj.bar, 'FOO');
    }

    ['@test Overriding a method several layers deep'](assert) {
      let SomeClass = _object.default.extend({
        fooCnt: 0,
        foo() {
          this.fooCnt++;
        },

        barCnt: 0,
        bar() {
          this.barCnt++;
        }
      });

      let AnotherClass = SomeClass.extend({
        barCnt: 0,
        bar() {
          this.barCnt++;
          this._super(...arguments);
        }
      });

      let FinalClass = AnotherClass.extend({
        fooCnt: 0,
        foo() {
          this.fooCnt++;
          this._super(...arguments);
        }
      });

      let obj = FinalClass.create();
      obj.foo();
      obj.bar();
      assert.equal(obj.fooCnt, 2, 'should invoke both');
      assert.equal(obj.barCnt, 2, 'should invoke both');

      // Try overriding on create also
      obj = FinalClass.extend({
        foo() {
          this.fooCnt++;
          this._super(...arguments);
        }
      }).create();

      obj.foo();
      obj.bar();
      assert.equal(obj.fooCnt, 3, 'should invoke final as well');
      assert.equal(obj.barCnt, 2, 'should invoke both');
    }

    ['@test With concatenatedProperties'](assert) {
      let SomeClass = _object.default.extend({
        things: 'foo',
        concatenatedProperties: ['things']
      });
      let AnotherClass = SomeClass.extend({ things: 'bar' });
      let YetAnotherClass = SomeClass.extend({ things: 'baz' });
      let some = SomeClass.create();
      let another = AnotherClass.create();
      let yetAnother = YetAnotherClass.create();
      assert.deepEqual(some.get('things'), ['foo'], 'base class should have just its value');
      assert.deepEqual(another.get('things'), ['foo', 'bar'], "subclass should have base class' and its own");
      assert.deepEqual(yetAnother.get('things'), ['foo', 'baz'], "subclass should have base class' and its own");
    }

    ['@test With concatenatedProperties class properties'](assert) {
      let SomeClass = _object.default.extend();
      SomeClass.reopenClass({
        concatenatedProperties: ['things'],
        things: 'foo'
      });
      let AnotherClass = SomeClass.extend();
      AnotherClass.reopenClass({ things: 'bar' });
      let YetAnotherClass = SomeClass.extend();
      YetAnotherClass.reopenClass({ things: 'baz' });
      let some = SomeClass.create();
      let another = AnotherClass.create();
      let yetAnother = YetAnotherClass.create();
      assert.deepEqual((0, _metal.get)(some.constructor, 'things'), ['foo'], 'base class should have just its value');
      assert.deepEqual((0, _metal.get)(another.constructor, 'things'), ['foo', 'bar'], "subclass should have base class' and its own");
      assert.deepEqual((0, _metal.get)(yetAnother.constructor, 'things'), ['foo', 'baz'], "subclass should have base class' and its own");
    }

    ['@test Overriding a computed property with an observer'](assert) {
      let Parent = _object.default.extend({
        foo: (0, _metal.computed)(function () {
          return 'FOO';
        })
      });

      let seen = [];

      let Child = Parent.extend({
        foo: (0, _metal.observer)('bar', function () {
          seen.push(this.get('bar'));
        })
      });

      let child = Child.create({ bar: 0 });

      assert.deepEqual(seen, []);

      child.set('bar', 1);

      assert.deepEqual(seen, [1]);

      child.set('bar', 2);

      assert.deepEqual(seen, [1, 2]);
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/observer_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_runloop, _metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('EmberObject observer', class extends _internalTestHelpers.AbstractTestCase {
    ['@test observer on class'](assert) {
      let MyClass = _object.default.extend({
        count: 0,

        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = MyClass.create();
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observer on subclass'](assert) {
      let MyClass = _object.default.extend({
        count: 0,

        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let Subclass = MyClass.extend({
        foo: (0, _metal.observer)('baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = Subclass.create();
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer after change');

      (0, _metal.set)(obj, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observer on instance'](assert) {
      let obj = _object.default.extend({
        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      }).create({
        count: 0
      });

      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observer on instance overriding class'](assert) {
      let MyClass = _object.default.extend({
        count: 0,

        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj = MyClass.extend({
        foo: (0, _metal.observer)('baz', function () {
          // <-- change property we observe
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      }).create();

      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer immediately');

      (0, _metal.set)(obj, 'bar', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer after change');

      (0, _metal.set)(obj, 'baz', 'BAZ');
      assert.equal((0, _metal.get)(obj, 'count'), 1, 'should invoke observer after change');
    }

    ['@test observer should not fire after being destroyed'](assert) {
      let obj = _object.default.extend({
        count: 0,
        foo: (0, _metal.observer)('bar', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      }).create();

      assert.equal((0, _metal.get)(obj, 'count'), 0, 'precond - should not invoke observer immediately');

      (0, _runloop.run)(() => obj.destroy());

      expectAssertion(function () {
        (0, _metal.set)(obj, 'bar', 'BAZ');
      }, `calling set on destroyed object: ${obj}.bar = BAZ`);

      assert.equal((0, _metal.get)(obj, 'count'), 0, 'should not invoke observer after change');
    }

    // ..........................................................
    // COMPLEX PROPERTIES
    //

    ['@test chain observer on class'](assert) {
      let MyClass = _object.default.extend({
        count: 0,

        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj1 = MyClass.create({
        bar: { baz: 'biff' }
      });

      let obj2 = MyClass.create({
        bar: { baz: 'biff2' }
      });

      assert.equal((0, _metal.get)(obj1, 'count'), 0, 'should not invoke yet');
      assert.equal((0, _metal.get)(obj2, 'count'), 0, 'should not invoke yet');

      (0, _metal.set)((0, _metal.get)(obj1, 'bar'), 'baz', 'BIFF1');
      assert.equal((0, _metal.get)(obj1, 'count'), 1, 'should invoke observer on obj1');
      assert.equal((0, _metal.get)(obj2, 'count'), 0, 'should not invoke yet');

      (0, _metal.set)((0, _metal.get)(obj2, 'bar'), 'baz', 'BIFF2');
      assert.equal((0, _metal.get)(obj1, 'count'), 1, 'should not invoke again');
      assert.equal((0, _metal.get)(obj2, 'count'), 1, 'should invoke observer on obj2');
    }

    ['@test chain observer on class'](assert) {
      let MyClass = _object.default.extend({
        count: 0,

        foo: (0, _metal.observer)('bar.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      });

      let obj1 = MyClass.extend().create({
        bar: { baz: 'biff' }
      });

      let obj2 = MyClass.extend({
        foo: (0, _metal.observer)('bar2.baz', function () {
          (0, _metal.set)(this, 'count', (0, _metal.get)(this, 'count') + 1);
        })
      }).create({
        bar: { baz: 'biff2' },
        bar2: { baz: 'biff3' }
      });

      assert.equal((0, _metal.get)(obj1, 'count'), 0, 'should not invoke yet');
      assert.equal((0, _metal.get)(obj2, 'count'), 0, 'should not invoke yet');

      (0, _metal.set)((0, _metal.get)(obj1, 'bar'), 'baz', 'BIFF1');
      assert.equal((0, _metal.get)(obj1, 'count'), 1, 'should invoke observer on obj1');
      assert.equal((0, _metal.get)(obj2, 'count'), 0, 'should not invoke yet');

      (0, _metal.set)((0, _metal.get)(obj2, 'bar'), 'baz', 'BIFF2');
      assert.equal((0, _metal.get)(obj1, 'count'), 1, 'should not invoke again');
      assert.equal((0, _metal.get)(obj2, 'count'), 0, 'should not invoke yet');

      (0, _metal.set)((0, _metal.get)(obj2, 'bar2'), 'baz', 'BIFF3');
      assert.equal((0, _metal.get)(obj1, 'count'), 1, 'should not invoke again');
      assert.equal((0, _metal.get)(obj2, 'count'), 1, 'should invoke observer on obj2');
    }

    ['@test chain observer on class that has a reference to an uninitialized object will finish chains that reference it'](assert) {
      let changed = false;

      let ChildClass = _object.default.extend({
        parent: null,
        parentOneTwoDidChange: (0, _metal.observer)('parent.one.two', function () {
          changed = true;
        })
      });

      let ParentClass = _object.default.extend({
        one: {
          two: 'old'
        },
        init() {
          this.child = ChildClass.create({
            parent: this
          });
        }
      });

      let parent = ParentClass.create();

      assert.equal(changed, false, 'precond');

      (0, _metal.set)(parent, 'one.two', 'new');

      assert.equal(changed, true, 'child should have been notified of change to path');

      (0, _metal.set)(parent, 'one', { two: 'newer' });

      assert.equal(changed, true, 'child should have been notified of change to path');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/reopenClass_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/object/reopenClass', class extends _internalTestHelpers.AbstractTestCase {
    ['@test adds new properties to subclass'](assert) {
      let Subclass = _object.default.extend();
      Subclass.reopenClass({
        foo() {
          return 'FOO';
        },
        bar: 'BAR'
      });

      assert.equal(Subclass.foo(), 'FOO', 'Adds method');
      assert.equal((0, _metal.get)(Subclass, 'bar'), 'BAR', 'Adds property');
    }

    ['@test class properties inherited by subclasses'](assert) {
      let Subclass = _object.default.extend();
      Subclass.reopenClass({
        foo() {
          return 'FOO';
        },
        bar: 'BAR'
      });

      let SubSub = Subclass.extend();

      assert.equal(SubSub.foo(), 'FOO', 'Adds method');
      assert.equal((0, _metal.get)(SubSub, 'bar'), 'BAR', 'Adds property');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/reopen_test', ['@ember/-internals/metal', '@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_metal, _object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/core_object/reopen', class extends _internalTestHelpers.AbstractTestCase {
    ['@test adds new properties to subclass instance'](assert) {
      let Subclass = _object.default.extend();
      Subclass.reopen({
        foo() {
          return 'FOO';
        },
        bar: 'BAR'
      });

      assert.equal(Subclass.create().foo(), 'FOO', 'Adds method');
      assert.equal((0, _metal.get)(Subclass.create(), 'bar'), 'BAR', 'Adds property');
    }

    ['@test reopened properties inherited by subclasses'](assert) {
      let Subclass = _object.default.extend();
      let SubSub = Subclass.extend();

      Subclass.reopen({
        foo() {
          return 'FOO';
        },
        bar: 'BAR'
      });

      assert.equal(SubSub.create().foo(), 'FOO', 'Adds method');
      assert.equal((0, _metal.get)(SubSub.create(), 'bar'), 'BAR', 'Adds property');
    }

    ['@test allows reopening already instantiated classes'](assert) {
      let Subclass = _object.default.extend();

      Subclass.create();

      Subclass.reopen({
        trololol: true
      });

      assert.equal(Subclass.create().get('trololol'), true, 'reopen works');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/strict-mode-test', ['@ember/-internals/runtime/lib/system/object', 'internal-test-helpers'], function (_object, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('strict mode tests', class extends _internalTestHelpers.AbstractTestCase {
    ['@test __superWrapper does not throw errors in strict mode'](assert) {
      let Foo = _object.default.extend({
        blah() {
          return 'foo';
        }
      });

      let Bar = Foo.extend({
        blah() {
          return 'bar';
        },

        callBlah() {
          let blah = this.blah;

          return blah();
        }
      });

      let bar = Bar.create();

      assert.equal(bar.callBlah(), 'bar', 'can call local function without call/apply');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object/toString_test', ['@ember/runloop', '@ember/-internals/utils', '@ember/-internals/environment', '@ember/-internals/runtime/lib/system/object', '@ember/-internals/runtime/lib/system/namespace', 'internal-test-helpers'], function (_runloop, _utils, _environment, _object, _namespace, _internalTestHelpers) {
  'use strict';

  let originalLookup = _environment.context.lookup;
  let lookup;

  (0, _internalTestHelpers.moduleFor)('system/object/toString', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      _environment.context.lookup = lookup = {};
    }

    afterEach() {
      _environment.context.lookup = originalLookup;
    }

    ['@test toString() returns the same value if called twice'](assert) {
      let Foo = _namespace.default.create();
      Foo.toString = function () {
        return 'Foo';
      };

      Foo.Bar = _object.default.extend();

      assert.equal(Foo.Bar.toString(), 'Foo.Bar');
      assert.equal(Foo.Bar.toString(), 'Foo.Bar');

      let obj = Foo.Bar.create();

      assert.equal(obj.toString(), '<Foo.Bar:' + (0, _utils.guidFor)(obj) + '>');
      assert.equal(obj.toString(), '<Foo.Bar:' + (0, _utils.guidFor)(obj) + '>');

      assert.equal(Foo.Bar.toString(), 'Foo.Bar');

      (0, _runloop.run)(Foo, 'destroy');
    }

    ['@test toString on a class returns a useful value when nested in a namespace'](assert) {
      let obj;

      let Foo = _namespace.default.create();
      Foo.toString = function () {
        return 'Foo';
      };

      Foo.Bar = _object.default.extend();
      assert.equal(Foo.Bar.toString(), 'Foo.Bar');

      obj = Foo.Bar.create();
      assert.equal(obj.toString(), '<Foo.Bar:' + (0, _utils.guidFor)(obj) + '>');

      Foo.Baz = Foo.Bar.extend();
      assert.equal(Foo.Baz.toString(), 'Foo.Baz');

      obj = Foo.Baz.create();
      assert.equal(obj.toString(), '<Foo.Baz:' + (0, _utils.guidFor)(obj) + '>');

      obj = Foo.Bar.create();
      assert.equal(obj.toString(), '<Foo.Bar:' + (0, _utils.guidFor)(obj) + '>');

      (0, _runloop.run)(Foo, 'destroy');
    }

    ['@test toString on a namespace finds the namespace in lookup'](assert) {
      let Foo = lookup.Foo = _namespace.default.create();

      assert.equal(Foo.toString(), 'Foo');

      (0, _runloop.run)(Foo, 'destroy');
    }

    ['@test toString on a namespace finds the namespace in lookup'](assert) {
      let Foo = lookup.Foo = _namespace.default.create();
      let obj;

      Foo.Bar = _object.default.extend();

      assert.equal(Foo.Bar.toString(), 'Foo.Bar');

      obj = Foo.Bar.create();
      assert.equal(obj.toString(), '<Foo.Bar:' + (0, _utils.guidFor)(obj) + '>');

      (0, _runloop.run)(Foo, 'destroy');
    }

    ['@test toString on a namespace falls back to modulePrefix, if defined'](assert) {
      let Foo = _namespace.default.create({ modulePrefix: 'foo' });

      assert.equal(Foo.toString(), 'foo');

      (0, _runloop.run)(Foo, 'destroy');
    }

    ['@test toString includes toStringExtension if defined'](assert) {
      let Foo = _object.default.extend({
        toStringExtension() {
          return 'fooey';
        }
      });
      let foo = Foo.create();
      let Bar = _object.default.extend({});
      let bar = Bar.create();

      // simulate these classes being defined on a Namespace
      (0, _utils.setName)(Foo, 'Foo');
      (0, _utils.setName)(Bar, 'Bar');

      assert.equal(bar.toString(), '<Bar:' + (0, _utils.guidFor)(bar) + '>', 'does not include toStringExtension part');
      assert.equal(foo.toString(), '<Foo:' + (0, _utils.guidFor)(foo) + ':fooey>', 'Includes toStringExtension result');
    }
  });
});
enifed('@ember/-internals/runtime/tests/system/object_proxy_test', ['@ember/-internals/metal', '@ember/-internals/utils', '@ember/-internals/runtime/lib/system/object_proxy', 'internal-test-helpers'], function (_metal, _utils, _object_proxy, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ObjectProxy', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should not proxy properties passed to create'](assert) {
      let Proxy = _object_proxy.default.extend({
        cp: (0, _metal.computed)({
          get() {
            return this._cp;
          },
          set(key, value) {
            this._cp = value;
            return this._cp;
          }
        })
      });
      let proxy = Proxy.create({
        prop: 'Foo',
        cp: 'Bar'
      });

      assert.equal((0, _metal.get)(proxy, 'prop'), 'Foo', 'should not have tried to proxy set');
      assert.equal(proxy._cp, 'Bar', 'should use CP setter');
    }

    ['@test should proxy properties to content'](assert) {
      let content = {
        firstName: 'Tom',
        lastName: 'Dale',
        unknownProperty(key) {
          return key + ' unknown';
        }
      };
      let proxy = _object_proxy.default.create();

      assert.equal((0, _metal.get)(proxy, 'firstName'), undefined, 'get on proxy without content should return undefined');
      expectAssertion(() => {
        (0, _metal.set)(proxy, 'firstName', 'Foo');
      }, /Cannot delegate set\('firstName', Foo\) to the 'content'/i);

      (0, _metal.set)(proxy, 'content', content);

      assert.equal((0, _metal.get)(proxy, 'firstName'), 'Tom', 'get on proxy with content should forward to content');
      assert.equal((0, _metal.get)(proxy, 'lastName'), 'Dale', 'get on proxy with content should forward to content');
      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo unknown', 'get on proxy with content should forward to content');

      (0, _metal.set)(proxy, 'lastName', 'Huda');

      assert.equal((0, _metal.get)(content, 'lastName'), 'Huda', 'content should have new value from set on proxy');
      assert.equal((0, _metal.get)(proxy, 'lastName'), 'Huda', 'proxy should have new value from set on proxy');

      (0, _metal.set)(proxy, 'content', { firstName: 'Yehuda', lastName: 'Katz' });

      assert.equal((0, _metal.get)(proxy, 'firstName'), 'Yehuda', 'proxy should reflect updated content');
      assert.equal((0, _metal.get)(proxy, 'lastName'), 'Katz', 'proxy should reflect updated content');
    }

    ['@test getting proxied properties with Ember.get should work'](assert) {
      let proxy = _object_proxy.default.create({
        content: {
          foo: 'FOO'
        }
      });

      assert.equal((0, _metal.get)(proxy, 'foo'), 'FOO');
    }

    [`@test JSON.stringify doens't assert`](assert) {
      let proxy = _object_proxy.default.create({
        content: {
          foo: 'FOO'
        }
      });

      assert.equal(JSON.stringify(proxy), JSON.stringify({ content: { foo: 'FOO' } }));
    }

    ['@test calling a function on the proxy avoids the assertion'](assert) {
      if (true /* DEBUG */ && _utils.HAS_NATIVE_PROXY) {
        let proxy = _object_proxy.default.extend({
          init() {
            if (!this.foobar) {
              this.foobar = function () {
                let content = (0, _metal.get)(this, 'content');
                return content.foobar.apply(content, []);
              };
            }
          }
        }).create({
          content: {
            foobar() {
              return 'xoxo';
            }
          }
        });

        assert.equal(proxy.foobar(), 'xoxo', 'should be able to use a function from a proxy');
      } else {
        assert.expect(0);
      }
    }

    [`@test setting a property on the proxy avoids the assertion`](assert) {
      let proxy = _object_proxy.default.create({
        toJSON: undefined,
        content: {
          toJSON() {
            return 'hello';
          }
        }
      });

      assert.equal(JSON.stringify(proxy), JSON.stringify({ content: 'hello' }));
    }

    [`@test setting a property on the proxy's prototype avoids the assertion`](assert) {
      let proxy = _object_proxy.default.extend({
        toJSON: null
      }).create({
        content: {
          toJSON() {
            return 'hello';
          }
        }
      });

      assert.equal(JSON.stringify(proxy), JSON.stringify({ content: 'hello' }));
    }

    ['@test getting proxied properties with [] should be an error'](assert) {
      if (true /* DEBUG */ && _utils.HAS_NATIVE_PROXY) {
        let proxy = _object_proxy.default.create({
          content: {
            foo: 'FOO'
          }
        });

        expectAssertion(() => proxy.foo, /\.get\('foo'\)/);
      } else {
        assert.expect(0);
      }
    }

    ['@test should work with watched properties'](assert) {
      let content1 = { firstName: 'Tom', lastName: 'Dale' };
      let content2 = { firstName: 'Yehuda', lastName: 'Katz' };
      let count = 0;
      let last;

      let Proxy = _object_proxy.default.extend({
        fullName: (0, _metal.computed)(function () {
          let firstName = this.get('firstName');
          let lastName = this.get('lastName');

          if (firstName && lastName) {
            return firstName + ' ' + lastName;
          }
          return firstName || lastName;
        }).property('firstName', 'lastName')
      });

      let proxy = Proxy.create();

      (0, _metal.addObserver)(proxy, 'fullName', function () {
        last = (0, _metal.get)(proxy, 'fullName');
        count++;
      });

      // proxy without content returns undefined
      assert.equal((0, _metal.get)(proxy, 'fullName'), undefined);

      // setting content causes all watched properties to change
      (0, _metal.set)(proxy, 'content', content1);
      // both dependent keys changed
      assert.equal(count, 2);
      assert.equal(last, 'Tom Dale');

      // setting property in content causes proxy property to change
      (0, _metal.set)(content1, 'lastName', 'Huda');
      assert.equal(count, 3);
      assert.equal(last, 'Tom Huda');

      // replacing content causes all watched properties to change
      (0, _metal.set)(proxy, 'content', content2);
      // both dependent keys changed
      assert.equal(count, 5);
      assert.equal(last, 'Yehuda Katz');
      // content1 is no longer watched
      assert.ok(!(0, _metal.isWatching)(content1, 'firstName'), 'not watching firstName');
      assert.ok(!(0, _metal.isWatching)(content1, 'lastName'), 'not watching lastName');

      // setting property in new content
      (0, _metal.set)(content2, 'firstName', 'Tomhuda');
      assert.equal(last, 'Tomhuda Katz');
      assert.equal(count, 6);

      // setting property in proxy syncs with new content
      (0, _metal.set)(proxy, 'lastName', 'Katzdale');
      assert.equal(count, 7);
      assert.equal(last, 'Tomhuda Katzdale');
      assert.equal((0, _metal.get)(content2, 'firstName'), 'Tomhuda');
      assert.equal((0, _metal.get)(content2, 'lastName'), 'Katzdale');
    }

    ['@test set and get should work with paths'](assert) {
      let content = { foo: { bar: 'baz' } };
      let proxy = _object_proxy.default.create({ content });
      let count = 0;

      proxy.set('foo.bar', 'hello');
      assert.equal(proxy.get('foo.bar'), 'hello');
      assert.equal(proxy.get('content.foo.bar'), 'hello');

      proxy.addObserver('foo.bar', function () {
        count++;
      });

      proxy.set('foo.bar', 'bye');

      assert.equal(count, 1);
      assert.equal(proxy.get('foo.bar'), 'bye');
      assert.equal(proxy.get('content.foo.bar'), 'bye');
    }

    ['@test should transition between watched and unwatched strategies'](assert) {
      let content = { foo: 'foo' };
      let proxy = _object_proxy.default.create({ content: content });
      let count = 0;

      function observer() {
        count++;
      }

      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo');

      (0, _metal.set)(content, 'foo', 'bar');

      assert.equal((0, _metal.get)(proxy, 'foo'), 'bar');

      (0, _metal.set)(proxy, 'foo', 'foo');

      assert.equal((0, _metal.get)(content, 'foo'), 'foo');
      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo');

      (0, _metal.addObserver)(proxy, 'foo', observer);

      assert.equal(count, 0);
      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo');

      (0, _metal.set)(content, 'foo', 'bar');

      assert.equal(count, 1);
      assert.equal((0, _metal.get)(proxy, 'foo'), 'bar');

      (0, _metal.set)(proxy, 'foo', 'foo');

      assert.equal(count, 2);
      assert.equal((0, _metal.get)(content, 'foo'), 'foo');
      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo');

      (0, _metal.removeObserver)(proxy, 'foo', observer);

      (0, _metal.set)(content, 'foo', 'bar');

      assert.equal((0, _metal.get)(proxy, 'foo'), 'bar');

      (0, _metal.set)(proxy, 'foo', 'foo');

      assert.equal((0, _metal.get)(content, 'foo'), 'foo');
      assert.equal((0, _metal.get)(proxy, 'foo'), 'foo');
    }

    ['@test setting `undefined` to a proxied content property should override its existing value'](assert) {
      let proxyObject = _object_proxy.default.create({
        content: {
          prop: 'emberjs'
        }
      });
      (0, _metal.set)(proxyObject, 'prop', undefined);
      assert.equal((0, _metal.get)(proxyObject, 'prop'), undefined, 'sets the `undefined` value to the proxied content');
    }
  });
});
enifed('@ember/-internals/utils/tests/cache_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Cache', class extends _internalTestHelpers.AbstractTestCase {
    ['@test basic'](assert) {
      let cache = new _utils.Cache(100, key => key.toUpperCase());

      assert.equal(cache.get('foo'), 'FOO');
      assert.equal(cache.get('bar'), 'BAR');
      assert.equal(cache.get('foo'), 'FOO');
    }

    ['@test explicit sets'](assert) {
      let cache = new _utils.Cache(100, key => key.toUpperCase());

      assert.equal(cache.get('foo'), 'FOO');

      assert.equal(cache.set('foo', 'FOO!!!'), 'FOO!!!');

      assert.equal(cache.get('foo'), 'FOO!!!');

      assert.strictEqual(cache.set('foo', undefined), undefined);

      assert.strictEqual(cache.get('foo'), undefined);
    }

    ['@test caches computation correctly'](assert) {
      let count = 0;
      let cache = new _utils.Cache(100, key => {
        count++;
        return key.toUpperCase();
      });

      assert.equal(count, 0);
      cache.get('foo');
      assert.equal(count, 1);
      cache.get('bar');
      assert.equal(count, 2);
      cache.get('bar');
      assert.equal(count, 2);
      cache.get('foo');
      assert.equal(count, 2);
    }

    ['@test handles undefined value correctly'](assert) {
      let count = 0;
      let cache = new _utils.Cache(100, () => {
        count++;
      });

      assert.equal(count, 0);
      assert.strictEqual(cache.get('foo'), undefined);
      assert.equal(count, 1);
      assert.strictEqual(cache.get('bar'), undefined);
      assert.equal(count, 2);
      assert.strictEqual(cache.get('bar'), undefined);
      assert.equal(count, 2);
      assert.strictEqual(cache.get('foo'), undefined);
      assert.equal(count, 2);
    }

    ['@test continues working after reaching cache limit'](assert) {
      let cache = new _utils.Cache(3, key => key.toUpperCase());

      cache.get('a');
      cache.get('b');
      cache.get('c');

      assert.equal(cache.get('d'), 'D');
      assert.equal(cache.get('a'), 'A');
      assert.equal(cache.get('b'), 'B');
      assert.equal(cache.get('c'), 'C');
    }
  });
});
enifed('@ember/-internals/utils/tests/can_invoke_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  let obj;

  (0, _internalTestHelpers.moduleFor)('Ember.canInvoke', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      obj = {
        foobar: 'foobar',
        aMethodThatExists() {}
      };
    }

    teardown() {
      obj = undefined;
    }

    ["@test should return false if the object doesn't exist"](assert) {
      assert.equal((0, _utils.canInvoke)(undefined, 'aMethodThatDoesNotExist'), false);
    }

    ['@test should return true for falsy values that have methods'](assert) {
      assert.equal((0, _utils.canInvoke)(false, 'valueOf'), true);
      assert.equal((0, _utils.canInvoke)('', 'charAt'), true);
      assert.equal((0, _utils.canInvoke)(0, 'toFixed'), true);
    }

    ['@test should return true if the method exists on the object'](assert) {
      assert.equal((0, _utils.canInvoke)(obj, 'aMethodThatExists'), true);
    }

    ["@test should return false if the method doesn't exist on the object"](assert) {
      assert.equal((0, _utils.canInvoke)(obj, 'aMethodThatDoesNotExist'), false);
    }

    ['@test should return false if the property exists on the object but is a non-function'](assert) {
      assert.equal((0, _utils.canInvoke)(obj, 'foobar'), false);
    }
  });
});
enifed('@ember/-internals/utils/tests/checkHasSuper_test', ['@ember/-internals/browser-environment', '@ember/-internals/utils', 'internal-test-helpers'], function (_browserEnvironment, _utils, _internalTestHelpers) {
  'use strict';

  // Only run this test on browsers that we are certain should have function
  // source available.  This allows the test suite to continue to pass on other
  // platforms that correctly (for them) fall back to the "always wrap" code.
  if (_browserEnvironment.isChrome || _browserEnvironment.isFirefox) {
    (0, _internalTestHelpers.moduleFor)('checkHasSuper', class extends _internalTestHelpers.AbstractTestCase {
      ['@test does not super wrap needlessly [GH #12462]'](assert) {
        assert.notOk((0, _utils.checkHasSuper)(function () {}), 'empty function does not have super');
      }
    });
  }
});
enifed('@ember/-internals/utils/tests/generate_guid_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.generateGuid', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Prefix'](assert) {
      let a = {};

      assert.ok((0, _utils.generateGuid)(a, 'tyrell').indexOf('tyrell') > -1, 'guid can be prefixed');
    }
  });
});
enifed('@ember/-internals/utils/tests/guid_for_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  function sameGuid(assert, a, b, message) {
    assert.equal((0, _utils.guidFor)(a), (0, _utils.guidFor)(b), message);
  }

  function diffGuid(assert, a, b, message) {
    assert.ok((0, _utils.guidFor)(a) !== (0, _utils.guidFor)(b), message);
  }

  function nanGuid(assert, obj) {
    let type = typeof obj;
    assert.ok(isNaN(parseInt((0, _utils.guidFor)(obj), 0)), 'guids for ' + type + "don't parse to numbers");
  }

  (0, _internalTestHelpers.moduleFor)('guidFor', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Object'](assert) {
      let a = {};
      let b = {};

      sameGuid(assert, a, a, 'same object always yields same guid');
      diffGuid(assert, a, b, 'different objects yield different guids');
      nanGuid(assert, a);
    }

    ['@test strings'](assert) {
      let a = 'string A';
      let aprime = 'string A';
      let b = 'String B';

      sameGuid(assert, a, a, 'same string always yields same guid');
      sameGuid(assert, a, aprime, 'identical strings always yield the same guid');
      diffGuid(assert, a, b, 'different strings yield different guids');
      nanGuid(assert, a);
    }

    ['@test numbers'](assert) {
      let a = 23;
      let aprime = 23;
      let b = 34;

      sameGuid(assert, a, a, 'same numbers always yields same guid');
      sameGuid(assert, a, aprime, 'identical numbers always yield the same guid');
      diffGuid(assert, a, b, 'different numbers yield different guids');
      nanGuid(assert, a);
    }

    ['@test symbols'](assert) {
      if (typeof Symbol === 'undefined') {
        assert.ok(true, 'symbols are not supported on this browser');
        return;
      }

      let a = Symbol('a');
      let b = Symbol('b');

      sameGuid(assert, a, a, 'same symbols always yields same guid');
      diffGuid(assert, a, b, 'different symbols yield different guids');
      nanGuid(assert, a);
    }

    ['@test booleans'](assert) {
      let a = true;
      let aprime = true;
      let b = false;

      sameGuid(assert, a, a, 'same booleans always yields same guid');
      sameGuid(assert, a, aprime, 'identical booleans always yield the same guid');
      diffGuid(assert, a, b, 'different boolean yield different guids');
      nanGuid(assert, a);
      nanGuid(assert, b);
    }

    ['@test null and undefined'](assert) {
      let a = null;
      let aprime = null;
      let b;

      sameGuid(assert, a, a, 'null always returns the same guid');
      sameGuid(assert, b, b, 'undefined always returns the same guid');
      sameGuid(assert, a, aprime, 'different nulls return the same guid');
      diffGuid(assert, a, b, 'null and undefined return different guids');
      nanGuid(assert, a);
      nanGuid(assert, b);
    }

    ['@test arrays'](assert) {
      let a = ['a', 'b', 'c'];
      let aprime = ['a', 'b', 'c'];
      let b = ['1', '2', '3'];

      sameGuid(assert, a, a, 'same instance always yields same guid');
      diffGuid(assert, a, aprime, 'identical arrays always yield the same guid');
      diffGuid(assert, a, b, 'different arrays yield different guids');
      nanGuid(assert, a);
    }
  });
});
enifed('@ember/-internals/utils/tests/inspect_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.inspect', class extends _internalTestHelpers.AbstractTestCase {
    ['@test strings'](assert) {
      assert.equal((0, _utils.inspect)('foo'), '"foo"');
    }

    ['@test numbers'](assert) {
      assert.equal((0, _utils.inspect)(2.6), '2.6');
    }

    ['@test null'](assert) {
      assert.equal((0, _utils.inspect)(null), 'null');
    }

    ['@test undefined'](assert) {
      assert.equal((0, _utils.inspect)(undefined), 'undefined');
    }

    ['@test true'](assert) {
      assert.equal((0, _utils.inspect)(true), 'true');
    }

    ['@test false'](assert) {
      assert.equal((0, _utils.inspect)(false), 'false');
    }

    ['@test object'](assert) {
      assert.equal((0, _utils.inspect)({}), '{ }');
      assert.equal((0, _utils.inspect)({ foo: 'bar' }), '{ foo: "bar" }');
      let obj = {
        foo() {
          return this;
        }
      };
      // IE 11 doesn't have function name
      if (obj.foo.name) {
        assert.equal((0, _utils.inspect)(obj), `{ foo: [Function:foo] }`);
      } else {
        assert.equal((0, _utils.inspect)(obj), `{ foo: [Function] }`);
      }
    }

    ['@test objects without a prototype'](assert) {
      let prototypelessObj = Object.create(null);
      prototypelessObj.a = 1;
      prototypelessObj.b = [Object.create(null)];
      assert.equal((0, _utils.inspect)({ foo: prototypelessObj }), '{ foo: { a: 1, b: [ { } ] } }');
    }

    ['@test array'](assert) {
      assert.equal((0, _utils.inspect)([1, 2, 3]), '[ 1, 2, 3 ]');
    }

    ['@test array list limit'](assert) {
      let a = [];
      for (let i = 0; i < 120; i++) {
        a.push(1);
      }
      assert.equal((0, _utils.inspect)(a), `[ ${a.slice(0, 100).join(', ')}, ... 20 more items ]`);
    }

    ['@test object list limit'](assert) {
      let obj = {};
      let pairs = [];
      for (let i = 0; i < 120; i++) {
        obj['key' + i] = i;
        pairs.push(`key${i}: ${i}`);
      }
      assert.equal((0, _utils.inspect)(obj), `{ ${pairs.slice(0, 100).join(', ')}, ... 20 more keys }`);
    }

    ['@test depth limit'](assert) {
      assert.equal((0, _utils.inspect)([[[['here', { a: 1 }, [1]]]]]), '[ [ [ [ "here", [Object], [Array] ] ] ] ]');
    }

    ['@test odd key'](assert) {
      assert.equal((0, _utils.inspect)({
        [`Hello world!
How are you?`]: 1
      }), '{ "Hello world!\\nHow are you?": 1 }');
    }

    ['@test node call'](assert) {
      let obj = { a: 1 };
      obj.inspect = _utils.inspect;
      let depth = 2;
      let options = {};
      assert.equal(obj.inspect(depth, options), obj);
    }

    ['@test cycle'](assert) {
      let obj = {};
      obj.a = obj;
      let arr = [obj];
      arr.push(arr);
      assert.equal((0, _utils.inspect)(arr), '[ { a: [Circular] }, [Circular] ]');
    }

    ['@test custom toString'](assert) {
      class Component {
        static toString() {
          return '@ember/component';
        }

        toString() {
          return `<${this.constructor}:ember234>`;
        }
      }

      assert.equal((0, _utils.inspect)([new Component(), Component]), '[ <@ember/component:ember234>, @ember/component ]');
    }

    ['@test regexp'](assert) {
      assert.equal((0, _utils.inspect)(/regexp/), '/regexp/');
    }

    ['@test date'](assert) {
      let inspected = (0, _utils.inspect)(new Date('Sat Apr 30 2011 13:24:11'));
      assert.ok(inspected.match(/Sat Apr 30/), 'The inspected date has its date');
      assert.ok(inspected.match(/2011/), 'The inspected date has its year');
      assert.ok(inspected.match(/13:24:11/), 'The inspected date has its time');
    }

    ['@test inspect outputs the toString() representation of Symbols'](assert) {
      if (_utils.HAS_NATIVE_SYMBOL) {
        let symbol = Symbol('test');
        assert.equal((0, _utils.inspect)(symbol), 'Symbol(test)');
      } else {
        assert.expect(0);
      }
    }
  });
});
enifed('@ember/-internals/utils/tests/is_proxy_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/utils isProxy', class extends _internalTestHelpers.AbstractTestCase {
    ['@test basic'](assert) {
      let proxy = {};
      (0, _utils.setProxy)(proxy);

      assert.equal((0, _utils.isProxy)(proxy), true);

      assert.equal((0, _utils.isProxy)({}), false);
      assert.equal((0, _utils.isProxy)(undefined), false);
      assert.equal((0, _utils.isProxy)(null), false);
    }
  });
});
enifed('@ember/-internals/utils/tests/make_array_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.makeArray', class extends _internalTestHelpers.AbstractTestCase {
    ['@test undefined'](assert) {
      assert.deepEqual((0, _utils.makeArray)(), []);
      assert.deepEqual((0, _utils.makeArray)(undefined), []);
    }

    ['@test null'](assert) {
      assert.deepEqual((0, _utils.makeArray)(null), []);
    }

    ['@test string'](assert) {
      assert.deepEqual((0, _utils.makeArray)('lindsay'), ['lindsay']);
    }

    ['@test number'](assert) {
      assert.deepEqual((0, _utils.makeArray)(0), [0]);
      assert.deepEqual((0, _utils.makeArray)(1), [1]);
    }

    ['@test array'](assert) {
      assert.deepEqual((0, _utils.makeArray)([1, 2, 42]), [1, 2, 42]);
    }

    ['@test true'](assert) {
      assert.deepEqual((0, _utils.makeArray)(true), [true]);
    }

    ['@test false'](assert) {
      assert.deepEqual((0, _utils.makeArray)(false), [false]);
    }

    ['@test object'](assert) {
      assert.deepEqual((0, _utils.makeArray)({}), [{}]);
    }
  });
});
enifed('@ember/-internals/utils/tests/to-string-test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('@ember/-internals/utils toString', class extends _internalTestHelpers.AbstractTestCase {
    [`@test toString uses an object's toString method when available`](assert) {
      let obj = {
        toString() {
          return 'bob';
        }
      };

      assert.strictEqual((0, _utils.toString)(obj), 'bob');
    }

    ['@test toString falls back to Object.prototype.toString'](assert) {
      let obj = Object.create(null);

      assert.strictEqual((0, _utils.toString)(obj), {}.toString());
    }

    ['@test toString does not fail when called on Arrays with objects without toString method'](assert) {
      let obj = Object.create(null);
      let arr = [obj, 2];

      assert.strictEqual((0, _utils.toString)(arr), `${{}.toString()},2`);
    }
  });
});
enifed('@ember/-internals/utils/tests/try_invoke_test', ['@ember/-internals/utils', 'internal-test-helpers'], function (_utils, _internalTestHelpers) {
  'use strict';

  let obj;

  (0, _internalTestHelpers.moduleFor)('Ember.tryInvoke', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      obj = {
        aMethodThatExists() {
          return true;
        },
        aMethodThatTakesArguments(arg1, arg2) {
          return arg1 === arg2;
        }
      };
    }

    teardown() {
      obj = undefined;
    }

    ["@test should return undefined when the object doesn't exist"](assert) {
      assert.equal((0, _utils.tryInvoke)(undefined, 'aMethodThatDoesNotExist'), undefined);
    }

    ["@test should return undefined when asked to perform a method that doesn't exist on the object"](assert) {
      assert.equal((0, _utils.tryInvoke)(obj, 'aMethodThatDoesNotExist'), undefined);
    }

    ['@test should return what the method returns when asked to perform a method that exists on the object'](assert) {
      assert.equal((0, _utils.tryInvoke)(obj, 'aMethodThatExists'), true);
    }

    ['@test should return what the method returns when asked to perform a method that takes arguments and exists on the object'](assert) {
      assert.equal((0, _utils.tryInvoke)(obj, 'aMethodThatTakesArguments', [true, true]), true);
    }
  });
});
enifed('@ember/application/tests/application_instance_test', ['@ember/engine', '@ember/application', '@ember/application/instance', '@ember/runloop', '@ember/-internals/container', 'internal-test-helpers', '@ember/-internals/runtime', '@ember/debug'], function (_engine, _application, _instance, _runloop, _container, _internalTestHelpers, _runtime, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');
  const noop = function () {};

  let application, appInstance;

  (0, _internalTestHelpers.moduleFor)('ApplicationInstance', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super();

      document.getElementById('qunit-fixture').innerHTML = `
      <div id='one'><div id='one-child'>HI</div></div><div id='two'>HI</div>
    `;
      application = (0, _runloop.run)(() => _application.default.create({ rootElement: '#one', router: null }));
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
      if (appInstance) {
        (0, _runloop.run)(appInstance, 'destroy');
        appInstance = null;
      }

      if (application) {
        (0, _runloop.run)(application, 'destroy');
        application = null;
      }

      document.getElementById('qunit-fixture').innerHTML = '';
    }

    ['@test an application instance can be created based upon an application'](assert) {
      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));

      assert.ok(appInstance, 'instance should be created');
      assert.equal(appInstance.application, application, 'application should be set to parent');
    }

    ['@test customEvents added to the application before setupEventDispatcher'](assert) {
      assert.expect(1);

      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));
      appInstance.setupRegistry();

      application.customEvents = {
        awesome: 'sauce'
      };

      let eventDispatcher = appInstance.lookup('event_dispatcher:main');
      eventDispatcher.setup = function (events) {
        assert.equal(events.awesome, 'sauce');
      };

      appInstance.setupEventDispatcher();
    }

    ['@test customEvents added to the application before setupEventDispatcher'](assert) {
      assert.expect(1);

      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));
      appInstance.setupRegistry();

      application.customEvents = {
        awesome: 'sauce'
      };

      let eventDispatcher = appInstance.lookup('event_dispatcher:main');
      eventDispatcher.setup = function (events) {
        assert.equal(events.awesome, 'sauce');
      };

      appInstance.setupEventDispatcher();
    }

    ['@test customEvents added to the application instance before setupEventDispatcher'](assert) {
      assert.expect(1);

      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));
      appInstance.setupRegistry();

      appInstance.customEvents = {
        awesome: 'sauce'
      };

      let eventDispatcher = appInstance.lookup('event_dispatcher:main');
      eventDispatcher.setup = function (events) {
        assert.equal(events.awesome, 'sauce');
      };

      appInstance.setupEventDispatcher();
    }

    ['@test unregistering a factory clears all cached instances of that factory'](assert) {
      assert.expect(5);

      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));

      let PostController1 = (0, _internalTestHelpers.factory)();
      let PostController2 = (0, _internalTestHelpers.factory)();

      appInstance.register('controller:post', PostController1);

      let postController1 = appInstance.lookup('controller:post');
      let postController1Factory = appInstance.factoryFor('controller:post');
      assert.ok(postController1 instanceof PostController1, 'precond - lookup creates instance');
      assert.equal(PostController1, postController1Factory.class, 'precond - factoryFor().class matches');

      appInstance.unregister('controller:post');
      appInstance.register('controller:post', PostController2);

      let postController2 = appInstance.lookup('controller:post');
      let postController2Factory = appInstance.factoryFor('controller:post');
      assert.ok(postController2 instanceof PostController2, 'lookup creates instance');
      assert.equal(PostController2, postController2Factory.class, 'factoryFor().class matches');

      assert.notStrictEqual(postController1, postController2, 'lookup creates a brand new instance, because the previous one was reset');
    }

    ['@skip unregistering a factory clears caches with source of that factory'](assert) {
      assert.expect(1);

      appInstance = (0, _runloop.run)(() => _instance.default.create({ application }));

      let PostController1 = (0, _internalTestHelpers.factory)();
      let PostController2 = (0, _internalTestHelpers.factory)();

      appInstance.register('controller:post', PostController1);

      appInstance.lookup('controller:post');
      let postControllerLookupWithSource = appInstance.lookup('controller:post', {
        source: 'doesnt-even-matter'
      });

      appInstance.unregister('controller:post');
      appInstance.register('controller:post', PostController2);

      // The cache that is source-specific is not cleared
      assert.ok(postControllerLookupWithSource !== appInstance.lookup('controller:post', {
        source: 'doesnt-even-matter'
      }), 'lookup with source creates a new instance');
    }

    ['@test can build and boot a registered engine'](assert) {
      assert.expect(11);

      let ChatEngine = _engine.default.extend();
      let chatEngineInstance;

      application.register('engine:chat', ChatEngine);

      (0, _runloop.run)(() => {
        appInstance = _instance.default.create({ application });
        appInstance.setupRegistry();
        chatEngineInstance = appInstance.buildChildEngineInstance('chat');
      });

      return chatEngineInstance.boot().then(() => {
        assert.ok(true, 'boot successful');

        let registrations = ['route:basic', 'service:-routing', 'service:-glimmer-environment'];

        registrations.forEach(key => {
          assert.strictEqual(chatEngineInstance.resolveRegistration(key), appInstance.resolveRegistration(key), `Engine and parent app share registrations for '${key}'`);
        });

        let singletons = ['router:main', _container.privatize`-bucket-cache:main`, '-view-registry:main', '-environment:main', 'service:-document', 'event_dispatcher:main'];

        let env = appInstance.lookup('-environment:main');
        singletons.push(env.isInteractive ? 'renderer:-dom' : 'renderer:-inert');

        singletons.forEach(key => {
          assert.strictEqual(chatEngineInstance.lookup(key), appInstance.lookup(key), `Engine and parent app share singleton '${key}'`);
        });
      });
    }

    ['@test can build a registry via ApplicationInstance.setupRegistry() -- simulates ember-test-helpers'](assert) {
      let namespace = _runtime.Object.create({
        Resolver: { create: function () {} }
      });

      let registry = _application.default.buildRegistry(namespace);

      _instance.default.setupRegistry(registry);

      assert.equal(registry.resolve('service:-document'), document);
    }
  });
});
enifed('@ember/application/tests/application_test', ['ember/version', '@ember/-internals/environment', '@ember/-internals/metal', '@ember/debug', '@ember/application', '@ember/-internals/routing', '@ember/-internals/views', '@ember/controller', '@ember/-internals/runtime', '@ember/-internals/glimmer', '@ember/-internals/container', '@ember/polyfills', 'internal-test-helpers', '@ember/runloop'], function (_version, _environment, _metal, _debug, _application, _routing, _views, _controller, _runtime, _glimmer, _container, _polyfills, _internalTestHelpers, _runloop) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application, autobooting multiple apps', class extends _internalTestHelpers.ApplicationTestCase {
    get fixture() {
      return `
      <div id="one">
        <div id="one-child">HI</div>
      </div>
      <div id="two">HI</div>
    `;
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        rootElement: '#one',
        router: null,
        autoboot: true
      });
    }

    createSecondApplication(options) {
      let myOptions = (0, _polyfills.assign)(this.applicationOptions, options);
      return this.secondApp = _application.default.create(myOptions);
    }

    teardown() {
      super.teardown();

      if (this.secondApp) {
        this.runTask(() => this.secondApp.destroy());
      }
    }

    [`@test you can make a new application in a non-overlapping element`](assert) {
      let app = this.runTask(() => this.createSecondApplication({
        rootElement: '#two'
      }));

      this.runTask(() => app.destroy());
      assert.ok(true, 'should not raise');
    }

    [`@test you cannot make a new application that is a parent of an existing application`]() {
      expectAssertion(() => {
        this.runTask(() => this.createSecondApplication({
          rootElement: this.applicationOptions.rootElement
        }));
      });
    }

    [`@test you cannot make a new application that is a descendant of an existing application`]() {
      expectAssertion(() => {
        this.runTask(() => this.createSecondApplication({
          rootElement: '#one-child'
        }));
      });
    }

    [`@test you cannot make a new application that is a duplicate of an existing application`]() {
      expectAssertion(() => {
        this.runTask(() => this.createSecondApplication({
          rootElement: '#one'
        }));
      });
    }

    [`@test you cannot make two default applications without a rootElement error`]() {
      expectAssertion(() => {
        this.runTask(() => this.createSecondApplication());
      });
    }
  }); /*globals EmberDev */


  (0, _internalTestHelpers.moduleFor)('Application', class extends _internalTestHelpers.ApplicationTestCase {
    [`@test builds a registry`](assert) {
      let { application } = this;
      assert.strictEqual(application.resolveRegistration('application:main'), application, `application:main is registered`);
      assert.deepEqual(application.registeredOptionsForType('component'), { singleton: false }, `optionsForType 'component'`);
      assert.deepEqual(application.registeredOptionsForType('view'), { singleton: false }, `optionsForType 'view'`);

      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'controller:basic');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, '-view-registry:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'view', '_viewRegistry', '-view-registry:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'route', '_topLevelViewTemplate', 'template:-outlet');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'route:basic');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'event_dispatcher:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'router:main', 'namespace', 'application:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'view:-outlet', 'namespace', 'application:main');

      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'location:auto');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'location:hash');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'location:history');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'location:none');

      (0, _internalTestHelpers.verifyInjection)(assert, application, 'controller', 'target', 'router:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'controller', 'namespace', 'application:main');

      (0, _internalTestHelpers.verifyRegistration)(assert, application, _container.privatize`-bucket-cache:main`);
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'router', '_bucketCache', _container.privatize`-bucket-cache:main`);
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'route', '_bucketCache', _container.privatize`-bucket-cache:main`);

      (0, _internalTestHelpers.verifyInjection)(assert, application, 'route', '_router', 'router:main');

      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'component:-text-field');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'component:-text-area');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'component:-checkbox');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'component:link-to');

      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'service:-routing');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'service:-routing', 'router', 'router:main');

      // DEBUGGING
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'resolver-for-debugging:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'container-debug-adapter:main', 'resolver', 'resolver-for-debugging:main');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'data-adapter:main', 'containerDebugAdapter', 'container-debug-adapter:main');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'container-debug-adapter:main');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'component-lookup:main');

      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'service:-glimmer-environment');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'service:-dom-changes');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'service:-dom-tree-construction');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'service:-glimmer-environment', 'appendOperations', 'service:-dom-tree-construction');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'service:-glimmer-environment', 'updateOperations', 'service:-dom-changes');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'renderer', 'env', 'service:-glimmer-environment');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'view:-outlet');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'renderer:-dom');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'renderer:-inert');
      (0, _internalTestHelpers.verifyRegistration)(assert, application, _container.privatize`template:components/-default`);
      (0, _internalTestHelpers.verifyRegistration)(assert, application, 'template:-outlet');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'view:-outlet', 'template', 'template:-outlet');
      (0, _internalTestHelpers.verifyInjection)(assert, application, 'template', 'compiler', _container.privatize`template-compiler:main`);

      assert.deepEqual(application.registeredOptionsForType('helper'), { instantiate: false }, `optionsForType 'helper'`);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application, default resolver with autoboot', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    constructor() {
      super(...arguments);
      this.originalLookup = _environment.context.lookup;
    }

    teardown() {
      _environment.context.lookup = this.originalLookup;
      super.teardown();
      (0, _glimmer.setTemplates)({});
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        autoboot: true
      });
    }

    [`@test acts like a namespace`](assert) {
      this.application = this.runTask(() => this.createApplication());
      let Foo = this.application.Foo = _runtime.Object.extend();
      assert.equal(Foo.toString(), 'TestApp.Foo', 'Classes pick up their parent namespace');
    }

    [`@test can specify custom router`](assert) {
      let MyRouter = _routing.Router.extend();
      this.runTask(() => {
        this.createApplication();
        this.application.Router = MyRouter;
      });

      assert.ok(this.application.__deprecatedInstance__.lookup('router:main') instanceof MyRouter, 'application resolved the correct router');
    }

    [`@test Minimal Application initialized with just an application template`]() {
      this.setupFixture('<script type="text/x-handlebars">Hello World</script>');
      this.runTask(() => this.createApplication());
      this.assertInnerHTML('Hello World');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application, autobooting', class extends _internalTestHelpers.AutobootApplicationTestCase {
    constructor() {
      super(...arguments);
      this.originalLogVersion = _environment.ENV.LOG_VERSION;
      this.originalDebug = (0, _debug.getDebugFunction)('debug');
      this.originalWarn = (0, _debug.getDebugFunction)('warn');
    }

    teardown() {
      (0, _debug.setDebugFunction)('warn', this.originalWarn);
      (0, _debug.setDebugFunction)('debug', this.originalDebug);
      _environment.ENV.LOG_VERSION = this.originalLogVersion;
      super.teardown();
    }

    [`@test initialized application goes to initial route`]() {
      this.runTask(() => {
        this.createApplication();
        this.addTemplate('application', '{{outlet}}');
        this.addTemplate('index', '<h1>Hi from index</h1>');
      });

      this.assertText('Hi from index');
    }

    [`@test ready hook is called before routing begins`](assert) {
      assert.expect(2);

      this.runTask(() => {
        function registerRoute(application, name, callback) {
          let route = _routing.Route.extend({
            activate: callback
          });

          application.register('route:' + name, route);
        }

        let MyApplication = _application.default.extend({
          ready() {
            registerRoute(this, 'index', () => {
              assert.ok(true, 'last-minute route is activated');
            });
          }
        });

        let app = this.createApplication({}, MyApplication);

        registerRoute(app, 'application', () => assert.ok(true, 'normal route is activated'));
      });
    }

    [`@test initialize application via initialize call`](assert) {
      this.runTask(() => this.createApplication());
      // This is not a public way to access the container; we just
      // need to make some assertions about the created router
      let router = this.applicationInstance.lookup('router:main');
      assert.equal(router instanceof _routing.Router, true, 'Router was set from initialize call');
      assert.equal(router.location instanceof _routing.NoneLocation, true, 'Location was set from location implementation name');
    }

    [`@test initialize application with stateManager via initialize call from Router class`](assert) {
      this.runTask(() => {
        this.createApplication();
        this.addTemplate('application', '<h1>Hello!</h1>');
      });
      // This is not a public way to access the container; we just
      // need to make some assertions about the created router
      let router = this.application.__deprecatedInstance__.lookup('router:main');
      assert.equal(router instanceof _routing.Router, true, 'Router was set from initialize call');
      this.assertText('Hello!');
    }

    [`@test Application Controller backs the appplication template`]() {
      this.runTask(() => {
        this.createApplication();
        this.addTemplate('application', '<h1>{{greeting}}</h1>');
        this.add('controller:application', _controller.default.extend({
          greeting: 'Hello!'
        }));
      });
      this.assertText('Hello!');
    }

    [`@test enable log of libraries with an ENV var`](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      let messages = [];

      _environment.ENV.LOG_VERSION = true;

      (0, _debug.setDebugFunction)('debug', message => messages.push(message));

      _metal.libraries.register('my-lib', '2.0.0a');

      this.runTask(() => this.createApplication());

      assert.equal(messages[1], 'Ember  : ' + _version.default);
      if (_views.jQueryDisabled) {
        assert.equal(messages[2], 'my-lib : ' + '2.0.0a');
      } else {
        assert.equal(messages[2], 'jQuery : ' + (0, _views.jQuery)().jquery);
        assert.equal(messages[3], 'my-lib : ' + '2.0.0a');
      }

      _metal.libraries.deRegister('my-lib');
    }

    [`@test disable log of version of libraries with an ENV var`](assert) {
      let logged = false;

      _environment.ENV.LOG_VERSION = false;

      (0, _debug.setDebugFunction)('debug', () => logged = true);

      this.runTask(() => this.createApplication());

      assert.ok(!logged, 'library version logging skipped');
    }

    [`@test can resolve custom router`](assert) {
      let CustomRouter = _routing.Router.extend();

      this.runTask(() => {
        this.createApplication();
        this.add('router:main', CustomRouter);
      });

      assert.ok(this.application.__deprecatedInstance__.lookup('router:main') instanceof CustomRouter, 'application resolved the correct router');
    }

    [`@test does not leak itself in onLoad._loaded`](assert) {
      assert.equal(_application._loaded.application, undefined);
      this.runTask(() => this.createApplication());
      assert.equal(_application._loaded.application, this.application);
      this.runTask(() => this.application.destroy());
      assert.equal(_application._loaded.application, undefined);
    }

    [`@test can build a registry via Application.buildRegistry() --- simulates ember-test-helpers`](assert) {
      let namespace = _runtime.Object.create({
        Resolver: { create: function () {} }
      });

      let registry = _application.default.buildRegistry(namespace);

      assert.equal(registry.resolve('application:main'), namespace);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application#buildRegistry', class extends _internalTestHelpers.AbstractTestCase {
    [`@test can build a registry via Application.buildRegistry() --- simulates ember-test-helpers`](assert) {
      let namespace = _runtime.Object.create({
        Resolver: { create() {} }
      });

      let registry = _application.default.buildRegistry(namespace);

      assert.equal(registry.resolve('application:main'), namespace);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application - instance tracking', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test tracks built instance'](assert) {
      let instance = this.application.buildInstance();
      (0, _runloop.run)(() => {
        this.application.destroy();
      });

      assert.ok(instance.isDestroyed, 'instance was destroyed');
    }

    ['@test tracks built instances'](assert) {
      let instanceA = this.application.buildInstance();
      let instanceB = this.application.buildInstance();
      (0, _runloop.run)(() => {
        this.application.destroy();
      });

      assert.ok(instanceA.isDestroyed, 'instanceA was destroyed');
      assert.ok(instanceB.isDestroyed, 'instanceB was destroyed');
    }
  });
});
enifed('@ember/application/tests/bootstrap-test', ['@ember/polyfills', 'internal-test-helpers'], function (_polyfills, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application with default resolver and autoboot', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    get fixture() {
      return `
      <div id="app"></div>

      <script type="text/x-handlebars">Hello {{outlet}}</script>
      <script type="text/x-handlebars" id="index">World!</script>
    `;
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        autoboot: true,
        rootElement: '#app'
      });
    }

    ['@test templates in script tags are extracted at application creation'](assert) {
      this.runTask(() => this.createApplication());
      assert.equal(document.getElementById('app').textContent, 'Hello World!');
    }
  });
});
enifed('@ember/application/tests/dependency_injection/custom_resolver_test', ['@ember/application/globals-resolver', '@ember/polyfills', 'internal-test-helpers'], function (_globalsResolver, _polyfills, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application with extended default resolver and autoboot', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    get applicationOptions() {
      let applicationTemplate = this.compile(`<h1>Fallback</h1>`);

      let Resolver = _globalsResolver.default.extend({
        resolveTemplate(resolvable) {
          if (resolvable.fullNameWithoutType === 'application') {
            return applicationTemplate;
          } else {
            return this._super(resolvable);
          }
        }
      });

      return (0, _polyfills.assign)(super.applicationOptions, {
        Resolver,
        autoboot: true
      });
    }

    [`@test a resolver can be supplied to application`]() {
      this.runTask(() => this.createApplication());
      this.assertText('Fallback');
    }
  });
});
enifed('@ember/application/tests/dependency_injection/default_resolver_test', ['internal-test-helpers', '@ember/-internals/environment', '@ember/controller', '@ember/service', '@ember/-internals/runtime', '@ember/-internals/routing', '@ember/-internals/glimmer', '@ember/debug'], function (_internalTestHelpers, _environment, _controller, _service, _runtime, _routing, _glimmer, _debug) {
  'use strict';

  /* globals EmberDev */
  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - Integration - default resolver', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    beforeEach() {
      this.runTask(() => this.createApplication());
      return this.visit('/');
    }

    get privateRegistry() {
      return this.application.__registry__;
    }

    /*
    * This first batch of tests are integration tests against the public
    * applicationInstance API.
    */

    [`@test the default resolver looks up templates in Ember.TEMPLATES`](assert) {
      let fooTemplate = this.addTemplate('foo', `foo template`);
      let fooBarTemplate = this.addTemplate('fooBar', `fooBar template`);
      let fooBarBazTemplate = this.addTemplate('fooBar/baz', `fooBar/baz template`);

      assert.equal(this.applicationInstance.factoryFor('template:foo').class, fooTemplate, 'resolves template:foo');
      assert.equal(this.applicationInstance.factoryFor('template:fooBar').class, fooBarTemplate, 'resolves template:foo_bar');
      assert.equal(this.applicationInstance.factoryFor('template:fooBar.baz').class, fooBarBazTemplate, 'resolves template:foo_bar.baz');
    }

    [`@test the default resolver looks up basic name as no prefix`](assert) {
      let instance = this.applicationInstance.lookup('controller:basic');
      assert.ok(_controller.default.detect(instance), 'locator looks up correct controller');
    }

    [`@test the default resolver looks up arbitrary types on the namespace`](assert) {
      let Class = this.application.FooManager = _runtime.Object.extend();
      let resolvedClass = this.application.resolveRegistration('manager:foo');
      assert.equal(Class, resolvedClass, 'looks up FooManager on application');
    }

    [`@test the default resolver resolves models on the namespace`](assert) {
      let Class = this.application.Post = _runtime.Object.extend();
      let factoryClass = this.applicationInstance.factoryFor('model:post').class;
      assert.equal(Class, factoryClass, 'looks up Post model on application');
    }

    [`@test the default resolver resolves *:main on the namespace`](assert) {
      let Class = this.application.FooBar = _runtime.Object.extend();
      let factoryClass = this.applicationInstance.factoryFor('foo-bar:main').class;
      assert.equal(Class, factoryClass, 'looks up FooBar type without name on application');
    }

    [`@test the default resolver resolves container-registered helpers`](assert) {
      let shorthandHelper = (0, _glimmer.helper)(() => {});
      let helper = _glimmer.Helper.extend();

      this.application.register('helper:shorthand', shorthandHelper);
      this.application.register('helper:complete', helper);

      let lookedUpShorthandHelper = this.applicationInstance.factoryFor('helper:shorthand').class;

      assert.ok(lookedUpShorthandHelper.isHelperFactory, 'shorthand helper isHelper');

      let lookedUpHelper = this.applicationInstance.factoryFor('helper:complete').class;

      assert.ok(lookedUpHelper.isHelperFactory, 'complete helper is factory');
      assert.ok(helper.detect(lookedUpHelper), 'looked up complete helper');
    }

    [`@test the default resolver resolves container-registered helpers via lookupFor`](assert) {
      let shorthandHelper = (0, _glimmer.helper)(() => {});
      let helper = _glimmer.Helper.extend();

      this.application.register('helper:shorthand', shorthandHelper);
      this.application.register('helper:complete', helper);

      let lookedUpShorthandHelper = this.applicationInstance.factoryFor('helper:shorthand').class;

      assert.ok(lookedUpShorthandHelper.isHelperFactory, 'shorthand helper isHelper');

      let lookedUpHelper = this.applicationInstance.factoryFor('helper:complete').class;

      assert.ok(lookedUpHelper.isHelperFactory, 'complete helper is factory');
      assert.ok(helper.detect(lookedUpHelper), 'looked up complete helper');
    }

    [`@test the default resolver resolves helpers on the namespace`](assert) {
      let ShorthandHelper = (0, _glimmer.helper)(() => {});
      let CompleteHelper = _glimmer.Helper.extend();

      this.application.ShorthandHelper = ShorthandHelper;
      this.application.CompleteHelper = CompleteHelper;

      let resolvedShorthand = this.application.resolveRegistration('helper:shorthand');
      let resolvedComplete = this.application.resolveRegistration('helper:complete');

      assert.equal(resolvedShorthand, ShorthandHelper, 'resolve fetches the shorthand helper factory');
      assert.equal(resolvedComplete, CompleteHelper, 'resolve fetches the complete helper factory');
    }

    [`@test the default resolver resolves to the same instance, no matter the notation `](assert) {
      this.application.NestedPostController = _controller.default.extend({});

      assert.equal(this.applicationInstance.lookup('controller:nested-post'), this.applicationInstance.lookup('controller:nested_post'), 'looks up NestedPost controller on application');
    }

    [`@test the default resolver throws an error if the fullName to resolve is invalid`]() {
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration(undefined);
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration(null);
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration('');
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration('');
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration(':');
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration('model');
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration('model:');
      }, /fullName must be a proper full name/);
      expectAssertion(() => {
        this.applicationInstance.resolveRegistration(':type');
      }, /fullName must be a proper full name/);
    }

    /*
    * The following are integration tests against the private registry API.
    */

    [`@test lookup description`](assert) {
      this.application.toString = () => 'App';

      assert.equal(this.privateRegistry.describe('controller:foo'), 'App.FooController', 'Type gets appended at the end');
      assert.equal(this.privateRegistry.describe('controller:foo.bar'), 'App.FooBarController', 'dots are removed');
      assert.equal(this.privateRegistry.describe('model:foo'), 'App.Foo', "models don't get appended at the end");
    }

    [`@test assertion for routes without isRouteFactory property`]() {
      this.application.FooRoute = _glimmer.Component.extend();

      expectAssertion(() => {
        this.privateRegistry.resolve(`route:foo`);
      }, /to resolve to an Ember.Route/, 'Should assert');
    }

    [`@test no assertion for routes that extend from Route`](assert) {
      assert.expect(0);
      this.application.FooRoute = _routing.Route.extend();
      this.privateRegistry.resolve(`route:foo`);
    }

    [`@test deprecation warning for service factories without isServiceFactory property`]() {
      expectAssertion(() => {
        this.application.FooService = _runtime.Object.extend();
        this.privateRegistry.resolve('service:foo');
      }, /Expected service:foo to resolve to an Ember.Service but instead it was TestApp\.FooService\./);
    }

    [`@test no deprecation warning for service factories that extend from Service`](assert) {
      assert.expect(0);
      this.application.FooService = _service.default.extend();
      this.privateRegistry.resolve('service:foo');
    }

    [`@test deprecation warning for component factories without isComponentFactory property`]() {
      expectAssertion(() => {
        this.application.FooComponent = _runtime.Object.extend();
        this.privateRegistry.resolve('component:foo');
      }, /Expected component:foo to resolve to an Ember\.Component but instead it was TestApp\.FooComponent\./);
    }

    [`@test no deprecation warning for component factories that extend from Component`]() {
      expectNoDeprecation();
      this.application.FooView = _glimmer.Component.extend();
      this.privateRegistry.resolve('component:foo');
    }

    [`@test knownForType returns each item for a given type found`](assert) {
      this.application.FooBarHelper = 'foo';
      this.application.BazQuxHelper = 'bar';

      let found = this.privateRegistry.resolver.knownForType('helper');

      assert.deepEqual(found, {
        'helper:foo-bar': true,
        'helper:baz-qux': true
      });
    }

    [`@test knownForType is not required to be present on the resolver`](assert) {
      delete this.privateRegistry.resolver.knownForType;

      this.privateRegistry.resolver.knownForType('helper', () => {});

      assert.ok(true, 'does not error');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - Integration - default resolver w/ other namespace', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    beforeEach() {
      this.UserInterface = _environment.context.lookup.UserInterface = _runtime.Namespace.create();
      this.runTask(() => this.createApplication());
      return this.visit('/');
    }

    teardown() {
      let UserInterfaceNamespace = _runtime.Namespace.NAMESPACES_BY_ID['UserInterface'];
      if (UserInterfaceNamespace) {
        this.runTask(() => {
          UserInterfaceNamespace.destroy();
        });
      }
      super.teardown();
    }

    [`@test the default resolver can look things up in other namespaces`](assert) {
      this.UserInterface.NavigationController = _controller.default.extend();

      let nav = this.applicationInstance.lookup('controller:userInterface/navigation');

      assert.ok(nav instanceof this.UserInterface.NavigationController, 'the result should be an instance of the specified class');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - Integration - default resolver', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    constructor() {
      super();
      this._originalLookup = _environment.context.lookup;
      this._originalInfo = (0, _debug.getDebugFunction)('info');
    }

    beforeEach() {
      this.runTask(() => this.createApplication());
      return this.visit('/');
    }

    teardown() {
      (0, _debug.setDebugFunction)('info', this._originalInfo);
      _environment.context.lookup = this._originalLookup;
      super.teardown();
    }

    [`@test the default resolver logs hits if 'LOG_RESOLVER' is set`](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      assert.expect(3);

      this.application.LOG_RESOLVER = true;
      this.application.ScoobyDoo = _runtime.Object.extend();
      this.application.toString = () => 'App';

      (0, _debug.setDebugFunction)('info', function (symbol, name, padding, lookupDescription) {
        assert.equal(symbol, '[✓]', 'proper symbol is printed when a module is found');
        assert.equal(name, 'doo:scooby', 'proper lookup value is logged');
        assert.equal(lookupDescription, 'App.ScoobyDoo');
      });

      this.applicationInstance.resolveRegistration('doo:scooby');
    }

    [`@test the default resolver logs misses if 'LOG_RESOLVER' is set`](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      assert.expect(3);

      this.application.LOG_RESOLVER = true;
      this.application.toString = () => 'App';

      (0, _debug.setDebugFunction)('info', function (symbol, name, padding, lookupDescription) {
        assert.equal(symbol, '[ ]', 'proper symbol is printed when a module is not found');
        assert.equal(name, 'doo:scooby', 'proper lookup value is logged');
        assert.equal(lookupDescription, 'App.ScoobyDoo');
      });

      this.applicationInstance.resolveRegistration('doo:scooby');
    }

    [`@test doesn't log without LOG_RESOLVER`](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      let infoCount = 0;

      this.application.ScoobyDoo = _runtime.Object.extend();

      (0, _debug.setDebugFunction)('info', () => infoCount = infoCount + 1);

      this.applicationInstance.resolveRegistration('doo:scooby');
      this.applicationInstance.resolveRegistration('doo:scrappy');
      assert.equal(infoCount, 0, 'console.info should not be called if LOG_RESOLVER is not set');
    }
  });
});
enifed('@ember/application/tests/dependency_injection/normalization_test', ['@ember/runloop', '@ember/application', 'internal-test-helpers'], function (_runloop, _application, _internalTestHelpers) {
  'use strict';

  let application, registry;

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - normalize', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      application = (0, _runloop.run)(_application.default, 'create');
      registry = application.__registry__;
    }

    teardown() {
      super.teardown();
      (0, _runloop.run)(application, 'destroy');
      application = undefined;
      registry = undefined;
    }

    ['@test normalization'](assert) {
      assert.ok(registry.normalize, 'registry#normalize is present');

      assert.equal(registry.normalize('foo:bar'), 'foo:bar');

      assert.equal(registry.normalize('controller:posts'), 'controller:posts');
      assert.equal(registry.normalize('controller:posts_index'), 'controller:postsIndex');
      assert.equal(registry.normalize('controller:posts.index'), 'controller:postsIndex');
      assert.equal(registry.normalize('controller:posts-index'), 'controller:postsIndex');
      assert.equal(registry.normalize('controller:posts.post.index'), 'controller:postsPostIndex');
      assert.equal(registry.normalize('controller:posts_post.index'), 'controller:postsPostIndex');
      assert.equal(registry.normalize('controller:posts.post_index'), 'controller:postsPostIndex');
      assert.equal(registry.normalize('controller:posts.post-index'), 'controller:postsPostIndex');
      assert.equal(registry.normalize('controller:postsIndex'), 'controller:postsIndex');
      assert.equal(registry.normalize('controller:blogPosts.index'), 'controller:blogPostsIndex');
      assert.equal(registry.normalize('controller:blog/posts.index'), 'controller:blog/postsIndex');
      assert.equal(registry.normalize('controller:blog/posts-index'), 'controller:blog/postsIndex');
      assert.equal(registry.normalize('controller:blog/posts.post.index'), 'controller:blog/postsPostIndex');
      assert.equal(registry.normalize('controller:blog/posts_post.index'), 'controller:blog/postsPostIndex');
      assert.equal(registry.normalize('controller:blog/posts_post-index'), 'controller:blog/postsPostIndex');

      assert.equal(registry.normalize('template:blog/posts_index'), 'template:blog/posts_index');
    }

    ['@test normalization is indempotent'](assert) {
      let examples = ['controller:posts', 'controller:posts.post.index', 'controller:blog/posts.post_index', 'template:foo_bar'];

      examples.forEach(example => {
        assert.equal(registry.normalize(registry.normalize(example)), registry.normalize(example));
      });
    }
  });
});
enifed('@ember/application/tests/dependency_injection/to_string_test', ['@ember/polyfills', '@ember/-internals/utils', '@ember/-internals/runtime', 'internal-test-helpers'], function (_polyfills, _utils, _runtime, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - DefaultResolver#toString', class extends _internalTestHelpers.DefaultResolverApplicationTestCase {
    constructor() {
      super();
      this.runTask(() => this.createApplication());
      this.application.Post = _runtime.Object.extend();
    }

    beforeEach() {
      return this.visit('/');
    }

    ['@test factories'](assert) {
      let PostFactory = this.applicationInstance.factoryFor('model:post').class;
      assert.equal(PostFactory.toString(), 'TestApp.Post', 'expecting the model to be post');
    }

    ['@test instances'](assert) {
      let post = this.applicationInstance.lookup('model:post');
      let guid = (0, _utils.guidFor)(post);

      assert.equal(post.toString(), '<TestApp.Post:' + guid + '>', 'expecting the model to be post');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection - Resolver#toString', class extends _internalTestHelpers.ApplicationTestCase {
    beforeEach() {
      return this.visit('/');
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        Resolver: class extends _internalTestHelpers.ModuleBasedTestResolver {
          makeToString(_, fullName) {
            return fullName;
          }
        }
      });
    }

    ['@test toString called on a resolver'](assert) {
      this.add('model:peter', _runtime.Object.extend());

      let peter = this.applicationInstance.lookup('model:peter');
      let guid = (0, _utils.guidFor)(peter);
      assert.equal(peter.toString(), `<model:peter:${guid}>`, 'expecting the supermodel to be peter');
    }
  });
});
enifed('@ember/application/tests/dependency_injection_test', ['@ember/-internals/environment', '@ember/runloop', '@ember/-internals/runtime', '@ember/application', 'internal-test-helpers'], function (_environment, _runloop, _runtime, _application, _internalTestHelpers) {
  'use strict';

  let originalLookup = _environment.context.lookup;
  let registry, locator, application;

  (0, _internalTestHelpers.moduleFor)('Application Dependency Injection', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      application = (0, _runloop.run)(_application.default, 'create');

      application.Person = _runtime.Object.extend({});
      application.Orange = _runtime.Object.extend({});
      application.Email = _runtime.Object.extend({});
      application.User = _runtime.Object.extend({});
      application.PostIndexController = _runtime.Object.extend({});

      application.register('model:person', application.Person, {
        singleton: false
      });
      application.register('model:user', application.User, {
        singleton: false
      });
      application.register('fruit:favorite', application.Orange);
      application.register('communication:main', application.Email, {
        singleton: false
      });
      application.register('controller:postIndex', application.PostIndexController, {
        singleton: true
      });

      registry = application.__registry__;
      locator = application.__container__;

      _environment.context.lookup = {};
    }

    teardown() {
      super.teardown();
      (0, _runloop.run)(application, 'destroy');
      registry = application = locator = null;
      _environment.context.lookup = originalLookup;
    }

    ['@test container lookup is normalized'](assert) {
      let dotNotationController = locator.lookup('controller:post.index');
      let camelCaseController = locator.lookup('controller:postIndex');

      assert.ok(dotNotationController instanceof application.PostIndexController);
      assert.ok(camelCaseController instanceof application.PostIndexController);

      assert.equal(dotNotationController, camelCaseController);
    }

    ['@test registered entities can be looked up later'](assert) {
      assert.equal(registry.resolve('model:person'), application.Person);
      assert.equal(registry.resolve('model:user'), application.User);
      assert.equal(registry.resolve('fruit:favorite'), application.Orange);
      assert.equal(registry.resolve('communication:main'), application.Email);
      assert.equal(registry.resolve('controller:postIndex'), application.PostIndexController);

      assert.equal(locator.lookup('fruit:favorite'), locator.lookup('fruit:favorite'), 'singleton lookup worked');
      assert.ok(locator.lookup('model:user') !== locator.lookup('model:user'), 'non-singleton lookup worked');
    }

    ['@test injections'](assert) {
      application.inject('model', 'fruit', 'fruit:favorite');
      application.inject('model:user', 'communication', 'communication:main');

      let user = locator.lookup('model:user');
      let person = locator.lookup('model:person');
      let fruit = locator.lookup('fruit:favorite');

      assert.equal(user.get('fruit'), fruit);
      assert.equal(person.get('fruit'), fruit);

      assert.ok(application.Email.detectInstance(user.get('communication')));
    }
  });
});
enifed('@ember/application/tests/initializers_test', ['@ember/polyfills', 'internal-test-helpers', '@ember/application'], function (_polyfills, _internalTestHelpers, _application) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application initializers', class extends _internalTestHelpers.AutobootApplicationTestCase {
    get fixture() {
      return `<div id="one">ONE</div>
      <div id="two">TWO</div>
    `;
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        rootElement: '#one'
      });
    }

    createSecondApplication(options, MyApplication = _application.default) {
      let myOptions = (0, _polyfills.assign)(this.applicationOptions, {
        rootElement: '#two'
      }, options);
      let secondApp = this.secondApp = MyApplication.create(myOptions);
      return secondApp;
    }

    teardown() {
      super.teardown();

      if (this.secondApp) {
        this.runTask(() => this.secondApp.destroy());
      }
    }

    [`@test initializers require proper 'name' and 'initialize' properties`]() {
      let MyApplication = _application.default.extend();

      expectAssertion(() => {
        MyApplication.initializer({ name: 'initializer' });
      });

      expectAssertion(() => {
        MyApplication.initializer({ initialize() {} });
      });
    }

    [`@test initializers that throw errors cause the boot promise to reject with the error`](assert) {
      assert.expect(2);

      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'initializer',
        initialize() {
          throw new Error('boot failure');
        }
      });

      this.runTask(() => {
        this.createApplication({
          autoboot: false
        }, MyApplication);
      });

      let app = this.application;

      try {
        this.runTask(() => {
          app.boot().then(() => {
            assert.ok(false, 'The boot promise should not resolve when there is a boot error');
          }, error => {
            assert.ok(error instanceof Error, 'The boot promise should reject with an error');
            assert.equal(error.message, 'boot failure');
          });
        });
      } catch (error) {
        assert.ok(false, 'The boot method should not throw');
        throw error;
      }
    }

    [`@test initializers are passed an App`](assert) {
      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'initializer',
        initialize(App) {
          assert.ok(App instanceof _application.default, 'initialize is passed an Application');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }

    [`@test initializers can be registered in a specified order`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'fourth',
        after: 'third',
        initialize() {
          order.push('fourth');
        }
      });

      MyApplication.initializer({
        name: 'second',
        after: 'first',
        before: 'third',
        initialize() {
          order.push('second');
        }
      });

      MyApplication.initializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyApplication.initializer({
        name: 'first',
        before: 'second',
        initialize() {
          order.push('first');
        }
      });

      MyApplication.initializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyApplication.initializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    [`@test initializers can be registered in a specified order as an array`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyApplication.initializer({
        name: 'second',
        after: 'first',
        before: ['third', 'fourth'],
        initialize() {
          order.push('second');
        }
      });

      MyApplication.initializer({
        name: 'fourth',
        after: ['second', 'third'],
        initialize() {
          order.push('fourth');
        }
      });

      MyApplication.initializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyApplication.initializer({
        name: 'first',
        before: ['second'],
        initialize() {
          order.push('first');
        }
      });

      MyApplication.initializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    [`@test initializers can have multiple dependencies`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();
      let a = {
        name: 'a',
        before: 'b',
        initialize() {
          order.push('a');
        }
      };
      let b = {
        name: 'b',
        initialize() {
          order.push('b');
        }
      };
      let c = {
        name: 'c',
        after: 'b',
        initialize() {
          order.push('c');
        }
      };
      let afterB = {
        name: 'after b',
        after: 'b',
        initialize() {
          order.push('after b');
        }
      };
      let afterC = {
        name: 'after c',
        after: 'c',
        initialize() {
          order.push('after c');
        }
      };

      MyApplication.initializer(b);
      MyApplication.initializer(a);
      MyApplication.initializer(afterC);
      MyApplication.initializer(afterB);
      MyApplication.initializer(c);

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.ok(order.indexOf(a.name) < order.indexOf(b.name), 'a < b');
      assert.ok(order.indexOf(b.name) < order.indexOf(c.name), 'b < c');
      assert.ok(order.indexOf(b.name) < order.indexOf(afterB.name), 'b < afterB');
      assert.ok(order.indexOf(c.name) < order.indexOf(afterC.name), 'c < afterC');
    }

    [`@test initializers set on Application subclasses are not shared between apps`](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstApp = _application.default.extend();

      FirstApp.initializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondApp = _application.default.extend();

      SecondApp.initializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      this.runTask(() => this.createApplication({}, FirstApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run');
      assert.equal(secondInitializerRunCount, 0, 'first initializer only was run');

      this.runTask(() => this.createSecondApplication({}, SecondApp));

      assert.equal(firstInitializerRunCount, 1, 'second initializer only was run');
      assert.equal(secondInitializerRunCount, 1, 'second initializer only was run');
    }

    [`@test initializers are concatenated`](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstApp = _application.default.extend();

      FirstApp.initializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondApp = FirstApp.extend();
      SecondApp.initializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      this.runTask(() => this.createApplication({}, FirstApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run when base class created');
      assert.equal(secondInitializerRunCount, 0, 'first initializer only was run when base class created');

      firstInitializerRunCount = 0;
      this.runTask(() => this.createSecondApplication({}, SecondApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer was run when subclass created');
      assert.equal(secondInitializerRunCount, 1, 'second initializers was run when subclass created');
    }

    [`@test initializers are per-app`](assert) {
      assert.expect(2);

      let FirstApp = _application.default.extend();

      FirstApp.initializer({
        name: 'abc',
        initialize() {}
      });

      expectAssertion(() => {
        FirstApp.initializer({
          name: 'abc',
          initialize() {}
        });
      });

      let SecondApp = _application.default.extend();
      SecondApp.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      assert.ok(true, 'Two apps can have initializers named the same.');
    }

    [`@test initializers are executed in their own context`](assert) {
      assert.expect(1);
      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'coolInitializer',
        myProperty: 'cool',
        initialize() {
          assert.equal(this.myProperty, 'cool', 'should have access to its own context');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }
  });
});
enifed('@ember/application/tests/instance_initializers_test', ['@ember/polyfills', 'internal-test-helpers', '@ember/application/instance', '@ember/application'], function (_polyfills, _internalTestHelpers, _instance, _application) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application instance initializers', class extends _internalTestHelpers.AutobootApplicationTestCase {
    get fixture() {
      return `<div id="one">ONE</div>
      <div id="two">TWO</div>
    `;
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        rootElement: '#one'
      });
    }

    createSecondApplication(options, MyApplication = _application.default) {
      let myOptions = (0, _polyfills.assign)(this.applicationOptions, {
        rootElement: '#two'
      }, options);
      let secondApp = this.secondApp = MyApplication.create(myOptions);
      return secondApp;
    }

    teardown() {
      super.teardown();

      if (this.secondApp) {
        this.runTask(() => this.secondApp.destroy());
      }
    }

    [`@test initializers require proper 'name' and 'initialize' properties`]() {
      let MyApplication = _application.default.extend();

      expectAssertion(() => {
        MyApplication.instanceInitializer({ name: 'initializer' });
      });

      expectAssertion(() => {
        MyApplication.instanceInitializer({ initialize() {} });
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }

    [`@test initializers are passed an app instance`](assert) {
      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'initializer',
        initialize(instance) {
          assert.ok(instance instanceof _instance.default, 'initialize is passed an application instance');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }

    [`@test initializers can be registered in a specified order`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'fourth',
        after: 'third',
        initialize() {
          order.push('fourth');
        }
      });

      MyApplication.instanceInitializer({
        name: 'second',
        after: 'first',
        before: 'third',
        initialize() {
          order.push('second');
        }
      });

      MyApplication.instanceInitializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyApplication.instanceInitializer({
        name: 'first',
        before: 'second',
        initialize() {
          order.push('first');
        }
      });

      MyApplication.instanceInitializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyApplication.instanceInitializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    [`@test initializers can be registered in a specified order as an array`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyApplication.instanceInitializer({
        name: 'second',
        after: 'first',
        before: ['third', 'fourth'],
        initialize() {
          order.push('second');
        }
      });

      MyApplication.instanceInitializer({
        name: 'fourth',
        after: ['second', 'third'],
        initialize() {
          order.push('fourth');
        }
      });

      MyApplication.instanceInitializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyApplication.instanceInitializer({
        name: 'first',
        before: ['second'],
        initialize() {
          order.push('first');
        }
      });

      MyApplication.instanceInitializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    [`@test initializers can have multiple dependencies`](assert) {
      let order = [];
      let MyApplication = _application.default.extend();
      let a = {
        name: 'a',
        before: 'b',
        initialize() {
          order.push('a');
        }
      };
      let b = {
        name: 'b',
        initialize() {
          order.push('b');
        }
      };
      let c = {
        name: 'c',
        after: 'b',
        initialize() {
          order.push('c');
        }
      };
      let afterB = {
        name: 'after b',
        after: 'b',
        initialize() {
          order.push('after b');
        }
      };
      let afterC = {
        name: 'after c',
        after: 'c',
        initialize() {
          order.push('after c');
        }
      };

      MyApplication.instanceInitializer(b);
      MyApplication.instanceInitializer(a);
      MyApplication.instanceInitializer(afterC);
      MyApplication.instanceInitializer(afterB);
      MyApplication.instanceInitializer(c);

      this.runTask(() => this.createApplication({}, MyApplication));

      assert.ok(order.indexOf(a.name) < order.indexOf(b.name), 'a < b');
      assert.ok(order.indexOf(b.name) < order.indexOf(c.name), 'b < c');
      assert.ok(order.indexOf(b.name) < order.indexOf(afterB.name), 'b < afterB');
      assert.ok(order.indexOf(c.name) < order.indexOf(afterC.name), 'c < afterC');
    }

    [`@test initializers set on Application subclasses should not be shared between apps`](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstApp = _application.default.extend();

      FirstApp.instanceInitializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondApp = _application.default.extend();
      SecondApp.instanceInitializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      this.runTask(() => this.createApplication({}, FirstApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run');
      assert.equal(secondInitializerRunCount, 0, 'first initializer only was run');

      this.runTask(() => this.createSecondApplication({}, SecondApp));

      assert.equal(firstInitializerRunCount, 1, 'second initializer only was run');
      assert.equal(secondInitializerRunCount, 1, 'second initializer only was run');
    }

    [`@test initializers are concatenated`](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstApp = _application.default.extend();

      FirstApp.instanceInitializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondApp = FirstApp.extend();
      SecondApp.instanceInitializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      this.runTask(() => this.createApplication({}, FirstApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run when base class created');
      assert.equal(secondInitializerRunCount, 0, 'first initializer only was run when base class created');

      firstInitializerRunCount = 0;
      this.runTask(() => this.createSecondApplication({}, SecondApp));

      assert.equal(firstInitializerRunCount, 1, 'first initializer was run when subclass created');
      assert.equal(secondInitializerRunCount, 1, 'second initializers was run when subclass created');
    }

    [`@test initializers are per-app`](assert) {
      assert.expect(2);

      let FirstApp = _application.default.extend();
      FirstApp.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      expectAssertion(function () {
        FirstApp.instanceInitializer({
          name: 'abc',
          initialize() {}
        });
      });

      this.runTask(() => this.createApplication({}, FirstApp));

      let SecondApp = _application.default.extend();
      SecondApp.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      this.runTask(() => this.createSecondApplication({}, SecondApp));

      assert.ok(true, 'Two apps can have initializers named the same.');
    }

    [`@test initializers are run before ready hook`](assert) {
      assert.expect(2);

      let MyApplication = _application.default.extend({
        ready() {
          assert.ok(true, 'ready is called');
          readyWasCalled = false;
        }
      });
      let readyWasCalled = false;

      MyApplication.instanceInitializer({
        name: 'initializer',
        initialize() {
          assert.ok(!readyWasCalled, 'ready is not yet called');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }

    [`@test initializers are executed in their own context`](assert) {
      assert.expect(1);

      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'coolInitializer',
        myProperty: 'cool',
        initialize() {
          assert.equal(this.myProperty, 'cool', 'should have access to its own context');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));
    }

    [`@test initializers get an instance on app reset`](assert) {
      assert.expect(2);

      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'giveMeAnInstance',
        initialize(instance) {
          assert.ok(!!instance, 'Initializer got an instance');
        }
      });

      this.runTask(() => this.createApplication({}, MyApplication));

      this.runTask(() => this.application.reset());
    }
  });
});
enifed('@ember/application/tests/lazy_load_test', ['@ember/runloop', '@ember/application', 'internal-test-helpers'], function (_runloop, _application, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Lazy Loading', class extends _internalTestHelpers.AbstractTestCase {
    afterEach() {
      let keys = Object.keys(_application._loaded);
      for (let i = 0; i < keys.length; i++) {
        delete _application._loaded[keys[i]];
      }
    }

    ['@test if a load hook is registered, it is executed when runLoadHooks are exected'](assert) {
      let count = 0;

      (0, _runloop.run)(function () {
        (0, _application.onLoad)('__test_hook__', function (object) {
          count += object;
        });
      });

      (0, _runloop.run)(function () {
        (0, _application.runLoadHooks)('__test_hook__', 1);
      });

      assert.equal(count, 1, 'the object was passed into the load hook');
    }

    ['@test if runLoadHooks was already run, it executes newly added hooks immediately'](assert) {
      let count = 0;
      (0, _runloop.run)(() => {
        (0, _application.onLoad)('__test_hook__', object => count += object);
      });

      (0, _runloop.run)(() => (0, _application.runLoadHooks)('__test_hook__', 1));

      count = 0;
      (0, _runloop.run)(() => {
        (0, _application.onLoad)('__test_hook__', object => count += object);
      });

      assert.equal(count, 1, 'the original object was passed into the load hook');
    }

    ["@test hooks in ENV.EMBER_LOAD_HOOKS['hookName'] get executed"](assert) {
      // Note that the necessary code to perform this test is run before
      // the Ember lib is loaded in tests/index.html

      (0, _runloop.run)(() => {
        (0, _application.runLoadHooks)('__before_ember_test_hook__', 1);
      });

      assert.equal(window.ENV.__test_hook_count__, 1, 'the object was passed into the load hook');
    }

    ['@test load hooks trigger a custom event'](assert) {
      if (typeof window === 'object' && typeof window.dispatchEvent === 'function' && typeof CustomEvent === 'function') {
        let eventObject = 'super duper awesome events';

        window.addEventListener('__test_hook_for_events__', function (e) {
          assert.ok(true, 'custom event was fired');
          assert.equal(e.detail, eventObject, 'event details are provided properly');
        });

        (0, _runloop.run)(() => {
          (0, _application.runLoadHooks)('__test_hook_for_events__', eventObject);
        });
      } else {
        assert.expect(0);
      }
    }
  });
});
enifed('@ember/application/tests/logging_test', ['internal-test-helpers', '@ember/controller', '@ember/-internals/routing', '@ember/polyfills'], function (_internalTestHelpers, _controller, _routing, _polyfills) {
  'use strict';

  /*globals EmberDev */

  class LoggingApplicationTestCase extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();

      this.logs = {};

      /* eslint-disable no-console */
      this._originalLogger = console.info;

      console.info = (_, { fullName }) => {
        if (!this.logs.hasOwnProperty(fullName)) {
          this.logs[fullName] = 0;
        }
        /* eslint-ensable no-console */
        this.logs[fullName]++;
      };

      this.router.map(function () {
        this.route('posts', { resetNamespace: true });
      });
    }

    teardown() {
      /* eslint-disable no-console */
      console.info = this._originalLogger;
      /* eslint-enable no-console */
      super.teardown();
    }
  }

  (0, _internalTestHelpers.moduleFor)('Application with LOG_ACTIVE_GENERATION=true', class extends LoggingApplicationTestCase {
    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        LOG_ACTIVE_GENERATION: true
      });
    }

    ['@test log class generation if logging enabled'](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      return this.visit('/posts').then(() => {
        assert.equal(Object.keys(this.logs).length, 4, 'expected logs');
      });
    }

    ['@test actively generated classes get logged'](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      return this.visit('/posts').then(() => {
        assert.equal(this.logs['controller:application'], 1, 'expected: ApplicationController was generated');
        assert.equal(this.logs['controller:posts'], 1, 'expected: PostsController was generated');

        assert.equal(this.logs['route:application'], 1, 'expected: ApplicationRoute was generated');
        assert.equal(this.logs['route:posts'], 1, 'expected: PostsRoute was generated');
      });
    }

    ['@test predefined classes do not get logged'](assert) {
      this.add('controller:application', _controller.default.extend());
      this.add('controller:posts', _controller.default.extend());
      this.add('route:application', _routing.Route.extend());
      this.add('route:posts', _routing.Route.extend());

      return this.visit('/posts').then(() => {
        assert.ok(!this.logs['controller:application'], 'did not expect: ApplicationController was generated');
        assert.ok(!this.logs['controller:posts'], 'did not expect: PostsController was generated');

        assert.ok(!this.logs['route:application'], 'did not expect: ApplicationRoute was generated');
        assert.ok(!this.logs['route:posts'], 'did not expect: PostsRoute was generated');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application when LOG_ACTIVE_GENERATION=false', class extends LoggingApplicationTestCase {
    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        LOG_ACTIVE_GENERATION: false
      });
    }

    [`@test do NOT log class generation if logging disabled`](assert) {
      return this.visit('/posts').then(() => {
        assert.equal(Object.keys(this.logs).length, 0, 'expected logs');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application with LOG_VIEW_LOOKUPS=true', class extends LoggingApplicationTestCase {
    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        LOG_VIEW_LOOKUPS: true
      });
    }

    [`@test log when template and view are missing when flag is active`](assert) {
      if (EmberDev && EmberDev.runningProdBuild) {
        assert.ok(true, 'Logging does not occur in production builds');
        return;
      }

      this.addTemplate('application', '{{outlet}}');

      return this.visit('/').then(() => this.visit('/posts')).then(() => {
        assert.equal(this.logs['template:application'], undefined, 'expected: Should not log template:application since it exists.');
        assert.equal(this.logs['template:index'], 1, 'expected: Could not find "index" template or view.');
        assert.equal(this.logs['template:posts'], 1, 'expected: Could not find "posts" template or view.');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application with LOG_VIEW_LOOKUPS=false', class extends LoggingApplicationTestCase {
    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        LOG_VIEW_LOOKUPS: false
      });
    }

    [`@test do not log when template and view are missing when flag is not true`](assert) {
      return this.visit('/posts').then(() => {
        assert.equal(Object.keys(this.logs).length, 0, 'expected no logs');
      });
    }

    [`@test do not log which views are used with templates when flag is not true`](assert) {
      return this.visit('/posts').then(() => {
        assert.equal(Object.keys(this.logs).length, 0, 'expected no logs');
      });
    }
  });
});
enifed('@ember/application/tests/readiness_test', ['internal-test-helpers', '@ember/runloop', '@ember/application'], function (_internalTestHelpers, _runloop, _application) {
  'use strict';

  let jQuery, application, Application;
  let readyWasCalled, domReady, readyCallbacks;

  // We are using a small mock of jQuery because jQuery is third-party code with
  // very well-defined semantics, and we want to confirm that a jQuery stub run
  // in a more minimal server environment that implements this behavior will be
  // sufficient for Ember's requirements.
  (0, _internalTestHelpers.moduleFor)('Application readiness', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();

      readyWasCalled = 0;
      readyCallbacks = [];

      let jQueryInstance = {
        ready(callback) {
          readyCallbacks.push(callback);
          if (jQuery.isReady) {
            domReady();
          }
        }
      };

      jQuery = function () {
        return jQueryInstance;
      };
      jQuery.isReady = false;

      let domReadyCalled = 0;
      domReady = function () {
        if (domReadyCalled !== 0) {
          return;
        }
        domReadyCalled++;
        for (let i = 0; i < readyCallbacks.length; i++) {
          readyCallbacks[i]();
        }
      };

      Application = _application.default.extend({
        $: jQuery,

        ready() {
          readyWasCalled++;
        }
      });
    }

    teardown() {
      if (application) {
        (0, _runloop.run)(() => application.destroy());
        jQuery = readyCallbacks = domReady = Application = application = undefined;
      }
    }

    // These tests are confirming that if the callbacks passed into jQuery's ready hook is called
    // synchronously during the application's initialization, we get the same behavior as if
    // it was triggered after initialization.

    ["@test Application's ready event is called right away if jQuery is already ready"](assert) {
      jQuery.isReady = true;

      (0, _runloop.run)(() => {
        application = Application.create({ router: false });

        assert.equal(readyWasCalled, 0, 'ready is not called until later');
      });

      assert.equal(readyWasCalled, 1, 'ready was called');

      domReady();

      assert.equal(readyWasCalled, 1, "application's ready was not called again");
    }

    ["@test Application's ready event is called after the document becomes ready"](assert) {
      (0, _runloop.run)(() => {
        application = Application.create({ router: false });
      });

      assert.equal(readyWasCalled, 0, "ready wasn't called yet");

      domReady();

      assert.equal(readyWasCalled, 1, 'ready was called now that DOM is ready');
    }

    ["@test Application's ready event can be deferred by other components"](assert) {
      (0, _runloop.run)(() => {
        application = Application.create({ router: false });
        application.deferReadiness();
      });

      assert.equal(readyWasCalled, 0, "ready wasn't called yet");

      domReady();

      assert.equal(readyWasCalled, 0, "ready wasn't called yet");

      (0, _runloop.run)(() => {
        application.advanceReadiness();
        assert.equal(readyWasCalled, 0);
      });

      assert.equal(readyWasCalled, 1, 'ready was called now all readiness deferrals are advanced');
    }

    ["@test Application's ready event can be deferred by other components"](assert) {
      jQuery.isReady = false;

      (0, _runloop.run)(() => {
        application = Application.create({ router: false });
        application.deferReadiness();
        assert.equal(readyWasCalled, 0, "ready wasn't called yet");
      });

      domReady();

      assert.equal(readyWasCalled, 0, "ready wasn't called yet");

      (0, _runloop.run)(() => {
        application.advanceReadiness();
      });

      assert.equal(readyWasCalled, 1, 'ready was called now all readiness deferrals are advanced');

      expectAssertion(() => {
        application.deferReadiness();
      });
    }
  });
});
enifed('@ember/application/tests/reset_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/controller', '@ember/-internals/routing', 'internal-test-helpers'], function (_runloop, _metal, _controller, _routing, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application - resetting', class extends _internalTestHelpers.AutobootApplicationTestCase {
    ['@test Brings its own run-loop if not provided'](assert) {
      assert.expect(0);
      (0, _runloop.run)(() => this.createApplication());
      this.application.reset();
    }

    ['@test Does not bring its own run loop if one is already provided'](assert) {
      assert.expect(3);

      let didBecomeReady = false;

      (0, _runloop.run)(() => this.createApplication());

      (0, _runloop.run)(() => {
        this.application.ready = () => {
          didBecomeReady = true;
        };

        this.application.reset();

        this.application.deferReadiness();
        assert.ok(!didBecomeReady, 'app is not ready');
      });

      assert.ok(!didBecomeReady, 'app is not ready');
      (0, _runloop.run)(this.application, 'advanceReadiness');
      assert.ok(didBecomeReady, 'app is ready');
    }

    ['@test When an application is reset, new instances of controllers are generated'](assert) {
      (0, _runloop.run)(() => {
        this.createApplication();
        this.add('controller:academic', _controller.default.extend());
      });

      let firstController = this.applicationInstance.lookup('controller:academic');
      let secondController = this.applicationInstance.lookup('controller:academic');

      this.application.reset();

      let thirdController = this.applicationInstance.lookup('controller:academic');

      assert.strictEqual(firstController, secondController, 'controllers looked up in succession should be the same instance');

      assert.ok(firstController.isDestroying, 'controllers are destroyed when their application is reset');

      assert.notStrictEqual(firstController, thirdController, 'controllers looked up after the application is reset should not be the same instance');
    }

    ['@test When an application is reset, the eventDispatcher is destroyed and recreated'](assert) {
      let eventDispatcherWasSetup = 0;
      let eventDispatcherWasDestroyed = 0;

      let mockEventDispatcher = {
        setup() {
          eventDispatcherWasSetup++;
        },
        destroy() {
          eventDispatcherWasDestroyed++;
        }
      };

      (0, _runloop.run)(() => {
        this.createApplication();
        this.add('event_dispatcher:main', {
          create: () => mockEventDispatcher
        });

        assert.equal(eventDispatcherWasSetup, 0);
        assert.equal(eventDispatcherWasDestroyed, 0);
      });

      assert.equal(eventDispatcherWasSetup, 1);
      assert.equal(eventDispatcherWasDestroyed, 0);

      this.application.reset();

      assert.equal(eventDispatcherWasDestroyed, 1);
      assert.equal(eventDispatcherWasSetup, 2, 'setup called after reset');
    }

    ['@test When an application is reset, the router URL is reset to `/`'](assert) {
      (0, _runloop.run)(() => {
        this.createApplication();

        this.add('router:main', _routing.Router.extend({
          location: 'none'
        }));

        this.router.map(function () {
          this.route('one');
          this.route('two');
        });
      });

      let initialRouter, initialApplicationController;
      return this.visit('/one').then(() => {
        initialApplicationController = this.applicationInstance.lookup('controller:application');
        initialRouter = this.applicationInstance.lookup('router:main');
        let location = initialRouter.get('location');

        assert.equal(location.getURL(), '/one');
        assert.equal((0, _metal.get)(initialApplicationController, 'currentPath'), 'one');

        this.application.reset();

        return this.application._bootPromise;
      }).then(() => {
        let applicationController = this.applicationInstance.lookup('controller:application');
        assert.strictEqual(applicationController, undefined, 'application controller no longer exists');

        return this.visit('/one');
      }).then(() => {
        let applicationController = this.applicationInstance.lookup('controller:application');
        let router = this.applicationInstance.lookup('router:main');
        let location = router.get('location');

        assert.notEqual(initialRouter, router, 'a different router instance was created');
        assert.notEqual(initialApplicationController, applicationController, 'a different application controller is created');

        assert.equal(location.getURL(), '/one');
        assert.equal((0, _metal.get)(applicationController, 'currentPath'), 'one');
      });
    }

    ['@test When an application with advance/deferReadiness is reset, the app does correctly become ready after reset'](assert) {
      let readyCallCount = 0;

      (0, _runloop.run)(() => {
        this.createApplication({
          ready() {
            readyCallCount++;
          }
        });

        this.application.deferReadiness();
        assert.equal(readyCallCount, 0, 'ready has not yet been called');
      });

      (0, _runloop.run)(this.application, 'advanceReadiness');

      assert.equal(readyCallCount, 1, 'ready was called once');

      this.application.reset();

      assert.equal(readyCallCount, 2, 'ready was called twice');
    }
  });
});
enifed('@ember/application/tests/visit_test', ['internal-test-helpers', '@ember/service', '@ember/-internals/runtime', '@ember/runloop', '@ember/application', '@ember/application/instance', '@ember/engine', '@ember/-internals/routing', '@ember/-internals/glimmer', 'ember-template-compiler', '@ember/-internals/environment'], function (_internalTestHelpers, _service, _runtime, _runloop, _application, _instance, _engine, _routing, _glimmer, _emberTemplateCompiler, _environment) {
  'use strict';

  function expectAsyncError() {
    _runtime.RSVP.off('error');
  }

  (0, _internalTestHelpers.moduleFor)('Application - visit()', class extends _internalTestHelpers.ApplicationTestCase {
    teardown() {
      _runtime.RSVP.on('error', _runtime.onerrorDefault);
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;
      super.teardown();
    }

    createApplication(options) {
      return super.createApplication(options, _application.default.extend());
    }

    assertEmptyFixture(message) {
      this.assert.strictEqual(document.getElementById('qunit-fixture').children.length, 0, `there are no elements in the fixture element ${message ? message : ''}`);
    }

    [`@test does not add serialize-mode markers by default`](assert) {
      let templateContent = '<div class="foo">Hi, Mom!</div>';
      this.addTemplate('index', templateContent);
      let rootElement = document.createElement('div');

      let bootOptions = {
        isBrowser: false,
        rootElement
      };

      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;
      return this.visit('/', bootOptions).then(() => {
        assert.equal(rootElement.innerHTML, templateContent, 'without serialize flag renders as expected');
      });
    }

    [`@test _renderMode: rehydration`](assert) {
      assert.expect(2);

      let indexTemplate = '<div class="foo">Hi, Mom!</div>';
      this.addTemplate('index', indexTemplate);
      let rootElement = document.createElement('div');

      let bootOptions = {
        isBrowser: false,
        rootElement,
        _renderMode: 'serialize'
      };

      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;

      return this.visit('/', bootOptions).then(instance => {
        assert.ok((0, _glimmer.isSerializationFirstNode)(instance.rootElement.firstChild), 'glimmer-vm comment node was not found');
      }).then(() => {
        return this.runTask(() => {
          this.applicationInstance.destroy();
          this.applicationInstance = null;
        });
      }).then(() => {
        bootOptions = {
          isBrowser: false,
          rootElement,
          _renderMode: 'rehydrate'
        };

        this.application.visit('/', bootOptions).then(instance => {
          assert.equal(instance.rootElement.innerHTML, indexTemplate, 'was not properly rehydrated');
        });
      });
    }

    // This tests whether the application is "autobooted" by registering an
    // instance initializer and asserting it never gets run. Since this is
    // inherently testing that async behavior *doesn't* happen, we set a
    // 500ms timeout to verify that when autoboot is set to false, the
    // instance initializer that would normally get called on DOM ready
    // does not fire.
    [`@test Applications with autoboot set to false do not autoboot`](assert) {
      function delay(time) {
        return new _runtime.RSVP.Promise(resolve => (0, _runloop.later)(resolve, time));
      }

      let appBooted = 0;
      let instanceBooted = 0;

      this.application.initializer({
        name: 'assert-no-autoboot',
        initialize() {
          appBooted++;
        }
      });

      this.application.instanceInitializer({
        name: 'assert-no-autoboot',
        initialize() {
          instanceBooted++;
        }
      });

      assert.ok(!this.applicationInstance, 'precond - no instance');
      assert.ok(appBooted === 0, 'precond - not booted');
      assert.ok(instanceBooted === 0, 'precond - not booted');

      // Continue after 500ms
      return delay(500).then(() => {
        assert.ok(appBooted === 0, '500ms elapsed without app being booted');
        assert.ok(instanceBooted === 0, '500ms elapsed without instances being booted');

        return this.runTask(() => this.application.boot());
      }).then(() => {
        assert.ok(appBooted === 1, 'app should boot when manually calling `app.boot()`');
        assert.ok(instanceBooted === 0, 'no instances should be booted automatically when manually calling `app.boot()');
      });
    }

    [`@test calling visit() on an app without first calling boot() should boot the app`](assert) {
      let appBooted = 0;
      let instanceBooted = 0;

      this.application.initializer({
        name: 'assert-no-autoboot',
        initialize() {
          appBooted++;
        }
      });

      this.application.instanceInitializer({
        name: 'assert-no-autoboot',
        initialize() {
          instanceBooted++;
        }
      });

      return this.visit('/').then(() => {
        assert.ok(appBooted === 1, 'the app should be booted`');
        assert.ok(instanceBooted === 1, 'an instances should be booted');
      });
    }

    [`@test calling visit() on an already booted app should not boot it again`](assert) {
      let appBooted = 0;
      let instanceBooted = 0;

      this.application.initializer({
        name: 'assert-no-autoboot',
        initialize() {
          appBooted++;
        }
      });

      this.application.instanceInitializer({
        name: 'assert-no-autoboot',
        initialize() {
          instanceBooted++;
        }
      });

      return this.runTask(() => this.application.boot()).then(() => {
        assert.ok(appBooted === 1, 'the app should be booted');
        assert.ok(instanceBooted === 0, 'no instances should be booted');

        return this.visit('/');
      }).then(() => {
        assert.ok(appBooted === 1, 'the app should not be booted again');
        assert.ok(instanceBooted === 1, 'an instance should be booted');

        /*
        * Destroy the instance.
        */
        return this.runTask(() => {
          this.applicationInstance.destroy();
          this.applicationInstance = null;
        });
      }).then(() => {
        /*
        * Visit on the application a second time. The application should remain
        * booted, but a new instance will be created.
        */
        return this.application.visit('/').then(instance => {
          this.applicationInstance = instance;
        });
      }).then(() => {
        assert.ok(appBooted === 1, 'the app should not be booted again');
        assert.ok(instanceBooted === 2, 'another instance should be booted');
      });
    }

    [`@test visit() rejects on application boot failure`](assert) {
      this.application.initializer({
        name: 'error',
        initialize() {
          throw new Error('boot failure');
        }
      });

      expectAsyncError();

      return this.visit('/').then(() => {
        assert.ok(false, 'It should not resolve the promise');
      }, error => {
        assert.ok(error instanceof Error, 'It should reject the promise with the boot error');
        assert.equal(error.message, 'boot failure');
      });
    }

    [`@test visit() rejects on instance boot failure`](assert) {
      this.application.instanceInitializer({
        name: 'error',
        initialize() {
          throw new Error('boot failure');
        }
      });

      expectAsyncError();

      return this.visit('/').then(() => {
        assert.ok(false, 'It should not resolve the promise');
      }, error => {
        assert.ok(error instanceof Error, 'It should reject the promise with the boot error');
        assert.equal(error.message, 'boot failure');
      });
    }

    [`@test visit() follows redirects`](assert) {
      this.router.map(function () {
        this.route('a');
        this.route('b', { path: '/b/:b' });
        this.route('c', { path: '/c/:c' });
      });

      this.add('route:a', _routing.Route.extend({
        afterModel() {
          this.replaceWith('b', 'zomg');
        }
      }));

      this.add('route:b', _routing.Route.extend({
        afterModel(params) {
          this.transitionTo('c', params.b);
        }
      }));

      /*
      * First call to `visit` is `this.application.visit` and returns the
      * applicationInstance.
      */
      return this.visit('/a').then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(instance.getURL(), '/c/zomg', 'It should follow all redirects');
      });
    }

    [`@test visit() rejects if an error occurred during a transition`](assert) {
      this.router.map(function () {
        this.route('a');
        this.route('b', { path: '/b/:b' });
        this.route('c', { path: '/c/:c' });
      });

      this.add('route:a', _routing.Route.extend({
        afterModel() {
          this.replaceWith('b', 'zomg');
        }
      }));

      this.add('route:b', _routing.Route.extend({
        afterModel(params) {
          this.transitionTo('c', params.b);
        }
      }));

      this.add('route:c', _routing.Route.extend({
        afterModel() {
          throw new Error('transition failure');
        }
      }));

      expectAsyncError();

      return this.visit('/a').then(() => {
        assert.ok(false, 'It should not resolve the promise');
      }, error => {
        assert.ok(error instanceof Error, 'It should reject the promise with the boot error');
        assert.equal(error.message, 'transition failure');
      });
    }

    [`@test visit() chain`](assert) {
      this.router.map(function () {
        this.route('a');
        this.route('b');
        this.route('c');
      });

      return this.visit('/').then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(instance.getURL(), '/');

        return instance.visit('/a');
      }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(instance.getURL(), '/a');

        return instance.visit('/b');
      }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(instance.getURL(), '/b');

        return instance.visit('/c');
      }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(instance.getURL(), '/c');
      });
    }

    [`@test visit() returns a promise that resolves when the view has rendered`](assert) {
      this.addTemplate('application', `<h1>Hello world</h1>`);

      this.assertEmptyFixture();

      return this.visit('/').then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.equal(this.element.textContent, 'Hello world', 'the application was rendered once the promise resolves');
      });
    }

    [`@test visit() returns a promise that resolves without rendering when shouldRender is set to false`](assert) {
      assert.expect(3);

      this.addTemplate('application', '<h1>Hello world</h1>');

      this.assertEmptyFixture();

      return this.visit('/', { shouldRender: false }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');

        this.assertEmptyFixture('after visit');
      });
    }

    [`@test visit() renders a template when shouldRender is set to true`](assert) {
      assert.expect(3);

      this.addTemplate('application', '<h1>Hello world</h1>');

      this.assertEmptyFixture();

      return this.visit('/', { shouldRender: true }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.strictEqual(document.querySelector('#qunit-fixture').children.length, 1, 'there is 1 element in the fixture element after visit');
      });
    }

    [`@test visit() returns a promise that resolves without rendering when shouldRender is set to false with Engines`](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.mount('blog');
      });

      this.addTemplate('application', '<h1>Hello world</h1>');

      // Register engine
      let BlogEngine = _engine.default.extend();
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let BlogMap = function () {};
      this.add('route-map:blog', BlogMap);

      this.assertEmptyFixture();

      return this.visit('/blog', { shouldRender: false }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');

        this.assertEmptyFixture('after visit');
      });
    }

    [`@test visit() does not setup the event_dispatcher:main if isInteractive is false (with Engines) GH#15615`](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.mount('blog');
      });

      this.addTemplate('application', '<h1>Hello world</h1>{{outlet}}');
      this.add('event_dispatcher:main', {
        create() {
          throw new Error('should not happen!');
        }
      });

      // Register engine
      let BlogEngine = _engine.default.extend({
        init(...args) {
          this._super.apply(this, args);
          this.register('template:application', (0, _emberTemplateCompiler.compile)('{{cache-money}}'));
          this.register('template:components/cache-money', (0, _emberTemplateCompiler.compile)(`
          <p>Dis cache money</p>
        `));
          this.register('component:cache-money', _glimmer.Component.extend({}));
        }
      });
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let BlogMap = function () {};
      this.add('route-map:blog', BlogMap);

      this.assertEmptyFixture();

      return this.visit('/blog', { isInteractive: false }).then(instance => {
        assert.ok(instance instanceof _instance.default, 'promise is resolved with an ApplicationInstance');
        assert.strictEqual(this.element.querySelector('p').textContent, 'Dis cache money', 'Engine component is resolved');
      });
    }

    [`@test visit() on engine resolves engine component`](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.mount('blog');
      });

      // Register engine
      let BlogEngine = _engine.default.extend({
        init(...args) {
          this._super.apply(this, args);
          this.register('template:application', (0, _emberTemplateCompiler.compile)('{{cache-money}}'));
          this.register('template:components/cache-money', (0, _emberTemplateCompiler.compile)(`
          <p>Dis cache money</p>
        `));
          this.register('component:cache-money', _glimmer.Component.extend({}));
        }
      });
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let BlogMap = function () {};
      this.add('route-map:blog', BlogMap);

      this.assertEmptyFixture();

      return this.visit('/blog', { shouldRender: true }).then(() => {
        assert.strictEqual(this.element.querySelector('p').textContent, 'Dis cache money', 'Engine component is resolved');
      });
    }

    [`@test visit() on engine resolves engine helper`](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.mount('blog');
      });

      // Register engine
      let BlogEngine = _engine.default.extend({
        init(...args) {
          this._super.apply(this, args);
          this.register('template:application', (0, _emberTemplateCompiler.compile)('{{swag}}'));
          this.register('helper:swag', (0, _glimmer.helper)(function () {
            return 'turnt up';
          }));
        }
      });
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let BlogMap = function () {};
      this.add('route-map:blog', BlogMap);

      this.assertEmptyFixture();

      return this.visit('/blog', { shouldRender: true }).then(() => {
        assert.strictEqual(this.element.textContent, 'turnt up', 'Engine component is resolved');
      });
    }

    [`@test Ember Islands-style setup`](assert) {
      let xFooInitCalled = false;
      let xFooDidInsertElementCalled = false;

      let xBarInitCalled = false;
      let xBarDidInsertElementCalled = false;

      this.router.map(function () {
        this.route('show', { path: '/:component_name' });
      });

      this.add('route:show', _routing.Route.extend({
        queryParams: {
          data: { refreshModel: true }
        },

        model(params) {
          return {
            componentName: params.component_name,
            componentData: params.data ? JSON.parse(params.data) : undefined
          };
        }
      }));

      let Counter = _runtime.Object.extend({
        value: 0,

        increment() {
          this.incrementProperty('value');
        }
      });

      this.add('service:isolatedCounter', Counter);
      this.add('service:sharedCounter', Counter.create());
      this.application.registerOptions('service:sharedCounter', {
        instantiate: false
      });

      this.addTemplate('show', '{{component model.componentName model=model.componentData}}');

      this.addTemplate('components/x-foo', `
      <h1>X-Foo</h1>
      <p>Hello {{model.name}}, I have been clicked {{isolatedCounter.value}} times ({{sharedCounter.value}} times combined)!</p>
    `);

      this.add('component:x-foo', _glimmer.Component.extend({
        tagName: 'x-foo',

        isolatedCounter: (0, _service.inject)(),
        sharedCounter: (0, _service.inject)(),

        init() {
          this._super();
          xFooInitCalled = true;
        },

        didInsertElement() {
          xFooDidInsertElementCalled = true;
        },

        click() {
          this.get('isolatedCounter').increment();
          this.get('sharedCounter').increment();
        }
      }));

      this.addTemplate('components/x-bar', `
      <h1>X-Bar</h1>
      <button {{action "incrementCounter"}}>Join {{counter.value}} others in clicking me!</button>
    `);

      this.add('component:x-bar', _glimmer.Component.extend({
        counter: (0, _service.inject)('sharedCounter'),

        actions: {
          incrementCounter() {
            this.get('counter').increment();
          }
        },

        init() {
          this._super();
          xBarInitCalled = true;
        },

        didInsertElement() {
          xBarDidInsertElementCalled = true;
        }
      }));

      let fixtureElement = document.querySelector('#qunit-fixture');
      let foo = document.createElement('div');
      let bar = document.createElement('div');
      fixtureElement.appendChild(foo);
      fixtureElement.appendChild(bar);

      let data = encodeURIComponent(JSON.stringify({ name: 'Godfrey' }));
      let instances = [];

      return _runtime.RSVP.all([this.runTask(() => {
        return this.application.visit(`/x-foo?data=${data}`, {
          rootElement: foo
        });
      }), this.runTask(() => {
        return this.application.visit('/x-bar', { rootElement: bar });
      })]).then(_instances => {
        instances = _instances;

        assert.ok(xFooInitCalled);
        assert.ok(xFooDidInsertElementCalled);

        assert.ok(xBarInitCalled);
        assert.ok(xBarDidInsertElementCalled);

        assert.equal(foo.querySelector('h1').textContent, 'X-Foo');
        assert.equal(foo.querySelector('p').textContent, 'Hello Godfrey, I have been clicked 0 times (0 times combined)!');
        assert.ok(foo.textContent.indexOf('X-Bar') === -1);

        assert.equal(bar.querySelector('h1').textContent, 'X-Bar');
        assert.equal(bar.querySelector('button').textContent, 'Join 0 others in clicking me!');
        assert.ok(bar.textContent.indexOf('X-Foo') === -1);

        this.runTask(() => {
          this.click(foo.querySelector('x-foo'));
        });

        assert.equal(foo.querySelector('p').textContent, 'Hello Godfrey, I have been clicked 1 times (1 times combined)!');
        assert.equal(bar.querySelector('button').textContent, 'Join 1 others in clicking me!');

        this.runTask(() => {
          this.click(bar.querySelector('button'));
          this.click(bar.querySelector('button'));
        });

        assert.equal(foo.querySelector('p').textContent, 'Hello Godfrey, I have been clicked 1 times (3 times combined)!');
        assert.equal(bar.querySelector('button').textContent, 'Join 3 others in clicking me!');
      }).finally(() => {
        this.runTask(() => {
          instances.forEach(instance => {
            instance.destroy();
          });
        });
      });
    }
  });
});
enifed('@ember/controller/tests/controller_test', ['@ember/controller', '@ember/service', '@ember/-internals/runtime', '@ember/-internals/metal', 'internal-test-helpers'], function (_controller, _service, _runtime, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Controller event handling', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Action can be handled by a function on actions object'](assert) {
      assert.expect(1);
      let TestController = _controller.default.extend({
        actions: {
          poke() {
            assert.ok(true, 'poked');
          }
        }
      });
      let controller = TestController.create();
      controller.send('poke');
    }

    ['@test A handled action can be bubbled to the target for continued processing'](assert) {
      assert.expect(2);
      let TestController = _controller.default.extend({
        actions: {
          poke() {
            assert.ok(true, 'poked 1');
            return true;
          }
        }
      });

      let controller = TestController.create({
        target: _controller.default.extend({
          actions: {
            poke() {
              assert.ok(true, 'poked 2');
            }
          }
        }).create()
      });
      controller.send('poke');
    }

    ["@test Action can be handled by a superclass' actions object"](assert) {
      assert.expect(4);

      let SuperController = _controller.default.extend({
        actions: {
          foo() {
            assert.ok(true, 'foo');
          },
          bar(msg) {
            assert.equal(msg, 'HELLO');
          }
        }
      });

      let BarControllerMixin = _metal.Mixin.create({
        actions: {
          bar(msg) {
            assert.equal(msg, 'HELLO');
            this._super(msg);
          }
        }
      });

      let IndexController = SuperController.extend(BarControllerMixin, {
        actions: {
          baz() {
            assert.ok(true, 'baz');
          }
        }
      });

      let controller = IndexController.create({});
      controller.send('foo');
      controller.send('bar', 'HELLO');
      controller.send('baz');
    }

    ['@test .send asserts if called on a destroyed controller']() {
      let owner = (0, _internalTestHelpers.buildOwner)();

      owner.register('controller:application', _controller.default.extend({
        toString() {
          return 'controller:rip-alley';
        }
      }));

      let controller = owner.lookup('controller:application');
      (0, _internalTestHelpers.runDestroy)(owner);

      expectAssertion(() => {
        controller.send('trigger-me-dead');
      }, "Attempted to call .send() with the action 'trigger-me-dead' on the destroyed object 'controller:rip-alley'.");
    }
  });

  (0, _internalTestHelpers.moduleFor)('Controller deprecations -> Controller Content -> Model Alias', class extends _internalTestHelpers.AbstractTestCase {
    ['@test `content` is not moved to `model` when `model` is unset'](assert) {
      assert.expect(2);
      let controller;

      ignoreDeprecation(function () {
        controller = _controller.default.extend({
          content: 'foo-bar'
        }).create();
      });

      assert.notEqual(controller.get('model'), 'foo-bar', 'model is set properly');
      assert.equal(controller.get('content'), 'foo-bar', 'content is not set properly');
    }

    ['@test specifying `content` (without `model` specified) does not result in deprecation'](assert) {
      assert.expect(2);
      expectNoDeprecation();

      let controller = _controller.default.extend({
        content: 'foo-bar'
      }).create();

      assert.equal((0, _metal.get)(controller, 'content'), 'foo-bar');
    }

    ['@test specifying `content` (with `model` specified) does not result in deprecation'](assert) {
      assert.expect(3);
      expectNoDeprecation();

      let controller = _controller.default.extend({
        content: 'foo-bar',
        model: 'blammo'
      }).create();

      assert.equal((0, _metal.get)(controller, 'content'), 'foo-bar');
      assert.equal((0, _metal.get)(controller, 'model'), 'blammo');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Controller deprecations -> Controller injected properties', class extends _internalTestHelpers.AbstractTestCase {
    ['@test defining a controller on a non-controller should fail assertion']() {
      expectAssertion(function () {
        let owner = (0, _internalTestHelpers.buildOwner)();

        let AnObject = _runtime.Object.extend({
          foo: (0, _controller.inject)('bar')
        });

        owner.register('controller:bar', _runtime.Object.extend());
        owner.register('foo:main', AnObject);

        owner.lookup('foo:main');
      }, /Defining `foo` as an injected controller property on a non-controller \(`foo:main`\) is not allowed/);
    }

    ['@test controllers can be injected into controllers'](assert) {
      let owner = (0, _internalTestHelpers.buildOwner)();

      owner.register('controller:post', _controller.default.extend({
        postsController: (0, _controller.inject)('posts')
      }));

      owner.register('controller:posts', _controller.default.extend());

      let postController = owner.lookup('controller:post');
      let postsController = owner.lookup('controller:posts');

      assert.equal(postsController, postController.get('postsController'), 'controller.posts is injected');
    }

    ['@test services can be injected into controllers'](assert) {
      let owner = (0, _internalTestHelpers.buildOwner)();

      owner.register('controller:application', _controller.default.extend({
        authService: (0, _service.inject)('auth')
      }));

      owner.register('service:auth', _service.default.extend());

      let appController = owner.lookup('controller:application');
      let authService = owner.lookup('service:auth');

      assert.equal(authService, appController.get('authService'), 'service.auth is injected');
    }
  });
});
enifed('@ember/debug/tests/handlers-test', ['@ember/debug/lib/handlers', 'internal-test-helpers'], function (_handlers, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-debug: registerHandler', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      delete _handlers.HANDLERS.blarz;
    }

    ['@test calls handler on `invoke` when `falsey`'](assert) {
      assert.expect(2);

      function handler(message) {
        assert.ok(true, 'called handler');
        assert.equal(message, 'Foo bar');
      }

      (0, _handlers.registerHandler)('blarz', handler);

      (0, _handlers.invoke)('blarz', 'Foo bar', false);
    }

    ['@test does not call handler on `invoke` when `truthy`'](assert) {
      assert.expect(0);

      function handler() {
        assert.ok(false, 'called handler');
      }

      (0, _handlers.registerHandler)('blarz', handler);

      (0, _handlers.invoke)('blarz', 'Foo bar', true);
    }

    ['@test calling `invoke` without handlers does not throw an error'](assert) {
      assert.expect(0);

      (0, _handlers.invoke)('blarz', 'Foo bar', false);
    }

    ['@test invoking `next` argument calls the next handler'](assert) {
      assert.expect(2);

      function handler1() {
        assert.ok(true, 'called handler1');
      }

      function handler2(message, options, next) {
        assert.ok(true, 'called handler2');
        next(message, options);
      }

      (0, _handlers.registerHandler)('blarz', handler1);
      (0, _handlers.registerHandler)('blarz', handler2);

      (0, _handlers.invoke)('blarz', 'Foo', false);
    }

    ['@test invoking `next` when no other handlers exists does not error'](assert) {
      assert.expect(1);

      function handler(message, options, next) {
        assert.ok(true, 'called handler1');

        next(message, options);
      }

      (0, _handlers.registerHandler)('blarz', handler);

      (0, _handlers.invoke)('blarz', 'Foo', false);
    }

    ['@test handlers are called in the proper order'](assert) {
      assert.expect(11);

      let expectedMessage = 'This is the message';
      let expectedOptions = { id: 'foo-bar' };
      let expected = ['first', 'second', 'third', 'fourth', 'fifth'];
      let actualCalls = [];

      function generateHandler(item) {
        return function (message, options, next) {
          assert.equal(message, expectedMessage, `message supplied to ${item} handler is correct`);
          assert.equal(options, expectedOptions, `options supplied to ${item} handler is correct`);

          actualCalls.push(item);

          next(message, options);
        };
      }

      expected.forEach(item => (0, _handlers.registerHandler)('blarz', generateHandler(item)));

      (0, _handlers.invoke)('blarz', expectedMessage, false, expectedOptions);

      assert.deepEqual(actualCalls, expected.reverse(), 'handlers were called in proper order');
    }

    ['@test not invoking `next` prevents further handlers from being called'](assert) {
      assert.expect(1);

      function handler1() {
        assert.ok(false, 'called handler1');
      }

      function handler2() {
        assert.ok(true, 'called handler2');
      }

      (0, _handlers.registerHandler)('blarz', handler1);
      (0, _handlers.registerHandler)('blarz', handler2);

      (0, _handlers.invoke)('blarz', 'Foo', false);
    }

    ['@test handlers can call `next` with custom message and/or options'](assert) {
      assert.expect(4);

      let initialMessage = 'initial message';
      let initialOptions = { id: 'initial-options' };

      let handler2Message = 'Handler2 Message';
      let handler2Options = { id: 'handler-2' };

      function handler1(message, options) {
        assert.equal(message, handler2Message, 'handler2 message provided to handler1');
        assert.equal(options, handler2Options, 'handler2 options provided to handler1');
      }

      function handler2(message, options, next) {
        assert.equal(message, initialMessage, 'initial message provided to handler2');
        assert.equal(options, initialOptions, 'initial options proivided to handler2');

        next(handler2Message, handler2Options);
      }

      (0, _handlers.registerHandler)('blarz', handler1);
      (0, _handlers.registerHandler)('blarz', handler2);

      (0, _handlers.invoke)('blarz', initialMessage, false, initialOptions);
    }
  });
});
enifed('@ember/debug/tests/main_test', ['@ember/-internals/environment', '@ember/-internals/runtime', '@ember/debug/lib/handlers', '@ember/debug/lib/deprecate', '@ember/debug/lib/warn', '@ember/debug/index', 'internal-test-helpers'], function (_environment, _runtime, _handlers, _deprecate, _warn, _index, _internalTestHelpers) {
  'use strict';

  let originalEnvValue;
  let originalDeprecateHandler;
  let originalWarnHandler;

  const originalConsoleWarn = console.warn; // eslint-disable-line no-console
  const noop = function () {};

  (0, _internalTestHelpers.moduleFor)('ember-debug', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      originalEnvValue = _environment.ENV.RAISE_ON_DEPRECATION;
      originalDeprecateHandler = _handlers.HANDLERS.deprecate;
      originalWarnHandler = _handlers.HANDLERS.warn;

      _environment.ENV.RAISE_ON_DEPRECATION = true;
    }

    teardown() {
      _handlers.HANDLERS.deprecate = originalDeprecateHandler;
      _handlers.HANDLERS.warn = originalWarnHandler;

      _environment.ENV.RAISE_ON_DEPRECATION = originalEnvValue;
    }

    afterEach() {
      console.warn = originalConsoleWarn; // eslint-disable-line no-console
    }

    ['@test deprecate does not throw if RAISE_ON_DEPRECATION is false'](assert) {
      assert.expect(1);
      console.warn = noop; // eslint-disable-line no-console

      _environment.ENV.RAISE_ON_DEPRECATION = false;

      try {
        (0, _index.deprecate)('Should not throw', false, { id: 'test', until: 'forever' });
        assert.ok(true, 'deprecate did not throw');
      } catch (e) {
        assert.ok(false, `Expected deprecate not to throw but it did: ${e.message}`);
      }
    }

    ['@test deprecate resets deprecation level to RAISE if ENV.RAISE_ON_DEPRECATION is set'](assert) {
      assert.expect(2);
      console.warn = noop; // eslint-disable-line no-console

      _environment.ENV.RAISE_ON_DEPRECATION = false;

      try {
        (0, _index.deprecate)('Should not throw', false, { id: 'test', until: 'forever' });
        assert.ok(true, 'deprecate did not throw');
      } catch (e) {
        assert.ok(false, `Expected deprecate not to throw but it did: ${e.message}`);
      }

      _environment.ENV.RAISE_ON_DEPRECATION = true;

      assert.throws(() => {
        (0, _index.deprecate)('Should throw', false, { id: 'test', until: 'forever' });
      }, /Should throw/);
    }

    ['@test When ENV.RAISE_ON_DEPRECATION is true, it is still possible to silence a deprecation by id'](assert) {
      assert.expect(3);

      _environment.ENV.RAISE_ON_DEPRECATION = true;
      (0, _deprecate.registerHandler)(function (message, options, next) {
        if (!options || options.id !== 'my-deprecation') {
          next(...arguments);
        }
      });

      try {
        (0, _index.deprecate)('should be silenced with matching id', false, {
          id: 'my-deprecation',
          until: 'forever'
        });
        assert.ok(true, 'Did not throw when level is set by id');
      } catch (e) {
        assert.ok(false, `Expected deprecate not to throw but it did: ${e.message}`);
      }

      assert.throws(() => {
        (0, _index.deprecate)('Should throw with no matching id', false, {
          id: 'test',
          until: 'forever'
        });
      }, /Should throw with no matching id/);

      assert.throws(() => {
        (0, _index.deprecate)('Should throw with non-matching id', false, {
          id: 'other-id',
          until: 'forever'
        });
      }, /Should throw with non-matching id/);
    }

    ['@test deprecate throws deprecation if second argument is falsy'](assert) {
      assert.expect(3);

      assert.throws(() => (0, _index.deprecate)('Deprecation is thrown', false, {
        id: 'test',
        until: 'forever'
      }));
      assert.throws(() => (0, _index.deprecate)('Deprecation is thrown', '', { id: 'test', until: 'forever' }));
      assert.throws(() => (0, _index.deprecate)('Deprecation is thrown', 0, { id: 'test', until: 'forever' }));
    }

    ['@test deprecate does not invoke a function as the second argument'](assert) {
      assert.expect(1);

      (0, _index.deprecate)('Deprecation is thrown', function () {
        assert.ok(false, 'this function should not be invoked');
      }, { id: 'test', until: 'forever' });

      assert.ok(true, 'deprecations were not thrown');
    }

    ['@test deprecate does not throw deprecations if second argument is truthy'](assert) {
      assert.expect(1);

      (0, _index.deprecate)('Deprecation is thrown', true, {
        id: 'test',
        until: 'forever'
      });
      (0, _index.deprecate)('Deprecation is thrown', '1', { id: 'test', until: 'forever' });
      (0, _index.deprecate)('Deprecation is thrown', 1, { id: 'test', until: 'forever' });

      assert.ok(true, 'deprecations were not thrown');
    }

    ['@test assert throws if second argument is falsy'](assert) {
      assert.expect(3);

      assert.throws(() => (0, _index.assert)('Assertion is thrown', false));
      assert.throws(() => (0, _index.assert)('Assertion is thrown', ''));
      assert.throws(() => (0, _index.assert)('Assertion is thrown', 0));
    }

    ['@test assert does not throw if second argument is a function'](assert) {
      assert.expect(1);

      (0, _index.assert)('Assertion is thrown', () => true);

      assert.ok(true, 'assertions were not thrown');
    }

    ['@test assert does not throw if second argument is falsy'](assert) {
      assert.expect(1);

      (0, _index.assert)('Assertion is thrown', true);
      (0, _index.assert)('Assertion is thrown', '1');
      (0, _index.assert)('Assertion is thrown', 1);

      assert.ok(true, 'assertions were not thrown');
    }

    ['@test assert does not throw if second argument is an object'](assert) {
      assert.expect(1);
      let Igor = _runtime.Object.extend();

      (0, _index.assert)('is truthy', Igor);
      (0, _index.assert)('is truthy', Igor.create());

      assert.ok(true, 'assertions were not thrown');
    }

    ['@test deprecate does not throw a deprecation at log and silence'](assert) {
      assert.expect(4);
      let id = 'ABC';
      let until = 'forever';
      let shouldThrow = false;

      (0, _deprecate.registerHandler)(function (message, options) {
        if (options && options.id === id) {
          if (shouldThrow) {
            throw new Error(message);
          }
        }
      });

      try {
        (0, _index.deprecate)('Deprecation for testing purposes', false, { id, until });
        assert.ok(true, 'Deprecation did not throw');
      } catch (e) {
        assert.ok(false, 'Deprecation was thrown despite being added to blacklist');
      }

      try {
        (0, _index.deprecate)('Deprecation for testing purposes', false, { id, until });
        assert.ok(true, 'Deprecation did not throw');
      } catch (e) {
        assert.ok(false, 'Deprecation was thrown despite being added to blacklist');
      }

      shouldThrow = true;

      assert.throws(function () {
        (0, _index.deprecate)('Deprecation is thrown', false, { id, until });
      });

      assert.throws(function () {
        (0, _index.deprecate)('Deprecation is thrown', false, { id, until });
      });
    }

    ['@test deprecate without options triggers an assertion'](assert) {
      assert.expect(2);

      assert.throws(() => (0, _index.deprecate)('foo'), new RegExp(_deprecate.missingOptionsDeprecation), 'proper assertion is triggered when options is missing');

      assert.throws(() => (0, _index.deprecate)('foo', false, {}), new RegExp(_deprecate.missingOptionsDeprecation), 'proper assertion is triggered when options is missing');
    }

    ['@test deprecate without options.id triggers an assertion'](assert) {
      assert.expect(1);

      assert.throws(() => (0, _index.deprecate)('foo', false, { until: 'forever' }), new RegExp(_deprecate.missingOptionsIdDeprecation), 'proper assertion is triggered when options.id is missing');
    }

    ['@test deprecate without options.until triggers an assertion'](assert) {
      assert.expect(1);

      assert.throws(() => (0, _index.deprecate)('foo', false, { id: 'test' }), new RegExp(_deprecate.missingOptionsUntilDeprecation), 'proper assertion is triggered when options.until is missing');
    }

    ['@test warn without options triggers an assert'](assert) {
      assert.expect(1);

      assert.throws(() => (0, _index.warn)('foo'), new RegExp(_warn.missingOptionsDeprecation), 'deprecation is triggered when options is missing');
    }

    ['@test warn without options.id triggers an assertion'](assert) {
      assert.expect(1);

      assert.throws(() => (0, _index.warn)('foo', false, {}), new RegExp(_warn.missingOptionsIdDeprecation), 'deprecation is triggered when options is missing');
    }

    ['@test warn without options.id nor test triggers an assertion'](assert) {
      assert.expect(1);

      assert.throws(() => (0, _index.warn)('foo', {}), new RegExp(_warn.missingOptionsIdDeprecation), 'deprecation is triggered when options is missing');
    }

    ['@test warn without test but with options does not trigger an assertion'](assert) {
      assert.expect(1);

      (0, _warn.registerHandler)(function (message) {
        assert.equal(message, 'foo', 'warning was triggered');
      });

      (0, _index.warn)('foo', { id: 'ember-debug.do-not-raise' });
    }
  });
});
enifed('@ember/engine/tests/engine_initializers_test', ['@ember/runloop', '@ember/engine', 'internal-test-helpers'], function (_runloop, _engine, _internalTestHelpers) {
  'use strict';

  let MyEngine, myEngine, myEngineInstance;

  (0, _internalTestHelpers.moduleFor)('Engine initializers', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      (0, _runloop.run)(() => {
        if (myEngineInstance) {
          myEngineInstance.destroy();
          myEngineInstance = null;
        }

        if (myEngine) {
          myEngine.destroy();
          myEngine = null;
        }
      });
    }

    ["@test initializers require proper 'name' and 'initialize' properties"]() {
      MyEngine = _engine.default.extend();

      expectAssertion(() => {
        (0, _runloop.run)(() => {
          MyEngine.initializer({ name: 'initializer' });
        });
      });

      expectAssertion(() => {
        (0, _runloop.run)(() => {
          MyEngine.initializer({ initialize() {} });
        });
      });
    }

    ['@test initializers are passed an Engine'](assert) {
      MyEngine = _engine.default.extend();

      MyEngine.initializer({
        name: 'initializer',
        initialize(engine) {
          assert.ok(engine instanceof _engine.default, 'initialize is passed an Engine');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = myEngine.buildInstance();
    }

    ['@test initializers can be registered in a specified order'](assert) {
      let order = [];

      MyEngine = _engine.default.extend();
      MyEngine.initializer({
        name: 'fourth',
        after: 'third',
        initialize() {
          order.push('fourth');
        }
      });

      MyEngine.initializer({
        name: 'second',
        after: 'first',
        before: 'third',
        initialize() {
          order.push('second');
        }
      });

      MyEngine.initializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyEngine.initializer({
        name: 'first',
        before: 'second',
        initialize() {
          order.push('first');
        }
      });

      MyEngine.initializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyEngine.initializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = myEngine.buildInstance();

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    ['@test initializers can be registered in a specified order as an array'](assert) {
      let order = [];

      MyEngine = _engine.default.extend();

      MyEngine.initializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyEngine.initializer({
        name: 'second',
        after: 'first',
        before: ['third', 'fourth'],
        initialize() {
          order.push('second');
        }
      });

      MyEngine.initializer({
        name: 'fourth',
        after: ['second', 'third'],
        initialize() {
          order.push('fourth');
        }
      });

      MyEngine.initializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyEngine.initializer({
        name: 'first',
        before: ['second'],
        initialize() {
          order.push('first');
        }
      });

      MyEngine.initializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = myEngine.buildInstance();

      assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
    }

    ['@test initializers can have multiple dependencies'](assert) {
      let order = [];

      MyEngine = _engine.default.extend();

      let a = {
        name: 'a',
        before: 'b',
        initialize() {
          order.push('a');
        }
      };
      let b = {
        name: 'b',
        initialize() {
          order.push('b');
        }
      };
      let c = {
        name: 'c',
        after: 'b',
        initialize() {
          order.push('c');
        }
      };
      let afterB = {
        name: 'after b',
        after: 'b',
        initialize() {
          order.push('after b');
        }
      };
      let afterC = {
        name: 'after c',
        after: 'c',
        initialize() {
          order.push('after c');
        }
      };

      MyEngine.initializer(b);
      MyEngine.initializer(a);
      MyEngine.initializer(afterC);
      MyEngine.initializer(afterB);
      MyEngine.initializer(c);

      myEngine = MyEngine.create();
      myEngineInstance = myEngine.buildInstance();

      assert.ok(order.indexOf(a.name) < order.indexOf(b.name), 'a < b');
      assert.ok(order.indexOf(b.name) < order.indexOf(c.name), 'b < c');
      assert.ok(order.indexOf(b.name) < order.indexOf(afterB.name), 'b < afterB');
      assert.ok(order.indexOf(c.name) < order.indexOf(afterC.name), 'c < afterC');
    }

    ['@test initializers set on Engine subclasses are not shared between engines'](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstEngine = _engine.default.extend();

      FirstEngine.initializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondEngine = _engine.default.extend();

      SecondEngine.initializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      let firstEngine = FirstEngine.create();
      let firstEngineInstance = firstEngine.buildInstance();

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run');
      assert.equal(secondInitializerRunCount, 0, 'first initializer only was run');

      let secondEngine = SecondEngine.create();
      let secondEngineInstance = secondEngine.buildInstance();

      assert.equal(firstInitializerRunCount, 1, 'second initializer only was run');
      assert.equal(secondInitializerRunCount, 1, 'second initializer only was run');

      (0, _runloop.run)(function () {
        firstEngineInstance.destroy();
        secondEngineInstance.destroy();

        firstEngine.destroy();
        secondEngine.destroy();
      });
    }

    ['@test initializers are concatenated'](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstEngine = _engine.default.extend();

      FirstEngine.initializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondEngine = FirstEngine.extend();

      SecondEngine.initializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      let firstEngine = FirstEngine.create();
      let firstEngineInstance = firstEngine.buildInstance();

      assert.equal(firstInitializerRunCount, 1, 'first initializer only was run when base class created');
      assert.equal(secondInitializerRunCount, 0, 'second initializer was not run when first base class created');
      firstInitializerRunCount = 0;

      let secondEngine = SecondEngine.create();
      let secondEngineInstance = secondEngine.buildInstance();

      assert.equal(firstInitializerRunCount, 1, 'first initializer was run when subclass created');
      assert.equal(secondInitializerRunCount, 1, 'second initializers was run when subclass created');

      (0, _runloop.run)(function () {
        firstEngineInstance.destroy();
        secondEngineInstance.destroy();

        firstEngine.destroy();
        secondEngine.destroy();
      });
    }

    ['@test initializers are per-engine'](assert) {
      assert.expect(2);

      let FirstEngine = _engine.default.extend();

      FirstEngine.initializer({
        name: 'abc',
        initialize() {}
      });

      expectAssertion(function () {
        FirstEngine.initializer({
          name: 'abc',
          initialize() {}
        });
      });

      let SecondEngine = _engine.default.extend();
      SecondEngine.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      assert.ok(true, 'Two engines can have initializers named the same.');
    }

    ['@test initializers are executed in their own context'](assert) {
      assert.expect(1);

      MyEngine = _engine.default.extend();

      MyEngine.initializer({
        name: 'coolInitializer',
        myProperty: 'cool',
        initialize() {
          assert.equal(this.myProperty, 'cool', 'should have access to its own context');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = myEngine.buildInstance();
    }
  });
});
enifed('@ember/engine/tests/engine_instance_initializers_test', ['@ember/runloop', '@ember/engine', '@ember/engine/instance', 'internal-test-helpers'], function (_runloop, _engine, _instance, _internalTestHelpers) {
  'use strict';

  let MyEngine, myEngine, myEngineInstance;

  function buildEngineInstance(EngineClass) {
    let engineInstance = EngineClass.buildInstance();
    (0, _engine.setEngineParent)(engineInstance, {
      lookup() {
        return {};
      },
      resolveRegistration() {
        return {};
      }
    });
    return engineInstance;
  }

  (0, _internalTestHelpers.moduleFor)('Engine instance initializers', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      super.teardown();
      (0, _runloop.run)(() => {
        if (myEngineInstance) {
          myEngineInstance.destroy();
        }

        if (myEngine) {
          myEngine.destroy();
        }
      });
      MyEngine = myEngine = myEngineInstance = undefined;
    }

    ["@test initializers require proper 'name' and 'initialize' properties"]() {
      MyEngine = _engine.default.extend();

      expectAssertion(() => {
        (0, _runloop.run)(() => {
          MyEngine.instanceInitializer({ name: 'initializer' });
        });
      });

      expectAssertion(() => {
        (0, _runloop.run)(() => {
          MyEngine.instanceInitializer({ initialize() {} });
        });
      });
    }

    ['@test initializers are passed an engine instance'](assert) {
      MyEngine = _engine.default.extend();

      MyEngine.instanceInitializer({
        name: 'initializer',
        initialize(instance) {
          assert.ok(instance instanceof _instance.default, 'initialize is passed an engine instance');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = buildEngineInstance(myEngine);
      return myEngineInstance.boot();
    }

    ['@test initializers can be registered in a specified order'](assert) {
      let order = [];

      MyEngine = _engine.default.extend();

      MyEngine.instanceInitializer({
        name: 'fourth',
        after: 'third',
        initialize() {
          order.push('fourth');
        }
      });

      MyEngine.instanceInitializer({
        name: 'second',
        after: 'first',
        before: 'third',
        initialize() {
          order.push('second');
        }
      });

      MyEngine.instanceInitializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyEngine.instanceInitializer({
        name: 'first',
        before: 'second',
        initialize() {
          order.push('first');
        }
      });

      MyEngine.instanceInitializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyEngine.instanceInitializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = buildEngineInstance(myEngine);

      return myEngineInstance.boot().then(() => {
        assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
      });
    }

    ['@test initializers can be registered in a specified order as an array'](assert) {
      let order = [];
      MyEngine = _engine.default.extend();

      MyEngine.instanceInitializer({
        name: 'third',
        initialize() {
          order.push('third');
        }
      });

      MyEngine.instanceInitializer({
        name: 'second',
        after: 'first',
        before: ['third', 'fourth'],
        initialize() {
          order.push('second');
        }
      });

      MyEngine.instanceInitializer({
        name: 'fourth',
        after: ['second', 'third'],
        initialize() {
          order.push('fourth');
        }
      });

      MyEngine.instanceInitializer({
        name: 'fifth',
        after: 'fourth',
        before: 'sixth',
        initialize() {
          order.push('fifth');
        }
      });

      MyEngine.instanceInitializer({
        name: 'first',
        before: ['second'],
        initialize() {
          order.push('first');
        }
      });

      MyEngine.instanceInitializer({
        name: 'sixth',
        initialize() {
          order.push('sixth');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = buildEngineInstance(myEngine);

      return myEngineInstance.boot().then(() => {
        assert.deepEqual(order, ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']);
      });
    }

    ['@test initializers can have multiple dependencies'](assert) {
      let order = [];

      MyEngine = _engine.default.extend();

      let a = {
        name: 'a',
        before: 'b',
        initialize() {
          order.push('a');
        }
      };
      let b = {
        name: 'b',
        initialize() {
          order.push('b');
        }
      };
      let c = {
        name: 'c',
        after: 'b',
        initialize() {
          order.push('c');
        }
      };
      let afterB = {
        name: 'after b',
        after: 'b',
        initialize() {
          order.push('after b');
        }
      };
      let afterC = {
        name: 'after c',
        after: 'c',
        initialize() {
          order.push('after c');
        }
      };

      MyEngine.instanceInitializer(b);
      MyEngine.instanceInitializer(a);
      MyEngine.instanceInitializer(afterC);
      MyEngine.instanceInitializer(afterB);
      MyEngine.instanceInitializer(c);

      myEngine = MyEngine.create();
      myEngineInstance = buildEngineInstance(myEngine);

      return myEngineInstance.boot().then(() => {
        assert.ok(order.indexOf(a.name) < order.indexOf(b.name), 'a < b');
        assert.ok(order.indexOf(b.name) < order.indexOf(c.name), 'b < c');
        assert.ok(order.indexOf(b.name) < order.indexOf(afterB.name), 'b < afterB');
        assert.ok(order.indexOf(c.name) < order.indexOf(afterC.name), 'c < afterC');
      });
    }

    ['@test initializers set on Engine subclasses should not be shared between engines'](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstEngine = _engine.default.extend();
      let firstEngine, firstEngineInstance;

      FirstEngine.instanceInitializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondEngine = _engine.default.extend();
      let secondEngine, secondEngineInstance;

      SecondEngine.instanceInitializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      firstEngine = FirstEngine.create();
      firstEngineInstance = buildEngineInstance(firstEngine);

      return firstEngineInstance.boot().then(() => {
        assert.equal(firstInitializerRunCount, 1, 'first initializer only was run');
        assert.equal(secondInitializerRunCount, 0, 'first initializer only was run');

        secondEngine = SecondEngine.create();
        secondEngineInstance = buildEngineInstance(secondEngine);
        return secondEngineInstance.boot();
      }).then(() => {
        assert.equal(firstInitializerRunCount, 1, 'second initializer only was run');
        assert.equal(secondInitializerRunCount, 1, 'second initializer only was run');

        (0, _runloop.run)(() => {
          firstEngineInstance.destroy();
          secondEngineInstance.destroy();

          firstEngine.destroy();
          secondEngine.destroy();
        });
      });
    }

    ['@test initializers are concatenated'](assert) {
      let firstInitializerRunCount = 0;
      let secondInitializerRunCount = 0;
      let FirstEngine = _engine.default.extend();

      FirstEngine.instanceInitializer({
        name: 'first',
        initialize() {
          firstInitializerRunCount++;
        }
      });

      let SecondEngine = FirstEngine.extend();

      SecondEngine.instanceInitializer({
        name: 'second',
        initialize() {
          secondInitializerRunCount++;
        }
      });

      let firstEngine = FirstEngine.create();
      let firstEngineInstance = buildEngineInstance(firstEngine);

      let secondEngine, secondEngineInstance;

      return firstEngineInstance.boot().then(() => {
        assert.equal(firstInitializerRunCount, 1, 'first initializer only was run when base class created');
        assert.equal(secondInitializerRunCount, 0, 'second initializer was not run when first base class created');
        firstInitializerRunCount = 0;

        secondEngine = SecondEngine.create();
        secondEngineInstance = buildEngineInstance(secondEngine);
        return secondEngineInstance.boot();
      }).then(() => {
        assert.equal(firstInitializerRunCount, 1, 'first initializer was run when subclass created');
        assert.equal(secondInitializerRunCount, 1, 'second initializers was run when subclass created');

        (0, _runloop.run)(() => {
          firstEngineInstance.destroy();
          secondEngineInstance.destroy();

          firstEngine.destroy();
          secondEngine.destroy();
        });
      });
    }

    ['@test initializers are per-engine'](assert) {
      assert.expect(2);

      let FirstEngine = _engine.default.extend();

      FirstEngine.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      expectAssertion(() => {
        FirstEngine.instanceInitializer({
          name: 'abc',
          initialize() {}
        });
      });

      let SecondEngine = _engine.default.extend();
      SecondEngine.instanceInitializer({
        name: 'abc',
        initialize() {}
      });

      assert.ok(true, 'Two engines can have initializers named the same.');
    }

    ['@test initializers are executed in their own context'](assert) {
      assert.expect(1);

      let MyEngine = _engine.default.extend();

      MyEngine.instanceInitializer({
        name: 'coolInitializer',
        myProperty: 'cool',
        initialize() {
          assert.equal(this.myProperty, 'cool', 'should have access to its own context');
        }
      });

      myEngine = MyEngine.create();
      myEngineInstance = buildEngineInstance(myEngine);

      return myEngineInstance.boot();
    }
  });
});
enifed('@ember/engine/tests/engine_instance_test', ['@ember/engine', '@ember/engine/instance', '@ember/runloop', 'internal-test-helpers'], function (_engine, _instance, _runloop, _internalTestHelpers) {
  'use strict';

  let engine, engineInstance;

  (0, _internalTestHelpers.moduleFor)('EngineInstance', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      (0, _runloop.run)(() => {
        engine = _engine.default.create({ router: null });
      });
    }

    teardown() {
      if (engineInstance) {
        (0, _runloop.run)(engineInstance, 'destroy');
        engineInstance = undefined;
      }

      if (engine) {
        (0, _runloop.run)(engine, 'destroy');
        engine = undefined;
      }
    }

    ['@test an engine instance can be created based upon a base engine'](assert) {
      (0, _runloop.run)(() => {
        engineInstance = _instance.default.create({ base: engine });
      });

      assert.ok(engineInstance, 'instance should be created');
      assert.equal(engineInstance.base, engine, 'base should be set to engine');
    }

    ['@test unregistering a factory clears all cached instances of that factory'](assert) {
      assert.expect(3);

      engineInstance = (0, _runloop.run)(() => _instance.default.create({ base: engine }));

      let PostComponent = (0, _internalTestHelpers.factory)();

      engineInstance.register('component:post', PostComponent);

      let postComponent1 = engineInstance.lookup('component:post');
      assert.ok(postComponent1, 'lookup creates instance');

      engineInstance.unregister('component:post');
      engineInstance.register('component:post', PostComponent);

      let postComponent2 = engineInstance.lookup('component:post');
      assert.ok(postComponent2, 'lookup creates instance');

      assert.notStrictEqual(postComponent1, postComponent2, 'lookup creates a brand new instance because previous one was reset');
    }

    ['@test can be booted when its parent has been set'](assert) {
      assert.expect(3);

      engineInstance = (0, _runloop.run)(() => _instance.default.create({ base: engine }));

      expectAssertion(() => {
        engineInstance._bootSync();
      }, "An engine instance's parent must be set via `setEngineParent(engine, parent)` prior to calling `engine.boot()`.");

      (0, _engine.setEngineParent)(engineInstance, {});

      // Stub `cloneParentDependencies`, the internals of which are tested along
      // with application instances.
      engineInstance.cloneParentDependencies = function () {
        assert.ok(true, 'parent dependencies are cloned');
      };

      return engineInstance.boot().then(() => {
        assert.ok(true, 'boot successful');
      });
    }

    ['@test can build a child instance of a registered engine'](assert) {
      let ChatEngine = _engine.default.extend();
      let chatEngineInstance;

      engine.register('engine:chat', ChatEngine);

      (0, _runloop.run)(() => {
        engineInstance = _instance.default.create({ base: engine });

        // Try to build an unregistered engine.
        assert.throws(() => {
          engineInstance.buildChildEngineInstance('fake');
        }, `You attempted to mount the engine 'fake', but it is not registered with its parent.`);

        // Build the `chat` engine, registered above.
        chatEngineInstance = engineInstance.buildChildEngineInstance('chat');
      });

      assert.ok(chatEngineInstance, 'child engine instance successfully created');

      assert.strictEqual((0, _engine.getEngineParent)(chatEngineInstance), engineInstance, 'child engine instance is assigned the correct parent');
    }
  });
});
enifed('@ember/engine/tests/engine_parent_test', ['@ember/engine', 'internal-test-helpers'], function (_engine, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('EngineParent', class extends _internalTestHelpers.AbstractTestCase {
    ["@test An engine's parent can be set with `setEngineParent` and retrieved with `getEngineParent`"](assert) {
      let engine = {};
      let parent = {};

      assert.strictEqual((0, _engine.getEngineParent)(engine), undefined, 'parent has not been set');

      (0, _engine.setEngineParent)(engine, parent);

      assert.strictEqual((0, _engine.getEngineParent)(engine), parent, 'parent has been set');
    }
  });
});
enifed('@ember/engine/tests/engine_test', ['@ember/-internals/environment', '@ember/runloop', '@ember/engine', '@ember/-internals/runtime', '@ember/-internals/container', 'internal-test-helpers'], function (_environment, _runloop, _engine, _runtime, _container, _internalTestHelpers) {
  'use strict';

  let engine;
  let originalLookup = _environment.context.lookup;

  (0, _internalTestHelpers.moduleFor)('Engine', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      (0, _runloop.run)(() => {
        engine = _engine.default.create();
        _environment.context.lookup = { TestEngine: engine };
      });
    }

    teardown() {
      _environment.context.lookup = originalLookup;
      if (engine) {
        (0, _runloop.run)(engine, 'destroy');
        engine = null;
      }
    }

    ['@test acts like a namespace'](assert) {
      engine.Foo = _runtime.Object.extend();
      assert.equal(engine.Foo.toString(), 'TestEngine.Foo', 'Classes pick up their parent namespace');
    }

    ['@test builds a registry'](assert) {
      assert.strictEqual(engine.resolveRegistration('application:main'), engine, `application:main is registered`);
      assert.deepEqual(engine.registeredOptionsForType('component'), { singleton: false }, `optionsForType 'component'`);
      assert.deepEqual(engine.registeredOptionsForType('view'), { singleton: false }, `optionsForType 'view'`);
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'controller:basic');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'view', '_viewRegistry', '-view-registry:main');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'route', '_topLevelViewTemplate', 'template:-outlet');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'view:-outlet', 'namespace', 'application:main');

      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'controller', 'target', 'router:main');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'controller', 'namespace', 'application:main');

      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'router', '_bucketCache', _container.privatize`-bucket-cache:main`);
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'route', '_bucketCache', _container.privatize`-bucket-cache:main`);

      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'route', '_router', 'router:main');

      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'component:-text-field');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'component:-text-area');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'component:-checkbox');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'component:link-to');

      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'service:-routing');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'service:-routing', 'router', 'router:main');

      // DEBUGGING
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'resolver-for-debugging:main');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'container-debug-adapter:main', 'resolver', 'resolver-for-debugging:main');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'data-adapter:main', 'containerDebugAdapter', 'container-debug-adapter:main');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'container-debug-adapter:main');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'component-lookup:main');

      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'service:-dom-changes', 'document', 'service:-document');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'service:-dom-tree-construction', 'document', 'service:-document');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'view:-outlet');
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, _container.privatize`template:components/-default`);
      (0, _internalTestHelpers.verifyRegistration)(assert, engine, 'template:-outlet');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'view:-outlet', 'template', 'template:-outlet');
      (0, _internalTestHelpers.verifyInjection)(assert, engine, 'template', 'compiler', _container.privatize`template-compiler:main`);
      assert.deepEqual(engine.registeredOptionsForType('helper'), { instantiate: false }, `optionsForType 'helper'`);
    }
  });
});
enifed('@ember/error/tests/index_test', ['@ember/error', 'internal-test-helpers'], function (_error, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember Error Throwing', class extends _internalTestHelpers.AbstractTestCase {
    ['@test new EmberError displays provided message'](assert) {
      assert.throws(() => {
        throw new _error.default('A Message');
      }, function (e) {
        return e.message === 'A Message';
      }, 'the assigned message was displayed');
    }
  });
});
enifed('@ember/instrumentation/tests/index-test', ['@ember/instrumentation', 'internal-test-helpers'], function (_instrumentation, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember Instrumentation', class extends _internalTestHelpers.AbstractTestCase {
    afterEach() {
      (0, _instrumentation.reset)();
    }

    ['@test execute block even if no listeners'](assert) {
      let result = (0, _instrumentation.instrument)('render', {}, function () {
        return 'hello';
      });
      assert.equal(result, 'hello', 'called block');
    }

    ['@test subscribing to a simple path receives the listener'](assert) {
      assert.expect(12);

      let sentPayload = {};
      let count = 0;

      (0, _instrumentation.subscribe)('render', {
        before(name, timestamp, payload) {
          if (count === 0) {
            assert.strictEqual(name, 'render');
          } else {
            assert.strictEqual(name, 'render.handlebars');
          }

          assert.ok(typeof timestamp === 'number');
          assert.strictEqual(payload, sentPayload);
        },

        after(name, timestamp, payload) {
          if (count === 0) {
            assert.strictEqual(name, 'render');
          } else {
            assert.strictEqual(name, 'render.handlebars');
          }

          assert.ok(typeof timestamp === 'number');
          assert.strictEqual(payload, sentPayload);

          count++;
        }
      });

      (0, _instrumentation.instrument)('render', sentPayload, function () {});

      (0, _instrumentation.instrument)('render.handlebars', sentPayload, function () {});
    }

    ['@test returning a value from the before callback passes it to the after callback'](assert) {
      assert.expect(2);

      let passthru1 = {};
      let passthru2 = {};

      (0, _instrumentation.subscribe)('render', {
        before() {
          return passthru1;
        },
        after(name, timestamp, payload, beforeValue) {
          assert.strictEqual(beforeValue, passthru1);
        }
      });

      (0, _instrumentation.subscribe)('render', {
        before() {
          return passthru2;
        },
        after(name, timestamp, payload, beforeValue) {
          assert.strictEqual(beforeValue, passthru2);
        }
      });

      (0, _instrumentation.instrument)('render', null, function () {});
    }

    ['@test instrument with 2 args (name, callback) no payload'](assert) {
      assert.expect(1);

      (0, _instrumentation.subscribe)('render', {
        before(name, timestamp, payload) {
          assert.deepEqual(payload, {});
        },
        after() {}
      });

      (0, _instrumentation.instrument)('render', function () {});
    }

    ['@test instrument with 3 args (name, callback, binding) no payload'](assert) {
      assert.expect(2);

      let binding = {};
      (0, _instrumentation.subscribe)('render', {
        before(name, timestamp, payload) {
          assert.deepEqual(payload, {});
        },
        after() {}
      });

      (0, _instrumentation.instrument)('render', function () {
        assert.deepEqual(this, binding);
      }, binding);
    }

    ['@test instrument with 3 args (name, payload, callback) with payload'](assert) {
      assert.expect(1);

      let expectedPayload = { hi: 1 };
      (0, _instrumentation.subscribe)('render', {
        before(name, timestamp, payload) {
          assert.deepEqual(payload, expectedPayload);
        },
        after() {}
      });

      (0, _instrumentation.instrument)('render', expectedPayload, function () {});
    }

    ['@test instrument with 4 args (name, payload, callback, binding) with payload'](assert) {
      assert.expect(2);

      let expectedPayload = { hi: 1 };
      let binding = {};
      (0, _instrumentation.subscribe)('render', {
        before(name, timestamp, payload) {
          assert.deepEqual(payload, expectedPayload);
        },
        after() {}
      });

      (0, _instrumentation.instrument)('render', expectedPayload, function () {
        assert.deepEqual(this, binding);
      }, binding);
    }

    ['@test raising an exception in the instrumentation attaches it to the payload'](assert) {
      assert.expect(2);

      let error = new Error('Instrumentation');

      (0, _instrumentation.subscribe)('render', {
        before() {},
        after(name, timestamp, payload) {
          assert.strictEqual(payload.exception, error);
        }
      });

      (0, _instrumentation.subscribe)('render', {
        before() {},
        after(name, timestamp, payload) {
          assert.strictEqual(payload.exception, error);
        }
      });

      (0, _instrumentation.instrument)('render.handlebars', null, function () {
        throw error;
      });
    }

    ['@test it is possible to add a new subscriber after the first instrument'](assert) {
      (0, _instrumentation.instrument)('render.handlebars', null, function () {});

      (0, _instrumentation.subscribe)('render', {
        before() {
          assert.ok(true, 'Before callback was called');
        },
        after() {
          assert.ok(true, 'After callback was called');
        }
      });

      (0, _instrumentation.instrument)('render.handlebars', null, function () {});
    }

    ['@test it is possible to remove a subscriber'](assert) {
      assert.expect(4);

      let count = 0;

      let subscriber = (0, _instrumentation.subscribe)('render', {
        before() {
          assert.equal(count, 0);
          assert.ok(true, 'Before callback was called');
        },
        after() {
          assert.equal(count, 0);
          assert.ok(true, 'After callback was called');
          count++;
        }
      });

      (0, _instrumentation.instrument)('render.handlebars', null, function () {});

      (0, _instrumentation.unsubscribe)(subscriber);

      (0, _instrumentation.instrument)('render.handlebars', null, function () {});
    }
  });
});
enifed('@ember/map/tests/map_test', ['@ember/map', '@ember/map/with-default', '@ember/map/lib/ordered-set', 'internal-test-helpers'], function (_map, _withDefault, _orderedSet, _internalTestHelpers) {
  'use strict';

  let object, number, string, map, variety;
  const varieties = [['Map', _map.default], ['MapWithDefault', _withDefault.default]];

  function testMap(nameAndFunc) {
    variety = nameAndFunc[0];

    (0, _internalTestHelpers.moduleFor)('Ember.' + variety + ' (forEach and get are implicitly tested)', class extends _internalTestHelpers.AbstractTestCase {
      beforeEach() {
        object = {};
        number = 42;
        string = 'foo';

        expectDeprecation(() => {
          map = nameAndFunc[1].create();
        }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
      }

      ['@test set'](assert) {
        map.set(object, 'winning');
        map.set(number, 'winning');
        map.set(string, 'winning');

        mapHasEntries(assert, [[object, 'winning'], [number, 'winning'], [string, 'winning']]);

        map.set(object, 'losing');
        map.set(number, 'losing');
        map.set(string, 'losing');

        mapHasEntries(assert, [[object, 'losing'], [number, 'losing'], [string, 'losing']]);

        assert.equal(map.has('nope'), false, 'expected the key `nope` to not be present');
        assert.equal(map.has({}), false, 'expected they key `{}` to not be present');
      }

      ['@test set chaining'](assert) {
        map.set(object, 'winning').set(number, 'winning').set(string, 'winning');

        mapHasEntries(assert, [[object, 'winning'], [number, 'winning'], [string, 'winning']]);

        map.set(object, 'losing').set(number, 'losing').set(string, 'losing');

        mapHasEntries(assert, [[object, 'losing'], [number, 'losing'], [string, 'losing']]);

        assert.equal(map.has('nope'), false, 'expected the key `nope` to not be present');
        assert.equal(map.has({}), false, 'expected they key `{}` to not be present');
      }

      ['@test with key with undefined value'](assert) {
        map.set('foo', undefined);

        map.forEach(function (value, key) {
          assert.equal(value, undefined);
          assert.equal(key, 'foo');
        });

        assert.ok(map.has('foo'), 'has key foo, even with undefined value');

        assert.equal(map.size, 1);
      }

      ['@test arity of forEach is 1 – es6 23.1.3.5'](assert) {
        assert.equal(map.forEach.length, 1, 'expected arity for map.forEach is 1');
      }

      ['@test forEach throws without a callback as the first argument'](assert) {
        assert.equal(map.forEach.length, 1, 'expected arity for map.forEach is 1');
      }

      ['@test has empty collection'](assert) {
        assert.equal(map.has('foo'), false);
        assert.equal(map.has(), false);
      }

      ['@test delete'](assert) {
        expectNoDeprecation();

        map.set(object, 'winning');
        map.set(number, 'winning');
        map.set(string, 'winning');

        map.delete(object);
        map.delete(number);
        map.delete(string);

        // doesn't explode
        map.delete({});

        mapHasEntries(assert, []);
      }

      ['@test copy and then update'](assert) {
        map.set(object, 'winning');
        map.set(number, 'winning');
        map.set(string, 'winning');

        let map2;
        expectDeprecation(() => {
          map2 = map.copy();
        }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

        map2.set(object, 'losing');
        map2.set(number, 'losing');
        map2.set(string, 'losing');

        mapHasEntries(assert, [[object, 'winning'], [number, 'winning'], [string, 'winning']]);

        mapHasEntries(assert, [[object, 'losing'], [number, 'losing'], [string, 'losing']], map2);
      }

      ['@test copy and then delete'](assert) {
        map.set(object, 'winning');
        map.set(number, 'winning');
        map.set(string, 'winning');

        let map2;
        expectDeprecation(() => {
          map2 = map.copy();
        }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

        map2.delete(object);
        map2.delete(number);
        map2.delete(string);

        mapHasEntries(assert, [[object, 'winning'], [number, 'winning'], [string, 'winning']]);

        mapHasEntries(assert, [], map2);
      }

      ['@test size'](assert) {
        //Add a key twice
        assert.equal(map.size, 0);
        map.set(string, 'a string');
        assert.equal(map.size, 1);
        map.set(string, 'the same string');
        assert.equal(map.size, 1);

        //Add another
        map.set(number, 'a number');
        assert.equal(map.size, 2);

        //Remove one that doesn't exist
        map.delete('does not exist');
        assert.equal(map.size, 2);

        //Check copy
        expectDeprecation(() => {
          let copy = map.copy();
          assert.equal(copy.size, 2);
        }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

        //Remove a key twice
        map.delete(number);
        assert.equal(map.size, 1);
        map.delete(number);
        assert.equal(map.size, 1);

        //Remove the last key
        map.delete(string);
        assert.equal(map.size, 0);
        map.delete(string);
        assert.equal(map.size, 0);
      }

      ['@test forEach without proper callback'](assert) {
        expectAssertion(function () {
          map.forEach();
        }, '[object Undefined] is not a function');

        expectAssertion(function () {
          map.forEach(undefined);
        }, '[object Undefined] is not a function');

        expectAssertion(function () {
          map.forEach(1);
        }, '[object Number] is not a function');

        expectAssertion(function () {
          map.forEach({});
        }, '[object Object] is not a function');

        map.forEach(function (value, key) {
          map.delete(key);
        });
        // ensure the error happens even if no data is present
        assert.equal(map.size, 0);
        expectAssertion(function () {
          map.forEach({});
        }, '[object Object] is not a function');
      }

      ['@test forEach basic'](assert) {
        map.set('a', 1);
        map.set('b', 2);
        map.set('c', 3);

        let iteration = 0;

        let expectations = [{ value: 1, key: 'a', context: unboundThis }, { value: 2, key: 'b', context: unboundThis }, { value: 3, key: 'c', context: unboundThis }];

        map.forEach(function (value, key, theMap) {
          let expectation = expectations[iteration];

          assert.equal(value, expectation.value, 'value should be correct');
          assert.equal(key, expectation.key, 'key should be correct');
          assert.equal(this, expectation.context, 'context should be as if it was unbound');
          assert.equal(map, theMap, 'map being iterated over should be passed in');

          iteration++;
        });

        assert.equal(iteration, 3, 'expected 3 iterations');
      }

      ['@test forEach basic /w context'](assert) {
        map.set('a', 1);
        map.set('b', 2);
        map.set('c', 3);

        let iteration = 0;
        let context = {};
        let expectations = [{ value: 1, key: 'a', context: context }, { value: 2, key: 'b', context: context }, { value: 3, key: 'c', context: context }];

        map.forEach(function (value, key, theMap) {
          let expectation = expectations[iteration];

          assert.equal(value, expectation.value, 'value should be correct');
          assert.equal(key, expectation.key, 'key should be correct');
          assert.equal(this, expectation.context, 'context should be as if it was unbound');
          assert.equal(map, theMap, 'map being iterated over should be passed in');

          iteration++;
        }, context);

        assert.equal(iteration, 3, 'expected 3 iterations');
      }

      ['@test forEach basic /w deletion while enumerating'](assert) {
        map.set('a', 1);
        map.set('b', 2);
        map.set('c', 3);

        let iteration = 0;

        let expectations = [{ value: 1, key: 'a', context: unboundThis }, { value: 2, key: 'b', context: unboundThis }];

        map.forEach(function (value, key, theMap) {
          if (iteration === 0) {
            map.delete('c');
          }

          let expectation = expectations[iteration];

          assert.equal(value, expectation.value, 'value should be correct');
          assert.equal(key, expectation.key, 'key should be correct');
          assert.equal(this, expectation.context, 'context should be as if it was unbound');
          assert.equal(map, theMap, 'map being iterated over should be passed in');

          iteration++;
        });

        assert.equal(iteration, 2, 'expected 3 iterations');
      }

      ['@test forEach basic /w addition while enumerating'](assert) {
        map.set('a', 1);
        map.set('b', 2);
        map.set('c', 3);

        let iteration = 0;

        let expectations = [{ value: 1, key: 'a', context: unboundThis }, { value: 2, key: 'b', context: unboundThis }, { value: 3, key: 'c', context: unboundThis }, { value: 4, key: 'd', context: unboundThis }];

        map.forEach(function (value, key, theMap) {
          if (iteration === 0) {
            map.set('d', 4);
          }

          let expectation = expectations[iteration];

          assert.equal(value, expectation.value, 'value should be correct');
          assert.equal(key, expectation.key, 'key should be correct');
          assert.equal(this, expectation.context, 'context should be as if it was unbound');
          assert.equal(map, theMap, 'map being iterated over should be passed in');

          iteration++;
        });

        assert.equal(iteration, 4, 'expected 3 iterations');
      }

      ['@test clear'](assert) {
        let iterations = 0;

        map.set('a', 1);
        map.set('b', 2);
        map.set('c', 3);
        map.set('d', 4);

        assert.equal(map.size, 4);

        map.forEach(function () {
          iterations++;
        });
        assert.equal(iterations, 4);

        map.clear();
        assert.equal(map.size, 0);
        iterations = 0;
        map.forEach(function () {
          iterations++;
        });
        assert.equal(iterations, 0);
      }

      ['@skip -0'](assert) {
        assert.equal(map.has(-0), false);
        assert.equal(map.has(0), false);

        map.set(-0, 'zero');

        assert.equal(map.has(-0), true);
        assert.equal(map.has(0), true);

        assert.equal(map.get(0), 'zero');
        assert.equal(map.get(-0), 'zero');

        map.forEach(function (value, key) {
          assert.equal(1 / key, Infinity, 'spec says key should be positive zero');
        });
      }

      ['@test NaN'](assert) {
        assert.equal(map.has(NaN), false);

        map.set(NaN, 'not-a-number');

        assert.equal(map.has(NaN), true);

        assert.equal(map.get(NaN), 'not-a-number');
      }

      ['@test NaN Boxed'](assert) {
        let boxed = new Number(NaN);
        assert.equal(map.has(boxed), false);

        map.set(boxed, 'not-a-number');

        assert.equal(map.has(boxed), true);
        assert.equal(map.has(NaN), false);

        assert.equal(map.get(NaN), undefined);
        assert.equal(map.get(boxed), 'not-a-number');
      }

      ['@test 0 value'](assert) {
        let obj = {};
        assert.equal(map.has(obj), false);

        assert.equal(map.size, 0);
        map.set(obj, 0);
        assert.equal(map.size, 1);

        assert.equal(map.has(obj), true);
        assert.equal(map.get(obj), 0);

        map.delete(obj);
        assert.equal(map.has(obj), false);
        assert.equal(map.get(obj), undefined);
        assert.equal(map.size, 0);
      }
    });

    let mapHasLength = function (assert, expected, theMap) {
      theMap = theMap || map;

      let length = 0;
      theMap.forEach(function () {
        length++;
      });

      assert.equal(length, expected, 'map should contain ' + expected + ' items');
    };

    let mapHasEntries = function (assert, entries, theMap) {
      theMap = theMap || map;

      for (let i = 0; i < entries.length; i++) {
        assert.equal(theMap.get(entries[i][0]), entries[i][1]);
        assert.equal(theMap.has(entries[i][0]), true);
      }

      mapHasLength(assert, entries.length, theMap);
    };

    let unboundThis;

    (function () {
      unboundThis = this;
    })();
  }

  for (let i = 0; i < varieties.length; i++) {
    testMap(varieties[i]);
  }

  (0, _internalTestHelpers.moduleFor)('MapWithDefault - default values', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Retrieving a value that has not been set returns and sets a default value'](assert) {
      let map;
      expectDeprecation(() => {
        map = _withDefault.default.create({
          defaultValue(key) {
            return [key];
          }
        });
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

      let value = map.get('ohai');
      assert.deepEqual(value, ['ohai']);

      assert.strictEqual(value, map.get('ohai'));
    }

    ['@test Map.prototype.constructor'](assert) {
      expectDeprecation(() => {
        let map = new _map.default();
        assert.equal(map.constructor, _map.default);
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
    }

    ['@test MapWithDefault.prototype.constructor'](assert) {
      expectDeprecation(() => {
        let map = new _withDefault.default({
          defaultValue(key) {
            return key;
          }
        });
        assert.equal(map.constructor, _withDefault.default);
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
    }

    ['@test Copying a MapWithDefault copies the default value'](assert) {
      let map;
      expectDeprecation(() => {
        map = _withDefault.default.create({
          defaultValue(key) {
            return [key];
          }
        });
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

      map.set('ohai', 1);
      map.get('bai');

      let map2;
      expectDeprecation(() => {
        map2 = map.copy();
      }, 'Use of @ember/Map is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });

      assert.equal(map2.get('ohai'), 1);
      assert.deepEqual(map2.get('bai'), ['bai']);

      map2.set('kthx', 3);

      assert.deepEqual(map.get('kthx'), ['kthx']);
      assert.equal(map2.get('kthx'), 3);

      assert.deepEqual(map2.get('default'), ['default']);

      map2.defaultValue = key => ['tom is on', key];

      assert.deepEqual(map2.get('drugs'), ['tom is on', 'drugs']);
    }
  });

  (0, _internalTestHelpers.moduleFor)('OrderedSet', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      object = {};
      number = 42;
      string = 'foo';

      expectDeprecation(() => {
        map = _orderedSet.default.create();
      }, 'Use of @ember/OrderedSet is deprecated. Please use native `Map` instead', { id: 'ember-map-deprecation', until: '3.5.0' });
    }

    ['@test add returns the set'](assert) {
      let obj = {};
      assert.equal(map.add(obj), map);
      assert.equal(map.add(obj), map, 'when it is already in the set');
    }
  });

  (0, _internalTestHelpers.moduleFor)('__OrderedSet__', class extends _internalTestHelpers.AbstractTestCase {
    ['@test private __OrderedSet__ can be created without deprecation']() {
      expectNoDeprecation();
      _orderedSet.__OrderedSet__.create();
    }
  });
});
enifed('@ember/object/tests/computed/computed_macros_test', ['@ember/-internals/metal', '@ember/object/computed', '@ember/-internals/runtime', 'internal-test-helpers'], function (_metal, _computed, _runtime, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('CP macros', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Ember.computed.empty'](assert) {
      let obj = _runtime.Object.extend({
        bestLannister: null,
        lannisters: null,

        bestLannisterUnspecified: (0, _computed.empty)('bestLannister'),
        noLannistersKnown: (0, _computed.empty)('lannisters')
      }).create({
        lannisters: (0, _runtime.A)()
      });

      assert.equal((0, _metal.get)(obj, 'bestLannisterUnspecified'), true, 'bestLannister initially empty');
      assert.equal((0, _metal.get)(obj, 'noLannistersKnown'), true, 'lannisters initially empty');

      (0, _metal.get)(obj, 'lannisters').pushObject('Tyrion');
      (0, _metal.set)(obj, 'bestLannister', 'Tyrion');

      assert.equal((0, _metal.get)(obj, 'bestLannisterUnspecified'), false, 'empty respects strings');
      assert.equal((0, _metal.get)(obj, 'noLannistersKnown'), false, 'empty respects array mutations');
    }

    ['@test Ember.computed.notEmpty'](assert) {
      let obj = _runtime.Object.extend({
        bestLannister: null,
        lannisters: null,

        bestLannisterSpecified: (0, _computed.notEmpty)('bestLannister'),
        LannistersKnown: (0, _computed.notEmpty)('lannisters')
      }).create({
        lannisters: (0, _runtime.A)()
      });

      assert.equal((0, _metal.get)(obj, 'bestLannisterSpecified'), false, 'bestLannister initially empty');
      assert.equal((0, _metal.get)(obj, 'LannistersKnown'), false, 'lannisters initially empty');

      (0, _metal.get)(obj, 'lannisters').pushObject('Tyrion');
      (0, _metal.set)(obj, 'bestLannister', 'Tyrion');

      assert.equal((0, _metal.get)(obj, 'bestLannisterSpecified'), true, 'empty respects strings');
      assert.equal((0, _metal.get)(obj, 'LannistersKnown'), true, 'empty respects array mutations');
    }

    ['@test computed.not'](assert) {
      let obj = { foo: true };
      (0, _metal.defineProperty)(obj, 'notFoo', (0, _computed.not)('foo'));
      assert.equal((0, _metal.get)(obj, 'notFoo'), false);

      obj = { foo: { bar: true } };
      (0, _metal.defineProperty)(obj, 'notFoo', (0, _computed.not)('foo.bar'));
      assert.equal((0, _metal.get)(obj, 'notFoo'), false);
    }

    ['@test computed.empty'](assert) {
      let obj = { foo: [], bar: undefined, baz: null, quz: '' };
      (0, _metal.defineProperty)(obj, 'fooEmpty', (0, _computed.empty)('foo'));
      (0, _metal.defineProperty)(obj, 'barEmpty', (0, _computed.empty)('bar'));
      (0, _metal.defineProperty)(obj, 'bazEmpty', (0, _computed.empty)('baz'));
      (0, _metal.defineProperty)(obj, 'quzEmpty', (0, _computed.empty)('quz'));

      assert.equal((0, _metal.get)(obj, 'fooEmpty'), true);
      (0, _metal.set)(obj, 'foo', [1]);
      assert.equal((0, _metal.get)(obj, 'fooEmpty'), false);
      assert.equal((0, _metal.get)(obj, 'barEmpty'), true);
      assert.equal((0, _metal.get)(obj, 'bazEmpty'), true);
      assert.equal((0, _metal.get)(obj, 'quzEmpty'), true);
      (0, _metal.set)(obj, 'quz', 'asdf');
      assert.equal((0, _metal.get)(obj, 'quzEmpty'), false);
    }

    ['@test computed.bool'](assert) {
      let obj = { foo() {}, bar: 'asdf', baz: null, quz: false };
      (0, _metal.defineProperty)(obj, 'fooBool', (0, _computed.bool)('foo'));
      (0, _metal.defineProperty)(obj, 'barBool', (0, _computed.bool)('bar'));
      (0, _metal.defineProperty)(obj, 'bazBool', (0, _computed.bool)('baz'));
      (0, _metal.defineProperty)(obj, 'quzBool', (0, _computed.bool)('quz'));
      assert.equal((0, _metal.get)(obj, 'fooBool'), true);
      assert.equal((0, _metal.get)(obj, 'barBool'), true);
      assert.equal((0, _metal.get)(obj, 'bazBool'), false);
      assert.equal((0, _metal.get)(obj, 'quzBool'), false);
    }

    ['@test computed.alias'](assert) {
      let obj = { bar: 'asdf', baz: null, quz: false };
      (0, _metal.defineProperty)(obj, 'bay', (0, _metal.computed)(function () {
        return 'apple';
      }));

      (0, _metal.defineProperty)(obj, 'barAlias', (0, _metal.alias)('bar'));
      (0, _metal.defineProperty)(obj, 'bazAlias', (0, _metal.alias)('baz'));
      (0, _metal.defineProperty)(obj, 'quzAlias', (0, _metal.alias)('quz'));
      (0, _metal.defineProperty)(obj, 'bayAlias', (0, _metal.alias)('bay'));

      assert.equal((0, _metal.get)(obj, 'barAlias'), 'asdf');
      assert.equal((0, _metal.get)(obj, 'bazAlias'), null);
      assert.equal((0, _metal.get)(obj, 'quzAlias'), false);
      assert.equal((0, _metal.get)(obj, 'bayAlias'), 'apple');

      (0, _metal.set)(obj, 'barAlias', 'newBar');
      (0, _metal.set)(obj, 'bazAlias', 'newBaz');
      (0, _metal.set)(obj, 'quzAlias', null);

      assert.equal((0, _metal.get)(obj, 'barAlias'), 'newBar');
      assert.equal((0, _metal.get)(obj, 'bazAlias'), 'newBaz');
      assert.equal((0, _metal.get)(obj, 'quzAlias'), null);

      assert.equal((0, _metal.get)(obj, 'bar'), 'newBar');
      assert.equal((0, _metal.get)(obj, 'baz'), 'newBaz');
      assert.equal((0, _metal.get)(obj, 'quz'), null);
    }

    ['@test computed.alias set'](assert) {
      let obj = {};
      let constantValue = 'always `a`';

      (0, _metal.defineProperty)(obj, 'original', (0, _metal.computed)({
        get: function () {
          return constantValue;
        },
        set: function () {
          return constantValue;
        }
      }));
      (0, _metal.defineProperty)(obj, 'aliased', (0, _metal.alias)('original'));

      assert.equal((0, _metal.get)(obj, 'original'), constantValue);
      assert.equal((0, _metal.get)(obj, 'aliased'), constantValue);

      (0, _metal.set)(obj, 'aliased', 'should not set to this value');

      assert.equal((0, _metal.get)(obj, 'original'), constantValue);
      assert.equal((0, _metal.get)(obj, 'aliased'), constantValue);
    }

    ['@test computed.match'](assert) {
      let obj = { name: 'Paul' };
      (0, _metal.defineProperty)(obj, 'isPaul', (0, _computed.match)('name', /Paul/));

      assert.equal((0, _metal.get)(obj, 'isPaul'), true, 'is Paul');

      (0, _metal.set)(obj, 'name', 'Pierre');

      assert.equal((0, _metal.get)(obj, 'isPaul'), false, 'is not Paul anymore');
    }

    ['@test computed.notEmpty'](assert) {
      let obj = { items: [1] };
      (0, _metal.defineProperty)(obj, 'hasItems', (0, _computed.notEmpty)('items'));

      assert.equal((0, _metal.get)(obj, 'hasItems'), true, 'is not empty');

      (0, _metal.set)(obj, 'items', []);

      assert.equal((0, _metal.get)(obj, 'hasItems'), false, 'is empty');
    }

    ['@test computed.equal'](assert) {
      let obj = { name: 'Paul' };
      (0, _metal.defineProperty)(obj, 'isPaul', (0, _computed.equal)('name', 'Paul'));

      assert.equal((0, _metal.get)(obj, 'isPaul'), true, 'is Paul');

      (0, _metal.set)(obj, 'name', 'Pierre');

      assert.equal((0, _metal.get)(obj, 'isPaul'), false, 'is not Paul anymore');
    }

    ['@test computed.gt'](assert) {
      let obj = { number: 2 };
      (0, _metal.defineProperty)(obj, 'isGreaterThenOne', (0, _computed.gt)('number', 1));

      assert.equal((0, _metal.get)(obj, 'isGreaterThenOne'), true, 'is gt');

      (0, _metal.set)(obj, 'number', 1);

      assert.equal((0, _metal.get)(obj, 'isGreaterThenOne'), false, 'is not gt');

      (0, _metal.set)(obj, 'number', 0);

      assert.equal((0, _metal.get)(obj, 'isGreaterThenOne'), false, 'is not gt');
    }

    ['@test computed.gte'](assert) {
      let obj = { number: 2 };
      (0, _metal.defineProperty)(obj, 'isGreaterOrEqualThenOne', (0, _computed.gte)('number', 1));

      assert.equal((0, _metal.get)(obj, 'isGreaterOrEqualThenOne'), true, 'is gte');

      (0, _metal.set)(obj, 'number', 1);

      assert.equal((0, _metal.get)(obj, 'isGreaterOrEqualThenOne'), true, 'is gte');

      (0, _metal.set)(obj, 'number', 0);

      assert.equal((0, _metal.get)(obj, 'isGreaterOrEqualThenOne'), false, 'is not gte');
    }

    ['@test computed.lt'](assert) {
      let obj = { number: 0 };
      (0, _metal.defineProperty)(obj, 'isLesserThenOne', (0, _computed.lt)('number', 1));

      assert.equal((0, _metal.get)(obj, 'isLesserThenOne'), true, 'is lt');

      (0, _metal.set)(obj, 'number', 1);

      assert.equal((0, _metal.get)(obj, 'isLesserThenOne'), false, 'is not lt');

      (0, _metal.set)(obj, 'number', 2);

      assert.equal((0, _metal.get)(obj, 'isLesserThenOne'), false, 'is not lt');
    }

    ['@test computed.lte'](assert) {
      let obj = { number: 0 };
      (0, _metal.defineProperty)(obj, 'isLesserOrEqualThenOne', (0, _computed.lte)('number', 1));

      assert.equal((0, _metal.get)(obj, 'isLesserOrEqualThenOne'), true, 'is lte');

      (0, _metal.set)(obj, 'number', 1);

      assert.equal((0, _metal.get)(obj, 'isLesserOrEqualThenOne'), true, 'is lte');

      (0, _metal.set)(obj, 'number', 2);

      assert.equal((0, _metal.get)(obj, 'isLesserOrEqualThenOne'), false, 'is not lte');
    }

    ['@test computed.and two properties'](assert) {
      let obj = { one: true, two: true };
      (0, _metal.defineProperty)(obj, 'oneAndTwo', (0, _computed.and)('one', 'two'));

      assert.equal((0, _metal.get)(obj, 'oneAndTwo'), true, 'one and two');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneAndTwo'), false, 'one and not two');

      (0, _metal.set)(obj, 'one', null);
      (0, _metal.set)(obj, 'two', 'Yes');

      assert.equal((0, _metal.get)(obj, 'oneAndTwo'), null, 'returns falsy value as in &&');

      (0, _metal.set)(obj, 'one', true);
      (0, _metal.set)(obj, 'two', 2);

      assert.equal((0, _metal.get)(obj, 'oneAndTwo'), 2, 'returns truthy value as in &&');
    }

    ['@test computed.and three properties'](assert) {
      let obj = { one: true, two: true, three: true };
      (0, _metal.defineProperty)(obj, 'oneTwoThree', (0, _computed.and)('one', 'two', 'three'));

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one and two and three');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), false, 'one and not two and not three');

      (0, _metal.set)(obj, 'one', true);
      (0, _metal.set)(obj, 'two', 2);
      (0, _metal.set)(obj, 'three', 3);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), 3, 'returns truthy value as in &&');
    }

    ['@test computed.and expand properties'](assert) {
      let obj = { one: true, two: true, three: true };
      (0, _metal.defineProperty)(obj, 'oneTwoThree', (0, _computed.and)('{one,two,three}'));

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one and two and three');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), false, 'one and not two and not three');

      (0, _metal.set)(obj, 'one', true);
      (0, _metal.set)(obj, 'two', 2);
      (0, _metal.set)(obj, 'three', 3);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), 3, 'returns truthy value as in &&');
    }

    ['@test computed.or two properties'](assert) {
      let obj = { one: true, two: true };
      (0, _metal.defineProperty)(obj, 'oneOrTwo', (0, _computed.or)('one', 'two'));

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), true, 'one or two');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), true, 'one or two');

      (0, _metal.set)(obj, 'two', false);

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), false, 'nor one nor two');

      (0, _metal.set)(obj, 'two', null);

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), null, 'returns last falsy value as in ||');

      (0, _metal.set)(obj, 'two', true);

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), true, 'one or two');

      (0, _metal.set)(obj, 'one', 1);

      assert.equal((0, _metal.get)(obj, 'oneOrTwo'), 1, 'returns truthy value as in ||');
    }

    ['@test computed.or three properties'](assert) {
      let obj = { one: true, two: true, three: true };
      (0, _metal.defineProperty)(obj, 'oneTwoThree', (0, _computed.or)('one', 'two', 'three'));

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'two', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'three', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), false, 'one or two or three');

      (0, _metal.set)(obj, 'three', null);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), null, 'returns last falsy value as in ||');

      (0, _metal.set)(obj, 'two', true);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'one', 1);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), 1, 'returns truthy value as in ||');
    }

    ['@test computed.or expand properties'](assert) {
      let obj = { one: true, two: true, three: true };
      (0, _metal.defineProperty)(obj, 'oneTwoThree', (0, _computed.or)('{one,two,three}'));

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'one', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'two', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'three', false);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), false, 'one or two or three');

      (0, _metal.set)(obj, 'three', null);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), null, 'returns last falsy value as in ||');

      (0, _metal.set)(obj, 'two', true);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), true, 'one or two or three');

      (0, _metal.set)(obj, 'one', 1);

      assert.equal((0, _metal.get)(obj, 'oneTwoThree'), 1, 'returns truthy value as in ||');
    }

    ['@test computed.or and computed.and warn about dependent keys with spaces']() {
      let obj = { one: true, two: true };
      expectAssertion(function () {
        (0, _metal.defineProperty)(obj, 'oneOrTwo', (0, _computed.or)('one', 'two three'));
      }, /Dependent keys passed to computed\.or\(\) can't have spaces\./);

      expectAssertion(function () {
        (0, _metal.defineProperty)(obj, 'oneAndTwo', (0, _computed.and)('one', 'two three'));
      }, /Dependent keys passed to computed\.and\(\) can't have spaces\./);
    }

    ['@test computed.oneWay'](assert) {
      let obj = {
        firstName: 'Teddy',
        lastName: 'Zeenny'
      };

      (0, _metal.defineProperty)(obj, 'nickName', (0, _computed.oneWay)('firstName'));

      assert.equal((0, _metal.get)(obj, 'firstName'), 'Teddy');
      assert.equal((0, _metal.get)(obj, 'lastName'), 'Zeenny');
      assert.equal((0, _metal.get)(obj, 'nickName'), 'Teddy');

      (0, _metal.set)(obj, 'nickName', 'TeddyBear');

      assert.equal((0, _metal.get)(obj, 'firstName'), 'Teddy');
      assert.equal((0, _metal.get)(obj, 'lastName'), 'Zeenny');

      assert.equal((0, _metal.get)(obj, 'nickName'), 'TeddyBear');

      (0, _metal.set)(obj, 'firstName', 'TEDDDDDDDDYYY');

      assert.equal((0, _metal.get)(obj, 'nickName'), 'TeddyBear');
    }

    ['@test computed.readOnly'](assert) {
      let obj = {
        firstName: 'Teddy',
        lastName: 'Zeenny'
      };

      (0, _metal.defineProperty)(obj, 'nickName', (0, _computed.readOnly)('firstName'));

      assert.equal((0, _metal.get)(obj, 'firstName'), 'Teddy');
      assert.equal((0, _metal.get)(obj, 'lastName'), 'Zeenny');
      assert.equal((0, _metal.get)(obj, 'nickName'), 'Teddy');

      assert.throws(function () {
        (0, _metal.set)(obj, 'nickName', 'TeddyBear');
      }, / /);

      assert.equal((0, _metal.get)(obj, 'firstName'), 'Teddy');
      assert.equal((0, _metal.get)(obj, 'lastName'), 'Zeenny');

      assert.equal((0, _metal.get)(obj, 'nickName'), 'Teddy');

      (0, _metal.set)(obj, 'firstName', 'TEDDDDDDDDYYY');

      assert.equal((0, _metal.get)(obj, 'nickName'), 'TEDDDDDDDDYYY');
    }

    ['@test computed.deprecatingAlias'](assert) {
      let obj = { bar: 'asdf', baz: null, quz: false };
      (0, _metal.defineProperty)(obj, 'bay', (0, _metal.computed)(function () {
        return 'apple';
      }));

      (0, _metal.defineProperty)(obj, 'barAlias', (0, _computed.deprecatingAlias)('bar', { id: 'bar-deprecation', until: 'some.version' }));
      (0, _metal.defineProperty)(obj, 'bazAlias', (0, _computed.deprecatingAlias)('baz', { id: 'baz-deprecation', until: 'some.version' }));
      (0, _metal.defineProperty)(obj, 'quzAlias', (0, _computed.deprecatingAlias)('quz', { id: 'quz-deprecation', until: 'some.version' }));
      (0, _metal.defineProperty)(obj, 'bayAlias', (0, _computed.deprecatingAlias)('bay', { id: 'bay-deprecation', until: 'some.version' }));

      expectDeprecation(function () {
        assert.equal((0, _metal.get)(obj, 'barAlias'), 'asdf');
      }, 'Usage of `barAlias` is deprecated, use `bar` instead.');

      expectDeprecation(function () {
        assert.equal((0, _metal.get)(obj, 'bazAlias'), null);
      }, 'Usage of `bazAlias` is deprecated, use `baz` instead.');

      expectDeprecation(function () {
        assert.equal((0, _metal.get)(obj, 'quzAlias'), false);
      }, 'Usage of `quzAlias` is deprecated, use `quz` instead.');

      expectDeprecation(function () {
        assert.equal((0, _metal.get)(obj, 'bayAlias'), 'apple');
      }, 'Usage of `bayAlias` is deprecated, use `bay` instead.');

      expectDeprecation(function () {
        (0, _metal.set)(obj, 'barAlias', 'newBar');
      }, 'Usage of `barAlias` is deprecated, use `bar` instead.');

      expectDeprecation(function () {
        (0, _metal.set)(obj, 'bazAlias', 'newBaz');
      }, 'Usage of `bazAlias` is deprecated, use `baz` instead.');

      expectDeprecation(function () {
        (0, _metal.set)(obj, 'quzAlias', null);
      }, 'Usage of `quzAlias` is deprecated, use `quz` instead.');

      assert.equal((0, _metal.get)(obj, 'barAlias'), 'newBar');
      assert.equal((0, _metal.get)(obj, 'bazAlias'), 'newBaz');
      assert.equal((0, _metal.get)(obj, 'quzAlias'), null);

      assert.equal((0, _metal.get)(obj, 'bar'), 'newBar');
      assert.equal((0, _metal.get)(obj, 'baz'), 'newBaz');
      assert.equal((0, _metal.get)(obj, 'quz'), null);
    }
  });
});
enifed('@ember/object/tests/computed/reduce_computed_macros_test', ['@ember/runloop', '@ember/-internals/metal', '@ember/-internals/runtime', '@ember/object/computed', '@ember/canary-features', 'internal-test-helpers'], function (_runloop, _metal, _runtime, _computed, _canaryFeatures, _internalTestHelpers) {
  'use strict';

  let obj;
  (0, _internalTestHelpers.moduleFor)('map', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        mapped: (0, _computed.map)('array.@each.v', item => item.v),
        mappedObjects: (0, _computed.map)('arrayObjects.@each.v', item => ({
          name: item.v.name
        }))
      }).create({
        arrayObjects: (0, _runtime.A)([{ v: { name: 'Robert' } }, { v: { name: 'Leanna' } }]),

        array: (0, _runtime.A)([{ v: 1 }, { v: 3 }, { v: 2 }, { v: 1 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test map is readOnly'](assert) {
      assert.throws(function () {
        obj.set('mapped', 1);
      }, /Cannot set read-only property "mapped" on object:/);
    }

    ['@test it maps simple properties'](assert) {
      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 1]);

      obj.get('array').pushObject({ v: 5 });

      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 1, 5]);

      (0, _runtime.removeAt)(obj.get('array'), 3);

      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 5]);
    }

    ['@test it maps simple unshifted properties'](assert) {
      let array = (0, _runtime.A)();

      obj = _runtime.Object.extend({
        mapped: (0, _computed.map)('array', item => item.toUpperCase())
      }).create({
        array
      });

      array.unshiftObject('c');
      array.unshiftObject('b');
      array.unshiftObject('a');

      array.popObject();

      assert.deepEqual(obj.get('mapped'), ['A', 'B'], 'properties unshifted in sequence are mapped correctly');
    }

    ['@test it has the correct `this`'](assert) {
      obj = _runtime.Object.extend({
        mapped: (0, _computed.map)('array', function (item) {
          assert.equal(this, obj, 'should have correct context');
          return this.upperCase(item);
        }),
        upperCase(string) {
          return string.toUpperCase();
        }
      }).create({
        array: ['a', 'b', 'c']
      });

      assert.deepEqual(obj.get('mapped'), ['A', 'B', 'C'], 'properties unshifted in sequence are mapped correctly');
    }

    ['@test it passes the index to the callback'](assert) {
      let array = ['a', 'b', 'c'];

      obj = _runtime.Object.extend({
        mapped: (0, _computed.map)('array', (item, index) => index)
      }).create({
        array
      });

      assert.deepEqual(obj.get('mapped'), [0, 1, 2], 'index is passed to callback correctly');
    }

    ['@test it maps objects'](assert) {
      assert.deepEqual(obj.get('mappedObjects'), [{ name: 'Robert' }, { name: 'Leanna' }]);

      obj.get('arrayObjects').pushObject({
        v: { name: 'Eddard' }
      });

      assert.deepEqual(obj.get('mappedObjects'), [{ name: 'Robert' }, { name: 'Leanna' }, { name: 'Eddard' }]);

      (0, _runtime.removeAt)(obj.get('arrayObjects'), 1);

      assert.deepEqual(obj.get('mappedObjects'), [{ name: 'Robert' }, { name: 'Eddard' }]);

      (0, _metal.set)(obj.get('arrayObjects')[0], 'v', { name: 'Stannis' });

      assert.deepEqual(obj.get('mappedObjects'), [{ name: 'Stannis' }, { name: 'Eddard' }]);
    }

    ['@test it maps unshifted objects with property observers'](assert) {
      let array = (0, _runtime.A)();
      let cObj = { v: 'c' };

      obj = _runtime.Object.extend({
        mapped: (0, _computed.map)('array.@each.v', item => (0, _metal.get)(item, 'v').toUpperCase())
      }).create({
        array
      });

      array.unshiftObject(cObj);
      array.unshiftObject({ v: 'b' });
      array.unshiftObject({ v: 'a' });

      (0, _metal.set)(cObj, 'v', 'd');

      assert.deepEqual(array.mapBy('v'), ['a', 'b', 'd'], 'precond - unmapped array is correct');
      assert.deepEqual(obj.get('mapped'), ['A', 'B', 'D'], 'properties unshifted in sequence are mapped correctly');
    }
  });

  (0, _internalTestHelpers.moduleFor)('mapBy', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        mapped: (0, _computed.mapBy)('array', 'v')
      }).create({
        array: (0, _runtime.A)([{ v: 1 }, { v: 3 }, { v: 2 }, { v: 1 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test mapBy is readOnly'](assert) {
      assert.throws(function () {
        obj.set('mapped', 1);
      }, /Cannot set read-only property "mapped" on object:/);
    }

    ['@test it maps properties'](assert) {
      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 1]);

      obj.get('array').pushObject({ v: 5 });

      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 1, 5]);

      (0, _runtime.removeAt)(obj.get('array'), 3);

      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 5]);
    }

    ['@test it is observable'](assert) {
      let calls = 0;

      assert.deepEqual(obj.get('mapped'), [1, 3, 2, 1]);

      (0, _metal.addObserver)(obj, 'mapped.@each', () => calls++);

      obj.get('array').pushObject({ v: 5 });

      assert.equal(calls, 1, 'mapBy is observable');
    }
  });

  (0, _internalTestHelpers.moduleFor)('filter', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        filtered: (0, _computed.filter)('array', item => item % 2 === 0)
      }).create({
        array: (0, _runtime.A)([1, 2, 3, 4, 5, 6, 7, 8])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test filter is readOnly'](assert) {
      assert.throws(function () {
        obj.set('filtered', 1);
      }, /Cannot set read-only property "filtered" on object:/);
    }

    ['@test it filters according to the specified filter function'](assert) {
      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'filter filters by the specified function');
    }

    ['@test it passes the index to the callback'](assert) {
      obj = _runtime.Object.extend({
        filtered: (0, _computed.filter)('array', (item, index) => index === 1)
      }).create({
        array: ['a', 'b', 'c']
      });

      assert.deepEqual((0, _metal.get)(obj, 'filtered'), ['b'], 'index is passed to callback correctly');
    }

    ['@test it has the correct `this`'](assert) {
      obj = _runtime.Object.extend({
        filtered: (0, _computed.filter)('array', function (item, index) {
          assert.equal(this, obj);
          return this.isOne(index);
        }),
        isOne(value) {
          return value === 1;
        }
      }).create({
        array: ['a', 'b', 'c']
      });

      assert.deepEqual((0, _metal.get)(obj, 'filtered'), ['b'], 'index is passed to callback correctly');
    }

    ['@test it passes the array to the callback'](assert) {
      obj = _runtime.Object.extend({
        filtered: (0, _computed.filter)('array', (item, index, array) => index === (0, _metal.get)(array, 'length') - 2)
      }).create({
        array: (0, _runtime.A)(['a', 'b', 'c'])
      });

      assert.deepEqual(obj.get('filtered'), ['b'], 'array is passed to callback correctly');
    }

    ['@test it caches properly'](assert) {
      let array = obj.get('array');

      let filtered = obj.get('filtered');
      assert.ok(filtered === obj.get('filtered'));

      array.addObject(11);
      let newFiltered = obj.get('filtered');

      assert.ok(filtered !== newFiltered);

      assert.ok(obj.get('filtered') === newFiltered);
    }

    ['@test it updates as the array is modified'](assert) {
      let array = obj.get('array');

      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'precond - filtered array is initially correct');

      array.addObject(11);
      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'objects not passing the filter are not added');

      array.addObject(12);
      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8, 12], 'objects passing the filter are added');

      array.removeObject(3);
      array.removeObject(4);

      assert.deepEqual(obj.get('filtered'), [2, 6, 8, 12], 'objects removed from the dependent array are removed from the computed array');
    }

    ['@test the dependent array can be cleared one at a time'](assert) {
      let array = (0, _metal.get)(obj, 'array');

      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'precond - filtered array is initially correct');

      // clear 1-8 but in a random order
      array.removeObject(3);
      array.removeObject(1);
      array.removeObject(2);
      array.removeObject(4);
      array.removeObject(8);
      array.removeObject(6);
      array.removeObject(5);
      array.removeObject(7);

      assert.deepEqual(obj.get('filtered'), [], 'filtered array cleared correctly');
    }

    ['@test the dependent array can be `clear`ed directly (#3272)'](assert) {
      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'precond - filtered array is initially correct');

      obj.get('array').clear();

      assert.deepEqual(obj.get('filtered'), [], 'filtered array cleared correctly');
    }

    ['@test it updates as the array is replaced'](assert) {
      assert.deepEqual(obj.get('filtered'), [2, 4, 6, 8], 'precond - filtered array is initially correct');

      obj.set('array', [20, 21, 22, 23, 24]);

      assert.deepEqual(obj.get('filtered'), [20, 22, 24], 'computed array is updated when array is changed');
    }

    ['@test it updates properly on @each with {} dependencies'](assert) {
      let item = _runtime.Object.create({ prop: true });

      obj = _runtime.Object.extend({
        filtered: (0, _computed.filter)('items.@each.{prop}', function (item) {
          return item.get('prop') === true;
        })
      }).create({
        items: (0, _runtime.A)([item])
      });

      assert.deepEqual(obj.get('filtered'), [item]);

      item.set('prop', false);

      assert.deepEqual(obj.get('filtered'), []);
    }
  });

  (0, _internalTestHelpers.moduleFor)('filterBy', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        a1s: (0, _computed.filterBy)('array', 'a', 1),
        as: (0, _computed.filterBy)('array', 'a'),
        bs: (0, _computed.filterBy)('array', 'b')
      }).create({
        array: (0, _runtime.A)([{ name: 'one', a: 1, b: false }, { name: 'two', a: 2, b: false }, { name: 'three', a: 1, b: true }, { name: 'four', b: true }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test filterBy is readOnly'](assert) {
      assert.throws(function () {
        obj.set('as', 1);
      }, /Cannot set read-only property "as" on object:/);
    }

    ['@test properties can be filtered by truthiness'](assert) {
      assert.deepEqual(obj.get('as').mapBy('name'), ['one', 'two', 'three'], 'properties can be filtered by existence');
      assert.deepEqual(obj.get('bs').mapBy('name'), ['three', 'four'], 'booleans can be filtered');

      (0, _metal.set)(obj.get('array')[0], 'a', undefined);
      (0, _metal.set)(obj.get('array')[3], 'a', true);

      (0, _metal.set)(obj.get('array')[0], 'b', true);
      (0, _metal.set)(obj.get('array')[3], 'b', false);

      assert.deepEqual(obj.get('as').mapBy('name'), ['two', 'three', 'four'], 'arrays computed by filter property respond to property changes');
      assert.deepEqual(obj.get('bs').mapBy('name'), ['one', 'three'], 'arrays computed by filtered property respond to property changes');

      obj.get('array').pushObject({ name: 'five', a: 6, b: true });

      assert.deepEqual(obj.get('as').mapBy('name'), ['two', 'three', 'four', 'five'], 'arrays computed by filter property respond to added objects');
      assert.deepEqual(obj.get('bs').mapBy('name'), ['one', 'three', 'five'], 'arrays computed by filtered property respond to added objects');

      obj.get('array').popObject();

      assert.deepEqual(obj.get('as').mapBy('name'), ['two', 'three', 'four'], 'arrays computed by filter property respond to removed objects');
      assert.deepEqual(obj.get('bs').mapBy('name'), ['one', 'three'], 'arrays computed by filtered property respond to removed objects');

      obj.set('array', [{ name: 'six', a: 12, b: true }]);

      assert.deepEqual(obj.get('as').mapBy('name'), ['six'], 'arrays computed by filter property respond to array changes');
      assert.deepEqual(obj.get('bs').mapBy('name'), ['six'], 'arrays computed by filtered property respond to array changes');
    }

    ['@test properties can be filtered by values'](assert) {
      assert.deepEqual(obj.get('a1s').mapBy('name'), ['one', 'three'], 'properties can be filtered by matching value');

      obj.get('array').pushObject({ name: 'five', a: 1 });

      assert.deepEqual(obj.get('a1s').mapBy('name'), ['one', 'three', 'five'], 'arrays computed by matching value respond to added objects');

      obj.get('array').popObject();

      assert.deepEqual(obj.get('a1s').mapBy('name'), ['one', 'three'], 'arrays computed by matching value respond to removed objects');

      (0, _metal.set)(obj.get('array')[1], 'a', 1);
      (0, _metal.set)(obj.get('array')[2], 'a', 2);

      assert.deepEqual(obj.get('a1s').mapBy('name'), ['one', 'two'], 'arrays computed by matching value respond to modified properties');
    }

    ['@test properties values can be replaced'](assert) {
      obj = _runtime.Object.extend({
        a1s: (0, _computed.filterBy)('array', 'a', 1),
        a1bs: (0, _computed.filterBy)('a1s', 'b')
      }).create({
        array: []
      });

      assert.deepEqual(obj.get('a1bs').mapBy('name'), [], 'properties can be filtered by matching value');

      (0, _metal.set)(obj, 'array', [{ name: 'item1', a: 1, b: true }]);

      assert.deepEqual(obj.get('a1bs').mapBy('name'), ['item1'], 'properties can be filtered by matching value');
    }
  });

  [['uniq', _computed.uniq], ['union', _computed.union]].forEach(tuple => {
    let [name, macro] = tuple;

    (0, _internalTestHelpers.moduleFor)(`computed.${name}`, class extends _internalTestHelpers.AbstractTestCase {
      beforeEach() {
        obj = _runtime.Object.extend({
          union: macro('array', 'array2', 'array3')
        }).create({
          array: (0, _runtime.A)([1, 2, 3, 4, 5, 6]),
          array2: (0, _runtime.A)([4, 5, 6, 7, 8, 9, 4, 5, 6, 7, 8, 9]),
          array3: (0, _runtime.A)([1, 8, 10])
        });
      }

      afterEach() {
        (0, _runloop.run)(obj, 'destroy');
      }

      [`@test ${name} is readOnly`](assert) {
        assert.throws(function () {
          obj.set('union', 1);
        }, /Cannot set read-only property "union" on object:/);
      }

      ['@test does not include duplicates'](assert) {
        let array = obj.get('array');
        let array2 = obj.get('array2');

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], name + ' does not include duplicates');

        array.pushObject(8);

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], name + ' does not add existing items');

        array.pushObject(11);

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], name + ' adds new items');

        (0, _runtime.removeAt)(array2, 6); // remove 7

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], name + ' does not remove items that are still in the dependent array');

        array2.removeObject(7);

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 8, 9, 10, 11], name + ' removes items when their last instance is gone');
      }

      ['@test has set-union semantics'](assert) {
        let array = obj.get('array');

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], name + ' is initially correct');

        array.removeObject(6);

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 'objects are not removed if they exist in other dependent arrays');

        array.clear();

        assert.deepEqual(obj.get('union').sort((x, y) => x - y), [1, 4, 5, 6, 7, 8, 9, 10], 'objects are removed when they are no longer in any dependent array');
      }
    });
  });

  (0, _internalTestHelpers.moduleFor)('computed.uniqBy', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        list: null,
        uniqueById: (0, _computed.uniqBy)('list', 'id')
      }).create({
        list: (0, _runtime.A)([{ id: 1, value: 'one' }, { id: 2, value: 'two' }, { id: 1, value: 'one' }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test uniqBy is readOnly'](assert) {
      assert.throws(function () {
        obj.set('uniqueById', 1);
      }, /Cannot set read-only property "uniqueById" on object:/);
    }
    ['@test does not include duplicates'](assert) {
      assert.deepEqual(obj.get('uniqueById'), [{ id: 1, value: 'one' }, { id: 2, value: 'two' }]);
    }

    ['@test it does not share state among instances'](assert) {
      let MyObject = _runtime.Object.extend({
        list: [],
        uniqueByName: (0, _computed.uniqBy)('list', 'name')
      });
      let a = MyObject.create({
        list: [{ name: 'bob' }, { name: 'mitch' }, { name: 'mitch' }]
      });
      let b = MyObject.create({
        list: [{ name: 'warren' }, { name: 'mitch' }]
      });

      assert.deepEqual(a.get('uniqueByName'), [{ name: 'bob' }, { name: 'mitch' }]);
      // Making sure that 'mitch' appears
      assert.deepEqual(b.get('uniqueByName'), [{ name: 'warren' }, { name: 'mitch' }]);
    }

    ['@test it handles changes to the dependent array'](assert) {
      obj.get('list').pushObject({ id: 3, value: 'three' });

      assert.deepEqual(obj.get('uniqueById'), [{ id: 1, value: 'one' }, { id: 2, value: 'two' }, { id: 3, value: 'three' }], 'The list includes three');

      obj.get('list').pushObject({ id: 3, value: 'three' });

      assert.deepEqual(obj.get('uniqueById'), [{ id: 1, value: 'one' }, { id: 2, value: 'two' }, { id: 3, value: 'three' }], 'The list does not include a duplicate three');
    }

    ['@test it returns an empty array when computed on a non-array'](assert) {
      let MyObject = _runtime.Object.extend({
        list: null,
        uniq: (0, _computed.uniqBy)('list', 'name')
      });
      let a = MyObject.create({ list: 'not an array' });

      assert.deepEqual(a.get('uniq'), []);
    }
  });

  (0, _internalTestHelpers.moduleFor)('computed.intersect', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        intersection: (0, _computed.intersect)('array', 'array2', 'array3')
      }).create({
        array: (0, _runtime.A)([1, 2, 3, 4, 5, 6]),
        array2: (0, _runtime.A)([3, 3, 3, 4, 5]),
        array3: (0, _runtime.A)([3, 5, 6, 7, 8])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test intersect is readOnly'](assert) {
      assert.throws(function () {
        obj.set('intersection', 1);
      }, /Cannot set read-only property "intersection" on object:/);
    }

    ['@test it has set-intersection semantics'](assert) {
      let array2 = obj.get('array2');
      let array3 = obj.get('array3');

      assert.deepEqual(obj.get('intersection').sort((x, y) => x - y), [3, 5], 'intersection is initially correct');

      array2.shiftObject();

      assert.deepEqual(obj.get('intersection').sort((x, y) => x - y), [3, 5], 'objects are not removed when they are still in all dependent arrays');

      array2.shiftObject();

      assert.deepEqual(obj.get('intersection').sort((x, y) => x - y), [3, 5], 'objects are not removed when they are still in all dependent arrays');

      array2.shiftObject();

      assert.deepEqual(obj.get('intersection'), [5], 'objects are removed once they are gone from all dependent arrays');

      array2.pushObject(1);

      assert.deepEqual(obj.get('intersection'), [5], 'objects are not added as long as they are missing from any dependent array');

      array3.pushObject(1);

      assert.deepEqual(obj.get('intersection').sort((x, y) => x - y), [1, 5], 'objects added once they belong to all dependent arrays');
    }
  });

  (0, _internalTestHelpers.moduleFor)('setDiff', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        diff: (0, _computed.setDiff)('array', 'array2')
      }).create({
        array: (0, _runtime.A)([1, 2, 3, 4, 5, 6, 7]),
        array2: (0, _runtime.A)([3, 4, 5, 10])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test setDiff is readOnly'](assert) {
      assert.throws(function () {
        obj.set('diff', 1);
      }, /Cannot set read-only property "diff" on object:/);
    }

    ['@test it asserts if given fewer or more than two dependent properties']() {
      expectAssertion(function () {
        _runtime.Object.extend({
          diff: (0, _computed.setDiff)('array')
        }).create({
          array: (0, _runtime.A)([1, 2, 3, 4, 5, 6, 7]),
          array2: (0, _runtime.A)([3, 4, 5])
        });
      }, /\`computed\.setDiff\` requires exactly two dependent arrays/, 'setDiff requires two dependent arrays');

      expectAssertion(function () {
        _runtime.Object.extend({
          diff: (0, _computed.setDiff)('array', 'array2', 'array3')
        }).create({
          array: (0, _runtime.A)([1, 2, 3, 4, 5, 6, 7]),
          array2: (0, _runtime.A)([3, 4, 5]),
          array3: (0, _runtime.A)([7])
        });
      }, /\`computed\.setDiff\` requires exactly two dependent arrays/, 'setDiff requires two dependent arrays');
    }

    ['@test it has set-diff semantics'](assert) {
      let array1 = obj.get('array');
      let array2 = obj.get('array2');

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 6, 7], 'set-diff is initially correct');

      array2.popObject();

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 6, 7], 'removing objects from the remove set has no effect if the object is not in the keep set');

      array2.shiftObject();

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 3, 6, 7], "removing objects from the remove set adds them if they're in the keep set");

      array1.removeObject(3);

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 6, 7], 'removing objects from the keep array removes them from the computed array');

      array1.pushObject(5);

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 6, 7], 'objects added to the keep array that are in the remove array are not added to the computed array');

      array1.pushObject(22);

      assert.deepEqual(obj.get('diff').sort((x, y) => x - y), [1, 2, 6, 7, 22], 'objects added to the keep array not in the remove array are added to the computed array');
    }
  });

  (0, _internalTestHelpers.moduleFor)('sort - sortProperties', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items', 'itemSorting')
      }).create({
        itemSorting: (0, _runtime.A)(['lname', 'fname']),
        items: (0, _runtime.A)([{ fname: 'Jaime', lname: 'Lannister', age: 34 }, { fname: 'Cersei', lname: 'Lannister', age: 34 }, { fname: 'Robb', lname: 'Stark', age: 16 }, { fname: 'Bran', lname: 'Stark', age: 8 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test sort is readOnly'](assert) {
      assert.throws(function () {
        obj.set('sortedItems', 1);
      }, /Cannot set read-only property "sortedItems" on object:/);
    }

    ['@test arrays are initially sorted'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'array is initially sorted');
    }

    ['@test default sort order is correct'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'array is initially sorted');
    }

    ['@test changing the dependent array updates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('items', [{ fname: 'Roose', lname: 'Bolton' }, { fname: 'Theon', lname: 'Greyjoy' }, { fname: 'Ramsey', lname: 'Bolton' }, { fname: 'Stannis', lname: 'Baratheon' }]);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Stannis', 'Ramsey', 'Roose', 'Theon'], 'changing dependent array updates sorted array');
    }

    ['@test adding to the dependent array updates the sorted array'](assert) {
      let items = obj.get('items');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      items.pushObject({
        fname: 'Tyrion',
        lname: 'Lannister'
      });

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb'], 'Adding to the dependent array updates the sorted array');
    }

    ['@test removing from the dependent array updates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.get('items').popObject();

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Robb'], 'Removing from the dependent array updates the sorted array');
    }

    ['@test distinct items may be sort-equal, although their relative order will not be guaranteed'](assert) {
      // We recreate jaime and "Cersei" here only for test stability: we want
      // their guid-ordering to be deterministic
      let jaimeInDisguise = {
        fname: 'Cersei',
        lname: 'Lannister',
        age: 34
      };

      let jaime = {
        fname: 'Jaime',
        lname: 'Lannister',
        age: 34
      };

      let items = obj.get('items');

      items.replace(0, 1, [jaime]);
      items.replace(1, 1, [jaimeInDisguise]);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      (0, _metal.set)(jaimeInDisguise, 'fname', 'Jaime');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Jaime', 'Bran', 'Robb'], 'sorted array is updated');

      (0, _metal.set)(jaimeInDisguise, 'fname', 'Cersei');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'sorted array is updated');
    }

    ['@test guid sort-order fallback with a search proxy is not confused by non-search ObjectProxys'](assert) {
      let tyrion = {
        fname: 'Tyrion',
        lname: 'Lannister'
      };

      let tyrionInDisguise = _runtime.ObjectProxy.create({
        fname: 'Yollo',
        lname: '',
        content: tyrion
      });

      let items = obj.get('items');

      items.pushObject(tyrion);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb']);

      items.pushObject(tyrionInDisguise);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Yollo', 'Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb']);
    }

    ['@test updating sort properties detaches observers for old sort properties'](assert) {
      let objectToRemove = obj.get('items')[3];

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('itemSorting', (0, _runtime.A)(['fname:desc']));

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Robb', 'Jaime', 'Cersei', 'Bran'], 'after updating sort properties array is updated');

      obj.get('items').removeObject(objectToRemove);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Robb', 'Jaime', 'Cersei'], 'after removing item array is updated');

      (0, _metal.set)(objectToRemove, 'lname', 'Updated-Stark');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Robb', 'Jaime', 'Cersei'], 'after changing removed item array is not updated');
    }

    ['@test sort works if array property is null (non array value) on first evaluation of computed prop'](assert) {
      obj.set('items', null);
      assert.deepEqual(obj.get('sortedItems'), []);
      obj.set('items', (0, _runtime.A)([{ fname: 'Cersei', lname: 'Lanister' }]));
      assert.deepEqual(obj.get('sortedItems'), [{ fname: 'Cersei', lname: 'Lanister' }]);
    }

    ['@test updating sort properties updates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('itemSorting', (0, _runtime.A)(['fname:desc']));

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Robb', 'Jaime', 'Cersei', 'Bran'], 'after updating sort properties array is updated');
    }

    ['@test updating sort properties invalidates the sorted array'](assert) {
      let sortProps = obj.get('itemSorting');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      sortProps.clear();
      sortProps.pushObject('fname');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Bran', 'Cersei', 'Jaime', 'Robb'], 'after updating sort properties array is updated');
    }

    ['@test updating new sort properties invalidates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('itemSorting', (0, _runtime.A)(['age:desc', 'fname:asc']));

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Robb', 'Bran'], 'precond - array is correct after item sorting is changed');

      (0, _metal.set)(obj.get('items')[1], 'age', 29);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Cersei', 'Robb', 'Bran'], 'after updating sort properties array is updated');
    }

    ['@test sort direction defaults to ascending'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb']);
    }

    ['@test sort direction defaults to ascending (with sort property change)'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('itemSorting', (0, _runtime.A)(['fname']));

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Bran', 'Cersei', 'Jaime', 'Robb'], 'sort direction defaults to ascending');
    }

    ["@test updating an item's sort properties updates the sorted array"](assert) {
      let tyrionInDisguise = obj.get('items')[1];

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      (0, _metal.set)(tyrionInDisguise, 'fname', 'Tyrion');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Tyrion', 'Bran', 'Robb'], "updating an item's sort properties updates the sorted array");
    }

    ["@test updating several of an item's sort properties updated the sorted array"](assert) {
      let sansaInDisguise = obj.get('items')[1];

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      (0, _metal.setProperties)(sansaInDisguise, {
        fname: 'Sansa',
        lname: 'Stark'
      });

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Bran', 'Robb', 'Sansa'], "updating an item's sort properties updates the sorted array");
    }

    ["@test updating an item's sort properties does not error when binary search does a self compare (#3273)"](assert) {
      let jaime = {
        name: 'Jaime',
        status: 1
      };

      let cersei = {
        name: 'Cersei',
        status: 2
      };

      let obj = _runtime.Object.extend({
        sortProps: ['status'],
        sortedPeople: (0, _computed.sort)('people', 'sortProps')
      }).create({
        people: [jaime, cersei]
      });

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei], 'precond - array is initially sorted');

      (0, _metal.set)(cersei, 'status', 3);

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei], 'array is sorted correctly');

      (0, _metal.set)(cersei, 'status', 2);

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei], 'array is sorted correctly');
    }

    ['@test array should not be sorted if sort properties array is empty'](assert) {
      var o = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items', 'itemSorting')
      }).create({
        itemSorting: (0, _runtime.A)([]),
        // This bug only manifests when array.sort(() => 0) is not equal to array.
        // In order for this to happen, the browser must use an unstable sort and the
        // array must be sufficient large. On Chrome, 12 items is currently sufficient.
        items: (0, _runtime.A)([6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5])
      });

      assert.deepEqual(o.get('sortedItems'), [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5], 'array is not changed');
    }

    ['@test array should update if items to be sorted is replaced when sort properties array is empty'](assert) {
      var o = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items', 'itemSorting')
      }).create({
        itemSorting: (0, _runtime.A)([]),
        items: (0, _runtime.A)([6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5])
      });

      assert.deepEqual(o.get('sortedItems'), [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5], 'array is not changed');

      (0, _metal.set)(o, 'items', (0, _runtime.A)([5, 6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4]));

      assert.deepEqual(o.get('sortedItems'), [5, 6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4], 'array was updated');
    }

    ['@test array should update if items to be sorted is mutated when sort properties array is empty'](assert) {
      var o = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items', 'itemSorting')
      }).create({
        itemSorting: (0, _runtime.A)([]),
        items: (0, _runtime.A)([6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5])
      });

      assert.deepEqual(o.get('sortedItems'), [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5], 'array is not changed');

      o.get('items').pushObject(12);

      assert.deepEqual(o.get('sortedItems'), [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5, 12], 'array was updated');
    }

    ['@test array observers do not leak'](assert) {
      let daria = { name: 'Daria' };
      let jane = { name: 'Jane' };

      let sisters = [jane, daria];

      let sortProps = (0, _runtime.A)(['name']);
      let jaime = _runtime.Object.extend({
        sortedPeople: (0, _computed.sort)('sisters', 'sortProps'),
        sortProps
      }).create({
        sisters
      });

      jaime.get('sortedPeople');
      (0, _runloop.run)(jaime, 'destroy');

      try {
        sortProps.pushObject({
          name: 'Anna'
        });
        assert.ok(true);
      } catch (e) {
        assert.ok(false, e);
      }
    }

    ['@test property paths in sort properties update the sorted array'](assert) {
      let jaime = {
        relatedObj: { status: 1, firstName: 'Jaime', lastName: 'Lannister' }
      };

      let cersei = {
        relatedObj: { status: 2, firstName: 'Cersei', lastName: 'Lannister' }
      };

      let sansa = _runtime.Object.create({
        relatedObj: { status: 3, firstName: 'Sansa', lastName: 'Stark' }
      });

      let obj = _runtime.Object.extend({
        sortProps: ['relatedObj.status'],
        sortedPeople: (0, _computed.sort)('people', 'sortProps')
      }).create({
        people: [jaime, cersei, sansa]
      });

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei, sansa], 'precond - array is initially sorted');

      (0, _metal.set)(cersei, 'status', 3);

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei, sansa], 'array is sorted correctly');

      (0, _metal.set)(cersei, 'status', 1);

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei, sansa], 'array is sorted correctly');

      sansa.set('status', 1);

      assert.deepEqual(obj.get('sortedPeople'), [jaime, cersei, sansa], 'array is sorted correctly');

      obj.set('sortProps', ['relatedObj.firstName']);

      assert.deepEqual(obj.get('sortedPeople'), [cersei, jaime, sansa], 'array is sorted correctly');
    }

    ['@test if the dependentKey is neither an array nor object, it will return an empty array'](assert) {
      (0, _metal.set)(obj, 'items', null);
      assert.ok((0, _runtime.isArray)(obj.get('sortedItems')), 'returns an empty arrays');

      (0, _metal.set)(obj, 'array', undefined);
      assert.ok((0, _runtime.isArray)(obj.get('sortedItems')), 'returns an empty arrays');

      (0, _metal.set)(obj, 'array', 'not an array');
      assert.ok((0, _runtime.isArray)(obj.get('sortedItems')), 'returns an empty arrays');
    }
  });

  function sortByLnameFname(a, b) {
    let lna = (0, _metal.get)(a, 'lname');
    let lnb = (0, _metal.get)(b, 'lname');

    if (lna !== lnb) {
      return lna > lnb ? 1 : -1;
    }

    return sortByFnameAsc(a, b);
  }

  function sortByFnameAsc(a, b) {
    let fna = (0, _metal.get)(a, 'fname');
    let fnb = (0, _metal.get)(b, 'fname');

    if (fna === fnb) {
      return 0;
    }
    return fna > fnb ? 1 : -1;
  }

  (0, _internalTestHelpers.moduleFor)('sort - sort function', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items.@each.fname', sortByLnameFname)
      }).create({
        items: (0, _runtime.A)([{ fname: 'Jaime', lname: 'Lannister', age: 34 }, { fname: 'Cersei', lname: 'Lannister', age: 34 }, { fname: 'Robb', lname: 'Stark', age: 16 }, { fname: 'Bran', lname: 'Stark', age: 8 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test sort has correct `this`'](assert) {
      let obj = _runtime.Object.extend({
        sortedItems: (0, _computed.sort)('items.@each.fname', function (a, b) {
          assert.equal(this, obj, 'expected the object to be `this`');
          return this.sortByLastName(a, b);
        }),
        sortByLastName(a, b) {
          return sortByFnameAsc(a, b);
        }
      }).create({
        items: (0, _runtime.A)([{ fname: 'Jaime', lname: 'Lannister', age: 34 }, { fname: 'Cersei', lname: 'Lannister', age: 34 }, { fname: 'Robb', lname: 'Stark', age: 16 }, { fname: 'Bran', lname: 'Stark', age: 8 }])
      });

      obj.get('sortedItems');
    }

    ['@test sort (with function) is readOnly'](assert) {
      assert.throws(function () {
        obj.set('sortedItems', 1);
      }, /Cannot set read-only property "sortedItems" on object:/);
    }

    ['@test arrays are initially sorted'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'array is initially sorted');
    }

    ['@test default sort order is correct'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'array is initially sorted');
    }

    ['@test changing the dependent array updates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.set('items', [{ fname: 'Roose', lname: 'Bolton' }, { fname: 'Theon', lname: 'Greyjoy' }, { fname: 'Ramsey', lname: 'Bolton' }, { fname: 'Stannis', lname: 'Baratheon' }]);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Stannis', 'Ramsey', 'Roose', 'Theon'], 'changing dependent array updates sorted array');
    }

    ['@test adding to the dependent array updates the sorted array'](assert) {
      let items = obj.get('items');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      items.pushObject({
        fname: 'Tyrion',
        lname: 'Lannister'
      });

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb'], 'Adding to the dependent array updates the sorted array');
    }

    ['@test removing from the dependent array updates the sorted array'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      obj.get('items').popObject();

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Robb'], 'Removing from the dependent array updates the sorted array');
    }

    ['@test distinct items may be sort-equal, although their relative order will not be guaranteed'](assert) {
      // We recreate jaime and "Cersei" here only for test stability: we want
      // their guid-ordering to be deterministic
      let jaimeInDisguise = {
        fname: 'Cersei',
        lname: 'Lannister',
        age: 34
      };

      let jaime = {
        fname: 'Jaime',
        lname: 'Lannister',
        age: 34
      };

      let items = obj.get('items');

      items.replace(0, 1, [jaime]);
      items.replace(1, 1, [jaimeInDisguise]);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      (0, _metal.set)(jaimeInDisguise, 'fname', 'Jaime');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Jaime', 'Bran', 'Robb'], 'sorted array is updated');

      (0, _metal.set)(jaimeInDisguise, 'fname', 'Cersei');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'sorted array is updated');
    }

    ['@test guid sort-order fallback with a search proxy is not confused by non-search ObjectProxys'](assert) {
      let tyrion = {
        fname: 'Tyrion',
        lname: 'Lannister'
      };

      let tyrionInDisguise = _runtime.ObjectProxy.create({
        fname: 'Yollo',
        lname: '',
        content: tyrion
      });

      let items = obj.get('items');

      items.pushObject(tyrion);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb']);

      items.pushObject(tyrionInDisguise);

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Yollo', 'Cersei', 'Jaime', 'Tyrion', 'Bran', 'Robb']);
    }

    ['@test changing item properties specified via @each triggers a resort of the modified item'](assert) {
      let items = (0, _metal.get)(obj, 'items');

      let tyrionInDisguise = items[1];

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

      (0, _metal.set)(tyrionInDisguise, 'fname', 'Tyrion');

      assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Jaime', 'Tyrion', 'Bran', 'Robb'], 'updating a specified property on an item resorts it');
    }

    ['@test changing item properties not specified via @each does not trigger a resort'](assert) {
      if (!_canaryFeatures.EMBER_METAL_TRACKED_PROPERTIES) {
        let items = obj.get('items');
        let cersei = items[1];

        assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'precond - array is initially sorted');

        (0, _metal.set)(cersei, 'lname', 'Stark'); // plot twist! (possibly not canon)

        // The array has become unsorted.  If your sort function is sensitive to
        // properties, they *must* be specified as dependent item property keys or
        // we'll be doing binary searches on unsorted arrays.
        assert.deepEqual(obj.get('sortedItems').mapBy('fname'), ['Cersei', 'Jaime', 'Bran', 'Robb'], 'updating an unspecified property on an item does not resort it');
      } else {
        assert.expect(0);
      }
    }
  });

  (0, _internalTestHelpers.moduleFor)('sort - stability', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        sortProps: ['count', 'name'],
        sortedItems: (0, _computed.sort)('items', 'sortProps')
      }).create({
        items: [{ name: 'A', count: 1, thing: 4 }, { name: 'B', count: 1, thing: 3 }, { name: 'C', count: 1, thing: 2 }, { name: 'D', count: 1, thing: 4 }]
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test sorts correctly as only one property changes'](assert) {
      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'initial');

      (0, _metal.set)(obj.get('items')[3], 'count', 2);

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'final');
    }
  });

  let klass;
  (0, _internalTestHelpers.moduleFor)('sort - concurrency', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      klass = _runtime.Object.extend({
        sortProps: ['count'],
        sortedItems: (0, _computed.sort)('items', 'sortProps'),
        customSortedItems: (0, _computed.sort)('items.@each.count', (a, b) => a.count - b.count)
      });
      obj = klass.create({
        items: (0, _runtime.A)([{ name: 'A', count: 1, thing: 4, id: 1 }, { name: 'B', count: 2, thing: 3, id: 2 }, { name: 'C', count: 3, thing: 2, id: 3 }, { name: 'D', count: 4, thing: 1, id: 4 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test sorts correctly after mutation to the sort properties'](assert) {
      let sorted = obj.get('sortedItems');
      assert.deepEqual(sorted.mapBy('name'), ['A', 'B', 'C', 'D'], 'initial');

      (0, _metal.set)(obj.get('items')[1], 'count', 5);
      (0, _metal.set)(obj.get('items')[2], 'count', 6);

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'final');
    }

    ['@test sort correctly after mutation to the sort'](assert) {
      assert.deepEqual(obj.get('customSortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'initial');

      (0, _metal.set)(obj.get('items')[1], 'count', 5);
      (0, _metal.set)(obj.get('items')[2], 'count', 6);

      assert.deepEqual(obj.get('customSortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'final');

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'final');
    }

    ['@test sort correctly on multiple instances of the same class'](assert) {
      let obj2 = klass.create({
        items: (0, _runtime.A)([{ name: 'W', count: 23, thing: 4 }, { name: 'X', count: 24, thing: 3 }, { name: 'Y', count: 25, thing: 2 }, { name: 'Z', count: 26, thing: 1 }])
      });

      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['W', 'X', 'Y', 'Z'], 'initial');
      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'initial');

      (0, _metal.set)(obj.get('items')[1], 'count', 5);
      (0, _metal.set)(obj.get('items')[2], 'count', 6);
      (0, _metal.set)(obj2.get('items')[1], 'count', 27);
      (0, _metal.set)(obj2.get('items')[2], 'count', 28);

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'final');
      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['W', 'Z', 'X', 'Y'], 'final');

      obj.set('sortProps', ['thing']);

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['D', 'C', 'B', 'A'], 'final');

      obj2.notifyPropertyChange('sortedItems'); // invalidate to flush, to get DK refreshed
      obj2.get('sortedItems'); // flush to get updated DK

      obj2.set('items.firstObject.count', 9999);

      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['Z', 'X', 'Y', 'W'], 'final');
    }

    ['@test sort correctly when multiple sorts are chained on the same instance of a class'](assert) {
      let obj2 = klass.extend({
        items: (0, _metal.computed)('sibling.sortedItems.[]', function () {
          return this.get('sibling.sortedItems');
        }),
        asdf: (0, _metal.observer)('sibling.sortedItems.[]', function () {
          this.get('sibling.sortedItems');
        })
      }).create({
        sibling: obj
      });

      /*
                                             ┌───────────┐                              ┌────────────┐
                                             │sortedProps│                              │sortedProps2│
                                             └───────────┘                              └────────────┘
                                                   ▲                                           ▲
                                                   │               ╔═══════════╗               │
                                                   │─ ─ ─ ─ ─ ─ ─ ▶║ CP (sort) ║◀─ ─ ─ ─ ─ ─ ─ ┤
                                                   │               ╚═══════════╝               │
                                                   │                                           │
      ┌───────────┐                            ┏━━━━━━━━━━━┓                              ┏━━━━━━━━━━━━┓
      │           │   ┌ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─    ┃           ┃    ┌ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─     ┃            ┃
      │   items   │◀──  items.@each.count  │◀──┃sortedItems┃◀───  items.@each.count  │◀───┃sortedItems2┃
      │           │   └ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─    ┃           ┃    └ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─     ┃            ┃
      └───────────┘                            ┗━━━━━━━━━━━┛                              ┗━━━━━━━━━━━━┛
      */

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'obj.sortedItems.name should be sorted alpha');
      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'obj2.sortedItems.name should be sorted alpha');

      (0, _metal.set)(obj.get('items')[1], 'count', 5);
      (0, _metal.set)(obj.get('items')[2], 'count', 6);

      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'obj.sortedItems.name should now have changed');
      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['A', 'D', 'B', 'C'], 'obj2.sortedItems.name should still mirror sortedItems2');

      obj.set('sortProps', ['thing']);
      obj2.set('sortProps', ['id']);

      assert.deepEqual(obj2.get('sortedItems').mapBy('name'), ['A', 'B', 'C', 'D'], 'we now sort obj2 by id, so we expect a b c d');
      assert.deepEqual(obj.get('sortedItems').mapBy('name'), ['D', 'C', 'B', 'A'], 'we now sort obj by thing');
    }
  });

  (0, _internalTestHelpers.moduleFor)('max', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        max: (0, _computed.max)('items')
      }).create({
        items: (0, _runtime.A)([1, 2, 3])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test max is readOnly'](assert) {
      assert.throws(function () {
        obj.set('max', 1);
      }, /Cannot set read-only property "max" on object:/);
    }

    ['@test max tracks the max number as objects are added'](assert) {
      assert.equal(obj.get('max'), 3, 'precond - max is initially correct');

      let items = obj.get('items');

      items.pushObject(5);

      assert.equal(obj.get('max'), 5, 'max updates when a larger number is added');

      items.pushObject(2);

      assert.equal(obj.get('max'), 5, 'max does not update when a smaller number is added');
    }

    ['@test max recomputes when the current max is removed'](assert) {
      assert.equal(obj.get('max'), 3, 'precond - max is initially correct');

      obj.get('items').removeObject(2);

      assert.equal(obj.get('max'), 3, 'max is unchanged when a non-max item is removed');

      obj.get('items').removeObject(3);

      assert.equal(obj.get('max'), 1, 'max is recomputed when the current max is removed');
    }
  });

  (0, _internalTestHelpers.moduleFor)('min', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        min: (0, _computed.min)('items')
      }).create({
        items: (0, _runtime.A)([1, 2, 3])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test min is readOnly'](assert) {
      assert.throws(function () {
        obj.set('min', 1);
      }, /Cannot set read-only property "min" on object:/);
    }

    ['@test min tracks the min number as objects are added'](assert) {
      assert.equal(obj.get('min'), 1, 'precond - min is initially correct');

      obj.get('items').pushObject(-2);

      assert.equal(obj.get('min'), -2, 'min updates when a smaller number is added');

      obj.get('items').pushObject(2);

      assert.equal(obj.get('min'), -2, 'min does not update when a larger number is added');
    }

    ['@test min recomputes when the current min is removed'](assert) {
      let items = obj.get('items');

      assert.equal(obj.get('min'), 1, 'precond - min is initially correct');

      items.removeObject(2);

      assert.equal(obj.get('min'), 1, 'min is unchanged when a non-min item is removed');

      items.removeObject(1);

      assert.equal(obj.get('min'), 3, 'min is recomputed when the current min is removed');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember.arrayComputed - mixed sugar', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        lannisters: (0, _computed.filterBy)('items', 'lname', 'Lannister'),
        lannisterSorting: (0, _runtime.A)(['fname']),
        sortedLannisters: (0, _computed.sort)('lannisters', 'lannisterSorting'),

        starks: (0, _computed.filterBy)('items', 'lname', 'Stark'),
        starkAges: (0, _computed.mapBy)('starks', 'age'),
        oldestStarkAge: (0, _computed.max)('starkAges')
      }).create({
        items: (0, _runtime.A)([{ fname: 'Jaime', lname: 'Lannister', age: 34 }, { fname: 'Cersei', lname: 'Lannister', age: 34 }, { fname: 'Robb', lname: 'Stark', age: 16 }, { fname: 'Bran', lname: 'Stark', age: 8 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test filtering and sorting can be combined'](assert) {
      let items = obj.get('items');

      assert.deepEqual(obj.get('sortedLannisters').mapBy('fname'), ['Cersei', 'Jaime'], 'precond - array is initially filtered and sorted');

      items.pushObject({ fname: 'Tywin', lname: 'Lannister' });
      items.pushObject({ fname: 'Lyanna', lname: 'Stark' });
      items.pushObject({ fname: 'Gerion', lname: 'Lannister' });

      assert.deepEqual(obj.get('sortedLannisters').mapBy('fname'), ['Cersei', 'Gerion', 'Jaime', 'Tywin'], 'updates propagate to array');
    }

    ['@test filtering, sorting and reduce (max) can be combined'](assert) {
      let items = obj.get('items');

      assert.equal(16, obj.get('oldestStarkAge'), 'precond - end of chain is initially correct');

      items.pushObject({ fname: 'Rickon', lname: 'Stark', age: 5 });

      assert.equal(16, obj.get('oldestStarkAge'), 'chain is updated correctly');

      items.pushObject({ fname: 'Eddard', lname: 'Stark', age: 35 });

      assert.equal(35, obj.get('oldestStarkAge'), 'chain is updated correctly');
    }
  });

  function todo(name, priority) {
    return _runtime.Object.create({ name: name, priority: priority });
  }

  function priorityComparator(todoA, todoB) {
    let pa = parseInt((0, _metal.get)(todoA, 'priority'), 10);
    let pb = parseInt((0, _metal.get)(todoB, 'priority'), 10);

    return pa - pb;
  }

  function evenPriorities(todo) {
    let p = parseInt((0, _metal.get)(todo, 'priority'), 10);

    return p % 2 === 0;
  }

  (0, _internalTestHelpers.moduleFor)('Ember.arrayComputed - chains', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        sorted: (0, _computed.sort)('todos.@each.priority', priorityComparator),
        filtered: (0, _computed.filter)('sorted.@each.priority', evenPriorities)
      }).create({
        todos: (0, _runtime.A)([todo('E', 4), todo('D', 3), todo('C', 2), todo('B', 1), todo('A', 0)])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test it can filter and sort when both depend on the same item property'](assert) {
      assert.deepEqual(obj.get('todos').mapBy('name'), ['E', 'D', 'C', 'B', 'A'], 'precond - todos initially correct');
      assert.deepEqual(obj.get('sorted').mapBy('name'), ['A', 'B', 'C', 'D', 'E'], 'precond - sorted initially correct');
      assert.deepEqual(obj.get('filtered').mapBy('name'), ['A', 'C', 'E'], 'precond - filtered initially correct');

      (0, _metal.set)(obj.get('todos')[1], 'priority', 6);

      assert.deepEqual(obj.get('todos').mapBy('name'), ['E', 'D', 'C', 'B', 'A'], 'precond - todos remain correct');
      assert.deepEqual(obj.get('sorted').mapBy('name'), ['A', 'B', 'C', 'E', 'D'], 'precond - sorted updated correctly');
      assert.deepEqual(obj.get('filtered').mapBy('name'), ['A', 'C', 'E', 'D'], 'filtered updated correctly');
    }
  });

  let userFnCalls;
  (0, _internalTestHelpers.moduleFor)('Chaining array and reduced CPs', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      userFnCalls = 0;
      obj = _runtime.Object.extend({
        mapped: (0, _computed.mapBy)('array', 'v'),
        max: (0, _computed.max)('mapped'),
        maxDidChange: (0, _metal.observer)('max', () => userFnCalls++)
      }).create({
        array: (0, _runtime.A)([{ v: 1 }, { v: 3 }, { v: 2 }, { v: 1 }])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test it computes interdependent array computed properties'](assert) {
      assert.equal(obj.get('max'), 3, 'sanity - it properly computes the maximum value');

      let calls = 0;

      (0, _metal.addObserver)(obj, 'max', () => calls++);

      obj.get('array').pushObject({ v: 5 });

      assert.equal(obj.get('max'), 5, 'maximum value is updated correctly');
      assert.equal(userFnCalls, 1, 'object defined observers fire');
      assert.equal(calls, 1, 'runtime created observers fire');
    }
  });

  (0, _internalTestHelpers.moduleFor)('sum', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      obj = _runtime.Object.extend({
        total: (0, _computed.sum)('array')
      }).create({
        array: (0, _runtime.A)([1, 2, 3])
      });
    }

    afterEach() {
      (0, _runloop.run)(obj, 'destroy');
    }

    ['@test sum is readOnly'](assert) {
      assert.throws(function () {
        obj.set('total', 1);
      }, /Cannot set read-only property "total" on object:/);
    }

    ['@test sums the values in the dependentKey'](assert) {
      assert.equal(obj.get('total'), 6, 'sums the values');
    }

    ['@test if the dependentKey is neither an array nor object, it will return `0`'](assert) {
      (0, _metal.set)(obj, 'array', null);
      assert.equal((0, _metal.get)(obj, 'total'), 0, 'returns 0');

      (0, _metal.set)(obj, 'array', undefined);
      assert.equal((0, _metal.get)(obj, 'total'), 0, 'returns 0');

      (0, _metal.set)(obj, 'array', 'not an array');
      assert.equal((0, _metal.get)(obj, 'total'), 0, 'returns 0');
    }

    ['@test updates when array is modified'](assert) {
      obj.get('array').pushObject(1);

      assert.equal(obj.get('total'), 7, 'recomputed when elements are added');

      obj.get('array').popObject();

      assert.equal(obj.get('total'), 6, 'recomputes when elements are removed');
    }
  });

  (0, _internalTestHelpers.moduleFor)('collect', class extends _internalTestHelpers.AbstractTestCase {
    ['@test works'](assert) {
      let obj = { one: 'foo', two: 'bar', three: null };
      (0, _metal.defineProperty)(obj, 'all', (0, _computed.collect)('one', 'two', 'three', 'four'));

      assert.deepEqual((0, _metal.get)(obj, 'all'), ['foo', 'bar', null, null], 'have all of them');

      (0, _metal.set)(obj, 'four', true);

      assert.deepEqual((0, _metal.get)(obj, 'all'), ['foo', 'bar', null, true], 'have all of them');

      let a = [];
      (0, _metal.set)(obj, 'one', 0);
      (0, _metal.set)(obj, 'three', a);

      assert.deepEqual((0, _metal.get)(obj, 'all'), [0, 'bar', a, true], 'have all of them');
    }
  });
});
enifed('@ember/polyfills/tests/assign_test', ['@ember/polyfills', 'internal-test-helpers'], function (_polyfills, _internalTestHelpers) {
  'use strict';

  class AssignTests extends _internalTestHelpers.AbstractTestCase {
    ['@test merging objects'](assert) {
      let trgt = { a: 1 };
      let src1 = { b: 2 };
      let src2 = { c: 3 };
      this.assign(trgt, src1, src2);

      assert.deepEqual(trgt, { a: 1, b: 2, c: 3 }, 'assign copies values from one or more source objects to a target object');
      assert.deepEqual(src1, { b: 2 }, 'assign does not change source object 1');
      assert.deepEqual(src2, { c: 3 }, 'assign does not change source object 2');
    }

    ['@test merging objects with same property'](assert) {
      let trgt = { a: 1, b: 1 };
      let src1 = { a: 2, b: 2 };
      let src2 = { a: 3 };
      this.assign(trgt, src1, src2);

      assert.deepEqual(trgt, { a: 3, b: 2 }, 'properties are overwritten by other objects that have the same properties later in the parameters order');
    }

    ['@test null'](assert) {
      let trgt = { a: 1 };
      this.assign(trgt, null);

      assert.deepEqual(trgt, { a: 1 }, 'null as a source parameter is ignored');
    }

    ['@test undefined'](assert) {
      let trgt = { a: 1 };
      this.assign(trgt, null);

      assert.deepEqual(trgt, { a: 1 }, 'undefined as a source parameter is ignored');
    }
  }

  (0, _internalTestHelpers.moduleFor)('Ember.assign (polyfill)', class extends AssignTests {
    assign() {
      return (0, _polyfills.assignPolyfill)(...arguments);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Ember.assign (maybe not-polyfill ;) )', class extends AssignTests {
    assign() {
      return (0, _polyfills.assign)(...arguments);
    }
  });
});
enifed('@ember/polyfills/tests/merge_test', ['@ember/polyfills', 'internal-test-helpers'], function (_polyfills, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Ember.merge', class extends _internalTestHelpers.AbstractTestCase {
    ['@test merging objects'](assert) {
      let src1 = { a: 1 };
      let src2 = { b: 2 };
      expectDeprecation(() => {
        (0, _polyfills.merge)(src1, src2);
      }, 'Use of `merge` has been deprecated. Please use `assign` instead.');

      assert.deepEqual(src1, { a: 1, b: 2 }, 'merge copies values from second source object to first object');
    }
  });
});
enifed('@ember/runloop/tests/debounce_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('debounce', class extends _internalTestHelpers.AbstractTestCase {
    ['@test debounce - with target, with method, without args'](assert) {
      let done = assert.async();

      let calledWith = [];
      let target = {
        someFunc(...args) {
          calledWith.push(args);
        }
      };

      (0, _runloop.debounce)(target, target.someFunc, 10);
      (0, _runloop.debounce)(target, target.someFunc, 10);
      (0, _runloop.debounce)(target, target.someFunc, 10);

      setTimeout(() => {
        assert.deepEqual(calledWith, [[]], 'someFunc called once with correct arguments');
        done();
      }, 20);
    }

    ['@test debounce - with target, with method name, without args'](assert) {
      let done = assert.async();

      let calledWith = [];
      let target = {
        someFunc(...args) {
          calledWith.push(args);
        }
      };

      (0, _runloop.debounce)(target, 'someFunc', 10);
      (0, _runloop.debounce)(target, 'someFunc', 10);
      (0, _runloop.debounce)(target, 'someFunc', 10);

      setTimeout(() => {
        assert.deepEqual(calledWith, [[]], 'someFunc called once with correct arguments');
        done();
      }, 20);
    }

    ['@test debounce - without target, without args'](assert) {
      let done = assert.async();

      let calledWith = [];
      function someFunc(...args) {
        calledWith.push(args);
      }

      (0, _runloop.debounce)(someFunc, 10);
      (0, _runloop.debounce)(someFunc, 10);
      (0, _runloop.debounce)(someFunc, 10);

      setTimeout(() => {
        assert.deepEqual(calledWith, [[]], 'someFunc called once with correct arguments');
        done();
      }, 20);
    }

    ['@test debounce - without target, with args'](assert) {
      let done = assert.async();

      let calledWith = [];
      function someFunc(...args) {
        calledWith.push(args);
      }

      (0, _runloop.debounce)(someFunc, { isFoo: true }, 10);
      (0, _runloop.debounce)(someFunc, { isBar: true }, 10);
      (0, _runloop.debounce)(someFunc, { isBaz: true }, 10);

      setTimeout(() => {
        assert.deepEqual(calledWith, [[{ isBaz: true }]], 'someFunc called once with correct arguments');
        done();
      }, 20);
    }
  });
});
enifed('@ember/runloop/tests/later_test', ['internal-test-helpers', '@ember/polyfills', '@ember/runloop'], function (_internalTestHelpers, _polyfills, _runloop) {
  'use strict';

  const originalSetTimeout = window.setTimeout;
  const originalDateValueOf = Date.prototype.valueOf;
  const originalPlatform = _runloop.backburner._platform;

  function wait(callback, maxWaitCount = 100) {
    originalSetTimeout(() => {
      if (maxWaitCount > 0 && ((0, _runloop.hasScheduledTimers)() || (0, _runloop.getCurrentRunLoop)())) {
        wait(callback, maxWaitCount - 1);

        return;
      }

      callback();
    }, 10);
  }

  // Synchronous "sleep". This simulates work being done
  // after later was called but before the run loop
  // has flushed. In previous versions, this would have
  // caused the later callback to have run from
  // within the run loop flush, since by the time the
  // run loop has to flush, it would have considered
  // the timer already expired.
  function pauseUntil(time) {
    while (+new Date() < time) {
      /* do nothing - sleeping */
    }
  }

  (0, _internalTestHelpers.moduleFor)('run.later', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      _runloop.backburner._platform = originalPlatform;
      window.setTimeout = originalSetTimeout;
      Date.prototype.valueOf = originalDateValueOf;
    }

    ['@test should invoke after specified period of time - function only'](assert) {
      let done = assert.async();
      let invoked = false;

      (0, _runloop.run)(() => {
        (0, _runloop.later)(() => invoked = true, 100);
      });

      wait(() => {
        assert.equal(invoked, true, 'should have invoked later item');
        done();
      });
    }

    ['@test should invoke after specified period of time - target/method'](assert) {
      let done = assert.async();
      let obj = { invoked: false };

      (0, _runloop.run)(() => {
        (0, _runloop.later)(obj, function () {
          this.invoked = true;
        }, 100);
      });

      wait(() => {
        assert.equal(obj.invoked, true, 'should have invoked later item');
        done();
      });
    }

    ['@test should invoke after specified period of time - target/method/args'](assert) {
      let done = assert.async();
      let obj = { invoked: 0 };

      (0, _runloop.run)(() => {
        (0, _runloop.later)(obj, function (amt) {
          this.invoked += amt;
        }, 10, 100);
      });

      wait(() => {
        assert.equal(obj.invoked, 10, 'should have invoked later item');
        done();
      });
    }

    ['@test should always invoke within a separate runloop'](assert) {
      let done = assert.async();
      let obj = { invoked: 0 };
      let firstRunLoop, secondRunLoop;

      (0, _runloop.run)(() => {
        firstRunLoop = (0, _runloop.getCurrentRunLoop)();

        (0, _runloop.later)(obj, function (amt) {
          this.invoked += amt;
          secondRunLoop = (0, _runloop.getCurrentRunLoop)();
        }, 10, 1);

        pauseUntil(+new Date() + 100);
      });

      assert.ok(firstRunLoop, 'first run loop captured');
      assert.ok(!(0, _runloop.getCurrentRunLoop)(), "shouldn't be in a run loop after flush");
      assert.equal(obj.invoked, 0, "shouldn't have invoked later item yet");

      wait(() => {
        assert.equal(obj.invoked, 10, 'should have invoked later item');
        assert.ok(secondRunLoop, 'second run loop took place');
        assert.ok(secondRunLoop !== firstRunLoop, 'two different run loops took place');
        done();
      });
    }

    // Our current implementation doesn't allow us to correctly enforce this ordering.
    // We should probably implement a queue to provide this guarantee.
    // See https://github.com/emberjs/ember.js/issues/3526 for more information.

    // asyncTest('callback order', function() {
    //   let array = [];
    //   function fn(val) { array.push(val); }

    //   run(function() {
    //     later(this, fn, 4, 5);
    //     later(this, fn, 1, 1);
    //     later(this, fn, 5, 10);
    //     later(this, fn, 2, 3);
    //     later(this, fn, 3, 3);
    //   });

    //   deepEqual(array, []);

    //   wait(function() {
    //     QUnit.start();
    //     deepEqual(array, [1,2,3,4,5], 'callbacks were called in expected order');
    //   });
    // });

    // Out current implementation doesn't allow us to properly enforce what is tested here.
    // We should probably fix it, but it's not technically a bug right now.
    // See https://github.com/emberjs/ember.js/issues/3522 for more information.

    // asyncTest('callbacks coalesce into same run loop if expiring at the same time', function() {
    //   let array = [];
    //   function fn(val) { array.push(getCurrentRunLoop()); }

    //   run(function() {

    //     // Force +new Date to return the same result while scheduling
    //     // later timers. Otherwise: non-determinism!
    //     let now = +new Date();
    //     Date.prototype.valueOf = function() { return now; };

    //     later(this, fn, 10);
    //     later(this, fn, 200);
    //     later(this, fn, 200);

    //     Date.prototype.valueOf = originalDateValueOf;
    //   });

    //   deepEqual(array, []);

    //   wait(function() {
    //     QUnit.start();
    //     equal(array.length, 3, 'all callbacks called');
    //     ok(array[0] !== array[1], 'first two callbacks have different run loops');
    //     ok(array[0], 'first runloop present');
    //     ok(array[1], 'second runloop present');
    //     equal(array[1], array[2], 'last two callbacks got the same run loop');
    //   });
    // });

    ['@test inception calls to later should run callbacks in separate run loops'](assert) {
      let done = assert.async();
      let runLoop, finished;

      (0, _runloop.run)(() => {
        runLoop = (0, _runloop.getCurrentRunLoop)();
        assert.ok(runLoop);

        (0, _runloop.later)(() => {
          assert.ok((0, _runloop.getCurrentRunLoop)() && (0, _runloop.getCurrentRunLoop)() !== runLoop, 'first later callback has own run loop');
          runLoop = (0, _runloop.getCurrentRunLoop)();

          (0, _runloop.later)(() => {
            assert.ok((0, _runloop.getCurrentRunLoop)() && (0, _runloop.getCurrentRunLoop)() !== runLoop, 'second later callback has own run loop');
            finished = true;
          }, 40);
        }, 40);
      });

      wait(() => {
        assert.ok(finished, 'all .later callbacks run');
        done();
      });
    }

    ['@test setTimeout should never run with a negative wait'](assert) {
      let done = assert.async();
      // Rationale: The old run loop code was susceptible to an occasional
      // bug where invokeLaterTimers would be scheduled with a setTimeout
      // with a negative wait. Modern browsers normalize this to 0, but
      // older browsers (IE <= 8) break with a negative wait, which
      // happens when an expired timer callback takes a while to run,
      // which is what we simulate here.
      let newSetTimeoutUsed;
      _runloop.backburner._platform = (0, _polyfills.assign)({}, originalPlatform, {
        setTimeout() {
          let wait = arguments[arguments.length - 1];
          newSetTimeoutUsed = true;
          assert.ok(!isNaN(wait) && wait >= 0, 'wait is a non-negative number');

          return originalPlatform.setTimeout.apply(originalPlatform, arguments);
        }
      });

      let count = 0;
      (0, _runloop.run)(() => {
        (0, _runloop.later)(() => {
          count++;

          // This will get run first. Waste some time.
          // This is intended to break invokeLaterTimers code by taking a
          // long enough time that other timers should technically expire. It's
          // fine that they're not called in this run loop; just need to
          // make sure that invokeLaterTimers doesn't end up scheduling
          // a negative setTimeout.
          pauseUntil(+new Date() + 60);
        }, 1);

        (0, _runloop.later)(() => {
          assert.equal(count, 1, 'callbacks called in order');
        }, 50);
      });

      wait(() => {
        assert.ok(newSetTimeoutUsed, 'stub setTimeout was used');
        done();
      });
    }
  });
});
enifed('@ember/runloop/tests/next_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('run.next', class extends _internalTestHelpers.AbstractTestCase {
    ['@test should invoke immediately on next timeout'](assert) {
      let done = assert.async();
      let invoked = false;

      (0, _runloop.run)(() => (0, _runloop.next)(() => invoked = true));

      assert.equal(invoked, false, 'should not have invoked yet');

      setTimeout(() => {
        assert.equal(invoked, true, 'should have invoked later item');
        done();
      }, 20);
    }

    ['@test callback should be called from within separate loop'](assert) {
      let done = assert.async();
      let firstRunLoop, secondRunLoop;
      (0, _runloop.run)(() => {
        firstRunLoop = (0, _runloop.getCurrentRunLoop)();
        (0, _runloop.next)(() => secondRunLoop = (0, _runloop.getCurrentRunLoop)());
      });

      setTimeout(() => {
        assert.ok(secondRunLoop, 'callback was called from within run loop');
        assert.ok(firstRunLoop && secondRunLoop !== firstRunLoop, 'two separate run loops were invoked');
        done();
      }, 20);
    }

    ['@test multiple calls to next share coalesce callbacks into same run loop'](assert) {
      let done = assert.async();
      let secondRunLoop, thirdRunLoop;
      (0, _runloop.run)(() => {
        (0, _runloop.next)(() => secondRunLoop = (0, _runloop.getCurrentRunLoop)());
        (0, _runloop.next)(() => thirdRunLoop = (0, _runloop.getCurrentRunLoop)());
      });

      setTimeout(() => {
        assert.ok(secondRunLoop && secondRunLoop === thirdRunLoop, 'callbacks coalesced into same run loop');
        done();
      }, 20);
    }
  });
});
enifed('@ember/runloop/tests/once_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/once_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test calling invokeOnce more than once invokes only once'](assert) {
      let count = 0;
      (0, _runloop.run)(() => {
        function F() {
          count++;
        }
        (0, _runloop.once)(F);
        (0, _runloop.once)(F);
        (0, _runloop.once)(F);
      });

      assert.equal(count, 1, 'should have invoked once');
    }

    ['@test should differentiate based on target'](assert) {
      let A = { count: 0 };
      let B = { count: 0 };
      (0, _runloop.run)(() => {
        function F() {
          this.count++;
        }
        (0, _runloop.once)(A, F);
        (0, _runloop.once)(B, F);
        (0, _runloop.once)(A, F);
        (0, _runloop.once)(B, F);
      });

      assert.equal(A.count, 1, 'should have invoked once on A');
      assert.equal(B.count, 1, 'should have invoked once on B');
    }

    ['@test should ignore other arguments - replacing previous ones'](assert) {
      let A = { count: 0 };
      let B = { count: 0 };

      (0, _runloop.run)(() => {
        function F(amt) {
          this.count += amt;
        }
        (0, _runloop.once)(A, F, 10);
        (0, _runloop.once)(B, F, 20);
        (0, _runloop.once)(A, F, 30);
        (0, _runloop.once)(B, F, 40);
      });

      assert.equal(A.count, 30, 'should have invoked once on A');
      assert.equal(B.count, 40, 'should have invoked once on B');
    }

    ['@test should be inside of a runloop when running'](assert) {
      (0, _runloop.run)(() => {
        (0, _runloop.once)(() => assert.ok(!!(0, _runloop.getCurrentRunLoop)(), 'should have a runloop'));
      });
    }
  });
});
enifed('@ember/runloop/tests/onerror_test', ['@ember/runloop', '@ember/-internals/error-handling', '@ember/debug', 'internal-test-helpers'], function (_runloop, _errorHandling, _debug, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/onerror_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test With Ember.onerror undefined, errors in run are thrown'](assert) {
      let thrown = new Error('Boom!');
      let original = (0, _errorHandling.getOnerror)();

      let caught;
      (0, _errorHandling.setOnerror)(undefined);
      try {
        (0, _runloop.run)(() => {
          throw thrown;
        });
      } catch (error) {
        caught = error;
      } finally {
        (0, _errorHandling.setOnerror)(original);
      }

      assert.deepEqual(caught, thrown);
    }

    ['@test With Ember.onerror set, errors in run are caught'](assert) {
      let thrown = new Error('Boom!');
      let original = (0, _errorHandling.getOnerror)();
      let originalDispatchOverride = (0, _errorHandling.getDispatchOverride)();
      let originalIsTesting = (0, _debug.isTesting)();

      let caught;
      (0, _errorHandling.setOnerror)(error => {
        caught = error;
      });
      (0, _errorHandling.setDispatchOverride)(null);
      (0, _debug.setTesting)(false);

      try {
        (0, _runloop.run)(() => {
          throw thrown;
        });
      } finally {
        (0, _errorHandling.setOnerror)(original);
        (0, _errorHandling.setDispatchOverride)(originalDispatchOverride);
        (0, _debug.setTesting)(originalIsTesting);
      }

      assert.deepEqual(caught, thrown);
    }
  });
});
enifed('@ember/runloop/tests/run_bind_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/run_bind_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test bind builds a run-loop wrapped callback handler'](assert) {
      assert.expect(3);

      let obj = {
        value: 0,
        increment(increment) {
          assert.ok((0, _runloop.getCurrentRunLoop)(), 'expected a run-loop');
          return this.value += increment;
        }
      };

      let proxiedFunction = (0, _runloop.bind)(obj, obj.increment, 1);
      assert.equal(proxiedFunction(), 1);
      assert.equal(obj.value, 1);
    }

    ['@test bind keeps the async callback arguments'](assert) {
      assert.expect(4);

      function asyncCallback(increment, increment2, increment3) {
        assert.ok((0, _runloop.getCurrentRunLoop)(), 'expected a run-loop');
        assert.equal(increment, 1);
        assert.equal(increment2, 2);
        assert.equal(increment3, 3);
      }

      function asyncFunction(fn) {
        fn(2, 3);
      }

      asyncFunction((0, _runloop.bind)(asyncCallback, asyncCallback, 1));
    }

    ['@test [GH#16652] bind throws an error if callback is undefined']() {
      let assertBindThrows = (msg, ...args) => {
        expectAssertion(function () {
          (0, _runloop.bind)(...args);
        }, /could not find a suitable method to bind/, msg);
      };
      assertBindThrows('without arguments');
      assertBindThrows('with one arguments that is not a function', 'myMethod');
      assertBindThrows('if second parameter is not a function and not a property in first parameter', Object.create(null), 'myMethod');
    }
  });
});
enifed('@ember/runloop/tests/run_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/run_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test run invokes passed function, returning value'](assert) {
      let obj = {
        foo() {
          return [this.bar, 'FOO'];
        },
        bar: 'BAR',
        checkArgs(arg1, arg2) {
          return [arg1, this.bar, arg2];
        }
      };

      assert.equal((0, _runloop.run)(() => 'FOO'), 'FOO', 'pass function only');
      assert.deepEqual((0, _runloop.run)(obj, obj.foo), ['BAR', 'FOO'], 'pass obj and obj.method');
      assert.deepEqual((0, _runloop.run)(obj, 'foo'), ['BAR', 'FOO'], 'pass obj and "method"');
      assert.deepEqual((0, _runloop.run)(obj, obj.checkArgs, 'hello', 'world'), ['hello', 'BAR', 'world'], 'pass obj, obj.method, and extra arguments');
    }
  });
});
enifed('@ember/runloop/tests/schedule_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/schedule_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test scheduling item in queue should defer until finished'](assert) {
      let cnt = 0;

      (0, _runloop.run)(() => {
        (0, _runloop.schedule)('actions', () => cnt++);
        (0, _runloop.schedule)('actions', () => cnt++);
        assert.equal(cnt, 0, 'should not run action yet');
      });

      assert.equal(cnt, 2, 'should flush actions now');
    }

    ['@test a scheduled item can be canceled'](assert) {
      let hasRan = false;

      (0, _runloop.run)(() => {
        let cancelId = (0, _runloop.schedule)('actions', () => hasRan = true);
        (0, _runloop.cancel)(cancelId);
      });

      assert.notOk(hasRan, 'should not have ran callback run');
    }

    ['@test nested runs should queue each phase independently'](assert) {
      let cnt = 0;

      (0, _runloop.run)(() => {
        (0, _runloop.schedule)('actions', () => cnt++);
        assert.equal(cnt, 0, 'should not run action yet');

        (0, _runloop.run)(() => {
          (0, _runloop.schedule)('actions', () => cnt++);
        });
        assert.equal(cnt, 1, 'should not run action yet');
      });

      assert.equal(cnt, 2, 'should flush actions now');
    }

    ['@test prior queues should be flushed before moving on to next queue'](assert) {
      let order = [];

      (0, _runloop.run)(() => {
        let runLoop = (0, _runloop.getCurrentRunLoop)();
        assert.ok(runLoop, 'run loop present');

        expectDeprecation(() => {
          (0, _runloop.schedule)('sync', () => {
            order.push('sync');
            assert.equal(runLoop, (0, _runloop.getCurrentRunLoop)(), 'same run loop used');
          });
        }, `Scheduling into the 'sync' run loop queue is deprecated.`);

        (0, _runloop.schedule)('actions', () => {
          order.push('actions');
          assert.equal(runLoop, (0, _runloop.getCurrentRunLoop)(), 'same run loop used');

          (0, _runloop.schedule)('actions', () => {
            order.push('actions');
            assert.equal(runLoop, (0, _runloop.getCurrentRunLoop)(), 'same run loop used');
          });

          expectDeprecation(() => {
            (0, _runloop.schedule)('sync', () => {
              order.push('sync');
              assert.equal(runLoop, (0, _runloop.getCurrentRunLoop)(), 'same run loop used');
            });
          }, `Scheduling into the 'sync' run loop queue is deprecated.`);
        });

        (0, _runloop.schedule)('destroy', () => {
          order.push('destroy');
          assert.equal(runLoop, (0, _runloop.getCurrentRunLoop)(), 'same run loop used');
        });
      });

      assert.deepEqual(order, ['sync', 'actions', 'sync', 'actions', 'destroy']);
    }
  });
});
enifed('@ember/runloop/tests/sync_test', ['@ember/runloop', 'internal-test-helpers'], function (_runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/sync_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test sync() will immediately flush the sync queue only'](assert) {
      let cnt = 0;

      (0, _runloop.run)(() => {
        function cntup() {
          cnt++;
        }

        function syncfunc() {
          if (++cnt < 5) {
            expectDeprecation(() => {
              (0, _runloop.schedule)('sync', syncfunc);
            }, `Scheduling into the 'sync' run loop queue is deprecated.`);
          }
          (0, _runloop.schedule)('actions', cntup);
        }

        syncfunc();

        assert.equal(cnt, 1, 'should not run action yet');
      });

      assert.equal(cnt, 10, 'should flush actions now too');
    }
  });
});
enifed('@ember/runloop/tests/unwind_test', ['@ember/runloop', '@ember/error', 'internal-test-helpers'], function (_runloop, _error, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('system/run_loop/unwind_test', class extends _internalTestHelpers.AbstractTestCase {
    ['@test RunLoop unwinds despite unhandled exception'](assert) {
      let initialRunLoop = (0, _runloop.getCurrentRunLoop)();

      assert.throws(() => {
        (0, _runloop.run)(() => {
          (0, _runloop.schedule)('actions', function () {
            throw new _error.default('boom!');
          });
        });
      }, Error, 'boom!');

      // The real danger at this point is that calls to autorun will stick
      // tasks into the already-dead runloop, which will never get
      // flushed. I can't easily demonstrate this in a unit test because
      // autorun explicitly doesn't work in test mode. - ef4
      assert.equal((0, _runloop.getCurrentRunLoop)(), initialRunLoop, 'Previous run loop should be cleaned up despite exception');
    }

    ['@test run unwinds despite unhandled exception'](assert) {
      var initialRunLoop = (0, _runloop.getCurrentRunLoop)();

      assert.throws(() => {
        (0, _runloop.run)(function () {
          throw new _error.default('boom!');
        });
      }, _error.default, 'boom!');

      assert.equal((0, _runloop.getCurrentRunLoop)(), initialRunLoop, 'Previous run loop should be cleaned up despite exception');
    }
  });
});
enifed('@ember/string/tests/camelize_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.camelize)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.camelize(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.camelize', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.camelize is not modified without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.camelize, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String camelize tests'](assert) {
      test(assert, 'my favorite items', 'myFavoriteItems', 'camelize normal string');
      test(assert, 'I Love Ramen', 'iLoveRamen', 'camelize capitalized string');
      test(assert, 'css-class-name', 'cssClassName', 'camelize dasherized string');
      test(assert, 'action_name', 'actionName', 'camelize underscored string');
      test(assert, 'action.name', 'actionName', 'camelize dot notation string');
      test(assert, 'innerHTML', 'innerHTML', 'does nothing with camelcased string');
      test(assert, 'PrivateDocs/OwnerInvoice', 'privateDocs/ownerInvoice', 'camelize namespaced classified string');
      test(assert, 'private_docs/owner_invoice', 'privateDocs/ownerInvoice', 'camelize namespaced underscored string');
      test(assert, 'private-docs/owner-invoice', 'privateDocs/ownerInvoice', 'camelize namespaced dasherized string');
    }
  });
});
enifed('@ember/string/tests/capitalize_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.capitalize)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.capitalize(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.capitalize', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.capitalize is not modified without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.capitalize, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String capitalize tests'](assert) {
      test(assert, 'my favorite items', 'My favorite items', 'capitalize normal string');
      test(assert, 'css-class-name', 'Css-class-name', 'capitalize dasherized string');
      test(assert, 'action_name', 'Action_name', 'capitalize underscored string');
      test(assert, 'innerHTML', 'InnerHTML', 'capitalize camelcased string');
      test(assert, 'Capitalized string', 'Capitalized string', 'does nothing with capitalized string');
      test(assert, 'privateDocs/ownerInvoice', 'PrivateDocs/OwnerInvoice', 'capitalize namespaced camelized string');
      test(assert, 'private_docs/owner_invoice', 'Private_docs/Owner_invoice', 'capitalize namespaced underscored string');
      test(assert, 'private-docs/owner-invoice', 'Private-docs/Owner-invoice', 'capitalize namespaced dasherized string');
      test(assert, 'šabc', 'Šabc', 'capitalize string with accent character');
    }
  });
});
enifed('@ember/string/tests/classify_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.classify)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.classify(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.classify', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.classify is not modified without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.classify, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String classify tests'](assert) {
      test(assert, 'my favorite items', 'MyFavoriteItems', 'classify normal string');
      test(assert, 'css-class-name', 'CssClassName', 'classify dasherized string');
      test(assert, 'action_name', 'ActionName', 'classify underscored string');
      test(assert, 'privateDocs/ownerInvoice', 'PrivateDocs/OwnerInvoice', 'classify namespaced camelized string');
      test(assert, 'private_docs/owner_invoice', 'PrivateDocs/OwnerInvoice', 'classify namespaced underscored string');
      test(assert, 'private-docs/owner-invoice', 'PrivateDocs/OwnerInvoice', 'classify namespaced dasherized string');
      test(assert, '-view-registry', '_ViewRegistry', 'classify prefixed dasherized string');
      test(assert, 'components/-text-field', 'Components/_TextField', 'classify namespaced prefixed dasherized string');
      test(assert, '_Foo_Bar', '_FooBar', 'classify underscore-prefixed underscored string');
      test(assert, '_Foo-Bar', '_FooBar', 'classify underscore-prefixed dasherized string');
      test(assert, '_foo/_bar', '_Foo/_Bar', 'classify underscore-prefixed-namespaced underscore-prefixed string');
      test(assert, '-foo/_bar', '_Foo/_Bar', 'classify dash-prefixed-namespaced underscore-prefixed string');
      test(assert, '-foo/-bar', '_Foo/_Bar', 'classify dash-prefixed-namespaced dash-prefixed string');
      test(assert, 'InnerHTML', 'InnerHTML', 'does nothing with classified string');
      test(assert, '_FooBar', '_FooBar', 'does nothing with classified prefixed string');
    }
  });
});
enifed('@ember/string/tests/dasherize_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.dasherize)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.dasherize(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.dasherize', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.dasherize is not modified without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.dasherize, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String dasherize tests'](assert) {
      test(assert, 'my favorite items', 'my-favorite-items', 'dasherize normal string');
      test(assert, 'css-class-name', 'css-class-name', 'does nothing with dasherized string');
      test(assert, 'action_name', 'action-name', 'dasherize underscored string');
      test(assert, 'innerHTML', 'inner-html', 'dasherize camelcased string');
      test(assert, 'toString', 'to-string', 'dasherize string that is the property name of Object.prototype');
      test(assert, 'PrivateDocs/OwnerInvoice', 'private-docs/owner-invoice', 'dasherize namespaced classified string');
      test(assert, 'privateDocs/ownerInvoice', 'private-docs/owner-invoice', 'dasherize namespaced camelized string');
      test(assert, 'private_docs/owner_invoice', 'private-docs/owner-invoice', 'dasherize namespaced underscored string');
    }
  });
});
enifed('@ember/string/tests/decamelize_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.decamelize)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.decamelize(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.decamelize', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.decamelize is not modified without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.decamelize, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String decamelize tests'](assert) {
      test(assert, 'my favorite items', 'my favorite items', 'does nothing with normal string');
      test(assert, 'css-class-name', 'css-class-name', 'does nothing with dasherized string');
      test(assert, 'action_name', 'action_name', 'does nothing with underscored string');
      test(assert, 'innerHTML', 'inner_html', 'converts a camelized string into all lower case separated by underscores.');
      test(assert, 'size160Url', 'size160_url', 'decamelizes strings with numbers');
      test(assert, 'PrivateDocs/OwnerInvoice', 'private_docs/owner_invoice', 'decamelize namespaced classified string');
      test(assert, 'privateDocs/ownerInvoice', 'private_docs/owner_invoice', 'decamelize namespaced camelized string');
    }
  });
});
enifed('@ember/string/tests/loc_test', ['@ember/-internals/environment', '@ember/string', '@ember/string/lib/string_registry', 'internal-test-helpers'], function (_environment, _string, _string_registry, _internalTestHelpers) {
  'use strict';

  let oldString;

  function test(assert, given, args, expected, description) {
    assert.equal((0, _string.loc)(given, args), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.loc(...args), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.loc', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      oldString = (0, _string_registry.getStrings)();
      (0, _string_registry.setStrings)({
        '_Hello World': 'Bonjour le monde',
        '_Hello %@': 'Bonjour %@',
        '_Hello %@ %@': 'Bonjour %@ %@',
        '_Hello %@# %@#': 'Bonjour %@2 %@1'
      });
    }

    afterEach() {
      (0, _string_registry.setStrings)(oldString);
    }

    ['@test String.prototype.loc is not available without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.loc, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String loc tests'](assert) {
      test(assert, '_Hello World', [], 'Bonjour le monde', `loc('_Hello World') => 'Bonjour le monde'`);
      test(assert, '_Hello %@ %@', ['John', 'Doe'], 'Bonjour John Doe', `loc('_Hello %@ %@', ['John', 'Doe']) => 'Bonjour John Doe'`);
      test(assert, '_Hello %@# %@#', ['John', 'Doe'], 'Bonjour Doe John', `loc('_Hello %@# %@#', ['John', 'Doe']) => 'Bonjour Doe John'`);
      test(assert, '_Not In Strings', [], '_Not In Strings', `loc('_Not In Strings') => '_Not In Strings'`);
    }

    ['@test works with argument form'](assert) {
      assert.equal((0, _string.loc)('_Hello %@', 'John'), 'Bonjour John');
      assert.equal((0, _string.loc)('_Hello %@ %@', ['John'], 'Doe'), 'Bonjour John Doe');
    }
  });
});
enifed('@ember/string/tests/underscore_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.underscore)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.underscore(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.underscore', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.underscore is not available without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.underscore, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String underscore tests'](assert) {
      test(assert, 'my favorite items', 'my_favorite_items', 'with normal string');
      test(assert, 'css-class-name', 'css_class_name', 'with dasherized string');
      test(assert, 'action_name', 'action_name', 'does nothing with underscored string');
      test(assert, 'innerHTML', 'inner_html', 'with camelcased string');
      test(assert, 'PrivateDocs/OwnerInvoice', 'private_docs/owner_invoice', 'underscore namespaced classified string');
      test(assert, 'privateDocs/ownerInvoice', 'private_docs/owner_invoice', 'underscore namespaced camelized string');
      test(assert, 'private-docs/owner-invoice', 'private_docs/owner_invoice', 'underscore namespaced dasherized string');
    }
  });
});
enifed('@ember/string/tests/w_test', ['@ember/-internals/environment', '@ember/string', 'internal-test-helpers'], function (_environment, _string, _internalTestHelpers) {
  'use strict';

  function test(assert, given, expected, description) {
    assert.deepEqual((0, _string.w)(given), expected, description);
    if (_environment.ENV.EXTEND_PROTOTYPES.String) {
      assert.deepEqual(given.w(), expected, description);
    }
  }

  (0, _internalTestHelpers.moduleFor)('EmberStringUtils.w', class extends _internalTestHelpers.AbstractTestCase {
    ['@test String.prototype.w is not available without EXTEND_PROTOTYPES'](assert) {
      if (!_environment.ENV.EXTEND_PROTOTYPES.String) {
        assert.ok('undefined' === typeof String.prototype.w, 'String.prototype helper disabled');
      } else {
        assert.expect(0);
      }
    }

    ['@test String w tests'](assert) {
      test(assert, 'one two three', ['one', 'two', 'three'], `w('one two three') => ['one','two','three']`);
      test(assert, 'one   two  three', ['one', 'two', 'three'], `w('one    two    three') with extra spaces between words => ['one','two','three']`);
      test(assert, 'one\ttwo  three', ['one', 'two', 'three'], `w('one two three') with tabs`);
    }
  });
});
enifed('ember-template-compiler/tests/plugins/assert-if-helper-without-arguments-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-if-helper-without-argument', class extends _internalTestHelpers.AbstractTestCase {
    [`@test block if helper expects one argument`]() {
      expectAssertion(() => {
        (0, _index.compile)(`{{#if}}aVal{{/if}}`, {
          moduleName: 'baz/foo-bar'
        });
      }, `#if requires a single argument. ('baz/foo-bar' @ L1:C0) `);

      expectAssertion(() => {
        (0, _index.compile)(`{{#if val1 val2}}aVal{{/if}}`, {
          moduleName: 'baz/foo-bar'
        });
      }, `#if requires a single argument. ('baz/foo-bar' @ L1:C0) `);

      expectAssertion(() => {
        (0, _index.compile)(`{{#if}}aVal{{/if}}`, {
          moduleName: 'baz/foo-bar'
        });
      }, `#if requires a single argument. ('baz/foo-bar' @ L1:C0) `);
    }

    [`@test inline if helper expects between one and three arguments`]() {
      expectAssertion(() => {
        (0, _index.compile)(`{{if}}`, {
          moduleName: 'baz/foo-bar'
        });
      }, `The inline form of the 'if' helper expects two or three arguments. ('baz/foo-bar' @ L1:C0) `);

      (0, _index.compile)(`{{if foo bar baz}}`, {
        moduleName: 'baz/foo-bar'
      });
    }

    ['@test subexpression if helper expects between one and three arguments']() {
      expectAssertion(() => {
        (0, _index.compile)(`{{input foo=(if)}}`, {
          moduleName: 'baz/foo-bar'
        });
      }, `The inline form of the 'if' helper expects two or three arguments. ('baz/foo-bar' @ L1:C12) `);

      (0, _index.compile)(`{{some-thing foo=(if foo bar baz)}}`, {
        moduleName: 'baz/foo-bar'
      });
    }
  });
});
enifed('ember-template-compiler/tests/plugins/assert-input-helper-without-block-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-input-helper-without-block', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Using {{#input}}{{/input}} is not valid']() {
      let expectedMessage = `The {{input}} helper cannot be used in block form. ('baz/foo-bar' @ L1:C0) `;

      expectAssertion(() => {
        (0, _index.compile)('{{#input value="123"}}Completely invalid{{/input}}', {
          moduleName: 'baz/foo-bar'
        });
      }, expectedMessage);
    }
  });
});
enifed('ember-template-compiler/tests/plugins/assert-local-variable-shadowing-helper-invocation-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-local-variable-shadowing-helper-invocation', class extends _internalTestHelpers.AbstractTestCase {
    [`@test block statements shadowing sub-expression invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            {{concat (foo)}}
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C21) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            {{concat (foo bar baz)}}
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C21) `);

      // Not shadowed

      (0, _index.compile)(`
        {{#let foo as |foo|}}{{/let}}
        {{concat (foo)}}
        {{concat (foo bar baz)}}`, { moduleName: 'baz/foo-bar' });

      // Not an invocation

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          {{concat foo}}
        {{/let}}`, { moduleName: 'baz/foo-bar' });
    }

    [`@test element nodes shadowing sub-expression invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          <Foo as |foo|>
            {{concat (foo)}}
          </Foo>`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C21) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          <Foo as |foo|>
            {{concat (foo bar baz)}}
          </Foo>`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C21) `);

      // Not shadowed

      (0, _index.compile)(`
        <Foo as |foo|></Foo>
        {{concat (foo)}}
        {{concat (foo bar baz)}}`, { moduleName: 'baz/foo-bar' });

      // Not an invocation

      (0, _index.compile)(`
        <Foo as |foo|>
          {{concat foo}}
        </Foo>`, { moduleName: 'baz/foo-bar' });
    }

    [`@test deeply nested sub-expression invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <FooBar as |bar|>
              {{#each items as |baz|}}
                {{concat (foo)}}
              {{/each}}
            </FooBar>
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L5:C25) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <FooBar as |bar|>
              {{#each items as |baz|}}
                {{concat (foo bar baz)}}
              {{/each}}
            </FooBar>
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` helper because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L5:C25) `);

      // Not shadowed

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          <FooBar as |bar|>
            {{#each items as |baz|}}
            {{/each}}
            {{concat (baz)}}
            {{concat (baz bat)}}
          </FooBar>
          {{concat (bar)}}
          {{concat (bar baz bat)}}
        {{/let}}
        {{concat (foo)}}
        {{concat (foo bar baz bat)}}`, { moduleName: 'baz/foo-bar' });

      // Not an invocation

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          <FooBar as |bar|>
            {{#each items as |baz|}}
              {{concat foo}}
            {{/each}}
          </FooBar>
        {{/let}}`, { moduleName: 'baz/foo-bar' });
    }

    [`@test block statements shadowing mustache invocations`](assert) {
      // These are fine, because they should already be considered contextual
      // component invocations, not helper invocations
      assert.expect(0);

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          {{foo}}
        {{/let}}`, { moduleName: 'baz/foo-bar' });

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          {{foo bar baz}}
        {{/let}}`, { moduleName: 'baz/foo-bar' });
    }

    [`@test element nodes shadowing mustache invocations`](assert) {
      // These are fine, because they should already be considered contextual
      // component invocations, not helper invocations
      assert.expect(0);

      (0, _index.compile)(`
        <Foo as |foo|>
          {{foo}}
        </Foo>`, { moduleName: 'baz/foo-bar' });

      (0, _index.compile)(`
        <Foo as |foo|>
          {{foo bar baz}}
        </Foo>`, { moduleName: 'baz/foo-bar' });
    }

    [`@test deeply nested mustache invocations`](assert) {
      // These are fine, because they should already be considered contextual
      // component invocations, not helper invocations
      assert.expect(0);

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          <FooBar as |bar|>
            {{#each items as |baz|}}
              {{foo}}
            {{/each}}
          </FooBar>
        {{/let}}`, { moduleName: 'baz/foo-bar' });

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          <FooBar as |bar|>
            {{#each items as |baz|}}
              {{foo bar baz}}
            {{/each}}
          </FooBar>
        {{/let}}`, { moduleName: 'baz/foo-bar' });
    }

    [`@test block statements shadowing modifier invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <div {{foo}} />
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C17) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <div {{foo bar baz}} />
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C17) `);

      // Not shadowed

      (0, _index.compile)(`
        {{#let foo as |foo|}}{{/let}}
        <div {{foo}} />
        <div {{foo bar baz}} />`, { moduleName: 'baz/foo-bar' });
    }

    [`@test element nodes shadowing modifier invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          <Foo as |foo|>
            <div {{foo}} />
          </Foo>`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C17) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          <Foo as |foo|>
            <div {{foo bar baz}} />
          </Foo>`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L3:C17) `);

      // Not shadowed

      (0, _index.compile)(`
        <Foo as |foo|></Foo>
        <div {{foo}} />
        <div {{foo bar baz}} />`, { moduleName: 'baz/foo-bar' });
    }

    [`@test deeply nested modifier invocations`]() {
      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <FooBar as |bar|>
              {{#each items as |baz|}}
                <div {{foo}} />
              {{/each}}
            </FooBar>
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L5:C21) `);

      expectAssertion(() => {
        (0, _index.compile)(`
          {{#let foo as |foo|}}
            <FooBar as |bar|>
              {{#each items as |baz|}}
                <div {{foo bar baz}} />
              {{/each}}
            </FooBar>
          {{/let}}`, { moduleName: 'baz/foo-bar' });
      }, `Cannot invoke the \`foo\` modifier because it was shadowed by a local variable (i.e. a block param) with the same name. Please rename the local variable to resolve the conflict. ('baz/foo-bar' @ L5:C21) `);

      // Not shadowed

      (0, _index.compile)(`
        {{#let foo as |foo|}}
          <FooBar as |bar|>
            {{#each items as |baz|}}
            {{/each}}
            <div {{baz}} />
            <div {{baz bat}} />
          </FooBar>
          <div {{bar}} />
          <div {{bar baz bat}} />
        {{/let}}
        <div {{foo}} />
        <div {{foo bar baz bat}} />`, { moduleName: 'baz/foo-bar' });
    }
  });
});
enifed('ember-template-compiler/tests/plugins/assert-reserved-named-arguments-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  if (true /* EMBER_GLIMMER_NAMED_ARGUMENTS */) {
      (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-reserved-named-arguments (EMBER_GLIMMER_NAMED_ARGUMENTS) ', class extends _internalTestHelpers.AbstractTestCase {
        [`@test '@arguments' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@arguments}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@arguments' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @arguments}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@arguments' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @arguments "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@arguments' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@args' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@args}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@args' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @args}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@args' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @args "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@args' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@block' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@block}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@block' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @block}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@block' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @block "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@block' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@else' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@else}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@else' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @else}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@else' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @else "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@else' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        // anything else that doesn't start with a lower case letter
        [`@test '@Arguments' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@Arguments}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Arguments' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @Arguments}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Arguments' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @Arguments "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Arguments' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@Args' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@Args}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Args' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @Args}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Args' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @Args "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Args' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@FOO' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@FOO}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@FOO' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @FOO}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@FOO' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @FOO "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@FOO' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@Foo' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@Foo}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Foo' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @Foo}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Foo' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @Foo "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@Foo' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@.' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@.}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@.' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @.}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@.' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @. "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@.' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@_' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@_}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@_' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @_}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@_' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @_ "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@_' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@-' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@-}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@-' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @-}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@-' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @- "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@-' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@$' is reserved`]() {
          expectAssertion(() => {
            (0, _index.compile)(`{{@$}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@$' is reserved. ('baz/foo-bar' @ L1:C2) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{#if @$}}Yup{{/if}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@$' is reserved. ('baz/foo-bar' @ L1:C6) `);

          expectAssertion(() => {
            (0, _index.compile)(`{{input type=(if @$ "bar" "baz")}}`, {
              moduleName: 'baz/foo-bar'
            });
          }, `'@$' is reserved. ('baz/foo-bar' @ L1:C17) `);
        }

        [`@test '@' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @ "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@0' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@0}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @0}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @0 "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@1' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@1}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @1}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @1 "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@2' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@2}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @2}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @2 "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@@' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@@}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @@}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @@ "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@=' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@=}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @=}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @= "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }

        [`@test '@!' is de facto reserved (parse error)`](assert) {
          assert.throws(() => {
            (0, _index.compile)('{{@!}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{#if @!}}Yup{{/if}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);

          assert.throws(() => {
            (0, _index.compile)('{{input type=(if @! "bar" "baz")}}', {
              moduleName: 'baz/foo-bar'
            });
          }, /Expecting 'ID'/);
        }
      });
    } else {
    (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-reserved-named-arguments', class extends _internalTestHelpers.AbstractTestCase {
      ['@test Paths beginning with @ are not valid']() {
        expectAssertion(() => {
          (0, _index.compile)('{{@foo}}', {
            moduleName: 'baz/foo-bar'
          });
        }, `'@foo' is not a valid path. ('baz/foo-bar' @ L1:C2) `);

        expectAssertion(() => {
          (0, _index.compile)('{{#if @foo}}Yup{{/if}}', {
            moduleName: 'baz/foo-bar'
          });
        }, `'@foo' is not a valid path. ('baz/foo-bar' @ L1:C6) `);

        expectAssertion(() => {
          (0, _index.compile)('{{input type=(if @foo "bar" "baz")}}', {
            moduleName: 'baz/foo-bar'
          });
        }, `'@foo' is not a valid path. ('baz/foo-bar' @ L1:C17) `);
      }
    });
  }
});
enifed('ember-template-compiler/tests/plugins/assert-splattribute-expression-test', ['internal-test-helpers', 'ember-template-compiler/index'], function (_internalTestHelpers, _index) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: assert-splattribute-expression', class extends _internalTestHelpers.AbstractTestCase {
    expectedMessage(locInfo) {
      return true /* EMBER_GLIMMER_ANGLE_BRACKET_INVOCATION */ ? `Using "...attributes" can only be used in the element position e.g. <div ...attributes />. It cannot be used as a path. (${locInfo}) ` : `...attributes is an invalid path (${locInfo}) `;
    }

    '@test ...attributes is in element space'(assert) {
      if (true /* EMBER_GLIMMER_ANGLE_BRACKET_INVOCATION */) {
          assert.expect(0);

          (0, _index.compile)('<div ...attributes>Foo</div>');
        } else {
        expectAssertion(() => {
          (0, _index.compile)('<div ...attributes>Foo</div>');
        }, this.expectedMessage('L1:C5'));
      }
    }

    '@test {{...attributes}} is not valid'() {
      expectAssertion(() => {
        (0, _index.compile)('<div>{{...attributes}}</div>', {
          moduleName: 'foo-bar'
        });
      }, this.expectedMessage(`'foo-bar' @ L1:C7`));
    }

    '@test {{...attributes}} is not valid path expression'() {
      expectAssertion(() => {
        (0, _index.compile)('<div>{{...attributes}}</div>', {
          moduleName: 'foo-bar'
        });
      }, this.expectedMessage(`'foo-bar' @ L1:C7`));
    }
    '@test {{...attributes}} is not valid modifier'() {
      expectAssertion(() => {
        (0, _index.compile)('<div {{...attributes}}>Wat</div>', {
          moduleName: 'foo-bar'
        });
      }, this.expectedMessage(`'foo-bar' @ L1:C7`));
    }

    '@test {{...attributes}} is not valid attribute'() {
      expectAssertion(() => {
        (0, _index.compile)('<div class={{...attributes}}>Wat</div>', {
          moduleName: 'foo-bar'
        });
      }, this.expectedMessage(`'foo-bar' @ L1:C13`));
    }
  });
});
enifed('ember-template-compiler/tests/plugins/deprecate-send-action-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  const EVENTS = ['insert-newline', 'enter', 'escape-press', 'focus-in', 'focus-out', 'key-press', 'key-up', 'key-down'];

  class DeprecateSendActionTest extends _internalTestHelpers.AbstractTestCase {}

  EVENTS.forEach(function (e) {
    DeprecateSendActionTest.prototype[`@test Using \`{{input ${e}="actionName"}}\` provides a deprecation`] = function () {
      let expectedMessage = `Please refactor \`{{input ${e}="foo-bar"}}\` to \`{{input ${e}=(action "foo-bar")}}\. ('baz/foo-bar' @ L1:C0) `;

      expectDeprecation(() => {
        (0, _index.compile)(`{{input ${e}="foo-bar"}}`, { moduleName: 'baz/foo-bar' });
      }, expectedMessage);
    };
  });

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: deprecate-send-action', DeprecateSendActionTest);
});
enifed('ember-template-compiler/tests/plugins/transform-component-invocation-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: transforms component invocation', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Does not throw a compiler error for component invocations'](assert) {
      assert.expect(0);

      ['{{this.modal open}}', '{{this.modal isOpen=true}}', '{{#this.modal}}Woot{{/this.modal}}', '{{@modal open}}', // RFC#311
      '{{@modal isOpen=true}}', // RFC#311
      '{{#@modal}}Woot{{/@modal}}', // RFC#311
      '{{c.modal open}}', '{{c.modal isOpen=true}}', '{{#c.modal}}Woot{{/c.modal}}', '{{#my-component as |c|}}{{c name="Chad"}}{{/my-component}}', // RFC#311
      '{{#my-component as |c|}}{{c "Chad"}}{{/my-component}}', // RFC#311
      '{{#my-component as |c|}}{{#c}}{{/c}}{{/my-component}}', // RFC#311
      '<input disabled={{true}}>', // GH#15740
      '<td colspan={{3}}></td>'].forEach((layout, i) => {
        (0, _index.compile)(layout, { moduleName: `example-${i}` });
      });
    }
  });
});
enifed('ember-template-compiler/tests/plugins/transform-inline-link-to-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: inline-link-to', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Can transform an inline {{link-to}} without error'](assert) {
      assert.expect(0);

      (0, _index.compile)(`{{link-to 'foo' 'index'}}`, {
        moduleName: 'foo/bar/baz'
      });
    }
  });
});
enifed('ember-template-compiler/tests/plugins/transform-input-type-syntax-test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: input type syntax', class extends _internalTestHelpers.AbstractTestCase {
    ['@test Can compile an {{input}} helper that has a sub-expression value as its type'](assert) {
      assert.expect(0);

      (0, _index.compile)(`{{input type=(if true 'password' 'text')}}`);
    }

    ['@test Can compile an {{input}} helper with a string literal type'](assert) {
      assert.expect(0);

      (0, _index.compile)(`{{input type='text'}}`);
    }

    ['@test Can compile an {{input}} helper with a type stored in a var'](assert) {
      assert.expect(0);

      (0, _index.compile)(`{{input type=_type}}`);
    }
  });
});
enifed('ember-template-compiler/tests/system/bootstrap-test', ['@ember/runloop', '@ember/-internals/glimmer', 'ember-template-compiler/lib/system/bootstrap', 'internal-test-helpers'], function (_runloop, _glimmer, _bootstrap, _internalTestHelpers) {
  'use strict';

  let component, fixture;

  function checkTemplate(templateName, assert) {
    (0, _runloop.run)(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }));

    let template = (0, _glimmer.getTemplate)(templateName);
    let qunitFixture = document.querySelector('#qunit-fixture');

    assert.ok(template, 'template is available on Ember.TEMPLATES');
    assert.notOk(qunitFixture.querySelector('script'), 'script removed');

    let owner = (0, _internalTestHelpers.buildOwner)();
    owner.register('template:-top-level', template);
    owner.register('component:-top-level', _glimmer.Component.extend({
      layoutName: '-top-level',
      firstName: 'Tobias',
      drug: 'teamocil'
    }));

    component = owner.lookup('component:-top-level');
    (0, _internalTestHelpers.runAppend)(component);

    assert.equal(qunitFixture.textContent.trim(), 'Tobias takes teamocil', 'template works');
    (0, _internalTestHelpers.runDestroy)(owner);
  }

  (0, _internalTestHelpers.moduleFor)('ember-templates: bootstrap', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();

      fixture = document.getElementById('qunit-fixture');
    }

    teardown() {
      (0, _glimmer.setTemplates)({});
      fixture = component = null;
    }

    ['@test template with data-template-name should add a new template to Ember.TEMPLATES'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars" data-template-name="funkyTemplate">{{firstName}} takes {{drug}}</script>';

      checkTemplate('funkyTemplate', assert);
    }

    ['@test template with id instead of data-template-name should add a new template to Ember.TEMPLATES'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars" id="funkyTemplate" >{{firstName}} takes {{drug}}</script>';

      checkTemplate('funkyTemplate', assert);
    }

    ['@test template without data-template-name or id should default to application'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars">{{firstName}} takes {{drug}}</script>';

      checkTemplate('application', assert);
    }

    // Add this test case, only for typeof Handlebars === 'object';
    [`${typeof Handlebars === 'object' ? '@test' : '@skip'} template with type text/x-raw-handlebars should be parsed`](assert) {
      fixture.innerHTML = '<script type="text/x-raw-handlebars" data-template-name="funkyTemplate">{{name}}</script>';

      (0, _runloop.run)(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }));

      let template = (0, _glimmer.getTemplate)('funkyTemplate');

      assert.ok(template, 'template with name funkyTemplate available');

      // This won't even work with Ember templates
      assert.equal(template({ name: 'Tobias' }).trim(), 'Tobias');
    }

    ['@test duplicated default application templates should throw exception'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars">first</script><script type="text/x-handlebars">second</script>';

      assert.throws(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }), /Template named "[^"]+" already exists\./, 'duplicate templates should not be allowed');
    }

    ['@test default default application template and id application template present should throw exception'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars">first</script><script type="text/x-handlebars" id="application">second</script>';

      assert.throws(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }), /Template named "[^"]+" already exists\./, 'duplicate templates should not be allowed');
    }

    ['@test default application template and data-template-name application template present should throw exception'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars">first</script><script type="text/x-handlebars" data-template-name="application">second</script>';

      assert.throws(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }), /Template named "[^"]+" already exists\./, 'duplicate templates should not be allowed');
    }

    ['@test duplicated template id should throw exception'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars" id="funkyTemplate">first</script><script type="text/x-handlebars" id="funkyTemplate">second</script>';

      assert.throws(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }), /Template named "[^"]+" already exists\./, 'duplicate templates should not be allowed');
    }

    ['@test duplicated template data-template-name should throw exception'](assert) {
      fixture.innerHTML = '<script type="text/x-handlebars" data-template-name="funkyTemplate">first</script><script type="text/x-handlebars" data-template-name="funkyTemplate">second</script>';

      assert.throws(() => (0, _bootstrap.default)({ context: fixture, hasTemplate: _glimmer.hasTemplate, setTemplate: _glimmer.setTemplate }), /Template named "[^"]+" already exists\./, 'duplicate templates should not be allowed');
    }
  });
});
enifed('ember-template-compiler/tests/system/compile_options_test', ['ember-template-compiler/index', 'internal-test-helpers'], function (_index, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: default compile options', class extends _internalTestHelpers.AbstractTestCase {
    ['@test default options are a new copy'](assert) {
      assert.notEqual((0, _index.compileOptions)(), (0, _index.compileOptions)());
    }

    ['@test has default AST plugins'](assert) {
      assert.expect(_index.defaultPlugins.length);

      let plugins = (0, _index.compileOptions)().plugins.ast;

      for (let i = 0; i < _index.defaultPlugins.length; i++) {
        let plugin = _index.defaultPlugins[i];
        assert.ok(plugins.indexOf(plugin) > -1, `includes ${plugin}`);
      }
    }
  });

  let customTransformCounter = 0;
  class LegacyCustomTransform {
    constructor(options) {
      customTransformCounter++;
      this.options = options;
      this.syntax = null;
    }

    transform(ast) {
      let walker = new this.syntax.Walker();

      walker.visit(ast, node => {
        if (node.type !== 'ElementNode') {
          return;
        }

        for (var i = 0; i < node.attributes.length; i++) {
          let attribute = node.attributes[i];

          if (attribute.name === 'data-test') {
            node.attributes.splice(i, 1);
          }
        }
      });

      return ast;
    }
  }

  function customTransform() {
    customTransformCounter++;

    return {
      name: 'remove-data-test',

      visitor: {
        ElementNode(node) {
          for (var i = 0; i < node.attributes.length; i++) {
            let attribute = node.attributes[i];

            if (attribute.name === 'data-test') {
              node.attributes.splice(i, 1);
            }
          }
        }
      }
    };
  }

  class CustomPluginsTests extends _internalTestHelpers.RenderingTestCase {
    afterEach() {
      customTransformCounter = 0;
      return super.afterEach();
    }

    ['@test custom plugins can be used']() {
      this.render('<div data-test="foo" data-blah="derp" class="hahaha"></div>');
      this.assertElement(this.firstChild, {
        tagName: 'div',
        attrs: { class: 'hahaha', 'data-blah': 'derp' },
        content: ''
      });
    }

    ['@test wrapped plugins are only invoked once per template'](assert) {
      this.render('<div>{{#if falsey}}nope{{/if}}</div>');
      assert.equal(customTransformCounter, 1, 'transform should only be instantiated once');
    }
  }

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: registerPlugin with a custom plugins in legacy format', class extends CustomPluginsTests {
    beforeEach() {
      (0, _index.registerPlugin)('ast', LegacyCustomTransform);
    }

    afterEach() {
      (0, _index.unregisterPlugin)('ast', LegacyCustomTransform);
      return super.afterEach();
    }

    ['@test custom registered plugins are deduplicated'](assert) {
      (0, _index.registerPlugin)('ast', LegacyCustomTransform);
      this.registerTemplate('application', '<div data-test="foo" data-blah="derp" class="hahaha"></div>');
      assert.equal(customTransformCounter, 1, 'transform should only be instantiated once');
    }
  });

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: registerPlugin with a custom plugins', class extends CustomPluginsTests {
    beforeEach() {
      (0, _index.registerPlugin)('ast', customTransform);
    }

    afterEach() {
      (0, _index.unregisterPlugin)('ast', customTransform);
      return super.afterEach();
    }

    ['@test custom registered plugins are deduplicated'](assert) {
      (0, _index.registerPlugin)('ast', customTransform);
      this.registerTemplate('application', '<div data-test="foo" data-blah="derp" class="hahaha"></div>');
      assert.equal(customTransformCounter, 1, 'transform should only be instantiated once');
    }
  });

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: custom plugins in legacy format passed to compile', class extends _internalTestHelpers.RenderingTestCase {
    // override so that we can provide custom AST plugins to compile
    compile(templateString) {
      return (0, _index.compile)(templateString, {
        plugins: {
          ast: [LegacyCustomTransform]
        }
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('ember-template-compiler: custom plugins passed to compile', class extends _internalTestHelpers.RenderingTestCase {
    // override so that we can provide custom AST plugins to compile
    compile(templateString) {
      return (0, _index.compile)(templateString, {
        plugins: {
          ast: [customTransform]
        }
      });
    }
  });
});
enifed('ember-template-compiler/tests/system/dasherize-component-name-test', ['ember-template-compiler/lib/system/dasherize-component-name', 'internal-test-helpers'], function (_dasherizeComponentName, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('dasherize-component-name', class extends _internalTestHelpers.AbstractTestCase {
    ['@test names are correctly dasherized'](assert) {
      assert.equal(_dasherizeComponentName.default.get('Foo'), 'foo');
      assert.equal(_dasherizeComponentName.default.get('foo-bar'), 'foo-bar');
      assert.equal(_dasherizeComponentName.default.get('FooBar'), 'foo-bar');
      assert.equal(_dasherizeComponentName.default.get('XBlah'), 'x-blah');
      assert.equal(_dasherizeComponentName.default.get('X-Blah'), 'x-blah');
      assert.equal(_dasherizeComponentName.default.get('Foo::BarBaz'), 'foo::bar-baz');
      assert.equal(_dasherizeComponentName.default.get('Foo::Bar-Baz'), 'foo::bar-baz');
      assert.equal(_dasherizeComponentName.default.get('Foo@BarBaz'), 'foo@bar-baz');
      assert.equal(_dasherizeComponentName.default.get('Foo@Bar-Baz'), 'foo@bar-baz');
    }
  });
});
enifed('ember-testing/tests/acceptance_test', ['internal-test-helpers', '@ember/runloop', 'ember-testing/lib/test', 'ember-testing/lib/adapters/qunit', '@ember/-internals/routing', '@ember/-internals/runtime', '@ember/-internals/views', '@ember/debug'], function (_internalTestHelpers, _runloop, _test, _qunit, _routing, _runtime, _views, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');

  var originalConsoleError = console.error; // eslint-disable-line no-console
  let testContext;

  if (!_views.jQueryDisabled) {
    (0, _internalTestHelpers.moduleFor)('ember-testing Acceptance', class extends _internalTestHelpers.AutobootApplicationTestCase {
      constructor() {
        (0, _debug.setDebugFunction)('debug', function () {});
        super();
        this._originalAdapter = _test.default.adapter;

        testContext = this;

        this.runTask(() => {
          this.createApplication();
          this.router.map(function () {
            this.route('posts');
            this.route('comments');

            this.route('abort_transition');

            this.route('redirect');
          });

          this.indexHitCount = 0;
          this.currentRoute = 'index';

          this.add('route:index', _routing.Route.extend({
            model() {
              testContext.indexHitCount += 1;
            }
          }));

          this.add('route:posts', _routing.Route.extend({
            renderTemplate() {
              testContext.currentRoute = 'posts';
              this._super(...arguments);
            }
          }));

          this.addTemplate('posts', `
          <div class="posts-view">
            <a class="dummy-link"></a>
            <div id="comments-link">
              {{#link-to \'comments\'}}Comments{{/link-to}}
            </div>
          </div>
        `);

          this.add('route:comments', _routing.Route.extend({
            renderTemplate() {
              testContext.currentRoute = 'comments';
              this._super(...arguments);
            }
          }));

          this.addTemplate('comments', `<div>{{input type="text"}}</div>`);

          this.add('route:abort_transition', _routing.Route.extend({
            beforeModel(transition) {
              transition.abort();
            }
          }));

          this.add('route:redirect', _routing.Route.extend({
            beforeModel() {
              this.transitionTo('comments');
            }
          }));

          this.application.setupForTesting();

          _test.default.registerAsyncHelper('slowHelper', () => {
            return new _runtime.RSVP.Promise(resolve => (0, _runloop.later)(resolve, 10));
          });

          this.application.injectTestHelpers();
        });
      }
      afterEach() {
        console.error = originalConsoleError; // eslint-disable-line no-console
        super.afterEach();
      }

      teardown() {
        (0, _debug.setDebugFunction)('debug', originalDebug);
        _test.default.adapter = this._originalAdapter;
        _test.default.unregisterHelper('slowHelper');
        window.slowHelper = undefined;
        testContext = undefined;
        super.teardown();
      }

      [`@test helpers can be chained with then`](assert) {
        assert.expect(6);

        window.visit('/posts').then(() => {
          assert.equal(this.currentRoute, 'posts', 'Successfully visited posts route');
          assert.equal(window.currentURL(), '/posts', 'posts URL is correct');
          return window.click('a:contains("Comments")');
        }).then(() => {
          assert.equal(this.currentRoute, 'comments', 'visit chained with click');
          return window.fillIn('.ember-text-field', 'yeah');
        }).then(() => {
          assert.equal(document.querySelector('.ember-text-field').value, 'yeah', 'chained with fillIn');
          return window.fillIn('.ember-text-field', '#qunit-fixture', 'context working');
        }).then(() => {
          assert.equal(document.querySelector('.ember-text-field').value, 'context working', 'chained with fillIn');
          return window.click('.does-not-exist');
        }).catch(e => {
          assert.equal(e.message, 'Element .does-not-exist not found.', 'Non-existent click exception caught');
        });
      }

      [`@test helpers can be chained to each other (legacy)`](assert) {
        assert.expect(7);

        window.visit('/posts').click('a:first', '#comments-link').fillIn('.ember-text-field', 'hello').then(() => {
          assert.equal(this.currentRoute, 'comments', 'Successfully visited comments route');
          assert.equal(window.currentURL(), '/comments', 'Comments URL is correct');
          assert.equal(document.querySelector('.ember-text-field').value, 'hello', 'Fillin successfully works');
          window.find('.ember-text-field').one('keypress', e => {
            assert.equal(e.keyCode, 13, 'keyevent chained with correct keyCode.');
            assert.equal(e.which, 13, 'keyevent chained with correct which.');
          });
        }).keyEvent('.ember-text-field', 'keypress', 13).visit('/posts').then(() => {
          assert.equal(this.currentRoute, 'posts', 'Thens can also be chained to helpers');
          assert.equal(window.currentURL(), '/posts', 'URL is set correct on chained helpers');
        });
      }

      [`@test helpers don't need to be chained`](assert) {
        assert.expect(5);

        window.visit('/posts');

        window.click('a:first', '#comments-link');

        window.fillIn('.ember-text-field', 'hello');

        window.andThen(() => {
          assert.equal(this.currentRoute, 'comments', 'Successfully visited comments route');
          assert.equal(window.currentURL(), '/comments', 'Comments URL is correct');
          assert.equal(window.find('.ember-text-field').val(), 'hello', 'Fillin successfully works');
        });

        window.visit('/posts');

        window.andThen(() => {
          assert.equal(this.currentRoute, 'posts');
          assert.equal(window.currentURL(), '/posts');
        });
      }

      [`@test Nested async helpers`](assert) {
        assert.expect(5);

        window.visit('/posts');

        window.andThen(() => {
          window.click('a:first', '#comments-link');
          window.fillIn('.ember-text-field', 'hello');
        });

        window.andThen(() => {
          assert.equal(this.currentRoute, 'comments', 'Successfully visited comments route');
          assert.equal(window.currentURL(), '/comments', 'Comments URL is correct');
          assert.equal(window.find('.ember-text-field').val(), 'hello', 'Fillin successfully works');
        });

        window.visit('/posts');

        window.andThen(() => {
          assert.equal(this.currentRoute, 'posts');
          assert.equal(window.currentURL(), '/posts');
        });
      }

      [`@test Multiple nested async helpers`](assert) {
        assert.expect(3);

        window.visit('/posts');

        window.andThen(() => {
          window.click('a:first', '#comments-link');

          window.fillIn('.ember-text-field', 'hello');
          window.fillIn('.ember-text-field', 'goodbye');
        });

        window.andThen(() => {
          assert.equal(window.find('.ember-text-field').val(), 'goodbye', 'Fillin successfully works');
          assert.equal(this.currentRoute, 'comments', 'Successfully visited comments route');
          assert.equal(window.currentURL(), '/comments', 'Comments URL is correct');
        });
      }

      [`@test Helpers nested in thens`](assert) {
        assert.expect(5);

        window.visit('/posts').then(() => {
          window.click('a:first', '#comments-link');
        });

        window.andThen(() => {
          window.fillIn('.ember-text-field', 'hello');
        });

        window.andThen(() => {
          assert.equal(this.currentRoute, 'comments', 'Successfully visited comments route');
          assert.equal(window.currentURL(), '/comments', 'Comments URL is correct');
          assert.equal(window.find('.ember-text-field').val(), 'hello', 'Fillin successfully works');
        });

        window.visit('/posts');

        window.andThen(() => {
          assert.equal(this.currentRoute, 'posts');
          assert.equal(window.currentURL(), '/posts', 'Posts URL is correct');
        });
      }

      [`@test Aborted transitions are not logged via Ember.Test.adapter#exception`](assert) {
        assert.expect(0);

        _test.default.adapter = _qunit.default.create({
          exception() {
            assert.ok(false, 'aborted transitions are not logged');
          }
        });

        window.visit('/abort_transition');
      }

      [`@test Unhandled exceptions are logged via Ember.Test.adapter#exception`](assert) {
        assert.expect(2);

        console.error = () => {}; // eslint-disable-line no-console
        let asyncHandled;
        _test.default.adapter = _qunit.default.create({
          exception(error) {
            assert.equal(error.message, 'Element .does-not-exist not found.', 'Exception successfully caught and passed to Ember.Test.adapter.exception');
            // handle the rejection so it doesn't leak later.
            asyncHandled.catch(() => {});
          }
        });

        window.visit('/posts');

        window.click('.invalid-element').catch(error => {
          assert.equal(error.message, 'Element .invalid-element not found.', 'Exception successfully handled in the rejection handler');
        });

        asyncHandled = window.click('.does-not-exist');
      }

      [`@test Unhandled exceptions in 'andThen' are logged via Ember.Test.adapter#exception`](assert) {
        assert.expect(1);

        console.error = () => {}; // eslint-disable-line no-console
        _test.default.adapter = _qunit.default.create({
          exception(error) {
            assert.equal(error.message, 'Catch me', 'Exception successfully caught and passed to Ember.Test.adapter.exception');
          }
        });

        window.visit('/posts');

        window.andThen(() => {
          throw new Error('Catch me');
        });
      }

      [`@test should not start routing on the root URL when visiting another`](assert) {
        assert.expect(4);

        window.visit('/posts');

        window.andThen(() => {
          assert.ok(window.find('#comments-link'), 'found comments-link');
          assert.equal(this.currentRoute, 'posts', 'Successfully visited posts route');
          assert.equal(window.currentURL(), '/posts', 'Posts URL is correct');
          assert.equal(this.indexHitCount, 0, 'should not hit index route when visiting another route');
        });
      }

      [`@test only enters the index route once when visiting `](assert) {
        assert.expect(1);

        window.visit('/');

        window.andThen(() => {
          assert.equal(this.indexHitCount, 1, 'should hit index once when visiting /');
        });
      }

      [`@test test must not finish while asyncHelpers are pending`](assert) {
        assert.expect(2);

        let async = 0;
        let innerRan = false;

        _test.default.adapter = _qunit.default.extend({
          asyncStart() {
            async++;
            this._super();
          },
          asyncEnd() {
            async--;
            this._super();
          }
        }).create();

        this.application.testHelpers.slowHelper();

        window.andThen(() => {
          innerRan = true;
        });

        assert.equal(innerRan, false, 'should not have run yet');
        assert.ok(async > 0, 'should have told the adapter to pause');

        if (async === 0) {
          // If we failed the test, prevent zalgo from escaping and breaking
          // our other tests.
          _test.default.adapter.asyncStart();
          _test.default.resolve().then(() => {
            _test.default.adapter.asyncEnd();
          });
        }
      }

      [`@test visiting a URL that causes another transition should yield the correct URL`](assert) {
        assert.expect(1);

        window.visit('/redirect');

        window.andThen(() => {
          assert.equal(window.currentURL(), '/comments', 'Redirected to Comments URL');
        });
      }

      [`@test visiting a URL and then visiting a second URL with a transition should yield the correct URL`](assert) {
        assert.expect(2);

        window.visit('/posts');

        window.andThen(function () {
          assert.equal(window.currentURL(), '/posts', 'First visited URL is correct');
        });

        window.visit('/redirect');

        window.andThen(() => {
          assert.equal(window.currentURL(), '/comments', 'Redirected to Comments URL');
        });
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing Acceptance - teardown', class extends _internalTestHelpers.AutobootApplicationTestCase {
      [`@test that the setup/teardown happens correctly`](assert) {
        assert.expect(2);

        this.runTask(() => {
          this.createApplication();
        });
        this.application.injectTestHelpers();

        assert.ok(typeof _test.default.Promise.prototype.click === 'function');

        this.runTask(() => {
          this.application.destroy();
        });

        assert.equal(_test.default.Promise.prototype.click, undefined);
      }
    });
  }
});
enifed('ember-testing/tests/adapters/adapter_test', ['@ember/runloop', 'ember-testing/lib/adapters/adapter', 'internal-test-helpers'], function (_runloop, _adapter, _internalTestHelpers) {
  'use strict';

  var adapter;

  (0, _internalTestHelpers.moduleFor)('ember-testing Adapter', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      adapter = _adapter.default.create();
    }

    teardown() {
      (0, _runloop.run)(adapter, adapter.destroy);
    }

    ['@test exception throws'](assert) {
      var error = 'Hai';
      var thrown;

      try {
        adapter.exception(error);
      } catch (e) {
        thrown = e;
      }
      assert.equal(thrown, error);
    }
  });
});
enifed('ember-testing/tests/adapters/qunit_test', ['@ember/runloop', 'ember-testing/lib/adapters/qunit', 'internal-test-helpers'], function (_runloop, _qunit, _internalTestHelpers) {
  'use strict';

  var adapter;

  (0, _internalTestHelpers.moduleFor)('ember-testing QUnitAdapter: QUnit 2.x', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      this.originalStart = QUnit.start;
      this.originalStop = QUnit.stop;

      delete QUnit.start;
      delete QUnit.stop;

      adapter = _qunit.default.create();
    }

    teardown() {
      (0, _runloop.run)(adapter, adapter.destroy);

      QUnit.start = this.originalStart;
      QUnit.stop = this.originalStop;
    }

    ['@test asyncStart waits for asyncEnd to finish a test'](assert) {
      adapter.asyncStart();

      setTimeout(function () {
        assert.ok(true);
        adapter.asyncEnd();
      }, 50);
    }

    ['@test asyncStart waits for equal numbers of asyncEnd to finish a test'](assert) {
      let adapter = _qunit.default.create();

      adapter.asyncStart();
      adapter.asyncStart();
      adapter.asyncEnd();

      setTimeout(function () {
        assert.ok(true);
        adapter.asyncEnd();
      }, 50);
    }
  });
});
enifed('ember-testing/tests/adapters_test', ['@ember/runloop', '@ember/-internals/error-handling', 'ember-testing/lib/test', 'ember-testing/lib/adapters/adapter', 'ember-testing/lib/adapters/qunit', '@ember/application', 'internal-test-helpers', '@ember/-internals/runtime', '@ember/debug'], function (_runloop, _errorHandling, _test, _adapter, _qunit, _application, _internalTestHelpers, _runtime, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');
  const noop = function () {};

  var App, originalAdapter, originalQUnit, originalWindowOnerror;

  var originalConsoleError = console.error; // eslint-disable-line no-console

  function runThatThrowsSync(message = 'Error for testing error handling') {
    return (0, _runloop.run)(() => {
      throw new Error(message);
    });
  }

  function runThatThrowsAsync(message = 'Error for testing error handling') {
    return (0, _runloop.next)(() => {
      throw new Error(message);
    });
  }

  class AdapterSetupAndTearDown extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super();
      originalAdapter = _test.default.adapter;
      originalQUnit = window.QUnit;
      originalWindowOnerror = window.onerror;
    }

    afterEach() {
      console.error = originalConsoleError; // eslint-disable-line no-console
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
      if (App) {
        (0, _runloop.run)(App, App.destroy);
        App.removeTestHelpers();
        App = null;
      }

      _test.default.adapter = originalAdapter;
      window.QUnit = originalQUnit;
      window.onerror = originalWindowOnerror;
      (0, _errorHandling.setOnerror)(undefined);
    }
  }

  (0, _internalTestHelpers.moduleFor)('ember-testing Adapters', class extends AdapterSetupAndTearDown {
    ['@test Setting a test adapter manually'](assert) {
      assert.expect(1);
      var CustomAdapter;

      CustomAdapter = _adapter.default.extend({
        asyncStart() {
          assert.ok(true, 'Correct adapter was used');
        }
      });

      (0, _runloop.run)(function () {
        App = _application.default.create();
        _test.default.adapter = CustomAdapter.create();
        App.setupForTesting();
      });

      _test.default.adapter.asyncStart();
    }

    ['@test QUnitAdapter is used by default (if QUnit is available)'](assert) {
      assert.expect(1);

      _test.default.adapter = null;

      (0, _runloop.run)(function () {
        App = _application.default.create();
        App.setupForTesting();
      });

      assert.ok(_test.default.adapter instanceof _qunit.default);
    }

    ['@test Adapter is used by default (if QUnit is not available)'](assert) {
      assert.expect(2);

      delete window.QUnit;

      _test.default.adapter = null;

      (0, _runloop.run)(function () {
        App = _application.default.create();
        App.setupForTesting();
      });

      assert.ok(_test.default.adapter instanceof _adapter.default);
      assert.ok(!(_test.default.adapter instanceof _qunit.default));
    }

    ['@test With Ember.Test.adapter set, errors in synchronous Ember.run are bubbled out'](assert) {
      let thrown = new Error('Boom!');

      let caughtInAdapter, caughtInCatch;
      _test.default.adapter = _qunit.default.create({
        exception(error) {
          caughtInAdapter = error;
        }
      });

      try {
        (0, _runloop.run)(() => {
          throw thrown;
        });
      } catch (e) {
        caughtInCatch = e;
      }

      assert.equal(caughtInAdapter, undefined, 'test adapter should never receive synchronous errors');
      assert.equal(caughtInCatch, thrown, 'a "normal" try/catch should catch errors in sync run');
    }

    ['@test when both Ember.onerror (which rethrows) and TestAdapter are registered - sync run'](assert) {
      assert.expect(2);

      _test.default.adapter = {
        exception() {
          assert.notOk(true, 'adapter is not called for errors thrown in sync run loops');
        }
      };

      (0, _errorHandling.setOnerror)(function (error) {
        assert.ok(true, 'onerror is called for sync errors even if TestAdapter is setup');
        throw error;
      });

      assert.throws(runThatThrowsSync, Error, 'error is thrown');
    }

    ['@test when both Ember.onerror (which does not rethrow) and TestAdapter are registered - sync run'](assert) {
      assert.expect(2);

      _test.default.adapter = {
        exception() {
          assert.notOk(true, 'adapter is not called for errors thrown in sync run loops');
        }
      };

      (0, _errorHandling.setOnerror)(function () {
        assert.ok(true, 'onerror is called for sync errors even if TestAdapter is setup');
      });

      runThatThrowsSync();
      assert.ok(true, 'no error was thrown, Ember.onerror can intercept errors');
    }

    ['@test when TestAdapter is registered and error is thrown - async run'](assert) {
      assert.expect(3);
      let done = assert.async();

      let caughtInAdapter, caughtInCatch, caughtByWindowOnerror;
      _test.default.adapter = {
        exception(error) {
          caughtInAdapter = error;
        }
      };

      window.onerror = function (message) {
        caughtByWindowOnerror = message;
        // prevent "bubbling" and therefore failing the test
        return true;
      };

      try {
        runThatThrowsAsync();
      } catch (e) {
        caughtInCatch = e;
      }

      setTimeout(() => {
        assert.equal(caughtInAdapter, undefined, 'test adapter should never catch errors in run loops');
        assert.equal(caughtInCatch, undefined, 'a "normal" try/catch should never catch errors in an async run');

        assert.pushResult({
          result: /Error for testing error handling/.test(caughtByWindowOnerror),
          actual: caughtByWindowOnerror,
          expected: 'to include `Error for testing error handling`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        done();
      }, 20);
    }

    ['@test when both Ember.onerror and TestAdapter are registered - async run'](assert) {
      assert.expect(1);
      let done = assert.async();

      _test.default.adapter = {
        exception() {
          assert.notOk(true, 'Adapter.exception is not called for errors thrown in next');
        }
      };

      (0, _errorHandling.setOnerror)(function () {
        assert.ok(true, 'onerror is invoked for errors thrown in next/later');
      });

      runThatThrowsAsync();
      setTimeout(done, 10);
    }
  });

  function testAdapter(message, generatePromise, timeout = 10) {
    return class PromiseFailureTests extends AdapterSetupAndTearDown {
      [`@test ${message} when TestAdapter without \`exception\` method is present - rsvp`](assert) {
        assert.expect(1);

        let thrown = new Error('the error');
        _test.default.adapter = _qunit.default.create({
          exception: undefined
        });

        window.onerror = function (message) {
          assert.pushResult({
            result: /the error/.test(message),
            actual: message,
            expected: 'to include `the error`',
            message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
          });

          // prevent "bubbling" and therefore failing the test
          return true;
        };

        generatePromise(thrown);

        // RSVP.Promise's are configured to settle within the run loop, this
        // ensures that run loop has completed
        return new _runtime.RSVP.Promise(resolve => setTimeout(resolve, timeout));
      }

      [`@test ${message} when both Ember.onerror and TestAdapter without \`exception\` method are present - rsvp`](assert) {
        assert.expect(1);

        let thrown = new Error('the error');
        _test.default.adapter = _qunit.default.create({
          exception: undefined
        });

        (0, _errorHandling.setOnerror)(function (error) {
          assert.pushResult({
            result: /the error/.test(error.message),
            actual: error.message,
            expected: 'to include `the error`',
            message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
          });
        });

        generatePromise(thrown);

        // RSVP.Promise's are configured to settle within the run loop, this
        // ensures that run loop has completed
        return new _runtime.RSVP.Promise(resolve => setTimeout(resolve, timeout));
      }

      [`@test ${message} when TestAdapter is present - rsvp`](assert) {
        assert.expect(1);

        console.error = () => {}; // eslint-disable-line no-console
        let thrown = new Error('the error');
        _test.default.adapter = _qunit.default.create({
          exception(error) {
            assert.strictEqual(error, thrown, 'Adapter.exception is called for errors thrown in RSVP promises');
          }
        });

        generatePromise(thrown);

        // RSVP.Promise's are configured to settle within the run loop, this
        // ensures that run loop has completed
        return new _runtime.RSVP.Promise(resolve => setTimeout(resolve, timeout));
      }

      [`@test ${message} when both Ember.onerror and TestAdapter are present - rsvp`](assert) {
        assert.expect(1);

        let thrown = new Error('the error');
        _test.default.adapter = _qunit.default.create({
          exception(error) {
            assert.strictEqual(error, thrown, 'Adapter.exception is called for errors thrown in RSVP promises');
          }
        });

        (0, _errorHandling.setOnerror)(function () {
          assert.notOk(true, 'Ember.onerror is not called if Test.adapter does not rethrow');
        });

        generatePromise(thrown);

        // RSVP.Promise's are configured to settle within the run loop, this
        // ensures that run loop has completed
        return new _runtime.RSVP.Promise(resolve => setTimeout(resolve, timeout));
      }

      [`@test ${message} when both Ember.onerror and TestAdapter are present - rsvp`](assert) {
        assert.expect(2);

        let thrown = new Error('the error');
        _test.default.adapter = _qunit.default.create({
          exception(error) {
            assert.strictEqual(error, thrown, 'Adapter.exception is called for errors thrown in RSVP promises');
            throw error;
          }
        });

        (0, _errorHandling.setOnerror)(function (error) {
          assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises if Test.adapter rethrows');
        });

        generatePromise(thrown);

        // RSVP.Promise's are configured to settle within the run loop, this
        // ensures that run loop has completed
        return new _runtime.RSVP.Promise(resolve => setTimeout(resolve, timeout));
      }
    };
  }

  (0, _internalTestHelpers.moduleFor)('Adapter Errors: .then callback', testAdapter('errors in promise constructor', error => {
    new _runtime.RSVP.Promise(() => {
      throw error;
    });
  }));

  (0, _internalTestHelpers.moduleFor)('Adapter Errors: Promise Contructor', testAdapter('errors in promise constructor', error => {
    _runtime.RSVP.resolve().then(() => {
      throw error;
    });
  }));

  (0, _internalTestHelpers.moduleFor)('Adapter Errors: Promise chain .then callback', testAdapter('errors in promise constructor', error => {
    new _runtime.RSVP.Promise(resolve => setTimeout(resolve, 10)).then(() => {
      throw error;
    });
  }, 20));
});
enifed('ember-testing/tests/ext/rsvp_test', ['ember-testing/lib/ext/rsvp', 'ember-testing/lib/test/adapter', 'ember-testing/lib/test/promise', '@ember/runloop', '@ember/debug', 'internal-test-helpers'], function (_rsvp, _adapter, _promise, _runloop, _debug, _internalTestHelpers) {
  'use strict';

  const originalTestAdapter = (0, _adapter.getAdapter)();
  const originalTestingFlag = (0, _debug.isTesting)();

  let asyncStarted = 0;
  let asyncEnded = 0;

  (0, _internalTestHelpers.moduleFor)('ember-testing RSVP', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      (0, _debug.setTesting)(true);
      (0, _adapter.setAdapter)({
        asyncStart() {
          asyncStarted++;
        },
        asyncEnd() {
          asyncEnded++;
        }
      });
    }

    teardown() {
      asyncStarted = 0;
      asyncEnded = 0;
      (0, _adapter.setAdapter)(originalTestAdapter);
      (0, _debug.setTesting)(originalTestingFlag);
    }

    ['@test given `Ember.testing = true`, correctly informs the test suite about async steps'](assert) {
      let done = assert.async();
      assert.expect(19);

      assert.ok(!(0, _runloop.getCurrentRunLoop)(), 'expect no run-loop');

      (0, _debug.setTesting)(true);

      assert.equal(asyncStarted, 0);
      assert.equal(asyncEnded, 0);

      let user = _rsvp.default.Promise.resolve({ name: 'tomster' });

      assert.equal(asyncStarted, 0);
      assert.equal(asyncEnded, 0);

      user.then(function (user) {
        assert.equal(asyncStarted, 1);
        assert.equal(asyncEnded, 1);

        assert.equal(user.name, 'tomster');

        return _rsvp.default.Promise.resolve(1).then(function () {
          assert.equal(asyncStarted, 1);
          assert.equal(asyncEnded, 1);
        });
      }).then(function () {
        assert.equal(asyncStarted, 1);
        assert.equal(asyncEnded, 1);

        return new _rsvp.default.Promise(function (resolve) {
          setTimeout(function () {
            assert.equal(asyncStarted, 1);
            assert.equal(asyncEnded, 1);

            resolve({ name: 'async tomster' });

            assert.equal(asyncStarted, 2);
            assert.equal(asyncEnded, 1);
          }, 0);
        });
      }).then(function (user) {
        assert.equal(user.name, 'async tomster');
        assert.equal(asyncStarted, 2);
        assert.equal(asyncEnded, 2);
        done();
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('TestPromise', class extends _internalTestHelpers.AbstractTestCase {
    ['does not throw error when falsy value passed to then'](assert) {
      assert.expect(1);
      return new _promise.default(function (resolve) {
        resolve();
      }).then(null).then(function () {
        assert.ok(true);
      });
    }

    ['able to get last Promise'](assert) {
      assert.expect(2);

      var p1 = new _promise.default(function (resolve) {
        resolve();
      }).then(function () {
        assert.ok(true);
      });

      var p2 = new _promise.default(function (resolve) {
        resolve();
      });

      assert.deepEqual((0, _promise.getLastPromise)(), p2);
      return p1;
    }
  });
});
enifed('ember-testing/tests/helper_registration_test', ['@ember/runloop', 'ember-testing/lib/test', '@ember/application', 'internal-test-helpers'], function (_runloop, _test, _application, _internalTestHelpers) {
  'use strict';

  var App, appBooted, helperContainer;

  function registerHelper() {
    _test.default.registerHelper('boot', function (app) {
      (0, _runloop.run)(app, app.advanceReadiness);
      appBooted = true;
      return app.testHelpers.wait();
    });
  }

  function unregisterHelper() {
    _test.default.unregisterHelper('boot');
  }

  var originalAdapter = _test.default.adapter;

  function setupApp() {
    appBooted = false;
    helperContainer = {};

    (0, _runloop.run)(function () {
      App = _application.default.create();
      App.setupForTesting();
      App.injectTestHelpers(helperContainer);
    });
  }

  function destroyApp() {
    if (App) {
      (0, _runloop.run)(App, 'destroy');
      App = null;
      helperContainer = null;
    }
  }

  (0, _internalTestHelpers.moduleFor)('Test - registerHelper/unregisterHelper', class extends _internalTestHelpers.AbstractTestCase {
    teardown() {
      _test.default.adapter = originalAdapter;
      destroyApp();
    }

    ['@test Helper gets registered'](assert) {
      assert.expect(2);

      registerHelper();
      setupApp();

      assert.ok(App.testHelpers.boot);
      assert.ok(helperContainer.boot);
    }

    ['@test Helper is ran when called'](assert) {
      let done = assert.async();
      assert.expect(1);

      registerHelper();
      setupApp();

      App.testHelpers.boot().then(function () {
        assert.ok(appBooted);
      }).finally(done);
    }

    ['@test Helper can be unregistered'](assert) {
      assert.expect(4);

      registerHelper();
      setupApp();

      assert.ok(App.testHelpers.boot);
      assert.ok(helperContainer.boot);

      unregisterHelper();

      (0, _runloop.run)(App, 'destroy');
      setupApp();

      assert.ok(!App.testHelpers.boot, 'once unregistered the helper is not added to App.testHelpers');
      assert.ok(!helperContainer.boot, 'once unregistered the helper is not added to the helperContainer');
    }
  });
});
enifed('ember-testing/tests/helpers_test', ['internal-test-helpers', '@ember/-internals/routing', '@ember/controller', '@ember/-internals/runtime', '@ember/runloop', '@ember/-internals/glimmer', '@ember/-internals/views', 'ember-testing/lib/test', 'ember-testing/lib/setup_for_testing', 'ember-testing/lib/test/pending_requests', 'ember-testing/lib/test/adapter', 'ember-testing/lib/test/waiters', '@ember/debug'], function (_internalTestHelpers, _routing, _controller, _runtime, _runloop, _glimmer, _views, _test, _setup_for_testing, _pending_requests, _adapter, _waiters, _debug) {
  'use strict';

  var originalInfo = (0, _debug.getDebugFunction)('info');
  var noop = function () {};

  function registerHelper() {
    _test.default.registerHelper('LeakyMcLeakLeak', () => {});
  }

  function assertHelpers(assert, application, helperContainer, expected) {
    if (!helperContainer) {
      helperContainer = window;
    }
    if (expected === undefined) {
      expected = true;
    }

    function checkHelperPresent(helper, expected) {
      var presentInHelperContainer = !!helperContainer[helper];
      var presentInTestHelpers = !!application.testHelpers[helper];

      assert.ok(presentInHelperContainer === expected, "Expected '" + helper + "' to be present in the helper container (defaults to window).");
      assert.ok(presentInTestHelpers === expected, "Expected '" + helper + "' to be present in App.testHelpers.");
    }

    checkHelperPresent('visit', expected);
    checkHelperPresent('click', expected);
    checkHelperPresent('keyEvent', expected);
    checkHelperPresent('fillIn', expected);
    checkHelperPresent('wait', expected);
    checkHelperPresent('triggerEvent', expected);
  }

  function assertNoHelpers(assert, application, helperContainer) {
    assertHelpers(assert, application, helperContainer, false);
  }

  class HelpersTestCase extends _internalTestHelpers.AutobootApplicationTestCase {
    constructor() {
      super();
      this._originalAdapter = (0, _adapter.getAdapter)();
    }

    teardown() {
      (0, _adapter.setAdapter)(this._originalAdapter);
      document.removeEventListener('ajaxSend', _pending_requests.incrementPendingRequests);
      document.removeEventListener('ajaxComplete', _pending_requests.decrementPendingRequests);
      (0, _pending_requests.clearPendingRequests)();
      if (this.application) {
        this.application.removeTestHelpers();
      }
      super.teardown();
    }
  }

  class HelpersApplicationTestCase extends HelpersTestCase {
    constructor() {
      super();
      this.runTask(() => {
        this.createApplication();
        this.application.setupForTesting();
        this.application.injectTestHelpers();
      });
    }
  }

  if (!_views.jQueryDisabled) {
    (0, _internalTestHelpers.moduleFor)('ember-testing: Helper setup', class extends HelpersTestCase {
      [`@test Ember.Application#injectTestHelpers/#removeTestHelper`](assert) {
        this.runTask(() => {
          this.createApplication();
        });

        assertNoHelpers(assert, this.application);

        registerHelper();

        this.application.injectTestHelpers();

        assertHelpers(assert, this.application);

        assert.ok(_test.default.Promise.prototype.LeakyMcLeakLeak, 'helper in question SHOULD be present');

        this.application.removeTestHelpers();

        assertNoHelpers(assert, this.application);

        assert.equal(_test.default.Promise.prototype.LeakyMcLeakLeak, undefined, 'should NOT leak test promise extensions');
      }

      [`@test Ember.Application#setupForTesting`](assert) {
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        let routerInstance = this.applicationInstance.lookup('router:main');
        assert.equal(routerInstance.location, 'none');
      }

      [`@test Ember.Application.setupForTesting sets the application to 'testing'`](assert) {
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        assert.equal(this.application.testing, true, 'Application instance is set to testing.');
      }

      [`@test Ember.Application.setupForTesting leaves the system in a deferred state.`](assert) {
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        assert.equal(this.application._readinessDeferrals, 1, 'App is in deferred state after setupForTesting.');
      }

      [`@test App.reset() after Application.setupForTesting leaves the system in a deferred state.`](assert) {
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        assert.equal(this.application._readinessDeferrals, 1, 'App is in deferred state after setupForTesting.');

        this.application.reset();

        assert.equal(this.application._readinessDeferrals, 1, 'App is in deferred state after setupForTesting.');
      }

      [`@test Ember.Application#injectTestHelpers calls callbacks registered with onInjectHelpers`](assert) {
        let injected = 0;

        _test.default.onInjectHelpers(() => {
          injected++;
        });

        // bind(this) so Babel doesn't leak _this
        // into the context onInjectHelpers.
        this.runTask(function () {
          this.createApplication();
          this.application.setupForTesting();
        }.bind(this));

        assert.equal(injected, 0, 'onInjectHelpers are not called before injectTestHelpers');

        this.application.injectTestHelpers();

        assert.equal(injected, 1, 'onInjectHelpers are called after injectTestHelpers');
      }

      [`@test Ember.Application#injectTestHelpers adds helpers to provided object.`](assert) {
        let helpers = {};

        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        this.application.injectTestHelpers(helpers);

        assertHelpers(assert, this.application, helpers);

        this.application.removeTestHelpers();

        assertNoHelpers(assert, this.application, helpers);
      }

      [`@test Ember.Application#removeTestHelpers resets the helperContainer\'s original values`](assert) {
        let helpers = { visit: 'snazzleflabber' };

        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });

        this.application.injectTestHelpers(helpers);

        assert.notEqual(helpers.visit, 'snazzleflabber', 'helper added to container');
        this.application.removeTestHelpers();

        assert.equal(helpers.visit, 'snazzleflabber', 'original value added back to container');
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: Helper methods', class extends HelpersApplicationTestCase {
      [`@test 'wait' respects registerWaiters`](assert) {
        assert.expect(3);

        let counter = 0;
        function waiter() {
          return ++counter > 2;
        }

        let other = 0;
        function otherWaiter() {
          return ++other > 2;
        }

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        (0, _waiters.registerWaiter)(waiter);
        (0, _waiters.registerWaiter)(otherWaiter);

        let { application: { testHelpers } } = this;
        return testHelpers.wait().then(() => {
          assert.equal(waiter(), true, 'should not resolve until our waiter is ready');
          (0, _waiters.unregisterWaiter)(waiter);
          counter = 0;
          return testHelpers.wait();
        }).then(() => {
          assert.equal(counter, 0, 'unregistered waiter was not checked');
          assert.equal(otherWaiter(), true, 'other waiter is still registered');
        }).finally(() => {
          (0, _waiters.unregisterWaiter)(otherWaiter);
        });
      }

      [`@test 'visit' advances readiness.`](assert) {
        assert.expect(2);

        assert.equal(this.application._readinessDeferrals, 1, 'App is in deferred state after setupForTesting.');

        return this.application.testHelpers.visit('/').then(() => {
          assert.equal(this.application._readinessDeferrals, 0, `App's readiness was advanced by visit.`);
        });
      }

      [`@test 'wait' helper can be passed a resolution value`](assert) {
        assert.expect(4);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let promiseObjectValue = {};
        let objectValue = {};
        let { application: { testHelpers } } = this;
        return testHelpers.wait('text').then(val => {
          assert.equal(val, 'text', 'can resolve to a string');
          return testHelpers.wait(1);
        }).then(val => {
          assert.equal(val, 1, 'can resolve to an integer');
          return testHelpers.wait(objectValue);
        }).then(val => {
          assert.equal(val, objectValue, 'can resolve to an object');
          return testHelpers.wait(_runtime.RSVP.resolve(promiseObjectValue));
        }).then(val => {
          assert.equal(val, promiseObjectValue, 'can resolve to a promise resolution value');
        });
      }

      [`@test 'click' triggers appropriate events in order`](assert) {
        assert.expect(5);

        this.add('component:index-wrapper', _glimmer.Component.extend({
          classNames: 'index-wrapper',

          didInsertElement() {
            let wrapper = document.querySelector('.index-wrapper');
            wrapper.addEventListener('mousedown', e => events.push(e.type));
            wrapper.addEventListener('mouseup', e => events.push(e.type));
            wrapper.addEventListener('click', e => events.push(e.type));
            wrapper.addEventListener('focusin', e => {
              // IE11 _sometimes_ triggers focusin **twice** in a row
              // (we believe this is when it is under higher load)
              //
              // the goal here is to only push a single focusin when running on
              // IE11
              if (_internalTestHelpers.isIE11) {
                if (events[events.length - 1] !== 'focusin') {
                  events.push(e.type);
                }
              } else {
                events.push(e.type);
              }
            });
          }
        }));

        this.add('component:x-checkbox', _glimmer.Component.extend({
          tagName: 'input',
          attributeBindings: ['type'],
          type: 'checkbox',
          click() {
            events.push('click:' + this.get('checked'));
          },
          change() {
            events.push('change:' + this.get('checked'));
          }
        }));

        this.addTemplate('index', `
        {{#index-wrapper}}
          {{input type="text"}}
          {{x-checkbox type="checkbox"}}
          {{textarea}}
          <div contenteditable="true"> </div>
        {{/index-wrapper}}'));
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let events;
        let { application: { testHelpers } } = this;
        return testHelpers.wait().then(() => {
          events = [];
          return testHelpers.click('.index-wrapper');
        }).then(() => {
          assert.deepEqual(events, ['mousedown', 'mouseup', 'click'], 'fires events in order');
        }).then(() => {
          events = [];
          return testHelpers.click('.index-wrapper input[type=text]');
        }).then(() => {
          assert.deepEqual(events, ['mousedown', 'focusin', 'mouseup', 'click'], 'fires focus events on inputs');
        }).then(() => {
          events = [];
          return testHelpers.click('.index-wrapper textarea');
        }).then(() => {
          assert.deepEqual(events, ['mousedown', 'focusin', 'mouseup', 'click'], 'fires focus events on textareas');
        }).then(() => {
          events = [];
          return testHelpers.click('.index-wrapper div');
        }).then(() => {
          assert.deepEqual(events, ['mousedown', 'focusin', 'mouseup', 'click'], 'fires focus events on contenteditable');
        }).then(() => {
          events = [];
          return testHelpers.click('.index-wrapper input[type=checkbox]');
        }).then(() => {
          // i.e. mousedown, mouseup, change:true, click, click:true
          // Firefox differs so we can't assert the exact ordering here.
          // See https://bugzilla.mozilla.org/show_bug.cgi?id=843554.
          assert.equal(events.length, 5, 'fires click and change on checkboxes');
        });
      }

      [`@test 'click' triggers native events with simulated X/Y coordinates`](assert) {
        assert.expect(15);

        this.add('component:index-wrapper', _glimmer.Component.extend({
          classNames: 'index-wrapper',

          didInsertElement() {
            let pushEvent = e => events.push(e);
            this.element.addEventListener('mousedown', pushEvent);
            this.element.addEventListener('mouseup', pushEvent);
            this.element.addEventListener('click', pushEvent);
          }
        }));

        this.addTemplate('index', `
        {{#index-wrapper}}some text{{/index-wrapper}}
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let events;
        let { application: { testHelpers: { wait, click } } } = this;
        return wait().then(() => {
          events = [];
          return click('.index-wrapper');
        }).then(() => {
          events.forEach(e => {
            assert.ok(e instanceof window.Event, 'The event is an instance of MouseEvent');
            assert.ok(typeof e.screenX === 'number', 'screenX is correct');
            assert.ok(typeof e.screenY === 'number', 'screenY is correct');
            assert.ok(typeof e.clientX === 'number', 'clientX is correct');
            assert.ok(typeof e.clientY === 'number', 'clientY is correct');
          });
        });
      }

      [`@test 'triggerEvent' with mouseenter triggers native events with simulated X/Y coordinates`](assert) {
        assert.expect(5);

        let evt;
        this.add('component:index-wrapper', _glimmer.Component.extend({
          classNames: 'index-wrapper',
          didInsertElement() {
            this.element.addEventListener('mouseenter', e => evt = e);
          }
        }));

        this.addTemplate('index', `{{#index-wrapper}}some text{{/index-wrapper}}`);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { wait, triggerEvent } } } = this;
        return wait().then(() => {
          return triggerEvent('.index-wrapper', 'mouseenter');
        }).then(() => {
          assert.ok(evt instanceof window.Event, 'The event is an instance of MouseEvent');
          assert.ok(typeof evt.screenX === 'number', 'screenX is correct');
          assert.ok(typeof evt.screenY === 'number', 'screenY is correct');
          assert.ok(typeof evt.clientX === 'number', 'clientX is correct');
          assert.ok(typeof evt.clientY === 'number', 'clientY is correct');
        });
      }

      [`@test 'wait' waits for outstanding timers`](assert) {
        assert.expect(1);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let waitDone = false;
        (0, _runloop.later)(() => {
          waitDone = true;
        }, 20);

        return this.application.testHelpers.wait().then(() => {
          assert.equal(waitDone, true, 'should wait for the timer to be fired.');
        });
      }

      [`@test 'wait' respects registerWaiters with optional context`](assert) {
        assert.expect(3);

        let obj = {
          counter: 0,
          ready() {
            return ++this.counter > 2;
          }
        };

        let other = 0;
        function otherWaiter() {
          return ++other > 2;
        }

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        (0, _waiters.registerWaiter)(obj, obj.ready);
        (0, _waiters.registerWaiter)(otherWaiter);

        let { application: { testHelpers: { wait } } } = this;
        return wait().then(() => {
          assert.equal(obj.ready(), true, 'should not resolve until our waiter is ready');
          (0, _waiters.unregisterWaiter)(obj, obj.ready);
          obj.counter = 0;
          return wait();
        }).then(() => {
          assert.equal(obj.counter, 0, 'the unregistered waiter should still be at 0');
          assert.equal(otherWaiter(), true, 'other waiter should still be registered');
        }).finally(() => {
          (0, _waiters.unregisterWaiter)(otherWaiter);
        });
      }

      [`@test 'wait' does not error if routing has not begun`](assert) {
        assert.expect(1);

        return this.application.testHelpers.wait().then(() => {
          assert.ok(true, 'should not error without `visit`');
        });
      }

      [`@test 'triggerEvent' accepts an optional options hash without context`](assert) {
        assert.expect(3);

        let event;
        this.add('component:index-wrapper', _glimmer.Component.extend({
          didInsertElement() {
            let domElem = document.querySelector('.input');
            domElem.addEventListener('change', e => event = e);
            domElem.addEventListener('keydown', e => event = e);
          }
        }));

        this.addTemplate('index', `{{index-wrapper}}`);
        this.addTemplate('components/index-wrapper', `
        {{input type="text" id="scope" class="input"}}
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { wait, triggerEvent } } } = this;
        return wait().then(() => {
          return triggerEvent('.input', 'keydown', { keyCode: 13 });
        }).then(() => {
          assert.equal(event.keyCode, 13, 'options were passed');
          assert.equal(event.type, 'keydown', 'correct event was triggered');
          assert.equal(event.target.getAttribute('id'), 'scope', 'triggered on the correct element');
        });
      }

      [`@test 'triggerEvent' can limit searching for a selector to a scope`](assert) {
        assert.expect(2);

        let event;
        this.add('component:index-wrapper', _glimmer.Component.extend({
          didInsertElement() {
            let firstInput = document.querySelector('.input');
            firstInput.addEventListener('blur', e => event = e);
            firstInput.addEventListener('change', e => event = e);
            let secondInput = document.querySelector('#limited .input');
            secondInput.addEventListener('blur', e => event = e);
            secondInput.addEventListener('change', e => event = e);
          }
        }));

        this.addTemplate('components/index-wrapper', `
        {{input type="text" id="outside-scope" class="input"}}
        <div id="limited">
          {{input type="text" id="inside-scope" class="input"}}
        </div>
      `);
        this.addTemplate('index', `{{index-wrapper}}`);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { wait, triggerEvent } } } = this;
        return wait().then(() => {
          return triggerEvent('.input', '#limited', 'blur');
        }).then(() => {
          assert.equal(event.type, 'blur', 'correct event was triggered');
          assert.equal(event.target.getAttribute('id'), 'inside-scope', 'triggered on the correct element');
        });
      }

      [`@test 'triggerEvent' can be used to trigger arbitrary events`](assert) {
        assert.expect(2);

        let event;
        this.add('component:index-wrapper', _glimmer.Component.extend({
          didInsertElement() {
            let foo = document.getElementById('foo');
            foo.addEventListener('blur', e => event = e);
            foo.addEventListener('change', e => event = e);
          }
        }));

        this.addTemplate('components/index-wrapper', `
        {{input type="text" id="foo"}}
      `);
        this.addTemplate('index', `{{index-wrapper}}`);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { wait, triggerEvent } } } = this;
        return wait().then(() => {
          return triggerEvent('#foo', 'blur');
        }).then(() => {
          assert.equal(event.type, 'blur', 'correct event was triggered');
          assert.equal(event.target.getAttribute('id'), 'foo', 'triggered on the correct element');
        });
      }

      [`@test 'fillIn' takes context into consideration`](assert) {
        assert.expect(2);

        this.addTemplate('index', `
        <div id="parent">
          {{input type="text" id="first" class="current"}}
        </div>
        {{input type="text" id="second" class="current"}}
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { visit, fillIn, andThen, find } } } = this;
        visit('/');
        fillIn('.current', '#parent', 'current value');

        return andThen(() => {
          assert.equal(find('#first')[0].value, 'current value');
          assert.equal(find('#second')[0].value, '');
        });
      }

      [`@test 'fillIn' focuses on the element`](assert) {
        let wasFocused = false;

        this.add('controller:index', _controller.default.extend({
          actions: {
            wasFocused() {
              wasFocused = true;
            }
          }
        }));

        this.addTemplate('index', `
        <div id="parent">
          {{input type="text" id="first" focus-in=(action "wasFocused")}}
        </div>'
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { visit, fillIn, andThen, find, wait } } } = this;
        visit('/');
        fillIn('#first', 'current value');
        andThen(() => {
          assert.ok(wasFocused, 'focusIn event was triggered');

          assert.equal(find('#first')[0].value, 'current value');
        });

        return wait();
      }

      [`@test 'fillIn' fires 'input' and 'change' events in the proper order`](assert) {
        assert.expect(1);

        let events = [];
        this.add('controller:index', _controller.default.extend({
          actions: {
            oninputHandler(e) {
              events.push(e.type);
            },
            onchangeHandler(e) {
              events.push(e.type);
            }
          }
        }));

        this.addTemplate('index', `
        <input type="text" id="first"
            oninput={{action "oninputHandler"}}
            onchange={{action "onchangeHandler"}}>
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { visit, fillIn, andThen, wait } } } = this;

        visit('/');
        fillIn('#first', 'current value');
        andThen(() => {
          assert.deepEqual(events, ['input', 'change'], '`input` and `change` events are fired in the proper order');
        });

        return wait();
      }

      [`@test 'fillIn' only sets the value in the first matched element`](assert) {
        this.addTemplate('index', `
        <input type="text" id="first" class="in-test">
        <input type="text" id="second" class="in-test">
      `);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { visit, fillIn, find, andThen, wait } } } = this;

        visit('/');
        fillIn('input.in-test', 'new value');
        andThen(() => {
          assert.equal(find('#first')[0].value, 'new value');
          assert.equal(find('#second')[0].value, '');
        });

        return wait();
      }

      [`@test 'triggerEvent' accepts an optional options hash and context`](assert) {
        assert.expect(3);

        let event;
        this.add('component:index-wrapper', _glimmer.Component.extend({
          didInsertElement() {
            let firstInput = document.querySelector('.input');
            firstInput.addEventListener('keydown', e => event = e, false);
            firstInput.addEventListener('change', e => event = e, false);
            let secondInput = document.querySelector('#limited .input');
            secondInput.addEventListener('keydown', e => event = e, false);
            secondInput.addEventListener('change', e => event = e, false);
          }
        }));

        this.addTemplate('components/index-wrapper', `
        {{input type="text" id="outside-scope" class="input"}}
        <div id="limited">
          {{input type="text" id="inside-scope" class="input"}}
        </div>
      `);
        this.addTemplate('index', `{{index-wrapper}}`);

        this.runTask(() => {
          this.application.advanceReadiness();
        });

        let { application: { testHelpers: { wait, triggerEvent } } } = this;
        return wait().then(() => {
          return triggerEvent('.input', '#limited', 'keydown', {
            keyCode: 13
          });
        }).then(() => {
          assert.equal(event.keyCode, 13, 'options were passed');
          assert.equal(event.type, 'keydown', 'correct event was triggered');
          assert.equal(event.target.getAttribute('id'), 'inside-scope', 'triggered on the correct element');
        });
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: debugging helpers', class extends HelpersApplicationTestCase {
      afterEach() {
        super.afterEach();
        (0, _debug.setDebugFunction)('info', originalInfo);
      }

      constructor() {
        super();
        this.runTask(() => {
          this.application.advanceReadiness();
        });
      }

      [`@test pauseTest pauses`](assert) {
        assert.expect(1);
        // overwrite info to supress the console output (see https://github.com/emberjs/ember.js/issues/16391)
        (0, _debug.setDebugFunction)('info', noop);

        let { andThen, pauseTest } = this.application.testHelpers;

        andThen(() => {
          _test.default.adapter.asyncStart = () => {
            assert.ok(true, 'Async start should be called after waiting for other helpers');
          };
        });

        pauseTest();
      }

      [`@test resumeTest resumes paused tests`](assert) {
        assert.expect(1);
        // overwrite info to supress the console output (see https://github.com/emberjs/ember.js/issues/16391)
        (0, _debug.setDebugFunction)('info', noop);

        let { application: { testHelpers: { pauseTest, resumeTest } } } = this;

        (0, _runloop.later)(() => resumeTest(), 20);
        return pauseTest().then(() => {
          assert.ok(true, 'pauseTest promise was resolved');
        });
      }

      [`@test resumeTest throws if nothing to resume`](assert) {
        assert.expect(1);

        assert.throws(() => {
          this.application.testHelpers.resumeTest();
        }, /Testing has not been paused. There is nothing to resume./);
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: routing helpers', class extends HelpersTestCase {
      constructor() {
        super();
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
          this.application.injectTestHelpers();
          this.router.map(function () {
            this.route('posts', { resetNamespace: true }, function () {
              this.route('new');
              this.route('edit', { resetNamespace: true });
            });
          });
        });
        this.runTask(() => {
          this.application.advanceReadiness();
        });
      }

      [`@test currentRouteName for '/'`](assert) {
        assert.expect(3);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'index', `should equal 'index'.`);
          assert.equal(testHelpers.currentPath(), 'index', `should equal 'index'.`);
          assert.equal(testHelpers.currentURL(), '/', `should equal '/'.`);
        });
      }

      [`@test currentRouteName for '/posts'`](assert) {
        assert.expect(3);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/posts').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'posts.index', `should equal 'posts.index'.`);
          assert.equal(testHelpers.currentPath(), 'posts.index', `should equal 'posts.index'.`);
          assert.equal(testHelpers.currentURL(), '/posts', `should equal '/posts'.`);
        });
      }

      [`@test currentRouteName for '/posts/new'`](assert) {
        assert.expect(3);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/posts/new').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'posts.new', `should equal 'posts.new'.`);
          assert.equal(testHelpers.currentPath(), 'posts.new', `should equal 'posts.new'.`);
          assert.equal(testHelpers.currentURL(), '/posts/new', `should equal '/posts/new'.`);
        });
      }

      [`@test currentRouteName for '/posts/edit'`](assert) {
        assert.expect(3);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/posts/edit').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'edit', `should equal 'edit'.`);
          assert.equal(testHelpers.currentPath(), 'posts.edit', `should equal 'posts.edit'.`);
          assert.equal(testHelpers.currentURL(), '/posts/edit', `should equal '/posts/edit'.`);
        });
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: pendingRequests', class extends HelpersApplicationTestCase {
      trigger(type, xhr) {
        (0, _views.jQuery)(document).trigger(type, xhr);
      }

      [`@test pendingRequests is maintained for ajaxSend and ajaxComplete events`](assert) {
        let done = assert.async();
        assert.equal((0, _pending_requests.pendingRequests)(), 0);

        let xhr = { some: 'xhr' };

        this.trigger('ajaxSend', xhr);
        assert.equal((0, _pending_requests.pendingRequests)(), 1, 'Ember.Test.pendingRequests was incremented');

        this.trigger('ajaxComplete', xhr);
        setTimeout(function () {
          assert.equal((0, _pending_requests.pendingRequests)(), 0, 'Ember.Test.pendingRequests was decremented');
          done();
        }, 0);
      }

      [`@test pendingRequests is ignores ajaxComplete events from past setupForTesting calls`](assert) {
        assert.equal((0, _pending_requests.pendingRequests)(), 0);

        let xhr = { some: 'xhr' };

        this.trigger('ajaxSend', xhr);
        assert.equal((0, _pending_requests.pendingRequests)(), 1, 'Ember.Test.pendingRequests was incremented');

        (0, _setup_for_testing.default)();

        assert.equal((0, _pending_requests.pendingRequests)(), 0, 'Ember.Test.pendingRequests was reset');

        let altXhr = { some: 'more xhr' };

        this.trigger('ajaxSend', altXhr);
        assert.equal((0, _pending_requests.pendingRequests)(), 1, 'Ember.Test.pendingRequests was incremented');

        this.trigger('ajaxComplete', xhr);
        assert.equal((0, _pending_requests.pendingRequests)(), 1, 'Ember.Test.pendingRequests is not impressed with your unexpected complete');
      }

      [`@test pendingRequests is reset by setupForTesting`](assert) {
        (0, _pending_requests.incrementPendingRequests)();

        (0, _setup_for_testing.default)();

        assert.equal((0, _pending_requests.pendingRequests)(), 0, 'pendingRequests is reset');
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: async router', class extends HelpersTestCase {
      constructor() {
        super();

        this.runTask(() => {
          this.createApplication();

          this.router.map(function () {
            this.route('user', { resetNamespace: true }, function () {
              this.route('profile');
              this.route('edit');
            });
          });

          // Emulate a long-running unscheduled async operation.
          let resolveLater = () => new _runtime.RSVP.Promise(resolve => {
            /*
            * The wait() helper has a 10ms tick. We should resolve() after
            * at least one tick to test whether wait() held off while the
            * async router was still loading. 20ms should be enough.
            */
            (0, _runloop.later)(resolve, { firstName: 'Tom' }, 20);
          });

          this.add('route:user', _routing.Route.extend({
            model() {
              return resolveLater();
            }
          }));

          this.add('route:user.profile', _routing.Route.extend({
            beforeModel() {
              return resolveLater().then(() => this.transitionTo('user.edit'));
            }
          }));

          this.application.setupForTesting();
        });

        this.application.injectTestHelpers();
        this.runTask(() => {
          this.application.advanceReadiness();
        });
      }

      [`@test currentRouteName for '/user'`](assert) {
        assert.expect(4);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/user').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'user.index', `should equal 'user.index'.`);
          assert.equal(testHelpers.currentPath(), 'user.index', `should equal 'user.index'.`);
          assert.equal(testHelpers.currentURL(), '/user', `should equal '/user'.`);
          let userRoute = this.applicationInstance.lookup('route:user');
          assert.equal(userRoute.get('controller.model.firstName'), 'Tom', `should equal 'Tom'.`);
        });
      }

      [`@test currentRouteName for '/user/profile'`](assert) {
        assert.expect(4);

        let { application: { testHelpers } } = this;
        return testHelpers.visit('/user/profile').then(() => {
          assert.equal(testHelpers.currentRouteName(), 'user.edit', `should equal 'user.edit'.`);
          assert.equal(testHelpers.currentPath(), 'user.edit', `should equal 'user.edit'.`);
          assert.equal(testHelpers.currentURL(), '/user/edit', `should equal '/user/edit'.`);
          let userRoute = this.applicationInstance.lookup('route:user');
          assert.equal(userRoute.get('controller.model.firstName'), 'Tom', `should equal 'Tom'.`);
        });
      }
    });

    (0, _internalTestHelpers.moduleFor)('ember-testing: can override built-in helpers', class extends HelpersTestCase {
      constructor() {
        super();
        this.runTask(() => {
          this.createApplication();
          this.application.setupForTesting();
        });
        this._originalVisitHelper = _test.default._helpers.visit;
        this._originalFindHelper = _test.default._helpers.find;
      }

      teardown() {
        _test.default._helpers.visit = this._originalVisitHelper;
        _test.default._helpers.find = this._originalFindHelper;
        super.teardown();
      }

      [`@test can override visit helper`](assert) {
        assert.expect(1);

        _test.default.registerHelper('visit', () => {
          assert.ok(true, 'custom visit helper was called');
        });

        this.application.injectTestHelpers();

        return this.application.testHelpers.visit();
      }

      [`@test can override find helper`](assert) {
        assert.expect(1);

        _test.default.registerHelper('find', () => {
          assert.ok(true, 'custom find helper was called');

          return ['not empty array'];
        });

        this.application.injectTestHelpers();

        return this.application.testHelpers.findWithAssert('.who-cares');
      }
    });
  }
});
enifed('ember-testing/tests/integration_test', ['internal-test-helpers', 'ember-testing/lib/test', '@ember/-internals/runtime', '@ember/-internals/routing', '@ember/-internals/views'], function (_internalTestHelpers, _test, _runtime, _routing, _views) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember-testing Integration tests of acceptance', class extends _internalTestHelpers.AutobootApplicationTestCase {
    constructor() {
      super();

      this.modelContent = [];
      this._originalAdapter = _test.default.adapter;

      this.runTask(() => {
        this.createApplication();

        this.addTemplate('people', `
        <div>
          {{#each model as |person|}}
            <div class="name">{{person.firstName}}</div>
          {{/each}}
        </div>
      `);

        this.router.map(function () {
          this.route('people', { path: '/' });
        });

        this.add('route:people', _routing.Route.extend({
          model: () => this.modelContent
        }));

        this.application.setupForTesting();
      });

      this.runTask(() => {
        this.application.reset();
      });

      this.application.injectTestHelpers();
    }

    teardown() {
      super.teardown();
      _test.default.adapter = this._originalAdapter;
    }

    [`@test template is bound to empty array of people`](assert) {
      if (!_views.jQueryDisabled) {
        this.runTask(() => this.application.advanceReadiness());
        window.visit('/').then(() => {
          let rows = window.find('.name').length;
          assert.equal(rows, 0, 'successfully stubbed an empty array of people');
        });
      } else {
        this.runTask(() => this.application.advanceReadiness());
        window.visit('/').then(() => {
          expectAssertion(() => window.find('.name'), 'If jQuery is disabled, please import and use helpers from @ember/test-helpers [https://github.com/emberjs/ember-test-helpers]. Note: `find` is not an available helper.');
        });
      }
    }

    [`@test template is bound to array of 2 people`](assert) {
      if (!_views.jQueryDisabled) {
        this.modelContent = (0, _runtime.A)([]);
        this.modelContent.pushObject({ firstName: 'x' });
        this.modelContent.pushObject({ firstName: 'y' });

        this.runTask(() => this.application.advanceReadiness());
        window.visit('/').then(() => {
          let rows = window.find('.name').length;
          assert.equal(rows, 2, 'successfully stubbed a non empty array of people');
        });
      } else {
        assert.expect(0);
      }
    }

    [`@test 'visit' can be called without advanceReadiness.`](assert) {
      if (!_views.jQueryDisabled) {
        window.visit('/').then(() => {
          let rows = window.find('.name').length;
          assert.equal(rows, 0, 'stubbed an empty array of people without calling advanceReadiness.');
        });
      } else {
        assert.expect(0);
      }
    }
  });
});
enifed('ember-testing/tests/reexports_test', ['ember', 'internal-test-helpers'], function (_ember, _internalTestHelpers) {
  'use strict';

  class ReexportsTestCase extends _internalTestHelpers.AbstractTestCase {}

  [
  // ember-testing
  ['Test', 'ember-testing'], ['Test.Adapter', 'ember-testing', 'Adapter'], ['Test.QUnitAdapter', 'ember-testing', 'QUnitAdapter'], ['setupForTesting', 'ember-testing']].forEach(reexport => {
    let [path, moduleId, exportName] = reexport;

    // default path === exportName if none present
    if (!exportName) {
      exportName = path;
    }

    ReexportsTestCase.prototype[`@test Ember.${path} exports correctly`] = function (assert) {
      (0, _internalTestHelpers.confirmExport)(_ember.default, assert, path, moduleId, exportName);
    };
  });

  (0, _internalTestHelpers.moduleFor)('ember-testing reexports', ReexportsTestCase);
});
enifed('ember-testing/tests/test/waiters-test', ['ember-testing/lib/test/waiters', 'internal-test-helpers'], function (_waiters, _internalTestHelpers) {
  'use strict';

  class Waiters {
    constructor() {
      this._waiters = [];
    }

    add() {
      this._waiters.push([...arguments]);
    }

    register() {
      this.forEach((...args) => {
        (0, _waiters.registerWaiter)(...args);
      });
    }

    unregister() {
      this.forEach((...args) => {
        (0, _waiters.unregisterWaiter)(...args);
      });
    }

    forEach(callback) {
      for (let i = 0; i < this._waiters.length; i++) {
        let args = this._waiters[i];

        callback(...args);
      }
    }

    check() {
      this.register();
      let result = (0, _waiters.checkWaiters)();
      this.unregister();

      return result;
    }
  }

  (0, _internalTestHelpers.moduleFor)('ember-testing: waiters', class extends _internalTestHelpers.AbstractTestCase {
    constructor() {
      super();
      this.waiters = new Waiters();
    }

    teardown() {
      this.waiters.unregister();
    }

    ['@test registering a waiter'](assert) {
      assert.expect(2);

      let obj = { foo: true };

      this.waiters.add(obj, function () {
        assert.ok(this.foo, 'has proper `this` context');
        return true;
      });

      this.waiters.add(function () {
        assert.ok(true, 'is called');
        return true;
      });

      this.waiters.check();
    }

    ['@test unregistering a waiter'](assert) {
      assert.expect(2);

      let obj = { foo: true };

      this.waiters.add(obj, function () {
        assert.ok(true, 'precond - waiter with context is registered');
        return true;
      });

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter without context is registered');
        return true;
      });

      this.waiters.check();
      this.waiters.unregister();

      (0, _waiters.checkWaiters)();
    }

    ['@test checkWaiters returns false if all waiters return true'](assert) {
      assert.expect(3);

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter is registered');

        return true;
      });

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter is registered');

        return true;
      });

      assert.notOk(this.waiters.check(), 'checkWaiters returns true if all waiters return true');
    }

    ['@test checkWaiters returns true if any waiters return false'](assert) {
      assert.expect(3);

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter is registered');

        return true;
      });

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter is registered');

        return false;
      });

      assert.ok(this.waiters.check(), 'checkWaiters returns false if any waiters return false');
    }

    ['@test checkWaiters short circuits after first falsey waiter'](assert) {
      assert.expect(2);

      this.waiters.add(function () {
        assert.ok(true, 'precond - waiter is registered');

        return false;
      });

      this.waiters.add(function () {
        assert.notOk(true, 'waiter should not be called');
      });

      assert.ok(this.waiters.check(), 'checkWaiters returns false if any waiters return false');
    }
  });
});
enifed('ember/tests/application_lifecycle_test', ['internal-test-helpers', '@ember/application', '@ember/-internals/routing', '@ember/-internals/glimmer', '@ember/debug'], function (_internalTestHelpers, _application, _routing, _glimmer, _debug) {
  'use strict';

  const originalDebug = (0, _debug.getDebugFunction)('debug');
  const noop = function () {};

  (0, _internalTestHelpers.moduleFor)('Application Lifecycle - route hooks', class extends _internalTestHelpers.AutobootApplicationTestCase {
    createApplication() {
      let application = super.createApplication(...arguments);
      this.add('router:main', _routing.Router.extend({
        location: 'none'
      }));
      return application;
    }

    constructor() {
      (0, _debug.setDebugFunction)('debug', noop);
      super();
      let menuItem = this.menuItem = {};

      this.runTask(() => {
        this.createApplication();

        let SettingRoute = _routing.Route.extend({
          setupController() {
            this.controller.set('selectedMenuItem', menuItem);
          },
          deactivate() {
            this.controller.set('selectedMenuItem', null);
          }
        });
        this.add('route:index', SettingRoute);
        this.add('route:application', SettingRoute);
      });
    }

    teardown() {
      (0, _debug.setDebugFunction)('debug', originalDebug);
    }

    get indexController() {
      return this.applicationInstance.lookup('controller:index');
    }

    get applicationController() {
      return this.applicationInstance.lookup('controller:application');
    }

    [`@test Resetting the application allows controller properties to be set when a route deactivates`](assert) {
      let { indexController, applicationController } = this;
      assert.equal(indexController.get('selectedMenuItem'), this.menuItem);
      assert.equal(applicationController.get('selectedMenuItem'), this.menuItem);

      this.application.reset();

      assert.equal(indexController.get('selectedMenuItem'), null);
      assert.equal(applicationController.get('selectedMenuItem'), null);
    }

    [`@test Destroying the application resets the router before the appInstance is destroyed`](assert) {
      let { indexController, applicationController } = this;
      assert.equal(indexController.get('selectedMenuItem'), this.menuItem);
      assert.equal(applicationController.get('selectedMenuItem'), this.menuItem);

      this.runTask(() => {
        this.application.destroy();
      });

      assert.equal(indexController.get('selectedMenuItem'), null);
      assert.equal(applicationController.get('selectedMenuItem'), null);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Application Lifecycle', class extends _internalTestHelpers.AutobootApplicationTestCase {
    createApplication() {
      let application = super.createApplication(...arguments);
      this.add('router:main', _routing.Router.extend({
        location: 'none'
      }));
      return application;
    }

    [`@test Destroying a route after the router does create an undestroyed 'toplevelView'`](assert) {
      this.runTask(() => {
        this.createApplication();
        this.addTemplate('index', `Index!`);
        this.addTemplate('application', `Application! {{outlet}}`);
      });

      let router = this.applicationInstance.lookup('router:main');
      let route = this.applicationInstance.lookup('route:index');

      this.runTask(() => router.destroy());
      assert.equal(router._toplevelView, null, 'the toplevelView was cleared');

      this.runTask(() => route.destroy());
      assert.equal(router._toplevelView, null, 'the toplevelView was not reinitialized');

      this.runTask(() => this.application.destroy());
      assert.equal(router._toplevelView, null, 'the toplevelView was not reinitialized');
    }

    [`@test initializers can augment an applications customEvents hash`](assert) {
      assert.expect(1);

      let MyApplication = _application.default.extend();

      MyApplication.initializer({
        name: 'customize-things',
        initialize(application) {
          application.customEvents = {
            wowza: 'wowza'
          };
        }
      });

      this.runTask(() => {
        this.createApplication({}, MyApplication);

        this.add('component:foo-bar', _glimmer.Component.extend({
          wowza() {
            assert.ok(true, 'fired the event!');
          }
        }));

        this.addTemplate('application', `{{foo-bar}}`);
        this.addTemplate('components/foo-bar', `<div id='wowza-thingy'></div>`);
      });

      this.$('#wowza-thingy').trigger('wowza');
    }

    [`@test instanceInitializers can augment an the customEvents hash`](assert) {
      assert.expect(1);

      let MyApplication = _application.default.extend();

      MyApplication.instanceInitializer({
        name: 'customize-things',
        initialize(application) {
          application.customEvents = {
            herky: 'jerky'
          };
        }
      });
      this.runTask(() => {
        this.createApplication({}, MyApplication);

        this.add('component:foo-bar', _glimmer.Component.extend({
          jerky() {
            assert.ok(true, 'fired the event!');
          }
        }));

        this.addTemplate('application', `{{foo-bar}}`);
        this.addTemplate('components/foo-bar', `<div id='herky-thingy'></div>`);
      });

      this.$('#herky-thingy').trigger('herky');
    }
  });
});
enifed('ember/tests/component_context_test', ['@ember/controller', '@ember/-internals/glimmer', 'internal-test-helpers'], function (_controller, _glimmer, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application Lifecycle - Component Context', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Components with a block should have the proper content when a template is provided'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>
        {{#my-component}}{{text}}{{/my-component}}
      </div>
    `);

      this.add('controller:application', _controller.default.extend({
        text: 'outer'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          text: 'inner'
        }),
        template: `{{text}}-{{yield}}`
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'inner-outer', 'The component is composed correctly');
      });
    }

    ['@test Components with a block should yield the proper content without a template provided'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>
        {{#my-component}}{{text}}{{/my-component}}
      </div>
    `);

      this.add('controller:application', _controller.default.extend({
        text: 'outer'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          text: 'inner'
        })
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'outer', 'The component is composed correctly');
      });
    }
    ['@test Components without a block should have the proper content when a template is provided'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>{{my-component}}</div>
    `);

      this.add('controller:application', _controller.default.extend({
        text: 'outer'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          text: 'inner'
        }),
        template: '{{text}}'
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'inner', 'The component is composed correctly');
      });
    }

    ['@test Components without a block should have the proper content'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>{{my-component}}</div>
    `);

      this.add('controller:application', _controller.default.extend({
        text: 'outer'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          didInsertElement() {
            this.element.innerHTML = 'Some text inserted';
          }
        })
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'Some text inserted', 'The component is composed correctly');
      });
    }

    ['@test properties of a component without a template should not collide with internal structures [DEPRECATED]'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>{{my-component data=foo}}</div>`);

      this.add('controller:application', _controller.default.extend({
        text: 'outer',
        foo: 'Some text inserted'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          didInsertElement() {
            this.element.innerHTML = this.get('data');
          }
        })
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'Some text inserted', 'The component is composed correctly');
      });
    }

    ['@test attrs property of a component without a template should not collide with internal structures'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>{{my-component attrs=foo}}</div>
    `);

      this.add('controller:application', _controller.default.extend({
        text: 'outer',
        foo: 'Some text inserted'
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          didInsertElement() {
            this.element.innerHTML = this.get('attrs.attrs.value');
          }
        })
      });

      return this.visit('/').then(() => {
        let text = (0, _internalTestHelpers.getTextOf)(this.element.querySelector('#wrapper'));
        assert.equal(text, 'Some text inserted', 'The component is composed correctly');
      });
    }

    ['@test Components trigger actions in the parents context when called from within a block'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>
        {{#my-component}}
          <a href='#' id='fizzbuzz' {{action 'fizzbuzz'}}>Fizzbuzz</a>
        {{/my-component}}
      </div>
    `);

      this.add('controller:application', _controller.default.extend({
        actions: {
          fizzbuzz() {
            assert.ok(true, 'action triggered on parent');
          }
        }
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({})
      });

      return this.visit('/').then(() => {
        this.$('#fizzbuzz', '#wrapper').click();
      });
    }

    ['@test Components trigger actions in the components context when called from within its template'](assert) {
      this.addTemplate('application', `
      <div id='wrapper'>{{#my-component}}{{text}}{{/my-component}}</div>
    `);

      this.add('controller:application', _controller.default.extend({
        actions: {
          fizzbuzz() {
            assert.ok(false, 'action on the wrong context');
          }
        }
      }));
      this.addComponent('my-component', {
        ComponentClass: _glimmer.Component.extend({
          actions: {
            fizzbuzz() {
              assert.ok(true, 'action triggered on component');
            }
          }
        }),
        template: `<a href='#' id='fizzbuzz' {{action 'fizzbuzz'}}>Fizzbuzz</a>`
      });

      return this.visit('/').then(() => {
        this.$('#fizzbuzz', '#wrapper').click();
      });
    }
  });
});
enifed('ember/tests/component_registration_test', ['@ember/application', '@ember/controller', '@ember/-internals/glimmer', 'ember-template-compiler', 'internal-test-helpers', '@ember/-internals/environment'], function (_application, _controller, _glimmer, _emberTemplateCompiler, _internalTestHelpers, _environment) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application Lifecycle - Component Registration', class extends _internalTestHelpers.ApplicationTestCase {
    // This is necessary for this.application.instanceInitializer to not leak between tests
    createApplication(options) {
      return super.createApplication(options, _application.default.extend());
    }

    ['@test The helper becomes the body of the component']() {
      this.addTemplate('components/expand-it', '<p>hello {{yield}}</p>');
      this.addTemplate('application', 'Hello world {{#expand-it}}world{{/expand-it}}');

      return this.visit('/').then(() => {
        this.assertText('Hello world hello world');
        this.assertComponentElement(this.element.firstElementChild, {
          tagName: 'div',
          content: '<p>hello world</p>'
        });
      });
    }

    ['@test The helper becomes the body of the component (ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = true;)']() {
      _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = true;
      this.addTemplate('components/expand-it', '<p>hello {{yield}}</p>');
      this.addTemplate('application', 'Hello world {{#expand-it}}world{{/expand-it}}');

      return this.visit('/').then(() => {
        this.assertInnerHTML('Hello world <p>hello world</p>');
        _environment.ENV._TEMPLATE_ONLY_GLIMMER_COMPONENTS = false;
      });
    }

    ['@test If a component is registered, it is used'](assert) {
      this.addTemplate('components/expand-it', '<p>hello {{yield}}</p>');
      this.addTemplate('application', `Hello world {{#expand-it}}world{{/expand-it}}`);

      this.application.instanceInitializer({
        name: 'expand-it-component',
        initialize(applicationInstance) {
          applicationInstance.register('component:expand-it', _glimmer.Component.extend({
            classNames: 'testing123'
          }));
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('div.testing123').text().trim();
        assert.equal(text, 'hello world', 'The component is composed correctly');
      });
    }

    ['@test Late-registered components can be rendered with custom `layout` property'](assert) {
      this.addTemplate('application', `<div id='wrapper'>there goes {{my-hero}}</div>`);

      this.application.instanceInitializer({
        name: 'my-hero-component',
        initialize(applicationInstance) {
          applicationInstance.register('component:my-hero', _glimmer.Component.extend({
            classNames: 'testing123',
            layout: (0, _emberTemplateCompiler.compile)('watch him as he GOES')
          }));
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('#wrapper').text().trim();
        assert.equal(text, 'there goes watch him as he GOES', 'The component is composed correctly');
      });
    }

    ['@test Late-registered components can be rendered with template registered on the container'](assert) {
      this.addTemplate('application', `<div id='wrapper'>hello world {{sally-rutherford}}-{{#sally-rutherford}}!!!{{/sally-rutherford}}</div>`);

      this.application.instanceInitializer({
        name: 'sally-rutherford-component-template',
        initialize(applicationInstance) {
          applicationInstance.register('template:components/sally-rutherford', (0, _emberTemplateCompiler.compile)('funkytowny{{yield}}'));
        }
      });
      this.application.instanceInitializer({
        name: 'sally-rutherford-component',
        initialize(applicationInstance) {
          applicationInstance.register('component:sally-rutherford', _glimmer.Component);
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('#wrapper').text().trim();
        assert.equal(text, 'hello world funkytowny-funkytowny!!!', 'The component is composed correctly');
      });
    }

    ['@test Late-registered components can be rendered with ONLY the template registered on the container'](assert) {
      this.addTemplate('application', `<div id='wrapper'>hello world {{borf-snorlax}}-{{#borf-snorlax}}!!!{{/borf-snorlax}}</div>`);

      this.application.instanceInitializer({
        name: 'borf-snorlax-component-template',
        initialize(applicationInstance) {
          applicationInstance.register('template:components/borf-snorlax', (0, _emberTemplateCompiler.compile)('goodfreakingTIMES{{yield}}'));
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('#wrapper').text().trim();
        assert.equal(text, 'hello world goodfreakingTIMES-goodfreakingTIMES!!!', 'The component is composed correctly');
      });
    }

    ['@test Assigning layoutName to a component should setup the template as a layout'](assert) {
      assert.expect(1);

      this.addTemplate('application', `<div id='wrapper'>{{#my-component}}{{text}}{{/my-component}}</div>`);
      this.addTemplate('foo-bar-baz', '{{text}}-{{yield}}');

      this.application.instanceInitializer({
        name: 'application-controller',
        initialize(applicationInstance) {
          applicationInstance.register('controller:application', _controller.default.extend({
            text: 'outer'
          }));
        }
      });
      this.application.instanceInitializer({
        name: 'my-component-component',
        initialize(applicationInstance) {
          applicationInstance.register('component:my-component', _glimmer.Component.extend({
            text: 'inner',
            layoutName: 'foo-bar-baz'
          }));
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('#wrapper').text().trim();
        assert.equal(text, 'inner-outer', 'The component is composed correctly');
      });
    }

    ['@test Assigning layoutName and layout to a component should use the `layout` value'](assert) {
      assert.expect(1);

      this.addTemplate('application', `<div id='wrapper'>{{#my-component}}{{text}}{{/my-component}}</div>`);
      this.addTemplate('foo-bar-baz', 'No way!');

      this.application.instanceInitializer({
        name: 'application-controller-layout',
        initialize(applicationInstance) {
          applicationInstance.register('controller:application', _controller.default.extend({
            text: 'outer'
          }));
        }
      });
      this.application.instanceInitializer({
        name: 'my-component-component-layout',
        initialize(applicationInstance) {
          applicationInstance.register('component:my-component', _glimmer.Component.extend({
            text: 'inner',
            layoutName: 'foo-bar-baz',
            layout: (0, _emberTemplateCompiler.compile)('{{text}}-{{yield}}')
          }));
        }
      });

      return this.visit('/').then(() => {
        let text = this.$('#wrapper').text().trim();
        assert.equal(text, 'inner-outer', 'The component is composed correctly');
      });
    }

    ['@test Using name of component that does not exist']() {
      this.addTemplate('application', `<div id='wrapper'>{{#no-good}} {{/no-good}}</div>`);

      // TODO: Use the async form of expectAssertion here when it is available
      expectAssertion(() => {
        this.visit('/');
      }, /.* named "no-good" .*/);

      return this.runLoopSettled();
    }
  });
});
enifed('ember/tests/controller_test', ['@ember/controller', 'internal-test-helpers', '@ember/-internals/glimmer'], function (_controller, _internalTestHelpers, _glimmer) {
  'use strict';

  /*
   In Ember 1.x, controllers subtly affect things like template scope
   and action targets in exciting and often inscrutable ways. This test
   file contains integration tests that verify the correct behavior of
   the many parts of the system that change and rely upon controller scope,
   from the runtime up to the templating layer.
  */

  (0, _internalTestHelpers.moduleFor)('Template scoping examples', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Actions inside an outlet go to the associated controller'](assert) {
      this.add('controller:index', _controller.default.extend({
        actions: {
          componentAction() {
            assert.ok(true, 'controller received the action');
          }
        }
      }));

      this.addComponent('component-with-action', {
        ComponentClass: _glimmer.Component.extend({
          classNames: ['component-with-action'],
          click() {
            this.action();
          }
        })
      });

      this.addTemplate('index', '{{component-with-action action=(action "componentAction")}}');

      return this.visit('/').then(() => {
        this.runTask(() => this.$('.component-with-action').click());
      });
    }
  });
});
enifed('ember/tests/error_handler_test', ['@ember/debug', '@ember/runloop', '@ember/-internals/error-handling', 'rsvp', 'internal-test-helpers'], function (_debug, _runloop, _errorHandling, _rsvp, _internalTestHelpers) {
  'use strict';

  let WINDOW_ONERROR;

  function runThatThrowsSync(message = 'Error for testing error handling') {
    return (0, _runloop.run)(() => {
      throw new Error(message);
    });
  }

  (0, _internalTestHelpers.moduleFor)('error_handler', class extends _internalTestHelpers.AbstractTestCase {
    beforeEach() {
      // capturing this outside of module scope to ensure we grab
      // the test frameworks own window.onerror to reset it
      WINDOW_ONERROR = window.onerror;
    }

    afterEach() {
      (0, _debug.setTesting)(_debug.isTesting);
      window.onerror = WINDOW_ONERROR;

      (0, _errorHandling.setOnerror)(undefined);
    }

    ['@test by default there is no onerror - sync run'](assert) {
      assert.strictEqual((0, _errorHandling.getOnerror)(), undefined, 'precond - there should be no Ember.onerror set by default');
      assert.throws(runThatThrowsSync, Error, 'errors thrown sync are catchable');
    }

    ['@test when Ember.onerror (which rethrows) is registered - sync run'](assert) {
      assert.expect(2);
      (0, _errorHandling.setOnerror)(function (error) {
        assert.ok(true, 'onerror called');
        throw error;
      });
      assert.throws(runThatThrowsSync, Error, 'error is thrown');
    }

    ['@test when Ember.onerror (which does not rethrow) is registered - sync run'](assert) {
      assert.expect(2);
      (0, _errorHandling.setOnerror)(function () {
        assert.ok(true, 'onerror called');
      });
      runThatThrowsSync();
      assert.ok(true, 'no error was thrown, Ember.onerror can intercept errors');
    }

    ['@test does not swallow exceptions by default (Ember.testing = true, no Ember.onerror) - sync run'](assert) {
      (0, _debug.setTesting)(true);

      let error = new Error('the error');
      assert.throws(() => {
        (0, _runloop.run)(() => {
          throw error;
        });
      }, error);
    }

    ['@test does not swallow exceptions by default (Ember.testing = false, no Ember.onerror) - sync run'](assert) {
      (0, _debug.setTesting)(false);
      let error = new Error('the error');
      assert.throws(() => {
        (0, _runloop.run)(() => {
          throw error;
        });
      }, error);
    }

    ['@test does not swallow exceptions (Ember.testing = false, Ember.onerror which rethrows) - sync run'](assert) {
      assert.expect(2);
      (0, _debug.setTesting)(false);

      (0, _errorHandling.setOnerror)(function (error) {
        assert.ok(true, 'Ember.onerror was called');
        throw error;
      });

      let error = new Error('the error');
      assert.throws(() => {
        (0, _runloop.run)(() => {
          throw error;
        });
      }, error);
    }

    ['@test Ember.onerror can intercept errors (aka swallow) by not rethrowing (Ember.testing = false) - sync run'](assert) {
      assert.expect(1);
      (0, _debug.setTesting)(false);

      (0, _errorHandling.setOnerror)(function () {
        assert.ok(true, 'Ember.onerror was called');
      });

      let error = new Error('the error');
      try {
        (0, _runloop.run)(() => {
          throw error;
        });
      } catch (e) {
        assert.notOk(true, 'Ember.onerror that does not rethrow is intentionally swallowing errors, try / catch wrapping does not see error');
      }
    }

    ['@test does not swallow exceptions by default (Ember.testing = true, no Ember.onerror) - async run'](assert) {
      let done = assert.async();
      let caughtByWindowOnerror;

      (0, _debug.setTesting)(true);

      window.onerror = function (message) {
        caughtByWindowOnerror = message;
        // prevent "bubbling" and therefore failing the test
        return true;
      };

      (0, _runloop.later)(() => {
        throw new Error('the error');
      }, 10);

      setTimeout(() => {
        assert.pushResult({
          result: /the error/.test(caughtByWindowOnerror),
          actual: caughtByWindowOnerror,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        done();
      }, 20);
    }

    ['@test does not swallow exceptions by default (Ember.testing = false, no Ember.onerror) - async run'](assert) {
      let done = assert.async();
      let caughtByWindowOnerror;

      (0, _debug.setTesting)(false);

      window.onerror = function (message) {
        caughtByWindowOnerror = message;
        // prevent "bubbling" and therefore failing the test
        return true;
      };

      (0, _runloop.later)(() => {
        throw new Error('the error');
      }, 10);

      setTimeout(() => {
        assert.pushResult({
          result: /the error/.test(caughtByWindowOnerror),
          actual: caughtByWindowOnerror,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        done();
      }, 20);
    }

    ['@test Ember.onerror can intercept errors (aka swallow) by not rethrowing (Ember.testing = false) - async run'](assert) {
      let done = assert.async();

      (0, _debug.setTesting)(false);

      window.onerror = function () {
        assert.notOk(true, 'window.onerror is never invoked when Ember.onerror intentionally swallows errors');
        // prevent "bubbling" and therefore failing the test
        return true;
      };

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called with the error');
      });

      (0, _runloop.later)(() => {
        throw thrown;
      }, 10);

      setTimeout(done, 20);
    }

    [`@test errors in promise constructor when Ember.onerror which does not rethrow is present - rsvp`](assert) {
      assert.expect(1);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      new _rsvp.default.Promise(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise constructor when Ember.onerror which does rethrow is present - rsvp`](assert) {
      assert.expect(2);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      new _rsvp.default.Promise(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise constructor when Ember.onerror which does not rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(1);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      new _rsvp.default.Promise(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise constructor when Ember.onerror which does rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(2);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      new _rsvp.default.Promise(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise .then callback when Ember.onerror which does not rethrow is present - rsvp`](assert) {
      assert.expect(1);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      _rsvp.default.resolve().then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise .then callback when Ember.onerror which does rethrow is present - rsvp`](assert) {
      assert.expect(2);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      _rsvp.default.resolve().then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise .then callback when Ember.onerror which does not rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(1);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      _rsvp.default.resolve().then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in promise .then callback when Ember.onerror which does rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(2);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      _rsvp.default.resolve().then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 10));
    }

    [`@test errors in async promise .then callback when Ember.onerror which does not rethrow is present - rsvp`](assert) {
      assert.expect(1);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      new _rsvp.default.Promise(resolve => setTimeout(resolve, 10)).then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 20));
    }

    [`@test errors in async promise .then callback when Ember.onerror which does rethrow is present - rsvp`](assert) {
      assert.expect(2);

      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      new _rsvp.default.Promise(resolve => setTimeout(resolve, 10)).then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 20));
    }

    [`@test errors in async promise .then callback when Ember.onerror which does not rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(1);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
      });

      new _rsvp.default.Promise(resolve => setTimeout(resolve, 10)).then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 20));
    }

    [`@test errors in async promise .then callback when Ember.onerror which does rethrow is present (Ember.testing = false) - rsvp`](assert) {
      assert.expect(2);

      (0, _debug.setTesting)(false);
      let thrown = new Error('the error');
      (0, _errorHandling.setOnerror)(function (error) {
        assert.strictEqual(error, thrown, 'Ember.onerror is called for errors thrown in RSVP promises');
        throw error;
      });

      window.onerror = function (message) {
        assert.pushResult({
          result: /the error/.test(message),
          actual: message,
          expected: 'to include `the error`',
          message: 'error should bubble out to window.onerror, and therefore fail tests (due to QUnit implementing window.onerror)'
        });

        // prevent "bubbling" and therefore failing the test
        return true;
      };

      new _rsvp.default.Promise(resolve => setTimeout(resolve, 10)).then(() => {
        throw thrown;
      });

      // RSVP.Promise's are configured to settle within the run loop, this
      // ensures that run loop has completed
      return new _rsvp.default.Promise(resolve => setTimeout(resolve, 20));
    }
  });
});
enifed('ember/tests/helpers/helper_registration_test', ['internal-test-helpers', '@ember/controller', '@ember/service', '@ember/-internals/glimmer'], function (_internalTestHelpers, _controller, _service, _glimmer) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Application Lifecycle - Helper Registration', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Unbound dashed helpers registered on the container can be late-invoked'](assert) {
      this.addTemplate('application', `<div id='wrapper'>{{x-borf}} {{x-borf 'YES'}}</div>`);

      let myHelper = (0, _glimmer.helper)(params => params[0] || 'BORF');
      this.application.register('helper:x-borf', myHelper);

      return this.visit('/').then(() => {
        assert.equal(this.$('#wrapper').text(), 'BORF YES', 'The helper was invoked from the container');
      });
    }

    ['@test Bound helpers registered on the container can be late-invoked'](assert) {
      this.addTemplate('application', `<div id='wrapper'>{{x-reverse}} {{x-reverse foo}}</div>`);

      this.add('controller:application', _controller.default.extend({
        foo: 'alex'
      }));

      this.application.register('helper:x-reverse', (0, _glimmer.helper)(function ([value]) {
        return value ? value.split('').reverse().join('') : '--';
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('#wrapper').text(), '-- xela', 'The bound helper was invoked from the container');
      });
    }

    ['@test Undashed helpers registered on the container can be invoked'](assert) {
      this.addTemplate('application', `<div id='wrapper'>{{omg}}|{{yorp 'boo'}}|{{yorp 'ya'}}</div>`);

      this.application.register('helper:omg', (0, _glimmer.helper)(() => 'OMG'));

      this.application.register('helper:yorp', (0, _glimmer.helper)(([value]) => value));

      return this.visit('/').then(() => {
        assert.equal(this.$('#wrapper').text(), 'OMG|boo|ya', 'The helper was invoked from the container');
      });
    }

    ['@test Helpers can receive injections'](assert) {
      this.addTemplate('application', `<div id='wrapper'>{{full-name}}</div>`);

      let serviceCalled = false;

      this.add('service:name-builder', _service.default.extend({
        build() {
          serviceCalled = true;
        }
      }));

      this.add('helper:full-name', _glimmer.Helper.extend({
        nameBuilder: (0, _service.inject)('name-builder'),
        compute() {
          this.get('nameBuilder').build();
        }
      }));

      return this.visit('/').then(() => {
        assert.ok(serviceCalled, 'service was injected, method called');
      });
    }
  });
});
enifed('ember/tests/helpers/link_to_test', ['internal-test-helpers', '@ember/controller', '@ember/-internals/runtime', '@ember/-internals/metal', '@ember/instrumentation', '@ember/-internals/routing', '@ember/canary-features'], function (_internalTestHelpers, _controller, _runtime, _metal, _instrumentation, _routing, _canaryFeatures) {
  'use strict';

  // IE includes the host name
  function normalizeUrl(url) {
    return url.replace(/https?:\/\/[^\/]+/, '');
  }

  function shouldNotBeActive(assert, element) {
    checkActive(assert, element, false);
  }

  function shouldBeActive(assert, element) {
    checkActive(assert, element, true);
  }

  function checkActive(assert, element, active) {
    let classList = element.attr('class');
    assert.equal(classList.indexOf('active') > -1, active, `${element} active should be ${active}`);
  }

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper - basic tests', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();

      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
      {{#link-to 'index' id='self-link'}}Self{{/link-to}}
    `);
      this.addTemplate('about', `
      <h3 class="about">About</h3>
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
      {{#link-to 'about' id='self-link'}}Self{{/link-to}}
    `);
    }

    ['@test The {{link-to}} helper moves into the named route'](assert) {
      return this.visit('/').then(() => {
        assert.equal(this.$('h3.home').length, 1, 'The home template was rendered');
        assert.equal(this.$('#self-link.active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#about-link:not(.active)').length, 1, 'The other link was rendered without active class');

        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.$('h3.about').length, 1, 'The about template was rendered');
        assert.equal(this.$('#self-link.active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#home-link:not(.active)').length, 1, 'The other link was rendered without active class');
      });
    }

    [`@test the {{link-to}} helper doesn't add an href when the tagName isn't 'a'`](assert) {
      this.addTemplate('index', `
      {{#link-to 'about' id='about-link' tagName='div'}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#about-link').attr('href'), undefined, 'there is no href attribute');
      });
    }

    [`@test the {{link-to}} applies a 'disabled' class when disabled`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link-static" disabledWhen="shouldDisable"}}About{{/link-to}}
      {{#link-to "about" id="about-link-dynamic" disabledWhen=dynamicDisabledWhen}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        shouldDisable: true,
        dynamicDisabledWhen: 'shouldDisable'
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('#about-link-static.disabled').length, 1, 'The static link is disabled when its disabledWhen is true');
        assert.equal(this.$('#about-link-dynamic.disabled').length, 1, 'The dynamic link is disabled when its disabledWhen is true');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => controller.set('dynamicDisabledWhen', false));

        assert.equal(this.$('#about-link-dynamic.disabled').length, 0, 'The dynamic link is re-enabled when its disabledWhen becomes false');
      });
    }

    [`@test the {{link-to}} doesn't apply a 'disabled' class if disabledWhen is not provided`](assert) {
      this.addTemplate('index', `{{#link-to "about" id="about-link"}}About{{/link-to}}`);

      return this.visit('/').then(() => {
        assert.ok(!this.$('#about-link').hasClass('disabled'), 'The link is not disabled if disabledWhen not provided');
      });
    }

    [`@test the {{link-to}} helper supports a custom disabledClass`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link" disabledWhen=true disabledClass="do-not-want"}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#about-link.do-not-want').length, 1, 'The link can apply a custom disabled class');
      });
    }

    [`@test the {{link-to}} helper supports a custom disabledClass set via bound param`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link" disabledWhen=true disabledClass=disabledClass}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        disabledClass: 'do-not-want'
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('#about-link.do-not-want').length, 1, 'The link can apply a custom disabled class via bound param');
      });
    }

    [`@test the {{link-to}} helper does not respond to clicks when disabledWhen`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link" disabledWhen=true}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.$('h3.about').length, 0, 'Transitioning did not occur');
      });
    }

    [`@test the {{link-to}} helper does not respond to clicks when disabled`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link" disabled=true}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.$('h3.about').length, 0, 'Transitioning did not occur');
      });
    }

    [`@test the {{link-to}} helper responds to clicks according to its disabledWhen bound param`](assert) {
      this.addTemplate('index', `
      {{#link-to "about" id="about-link" disabledWhen=disabledWhen}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        disabledWhen: true
      }));

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.$('h3.about').length, 0, 'Transitioning did not occur');

        let controller = this.applicationInstance.lookup('controller:index');
        controller.set('disabledWhen', false);

        return this.runLoopSettled();
      }).then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.$('h3.about').length, 1, 'Transitioning did occur when disabledWhen became false');
      });
    }

    [`@test The {{link-to}} helper supports a custom activeClass`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
      {{#link-to 'index' id='self-link' activeClass='zomg-active'}}Self{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('h3.home').length, 1, 'The home template was rendered');
        assert.equal(this.$('#self-link.zomg-active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#about-link:not(.active)').length, 1, 'The other link was rendered without active class');
      });
    }

    [`@test The {{link-to}} helper supports a custom activeClass from a bound param`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
      {{#link-to 'index' id='self-link' activeClass=activeClass}}Self{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        activeClass: 'zomg-active'
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('h3.home').length, 1, 'The home template was rendered');
        assert.equal(this.$('#self-link.zomg-active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#about-link:not(.active)').length, 1, 'The other link was rendered without active class');
      });
    }

    [`@test The {{link-to}} helper supports 'classNameBindings' with custom values [GH #11699]`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link' classNameBindings='foo:foo-is-true:foo-is-false'}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        foo: false
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('#about-link.foo-is-false').length, 1, 'The about-link was rendered with the falsy class');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => controller.set('foo', true));

        assert.equal(this.$('#about-link.foo-is-true').length, 1, 'The about-link was rendered with the truthy class after toggling the property');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper - location hooks', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();

      this.updateCount = 0;
      this.replaceCount = 0;

      let testContext = this;
      this.add('location:none', _routing.NoneLocation.extend({
        setURL() {
          testContext.updateCount++;
          return this._super(...arguments);
        },
        replaceURL() {
          testContext.replaceCount++;
          return this._super(...arguments);
        }
      }));

      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
      {{#link-to 'index' id='self-link'}}Self{{/link-to}}
    `);
      this.addTemplate('about', `
      <h3 class="about">About</h3>
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
      {{#link-to 'about' id='self-link'}}Self{{/link-to}}
    `);
    }

    visit() {
      return super.visit(...arguments).then(() => {
        this.updateCountAfterVisit = this.updateCount;
        this.replaceCountAfterVisit = this.replaceCount;
      });
    }

    ['@test The {{link-to}} helper supports URL replacement'](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link' replace=true}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.updateCount, this.updateCountAfterVisit, 'setURL should not be called');
        assert.equal(this.replaceCount, this.replaceCountAfterVisit + 1, 'replaceURL should be called once');
      });
    }

    ['@test The {{link-to}} helper supports URL replacement via replace=boundTruthyThing'](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link' replace=boundTruthyThing}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        boundTruthyThing: true
      }));

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.updateCount, this.updateCountAfterVisit, 'setURL should not be called');
        assert.equal(this.replaceCount, this.replaceCountAfterVisit + 1, 'replaceURL should be called once');
      });
    }

    ['@test The {{link-to}} helper supports setting replace=boundFalseyThing'](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link' replace=boundFalseyThing}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        boundFalseyThing: false
      }));

      return this.visit('/').then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(this.updateCount, this.updateCountAfterVisit + 1, 'setURL should be called');
        assert.equal(this.replaceCount, this.replaceCountAfterVisit, 'replaceURL should not be called');
      });
    }
  });

  if (_canaryFeatures.EMBER_IMPROVED_INSTRUMENTATION) {
    (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper with EMBER_IMPROVED_INSTRUMENTATION', class extends _internalTestHelpers.ApplicationTestCase {
      constructor() {
        super();

        this.router.map(function () {
          this.route('about');
        });

        this.addTemplate('index', `
        <h3 class="home">Home</h3>
        {{#link-to 'about' id='about-link'}}About{{/link-to}}
        {{#link-to 'index' id='self-link'}}Self{{/link-to}}
      `);
        this.addTemplate('about', `
        <h3 class="about">About</h3>
        {{#link-to 'index' id='home-link'}}Home{{/link-to}}
        {{#link-to 'about' id='self-link'}}Self{{/link-to}}
      `);
      }

      beforeEach() {
        return this.visit('/');
      }

      afterEach() {
        (0, _instrumentation.reset)();

        return super.afterEach();
      }

      ['@test The {{link-to}} helper fires an interaction event'](assert) {
        assert.expect(2);

        (0, _instrumentation.subscribe)('interaction.link-to', {
          before() {
            assert.ok(true, 'instrumentation subscriber was called');
          },
          after() {
            assert.ok(true, 'instrumentation subscriber was called');
          }
        });

        return this.click('#about-link');
      }

      ['@test The {{link-to}} helper interaction event includes the route name'](assert) {
        assert.expect(2);

        (0, _instrumentation.subscribe)('interaction.link-to', {
          before(name, timestamp, { routeName }) {
            assert.equal(routeName, 'about', 'instrumentation subscriber was passed route name');
          },
          after(name, timestamp, { routeName }) {
            assert.equal(routeName, 'about', 'instrumentation subscriber was passed route name');
          }
        });

        return this.click('#about-link');
      }

      ['@test The {{link-to}} helper interaction event includes the transition in the after hook'](assert) {
        assert.expect(1);

        (0, _instrumentation.subscribe)('interaction.link-to', {
          before() {},
          after(name, timestamp, { transition }) {
            assert.equal(transition.targetName, 'about', 'instrumentation subscriber was passed route name');
          }
        });

        return this.click('#about-link');
      }
    });
  }

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper - nested routes and link-to arguments', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test The {{link-to}} helper supports leaving off .index for nested routes'](assert) {
      this.router.map(function () {
        this.route('about', function () {
          this.route('item');
        });
      });

      this.addTemplate('about', `<h1>About</h1>{{outlet}}`);
      this.addTemplate('about.index', `<div id='index'>Index</div>`);
      this.addTemplate('about.item', `<div id='item'>{{#link-to 'about'}}About{{/link-to}}</div>`);

      return this.visit('/about/item').then(() => {
        assert.equal(normalizeUrl(this.$('#item a').attr('href')), '/about');
      });
    }

    [`@test The {{link-to}} helper supports custom, nested, current-when`](assert) {
      this.router.map(function () {
        this.route('index', { path: '/' }, function () {
          this.route('about');
        });

        this.route('item');
      });

      this.addTemplate('index', `<h3 class="home">Home</h3>{{outlet}}`);
      this.addTemplate('index.about', `
      {{#link-to 'item' id='other-link' current-when='index'}}ITEM{{/link-to}}
    `);

      return this.visit('/about').then(() => {
        assert.equal(this.$('#other-link.active').length, 1, 'The link is active since current-when is a parent route');
      });
    }

    [`@test The {{link-to}} helper does not disregard current-when when it is given explicitly for a route`](assert) {
      this.router.map(function () {
        this.route('index', { path: '/' }, function () {
          this.route('about');
        });

        this.route('items', function () {
          this.route('item');
        });
      });

      this.addTemplate('index', `<h3 class="home">Home</h3>{{outlet}}`);
      this.addTemplate('index.about', `
      {{#link-to 'items' id='other-link' current-when='index'}}ITEM{{/link-to}}
    `);

      return this.visit('/about').then(() => {
        assert.equal(this.$('#other-link.active').length, 1, 'The link is active when current-when is given for explicitly for a route');
      });
    }

    ['@test The {{link-to}} helper does not disregard current-when when it is set via a bound param'](assert) {
      this.router.map(function () {
        this.route('index', { path: '/' }, function () {
          this.route('about');
        });

        this.route('items', function () {
          this.route('item');
        });
      });

      this.add('controller:index.about', _controller.default.extend({
        currentWhen: 'index'
      }));

      this.addTemplate('index', `<h3 class="home">Home</h3>{{outlet}}`);
      this.addTemplate('index.about', `{{#link-to 'items' id='other-link' current-when=currentWhen}}ITEM{{/link-to}}`);

      return this.visit('/about').then(() => {
        assert.equal(this.$('#other-link.active').length, 1, 'The link is active when current-when is given for explicitly for a route');
      });
    }

    ['@test The {{link-to}} helper supports multiple current-when routes'](assert) {
      this.router.map(function () {
        this.route('index', { path: '/' }, function () {
          this.route('about');
        });
        this.route('item');
        this.route('foo');
      });

      this.addTemplate('index', `<h3 class="home">Home</h3>{{outlet}}`);
      this.addTemplate('index.about', `{{#link-to 'item' id='link1' current-when='item index'}}ITEM{{/link-to}}`);
      this.addTemplate('item', `{{#link-to 'item' id='link2' current-when='item index'}}ITEM{{/link-to}}`);
      this.addTemplate('foo', `{{#link-to 'item' id='link3' current-when='item index'}}ITEM{{/link-to}}`);

      return this.visit('/about').then(() => {
        assert.equal(this.$('#link1.active').length, 1, 'The link is active since current-when contains the parent route');

        return this.visit('/item');
      }).then(() => {
        assert.equal(this.$('#link2.active').length, 1, 'The link is active since you are on the active route');

        return this.visit('/foo');
      }).then(() => {
        assert.equal(this.$('#link3.active').length, 0, 'The link is not active since current-when does not contain the active route');
      });
    }

    ['@test The {{link-to}} helper supports boolean values for current-when'](assert) {
      this.router.map(function () {
        this.route('index', { path: '/' }, function () {
          this.route('about');
        });
        this.route('item');
      });

      this.addTemplate('index.about', `
      {{#link-to 'index' id='index-link' current-when=isCurrent}}index{{/link-to}}
      {{#link-to 'item' id='about-link' current-when=true}}ITEM{{/link-to}}
    `);

      this.add('controller:index.about', _controller.default.extend({ isCurrent: false }));

      return this.visit('/about').then(() => {
        assert.ok(this.$('#about-link').hasClass('active'), 'The link is active since current-when is true');
        assert.notOk(this.$('#index-link').hasClass('active'), 'The link is not active since current-when is false');

        let controller = this.applicationInstance.lookup('controller:index.about');
        this.runTask(() => controller.set('isCurrent', true));

        assert.ok(this.$('#index-link').hasClass('active'), 'The link is active since current-when is true');
      });
    }

    ['@test The {{link-to}} helper defaults to bubbling'](assert) {
      this.addTemplate('about', `
      <div {{action 'hide'}}>
        {{#link-to 'about.contact' id='about-contact'}}About{{/link-to}}
      </div>
      {{outlet}}
    `);
      this.addTemplate('about.contact', `
      <h1 id='contact'>Contact</h1>
    `);

      this.router.map(function () {
        this.route('about', function () {
          this.route('contact');
        });
      });

      let hidden = 0;

      this.add('route:about', _routing.Route.extend({
        actions: {
          hide() {
            hidden++;
          }
        }
      }));

      return this.visit('/about').then(() => {
        return this.click('#about-contact');
      }).then(() => {
        assert.equal(this.$('#contact').text(), 'Contact', 'precond - the link worked');

        assert.equal(hidden, 1, 'The link bubbles');
      });
    }

    [`@test The {{link-to}} helper supports bubbles=false`](assert) {
      this.addTemplate('about', `
      <div {{action 'hide'}}>
        {{#link-to 'about.contact' id='about-contact' bubbles=false}}
          About
        {{/link-to}}
      </div>
      {{outlet}}
    `);
      this.addTemplate('about.contact', `<h1 id='contact'>Contact</h1>`);

      this.router.map(function () {
        this.route('about', function () {
          this.route('contact');
        });
      });

      let hidden = 0;

      this.add('route:about', _routing.Route.extend({
        actions: {
          hide() {
            hidden++;
          }
        }
      }));

      return this.visit('/about').then(() => {
        return this.click('#about-contact');
      }).then(() => {
        assert.equal(this.$('#contact').text(), 'Contact', 'precond - the link worked');

        assert.equal(hidden, 0, "The link didn't bubble");
      });
    }

    [`@test The {{link-to}} helper supports bubbles=boundFalseyThing`](assert) {
      this.addTemplate('about', `
      <div {{action 'hide'}}>
        {{#link-to 'about.contact' id='about-contact' bubbles=boundFalseyThing}}
          About
        {{/link-to}}
      </div>
      {{outlet}}
    `);
      this.addTemplate('about.contact', `<h1 id='contact'>Contact</h1>`);

      this.add('controller:about', _controller.default.extend({
        boundFalseyThing: false
      }));

      this.router.map(function () {
        this.route('about', function () {
          this.route('contact');
        });
      });

      let hidden = 0;

      this.add('route:about', _routing.Route.extend({
        actions: {
          hide() {
            hidden++;
          }
        }
      }));

      return this.visit('/about').then(() => {
        return this.click('#about-contact');
      }).then(() => {
        assert.equal(this.$('#contact').text(), 'Contact', 'precond - the link worked');
        assert.equal(hidden, 0, "The link didn't bubble");
      });
    }

    [`@test The {{link-to}} helper moves into the named route with context`](assert) {
      this.router.map(function () {
        this.route('about');
        this.route('item', { path: '/item/:id' });
      });

      this.addTemplate('about', `
      <h3 class="list">List</h3>
      <ul>
        {{#each model as |person|}}
          <li>
            {{#link-to 'item' person id=person.id}}
              {{person.name}}
            {{/link-to}}
          </li>
        {{/each}}
      </ul>
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
    `);

      this.addTemplate('item', `
      <h3 class="item">Item</h3>
      <p>{{model.name}}</p>
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
    `);

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
    `);

      this.add('route:about', _routing.Route.extend({
        model() {
          return [{ id: 'yehuda', name: 'Yehuda Katz' }, { id: 'tom', name: 'Tom Dale' }, { id: 'erik', name: 'Erik Brynroflsson' }];
        }
      }));

      return this.visit('/about').then(() => {
        assert.equal(this.$('h3.list').length, 1, 'The home template was rendered');
        assert.equal(normalizeUrl(this.$('#home-link').attr('href')), '/', 'The home link points back at /');

        return this.click('#yehuda');
      }).then(() => {
        assert.equal(this.$('h3.item').length, 1, 'The item template was rendered');
        assert.equal(this.$('p').text(), 'Yehuda Katz', 'The name is correct');

        return this.click('#home-link');
      }).then(() => {
        return this.click('#about-link');
      }).then(() => {
        assert.equal(normalizeUrl(this.$('li a#yehuda').attr('href')), '/item/yehuda');
        assert.equal(normalizeUrl(this.$('li a#tom').attr('href')), '/item/tom');
        assert.equal(normalizeUrl(this.$('li a#erik').attr('href')), '/item/erik');

        return this.click('#erik');
      }).then(() => {
        assert.equal(this.$('h3.item').length, 1, 'The item template was rendered');
        assert.equal(this.$('p').text(), 'Erik Brynroflsson', 'The name is correct');
      });
    }

    [`@test The {{link-to}} helper binds some anchor html tag common attributes`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'index' id='self-link' title='title-attr' rel='rel-attr' tabindex='-1'}}
        Self
      {{/link-to}}
    `);

      return this.visit('/').then(() => {
        let link = this.$('#self-link');
        assert.equal(link.attr('title'), 'title-attr', 'The self-link contains title attribute');
        assert.equal(link.attr('rel'), 'rel-attr', 'The self-link contains rel attribute');
        assert.equal(link.attr('tabindex'), '-1', 'The self-link contains tabindex attribute');
      });
    }

    [`@test The {{link-to}} helper supports 'target' attribute`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'index' id='self-link' target='_blank'}}Self{{/link-to}}
    `);

      return this.visit('/').then(() => {
        let link = this.$('#self-link');
        assert.equal(link.attr('target'), '_blank', 'The self-link contains `target` attribute');
      });
    }

    [`@test The {{link-to}} helper supports 'target' attribute specified as a bound param`](assert) {
      this.addTemplate('index', `<h3 class="home">Home</h3>{{#link-to 'index' id='self-link' target=boundLinkTarget}}Self{{/link-to}}`);

      this.add('controller:index', _controller.default.extend({
        boundLinkTarget: '_blank'
      }));

      return this.visit('/').then(() => {
        let link = this.$('#self-link');
        assert.equal(link.attr('target'), '_blank', 'The self-link contains `target` attribute');
      });
    }

    [`@test the {{link-to}} helper calls preventDefault`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to 'about' id='about-link'}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assertNav({ prevented: true }, () => this.$('#about-link').click(), assert);
      });
    }

    [`@test the {{link-to}} helper does not call preventDefault if 'preventDefault=false' is passed as an option`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to 'about' id='about-link' preventDefault=false}}About{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assertNav({ prevented: false }, () => this.$('#about-link').trigger('click'), assert);
      });
    }

    [`@test the {{link-to}} helper does not call preventDefault if 'preventDefault=boundFalseyThing' is passed as an option`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to 'about' id='about-link' preventDefault=boundFalseyThing}}About{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        boundFalseyThing: false
      }));

      return this.visit('/').then(() => {
        assertNav({ prevented: false }, () => this.$('#about-link').trigger('click'), assert);
      });
    }

    [`@test The {{link-to}} helper does not call preventDefault if 'target' attribute is provided`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'index' id='self-link' target='_blank'}}Self{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assertNav({ prevented: false }, () => this.$('#self-link').click(), assert);
      });
    }

    [`@test The {{link-to}} helper should preventDefault when 'target = _self'`](assert) {
      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to 'index' id='self-link' target='_self'}}Self{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assertNav({ prevented: true }, () => this.$('#self-link').click(), assert);
      });
    }

    [`@test The {{link-to}} helper should not transition if target is not equal to _self or empty`](assert) {
      this.addTemplate('index', `
      {{#link-to 'about' id='about-link' replace=true target='_blank'}}
        About
      {{/link-to}}
    `);

      this.router.map(function () {
        this.route('about');
      });

      return this.visit('/').then(() => this.click('#about-link')).then(() => {
        let currentRouteName = this.applicationInstance.lookup('controller:application').get('currentRouteName');
        assert.notEqual(currentRouteName, 'about', 'link-to should not transition if target is not equal to _self or empty');
      });
    }

    [`@test The {{link-to}} helper accepts string/numeric arguments`](assert) {
      this.router.map(function () {
        this.route('filter', { path: '/filters/:filter' });
        this.route('post', { path: '/post/:post_id' });
        this.route('repo', { path: '/repo/:owner/:name' });
      });

      this.add('controller:filter', _controller.default.extend({
        filter: 'unpopular',
        repo: { owner: 'ember', name: 'ember.js' },
        post_id: 123
      }));

      this.addTemplate('filter', `
      <p>{{filter}}</p>
      {{#link-to "filter" "unpopular" id="link"}}Unpopular{{/link-to}}
      {{#link-to "filter" filter id="path-link"}}Unpopular{{/link-to}}
      {{#link-to "post" post_id id="post-path-link"}}Post{{/link-to}}
      {{#link-to "post" 123 id="post-number-link"}}Post{{/link-to}}
      {{#link-to "repo" repo id="repo-object-link"}}Repo{{/link-to}}
    `);

      return this.visit('/filters/popular').then(() => {
        assert.equal(normalizeUrl(this.$('#link').attr('href')), '/filters/unpopular');
        assert.equal(normalizeUrl(this.$('#path-link').attr('href')), '/filters/unpopular');
        assert.equal(normalizeUrl(this.$('#post-path-link').attr('href')), '/post/123');
        assert.equal(normalizeUrl(this.$('#post-number-link').attr('href')), '/post/123');
        assert.equal(normalizeUrl(this.$('#repo-object-link').attr('href')), '/repo/ember/ember.js');
      });
    }

    [`@test Issue 4201 - Shorthand for route.index shouldn't throw errors about context arguments`](assert) {
      assert.expect(2);
      this.router.map(function () {
        this.route('lobby', function () {
          this.route('index', { path: ':lobby_id' });
          this.route('list');
        });
      });

      this.add('route:lobby.index', _routing.Route.extend({
        model(params) {
          assert.equal(params.lobby_id, 'foobar');
          return params.lobby_id;
        }
      }));

      this.addTemplate('lobby.index', `
      {{#link-to 'lobby' 'foobar' id='lobby-link'}}Lobby{{/link-to}}
    `);
      this.addTemplate('lobby.list', `
      {{#link-to 'lobby' 'foobar' id='lobby-link'}}Lobby{{/link-to}}
    `);

      return this.visit('/lobby/list').then(() => this.click('#lobby-link')).then(() => shouldBeActive(assert, this.$('#lobby-link')));
    }

    [`@test Quoteless route param performs property lookup`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to 'index' id='string-link'}}string{{/link-to}}
      {{#link-to foo id='path-link'}}path{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        foo: 'index'
      }));

      let assertEquality = href => {
        assert.equal(normalizeUrl(this.$('#string-link').attr('href')), '/');
        assert.equal(normalizeUrl(this.$('#path-link').attr('href')), href);
      };

      return this.visit('/').then(() => {
        assertEquality('/');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => controller.set('foo', 'about'));

        assertEquality('/about');
      });
    }

    [`@test The {{link-to}} helper refreshes href element when one of params changes`](assert) {
      this.router.map(function () {
        this.route('post', { path: '/posts/:post_id' });
      });

      let post = { id: '1' };
      let secondPost = { id: '2' };

      this.addTemplate('index', `
      {{#link-to "post" post id="post"}}post{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend());

      return this.visit('/').then(() => {
        let indexController = this.applicationInstance.lookup('controller:index');
        this.runTask(() => indexController.set('post', post));

        assert.equal(normalizeUrl(this.$('#post').attr('href')), '/posts/1', 'precond - Link has rendered href attr properly');

        this.runTask(() => indexController.set('post', secondPost));

        assert.equal(this.$('#post').attr('href'), '/posts/2', 'href attr was updated after one of the params had been changed');

        this.runTask(() => indexController.set('post', null));

        assert.equal(this.$('#post').attr('href'), '#', 'href attr becomes # when one of the arguments in nullified');
      });
    }

    [`@test The {{link-to}} helper is active when a route is active`](assert) {
      this.router.map(function () {
        this.route('about', function () {
          this.route('item');
        });
      });

      this.addTemplate('about', `
      <div id='about'>
        {{#link-to 'about' id='about-link'}}About{{/link-to}}
        {{#link-to 'about.item' id='item-link'}}Item{{/link-to}}
        {{outlet}}
      </div>
    `);

      return this.visit('/about').then(() => {
        assert.equal(this.$('#about-link.active').length, 1, 'The about route link is active');
        assert.equal(this.$('#item-link.active').length, 0, 'The item route link is inactive');

        return this.visit('/about/item');
      }).then(() => {
        assert.equal(this.$('#about-link.active').length, 1, 'The about route link is active');
        assert.equal(this.$('#item-link.active').length, 1, 'The item route link is active');
      });
    }

    [`@test The {{link-to}} helper works in an #each'd array of string route names`](assert) {
      this.router.map(function () {
        this.route('foo');
        this.route('bar');
        this.route('rar');
      });

      this.add('controller:index', _controller.default.extend({
        routeNames: (0, _runtime.A)(['foo', 'bar', 'rar']),
        route1: 'bar',
        route2: 'foo'
      }));

      this.addTemplate('index', `
      {{#each routeNames as |routeName|}}
        {{#link-to routeName}}{{routeName}}{{/link-to}}
      {{/each}}
      {{#each routeNames as |r|}}
        {{#link-to r}}{{r}}{{/link-to}}
      {{/each}}
      {{#link-to route1}}a{{/link-to}}
      {{#link-to route2}}b{{/link-to}}
    `);

      let linksEqual = (links, expected) => {
        assert.equal(links.length, expected.length, 'Has correct number of links');

        let idx;
        for (idx = 0; idx < links.length; idx++) {
          let href = this.$(links[idx]).attr('href');
          // Old IE includes the whole hostname as well
          assert.equal(href.slice(-expected[idx].length), expected[idx], `Expected link to be '${expected[idx]}', but was '${href}'`);
        }
      };

      return this.visit('/').then(() => {
        linksEqual(this.$('a'), ['/foo', '/bar', '/rar', '/foo', '/bar', '/rar', '/bar', '/foo']);

        let indexController = this.applicationInstance.lookup('controller:index');
        this.runTask(() => indexController.set('route1', 'rar'));

        linksEqual(this.$('a'), ['/foo', '/bar', '/rar', '/foo', '/bar', '/rar', '/rar', '/foo']);

        this.runTask(() => indexController.routeNames.shiftObject());

        linksEqual(this.$('a'), ['/bar', '/rar', '/bar', '/rar', '/rar', '/foo']);
      });
    }

    [`@test The non-block form {{link-to}} helper moves into the named route`](assert) {
      assert.expect(3);
      this.router.map(function () {
        this.route('contact');
      });

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{link-to 'Contact us' 'contact' id='contact-link'}}
      {{#link-to 'index' id='self-link'}}Self{{/link-to}}
    `);
      this.addTemplate('contact', `
      <h3 class="contact">Contact</h3>
      {{link-to 'Home' 'index' id='home-link'}}
      {{link-to 'Self' 'contact' id='self-link'}}
    `);

      return this.visit('/').then(() => {
        return this.click('#contact-link');
      }).then(() => {
        assert.equal(this.$('h3.contact').length, 1, 'The contact template was rendered');
        assert.equal(this.$('#self-link.active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#home-link:not(.active)').length, 1, 'The other link was rendered without active class');
      });
    }

    [`@test The non-block form {{link-to}} helper updates the link text when it is a binding`](assert) {
      assert.expect(8);
      this.router.map(function () {
        this.route('contact');
      });

      this.add('controller:index', _controller.default.extend({
        contactName: 'Jane'
      }));

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{link-to contactName 'contact' id='contact-link'}}
      {{#link-to 'index' id='self-link'}}Self{{/link-to}}
    `);
      this.addTemplate('contact', `
      <h3 class="contact">Contact</h3>
      {{link-to 'Home' 'index' id='home-link'}}
      {{link-to 'Self' 'contact' id='self-link'}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#contact-link').text(), 'Jane', 'The link title is correctly resolved');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => controller.set('contactName', 'Joe'));

        assert.equal(this.$('#contact-link').text(), 'Joe', 'The link title is correctly updated when the bound property changes');

        this.runTask(() => controller.set('contactName', 'Robert'));

        assert.equal(this.$('#contact-link').text(), 'Robert', 'The link title is correctly updated when the bound property changes a second time');

        return this.click('#contact-link');
      }).then(() => {
        assert.equal(this.$('h3.contact').length, 1, 'The contact template was rendered');
        assert.equal(this.$('#self-link.active').length, 1, 'The self-link was rendered with active class');
        assert.equal(this.$('#home-link:not(.active)').length, 1, 'The other link was rendered without active class');

        return this.click('#home-link');
      }).then(() => {
        assert.equal(this.$('h3.home').length, 1, 'The index template was rendered');
        assert.equal(this.$('#contact-link').text(), 'Robert', 'The link title is correctly updated when the route changes');
      });
    }

    [`@test The non-block form {{link-to}} helper moves into the named route with context`](assert) {
      assert.expect(5);

      this.router.map(function () {
        this.route('item', { path: '/item/:id' });
      });

      this.add('route:index', _routing.Route.extend({
        model() {
          return [{ id: 'yehuda', name: 'Yehuda Katz' }, { id: 'tom', name: 'Tom Dale' }, { id: 'erik', name: 'Erik Brynroflsson' }];
        }
      }));

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      <ul>
        {{#each model as |person|}}
          <li>
            {{link-to person.name 'item' person id=person.id}}
          </li>
        {{/each}}
      </ul>
    `);
      this.addTemplate('item', `
      <h3 class="item">Item</h3>
      <p>{{model.name}}</p>
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
    `);

      return this.visit('/').then(() => {
        return this.click('#yehuda');
      }).then(() => {
        assert.equal(this.$('h3.item').length, 1, 'The item template was rendered');
        assert.equal(this.$('p').text(), 'Yehuda Katz', 'The name is correct');

        return this.click('#home-link');
      }).then(() => {
        assert.equal(normalizeUrl(this.$('li a#yehuda').attr('href')), '/item/yehuda');
        assert.equal(normalizeUrl(this.$('li a#tom').attr('href')), '/item/tom');
        assert.equal(normalizeUrl(this.$('li a#erik').attr('href')), '/item/erik');
      });
    }

    [`@test The non-block form {{link-to}} performs property lookup`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{link-to 'string' 'index' id='string-link'}}
      {{link-to path foo id='path-link'}}
    `);

      this.add('controller:index', _controller.default.extend({
        foo: 'index'
      }));

      return this.visit('/').then(() => {
        let assertEquality = href => {
          assert.equal(normalizeUrl(this.$('#string-link').attr('href')), '/');
          assert.equal(normalizeUrl(this.$('#path-link').attr('href')), href);
        };

        assertEquality('/');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => controller.set('foo', 'about'));

        assertEquality('/about');
      });
    }

    [`@test The non-block form {{link-to}} protects against XSS`](assert) {
      this.addTemplate('application', `{{link-to display 'index' id='link'}}`);

      this.add('controller:application', _controller.default.extend({
        display: 'blahzorz'
      }));

      return this.visit('/').then(() => {
        assert.equal(this.$('#link').text(), 'blahzorz');

        let controller = this.applicationInstance.lookup('controller:application');
        this.runTask(() => controller.set('display', '<b>BLAMMO</b>'));

        assert.equal(this.$('#link').text(), '<b>BLAMMO</b>');
        assert.equal(this.$('b').length, 0);
      });
    }

    [`@test the {{link-to}} helper throws a useful error if you invoke it wrong`](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('post', { path: 'post/:post_id' });
      });

      this.addTemplate('application', `{{#link-to 'post'}}Post{{/link-to}}`);

      assert.throws(() => {
        this.visit('/');
      }, /(You attempted to define a `\{\{link-to "post"\}\}` but did not pass the parameters required for generating its dynamic segments.|You must provide param `post_id` to `generate`)/);

      return this.runLoopSettled();
    }

    [`@test the {{link-to}} helper does not throw an error if its route has exited`](assert) {
      assert.expect(0);

      this.router.map(function () {
        this.route('post', { path: 'post/:post_id' });
      });

      this.addTemplate('application', `
      {{#link-to 'index' id='home-link'}}Home{{/link-to}}
      {{#link-to 'post' defaultPost id='default-post-link'}}Default Post{{/link-to}}
      {{#if currentPost}}
        {{#link-to 'post' currentPost id='current-post-link'}}Current Post{{/link-to}}
      {{/if}}
    `);

      this.add('controller:application', _controller.default.extend({
        defaultPost: { id: 1 },
        postController: (0, _controller.inject)('post'),
        currentPost: (0, _metal.alias)('postController.model')
      }));

      this.add('controller:post', _controller.default.extend());

      this.add('route:post', _routing.Route.extend({
        model() {
          return { id: 2 };
        },
        serialize(model) {
          return { post_id: model.id };
        }
      }));

      return this.visit('/').then(() => this.click('#default-post-link')).then(() => this.click('#home-link')).then(() => this.click('#current-post-link')).then(() => this.click('#home-link'));
    }

    [`@test {{link-to}} active property respects changing parent route context`](assert) {
      this.router.map(function () {
        this.route('things', { path: '/things/:name' }, function () {
          this.route('other');
        });
      });

      this.addTemplate('application', `
      {{link-to 'OMG' 'things' 'omg' id='omg-link'}}
      {{link-to 'LOL' 'things' 'lol' id='lol-link'}}
    `);

      return this.visit('/things/omg').then(() => {
        shouldBeActive(assert, this.$('#omg-link'));
        shouldNotBeActive(assert, this.$('#lol-link'));

        return this.visit('/things/omg/other');
      }).then(() => {
        shouldBeActive(assert, this.$('#omg-link'));
        shouldNotBeActive(assert, this.$('#lol-link'));
      });
    }

    [`@test {{link-to}} populates href with default query param values even without query-params object`](assert) {
      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo'],
        foo: '123'
      }));

      this.addTemplate('index', `{{#link-to 'index' id='the-link'}}Index{{/link-to}}`);

      return this.visit('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/', 'link has right href');
      });
    }

    [`@test {{link-to}} populates href with default query param values with empty query-params object`](assert) {
      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo'],
        foo: '123'
      }));

      this.addTemplate('index', `
      {{#link-to 'index' (query-params) id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/', 'link has right href');
      });
    }

    [`@test {{link-to}} with only query-params and a block updates when route changes`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.add('controller:application', _controller.default.extend({
        queryParams: ['foo', 'bar'],
        foo: '123',
        bar: 'yes'
      }));

      this.addTemplate('application', `
      {{#link-to (query-params foo='456' bar='NAW') id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/?bar=NAW&foo=456', 'link has right href');

        return this.visit('/about');
      }).then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/about?bar=NAW&foo=456', 'link has right href');
      });
    }

    [`@test Block-less {{link-to}} with only query-params updates when route changes`](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.add('controller:application', _controller.default.extend({
        queryParams: ['foo', 'bar'],
        foo: '123',
        bar: 'yes'
      }));

      this.addTemplate('application', `
      {{link-to "Index" (query-params foo='456' bar='NAW') id='the-link'}}
    `);

      return this.visit('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/?bar=NAW&foo=456', 'link has right href');

        return this.visit('/about');
      }).then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/about?bar=NAW&foo=456', 'link has right href');
      });
    }

    [`@test The {{link-to}} helper can use dynamic params`](assert) {
      this.router.map(function () {
        this.route('foo', { path: 'foo/:some/:thing' });
        this.route('bar', { path: 'bar/:some/:thing/:else' });
      });

      this.add('controller:index', _controller.default.extend({
        init() {
          this._super(...arguments);
          this.dynamicLinkParams = ['foo', 'one', 'two'];
        }
      }));

      this.addTemplate('index', `
      <h3 class="home">Home</h3>
      {{#link-to params=dynamicLinkParams id="dynamic-link"}}Dynamic{{/link-to}}
    `);

      return this.visit('/').then(() => {
        let link = this.$('#dynamic-link');

        assert.equal(link.attr('href'), '/foo/one/two');

        let controller = this.applicationInstance.lookup('controller:index');
        this.runTask(() => {
          controller.set('dynamicLinkParams', ['bar', 'one', 'two', 'three']);
        });

        assert.equal(link.attr('href'), '/bar/one/two/three');
      });
    }

    [`@test GJ: {{link-to}} to a parent root model hook which performs a 'transitionTo' has correct active class #13256`](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('parent', function () {
          this.route('child');
        });
      });

      this.add('route:parent', _routing.Route.extend({
        afterModel() {
          this.transitionTo('parent.child');
        }
      }));

      this.addTemplate('application', `
      {{link-to 'Parent' 'parent' id='parent-link'}}
    `);

      return this.visit('/').then(() => {
        return this.click('#parent-link');
      }).then(() => {
        shouldBeActive(assert, this.$('#parent-link'));
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper - loading states and warnings', class extends _internalTestHelpers.ApplicationTestCase {
    [`@test link-to with null/undefined dynamic parameters are put in a loading state`](assert) {
      assert.expect(19);
      let warningMessage = 'This link-to is in an inactive loading state because at least one of its parameters presently has a null/undefined value, or the provided route name is invalid.';

      this.router.map(function () {
        this.route('thing', { path: '/thing/:thing_id' });
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to destinationRoute routeContext loadingClass='i-am-loading' id='context-link'}}
        string
      {{/link-to}}
      {{#link-to secondRoute loadingClass=loadingClass id='static-link'}}
        string
      {{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        destinationRoute: null,
        routeContext: null,
        loadingClass: 'i-am-loading'
      }));

      this.add('route:about', _routing.Route.extend({
        activate() {
          assert.ok(true, 'About was entered');
        }
      }));

      function assertLinkStatus(link, url) {
        if (url) {
          assert.equal(normalizeUrl(link.attr('href')), url, 'loaded link-to has expected href');
          assert.ok(!link.hasClass('i-am-loading'), 'loaded linkComponent has no loadingClass');
        } else {
          assert.equal(normalizeUrl(link.attr('href')), '#', "unloaded link-to has href='#'");
          assert.ok(link.hasClass('i-am-loading'), 'loading linkComponent has loadingClass');
        }
      }

      let contextLink, staticLink, controller;

      return this.visit('/').then(() => {
        contextLink = this.$('#context-link');
        staticLink = this.$('#static-link');
        controller = this.applicationInstance.lookup('controller:index');

        assertLinkStatus(contextLink);
        assertLinkStatus(staticLink);

        return expectWarning(() => {
          return this.click(contextLink[0]);
        }, warningMessage);
      }).then(() => {
        // Set the destinationRoute (context is still null).
        this.runTask(() => controller.set('destinationRoute', 'thing'));
        assertLinkStatus(contextLink);

        // Set the routeContext to an id
        this.runTask(() => controller.set('routeContext', '456'));
        assertLinkStatus(contextLink, '/thing/456');

        // Test that 0 isn't interpreted as falsy.
        this.runTask(() => controller.set('routeContext', 0));
        assertLinkStatus(contextLink, '/thing/0');

        // Set the routeContext to an object
        this.runTask(() => {
          controller.set('routeContext', { id: 123 });
        });
        assertLinkStatus(contextLink, '/thing/123');

        // Set the destinationRoute back to null.
        this.runTask(() => controller.set('destinationRoute', null));
        assertLinkStatus(contextLink);

        return expectWarning(() => {
          return this.click(staticLink[0]);
        }, warningMessage);
      }).then(() => {
        this.runTask(() => controller.set('secondRoute', 'about'));
        assertLinkStatus(staticLink, '/about');

        // Click the now-active link
        return this.click(staticLink[0]);
      });
    }
  });

  function assertNav(options, callback, assert) {
    let nav = false;

    function check(event) {
      assert.equal(event.defaultPrevented, options.prevented, `expected defaultPrevented=${options.prevented}`);
      nav = true;
      event.preventDefault();
    }

    try {
      document.addEventListener('click', check);
      callback();
    } finally {
      document.removeEventListener('click', check);
      assert.ok(nav, 'Expected a link to be clicked');
    }
  }
});
enifed('ember/tests/helpers/link_to_test/link_to_transitioning_classes_test', ['@ember/-internals/runtime', '@ember/-internals/routing', 'internal-test-helpers'], function (_runtime, _routing, _internalTestHelpers) {
  'use strict';

  function assertHasClass(assert, selector, label) {
    let testLabel = `${selector.attr('id')} should have class ${label}`;

    assert.equal(selector.hasClass(label), true, testLabel);
  }

  function assertHasNoClass(assert, selector, label) {
    let testLabel = `${selector.attr('id')} should not have class ${label}`;

    assert.equal(selector.hasClass(label), false, testLabel);
  }

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper: .transitioning-in .transitioning-out CSS classes', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();

      this.aboutDefer = _runtime.RSVP.defer();
      this.otherDefer = _runtime.RSVP.defer();
      this.newsDefer = _runtime.RSVP.defer();
      let _this = this;

      this.router.map(function () {
        this.route('about');
        this.route('other');
        this.route('news');
      });

      this.add('route:about', _routing.Route.extend({
        model() {
          return _this.aboutDefer.promise;
        }
      }));

      this.add('route:other', _routing.Route.extend({
        model() {
          return _this.otherDefer.promise;
        }
      }));

      this.add('route:news', _routing.Route.extend({
        model() {
          return _this.newsDefer.promise;
        }
      }));

      this.addTemplate('application', `
      {{outlet}}
      {{link-to 'Index' 'index' id='index-link'}}
      {{link-to 'About' 'about' id='about-link'}}
      {{link-to 'Other' 'other' id='other-link'}}
      {{link-to 'News' 'news' activeClass=false id='news-link'}}
    `);
    }

    beforeEach() {
      return this.visit('/');
    }

    afterEach() {
      super.afterEach();
      this.aboutDefer = null;
      this.otherDefer = null;
      this.newsDefer = null;
    }

    ['@test while a transition is underway'](assert) {
      let $index = this.$('#index-link');
      let $about = this.$('#about-link');
      let $other = this.$('#other-link');

      $about.click();

      assertHasClass(assert, $index, 'active');
      assertHasNoClass(assert, $about, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      this.runTask(() => this.aboutDefer.resolve());

      assertHasNoClass(assert, $index, 'active');
      assertHasClass(assert, $about, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');
    }

    ['@test while a transition is underway with activeClass is false'](assert) {
      let $index = this.$('#index-link');
      let $news = this.$('#news-link');
      let $other = this.$('#other-link');

      $news.click();

      assertHasClass(assert, $index, 'active');
      assertHasNoClass(assert, $news, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasClass(assert, $news, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $news, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      this.runTask(() => this.newsDefer.resolve());

      assertHasNoClass(assert, $index, 'active');
      assertHasNoClass(assert, $news, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $news, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $news, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');
    }
  });

  (0, _internalTestHelpers.moduleFor)(`The {{link-to}} helper: .transitioning-in .transitioning-out CSS classes - nested link-to's`, class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();
      this.aboutDefer = _runtime.RSVP.defer();
      this.otherDefer = _runtime.RSVP.defer();
      let _this = this;

      this.router.map(function () {
        this.route('parent-route', function () {
          this.route('about');
          this.route('other');
        });
      });
      this.add('route:parent-route.about', _routing.Route.extend({
        model() {
          return _this.aboutDefer.promise;
        }
      }));

      this.add('route:parent-route.other', _routing.Route.extend({
        model() {
          return _this.otherDefer.promise;
        }
      }));

      this.addTemplate('application', `
      {{outlet}}
      {{#link-to 'index' tagName='li'}}
        {{link-to 'Index' 'index' id='index-link'}}
      {{/link-to}}
      {{#link-to 'parent-route.about' tagName='li'}}
        {{link-to 'About' 'parent-route.about' id='about-link'}}
      {{/link-to}}
      {{#link-to 'parent-route.other' tagName='li'}}
        {{link-to 'Other' 'parent-route.other' id='other-link'}}
      {{/link-to}}
    `);
    }

    beforeEach() {
      return this.visit('/');
    }

    resolveAbout() {
      return this.runTask(() => {
        this.aboutDefer.resolve();
        this.aboutDefer = _runtime.RSVP.defer();
      });
    }

    resolveOther() {
      return this.runTask(() => {
        this.otherDefer.resolve();
        this.otherDefer = _runtime.RSVP.defer();
      });
    }

    teardown() {
      super.teardown();
      this.aboutDefer = null;
      this.otherDefer = null;
    }

    [`@test while a transition is underway with nested link-to's`](assert) {
      // TODO undo changes to this test but currently this test navigates away if navigation
      // outlet is not stable and the second $about.click() is triggered.
      let $about = this.$('#about-link');

      $about.click();

      let $index = this.$('#index-link');
      $about = this.$('#about-link');
      let $other = this.$('#other-link');

      assertHasClass(assert, $index, 'active');
      assertHasNoClass(assert, $about, 'active');
      assertHasNoClass(assert, $about, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      this.resolveAbout();

      $index = this.$('#index-link');
      $about = this.$('#about-link');
      $other = this.$('#other-link');

      assertHasNoClass(assert, $index, 'active');
      assertHasClass(assert, $about, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      $other.click();

      $index = this.$('#index-link');
      $about = this.$('#about-link');
      $other = this.$('#other-link');

      assertHasNoClass(assert, $index, 'active');
      assertHasClass(assert, $about, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $about, 'ember-transitioning-in');
      assertHasClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      this.resolveOther();

      $index = this.$('#index-link');
      $about = this.$('#about-link');
      $other = this.$('#other-link');

      assertHasNoClass(assert, $index, 'active');
      assertHasNoClass(assert, $about, 'active');
      assertHasClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');

      $about.click();

      $index = this.$('#index-link');
      $about = this.$('#about-link');
      $other = this.$('#other-link');

      assertHasNoClass(assert, $index, 'active');
      assertHasNoClass(assert, $about, 'active');
      assertHasClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasClass(assert, $other, 'ember-transitioning-out');

      this.resolveAbout();

      $index = this.$('#index-link');
      $about = this.$('#about-link');
      $other = this.$('#other-link');

      assertHasNoClass(assert, $index, 'active');
      assertHasClass(assert, $about, 'active');
      assertHasNoClass(assert, $other, 'active');

      assertHasNoClass(assert, $index, 'ember-transitioning-in');
      assertHasNoClass(assert, $about, 'ember-transitioning-in');
      assertHasNoClass(assert, $other, 'ember-transitioning-in');

      assertHasNoClass(assert, $index, 'ember-transitioning-out');
      assertHasNoClass(assert, $about, 'ember-transitioning-out');
      assertHasNoClass(assert, $other, 'ember-transitioning-out');
    }
  });
});
enifed('ember/tests/helpers/link_to_test/link_to_with_query_params_test', ['@ember/controller', '@ember/-internals/runtime', '@ember/-internals/routing', 'internal-test-helpers'], function (_controller, _runtime, _routing, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('The {{link-to}} helper: invoking with query params', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();
      let indexProperties = {
        foo: '123',
        bar: 'abc'
      };
      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo', 'bar', 'abool'],
        foo: indexProperties.foo,
        bar: indexProperties.bar,
        boundThing: 'OMG',
        abool: true
      }));
      this.add('controller:about', _controller.default.extend({
        queryParams: ['baz', 'bat'],
        baz: 'alex',
        bat: 'borf'
      }));
      this.indexProperties = indexProperties;
    }

    shouldNotBeActive(assert, selector) {
      this.checkActive(assert, selector, false);
    }

    shouldBeActive(assert, selector) {
      this.checkActive(assert, selector, true);
    }

    getController(name) {
      return this.applicationInstance.lookup(`controller:${name}`);
    }

    checkActive(assert, selector, active) {
      let classList = this.$(selector)[0].className;
      assert.equal(classList.indexOf('active') > -1, active, selector + ' active should be ' + active.toString());
    }

    [`@test doesn't update controller QP properties on current route when invoked`](assert) {
      this.addTemplate('index', `
      {{#link-to 'index' id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        this.click('#the-link');
        let indexController = this.getController('index');

        assert.deepEqual(indexController.getProperties('foo', 'bar'), this.indexProperties, 'controller QP properties do not update');
      });
    }

    [`@test doesn't update controller QP properties on current route when invoked (empty query-params obj)`](assert) {
      this.addTemplate('index', `
      {{#link-to 'index' (query-params) id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        this.click('#the-link');
        let indexController = this.getController('index');

        assert.deepEqual(indexController.getProperties('foo', 'bar'), this.indexProperties, 'controller QP properties do not update');
      });
    }

    [`@test doesn't update controller QP properties on current route when invoked (empty query-params obj, inferred route)`](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params) id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        this.click('#the-link');
        let indexController = this.getController('index');

        assert.deepEqual(indexController.getProperties('foo', 'bar'), this.indexProperties, 'controller QP properties do not update');
      });
    }

    ['@test updates controller QP properties on current route when invoked'](assert) {
      this.addTemplate('index', `
      {{#link-to 'index' (query-params foo='456') id="the-link"}}
        Index
      {{/link-to}}
    `);

      return this.visit('/').then(() => {
        this.click('#the-link');
        let indexController = this.getController('index');

        assert.deepEqual(indexController.getProperties('foo', 'bar'), { foo: '456', bar: 'abc' }, 'controller QP properties updated');
      });
    }

    ['@test updates controller QP properties on current route when invoked (inferred route)'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params foo='456') id="the-link"}}
        Index
      {{/link-to}}
    `);

      return this.visit('/').then(() => {
        this.click('#the-link');
        let indexController = this.getController('index');

        assert.deepEqual(indexController.getProperties('foo', 'bar'), { foo: '456', bar: 'abc' }, 'controller QP properties updated');
      });
    }

    ['@test updates controller QP properties on other route after transitioning to that route'](assert) {
      this.router.map(function () {
        this.route('about');
      });

      this.addTemplate('index', `
      {{#link-to 'about' (query-params baz='lol') id='the-link'}}
        About
      {{/link-to}}
    `);

      return this.visit('/').then(() => {
        let theLink = this.$('#the-link');
        assert.equal(theLink.attr('href'), '/about?baz=lol');

        this.runTask(() => this.click('#the-link'));

        let aboutController = this.getController('about');

        assert.deepEqual(aboutController.getProperties('baz', 'bat'), { baz: 'lol', bat: 'borf' }, 'about controller QP properties updated');
      });
    }

    ['@test supplied QP properties can be bound'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params foo=boundThing) id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        let indexController = this.getController('index');
        let theLink = this.$('#the-link');

        assert.equal(theLink.attr('href'), '/?foo=OMG');

        this.runTask(() => indexController.set('boundThing', 'ASL'));

        assert.equal(theLink.attr('href'), '/?foo=ASL');
      });
    }

    ['@test supplied QP properties can be bound (booleans)'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params abool=boundThing) id='the-link'}}
        Index
      {{/link-to}}
    `);

      return this.visit('/').then(() => {
        let indexController = this.getController('index');
        let theLink = this.$('#the-link');

        assert.equal(theLink.attr('href'), '/?abool=OMG');

        this.runTask(() => indexController.set('boundThing', false));

        assert.equal(theLink.attr('href'), '/?abool=false');

        this.click('#the-link');

        assert.deepEqual(indexController.getProperties('foo', 'bar', 'abool'), { foo: '123', bar: 'abc', abool: false }, 'bound bool QP properties update');
      });
    }
    ['@test href updates when unsupplied controller QP props change'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params foo='lol') id='the-link'}}Index{{/link-to}}
    `);

      return this.visit('/').then(() => {
        let indexController = this.getController('index');
        let theLink = this.$('#the-link');

        assert.equal(theLink.attr('href'), '/?foo=lol');

        this.runTask(() => indexController.set('bar', 'BORF'));

        assert.equal(theLink.attr('href'), '/?bar=BORF&foo=lol');

        this.runTask(() => indexController.set('foo', 'YEAH'));

        assert.equal(theLink.attr('href'), '/?bar=BORF&foo=lol');
      });
    }

    ['@test The {{link-to}} with only query params always transitions to the current route with the query params applied'](assert) {
      // Test harness for bug #12033
      this.addTemplate('cars', `
      {{#link-to 'cars.create' id='create-link'}}Create new car{{/link-to}}
      {{#link-to (query-params page='2') id='page2-link'}}Page 2{{/link-to}}
      {{outlet}}
    `);
      this.addTemplate('cars.create', `{{#link-to 'cars' id='close-link'}}Close create form{{/link-to}}`);

      this.router.map(function () {
        this.route('cars', function () {
          this.route('create');
        });
      });

      this.add('controller:cars', _controller.default.extend({
        queryParams: ['page'],
        page: 1
      }));

      return this.visit('/cars/create').then(() => {
        let router = this.appRouter;
        let carsController = this.getController('cars');

        assert.equal(router.currentRouteName, 'cars.create');

        this.runTask(() => this.click('#close-link'));

        assert.equal(router.currentRouteName, 'cars.index');
        assert.equal(router.get('url'), '/cars');
        assert.equal(carsController.get('page'), 1, 'The page query-param is 1');

        this.runTask(() => this.click('#page2-link'));

        assert.equal(router.currentRouteName, 'cars.index', 'The active route is still cars');
        assert.equal(router.get('url'), '/cars?page=2', 'The url has been updated');
        assert.equal(carsController.get('page'), 2, 'The query params have been updated');
      });
    }

    ['@test the {{link-to}} applies activeClass when query params are not changed'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params foo='cat') id='cat-link'}}Index{{/link-to}}
      {{#link-to (query-params foo='dog') id='dog-link'}}Index{{/link-to}}
      {{#link-to 'index' id='change-nothing'}}Index{{/link-to}}
    `);
      this.addTemplate('search', `
      {{#link-to (query-params search='same') id='same-search'}}Index{{/link-to}}
      {{#link-to (query-params search='change') id='change-search'}}Index{{/link-to}}
      {{#link-to (query-params search='same' archive=true) id='same-search-add-archive'}}Index{{/link-to}}
      {{#link-to (query-params archive=true) id='only-add-archive'}}Index{{/link-to}}
      {{#link-to (query-params search='same' archive=true) id='both-same'}}Index{{/link-to}}
      {{#link-to (query-params search='different' archive=true) id='change-one'}}Index{{/link-to}}
      {{#link-to (query-params search='different' archive=false) id='remove-one'}}Index{{/link-to}}
      {{outlet}}
    `);
      this.addTemplate('search.results', `
      {{#link-to (query-params sort='title') id='same-sort-child-only'}}Index{{/link-to}}
      {{#link-to (query-params search='same') id='same-search-parent-only'}}Index{{/link-to}}
      {{#link-to (query-params search='change') id='change-search-parent-only'}}Index{{/link-to}}
      {{#link-to (query-params search='same' sort='title') id='same-search-same-sort-child-and-parent'}}Index{{/link-to}}
      {{#link-to (query-params search='same' sort='author') id='same-search-different-sort-child-and-parent'}}Index{{/link-to}}
      {{#link-to (query-params search='change' sort='title') id='change-search-same-sort-child-and-parent'}}Index{{/link-to}}
      {{#link-to (query-params foo='dog') id='dog-link'}}Index{{/link-to}}
    `);

      this.router.map(function () {
        this.route('search', function () {
          this.route('results');
        });
      });

      this.add('controller:search', _controller.default.extend({
        queryParams: ['search', 'archive'],
        search: '',
        archive: false
      }));

      this.add('controller:search.results', _controller.default.extend({
        queryParams: ['sort', 'showDetails'],
        sort: 'title',
        showDetails: true
      }));

      return this.visit('/').then(() => {
        this.shouldNotBeActive(assert, '#cat-link');
        this.shouldNotBeActive(assert, '#dog-link');

        return this.visit('/?foo=cat');
      }).then(() => {
        this.shouldBeActive(assert, '#cat-link');
        this.shouldNotBeActive(assert, '#dog-link');

        return this.visit('/?foo=dog');
      }).then(() => {
        this.shouldBeActive(assert, '#dog-link');
        this.shouldNotBeActive(assert, '#cat-link');
        this.shouldBeActive(assert, '#change-nothing');

        return this.visit('/search?search=same');
      }).then(() => {
        this.shouldBeActive(assert, '#same-search');
        this.shouldNotBeActive(assert, '#change-search');
        this.shouldNotBeActive(assert, '#same-search-add-archive');
        this.shouldNotBeActive(assert, '#only-add-archive');
        this.shouldNotBeActive(assert, '#remove-one');

        return this.visit('/search?search=same&archive=true');
      }).then(() => {
        this.shouldBeActive(assert, '#both-same');
        this.shouldNotBeActive(assert, '#change-one');

        return this.visit('/search/results?search=same&sort=title&showDetails=true');
      }).then(() => {
        this.shouldBeActive(assert, '#same-sort-child-only');
        this.shouldBeActive(assert, '#same-search-parent-only');
        this.shouldNotBeActive(assert, '#change-search-parent-only');
        this.shouldBeActive(assert, '#same-search-same-sort-child-and-parent');
        this.shouldNotBeActive(assert, '#same-search-different-sort-child-and-parent');
        this.shouldNotBeActive(assert, '#change-search-same-sort-child-and-parent');
      });
    }

    ['@test the {{link-to}} applies active class when query-param is a number'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params page=pageNumber) id='page-link'}}
        Index
      {{/link-to}}
    `);
      this.add('controller:index', _controller.default.extend({
        queryParams: ['page'],
        page: 1,
        pageNumber: 5
      }));

      return this.visit('/').then(() => {
        this.shouldNotBeActive(assert, '#page-link');
        return this.visit('/?page=5');
      }).then(() => {
        this.shouldBeActive(assert, '#page-link');
      });
    }

    ['@test the {{link-to}} applies active class when query-param is an array'](assert) {
      this.addTemplate('index', `
      {{#link-to (query-params pages=pagesArray) id='array-link'}}Index{{/link-to}}
      {{#link-to (query-params pages=biggerArray) id='bigger-link'}}Index{{/link-to}}
      {{#link-to (query-params pages=emptyArray) id='empty-link'}}Index{{/link-to}}
    `);

      this.add('controller:index', _controller.default.extend({
        queryParams: ['pages'],
        pages: [],
        pagesArray: [1, 2],
        biggerArray: [1, 2, 3],
        emptyArray: []
      }));

      return this.visit('/').then(() => {
        this.shouldNotBeActive(assert, '#array-link');

        return this.visit('/?pages=%5B1%2C2%5D');
      }).then(() => {
        this.shouldBeActive(assert, '#array-link');
        this.shouldNotBeActive(assert, '#bigger-link');
        this.shouldNotBeActive(assert, '#empty-link');

        return this.visit('/?pages=%5B2%2C1%5D');
      }).then(() => {
        this.shouldNotBeActive(assert, '#array-link');
        this.shouldNotBeActive(assert, '#bigger-link');
        this.shouldNotBeActive(assert, '#empty-link');

        return this.visit('/?pages=%5B1%2C2%2C3%5D');
      }).then(() => {
        this.shouldBeActive(assert, '#bigger-link');
        this.shouldNotBeActive(assert, '#array-link');
        this.shouldNotBeActive(assert, '#empty-link');
      });
    }
    ['@test the {{link-to}} helper applies active class to the parent route'](assert) {
      this.router.map(function () {
        this.route('parent', function () {
          this.route('child');
        });
      });

      this.addTemplate('application', `
      {{#link-to 'parent' id='parent-link'}}Parent{{/link-to}}
      {{#link-to 'parent.child' id='parent-child-link'}}Child{{/link-to}}
      {{#link-to 'parent' (query-params foo=cat) id='parent-link-qp'}}Parent{{/link-to}}
      {{outlet}}
    `);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['foo'],
        foo: 'bar'
      }));

      return this.visit('/').then(() => {
        this.shouldNotBeActive(assert, '#parent-link');
        this.shouldNotBeActive(assert, '#parent-child-link');
        this.shouldNotBeActive(assert, '#parent-link-qp');
        return this.visit('/parent/child?foo=dog');
      }).then(() => {
        this.shouldBeActive(assert, '#parent-link');
        this.shouldNotBeActive(assert, '#parent-link-qp');
      });
    }

    ['@test The {{link-to}} helper disregards query-params in activeness computation when current-when is specified'](assert) {
      let appLink;

      this.router.map(function () {
        this.route('parent');
      });
      this.addTemplate('application', `
      {{#link-to 'parent' (query-params page=1) current-when='parent' id='app-link'}}
        Parent
      {{/link-to}}
      {{outlet}}
    `);
      this.addTemplate('parent', `
      {{#link-to 'parent' (query-params page=1) current-when='parent' id='parent-link'}}
        Parent
      {{/link-to}}
      {{outlet}}
    `);
      this.add('controller:parent', _controller.default.extend({
        queryParams: ['page'],
        page: 1
      }));

      return this.visit('/').then(() => {
        appLink = this.$('#app-link');

        assert.equal(appLink.attr('href'), '/parent');
        this.shouldNotBeActive(assert, '#app-link');

        return this.visit('/parent?page=2');
      }).then(() => {
        appLink = this.$('#app-link');
        let router = this.appRouter;

        assert.equal(appLink.attr('href'), '/parent');
        this.shouldBeActive(assert, '#app-link');
        assert.equal(this.$('#parent-link').attr('href'), '/parent');
        this.shouldBeActive(assert, '#parent-link');

        let parentController = this.getController('parent');

        assert.equal(parentController.get('page'), 2);

        this.runTask(() => parentController.set('page', 3));

        assert.equal(router.get('location.path'), '/parent?page=3');
        this.shouldBeActive(assert, '#app-link');
        this.shouldBeActive(assert, '#parent-link');

        this.runTask(() => this.click('#app-link'));

        assert.equal(router.get('location.path'), '/parent');
      });
    }

    ['@test link-to default query params while in active transition regression test'](assert) {
      this.router.map(function () {
        this.route('foos');
        this.route('bars');
      });
      let foos = _runtime.RSVP.defer();
      let bars = _runtime.RSVP.defer();

      this.addTemplate('application', `
      {{link-to 'Foos' 'foos' id='foos-link'}}
      {{link-to 'Baz Foos' 'foos' (query-params baz=true) id='baz-foos-link'}}
      {{link-to 'Quux Bars' 'bars' (query-params quux=true) id='bars-link'}}
    `);
      this.add('controller:foos', _controller.default.extend({
        queryParams: ['status'],
        baz: false
      }));
      this.add('route:foos', _routing.Route.extend({
        model() {
          return foos.promise;
        }
      }));
      this.add('controller:bars', _controller.default.extend({
        queryParams: ['status'],
        quux: false
      }));
      this.add('route:bars', _routing.Route.extend({
        model() {
          return bars.promise;
        }
      }));

      return this.visit('/').then(() => {
        let router = this.appRouter;
        let foosLink = this.$('#foos-link');
        let barsLink = this.$('#bars-link');
        let bazLink = this.$('#baz-foos-link');

        assert.equal(foosLink.attr('href'), '/foos');
        assert.equal(bazLink.attr('href'), '/foos?baz=true');
        assert.equal(barsLink.attr('href'), '/bars?quux=true');
        assert.equal(router.get('location.path'), '/');
        this.shouldNotBeActive(assert, '#foos-link');
        this.shouldNotBeActive(assert, '#baz-foos-link');
        this.shouldNotBeActive(assert, '#bars-link');

        this.runTask(() => barsLink.click());
        this.shouldNotBeActive(assert, '#bars-link');

        this.runTask(() => foosLink.click());
        this.shouldNotBeActive(assert, '#foos-link');

        this.runTask(() => foos.resolve());

        assert.equal(router.get('location.path'), '/foos');
        this.shouldBeActive(assert, '#foos-link');
      });
    }

    [`@test the {{link-to}} helper throws a useful error if you invoke it wrong`](assert) {
      assert.expect(1);

      this.addTemplate('application', `{{#link-to id='the-link'}}Index{{/link-to}}`);

      expectAssertion(() => {
        this.visit('/');
      }, /You must provide one or more parameters to the link-to component/);

      return this.runLoopSettled();
    }
  });
});
enifed('ember/tests/homepage_example_test', ['@ember/-internals/routing', '@ember/-internals/metal', '@ember/-internals/runtime', 'internal-test-helpers'], function (_routing, _metal, _runtime, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('The example renders correctly', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Render index template into application outlet'](assert) {
      this.addTemplate('application', '{{outlet}}');
      this.addTemplate('index', '<h1>People</h1><ul>{{#each model as |person|}}<li>Hello, <b>{{person.fullName}}</b>!</li>{{/each}}</ul>');

      let Person = _runtime.Object.extend({
        firstName: null,
        lastName: null,
        fullName: (0, _metal.computed)('firstName', 'lastName', function () {
          return `${this.get('firstName')} ${this.get('lastName')}`;
        })
      });

      this.add('route:index', _routing.Route.extend({
        model() {
          return (0, _runtime.A)([Person.create({ firstName: 'Tom', lastName: 'Dale' }), Person.create({ firstName: 'Yehuda', lastName: 'Katz' })]);
        }
      }));

      return this.visit('/').then(() => {
        let $ = this.$();

        assert.equal($.findAll('h1').text(), 'People');
        assert.equal($.findAll('li').length, 2);
        assert.equal($.findAll('li:nth-of-type(1)').text(), 'Hello, Tom Dale!');
        assert.equal($.findAll('li:nth-of-type(2)').text(), 'Hello, Yehuda Katz!');
      });
    }
  });
});
enifed('ember/tests/integration/multiple-app-test', ['internal-test-helpers', '@ember/application', '@ember/-internals/glimmer', '@ember/-internals/owner', '@ember/polyfills', 'rsvp'], function (_internalTestHelpers, _application, _glimmer, _owner, _polyfills, _rsvp) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('View Integration', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      document.getElementById('qunit-fixture').innerHTML = `
      <div id="one"></div>
      <div id="two"></div>
    `;
      super();
      this.runTask(() => {
        this.createSecondApplication();
      });
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        rootElement: '#one',
        router: null
      });
    }

    createSecondApplication(options) {
      let { applicationOptions } = this;
      let secondApplicationOptions = { rootElement: '#two' };
      let myOptions = (0, _polyfills.assign)(applicationOptions, secondApplicationOptions, options);
      this.secondApp = _application.default.create(myOptions);
      this.secondResolver = this.secondApp.__registry__.resolver;
      return this.secondApp;
    }

    teardown() {
      super.teardown();

      if (this.secondApp) {
        this.runTask(() => {
          this.secondApp.destroy();
        });
      }
    }

    addFactoriesToResolver(actions, resolver) {
      resolver.add('component:special-button', _glimmer.Component.extend({
        actions: {
          doStuff() {
            let rootElement = (0, _owner.getOwner)(this).application.rootElement;
            actions.push(rootElement);
          }
        }
      }));

      resolver.add('template:index', this.compile(`
        <h1>Node 1</h1>{{special-button}}
      `, {
        moduleName: 'my-app/templates/index.hbs'
      }));
      resolver.add('template:components/special-button', this.compile(`
        <button class='do-stuff' {{action 'doStuff'}}>Button</button>
      `, {
        moduleName: 'my-app/templates/components/special-button.hbs'
      }));
    }

    [`@test booting multiple applications can properly handle events`](assert) {
      let actions = [];
      this.addFactoriesToResolver(actions, this.resolver);
      this.addFactoriesToResolver(actions, this.secondResolver);

      return (0, _rsvp.resolve)().then(() => this.application.visit('/')).then(() => this.secondApp.visit('/')).then(() => {
        document.querySelector('#two .do-stuff').click();
        document.querySelector('#one .do-stuff').click();

        assert.deepEqual(actions, ['#two', '#one']);
      });
    }
  });
});
enifed('ember/tests/production_build_test', ['@ember/debug', 'internal-test-helpers'], function (_debug, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('production builds', class extends _internalTestHelpers.AbstractTestCase {
    ['@test assert does not throw in production builds'](assert) {
      if (!true /* DEBUG */) {
          assert.expect(1);

          try {
            true && !false && (0, _debug.assert)('Should not throw');

            assert.ok(true, 'Ember.assert did not throw');
          } catch (e) {
            assert.ok(false, `Expected assert not to throw but it did: ${e.message}`);
          }
        } else {
        assert.expect(0);
      }
    }

    ['@test runInDebug does not run the callback in production builds'](assert) {
      if (!true /* DEBUG */) {
          let fired = false;
          (0, _debug.runInDebug)(() => fired = true);

          assert.equal(fired, false, 'runInDebug callback should not be ran');
        } else {
        assert.expect(0);
      }
    }
  });
});
enifed('ember/tests/reexports_test', ['ember/index', '@ember/canary-features', 'internal-test-helpers', '@ember/-internals/views'], function (_index, _canaryFeatures, _internalTestHelpers, _views) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('ember reexports', class extends _internalTestHelpers.AbstractTestCase {
    [`@test Ember exports correctly`](assert) {
      allExports.forEach(reexport => {
        let [path, moduleId, exportName] = reexport;

        // default path === exportName if none present
        if (!exportName) {
          exportName = path;
        }

        (0, _internalTestHelpers.confirmExport)(_index.default, assert, path, moduleId, exportName, `Ember.${path} exports correctly`);
      });
    }

    ['@test Ember.String.isHTMLSafe exports correctly'](assert) {
      (0, _internalTestHelpers.confirmExport)(_index.default, assert, 'String.isHTMLSafe', '@ember/-internals/glimmer', 'isHTMLSafe');
    }

    ['@test Ember.EXTEND_PROTOTYPES is present (but deprecated)'](assert) {
      expectDeprecation(() => {
        assert.strictEqual(_index.default.ENV.EXTEND_PROTOTYPES, _index.default.EXTEND_PROTOTYPES, 'Ember.EXTEND_PROTOTYPES exists');
      }, /EXTEND_PROTOTYPES is deprecated/);
    }

    '@test Ember.FEATURES is exported'(assert) {
      for (let feature in _canaryFeatures.FEATURES) {
        assert.equal(_index.default.FEATURES[feature], _canaryFeatures.FEATURES[feature], 'Ember.FEATURES contains ${feature} with correct value');
      }
    }
  });

  let allExports = [
  // @ember/-internals/environment
  ['ENV', '@ember/-internals/environment', { get: 'getENV' }], ['lookup', '@ember/-internals/environment', { get: 'getLookup', set: 'setLookup' }], ['getOwner', '@ember/application', 'getOwner'], ['setOwner', '@ember/application', 'setOwner'], ['assign', '@ember/polyfills'],

  // @ember/-internals/utils
  ['GUID_KEY', '@ember/-internals/utils'], ['uuid', '@ember/-internals/utils'], ['generateGuid', '@ember/-internals/utils'], ['guidFor', '@ember/-internals/utils'], ['inspect', '@ember/-internals/utils'], ['makeArray', '@ember/-internals/utils'], ['canInvoke', '@ember/-internals/utils'], ['tryInvoke', '@ember/-internals/utils'], ['wrap', '@ember/-internals/utils'], ['NAME_KEY', '@ember/-internals/utils'],

  // @ember/-internals/container
  ['Registry', '@ember/-internals/container', 'Registry'], ['Container', '@ember/-internals/container', 'Container'],

  // @ember/debug
  ['deprecateFunc', '@ember/debug'], ['deprecate', '@ember/debug'], ['assert', '@ember/debug'], ['warn', '@ember/debug'], ['debug', '@ember/debug'], ['runInDebug', '@ember/debug'], ['Debug.registerDeprecationHandler', '@ember/debug', 'registerDeprecationHandler'], ['Debug.registerWarnHandler', '@ember/debug', 'registerWarnHandler'], ['Error', '@ember/error', 'default'],

  // @ember/-internals/metal
  ['computed', '@ember/-internals/metal', '_globalsComputed'], ['computed.alias', '@ember/-internals/metal', 'alias'], ['ComputedProperty', '@ember/-internals/metal'], ['cacheFor', '@ember/-internals/metal', 'getCachedValueFor'], ['merge', '@ember/polyfills'], ['instrument', '@ember/instrumentation'], ['subscribe', '@ember/instrumentation', 'subscribe'], ['Instrumentation.instrument', '@ember/instrumentation', 'instrument'], ['Instrumentation.subscribe', '@ember/instrumentation', 'subscribe'], ['Instrumentation.unsubscribe', '@ember/instrumentation', 'unsubscribe'], ['Instrumentation.reset', '@ember/instrumentation', 'reset'], ['testing', '@ember/debug', { get: 'isTesting', set: 'setTesting' }], ['onerror', '@ember/-internals/error-handling', { get: 'getOnerror', set: 'setOnerror' }], ['FEATURES.isEnabled', '@ember/canary-features', 'isEnabled'], ['meta', '@ember/-internals/meta'], ['get', '@ember/-internals/metal'], ['set', '@ember/-internals/metal'], ['_getPath', '@ember/-internals/metal'], ['getWithDefault', '@ember/-internals/metal'], ['trySet', '@ember/-internals/metal'], ['_Cache', '@ember/-internals/utils', 'Cache'], ['on', '@ember/-internals/metal'], ['addListener', '@ember/-internals/metal'], ['removeListener', '@ember/-internals/metal'], ['sendEvent', '@ember/-internals/metal'], ['hasListeners', '@ember/-internals/metal'], ['isNone', '@ember/-internals/metal'], ['isEmpty', '@ember/-internals/metal'], ['isBlank', '@ember/-internals/metal'], ['isPresent', '@ember/-internals/metal'], ['_Backburner', 'backburner', 'default'], ['run', '@ember/runloop', '_globalsRun'], ['run.backburner', '@ember/runloop', 'backburner'], ['run.begin', '@ember/runloop', 'begin'], ['run.bind', '@ember/runloop', 'bind'], ['run.cancel', '@ember/runloop', 'cancel'], ['run.debounce', '@ember/runloop', 'debounce'], ['run.end', '@ember/runloop', 'end'], ['run.hasScheduledTimers', '@ember/runloop', 'hasScheduledTimers'], ['run.join', '@ember/runloop', 'join'], ['run.later', '@ember/runloop', 'later'], ['run.next', '@ember/runloop', 'next'], ['run.once', '@ember/runloop', 'once'], ['run.schedule', '@ember/runloop', 'schedule'], ['run.scheduleOnce', '@ember/runloop', 'scheduleOnce'], ['run.throttle', '@ember/runloop', 'throttle'], ['run.currentRunLoop', '@ember/runloop', { get: 'getCurrentRunLoop' }], ['run.cancelTimers', '@ember/runloop', 'cancelTimers'], ['notifyPropertyChange', '@ember/-internals/metal'], ['overrideChains', '@ember/-internals/metal'], ['beginPropertyChanges', '@ember/-internals/metal'], ['endPropertyChanges', '@ember/-internals/metal'], ['changeProperties', '@ember/-internals/metal'], ['platform.defineProperty', null, { value: true }], ['platform.hasPropertyAccessors', null, { value: true }], ['defineProperty', '@ember/-internals/metal'], ['watchKey', '@ember/-internals/metal'], ['unwatchKey', '@ember/-internals/metal'], ['removeChainWatcher', '@ember/-internals/metal'], ['_ChainNode', '@ember/-internals/metal', 'ChainNode'], ['finishChains', '@ember/-internals/metal'], ['watchPath', '@ember/-internals/metal'], ['unwatchPath', '@ember/-internals/metal'], ['watch', '@ember/-internals/metal'], ['isWatching', '@ember/-internals/metal'], ['unwatch', '@ember/-internals/metal'], ['destroy', '@ember/-internals/meta', 'deleteMeta'], ['libraries', '@ember/-internals/metal'], ['OrderedSet', '@ember/map/lib/ordered-set', 'default'], ['Map', '@ember/map', 'default'], ['MapWithDefault', '@ember/map/with-default', 'default'], ['getProperties', '@ember/-internals/metal'], ['setProperties', '@ember/-internals/metal'], ['expandProperties', '@ember/-internals/metal'], ['addObserver', '@ember/-internals/metal'], ['removeObserver', '@ember/-internals/metal'], ['aliasMethod', '@ember/-internals/metal'], ['observer', '@ember/-internals/metal'], ['mixin', '@ember/-internals/metal'], ['Mixin', '@ember/-internals/metal'],

  // @ember/-internals/console
  ['Logger', '@ember/-internals/console', 'default'],

  // @ember/-internals/views
  !_views.jQueryDisabled && ['$', '@ember/-internals/views', 'jQuery'], ['ViewUtils.isSimpleClick', '@ember/-internals/views', 'isSimpleClick'], ['ViewUtils.getViewElement', '@ember/-internals/views', 'getViewElement'], ['ViewUtils.getViewBounds', '@ember/-internals/views', 'getViewBounds'], ['ViewUtils.getViewClientRects', '@ember/-internals/views', 'getViewClientRects'], ['ViewUtils.getViewBoundingClientRect', '@ember/-internals/views', 'getViewBoundingClientRect'], ['ViewUtils.getRootViews', '@ember/-internals/views', 'getRootViews'], ['ViewUtils.getChildViews', '@ember/-internals/views', 'getChildViews'], ['ViewUtils.isSerializationFirstNode', '@ember/-internals/glimmer', 'isSerializationFirstNode'], ['TextSupport', '@ember/-internals/views'], ['ComponentLookup', '@ember/-internals/views'], ['EventDispatcher', '@ember/-internals/views'],

  // @ember/-internals/glimmer
  ['Component', '@ember/-internals/glimmer', 'Component'], ['Helper', '@ember/-internals/glimmer', 'Helper'], ['Helper.helper', '@ember/-internals/glimmer', 'helper'], ['Checkbox', '@ember/-internals/glimmer', 'Checkbox'], ['LinkComponent', '@ember/-internals/glimmer', 'LinkComponent'], ['TextArea', '@ember/-internals/glimmer', 'TextArea'], ['TextField', '@ember/-internals/glimmer', 'TextField'], ['TEMPLATES', '@ember/-internals/glimmer', { get: 'getTemplates', set: 'setTemplates' }], ['Handlebars.template', '@ember/-internals/glimmer', 'template'], ['HTMLBars.template', '@ember/-internals/glimmer', 'template'], ['Handlebars.Utils.escapeExpression', '@ember/-internals/glimmer', 'escapeExpression'], ['String.htmlSafe', '@ember/-internals/glimmer', 'htmlSafe'], ['_setComponentManager', '@ember/-internals/glimmer', 'setComponentManager'], ['_componentManagerCapabilities', '@ember/-internals/glimmer', 'capabilities'],

  // @ember/-internals/runtime
  ['A', '@ember/-internals/runtime'], ['_RegistryProxyMixin', '@ember/-internals/runtime', 'RegistryProxyMixin'], ['_ContainerProxyMixin', '@ember/-internals/runtime', 'ContainerProxyMixin'], ['Object', '@ember/-internals/runtime'], ['String.loc', '@ember/string', 'loc'], ['String.w', '@ember/string', 'w'], ['String.dasherize', '@ember/string', 'dasherize'], ['String.decamelize', '@ember/string', 'decamelize'], ['String.camelize', '@ember/string', 'camelize'], ['String.classify', '@ember/string', 'classify'], ['String.underscore', '@ember/string', 'underscore'], ['String.capitalize', '@ember/string', 'capitalize'], ['compare', '@ember/-internals/runtime'], ['copy', '@ember/-internals/runtime'], ['isEqual', '@ember/-internals/runtime'], ['inject.controller', '@ember/controller', 'inject'], ['inject.service', '@ember/service', 'inject'], ['Array', '@ember/-internals/runtime'], ['Comparable', '@ember/-internals/runtime'], ['Namespace', '@ember/-internals/runtime'], ['Enumerable', '@ember/-internals/runtime'], ['ArrayProxy', '@ember/-internals/runtime'], ['ObjectProxy', '@ember/-internals/runtime'], ['ActionHandler', '@ember/-internals/runtime'], ['CoreObject', '@ember/-internals/runtime'], ['NativeArray', '@ember/-internals/runtime'], ['Copyable', '@ember/-internals/runtime'], ['MutableEnumerable', '@ember/-internals/runtime'], ['MutableArray', '@ember/-internals/runtime'], ['TargetActionSupport', '@ember/-internals/runtime'], ['Evented', '@ember/-internals/runtime'], ['PromiseProxyMixin', '@ember/-internals/runtime'], ['Observable', '@ember/-internals/runtime'], ['typeOf', '@ember/-internals/runtime'], ['isArray', '@ember/-internals/runtime'], ['Object', '@ember/-internals/runtime'], ['onLoad', '@ember/application'], ['runLoadHooks', '@ember/application'], ['Controller', '@ember/controller', 'default'], ['ControllerMixin', '@ember/controller/lib/controller_mixin', 'default'], ['Service', '@ember/service', 'default'], ['_ProxyMixin', '@ember/-internals/runtime'], ['RSVP', '@ember/-internals/runtime'], ['STRINGS', '@ember/string', { get: '_getStrings', set: '_setStrings' }], ['BOOTED', '@ember/-internals/metal', { get: 'isNamespaceSearchDisabled', set: 'setNamespaceSearchDisabled' }], ['computed.empty', '@ember/object/computed', 'empty'], ['computed.notEmpty', '@ember/object/computed', 'notEmpty'], ['computed.none', '@ember/object/computed', 'none'], ['computed.not', '@ember/object/computed', 'not'], ['computed.bool', '@ember/object/computed', 'bool'], ['computed.match', '@ember/object/computed', 'match'], ['computed.equal', '@ember/object/computed', 'equal'], ['computed.gt', '@ember/object/computed', 'gt'], ['computed.gte', '@ember/object/computed', 'gte'], ['computed.lt', '@ember/object/computed', 'lt'], ['computed.lte', '@ember/object/computed', 'lte'], ['computed.oneWay', '@ember/object/computed', 'oneWay'], ['computed.reads', '@ember/object/computed', 'oneWay'], ['computed.readOnly', '@ember/object/computed', 'readOnly'], ['computed.deprecatingAlias', '@ember/object/computed', 'deprecatingAlias'], ['computed.and', '@ember/object/computed', 'and'], ['computed.or', '@ember/object/computed', 'or'], ['computed.sum', '@ember/object/computed', 'sum'], ['computed.min', '@ember/object/computed', 'min'], ['computed.max', '@ember/object/computed', 'max'], ['computed.map', '@ember/object/computed', 'map'], ['computed.sort', '@ember/object/computed', 'sort'], ['computed.setDiff', '@ember/object/computed', 'setDiff'], ['computed.mapBy', '@ember/object/computed', 'mapBy'], ['computed.filter', '@ember/object/computed', 'filter'], ['computed.filterBy', '@ember/object/computed', 'filterBy'], ['computed.uniq', '@ember/object/computed', 'uniq'], ['computed.uniqBy', '@ember/object/computed', 'uniqBy'], ['computed.union', '@ember/object/computed', 'union'], ['computed.intersect', '@ember/object/computed', 'intersect'], ['computed.collect', '@ember/object/computed', 'collect'],

  // @ember/-internals/routing
  ['Location', '@ember/-internals/routing'], ['AutoLocation', '@ember/-internals/routing'], ['HashLocation', '@ember/-internals/routing'], ['HistoryLocation', '@ember/-internals/routing'], ['NoneLocation', '@ember/-internals/routing'], ['controllerFor', '@ember/-internals/routing'], ['generateControllerFactory', '@ember/-internals/routing'], ['generateController', '@ember/-internals/routing'], ['RouterDSL', '@ember/-internals/routing'], ['Router', '@ember/-internals/routing'], ['Route', '@ember/-internals/routing'],

  // ember-application
  ['Application', '@ember/application', 'default'], ['ApplicationInstance', '@ember/application/instance', 'default'], ['Engine', '@ember/engine', 'default'], ['EngineInstance', '@ember/engine/instance', 'default'], ['Resolver', '@ember/application/globals-resolver', 'default'], ['DefaultResolver', '@ember/application/globals-resolver', 'default'],

  // @ember/-internals/extension-support
  ['DataAdapter', '@ember/-internals/extension-support'], ['ContainerDebugAdapter', '@ember/-internals/extension-support']].filter(Boolean);
});
enifed('ember/tests/routing/decoupled_basic_test', ['@ember/-internals/owner', 'rsvp', 'ember-template-compiler', '@ember/-internals/routing', '@ember/controller', '@ember/-internals/runtime', 'internal-test-helpers', '@ember/runloop', '@ember/-internals/metal', '@ember/-internals/glimmer', '@ember/engine', 'router_js'], function (_owner, _rsvp, _emberTemplateCompiler, _routing, _controller, _runtime, _internalTestHelpers, _runloop, _metal, _glimmer, _engine, _router_js) {
  'use strict';

  let originalConsoleError; /* eslint-disable no-console */


  (0, _internalTestHelpers.moduleFor)('Basic Routing - Decoupled from global resolver', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super(...arguments);
      this.addTemplate('home', '<h3 class="hours">Hours</h3>');
      this.addTemplate('camelot', '<section id="camelot"><h3>Is a silly place</h3></section>');
      this.addTemplate('homepage', '<h3 id="troll">Megatroll</h3><p>{{model.home}}</p>');

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      originalConsoleError = console.error;
    }

    teardown() {
      super.teardown();
      console.error = originalConsoleError;
    }

    getController(name) {
      return this.applicationInstance.lookup(`controller:${name}`);
    }

    handleURLAborts(assert, path) {
      (0, _runloop.run)(() => {
        let router = this.applicationInstance.lookup('router:main');
        router.handleURL(path).then(function () {
          assert.ok(false, 'url: `' + path + '` was NOT to be handled');
        }, function (reason) {
          assert.ok(reason && reason.message === 'TransitionAborted', 'url: `' + path + '` was to be aborted');
        });
      });
    }

    get currentPath() {
      return this.getController('application').get('currentPath');
    }

    get currentURL() {
      return this.appRouter.get('currentURL');
    }

    handleURLRejectsWith(context, assert, path, expectedReason) {
      return context.visit(path).then(() => {
        assert.ok(false, 'expected handleURLing: `' + path + '` to fail');
      }).catch(reason => {
        assert.equal(reason.message, expectedReason);
      });
    }

    ['@test warn on URLs not included in the route set']() {
      return this.visit('/').then(() => {
        expectAssertion(() => {
          this.visit('/what-is-this-i-dont-even');
        }, /'\/what-is-this-i-dont-even' did not match any routes/);
      });
    }

    ['@test The Homepage'](assert) {
      return this.visit('/').then(() => {
        assert.equal(this.currentPath, 'home', 'currently on the home route');

        let text = this.$('.hours').text();
        assert.equal(text, 'Hours', 'the home template was rendered');
      });
    }

    [`@test The Homepage and the Camelot page with multiple Router.map calls`](assert) {
      this.router.map(function () {
        this.route('camelot', { path: '/camelot' });
      });

      return this.visit('/camelot').then(() => {
        assert.equal(this.currentPath, 'camelot');

        let text = this.$('#camelot').text();
        assert.equal(text, 'Is a silly place', 'the camelot template was rendered');

        return this.visit('/');
      }).then(() => {
        assert.equal(this.currentPath, 'home');

        let text = this.$('.hours').text();
        assert.equal(text, 'Hours', 'the home template was rendered');
      });
    }

    [`@test The Homepage with explicit template name in renderTemplate`](assert) {
      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          this.render('homepage');
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('#troll').text();
        assert.equal(text, 'Megatroll', 'the homepage template was rendered');
      });
    }

    [`@test an alternate template will pull in an alternate controller`](assert) {
      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          this.render('homepage');
        }
      }));
      this.add('controller:homepage', _controller.default.extend({
        model: {
          home: 'Comes from homepage'
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'Comes from homepage', 'the homepage template was rendered');
      });
    }

    [`@test An alternate template will pull in an alternate controller instead of controllerName`](assert) {
      this.add('route:home', _routing.Route.extend({
        controllerName: 'foo',
        renderTemplate() {
          this.render('homepage');
        }
      }));
      this.add('controller:foo', _controller.default.extend({
        model: {
          home: 'Comes from foo'
        }
      }));
      this.add('controller:homepage', _controller.default.extend({
        model: {
          home: 'Comes from homepage'
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'Comes from homepage', 'the homepage template was rendered');
      });
    }

    [`@test The template will pull in an alternate controller via key/value`](assert) {
      this.router.map(function () {
        this.route('homepage', { path: '/' });
      });

      this.add('route:homepage', _routing.Route.extend({
        renderTemplate() {
          this.render({ controller: 'home' });
        }
      }));
      this.add('controller:home', _controller.default.extend({
        model: {
          home: 'Comes from home.'
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'Comes from home.', 'the homepage template was rendered from data from the HomeController');
      });
    }

    [`@test The Homepage with explicit template name in renderTemplate and controller`](assert) {
      this.add('controller:home', _controller.default.extend({
        model: {
          home: 'YES I AM HOME'
        }
      }));
      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          this.render('homepage');
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'YES I AM HOME', 'The homepage template was rendered');
      });
    }

    [`@test Model passed via renderTemplate model is set as controller's model`](assert) {
      this.addTemplate('bio', '<p>{{model.name}}</p>');
      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          this.render('bio', {
            model: { name: 'emberjs' }
          });
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'emberjs', `Passed model was set as controller's model`);
      });
    }

    ['@test render uses templateName from route'](assert) {
      this.addTemplate('the_real_home_template', '<p>THIS IS THE REAL HOME</p>');
      this.add('route:home', _routing.Route.extend({
        templateName: 'the_real_home_template'
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();

        assert.equal(text, 'THIS IS THE REAL HOME', 'the homepage template was rendered');
      });
    }

    ['@test defining templateName allows other templates to be rendered'](assert) {
      this.addTemplate('alert', `<div class='alert-box'>Invader!</div>`);
      this.addTemplate('the_real_home_template', `<p>THIS IS THE REAL HOME</p>{{outlet 'alert'}}`);
      this.add('route:home', _routing.Route.extend({
        templateName: 'the_real_home_template',
        actions: {
          showAlert() {
            this.render('alert', {
              into: 'home',
              outlet: 'alert'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('p').text();
        assert.equal(text, 'THIS IS THE REAL HOME', 'the homepage template was rendered');

        return this.runTask(() => this.appRouter.send('showAlert'));
      }).then(() => {
        let text = this.$('.alert-box').text();

        assert.equal(text, 'Invader!', 'Template for alert was rendered into the outlet');
      });
    }

    ['@test templateName is still used when calling render with no name and options'](assert) {
      this.addTemplate('alert', `<div class='alert-box'>Invader!</div>`);
      this.addTemplate('home', `<p>THIS IS THE REAL HOME</p>{{outlet 'alert'}}`);

      this.add('route:home', _routing.Route.extend({
        templateName: 'alert',
        renderTemplate() {
          this.render({});
        }
      }));

      return this.visit('/').then(() => {
        let text = this.$('.alert-box').text();

        assert.equal(text, 'Invader!', 'default templateName was rendered into outlet');
      });
    }

    ['@test The Homepage with a `setupController` hook'](assert) {
      this.addTemplate('home', `<ul>{{#each hours as |entry|}}
        <li>{{entry}}</li>
      {{/each}}
      </ul>
    `);

      this.add('route:home', _routing.Route.extend({
        setupController(controller) {
          controller.set('hours', ['Monday through Friday: 9am to 5pm', 'Saturday: Noon to Midnight', 'Sunday: Noon to 6pm']);
        }
      }));
      return this.visit('/').then(() => {
        let text = this.$('ul li:nth-child(3)').text();

        assert.equal(text, 'Sunday: Noon to 6pm', 'The template was rendered with the hours context');
      });
    }

    [`@test The route controller is still set when overriding the setupController hook`](assert) {
      this.add('route:home', _routing.Route.extend({
        setupController() {
          // no-op
          // importantly, we are not calling this._super
        }
      }));

      this.add('controller:home', _controller.default.extend());

      return this.visit('/').then(() => {
        let homeRoute = this.applicationInstance.lookup('route:home');
        let homeController = this.applicationInstance.lookup('controller:home');

        assert.equal(homeRoute.controller, homeController, 'route controller is the home controller');
      });
    }

    ['@test the route controller can be specified via controllerName'](assert) {
      this.addTemplate('home', '<p>{{myValue}}</p>');
      this.add('route:home', _routing.Route.extend({
        controllerName: 'myController'
      }));
      this.add('controller:myController', _controller.default.extend({
        myValue: 'foo'
      }));

      return this.visit('/').then(() => {
        let homeRoute = this.applicationInstance.lookup('route:home');
        let myController = this.applicationInstance.lookup('controller:myController');
        let text = this.$('p').text();

        assert.equal(homeRoute.controller, myController, 'route controller is set by controllerName');
        assert.equal(text, 'foo', 'The homepage template was rendered with data from the custom controller');
      });
    }

    [`@test The route controller specified via controllerName is used in render`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('route:home', _routing.Route.extend({
        controllerName: 'myController',
        renderTemplate() {
          this.render('alternative_home');
        }
      }));

      this.add('controller:myController', _controller.default.extend({
        myValue: 'foo'
      }));

      this.addTemplate('alternative_home', '<p>alternative home: {{myValue}}</p>');

      return this.visit('/').then(() => {
        let homeRoute = this.applicationInstance.lookup('route:home');
        let myController = this.applicationInstance.lookup('controller:myController');
        let text = this.$('p').text();

        assert.equal(homeRoute.controller, myController, 'route controller is set by controllerName');

        assert.equal(text, 'alternative home: foo', 'The homepage template was rendered with data from the custom controller');
      });
    }

    [`@test The route controller specified via controllerName is used in render even when a controller with the routeName is available`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.addTemplate('home', '<p>home: {{myValue}}</p>');

      this.add('route:home', _routing.Route.extend({
        controllerName: 'myController'
      }));

      this.add('controller:home', _controller.default.extend({
        myValue: 'home'
      }));

      this.add('controller:myController', _controller.default.extend({
        myValue: 'myController'
      }));

      return this.visit('/').then(() => {
        let homeRoute = this.applicationInstance.lookup('route:home');
        let myController = this.applicationInstance.lookup('controller:myController');
        let text = this.$('p').text();

        assert.equal(homeRoute.controller, myController, 'route controller is set by controllerName');

        assert.equal(text, 'home: myController', 'The homepage template was rendered with data from the custom controller');
      });
    }

    [`@test The Homepage with a 'setupController' hook modifying other controllers`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('route:home', _routing.Route.extend({
        setupController() /* controller */{
          this.controllerFor('home').set('hours', ['Monday through Friday: 9am to 5pm', 'Saturday: Noon to Midnight', 'Sunday: Noon to 6pm']);
        }
      }));

      this.addTemplate('home', '<ul>{{#each hours as |entry|}}<li>{{entry}}</li>{{/each}}</ul>');

      return this.visit('/').then(() => {
        let text = this.$('ul li:nth-child(3)').text();

        assert.equal(text, 'Sunday: Noon to 6pm', 'The template was rendered with the hours context');
      });
    }

    [`@test The Homepage with a computed model that does not get overridden`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('controller:home', _controller.default.extend({
        model: (0, _metal.computed)(function () {
          return ['Monday through Friday: 9am to 5pm', 'Saturday: Noon to Midnight', 'Sunday: Noon to 6pm'];
        })
      }));

      this.addTemplate('home', '<ul>{{#each model as |passage|}}<li>{{passage}}</li>{{/each}}</ul>');

      return this.visit('/').then(() => {
        let text = this.$('ul li:nth-child(3)').text();

        assert.equal(text, 'Sunday: Noon to 6pm', 'The template was rendered with the context intact');
      });
    }

    [`@test The Homepage getting its controller context via model`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('route:home', _routing.Route.extend({
        model() {
          return ['Monday through Friday: 9am to 5pm', 'Saturday: Noon to Midnight', 'Sunday: Noon to 6pm'];
        },

        setupController(controller, model) {
          assert.equal(this.controllerFor('home'), controller);

          this.controllerFor('home').set('hours', model);
        }
      }));

      this.addTemplate('home', '<ul>{{#each hours as |entry|}}<li>{{entry}}</li>{{/each}}</ul>');

      return this.visit('/').then(() => {
        let text = this.$('ul li:nth-child(3)').text();

        assert.equal(text, 'Sunday: Noon to 6pm', 'The template was rendered with the hours context');
      });
    }

    [`@test The Specials Page getting its controller context by deserializing the params hash`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      this.add('route:special', _routing.Route.extend({
        model(params) {
          return _runtime.Object.create({
            menuItemId: params.menu_item_id
          });
        }
      }));

      this.addTemplate('special', '<p>{{model.menuItemId}}</p>');

      return this.visit('/specials/1').then(() => {
        let text = this.$('p').text();

        assert.equal(text, '1', 'The model was used to render the template');
      });
    }

    ['@test The Specials Page defaults to looking models up via `find`']() {
      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          return MenuItem.create({ id });
        }
      });
      this.add('model:menu_item', MenuItem);

      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      this.addTemplate('special', '{{model.id}}');

      return this.visit('/specials/1').then(() => {
        this.assertText('1', 'The model was used to render the template');
      });
    }

    ['@test The Special Page returning a promise puts the app into a loading state until the promise is resolved']() {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      let menuItem, resolve;

      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          menuItem = MenuItem.create({ id: id });

          return new _rsvp.default.Promise(function (res) {
            resolve = res;
          });
        }
      });

      this.add('model:menu_item', MenuItem);

      this.addTemplate('special', '<p>{{model.id}}</p>');
      this.addTemplate('loading', '<p>LOADING!</p>');

      let visited = this.visit('/specials/1');
      this.assertText('LOADING!', 'The app is in the loading state');

      resolve(menuItem);

      return visited.then(() => {
        this.assertText('1', 'The app is now in the specials state');
      });
    }

    [`@test The loading state doesn't get entered for promises that resolve on the same run loop`](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          return { id: id };
        }
      });

      this.add('model:menu_item', MenuItem);

      this.add('route:loading', _routing.Route.extend({
        enter() {
          assert.ok(false, "LoadingRoute shouldn't have been entered.");
        }
      }));

      this.addTemplate('special', '<p>{{model.id}}</p>');
      this.addTemplate('loading', '<p>LOADING!</p>');

      return this.visit('/specials/1').then(() => {
        let text = this.$('p').text();

        assert.equal(text, '1', 'The app is now in the specials state');
      });
    }

    ["@test The Special page returning an error invokes SpecialRoute's error handler"](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      let menuItem, promise, resolve;

      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          menuItem = MenuItem.create({ id: id });
          promise = new _rsvp.default.Promise(res => resolve = res);

          return promise;
        }
      });

      this.add('model:menu_item', MenuItem);

      this.add('route:special', _routing.Route.extend({
        setup() {
          throw new Error('Setup error');
        },
        actions: {
          error(reason) {
            assert.equal(reason.message, 'Setup error', 'SpecialRoute#error received the error thrown from setup');
            return true;
          }
        }
      }));

      this.handleURLRejectsWith(this, assert, 'specials/1', 'Setup error');

      (0, _runloop.run)(() => resolve(menuItem));
    }

    ["@test ApplicationRoute's default error handler can be overridden"](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      let menuItem, resolve;

      let MenuItem = _runtime.Object.extend();

      MenuItem.reopenClass({
        find(id) {
          menuItem = MenuItem.create({ id: id });
          return new _rsvp.default.Promise(res => resolve = res);
        }
      });
      this.add('model:menu_item', MenuItem);

      this.add('route:application', _routing.Route.extend({
        actions: {
          error(reason) {
            assert.equal(reason.message, 'Setup error', 'error was correctly passed to custom ApplicationRoute handler');
            return true;
          }
        }
      }));

      this.add('route:special', _routing.Route.extend({
        setup() {
          throw new Error('Setup error');
        }
      }));

      this.handleURLRejectsWith(this, assert, '/specials/1', 'Setup error');

      (0, _runloop.run)(() => resolve(menuItem));
    }

    ['@test Moving from one page to another triggers the correct callbacks'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('special', { path: '/specials/:menu_item_id' });
      });

      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          return MenuItem.create({ id: id });
        }
      });
      this.add('model:menu_item', MenuItem);

      this.addTemplate('home', '<h3>Home</h3>');
      this.addTemplate('special', '<p>{{model.id}}</p>');

      return this.visit('/').then(() => {
        this.assertText('Home', 'The app is now in the initial state');

        let promiseContext = MenuItem.create({ id: 1 });

        return this.visit('/specials/1', promiseContext);
      }).then(() => {
        assert.equal(this.currentURL, '/specials/1');
        this.assertText('1', 'The app is now transitioned');
      });
    }

    ['@test Nested callbacks are not exited when moving to siblings'](assert) {
      let rootSetup = 0;
      let rootRender = 0;
      let rootModel = 0;
      let rootSerialize = 0;
      let menuItem;
      let rootElement;

      let MenuItem = _runtime.Object.extend();
      MenuItem.reopenClass({
        find(id) {
          menuItem = MenuItem.create({ id: id });
          return menuItem;
        }
      });

      this.router.map(function () {
        this.route('root', { path: '/' }, function () {
          this.route('special', {
            path: '/specials/:menu_item_id',
            resetNamespace: true
          });
        });
      });

      this.add('route:root', _routing.Route.extend({
        model() {
          rootModel++;
          return this._super(...arguments);
        },

        setupController() {
          rootSetup++;
        },

        renderTemplate() {
          rootRender++;
        },

        serialize() {
          rootSerialize++;
          return this._super(...arguments);
        }
      }));

      this.add('route:loading', _routing.Route.extend({}));
      this.add('route:home', _routing.Route.extend({}));
      this.add('route:special', _routing.Route.extend({
        model({ menu_item_id }) {
          return MenuItem.find(menu_item_id);
        },
        setupController(controller, model) {
          (0, _metal.set)(controller, 'model', model);
        }
      }));

      this.addTemplate('root.index', '<h3>Home</h3>');
      this.addTemplate('special', '<p>{{model.id}}</p>');
      this.addTemplate('loading', '<p>LOADING!</p>');

      return this.visit('/').then(() => {
        rootElement = document.getElementById('qunit-fixture');

        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('h3')), 'Home', 'The app is now in the initial state');
        assert.equal(rootSetup, 1, 'The root setup was triggered');
        assert.equal(rootRender, 1, 'The root render was triggered');
        assert.equal(rootSerialize, 0, 'The root serialize was not called');
        assert.equal(rootModel, 1, 'The root model was called');

        let router = this.applicationInstance.lookup('router:main');
        let menuItem = MenuItem.create({ id: 1 });

        return router.transitionTo('special', menuItem).then(function () {
          assert.equal(rootSetup, 1, 'The root setup was not triggered again');
          assert.equal(rootRender, 1, 'The root render was not triggered again');
          assert.equal(rootSerialize, 0, 'The root serialize was not called');

          // TODO: Should this be changed?
          assert.equal(rootModel, 1, 'The root model was called again');

          assert.deepEqual(router.location.path, '/specials/1');
          assert.equal(router.currentPath, 'root.special');
        });
      });
    }

    ['@test Events are triggered on the controller if a matching action name is implemented'](assert) {
      let done = assert.async();

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      let model = { name: 'Tom Dale' };
      let stateIsNotCalled = true;

      this.add('route:home', _routing.Route.extend({
        model() {
          return model;
        },

        actions: {
          showStuff() {
            stateIsNotCalled = false;
          }
        }
      }));

      this.addTemplate('home', '<a {{action "showStuff" model}}>{{name}}</a>');
      this.add('controller:home', _controller.default.extend({
        actions: {
          showStuff(context) {
            assert.ok(stateIsNotCalled, 'an event on the state is not triggered');
            assert.deepEqual(context, { name: 'Tom Dale' }, 'an event with context is passed');
            done();
          }
        }
      }));

      this.visit('/').then(() => {
        document.getElementById('qunit-fixture').querySelector('a').click();
      });
    }

    ['@test Events are triggered on the current state when defined in `actions` object'](assert) {
      let done = assert.async();

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      let model = { name: 'Tom Dale' };
      let HomeRoute = _routing.Route.extend({
        model() {
          return model;
        },

        actions: {
          showStuff(obj) {
            assert.ok(this instanceof HomeRoute, 'the handler is an App.HomeRoute');
            assert.deepEqual(Object.assign({}, obj), { name: 'Tom Dale' }, 'the context is correct');
            done();
          }
        }
      });

      this.add('route:home', HomeRoute);
      this.addTemplate('home', '<a {{action "showStuff" model}}>{{model.name}}</a>');

      this.visit('/').then(() => {
        document.getElementById('qunit-fixture').querySelector('a').click();
      });
    }

    ['@test Events defined in `actions` object are triggered on the current state when routes are nested'](assert) {
      let done = assert.async();

      this.router.map(function () {
        this.route('root', { path: '/' }, function () {
          this.route('index', { path: '/' });
        });
      });

      let model = { name: 'Tom Dale' };

      let RootRoute = _routing.Route.extend({
        actions: {
          showStuff(obj) {
            assert.ok(this instanceof RootRoute, 'the handler is an App.HomeRoute');
            assert.deepEqual(Object.assign({}, obj), { name: 'Tom Dale' }, 'the context is correct');
            done();
          }
        }
      });
      this.add('route:root', RootRoute);
      this.add('route:root.index', _routing.Route.extend({
        model() {
          return model;
        }
      }));

      this.addTemplate('root.index', '<a {{action "showStuff" model}}>{{model.name}}</a>');

      this.visit('/').then(() => {
        document.getElementById('qunit-fixture').querySelector('a').click();
      });
    }

    ['@test Events can be handled by inherited event handlers'](assert) {
      assert.expect(4);

      let SuperRoute = _routing.Route.extend({
        actions: {
          foo() {
            assert.ok(true, 'foo');
          },
          bar(msg) {
            assert.equal(msg, 'HELLO', 'bar hander in super route');
          }
        }
      });

      let RouteMixin = _metal.Mixin.create({
        actions: {
          bar(msg) {
            assert.equal(msg, 'HELLO', 'bar handler in mixin');
            this._super(msg);
          }
        }
      });

      this.add('route:home', SuperRoute.extend(RouteMixin, {
        actions: {
          baz() {
            assert.ok(true, 'baz', 'baz hander in route');
          }
        }
      }));
      this.addTemplate('home', `
      <a class="do-foo" {{action "foo"}}>Do foo</a>
      <a class="do-bar-with-arg" {{action "bar" "HELLO"}}>Do bar with arg</a>
      <a class="do-baz" {{action "baz"}}>Do bar</a>
    `);

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        rootElement.querySelector('.do-foo').click();
        rootElement.querySelector('.do-bar-with-arg').click();
        rootElement.querySelector('.do-baz').click();
      });
    }

    ['@test Actions are not triggered on the controller if a matching action name is implemented as a method'](assert) {
      let done = assert.async();

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      let model = { name: 'Tom Dale' };
      let stateIsNotCalled = true;

      this.add('route:home', _routing.Route.extend({
        model() {
          return model;
        },

        actions: {
          showStuff(context) {
            assert.ok(stateIsNotCalled, 'an event on the state is not triggered');
            assert.deepEqual(context, { name: 'Tom Dale' }, 'an event with context is passed');
            done();
          }
        }
      }));

      this.addTemplate('home', '<a {{action "showStuff" model}}>{{name}}</a>');

      this.add('controller:home', _controller.default.extend({
        showStuff() {
          stateIsNotCalled = false;
          assert.ok(stateIsNotCalled, 'an event on the state is not triggered');
        }
      }));

      this.visit('/').then(() => {
        document.getElementById('qunit-fixture').querySelector('a').click();
      });
    }

    ['@test actions can be triggered with multiple arguments'](assert) {
      let done = assert.async();
      this.router.map(function () {
        this.route('root', { path: '/' }, function () {
          this.route('index', { path: '/' });
        });
      });

      let model1 = { name: 'Tilde' };
      let model2 = { name: 'Tom Dale' };

      let RootRoute = _routing.Route.extend({
        actions: {
          showStuff(obj1, obj2) {
            assert.ok(this instanceof RootRoute, 'the handler is an App.HomeRoute');
            assert.deepEqual(Object.assign({}, obj1), { name: 'Tilde' }, 'the first context is correct');
            assert.deepEqual(Object.assign({}, obj2), { name: 'Tom Dale' }, 'the second context is correct');
            done();
          }
        }
      });

      this.add('route:root', RootRoute);

      this.add('controller:root.index', _controller.default.extend({
        model1: model1,
        model2: model2
      }));

      this.addTemplate('root.index', '<a {{action "showStuff" model1 model2}}>{{model1.name}}</a>');

      this.visit('/').then(() => {
        document.getElementById('qunit-fixture').querySelector('a').click();
      });
    }

    ['@test transitioning multiple times in a single run loop only sets the URL once'](assert) {
      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('foo');
        this.route('bar');
      });

      return this.visit('/').then(() => {
        let urlSetCount = 0;
        let router = this.applicationInstance.lookup('router:main');

        router.get('location').setURL = function (path) {
          urlSetCount++;
          (0, _metal.set)(this, 'path', path);
        };

        assert.equal(urlSetCount, 0);

        (0, _runloop.run)(function () {
          router.transitionTo('foo');
          router.transitionTo('bar');
        });

        assert.equal(urlSetCount, 1);
        assert.equal(router.get('location').getURL(), '/bar');
      });
    }

    ['@test navigating away triggers a url property change'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('foo', { path: '/foo' });
        this.route('bar', { path: '/bar' });
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');

        (0, _metal.addObserver)(router, 'url', function () {
          assert.ok(true, 'url change event was fired');
        });
        ['foo', 'bar', '/foo'].forEach(destination => (0, _runloop.run)(router, 'transitionTo', destination));
      });
    }

    ['@test using replaceWith calls location.replaceURL if available'](assert) {
      let setCount = 0;
      let replaceCount = 0;
      this.router.reopen({
        location: _routing.NoneLocation.create({
          setURL(path) {
            setCount++;
            (0, _metal.set)(this, 'path', path);
          },

          replaceURL(path) {
            replaceCount++;
            (0, _metal.set)(this, 'path', path);
          }
        })
      });

      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('foo');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(setCount, 1);
        assert.equal(replaceCount, 0);

        (0, _runloop.run)(() => router.replaceWith('foo'));

        assert.equal(setCount, 1, 'should not call setURL');
        assert.equal(replaceCount, 1, 'should call replaceURL once');
        assert.equal(router.get('location').getURL(), '/foo');
      });
    }

    ['@test using replaceWith calls setURL if location.replaceURL is not defined'](assert) {
      let setCount = 0;

      this.router.reopen({
        location: _routing.NoneLocation.create({
          setURL(path) {
            setCount++;
            (0, _metal.set)(this, 'path', path);
          }
        })
      });

      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('foo');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');

        assert.equal(setCount, 1);
        (0, _runloop.run)(() => router.replaceWith('foo'));
        assert.equal(setCount, 2, 'should call setURL once');
        assert.equal(router.get('location').getURL(), '/foo');
      });
    }

    ['@test Route inherits model from parent route'](assert) {
      assert.expect(9);

      this.router.map(function () {
        this.route('the-post', { path: '/posts/:post_id' }, function () {
          this.route('comments');

          this.route('shares', { path: '/shares/:share_id', resetNamespace: true }, function () {
            this.route('share');
          });
        });
      });

      let post1 = {};
      let post2 = {};
      let post3 = {};
      let share1 = {};
      let share2 = {};
      let share3 = {};

      let posts = {
        1: post1,
        2: post2,
        3: post3
      };
      let shares = {
        1: share1,
        2: share2,
        3: share3
      };

      this.add('route:the-post', _routing.Route.extend({
        model(params) {
          return posts[params.post_id];
        }
      }));

      this.add('route:the-post.comments', _routing.Route.extend({
        afterModel(post /*, transition */) {
          let parent_model = this.modelFor('the-post');

          assert.equal(post, parent_model);
        }
      }));

      this.add('route:shares', _routing.Route.extend({
        model(params) {
          return shares[params.share_id];
        }
      }));

      this.add('route:shares.share', _routing.Route.extend({
        afterModel(share /*, transition */) {
          let parent_model = this.modelFor('shares');

          assert.equal(share, parent_model);
        }
      }));

      return this.visit('/posts/1/comments').then(() => {
        assert.ok(true, 'url: /posts/1/comments was handled');
        return this.visit('/posts/1/shares/1');
      }).then(() => {
        assert.ok(true, 'url: /posts/1/shares/1 was handled');
        return this.visit('/posts/2/comments');
      }).then(() => {
        assert.ok(true, 'url: /posts/2/comments was handled');
        return this.visit('/posts/2/shares/2');
      }).then(() => {
        assert.ok(true, 'url: /posts/2/shares/2 was handled');
        return this.visit('/posts/3/comments');
      }).then(() => {
        assert.ok(true, 'url: /posts/3/shares was handled');
        return this.visit('/posts/3/shares/3');
      }).then(() => {
        assert.ok(true, 'url: /posts/3/shares/3 was handled');
      });
    }

    ['@test Routes with { resetNamespace: true } inherits model from parent route'](assert) {
      assert.expect(6);

      this.router.map(function () {
        this.route('the-post', { path: '/posts/:post_id' }, function () {
          this.route('comments', { resetNamespace: true }, function () {});
        });
      });

      let post1 = {};
      let post2 = {};
      let post3 = {};

      let posts = {
        1: post1,
        2: post2,
        3: post3
      };

      this.add('route:the-post', _routing.Route.extend({
        model(params) {
          return posts[params.post_id];
        }
      }));

      this.add('route:comments', _routing.Route.extend({
        afterModel(post /*, transition */) {
          let parent_model = this.modelFor('the-post');

          assert.equal(post, parent_model);
        }
      }));

      return this.visit('/posts/1/comments').then(() => {
        assert.ok(true, '/posts/1/comments');
        return this.visit('/posts/2/comments');
      }).then(() => {
        assert.ok(true, '/posts/2/comments');
        return this.visit('/posts/3/comments');
      }).then(() => {
        assert.ok(true, '/posts/3/comments');
      });
    }

    ['@test It is possible to get the model from a parent route'](assert) {
      assert.expect(6);

      this.router.map(function () {
        this.route('the-post', { path: '/posts/:post_id' }, function () {
          this.route('comments', { resetNamespace: true });
        });
      });

      let post1 = {};
      let post2 = {};
      let post3 = {};
      let currentPost;

      let posts = {
        1: post1,
        2: post2,
        3: post3
      };

      this.add('route:the-post', _routing.Route.extend({
        model(params) {
          return posts[params.post_id];
        }
      }));

      this.add('route:comments', _routing.Route.extend({
        model() {
          assert.equal(this.modelFor('the-post'), currentPost);
        }
      }));

      currentPost = post1;
      return this.visit('/posts/1/comments').then(() => {
        assert.ok(true, '/posts/1/comments has been handled');
        currentPost = post2;
        return this.visit('/posts/2/comments');
      }).then(() => {
        assert.ok(true, '/posts/2/comments has been handled');
        currentPost = post3;
        return this.visit('/posts/3/comments');
      }).then(() => {
        assert.ok(true, '/posts/3/comments has been handled');
      });
    }

    ['@test A redirection hook is provided'](assert) {
      this.router.map(function () {
        this.route('choose', { path: '/' });
        this.route('home');
      });

      let chooseFollowed = 0;
      let destination = 'home';

      this.add('route:choose', _routing.Route.extend({
        redirect() {
          if (destination) {
            this.transitionTo(destination);
          }
        },

        setupController() {
          chooseFollowed++;
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(chooseFollowed, 0, "The choose route wasn't entered since a transition occurred");
        assert.equal(rootElement.querySelectorAll('h3.hours').length, 1, 'The home template was rendered');
        assert.equal(this.applicationInstance.lookup('controller:application').get('currentPath'), 'home');
      });
    }

    ['@test Redirecting from the middle of a route aborts the remainder of the routes'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('home');
        this.route('foo', function () {
          this.route('bar', { resetNamespace: true }, function () {
            this.route('baz');
          });
        });
      });

      this.add('route:bar', _routing.Route.extend({
        redirect() {
          this.transitionTo('home');
        },
        setupController() {
          assert.ok(false, 'Should transition before setupController');
        }
      }));

      this.add('route:bar-baz', _routing.Route.extend({
        enter() {
          assert.ok(false, 'Should abort transition getting to next route');
        }
      }));

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        this.handleURLAborts(assert, '/foo/bar/baz');
        assert.equal(this.applicationInstance.lookup('controller:application').get('currentPath'), 'home');
        assert.equal(router.get('location').getURL(), '/home');
      });
    }

    ['@test Redirecting to the current target in the middle of a route does not abort initial routing'](assert) {
      assert.expect(5);

      this.router.map(function () {
        this.route('home');
        this.route('foo', function () {
          this.route('bar', { resetNamespace: true }, function () {
            this.route('baz');
          });
        });
      });

      let successCount = 0;

      this.add('route:bar', _routing.Route.extend({
        redirect() {
          return this.transitionTo('bar.baz').then(function () {
            successCount++;
          });
        },

        setupController() {
          assert.ok(true, "Should still invoke bar's setupController");
        }
      }));

      this.add('route:bar.baz', _routing.Route.extend({
        setupController() {
          assert.ok(true, "Should still invoke bar.baz's setupController");
        }
      }));

      return this.visit('/foo/bar/baz').then(() => {
        assert.ok(true, '/foo/bar/baz has been handled');
        assert.equal(this.applicationInstance.lookup('controller:application').get('currentPath'), 'foo.bar.baz');
        assert.equal(successCount, 1, 'transitionTo success handler was called once');
      });
    }

    ['@test Redirecting to the current target with a different context aborts the remainder of the routes'](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('home');
        this.route('foo', function () {
          this.route('bar', { path: 'bar/:id', resetNamespace: true }, function () {
            this.route('baz');
          });
        });
      });

      let model = { id: 2 };

      let count = 0;

      this.add('route:bar', _routing.Route.extend({
        afterModel() {
          if (count++ > 10) {
            assert.ok(false, 'infinite loop');
          } else {
            this.transitionTo('bar.baz', model);
          }
        }
      }));

      this.add('route:bar.baz', _routing.Route.extend({
        setupController() {
          assert.ok(true, 'Should still invoke setupController');
        }
      }));

      return this.visit('/').then(() => {
        this.handleURLAborts(assert, '/foo/bar/1/baz');
        assert.equal(this.applicationInstance.lookup('controller:application').get('currentPath'), 'foo.bar.baz');
        assert.equal(this.applicationInstance.lookup('router:main').get('location').getURL(), '/foo/bar/2/baz');
      });
    }

    ['@test Transitioning from a parent event does not prevent currentPath from being set'](assert) {
      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar', { resetNamespace: true }, function () {
            this.route('baz');
          });
          this.route('qux');
        });
      });

      this.add('route:foo', _routing.Route.extend({
        actions: {
          goToQux() {
            this.transitionTo('foo.qux');
          }
        }
      }));

      return this.visit('/foo/bar/baz').then(() => {
        assert.ok(true, '/foo/bar/baz has been handled');
        let applicationController = this.applicationInstance.lookup('controller:application');
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(applicationController.get('currentPath'), 'foo.bar.baz');
        (0, _runloop.run)(() => router.send('goToQux'));
        assert.equal(applicationController.get('currentPath'), 'foo.qux');
        assert.equal(router.get('location').getURL(), '/foo/qux');
      });
    }

    ['@test Generated names can be customized when providing routes with dot notation'](assert) {
      assert.expect(4);

      this.addTemplate('index', '<div>Index</div>');
      this.addTemplate('application', "<h1>Home</h1><div class='main'>{{outlet}}</div>");
      this.addTemplate('foo', "<div class='middle'>{{outlet}}</div>");
      this.addTemplate('bar', "<div class='bottom'>{{outlet}}</div>");
      this.addTemplate('bar.baz', '<p>{{name}}Bottom!</p>');

      this.router.map(function () {
        this.route('foo', { path: '/top' }, function () {
          this.route('bar', { path: '/middle', resetNamespace: true }, function () {
            this.route('baz', { path: '/bottom' });
          });
        });
      });

      this.add('route:foo', _routing.Route.extend({
        renderTemplate() {
          assert.ok(true, 'FooBarRoute was called');
          return this._super(...arguments);
        }
      }));

      this.add('route:bar.baz', _routing.Route.extend({
        renderTemplate() {
          assert.ok(true, 'BarBazRoute was called');
          return this._super(...arguments);
        }
      }));

      this.add('controller:bar', _controller.default.extend({
        name: 'Bar'
      }));

      this.add('controller:bar.baz', _controller.default.extend({
        name: 'BarBaz'
      }));

      return this.visit('/top/middle/bottom').then(() => {
        assert.ok(true, '/top/middle/bottom has been handled');
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('.main .middle .bottom p')), 'BarBazBottom!', 'The templates were rendered into their appropriate parents');
      });
    }

    ["@test Child routes render into their parent route's template by default"](assert) {
      this.addTemplate('index', '<div>Index</div>');
      this.addTemplate('application', "<h1>Home</h1><div class='main'>{{outlet}}</div>");
      this.addTemplate('top', "<div class='middle'>{{outlet}}</div>");
      this.addTemplate('middle', "<div class='bottom'>{{outlet}}</div>");
      this.addTemplate('middle.bottom', '<p>Bottom!</p>');

      this.router.map(function () {
        this.route('top', function () {
          this.route('middle', { resetNamespace: true }, function () {
            this.route('bottom');
          });
        });
      });

      return this.visit('/top/middle/bottom').then(() => {
        assert.ok(true, '/top/middle/bottom has been handled');
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('.main .middle .bottom p')), 'Bottom!', 'The templates were rendered into their appropriate parents');
      });
    }

    ['@test Child routes render into specified template'](assert) {
      this.addTemplate('index', '<div>Index</div>');
      this.addTemplate('application', "<h1>Home</h1><div class='main'>{{outlet}}</div>");
      this.addTemplate('top', "<div class='middle'>{{outlet}}</div>");
      this.addTemplate('middle', "<div class='bottom'>{{outlet}}</div>");
      this.addTemplate('middle.bottom', '<p>Bottom!</p>');

      this.router.map(function () {
        this.route('top', function () {
          this.route('middle', { resetNamespace: true }, function () {
            this.route('bottom');
          });
        });
      });

      this.add('route:middle.bottom', _routing.Route.extend({
        renderTemplate() {
          this.render('middle/bottom', { into: 'top' });
        }
      }));

      return this.visit('/top/middle/bottom').then(() => {
        assert.ok(true, '/top/middle/bottom has been handled');
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.querySelectorAll('.main .middle .bottom p').length, 0, 'should not render into the middle template');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('.main .middle > p')), 'Bottom!', 'The template was rendered into the top template');
      });
    }

    ['@test Rendering into specified template with slash notation'](assert) {
      this.addTemplate('person.profile', 'profile {{outlet}}');
      this.addTemplate('person.details', 'details!');

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          this.render('person/profile');
          this.render('person/details', { into: 'person/profile' });
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'profile details!', 'The templates were rendered');
      });
    }

    ['@test Parent route context change'](assert) {
      let editCount = 0;
      let editedPostIds = (0, _runtime.A)();

      this.addTemplate('application', '{{outlet}}');
      this.addTemplate('posts', '{{outlet}}');
      this.addTemplate('post', '{{outlet}}');
      this.addTemplate('post/index', 'showing');
      this.addTemplate('post/edit', 'editing');

      this.router.map(function () {
        this.route('posts', function () {
          this.route('post', { path: '/:postId', resetNamespace: true }, function () {
            this.route('edit');
          });
        });
      });

      this.add('route:posts', _routing.Route.extend({
        actions: {
          showPost(context) {
            this.transitionTo('post', context);
          }
        }
      }));

      this.add('route:post', _routing.Route.extend({
        model(params) {
          return { id: params.postId };
        },

        serialize(model) {
          return { postId: model.id };
        },

        actions: {
          editPost() {
            this.transitionTo('post.edit');
          }
        }
      }));

      this.add('route:post.edit', _routing.Route.extend({
        model() {
          let postId = this.modelFor('post').id;
          editedPostIds.push(postId);
          return null;
        },
        setup() {
          this._super(...arguments);
          editCount++;
        }
      }));

      return this.visit('/posts/1').then(() => {
        assert.ok(true, '/posts/1 has been handled');
        let router = this.applicationInstance.lookup('router:main');
        (0, _runloop.run)(() => router.send('editPost'));
        (0, _runloop.run)(() => router.send('showPost', { id: '2' }));
        (0, _runloop.run)(() => router.send('editPost'));
        assert.equal(editCount, 2, 'set up the edit route twice without failure');
        assert.deepEqual(editedPostIds, ['1', '2'], 'modelFor posts.post returns the right context');
      });
    }

    ['@test Router accounts for rootURL on page load when using history location'](assert) {
      let rootURL = window.location.pathname + '/app';
      let postsTemplateRendered = false;
      let setHistory;

      setHistory = function (obj, path) {
        obj.set('history', { state: { path: path } });
      };

      let location = _routing.HistoryLocation.create({
        initState() {
          let path = rootURL + '/posts';

          setHistory(this, path);
          this.set('location', {
            pathname: path,
            href: 'http://localhost/' + path
          });
        },

        replaceState(path) {
          setHistory(this, path);
        },

        pushState(path) {
          setHistory(this, path);
        }
      });

      this.router.reopen({
        // location: 'historyTest',
        location,
        rootURL: rootURL
      });

      this.router.map(function () {
        this.route('posts', { path: '/posts' });
      });

      this.add('route:posts', _routing.Route.extend({
        model() {},
        renderTemplate() {
          postsTemplateRendered = true;
        }
      }));

      return this.visit('/').then(() => {
        assert.ok(postsTemplateRendered, 'Posts route successfully stripped from rootURL');

        (0, _internalTestHelpers.runDestroy)(location);
        location = null;
      });
    }

    ['@test The rootURL is passed properly to the location implementation'](assert) {
      assert.expect(1);
      let rootURL = '/blahzorz';
      this.add('location:history-test', _routing.HistoryLocation.extend({
        rootURL: 'this is not the URL you are looking for',
        history: {
          pushState() {}
        },
        initState() {
          assert.equal(this.get('rootURL'), rootURL);
        }
      }));

      this.router.reopen({
        location: 'history-test',
        rootURL: rootURL,
        // if we transition in this test we will receive failures
        // if the tests are run from a static file
        _doURLTransition() {
          return _rsvp.default.resolve('');
        }
      });

      return this.visit('/');
    }

    ['@test Only use route rendered into main outlet for default into property on child'](assert) {
      this.addTemplate('application', "{{outlet 'menu'}}{{outlet}}");
      this.addTemplate('posts', '{{outlet}}');
      this.addTemplate('posts.index', '<p class="posts-index">postsIndex</p>');
      this.addTemplate('posts.menu', '<div class="posts-menu">postsMenu</div>');

      this.router.map(function () {
        this.route('posts', function () {});
      });

      this.add('route:posts', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('posts/menu', {
            into: 'application',
            outlet: 'menu'
          });
        }
      }));

      return this.visit('/posts').then(() => {
        assert.ok(true, '/posts has been handled');
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-menu')), 'postsMenu', 'The posts/menu template was rendered');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p.posts-index')), 'postsIndex', 'The posts/index template was rendered');
      });
    }

    ['@test Generating a URL should not affect currentModel'](assert) {
      this.router.map(function () {
        this.route('post', { path: '/posts/:post_id' });
      });

      let posts = {
        1: { id: 1 },
        2: { id: 2 }
      };

      this.add('route:post', _routing.Route.extend({
        model(params) {
          return posts[params.post_id];
        }
      }));

      return this.visit('/posts/1').then(() => {
        assert.ok(true, '/posts/1 has been handled');

        let route = this.applicationInstance.lookup('route:post');
        assert.equal(route.modelFor('post'), posts[1]);

        let url = this.applicationInstance.lookup('router:main').generate('post', posts[2]);
        assert.equal(url, '/posts/2');
        assert.equal(route.modelFor('post'), posts[1]);
      });
    }

    ["@test Nested index route is not overridden by parent's implicit index route"](assert) {
      this.router.map(function () {
        this.route('posts', function () {
          this.route('index', { path: ':category' });
        });
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        return router.transitionTo('posts', { category: 'emberjs' });
      }).then(() => {
        let router = this.applicationInstance.lookup('router:main');
        assert.deepEqual(router.location.path, '/posts/emberjs');
      });
    }

    ['@test Application template does not duplicate when re-rendered'](assert) {
      this.addTemplate('application', '<h3 class="render-once">I render once</h3>{{outlet}}');

      this.router.map(function () {
        this.route('posts');
      });

      this.add('route:application', _routing.Route.extend({
        model() {
          return (0, _runtime.A)();
        }
      }));

      return this.visit('/posts').then(() => {
        assert.ok(true, '/posts has been handled');
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('h3.render-once')), 'I render once');
      });
    }

    ['@test Child routes should render inside the application template if the application template causes a redirect'](assert) {
      this.addTemplate('application', '<h3>App</h3> {{outlet}}');
      this.addTemplate('posts', 'posts');

      this.router.map(function () {
        this.route('posts');
        this.route('photos');
      });

      this.add('route:application', _routing.Route.extend({
        afterModel() {
          this.transitionTo('posts');
        }
      }));

      return this.visit('/posts').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'App posts');
      });
    }

    ["@test The template is not re-rendered when the route's context changes"](assert) {
      this.router.map(function () {
        this.route('page', { path: '/page/:name' });
      });

      this.add('route:page', _routing.Route.extend({
        model(params) {
          return _runtime.Object.create({ name: params.name });
        }
      }));

      let insertionCount = 0;
      this.add('component:foo-bar', _glimmer.Component.extend({
        didInsertElement() {
          insertionCount += 1;
        }
      }));

      this.addTemplate('page', '<p>{{model.name}}{{foo-bar}}</p>');

      let rootElement = document.getElementById('qunit-fixture');
      return this.visit('/page/first').then(() => {
        assert.ok(true, '/page/first has been handled');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'first');
        assert.equal(insertionCount, 1);
        return this.visit('/page/second');
      }).then(() => {
        assert.ok(true, '/page/second has been handled');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'second');
        assert.equal(insertionCount, 1, 'view should have inserted only once');
        let router = this.applicationInstance.lookup('router:main');
        return (0, _runloop.run)(() => router.transitionTo('page', _runtime.Object.create({ name: 'third' })));
      }).then(() => {
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'third');
        assert.equal(insertionCount, 1, 'view should still have inserted only once');
      });
    }

    ['@test The template is not re-rendered when two routes present the exact same template & controller'](assert) {
      this.router.map(function () {
        this.route('first');
        this.route('second');
        this.route('third');
        this.route('fourth');
      });

      // Note add a component to test insertion

      let insertionCount = 0;
      this.add('component:x-input', _glimmer.Component.extend({
        didInsertElement() {
          insertionCount += 1;
        }
      }));

      let SharedRoute = _routing.Route.extend({
        setupController() {
          this.controllerFor('shared').set('message', 'This is the ' + this.routeName + ' message');
        },

        renderTemplate() {
          this.render('shared', { controller: 'shared' });
        }
      });

      this.add('route:shared', SharedRoute);
      this.add('route:first', SharedRoute.extend());
      this.add('route:second', SharedRoute.extend());
      this.add('route:third', SharedRoute.extend());
      this.add('route:fourth', SharedRoute.extend());

      this.add('controller:shared', _controller.default.extend());

      this.addTemplate('shared', '<p>{{message}}{{x-input}}</p>');

      let rootElement = document.getElementById('qunit-fixture');
      return this.visit('/first').then(() => {
        assert.ok(true, '/first has been handled');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'This is the first message');
        assert.equal(insertionCount, 1, 'expected one assertion');
        return this.visit('/second');
      }).then(() => {
        assert.ok(true, '/second has been handled');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'This is the second message');
        assert.equal(insertionCount, 1, 'expected one assertion');
        return (0, _runloop.run)(() => {
          this.applicationInstance.lookup('router:main').transitionTo('third').then(function () {
            assert.ok(true, 'expected transition');
          }, function (reason) {
            assert.ok(false, 'unexpected transition failure: ', QUnit.jsDump.parse(reason));
          });
        });
      }).then(() => {
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'This is the third message');
        assert.equal(insertionCount, 1, 'expected one assertion');
        return this.visit('fourth');
      }).then(() => {
        assert.ok(true, '/fourth has been handled');
        assert.equal(insertionCount, 1, 'expected one assertion');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'This is the fourth message');
      });
    }

    ['@test ApplicationRoute with model does not proxy the currentPath'](assert) {
      let model = {};
      let currentPath;

      this.router.map(function () {
        this.route('index', { path: '/' });
      });

      this.add('route:application', _routing.Route.extend({
        model() {
          return model;
        }
      }));

      this.add('controller:application', _controller.default.extend({
        currentPathDidChange: (0, _metal.observer)('currentPath', function () {
          currentPath = this.currentPath;
        })
      }));

      return this.visit('/').then(() => {
        assert.equal(currentPath, 'index', 'currentPath is index');
        assert.equal('currentPath' in model, false, 'should have defined currentPath on controller');
      });
    }

    ['@test Promises encountered on app load put app into loading state until resolved'](assert) {
      assert.expect(2);

      let deferred = _rsvp.default.defer();
      this.router.map(function () {
        this.route('index', { path: '/' });
      });

      this.add('route:index', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      this.addTemplate('index', '<p>INDEX</p>');
      this.addTemplate('loading', '<p>LOADING</p>');

      (0, _runloop.run)(() => this.visit('/'));
      let rootElement = document.getElementById('qunit-fixture');
      assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'LOADING', 'The loading state is displaying.');
      (0, _runloop.run)(deferred.resolve);
      assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p')), 'INDEX', 'The index route is display.');
    }

    ['@test Route should tear down multiple outlets'](assert) {
      this.addTemplate('application', "{{outlet 'menu'}}{{outlet}}{{outlet 'footer'}}");
      this.addTemplate('posts', '{{outlet}}');
      this.addTemplate('users', 'users');
      this.addTemplate('posts.index', '<p class="posts-index">postsIndex</p>');
      this.addTemplate('posts.menu', '<div class="posts-menu">postsMenu</div>');
      this.addTemplate('posts.footer', '<div class="posts-footer">postsFooter</div>');

      this.router.map(function () {
        this.route('posts', function () {});
        this.route('users', function () {});
      });

      this.add('route:posts', _routing.Route.extend({
        renderTemplate() {
          this.render('posts/menu', {
            into: 'application',
            outlet: 'menu'
          });

          this.render();

          this.render('posts/footer', {
            into: 'application',
            outlet: 'footer'
          });
        }
      }));

      let rootElement = document.getElementById('qunit-fixture');
      return this.visit('/posts').then(() => {
        assert.ok(true, '/posts has been handled');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-menu')), 'postsMenu', 'The posts/menu template was rendered');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('p.posts-index')), 'postsIndex', 'The posts/index template was rendered');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-footer')), 'postsFooter', 'The posts/footer template was rendered');

        return this.visit('/users');
      }).then(() => {
        assert.ok(true, '/users has been handled');
        assert.equal(rootElement.querySelector('div.posts-menu'), null, 'The posts/menu template was removed');
        assert.equal(rootElement.querySelector('p.posts-index'), null, 'The posts/index template was removed');
        assert.equal(rootElement.querySelector('div.posts-footer'), null, 'The posts/footer template was removed');
      });
    }

    ['@test Route supports clearing outlet explicitly'](assert) {
      this.addTemplate('application', "{{outlet}}{{outlet 'modal'}}");
      this.addTemplate('posts', '{{outlet}}');
      this.addTemplate('users', 'users');
      this.addTemplate('posts.index', '<div class="posts-index">postsIndex {{outlet}}</div>');
      this.addTemplate('posts.modal', '<div class="posts-modal">postsModal</div>');
      this.addTemplate('posts.extra', '<div class="posts-extra">postsExtra</div>');

      this.router.map(function () {
        this.route('posts', function () {});
        this.route('users', function () {});
      });

      this.add('route:posts', _routing.Route.extend({
        actions: {
          showModal() {
            this.render('posts/modal', {
              into: 'application',
              outlet: 'modal'
            });
          },
          hideModal() {
            this.disconnectOutlet({
              outlet: 'modal',
              parentView: 'application'
            });
          }
        }
      }));

      this.add('route:posts.index', _routing.Route.extend({
        actions: {
          showExtra() {
            this.render('posts/extra', {
              into: 'posts/index'
            });
          },
          hideExtra() {
            this.disconnectOutlet({ parentView: 'posts/index' });
          }
        }
      }));

      let rootElement = document.getElementById('qunit-fixture');

      return this.visit('/posts').then(() => {
        let router = this.applicationInstance.lookup('router:main');

        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-index')), 'postsIndex', 'The posts/index template was rendered');
        (0, _runloop.run)(() => router.send('showModal'));
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-modal')), 'postsModal', 'The posts/modal template was rendered');
        (0, _runloop.run)(() => router.send('showExtra'));

        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-extra')), 'postsExtra', 'The posts/extra template was rendered');
        (0, _runloop.run)(() => router.send('hideModal'));

        assert.equal(rootElement.querySelector('div.posts-modal'), null, 'The posts/modal template was removed');
        (0, _runloop.run)(() => router.send('hideExtra'));

        assert.equal(rootElement.querySelector('div.posts-extra'), null, 'The posts/extra template was removed');
        (0, _runloop.run)(function () {
          router.send('showModal');
        });
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-modal')), 'postsModal', 'The posts/modal template was rendered');
        (0, _runloop.run)(function () {
          router.send('showExtra');
        });
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-extra')), 'postsExtra', 'The posts/extra template was rendered');
        return this.visit('/users');
      }).then(() => {
        assert.equal(rootElement.querySelector('div.posts-index'), null, 'The posts/index template was removed');
        assert.equal(rootElement.querySelector('div.posts-modal'), null, 'The posts/modal template was removed');
        assert.equal(rootElement.querySelector('div.posts-extra'), null, 'The posts/extra template was removed');
      });
    }

    ['@test Route supports clearing outlet using string parameter'](assert) {
      this.addTemplate('application', "{{outlet}}{{outlet 'modal'}}");
      this.addTemplate('posts', '{{outlet}}');
      this.addTemplate('users', 'users');
      this.addTemplate('posts.index', '<div class="posts-index">postsIndex {{outlet}}</div>');
      this.addTemplate('posts.modal', '<div class="posts-modal">postsModal</div>');

      this.router.map(function () {
        this.route('posts', function () {});
        this.route('users', function () {});
      });

      this.add('route:posts', _routing.Route.extend({
        actions: {
          showModal() {
            this.render('posts/modal', {
              into: 'application',
              outlet: 'modal'
            });
          },
          hideModal() {
            this.disconnectOutlet('modal');
          }
        }
      }));

      let rootElement = document.getElementById('qunit-fixture');
      return this.visit('/posts').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-index')), 'postsIndex', 'The posts/index template was rendered');
        (0, _runloop.run)(() => router.send('showModal'));
        assert.equal((0, _internalTestHelpers.getTextOf)(rootElement.querySelector('div.posts-modal')), 'postsModal', 'The posts/modal template was rendered');
        (0, _runloop.run)(() => router.send('hideModal'));
        assert.equal(rootElement.querySelector('div.posts-modal'), null, 'The posts/modal template was removed');
        return this.visit('/users');
      }).then(() => {
        assert.equal(rootElement.querySelector('div.posts-index'), null, 'The posts/index template was removed');
        assert.equal(rootElement.querySelector('div.posts-modal'), null, 'The posts/modal template was removed');
      });
    }

    ['@test Route silently fails when cleaning an outlet from an inactive view'](assert) {
      assert.expect(1); // handleURL

      this.addTemplate('application', '{{outlet}}');
      this.addTemplate('posts', "{{outlet 'modal'}}");
      this.addTemplate('modal', 'A Yo.');

      this.router.map(function () {
        this.route('posts');
      });

      this.add('route:posts', _routing.Route.extend({
        actions: {
          hideSelf() {
            this.disconnectOutlet({
              outlet: 'main',
              parentView: 'application'
            });
          },
          showModal() {
            this.render('modal', { into: 'posts', outlet: 'modal' });
          },
          hideModal() {
            this.disconnectOutlet({ outlet: 'modal', parentView: 'posts' });
          }
        }
      }));

      return this.visit('/posts').then(() => {
        assert.ok(true, '/posts has been handled');
        let router = this.applicationInstance.lookup('router:main');
        (0, _runloop.run)(() => router.send('showModal'));
        (0, _runloop.run)(() => router.send('hideSelf'));
        (0, _runloop.run)(() => router.send('hideModal'));
      });
    }

    ['@test Router `willTransition` hook passes in cancellable transition'](assert) {
      // Should hit willTransition 3 times, once for the initial route, and then 2 more times
      // for the two handleURL calls below
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.expect(7);

          this.router.reopen({
            init() {
              this._super(...arguments);
              this.on('routeWillChange', transition => {
                assert.ok(true, 'routeWillChange was called');
                if (transition.intent && transition.intent.url !== '/') {
                  transition.abort();
                }
              });
            }
          });
        } else {
        assert.expect(5);
        this.router.reopen({
          willTransition(_, _2, transition) {
            assert.ok(true, 'willTransition was called');
            if (transition.intent.url !== '/') {
              transition.abort();
            }
          }
        });
      }

      this.router.map(function () {
        this.route('nork');
        this.route('about');
      });

      this.add('route:loading', _routing.Route.extend({
        activate() {
          assert.ok(false, 'LoadingRoute was not entered');
        }
      }));

      this.add('route:nork', _routing.Route.extend({
        activate() {
          assert.ok(false, 'NorkRoute was not entered');
        }
      }));

      this.add('route:about', _routing.Route.extend({
        activate() {
          assert.ok(false, 'AboutRoute was not entered');
        }
      }));

      return this.visit('/').then(() => {
        this.handleURLAborts(assert, '/nork');
        this.handleURLAborts(assert, '/about');
      });
    }

    ['@test Aborting/redirecting the transition in `willTransition` prevents LoadingRoute from being entered'](assert) {
      assert.expect(5);

      this.router.map(function () {
        this.route('index');
        this.route('nork');
        this.route('about');
      });

      let redirect = false;

      this.add('route:index', _routing.Route.extend({
        actions: {
          willTransition(transition) {
            assert.ok(true, 'willTransition was called');
            if (redirect) {
              // router.js won't refire `willTransition` for this redirect
              this.transitionTo('about');
            } else {
              transition.abort();
            }
          }
        }
      }));

      let deferred = null;

      this.add('route:loading', _routing.Route.extend({
        activate() {
          assert.ok(deferred, 'LoadingRoute should be entered at this time');
        },
        deactivate() {
          assert.ok(true, 'LoadingRoute was exited');
        }
      }));

      this.add('route:nork', _routing.Route.extend({
        activate() {
          assert.ok(true, 'NorkRoute was entered');
        }
      }));

      this.add('route:about', _routing.Route.extend({
        activate() {
          assert.ok(true, 'AboutRoute was entered');
        },
        model() {
          if (deferred) {
            return deferred.promise;
          }
        }
      }));

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        // Attempted transitions out of index should abort.
        (0, _runloop.run)(router, 'transitionTo', 'nork');
        (0, _runloop.run)(router, 'handleURL', '/nork');

        // Attempted transitions out of index should redirect to about
        redirect = true;
        (0, _runloop.run)(router, 'transitionTo', 'nork');
        (0, _runloop.run)(router, 'transitionTo', 'index');

        // Redirected transitions out of index to a route with a
        // promise model should pause the transition and
        // activate LoadingRoute
        deferred = _rsvp.default.defer();
        (0, _runloop.run)(router, 'transitionTo', 'nork');
        (0, _runloop.run)(deferred.resolve);
      });
    }

    ['@test `didTransition` event fires on the router'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('nork');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        router.one('didTransition', function () {
          assert.ok(true, 'didTransition fired on initial routing');
        });
        this.visit('/');
      }).then(() => {
        let router = this.applicationInstance.lookup('router:main');
        router.one('didTransition', function () {
          assert.ok(true, 'didTransition fired on the router');
          assert.equal(router.get('url'), '/nork', 'The url property is updated by the time didTransition fires');
        });

        return this.visit('/nork');
      });
    }

    ['@test `didTransition` can be reopened'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('nork');
      });
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.ok(true, 'no longer a valid test');
          return;
        } else {
        this.router.reopen({
          didTransition() {
            this._super(...arguments);
            assert.ok(true, 'reopened didTransition was called');
          }
        });
      }

      return this.visit('/');
    }

    ['@test `activate` event fires on the route'](assert) {
      assert.expect(2);

      let eventFired = 0;

      this.router.map(function () {
        this.route('nork');
      });

      this.add('route:nork', _routing.Route.extend({
        init() {
          this._super(...arguments);

          this.on('activate', function () {
            assert.equal(++eventFired, 1, 'activate event is fired once');
          });
        },

        activate() {
          assert.ok(true, 'activate hook is called');
        }
      }));

      return this.visit('/nork');
    }

    ['@test `deactivate` event fires on the route'](assert) {
      assert.expect(2);

      let eventFired = 0;

      this.router.map(function () {
        this.route('nork');
        this.route('dork');
      });

      this.add('route:nork', _routing.Route.extend({
        init() {
          this._super(...arguments);

          this.on('deactivate', function () {
            assert.equal(++eventFired, 1, 'deactivate event is fired once');
          });
        },

        deactivate() {
          assert.ok(true, 'deactivate hook is called');
        }
      }));

      return this.visit('/nork').then(() => this.visit('/dork'));
    }

    ['@test Actions can be handled by inherited action handlers'](assert) {
      assert.expect(4);

      let SuperRoute = _routing.Route.extend({
        actions: {
          foo() {
            assert.ok(true, 'foo');
          },
          bar(msg) {
            assert.equal(msg, 'HELLO');
          }
        }
      });

      let RouteMixin = _metal.Mixin.create({
        actions: {
          bar(msg) {
            assert.equal(msg, 'HELLO');
            this._super(msg);
          }
        }
      });

      this.add('route:home', SuperRoute.extend(RouteMixin, {
        actions: {
          baz() {
            assert.ok(true, 'baz');
          }
        }
      }));

      this.addTemplate('home', `
      <a class="do-foo" {{action "foo"}}>Do foo</a>
      <a class="do-bar-with-arg" {{action "bar" "HELLO"}}>Do bar with arg</a>
      <a class="do-baz" {{action "baz"}}>Do bar</a>
    `);

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        rootElement.querySelector('.do-foo').click();
        rootElement.querySelector('.do-bar-with-arg').click();
        rootElement.querySelector('.do-baz').click();
      });
    }

    ['@test transitionTo returns Transition when passed a route name'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('bar');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        let transition = (0, _runloop.run)(() => router.transitionTo('bar'));
        assert.equal(transition instanceof _router_js.InternalTransition, true);
      });
    }

    ['@test transitionTo returns Transition when passed a url'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('root', { path: '/' });
        this.route('bar', function () {
          this.route('baz');
        });
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        let transition = (0, _runloop.run)(() => router.transitionTo('/bar/baz'));
        assert.equal(transition instanceof _router_js.InternalTransition, true);
      });
    }

    ['@test currentRouteName is a property installed on ApplicationController that can be used in transitionTo'](assert) {
      assert.expect(24);

      this.router.map(function () {
        this.route('index', { path: '/' });
        this.route('be', function () {
          this.route('excellent', { resetNamespace: true }, function () {
            this.route('to', { resetNamespace: true }, function () {
              this.route('each', { resetNamespace: true }, function () {
                this.route('other');
              });
            });
          });
        });
      });

      return this.visit('/').then(() => {
        let appController = this.applicationInstance.lookup('controller:application');
        let router = this.applicationInstance.lookup('router:main');

        function transitionAndCheck(path, expectedPath, expectedRouteName) {
          if (path) {
            (0, _runloop.run)(router, 'transitionTo', path);
          }
          assert.equal(appController.get('currentPath'), expectedPath);
          assert.equal(appController.get('currentRouteName'), expectedRouteName);
        }

        transitionAndCheck(null, 'index', 'index');
        transitionAndCheck('/be', 'be.index', 'be.index');
        transitionAndCheck('/be/excellent', 'be.excellent.index', 'excellent.index');
        transitionAndCheck('/be/excellent/to', 'be.excellent.to.index', 'to.index');
        transitionAndCheck('/be/excellent/to/each', 'be.excellent.to.each.index', 'each.index');
        transitionAndCheck('/be/excellent/to/each/other', 'be.excellent.to.each.other', 'each.other');

        transitionAndCheck('index', 'index', 'index');
        transitionAndCheck('be', 'be.index', 'be.index');
        transitionAndCheck('excellent', 'be.excellent.index', 'excellent.index');
        transitionAndCheck('to.index', 'be.excellent.to.index', 'to.index');
        transitionAndCheck('each', 'be.excellent.to.each.index', 'each.index');
        transitionAndCheck('each.other', 'be.excellent.to.each.other', 'each.other');
      });
    }

    ['@test Route model hook finds the same model as a manual find'](assert) {
      let post;
      let Post = _runtime.Object.extend();
      this.add('model:post', Post);
      Post.reopenClass({
        find() {
          post = this;
          return {};
        }
      });

      this.router.map(function () {
        this.route('post', { path: '/post/:post_id' });
      });

      return this.visit('/post/1').then(() => {
        assert.equal(Post, post);
      });
    }

    ['@test Routes can refresh themselves causing their model hooks to be re-run'](assert) {
      this.router.map(function () {
        this.route('parent', { path: '/parent/:parent_id' }, function () {
          this.route('child');
        });
      });

      let appcount = 0;
      this.add('route:application', _routing.Route.extend({
        model() {
          ++appcount;
        }
      }));

      let parentcount = 0;
      this.add('route:parent', _routing.Route.extend({
        model(params) {
          assert.equal(params.parent_id, '123');
          ++parentcount;
        },
        actions: {
          refreshParent() {
            this.refresh();
          }
        }
      }));

      let childcount = 0;
      this.add('route:parent.child', _routing.Route.extend({
        model() {
          ++childcount;
        }
      }));

      let router;
      return this.visit('/').then(() => {
        router = this.applicationInstance.lookup('router:main');
        assert.equal(appcount, 1);
        assert.equal(parentcount, 0);
        assert.equal(childcount, 0);
        return (0, _runloop.run)(router, 'transitionTo', 'parent.child', '123');
      }).then(() => {
        assert.equal(appcount, 1);
        assert.equal(parentcount, 1);
        assert.equal(childcount, 1);
        return (0, _runloop.run)(router, 'send', 'refreshParent');
      }).then(() => {
        assert.equal(appcount, 1);
        assert.equal(parentcount, 2);
        assert.equal(childcount, 2);
      });
    }

    ['@test Specifying non-existent controller name in route#render throws'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.add('route:home', _routing.Route.extend({
        renderTemplate() {
          expectAssertion(() => {
            this.render('homepage', {
              controller: 'stefanpenneristhemanforme'
            });
          }, "You passed `controller: 'stefanpenneristhemanforme'` into the `render` method, but no such controller could be found.");
        }
      }));

      return this.visit('/');
    }

    ["@test Redirecting with null model doesn't error out"](assert) {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('about', { path: '/about/:hurhurhur' });
      });

      this.add('route:about', _routing.Route.extend({
        serialize: function (model) {
          if (model === null) {
            return { hurhurhur: 'TreeklesMcGeekles' };
          }
        }
      }));

      this.add('route:home', _routing.Route.extend({
        beforeModel() {
          this.transitionTo('about', null);
        }
      }));

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(router.get('location.path'), '/about/TreeklesMcGeekles');
      });
    }

    ['@test rejecting the model hooks promise with a non-error prints the `message` property'](assert) {
      assert.expect(5);

      let rejectedMessage = 'OMG!! SOOOOOO BAD!!!!';
      let rejectedStack = 'Yeah, buddy: stack gets printed too.';

      this.router.map(function () {
        this.route('yippie', { path: '/' });
      });

      console.error = function (initialMessage, errorMessage, errorStack) {
        assert.equal(initialMessage, 'Error while processing route: yippie', 'a message with the current route name is printed');
        assert.equal(errorMessage, rejectedMessage, "the rejected reason's message property is logged");
        assert.equal(errorStack, rejectedStack, "the rejected reason's stack property is logged");
      };

      this.add('route:yippie', _routing.Route.extend({
        model() {
          return _rsvp.default.reject({
            message: rejectedMessage,
            stack: rejectedStack
          });
        }
      }));

      return assert.throws(() => {
        return this.visit('/');
      }, function (err) {
        assert.equal(err.message, rejectedMessage);
        return true;
      }, 'expected an exception');
    }

    ['@test rejecting the model hooks promise with an error with `errorThrown` property prints `errorThrown.message` property'](assert) {
      assert.expect(5);
      let rejectedMessage = 'OMG!! SOOOOOO BAD!!!!';
      let rejectedStack = 'Yeah, buddy: stack gets printed too.';

      this.router.map(function () {
        this.route('yippie', { path: '/' });
      });

      console.error = function (initialMessage, errorMessage, errorStack) {
        assert.equal(initialMessage, 'Error while processing route: yippie', 'a message with the current route name is printed');
        assert.equal(errorMessage, rejectedMessage, "the rejected reason's message property is logged");
        assert.equal(errorStack, rejectedStack, "the rejected reason's stack property is logged");
      };

      this.add('route:yippie', _routing.Route.extend({
        model() {
          return _rsvp.default.reject({
            errorThrown: { message: rejectedMessage, stack: rejectedStack }
          });
        }
      }));

      assert.throws(() => this.visit('/'), function (err) {
        assert.equal(err.message, rejectedMessage);
        return true;
      }, 'expected an exception');
    }

    ['@test rejecting the model hooks promise with no reason still logs error'](assert) {
      assert.expect(2);
      this.router.map(function () {
        this.route('wowzers', { path: '/' });
      });

      console.error = function (initialMessage) {
        assert.equal(initialMessage, 'Error while processing route: wowzers', 'a message with the current route name is printed');
      };

      this.add('route:wowzers', _routing.Route.extend({
        model() {
          return _rsvp.default.reject();
        }
      }));

      return assert.throws(() => this.visit('/'));
    }

    ['@test rejecting the model hooks promise with a string shows a good error'](assert) {
      assert.expect(3);
      let rejectedMessage = 'Supercalifragilisticexpialidocious';

      this.router.map(function () {
        this.route('yondo', { path: '/' });
      });

      console.error = function (initialMessage, errorMessage) {
        assert.equal(initialMessage, 'Error while processing route: yondo', 'a message with the current route name is printed');
        assert.equal(errorMessage, rejectedMessage, "the rejected reason's message property is logged");
      };

      this.add('route:yondo', _routing.Route.extend({
        model() {
          return _rsvp.default.reject(rejectedMessage);
        }
      }));

      assert.throws(() => this.visit('/'), new RegExp(rejectedMessage), 'expected an exception');
    }

    ["@test willLeave, willChangeContext, willChangeModel actions don't fire unless feature flag enabled"](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('about');
      });

      function shouldNotFire() {
        assert.ok(false, "this action shouldn't have been received");
      }

      this.add('route:index', _routing.Route.extend({
        actions: {
          willChangeModel: shouldNotFire,
          willChangeContext: shouldNotFire,
          willLeave: shouldNotFire
        }
      }));

      this.add('route:about', _routing.Route.extend({
        setupController() {
          assert.ok(true, 'about route was entered');
        }
      }));

      return this.visit('/about');
    }

    ['@test Errors in transitionTo within redirect hook are logged'](assert) {
      assert.expect(4);
      let actual = [];

      this.router.map(function () {
        this.route('yondo', { path: '/' });
        this.route('stink-bomb');
      });

      this.add('route:yondo', _routing.Route.extend({
        redirect() {
          this.transitionTo('stink-bomb', { something: 'goes boom' });
        }
      }));

      console.error = function () {
        // push the arguments onto an array so we can detect if the error gets logged twice
        actual.push(arguments);
      };

      assert.throws(() => this.visit('/'), /More context objects were passed/);

      assert.equal(actual.length, 1, 'the error is only logged once');
      assert.equal(actual[0][0], 'Error while processing route: yondo', 'source route is printed');
      assert.ok(actual[0][1].match(/More context objects were passed than there are dynamic segments for the route: stink-bomb/), 'the error is printed');
    }

    ['@test Errors in transition show error template if available'](assert) {
      this.addTemplate('error', "<div id='error'>Error!</div>");

      this.router.map(function () {
        this.route('yondo', { path: '/' });
        this.route('stink-bomb');
      });

      this.add('route:yondo', _routing.Route.extend({
        redirect() {
          this.transitionTo('stink-bomb', { something: 'goes boom' });
        }
      }));
      console.error = () => {};

      return this.visit('/').then(() => {
        let rootElement = document.querySelector('#qunit-fixture');
        assert.equal(rootElement.querySelectorAll('#error').length, 1, 'Error template was rendered.');
      });
    }

    ['@test Route#resetController gets fired when changing models and exiting routes'](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('a', function () {
          this.route('b', { path: '/b/:id', resetNamespace: true }, function () {});
          this.route('c', { path: '/c/:id', resetNamespace: true }, function () {});
        });
        this.route('out');
      });

      let calls = [];

      let SpyRoute = _routing.Route.extend({
        setupController() /* controller, model, transition */{
          calls.push(['setup', this.routeName]);
        },

        resetController() /* controller */{
          calls.push(['reset', this.routeName]);
        }
      });

      this.add('route:a', SpyRoute.extend());
      this.add('route:b', SpyRoute.extend());
      this.add('route:c', SpyRoute.extend());
      this.add('route:out', SpyRoute.extend());

      let router;
      return this.visit('/').then(() => {
        router = this.applicationInstance.lookup('router:main');
        assert.deepEqual(calls, []);
        return (0, _runloop.run)(router, 'transitionTo', 'b', 'b-1');
      }).then(() => {
        assert.deepEqual(calls, [['setup', 'a'], ['setup', 'b']]);
        calls.length = 0;
        return (0, _runloop.run)(router, 'transitionTo', 'c', 'c-1');
      }).then(() => {
        assert.deepEqual(calls, [['reset', 'b'], ['setup', 'c']]);
        calls.length = 0;
        return (0, _runloop.run)(router, 'transitionTo', 'out');
      }).then(() => {
        assert.deepEqual(calls, [['reset', 'c'], ['reset', 'a'], ['setup', 'out']]);
      });
    }

    ['@test Exception during initialization of non-initial route is not swallowed'](assert) {
      this.router.map(function () {
        this.route('boom');
      });
      this.add('route:boom', _routing.Route.extend({
        init() {
          throw new Error('boom!');
        }
      }));

      return assert.throws(() => this.visit('/boom'), /\bboom\b/);
    }

    ['@test Exception during initialization of initial route is not swallowed'](assert) {
      this.router.map(function () {
        this.route('boom', { path: '/' });
      });
      this.add('route:boom', _routing.Route.extend({
        init() {
          throw new Error('boom!');
        }
      }));
      return assert.throws(() => this.visit('/'), /\bboom\b/);
    }

    ['@test {{outlet}} works when created after initial render'](assert) {
      this.addTemplate('sample', 'Hi{{#if showTheThing}}{{outlet}}{{/if}}Bye');
      this.addTemplate('sample.inner', 'Yay');
      this.addTemplate('sample.inner2', 'Boo');
      this.router.map(function () {
        this.route('sample', { path: '/' }, function () {
          this.route('inner', { path: '/' });
          this.route('inner2', { path: '/2' });
        });
      });

      let rootElement;
      return this.visit('/').then(() => {
        rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'HiBye', 'initial render');

        (0, _runloop.run)(() => this.applicationInstance.lookup('controller:sample').set('showTheThing', true));

        assert.equal(rootElement.textContent.trim(), 'HiYayBye', 'second render');
        return this.visit('/2');
      }).then(() => {
        assert.equal(rootElement.textContent.trim(), 'HiBooBye', 'third render');
      });
    }

    ['@test Can render into a named outlet at the top level'](assert) {
      this.addTemplate('application', 'A-{{outlet}}-B-{{outlet "other"}}-C');
      this.addTemplate('modal', 'Hello world');
      this.addTemplate('index', 'The index');
      this.router.map(function () {
        this.route('index', { path: '/' });
      });
      this.add('route:application', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('modal', {
            into: 'application',
            outlet: 'other'
          });
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'A-The index-B-Hello world-C', 'initial render');
      });
    }

    ['@test Can disconnect a named outlet at the top level'](assert) {
      this.addTemplate('application', 'A-{{outlet}}-B-{{outlet "other"}}-C');
      this.addTemplate('modal', 'Hello world');
      this.addTemplate('index', 'The index');
      this.router.map(function () {
        this.route('index', { path: '/' });
      });
      this.add('route:application', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('modal', {
            into: 'application',
            outlet: 'other'
          });
        },
        actions: {
          banish() {
            this.disconnectOutlet({
              parentView: 'application',
              outlet: 'other'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'A-The index-B-Hello world-C', 'initial render');

        (0, _runloop.run)(this.applicationInstance.lookup('router:main'), 'send', 'banish');

        assert.equal(rootElement.textContent.trim(), 'A-The index-B--C', 'second render');
      });
    }

    ['@test Can render into a named outlet at the top level, with empty main outlet'](assert) {
      this.addTemplate('application', 'A-{{outlet}}-B-{{outlet "other"}}-C');
      this.addTemplate('modal', 'Hello world');

      this.router.map(function () {
        this.route('hasNoTemplate', { path: '/' });
      });

      this.add('route:application', _routing.Route.extend({
        renderTemplate() {
          this.render();
          this.render('modal', {
            into: 'application',
            outlet: 'other'
          });
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'A--B-Hello world-C', 'initial render');
      });
    }

    ['@test Can render into a named outlet at the top level, later'](assert) {
      this.addTemplate('application', 'A-{{outlet}}-B-{{outlet "other"}}-C');
      this.addTemplate('modal', 'Hello world');
      this.addTemplate('index', 'The index');
      this.router.map(function () {
        this.route('index', { path: '/' });
      });
      this.add('route:application', _routing.Route.extend({
        actions: {
          launch() {
            this.render('modal', {
              into: 'application',
              outlet: 'other'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.textContent.trim(), 'A-The index-B--C', 'initial render');
        (0, _runloop.run)(this.applicationInstance.lookup('router:main'), 'send', 'launch');
        assert.equal(rootElement.textContent.trim(), 'A-The index-B-Hello world-C', 'second render');
      });
    }

    ["@test Can render routes with no 'main' outlet and their children"](assert) {
      this.addTemplate('application', '<div id="application">{{outlet "app"}}</div>');
      this.addTemplate('app', '<div id="app-common">{{outlet "common"}}</div><div id="app-sub">{{outlet "sub"}}</div>');
      this.addTemplate('common', '<div id="common"></div>');
      this.addTemplate('sub', '<div id="sub"></div>');

      this.router.map(function () {
        this.route('app', { path: '/app' }, function () {
          this.route('sub', { path: '/sub', resetNamespace: true });
        });
      });

      this.add('route:app', _routing.Route.extend({
        renderTemplate() {
          this.render('app', {
            outlet: 'app',
            into: 'application'
          });
          this.render('common', {
            outlet: 'common',
            into: 'app'
          });
        }
      }));

      this.add('route:sub', _routing.Route.extend({
        renderTemplate() {
          this.render('sub', {
            outlet: 'sub',
            into: 'app'
          });
        }
      }));

      let rootElement;
      return this.visit('/app').then(() => {
        rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.querySelectorAll('#app-common #common').length, 1, 'Finds common while viewing /app');
        return this.visit('/app/sub');
      }).then(() => {
        assert.equal(rootElement.querySelectorAll('#app-common #common').length, 1, 'Finds common while viewing /app/sub');
        assert.equal(rootElement.querySelectorAll('#app-sub #sub').length, 1, 'Finds sub while viewing /app/sub');
      });
    }

    ['@test Tolerates stacked renders'](assert) {
      this.addTemplate('application', '{{outlet}}{{outlet "modal"}}');
      this.addTemplate('index', 'hi');
      this.addTemplate('layer', 'layer');
      this.router.map(function () {
        this.route('index', { path: '/' });
      });
      this.add('route:application', _routing.Route.extend({
        actions: {
          openLayer() {
            this.render('layer', {
              into: 'application',
              outlet: 'modal'
            });
          },
          close() {
            this.disconnectOutlet({
              outlet: 'modal',
              parentView: 'application'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(rootElement.textContent.trim(), 'hi');
        (0, _runloop.run)(router, 'send', 'openLayer');
        assert.equal(rootElement.textContent.trim(), 'hilayer');
        (0, _runloop.run)(router, 'send', 'openLayer');
        assert.equal(rootElement.textContent.trim(), 'hilayer');
        (0, _runloop.run)(router, 'send', 'close');
        assert.equal(rootElement.textContent.trim(), 'hi');
      });
    }

    ['@test Renders child into parent with non-default template name'](assert) {
      this.addTemplate('application', '<div class="a">{{outlet}}</div>');
      this.addTemplate('exports.root', '<div class="b">{{outlet}}</div>');
      this.addTemplate('exports.index', '<div class="c"></div>');

      this.router.map(function () {
        this.route('root', function () {});
      });

      this.add('route:root', _routing.Route.extend({
        renderTemplate() {
          this.render('exports/root');
        }
      }));

      this.add('route:root.index', _routing.Route.extend({
        renderTemplate() {
          this.render('exports/index');
        }
      }));

      return this.visit('/root').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        assert.equal(rootElement.querySelectorAll('.a .b .c').length, 1);
      });
    }

    ["@test Allows any route to disconnectOutlet another route's templates"](assert) {
      this.addTemplate('application', '{{outlet}}{{outlet "modal"}}');
      this.addTemplate('index', 'hi');
      this.addTemplate('layer', 'layer');
      this.router.map(function () {
        this.route('index', { path: '/' });
      });
      this.add('route:application', _routing.Route.extend({
        actions: {
          openLayer() {
            this.render('layer', {
              into: 'application',
              outlet: 'modal'
            });
          }
        }
      }));
      this.add('route:index', _routing.Route.extend({
        actions: {
          close() {
            this.disconnectOutlet({
              parentView: 'application',
              outlet: 'modal'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let rootElement = document.getElementById('qunit-fixture');
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(rootElement.textContent.trim(), 'hi');
        (0, _runloop.run)(router, 'send', 'openLayer');
        assert.equal(rootElement.textContent.trim(), 'hilayer');
        (0, _runloop.run)(router, 'send', 'close');
        assert.equal(rootElement.textContent.trim(), 'hi');
      });
    }

    ['@test Components inside an outlet have their didInsertElement hook invoked when the route is displayed'](assert) {
      this.addTemplate('index', '{{#if showFirst}}{{my-component}}{{else}}{{other-component}}{{/if}}');

      let myComponentCounter = 0;
      let otherComponentCounter = 0;
      let indexController;

      this.router.map(function () {
        this.route('index', { path: '/' });
      });

      this.add('controller:index', _controller.default.extend({
        showFirst: true
      }));

      this.add('route:index', _routing.Route.extend({
        setupController(controller) {
          indexController = controller;
        }
      }));

      this.add('component:my-component', _glimmer.Component.extend({
        didInsertElement() {
          myComponentCounter++;
        }
      }));

      this.add('component:other-component', _glimmer.Component.extend({
        didInsertElement() {
          otherComponentCounter++;
        }
      }));

      return this.visit('/').then(() => {
        assert.strictEqual(myComponentCounter, 1, 'didInsertElement invoked on displayed component');
        assert.strictEqual(otherComponentCounter, 0, 'didInsertElement not invoked on displayed component');

        (0, _runloop.run)(() => indexController.set('showFirst', false));

        assert.strictEqual(myComponentCounter, 1, 'didInsertElement not invoked on displayed component');
        assert.strictEqual(otherComponentCounter, 1, 'didInsertElement invoked on displayed component');
      });
    }

    ['@test Doesnt swallow exception thrown from willTransition'](assert) {
      assert.expect(1);
      this.addTemplate('application', '{{outlet}}');
      this.addTemplate('index', 'index');
      this.addTemplate('other', 'other');

      this.router.map(function () {
        this.route('index', { path: '/' });
        this.route('other', function () {});
      });

      this.add('route:index', _routing.Route.extend({
        actions: {
          willTransition() {
            throw new Error('boom');
          }
        }
      }));

      return this.visit('/').then(() => {
        return assert.throws(() => {
          return this.visit('/other');
        }, /boom/, 'expected an exception but none was thrown');
      });
    }

    ['@test Exception if outlet name is undefined in render and disconnectOutlet']() {
      this.add('route:application', _routing.Route.extend({
        actions: {
          showModal() {
            this.render({
              outlet: undefined,
              parentView: 'application'
            });
          },
          hideModal() {
            this.disconnectOutlet({
              outlet: undefined,
              parentView: 'application'
            });
          }
        }
      }));

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        expectAssertion(() => {
          (0, _runloop.run)(() => router.send('showModal'));
        }, /You passed undefined as the outlet name/);

        expectAssertion(() => {
          (0, _runloop.run)(() => router.send('hideModal'));
        }, /You passed undefined as the outlet name/);
      });
    }

    ['@test Route serializers work for Engines'](assert) {
      assert.expect(2);

      // Register engine
      let BlogEngine = _engine.default.extend();
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let postSerialize = function (params) {
        assert.ok(true, 'serialize hook runs');
        return {
          post_id: params.id
        };
      };
      let BlogMap = function () {
        this.route('post', {
          path: '/post/:post_id',
          serialize: postSerialize
        });
      };
      this.add('route-map:blog', BlogMap);

      this.router.map(function () {
        this.mount('blog');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        assert.equal(router._routerMicrolib.generate('blog.post', { id: '13' }), '/blog/post/13', 'url is generated properly');
      });
    }

    ['@test Defining a Route#serialize method in an Engine throws an error'](assert) {
      assert.expect(1);

      // Register engine
      let BlogEngine = _engine.default.extend();
      this.add('engine:blog', BlogEngine);

      // Register engine route map
      let BlogMap = function () {
        this.route('post');
      };
      this.add('route-map:blog', BlogMap);

      this.router.map(function () {
        this.mount('blog');
      });

      return this.visit('/').then(() => {
        let router = this.applicationInstance.lookup('router:main');
        let PostRoute = _routing.Route.extend({ serialize() {} });
        this.applicationInstance.lookup('engine:blog').register('route:post', PostRoute);

        assert.throws(() => router.transitionTo('blog.post'), /Defining a custom serialize method on an Engine route is not supported/);
      });
    }

    ['@test App.destroy does not leave undestroyed views after clearing engines'](assert) {
      assert.expect(4);

      let engineInstance;
      // Register engine
      let BlogEngine = _engine.default.extend();
      this.add('engine:blog', BlogEngine);
      let EngineIndexRoute = _routing.Route.extend({
        init() {
          this._super(...arguments);
          engineInstance = (0, _owner.getOwner)(this);
        }
      });

      // Register engine route map
      let BlogMap = function () {
        this.route('post');
      };
      this.add('route-map:blog', BlogMap);

      this.router.map(function () {
        this.mount('blog');
      });

      return this.visit('/').then(() => {
        let engine = this.applicationInstance.lookup('engine:blog');
        engine.register('route:index', EngineIndexRoute);
        engine.register('template:index', (0, _emberTemplateCompiler.compile)('Engine Post!'));
        return this.visit('/blog');
      }).then(() => {
        assert.ok(true, '/blog has been handled');
        let route = engineInstance.lookup('route:index');
        let router = this.applicationInstance.lookup('router:main');

        (0, _runloop.run)(router, 'destroy');
        assert.equal(router._toplevelView, null, 'the toplevelView was cleared');

        (0, _runloop.run)(route, 'destroy');
        assert.equal(router._toplevelView, null, 'the toplevelView was not reinitialized');

        (0, _runloop.run)(this.applicationInstance, 'destroy');
        assert.equal(router._toplevelView, null, 'the toplevelView was not reinitialized');
      });
    }

    ["@test Generated route should be an instance of App's default route if provided"](assert) {
      let generatedRoute;

      this.router.map(function () {
        this.route('posts');
      });

      let AppRoute = _routing.Route.extend();
      this.add('route:basic', AppRoute);

      return this.visit('/posts').then(() => {
        generatedRoute = this.applicationInstance.lookup('route:posts');

        assert.ok(generatedRoute instanceof AppRoute, 'should extend the correct route');
      });
    }
  });
});
enifed('ember/tests/routing/deprecated_handler_infos_test', ['internal-test-helpers'], function (_internalTestHelpers) {
  'use strict';

  if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
      (0, _internalTestHelpers.moduleFor)('Deprecated HandlerInfos', class extends _internalTestHelpers.ApplicationTestCase {
        constructor() {
          super(...arguments);
          this.router.map(function () {
            this.route('parent', function () {
              this.route('child');
              this.route('sibling');
            });
          });
        }

        get routerOptions() {
          return {
            willTransition(oldHandlers, newHandlers, transition) {
              expectDeprecation(() => {
                this._routerMicrolib.currentHandlerInfos;
              }, 'You attempted to use "_routerMicrolib.currentHandlerInfos" which is a private API that will be removed.');

              expectDeprecation(() => {
                this._routerMicrolib.getHandler('parent');
              }, 'You attempted to use "_routerMicrolib.getHandler" which is a private API that will be removed.');

              oldHandlers.forEach(handler => {
                expectDeprecation(() => {
                  handler.handler;
                }, 'You attempted to read "handlerInfo.handler" which is a private API that will be removed.');
              });
              newHandlers.forEach(handler => {
                expectDeprecation(() => {
                  handler.handler;
                }, 'You attempted to read "handlerInfo.handler" which is a private API that will be removed.');
              });

              expectDeprecation(() => {
                transition.handlerInfos;
              }, 'You attempted to use "transition.handlerInfos" which is a private API that will be removed.');

              expectDeprecation(() => {
                transition.state.handlerInfos;
              }, 'You attempted to use "transition.state.handlerInfos" which is a private API that will be removed.');
              QUnit.assert.ok(true, 'willTransition');
            },

            didTransition(newHandlers) {
              newHandlers.forEach(handler => {
                expectDeprecation(() => {
                  handler.handler;
                }, 'You attempted to read "handlerInfo.handler" which is a private API that will be removed.');
              });
              QUnit.assert.ok(true, 'didTransition');
            }
          };
        }

        '@test handlerInfos are deprecated and associated private apis'(assert) {
          let done = assert.async();
          expectDeprecation(() => {
            return this.visit('/parent').then(() => {
              done();
            });
          }, /You attempted to override the \"(willTransition|didTransition)\" method which is deprecated. Please inject the router service and listen to the \"(routeWillChange|routeDidChange)\" event\./);
        }
      });
    }
});
enifed('ember/tests/routing/deprecated_transition_state_test', ['internal-test-helpers'], function (_internalTestHelpers) {
  'use strict';

  if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
      (0, _internalTestHelpers.moduleFor)('Deprecated Transition State', class extends _internalTestHelpers.RouterTestCase {
        '@test touching transition.state is deprecated'(assert) {
          assert.expect(1);
          return this.visit('/').then(() => {
            this.routerService.on('routeWillChange', transition => {
              expectDeprecation(() => {
                transition.state;
              }, 'You attempted to read "transition.state" which is a private API. You should read the `RouteInfo` object on "transition.to" or "transition.from" which has the public state on it.');
            });
            return this.routerService.transitionTo('/child');
          });
        }

        '@test touching transition.queryParams is deprecated'(assert) {
          assert.expect(1);
          return this.visit('/').then(() => {
            this.routerService.on('routeWillChange', transition => {
              expectDeprecation(() => {
                transition.queryParams;
              }, 'You attempted to read "transition.queryParams" which is a private API. You should read the `RouteInfo` object on "transition.to" or "transition.from" which has the queryParams on it.');
            });
            return this.routerService.transitionTo('/child');
          });
        }

        '@test touching transition.params is deprecated'(assert) {
          assert.expect(1);
          return this.visit('/').then(() => {
            this.routerService.on('routeWillChange', transition => {
              expectDeprecation(() => {
                transition.params;
              }, 'You attempted to read "transition.params" which is a private API. You should read the `RouteInfo` object on "transition.to" or "transition.from" which has the params on it.');
            });
            return this.routerService.transitionTo('/child');
          });
        }
      });
    }
});
enifed('ember/tests/routing/query_params_test', ['@ember/controller', '@ember/string', '@ember/-internals/runtime', '@ember/runloop', '@ember/-internals/meta', '@ember/-internals/metal', '@ember/-internals/routing', 'router_js', 'internal-test-helpers'], function (_controller, _string, _runtime, _runloop, _meta, _metal, _routing, _router_js, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Query Params - main', class extends _internalTestHelpers.QueryParamTestCase {
    refreshModelWhileLoadingTest(loadingReturn) {
      let assert = this.assert;

      assert.expect(9);

      let appModelCount = 0;
      let promiseResolve;

      this.add('route:application', _routing.Route.extend({
        queryParams: {
          appomg: {
            defaultValue: 'applol'
          }
        },
        model() /* params */{
          appModelCount++;
        }
      }));

      this.setSingleQPController('index', 'omg', undefined, {
        omg: undefined
      });

      let actionName = typeof loadingReturn !== 'undefined' ? 'loading' : 'ignore';
      let indexModelCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: {
          omg: {
            refreshModel: true
          }
        },
        actions: {
          [actionName]: function () {
            return loadingReturn;
          }
        },
        model(params) {
          indexModelCount++;
          if (indexModelCount === 2) {
            assert.deepEqual(params, { omg: 'lex' });
            return new _runtime.RSVP.Promise(function (resolve) {
              promiseResolve = resolve;
              return;
            });
          } else if (indexModelCount === 3) {
            assert.deepEqual(params, { omg: 'hello' }, "Model hook reruns even if the previous one didn't finish");
          }
        }
      }));

      return this.visit('/').then(() => {
        assert.equal(appModelCount, 1, 'appModelCount is 1');
        assert.equal(indexModelCount, 1);

        let indexController = this.getController('index');
        this.setAndFlush(indexController, 'omg', 'lex');

        assert.equal(appModelCount, 1, 'appModelCount is 1');
        assert.equal(indexModelCount, 2);

        this.setAndFlush(indexController, 'omg', 'hello');
        assert.equal(appModelCount, 1, 'appModelCount is 1');
        assert.equal(indexModelCount, 3);

        (0, _runloop.run)(function () {
          promiseResolve();
        });

        assert.equal((0, _metal.get)(indexController, 'omg'), 'hello', 'At the end last value prevails');
      });
    }

    ["@test No replaceURL occurs on startup because default values don't show up in URL"](assert) {
      assert.expect(1);

      this.setSingleQPController('index');

      return this.visitAndAssert('/');
    }

    ['@test Calling transitionTo does not lose query params already on the activeTransition'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('parent', function () {
          this.route('child');
          this.route('sibling');
        });
      });

      this.add('route:parent.child', _routing.Route.extend({
        afterModel() {
          this.transitionTo('parent.sibling');
        }
      }));

      this.setSingleQPController('parent');

      return this.visit('/parent/child?foo=lol').then(() => {
        this.assertCurrentPath('/parent/sibling?foo=lol', 'redirected to the sibling route, instead of child route');
        assert.equal(this.getController('parent').get('foo'), 'lol', 'controller has value from the active transition');
      });
    }

    ['@test Single query params can be set on the controller and reflected in the url'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.setSingleQPController('home');

      return this.visitAndAssert('/').then(() => {
        let controller = this.getController('home');

        this.setAndFlush(controller, 'foo', '456');
        this.assertCurrentPath('/?foo=456');

        this.setAndFlush(controller, 'foo', '987');
        this.assertCurrentPath('/?foo=987');
      });
    }

    ['@test Query params can map to different url keys configured on the controller'](assert) {
      assert.expect(6);

      this.add('controller:index', _controller.default.extend({
        queryParams: [{ foo: 'other_foo', bar: { as: 'other_bar' } }],
        foo: 'FOO',
        bar: 'BAR'
      }));

      return this.visitAndAssert('/').then(() => {
        let controller = this.getController('index');

        this.setAndFlush(controller, 'foo', 'LEX');
        this.assertCurrentPath('/?other_foo=LEX', "QP mapped correctly without 'as'");

        this.setAndFlush(controller, 'foo', 'WOO');
        this.assertCurrentPath('/?other_foo=WOO', "QP updated correctly without 'as'");

        this.transitionTo('/?other_foo=NAW');
        assert.equal(controller.get('foo'), 'NAW', 'QP managed correctly on URL transition');

        this.setAndFlush(controller, 'bar', 'NERK');
        this.assertCurrentPath('/?other_bar=NERK&other_foo=NAW', "QP mapped correctly with 'as'");

        this.setAndFlush(controller, 'bar', 'NUKE');
        this.assertCurrentPath('/?other_bar=NUKE&other_foo=NAW', "QP updated correctly with 'as'");
      });
    }

    ['@test Routes have a private overridable serializeQueryParamKey hook'](assert) {
      assert.expect(2);

      this.add('route:index', _routing.Route.extend({
        serializeQueryParamKey: _string.dasherize
      }));

      this.setSingleQPController('index', 'funTimes', '');

      return this.visitAndAssert('/').then(() => {
        let controller = this.getController('index');

        this.setAndFlush(controller, 'funTimes', 'woot');
        this.assertCurrentPath('/?fun-times=woot');
      });
    }

    ['@test Can override inherited QP behavior by specifying queryParams as a computed property'](assert) {
      assert.expect(3);

      this.setSingleQPController('index', 'a', 0, {
        queryParams: (0, _metal.computed)(function () {
          return ['c'];
        }),
        c: true
      });

      return this.visitAndAssert('/').then(() => {
        let indexController = this.getController('index');

        this.setAndFlush(indexController, 'a', 1);
        this.assertCurrentPath('/', 'QP did not update due to being overriden');

        this.setAndFlush(indexController, 'c', false);
        this.assertCurrentPath('/?c=false', 'QP updated with overridden param');
      });
    }

    ['@test Can concatenate inherited QP behavior by specifying queryParams as an array'](assert) {
      assert.expect(3);

      this.setSingleQPController('index', 'a', 0, {
        queryParams: ['c'],
        c: true
      });

      return this.visitAndAssert('/').then(() => {
        let indexController = this.getController('index');

        this.setAndFlush(indexController, 'a', 1);
        this.assertCurrentPath('/?a=1', 'Inherited QP did update');

        this.setAndFlush(indexController, 'c', false);
        this.assertCurrentPath('/?a=1&c=false', 'New QP did update');
      });
    }

    ['@test model hooks receives query params'](assert) {
      assert.expect(2);

      this.setSingleQPController('index');

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { foo: 'bar' });
        }
      }));

      return this.visitAndAssert('/');
    }

    ['@test model hooks receives query params with dynamic segment params'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:id' });
      });

      this.setSingleQPController('index');

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { foo: 'bar', id: 'baz' });
        }
      }));

      return this.visitAndAssert('/baz');
    }

    ['@test model hooks receives query params (overridden by incoming url value)'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:id' });
      });

      this.setSingleQPController('index');

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { foo: 'baz', id: 'boo' });
        }
      }));

      return this.visitAndAssert('/boo?foo=baz');
    }

    ['@test error is thrown if dynamic segment and query param have same name'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('index', { path: '/:foo' });
      });

      this.setSingleQPController('index');

      expectAssertion(() => {
        this.visitAndAssert('/boo?foo=baz');
      }, `The route 'index' has both a dynamic segment and query param with name 'foo'. Please rename one to avoid collisions.`);
    }

    ['@test query params have been set by the time setupController is called'](assert) {
      assert.expect(2);

      this.setSingleQPController('application');

      this.add('route:application', _routing.Route.extend({
        setupController(controller) {
          assert.equal(controller.get('foo'), 'YEAH', "controller's foo QP property set before setupController called");
        }
      }));

      return this.visitAndAssert('/?foo=YEAH');
    }

    ['@test mapped query params have been set by the time setupController is called'](assert) {
      assert.expect(2);

      this.setSingleQPController('application', { faz: 'foo' });

      this.add('route:application', _routing.Route.extend({
        setupController(controller) {
          assert.equal(controller.get('faz'), 'YEAH', "controller's foo QP property set before setupController called");
        }
      }));

      return this.visitAndAssert('/?foo=YEAH');
    }

    ['@test Route#paramsFor fetches query params with default value'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:something' });
      });

      this.setSingleQPController('index');

      this.add('route:index', _routing.Route.extend({
        model() /* params, transition */{
          assert.deepEqual(this.paramsFor('index'), { something: 'baz', foo: 'bar' }, 'could retrieve params for index');
        }
      }));

      return this.visitAndAssert('/baz');
    }

    ['@test Route#paramsFor fetches query params with non-default value'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:something' });
      });

      this.setSingleQPController('index');

      this.add('route:index', _routing.Route.extend({
        model() /* params, transition */{
          assert.deepEqual(this.paramsFor('index'), { something: 'baz', foo: 'boo' }, 'could retrieve params for index');
        }
      }));

      return this.visitAndAssert('/baz?foo=boo');
    }

    ['@test Route#paramsFor fetches default falsy query params'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:something' });
      });

      this.setSingleQPController('index', 'foo', false);

      this.add('route:index', _routing.Route.extend({
        model() /* params, transition */{
          assert.deepEqual(this.paramsFor('index'), { something: 'baz', foo: false }, 'could retrieve params for index');
        }
      }));

      return this.visitAndAssert('/baz');
    }

    ['@test Route#paramsFor fetches non-default falsy query params'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('index', { path: '/:something' });
      });

      this.setSingleQPController('index', 'foo', true);

      this.add('route:index', _routing.Route.extend({
        model() /* params, transition */{
          assert.deepEqual(this.paramsFor('index'), { something: 'baz', foo: false }, 'could retrieve params for index');
        }
      }));

      return this.visitAndAssert('/baz?foo=false');
    }

    ['@test model hook can query prefix-less application params'](assert) {
      assert.expect(4);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      this.add('route:application', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { appomg: 'applol' });
        }
      }));

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { omg: 'lol' });
          assert.deepEqual(this.paramsFor('application'), {
            appomg: 'applol'
          });
        }
      }));

      return this.visitAndAssert('/');
    }

    ['@test model hook can query prefix-less application params (overridden by incoming url value)'](assert) {
      assert.expect(4);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      this.add('route:application', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { appomg: 'appyes' });
        }
      }));

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { omg: 'yes' });
          assert.deepEqual(this.paramsFor('application'), {
            appomg: 'appyes'
          });
        }
      }));

      return this.visitAndAssert('/?appomg=appyes&omg=yes');
    }

    ['@test can opt into full transition by setting refreshModel in route queryParams'](assert) {
      assert.expect(7);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      let appModelCount = 0;
      this.add('route:application', _routing.Route.extend({
        model() /* params, transition */{
          appModelCount++;
        }
      }));

      let indexModelCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: {
          omg: {
            refreshModel: true
          }
        },
        model(params) {
          indexModelCount++;

          if (indexModelCount === 1) {
            assert.deepEqual(params, { omg: 'lol' }, 'params are correct on first pass');
          } else if (indexModelCount === 2) {
            assert.deepEqual(params, { omg: 'lex' }, 'params are correct on second pass');
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal(appModelCount, 1, 'app model hook ran');
        assert.equal(indexModelCount, 1, 'index model hook ran');

        let indexController = this.getController('index');
        this.setAndFlush(indexController, 'omg', 'lex');

        assert.equal(appModelCount, 1, 'app model hook did not run again');
        assert.equal(indexModelCount, 2, 'index model hook ran again due to refreshModel');
      });
    }

    ['@test refreshModel and replace work together'](assert) {
      assert.expect(8);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      let appModelCount = 0;
      this.add('route:application', _routing.Route.extend({
        model() /* params */{
          appModelCount++;
        }
      }));

      let indexModelCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: {
          omg: {
            refreshModel: true,
            replace: true
          }
        },
        model(params) {
          indexModelCount++;

          if (indexModelCount === 1) {
            assert.deepEqual(params, { omg: 'lol' }, 'params are correct on first pass');
          } else if (indexModelCount === 2) {
            assert.deepEqual(params, { omg: 'lex' }, 'params are correct on second pass');
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal(appModelCount, 1, 'app model hook ran');
        assert.equal(indexModelCount, 1, 'index model hook ran');

        let indexController = this.getController('index');
        this.expectedReplaceURL = '/?omg=lex';
        this.setAndFlush(indexController, 'omg', 'lex');

        assert.equal(appModelCount, 1, 'app model hook did not run again');
        assert.equal(indexModelCount, 2, 'index model hook ran again due to refreshModel');
      });
    }

    ['@test multiple QP value changes only cause a single model refresh'](assert) {
      assert.expect(2);

      this.setSingleQPController('index', 'alex', 'lol');
      this.setSingleQPController('index', 'steely', 'lel');

      let refreshCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: {
          alex: {
            refreshModel: true
          },
          steely: {
            refreshModel: true
          }
        },
        refresh() {
          refreshCount++;
        }
      }));

      return this.visitAndAssert('/').then(() => {
        let indexController = this.getController('index');
        (0, _runloop.run)(indexController, 'setProperties', {
          alex: 'fran',
          steely: 'david'
        });
        assert.equal(refreshCount, 1, 'index refresh hook only run once');
      });
    }

    ['@test refreshModel does not cause a second transition during app boot '](assert) {
      assert.expect(1);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      this.add('route:index', _routing.Route.extend({
        queryParams: {
          omg: {
            refreshModel: true
          }
        },
        refresh() {
          assert.ok(false);
        }
      }));

      return this.visitAndAssert('/?appomg=hello&omg=world');
    }

    ['@test queryParams are updated when a controller property is set and the route is refreshed. Issue #13263  '](assert) {
      this.addTemplate('application', '<button id="test-button" {{action \'increment\'}}>Increment</button><span id="test-value">{{foo}}</span>{{outlet}}');

      this.setSingleQPController('application', 'foo', 1, {
        actions: {
          increment() {
            this.incrementProperty('foo');
            this.send('refreshRoute');
          }
        }
      });

      this.add('route:application', _routing.Route.extend({
        actions: {
          refreshRoute() {
            this.refresh();
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal((0, _internalTestHelpers.getTextOf)(document.getElementById('test-value')), '1');

        (0, _runloop.run)(document.getElementById('test-button'), 'click');
        assert.equal((0, _internalTestHelpers.getTextOf)(document.getElementById('test-value')), '2');
        this.assertCurrentPath('/?foo=2');

        (0, _runloop.run)(document.getElementById('test-button'), 'click');
        assert.equal((0, _internalTestHelpers.getTextOf)(document.getElementById('test-value')), '3');
        this.assertCurrentPath('/?foo=3');
      });
    }

    ["@test Use Ember.get to retrieve query params 'refreshModel' configuration"](assert) {
      assert.expect(7);

      this.setSingleQPController('application', 'appomg', 'applol');
      this.setSingleQPController('index', 'omg', 'lol');

      let appModelCount = 0;
      this.add('route:application', _routing.Route.extend({
        model() /* params */{
          appModelCount++;
        }
      }));

      let indexModelCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: _runtime.Object.create({
          unknownProperty() {
            return { refreshModel: true };
          }
        }),
        model(params) {
          indexModelCount++;

          if (indexModelCount === 1) {
            assert.deepEqual(params, { omg: 'lol' });
          } else if (indexModelCount === 2) {
            assert.deepEqual(params, { omg: 'lex' });
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal(appModelCount, 1);
        assert.equal(indexModelCount, 1);

        let indexController = this.getController('index');
        this.setAndFlush(indexController, 'omg', 'lex');

        assert.equal(appModelCount, 1);
        assert.equal(indexModelCount, 2);
      });
    }

    ['@test can use refreshModel even with URL changes that remove QPs from address bar'](assert) {
      assert.expect(4);

      this.setSingleQPController('index', 'omg', 'lol');

      let indexModelCount = 0;
      this.add('route:index', _routing.Route.extend({
        queryParams: {
          omg: {
            refreshModel: true
          }
        },
        model(params) {
          indexModelCount++;

          let data;
          if (indexModelCount === 1) {
            data = 'foo';
          } else if (indexModelCount === 2) {
            data = 'lol';
          }

          assert.deepEqual(params, { omg: data }, 'index#model receives right data');
        }
      }));

      return this.visitAndAssert('/?omg=foo').then(() => {
        this.transitionTo('/');

        let indexController = this.getController('index');
        assert.equal(indexController.get('omg'), 'lol');
      });
    }

    ['@test can opt into a replace query by specifying replace:true in the Route config hash'](assert) {
      assert.expect(2);

      this.setSingleQPController('application', 'alex', 'matchneer');

      this.add('route:application', _routing.Route.extend({
        queryParams: {
          alex: {
            replace: true
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        let appController = this.getController('application');
        this.expectedReplaceURL = '/?alex=wallace';
        this.setAndFlush(appController, 'alex', 'wallace');
      });
    }

    ['@test Route query params config can be configured using property name instead of URL key'](assert) {
      assert.expect(2);

      this.add('controller:application', _controller.default.extend({
        queryParams: [{ commitBy: 'commit_by' }]
      }));

      this.add('route:application', _routing.Route.extend({
        queryParams: {
          commitBy: {
            replace: true
          }
        }
      }));

      return this.visitAndAssert('/').then(() => {
        let appController = this.getController('application');
        this.expectedReplaceURL = '/?commit_by=igor_seb';
        this.setAndFlush(appController, 'commitBy', 'igor_seb');
      });
    }

    ['@test An explicit replace:false on a changed QP always wins and causes a pushState'](assert) {
      assert.expect(3);

      this.add('controller:application', _controller.default.extend({
        queryParams: ['alex', 'steely'],
        alex: 'matchneer',
        steely: 'dan'
      }));

      this.add('route:application', _routing.Route.extend({
        queryParams: {
          alex: {
            replace: true
          },
          steely: {
            replace: false
          }
        }
      }));

      return this.visit('/').then(() => {
        let appController = this.getController('application');
        this.expectedPushURL = '/?alex=wallace&steely=jan';
        (0, _runloop.run)(appController, 'setProperties', { alex: 'wallace', steely: 'jan' });

        this.expectedPushURL = '/?alex=wallace&steely=fran';
        (0, _runloop.run)(appController, 'setProperties', { steely: 'fran' });

        this.expectedReplaceURL = '/?alex=sriracha&steely=fran';
        (0, _runloop.run)(appController, 'setProperties', { alex: 'sriracha' });
      });
    }

    ['@test can opt into full transition by setting refreshModel in route queryParams when transitioning from child to parent'](assert) {
      this.addTemplate('parent', '{{outlet}}');
      this.addTemplate('parent.child', "{{link-to 'Parent' 'parent' (query-params foo='change') id='parent-link'}}");

      this.router.map(function () {
        this.route('parent', function () {
          this.route('child');
        });
      });

      let parentModelCount = 0;
      this.add('route:parent', _routing.Route.extend({
        model() {
          parentModelCount++;
        },
        queryParams: {
          foo: {
            refreshModel: true
          }
        }
      }));

      this.setSingleQPController('parent', 'foo', 'abc');

      return this.visit('/parent/child?foo=lol').then(() => {
        assert.equal(parentModelCount, 1);

        (0, _runloop.run)(document.getElementById('parent-link'), 'click');
        assert.equal(parentModelCount, 2);
      });
    }

    ["@test Use Ember.get to retrieve query params 'replace' configuration"](assert) {
      assert.expect(2);

      this.setSingleQPController('application', 'alex', 'matchneer');

      this.add('route:application', _routing.Route.extend({
        queryParams: _runtime.Object.create({
          unknownProperty() /* keyName */{
            // We are simulating all qps requiring refresh
            return { replace: true };
          }
        })
      }));

      return this.visitAndAssert('/').then(() => {
        let appController = this.getController('application');
        this.expectedReplaceURL = '/?alex=wallace';
        this.setAndFlush(appController, 'alex', 'wallace');
      });
    }

    ['@test can override incoming QP values in setupController'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('about');
      });

      this.setSingleQPController('index', 'omg', 'lol');

      this.add('route:index', _routing.Route.extend({
        setupController(controller) {
          assert.ok(true, 'setupController called');
          controller.set('omg', 'OVERRIDE');
        },
        actions: {
          queryParamsDidChange() {
            assert.ok(false, "queryParamsDidChange shouldn't fire");
          }
        }
      }));

      return this.visitAndAssert('/about').then(() => {
        this.transitionTo('index');
        this.assertCurrentPath('/?omg=OVERRIDE');
      });
    }

    ['@test can override incoming QP array values in setupController'](assert) {
      assert.expect(3);

      this.router.map(function () {
        this.route('about');
      });

      this.setSingleQPController('index', 'omg', ['lol']);

      this.add('route:index', _routing.Route.extend({
        setupController(controller) {
          assert.ok(true, 'setupController called');
          controller.set('omg', ['OVERRIDE']);
        },
        actions: {
          queryParamsDidChange() {
            assert.ok(false, "queryParamsDidChange shouldn't fire");
          }
        }
      }));

      return this.visitAndAssert('/about').then(() => {
        this.transitionTo('index');
        this.assertCurrentPath('/?omg=' + encodeURIComponent(JSON.stringify(['OVERRIDE'])));
      });
    }

    ['@test URL transitions that remove QPs still register as QP changes'](assert) {
      assert.expect(2);

      this.setSingleQPController('index', 'omg', 'lol');

      return this.visit('/?omg=borf').then(() => {
        let indexController = this.getController('index');
        assert.equal(indexController.get('omg'), 'borf');

        this.transitionTo('/');
        assert.equal(indexController.get('omg'), 'lol');
      });
    }

    ['@test Subresource naming style is supported'](assert) {
      assert.expect(5);

      this.router.map(function () {
        this.route('abc.def', { path: '/abcdef' }, function () {
          this.route('zoo');
        });
      });

      this.addTemplate('application', "{{link-to 'A' 'abc.def' (query-params foo='123') id='one'}}{{link-to 'B' 'abc.def.zoo' (query-params foo='123' bar='456') id='two'}}{{outlet}}");

      this.setSingleQPController('abc.def', 'foo', 'lol');
      this.setSingleQPController('abc.def.zoo', 'bar', 'haha');

      return this.visitAndAssert('/').then(() => {
        assert.equal(this.$('#one').attr('href'), '/abcdef?foo=123');
        assert.equal(this.$('#two').attr('href'), '/abcdef/zoo?bar=456&foo=123');

        (0, _runloop.run)(this.$('#one'), 'click');
        this.assertCurrentPath('/abcdef?foo=123');

        (0, _runloop.run)(this.$('#two'), 'click');
        this.assertCurrentPath('/abcdef/zoo?bar=456&foo=123');
      });
    }

    ['@test transitionTo supports query params']() {
      this.setSingleQPController('index', 'foo', 'lol');

      return this.visitAndAssert('/').then(() => {
        this.transitionTo({ queryParams: { foo: 'borf' } });
        this.assertCurrentPath('/?foo=borf', 'shorthand supported');

        this.transitionTo({ queryParams: { 'index:foo': 'blaf' } });
        this.assertCurrentPath('/?foo=blaf', 'longform supported');

        this.transitionTo({ queryParams: { 'index:foo': false } });
        this.assertCurrentPath('/?foo=false', 'longform supported (bool)');

        this.transitionTo({ queryParams: { foo: false } });
        this.assertCurrentPath('/?foo=false', 'shorhand supported (bool)');
      });
    }

    ['@test transitionTo supports query params (multiple)']() {
      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo', 'bar'],
        foo: 'lol',
        bar: 'wat'
      }));

      return this.visitAndAssert('/').then(() => {
        this.transitionTo({ queryParams: { foo: 'borf' } });
        this.assertCurrentPath('/?foo=borf', 'shorthand supported');

        this.transitionTo({ queryParams: { 'index:foo': 'blaf' } });
        this.assertCurrentPath('/?foo=blaf', 'longform supported');

        this.transitionTo({ queryParams: { 'index:foo': false } });
        this.assertCurrentPath('/?foo=false', 'longform supported (bool)');

        this.transitionTo({ queryParams: { foo: false } });
        this.assertCurrentPath('/?foo=false', 'shorhand supported (bool)');
      });
    }

    ["@test setting controller QP to empty string doesn't generate null in URL"](assert) {
      assert.expect(1);

      this.setSingleQPController('index', 'foo', '123');

      return this.visit('/').then(() => {
        let controller = this.getController('index');

        this.expectedPushURL = '/?foo=';
        this.setAndFlush(controller, 'foo', '');
      });
    }

    ["@test setting QP to empty string doesn't generate null in URL"](assert) {
      assert.expect(1);

      this.add('route:index', _routing.Route.extend({
        queryParams: {
          foo: {
            defaultValue: '123'
          }
        }
      }));

      return this.visit('/').then(() => {
        let controller = this.getController('index');

        this.expectedPushURL = '/?foo=';
        this.setAndFlush(controller, 'foo', '');
      });
    }

    ['@test A default boolean value deserializes QPs as booleans rather than strings'](assert) {
      assert.expect(3);

      this.setSingleQPController('index', 'foo', false);

      this.add('route:index', _routing.Route.extend({
        model(params) {
          assert.equal(params.foo, true, 'model hook received foo as boolean true');
        }
      }));

      return this.visit('/?foo=true').then(() => {
        let controller = this.getController('index');
        assert.equal(controller.get('foo'), true);

        this.transitionTo('/?foo=false');
        assert.equal(controller.get('foo'), false);
      });
    }

    ['@test Query param without value are empty string'](assert) {
      assert.expect(1);

      this.add('controller:index', _controller.default.extend({
        queryParams: ['foo'],
        foo: ''
      }));

      return this.visit('/?foo=').then(() => {
        let controller = this.getController('index');
        assert.equal(controller.get('foo'), '');
      });
    }

    ['@test Array query params can be set'](assert) {
      assert.expect(2);

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.setSingleQPController('home', 'foo', []);

      return this.visit('/').then(() => {
        let controller = this.getController('home');

        this.setAndFlush(controller, 'foo', [1, 2]);
        this.assertCurrentPath('/?foo=%5B1%2C2%5D');

        this.setAndFlush(controller, 'foo', [3, 4]);
        this.assertCurrentPath('/?foo=%5B3%2C4%5D');
      });
    }

    ['@test (de)serialization: arrays'](assert) {
      assert.expect(4);

      this.setSingleQPController('index', 'foo', [1]);

      return this.visitAndAssert('/').then(() => {
        this.transitionTo({ queryParams: { foo: [2, 3] } });
        this.assertCurrentPath('/?foo=%5B2%2C3%5D', 'shorthand supported');
        this.transitionTo({ queryParams: { 'index:foo': [4, 5] } });
        this.assertCurrentPath('/?foo=%5B4%2C5%5D', 'longform supported');
        this.transitionTo({ queryParams: { foo: [] } });
        this.assertCurrentPath('/?foo=%5B%5D', 'longform supported');
      });
    }

    ['@test Url with array query param sets controller property to array'](assert) {
      assert.expect(1);

      this.setSingleQPController('index', 'foo', '');

      return this.visit('/?foo[]=1&foo[]=2&foo[]=3').then(() => {
        let controller = this.getController('index');
        assert.deepEqual(controller.get('foo'), ['1', '2', '3']);
      });
    }

    ['@test Array query params can be pushed/popped'](assert) {
      assert.expect(17);

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      this.setSingleQPController('home', 'foo', (0, _runtime.A)());

      return this.visitAndAssert('/').then(() => {
        let controller = this.getController('home');

        (0, _runloop.run)(controller.foo, 'pushObject', 1);
        this.assertCurrentPath('/?foo=%5B1%5D');
        assert.deepEqual(controller.foo, [1]);

        (0, _runloop.run)(controller.foo, 'popObject');
        this.assertCurrentPath('/');
        assert.deepEqual(controller.foo, []);

        (0, _runloop.run)(controller.foo, 'pushObject', 1);
        this.assertCurrentPath('/?foo=%5B1%5D');
        assert.deepEqual(controller.foo, [1]);

        (0, _runloop.run)(controller.foo, 'popObject');
        this.assertCurrentPath('/');
        assert.deepEqual(controller.foo, []);

        (0, _runloop.run)(controller.foo, 'pushObject', 1);
        this.assertCurrentPath('/?foo=%5B1%5D');
        assert.deepEqual(controller.foo, [1]);

        (0, _runloop.run)(controller.foo, 'pushObject', 2);
        this.assertCurrentPath('/?foo=%5B1%2C2%5D');
        assert.deepEqual(controller.foo, [1, 2]);

        (0, _runloop.run)(controller.foo, 'popObject');
        this.assertCurrentPath('/?foo=%5B1%5D');
        assert.deepEqual(controller.foo, [1]);

        (0, _runloop.run)(controller.foo, 'unshiftObject', 'lol');
        this.assertCurrentPath('/?foo=%5B%22lol%22%2C1%5D');
        assert.deepEqual(controller.foo, ['lol', 1]);
      });
    }

    ["@test Overwriting with array with same content shouldn't refire update"](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('home', { path: '/' });
      });

      let modelCount = 0;
      this.add('route:home', _routing.Route.extend({
        model() {
          modelCount++;
        }
      }));

      this.setSingleQPController('home', 'foo', (0, _runtime.A)([1]));

      return this.visitAndAssert('/').then(() => {
        assert.equal(modelCount, 1);

        let controller = this.getController('home');
        this.setAndFlush(controller, 'model', (0, _runtime.A)([1]));

        assert.equal(modelCount, 1);
        this.assertCurrentPath('/');
      });
    }

    ['@test Defaulting to params hash as the model should not result in that params object being watched'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('other');
      });

      // This causes the params hash, which is returned as a route's
      // model if no other model could be resolved given the provided
      // params (and no custom model hook was defined), to be watched,
      // unless we return a copy of the params hash.
      this.setSingleQPController('application', 'woot', 'wat');

      this.add('route:other', _routing.Route.extend({
        model(p, trans) {
          let m = (0, _meta.peekMeta)(trans[_router_js.PARAMS_SYMBOL].application);
          assert.ok(m === undefined, "A meta object isn't constructed for this params POJO");
        }
      }));

      return this.visit('/').then(() => {
        this.transitionTo('other');
      });
    }

    ['@test Setting bound query param property to null or undefined does not serialize to url'](assert) {
      assert.expect(9);

      this.router.map(function () {
        this.route('home');
      });

      this.setSingleQPController('home', 'foo', [1, 2]);

      return this.visitAndAssert('/home').then(() => {
        var controller = this.getController('home');

        assert.deepEqual(controller.get('foo'), [1, 2]);
        this.assertCurrentPath('/home');

        this.setAndFlush(controller, 'foo', (0, _runtime.A)([1, 3]));
        this.assertCurrentPath('/home?foo=%5B1%2C3%5D');

        return this.transitionTo('/home').then(() => {
          assert.deepEqual(controller.get('foo'), [1, 2]);
          this.assertCurrentPath('/home');

          this.setAndFlush(controller, 'foo', null);
          this.assertCurrentPath('/home', 'Setting property to null');

          this.setAndFlush(controller, 'foo', (0, _runtime.A)([1, 3]));
          this.assertCurrentPath('/home?foo=%5B1%2C3%5D');

          this.setAndFlush(controller, 'foo', undefined);
          this.assertCurrentPath('/home', 'Setting property to undefined');
        });
      });
    }

    ['@test {{link-to}} with null or undefined QPs does not get serialized into url'](assert) {
      assert.expect(3);

      this.addTemplate('home', "{{link-to 'Home' 'home' (query-params foo=nullValue) id='null-link'}}{{link-to 'Home' 'home' (query-params foo=undefinedValue) id='undefined-link'}}");

      this.router.map(function () {
        this.route('home');
      });

      this.setSingleQPController('home', 'foo', [], {
        nullValue: null,
        undefinedValue: undefined
      });

      return this.visitAndAssert('/home').then(() => {
        assert.equal(this.$('#null-link').attr('href'), '/home');
        assert.equal(this.$('#undefined-link').attr('href'), '/home');
      });
    }

    ["@test A child of a resource route still defaults to parent route's model even if the child route has a query param"](assert) {
      assert.expect(2);

      this.setSingleQPController('index', 'woot', undefined, {
        woot: undefined
      });

      this.add('route:application', _routing.Route.extend({
        model() /* p, trans */{
          return { woot: true };
        }
      }));

      this.add('route:index', _routing.Route.extend({
        setupController(controller, model) {
          assert.deepEqual(model, { woot: true }, 'index route inherited model route from parent route');
        }
      }));

      return this.visitAndAssert('/');
    }

    ['@test opting into replace does not affect transitions between routes'](assert) {
      assert.expect(5);

      this.addTemplate('application', "{{link-to 'Foo' 'foo' id='foo-link'}}{{link-to 'Bar' 'bar' id='bar-no-qp-link'}}{{link-to 'Bar' 'bar' (query-params raytiley='isthebest') id='bar-link'}}{{outlet}}");

      this.router.map(function () {
        this.route('foo');
        this.route('bar');
      });

      this.setSingleQPController('bar', 'raytiley', 'israd');

      this.add('route:bar', _routing.Route.extend({
        queryParams: {
          raytiley: {
            replace: true
          }
        }
      }));

      return this.visit('/').then(() => {
        let controller = this.getController('bar');

        this.expectedPushURL = '/foo';
        (0, _runloop.run)(document.getElementById('foo-link'), 'click');

        this.expectedPushURL = '/bar';
        (0, _runloop.run)(document.getElementById('bar-no-qp-link'), 'click');

        this.expectedReplaceURL = '/bar?raytiley=woot';
        this.setAndFlush(controller, 'raytiley', 'woot');

        this.expectedPushURL = '/foo';
        (0, _runloop.run)(document.getElementById('foo-link'), 'click');

        this.expectedPushURL = '/bar?raytiley=isthebest';
        (0, _runloop.run)(document.getElementById('bar-link'), 'click');
      });
    }

    ["@test undefined isn't serialized or deserialized into a string"](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('example');
      });

      this.addTemplate('application', "{{link-to 'Example' 'example' (query-params foo=undefined) id='the-link'}}");

      this.setSingleQPController('example', 'foo', undefined, {
        foo: undefined
      });

      this.add('route:example', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { foo: undefined });
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/example', 'renders without undefined qp serialized');

        return this.transitionTo('example', {
          queryParams: { foo: undefined }
        }).then(() => {
          this.assertCurrentPath('/example');
        });
      });
    }

    ['@test when refreshModel is true and loading hook is undefined, model hook will rerun when QPs change even if previous did not finish']() {
      return this.refreshModelWhileLoadingTest();
    }

    ['@test when refreshModel is true and loading hook returns false, model hook will rerun when QPs change even if previous did not finish']() {
      return this.refreshModelWhileLoadingTest(false);
    }

    ['@test when refreshModel is true and loading hook returns true, model hook will rerun when QPs change even if previous did not finish']() {
      return this.refreshModelWhileLoadingTest(true);
    }

    ["@test warn user that Route's queryParams configuration must be an Object, not an Array"](assert) {
      assert.expect(1);

      this.add('route:application', _routing.Route.extend({
        queryParams: [{ commitBy: { replace: true } }]
      }));

      expectAssertion(() => {
        this.visit('/');
      }, 'You passed in `[{"commitBy":{"replace":true}}]` as the value for `queryParams` but `queryParams` cannot be an Array');
    }

    ['@test handle route names that clash with Object.prototype properties'](assert) {
      assert.expect(1);

      this.router.map(function () {
        this.route('constructor');
      });

      this.add('route:constructor', _routing.Route.extend({
        queryParams: {
          foo: {
            defaultValue: '123'
          }
        }
      }));

      return this.visit('/').then(() => {
        this.transitionTo('constructor', { queryParams: { foo: '999' } });
        let controller = this.getController('constructor');
        assert.equal((0, _metal.get)(controller, 'foo'), '999');
      });
    }
  });
});
enifed('ember/tests/routing/query_params_test/model_dependent_state_with_query_params_test', ['@ember/controller', '@ember/-internals/runtime', '@ember/-internals/routing', '@ember/runloop', '@ember/-internals/metal', 'internal-test-helpers'], function (_controller, _runtime, _routing, _runloop, _metal, _internalTestHelpers) {
  'use strict';

  class ModelDependentQPTestCase extends _internalTestHelpers.QueryParamTestCase {
    boot() {
      this.setupApplication();
      return this.visitApplication();
    }

    teardown() {
      super.teardown(...arguments);
      this.assert.ok(!this.expectedModelHookParams, 'there should be no pending expectation of expected model hook params');
    }

    reopenController(name, options) {
      this.application.resolveRegistration(`controller:${name}`).reopen(options);
    }

    reopenRoute(name, options) {
      this.application.resolveRegistration(`route:${name}`).reopen(options);
    }

    queryParamsStickyTest1(urlPrefix) {
      let assert = this.assert;

      assert.expect(14);

      return this.boot().then(() => {
        (0, _runloop.run)(this.$link1, 'click');
        this.assertCurrentPath(`${urlPrefix}/a-1`);

        this.setAndFlush(this.controller, 'q', 'lol');

        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);

        (0, _runloop.run)(this.$link2, 'click');

        assert.equal(this.controller.get('q'), 'wat');
        assert.equal(this.controller.get('z'), 0);
        assert.deepEqual(this.controller.get('model'), { id: 'a-2' });
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);
      });
    }

    queryParamsStickyTest2(urlPrefix) {
      let assert = this.assert;

      assert.expect(24);

      return this.boot().then(() => {
        this.expectedModelHookParams = { id: 'a-1', q: 'lol', z: 0 };
        this.transitionTo(`${urlPrefix}/a-1?q=lol`);

        assert.deepEqual(this.controller.get('model'), { id: 'a-1' });
        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 0);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);

        this.expectedModelHookParams = { id: 'a-2', q: 'lol', z: 0 };
        this.transitionTo(`${urlPrefix}/a-2?q=lol`);

        assert.deepEqual(this.controller.get('model'), { id: 'a-2' }, "controller's model changed to a-2");
        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 0);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);

        this.expectedModelHookParams = { id: 'a-3', q: 'lol', z: 123 };
        this.transitionTo(`${urlPrefix}/a-3?q=lol&z=123`);

        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 123);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=lol&z=123`);
      });
    }

    queryParamsStickyTest3(urlPrefix, articleLookup) {
      let assert = this.assert;

      assert.expect(32);

      this.addTemplate('application', `{{#each articles as |a|}} {{link-to 'Article' '${articleLookup}' a.id id=a.id}} {{/each}}`);

      return this.boot().then(() => {
        this.expectedModelHookParams = { id: 'a-1', q: 'wat', z: 0 };
        this.transitionTo(articleLookup, 'a-1');

        assert.deepEqual(this.controller.get('model'), { id: 'a-1' });
        assert.equal(this.controller.get('q'), 'wat');
        assert.equal(this.controller.get('z'), 0);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);

        this.expectedModelHookParams = { id: 'a-2', q: 'lol', z: 0 };
        this.transitionTo(articleLookup, 'a-2', { queryParams: { q: 'lol' } });

        assert.deepEqual(this.controller.get('model'), { id: 'a-2' });
        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 0);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3`);

        this.expectedModelHookParams = { id: 'a-3', q: 'hay', z: 0 };
        this.transitionTo(articleLookup, 'a-3', { queryParams: { q: 'hay' } });

        assert.deepEqual(this.controller.get('model'), { id: 'a-3' });
        assert.equal(this.controller.get('q'), 'hay');
        assert.equal(this.controller.get('z'), 0);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=hay`);

        this.expectedModelHookParams = { id: 'a-2', q: 'lol', z: 1 };
        this.transitionTo(articleLookup, 'a-2', { queryParams: { z: 1 } });

        assert.deepEqual(this.controller.get('model'), { id: 'a-2' });
        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 1);
        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol&z=1`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=hay`);
      });
    }

    queryParamsStickyTest4(urlPrefix, articleLookup) {
      let assert = this.assert;

      assert.expect(24);

      this.setupApplication();

      this.reopenController(articleLookup, {
        queryParams: { q: { scope: 'controller' } }
      });

      return this.visitApplication().then(() => {
        (0, _runloop.run)(this.$link1, 'click');
        this.assertCurrentPath(`${urlPrefix}/a-1`);

        this.setAndFlush(this.controller, 'q', 'lol');

        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=lol`);

        (0, _runloop.run)(this.$link2, 'click');

        assert.equal(this.controller.get('q'), 'lol');
        assert.equal(this.controller.get('z'), 0);
        assert.deepEqual(this.controller.get('model'), { id: 'a-2' });

        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=lol`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=lol`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=lol`);

        this.expectedModelHookParams = { id: 'a-3', q: 'haha', z: 123 };
        this.transitionTo(`${urlPrefix}/a-3?q=haha&z=123`);

        assert.deepEqual(this.controller.get('model'), { id: 'a-3' });
        assert.equal(this.controller.get('q'), 'haha');
        assert.equal(this.controller.get('z'), 123);

        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=haha`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=haha`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=haha&z=123`);

        this.setAndFlush(this.controller, 'q', 'woot');

        assert.equal(this.$link1.getAttribute('href'), `${urlPrefix}/a-1?q=woot`);
        assert.equal(this.$link2.getAttribute('href'), `${urlPrefix}/a-2?q=woot`);
        assert.equal(this.$link3.getAttribute('href'), `${urlPrefix}/a-3?q=woot&z=123`);
      });
    }

    queryParamsStickyTest5(urlPrefix, commentsLookupKey) {
      let assert = this.assert;

      assert.expect(12);

      return this.boot().then(() => {
        this.transitionTo(commentsLookupKey, 'a-1');

        let commentsCtrl = this.getController(commentsLookupKey);
        assert.equal(commentsCtrl.get('page'), 1);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments`);

        this.setAndFlush(commentsCtrl, 'page', 2);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments?page=2`);

        this.setAndFlush(commentsCtrl, 'page', 3);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments?page=3`);

        this.transitionTo(commentsLookupKey, 'a-2');
        assert.equal(commentsCtrl.get('page'), 1);
        this.assertCurrentPath(`${urlPrefix}/a-2/comments`);

        this.transitionTo(commentsLookupKey, 'a-1');
        assert.equal(commentsCtrl.get('page'), 3);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments?page=3`);
      });
    }

    queryParamsStickyTest6(urlPrefix, articleLookup, commentsLookup) {
      let assert = this.assert;

      assert.expect(13);

      this.setupApplication();

      this.reopenRoute(articleLookup, {
        resetController(controller, isExiting) {
          this.controllerFor(commentsLookup).set('page', 1);
          if (isExiting) {
            controller.set('q', 'imdone');
          }
        }
      });

      this.addTemplate('about', `{{link-to 'A' '${commentsLookup}' 'a-1' id='one'}} {{link-to 'B' '${commentsLookup}' 'a-2' id='two'}}`);

      return this.visitApplication().then(() => {
        this.transitionTo(commentsLookup, 'a-1');

        let commentsCtrl = this.getController(commentsLookup);
        assert.equal(commentsCtrl.get('page'), 1);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments`);

        this.setAndFlush(commentsCtrl, 'page', 2);
        this.assertCurrentPath(`${urlPrefix}/a-1/comments?page=2`);

        this.transitionTo(commentsLookup, 'a-2');
        assert.equal(commentsCtrl.get('page'), 1);
        assert.equal(this.controller.get('q'), 'wat');

        this.transitionTo(commentsLookup, 'a-1');

        this.assertCurrentPath(`${urlPrefix}/a-1/comments`);
        assert.equal(commentsCtrl.get('page'), 1);

        this.transitionTo('about');
        assert.equal(document.getElementById('one').getAttribute('href'), `${urlPrefix}/a-1/comments?q=imdone`);
        assert.equal(document.getElementById('two').getAttribute('href'), `${urlPrefix}/a-2/comments`);
      });
    }
  }

  (0, _internalTestHelpers.moduleFor)('Query Params - model-dependent state', class extends ModelDependentQPTestCase {
    setupApplication() {
      this.router.map(function () {
        this.route('article', { path: '/a/:id' }, function () {
          this.route('comments', { resetNamespace: true });
        });
        this.route('about');
      });

      let articles = (0, _runtime.A)([{ id: 'a-1' }, { id: 'a-2' }, { id: 'a-3' }]);

      this.add('controller:application', _controller.default.extend({
        articles
      }));

      let self = this;
      let assert = this.assert;
      this.add('route:article', _routing.Route.extend({
        model(params) {
          if (self.expectedModelHookParams) {
            assert.deepEqual(params, self.expectedModelHookParams, 'the ArticleRoute model hook received the expected merged dynamic segment + query params hash');
            self.expectedModelHookParams = null;
          }
          return articles.findBy('id', params.id);
        }
      }));

      this.add('controller:article', _controller.default.extend({
        queryParams: ['q', 'z'],
        q: 'wat',
        z: 0
      }));

      this.add('controller:comments', _controller.default.extend({
        queryParams: 'page',
        page: 1
      }));

      this.addTemplate('application', "{{#each articles as |a|}} 1{{link-to 'Article' 'article' a id=a.id}} {{/each}} {{outlet}}");
    }

    visitApplication() {
      return this.visit('/').then(() => {
        let assert = this.assert;

        this.$link1 = document.getElementById('a-1');
        this.$link2 = document.getElementById('a-2');
        this.$link3 = document.getElementById('a-3');

        assert.equal(this.$link1.getAttribute('href'), '/a/a-1');
        assert.equal(this.$link2.getAttribute('href'), '/a/a-2');
        assert.equal(this.$link3.getAttribute('href'), '/a/a-3');

        this.controller = this.getController('article');
      });
    }

    ["@test query params have 'model' stickiness by default"]() {
      return this.queryParamsStickyTest1('/a');
    }

    ["@test query params have 'model' stickiness by default (url changes)"]() {
      return this.queryParamsStickyTest2('/a');
    }

    ["@test query params have 'model' stickiness by default (params-based transitions)"]() {
      return this.queryParamsStickyTest3('/a', 'article');
    }

    ["@test 'controller' stickiness shares QP state between models"]() {
      return this.queryParamsStickyTest4('/a', 'article');
    }

    ["@test 'model' stickiness is scoped to current or first dynamic parent route"]() {
      return this.queryParamsStickyTest5('/a', 'comments');
    }

    ['@test can reset query params using the resetController hook']() {
      return this.queryParamsStickyTest6('/a', 'article', 'comments');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Query Params - model-dependent state (nested)', class extends ModelDependentQPTestCase {
    setupApplication() {
      this.router.map(function () {
        this.route('site', function () {
          this.route('article', { path: '/a/:id' }, function () {
            this.route('comments');
          });
        });
        this.route('about');
      });

      let site_articles = (0, _runtime.A)([{ id: 'a-1' }, { id: 'a-2' }, { id: 'a-3' }]);

      this.add('controller:application', _controller.default.extend({
        articles: site_articles
      }));

      let self = this;
      let assert = this.assert;
      this.add('route:site.article', _routing.Route.extend({
        model(params) {
          if (self.expectedModelHookParams) {
            assert.deepEqual(params, self.expectedModelHookParams, 'the ArticleRoute model hook received the expected merged dynamic segment + query params hash');
            self.expectedModelHookParams = null;
          }
          return site_articles.findBy('id', params.id);
        }
      }));

      this.add('controller:site.article', _controller.default.extend({
        queryParams: ['q', 'z'],
        q: 'wat',
        z: 0
      }));

      this.add('controller:site.article.comments', _controller.default.extend({
        queryParams: 'page',
        page: 1
      }));

      this.addTemplate('application', "{{#each articles as |a|}} {{link-to 'Article' 'site.article' a id=a.id}} {{/each}} {{outlet}}");
    }

    visitApplication() {
      return this.visit('/').then(() => {
        let assert = this.assert;

        this.$link1 = document.getElementById('a-1');
        this.$link2 = document.getElementById('a-2');
        this.$link3 = document.getElementById('a-3');

        assert.equal(this.$link1.getAttribute('href'), '/site/a/a-1');
        assert.equal(this.$link2.getAttribute('href'), '/site/a/a-2');
        assert.equal(this.$link3.getAttribute('href'), '/site/a/a-3');

        this.controller = this.getController('site.article');
      });
    }

    ["@test query params have 'model' stickiness by default"]() {
      return this.queryParamsStickyTest1('/site/a');
    }

    ["@test query params have 'model' stickiness by default (url changes)"]() {
      return this.queryParamsStickyTest2('/site/a');
    }

    ["@test query params have 'model' stickiness by default (params-based transitions)"]() {
      return this.queryParamsStickyTest3('/site/a', 'site.article');
    }

    ["@test 'controller' stickiness shares QP state between models"]() {
      return this.queryParamsStickyTest4('/site/a', 'site.article');
    }

    ["@test 'model' stickiness is scoped to current or first dynamic parent route"]() {
      return this.queryParamsStickyTest5('/site/a', 'site.article.comments');
    }

    ['@test can reset query params using the resetController hook']() {
      return this.queryParamsStickyTest6('/site/a', 'site.article', 'site.article.comments');
    }
  });

  (0, _internalTestHelpers.moduleFor)('Query Params - model-dependent state (nested & more than 1 dynamic segment)', class extends ModelDependentQPTestCase {
    setupApplication() {
      this.router.map(function () {
        this.route('site', { path: '/site/:site_id' }, function () {
          this.route('article', { path: '/a/:article_id' }, function () {
            this.route('comments');
          });
        });
      });

      let sites = (0, _runtime.A)([{ id: 's-1' }, { id: 's-2' }, { id: 's-3' }]);
      let site_articles = (0, _runtime.A)([{ id: 'a-1' }, { id: 'a-2' }, { id: 'a-3' }]);

      this.add('controller:application', _controller.default.extend({
        siteArticles: site_articles,
        sites,
        allSitesAllArticles: (0, _metal.computed)({
          get() {
            let ret = [];
            let siteArticles = this.siteArticles;
            let sites = this.sites;
            sites.forEach(site => {
              ret = ret.concat(siteArticles.map(article => {
                return {
                  id: `${site.id}-${article.id}`,
                  site_id: site.id,
                  article_id: article.id
                };
              }));
            });
            return ret;
          }
        })
      }));

      let self = this;
      let assert = this.assert;
      this.add('route:site', _routing.Route.extend({
        model(params) {
          if (self.expectedSiteModelHookParams) {
            assert.deepEqual(params, self.expectedSiteModelHookParams, 'the SiteRoute model hook received the expected merged dynamic segment + query params hash');
            self.expectedSiteModelHookParams = null;
          }
          return sites.findBy('id', params.site_id);
        }
      }));

      this.add('route:site.article', _routing.Route.extend({
        model(params) {
          if (self.expectedArticleModelHookParams) {
            assert.deepEqual(params, self.expectedArticleModelHookParams, 'the SiteArticleRoute model hook received the expected merged dynamic segment + query params hash');
            self.expectedArticleModelHookParams = null;
          }
          return site_articles.findBy('id', params.article_id);
        }
      }));

      this.add('controller:site', _controller.default.extend({
        queryParams: ['country'],
        country: 'au'
      }));

      this.add('controller:site.article', _controller.default.extend({
        queryParams: ['q', 'z'],
        q: 'wat',
        z: 0
      }));

      this.add('controller:site.article.comments', _controller.default.extend({
        queryParams: ['page'],
        page: 1
      }));

      this.addTemplate('application', "{{#each allSitesAllArticles as |a|}} {{#link-to 'site.article' a.site_id a.article_id id=a.id}}Article [{{a.site_id}}] [{{a.article_id}}]{{/link-to}} {{/each}} {{outlet}}");
    }

    visitApplication() {
      return this.visit('/').then(() => {
        let assert = this.assert;

        this.links = {};
        this.links['s-1-a-1'] = document.getElementById('s-1-a-1');
        this.links['s-1-a-2'] = document.getElementById('s-1-a-2');
        this.links['s-1-a-3'] = document.getElementById('s-1-a-3');
        this.links['s-2-a-1'] = document.getElementById('s-2-a-1');
        this.links['s-2-a-2'] = document.getElementById('s-2-a-2');
        this.links['s-2-a-3'] = document.getElementById('s-2-a-3');
        this.links['s-3-a-1'] = document.getElementById('s-3-a-1');
        this.links['s-3-a-2'] = document.getElementById('s-3-a-2');
        this.links['s-3-a-3'] = document.getElementById('s-3-a-3');

        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.site_controller = this.getController('site');
        this.article_controller = this.getController('site.article');
      });
    }

    ["@test query params have 'model' stickiness by default"](assert) {
      assert.expect(59);

      return this.boot().then(() => {
        (0, _runloop.run)(this.links['s-1-a-1'], 'click');
        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-1' });
        this.assertCurrentPath('/site/s-1/a/a-1');

        this.setAndFlush(this.article_controller, 'q', 'lol');

        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.setAndFlush(this.site_controller, 'country', 'us');

        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?country=us');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?country=us');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        (0, _runloop.run)(this.links['s-1-a-2'], 'click');

        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'wat');
        assert.equal(this.article_controller.get('z'), 0);
        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' });
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?country=us');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?country=us');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        (0, _runloop.run)(this.links['s-2-a-2'], 'click');

        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'wat');
        assert.equal(this.article_controller.get('z'), 0);
        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' });
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?country=us');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?country=us');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');
      });
    }

    ["@test query params have 'model' stickiness by default (url changes)"](assert) {
      assert.expect(88);

      return this.boot().then(() => {
        this.expectedSiteModelHookParams = { site_id: 's-1', country: 'au' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-1',
          q: 'lol',
          z: 0
        };
        this.transitionTo('/site/s-1/a/a-1?q=lol');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' }, "site controller's model is s-1");
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-1' }, "article controller's model is a-1");
        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.expectedSiteModelHookParams = { site_id: 's-2', country: 'us' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-1',
          q: 'lol',
          z: 0
        };
        this.transitionTo('/site/s-2/a/a-1?country=us&q=lol');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' }, "site controller's model is s-2");
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-1' }, "article controller's model is a-1");
        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.expectedSiteModelHookParams = { site_id: 's-2', country: 'us' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-2',
          q: 'lol',
          z: 0
        };
        this.transitionTo('/site/s-2/a/a-2?country=us&q=lol');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' }, "site controller's model is s-2");
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' }, "article controller's model is a-2");
        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.expectedSiteModelHookParams = { site_id: 's-2', country: 'us' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-3',
          q: 'lol',
          z: 123
        };
        this.transitionTo('/site/s-2/a/a-3?country=us&q=lol&z=123');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' }, "site controller's model is s-2");
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-3' }, "article controller's model is a-3");
        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 123);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=lol&z=123');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us&q=lol&z=123');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?q=lol&z=123');

        this.expectedSiteModelHookParams = { site_id: 's-3', country: 'nz' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-3',
          q: 'lol',
          z: 123
        };
        this.transitionTo('/site/s-3/a/a-3?country=nz&q=lol&z=123');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-3' }, "site controller's model is s-3");
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-3' }, "article controller's model is a-3");
        assert.equal(this.site_controller.get('country'), 'nz');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 123);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=lol');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=lol&z=123');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=lol');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us&q=lol&z=123');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?country=nz&q=lol');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?country=nz&q=lol');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?country=nz&q=lol&z=123');
      });
    }

    ["@test query params have 'model' stickiness by default (params-based transitions)"](assert) {
      assert.expect(118);

      return this.boot().then(() => {
        this.expectedSiteModelHookParams = { site_id: 's-1', country: 'au' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-1',
          q: 'wat',
          z: 0
        };
        this.transitionTo('site.article', 's-1', 'a-1');

        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-1' });
        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'wat');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.expectedSiteModelHookParams = { site_id: 's-1', country: 'au' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-2',
          q: 'lol',
          z: 0
        };
        this.transitionTo('site.article', 's-1', 'a-2', {
          queryParams: { q: 'lol' }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' });
        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?q=lol');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3');

        this.expectedSiteModelHookParams = { site_id: 's-1', country: 'au' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-3',
          q: 'hay',
          z: 0
        };
        this.transitionTo('site.article', 's-1', 'a-3', {
          queryParams: { q: 'hay' }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-3' });
        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'hay');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=hay');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?q=lol');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?q=hay');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?q=hay');

        this.expectedSiteModelHookParams = { site_id: 's-1', country: 'au' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-2',
          q: 'lol',
          z: 1
        };
        this.transitionTo('site.article', 's-1', 'a-2', {
          queryParams: { z: 1 }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-1' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' });
        assert.equal(this.site_controller.get('country'), 'au');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 1);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=hay');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?q=hay');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?q=hay');

        this.expectedSiteModelHookParams = { site_id: 's-2', country: 'us' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-2',
          q: 'lol',
          z: 1
        };
        this.transitionTo('site.article', 's-2', 'a-2', {
          queryParams: { country: 'us' }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-2' });
        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'lol');
        assert.equal(this.article_controller.get('z'), 1);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=hay');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol&z=1');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us&q=hay');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?q=hay');

        this.expectedSiteModelHookParams = { site_id: 's-2', country: 'us' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-1',
          q: 'yeah',
          z: 0
        };
        this.transitionTo('site.article', 's-2', 'a-1', {
          queryParams: { q: 'yeah' }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-2' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-1' });
        assert.equal(this.site_controller.get('country'), 'us');
        assert.equal(this.article_controller.get('q'), 'yeah');
        assert.equal(this.article_controller.get('z'), 0);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=yeah');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=hay');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=yeah');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol&z=1');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us&q=hay');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?q=yeah');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?q=hay');

        this.expectedSiteModelHookParams = { site_id: 's-3', country: 'nz' };
        this.expectedArticleModelHookParams = {
          article_id: 'a-3',
          q: 'hay',
          z: 3
        };
        this.transitionTo('site.article', 's-3', 'a-3', {
          queryParams: { country: 'nz', z: 3 }
        });

        assert.deepEqual(this.site_controller.get('model'), { id: 's-3' });
        assert.deepEqual(this.article_controller.get('model'), { id: 'a-3' });
        assert.equal(this.site_controller.get('country'), 'nz');
        assert.equal(this.article_controller.get('q'), 'hay');
        assert.equal(this.article_controller.get('z'), 3);
        assert.equal(this.links['s-1-a-1'].getAttribute('href'), '/site/s-1/a/a-1?q=yeah');
        assert.equal(this.links['s-1-a-2'].getAttribute('href'), '/site/s-1/a/a-2?q=lol&z=1');
        assert.equal(this.links['s-1-a-3'].getAttribute('href'), '/site/s-1/a/a-3?q=hay&z=3');
        assert.equal(this.links['s-2-a-1'].getAttribute('href'), '/site/s-2/a/a-1?country=us&q=yeah');
        assert.equal(this.links['s-2-a-2'].getAttribute('href'), '/site/s-2/a/a-2?country=us&q=lol&z=1');
        assert.equal(this.links['s-2-a-3'].getAttribute('href'), '/site/s-2/a/a-3?country=us&q=hay&z=3');
        assert.equal(this.links['s-3-a-1'].getAttribute('href'), '/site/s-3/a/a-1?country=nz&q=yeah');
        assert.equal(this.links['s-3-a-2'].getAttribute('href'), '/site/s-3/a/a-2?country=nz&q=lol&z=1');
        assert.equal(this.links['s-3-a-3'].getAttribute('href'), '/site/s-3/a/a-3?country=nz&q=hay&z=3');
      });
    }
  });
});
enifed('ember/tests/routing/query_params_test/overlapping_query_params_test', ['@ember/controller', '@ember/-internals/routing', '@ember/runloop', '@ember/-internals/metal', 'internal-test-helpers'], function (_controller, _routing, _runloop, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Query Params - overlapping query param property names', class extends _internalTestHelpers.QueryParamTestCase {
    setupBase() {
      this.router.map(function () {
        this.route('parent', function () {
          this.route('child');
        });
      });

      return this.visit('/parent/child');
    }

    ['@test can remap same-named qp props'](assert) {
      assert.expect(7);

      this.setMappedQPController('parent');
      this.setMappedQPController('parent.child', 'page', 'childPage');

      return this.setupBase().then(() => {
        this.assertCurrentPath('/parent/child');

        let parentController = this.getController('parent');
        let parentChildController = this.getController('parent.child');

        this.setAndFlush(parentController, 'page', 2);
        this.assertCurrentPath('/parent/child?parentPage=2');
        this.setAndFlush(parentController, 'page', 1);
        this.assertCurrentPath('/parent/child');

        this.setAndFlush(parentChildController, 'page', 2);
        this.assertCurrentPath('/parent/child?childPage=2');
        this.setAndFlush(parentChildController, 'page', 1);
        this.assertCurrentPath('/parent/child');

        (0, _runloop.run)(() => {
          parentController.set('page', 2);
          parentChildController.set('page', 2);
        });

        this.assertCurrentPath('/parent/child?childPage=2&parentPage=2');

        (0, _runloop.run)(() => {
          parentController.set('page', 1);
          parentChildController.set('page', 1);
        });

        this.assertCurrentPath('/parent/child');
      });
    }

    ['@test query params can be either controller property or url key'](assert) {
      assert.expect(3);

      this.setMappedQPController('parent');

      return this.setupBase().then(() => {
        this.assertCurrentPath('/parent/child');

        this.transitionTo('parent.child', { queryParams: { page: 2 } });
        this.assertCurrentPath('/parent/child?parentPage=2');

        this.transitionTo('parent.child', { queryParams: { parentPage: 3 } });
        this.assertCurrentPath('/parent/child?parentPage=3');
      });
    }

    ['@test query param matching a url key and controller property'](assert) {
      assert.expect(3);

      this.setMappedQPController('parent', 'page', 'parentPage');
      this.setMappedQPController('parent.child', 'index', 'page');

      return this.setupBase().then(() => {
        this.transitionTo('parent.child', { queryParams: { page: 2 } });
        this.assertCurrentPath('/parent/child?parentPage=2');

        this.transitionTo('parent.child', { queryParams: { parentPage: 3 } });
        this.assertCurrentPath('/parent/child?parentPage=3');

        this.transitionTo('parent.child', {
          queryParams: { index: 2, page: 2 }
        });
        this.assertCurrentPath('/parent/child?page=2&parentPage=2');
      });
    }

    ['@test query param matching same property on two controllers use the urlKey higher in the chain'](assert) {
      assert.expect(4);

      this.setMappedQPController('parent', 'page', 'parentPage');
      this.setMappedQPController('parent.child', 'page', 'childPage');

      return this.setupBase().then(() => {
        this.transitionTo('parent.child', { queryParams: { page: 2 } });
        this.assertCurrentPath('/parent/child?parentPage=2');

        this.transitionTo('parent.child', { queryParams: { parentPage: 3 } });
        this.assertCurrentPath('/parent/child?parentPage=3');

        this.transitionTo('parent.child', {
          queryParams: { childPage: 2, page: 2 }
        });
        this.assertCurrentPath('/parent/child?childPage=2&parentPage=2');

        this.transitionTo('parent.child', {
          queryParams: { childPage: 3, parentPage: 4 }
        });
        this.assertCurrentPath('/parent/child?childPage=3&parentPage=4');
      });
    }

    ['@test query params does not error when a query parameter exists for route instances that share a controller'](assert) {
      assert.expect(1);

      let parentController = _controller.default.extend({
        queryParams: { page: 'page' }
      });
      this.add('controller:parent', parentController);
      this.add('route:parent.child', _routing.Route.extend({ controllerName: 'parent' }));

      return this.setupBase('/parent').then(() => {
        this.transitionTo('parent.child', { queryParams: { page: 2 } });
        this.assertCurrentPath('/parent/child?page=2');
      });
    }
    ['@test query params in the same route hierarchy with the same url key get auto-scoped'](assert) {
      assert.expect(1);

      this.setMappedQPController('parent');
      this.setMappedQPController('parent.child');

      expectAssertion(() => {
        this.setupBase();
      }, "You're not allowed to have more than one controller property map to the same query param key, but both `parent:page` and `parent.child:page` map to `parentPage`. You can fix this by mapping one of the controller properties to a different query param key via the `as` config option, e.g. `page: { as: 'other-page' }`");
    }

    ['@test Support shared but overridable mixin pattern'](assert) {
      assert.expect(7);

      let HasPage = _metal.Mixin.create({
        queryParams: 'page',
        page: 1
      });

      this.add('controller:parent', _controller.default.extend(HasPage, {
        queryParams: { page: 'yespage' }
      }));

      this.add('controller:parent.child', _controller.default.extend(HasPage));

      return this.setupBase().then(() => {
        this.assertCurrentPath('/parent/child');

        let parentController = this.getController('parent');
        let parentChildController = this.getController('parent.child');

        this.setAndFlush(parentChildController, 'page', 2);
        this.assertCurrentPath('/parent/child?page=2');
        assert.equal(parentController.get('page'), 1);
        assert.equal(parentChildController.get('page'), 2);

        this.setAndFlush(parentController, 'page', 2);
        this.assertCurrentPath('/parent/child?page=2&yespage=2');
        assert.equal(parentController.get('page'), 2);
        assert.equal(parentChildController.get('page'), 2);
      });
    }
  });
});
enifed('ember/tests/routing/query_params_test/query_param_async_get_handler_test', ['@ember/-internals/metal', '@ember/-internals/runtime', '@ember/-internals/routing', 'internal-test-helpers'], function (_metal, _runtime, _routing, _internalTestHelpers) {
  'use strict';

  // These tests mimic what happens with lazily loaded Engines.
  (0, _internalTestHelpers.moduleFor)('Query Params - async get handler', class extends _internalTestHelpers.QueryParamTestCase {
    get routerOptions() {
      let fetchedHandlers = this.fetchedHandlers = [];

      return {
        location: 'test',

        init() {
          this._super(...arguments);
          this._seenHandlers = Object.create(null);
          this._handlerPromises = Object.create(null);
        },

        setupRouter() {
          this._super(...arguments);
          let { _handlerPromises: handlerPromises, _seenHandlers: seenHandlers } = this;
          let getRoute = this._routerMicrolib.getRoute;

          this._routerMicrolib.getRoute = function (routeName) {
            fetchedHandlers.push(routeName);

            // Cache the returns so we don't have more than one Promise for a
            // given handler.
            return handlerPromises[routeName] || (handlerPromises[routeName] = new _runtime.RSVP.Promise(resolve => {
              setTimeout(() => {
                let handler = getRoute(routeName);

                seenHandlers[routeName] = handler;

                resolve(handler);
              }, 10);
            }));
          };
        },

        _getQPMeta(routeInfo) {
          let handler = this._seenHandlers[routeInfo.name];
          if (handler) {
            return (0, _metal.get)(handler, '_qp');
          }
        }
      };
    }

    ['@test can render a link to an asynchronously loaded route without fetching the route'](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('post', { path: '/post/:id' });
      });

      this.setSingleQPController('post');

      let setupAppTemplate = () => {
        this.addTemplate('application', `
        {{link-to 'Post' 'post' 1337 (query-params foo='bar') class='post-link is-1337'}}
        {{link-to 'Post' 'post' 7331 (query-params foo='boo') class='post-link is-7331'}}
        {{outlet}}
      `);
      };

      setupAppTemplate();

      return this.visitAndAssert('/').then(() => {
        assert.equal(this.$('.post-link.is-1337').attr('href'), '/post/1337?foo=bar', 'renders correctly with default QP value');
        assert.equal(this.$('.post-link.is-7331').attr('href'), '/post/7331?foo=boo', 'renders correctly with non-default QP value');
        assert.deepEqual(this.fetchedHandlers, ['application', 'index'], `only fetched the handlers for the route we're on`);
      });
    }

    ['@test can transitionTo to an asynchronously loaded route with simple query params'](assert) {
      assert.expect(6);

      this.router.map(function () {
        this.route('post', { path: '/post/:id' });
        this.route('posts');
      });

      this.setSingleQPController('post');

      let postController;
      return this.visitAndAssert('/').then(() => {
        postController = this.getController('post');

        return this.transitionTo('posts').then(() => {
          this.assertCurrentPath('/posts');
        });
      }).then(() => {
        return this.transitionTo('post', 1337, {
          queryParams: { foo: 'boo' }
        }).then(() => {
          assert.equal(postController.get('foo'), 'boo', 'simple QP is correctly set on controller');
          this.assertCurrentPath('/post/1337?foo=boo');
        });
      }).then(() => {
        return this.transitionTo('post', 1337, {
          queryParams: { foo: 'bar' }
        }).then(() => {
          assert.equal(postController.get('foo'), 'bar', 'simple QP is correctly set with default value');
          this.assertCurrentPath('/post/1337');
        });
      });
    }

    ['@test can transitionTo to an asynchronously loaded route with array query params'](assert) {
      assert.expect(5);

      this.router.map(function () {
        this.route('post', { path: '/post/:id' });
      });

      this.setSingleQPController('post', 'comments', []);

      let postController;
      return this.visitAndAssert('/').then(() => {
        postController = this.getController('post');
        return this.transitionTo('post', 1337, {
          queryParams: { comments: [1, 2] }
        }).then(() => {
          assert.deepEqual(postController.get('comments'), [1, 2], 'array QP is correctly set with default value');
          this.assertCurrentPath('/post/1337?comments=%5B1%2C2%5D');
        });
      }).then(() => {
        return this.transitionTo('post', 1338).then(() => {
          assert.deepEqual(postController.get('comments'), [], 'array QP is correctly set on controller');
          this.assertCurrentPath('/post/1338');
        });
      });
    }

    ['@test can transitionTo to an asynchronously loaded route with mapped query params'](assert) {
      assert.expect(7);

      this.router.map(function () {
        this.route('post', { path: '/post/:id' }, function () {
          this.route('index', { path: '/' });
        });
      });

      this.setSingleQPController('post');
      this.setMappedQPController('post.index', 'comment', 'note');

      let postController;
      let postIndexController;

      return this.visitAndAssert('/').then(() => {
        postController = this.getController('post');
        postIndexController = this.getController('post.index');

        return this.transitionTo('post.index', 1337, {
          queryParams: { note: 6, foo: 'boo' }
        }).then(() => {
          assert.equal(postController.get('foo'), 'boo', 'simple QP is correctly set on controller');
          assert.equal(postIndexController.get('comment'), 6, 'mapped QP is correctly set on controller');
          this.assertCurrentPath('/post/1337?foo=boo&note=6');
        });
      }).then(() => {
        return this.transitionTo('post', 1337, {
          queryParams: { foo: 'bar' }
        }).then(() => {
          assert.equal(postController.get('foo'), 'bar', 'simple QP is correctly set with default value');
          assert.equal(postIndexController.get('comment'), 6, 'mapped QP retains value scoped to model');
          this.assertCurrentPath('/post/1337?note=6');
        });
      });
    }

    ['@test can transitionTo with a URL'](assert) {
      assert.expect(7);

      this.router.map(function () {
        this.route('post', { path: '/post/:id' }, function () {
          this.route('index', { path: '/' });
        });
      });

      this.setSingleQPController('post');
      this.setMappedQPController('post.index', 'comment', 'note');

      let postController;
      let postIndexController;

      return this.visitAndAssert('/').then(() => {
        postController = this.getController('post');
        postIndexController = this.getController('post.index');

        return this.transitionTo('/post/1337?foo=boo&note=6').then(() => {
          assert.equal(postController.get('foo'), 'boo', 'simple QP is correctly deserialized on controller');
          assert.equal(postIndexController.get('comment'), 6, 'mapped QP is correctly deserialized on controller');
          this.assertCurrentPath('/post/1337?foo=boo&note=6');
        });
      }).then(() => {
        return this.transitionTo('/post/1337?note=6').then(() => {
          assert.equal(postController.get('foo'), 'bar', 'simple QP is correctly deserialized with default value');
          assert.equal(postIndexController.get('comment'), 6, 'mapped QP retains value scoped to model');
          this.assertCurrentPath('/post/1337?note=6');
        });
      });
    }

    ["@test undefined isn't serialized or deserialized into a string"](assert) {
      assert.expect(4);

      this.router.map(function () {
        this.route('example');
      });

      this.addTemplate('application', "{{link-to 'Example' 'example' (query-params foo=undefined) id='the-link'}}");

      this.setSingleQPController('example', 'foo', undefined, {
        foo: undefined
      });

      this.add('route:example', _routing.Route.extend({
        model(params) {
          assert.deepEqual(params, { foo: undefined });
        }
      }));

      return this.visitAndAssert('/').then(() => {
        assert.equal(this.$('#the-link').attr('href'), '/example', 'renders without undefined qp serialized');

        return this.transitionTo('example', {
          queryParams: { foo: undefined }
        }).then(() => {
          this.assertCurrentPath('/example');
        });
      });
    }
  });
});
enifed('ember/tests/routing/query_params_test/query_params_paramless_link_to_test', ['@ember/controller', 'internal-test-helpers'], function (_controller, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Query Params - paramless link-to', class extends _internalTestHelpers.QueryParamTestCase {
    testParamlessLinks(assert, routeName) {
      assert.expect(1);

      this.addTemplate(routeName, "{{link-to 'index' 'index' id='index-link'}}");

      this.add(`controller:${routeName}`, _controller.default.extend({
        queryParams: ['foo'],
        foo: 'wat'
      }));

      return this.visit('/?foo=YEAH').then(() => {
        assert.equal(document.getElementById('index-link').getAttribute('href'), '/?foo=YEAH');
      });
    }

    ["@test param-less links in an app booted with query params in the URL don't reset the query params: application"](assert) {
      return this.testParamlessLinks(assert, 'application');
    }

    ["@test param-less links in an app booted with query params in the URL don't reset the query params: index"](assert) {
      return this.testParamlessLinks(assert, 'index');
    }
  });
});
enifed('ember/tests/routing/query_params_test/shared_state_test', ['@ember/controller', '@ember/service', '@ember/runloop', 'internal-test-helpers'], function (_controller, _service, _runloop, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Query Params - shared service state', class extends _internalTestHelpers.QueryParamTestCase {
    boot() {
      this.setupApplication();
      return this.visitApplication();
    }

    setupApplication() {
      this.router.map(function () {
        this.route('home', { path: '/' });
        this.route('dashboard');
      });

      this.add('service:filters', _service.default.extend({
        shared: true
      }));

      this.add('controller:home', _controller.default.extend({
        filters: (0, _service.inject)()
      }));

      this.add('controller:dashboard', _controller.default.extend({
        filters: (0, _service.inject)(),
        queryParams: [{ 'filters.shared': 'shared' }]
      }));

      this.addTemplate('application', `{{link-to 'Home' 'home' }} <div> {{outlet}} </div>`);
      this.addTemplate('home', `{{link-to 'Dashboard' 'dashboard' }}{{input type="checkbox" id='filters-checkbox' checked=(mut filters.shared) }}`);
      this.addTemplate('dashboard', `{{link-to 'Home' 'home' }}`);
    }
    visitApplication() {
      return this.visit('/');
    }

    ['@test can modify shared state before transition'](assert) {
      assert.expect(1);

      return this.boot().then(() => {
        this.$input = document.getElementById('filters-checkbox');

        // click the checkbox once to set filters.shared to false
        (0, _runloop.run)(this.$input, 'click');

        return this.visit('/dashboard').then(() => {
          assert.ok(true, 'expecting navigating to dashboard to succeed');
        });
      });
    }

    ['@test can modify shared state back to the default value before transition'](assert) {
      assert.expect(1);

      return this.boot().then(() => {
        this.$input = document.getElementById('filters-checkbox');

        // click the checkbox twice to set filters.shared to false and back to true
        (0, _runloop.run)(this.$input, 'click');
        (0, _runloop.run)(this.$input, 'click');

        return this.visit('/dashboard').then(() => {
          assert.ok(true, 'expecting navigating to dashboard to succeed');
        });
      });
    }
  });
});
enifed('ember/tests/routing/router_map_test', ['internal-test-helpers', '@ember/runloop', '@ember/-internals/routing'], function (_internalTestHelpers, _runloop, _routing) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Router.map', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Router.map returns an Ember Router class'](assert) {
      assert.expect(1);

      let ret = this.router.map(function () {
        this.route('hello');
      });

      assert.ok(_routing.Router.detect(ret));
    }

    ['@test Router.map can be called multiple times'](assert) {
      assert.expect(2);

      this.addTemplate('hello', 'Hello!');
      this.addTemplate('goodbye', 'Goodbye!');

      this.router.map(function () {
        this.route('hello');
      });

      this.router.map(function () {
        this.route('goodbye');
      });

      return (0, _runloop.run)(() => {
        return this.visit('/hello').then(() => {
          this.assertText('Hello!');
        }).then(() => {
          return this.visit('/goodbye');
        }).then(() => {
          this.assertText('Goodbye!');
        });
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/basic_test', ['@ember/-internals/routing', '@ember/-internals/metal', 'internal-test-helpers'], function (_routing, _metal, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Router Service - main', class extends _internalTestHelpers.RouterTestCase {
    ['@test RouterService#currentRouteName is correctly set for top level route'](assert) {
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.expect(6);
        } else {
        assert.expect(1);
      }

      return this.visit('/').then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName, params, paramNames, queryParams } = currentRoute;
            assert.equal(name, 'parent.index');
            assert.equal(localName, 'index');
            assert.deepEqual(params, {});
            assert.deepEqual(queryParams, {});
            assert.deepEqual(paramNames, []);
          }

        assert.equal(this.routerService.get('currentRouteName'), 'parent.index');
      });
    }

    ['@test RouterService#currentRouteName is correctly set for child route'](assert) {
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.expect(6);
        } else {
        assert.expect(1);
      }

      return this.visit('/child').then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName, params, paramNames, queryParams } = currentRoute;
            assert.equal(name, 'parent.child');
            assert.equal(localName, 'child');
            assert.deepEqual(params, {});
            assert.deepEqual(queryParams, {});
            assert.deepEqual(paramNames, []);
          }

        assert.equal(this.routerService.get('currentRouteName'), 'parent.child');
      });
    }

    ['@test RouterService#currentRouteName is correctly set after transition'](assert) {
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.expect(5);
        } else {
        assert.expect(1);
      }

      return this.visit('/child').then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName } = currentRoute;
            assert.equal(name, 'parent.child');
            assert.equal(localName, 'child');
          }

        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName } = currentRoute;
            assert.equal(name, 'parent.sister');
            assert.equal(localName, 'sister');
          }
        assert.equal(this.routerService.get('currentRouteName'), 'parent.sister');
      });
    }

    ['@test RouterService#currentRouteName is correctly set on each transition'](assert) {
      if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
          assert.expect(9);
        } else {
        assert.expect(3);
      }

      return this.visit('/child').then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName } = currentRoute;
            assert.equal(name, 'parent.child');
            assert.equal(localName, 'child');
          }
        assert.equal(this.routerService.get('currentRouteName'), 'parent.child');

        return this.visit('/sister');
      }).then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName } = currentRoute;
            assert.equal(name, 'parent.sister');
            assert.equal(localName, 'sister');
          }
        assert.equal(this.routerService.get('currentRouteName'), 'parent.sister');

        return this.visit('/brother');
      }).then(() => {
        if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
            let currentRoute = this.routerService.currentRoute;
            let { name, localName } = currentRoute;
            assert.equal(name, 'parent.brother');
            assert.equal(localName, 'brother');
          }
        assert.equal(this.routerService.get('currentRouteName'), 'parent.brother');
      });
    }

    ['@test RouterService#rootURL is correctly set to the default value'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        assert.equal(this.routerService.get('rootURL'), '/');
      });
    }

    ['@test RouterService#rootURL is correctly set to a custom value'](assert) {
      assert.expect(1);

      this.add('route:parent.index', _routing.Route.extend({
        init() {
          this._super();
          (0, _metal.set)(this._router, 'rootURL', '/homepage');
        }
      }));

      return this.visit('/').then(() => {
        assert.equal(this.routerService.get('rootURL'), '/homepage');
      });
    }

    ['@test RouterService#location is correctly delegated from router:main'](assert) {
      assert.expect(2);

      return this.visit('/').then(() => {
        let location = this.routerService.get('location');
        assert.ok(location);
        assert.ok(location instanceof _routing.NoneLocation);
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/currenturl_lifecycle_test', ['@ember/service', '@ember/object/computed', '@ember/-internals/glimmer', '@ember/-internals/routing', '@ember/-internals/metal', 'internal-test-helpers'], function (_service, _computed, _glimmer, _routing, _metal, _internalTestHelpers) {
  'use strict';

  let results = [];
  let ROUTE_NAMES = ['index', 'child', 'sister', 'brother'];

  let InstrumentedRoute = _routing.Route.extend({
    routerService: (0, _service.inject)('router'),

    beforeModel() {
      let service = (0, _metal.get)(this, 'routerService');
      results.push([service.get('currentRouteName'), 'beforeModel', service.get('currentURL')]);
    },

    model() {
      let service = (0, _metal.get)(this, 'routerService');
      results.push([service.get('currentRouteName'), 'model', service.get('currentURL')]);
    },

    afterModel() {
      let service = (0, _metal.get)(this, 'routerService');
      results.push([service.get('currentRouteName'), 'afterModel', service.get('currentURL')]);
    }
  });

  (0, _internalTestHelpers.moduleFor)('Router Service - currentURL', class extends _internalTestHelpers.RouterTestCase {
    constructor() {
      super(...arguments);

      results = [];

      ROUTE_NAMES.forEach(name => {
        let routeName = `parent.${name}`;
        this.add(`route:${routeName}`, InstrumentedRoute.extend());
        this.addTemplate(routeName, '{{current-url}}');
      });

      this.addComponent('current-url', {
        ComponentClass: _glimmer.Component.extend({
          routerService: (0, _service.inject)('router'),
          currentURL: (0, _computed.readOnly)('routerService.currentURL')
        }),
        template: '{{currentURL}}'
      });
    }

    ['@test RouterService#currentURL is correctly set for top level route'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        assert.equal(this.routerService.get('currentURL'), '/');
      });
    }

    ['@test RouterService#currentURL is correctly set for child route'](assert) {
      assert.expect(1);

      return this.visit('/child').then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child');
      });
    }

    ['@test RouterService#currentURL is correctly set after transition'](assert) {
      assert.expect(1);

      return this.visit('/child').then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/sister');
      });
    }

    ['@test RouterService#currentURL is correctly set on each transition'](assert) {
      assert.expect(3);

      return this.visit('/child').then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child');

        return this.visit('/sister');
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/sister');

        return this.visit('/brother');
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/brother');
      });
    }

    ['@test RouterService#currentURL is not set during lifecycle hooks'](assert) {
      assert.expect(2);

      return this.visit('/').then(() => {
        assert.deepEqual(results, [[null, 'beforeModel', null], [null, 'model', null], [null, 'afterModel', null]]);

        results = [];

        return this.visit('/child');
      }).then(() => {
        assert.deepEqual(results, [['parent.index', 'beforeModel', '/'], ['parent.index', 'model', '/'], ['parent.index', 'afterModel', '/']]);
      });
    }

    ['@test RouterService#currentURL is correctly set with component after consecutive visits'](assert) {
      assert.expect(3);

      return this.visit('/').then(() => {
        this.assertText('/');

        return this.visit('/child');
      }).then(() => {
        this.assertText('/child');

        return this.visit('/');
      }).then(() => {
        this.assertText('/');
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/events_test', ['internal-test-helpers', '@ember/service', '@ember/-internals/routing'], function (_internalTestHelpers, _service, _routing) {
  'use strict';

  if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
      (0, _internalTestHelpers.moduleFor)('Router Service - events', class extends _internalTestHelpers.RouterTestCase {
        '@test initial render'(assert) {
          assert.expect(12);
          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);
              this.router.on('routeWillChange', transition => {
                assert.ok(transition);
                assert.equal(transition.from, undefined);
                assert.equal(transition.to.name, 'parent.index');
                assert.equal(transition.to.localName, 'index');
              });

              this.router.on('routeDidChange', transition => {
                assert.ok(transition);
                assert.ok(this.router.currentURL, `has URL ${this.router.currentURL}`);
                assert.equal(this.router.currentURL, '/');
                assert.ok(this.router.currentRouteName, `has route name ${this.router.currentRouteName}`);
                assert.equal(this.router.currentRouteName, 'parent.index');
                assert.equal(transition.from, undefined);
                assert.equal(transition.to.name, 'parent.index');
                assert.equal(transition.to.localName, 'index');
              });
            }
          }));
          return this.visit('/');
        }

        '@test subsequent visits'(assert) {
          assert.expect(24);
          let toParent = true;

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);
              this.router.on('routeWillChange', transition => {
                if (toParent) {
                  assert.equal(this.router.currentURL, null, 'starts as null');
                  assert.equal(transition.from, undefined);
                  assert.equal(transition.to.name, 'parent.child');
                  assert.equal(transition.to.localName, 'child');
                  assert.equal(transition.to.parent.name, 'parent', 'parent node');
                  assert.equal(transition.to.parent.child, transition.to, 'parents child node is the `to`');
                  assert.equal(transition.to.parent.parent.name, 'application', 'top level');
                  assert.equal(transition.to.parent.parent.parent, null, 'top level');
                } else {
                  assert.equal(this.router.currentURL, '/child', 'not changed until transition');
                  assert.notEqual(transition.from, undefined);
                  assert.equal(transition.from.name, 'parent.child');
                  assert.equal(transition.from.localName, 'child');
                  assert.equal(transition.to.localName, 'sister');
                  assert.equal(transition.to.name, 'parent.sister');
                }
              });

              this.router.on('routeDidChange', transition => {
                if (toParent) {
                  assert.equal(this.router.currentURL, '/child');
                  assert.equal(transition.from, undefined);
                  assert.equal(transition.to.name, 'parent.child');
                  assert.equal(transition.to.localName, 'child');
                } else {
                  assert.equal(this.router.currentURL, '/sister');
                  assert.notEqual(transition.from, undefined);
                  assert.equal(transition.from.name, 'parent.child');
                  assert.equal(transition.from.localName, 'child');
                  assert.equal(transition.to.localName, 'sister');
                  assert.equal(transition.to.name, 'parent.sister');
                }
              });
            }
          }));
          return this.visit('/child').then(() => {
            toParent = false;
            return this.routerService.transitionTo('parent.sister');
          });
        }

        '@test redirection with `transitionTo`'(assert) {
          assert.expect(8);
          let toChild = false;
          let toSister = false;

          this.add(`route:parent`, _routing.Route.extend({
            model() {
              this.transitionTo('parent.child');
            }
          }));

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              this.transitionTo('parent.sister');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                assert.equal(transition.from, undefined, 'initial');
                if (toChild) {
                  if (toSister) {
                    assert.equal(transition.to.name, 'parent.sister', 'going to /sister');
                  } else {
                    assert.equal(transition.to.name, 'parent.child', 'going to /child');
                    toSister = true;
                  }
                } else {
                  // Going to `/`
                  assert.equal(transition.to.name, 'parent.index', 'going to /');
                  toChild = true;
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.from, undefined, 'initial');
                assert.equal(transition.to.name, 'parent.sister', 'landed on /sister');
              });
            }
          }));
          return this.visit('/');
        }

        '@test redirection with `replaceWith`'(assert) {
          assert.expect(8);
          let toChild = false;
          let toSister = false;

          this.add(`route:parent`, _routing.Route.extend({
            model() {
              this.replaceWith('parent.child');
            }
          }));

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              this.replaceWith('parent.sister');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                assert.equal(transition.from, undefined, 'initial');
                if (toChild) {
                  if (toSister) {
                    assert.equal(transition.to.name, 'parent.sister', 'going to /sister');
                  } else {
                    assert.equal(transition.to.name, 'parent.child', 'going to /child');
                    toSister = true;
                  }
                } else {
                  // Going to `/`
                  assert.equal(transition.to.name, 'parent.index', 'going to /');
                  toChild = true;
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.from, undefined, 'initial');
                assert.equal(transition.to.name, 'parent.sister', 'landed on /sister');
              });
            }
          }));
          return this.visit('/');
        }

        '@test nested redirection with `transitionTo`'(assert) {
          assert.expect(11);
          let toChild = false;
          let toSister = false;

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              this.transitionTo('parent.sister');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                if (toChild) {
                  assert.equal(transition.from.name, 'parent.index');
                  if (toSister) {
                    assert.equal(transition.to.name, 'parent.sister', 'going to /sister');
                  } else {
                    assert.equal(transition.to.name, 'parent.child', 'going to /child');
                    toSister = true;
                  }
                } else {
                  // Going to `/`
                  assert.equal(transition.to.name, 'parent.index', 'going to /');
                  assert.equal(transition.from, undefined, 'initial');
                }
              });

              this.router.on('routeDidChange', transition => {
                if (toSister) {
                  assert.equal(transition.from.name, 'parent.index', 'initial');
                  assert.equal(transition.to.name, 'parent.sister', 'landed on /sister');
                } else {
                  assert.equal(transition.from, undefined, 'initial');
                  assert.equal(transition.to.name, 'parent.index', 'landed on /');
                }
              });
            }
          }));
          return this.visit('/').then(() => {
            toChild = true;
            return this.routerService.transitionTo('/child').catch(e => {
              assert.equal(e.name, 'TransitionAborted', 'Transition aborted');
            });
          });
        }

        '@test nested redirection with `replaceWith`'(assert) {
          assert.expect(11);
          let toChild = false;
          let toSister = false;

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              this.replaceWith('parent.sister');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                if (toChild) {
                  assert.equal(transition.from.name, 'parent.index');
                  if (toSister) {
                    assert.equal(transition.to.name, 'parent.sister', 'going to /sister');
                  } else {
                    assert.equal(transition.to.name, 'parent.child', 'going to /child');
                    toSister = true;
                  }
                } else {
                  // Going to `/`
                  assert.equal(transition.to.name, 'parent.index', 'going to /');
                  assert.equal(transition.from, undefined, 'initial');
                }
              });

              this.router.on('routeDidChange', transition => {
                if (toSister) {
                  assert.equal(transition.from.name, 'parent.index', 'initial');
                  assert.equal(transition.to.name, 'parent.sister', 'landed on /sister');
                } else {
                  assert.equal(transition.from, undefined, 'initial');
                  assert.equal(transition.to.name, 'parent.index', 'landed on /');
                }
              });
            }
          }));
          return this.visit('/').then(() => {
            toChild = true;
            return this.routerService.transitionTo('/child').catch(e => {
              assert.equal(e.name, 'TransitionAborted', 'Transition aborted');
            });
          });
        }

        '@test aborted transition'(assert) {
          assert.expect(11);
          let didAbort = false;
          let toChild = false;

          this.add(`route:parent.child`, _routing.Route.extend({
            model(_model, transition) {
              didAbort = true;
              transition.abort();
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                if (didAbort) {
                  assert.equal(transition.to.name, 'parent.index', 'transition aborted');
                  assert.equal(transition.from.name, 'parent.index', 'transition aborted');
                } else if (toChild) {
                  assert.equal(transition.from.name, 'parent.index', 'from /');
                  assert.equal(transition.to.name, 'parent.child', 'to /child');
                } else {
                  assert.equal(transition.to.name, 'parent.index', 'going to /');
                  assert.equal(transition.from, undefined, 'initial');
                }
              });

              this.router.on('routeDidChange', transition => {
                if (didAbort) {
                  assert.equal(transition.to.name, 'parent.index', 'landed on /');
                  assert.equal(transition.from.name, 'parent.index', 'initial');
                } else {
                  assert.equal(transition.to.name, 'parent.index', 'transition aborted');
                  assert.equal(transition.from, undefined, 'transition aborted');
                }
              });
            }
          }));
          return this.visit('/').then(() => {
            toChild = true;
            return this.routerService.transitionTo('/child').catch(e => {
              assert.equal(e.name, 'TransitionAborted', 'Transition aborted');
            });
          });
        }

        '@test query param transitions'(assert) {
          assert.expect(15);
          let initial = true;
          let addQP = false;
          let removeQP = false;

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                assert.equal(transition.to.name, 'parent.index');
                if (initial) {
                  assert.equal(transition.from, null);
                  assert.deepEqual(transition.to.queryParams, { a: 'true' });
                } else if (addQP) {
                  assert.deepEqual(transition.from.queryParams, { a: 'true' });
                  assert.deepEqual(transition.to.queryParams, { a: 'false', b: 'b' });
                } else if (removeQP) {
                  assert.deepEqual(transition.from.queryParams, { a: 'false', b: 'b' });
                  assert.deepEqual(transition.to.queryParams, { a: 'false' });
                } else {
                  assert.ok(false, 'never');
                }
              });

              this.router.on('routeDidChange', transition => {
                if (initial) {
                  assert.equal(transition.from, null);
                  assert.deepEqual(transition.to.queryParams, { a: 'true' });
                } else if (addQP) {
                  assert.deepEqual(transition.from.queryParams, { a: 'true' });
                  assert.deepEqual(transition.to.queryParams, { a: 'false', b: 'b' });
                } else if (removeQP) {
                  assert.deepEqual(transition.from.queryParams, { a: 'false', b: 'b' });
                  assert.deepEqual(transition.to.queryParams, { a: 'false' });
                } else {
                  assert.ok(false, 'never');
                }
              });
            }
          }));

          return this.visit('/?a=true').then(() => {
            addQP = true;
            initial = false;
            return this.routerService.transitionTo('/?a=false&b=b');
          }).then(() => {
            removeQP = true;
            addQP = false;
            return this.routerService.transitionTo('/?a=false');
          });
        }

        '@test query param redirects with `transitionTo`'(assert) {
          assert.expect(6);
          let toSister = false;

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              toSister = true;
              this.transitionTo('/sister?a=a');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                if (toSister) {
                  assert.equal(transition.to.name, 'parent.sister');
                  assert.deepEqual(transition.to.queryParams, { a: 'a' });
                } else {
                  assert.equal(transition.to.name, 'parent.child');
                  assert.deepEqual(transition.to.queryParams, {});
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.to.name, 'parent.sister');
                assert.deepEqual(transition.to.queryParams, { a: 'a' });
              });
            }
          }));

          return this.visit('/child');
        }
        '@test query param redirects with `replaceWith`'(assert) {
          assert.expect(6);
          let toSister = false;

          this.add(`route:parent.child`, _routing.Route.extend({
            model() {
              toSister = true;
              this.replaceWith('/sister?a=a');
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                if (toSister) {
                  assert.equal(transition.to.name, 'parent.sister');
                  assert.deepEqual(transition.to.queryParams, { a: 'a' });
                } else {
                  assert.equal(transition.to.name, 'parent.child');
                  assert.deepEqual(transition.to.queryParams, {});
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.to.name, 'parent.sister');
                assert.deepEqual(transition.to.queryParams, { a: 'a' });
              });
            }
          }));

          return this.visit('/child');
        }

        '@test params'(assert) {
          assert.expect(14);

          let inital = true;

          this.add('route:dynamic', _routing.Route.extend({
            model(params) {
              if (inital) {
                assert.deepEqual(params, { dynamic_id: '123' });
              } else {
                assert.deepEqual(params, { dynamic_id: '1' });
              }
              return params;
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                assert.equal(transition.to.name, 'dynamic');
                if (inital) {
                  assert.deepEqual(transition.to.paramNames, ['dynamic_id']);
                  assert.deepEqual(transition.to.params, { dynamic_id: '123' });
                } else {
                  assert.deepEqual(transition.to.paramNames, ['dynamic_id']);
                  assert.deepEqual(transition.to.params, { dynamic_id: '1' });
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.to.name, 'dynamic');
                assert.deepEqual(transition.to.paramNames, ['dynamic_id']);
                if (inital) {
                  assert.deepEqual(transition.to.params, { dynamic_id: '123' });
                } else {
                  assert.deepEqual(transition.to.params, { dynamic_id: '1' });
                }
              });
            }
          }));

          return this.visit('/dynamic/123').then(() => {
            inital = false;
            return this.routerService.transitionTo('dynamic', 1);
          });
        }

        '@test nested params'(assert) {
          assert.expect(24);
          let initial = true;

          this.add('route:dynamicWithChild', _routing.Route.extend({
            model(params) {
              if (initial) {
                assert.deepEqual(params, { dynamic_id: '123' });
              } else {
                assert.deepEqual(params, { dynamic_id: '456' });
              }
              return params;
            }
          }));

          this.add('route:dynamicWithChild.child', _routing.Route.extend({
            model(params) {
              assert.deepEqual(params, { child_id: '456' });
              return params;
            }
          }));

          this.add(`route:application`, _routing.Route.extend({
            router: (0, _service.inject)('router'),
            init() {
              this._super(...arguments);

              this.router.on('routeWillChange', transition => {
                assert.equal(transition.to.name, 'dynamicWithChild.child');
                assert.deepEqual(transition.to.paramNames, ['child_id']);
                assert.deepEqual(transition.to.params, { child_id: '456' });
                assert.deepEqual(transition.to.parent.paramNames, ['dynamic_id']);
                if (initial) {
                  assert.deepEqual(transition.to.parent.params, { dynamic_id: '123' });
                } else {
                  assert.deepEqual(transition.to.parent.params, { dynamic_id: '456' });
                }
              });

              this.router.on('routeDidChange', transition => {
                assert.equal(transition.to.name, 'dynamicWithChild.child');
                assert.deepEqual(transition.to.paramNames, ['child_id']);
                assert.deepEqual(transition.to.params, { child_id: '456' });
                assert.deepEqual(transition.to.parent.paramNames, ['dynamic_id']);
                if (initial) {
                  assert.deepEqual(transition.to.parent.params, { dynamic_id: '123' });
                } else {
                  assert.deepEqual(transition.to.parent.params, { dynamic_id: '456' });
                }
              });
            }
          }));

          return this.visit('/dynamic-with-child/123/456').then(() => {
            initial = false;
            return this.routerService.transitionTo('/dynamic-with-child/456/456');
          });
        }
      });

      (0, _internalTestHelpers.moduleFor)('Router Service - deprecated events', class extends _internalTestHelpers.RouterTestCase {
        '@test willTransition events are deprecated'() {
          return this.visit('/').then(() => {
            expectDeprecation(() => {
              this.routerService['_router'].on('willTransition', () => {});
            }, 'You attempted to listen to the "willTransition" event which is deprecated. Please inject the router service and listen to the "routeWillChange" event.');
          });
        }

        '@test willTransition events are deprecated on routes'() {
          this.add('route:application', _routing.Route.extend({
            init() {
              this._super(...arguments);
              this.on('willTransition', () => {});
            }
          }));
          expectDeprecation(() => {
            return this.visit('/');
          }, 'You attempted to listen to the "willTransition" event which is deprecated. Please inject the router service and listen to the "routeWillChange" event.');
        }

        '@test didTransition events are deprecated on routes'() {
          this.add('route:application', _routing.Route.extend({
            init() {
              this._super(...arguments);
              this.on('didTransition', () => {});
            }
          }));
          expectDeprecation(() => {
            return this.visit('/');
          }, 'You attempted to listen to the "didTransition" event which is deprecated. Please inject the router service and listen to the "routeDidChange" event.');
        }

        '@test other events are not deprecated on routes'() {
          this.add('route:application', _routing.Route.extend({
            init() {
              this._super(...arguments);
              this.on('fixx', () => {});
            }
          }));
          expectNoDeprecation(() => {
            return this.visit('/');
          });
        }

        '@test didTransition events are deprecated'() {
          return this.visit('/').then(() => {
            expectDeprecation(() => {
              this.routerService['_router'].on('didTransition', () => {});
            }, 'You attempted to listen to the "didTransition" event which is deprecated. Please inject the router service and listen to the "routeDidChange" event.');
          });
        }

        '@test other events are not deprecated'() {
          return this.visit('/').then(() => {
            expectNoDeprecation(() => {
              this.routerService['_router'].on('wat', () => {});
            });
          });
        }
      });

      (0, _internalTestHelpers.moduleFor)('Router Service: deprecated willTransition hook', class extends _internalTestHelpers.RouterTestCase {
        get routerOptions() {
          return {
            willTransition() {
              this._super(...arguments);
              // Overrides
            }
          };
        }

        '@test willTransition hook is deprecated'() {
          expectDeprecation(() => {
            return this.visit('/');
          }, 'You attempted to override the "willTransition" method which is deprecated. Please inject the router service and listen to the "routeWillChange" event.');
        }
      });
      (0, _internalTestHelpers.moduleFor)('Router Service: deprecated didTransition hook', class extends _internalTestHelpers.RouterTestCase {
        get routerOptions() {
          return {
            didTransition() {
              this._super(...arguments);
              // Overrides
            }
          };
        }

        '@test didTransition hook is deprecated'() {
          expectDeprecation(() => {
            return this.visit('/');
          }, 'You attempted to override the "didTransition" method which is deprecated. Please inject the router service and listen to the "routeDidChange" event.');
        }
      });
    }
});
enifed('ember/tests/routing/router_service_test/isActive_test', ['@ember/controller', 'internal-test-helpers'], function (_controller, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Router Service - isActive', class extends _internalTestHelpers.RouterTestCase {
    ['@test RouterService#isActive returns true for simple route'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        assert.ok(this.routerService.isActive('parent.sister'));
      });
    }

    ['@test RouterService#isActive returns true for simple route with dynamic segments'](assert) {
      assert.expect(1);

      let dynamicModel = { id: 1 };

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('dynamic', dynamicModel);
      }).then(() => {
        assert.ok(this.routerService.isActive('dynamic', dynamicModel));
      });
    }

    ['@test RouterService#isActive does not eagerly instantiate controller for query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      this.add('controller:parent.sister', _controller.default.extend({
        queryParams: ['sort'],
        sort: 'ASC',

        init() {
          assert.ok(false, 'should never create');
          this._super(...arguments);
        }
      }));

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.brother');
      }).then(() => {
        assert.notOk(this.routerService.isActive('parent.sister', queryParams));
      });
    }

    ['@test RouterService#isActive is correct for simple route with basic query params'](assert) {
      assert.expect(2);

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort'],
        sort: 'ASC'
      }));

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child', queryParams);
      }).then(() => {
        assert.ok(this.routerService.isActive('parent.child', queryParams));
        assert.notOk(this.routerService.isActive('parent.child', this.buildQueryParams({ sort: 'DESC' })));
      });
    }

    ['@test RouterService#isActive for simple route with array as query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ sort: ['ascending'] });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child', queryParams);
      }).then(() => {
        assert.notOk(this.routerService.isActive('parent.child', this.buildQueryParams({ sort: 'descending' })));
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/recognize_test', ['internal-test-helpers', '@ember/-internals/routing'], function (_internalTestHelpers, _routing) {
  'use strict';

  if (true /* EMBER_ROUTING_ROUTER_SERVICE */) {
      (0, _internalTestHelpers.moduleFor)('Router Service - recognize', class extends _internalTestHelpers.RouterTestCase {
        '@test returns a RouteInfo for recognized URL'(assert) {
          return this.visit('/').then(() => {
            let routeInfo = this.routerService.recognize('/dynamic-with-child/123/1?a=b');
            assert.ok(routeInfo);
            let { name, localName, parent, child, params, queryParams, paramNames } = routeInfo;
            assert.equal(name, 'dynamicWithChild.child');
            assert.equal(localName, 'child');
            assert.ok(parent);
            assert.equal(parent.name, 'dynamicWithChild');
            assert.notOk(child);
            assert.deepEqual(params, { child_id: '1' });
            assert.deepEqual(queryParams, { a: 'b' });
            assert.deepEqual(paramNames, ['child_id']);
          });
        }

        '@test does not transition'(assert) {
          this.addTemplate('parent', 'Parent');
          this.addTemplate('dynamic-with-child.child', 'Dynamic Child');

          return this.visit('/').then(() => {
            this.routerService.recognize('/dynamic-with-child/123/1?a=b');
            this.assertText('Parent', 'Did not transition and cause render');
            assert.equal(this.routerService.currentURL, '/', 'Did not transition');
          });
        }

        '@test respects the usage of a different rootURL'(assert) {
          this.router.reopen({
            rootURL: '/app/'
          });

          return this.visit('/app').then(() => {
            let routeInfo = this.routerService.recognize('/app/child/');
            assert.ok(routeInfo);
            let { name, localName, parent } = routeInfo;
            assert.equal(name, 'parent.child');
            assert.equal(localName, 'child');
            assert.equal(parent.name, 'parent');
          });
        }

        '@test must include rootURL'() {
          this.addTemplate('parent', 'Parent');
          this.addTemplate('dynamic-with-child.child', 'Dynamic Child');

          this.router.reopen({
            rootURL: '/app/'
          });

          return this.visit('/app').then(() => {
            expectAssertion(() => {
              this.routerService.recognize('/dynamic-with-child/123/1?a=b');
            }, 'You must pass a url that begins with the application\'s rootURL "/app/"');
          });
        }

        '@test returns `null` if URL is not recognized'(assert) {
          return this.visit('/').then(() => {
            let routeInfo = this.routerService.recognize('/foo');
            assert.equal(routeInfo, null);
          });
        }
      });

      (0, _internalTestHelpers.moduleFor)('Router Service - recognizeAndLoad', class extends _internalTestHelpers.RouterTestCase {
        '@test returns a RouteInfoWithAttributes for recognized URL'(assert) {
          this.add('route:dynamicWithChild', _routing.Route.extend({
            model(params) {
              return { name: 'dynamicWithChild', data: params.dynamic_id };
            }
          }));
          this.add('route:dynamicWithChild.child', _routing.Route.extend({
            model(params) {
              return { name: 'dynamicWithChild.child', data: params.child_id };
            }
          }));

          return this.visit('/').then(() => {
            return this.routerService.recognizeAndLoad('/dynamic-with-child/123/1?a=b');
          }).then(routeInfoWithAttributes => {
            assert.ok(routeInfoWithAttributes);
            let {
              name,
              localName,
              parent,
              attributes,
              paramNames,
              params,
              queryParams
            } = routeInfoWithAttributes;
            assert.equal(name, 'dynamicWithChild.child');
            assert.equal(localName, 'child');
            assert.equal(parent.name, 'dynamicWithChild');
            assert.deepEqual(params, { child_id: '1' });
            assert.deepEqual(queryParams, { a: 'b' });
            assert.deepEqual(paramNames, ['child_id']);
            assert.deepEqual(attributes, { name: 'dynamicWithChild.child', data: '1' });
            assert.deepEqual(parent.attributes, { name: 'dynamicWithChild', data: '123' });
            assert.deepEqual(parent.paramNames, ['dynamic_id']);
            assert.deepEqual(parent.params, { dynamic_id: '123' });
          });
        }

        '@test does not transition'(assert) {
          this.addTemplate('parent', 'Parent{{outlet}}');
          this.addTemplate('parent.child', 'Child');

          this.add('route:parent.child', _routing.Route.extend({
            model() {
              return { name: 'child', data: ['stuff'] };
            }
          }));
          return this.visit('/').then(() => {
            this.routerService.on('routeWillChange', () => assert.ok(false));
            this.routerService.on('routeDidChange', () => assert.ok(false));
            return this.routerService.recognizeAndLoad('/child');
          }).then(() => {
            assert.equal(this.routerService.currentURL, '/');
            this.assertText('Parent');
          });
        }

        '@test respects the usage of a different rootURL'(assert) {
          this.router.reopen({
            rootURL: '/app/'
          });

          return this.visit('/app').then(() => {
            return this.routerService.recognizeAndLoad('/app/child/');
          }).then(routeInfoWithAttributes => {
            assert.ok(routeInfoWithAttributes);
            let { name, localName, parent } = routeInfoWithAttributes;
            assert.equal(name, 'parent.child');
            assert.equal(localName, 'child');
            assert.equal(parent.name, 'parent');
          });
        }

        '@test must include rootURL'() {
          this.router.reopen({
            rootURL: '/app/'
          });

          return this.visit('/app').then(() => {
            expectAssertion(() => {
              this.routerService.recognizeAndLoad('/dynamic-with-child/123/1?a=b');
            }, 'You must pass a url that begins with the application\'s rootURL "/app/"');
          });
        }

        '@test rejects if url is not recognized'(assert) {
          this.addTemplate('parent', 'Parent{{outlet}}');
          this.addTemplate('parent.child', 'Child');

          this.add('route:parent.child', _routing.Route.extend({
            model() {
              return { name: 'child', data: ['stuff'] };
            }
          }));
          return this.visit('/').then(() => {
            return this.routerService.recognizeAndLoad('/foo');
          }).then(() => {
            assert.ok(false, 'never');
          }, reason => {
            assert.equal(reason, 'URL /foo was not recognized');
          });
        }

        '@test rejects if there is an unhandled error'(assert) {
          this.addTemplate('parent', 'Parent{{outlet}}');
          this.addTemplate('parent.child', 'Child');

          this.add('route:parent.child', _routing.Route.extend({
            model() {
              throw Error('Unhandled');
            }
          }));
          return this.visit('/').then(() => {
            return this.routerService.recognizeAndLoad('/child');
          }).then(() => {
            assert.ok(false, 'never');
          }, err => {
            assert.equal(err.message, 'Unhandled');
          });
        }
      });
    }
});
enifed('ember/tests/routing/router_service_test/replaceWith_test', ['@ember/-internals/routing', 'internal-test-helpers', 'router_js', '@ember/controller'], function (_routing, _internalTestHelpers, _router_js, _controller) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Router Service - replaceWith', class extends _internalTestHelpers.RouterTestCase {
    constructor() {
      super(...arguments);

      let testCase = this;
      testCase.state = [];

      this.add('location:test', _routing.NoneLocation.extend({
        setURL(path) {
          testCase.state.push(path);
          this.set('path', path);
        },

        replaceURL(path) {
          testCase.state.splice(testCase.state.length - 1, 1, path);
          this.set('path', path);
        }
      }));
    }

    get routerOptions() {
      return {
        location: 'test'
      };
    }

    ['@test RouterService#replaceWith returns a Transition'](assert) {
      assert.expect(1);

      let transition;

      return this.visit('/').then(() => {
        transition = this.routerService.replaceWith('parent.child');

        assert.ok(transition instanceof _router_js.InternalTransition);

        return transition;
      });
    }

    ['@test RouterService#replaceWith with basic route replaces location'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        return this.routerService.replaceWith('parent.brother');
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child', '/brother']);
      });
    }

    ['@test RouterService#replaceWith with basic route using URLs replaces location'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('/child');
      }).then(() => {
        return this.routerService.transitionTo('/sister');
      }).then(() => {
        return this.routerService.replaceWith('/brother');
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child', '/brother']);
      });
    }

    ['@test RouterService#replaceWith transitioning back to previously visited route replaces location'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        return this.routerService.transitionTo('parent.brother');
      }).then(() => {
        return this.routerService.replaceWith('parent.sister');
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child', '/sister', '/sister']);
      });
    }

    ['@test RouterService#replaceWith with basic query params does not remove query param defaults'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort'],
        sort: 'ASC'
      }));

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.brother');
      }).then(() => {
        return this.routerService.replaceWith('parent.sister');
      }).then(() => {
        return this.routerService.replaceWith('parent.child', queryParams);
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child?sort=ASC']);
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/transitionTo_test', ['@ember/service', '@ember/-internals/glimmer', '@ember/-internals/routing', '@ember/controller', '@ember/runloop', '@ember/-internals/metal', 'internal-test-helpers', 'router_js'], function (_service, _glimmer, _routing, _controller, _runloop, _metal, _internalTestHelpers, _router_js) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Router Service - transitionTo', class extends _internalTestHelpers.RouterTestCase {
    constructor() {
      super(...arguments);

      let testCase = this;
      testCase.state = [];

      this.add('location:test', _routing.NoneLocation.extend({
        setURL(path) {
          testCase.state.push(path);
          this.set('path', path);
        },

        replaceURL(path) {
          testCase.state.splice(testCase.state.length - 1, 1, path);
          this.set('path', path);
        }
      }));
    }

    get routerOptions() {
      return {
        location: 'test'
      };
    }

    ['@test RouterService#transitionTo returns a Transition'](assert) {
      assert.expect(1);

      let transition;

      return this.visit('/').then(() => {
        transition = this.routerService.transitionTo('parent.child');

        assert.ok(transition instanceof _router_js.InternalTransition);

        return transition;
      });
    }

    ['@test RouterService#transitionTo with basic route updates location'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        return this.routerService.transitionTo('parent.brother');
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child', '/sister', '/brother']);
      });
    }

    ['@test RouterService#transitionTo transitioning back to previously visited route updates location'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        return this.routerService.transitionTo('parent.brother');
      }).then(() => {
        return this.routerService.transitionTo('parent.sister');
      }).then(() => {
        assert.deepEqual(this.state, ['/', '/child', '/sister', '/brother', '/sister']);
      });
    }

    ['@test RouterService#transitionTo with basic route'](assert) {
      assert.expect(1);

      let componentInstance;

      this.addTemplate('parent.index', '{{foo-bar}}');

      this.addComponent('foo-bar', {
        ComponentClass: _glimmer.Component.extend({
          routerService: (0, _service.inject)('router'),
          init() {
            this._super();
            componentInstance = this;
          },
          actions: {
            transitionToSister() {
              (0, _metal.get)(this, 'routerService').transitionTo('parent.sister');
            }
          }
        }),
        template: `foo-bar`
      });

      return this.visit('/').then(() => {
        (0, _runloop.run)(function () {
          componentInstance.send('transitionToSister');
        });

        assert.equal(this.routerService.get('currentRouteName'), 'parent.sister');
      });
    }

    ['@test RouterService#transitionTo with basic route using URL'](assert) {
      assert.expect(1);

      let componentInstance;

      this.addTemplate('parent.index', '{{foo-bar}}');

      this.addComponent('foo-bar', {
        ComponentClass: _glimmer.Component.extend({
          routerService: (0, _service.inject)('router'),
          init() {
            this._super();
            componentInstance = this;
          },
          actions: {
            transitionToSister() {
              (0, _metal.get)(this, 'routerService').transitionTo('/sister');
            }
          }
        }),
        template: `foo-bar`
      });

      return this.visit('/').then(() => {
        (0, _runloop.run)(function () {
          componentInstance.send('transitionToSister');
        });

        assert.equal(this.routerService.get('currentRouteName'), 'parent.sister');
      });
    }

    ['@test RouterService#transitionTo with dynamic segment'](assert) {
      assert.expect(3);

      let componentInstance;
      let dynamicModel = { id: 1, contents: 'much dynamicism' };

      this.addTemplate('parent.index', '{{foo-bar}}');
      this.addTemplate('dynamic', '{{model.contents}}');

      this.addComponent('foo-bar', {
        ComponentClass: _glimmer.Component.extend({
          routerService: (0, _service.inject)('router'),
          init() {
            this._super();
            componentInstance = this;
          },
          actions: {
            transitionToDynamic() {
              (0, _metal.get)(this, 'routerService').transitionTo('dynamic', dynamicModel);
            }
          }
        }),
        template: `foo-bar`
      });

      return this.visit('/').then(() => {
        (0, _runloop.run)(function () {
          componentInstance.send('transitionToDynamic');
        });

        assert.equal(this.routerService.get('currentRouteName'), 'dynamic');
        assert.equal(this.routerService.get('currentURL'), '/dynamic/1');
        this.assertText('much dynamicism');
      });
    }

    ['@test RouterService#transitionTo with dynamic segment and model hook'](assert) {
      assert.expect(3);

      let componentInstance;
      let dynamicModel = { id: 1, contents: 'much dynamicism' };

      this.add('route:dynamic', _routing.Route.extend({
        model() {
          return dynamicModel;
        }
      }));

      this.addTemplate('parent.index', '{{foo-bar}}');
      this.addTemplate('dynamic', '{{model.contents}}');

      this.addComponent('foo-bar', {
        ComponentClass: _glimmer.Component.extend({
          routerService: (0, _service.inject)('router'),
          init() {
            this._super();
            componentInstance = this;
          },
          actions: {
            transitionToDynamic() {
              (0, _metal.get)(this, 'routerService').transitionTo('dynamic', 1);
            }
          }
        }),
        template: `foo-bar`
      });

      return this.visit('/').then(() => {
        (0, _runloop.run)(function () {
          componentInstance.send('transitionToDynamic');
        });

        assert.equal(this.routerService.get('currentRouteName'), 'dynamic');
        assert.equal(this.routerService.get('currentURL'), '/dynamic/1');
        this.assertText('much dynamicism');
      });
    }

    ['@test RouterService#transitionTo with basic query params does not remove query param defaults'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort'],
        sort: 'ASC'
      }));

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child', queryParams);
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child?sort=ASC');
      });
    }

    ['@test RouterService#transitionTo passing only queryParams works'](assert) {
      assert.expect(2);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort']
      }));

      let queryParams = this.buildQueryParams({ sort: 'DESC' });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child');
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child');
      }).then(() => {
        return this.routerService.transitionTo(queryParams);
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child?sort=DESC');
      });
    }

    ['@test RouterService#transitionTo with unspecified query params'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort', 'page', 'category', 'extra'],
        sort: 'ASC',
        page: null,
        category: undefined
      }));

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child', queryParams);
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child?sort=ASC');
      });
    }

    ['@test RouterService#transitionTo with aliased query params uses the original provided key'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: {
          cont_sort: 'url_sort'
        },
        cont_sort: 'ASC'
      }));

      let queryParams = this.buildQueryParams({ url_sort: 'ASC' });

      return this.visit('/').then(() => {
        return this.routerService.transitionTo('parent.child', queryParams);
      }).then(() => {
        assert.equal(this.routerService.get('currentURL'), '/child?url_sort=ASC');
      });
    }

    ['@test RouterService#transitionTo with aliased query params uses the original provided key when controller property name'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: {
          cont_sort: 'url_sort'
        },
        cont_sort: 'ASC'
      }));

      let queryParams = this.buildQueryParams({ cont_sort: 'ASC' });

      return this.visit('/').then(() => {
        expectAssertion(() => {
          return this.routerService.transitionTo('parent.child', queryParams);
        }, 'You passed the `cont_sort` query parameter during a transition into parent.child, please update to url_sort');
      });
    }
  });
});
enifed('ember/tests/routing/router_service_test/urlFor_test', ['@ember/controller', '@ember/string', '@ember/-internals/routing', '@ember/-internals/metal', 'internal-test-helpers'], function (_controller, _string, _routing, _metal, _internalTestHelpers) {
  'use strict';

  function setupController(app, name) {
    let controllerName = `${(0, _string.capitalize)(name)}Controller`;

    Object.defineProperty(app, controllerName, {
      get() {
        throw new Error(`Generating a URL should not require instantiation of a ${controllerName}.`);
      }
    });
  }

  (0, _internalTestHelpers.moduleFor)('Router Service - urlFor', class extends _internalTestHelpers.RouterTestCase {
    ['@test RouterService#urlFor returns URL for simple route'](assert) {
      assert.expect(1);

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child');

        assert.equal('/child', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with dynamic segments'](assert) {
      assert.expect(1);

      setupController(this.application, 'dynamic');

      let dynamicModel = { id: 1, contents: 'much dynamicism' };

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('dynamic', dynamicModel);

        assert.equal('/dynamic/1', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with basic query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: 'bar' });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child', queryParams);

        assert.equal('/child?foo=bar', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with basic query params and default value'](assert) {
      assert.expect(1);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort'],
        sort: 'ASC'
      }));

      let queryParams = this.buildQueryParams({ sort: 'ASC' });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child', queryParams);

        assert.equal('/child?sort=ASC', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with basic query params and default value with stickyness'](assert) {
      assert.expect(2);

      this.add('controller:parent.child', _controller.default.extend({
        queryParams: ['sort', 'foo'],
        sort: 'ASC'
      }));

      return this.visit('/child/?sort=DESC').then(() => {
        let controller = this.applicationInstance.lookup('controller:parent.child');
        assert.equal((0, _metal.get)(controller, 'sort'), 'DESC', 'sticky is set');

        let queryParams = this.buildQueryParams({ foo: 'derp' });
        let actual = this.routerService.urlFor('parent.child', queryParams);

        assert.equal(actual, '/child?foo=derp', 'does not use "stickiness"');
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with array as query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({
        selectedItems: ['a', 'b', 'c']
      });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child', queryParams);

        assert.equal('/child?selectedItems[]=a&selectedItems[]=b&selectedItems[]=c', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with null query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: null });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child', queryParams);

        assert.equal('/child', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with undefined query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: undefined });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('parent.child', queryParams);

        assert.equal('/child', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with dynamic segments and basic query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: 'bar' });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('dynamic', { id: 1 }, queryParams);

        assert.equal('/dynamic/1?foo=bar', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with dynamic segments and array as query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({
        selectedItems: ['a', 'b', 'c']
      });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('dynamic', { id: 1 }, queryParams);

        assert.equal('/dynamic/1?selectedItems[]=a&selectedItems[]=b&selectedItems[]=c', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with dynamic segments and null query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: null });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('dynamic', { id: 1 }, queryParams);

        assert.equal('/dynamic/1', expectedURL);
      });
    }

    ['@test RouterService#urlFor returns URL for simple route with dynamic segments and undefined query params'](assert) {
      assert.expect(1);

      let queryParams = this.buildQueryParams({ foo: undefined });

      return this.visit('/').then(() => {
        let expectedURL = this.routerService.urlFor('dynamic', { id: 1 }, queryParams);

        assert.equal('/dynamic/1', expectedURL);
      });
    }

    ['@test RouterService#urlFor correctly transitions to route via generated path'](assert) {
      assert.expect(1);

      let expectedURL;

      return this.visit('/').then(() => {
        expectedURL = this.routerService.urlFor('parent.child');

        return this.routerService.transitionTo(expectedURL);
      }).then(() => {
        assert.equal(expectedURL, this.routerService.get('currentURL'));
      });
    }

    ['@test RouterService#urlFor correctly transitions to route via generated path with dynamic segments'](assert) {
      assert.expect(1);

      let expectedURL;
      let dynamicModel = { id: 1 };

      this.add('route:dynamic', _routing.Route.extend({
        model() {
          return dynamicModel;
        }
      }));

      return this.visit('/').then(() => {
        expectedURL = this.routerService.urlFor('dynamic', dynamicModel);

        return this.routerService.transitionTo(expectedURL);
      }).then(() => {
        assert.equal(expectedURL, this.routerService.get('currentURL'));
      });
    }

    ['@test RouterService#urlFor correctly transitions to route via generated path with query params'](assert) {
      assert.expect(1);

      let expectedURL;
      let actualURL;
      let queryParams = this.buildQueryParams({ foo: 'bar' });

      return this.visit('/').then(() => {
        expectedURL = this.routerService.urlFor('parent.child', queryParams);

        return this.routerService.transitionTo(expectedURL);
      }).then(() => {
        actualURL = `${this.routerService.get('currentURL')}?foo=bar`;

        assert.equal(expectedURL, actualURL);
      });
    }

    ['@test RouterService#urlFor correctly transitions to route via generated path with dynamic segments and query params'](assert) {
      assert.expect(1);

      let expectedURL;
      let actualURL;
      let queryParams = this.buildQueryParams({ foo: 'bar' });
      let dynamicModel = { id: 1 };

      this.add('route:dynamic', _routing.Route.extend({
        model() {
          return dynamicModel;
        }
      }));

      return this.visit('/').then(() => {
        expectedURL = this.routerService.urlFor('dynamic', dynamicModel, queryParams);

        return this.routerService.transitionTo(expectedURL);
      }).then(() => {
        actualURL = `${this.routerService.get('currentURL')}?foo=bar`;

        assert.equal(expectedURL, actualURL);
      });
    }
  });
});
enifed('ember/tests/routing/substates_test', ['@ember/-internals/runtime', '@ember/-internals/routing', 'internal-test-helpers'], function (_runtime, _routing, _internalTestHelpers) {
  'use strict';

  let counter;

  function step(assert, expectedValue, description) {
    assert.equal(counter, expectedValue, 'Step ' + expectedValue + ': ' + description);
    counter++;
  }

  (0, _internalTestHelpers.moduleFor)('Loading/Error Substates', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super(...arguments);
      counter = 1;

      this.addTemplate('application', `<div id="app">{{outlet}}</div>`);
      this.addTemplate('index', 'INDEX');
    }

    getController(name) {
      return this.applicationInstance.lookup(`controller:${name}`);
    }

    get currentPath() {
      return this.getController('application').get('currentPath');
    }

    ['@test Slow promise from a child route of application enters nested loading state'](assert) {
      let turtleDeferred = _runtime.RSVP.defer();

      this.router.map(function () {
        this.route('turtle');
      });

      this.add('route:application', _routing.Route.extend({
        setupController() {
          step(assert, 2, 'ApplicationRoute#setupController');
        }
      }));

      this.add('route:turtle', _routing.Route.extend({
        model() {
          step(assert, 1, 'TurtleRoute#model');
          return turtleDeferred.promise;
        }
      }));
      this.addTemplate('turtle', 'TURTLE');
      this.addTemplate('loading', 'LOADING');

      let promise = this.visit('/turtle').then(() => {
        text = this.$('#app').text();
        assert.equal(text, 'TURTLE', `turtle template has loaded and replaced the loading template`);
      });

      let text = this.$('#app').text();
      assert.equal(text, 'LOADING', `The Loading template is nested in application template's outlet`);

      turtleDeferred.resolve();
      return promise;
    }

    [`@test Slow promises returned from ApplicationRoute#model don't enter LoadingRoute`](assert) {
      let appDeferred = _runtime.RSVP.defer();

      this.add('route:application', _routing.Route.extend({
        model() {
          return appDeferred.promise;
        }
      }));
      this.add('route:loading', _routing.Route.extend({
        setupController() {
          assert.ok(false, `shouldn't get here`);
        }
      }));

      let promise = this.visit('/').then(() => {
        let text = this.$('#app').text();

        assert.equal(text, 'INDEX', `index template has been rendered`);
      });

      if (this.element) {
        assert.equal(this.element.textContent, '');
      }

      appDeferred.resolve();

      return promise;
    }

    [`@test Don't enter loading route unless either route or template defined`](assert) {
      let deferred = _runtime.RSVP.defer();

      this.router.map(function () {
        this.route('dummy');
      });
      this.add('route:dummy', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));
      this.addTemplate('dummy', 'DUMMY');

      return this.visit('/').then(() => {
        let promise = this.visit('/dummy').then(() => {
          let text = this.$('#app').text();

          assert.equal(text, 'DUMMY', `dummy template has been rendered`);
        });

        assert.ok(this.currentPath !== 'loading', `
        loading state not entered
      `);
        deferred.resolve();

        return promise;
      });
    }

    ['@test Enter loading route only if loadingRoute is defined'](assert) {
      let deferred = _runtime.RSVP.defer();

      this.router.map(function () {
        this.route('dummy');
      });

      this.add('route:dummy', _routing.Route.extend({
        model() {
          step(assert, 1, 'DummyRoute#model');
          return deferred.promise;
        }
      }));
      this.add('route:loading', _routing.Route.extend({
        setupController() {
          step(assert, 2, 'LoadingRoute#setupController');
        }
      }));
      this.addTemplate('dummy', 'DUMMY');

      return this.visit('/').then(() => {
        let promise = this.visit('/dummy').then(() => {
          let text = this.$('#app').text();

          assert.equal(text, 'DUMMY', `dummy template has been rendered`);
        });

        assert.equal(this.currentPath, 'loading', `loading state entered`);
        deferred.resolve();

        return promise;
      });
    }

    ['@test Slow promises returned from ApplicationRoute#model enter ApplicationLoadingRoute if present'](assert) {
      let appDeferred = _runtime.RSVP.defer();

      this.add('route:application', _routing.Route.extend({
        model() {
          return appDeferred.promise;
        }
      }));
      let loadingRouteEntered = false;
      this.add('route:application_loading', _routing.Route.extend({
        setupController() {
          loadingRouteEntered = true;
        }
      }));

      let promise = this.visit('/').then(() => {
        assert.equal(this.$('#app').text(), 'INDEX', 'index route loaded');
      });
      assert.ok(loadingRouteEntered, 'ApplicationLoadingRoute was entered');
      appDeferred.resolve();

      return promise;
    }

    ['@test Slow promises returned from ApplicationRoute#model enter application_loading if template present'](assert) {
      let appDeferred = _runtime.RSVP.defer();

      this.addTemplate('application_loading', `
      <div id="toplevel-loading">TOPLEVEL LOADING</div>
    `);
      this.add('route:application', _routing.Route.extend({
        model() {
          return appDeferred.promise;
        }
      }));

      let promise = this.visit('/').then(() => {
        let length = this.$('#toplevel-loading').length;
        text = this.$('#app').text();

        assert.equal(length, 0, `top-level loading view has been entirely removed from the DOM`);
        assert.equal(text, 'INDEX', 'index has fully rendered');
      });
      let text = this.$('#toplevel-loading').text();

      assert.equal(text, 'TOPLEVEL LOADING', 'still loading the top level');
      appDeferred.resolve();

      return promise;
    }

    ['@test Prioritized substate entry works with preserved-namespace nested routes'](assert) {
      let deferred = _runtime.RSVP.defer();

      this.addTemplate('foo.bar_loading', 'FOOBAR LOADING');
      this.addTemplate('foo.bar.index', 'YAY');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar', { path: '/bar' }, function () {});
        });
      });

      this.add('route:foo.bar', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      return this.visit('/').then(() => {
        let promise = this.visit('/foo/bar').then(() => {
          text = this.$('#app').text();

          assert.equal(text, 'YAY', 'foo.bar.index fully loaded');
        });
        let text = this.$('#app').text();

        assert.equal(text, 'FOOBAR LOADING', `foo.bar_loading was entered (as opposed to something like foo/foo/bar_loading)`);
        deferred.resolve();

        return promise;
      });
    }

    ['@test Prioritized substate entry works with reset-namespace nested routes'](assert) {
      let deferred = _runtime.RSVP.defer();

      this.addTemplate('bar_loading', 'BAR LOADING');
      this.addTemplate('bar.index', 'YAY');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar', { path: '/bar', resetNamespace: true }, function () {});
        });
      });

      this.add('route:bar', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      return this.visit('/').then(() => {
        let promise = this.visit('/foo/bar').then(() => {
          text = this.$('#app').text();

          assert.equal(text, 'YAY', 'bar.index fully loaded');
        });

        let text = this.$('#app').text();

        assert.equal(text, 'BAR LOADING', `foo.bar_loading was entered (as opposed to something likefoo/foo/bar_loading)`);
        deferred.resolve();

        return promise;
      });
    }

    ['@test Prioritized loading substate entry works with preserved-namespace nested routes'](assert) {
      let deferred = _runtime.RSVP.defer();

      this.addTemplate('foo.bar_loading', 'FOOBAR LOADING');
      this.addTemplate('foo.bar', 'YAY');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar');
        });
      });

      this.add('route:foo.bar', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      let promise = this.visit('/foo/bar').then(() => {
        text = this.$('#app').text();

        assert.equal(text, 'YAY', 'foo.bar has rendered');
      });
      let text = this.$('#app').text();

      assert.equal(text, 'FOOBAR LOADING', `foo.bar_loading was entered (as opposed to something like foo/foo/bar_loading)`);
      deferred.resolve();

      return promise;
    }

    ['@test Prioritized error substate entry works with preserved-namespaec nested routes'](assert) {
      this.addTemplate('foo.bar_error', 'FOOBAR ERROR: {{model.msg}}');
      this.addTemplate('foo.bar', 'YAY');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar');
        });
      });

      this.add('route:foo.bar', _routing.Route.extend({
        model() {
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        }
      }));

      return this.visit('/').then(() => {
        return this.visit('/foo/bar').then(() => {
          let text = this.$('#app').text();
          assert.equal(text, 'FOOBAR ERROR: did it broke?', `foo.bar_error was entered (as opposed to something like foo/foo/bar_error)`);
        });
      });
    }
    ['@test Prioritized loading substate entry works with auto-generated index routes'](assert) {
      let deferred = _runtime.RSVP.defer();
      this.addTemplate('foo.index_loading', 'FOO LOADING');
      this.addTemplate('foo.index', 'YAY');
      this.addTemplate('foo', '{{outlet}}');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar');
        });
      });

      this.add('route:foo.index', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));
      this.add('route:foo', _routing.Route.extend({
        model() {
          return true;
        }
      }));

      let promise = this.visit('/foo').then(() => {
        text = this.$('#app').text();

        assert.equal(text, 'YAY', 'foo.index was rendered');
      });
      let text = this.$('#app').text();
      assert.equal(text, 'FOO LOADING', 'foo.index_loading was entered');

      deferred.resolve();

      return promise;
    }

    ['@test Prioritized error substate entry works with auto-generated index routes'](assert) {
      this.addTemplate('foo.index_error', 'FOO ERROR: {{model.msg}}');
      this.addTemplate('foo.index', 'YAY');
      this.addTemplate('foo', '{{outlet}}');

      this.router.map(function () {
        this.route('foo', function () {
          this.route('bar');
        });
      });

      this.add('route:foo.index', _routing.Route.extend({
        model() {
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        }
      }));
      this.add('route:foo', _routing.Route.extend({
        model() {
          return true;
        }
      }));

      return this.visit('/').then(() => {
        return this.visit('/foo').then(() => {
          let text = this.$('#app').text();

          assert.equal(text, 'FOO ERROR: did it broke?', 'foo.index_error was entered');
        });
      });
    }

    ['@test Rejected promises returned from ApplicationRoute transition into top-level application_error'](assert) {
      let reject = true;

      this.addTemplate('index', '<div id="index">INDEX</div>');
      this.add('route:application', _routing.Route.extend({
        init() {
          this._super(...arguments);
        },
        model() {
          if (reject) {
            return _runtime.RSVP.reject({ msg: 'BAD NEWS BEARS' });
          } else {
            return {};
          }
        }
      }));

      this.addTemplate('application_error', `
      <p id="toplevel-error">TOPLEVEL ERROR: {{model.msg}}</p>
    `);

      return this.visit('/').then(() => {
        let text = this.$('#toplevel-error').text();
        assert.equal(text, 'TOPLEVEL ERROR: BAD NEWS BEARS', 'toplevel error rendered');
        reject = false;
      }).then(() => {
        return this.visit('/');
      }).then(() => {
        let text = this.$('#index').text();
        assert.equal(text, 'INDEX', 'the index route resolved');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Loading/Error Substates - nested routes', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super(...arguments);

      counter = 1;

      this.addTemplate('application', `<div id="app">{{outlet}}</div>`);
      this.addTemplate('index', 'INDEX');
      this.addTemplate('grandma', 'GRANDMA {{outlet}}');
      this.addTemplate('mom', 'MOM');

      this.router.map(function () {
        this.route('grandma', function () {
          this.route('mom', { resetNamespace: true }, function () {
            this.route('sally');
            this.route('this-route-throws');
          });
          this.route('puppies');
        });
        this.route('memere', { path: '/memere/:seg' }, function () {});
      });

      this.visit('/');
    }

    getController(name) {
      return this.applicationInstance.lookup(`controller:${name}`);
    }

    get currentPath() {
      return this.getController('application').get('currentPath');
    }

    ['@test ApplicationRoute#currentPath reflects loading state path'](assert) {
      let momDeferred = _runtime.RSVP.defer();

      this.addTemplate('grandma.loading', 'GRANDMALOADING');

      this.add('route:mom', _routing.Route.extend({
        model() {
          return momDeferred.promise;
        }
      }));

      let promise = this.visit('/grandma/mom').then(() => {
        text = this.$('#app').text();

        assert.equal(text, 'GRANDMA MOM', `Grandma.mom loaded text is displayed`);
        assert.equal(this.currentPath, 'grandma.mom.index', `currentPath reflects final state`);
      });
      let text = this.$('#app').text();

      assert.equal(text, 'GRANDMA GRANDMALOADING', `Grandma.mom loading text displayed`);

      assert.equal(this.currentPath, 'grandma.loading', `currentPath reflects loading state`);

      momDeferred.resolve();

      return promise;
    }

    [`@test Loading actions bubble to root but don't enter substates above pivot `](assert) {
      let sallyDeferred = _runtime.RSVP.defer();
      let puppiesDeferred = _runtime.RSVP.defer();

      this.add('route:application', _routing.Route.extend({
        actions: {
          loading() {
            assert.ok(true, 'loading action received on ApplicationRoute');
          }
        }
      }));

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          return sallyDeferred.promise;
        }
      }));

      this.add('route:grandma.puppies', _routing.Route.extend({
        model() {
          return puppiesDeferred.promise;
        }
      }));

      let promise = this.visit('/grandma/mom/sally');
      assert.equal(this.currentPath, 'index', 'Initial route fully loaded');

      sallyDeferred.resolve();

      promise.then(() => {
        assert.equal(this.currentPath, 'grandma.mom.sally', 'transition completed');

        let visit = this.visit('/grandma/puppies');
        assert.equal(this.currentPath, 'grandma.mom.sally', 'still in initial state because the only loading state is above the pivot route');

        return visit;
      }).then(() => {
        this.runTask(() => puppiesDeferred.resolve());

        assert.equal(this.currentPath, 'grandma.puppies', 'Finished transition');
      });

      return promise;
    }

    ['@test Default error event moves into nested route'](assert) {
      this.addTemplate('grandma.error', 'ERROR: {{model.msg}}');

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error() {
            step(assert, 2, 'MomSallyRoute#actions.error');
            return true;
          }
        }
      }));

      return this.visit('/grandma/mom/sally').then(() => {
        step(assert, 3, 'App finished loading');

        let text = this.$('#app').text();

        assert.equal(text, 'GRANDMA ERROR: did it broke?', 'error bubbles');
        assert.equal(this.currentPath, 'grandma.error', 'Initial route fully loaded');
      });
    }

    [`@test Non-bubbled errors that re-throw aren't swallowed`](assert) {
      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error(err) {
            // returns undefined which is falsey
            throw err;
          }
        }
      }));

      assert.throws(() => {
        this.visit('/grandma/mom/sally');
      }, function (err) {
        return err.msg === 'did it broke?';
      }, 'it broke');

      return this.runLoopSettled();
    }

    [`@test Handled errors that re-throw aren't swallowed`](assert) {
      let handledError;

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error(err) {
            step(assert, 2, 'MomSallyRoute#actions.error');
            handledError = err;
            this.transitionTo('mom.this-route-throws');

            return false;
          }
        }
      }));

      this.add('route:mom.this-route-throws', _routing.Route.extend({
        model() {
          step(assert, 3, 'MomThisRouteThrows#model');
          throw handledError;
        }
      }));

      assert.throws(() => {
        this.visit('/grandma/mom/sally');
      }, function (err) {
        return err.msg === 'did it broke?';
      }, `it broke`);

      return this.runLoopSettled();
    }

    ['@test errors that are bubbled are thrown at a higher level if not handled'](assert) {
      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error() {
            step(assert, 2, 'MomSallyRoute#actions.error');
            return true;
          }
        }
      }));

      assert.throws(() => {
        this.visit('/grandma/mom/sally');
      }, function (err) {
        return err.msg == 'did it broke?';
      }, 'Correct error was thrown');

      return this.runLoopSettled();
    }

    [`@test Handled errors that are thrown through rejection aren't swallowed`](assert) {
      let handledError;

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error(err) {
            step(assert, 2, 'MomSallyRoute#actions.error');
            handledError = err;
            this.transitionTo('mom.this-route-throws');

            return false;
          }
        }
      }));

      this.add('route:mom.this-route-throws', _routing.Route.extend({
        model() {
          step(assert, 3, 'MomThisRouteThrows#model');
          return _runtime.RSVP.reject(handledError);
        }
      }));

      assert.throws(() => {
        this.visit('/grandma/mom/sally');
      }, function (err) {
        return err.msg === 'did it broke?';
      }, 'it broke');

      return this.runLoopSettled();
    }

    ['@test Default error events move into nested route, prioritizing more specifically named error routes - NEW'](assert) {
      this.addTemplate('grandma.error', 'ERROR: {{model.msg}}');
      this.addTemplate('mom_error', 'MOM ERROR: {{model.msg}}');

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error() {
            step(assert, 2, 'MomSallyRoute#actions.error');
            return true;
          }
        }
      }));

      return this.visit('/grandma/mom/sally').then(() => {
        step(assert, 3, 'Application finished booting');

        assert.equal(this.$('#app').text(), 'GRANDMA MOM ERROR: did it broke?', 'the more specifically named mome error substate was entered over the other error route');

        assert.equal(this.currentPath, 'grandma.mom_error', 'Initial route fully loaded');
      });
    }

    ['@test Slow promises waterfall on startup'](assert) {
      let grandmaDeferred = _runtime.RSVP.defer();
      let sallyDeferred = _runtime.RSVP.defer();

      this.addTemplate('loading', 'LOADING');
      this.addTemplate('mom', 'MOM {{outlet}}');
      this.addTemplate('mom.loading', 'MOMLOADING');
      this.addTemplate('mom.sally', 'SALLY');

      this.add('route:grandma', _routing.Route.extend({
        model() {
          step(assert, 1, 'GrandmaRoute#model');
          return grandmaDeferred.promise;
        }
      }));

      this.add('route:mom', _routing.Route.extend({
        model() {
          step(assert, 2, 'MomRoute#model');
          return {};
        }
      }));

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 3, 'SallyRoute#model');
          return sallyDeferred.promise;
        },
        setupController() {
          step(assert, 4, 'SallyRoute#setupController');
        }
      }));

      let promise = this.visit('/grandma/mom/sally').then(() => {
        text = this.$('#app').text();

        assert.equal(text, 'GRANDMA MOM SALLY', `Sally template displayed`);
      });
      let text = this.$('#app').text();

      assert.equal(text, 'LOADING', `The loading template is nested in application template's outlet`);

      this.runTask(() => grandmaDeferred.resolve());
      text = this.$('#app').text();

      assert.equal(text, 'GRANDMA MOM MOMLOADING', `Mom's child loading route is displayed due to sally's slow promise`);

      sallyDeferred.resolve();

      return promise;
    }
    ['@test Enter child loading state of pivot route'](assert) {
      let deferred = _runtime.RSVP.defer();
      this.addTemplate('grandma.loading', 'GMONEYLOADING');

      this.add('route:mom.sally', _routing.Route.extend({
        setupController() {
          step(assert, 1, 'SallyRoute#setupController');
        }
      }));

      this.add('route:grandma.puppies', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      return this.visit('/grandma/mom/sally').then(() => {
        assert.equal(this.currentPath, 'grandma.mom.sally', 'Initial route fully loaded');

        let promise = this.visit('/grandma/puppies').then(() => {
          assert.equal(this.currentPath, 'grandma.puppies', 'Finished transition');
        });

        assert.equal(this.currentPath, 'grandma.loading', `in pivot route's child loading state`);
        deferred.resolve();

        return promise;
      });
    }

    [`@test Error events that aren't bubbled don't throw application assertions`](assert) {
      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error(err) {
            step(assert, 2, 'MomSallyRoute#actions.error');
            assert.equal(err.msg, 'did it broke?', `it didn't break`);
            return false;
          }
        }
      }));

      return this.visit('/grandma/mom/sally');
    }

    ['@test Handled errors that bubble can be handled at a higher level'](assert) {
      let handledError;

      this.add('route:mom', _routing.Route.extend({
        actions: {
          error(err) {
            step(assert, 3, 'MomRoute#actions.error');
            assert.equal(err, handledError, `error handled and rebubbled is handleable at higher route`);
          }
        }
      }));

      this.add('route:mom.sally', _routing.Route.extend({
        model() {
          step(assert, 1, 'MomSallyRoute#model');
          return _runtime.RSVP.reject({
            msg: 'did it broke?'
          });
        },
        actions: {
          error(err) {
            step(assert, 2, 'MomSallyRoute#actions.error');
            handledError = err;

            return true;
          }
        }
      }));

      return this.visit('/grandma/mom/sally');
    }

    ['@test Setting a query param during a slow transition should work'](assert) {
      let deferred = _runtime.RSVP.defer();
      this.addTemplate('memere.loading', 'MMONEYLOADING');

      this.add('route:grandma', _routing.Route.extend({
        beforeModel: function () {
          this.transitionTo('memere', 1);
        }
      }));

      this.add('route:memere', _routing.Route.extend({
        queryParams: {
          test: { defaultValue: 1 }
        }
      }));

      this.add('route:memere.index', _routing.Route.extend({
        model() {
          return deferred.promise;
        }
      }));

      let promise = this.visit('/grandma').then(() => {
        assert.equal(this.currentPath, 'memere.index', 'Transition should be complete');
      });
      let memereController = this.getController('memere');

      assert.equal(this.currentPath, 'memere.loading', 'Initial route should be loading');

      memereController.set('test', 3);

      assert.equal(this.currentPath, 'memere.loading', 'Initial route should still be loading');

      assert.equal(memereController.get('test'), 3, 'Controller query param value should have changed');
      deferred.resolve();

      return promise;
    }
  });
});
enifed('ember/tests/routing/toplevel_dom_test', ['@ember/-internals/environment', 'internal-test-helpers'], function (_environment, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Top Level DOM Structure', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super(...arguments);
      this._APPLICATION_TEMPLATE_WRAPPER = _environment.ENV._APPLICATION_TEMPLATE_WRAPPER;
    }

    teardown() {
      super.teardown();
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = this._APPLICATION_TEMPLATE_WRAPPER;
    }

    ['@test topmost template with wrapper']() {
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = true;

      this.addTemplate('application', 'hello world');

      return this.visit('/').then(() => {
        this.assertComponentElement(this.element, { content: 'hello world' });
      });
    }

    ['@test topmost template without wrapper']() {
      _environment.ENV._APPLICATION_TEMPLATE_WRAPPER = false;

      this.addTemplate('application', 'hello world');

      return this.visit('/').then(() => {
        this.assertInnerHTML('hello world');
      });
    }
  });
});
enifed('ember/tests/service_injection_test', ['@ember/-internals/owner', '@ember/controller', '@ember/service', '@ember/-internals/runtime', 'internal-test-helpers', '@ember/-internals/metal', '@ember/canary-features'], function (_owner, _controller, _service, _runtime, _internalTestHelpers, _metal, _canaryFeatures) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('Service Injection', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Service can be injected and is resolved'](assert) {
      this.add('controller:application', _controller.default.extend({
        myService: (0, _service.inject)('my-service')
      }));
      let MyService = _service.default.extend();
      this.add('service:my-service', MyService);
      this.addTemplate('application', '');

      this.visit('/').then(() => {
        let controller = this.applicationInstance.lookup('controller:application');
        assert.ok(controller.get('myService') instanceof MyService);
      });
    }

    ['@test Service can be an object proxy and access owner in init GH#16484'](assert) {
      let serviceOwner;

      this.add('controller:application', _controller.default.extend({
        myService: (0, _service.inject)('my-service')
      }));
      let MyService = _service.default.extend(_runtime._ProxyMixin, {
        init() {
          this._super(...arguments);

          serviceOwner = (0, _owner.getOwner)(this);
        }
      });
      this.add('service:my-service', MyService);
      this.addTemplate('application', '');

      this.visit('/').then(instance => {
        let controller = this.applicationInstance.lookup('controller:application');
        assert.ok(controller.get('myService') instanceof MyService);
        assert.equal(serviceOwner, instance, 'should be able to `getOwner` in init');
      });
    }
  });

  (0, _internalTestHelpers.moduleFor)('Service Injection with ES5 Getters', class extends _internalTestHelpers.ApplicationTestCase {
    ['@test Service can be injected and is resolved without calling `get`'](assert) {
      this.add('controller:application', _controller.default.extend({
        myService: (0, _service.inject)('my-service')
      }));
      let MyService = _service.default.extend({
        name: (0, _metal.computed)(function () {
          return 'The service name';
        })
      });
      this.add('service:my-service', MyService);
      this.addTemplate('application', '');

      this.visit('/').then(() => {
        let controller = this.applicationInstance.lookup('controller:application');
        assert.ok(controller.myService instanceof MyService);
        assert.equal(controller.myService.name, 'The service name', 'service property accessible');
      });
    }
  });

  if (_canaryFeatures.EMBER_MODULE_UNIFICATION) {
    (0, _internalTestHelpers.moduleFor)('Service Injection (MU)', class extends _internalTestHelpers.ApplicationTestCase {
      ['@test Service can be injected with source and is resolved'](assert) {
        let source = 'controller:src/ui/routes/application/controller';
        this.add('controller:application', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source })
        }));
        let MyService = _service.default.extend();
        this.add({
          specifier: 'service:my-service',
          source
        }, MyService);

        return this.visit('/').then(() => {
          let controller = this.applicationInstance.lookup('controller:application');

          assert.ok(controller.get('myService') instanceof MyService);
        });
      }

      ['@test Services can be injected with same name, different source, and resolve different instances'](assert) {
        // This test implies that there is a file src/ui/routes/route-a/-services/my-service
        let routeASource = 'controller:src/ui/routes/route-a/controller';
        // This test implies that there is a file src/ui/routes/route-b/-services/my-service
        let routeBSource = 'controller:src/ui/routes/route-b/controller';

        this.add('controller:route-a', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source: routeASource })
        }));

        this.add('controller:route-b', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source: routeBSource })
        }));

        let LocalLookupService = _service.default.extend();
        this.add({
          specifier: 'service:my-service',
          source: routeASource
        }, LocalLookupService);

        let MyService = _service.default.extend();
        this.add({
          specifier: 'service:my-service',
          source: routeBSource
        }, MyService);

        return this.visit('/').then(() => {
          let controllerA = this.applicationInstance.lookup('controller:route-a');
          let serviceFromControllerA = controllerA.get('myService');
          assert.ok(serviceFromControllerA instanceof LocalLookupService, 'local lookup service is returned');

          let controllerB = this.applicationInstance.lookup('controller:route-b');
          let serviceFromControllerB = controllerB.get('myService');
          assert.ok(serviceFromControllerB instanceof MyService, 'global service is returned');

          assert.notStrictEqual(serviceFromControllerA, serviceFromControllerB);
        });
      }

      ['@test Services can be injected with same name, different source, but same resolution result, and share an instance'](assert) {
        let routeASource = 'controller:src/ui/routes/route-a/controller';
        let routeBSource = 'controller:src/ui/routes/route-b/controller';

        this.add('controller:route-a', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source: routeASource })
        }));

        this.add('controller:route-b', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source: routeBSource })
        }));

        let MyService = _service.default.extend();
        this.add({
          specifier: 'service:my-service'
        }, MyService);

        return this.visit('/').then(() => {
          let controllerA = this.applicationInstance.lookup('controller:route-a');
          let serviceFromControllerA = controllerA.get('myService');
          assert.ok(serviceFromControllerA instanceof MyService);

          let controllerB = this.applicationInstance.lookup('controller:route-b');
          assert.strictEqual(serviceFromControllerA, controllerB.get('myService'));
        });
      }

      /*
      * This test demonstrates a failure in the caching system of ember's
      * container around singletons and and local lookup. The local lookup
      * is cached and the global injection is then looked up incorrectly.
      *
      * The paractical rules of Ember's module unification config are such
      * that services cannot be locally looked up, thus this case is really
      * just a demonstration of what could go wrong if we permit arbitrary
      * configuration (such as a singleton type that has local lookup).
      */
      ['@test Services can be injected with same name, one with source one without, and share an instance'](assert) {
        let routeASource = 'controller:src/ui/routes/route-a/controller';
        this.add('controller:route-a', _controller.default.extend({
          myService: (0, _service.inject)('my-service', { source: routeASource })
        }));

        this.add('controller:route-b', _controller.default.extend({
          myService: (0, _service.inject)('my-service')
        }));

        let MyService = _service.default.extend();
        this.add({
          specifier: 'service:my-service'
        }, MyService);

        return this.visit('/').then(() => {
          let controllerA = this.applicationInstance.lookup('controller:route-a');
          let serviceFromControllerA = controllerA.get('myService');
          assert.ok(serviceFromControllerA instanceof MyService, 'global service is returned');

          let controllerB = this.applicationInstance.lookup('controller:route-b');
          let serviceFromControllerB = controllerB.get('myService');
          assert.ok(serviceFromControllerB instanceof MyService, 'global service is returned');

          assert.strictEqual(serviceFromControllerA, serviceFromControllerB);
        });
      }

      ['@test Service with namespace can be injected and is resolved'](assert) {
        this.add('controller:application', _controller.default.extend({
          myService: (0, _service.inject)('my-namespace::my-service')
        }));
        let MyService = _service.default.extend();
        this.add({
          specifier: 'service:my-service',
          namespace: 'my-namespace'
        }, MyService);

        this.visit('/').then(() => {
          let controller = this.applicationInstance.lookup('controller:application');
          assert.ok(controller.get('myService') instanceof MyService);
        });
      }
    });
  }
});
enifed('ember/tests/view_instrumentation_test', ['@ember/instrumentation', 'internal-test-helpers'], function (_instrumentation, _internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('View Instrumentation', class extends _internalTestHelpers.ApplicationTestCase {
    constructor() {
      super();
      this.addTemplate('application', `{{outlet}}`);
      this.addTemplate('index', `<h1>Index</h1>`);
      this.addTemplate('posts', `<h1>Posts</h1>`);

      this.router.map(function () {
        this.route('posts');
      });
    }
    teardown() {
      (0, _instrumentation.reset)();
      super.teardown();
    }

    ['@test Nodes without view instances are instrumented'](assert) {
      let called = false;

      (0, _instrumentation.subscribe)('render', {
        before() {
          called = true;
        },
        after() {}
      });

      return this.visit('/').then(() => {
        assert.equal(this.textValue(), 'Index', 'It rendered the correct template');

        assert.ok(called, 'Instrumentation called on first render');
        called = false;

        return this.visit('/posts');
      }).then(() => {
        assert.equal(this.textValue(), 'Posts', 'It rendered the correct template');
        assert.ok(called, 'Instrumentation called on transition to non-view backed route');
      });
    }
  });
});
enifed('internal-test-helpers/index', ['exports', 'internal-test-helpers/lib/factory', 'internal-test-helpers/lib/build-owner', 'internal-test-helpers/lib/confirm-export', 'internal-test-helpers/lib/equal-inner-html', 'internal-test-helpers/lib/equal-tokens', 'internal-test-helpers/lib/module-for', 'internal-test-helpers/lib/strip', 'internal-test-helpers/lib/apply-mixins', 'internal-test-helpers/lib/get-text-of', 'internal-test-helpers/lib/matchers', 'internal-test-helpers/lib/run', 'internal-test-helpers/lib/test-cases/abstract', 'internal-test-helpers/lib/test-cases/abstract-application', 'internal-test-helpers/lib/test-cases/application', 'internal-test-helpers/lib/test-cases/query-param', 'internal-test-helpers/lib/test-cases/abstract-rendering', 'internal-test-helpers/lib/test-cases/rendering', 'internal-test-helpers/lib/test-cases/router', 'internal-test-helpers/lib/test-cases/autoboot-application', 'internal-test-helpers/lib/test-cases/default-resolver-application', 'internal-test-helpers/lib/test-resolver', 'internal-test-helpers/lib/browser-detect', 'internal-test-helpers/lib/registry-check'], function (exports, _factory, _buildOwner, _confirmExport, _equalInnerHtml, _equalTokens, _moduleFor, _strip, _applyMixins, _getTextOf, _matchers, _run, _abstract, _abstractApplication, _application, _queryParam, _abstractRendering, _rendering, _router, _autobootApplication, _defaultResolverApplication, _testResolver, _browserDetect, _registryCheck) {
  'use strict';

  Object.defineProperty(exports, 'factory', {
    enumerable: true,
    get: function () {
      return _factory.default;
    }
  });
  Object.defineProperty(exports, 'buildOwner', {
    enumerable: true,
    get: function () {
      return _buildOwner.default;
    }
  });
  Object.defineProperty(exports, 'confirmExport', {
    enumerable: true,
    get: function () {
      return _confirmExport.default;
    }
  });
  Object.defineProperty(exports, 'equalInnerHTML', {
    enumerable: true,
    get: function () {
      return _equalInnerHtml.default;
    }
  });
  Object.defineProperty(exports, 'equalTokens', {
    enumerable: true,
    get: function () {
      return _equalTokens.default;
    }
  });
  Object.defineProperty(exports, 'moduleFor', {
    enumerable: true,
    get: function () {
      return _moduleFor.default;
    }
  });
  Object.defineProperty(exports, 'strip', {
    enumerable: true,
    get: function () {
      return _strip.default;
    }
  });
  Object.defineProperty(exports, 'applyMixins', {
    enumerable: true,
    get: function () {
      return _applyMixins.default;
    }
  });
  Object.defineProperty(exports, 'getTextOf', {
    enumerable: true,
    get: function () {
      return _getTextOf.default;
    }
  });
  Object.defineProperty(exports, 'equalsElement', {
    enumerable: true,
    get: function () {
      return _matchers.equalsElement;
    }
  });
  Object.defineProperty(exports, 'classes', {
    enumerable: true,
    get: function () {
      return _matchers.classes;
    }
  });
  Object.defineProperty(exports, 'styles', {
    enumerable: true,
    get: function () {
      return _matchers.styles;
    }
  });
  Object.defineProperty(exports, 'regex', {
    enumerable: true,
    get: function () {
      return _matchers.regex;
    }
  });
  Object.defineProperty(exports, 'runAppend', {
    enumerable: true,
    get: function () {
      return _run.runAppend;
    }
  });
  Object.defineProperty(exports, 'runDestroy', {
    enumerable: true,
    get: function () {
      return _run.runDestroy;
    }
  });
  Object.defineProperty(exports, 'AbstractTestCase', {
    enumerable: true,
    get: function () {
      return _abstract.default;
    }
  });
  Object.defineProperty(exports, 'AbstractApplicationTestCase', {
    enumerable: true,
    get: function () {
      return _abstractApplication.default;
    }
  });
  Object.defineProperty(exports, 'ApplicationTestCase', {
    enumerable: true,
    get: function () {
      return _application.default;
    }
  });
  Object.defineProperty(exports, 'QueryParamTestCase', {
    enumerable: true,
    get: function () {
      return _queryParam.default;
    }
  });
  Object.defineProperty(exports, 'AbstractRenderingTestCase', {
    enumerable: true,
    get: function () {
      return _abstractRendering.default;
    }
  });
  Object.defineProperty(exports, 'RenderingTestCase', {
    enumerable: true,
    get: function () {
      return _rendering.default;
    }
  });
  Object.defineProperty(exports, 'RouterTestCase', {
    enumerable: true,
    get: function () {
      return _router.default;
    }
  });
  Object.defineProperty(exports, 'AutobootApplicationTestCase', {
    enumerable: true,
    get: function () {
      return _autobootApplication.default;
    }
  });
  Object.defineProperty(exports, 'DefaultResolverApplicationTestCase', {
    enumerable: true,
    get: function () {
      return _defaultResolverApplication.default;
    }
  });
  Object.defineProperty(exports, 'TestResolver', {
    enumerable: true,
    get: function () {
      return _testResolver.default;
    }
  });
  Object.defineProperty(exports, 'ModuleBasedTestResolver', {
    enumerable: true,
    get: function () {
      return _testResolver.ModuleBasedResolver;
    }
  });
  Object.defineProperty(exports, 'isIE11', {
    enumerable: true,
    get: function () {
      return _browserDetect.isIE11;
    }
  });
  Object.defineProperty(exports, 'isEdge', {
    enumerable: true,
    get: function () {
      return _browserDetect.isEdge;
    }
  });
  Object.defineProperty(exports, 'verifyInjection', {
    enumerable: true,
    get: function () {
      return _registryCheck.verifyInjection;
    }
  });
  Object.defineProperty(exports, 'verifyRegistration', {
    enumerable: true,
    get: function () {
      return _registryCheck.verifyRegistration;
    }
  });
});
enifed('internal-test-helpers/lib/apply-mixins', ['exports', '@ember/polyfills', 'internal-test-helpers/lib/get-all-property-names'], function (exports, _polyfills, _getAllPropertyNames) {
  'use strict';

  exports.default = applyMixins;


  function isGenerator(mixin) {
    return Array.isArray(mixin.cases) && typeof mixin.generate === 'function';
  }

  function applyMixins(TestClass, ...mixins) {
    mixins.forEach(mixinOrGenerator => {
      let mixin;

      if (isGenerator(mixinOrGenerator)) {
        let generator = mixinOrGenerator;
        mixin = {};

        generator.cases.forEach((value, idx) => {
          (0, _polyfills.assign)(mixin, generator.generate(value, idx));
        });

        (0, _polyfills.assign)(TestClass.prototype, mixin);
      } else if (typeof mixinOrGenerator === 'function') {
        let properties = (0, _getAllPropertyNames.default)(mixinOrGenerator);
        mixin = new mixinOrGenerator();

        properties.forEach(name => {
          TestClass.prototype[name] = function () {
            return mixin[name].apply(mixin, arguments);
          };
        });
      } else {
        mixin = mixinOrGenerator;
        (0, _polyfills.assign)(TestClass.prototype, mixin);
      }
    });

    return TestClass;
  }
});
enifed('internal-test-helpers/lib/browser-detect', ['exports'], function (exports) {
  'use strict';

  // `window.ActiveXObject` is "falsey" in IE11 (but not `undefined` or `false`)
  // `"ActiveXObject" in window` returns `true` in all IE versions
  // only IE11 will pass _both_ of these conditions
  const isIE11 = exports.isIE11 = !window.ActiveXObject && 'ActiveXObject' in window;
  const isEdge = exports.isEdge = /Edge/.test(navigator.userAgent);
});
enifed('internal-test-helpers/lib/build-owner', ['exports', '@ember/-internals/container', '@ember/-internals/routing', '@ember/application/instance', '@ember/application', '@ember/-internals/runtime'], function (exports, _container, _routing, _instance, _application, _runtime) {
  'use strict';

  exports.default = buildOwner;


  class ResolverWrapper {
    constructor(resolver) {
      this.resolver = resolver;
    }

    create() {
      return this.resolver;
    }
  }

  function buildOwner(options = {}) {
    let ownerOptions = options.ownerOptions || {};
    let resolver = options.resolver;
    let bootOptions = options.bootOptions || {};

    let Owner = _runtime.Object.extend(_runtime.RegistryProxyMixin, _runtime.ContainerProxyMixin);

    let namespace = _runtime.Object.create({
      Resolver: new ResolverWrapper(resolver)
    });

    let fallbackRegistry = _application.default.buildRegistry(namespace);
    fallbackRegistry.register('router:main', _routing.Router);

    let registry = new _container.Registry({
      fallback: fallbackRegistry
    });

    _instance.default.setupRegistry(registry, bootOptions);

    let owner = Owner.create({
      __registry__: registry,
      __container__: null
    }, ownerOptions);

    let container = registry.container({ owner });
    owner.__container__ = container;

    return owner;
  }
});
enifed('internal-test-helpers/lib/confirm-export', ['exports', 'require'], function (exports, _require2) {
  'use strict';

  exports.default = confirmExport;


  function getDescriptor(obj, path) {
    let parts = path.split('.');
    let value = obj;
    for (let i = 0; i < parts.length - 1; i++) {
      let part = parts[i];
      value = value[part];
      if (!value) {
        return undefined;
      }
    }
    let last = parts[parts.length - 1];
    return Object.getOwnPropertyDescriptor(value, last);
  }

  function confirmExport(Ember, assert, path, moduleId, exportName) {
    try {
      let desc = getDescriptor(Ember, path);
      assert.ok(desc, `the ${path} property exists on the Ember global`);

      if (typeof exportName === 'string') {
        let mod = (0, _require2.default)(moduleId);
        assert.equal(desc.value, mod[exportName], `Ember.${path} is exported correctly`);
        assert.notEqual(mod[exportName], undefined, `Ember.${path} is not \`undefined\``);
      } else if ('value' in desc) {
        assert.equal(desc.value, exportName.value, `Ember.${path} is exported correctly`);
      } else {
        let mod = (0, _require2.default)(moduleId);
        assert.equal(desc.get, mod[exportName.get], `Ember.${path} getter is exported correctly`);
        assert.notEqual(desc.get, undefined, `Ember.${path} getter is not undefined`);

        if (exportName.set) {
          assert.equal(desc.set, mod[exportName.set], `Ember.${path} setter is exported correctly`);
          assert.notEqual(desc.set, undefined, `Ember.${path} setter is not undefined`);
        }
      }
    } catch (error) {
      assert.pushResult({
        result: false,
        message: `An error occured while testing ${path} is exported from ${moduleId}.`,
        source: error
      });
    }
  }
});
enifed('internal-test-helpers/lib/ember-dev/assertion', ['exports', 'internal-test-helpers/lib/ember-dev/utils'], function (exports, _utils) {
    'use strict';

    const BREAK = {};
    /*
      This assertion class is used to test assertions made using Ember.assert.
      It injects two helpers onto `window`:
    
      - expectAssertion(func: Function, [expectedMessage: String | RegExp])
    
      This function calls `func` and asserts that `Ember.assert` is invoked during
      the execution. Moreover, it takes a String or a RegExp as a second optional
      argument that can be used to test if a specific assertion message was
      generated.
    
      - ignoreAssertion(func: Function)
    
      This function calls `func` and disables `Ember.assert` during the execution.
      In particular, this prevents `Ember.assert` from throw errors that would
      disrupt the control flow.
    */
    class AssertionAssert {
        constructor(env) {
            this.env = env;
        }
        reset() {}
        assert() {}
        inject() {
            let expectAssertion = (func, expectedMessage) => {
                let { assert } = QUnit.config.current;
                if (this.env.runningProdBuild) {
                    assert.ok(true, 'Assertions disabled in production builds.');
                    return;
                }
                let sawCall = false;
                let actualMessage = undefined;
                // The try-catch statement is used to "exit" `func` as soon as
                // the first useful assertion has been produced.
                try {
                    (0, _utils.callWithStub)(this.env, 'assert', func, (message, test) => {
                        sawCall = true;
                        if ((0, _utils.checkTest)(test)) {
                            return;
                        }
                        actualMessage = message;
                        throw BREAK;
                    });
                } catch (e) {
                    if (e !== BREAK) {
                        throw e;
                    }
                }
                check(assert, sawCall, actualMessage, expectedMessage);
            };
            let ignoreAssertion = func => {
                (0, _utils.callWithStub)(this.env, 'assert', func);
            };
            window.expectAssertion = expectAssertion;
            window.ignoreAssertion = ignoreAssertion;
        }
        restore() {
            window.expectAssertion = null;
            window.ignoreAssertion = null;
        }
    }
    exports.default = AssertionAssert;
    function check(assert, sawCall, actualMessage, expectedMessage) {
        // Run assertions in an order that is useful when debugging a test failure.
        if (!sawCall) {
            assert.ok(false, `Expected Ember.assert to be called (Not called with any value).`);
        } else if (!actualMessage) {
            assert.ok(false, `Expected a failing Ember.assert (Ember.assert called, but without a failing test).`);
        } else {
            if (expectedMessage) {
                if (expectedMessage instanceof RegExp) {
                    assert.ok(expectedMessage.test(actualMessage), `Expected failing Ember.assert: '${expectedMessage}', but got '${actualMessage}'.`);
                } else {
                    assert.equal(actualMessage, expectedMessage, `Expected failing Ember.assert: '${expectedMessage}', but got '${actualMessage}'.`);
                }
            } else {
                // Positive assertion that assert was called
                assert.ok(true, 'Expected a failing Ember.assert.');
            }
        }
    }
});
enifed('internal-test-helpers/lib/ember-dev/containers', ['exports', '@ember/-internals/container'], function (exports, _container) {
  'use strict';

  function ContainersAssert(env) {
    this.env = env;
  }

  const { _leakTracking: containerLeakTracking } = _container.Container;

  ContainersAssert.prototype = {
    reset: function () {},
    inject: function () {},
    assert: function () {
      if (containerLeakTracking === undefined) return;
      let { config } = QUnit;
      let { testName, testId, module: { name: moduleName }, finish: originalFinish } = config.current;
      config.current.finish = function () {
        originalFinish.call(this);
        originalFinish = undefined;
        config.queue.unshift(function () {
          if (containerLeakTracking.hasContainers()) {
            containerLeakTracking.reset();
            // eslint-disable-next-line no-console
            console.assert(false, `Leaked container after test ${moduleName}: ${testName} testId=${testId}`);
          }
        });
      };
    },
    restore: function () {}
  };

  exports.default = ContainersAssert;
});
enifed('internal-test-helpers/lib/ember-dev/debug', ['exports', 'internal-test-helpers/lib/ember-dev/method-call-tracker'], function (exports, _methodCallTracker) {
    'use strict';

    class DebugAssert {
        constructor(methodName, env) {
            this.methodName = methodName;
            this.env = env;
            this.tracker = null;
        }
        inject() {}
        restore() {
            this.reset();
        }
        reset() {
            if (this.tracker) {
                this.tracker.restoreMethod();
            }
            this.tracker = null;
        }
        assert() {
            if (this.tracker) {
                this.tracker.assert();
            }
        }
        // Run an expectation callback within the context of a new tracker, optionally
        // accepting a function to run, which asserts immediately
        runExpectation(func, callback) {
            let originalTracker = null;
            // When helpers are passed a callback, they get a new tracker context
            if (func) {
                originalTracker = this.tracker;
                this.tracker = null;
            }
            if (!this.tracker) {
                this.tracker = new _methodCallTracker.default(this.env, this.methodName);
            }
            // Yield to caller with tracker instance
            callback(this.tracker);
            // Once the given callback is invoked, the pending assertions should be
            // flushed immediately
            if (func) {
                func();
                this.assert();
                this.reset();
                this.tracker = originalTracker;
            }
        }
    }
    exports.default = DebugAssert;
});
enifed('internal-test-helpers/lib/ember-dev/deprecation', ['exports', 'internal-test-helpers/lib/ember-dev/debug', 'internal-test-helpers/lib/ember-dev/utils'], function (exports, _debug, _utils) {
    'use strict';

    class DeprecationAssert extends _debug.default {
        constructor(env) {
            super('deprecate', env);
        }
        inject() {
            // Expects no deprecation to happen within a function, or if no function is
            // passed, from the time of calling until the end of the test.
            //
            // expectNoDeprecation(function() {
            //   fancyNewThing();
            // });
            //
            // expectNoDeprecation();
            // Ember.deprecate("Old And Busted");
            //
            let expectNoDeprecation = func => {
                if (typeof func !== 'function') {
                    func = undefined;
                }
                this.runExpectation(func, tracker => {
                    if (tracker.isExpectingCalls()) {
                        throw new Error('expectNoDeprecation was called after expectDeprecation was called!');
                    }
                    tracker.expectNoCalls();
                });
            };
            // Expect a deprecation to happen within a function, or if no function
            // is pass, from the time of calling until the end of the test. Can be called
            // multiple times to assert deprecations with different specific messages
            // were fired.
            //
            // expectDeprecation(function() {
            //   Ember.deprecate("Old And Busted");
            // }, /* optionalStringOrRegex */);
            //
            // expectDeprecation(/* optionalStringOrRegex */);
            // Ember.deprecate("Old And Busted");
            //
            let expectDeprecation = (func, message) => {
                let actualFunc;
                if (typeof func !== 'function') {
                    message = func;
                    actualFunc = undefined;
                } else {
                    actualFunc = func;
                }
                this.runExpectation(actualFunc, tracker => {
                    if (tracker.isExpectingNoCalls()) {
                        throw new Error('expectDeprecation was called after expectNoDeprecation was called!');
                    }
                    tracker.expectCall(message, ['id', 'until']);
                });
            };
            let ignoreDeprecation = func => {
                (0, _utils.callWithStub)(this.env, 'deprecate', func);
            };
            window.expectNoDeprecation = expectNoDeprecation;
            window.expectDeprecation = expectDeprecation;
            window.ignoreDeprecation = ignoreDeprecation;
        }
        restore() {
            super.restore();
            window.expectDeprecation = null;
            window.expectNoDeprecation = null;
            window.ignoreDeprecation = null;
        }
    }
    exports.default = DeprecationAssert;
});
enifed('internal-test-helpers/lib/ember-dev/index', ['exports', 'internal-test-helpers/lib/ember-dev/deprecation', 'internal-test-helpers/lib/ember-dev/warning', 'internal-test-helpers/lib/ember-dev/assertion', 'internal-test-helpers/lib/ember-dev/run-loop', 'internal-test-helpers/lib/ember-dev/namespaces', 'internal-test-helpers/lib/ember-dev/containers', 'internal-test-helpers/lib/ember-dev/utils'], function (exports, _deprecation, _warning, _assertion, _runLoop, _namespaces, _containers, _utils) {
  'use strict';

  var EmberDevTestHelperAssert = (0, _utils.buildCompositeAssert)([_deprecation.default, _warning.default, _assertion.default, _runLoop.default, _namespaces.default, _containers.default]);

  exports.default = EmberDevTestHelperAssert;
});
enifed('internal-test-helpers/lib/ember-dev/method-call-tracker', ['exports', 'internal-test-helpers/lib/ember-dev/utils'], function (exports, _utils) {
    'use strict';

    class MethodCallTracker {
        constructor(env, methodName) {
            this._env = env;
            this._methodName = methodName;
            this._isExpectingNoCalls = false;
            this._expectedMessages = [];
            this._expectedOptionLists = [];
            this._actuals = [];
            this._originalMethod = undefined;
        }
        stubMethod() {
            if (this._originalMethod) {
                // Method is already stubbed
                return;
            }
            let env = this._env;
            let methodName = this._methodName;
            this._originalMethod = env.getDebugFunction(methodName);
            env.setDebugFunction(methodName, (message, test, options) => {
                let resultOfTest = (0, _utils.checkTest)(test);
                this._actuals.push([message, resultOfTest, options]);
            });
        }
        restoreMethod() {
            if (this._originalMethod) {
                this._env.setDebugFunction(this._methodName, this._originalMethod);
            }
        }
        expectCall(message, options) {
            this.stubMethod();
            this._expectedMessages.push(message || /.*/);
            this._expectedOptionLists.push(options);
        }
        expectNoCalls() {
            this.stubMethod();
            this._isExpectingNoCalls = true;
        }
        isExpectingNoCalls() {
            return this._isExpectingNoCalls;
        }
        isExpectingCalls() {
            return !this._isExpectingNoCalls && this._expectedMessages.length;
        }
        assert() {
            let { assert } = QUnit.config.current;
            let env = this._env;
            let methodName = this._methodName;
            let isExpectingNoCalls = this._isExpectingNoCalls;
            let expectedMessages = this._expectedMessages;
            let expectedOptionLists = this._expectedOptionLists;
            let actuals = this._actuals;
            let o, i, j;
            if (!isExpectingNoCalls && expectedMessages.length === 0 && actuals.length === 0) {
                return;
            }
            if (env.runningProdBuild) {
                assert.ok(true, `calls to Ember.${methodName} disabled in production builds.`);
                return;
            }
            if (isExpectingNoCalls) {
                let actualMessages = [];
                for (i = 0; i < actuals.length; i++) {
                    if (!actuals[i][1]) {
                        actualMessages.push(actuals[i][0]);
                    }
                }
                assert.ok(actualMessages.length === 0, `Expected no Ember.${methodName} calls, got ${actuals.length}: ${actualMessages.join(', ')}`);
                return;
            }
            let actual;
            let match = undefined;
            for (o = 0; o < expectedMessages.length; o++) {
                const expectedMessage = expectedMessages[o];
                const expectedOptionList = expectedOptionLists[o];
                for (i = 0; i < actuals.length; i++) {
                    let matchesMessage = false;
                    let matchesOptionList = false;
                    actual = actuals[i];
                    if (actual[1] === true) {
                        continue;
                    }
                    if (expectedMessage instanceof RegExp && expectedMessage.test(actual[0])) {
                        matchesMessage = true;
                    } else if (expectedMessage === actual[0]) {
                        matchesMessage = true;
                    }
                    if (expectedOptionList === undefined) {
                        matchesOptionList = true;
                    } else if (actual[2]) {
                        matchesOptionList = true;
                        for (j = 0; j < expectedOptionList.length; j++) {
                            matchesOptionList = matchesOptionList && actual[2].hasOwnProperty(expectedOptionList[j]);
                        }
                    }
                    if (matchesMessage && matchesOptionList) {
                        match = actual;
                        break;
                    }
                }
                const expectedOptionsMessage = expectedOptionList ? `and options: { ${expectedOptionList.join(', ')} }` : 'and no options';
                const actualOptionsMessage = actual && actual[2] ? `and options: { ${Object.keys(actual[2]).join(', ')} }` : 'and no options';
                if (!actual) {
                    assert.ok(false, `Received no Ember.${methodName} calls at all, expecting: ${expectedMessage}`);
                } else if (match && !match[1]) {
                    assert.ok(true, `Received failing Ember.${methodName} call with message: ${match[0]}`);
                } else if (match && match[1]) {
                    assert.ok(false, `Expected failing Ember.${methodName} call, got succeeding with message: ${match[0]}`);
                } else if (actual[1]) {
                    assert.ok(false, `Did not receive failing Ember.${methodName} call matching '${expectedMessage}' ${expectedOptionsMessage}, last was success with '${actual[0]}' ${actualOptionsMessage}`);
                } else if (!actual[1]) {
                    assert.ok(false, `Did not receive failing Ember.${methodName} call matching '${expectedMessage}' ${expectedOptionsMessage}, last was failure with '${actual[0]}' ${actualOptionsMessage}`);
                }
            }
        }
    }
    exports.default = MethodCallTracker;
});
enifed('internal-test-helpers/lib/ember-dev/namespaces', ['exports', '@ember/runloop', '@ember/-internals/metal'], function (exports, _runloop, _metal) {
  'use strict';

  function NamespacesAssert(env) {
    this.env = env;
  }

  NamespacesAssert.prototype = {
    reset: function () {},
    inject: function () {},
    assert: function () {
      let { assert } = QUnit.config.current;

      if (_metal.NAMESPACES.length > 0) {
        assert.ok(false, 'Should not have any NAMESPACES after tests');
        (0, _runloop.run)(() => {
          let namespaces = _metal.NAMESPACES.slice();
          for (let i = 0; i < namespaces.length; i++) {
            namespaces[i].destroy();
          }
        });
      }
      let keys = Object.keys(_metal.NAMESPACES_BY_ID);
      if (keys.length > 0) {
        assert.ok(false, 'Should not have any NAMESPACES_BY_ID after tests');
        for (let i = 0; i < keys.length; i++) {
          delete _metal.NAMESPACES_BY_ID[keys[i]];
        }
      }
    },
    restore: function () {}
  };

  exports.default = NamespacesAssert;
});
enifed('internal-test-helpers/lib/ember-dev/run-loop', ['exports', '@ember/runloop'], function (exports, _runloop) {
  'use strict';

  function RunLoopAssertion(env) {
    this.env = env;
  }

  RunLoopAssertion.prototype = {
    reset: function () {},
    inject: function () {},
    assert: function () {
      let { assert } = QUnit.config.current;

      if ((0, _runloop.getCurrentRunLoop)()) {
        assert.ok(false, 'Should not be in a run loop at end of test');
        while ((0, _runloop.getCurrentRunLoop)()) {
          (0, _runloop.end)();
        }
      }

      if ((0, _runloop.hasScheduledTimers)()) {
        assert.ok(false, 'Ember run should not have scheduled timers at end of test');
        (0, _runloop.cancelTimers)();
      }
    },
    restore: function () {}
  };

  exports.default = RunLoopAssertion;
});
enifed("internal-test-helpers/lib/ember-dev/setup-qunit", ["exports"], function (exports) {
    "use strict";

    exports.default = setupQUnit;
    function setupQUnit(assertion, _qunitGlobal) {
        let qunitGlobal = QUnit;
        if (_qunitGlobal) {
            qunitGlobal = _qunitGlobal;
        }
        let originalModule = qunitGlobal.module;
        qunitGlobal.module = function (name, _options) {
            let options = _options || {};
            let originalSetup = options.setup || options.beforeEach || function () {};
            let originalTeardown = options.teardown || options.afterEach || function () {};
            delete options.setup;
            delete options.teardown;
            options.beforeEach = function () {
                assertion.reset();
                assertion.inject();
                return originalSetup.apply(this, arguments);
            };
            options.afterEach = function () {
                let result = originalTeardown.apply(this, arguments);
                assertion.assert();
                assertion.restore();
                return result;
            };
            return originalModule(name, options);
        };
    }
});
enifed('internal-test-helpers/lib/ember-dev/utils', ['exports'], function (exports) {
    'use strict';

    exports.buildCompositeAssert = buildCompositeAssert;
    exports.callWithStub = callWithStub;
    exports.checkTest = checkTest;
    function callForEach(prop, func) {
        return function () {
            for (let i = 0, l = this[prop].length; i < l; i++) {
                this[prop][i][func]();
            }
        };
    }
    function buildCompositeAssert(assertClasses) {
        function Composite(env) {
            this.asserts = assertClasses.map(Assert => new Assert(env));
        }
        Composite.prototype = {
            reset: callForEach('asserts', 'reset'),
            inject: callForEach('asserts', 'inject'),
            assert: callForEach('asserts', 'assert'),
            restore: callForEach('asserts', 'restore')
        };
        return Composite;
    }
    function noop() {}
    function callWithStub(env, name, func, debugStub = noop) {
        let originalFunc = env.getDebugFunction(name);
        try {
            env.setDebugFunction(name, debugStub);
            func();
        } finally {
            env.setDebugFunction(name, originalFunc);
        }
    }
    function checkTest(test) {
        return typeof test === 'function' ? test() : test;
    }
});
enifed('internal-test-helpers/lib/ember-dev/warning', ['exports', 'internal-test-helpers/lib/ember-dev/debug', 'internal-test-helpers/lib/ember-dev/utils'], function (exports, _debug, _utils) {
    'use strict';

    class WarningAssert extends _debug.default {
        constructor(env) {
            super('warn', env);
        }
        inject() {
            // Expects no warning to happen within a function, or if no function is
            // passed, from the time of calling until the end of the test.
            //
            // expectNoWarning(function() {
            //   fancyNewThing();
            // });
            //
            // expectNoWarning();
            // Ember.warn("Oh snap, didn't expect that");
            //
            let expectNoWarning = func => {
                if (typeof func !== 'function') {
                    func = undefined;
                }
                this.runExpectation(func, tracker => {
                    if (tracker.isExpectingCalls()) {
                        throw new Error('expectNoWarning was called after expectWarning was called!');
                    }
                    tracker.expectNoCalls();
                });
            };
            // Expect a warning to happen within a function, or if no function is
            // passed, from the time of calling until the end of the test. Can be called
            // multiple times to assert warnings with different specific messages
            // happened.
            //
            // expectWarning(function() {
            //   Ember.warn("Times they are a-changin'");
            // }, /* optionalStringOrRegex */);
            //
            // expectWarning(/* optionalStringOrRegex */);
            // Ember.warn("Times definitely be changin'");
            //
            let expectWarning = (func, message) => {
                let actualFunc;
                if (typeof func !== 'function') {
                    message = func;
                    actualFunc = undefined;
                } else {
                    actualFunc = func;
                }
                this.runExpectation(actualFunc, tracker => {
                    if (tracker.isExpectingNoCalls()) {
                        throw new Error('expectWarning was called after expectNoWarning was called!');
                    }
                    tracker.expectCall(message);
                });
            };
            let ignoreWarning = func => {
                (0, _utils.callWithStub)(this.env, 'warn', func);
            };
            window.expectNoWarning = expectNoWarning;
            window.expectWarning = expectWarning;
            window.ignoreWarning = ignoreWarning;
        }
        restore() {
            super.restore();
            window.expectWarning = null;
            window.expectNoWarning = null;
            window.ignoreWarning = null;
        }
    }
    exports.default = WarningAssert;
});
enifed('internal-test-helpers/lib/equal-inner-html', ['exports'], function (exports) {
  'use strict';

  exports.default = equalInnerHTML;
  // detect side-effects of cloning svg elements in IE9-11
  let ieSVGInnerHTML = (() => {
    if (!document.createElementNS) {
      return false;
    }
    let div = document.createElement('div');
    let node = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
    div.appendChild(node);
    let clone = div.cloneNode(true);
    return clone.innerHTML === '<svg xmlns="http://www.w3.org/2000/svg" />';
  })();

  function normalizeInnerHTML(actualHTML) {
    if (ieSVGInnerHTML) {
      // Replace `<svg xmlns="http://www.w3.org/2000/svg" height="50%" />` with `<svg height="50%"></svg>`, etc.
      // drop namespace attribute
      // replace self-closing elements
      actualHTML = actualHTML.replace(/ xmlns="[^"]+"/, '').replace(/<([^ >]+) [^\/>]*\/>/gi, (tag, tagName) => `${tag.slice(0, tag.length - 3)}></${tagName}>`);
    }

    return actualHTML;
  }

  function equalInnerHTML(assert, fragment, html) {
    let actualHTML = normalizeInnerHTML(fragment.innerHTML);

    assert.pushResult({
      result: actualHTML === html,
      actual: actualHTML,
      expected: html
    });
  }
});
enifed('internal-test-helpers/lib/equal-tokens', ['exports', 'simple-html-tokenizer'], function (exports, _simpleHtmlTokenizer) {
  'use strict';

  exports.default = equalTokens;


  function generateTokens(containerOrHTML) {
    if (typeof containerOrHTML === 'string') {
      return {
        tokens: (0, _simpleHtmlTokenizer.tokenize)(containerOrHTML),
        html: containerOrHTML
      };
    } else {
      return {
        tokens: (0, _simpleHtmlTokenizer.tokenize)(containerOrHTML.innerHTML),
        html: containerOrHTML.innerHTML
      };
    }
  }

  function normalizeTokens(tokens) {
    tokens.forEach(token => {
      if (token.type === 'StartTag') {
        token.attributes = token.attributes.sort((a, b) => {
          if (a[0] > b[0]) {
            return 1;
          }
          if (a[0] < b[0]) {
            return -1;
          }
          return 0;
        });
      }
    });
  }

  function equalTokens(actualContainer, expectedHTML, message = null) {
    let actual = generateTokens(actualContainer);
    let expected = generateTokens(expectedHTML);

    normalizeTokens(actual.tokens);
    normalizeTokens(expected.tokens);

    let { assert } = QUnit.config.current;
    let equiv = QUnit.equiv(actual.tokens, expected.tokens);

    if (equiv && expected.html !== actual.html) {
      assert.deepEqual(actual.tokens, expected.tokens, message);
    } else {
      assert.pushResult({
        result: QUnit.equiv(actual.tokens, expected.tokens),
        actual: actual.html,
        expected: expected.html,
        message
      });
    }
  }
});
enifed('internal-test-helpers/lib/factory', ['exports'], function (exports) {
  'use strict';

  exports.default = factory;
  function setProperties(object, properties) {
    for (let key in properties) {
      if (properties.hasOwnProperty(key)) {
        object[key] = properties[key];
      }
    }
  }

  let guids = 0;

  function factory() {
    function Klass(options) {
      setProperties(this, options);
      this._guid = guids++;
      this.isDestroyed = false;
    }

    Klass.prototype.constructor = Klass;
    Klass.prototype.destroy = function () {
      this.isDestroyed = true;
    };

    Klass.prototype.toString = function () {
      return '<Factory:' + this._guid + '>';
    };

    Klass.create = create;
    Klass.extend = extend;
    Klass.reopen = extend;
    Klass.reopenClass = reopenClass;

    return Klass;

    function create(options) {
      return new this.prototype.constructor(options);
    }

    function reopenClass(options) {
      setProperties(this, options);
    }

    function extend(options) {
      function Child(options) {
        Klass.call(this, options);
      }

      let Parent = this;

      Child.prototype = new Parent();
      Child.prototype.constructor = Child;

      setProperties(Child, Klass);
      setProperties(Child.prototype, options);

      Child.create = create;
      Child.extend = extend;
      Child.reopen = extend;

      Child.reopenClass = reopenClass;

      return Child;
    }
  }
});
enifed("internal-test-helpers/lib/get-all-property-names", ["exports"], function (exports) {
  "use strict";

  exports.default = getAllPropertyNames;
  function getAllPropertyNames(Klass) {
    let proto = Klass.prototype;
    let properties = new Set();

    while (proto !== Object.prototype) {
      let names = Object.getOwnPropertyNames(proto);
      names.forEach(name => properties.add(name));
      proto = Object.getPrototypeOf(proto);
    }

    return properties;
  }
});
enifed("internal-test-helpers/lib/get-text-of", ["exports"], function (exports) {
    "use strict";

    exports.default = getTextOf;
    function getTextOf(elem) {
        return elem.textContent.trim();
    }
});
enifed('internal-test-helpers/lib/matchers', ['exports'], function (exports) {
  'use strict';

  exports.regex = regex;
  exports.classes = classes;
  exports.styles = styles;
  exports.equalsElement = equalsElement;
  const HTMLElement = window.HTMLElement;
  const MATCHER_BRAND = '3d4ef194-13be-4ccf-8dc7-862eea02c93e';

  function isMatcher(obj) {
    return typeof obj === 'object' && obj !== null && MATCHER_BRAND in obj;
  }

  function equalsAttr(expected) {
    return {
      [MATCHER_BRAND]: true,

      match(actual) {
        return expected === actual;
      },

      expected() {
        return expected;
      },

      message() {
        return `should equal ${this.expected()}`;
      }
    };
  }

  function regex(r) {
    return {
      [MATCHER_BRAND]: true,

      match(v) {
        return r.test(v);
      },

      expected() {
        return r.toString();
      },

      message() {
        return `should match ${this.expected()}`;
      }
    };
  }

  function classes(expected) {
    return {
      [MATCHER_BRAND]: true,

      match(actual) {
        actual = actual.trim();
        return actual && expected.split(/\s+/).sort().join(' ') === actual.trim().split(/\s+/).sort().join(' ');
      },

      expected() {
        return expected;
      },

      message() {
        return `should match ${this.expected()}`;
      }
    };
  }

  function styles(expected) {
    return {
      [MATCHER_BRAND]: true,

      match(actual) {
        // coerce `null` or `undefined` to an empty string
        // needed for matching empty styles on IE9 - IE11
        actual = actual || '';
        actual = actual.trim();

        return expected.split(';').map(s => s.trim()).filter(s => s).sort().join('; ') === actual.split(';').map(s => s.trim()).filter(s => s).sort().join('; ');
      },

      expected() {
        return expected;
      },

      message() {
        return `should match ${this.expected()}`;
      }
    };
  }

  function equalsElement(assert, element, tagName, attributes, content) {
    assert.pushResult({
      result: element.tagName === tagName.toUpperCase(),
      actual: element.tagName.toLowerCase(),
      expected: tagName,
      message: `expect tagName to be ${tagName}`
    });

    let expectedAttrs = {};
    let expectedCount = 0;

    for (let name in attributes) {
      let expected = attributes[name];
      if (expected !== null) {
        expectedCount++;
      }

      let matcher = isMatcher(expected) ? expected : equalsAttr(expected);

      expectedAttrs[name] = matcher;

      assert.pushResult({
        result: expectedAttrs[name].match(element.getAttribute(name)),
        actual: element.getAttribute(name),
        expected: matcher.expected(),
        message: `Element's ${name} attribute ${matcher.message()}`
      });
    }

    let actualAttributes = {};

    for (let i = 0, l = element.attributes.length; i < l; i++) {
      actualAttributes[element.attributes[i].name] = element.attributes[i].value;
    }

    if (!(element instanceof HTMLElement)) {
      assert.pushResult({
        result: element instanceof HTMLElement,
        message: 'Element must be an HTML Element, not an SVG Element'
      });
    } else {
      assert.pushResult({
        result: element.attributes.length === expectedCount || !attributes,
        actual: element.attributes.length,
        expected: expectedCount,
        message: `Expected ${expectedCount} attributes; got ${element.outerHTML}`
      });

      if (content !== null) {
        assert.pushResult({
          result: element.innerHTML === content,
          actual: element.innerHTML,
          expected: content,
          message: `The element had '${content}' as its content`
        });
      }
    }
  }
});
enifed('internal-test-helpers/lib/module-for', ['exports', '@ember/canary-features', 'internal-test-helpers/lib/apply-mixins', 'internal-test-helpers/lib/get-all-property-names', 'rsvp'], function (exports, _canaryFeatures, _applyMixins, _getAllPropertyNames, _rsvp) {
  'use strict';

  exports.default = moduleFor;
  function moduleFor(description, TestClass, ...mixins) {
    QUnit.module(description, {
      beforeEach: function (...args) {
        let instance = new TestClass(...args);
        this.instance = instance;
        if (instance.beforeEach) {
          return instance.beforeEach(...args);
        }
      },

      afterEach: function () {
        let promises = [];
        let instance = this.instance;
        this.instance = null;
        if (instance.teardown) {
          promises.push(instance.teardown());
        }
        if (instance.afterEach) {
          promises.push(instance.afterEach());
        }

        // this seems odd, but actually saves significant time
        // in the test suite
        //
        // returning a promise from a QUnit test always adds a 13ms
        // delay to the test, this filtering prevents returning a
        // promise when it is not needed
        //
        // Remove after we can update to QUnit that includes
        // https://github.com/qunitjs/qunit/pull/1246
        let filteredPromises = promises.filter(Boolean);
        if (filteredPromises.length > 0) {
          return (0, _rsvp.all)(filteredPromises);
        }
      }
    });

    if (mixins.length > 0) {
      (0, _applyMixins.default)(TestClass, ...mixins);
    }

    let properties = (0, _getAllPropertyNames.default)(TestClass);
    properties.forEach(generateTest);

    function shouldTest(features) {
      return features.every(feature => {
        if (feature[0] === '!' && (0, _canaryFeatures.isEnabled)(feature.slice(1))) {
          return false;
        } else if (!(0, _canaryFeatures.isEnabled)(feature)) {
          return false;
        } else {
          return true;
        }
      });
    }

    function generateTest(name) {
      if (name.indexOf('@test ') === 0) {
        QUnit.test(name.slice(5), function (assert) {
          return this.instance[name](assert);
        });
      } else if (name.indexOf('@only ') === 0) {
        QUnit.only(name.slice(5), function (assert) {
          return this.instance[name](assert);
        });
      } else if (name.indexOf('@skip ') === 0) {
        QUnit.skip(name.slice(5), function (assert) {
          return this.instance[name](assert);
        });
      } else {
        let match = /^@feature\(([A-Z_a-z-!]+)\) /.exec(name);

        if (match) {
          let features = match[1].replace(/ /g, '').split(',');

          if (shouldTest(features)) {
            QUnit.test(name.slice(match[0].length), function (assert) {
              return this.instance[name](assert);
            });
          }
        }
      }
    }
  }
});
enifed('internal-test-helpers/lib/registry-check', ['exports'], function (exports) {
  'use strict';

  exports.verifyRegistration = verifyRegistration;
  exports.verifyInjection = verifyInjection;
  function verifyRegistration(assert, owner, fullName) {
    assert.ok(owner.resolveRegistration(fullName), `has registration: ${fullName}`);
  }

  function verifyInjection(assert, owner, fullName, property, injectionName) {
    let registry = owner.__registry__;
    let injections;

    if (fullName.indexOf(':') === -1) {
      injections = registry.getTypeInjections(fullName);
    } else {
      injections = registry.getInjections(registry.normalize(fullName));
    }

    let normalizedName = registry.normalize(injectionName);
    let hasInjection = false;
    let injection;

    for (let i = 0, l = injections.length; i < l; i++) {
      injection = injections[i];
      if (injection.property === property && injection.specifier === normalizedName) {
        hasInjection = true;
        break;
      }
    }

    assert.ok(hasInjection, `has injection: ${fullName}.${property} = ${injectionName}`);
  }
});
enifed('internal-test-helpers/lib/run', ['exports', '@ember/runloop'], function (exports, _runloop) {
  'use strict';

  exports.runAppend = runAppend;
  exports.runDestroy = runDestroy;
  function runAppend(view) {
    (0, _runloop.run)(view, 'appendTo', document.getElementById('qunit-fixture'));
  }

  function runDestroy(toDestroy) {
    if (toDestroy) {
      (0, _runloop.run)(toDestroy, 'destroy');
    }
  }
});
enifed('internal-test-helpers/lib/strip', ['exports'], function (exports) {
  'use strict';

  exports.default = strip;
  function strip([...strings], ...values) {
    let str = strings.map((string, index) => {
      let interpolated = values[index];
      return string + (interpolated !== undefined ? interpolated : '');
    }).join('');
    return str.split('\n').map(s => s.trim()).join('');
  }
});
enifed('internal-test-helpers/lib/system/synthetic-events', ['exports', '@ember/runloop', '@ember/polyfills'], function (exports, _runloop, _polyfills) {
  'use strict';

  exports.elMatches = undefined;
  exports.matches = matches;
  exports.click = click;
  exports.focus = focus;
  exports.blur = blur;
  exports.fireEvent = fireEvent;


  const DEFAULT_EVENT_OPTIONS = { canBubble: true, cancelable: true };
  /* globals Element */

  const KEYBOARD_EVENT_TYPES = ['keydown', 'keypress', 'keyup'];
  const MOUSE_EVENT_TYPES = ['click', 'mousedown', 'mouseup', 'dblclick', 'mouseenter', 'mouseleave', 'mousemove', 'mouseout', 'mouseover'];

  const elMatches = exports.elMatches = typeof Element !== 'undefined' && (Element.prototype.matches || Element.prototype.matchesSelector || Element.prototype.mozMatchesSelector || Element.prototype.msMatchesSelector || Element.prototype.oMatchesSelector || Element.prototype.webkitMatchesSelector);

  function matches(el, selector) {
    return elMatches.call(el, selector);
  }

  function isFocusable(el) {
    let focusableTags = ['INPUT', 'BUTTON', 'LINK', 'SELECT', 'A', 'TEXTAREA'];
    let { tagName, type } = el;

    if (type === 'hidden') {
      return false;
    }

    return focusableTags.indexOf(tagName) > -1 || el.contentEditable === 'true';
  }

  function click(el, options = {}) {
    (0, _runloop.run)(() => fireEvent(el, 'mousedown', options));
    focus(el);
    (0, _runloop.run)(() => fireEvent(el, 'mouseup', options));
    (0, _runloop.run)(() => fireEvent(el, 'click', options));
  }

  function focus(el) {
    if (!el) {
      return;
    }
    if (isFocusable(el)) {
      (0, _runloop.run)(null, function () {
        let browserIsNotFocused = document.hasFocus && !document.hasFocus();

        // Firefox does not trigger the `focusin` event if the window
        // does not have focus. If the document doesn't have focus just
        // use trigger('focusin') instead.
        if (browserIsNotFocused) {
          fireEvent(el, 'focusin');
        }

        // makes `document.activeElement` be `el`. If the browser is focused, it also fires a focus event
        el.focus();

        // if the browser is not focused the previous `el.focus()` didn't fire an event, so we simulate it
        if (browserIsNotFocused) {
          fireEvent(el, 'focus');
        }
      });
    }
  }

  function blur(el) {
    if (isFocusable(el)) {
      (0, _runloop.run)(null, function () {
        let browserIsNotFocused = document.hasFocus && !document.hasFocus();

        fireEvent(el, 'focusout');

        // makes `document.activeElement` be `body`.
        // If the browser is focused, it also fires a blur event
        el.blur();

        // Chrome/Firefox does not trigger the `blur` event if the window
        // does not have focus. If the document does not have focus then
        // fire `blur` event via native event.
        if (browserIsNotFocused) {
          fireEvent(el, 'blur');
        }
      });
    }
  }

  function fireEvent(element, type, options = {}) {
    if (!element) {
      return;
    }
    let event;
    if (KEYBOARD_EVENT_TYPES.indexOf(type) > -1) {
      event = buildKeyboardEvent(type, options);
    } else if (MOUSE_EVENT_TYPES.indexOf(type) > -1) {
      let rect = element.getBoundingClientRect();
      let x = rect.left + 1;
      let y = rect.top + 1;
      let simulatedCoordinates = {
        screenX: x + 5,
        screenY: y + 95,
        clientX: x,
        clientY: y
      };
      event = buildMouseEvent(type, (0, _polyfills.assign)(simulatedCoordinates, options));
    } else {
      event = buildBasicEvent(type, options);
    }
    element.dispatchEvent(event);

    return event;
  }

  function buildBasicEvent(type, options = {}) {
    let event = document.createEvent('Events');
    event.initEvent(type, true, true);
    (0, _polyfills.assign)(event, options);
    return event;
  }

  function buildMouseEvent(type, options = {}) {
    let event;
    try {
      event = document.createEvent('MouseEvents');
      let eventOpts = (0, _polyfills.assign)({}, DEFAULT_EVENT_OPTIONS, options);

      event.initMouseEvent(type, eventOpts.canBubble, eventOpts.cancelable, window, eventOpts.detail, eventOpts.screenX, eventOpts.screenY, eventOpts.clientX, eventOpts.clientY, eventOpts.ctrlKey, eventOpts.altKey, eventOpts.shiftKey, eventOpts.metaKey, eventOpts.button, eventOpts.relatedTarget);
    } catch (e) {
      event = buildBasicEvent(type, options);
    }
    return event;
  }

  function buildKeyboardEvent(type, options = {}) {
    let event;
    try {
      event = document.createEvent('KeyEvents');
      let eventOpts = (0, _polyfills.assign)({}, DEFAULT_EVENT_OPTIONS, options);
      event.initKeyEvent(type, eventOpts.canBubble, eventOpts.cancelable, window, eventOpts.ctrlKey, eventOpts.altKey, eventOpts.shiftKey, eventOpts.metaKey, eventOpts.keyCode, eventOpts.charCode);
    } catch (e) {
      event = buildBasicEvent(type, options);
    }
    return event;
  }
});
enifed('internal-test-helpers/lib/test-cases/abstract-application', ['exports', 'ember-template-compiler', '@ember/-internals/environment', 'internal-test-helpers/lib/test-cases/abstract', 'internal-test-helpers/lib/run'], function (exports, _emberTemplateCompiler, _environment, _abstract, _run) {
  'use strict';

  class AbstractApplicationTestCase extends _abstract.default {
    _ensureInstance(bootOptions) {
      if (this._applicationInstancePromise) {
        return this._applicationInstancePromise;
      }

      return this._applicationInstancePromise = this.runTask(() => this.application.boot()).then(app => {
        this.applicationInstance = app.buildInstance();

        return this.applicationInstance.boot(bootOptions);
      });
    }

    visit(url, options) {
      // TODO: THIS IS HORRIBLE
      // the promise returned by `ApplicationInstance.protoype.visit` does **not**
      // currently guarantee rendering is completed
      return this.runTask(() => {
        return this._ensureInstance(options).then(instance => instance.visit(url));
      });
    }

    get element() {
      if (this._element) {
        return this._element;
      } else if (_environment.ENV._APPLICATION_TEMPLATE_WRAPPER) {
        return this._element = document.querySelector('#qunit-fixture > div.ember-view');
      } else {
        return this._element = document.querySelector('#qunit-fixture');
      }
    }

    set element(element) {
      this._element = element;
    }

    afterEach() {
      (0, _run.runDestroy)(this.applicationInstance);
      (0, _run.runDestroy)(this.application);

      super.teardown();
    }

    get applicationOptions() {
      return {
        rootElement: '#qunit-fixture'
      };
    }

    get routerOptions() {
      return {
        location: 'none'
      };
    }

    get router() {
      return this.application.resolveRegistration('router:main');
    }

    compile() /* string, options */{
      return (0, _emberTemplateCompiler.compile)(...arguments);
    }
  }
  exports.default = AbstractApplicationTestCase;
});
enifed('internal-test-helpers/lib/test-cases/abstract-rendering', ['exports', '@ember/polyfills', 'ember-template-compiler', '@ember/-internals/views', '@ember/-internals/glimmer', 'internal-test-helpers/lib/test-resolver', 'internal-test-helpers/lib/test-cases/abstract', 'internal-test-helpers/lib/build-owner', 'internal-test-helpers/lib/run'], function (exports, _polyfills, _emberTemplateCompiler, _views, _glimmer, _testResolver, _abstract, _buildOwner, _run) {
  'use strict';

  const TextNode = window.Text;

  class AbstractRenderingTestCase extends _abstract.default {
    constructor() {
      super(...arguments);
      let bootOptions = this.getBootOptions();

      let owner = this.owner = (0, _buildOwner.default)({
        ownerOptions: this.getOwnerOptions(),
        resolver: this.getResolver(),
        bootOptions
      });

      this.renderer = this.owner.lookup('renderer:-dom');
      this.element = document.querySelector('#qunit-fixture');
      this.component = null;

      owner.register('event_dispatcher:main', _views.EventDispatcher);
      owner.inject('event_dispatcher:main', '_viewRegistry', '-view-registry:main');
      if (!bootOptions || bootOptions.isInteractive !== false) {
        owner.lookup('event_dispatcher:main').setup(this.getCustomDispatcherEvents(), this.element);
      }
    }

    compile() {
      return (0, _emberTemplateCompiler.compile)(...arguments);
    }

    getCustomDispatcherEvents() {
      return {};
    }

    getOwnerOptions() {}
    getBootOptions() {}

    get resolver() {
      return this.owner.__registry__.fallback.resolver;
    }

    getResolver() {
      return new _testResolver.ModuleBasedResolver();
    }

    add(specifier, factory) {
      this.resolver.add(specifier, factory);
    }

    addTemplate(templateName, templateString) {
      if (typeof templateName === 'string') {
        this.resolver.add(`template:${templateName}`, this.compile(templateString, {
          moduleName: templateName
        }));
      } else {
        this.resolver.add(templateName, this.compile(templateString, {
          moduleName: templateName.moduleName
        }));
      }
    }

    addComponent(name, { ComponentClass = null, template = null }) {
      if (ComponentClass) {
        this.resolver.add(`component:${name}`, ComponentClass);
      }

      if (typeof template === 'string') {
        this.resolver.add(`template:components/${name}`, this.compile(template, {
          moduleName: `components/${name}`
        }));
      }
    }

    afterEach() {
      try {
        if (this.component) {
          (0, _run.runDestroy)(this.component);
        }
        if (this.owner) {
          (0, _run.runDestroy)(this.owner);
        }
      } finally {
        (0, _glimmer._resetRenderers)();
      }
    }

    get context() {
      return this.component;
    }

    render(templateStr, context = {}) {
      let { owner } = this;

      owner.register('template:-top-level', this.compile(templateStr, {
        moduleName: '-top-level'
      }));

      let attrs = (0, _polyfills.assign)({}, context, {
        tagName: '',
        layoutName: '-top-level'
      });

      owner.register('component:-top-level', _glimmer.Component.extend(attrs));

      this.component = owner.lookup('component:-top-level');

      (0, _run.runAppend)(this.component);
    }

    rerender() {
      this.component.rerender();
    }

    registerHelper(name, funcOrClassBody) {
      let type = typeof funcOrClassBody;

      if (type === 'function') {
        this.owner.register(`helper:${name}`, (0, _glimmer.helper)(funcOrClassBody));
      } else if (type === 'object' && type !== null) {
        this.owner.register(`helper:${name}`, _glimmer.Helper.extend(funcOrClassBody));
      } else {
        throw new Error(`Cannot register ${funcOrClassBody} as a helper`);
      }
    }

    registerPartial(name, template) {
      let owner = this.env.owner || this.owner;
      if (typeof template === 'string') {
        owner.register(`template:${name}`, this.compile(template, { moduleName: `my-app/templates/-${name}.hbs` }));
      }
    }

    registerComponent(name, { ComponentClass = _glimmer.Component, template = null }) {
      let { owner } = this;

      if (ComponentClass) {
        owner.register(`component:${name}`, ComponentClass);
      }

      if (typeof template === 'string') {
        owner.register(`template:components/${name}`, this.compile(template, {
          moduleName: `my-app/templates/components/${name}.hbs`
        }));
      }
    }

    registerModifier(name, ModifierClass) {
      let { owner } = this;

      owner.register(`modifier:${name}`, ModifierClass);
    }

    registerComponentManager(name, manager) {
      let owner = this.env.owner || this.owner;
      owner.register(`component-manager:${name}`, manager);
    }

    registerTemplate(name, template) {
      let { owner } = this;
      if (typeof template === 'string') {
        owner.register(`template:${name}`, this.compile(template, {
          moduleName: `my-app/templates/${name}.hbs`
        }));
      } else {
        throw new Error(`Registered template "${name}" must be a string`);
      }
    }

    registerService(name, klass) {
      this.owner.register(`service:${name}`, klass);
    }

    assertTextNode(node, text) {
      if (!(node instanceof TextNode)) {
        throw new Error(`Expecting a text node, but got ${node}`);
      }

      this.assert.strictEqual(node.textContent, text, 'node.textContent');
    }
  }
  exports.default = AbstractRenderingTestCase;
});
enifed('internal-test-helpers/lib/test-cases/abstract', ['exports', '@ember/polyfills', '@ember/runloop', 'internal-test-helpers/lib/test-cases/node-query', 'internal-test-helpers/lib/equal-inner-html', 'internal-test-helpers/lib/equal-tokens', 'internal-test-helpers/lib/matchers', 'rsvp'], function (exports, _polyfills, _runloop, _nodeQuery, _equalInnerHtml, _equalTokens, _matchers, _rsvp) {
  'use strict';

  const TextNode = window.Text; /* global Element */

  const HTMLElement = window.HTMLElement;
  const Comment = window.Comment;

  function isMarker(node) {
    if (node instanceof Comment && node.textContent === '') {
      return true;
    }

    if (node instanceof TextNode && node.textContent === '') {
      return true;
    }

    return false;
  }

  class AbstractTestCase {
    constructor(assert) {
      this.element = null;
      this.snapshot = null;
      this.assert = assert;

      let { fixture } = this;
      if (fixture) {
        this.setupFixture(fixture);
      }
    }

    teardown() {}
    afterEach() {}

    runTask(callback) {
      return (0, _runloop.run)(callback);
    }

    runTaskNext() {
      return new _rsvp.Promise(resolve => {
        return (0, _runloop.next)(resolve);
      });
    }

    setupFixture(innerHTML) {
      let fixture = document.getElementById('qunit-fixture');
      fixture.innerHTML = innerHTML;
    }

    // The following methods require `this.element` to work

    get firstChild() {
      return this.nthChild(0);
    }

    nthChild(n) {
      let i = 0;
      let node = this.element.firstChild;

      while (node) {
        if (!isMarker(node)) {
          i++;
        }

        if (i > n) {
          break;
        } else {
          node = node.nextSibling;
        }
      }

      return node;
    }

    get nodesCount() {
      let count = 0;
      let node = this.element.firstChild;

      while (node) {
        if (!isMarker(node)) {
          count++;
        }

        node = node.nextSibling;
      }

      return count;
    }

    $(sel) {
      if (sel instanceof Element) {
        return _nodeQuery.default.element(sel);
      } else if (typeof sel === 'string') {
        return _nodeQuery.default.query(sel, this.element);
      } else if (sel !== undefined) {
        throw new Error(`Invalid this.$(${sel})`);
      } else {
        return _nodeQuery.default.element(this.element);
      }
    }

    wrap(element) {
      return _nodeQuery.default.element(element);
    }

    click(selector) {
      let element;
      if (typeof selector === 'string') {
        element = this.element.querySelector(selector);
      } else {
        element = selector;
      }

      let event = element.click();

      return this.runLoopSettled(event);
    }

    // TODO: Find a better name 😎
    runLoopSettled(value) {
      return new _rsvp.Promise(function (resolve) {
        // Every 5ms, poll for the async thing to have finished
        let watcher = setInterval(() => {
          // If there are scheduled timers or we are inside of a run loop, keep polling
          if ((0, _runloop.hasScheduledTimers)() || (0, _runloop.getCurrentRunLoop)()) {
            return;
          }

          // Stop polling
          clearInterval(watcher);

          // Synchronously resolve the promise
          resolve(value);
        }, 5);
      });
    }

    textValue() {
      return this.element.textContent;
    }

    takeSnapshot() {
      let snapshot = this.snapshot = [];

      let node = this.element.firstChild;

      while (node) {
        if (!isMarker(node)) {
          snapshot.push(node);
        }

        node = node.nextSibling;
      }

      return snapshot;
    }

    assertText(text) {
      this.assert.strictEqual(this.textValue(), text, `#qunit-fixture content should be: \`${text}\``);
    }

    assertInnerHTML(html) {
      (0, _equalInnerHtml.default)(this.assert, this.element, html);
    }

    assertHTML(html) {
      (0, _equalTokens.default)(this.element, html, `#qunit-fixture content should be: \`${html}\``);
    }

    assertElement(node, { ElementType = HTMLElement, tagName, attrs = null, content = null }) {
      if (!(node instanceof ElementType)) {
        throw new Error(`Expecting a ${ElementType.name}, but got ${node}`);
      }

      (0, _matchers.equalsElement)(this.assert, node, tagName, attrs, content);
    }

    assertComponentElement(node, { ElementType = HTMLElement, tagName = 'div', attrs = null, content = null }) {
      attrs = (0, _polyfills.assign)({}, { id: (0, _matchers.regex)(/^ember\d*$/), class: (0, _matchers.classes)('ember-view') }, attrs || {});
      this.assertElement(node, { ElementType, tagName, attrs, content });
    }

    assertSameNode(actual, expected) {
      this.assert.strictEqual(actual, expected, 'DOM node stability');
    }

    assertInvariants(oldSnapshot, newSnapshot) {
      oldSnapshot = oldSnapshot || this.snapshot;
      newSnapshot = newSnapshot || this.takeSnapshot();

      this.assert.strictEqual(newSnapshot.length, oldSnapshot.length, 'Same number of nodes');

      for (let i = 0; i < oldSnapshot.length; i++) {
        this.assertSameNode(newSnapshot[i], oldSnapshot[i]);
      }
    }

    assertPartialInvariants(start, end) {
      this.assertInvariants(this.snapshot, this.takeSnapshot().slice(start, end));
    }

    assertStableRerender() {
      this.takeSnapshot();
      this.runTask(() => this.rerender());
      this.assertInvariants();
    }
  }
  exports.default = AbstractTestCase;
});
enifed('internal-test-helpers/lib/test-cases/application', ['exports', 'internal-test-helpers/lib/test-cases/test-resolver-application', '@ember/application', '@ember/-internals/routing', '@ember/polyfills'], function (exports, _testResolverApplication, _application, _routing, _polyfills) {
  'use strict';

  class ApplicationTestCase extends _testResolverApplication.default {
    constructor() {
      super(...arguments);

      let { applicationOptions } = this;
      this.application = this.runTask(this.createApplication.bind(this, applicationOptions));

      this.resolver = this.application.__registry__.resolver;

      if (this.resolver) {
        this.resolver.add('router:main', _routing.Router.extend(this.routerOptions));
      }
    }

    createApplication(myOptions = {}, MyApplication = _application.default) {
      return MyApplication.create(myOptions);
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        autoboot: false
      });
    }

    get appRouter() {
      return this.applicationInstance.lookup('router:main');
    }

    transitionTo() {
      return this.runTask(() => {
        return this.appRouter.transitionTo(...arguments);
      });
    }
  }
  exports.default = ApplicationTestCase;
});
enifed('internal-test-helpers/lib/test-cases/autoboot-application', ['exports', 'internal-test-helpers/lib/test-cases/test-resolver-application', '@ember/application', '@ember/polyfills', '@ember/-internals/routing'], function (exports, _testResolverApplication, _application, _polyfills, _routing) {
  'use strict';

  class AutobootApplicationTestCase extends _testResolverApplication.default {
    createApplication(options, MyApplication = _application.default) {
      let myOptions = (0, _polyfills.assign)(this.applicationOptions, options);
      let application = this.application = MyApplication.create(myOptions);
      this.resolver = application.__registry__.resolver;

      if (this.resolver) {
        this.resolver.add('router:main', _routing.Router.extend(this.routerOptions));
      }

      return application;
    }

    visit(url) {
      return this.application.boot().then(() => {
        return this.applicationInstance.visit(url);
      });
    }

    get applicationInstance() {
      let { application } = this;

      if (!application) {
        return undefined;
      }

      return application.__deprecatedInstance__;
    }
  }
  exports.default = AutobootApplicationTestCase;
});
enifed('internal-test-helpers/lib/test-cases/default-resolver-application', ['exports', 'internal-test-helpers/lib/test-cases/abstract-application', '@ember/application/globals-resolver', '@ember/application', '@ember/-internals/glimmer', '@ember/polyfills', '@ember/-internals/routing'], function (exports, _abstractApplication, _globalsResolver, _application, _glimmer, _polyfills, _routing) {
  'use strict';

  class ApplicationTestCase extends _abstractApplication.default {
    createApplication() {
      let application = this.application = _application.default.create(this.applicationOptions);
      application.Router = _routing.Router.extend(this.routerOptions);
      return application;
    }

    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        name: 'TestApp',
        autoboot: false,
        Resolver: _globalsResolver.default
      });
    }

    afterEach() {
      (0, _glimmer.setTemplates)({});
      return super.afterEach();
    }

    get appRouter() {
      return this.applicationInstance.lookup('router:main');
    }

    transitionTo() {
      return this.runTask(() => {
        return this.appRouter.transitionTo(...arguments);
      });
    }

    addTemplate(name, templateString) {
      let compiled = this.compile(templateString);
      (0, _glimmer.setTemplate)(name, compiled);
      return compiled;
    }
  }
  exports.default = ApplicationTestCase;
});
enifed('internal-test-helpers/lib/test-cases/node-query', ['exports', '@ember/debug', 'internal-test-helpers/lib/system/synthetic-events'], function (exports, _debug, _syntheticEvents) {
  'use strict';

  /* global Node */

  class NodeQuery {
    static query(selector, context = document) {
      true && !(context && context instanceof Node) && (0, _debug.assert)(`Invalid second parameter to NodeQuery.query`, context && context instanceof Node);

      return new NodeQuery(toArray(context.querySelectorAll(selector)));
    }

    static element(element) {
      return new NodeQuery([element]);
    }

    constructor(nodes) {
      true && !Array.isArray(nodes) && (0, _debug.assert)('NodeQuery must be initialized with a literal array', Array.isArray(nodes));

      this.nodes = nodes;

      for (let i = 0; i < nodes.length; i++) {
        this[i] = nodes[i];
      }

      this.length = nodes.length;

      Object.freeze(this);
    }

    find(selector) {
      assertSingle(this);

      return this[0].querySelector(selector);
    }

    findAll(selector) {
      let nodes = [];

      this.nodes.forEach(node => {
        nodes.push(...node.querySelectorAll(selector));
      });

      return new NodeQuery(nodes);
    }

    trigger(eventName, options) {
      return this.nodes.map(node => (0, _syntheticEvents.fireEvent)(node, eventName, options));
    }

    click() {
      return this.trigger('click');
    }

    focus() {
      this.nodes.forEach(_syntheticEvents.focus);
    }

    text() {
      return this.nodes.map(node => node.textContent).join('');
    }

    attr(name) {
      if (arguments.length !== 1) {
        throw new Error('not implemented');
      }

      assertSingle(this);

      return this.nodes[0].getAttribute(name);
    }

    prop(name, value) {
      if (arguments.length > 1) {
        return this.setProp(name, value);
      }

      assertSingle(this);

      return this.nodes[0][name];
    }

    setProp(name, value) {
      this.nodes.forEach(node => node[name] = value);

      return this;
    }

    val(value) {
      if (arguments.length === 1) {
        return this.setProp('value', value);
      }

      return this.prop('value');
    }

    is(selector) {
      return this.nodes.every(node => (0, _syntheticEvents.matches)(node, selector));
    }

    hasClass(className) {
      return this.is(`.${className}`);
    }
  }

  exports.default = NodeQuery;
  function assertSingle(nodeQuery) {
    if (nodeQuery.length !== 1) {
      throw new Error(`attr(name) called on a NodeQuery with ${this.nodes.length} elements. Expected one element.`);
    }
  }

  function toArray(nodes) {
    let out = [];

    for (let i = 0; i < nodes.length; i++) {
      out.push(nodes[i]);
    }

    return out;
  }
});
enifed('internal-test-helpers/lib/test-cases/query-param', ['exports', '@ember/controller', '@ember/-internals/routing', '@ember/runloop', 'internal-test-helpers/lib/test-cases/application'], function (exports, _controller, _routing, _runloop, _application) {
  'use strict';

  class QueryParamTestCase extends _application.default {
    constructor() {
      super(...arguments);

      let testCase = this;
      testCase.expectedPushURL = null;
      testCase.expectedReplaceURL = null;

      this.add('location:test', _routing.NoneLocation.extend({
        setURL(path) {
          if (testCase.expectedReplaceURL) {
            testCase.assert.ok(false, 'pushState occurred but a replaceState was expected');
          }

          if (testCase.expectedPushURL) {
            testCase.assert.equal(path, testCase.expectedPushURL, 'an expected pushState occurred');
            testCase.expectedPushURL = null;
          }

          this.set('path', path);
        },

        replaceURL(path) {
          if (testCase.expectedPushURL) {
            testCase.assert.ok(false, 'replaceState occurred but a pushState was expected');
          }

          if (testCase.expectedReplaceURL) {
            testCase.assert.equal(path, testCase.expectedReplaceURL, 'an expected replaceState occurred');
            testCase.expectedReplaceURL = null;
          }

          this.set('path', path);
        }
      }));
    }

    visitAndAssert(path) {
      return this.visit(...arguments).then(() => {
        this.assertCurrentPath(path);
      });
    }

    getController(name) {
      return this.applicationInstance.lookup(`controller:${name}`);
    }

    getRoute(name) {
      return this.applicationInstance.lookup(`route:${name}`);
    }

    get routerOptions() {
      return {
        location: 'test'
      };
    }

    setAndFlush(obj, prop, value) {
      return (0, _runloop.run)(obj, 'set', prop, value);
    }

    assertCurrentPath(path, message = `current path equals '${path}'`) {
      this.assert.equal(this.appRouter.get('location.path'), path, message);
    }

    /**
      Sets up a Controller for a given route with a single query param and default
      value. Can optionally extend the controller with an object.
       @public
      @method setSingleQPController
    */
    setSingleQPController(routeName, param = 'foo', defaultValue = 'bar', options = {}) {
      this.add(`controller:${routeName}`, _controller.default.extend({
        queryParams: [param],
        [param]: defaultValue
      }, options));
    }

    /**
      Sets up a Controller for a given route with a custom property/url key mapping.
       @public
      @method setMappedQPController
    */
    setMappedQPController(routeName, prop = 'page', urlKey = 'parentPage', defaultValue = 1, options = {}) {
      this.add(`controller:${routeName}`, _controller.default.extend({
        queryParams: {
          [prop]: urlKey
        },
        [prop]: defaultValue
      }, options));
    }
  }
  exports.default = QueryParamTestCase;
});
enifed('internal-test-helpers/lib/test-cases/rendering', ['exports', 'internal-test-helpers/lib/test-cases/abstract-rendering', '@ember/-internals/container'], function (exports, _abstractRendering, _container) {
  'use strict';

  class RenderingTestCase extends _abstractRendering.default {
    constructor() {
      super(...arguments);
      let { owner } = this;

      this.env = owner.lookup('service:-glimmer-environment');
      this.templateOptions = owner.lookup(_container.privatize`template-compiler:main`);
      this.compileTimeLookup = this.templateOptions.resolver;
      this.runtimeResolver = this.compileTimeLookup.resolver;
    }
  }
  exports.default = RenderingTestCase;
});
enifed('internal-test-helpers/lib/test-cases/router', ['exports', 'internal-test-helpers/lib/test-cases/application'], function (exports, _application) {
  'use strict';

  class RouterTestCase extends _application.default {
    constructor() {
      super(...arguments);

      this.router.map(function () {
        this.route('parent', { path: '/' }, function () {
          this.route('child');
          this.route('sister');
          this.route('brother');
        });
        this.route('dynamic', { path: '/dynamic/:dynamic_id' });
        this.route('dynamicWithChild', { path: '/dynamic-with-child/:dynamic_id' }, function () {
          this.route('child', { path: '/:child_id' });
        });
      });
    }

    get routerService() {
      return this.applicationInstance.lookup('service:router');
    }

    buildQueryParams(queryParams) {
      return {
        queryParams
      };
    }
  }
  exports.default = RouterTestCase;
});
enifed('internal-test-helpers/lib/test-cases/test-resolver-application', ['exports', 'internal-test-helpers/lib/test-cases/abstract-application', 'internal-test-helpers/lib/test-resolver', '@ember/-internals/glimmer', '@ember/polyfills'], function (exports, _abstractApplication, _testResolver, _glimmer, _polyfills) {
  'use strict';

  class TestResolverApplicationTestCase extends _abstractApplication.default {
    get applicationOptions() {
      return (0, _polyfills.assign)(super.applicationOptions, {
        Resolver: _testResolver.ModuleBasedResolver
      });
    }

    add(specifier, factory) {
      this.resolver.add(specifier, factory);
    }

    addTemplate(templateName, templateString) {
      this.resolver.add(`template:${templateName}`, this.compile(templateString, {
        moduleName: `my-app/templates/${templateName}.hbs`
      }));
    }

    addComponent(name, { ComponentClass = _glimmer.Component, template = null }) {
      if (ComponentClass) {
        this.resolver.add(`component:${name}`, ComponentClass);
      }

      if (typeof template === 'string') {
        this.resolver.add(`template:components/${name}`, this.compile(template, {
          moduleName: `my-app/templates/components/${name}.hbs`
        }));
      }
    }
  }
  exports.default = TestResolverApplicationTestCase;
});
enifed('internal-test-helpers/lib/test-resolver', ['exports', 'ember-template-compiler'], function (exports, _emberTemplateCompiler) {
  'use strict';

  exports.ModuleBasedResolver = undefined;


  const DELIMITER = '%';

  function serializeKey(specifier, source, namespace) {
    let [type, name] = specifier.split(':');
    return `${type}://${[name, namespace ? '[source invalid due to namespace]' : source, namespace].join(DELIMITER)}`;
  }

  class Resolver {
    constructor() {
      this._registered = {};
    }
    resolve(specifier) {
      return this._registered[specifier] || this._registered[serializeKey(specifier)];
    }
    expandLocalLookup(specifier, source, namespace) {
      if (specifier.indexOf('://') !== -1) {
        return specifier; // an already expanded specifier
      }

      if (source || namespace) {
        let key = serializeKey(specifier, source, namespace);
        if (this._registered[key]) {
          return key; // like local lookup
        }

        key = serializeKey(specifier);
        if (this._registered[key]) {
          return specifier; // top level resolution despite source/namespace
        }
      }

      return specifier; // didn't know how to expand it
    }
    add(lookup, factory) {
      let key;
      switch (typeof lookup) {
        case 'string':
          if (lookup.indexOf(':') === -1) {
            throw new Error('Specifiers added to the resolver must be in the format of type:name');
          }
          key = serializeKey(lookup);
          break;
        case 'object':
          key = serializeKey(lookup.specifier, lookup.source, lookup.namespace);
          break;
        default:
          throw new Error('Specifier string has an unknown type');
      }

      return this._registered[key] = factory;
    }
    addTemplate(templateName, template) {
      let templateType = typeof template;
      if (templateType !== 'string') {
        throw new Error(`You called addTemplate for "${templateName}" with a template argument of type of '${templateType}'. addTemplate expects an argument of an uncompiled template as a string.`);
      }
      return this._registered[serializeKey(`template:${templateName}`)] = (0, _emberTemplateCompiler.compile)(template, {
        moduleName: `my-app/templates/${templateName}.hbs`
      });
    }
    static create() {
      return new this();
    }
  }

  exports.default = Resolver;


  /*
   * A resolver with moduleBasedResolver = true handles error and loading
   * substates differently than a standard resolver.
   */
  class ModuleBasedResolver extends Resolver {
    get moduleBasedResolver() {
      return true;
    }
  }

  exports.ModuleBasedResolver = ModuleBasedResolver;
});
enifed('internal-test-helpers/tests/index-test', ['internal-test-helpers'], function (_internalTestHelpers) {
  'use strict';

  (0, _internalTestHelpers.moduleFor)('internal-test-helpers', class extends _internalTestHelpers.AbstractTestCase {
    ['@test module present'](assert) {
      assert.ok(true, 'each package needs at least one test to be able to run through `npm test`');
    }
  });
});
/*global enifed, module */
enifed('node-module', ['exports'], function(_exports) {
  var IS_NODE = typeof module === 'object' && typeof module.require === 'function';
  if (IS_NODE) {
    _exports.require = module.require;
    _exports.module = module;
    _exports.IS_NODE = IS_NODE;
  } else {
    _exports.require = null;
    _exports.module = null;
    _exports.IS_NODE = IS_NODE;
  }
});



}());
//# sourceMappingURL=ember-tests.map
